//---------------------------------------------------------------------------
#ifndef AnimationBSRContextH
#define AnimationBSRContextH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PatchGraphContainer.h"
//---------------------------------------------------------------------------

namespace geoxl {

class RegistrationContext;
class PGOpt2Config;

/// Templates
#define TWO_CUBES 0
#define BUNNY 1
#define BUST 2
#define THEO_GYM 3
#define FACE 4
#define THEO_DANCE 5
#define FINGERS 6
#define DANCER_ONLY 7
#define MAX_PLANCK 8
#define SPOK 9
#define PMD 10
#define VENUS 11
#define FACE_BENNI 12
#define BEN 13
#define ELEPHANT 14
#define TORSO 15
#define ARM 16
#define BAER 17

/// ICP-based merge
const int ICP_REGISTER_RIGID  = 0;
const int ICP_REGISTER_SUBDIVIDED  = 1;

/// Merge candidate selection mode
const int SELECTION_MODE_MANUAL = 0;
const int SELECTION_MODE_AUTOMATIC = 1;



/// Merge patch graphs
class ANIMATIONRECONSTRUCTION_API AnimationBSRContext : public Persistent {
	GEOXL_CLASS( AnimationBSRContext )

   public:
      static const card32 MERGE_MIN_DEFORM_ENERGY = 0;
      static const card32 MERGE_MIN_TOTAL_ENERGY = 1;
      static const card32 MERGE_MIN_DEFORM_PLUS_DISTANCE_ENERGY = 2;

	private:
		/// Sampling spacing
		float32 samplingSpacing;
		/// Average number of 
		/// Patch graph container
		PatchGraphContainer* pgContainer;
		/// Index of current test dataset (used to set the appropriate parameters)
		card32 dataSet;
		/// Tensor voting settings
		float32 tvSigma;
		float32 tvCurvaturePenalty;
		float32 tvSurfacenessThreshold;
		/// ICP merge settings
		card32 icpRegistrationMode;
		/// Data points influence (times sampling spacing)
		float32 dataPointsInfluence;
		// Number of merge steps applied; negative values indicate goal number of patch graphs
		int32 numMergeSteps;
		/// Minimum size for a patch graph
		card32 minPatchGraphSize;
		/// Subsampling of the icp registration
		int32 icpSubsampling;
		/// Print debug information to console
		bool verbose;
		/// Distance: max factor of tvSigma in pg extract region growing
		float32 pgRegionGrowScale;
		/// Factor of sampling spacing used int initial knn topology graph creation
		float32 extractKNNFactor;
		
		/// distance up to which neighbors are connected during merging (multiple of sampleSpacing)
		float32 topMergePruneDistFactor;
		/// radius that is used to look for candidate neighbors after merging (multiple of sampleSpacing)
		float32 topEdgeResampleSearchRadiusFactor;
		/// radius in which neighbors are kept after topology merging (multiple of sampleSpacing)
		float32 topEdgeResamplePruneRadiusFactor;
		
		/// Scaling factor for debug visualizations
		float32 renderScale;
		/// holds settings for feature registration
		RegistrationContext * registrationContext;

		/// Merge candidate selector selection mode
		card32 mergeCandSelectionMode;
		
		/// not used anymore
		bool urshapeOpt;
		/// not used anymore
		bool urshapeMerge;
		
		/// use urshape in global opt
		bool urshapeGlobalOpt;
		/// use urshape in global opt - affect scene
		bool urshapeToScene;
		/// use urshape in global opt - create from scene
		bool sceneToUrshape;
		/// cyclic time boundary conditions in global opt 
		bool cyclicTime;

		/// not use refactored anim-rec pipeline
		bool useOldPipeline;


      // new pipeline parameters ------------
      
      // -- topology2
      
      /// k in kNN for topology computation (invariant to be kept over iterations)
      card32 topology2NumNearestNeighbors;
      /// eps for topology computation (multiple of sampleSpacing, invariant to be kept over iterations)
      float32 topology2MaxDistanceRelSampleSpacing;

      /// max additive length change before the edge is removed
      float32 topology2MaxAdditiveEdgeLengthDev;
      /// max multiplicative factor by which the edge length might change
      float32 topology2MaxMultiplicativeEdgeLengthDev;
      /// minimum length of an edge before the multiplicative criterion is used to
      /// remove the edge (as a fraction of sample spacing)
      float32 topology2MinMultRejectLengthFraction;
      card32 topology2MergeSelCriterion;
      card32 topology2minimumVertexDegree;

      // -- opt graph topology 2

		bool optGraph2CyclicTime;
		bool optGraph2twoWayTemporalEdges;
		bool optGraph2useUrshape;
		bool alignment2FeaturesFirstDataSecond;

      // ------------------------------------


      // --- optimization parameters
	   PGOpt2Config *pairwiseAlignmentConfig;
	   PGOpt2Config *holeFillingConfig;
	   PGOpt2Config *globalOptConfig;
	   PGOpt2Config *initialOptConfig;

	   /// Save container after each merge
	   bool saveAfterEachMerge;


   public:		
		float32 maxRelEdgeLength;
      float32 maxEdgeStretchFactor;
		
	public:
		/*** CLASS MANAGEMENT ***/
		/// Constructor
		AnimationBSRContext();   
		/// Destructor
		~AnimationBSRContext();

		/// Set the parameters required for each test dataset
		void setParameters();
		/// Load patch graph container from file
		void loadPGContainer( const QString& filename );
		/// Save patch graph container from file
		void savePGContainer( const QString& filename );

		/*** MEMBER ACCESS ***/
		inline float32 getSamplingSpacing() const { return samplingSpacing; }
		inline PatchGraphContainer* getPGContainer() { return pgContainer; }
		inline void setPGContainer(PatchGraphContainer *c) { if (c!=pgContainer) {delete pgContainer;} pgContainer = c; }
		inline float32 getTVSigma() const { return tvSigma; }
		inline float32 getTVCurvaturePenalty() const { return tvCurvaturePenalty; }
		inline float32 getTVSurfacenessThreshold() const { return tvSurfacenessThreshold; }
		inline card32 getICPRegistrationMode() const { return icpRegistrationMode; }
		inline float32 getDataPointsInfluence() const { return dataPointsInfluence; }
		inline float32 getRenderScale() const { return renderScale; }
		RegistrationContext * getRegistrationContext() const {return registrationContext;}
		inline card32 getDataSet() const { return dataSet; }
		inline card32 getMergeCandSelectionMode() const { return mergeCandSelectionMode; }
		inline float32 getTopMergePruneDistFactor() const { return topMergePruneDistFactor;}
		inline float32 getTopEdgeResampleSearchRadiusFactor() const { return topEdgeResampleSearchRadiusFactor;}
		inline float32 getTopEdgeResamplePruneRadiusFactor() const { return topEdgeResamplePruneRadiusFactor;}
		inline void setMergeCandSelectionMode( const card32 mode ){ mergeCandSelectionMode = mode; }
		inline void setNumMergeSteps( const card32 n ){ numMergeSteps = n; }
		inline int32 getNumMergeSteps() const { return numMergeSteps; }
		inline card32 getMinPatchGraphSize() const { return minPatchGraphSize; }
		inline int32 getICPSubsampling() const { return icpSubsampling; }
		inline bool getVerbose() const { return verbose; }
		inline void setVerbose( bool b ){ verbose = b; }
		inline float32 getPGRegionGrowScale() const { return pgRegionGrowScale; }
		inline float32 getExtractKNNFactor() const { return extractKNNFactor; }
   	inline bool getUseOldPipeline() const {return useOldPipeline;}
	   inline void setUseOldPipeline(bool v) {useOldPipeline=v;}
	   inline bool getUrshapeGlobalOpt() const {return urshapeGlobalOpt;}
	   inline void setUrshapeGlobalOpt(bool v) {urshapeGlobalOpt=v;}
	   inline bool getUrshapeToScene() const {return urshapeToScene;}
	   inline void setUrshapeToScene(bool v) {urshapeToScene=v;}
	   inline bool getSceneToUrshape() const {return sceneToUrshape;}
	   inline void setSceneToUrshape(bool v) {sceneToUrshape=v;}
	   inline bool getCyclicTime() const {return cyclicTime;}
	   inline void setCyclicTime(bool v) {cyclicTime=v;}
      
	   PGOpt2Config* getPairwiseAlignmentConfig() const {return pairwiseAlignmentConfig;}
	   void setPairwiseAlignmentConfig(PGOpt2Config* v) {pairwiseAlignmentConfig=v;}
	   PGOpt2Config* getHoleFillingConfig() const {return holeFillingConfig;}
	   void setHoleFillingConfig(PGOpt2Config* v) {holeFillingConfig=v;}
	   PGOpt2Config* getGlobalOptConfig() const {return globalOptConfig;}
	   void setGlobalOptConfig(PGOpt2Config* v) {globalOptConfig=v;}
	   PGOpt2Config* getInitialOptConfig() const {return initialOptConfig;}
	   void setInitialOptConfig(PGOpt2Config *v) {initialOptConfig=v;}

	   
	   card32 getTopology2NumNearestNeighbors() const {return topology2NumNearestNeighbors;}
	   void setTopology2NumNearestNeighbors(card32 v) {topology2NumNearestNeighbors=v;}

	   float32 getTopology2MaxDistanceRelSampleSpacing() const {return topology2MaxDistanceRelSampleSpacing;}
	   void setTopology2MaxDistanceRelSampleSpacing(float32 v) {topology2MaxDistanceRelSampleSpacing=v;}

	   float32 getTopology2MaxAdditiveEdgeLengthDev() const {return topology2MaxAdditiveEdgeLengthDev;}
	   void setTopology2MaxAdditiveEdgeLengthDev(float32 v) {topology2MaxAdditiveEdgeLengthDev=v;}

	   float32 getTopology2MaxMultiplicativeEdgeLengthDev() const {return topology2MaxMultiplicativeEdgeLengthDev;}
	   void setTopology2MaxMultiplicativeEdgeLengthDev(float32 v) {topology2MaxMultiplicativeEdgeLengthDev=v;}

	   float32 getTopology2MinMultRejectLengthFraction() const {return topology2MinMultRejectLengthFraction;}
	   void setTopology2MinMultRejectLengthFraction(float32 v) {topology2MinMultRejectLengthFraction=v;}

	   card32 getTopology2MergeSelCriterion() const {return topology2MergeSelCriterion;}
	   void setTopology2MergeSelCriterion(card32 v) {topology2MergeSelCriterion=v;}

	   card32 getTopology2MinimumVertexDegree() const {return topology2minimumVertexDegree;}
   	void setTopology2MinimumVertexDegree(card32 v) {topology2minimumVertexDegree=v;}

		bool getOptGraph2CyclicTime() const {return optGraph2CyclicTime;}
		void setOptGraph2CyclicTime(bool v) {optGraph2CyclicTime=v;}

		bool getOptGraph2twoWayTemporalEdges() const {return optGraph2twoWayTemporalEdges;}
		void setOptGraph2twoWayTemporalEdges(bool v) {optGraph2twoWayTemporalEdges=v;}

		bool getOptGraph2useUrshape() const {return optGraph2useUrshape;}
		void setOptGraph2useUrshape(bool v) {optGraph2useUrshape=v;}


      static string getContextDataSetDescr(const card32 index);

	  inline void setSaveAfterEachMerge( bool b ){ saveAfterEachMerge = b; }
	  inline bool getSaveAfterEachMerge() const { return saveAfterEachMerge; }

      bool getAlignment2FeaturesFirstDataSecond() const {return alignment2FeaturesFirstDataSecond;}
      void setAlignment2FeaturesFirstDataSecond(bool v) {alignment2FeaturesFirstDataSecond=v;}


};
}

#endif
