//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "AnimationBSRDebugger.h"
#include "StandardSceneGraphIterator.h"
#include "ObjectClassProperty.h"
#include "PointSet_templCode.h"
#include "BSROptPotential.h"
#include "BSRFitting.h"
#include "UnifyNormals.h"
#include "DebugOutput.h"
#include "Timer.h"
#include "SeparatorClassProperty.h"
#include "SGObjectNode.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

string getColorizeDataMode( const card32 index ){
	if ( index == COLORIZE_DATA_BLUE ) return "Blue";
	else if ( index == COLORIZE_DATA_SURFACENESS ) return "Surfaceness";
	else return "Undefined";
}

IMPLEMENT_GEOXL_CLASS( AnimationBSRDebugger, 3 )
{
	BEGIN_CLASS_INIT( AnimationBSRDebugger )
	ADD_CARD32_PROP_UPDATE_DESCR( colorizeDataMode, 0, NULL, getColorizeDataMode, 2 )
	ADD_CARD32_PROP( patchGraphIndex, 2 )
	ADD_SEPARATOR( "Outliers" );
	ADD_CARD32_PROP( numOutliers, 3 )
	ADD_NOARGS_METHOD( colorizeData )
	ADD_NOARGS_METHOD( checkAttributes )
	ADD_NOARGS_METHOD( computeReferenceFrames )
	ADD_NOARGS_METHOD( computeHeightFieldMonomials )
	ADD_NOARGS_METHOD( optimizeAllPatchGraphs )
	ADD_NOARGS_METHOD( addOutliers )
}

/**
 * Constructor
 */
AnimationBSRDebugger::AnimationBSRDebugger()
{
	colorizeDataMode = COLORIZE_DATA_SURFACENESS;
	scene = NULL;
	context = NULL;
	patchGraphIndex = 0;
	numOutliers = 1000;;	
}

/**
* Destructor
*/
AnimationBSRDebugger::~AnimationBSRDebugger()
{
}

/** 
 * Colorize data
 */
void AnimationBSRDebugger::colorizeData()
{
	if ( context == NULL ) return;
	if ( scene == NULL ){
		debugOutput << "coloizeData: invalid scene object.\n"; debugOutput.flush();
		return;
	}
	UnstructuredInCorePointCloud* pc = getDataFrame( getCurrentFrameIndex() );
	if ( ( pc == NULL ) || !pc->providesAttribute( "color" ) || !pc->providesAttribute( "tensor" ) ){
		debugOutput << "colorizeData: invalid point cloud.\n"; debugOutput.flush();
		return;
	}

	PointSet* ps = pc->getPointSet();
	AAT COLOR = ps->getAAT( "color" );
	AAT SURFACENESS = ps->getAAT( "surfaceness" );
	for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
		if ( colorizeDataMode == COLORIZE_DATA_BLUE ){
			ps->set3f( pointIndex, COLOR, makeVector3f( 0.2f, 0.2f, 0.8f ) );
		} else if ( colorizeDataMode == COLORIZE_DATA_SURFACENESS ){
			float32 surfaceness = ps->get1f( pointIndex, SURFACENESS );
			Vector3f c = ( surfaceness < context->getTVSurfacenessThreshold() )? makeVector3f( 0.2f, 0.2f, 0.8f ) : makeVector3f( 0.8f, 0.2f, 0.2f );
			ps->set3f( pointIndex, COLOR, c );
		}
	}
}

void AnimationBSRDebugger::setScene( Scene* scene )
{ 
	this->scene = scene; 
}

/** 
* Get number of frames
*/
card32 AnimationBSRDebugger::getNumFrames() const
{
	SGSimpleAnimationNode* animDataNode = getDataAnimationNode();
	if ( animDataNode == NULL ){
		return 0;
	}
	return animDataNode->getNumChildNodes( scene->getRootState() );
}

/**
* Get the data animation node
*/
SGSimpleAnimationNode* AnimationBSRDebugger::getDataAnimationNode() const
{
	// Find data node
	StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
	SGSimpleAnimationNode* animDataNode = NULL;
	it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
	while ( !it->atEnd() ){
		if ( it->getNameWithPath() == "root/data" ){
			animDataNode = dynamic_cast<SGSimpleAnimationNode*>( it->getNode() );
			if ( animDataNode != NULL ){
				break;            
			}
		}
		it->next();
	}
	delete it;
	return animDataNode;
}

/**
* Return index of current frame
*/
int32 AnimationBSRDebugger::getCurrentFrameIndex() const
{
	return scene->getRootState()->dynamicState->time;
}

/**
* Get data point cloud for frame index
*/
UnstructuredInCorePointCloud* AnimationBSRDebugger::getDataFrame( card32 index )
{
	return dynamic_cast<UnstructuredInCorePointCloud*>( getPointCloud( scene, getDataFrameName( index ) ) ); 
}

/***
 * Check for required attributes
 */
void AnimationBSRDebugger::checkAttributes()
{
	 //attribute missing; intAtt, tensor, color, position
	for ( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		checkAttribute( scene, getDataFrameName( frameIndex ), "position", 3, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( scene, getDataFrameName( frameIndex ), "color", 3, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( scene, getDataFrameName( frameIndex ), "flags", 1, VAD::DATA_FORMAT_INT32 );
		checkAttribute( scene, getDataFrameName( frameIndex ), "intAtt", 1, VAD::DATA_FORMAT_INT32 );
		checkAttribute( scene, getDataFrameName( frameIndex ), "tensor", 6, VAD::DATA_FORMAT_FLOAT32 );
	}
}

/** 
* Compute reference frames
*/
void AnimationBSRDebugger::computeReferenceFrames()
{
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	if ( pg == NULL ){
		debugOutput << "computeReferenceFrames: invalid settings.\n"; debugOutput.flush();
		return;
	}
	// Compute reference frames
	BSROptVariableIndexSet indices;
	BSROptPointIndex index;
	for ( unsigned int frameIndex = 0; frameIndex < pg->getTrajectorySet()->getNumFrames(); frameIndex++ ){
		for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			index.frame = frameIndex;
			index.point = patchIndex;
			indices.push_back( index );
		}
	}
	BSROptContext* optContext = new BSROptContext();
	BSRFitting::fitRefFrames( optContext, pg, indices );
	UnifyNormals::unifyNormals( pg );
	delete optContext;
}

/**  
 * Compute height field monomials
 */
void AnimationBSRDebugger::computeHeightFieldMonomials()
{
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	if ( pg == NULL ){
		debugOutput << "computeReferenceFrames: invalid settings.\n"; debugOutput.flush();
		return;
	}
	// Compute monomials
	BSROptVariableIndexSet indices;
	BSROptPointIndex index;
	for ( unsigned int frameIndex = 0; frameIndex < pg->getTrajectorySet()->getNumFrames(); frameIndex++ ){
		for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			index.frame = frameIndex;
			index.point = patchIndex;
			indices.push_back( index );
		}
	}
	BSRFitting::fit2ndOrderHeightf( new BSROptContext(), pg, indices );
}

/**
 * Optimize all patch graphs
 */
void AnimationBSRDebugger::optimizeAllPatchGraphs()
{
   debugOutput << "Optimize all frames\n";
   Timer timer;
	// Optimizer
	PatchGraphBSROpt* optimizer = new PatchGraphBSROpt();
	BOPPrjSmoothness* smoothness = new BOPPrjSmoothness();
	optimizer->addPotential( smoothness );
	optimizer->getContext()->weightSmooth = 1.0f;
	optimizer->getContext()->simpleSmoothness = true;
	BOPSimpleICPLikelihood* likelihood = new BOPSimpleICPLikelihood();
	optimizer->addPotential( likelihood );
	optimizer->getContext()->weightICPLikelihood = 1.0f;
	optimizer->getContext()->solver = BSROptContext::SOLVER_NEWTON_CG;
	optimizer->getContext()->stepSize = 0.5f;
	for ( unsigned int index = 0; index < context->getPGContainer()->getNumPatchGraphs(); index++ ){
		PatchGraph* pg = context->getPGContainer()->getPatchGraph( index );
		// Extract indices
		BSROptVariableIndexSet indices;
		card32 numPatches = pg->getTrajectorySet()->getNumPatches();
		card32 numFrames = pg->getTrajectorySet()->getNumFrames();
		for ( unsigned int frameIndex = 0; frameIndex < numFrames; frameIndex++ ){
			for ( unsigned int patchIndex = 0; patchIndex < numPatches; patchIndex++ ){
				BSROptPointIndex i;
				i.frame = frameIndex;
				i.point = patchIndex;
				indices.push_back( i );
			}
		}
		optimizer->getContext()->stepSize = 0.5f;
		optimizer->getContext()->numIterations = 5;
		optimizer->compute( &indices, context->getPGContainer()->getAnimSeq(), pg, NULL );
	}	
	delete optimizer;
	debugOutput << "Optimize time: " << timer.getDeltaValue() << "\n\n";
}

/** 
 * Add Outliers to all data points
 */
void AnimationBSRDebugger::addOutliers()
{
	StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
	SGSimpleAnimationNode* animDataNode = NULL;
	it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
	while ( !it->atEnd() ){
		if ( it->getNameWithPath() == "root/data" ){
			animDataNode = dynamic_cast<SGSimpleAnimationNode*>( it->getNode() );
			if ( animDataNode != NULL ){
				for ( unsigned int childIndex = 0; childIndex < animDataNode->getNumChildNodes( scene->getRootState() ); childIndex++ ){
					SGObjectNode* o = dynamic_cast<SGObjectNode*>( animDataNode->getChildNode( scene->getRootState(), childIndex ) );
					if ( o != NULL ){
						UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( o->getSceneObject() );
						if ( upc != NULL ){
							UnstructuredInCorePointCloud* newPC = new UnstructuredInCorePointCloud();
							newPC->clearAndSetup( upc->getPointSet()->getDescr(), upc->getNumPoints() + numOutliers );
							AAT POSITION = upc->getPointSet()->getAAT( "position" );
							char *data = new char[upc->getPointSet()->getDescr()->getSize()];
							// Copy old points
							for ( unsigned int n = 0; n < upc->getNumPoints(); n++ ){
								upc->getPointSet()->getVertex( n, data );
								newPC->getPointSet()->setVertex( n, data );
							}
							for ( unsigned int n = 0; n < numOutliers; n++ ){
								Vector3f outlierPos = makeVector3f( rand()%200/100.0f-1.0f, rand()%200/100.0f-1.0f, rand()%200/100.0f-1.0f );
								newPC->getPointSet()->set3f( upc->getNumPoints() + n, POSITION, outlierPos );
							}
							o->clearAndSetup( newPC );
						}
					}
				}
			           
			}
		}
		it->next();
	}
	delete it;

	
}