//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ComputeSampleSpacing.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "VertexDescriptor.h"
#include "TrajectorySet.h"
#include "NumericalClassProperty.h"
#include "TGNeighborhoodKNNSearch.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



static string describeTechnique(card32 v) {
	if (v==0) return "TECHNIQUE_MAX_NEIGHBOR_DIST";
	if (v==1) return "TECHNIQUE_AVERAGE_NEIGHBOR_DIST";
	if (v==2) return "TECHNIQUE_KTH_SMALLEST";
	if (v==3) return "TECHNIQUE_KNN_AREA";
	return "undefined";
};

IMPLEMENT_GEOXL_CLASS(ComputeSampleSpacing, 2)
{
	BEGIN_CLASS_INIT(ComputeSampleSpacing)
	ADD_CARD32_PROP_UPDATE_DESCR(technique, 0, NULL, describeTechnique, 4)
	ADD_CARD32_PROP(kthSmallestIndex, 1)
	ADD_CARD32_PROP(numKNN, 2)
	ADD_FLOAT32_PROP(corrrectionFactor, 2)
}


// --


ComputeSampleSpacing::ComputeSampleSpacing() {
	technique = TECHNIQUE_KNN_AREA;
	kthSmallestIndex = 6;
	numKNN = 20;
	corrrectionFactor = 1.0f;
}

void ComputeSampleSpacing::setup(card32 technique) {
	this->technique = technique;
}

void ComputeSampleSpacing::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

ComputeSampleSpacing::~ComputeSampleSpacing() {}



void ComputeSampleSpacing::estimatePatchSamplingDistance(PatchGraph *pg) {
	TrajectorySet *ts = pg->getTrajectorySet();
	InCorePCTopologyGraph *tg = pg->getTopology();
	for (card32 i=0; i<ts->getNumFrames(); i++) {
		
		PointSet *ps = ts->copySingleFrame(i);
		
		const card32 numPatches = ps->getNumEntries();
		AAT POS = ps->getAAT("position");
		AAT SAMPLE_SPACING = ps->getAAT("sampleSpacing");
		
		if (technique == TECHNIQUE_KNN_AREA) {
			if (numPatches > 0) {
				TGNeighborhoodKNNSearch knn(ps, tg, 0, 0, "position");	
				for (card32 j=0; j<numPatches; j++) {
					knn.reset(0, j);
					card32 numNeighbors=1;
					card32 lastIndex = j;
					while (numNeighbors<numKNN && !knn.atEnd()) {
						lastIndex = knn.getCurrentIndex();
						knn.next();
						numNeighbors++;
					}
					float32 dist = norm(ps->get3f(lastIndex, POS) - ps->get3f(j, POS));
					float32 sampleSpacing;
					if (dist > 0) {
						float32 area = dist*dist*M_PI;
						sampleSpacing = sqrt(fabs(1.21*area/(M_PI*numNeighbors)));
						/// correct for border points (not exact)
						area += (dist*M_PI*2 / sampleSpacing)*sampleSpacing*sampleSpacing*M_PI/2;
						sampleSpacing = sqrt(fabs(1.21*area/(M_PI*numNeighbors)));
					} else {
						sampleSpacing = 1E36f;
					}
					ps->set<float32>(j, SAMPLE_SPACING, sampleSpacing*corrrectionFactor);
				}
			}					
		} else {
			for (card32 j=0; j<numPatches; j++) {
		
				Vector3f pos = ps->get3f(j, POS);
				const card32 numNeighbors = tg->getNumAdjacentVertices(j);
				float32 maxDist = 0;
				float32 distSum = 0;
				vector<float32> distVec;
				for (card32 t=0; t<numNeighbors; t++) {
					card32 nIndex = tg->getAdjacentVertex(j, t);
					Vector3f nPos = ps->get3f(nIndex, POS);
					float32 nDist = norm(pos-nPos);
					if (technique != TECHNIQUE_KTH_SMALLEST) {
						if (maxDist < nDist) maxDist = nDist;
						distSum += nDist;
					} else {
						distVec.push_back(nDist);
					}
				}
				if (technique != TECHNIQUE_KTH_SMALLEST) {
					if (technique == TECHNIQUE_MAX_NEIGHBOR_DIST) {
						ps->set<float32>(j, SAMPLE_SPACING, maxDist*corrrectionFactor);
					} else {
						ps->set<float32>(j, SAMPLE_SPACING, (numNeighbors>0?distSum/numNeighbors:0)*corrrectionFactor);
					}
				} else {
					sort(distVec.begin(), distVec.end());
					float32 val = 0;
					if (kthSmallestIndex == 0) kthSmallestIndex = 1;
					if (distVec.size() >= kthSmallestIndex) {
						val = distVec[kthSmallestIndex-1];
					} else if (!distVec.empty()) {
						val = distVec.back();
					}
					ps->set<float32>(j, SAMPLE_SPACING, val*corrrectionFactor);
				}
			}
		}
		ts->setSingleFrame(i, ps);
		delete ps;
	}
}
