//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DetectTopologicalErrorMerge.h"
#include "PGDataPointIndices.h"
#include "PointSetKNNQuery.h"
#include "InCorePCTopologyGraph.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( DetectTopologicalErrorMerge, 0 )
{
	BEGIN_CLASS_INIT( DetectTopologicalErrorMerge )
	ADD_INT32_PROP( numDataPointsThreshold, 0 )
	ADD_FLOAT32_PROP( invalidTrajThreshold, 0 )
	ADD_FLOAT32_PROP( invalidTopoThreshold, 0 )
}

/** 
 * Constructor
 */
DetectTopologicalErrorMerge::DetectTopologicalErrorMerge()
{
	numDataPointsThreshold = 0;
	invalidTrajThreshold = 0.2f;
	invalidTopoThreshold = 0.075f;
	context = NULL;
	knns.clear();
	pss.clear();
	octrees.clear();
}

/** 
 * Destructor
 */
DetectTopologicalErrorMerge::~DetectTopologicalErrorMerge()
{
	freeMemory();
}

/***
 * Check if a trajectory is invalid
 */
bool DetectTopologicalErrorMerge::isInvalidTrajectory( 
	const PatchGraph* pg, 
	card32 patchIndex, 
	const card32 startFrame, 
	const card32 endFrame,
	const card32 patchOffset
	)
{
	card32 numFrames = pg->getTrajectorySet()->getNumFrames();
	card32 numInvalidFrames = 0;
	AAT POSITION = pg->getTrajectorySet()->getTrajectories()->getAAT( "position" );
	card32 nIndex;
	bool found;
	for ( unsigned int frameIndex = startFrame; frameIndex <= endFrame; frameIndex++ ){
		// Find closest patch from other point cloud
		Vector3f pos = pss[frameIndex]->get3f( patchIndex, POSITION );
		knns[frameIndex].setTargetPoint( pos );
		knns[frameIndex].getNextNearestPointIndex( nIndex, found );
		float32 d = 0;
		bool foundNeighborFromOtherPatchGraph = false;
		while( found && !foundNeighborFromOtherPatchGraph && d < context->getSamplingSpacing() * 2.0f ){
			Vector3f nPos = pss[frameIndex]->get3f( nIndex, POSITION );
			d = norm( nPos - pos );
			if ( ( patchIndex < patchOffset ) && ( nIndex >= patchOffset ) ) foundNeighborFromOtherPatchGraph = true;
			else if ( ( patchIndex >= patchOffset ) && ( nIndex < patchOffset ) ) foundNeighborFromOtherPatchGraph = true;
			knns[frameIndex].getNextNearestPointIndex( nIndex, found );
		}

		if ( !foundNeighborFromOtherPatchGraph ) numInvalidFrames++;	
	}
	return ( float32( numInvalidFrames ) / float32( numFrames ) > invalidTrajThreshold );
}

/**
 * Compute ratio of invalid patches in a part of a patch graph
 */
float32 DetectTopologicalErrorMerge::computeInvalids( 
	const PatchGraph* pg, 
	const card32 startFrame, 
	const card32 endFrame,
	const card32 patchOffset
	)
{
	card32 numInvalids = 0;
	for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
		if ( isInvalidTrajectory( pg, patchIndex, startFrame, endFrame, patchOffset ) ) numInvalids++;
	}
	return float32( numInvalids) / float32( pg->getTrajectorySet()->getNumPatches() );
}

/** 
 * Free mem
 */
void DetectTopologicalErrorMerge::freeMemory()
{
	for ( unsigned int n = 0; n < octrees.size(); n++ ){
		delete octrees[n];
	}
	octrees.clear();
	knns.clear();
	for( unsigned int n = 0; n < pss.size(); n++ ){
		delete pss[n];
	}
	pss.clear();
}

/** 
 * Detect topological error in patch graph in merge operation; return true when error was detected
 */
bool DetectTopologicalErrorMerge::detectTopoError( 
	PatchGraph* combinedPG,	
	PatchGraph* pg[2],
	card32& replacePG,						
	vector<PatchGraph*>& newPatchGraphs )
{
	// Invalid topology means that the assignment of data points has changed drastically.
	card32 frameOffset = pg[0]->getTrajectorySet()->getNumFrames();
	card32 patchOffset = pg[0]->getTrajectorySet()->getNumPatches();
	card32 numFrames = combinedPG->getTrajectorySet()->getNumFrames();
	newPatchGraphs.clear();

	// Compute octree for each frames
	freeMemory();
	octrees.clear(); knns.clear(); pss.clear();
	for ( unsigned int frameIndex = 0; frameIndex < numFrames; frameIndex++ ){
		PointSet* ps = combinedPG->getTrajectorySet()->copySingleFrame( frameIndex );
		AAT POSITION = ps->getAAT( "position" );
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps );
		octrees.push_back( octree );
		knns.push_back( PointSetKNNQuery( ps, octree, NULL_VECTOR3F, POSITION ) );
		pss.push_back( ps );
	}
	
	float32 ratioPG1 = computeInvalids( pg[0], 0, pg[0]->getTrajectorySet()->getNumFrames()-1, patchOffset );
	float32 ratioPG2 = computeInvalids( pg[1], 0, pg[1]->getTrajectorySet()->getNumFrames()-1, patchOffset );
	float32 ratioCombined1 = computeInvalids( combinedPG, 0, pg[0]->getTrajectorySet()->getNumFrames()-1, patchOffset );
	float32 ratioCombined2 = computeInvalids( combinedPG, pg[0]->getTrajectorySet()->getNumFrames(), combinedPG->getTrajectorySet()->getNumFrames()-1, patchOffset );

	//debugOutput << " pg1 invalid patches ratio: " << ratioPG1 << " -> " << ratioCombined1 << ".\n"; debugOutput.flush();
	//debugOutput << " pg2 invalid patches ratio: " << ratioPG2 << " -> " << ratioCombined2 << ".\n"; debugOutput.flush();

	// Check for invalidness
	card32 startFrame, endFrame;
	float32 error;
	if( ratioCombined1 - ratioPG1 > invalidTopoThreshold ){
		// Error with pg[0] merge
		error = ratioCombined1 - ratioPG1;
		replacePG = 1;
		startFrame = 0;
		endFrame = pg[0]->getTrajectorySet()->getNumFrames()-1;
	} else if ( ratioCombined2 - ratioPG2 > invalidTopoThreshold ){
		// Error with pg[1] merge
		error = ratioCombined2 - ratioPG2;
		replacePG = 0;
		startFrame = pg[0]->getTrajectorySet()->getNumFrames();
		endFrame = combinedPG->getTrajectorySet()->getNumFrames()-1;
	} else {
		error = 0;
		freeMemory();
		return false;
	}
		
	if ( context->getVerbose() ) { debugOutput << " Breaking up topology in patch graph (invalid trajectories increase: " << ( ratioCombined2 - ratioPG2 )* 100.0f << "%).\n"; debugOutput.flush(); }

	if ( replacePG < 2){
		card32 frameOffset = pg[0]->getTrajectorySet()->getNumFrames();
		PointSet* combinedPGPS = NULL;
		card32 replaceFrameIndex;
		if ( replacePG == 0 ){
			combinedPGPS = combinedPG->getTrajectorySet()->copySingleFrame( frameOffset-1 );
			replaceFrameIndex = frameOffset-1;
		} else {
			combinedPGPS = combinedPG->getTrajectorySet()->copySingleFrame( frameOffset );
			replaceFrameIndex = 0;
		}
		AAT COMBINED_POSITION = combinedPGPS->getAAT( "position" );
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( combinedPGPS );
		PointSetKNNQuery knn( combinedPGPS, octree, NULL_VECTOR3F, COMBINED_POSITION );
		card32 nIndex;
		bool found;

		// Create two new patch graphs: split pg[replacePG] in half
		// newPG1: patch graph with valid trajectories
		// newPG2: patch graph with invalids
		set<card32> newPG1Set; newPG1Set.clear();
		set<card32> newPG2Set; newPG2Set.clear();
		vector<int32> mapping1; mapping1.resize( pg[replacePG]->getTrajectorySet()->getNumPatches() );
		vector<int32> mapping2; mapping2.resize( pg[replacePG]->getTrajectorySet()->getNumPatches() );
		AAT POSITION = pg[replacePG]->getTrajectorySet()->getTrajectories()->getAAT( "position" );
		
		for ( unsigned int patchIndex = 0; patchIndex < pg[replacePG]->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			// Find closest patch in combinedPG
			Vector3f pos = pg[replacePG]->getTrajectorySet()->getTrajectories()->get3f( replaceFrameIndex, patchIndex, POSITION );
			knn.setTargetPoint( pos );
			knn.getNextNearestPointIndex( nIndex, found );
			if ( isInvalidTrajectory( combinedPG, nIndex, startFrame, endFrame, patchOffset ) ){
				newPG2Set.insert( patchIndex );
				mapping2[patchIndex] = (int32)newPG2Set.size()-1;
				mapping1[patchIndex] = -1;
			} else {
				newPG1Set.insert( patchIndex );
				mapping1[patchIndex] = (int32)newPG1Set.size()-1;
				mapping2[patchIndex] = -1;
			}
		}

		// Create new patch graph (valid points)
		PatchGraph* newPG1 = new PatchGraph();
		TrajectorySet* newTS1 = new TrajectorySet();
		PointSet* newTSPS1 = new PointSet();
		InCorePCTopologyGraph* newTG1 = new InCorePCTopologyGraph();
		VertexDescriptor tgvd;
		tgvd.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
		tgvd.pushAttrib( VAD( "color", 4, VAD::DATA_FORMAT_FLOAT32 ) );
		newTG1->clearAndSetup( &tgvd );
		newTSPS1->clearAndSetup( pg[replacePG]->getTrajectorySet()->getNumFrames(), newPG1Set.size(), pg[replacePG]->getTrajectorySet()->getTrajectories()->getDescr() );
		newTS1->clearAndSetup( newTSPS1 );
		newPG1->clearAndSetup( newTS1, newTG1 );
		// Copy data
		card32 patchIndex = 0;
		for ( set<card32>::iterator it = newPG1Set.begin(); it != newPG1Set.end(); it++ ){
			card32 replacePatchIndex = card32( *it );
			for ( unsigned int frameIndex = 0; frameIndex < newPG1->getTrajectorySet()->getNumFrames(); frameIndex++ ){
				Vector3f pos = pg[replacePG]->getTrajectorySet()->getTrajectories()->get3f( frameIndex, replacePatchIndex, POSITION );
				newPG1->getTrajectorySet()->getTrajectories()->set3f( frameIndex, patchIndex, POSITION, pos );
				vector<card32>& indices = pg[replacePG]->getTrajectorySet()->getDataPoints()->getIndices( frameIndex, replacePatchIndex );
				newPG1->getTrajectorySet()->getDataPoints()->setIndices( frameIndex, patchIndex, indices );
			}
			patchIndex++;
		}
		newPG1->setStartFrameIndex( pg[replacePG]->getStartFrameIndex() );
		newPG1->computeBBox();
		newPatchGraphs.push_back( newPG1 );

		// Create new patch graph (invalid points)
		PatchGraph* newPG2 = new PatchGraph();
		TrajectorySet* newTS2 = new TrajectorySet();
		PointSet* newTSPS2 = new PointSet();
		InCorePCTopologyGraph* newTG2 = new InCorePCTopologyGraph();
		newTG2->clearAndSetup( &tgvd );
		newTSPS2->clearAndSetup( pg[replacePG]->getTrajectorySet()->getNumFrames(), newPG2Set.size(), pg[replacePG]->getTrajectorySet()->getTrajectories()->getDescr() );
		newTS2->clearAndSetup( newTSPS2 );
		newPG2->clearAndSetup( newTS2, newTG2 );
		// Copy data
		patchIndex = 0;
		for ( set<card32>::iterator it = newPG2Set.begin(); it != newPG2Set.end(); it++ ){
			card32 replacePatchIndex = card32( *it );
			for ( unsigned int frameIndex = 0; frameIndex < newPG2->getTrajectorySet()->getNumFrames(); frameIndex++ ){
				Vector3f pos = pg[replacePG]->getTrajectorySet()->getTrajectories()->get3f( frameIndex, replacePatchIndex, POSITION );
				newPG2->getTrajectorySet()->getTrajectories()->set3f( frameIndex, patchIndex, POSITION, pos );
				vector<card32>& indices = pg[replacePG]->getTrajectorySet()->getDataPoints()->getIndices( frameIndex, replacePatchIndex );
				newPG2->getTrajectorySet()->getDataPoints()->setIndices( frameIndex, patchIndex, indices );
			}
			patchIndex++;
		}
		newPG2->setStartFrameIndex( pg[replacePG]->getStartFrameIndex() );
		newPG2->computeBBox();
		newPatchGraphs.push_back( newPG2 );

		// Copy topology
		for ( unsigned int patchIndex = 0; patchIndex < pg[replacePG]->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			for ( unsigned int n = 0; n < pg[replacePG]->getTopology()->getNumAdjacentVertices( patchIndex ); n++ ){
				card32 nPatchIndex = pg[replacePG]->getTopology()->getAdjacentVertex( patchIndex, n );
				if ( ( mapping1[patchIndex] >= 0 ) && ( mapping1[nPatchIndex] >= 0 ) ){
					newPG1->getTopology()->addEdge( mapping1[patchIndex], mapping1[nPatchIndex] );
				} else if ( ( mapping2[patchIndex] >= 0 ) && ( mapping2[nPatchIndex] >= 0 ) ){
					newPG2->getTopology()->addEdge( mapping2[patchIndex], mapping2[nPatchIndex] );
				}
			}
		}

		delete combinedPGPS;
		delete octree;
		freeMemory();
		return true;
	} else {
		freeMemory();
		return false;
	}
	
}