//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ImplicitPatchGraphMeshing.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "VertexDescriptor.h"
#include "TrajectorySet.h"
#include "PatchTypes.h"
#include "UnstructuredInCoreTriangleMesh.h"	 
#include "MarchingCubes.h"
#include "PatchGraphRoutines.h"
#include "StandardChannels.h"
#include "PointSetKNNQuery.h"
#include "Timer.h"
#include "ProgressWindow.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



static string describeInterpKernel(card32 v) {
	if (v==0) return "PATCH_INTERP_LIN";
	if (v==1) return "PATCH_INTERP_GAUSSIAN";
	if (v==2) return "PATCH_INTERP_COS";
	return "undefined";
};


IMPLEMENT_GEOXL_CLASS(ImplicitPatchGraphMeshing, 3)
{
	BEGIN_CLASS_INIT(ImplicitPatchGraphMeshing)
	ADD_FLOAT32_PROP(marchingCubesSideLength, 0)
	ADD_BOOLEAN_PROP(flipNormals, 3)
	ADD_CARD32_PROP_UPDATE_DESCR(patchInterpolation, 0, NULL, describeInterpKernel, 3)
	ADD_FLOAT32_PROP(gaussianFilterCutoffValue, 0)
	ADD_CARD32_PROP(startGrid, 0)
	ADD_CARD32_PROP(numNewton, 0)
	ADD_FLOAT32_PROP(relativeFilterSize, 2)
	ADD_FLOAT32_PROP(minWeightSum, 1)
	ADD_BOOLEAN_PROP(localMeshing, 1)
	ADD_FLOAT32_PROP(mcThickness, 2)
}


// --


ImplicitPatchGraphMeshing::ImplicitPatchGraphMeshing() {
	marchingCubesSideLength    = 0.1f;
	flipNormals                = false;
	patchInterpolation		   = PATCH_INTERP_GAUSSIAN;
	gaussianFilterCutoffValue	= 0.001f;
	startGrid                  = 5;
	numNewton                  = 3;
	relativeFilterSize         = 1;
	minWeightSum               = 0.1f;
	localMeshing               = false;
	mcThickness                = 2.0f;
}

void ImplicitPatchGraphMeshing::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

ImplicitPatchGraphMeshing::~ImplicitPatchGraphMeshing() {}


// -


void ImplicitPatchGraphMeshing::setup(float32 marchingCubesSideLength, card32 patchInterpolation, float32 gaussianFilterCutoffValue,
                                      float32 relativeFilterSize, card32 startGrid, card32 numNewton)  
{
	this->marchingCubesSideLength   = marchingCubesSideLength;
	this->patchInterpolation	     = patchInterpolation;
	this->relativeFilterSize		  = relativeFilterSize;
	this->gaussianFilterCutoffValue = gaussianFilterCutoffValue;
	this->startGrid                 = startGrid;
	this->numNewton                 = numNewton;
}

static void quantize(BoundingBox3f &bb, const float32 cubeSideLength) {
	for (card32 i=0; i<3; i++) {
		bb.lowerCorner[i] = floor(bb.lowerCorner[i]/cubeSideLength) * cubeSideLength;
		bb.upperCorner[i] = ceil(bb.upperCorner[i]/cubeSideLength) * cubeSideLength;
	}
}

float32 ImplicitPatchGraphMeshing::computeWeight(const Vector3f &pos1, const Vector3f &pos2, float32 sigma) const{
   float32 dQuad = normQuad(pos1 - pos2);
   switch (patchInterpolation) {
      default:
      case PATCH_INTERP_LIN: {
         float32 dist = sqrt(dQuad);
         return dist > sigma ? 0 : 1.0f - dist / sigma;
      }
      case PATCH_INTERP_GAUSSIAN: {
         float32 v = exp(-dQuad / (sigma*sigma));
         if (v <= gaussianFilterCutoffValue) return 0;
         return v - gaussianFilterCutoffValue;
      }
      case PATCH_INTERP_COS: {
         float32 dist = sqrt(dQuad);
         return dist > sigma ? 0 : 0.5f*(1.0f+cos(dist/sigma*M_PI));
      }
   }
}


struct QuadraticPatch {
	Vector3f pos;
	ReferenceFramePatch ref;
	CoefficientSet qCoeffs;
	float32 sampleDistance;
	
	QuadraticPatch() : qCoeffs(6) {} 
	
	inline void get(
		const PointSet *ps, card32 patchIndex, 
		const AAT positon_aat, const AAT normal_aat, 
		const AAT tangent_u_aat, const AAT hfield_aat, 
		const AAT sample_spacing_aat) 
	{
		pos = ps->get3f(patchIndex, positon_aat);
		sampleDistance = ps->get<float32>(patchIndex, sample_spacing_aat);
		ref.get(ps, patchIndex, normal_aat, tangent_u_aat);
		qCoeffs.get(ps, patchIndex, hfield_aat);
	}
};


static inline float32 evaluate(const QuadraticPatch &patch, const Vector3f &pos, card32 startGrid, card32 numNewton) {
	Vector2f minUV;
	float32 minDist;			
	computeClosestPointQuadPatch(patch.pos, patch.ref, patch.qCoeffs, pos, patch.sampleDistance*2, minUV, minDist, startGrid, numNewton);
	return minDist;	
}

bool ImplicitPatchGraphMeshing::MyNeighborhoodSearch::inRange(const PointSet *ps, card32 index) {
   float32 sDist = ps->get<float32>(index, samplingAAT);
   Vector3f npos = ps->get3f(index, posAAT);
   if (owner->computeWeight(pos, npos, sDist * owner->getRelativeFilterSize()) > 0) {
      return true;
   } else {
      return false;
   }
   
}

void ImplicitPatchGraphMeshing::MyNeighborhoodSearch::setupMNS(
   const InCorePCTopologyGraph *tg, const PointSet *ps, card32 startIndex, 
   const Vector3f &pos, const AAT &samplingAAT, const AAT &posAAT,
   const ImplicitPatchGraphMeshing *owner)
{
   this->pos = pos;
   this->samplingAAT = samplingAAT;
   this->posAAT = posAAT;
   this->owner = owner;
   setup(tg, ps, startIndex);
}


const float32 UNDEF = 9.87654321E22f;

static inline bool ntransf(float32 v) {
   return (v != UNDEF) && (v != 0.0f);
}

static inline float32 compSign(float32 s) {
   if (s > 0) return 1e5;
   if (s < 0) return -1e5;
   pAssert(false);
   return 0;
}

struct CoordsType {
   Vector3f pos;
   CoordsType(Vector3f pos) {this->pos = pos;}
   inline bool operator<(const CoordsType &other) const {
      if (pos[0]<other.pos[0]) {
         return true;
      } else if (pos[0] == other.pos[0]) {
         if (pos[1]<other.pos[1]) {
            return true;
         } else if (pos[1] == other.pos[1]) {
            if (pos[2]<other.pos[2]) {
               return true;
            } else {
               return false;
            }
         } else {
            return false;
         }
      } else {
         return false;
      }
   }
};

struct EvalContext {
   PointSetKNNQuery *knn;
   const InCorePCTopologyGraph *tg;
   const PointSet *ps;
   AAT SAMPLE_SPACING_AAT;
   AAT POSITION_AAT;
   AAT NORMAL_AAT;
   AAT TANGENT_U_AAT;
   AAT HFIELD_AAT;
   float32 mcThickness;
   float32 relativeFilterSize;
   card32 startGrid;
   card32 numNewton;
   float32 minWeightSum;
   ImplicitPatchGraphMeshing *thisPtr;
   map<CoordsType, float> cache;
   card32 local__patchIndex;
   bool flipNormal;
};

float ImplicitPatchGraphMeshing::evalVol(void *context, float x, float y, float z) {
   EvalContext *ctx = (EvalContext*)context;

   Vector3f pos = makeVector3f(x,y,z);
   map<CoordsType, float>::iterator cachePos = ctx->cache.find(CoordsType(pos));
   if (cachePos == ctx->cache.end()) {
      
      float result = 0;
      float32 weightSum = 0;
      
      ctx->knn->setTargetPoint(pos);
      card32 startIndex;
      bool found;
      
      ctx->knn->getNextNearestPointIndex(startIndex, found);
      
      if (found) {
	      MyNeighborhoodSearch nr;
         nr.setupMNS(ctx->tg, ctx->ps, startIndex, pos, ctx->SAMPLE_SPACING_AAT, ctx->POSITION_AAT, ctx->thisPtr);
            
         if (norm(pos - ctx->ps->get3f(startIndex, ctx->POSITION_AAT)) <= ctx->mcThickness*ctx->ps->get<float32>(startIndex, ctx->SAMPLE_SPACING_AAT)) {

            while(!nr.atEnd()) {
	            card32 nIndex = nr.getCurrentIndex();
	            nr.next();
		         QuadraticPatch patch;
	            patch.get(ctx->ps, nIndex, ctx->POSITION_AAT, ctx->NORMAL_AAT, ctx->TANGENT_U_AAT, ctx->HFIELD_AAT, ctx->SAMPLE_SPACING_AAT);
	            float32 w = ctx->thisPtr->computeWeight(pos, patch.pos, patch.sampleDistance * ctx->relativeFilterSize);
	            result += w*evaluate(patch, pos, ctx->startGrid, ctx->numNewton);
	            weightSum += w;
            }
         }
      }
      if (weightSum >= ctx->minWeightSum) {
         result /= weightSum;
	      if (ctx->flipNormal) result = -result;
      } else {
         result = UNDEF;
      }   

      ctx->cache.insert(pair<CoordsType, float>(CoordsType(pos), result));
      return result;
   } else {
      return cachePos->second;
   }
}


bool ImplicitPatchGraphMeshing::evalVolIsEmpty(void *context, const BoundingBox3f &bb) {
   return false;
} 

float ImplicitPatchGraphMeshing::evalVolLocal(void *context, float x, float y, float z) {
   EvalContext *ctx = (EvalContext*)context;

   Vector3f pos = makeVector3f(x,y,z);
   map<CoordsType, float>::iterator cachePos = ctx->cache.find(CoordsType(pos));
   if (cachePos == ctx->cache.end()) {
      float32 result = 0;
      Vector3f pos = makeVector3f(x,y,z);
      
      MyNeighborhoodSearch nr;
      nr.setupMNS(ctx->tg, ctx->ps, ctx->local__patchIndex, pos, ctx->SAMPLE_SPACING_AAT, ctx->POSITION_AAT, ctx->thisPtr);
      float32 weightSum = 0;
      while (!nr.atEnd()) {
         card32 nIndex = nr.getCurrentIndex();
         nr.next();
	      QuadraticPatch npatch;
	      npatch.get(ctx->ps, nIndex, ctx->POSITION_AAT, ctx->NORMAL_AAT, ctx->TANGENT_U_AAT, ctx->HFIELD_AAT, ctx->SAMPLE_SPACING_AAT);
	      float32 w = ctx->thisPtr->computeWeight(pos, npatch.pos, npatch.sampleDistance * ctx->relativeFilterSize);
	      result += w*evaluate(npatch, pos, ctx->startGrid, ctx->numNewton);
	      weightSum += w;
      }
      if (weightSum >= ctx->minWeightSum) {
         result /= weightSum;
	      if (ctx->flipNormal) result = -result;
      } else {
         result  = UNDEF;
      }  

      ctx->cache.insert(pair<CoordsType, float>(CoordsType(pos), result));
      return result;
   } else {
      return cachePos->second;
   }
}

bool ImplicitPatchGraphMeshing::evalVolIsEmptyLocal(void *context, const BoundingBox3f &bb) {
   return false;
} 


UnstructuredInCoreTriangleMesh *ImplicitPatchGraphMeshing::createMesh(PatchGraph *pg, const card32 frame) {
	MarchingCubes mc;
	mc.setGridSnappingTolerance(0.0f);
	
	PointSet *ps = pg->getTrajectorySet()->copySingleFrame(frame);
	if ( !ps->providesAttribute( "position" ) ||
		!ps->providesAttribute( "normal" ) ||
		!ps->providesAttribute( "tangent_u" ) ||
		!ps->providesAttribute( "sampleSpacing" ) ||
		!ps->providesAttribute( HEIGHTFIELD_COEFFS_NAME ) ){
		debugOutput << "createMesh: attribute channel missing.\n"; debugOutput.flush(); 
		return NULL;
	}
	const InCorePCTopologyGraph *tg = pg->getTopology();
	const card32 numPatches = ps->getNumEntries();
	const AAT POSITION_AAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	const AAT NORMAL_AAT = ps->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
	const AAT TANGENT_U_AAT = ps->getAAT("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32);
	const AAT SAMPLE_SPACING_AAT = ps->getAAT("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32);
	const AAT HFIELD_AAT = ps->getAAT(HEIGHTFIELD_COEFFS_NAME, 6, VAD::DATA_FORMAT_FLOAT32);
	
	Timer timer;
	
   if (localMeshing) {
   	//progressWindow->pushStep( true, "local meshing" );
	   for (card32 patchIndex=0; patchIndex<numPatches; patchIndex++) {

		   QuadraticPatch patch;
		   patch.get(ps, patchIndex, POSITION_AAT, NORMAL_AAT, TANGENT_U_AAT, HFIELD_AAT, SAMPLE_SPACING_AAT);

		   BoundingBox3f bb;
		   bb.lowerCorner = patch.pos - makeVector3f(patch.sampleDistance, patch.sampleDistance, patch.sampleDistance);
		   bb.upperCorner = patch.pos + makeVector3f(patch.sampleDistance, patch.sampleDistance, patch.sampleDistance);
		   Vector3i dim;
		   quantize(bb, marchingCubesSideLength);

         for (card32 i=0; i<3; i++) {
            dim[i] = bb.getSideLength(0)/marchingCubesSideLength;
         }

         EvalContext ctx;
         ctx.knn                = NULL;
         ctx.tg                 = tg;
         ctx.ps                 = ps;
         ctx.SAMPLE_SPACING_AAT = SAMPLE_SPACING_AAT;
         ctx.POSITION_AAT       = POSITION_AAT;
         ctx.NORMAL_AAT         = NORMAL_AAT;
         ctx.TANGENT_U_AAT      = TANGENT_U_AAT;
         ctx.HFIELD_AAT         = HFIELD_AAT;
         ctx.mcThickness        = mcThickness;
         ctx.relativeFilterSize = relativeFilterSize;
         ctx.startGrid          = startGrid;
         ctx.numNewton          = numNewton;
         ctx.minWeightSum       = minWeightSum;
         ctx.thisPtr            = this;
         ctx.cache.clear();
         ctx.local__patchIndex  = patchIndex;
         ctx.flipNormal         = flipNormals;

	      mc.calculateHierarchical(evalVolLocal, &ctx, evalVolIsEmpty, bb, dim, 0, false, UNDEF);
		   //bool aborted = !progressWindow->progress( (float)patchIndex/(float)numPatches*100.0f );
		   //if (aborted) break;
	   }
   	//bool aborted;
   	//progressWindow->popSteps( aborted );
   } else {
      BoundingBox3f bb;
      bool init = false;
	   for (card32 patchIndex=0; patchIndex<numPatches; patchIndex++) {
	      if (!init) {
	         bb.lowerCorner = ps->get3f(patchIndex, POSITION_AAT);
	         bb.upperCorner = ps->get3f(patchIndex, POSITION_AAT);
	         init = true;
	      }
	      float32 sDist = ps->get<float>(patchIndex, SAMPLE_SPACING_AAT);
         bb.addPoint(ps->get3f(patchIndex, POSITION_AAT) + makeVector3f(sDist*2, sDist*2, sDist*2));
         bb.addPoint(ps->get3f(patchIndex, POSITION_AAT) - makeVector3f(sDist*2, sDist*2, sDist*2));
	   }
	   
	   quantize(bb, marchingCubesSideLength);
	   
	   Vector3i dim;
      for (card32 i=0; i<3; i++) {
         dim[i] = bb.getSideLength(0)/marchingCubesSideLength;
      }

      PointSetKNNQuery knn(ps, pg->getTrajectorySet()->fetchOctree(frame), NULL_VECTOR3F, POSITION_AAT);

      EvalContext ctx;
      ctx.knn                = &knn;
      ctx.tg                 = tg;
      ctx.ps                 = ps;
      ctx.SAMPLE_SPACING_AAT = SAMPLE_SPACING_AAT;
      ctx.POSITION_AAT       = POSITION_AAT;
      ctx.NORMAL_AAT         = NORMAL_AAT;
      ctx.TANGENT_U_AAT      = TANGENT_U_AAT;
      ctx.HFIELD_AAT         = HFIELD_AAT;
      ctx.mcThickness        = mcThickness;
      ctx.relativeFilterSize = relativeFilterSize;
      ctx.startGrid          = startGrid;
      ctx.numNewton          = numNewton;
      ctx.minWeightSum       = minWeightSum;
      ctx.thisPtr            = this;
      ctx.flipNormal         = flipNormals;
      ctx.cache.clear();

	   mc.calculateHierarchical(evalVol, &ctx, evalVolIsEmpty, bb, dim, 0, true, UNDEF);
   
   }
   
	delete ps;

	UnstructuredInCoreTriangleMesh *result = new UnstructuredInCoreTriangleMesh();
	VertexDescriptor vd;
	vd.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( VAD( "texcoord", 2, VAD::DATA_FORMAT_FLOAT32 ) );
	result->clearAndSetup(&vd, mc.getNumVertices(), &INDEX_TRI_DESCR, mc.getNumTriangles());

	PointSet *vertices = result->getPointSet();
	PointSet *triangles = result->getTriangles();

	AAT V_POS = vertices->getAAT("position");
	AAT T_INDEX = triangles->getAAT("index");

	const card32 numVertices = vertices->getNumEntries();
	for (card32 i=0; i<numVertices; i++) {
		vertices->set3f(i, V_POS, mc.getVertex(i));
	}
	const card32 numTriangles = triangles->getNumEntries();
	for (card32 i=0; i<numTriangles; i++) {
		triangles->set<int32, 3>(i, T_INDEX, mc.getTriangle(i));
	}
	
	card32 d = timer.getDeltaValue();
	
	debugOutput << "implicit meshing " << d << " ms.\n";

   if (localMeshing) {
      result->closeMesh();
      result->clearGarbage();
   }

   result->snapMesh(marchingCubesSideLength/20.0f);
   result->snapMesh(marchingCubesSideLength/20.0f);
   result->snapMesh(marchingCubesSideLength/20.0f);

   result->closeMesh();
   result->clearGarbage();

	return result;
}





