//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "NearestNeighborTopologyMerging.h"
//---------------------------------------------------------------------------
#include "PointSetKNNQuery.h"
#include "InCorePCTopologyGraph.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

void NearestNeighborTopologyMerging::stich_Modified(
	const PointSet* source, 
	const PointSet* target, 
	const float32 pruneDist,
	const card32 sourcePatchIndexOffset,
	const card32 targetPatchIndexOffset,
	AttachedIndexedOctree *targetOctree,
	set< pair<card32, card32> > &result) 
{
	const AAT SOURCE_POS_AAT = source->getAAT("position");
	const AAT TARGET_POS_AAT = target->getAAT("position");
	PointSetKNNQuery targetKNN(target, targetOctree, NULL_VECTOR3F, TARGET_POS_AAT);

	for (unsigned i=0; i<source->getNumEntries(); i++) 
	{
		Vector3f pos = source->get<float32, 3>(i, SOURCE_POS_AAT);
		targetKNN.setTargetPoint(pos);
		bool found = false;
		card32 match;
		card32	numOfFoundNeighbors = 0;
		do 
		{
			targetKNN.getNextNearestPointIndex(match, found);
			if (found) 
			{
				Vector3f matchPos = target->get<float32, 3>(match, TARGET_POS_AAT);
				float32 dist = norm(pos-matchPos);
				if (dist <= pruneDist)  
				{
					card32 v1 = i+sourcePatchIndexOffset;
					card32 v2 = match+targetPatchIndexOffset;
					pair<card32, card32> e(min(v1,v2), max(v1,v2));
					result.insert(e);
				}
				numOfFoundNeighbors++;
			}
			else
			{
				break;
			}
		} while(numOfFoundNeighbors < 2);
	}
}


void NearestNeighborTopologyMerging::stich(
	const PointSet* source, 
	const PointSet* target, 
	const float32 pruneDist,
	const card32 sourcePatchIndexOffset,
	const card32 targetPatchIndexOffset,
	AttachedIndexedOctree *targetOctree,
	set< pair<card32, card32> > &result) 
{
	const AAT SOURCE_POS_AAT = source->getAAT("position");
	const AAT TARGET_POS_AAT = target->getAAT("position");
	PointSetKNNQuery targetKNN(target, targetOctree, NULL_VECTOR3F, TARGET_POS_AAT);
	
	for (unsigned i=0; i<source->getNumEntries(); i++) 
	{
		Vector3f pos = source->get<float32, 3>(i, SOURCE_POS_AAT);
		targetKNN.setTargetPoint(pos);
		bool found = false;
		card32 match;
		targetKNN.getNextNearestPointIndex(match, found);
		if (found) 
		{
			Vector3f matchPos = target->get<float32, 3>(match, TARGET_POS_AAT);
			float32 dist = norm(pos-matchPos);
			if (dist <= pruneDist)  
			{
				card32 v1 = i+sourcePatchIndexOffset;
				card32 v2 = match+targetPatchIndexOffset;
				pair<card32, card32> e(min(v1,v2), max(v1,v2));
				result.insert(e);
			}
		}
	}
}


void NearestNeighborTopologyMerging::connectNeighbors
  (AnimationBSRContext* context,
	const PointSet* interpolatedFrame, const PointSet* targetFrame, 
	PatchGraph* combinedPG,
	const card32 sourcePatchIndexOffset,
	const card32 targetPatchIndexOffset,
	DebugRenderer *debugRenderer,  
	const bool verbose)
{
	AttachedIndexedOctree *interpolatedOctree = PointSetKNNQuery::createOctree(interpolatedFrame);
	AttachedIndexedOctree *targetOctree			= PointSetKNNQuery::createOctree(targetFrame);
	
	set< pair<card32, card32> > edges;
	
	float32 pruneDist = context->getSamplingSpacing()*context->getTopMergePruneDistFactor();
	stich(interpolatedFrame, targetFrame, pruneDist, sourcePatchIndexOffset, targetPatchIndexOffset, targetOctree, edges);
	stich(targetFrame, interpolatedFrame, pruneDist, targetPatchIndexOffset, sourcePatchIndexOffset, interpolatedOctree, edges);
	
	set< pair<card32, card32> >::iterator pos = edges.begin();
	set< pair<card32, card32> >::iterator end = edges.end();
	
	InCorePCTopologyGraph *tg = combinedPG->getTopology();
	
	while (pos != end) {
		/// addEdge is symmetric by design (undirected graph)
		tg->addEdge(pos->first, pos->second);
		++pos;
	}
	
	delete interpolatedOctree;
	delete targetOctree;
}

void NearestNeighborTopologyMerging::connectNeighbors_Modified(
 AnimationBSRContext* context,
 const PointSet* interpolatedFrame, const PointSet* targetFrame, 
 InCorePCTopologyGraph	*topology,
 const card32 sourcePatchIndexOffset,
 const card32 targetPatchIndexOffset,
 DebugRenderer *debugRenderer,  
 const bool verbose)
{
	AttachedIndexedOctree *interpolatedOctree = PointSetKNNQuery::createOctree(interpolatedFrame);
	AttachedIndexedOctree *targetOctree			= PointSetKNNQuery::createOctree(targetFrame);

	set< pair<card32, card32> > edges;

	float32 pruneDist = context->getSamplingSpacing()*context->getTopMergePruneDistFactor();
	stich_Modified(interpolatedFrame, targetFrame, pruneDist, sourcePatchIndexOffset, targetPatchIndexOffset, targetOctree, edges);
	stich_Modified(targetFrame, interpolatedFrame, pruneDist, targetPatchIndexOffset, sourcePatchIndexOffset, interpolatedOctree, edges);

	set< pair<card32, card32> >::iterator pos = edges.begin();
	set< pair<card32, card32> >::iterator end = edges.end();


	while (pos != end) {
		/// addEdge is symmetric by design (undirected graph)
		topology->addEdge(pos->first, pos->second);
		++pos;
	}

	delete interpolatedOctree;
	delete targetOctree;
}

struct edge_MG{
public:
	float32	weight;
	card32	indexMin;
	card32	indexMax;
public:
	edge_MG(){}
	inline edge_MG(card32 v1, card32 v2, float32 w)
	{
		pAssert(v1 != v2);
		indexMin		= min(v1, v2);
		indexMax	= max(v1, v2);
		this->weight = w;
	}
	inline const bool operator < (const edge_MG &e) const
	{
		if(weight < e.weight)
		{
			return true;
		}
		else if(weight > e.weight)
		{
			return false;
		}
		else if(indexMax < e.indexMax)
		{
			return true;
		}
		else if(indexMax > e.indexMax)
		{
			return false;
		}
		else
		{
			return indexMin < e.indexMin;
		}
	}
};

static bool isGoodEdge(const PatchGraph* pg, card32 v1, card32 v2, float32 stretch_Threshold, float32 max_dis)
{
	const	PointSet	*ps = pg->getTrajectorySet()->getTrajectories();
	AAT		POSITION		= ps->getAAT("position");

	card32	width = ps->getWidth();

	float32	min_sqr = 1e10, max_sqr = -1e10;

	for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
	{
		Vector3f	pik = ps->get3f(frameIndex, v1, POSITION);
		Vector3f	pjk = ps->get3f(frameIndex, v2, POSITION);
		Vector3f	rijk = pik - pjk;
		if(min_sqr > rijk.getSqrNorm())
			min_sqr = rijk.getSqrNorm();
		if(max_sqr < rijk.getSqrNorm())
			max_sqr = rijk.getSqrNorm();
	}
	if(max_sqr/min_sqr < stretch_Threshold*stretch_Threshold && max_sqr < max_dis*max_dis)
		return true;
	else
		return false;
}

static	bool	isSource(card32 v1, card32 sourcePatchIndexOffset, card32 sourceHeight)
{
	if(sourcePatchIndexOffset <=v1 && sourcePatchIndexOffset + sourceHeight > v1)
	{
		return true;
	}
	else
	{
		return false;
	}
}

static	void	insert_A_Edge(vector<set<card32> >	*edges, card32 v1, card32 v2, 
							   const PatchGraph *sourcePG, card32 sourcePatchIndexOffset, card32 sourceHeight,
							   const PatchGraph *targetPG, card32 targetPatchIndexOffset, card32 targetHeight,
							   float32 max_dist, float32 maxRelEdgeLength)
{
	bool flag = true;
	
	bool	v1_source = isSource(v1, sourcePatchIndexOffset, sourceHeight),
			v2_source = isSource(v2, sourcePatchIndexOffset, sourceHeight);
	
	set<card32>::iterator	iter_v1, iter_v2;
	if(v1_source)
	{
		if(v2_source)
		{
			//traverse the neighbors of v1 and v2
			for(iter_v1 = (*edges)[v1].begin();	iter_v1 != (*edges)[v1].end(); iter_v1++)
			{
				card32 neighbor_id1 = *iter_v1;
				for(iter_v2 = (*edges)[v2].begin();	iter_v2 != (*edges)[v2].end(); iter_v2++)
				{
					card32 neighbor_id2 = *iter_v2;
					if(!isSource(neighbor_id1, sourcePatchIndexOffset, sourceHeight)
						&&!isSource(neighbor_id2, sourcePatchIndexOffset, sourceHeight))
					{
						if(!isGoodEdge(targetPG, neighbor_id1 - targetPatchIndexOffset,
							neighbor_id2 - targetPatchIndexOffset, maxRelEdgeLength, max_dist))
						{
							flag = false;
							break;
						}
					}
				}
				if(!flag)
				{
					break;
				}
			}
		}
		else
		{
			for(iter_v2 = (*edges)[v2].begin();	iter_v2 != (*edges)[v2].end(); iter_v2++)
			{
				card32 neighbor_id2 = *iter_v2;
				if(isSource(neighbor_id2, sourcePatchIndexOffset, sourceHeight))
				{
					if(!isGoodEdge(sourcePG, v1 - sourcePatchIndexOffset,
						neighbor_id2 - sourcePatchIndexOffset, maxRelEdgeLength, max_dist))
					{
						flag = false;
						break;
					}
				}
			}
			for(iter_v1 = (*edges)[v1].begin();	iter_v1 != (*edges)[v1].end(); iter_v1++)
			{
				card32 neighbor_id1 = *iter_v1;
				if(!isSource(neighbor_id1, sourcePatchIndexOffset, sourceHeight))
				{
					if(!isGoodEdge(targetPG, neighbor_id1 - targetPatchIndexOffset,
						v2 - targetPatchIndexOffset, maxRelEdgeLength, max_dist))
					{
						flag = false;
						break;
					}
				}
			}
		}
	}
	else 
	{
		if(v2_source)
		{

			for(iter_v2 = (*edges)[v2].begin();	iter_v2 != (*edges)[v2].end(); iter_v2++)
			{
				card32 neighbor_id2 = *iter_v2;
				if(!isSource(neighbor_id2, sourcePatchIndexOffset, sourceHeight))
				{
					if(!isGoodEdge(targetPG, v1 - targetPatchIndexOffset,
						neighbor_id2 - targetPatchIndexOffset, maxRelEdgeLength, max_dist))
					{
						flag = false;
						break;
					}
				}
			}
			for(iter_v1 = (*edges)[v1].begin();	iter_v1 != (*edges)[v1].end(); iter_v1++)
			{
				card32 neighbor_id1 = *iter_v1;
				if(isSource(neighbor_id1, sourcePatchIndexOffset, sourceHeight))
				{
					if(!isGoodEdge(sourcePG, neighbor_id1 - sourcePatchIndexOffset,
						v2 - sourcePatchIndexOffset, maxRelEdgeLength, max_dist))
					{
						flag = false;
						break;
					}
				}
			}
		}
		else
		{
			//traverse the neighbors of v1 and v2
			for(iter_v1 = (*edges)[v1].begin();	iter_v1 != (*edges)[v1].end(); iter_v1++)
			{
				card32 neighbor_id1 = *iter_v1;
				for(iter_v2 = (*edges)[v2].begin();	iter_v2 != (*edges)[v2].end(); iter_v2++)
				{
					card32 neighbor_id2 = *iter_v2;
					if(isSource(neighbor_id1, sourcePatchIndexOffset, sourceHeight)
						&&isSource(neighbor_id2, sourcePatchIndexOffset, sourceHeight))
					{
						if(!isGoodEdge(sourcePG, neighbor_id1 - sourcePatchIndexOffset,
							neighbor_id2 - sourcePatchIndexOffset, maxRelEdgeLength, max_dist))
						{
							flag = false;
							break;
						}
					}
				}
				if(!flag)
				{
					break;
				}
			}
		}
	}
	if(flag)
	{
		(*edges)[v1].insert(v2);
		(*edges)[v2].insert(v1);
	}
}

void	NearestNeighborTopologyMerging::connectNeighbors_Final(
	AnimationBSRContext* context, 
	const PointSet* interpolatedFrame, 
	const PointSet* targetFrame, 
	const PatchGraph* interpolatedPatchGraph, 
	const PatchGraph* targetPatchGraph, 
	InCorePCTopologyGraph *topology,
	const card32 sourcePatchIndexOffset,
	const card32 targetPatchIndexOffset,
	DebugRenderer *debugRenderer, 
	const bool verbose /* = false  */)
{
	//create  knn structure for nearest neighbor search
	AttachedIndexedOctree *interpolatedOctree	= PointSetKNNQuery::createOctree(interpolatedFrame);
	AttachedIndexedOctree *targetOctree			= PointSetKNNQuery::createOctree(targetFrame);

	AAT		SOURCE_POS	= interpolatedFrame->getAAT("position");
	AAT		TARGET_POS	= targetFrame->getAAT("position");
	PointSetKNNQuery targetKNN(targetFrame,		  targetOctree, NULL_VECTOR3F, TARGET_POS);
	PointSetKNNQuery interpolatedKNN(interpolatedFrame, interpolatedOctree, NULL_VECTOR3F, SOURCE_POS);

	vector<set<card32> >		edges;
	set<card32>::iterator	iter_set;
	card32	height = interpolatedFrame->getHeight() + targetFrame->getHeight();
	edges.resize(height);
	
	float32 pruneDist = context->getSamplingSpacing()*context->getTopMergePruneDistFactor();

	for (card32 surfelIndex=0; surfelIndex < interpolatedFrame->getHeight(); surfelIndex++) 
	{
		Vector3f pos = interpolatedFrame->get<float32, 3>(surfelIndex, SOURCE_POS);
		targetKNN.setTargetPoint(pos);
		bool found = false;
		card32 match;
		card32	numOfFoundNeighbors = 0;
		do 
		{
			targetKNN.getNextNearestPointIndex(match, found);
			if (found) 
			{
				Vector3f matchPos = targetFrame->get<float32, 3>(match, TARGET_POS);
				float32 dist = norm(pos-matchPos);
				if (dist <= pruneDist)  
				{
					insert_A_Edge(&edges, surfelIndex + sourcePatchIndexOffset, match + targetPatchIndexOffset, 
						interpolatedPatchGraph, sourcePatchIndexOffset, interpolatedFrame->getHeight(), 
						targetPatchGraph, targetPatchIndexOffset, targetFrame->getHeight(), context->maxEdgeStretchFactor*pruneDist,
						context->maxRelEdgeLength
						);
				}
				numOfFoundNeighbors++;
			}
			else
			{
				break;
			}
		} while(numOfFoundNeighbors < 4);
	}

	for (card32 surfelIndex=0; surfelIndex < targetFrame->getHeight(); surfelIndex++) 
	{
		Vector3f pos = targetFrame->get<float32, 3>(surfelIndex, TARGET_POS);
		targetKNN.setTargetPoint(pos);
		bool found = false;
		card32 match;
		card32	numOfFoundNeighbors = 0;
		do 
		{
			interpolatedKNN.getNextNearestPointIndex(match, found);
			if (found) 
			{
				Vector3f matchPos = interpolatedFrame->get<float32, 3>(match, SOURCE_POS);
				float32 dist = norm(pos-matchPos);
				if (dist <= pruneDist)  
				{
					insert_A_Edge(&edges, surfelIndex + targetPatchIndexOffset, match + sourcePatchIndexOffset, 
						interpolatedPatchGraph, sourcePatchIndexOffset, interpolatedFrame->getHeight(), 
						targetPatchGraph, targetPatchIndexOffset, targetFrame->getHeight(), context->maxEdgeStretchFactor*pruneDist,
						context->maxRelEdgeLength
						);
				}
				numOfFoundNeighbors++;
			}
			else
			{
				break;
			}
		} while(numOfFoundNeighbors < 4);
	}

	
	//Add old topology to the list
	const InCorePCTopologyGraph	*interpolated_tg = interpolatedPatchGraph->getTopology();
	for(card32 surfelIndex = 0; surfelIndex < interpolatedFrame->getHeight(); surfelIndex++)
	{
		Vector3f	pik = interpolatedFrame->get3f(0, surfelIndex, SOURCE_POS);
		card32 numOfEntries = interpolated_tg->getNumAdjacentVertices(surfelIndex);
		for(card32 ni = 0; ni < numOfEntries; ni++)
		{
			card32 neighbor_id = interpolated_tg->getAdjacentVertex(surfelIndex, ni);
			if(surfelIndex < neighbor_id)
			{
				insert_A_Edge(&edges, surfelIndex + sourcePatchIndexOffset, neighbor_id + sourcePatchIndexOffset, 
					interpolatedPatchGraph, sourcePatchIndexOffset, interpolatedFrame->getHeight(), 
					targetPatchGraph, targetPatchIndexOffset, targetFrame->getHeight(), context->maxEdgeStretchFactor*pruneDist,
					context->maxRelEdgeLength
					);
			}
		}
	}

	const InCorePCTopologyGraph	*target_tg		 = targetPatchGraph->getTopology();
	for(card32 surfelIndex = 0; surfelIndex < targetFrame->getHeight(); surfelIndex++)
	{
		Vector3f	pik = targetFrame->get3f(0, surfelIndex, TARGET_POS);
		card32 numOfEntries = target_tg->getNumAdjacentVertices(surfelIndex);
		for(card32 ni = 0; ni < numOfEntries; ni++)
		{
			card32 neighbor_id = target_tg->getAdjacentVertex(surfelIndex, ni);
			if(surfelIndex < neighbor_id)
			{
				insert_A_Edge(&edges, surfelIndex + targetPatchIndexOffset, neighbor_id + targetPatchIndexOffset, 
					interpolatedPatchGraph, sourcePatchIndexOffset, interpolatedFrame->getHeight(), 
					targetPatchGraph, targetPatchIndexOffset, targetFrame->getHeight(), context->maxEdgeStretchFactor*pruneDist,
					context->maxRelEdgeLength
					);
			}
		}
	}


	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		for(iter_set = edges[surfelIndex].begin();
			iter_set != edges[surfelIndex].end();
			iter_set++)
		{
			card32 neighbor_id = *iter_set;
			if(surfelIndex < neighbor_id)
			{
				topology->addEdge(surfelIndex, neighbor_id);
			}
		}
	}

	delete interpolatedOctree;
	delete targetOctree;
}

