//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchApprox3D.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "VertexDescriptor.h"
#include "TrajectorySet.h"
#include "PatchTypes.h"
#include "PointCloudAnimationSequence.h"
#include "PGDataPointIndices.h"
#include "HeightFieldBasisFunctionSet.h"
#include "PatchTypes.h"
#include "LeastSquaresSystemSetup.h"
#include "LeastSquaresSystemSetup_templCode.h"
#include "VariablesNaming.h"
#include "VariablesNaming_templCode.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers_templCode.h"
#include "Timer.h"
#include "SparseScatterOperations.h"
#include "SparseScatterOperations_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



static string describeWindow(card32 v) {
	if (v==0) return "WINDOW_NONE";
	if (v==1) return "WINDOW_GAUSSIAN";
	if (v==2) return "WINDOW_LIN";
	return "undefined";
};

static string describeSizing(card32 v) {
	if (v==0) return "WINDOW_FIXED";
	if (v==1) return "WINDOW_CHANNEL";
	return "undefined";
};

IMPLEMENT_GEOXL_CLASS(PatchApprox3D, 1)
{
	BEGIN_CLASS_INIT(PatchApprox3D)
	ADD_CARD32_PROP_UPDATE_DESCR(windowingFunction, 1, NULL, describeWindow, 3)
   ADD_CARD32_PROP_UPDATE_DESCR(windowSizing, 1, NULL, describeSizing, 2)
   ADD_FLOAT32_PROP(constantWindowSize, 1)
}


// --


PatchApprox3D::PatchApprox3D() {
	bfs = NULL;
	windowingFunction = WINDOW_NONE;
	windowSizing = WINDOW_FIXED;
	constantWindowSize = 1;
}

void PatchApprox3D::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

PatchApprox3D::~PatchApprox3D() {
}


float32 PatchApprox3D::winGaussian(const Vector2f &pos, float32 windowSize) {
	return exp(-pos*pos/(windowSize*windowSize));
}

float32 PatchApprox3D::winLin(const Vector2f &pos, float32 windowSize) {
	float32 v = 1.0f - norm(pos)/windowSize;
	if (v>=0.0f) 
		return v; 
	else 
		return 0.0f;
}

void PatchApprox3D::setup(
	HeightFieldBasisFunctionSet *bfs, 
	card32 windowingFunction, 
	card32 windowSizing, 
	float32 constantWindowSize)
{
	this->bfs = bfs;
	this->windowingFunction = windowingFunction;
	this->windowSizing = windowSizing;
	this->constantWindowSize = constantWindowSize;
}
																    
void PatchApprox3D::computeCoeffsUnregularized(PatchGraph *pg, PointCloudAnimationSequence *data)
{
   Timer timer;
	pAssert(bfs != NULL);
	WeightFunction wFunc = NULL;
	if (windowingFunction == WINDOW_GAUSSIAN) {
		wFunc = winGaussian;
	} else if (windowingFunction == WINDOW_LIN) {
		wFunc = winLin;
	}

	DVectorF targetValues;
	DMatrixF basisFunctionValues;
	DVectorF targetValuesWeights;
	DMatrixF mat, invMat;
	DVectorF rhs;
	CoefficientSet coeffs(bfs->getNumBasisFunctions());
			
	TrajectorySet *ts = pg->getTrajectorySet();
	InCorePCTopologyGraph *tg = pg->getTopology();
	PGDataPointIndices *dataIndices = ts->getDataPoints();
	for (card32 frame=0; frame<ts->getNumFrames(); frame++) {
		PointSet *patchPS = ts->copySingleFrame(frame);
		PointSet *dataPS = data->getFrame(frame);
		AAT DATA_POS = dataPS->getAAT("position");
		AAT PATCH_POS = patchPS->getAAT("position");
		AAT NORMAL = patchPS->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
		AAT TANGENT_U = patchPS->getAAT("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32);
		AAT COEFFS_AAT = patchPS->getAAT(HEIGHTFIELD_COEFFS_NAME);
		AAT SAMPLE_SPACING;
		if (windowSizing == WINDOW_CHANNEL) {
			SAMPLE_SPACING = patchPS->getAAT("sampleSpacing");
		}
		
		const card32 numPatches = ts->getNumPatches();
		for (card32 patch=0; patch<numPatches; patch++) {
			ReferenceFramePatch refFrame;
			refFrame.get(patchPS, patch, NORMAL, TANGENT_U);
			Matrix3f transf;
			transf[0] = refFrame.tangent_u;
			transf[1] = refFrame.computeTangentV();
			transf[2] = refFrame.normal;
			transf = transf.transpose();
			Vector3f origin = patchPS->get3f(patch, PATCH_POS);
		
			float32 windowSize = constantWindowSize;
			if (windowSizing == WINDOW_CHANNEL) {
				windowSizing *= patchPS->get<float32, 1>(patch, SAMPLE_SPACING)[0];
			}
		
			const card32 numDataPoints = dataIndices->getNumDataPoints(frame, patch);
			vector<Vector3f> points;
			for (card32 dp=0; dp<numDataPoints; dp++) {
				points.push_back(dataPS->get3f(dataIndices->getIndex(frame, patch, dp), DATA_POS));
			}
			
			setupDenseLS(bfs, wFunc, points, origin, transf,
						    targetValues, basisFunctionValues, &targetValuesWeights, windowSize);

			DenseLeastSquaresSystemSetup<float32> dlss;
			dlss.systemSetup(targetValues, basisFunctionValues, mat, rhs, wFunc?&targetValuesWeights:NULL);
			//#define TEST_CG
			#ifdef TEST_CG						  
			SparseMatrix<float32> testCGMat;
			convertDenseToSparse(mat, testCGMat);		   
	      conjugateGradientsSolve<float32>(testCGMat, coeffs.coeffs, rhs, 1.0E-5f, 10, true);
		   //gaussSeidelSolve<float32>(testCGMat, coeffs.coeffs, rhs, 1E-20f, 1E-7f, 1000, true);
			#else
			invMat = invertMatrix(mat);
			coeffs.coeffs = invMat * rhs;
			//debugOutput << coeffs.coeffs.toString() << "\n";
			#endif
			coeffs.set(patchPS, patch, COEFFS_AAT);
		}
		
		ts->setSingleFrame(frame, patchPS);
		delete patchPS;
	}
	card32 t = timer.getDeltaValue();
	debugOutput <<"computeCoeffsUnregularized: " << t << "ms.\n";
}


void PatchApprox3D::computeCoeffsRegularized(PatchGraph *pg, PointCloudAnimationSequence *data)
{
   Timer timer;
   card32 timerValue = timer.getCurrentValue();
	pAssert(bfs != NULL);
	WeightFunction wFunc = NULL;
	if (windowingFunction == WINDOW_GAUSSIAN) {
		wFunc = winGaussian;
	} else if (windowingFunction == WINDOW_LIN) {
		wFunc = winLin;
	}

	DVectorF targetValues;
	DMatrixF basisFunctionValues;
	DVectorF targetValuesWeights;
	DMatrixF mat, invMat;
	DVectorF rhs;
	CoefficientSet coeffs(bfs->getNumBasisFunctions());
			
	TrajectorySet *ts = pg->getTrajectorySet();
	InCorePCTopologyGraph *tg = pg->getTopology();
	PGDataPointIndices *dataIndices = ts->getDataPoints();  
	
	for (card32 frame=0; frame<ts->getNumFrames(); frame++) {
		PointSet *patchPS = ts->copySingleFrame(frame);
		PointSet *dataPS = data->getFrame(frame);
		AAT DATA_POS = dataPS->getAAT("position");
		AAT PATCH_POS = patchPS->getAAT("position");
		AAT NORMAL = patchPS->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
		AAT TANGENT_U = patchPS->getAAT("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32);
		AAT COEFFS_AAT = patchPS->getAAT(HEIGHTFIELD_COEFFS_NAME);
		AAT SAMPLE_SPACING;
		const card32 NUM_BASIS_FUNCTIONS = bfs->getNumBasisFunctions();
		if (windowSizing == WINDOW_CHANNEL) {
			SAMPLE_SPACING = patchPS->getAAT("sampleSpacing");
		}
							  
		const card32 numPatches = ts->getNumPatches();	  

      SparseMatrixF systemMatrix(numPatches*NUM_BASIS_FUNCTIONS);
      DVectorF systemRHS = nullDVector<float32>(numPatches*NUM_BASIS_FUNCTIONS);
      VariablesNamingScheme naming;
      naming.addAttrib(VAR_AD("hfield", NUM_BASIS_FUNCTIONS));
      VAR_AAT vaatHField = naming.getAAT("hfield");
		
		for (card32 patch=0; patch<numPatches; patch++) {
			ReferenceFramePatch refFrame;
			refFrame.get(patchPS, patch, NORMAL, TANGENT_U);
			Matrix3f transf;
			transf[0] = refFrame.tangent_u;
			transf[1] = refFrame.computeTangentV();
			transf[2] = refFrame.normal;
			transf = transf.transpose();
			Vector3f origin = patchPS->get3f(patch, PATCH_POS);
		
			float32 windowSize = constantWindowSize;
			if (windowSizing == WINDOW_CHANNEL) {
				windowSizing *= patchPS->get<float32, 1>(patch, SAMPLE_SPACING)[0];
			}
		
			const card32 numDataPoints = dataIndices->getNumDataPoints(frame, patch);
			vector<Vector3f> points;
			for (card32 dp=0; dp<numDataPoints; dp++) {
				points.push_back(dataPS->get3f(dataIndices->getIndex(frame, patch, dp), DATA_POS));
			}
			
			setupDenseLS(bfs, wFunc, points, origin, transf,
						    targetValues, basisFunctionValues, &targetValuesWeights, windowSize);

			DenseLeastSquaresSystemSetup<float32> dlss;
			dlss.systemSetup(targetValues, basisFunctionValues, mat, rhs, wFunc?&targetValuesWeights:NULL);
			
         mat = mat;
			scatterAndAddDenseMatrixByName(systemMatrix, mat, vaatHField, patch);
			rhs = rhs;
			scatterAndAddVectorByName(systemRHS, rhs, vaatHField, patch);
		}
   	
   	card32 t = timer.getDeltaValue();

      DVectorF solution = nullDVector<float32>(systemRHS.getDim());
      conjugateGradientsSolveOtherPrecision(systemMatrix, solution, systemRHS, 1.0E-5f, 1000, false);
      //conjugateGradientsSolveOtherPrecision<float32, float64>(systemMatrix, solution, systemRHS, 1.0E-5f, 1000, false);
      //gaussSeidelSolve<float32>(systemMatrix, solution, systemRHS, 1E-20f, 1E-7f, 1000, true);

      NamedDynamicVector<float32> namedSolution(solution);

		for (card32 patch=0; patch<numPatches; patch++) {
         for (card32 i=0; i<NUM_BASIS_FUNCTIONS; i++) {
            coeffs[i] = namedSolution.get(vaatHField, i, patch);
    		}
  			coeffs.set(patchPS, patch, COEFFS_AAT);
		}

   	t = timer.getDeltaValue();
	   debugOutput <<"computeCoeffsRegularized (solve): " << t << "ms.\n";

		ts->setSingleFrame(frame, patchPS);
		delete patchPS;
	}

	card32 t = timer.getCurrentValue() - timerValue;
	debugOutput <<"computeCoeffsRegularized: " << t << "ms.\n";

}


// ----


void geoxl::setupDenseLS(HeightFieldBasisFunctionSet *basisFunctions,
		 					  WeightFunction weightFunction,
							  const vector<Vector3f> &points,
						     const Vector3f &origin,
						     const Matrix3f &transf,
						     DynamicVector<float32> &targetValues,
						     DynamicMatrix<float32> &basisFunctionValues,
							  DynamicVector<float32> *targetValuesWeights,
							  float32 windowSize)
{
	const unsigned numBasisFunctions = basisFunctions->getNumBasisFunctions();
	const mpcard numPoints = points.size();
	pAssert(numBasisFunctions != 0);
	pAssert(numPoints !=0);
	pAssert(basisFunctions);
	targetValues.setDim(numPoints);
	basisFunctionValues.setDimension(numBasisFunctions, numPoints, false);
	if (targetValuesWeights) {
		targetValuesWeights->setDim(numPoints);
	}
	for (unsigned point=0; point<numPoints; point++) {
		Vector3f transformedPoint = transf * (points[point] - origin);
		Vector2f uv = makeVector2f(transformedPoint[0], transformedPoint[1]);
		targetValues[point] = transformedPoint[2];
		if (targetValuesWeights && weightFunction) {
			(*targetValuesWeights)[point] = (*weightFunction)(uv, windowSize);
		}
		for (unsigned bf=0; bf<numBasisFunctions; bf++) {
			basisFunctionValues[bf][point] = basisFunctions->evaluate(bf, uv);
		}
	}
}
