//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphExtract.h"
//---------------------------------------------------------------------------
#include "PatchGraphContainer.h"
#include "CopyObjectProperties.h"
#include "PGDataPointIndices.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "TopologicalKNNIterator.h"
#include "PointSetKNNQuery.h"
#include "ClassMethods.h"
#include "StandardSceneGraphIterator.h"
#include "ProgressWindow.h"
#include "BSROptContext.h"
#include "PatchTypes.h"
#include "DataPointAssignment.h"
#include "SGObjectNode.h"
#include "FifoStack.h"
#include <time.h>
#include <stdio.h>
#include <dos.h>
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphExtract, 3 )
{
	BEGIN_CLASS_INIT( PatchGraphExtract )
	ADD_OBJECT_PROP( tv, 0, TensorVoting::getClass(), true )
	ADD_OBJECT_PROP( resampler, 0, PatchGraphResample::getClass(), true )
	ADD_NOARGS_METHOD( checkAttributes )
	ADD_NOARGS_METHOD( align )
	ADD_NOARGS_METHOD( tensorVoteFrame )
	ADD_NOARGS_METHOD( extractPatchGraph )
	ADD_NOARGS_METHOD( tensorVoteAllFrames )
	ADD_NOARGS_METHOD( extractPatchGraphAllFrames )
	ADD_NOARGS_METHOD( removeOutliers );
	//ADD_NOARGS_METHOD( detectComponents )
	//ADD_OBJECT_PROP( altResampler, 0, PatchGraphTopologicalResampling::getClass(), true )
}

/**
 * Constructor
 */
PatchGraphExtract::PatchGraphExtract() 
{
	context = NULL;
	scene = NULL;
	tv = new TensorVoting();
	resampler = new PatchGraphResample();
}

/**
 * Destructor
 */
PatchGraphExtract::~PatchGraphExtract() 
{
	delete tv;
	delete resampler;
}

/** 
 * Object assignment
 */
void PatchGraphExtract::assign( const Object* obj, CopyContext *context ) 
{
	copyObjectProperties( obj, this );
}

/** 
 * Check for all required attributes in input data
 */
void PatchGraphExtract::checkAttributes()
{
	debugOutput << "Checking for required attributes.\n"; debugOutput.flush();
	for( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		char buffer[10];
		string name = std::string( "root/data/frame" ) + itoa( frameIndex, buffer, 10 );
		checkAttribute( scene, name, "position", 3, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( scene, name, "color", 3, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( scene, name, "tensor", 6, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( scene, name, "intAtt", 1, VAD::DATA_FORMAT_INT32 );
		checkAttribute( scene, name, "surfaceness", 1, VAD::DATA_FORMAT_FLOAT32 );
	}	
}

/** 
 * Align dataset to [-1;1]
 */
void PatchGraphExtract::align()
{
	debugOutput << "Aligning dataset to [-1;1].\n"; debugOutput.flush();
	// Determine bounding box
	BoundingBox3f bbox;
	bool firstPoint = true;
	for( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		UnstructuredInCorePointCloud* pc = getDataFrame( frameIndex );
		AAT POSITION = pc->getPointSet()->getAAT( "position" );
		for ( unsigned int pointIndex = 0; pointIndex < pc->getNumPoints(); pointIndex++ ){
			Vector3f pos = pc->getPointSet()->get3f( pointIndex, POSITION );
			if ( firstPoint ){ bbox = BoundingBox3f( pos ); firstPoint = false; }
			bbox.addPoint( pos );
		}
	}
	float32 scale = 2.0f / bbox.getMaxSideLength();
	Vector3f t = bbox.getCenter();

	// Transform points
	for( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		UnstructuredInCorePointCloud* pc = getDataFrame( frameIndex );
		AAT POSITION = pc->getPointSet()->getAAT( "position" );
		for ( unsigned int pointIndex = 0; pointIndex < pc->getNumPoints(); pointIndex++ ){
			Vector3f pos = pc->getPointSet()->get3f( pointIndex, POSITION );
			Vector3f newPos = ( pos - t ) * scale;
			pc->getPointSet()->set3f( pointIndex, POSITION, newPos );
		}
	}
}

/**
* Compute a topology graph
*/
void PatchGraphExtract::computeKNNTopologyGraph( InCorePCTopologyGraph* tg, PointSet *ps, const card32 startIndex, const card32 endIndex, AttachedIndexedOctree* octree )
{
	AAT POSITION = ps->getAAT( "position" );
	PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );
	card32 nIndex;
	bool found;
	for( card32 pointIndex = startIndex; pointIndex <= endIndex; pointIndex++) {
		Vector3f pos = ps->get3f( pointIndex, POSITION );
		knn.setTargetPoint( pos );
		// Self
		knn.getNextNearestPointIndex( nIndex, found );

		// Using nearest neighbors
		for ( unsigned int n = 0; n < 8; n++ ){
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = ps->get3f( nIndex, POSITION );
			tg->addEdge( pointIndex, nIndex );
		}

		/*
		// Using subsampling
		
		knn.getNextNearestPointIndex( nIndex, found );
		Vector3f nPos = ps->get3f( nIndex, POSITION );
		float32 d = norm( pos - nPos ) ;
		vector<card32> candidates; candidates.clear();
		while ( d < context->getSamplingSpacing() * context->getExtractKNNFactor() ){
			if ( ( pointIndex != nIndex ) && ( nIndex >= startIndex ) && ( nIndex <= endIndex ) ) {
				candidates.push_back( nIndex );
			}
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = ps->get3f( nIndex, POSITION );
			d = norm( pos - nPos );
		}
		for ( unsigned int n = 0; n < 8; n++ ){
			if ( candidates.size() < 1 ) break;
			card32 index = rand()%card32( candidates.size() );
			tg->addEdge( pointIndex, candidates[index] );
			candidates.erase( candidates.begin() + index );
		}
		*/
	}
}


/**
* Apply tensor voting in order to find salient surface regions
*/
void PatchGraphExtract::applyTensorVoting( const card32 frameIndex, bool verbose )
{
	UnstructuredInCorePointCloud* pc = getDataFrame( frameIndex );
	if ( pc == NULL ) return;
	float32 start = clock();
	if ( verbose ) { debugOutput << "  Computing tensor voting ... "; debugOutput.flush(); }
	tv->tensorVote( true, pc, context->getTVSigma() );
	if ( verbose ) { debugOutput << "time required: " << float32( ( clock() - start ) / CLK_TCK ) << ".\n"; debugOutput.flush(); }
}

/** 
* Extract patch graphs from a point cloud
*/
void PatchGraphExtract::extractPatchGraphs( const card32 frameIndex, bool verbose )
{
	float32 start = clock();

	UnstructuredInCorePointCloud* pc = getDataFrame( frameIndex );
	if ( pc == NULL ){
		return;
	}
	PointSet* ps = pc->getPointSet();
	if ( ps == NULL ){
		debugOutput << "extractPatchGraphs: invalid point set.\n"; debugOutput.flush();
		return;
	}
	if ( context->getPGContainer() == NULL ){
		debugOutput << "extractPatchGraphs: invalid container.\n"; debugOutput.flush();
		return;
	}
	if ( !ps->providesAttribute( "intAtt" ) ||
		!ps->providesAttribute( "tensor" ) ||
		!ps->providesAttribute( "color" ) ||
		!ps->providesAttribute( "position" ) ){
		debugOutput << "extractPatchGraphs: attribute missing; intAtt, tensor, color, position required.\n"; debugOutput.flush();
		return;
	}

	if ( verbose ) { debugOutput << "  Creating patch graph.\n"; debugOutput.flush(); }
	// INT_ATT meaning
	// -1: not surface
	// 0 surface
	// 1 assigned to patch graph
	AAT COLOR = ps->getAAT( "color" );
	AAT SURFACENESS = ps->getAAT( "surfaceness" );
	AAT INT_ATT = ps->getAAT( "intAtt" );
	AAT POSITION = ps->getAAT( "position" );
	for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex ++ ){
		// Extract tensor
		float32 surfaceness = ps->get1f( pointIndex, SURFACENESS );
		StaticVector<int32,1> v; v[0] = ( surfaceness > context->getTVSurfacenessThreshold () )? 0 : -1;
		ps->set1i( pointIndex, INT_ATT, v );
	}

	if ( verbose ) { debugOutput << "  Detecting components.\n"; debugOutput.flush(); }
	// Start region growing in order to find surface regions in the patches
	FifoStack<card32> stack;
	AttachedIndexedOctree *octree = PointSetKNNQuery::createOctree( ps );
	PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, ps->getAAT("position") );
	card32 nIndex;
	bool found;
	vector< vector<card32> > components; components.clear();
	for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
		if ( ps->get1i( pointIndex, INT_ATT ) == 0 ){
			// Grow new region
			vector<card32> componentIndices; componentIndices.clear();
			stack.clear();
			stack.push( pointIndex );
			componentIndices.push_back( pointIndex );
			StaticVector<int32,1> v; v[0] = 1;
			ps->set1i( pointIndex, INT_ATT, v );
			while( !stack.empty() ){
				Vector3f seedPoint = ps->get3f( stack.pop(), POSITION );
				knn.setTargetPoint( seedPoint );

				// Look in local feature size
				float32 dist = 0;
				float32 discardDist = min( context->getSamplingSpacing(), context->getTVSigma() * context->getPGRegionGrowScale() * 0.5 );
				while( dist < context->getTVSigma() * context->getPGRegionGrowScale() ){
					knn.getNextNearestPointIndex( nIndex, found );
					Vector3f nPos = ps->get3f( nIndex, POSITION );
					dist = norm( seedPoint - nPos );
					if ( ps->get1i( nIndex, INT_ATT ) == 0 ){
						StaticVector<int32,1> v; v[0] = 1;
						ps->set1i( nIndex, INT_ATT, v );
						if ( dist >= discardDist ){
							componentIndices.push_back( nIndex );
							stack.push( nIndex );
						}
					}
				}
			}
			// Save component indices
			if ( componentIndices.size() > context->getMinPatchGraphSize() ){
				components.push_back( componentIndices );
			}
		}
	}

	// Create patches point set
	size_t numPatches = 0;
	for ( unsigned int n = 0; n <components.size(); n++ ){
		numPatches += components[n].size();
	}
	if ( verbose ) { debugOutput << "  #points: " << numPatches << ".\n"; debugOutput.flush(); }
	PointSet* tsPS = new PointSet();
	VertexDescriptor vdPatches;
	vdPatches.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vdPatches.pushAttrib( mVAD( "gradient", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( mVAD( "hfieldCoeffs", 6, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( mVAD( "tangent_u", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdPatches.pushAttrib( ROTATION_MATRIX_PATCH_VAD );
	tsPS->clearAndSetup( 1, numPatches, &vdPatches );
	AAT PG_POSITION = tsPS->getAAT( "position" );
	AAT PG_COLOR = tsPS->getAAT( "color" );
	AAT PG_FLAGS = tsPS->getAAT( "flags" );
	// Create topology graph
	InCorePCTopologyGraph* tg = new InCorePCTopologyGraph();
	VertexDescriptor vdTG;
	vdTG.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
	vdTG.pushAttrib( VAD( "color", 4, VAD::DATA_FORMAT_FLOAT32 ) );
	tg->clearAndSetup( &vdTG );
	card32 pIndex = 0;
	for ( unsigned int n = 0; n < components.size(); n++ ){
		Vector3f color = makeVector3f( rand()%100/100.0f, rand()%100/100.0f, rand()%100/100.0f );
		for ( unsigned int compPointIndex = 0; compPointIndex < components[n].size(); compPointIndex++ ){
			Vector3f pos = ps->get3f( components[n][compPointIndex], POSITION );
			//Vector3f color = ps->get3f( components[n][compPointIndex], COLOR );
			tsPS->set3f( pIndex, PG_POSITION, pos );
			tsPS->set3f( pIndex, PG_COLOR, color );
			tsPS->set<int>( pIndex, PG_FLAGS, FLAG_PATCH_INITIALIZED);
			pIndex++;
		}
	}
	// Compute topology for components
	AttachedIndexedOctree *tsPSOctree = PointSetKNNQuery::createOctree( tsPS );
	size_t componentOffset = 0;
	for ( unsigned int n = 0; n < components.size(); n++ ){
		if ( verbose ) { debugOutput << "  Creating topology graph (" << (n+1) << "/" <<  (mpcard) components.size() << ")\n"; debugOutput.flush(); }
		computeKNNTopologyGraph( tg, tsPS, (card32)componentOffset, (card32)(componentOffset + components[n].size()-1), tsPSOctree );
		componentOffset += components[n].size();
	}
	delete tsPSOctree;
	// Create patch graph
	PatchGraph* patchGraph = new PatchGraph();
	TrajectorySet* ts = new TrajectorySet();
	PGDataPointIndices* dataPointIndices = new PGDataPointIndices();
	dataPointIndices->setup( 1, tsPS->getNumEntries() );
	ts->clearAndSetup( tsPS, dataPointIndices );
	patchGraph->clearAndSetup( ts, tg, false );
	patchGraph->setStartFrameIndex( frameIndex );

	// Re-sample
	float32 s = clock();
	if ( verbose ) { debugOutput << "  Resampling patch graph ... "; debugOutput.flush(); }
	resampler->resample( patchGraph );
	if ( verbose ) { debugOutput << "time required: " << float32( ( clock() - s ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush(); }

	/*
	patchGraph->createPatchAttribute(ROTATION_MATRIX_PATCH_VAD);
	//Re-sample the geometry
	PatchGraph *newPG = altResampler->fullResampling_Modified(patchGraph, 2*context->getSamplingSpacing(), 3*context->getSamplingSpacing(), 8, 2);
	patchGraph->assign(newPG);
	delete newPG;
	*/	

	s = clock();
	if ( verbose ) { debugOutput << "  Assigning data points ... "; debugOutput.flush(); }
	DataPointAssignment::assignDataPoints( patchGraph, context->getPGContainer()->getAnimSeq(), context->getSamplingSpacing() * context->getDataPointsInfluence() );
	if ( verbose ) { debugOutput << "time required: " << float32( ( clock() - s ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush(); }

	/*
	// Optimize
	s = clock();
	if ( verbose ) { debugOutput << "  Optimizing created patch graph ... "; debugOutput.flush(); }
	optimize( patchGraph, context->getPGContainer()->getAnimSeq() );
	if ( verbose ) { debugOutput << "time required: " << float32( ( clock() - s ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush(); }
	*/

	// Computing bounding box
	patchGraph->computeBBox();

	// Remove isolated points
	detectAndRemoveIsolatedVertices( patchGraph );

	// Save patch graph
	context->getPGContainer()->addPatchGraph( patchGraph );

	if ( verbose ) { debugOutput << "  Created patch graph with " << patchGraph->getTrajectorySet()->getNumPatches() << " patches.\n"; debugOutput.flush(); }

	// Free allocated mem
	delete octree;

	if ( verbose ) { debugOutput << " Time required: " << float32( ( clock() - start ) / CLK_TCK ) << ".\n"; debugOutput.flush(); }
}

/** 
 * Apply tensor voting to current frame
 */
void PatchGraphExtract::tensorVoteFrame()
{
	context->getPGContainer()->clear();
	setAnimationSequence();
	int32 frameIndex = getCurrentFrameIndex();
	applyTensorVoting( frameIndex, context->getVerbose() );
}

/**
 * Extract single patch graph from point cloud
 */
void PatchGraphExtract::extractPatchGraph()
{
	context->getPGContainer()->clear();
	setAnimationSequence();
	int32 frameIndex = getCurrentFrameIndex();
	extractPatchGraphs( frameIndex,context->getVerbose() );
}

/** 
* Extract all patch graphs from point cloud
*/
void PatchGraphExtract::tensorVoteAllFrames()
{
	progressWindow->pushStep( true, "Tensor voting" );
	progressWindow->progress( 0 );	
	context->getPGContainer()->clear();
	setAnimationSequence();
	float32 start = clock();
	for ( int32 frameIndex = 0; frameIndex < int32( getNumFrames() ); frameIndex++ ) {
		PointCloud *pc = getPointCloud(scene, getDataFrameName(frameIndex));
		if (pc == NULL) {
			warning("tensorVoteAllFrames: data frame not found.");
			continue;
		}
		checkAttribute(pc, "intAtt", 1, VAD::DATA_FORMAT_INT32);
		checkAttribute(pc, "tensor", 6, VAD::DATA_FORMAT_FLOAT32);
		checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);

		debugOutput << "Tensor voting in frame " << card32( frameIndex + 1 ) << "/" << card32( getNumFrames() ) << ".\n"; debugOutput.flush();
		applyTensorVoting( frameIndex, context->getVerbose() );
		progressWindow->progress( card32( float32( frameIndex + 1 ) * 100.0f / float32( getNumFrames() ) ) );
	}
	progressWindow->popSteps( );
	debugOutput << "Time required: " << float32( ( clock() - start ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush();
}

/** 
 * Extract all patch graphs from point cloud
 */
void PatchGraphExtract::extractPatchGraphAllFrames()
{
	progressWindow->pushStep( true, "Extract patch graphs" );
	progressWindow->progress( 0 );	
	context->getPGContainer()->clear();
	setAnimationSequence();
	float32 start = clock();
	for ( int32 frameIndex = 0; frameIndex < int32( getNumFrames() ); frameIndex++ ) {
		PointCloud *pc = getPointCloud(scene, getDataFrameName(frameIndex));
		if (pc == NULL) {
			warning("extractPatchGraphAllFrames: data frame not found.");
			continue;
		}
		checkAttribute(pc, "intAtt", 1, VAD::DATA_FORMAT_INT32);
		checkAttribute(pc, "tensor", 6, VAD::DATA_FORMAT_FLOAT32);
		checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);

		debugOutput << "Extracting patch graphs in frame " << card32( frameIndex + 1 ) << "/" << card32( getNumFrames() ) << ".\n"; debugOutput.flush();
		extractPatchGraphs( frameIndex, context->getVerbose() );
		progressWindow->progress( card32( float32( frameIndex + 1 ) * 100.0f / float32( getNumFrames() ) ) );
	}
	progressWindow->popSteps( );
	debugOutput << "Time required: " << float32( ( clock() - start ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush();
}

/** 
* Get number of frames
*/
card32 PatchGraphExtract::getNumFrames() const
{
	SGSimpleAnimationNode* animDataNode = getDataAnimationNode();
	if ( animDataNode == NULL ){
		return 0;
	}
	return animDataNode->getNumChildNodes( scene->getRootState() );
}

/**
* Get the data animation node
*/
SGSimpleAnimationNode* PatchGraphExtract::getDataAnimationNode() const
{
	// Find data node
	StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
	SGSimpleAnimationNode* animDataNode = NULL;
	it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
	while ( !it->atEnd() ){
		if ( it->getNameWithPath() == "root/data" ){
			animDataNode = dynamic_cast<SGSimpleAnimationNode*>( it->getNode() );
			if ( animDataNode != NULL ){
				break;            
			}
		}
		it->next();
	}
	delete it;
	return animDataNode;
}

/**
* Return index of current frame
*/
int32 PatchGraphExtract::getCurrentFrameIndex() const
{
	return scene->getRootState()->dynamicState->time;
}

/**
* Get data point cloud for frame index
*/
UnstructuredInCorePointCloud* PatchGraphExtract::getDataFrame( card32 index )
{
	string frameName = getDataFrameName( index );
	PointCloud *pc = getPointCloud( scene, frameName );
	return dynamic_cast<UnstructuredInCorePointCloud*>( pc ); 
}

/// Set the animation sequence
void PatchGraphExtract::setAnimationSequence()
{
	context->getPGContainer()->getAnimSeq()->clear();
	for ( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		PointSet* ps = getDataFrame( frameIndex )->getPointSet();
		if ( ps == NULL ){
			debugOutput << "setAnimationSequence: invalid data.\n"; debugOutput.flush();
		}
		context->getPGContainer()->getAnimSeq()->addFrame( ps );
	}
}

/** 
 * Detect and remove isolated vertices
 */
void PatchGraphExtract::detectAndRemoveIsolatedVertices( PatchGraph* pg )
{
	if ( pg == NULL ){
		debugOutput << "detectAndRemoveIsolatedVertices: invalid patch graph.\n"; debugOutput.flush();
		return;
	}
	if ( pg->getTrajectorySet()->getNumFrames() != 1 ){
		debugOutput << "detectAndRemoveIsolatedVertices: only works on single frame patch graphs.\n"; debugOutput.flush();
	}
	PointSet* traj = pg->getTrajectorySet()->getTrajectories();
	InCorePCTopologyGraph* tg = pg->getTopology();
	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	card32 numIsolatedPatches = 0;
	vector<int32> map; map.resize( numPatches );

	// Detect isolated patches
	for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
		if ( tg->getNumAdjacentVertices( pointIndex ) < 1 ){
			numIsolatedPatches++;
			map[pointIndex] = -1;
		} else {
			map[pointIndex] = pointIndex - numIsolatedPatches;
		}
	}
	
	if ( numIsolatedPatches > 0 ){
		debugOutput << "[Should not occur any more -> Tell Philipp] Removing " << numIsolatedPatches << " isolated patch point(s).\n"; debugOutput.flush();
		// Create new trajectory set
		PointSet* newTraj = new PointSet();
		card32 newNumPatches = numPatches - numIsolatedPatches;
		newTraj->clearAndSetup( 1, newNumPatches, traj->getDescr() );
		for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
			int32 newPatchIndex = map[pointIndex];
			if ( newPatchIndex >= 0 ){
				newTraj->setVertex( newPatchIndex, traj->getDataPointer( pointIndex ) );
			}
		}
		// Create new topology graph
		InCorePCTopologyGraph* newTG = new InCorePCTopologyGraph();
		newTG->clearAndSetup( tg->getEdges()->getDescr() );
		for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
			for ( unsigned int n = 0; n < tg->getNumAdjacentVertices( pointIndex ); n++ ){
				card32 nIndex = tg->getAdjacentVertex( pointIndex, n );
				if ( ( map[pointIndex] >= 0 ) &&( map[nIndex] >= 0 ) ){
					newTG->addEdge( map[pointIndex], map[nIndex] );
				}
			}
		}
		// Create new data point indices
		PGDataPointIndices* newDP = new PGDataPointIndices();
		newDP->setup( 1, newTraj->getNumEntries() );
		// Reset patch graph
		pg->setTopology( newTG );
		pg->getTrajectorySet()->clearAndSetup( newTraj, newDP );
		// Assign data points
		DataPointAssignment::assignDataPoints( pg, context->getPGContainer()->getAnimSeq(), context->getSamplingSpacing() * context->getDataPointsInfluence() );
	} else {
		//debugOutput << "No isolated patch points detected.\n"; debugOutput.flush();
	}
}


/** 
 * Remove outliers
 */
void PatchGraphExtract::removeOutliers()
{
	debugOutput << "Not implemented yet!\n"; debugOutput.flush();
	return;
	for( unsigned int frameIndex = 0; frameIndex < getNumFrames(); frameIndex++ ){
		char buffer[10];
		string name = std::string( "root/data/frame" ) + itoa( frameIndex, buffer, 10 );
		// Search for point cloud
		StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
		it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
		while ( !it->atEnd() ){
			if ( it->getNameWithPath() == name ){
				SGObjectNode* sgo = dynamic_cast<SGObjectNode*>( it->getNode() );
				UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( sgo->getSceneObject() );
				PointSet* ps = upc->getPointSet();
				PointSet* ps2 = (PointSet*)(ps->copy());
			}
			it->next();
		}
		delete it;
	}
}