//---------------------------------------------------------------------------
#ifndef PatchGraphExtractH
#define PatchGraphExtractH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PatchGraphContainer.h"
#include "TensorVoting.h"
#include "UniformPointCloudResampling.h"
#include "InCorePCTopologyGraph.h"
#include "PatchGraphResample.h"
#include "AnimationBSRContext.h"
#include "SGSimpleAnimationNode.h"
#include "PatchGraphBSROpt.h"
#include "UnifyNormals.h"
#include "PatchGraphComponentsHandler.h"
#include "PatchGraphTopologicalResampling.h"
//---------------------------------------------------------------------------

namespace geoxl {

/// Extract initial patch graphs from point clouds.
class ANIMATIONRECONSTRUCTION_API PatchGraphExtract : public Persistent {
	GEOXL_CLASS( PatchGraphExtract )

	private:
		/// Tensor voting
		TensorVoting* tv;
		/// Resampler
		PatchGraphResample* resampler;
		/// Context
		AnimationBSRContext* context;
		/// Scene object
		Scene* scene;

		/*** HELPERS ***/
		/// Compute a topology graph; use spatial nearest neighbors on point positions
		void computeKNNTopologyGraph( InCorePCTopologyGraph* tg, PointSet *ps, const card32 startIndex, const card32 endIndex, AttachedIndexedOctree* octree );
		/// Apply tensor voting in order to find salient surface regions
		void applyTensorVoting( const card32 frameIndex, bool verbose );
		/// Extract all patch graphs for given frame
		void extractPatchGraphs( const card32 frameIndex, bool verbose );

		/*** INPUT DATA ACCESS ***/
		/// Get number of frames
		card32 getNumFrames() const;
		/// Get the data animation node
		SGSimpleAnimationNode* getDataAnimationNode() const;
		/// Return index of current frame
		int32 getCurrentFrameIndex() const;
		/// Get data point cloud for frame index
		UnstructuredInCorePointCloud* getDataFrame(card32 index );
		/// Get name of data frame point cloud 
		string getDataFrameName( card32 index ){ char buffer[10]; return std::string( "root/data/frame" ) + itoa( index, buffer, 10 ); }
		/// Set the animation sequence
		void setAnimationSequence();


		/*
		/// Alternative resampler
		PatchGraphTopologicalResampling* altResampler;
		*/
	
	public:
		/*** CLASS MANAGEMENT ***/
		/// Constructor
		PatchGraphExtract();   
		/// Destructor
		~PatchGraphExtract();

		/// Check for all required attributes in input data
		void checkAttributes();
		/// Align dataset to [-1;1]
		void align();
		/// Remove outliers
		void removeOutliers();
		/// Apply tensor voting to current frame
		void tensorVoteFrame();
		/// Apply tensor voting to current frame
		void tensorVoteAllFrames();
		/// Extract single patch graph from point cloud
		void extractPatchGraph();
		/// Extract all patch graphs from point cloud
		void extractPatchGraphAllFrames();
		/// Detect and remove isolated vertices
		void detectAndRemoveIsolatedVertices( PatchGraph* pg );

		/*** MEMBER ACCESS ***/
		inline TensorVoting* getTensorVoting() const { return tv; }
		inline void setContext( AnimationBSRContext* context ){ this->context = context; resampler->setContext( context ); }
		inline void setScene( Scene* scene ){ this->scene = scene; }

		/*** PCInteractable ***/
		/// Object assignment
		void assign(const Object* obj, CopyContext *context = NULL);

};

}

#endif