//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphMerge.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "PatchGraphContainer.h"
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "PointSetKNNQuery.h"
#include "PGDataPointIndices.h"
#include "BSROptContext.h"
#include "TGNeighborhoodKNNSearch.h"
#include "PatchGraphTopologicalResampling.h"
#include "DataPointAssignment.h"
#include "PatchTypes.h"
#include "Correspondence.h"
#include "FeatureCorrespondences.h"
//---------------------------------------------------------------------------
#include <time.h>
#include <stdio.h>
#include <dos.h>
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphMerge, 10 )
{
	BEGIN_CLASS_INIT( PatchGraphMerge )	
	ADD_OBJECT_PROP( altResampler, 7, PatchGraphTopologicalResampling::getClass(), true )
	ADD_VAR_ARRAY_PROP( alignMergeExtrapolate, 2, 
		new ObjectClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		AbstractMergeExtrapolator::getClass(), true, NULL )
	)
}

/**
 * Constructor
 */
PatchGraphMerge::PatchGraphMerge() 
{
	altResampler = new PatchGraphTopologicalResampling();
	context = NULL;
}

/**
 * Destructor
 */
PatchGraphMerge::~PatchGraphMerge() 
{ 
	delete altResampler;
	for (unsigned i=0; i<alignMergeExtrapolate.size(); i++) {
		delete alignMergeExtrapolate[i];
	}
}

/**
 * Member access: set context
 */
void PatchGraphMerge::setContext( AnimationBSRContext* context )
{ 
	this->context = context; 
}

/** 
 * Object assignment
 */
void PatchGraphMerge::assign( const Object* obj, CopyContext *context ) 
{
	copyObjectProperties( obj, this );
}

void PatchGraphMerge::addAlignMergeExtrapolateStep(AbstractMergeExtrapolator *step) {
	alignMergeExtrapolate.push_back(step);
}

void PatchGraphMerge::clearAMESteps() {
	for (unsigned i=0; i<alignMergeExtrapolate.size(); i++) {
		delete alignMergeExtrapolate[i];
	}
	alignMergeExtrapolate.clear();
}

/// helper function to extract feature correspondences
static void buildFeatureCorrespondences(
	map<card32, Vector3f> &featureCorrespondencesLeft,
	map<card32, Vector3f> &featureCorrespondencesRight,
	AnimationBSRContext* context, 
	const int32 sourceFrame, 
	PatchGraph *left,
	PatchGraph *right)
{
	PatchGraphContainer *pgc = context->getPGContainer();
	FeatureCorrespondences *corr = pgc->getFeatureCorrespondences();
	
	if (!corr) return;
	if (corr->getNumFrames() <= 0) return;

	TrajectorySet *leftTS = left->getTrajectorySet();
	TrajectorySet *rightTS = right->getTrajectorySet();
	leftTS->invalidateAllAttachments();
	rightTS->invalidateAllAttachments();

	AttachedIndexedOctree *oLeft = leftTS->fetchOctree(leftTS->getNumFrames()-1);		
	AttachedIndexedOctree *oRight = rightTS->fetchOctree(0);		
	
	PointSet *leftPSCopy = leftTS->copySingleFrame(leftTS->getNumFrames()-1);
	PointSet *rightPSCopy = rightTS->copySingleFrame(0);

	PointSetKNNQuery knnLeft(leftPSCopy, oLeft, NULL_VECTOR3F, leftPSCopy->getAAT("position"));
	PointSetKNNQuery knnRight(rightPSCopy, oRight, NULL_VECTOR3F, rightPSCopy->getAAT("position"));
	
	const card32 numCorr = corr->getNumCorrespondenciesInFrame(sourceFrame);
	for (card32 i=0; i<numCorr; i++) {
		if (corr->getCorrespondence(sourceFrame, i)->getFlags() == Correspondence::CF_Initial) continue;
		Vector3f leftPos = corr->getCorrespondence(sourceFrame, i)->getPosition1();
		Vector3f rightPos = corr->getCorrespondence(sourceFrame, i)->getPosition2();
		knnLeft.setTargetPoint(leftPos);
		knnRight.setTargetPoint(rightPos);
		
		card32 leftIndex;
		card32 rightIndex;
		
		bool found;
		knnLeft.getNextNearestPointIndex(leftIndex, found);
		if (!found) {
			warning("buildFeatureCorrespondences - something's wrong");
			continue;
		}
		knnRight.getNextNearestPointIndex(rightIndex, found);
		if (!found) {
			warning("buildFeatureCorrespondences - something's wrong");
			continue;
		}
		
		pair<card32, Vector3f> vl(leftIndex, rightPos);
		featureCorrespondencesLeft.insert(vl);
		pair<card32, Vector3f> vr(rightIndex, leftPos);
		featureCorrespondencesRight.insert(vr);
	}
	
	delete leftPSCopy;
	delete rightPSCopy;
}

bool PatchGraphMerge::merge_optimized2(
									  AnimationBSRContext* context, 
									  FeatureRegistration *featureMatcher,
									  const int32 pgIndex1, 
									  const int32 pgIndex2, 
									  DebugRenderer *debugRenderer, 
									  bool verbose )
{
	float32 mergeStartTime = clock();
	// Test for valid index
	PatchGraph* pg1 = context->getPGContainer()->getPatchGraph( pgIndex1 );
	PointCloudAnimationSequence* animSeq = context->getPGContainer()->getAnimSeq();
	PatchGraph* pg2 = context->getPGContainer()->getPatchGraph( pgIndex2 );
	if ( ( pg1 == NULL ) || ( pg2 == NULL ) )
	{
		debugOutput << "merge: invalid indices!\n"; debugOutput.flush();
		return false;
	}
	debugOutput << " Merging patches "
		<< pg1->getStartFrameIndex() << ".." << ( pg1->getStartFrameIndex() + pg1->getTrajectorySet()->getNumFrames() - 1 )
		<< " with " << pg2->getStartFrameIndex() << ".." << ( pg2->getStartFrameIndex() + pg2->getTrajectorySet()->getNumFrames() - 1 ) << ".\n"; 
	debugOutput.flush();

	pg1->createPatchAttribute(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
	pg1->createPatchAttribute(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
	pg1->createPatchAttribute(mVAD("varIndex_pos", 1, VAD::DATA_FORMAT_INT32));
	pg1->createPatchAttribute(mVAD("varIndex_nor", 1, VAD::DATA_FORMAT_INT32));

	pg2->createPatchAttribute(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
	pg2->createPatchAttribute(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
	pg2->createPatchAttribute(mVAD("varIndex_nor", 1, VAD::DATA_FORMAT_INT32));
	pg2->createPatchAttribute(mVAD("varIndex_pos", 1, VAD::DATA_FORMAT_INT32));
	
	// Test for valid patch graphs
	if ( ( pg1 == NULL ) || ( pg2 == NULL ) ){
		debugOutput << "merge: invalid patch graphs.\n";
		debugOutput.flush();
		return false;
	}

	// Test for patch graph adjacency
	if ( pg1->getTrajectorySet()->getNumFrames() + pg1->getStartFrameIndex() != pg2->getStartFrameIndex() ){
		debugOutput << "merge: can only merge adjacent patch graphs.\n"; 
		debugOutput.flush();
		return false;
	}

	card32 frameOffset = pg1->getTrajectorySet()->getNumFrames();
	int32 patchOffset = pg1->getTrajectorySet()->getNumPatches();
	card32 numFrames = pg1->getTrajectorySet()->getNumFrames() + pg2->getTrajectorySet()->getNumFrames();
	card32 numPatches = pg1->getTrajectorySet()->getNumPatches() + pg2->getTrajectorySet()->getNumPatches();
	card32 localFrameLeft = pg1->getTrajectorySet()->getNumFrames() - 1;
	card32 localFrameRight = localFrameLeft + 1;

	// Create combined patch graph data structure
	TrajectorySet* ts = new TrajectorySet();
	PointSet* trajectories = new PointSet();
	if(pg1->getTrajectorySet()->getTrajectories()->getWidth() >  pg2->getTrajectorySet()->getTrajectories()->getWidth()){
		trajectories->clearAndSetup( numFrames, numPatches, pg1->getTrajectorySet()->getTrajectories()->getDescr() );
	} else {
		trajectories->clearAndSetup( numFrames, numPatches, pg2->getTrajectorySet()->getTrajectories()->getDescr() );
	}
	ts->setStartFrameIndex( pg1->getStartFrameIndex() );
	ts->clearAndSetup( trajectories );
	PointCloudAnimationSequence	*localDataPoints = new PointCloudAnimationSequence();
	InCorePCTopologyGraph* tg = new InCorePCTopologyGraph();
	PatchGraph* combinedPG = new PatchGraph();
	combinedPG->clearAndSetup( ts, tg, false );

	// Copy all information existent in pg1 and pg2 to combined data structure
	float32 start = clock();
	createCombinedTrajectorySet_Modified(context, ts, tg, pg1, pg2);

	/// Correspondences
	map<card32, Vector3f> featureCorrespondences12;
	map<card32, Vector3f> featureCorrespondences21;
	buildFeatureCorrespondences(
		featureCorrespondences12, 
		featureCorrespondences21, 
		context, 
		pg2->getStartFrameIndex()-1, 
		pg1,
		pg2
	);	

	/// Extrapolate information from pg1 into frames in pg2 and vice versa
	for (unsigned i=0; i<alignMergeExtrapolate.size(); i++) {
		alignMergeExtrapolate[i]->compute_Modified(context, featureMatcher, pg1, pg2, combinedPG, featureCorrespondences12, featureCorrespondences21, debugRenderer, verbose);
	}

	// Re-sample patch graph
	start = clock();
	PatchGraph *newPG = altResampler->fullResampling_Modified(combinedPG,
			2*context->getSamplingSpacing(), 
			context->maxEdgeStretchFactor, 
			8, 2);
	combinedPG->assign(newPG);
	delete newPG;

	/// Extrapolate information from pg1 into frames in pg2 and vice versa
	for (unsigned i=0; i<alignMergeExtrapolate.size(); i++) 
	{
		alignMergeExtrapolate[i]->globalAssembly(context, combinedPG, debugRenderer, verbose);
	}

	// Update bounding box
	combinedPG->computeBBox();

	/// Delete merged patch graphs
	context->getPGContainer()->deletePatchGraph( pgIndex1 );
	context->getPGContainer()->deletePatchGraph( ( pgIndex1 < pgIndex2 )? pgIndex2 - 1 : pgIndex2 );
	context->getPGContainer()->addPatchGraph( combinedPG );

	if ( context->getVerbose() ) { debugOutput << " Time required: " << float32( ( clock() - mergeStartTime ) / CLK_TCK ) << " seconds.\n"; debugOutput.flush(); }

	delete localDataPoints;
	return true;
}


void	PatchGraphMerge::createCombinedTrajectorySet_Modified(
	geoxl::AnimationBSRContext *context, 
	geoxl::TrajectorySet *ts, 
	geoxl::InCorePCTopologyGraph *tg,
	const geoxl::PatchGraph *pg1, 
	const geoxl::PatchGraph *pg2)
{
	card32 patchOffset	= pg1->getTrajectorySet()->getNumPatches();
	card32 frameOffset	= pg1->getTrajectorySet()->getNumFrames();
	int32	frameLeft	= pg1->getTrajectorySet()->getNumFrames() - 1;
	int32	frameRight	= frameLeft + 1;
	AAT		POSITION	= ts->getTrajectories()->getAAT("position");
	PointSet *psSource	= NULL;

	/// We need the folloiwng channel for iterative extrapolation
	ts->createPatchAttribute(mVAD("topology_depth", 1,	VAD::DATA_FORMAT_INT32));
	AAT			TOP_DEPTH			= ts->getTrajectories()->getAAT("topology_depth");

	// Frame 0 .. frameLeft: 
	//  - patch 0 .. offset-1: copy positions, data point assignments
	//  - patch offset .. ts->getNumPatches: reset
	for(int32 frameIndex = 0; frameIndex <= frameLeft; frameIndex++)
	{
		psSource = pg1->getTrajectorySet()->copySingleFrame(frameIndex);
		for(unsigned int patchIndex = 0; patchIndex < patchOffset; patchIndex++)
		{
			/// // Position
			ts->getTrajectories()->setVertex(frameIndex, patchIndex, psSource->getDataPointer(0, patchIndex));
			/// the geometries of these points are fixed
			ts->getTrajectories()->set<int32>(frameIndex, patchIndex, TOP_DEPTH, 0);
			/// Data point assignment
			vector<card32> indices = pg1->getTrajectorySet()->getDataPoints()->getIndices(frameIndex, patchIndex);
			ts->getDataPoints()->setIndices(frameIndex, patchIndex, indices);
		}
		delete psSource; psSource = NULL;
		for (unsigned int patchIndex = patchOffset; patchIndex < ts->getNumPatches(); patchIndex++)
		{
			// Position
			ts->getTrajectories()->set3f(frameIndex, patchIndex, POSITION, makeVector3f( 0,0,0 ));
			/// the geometries of these points are unknown
			ts->getTrajectories()->set<int32>(frameIndex, patchIndex, TOP_DEPTH, -1);
			// Data point assignment
			vector<card32> indices; indices.clear();
			ts->getDataPoints()->setIndices(frameIndex, patchIndex, indices);
		}
		///Create DataPoints.

	}

	// Frame frameRight .. ts->getNumFrames()-1: 
	// - patch 0: .. offset-1: reset
	// - patch offset .. ts->getNumPatches): copy positions, data point assignments
	for(int32 frameIndex = frameRight; frameIndex < int32( ts->getNumFrames()); frameIndex++)
	{
		for(unsigned int patchIndex = 0; patchIndex < patchOffset; patchIndex++)
		{
			// Position
			ts->getTrajectories()->set3f( frameIndex, patchIndex, POSITION, makeVector3f(0,0,0 ));
			/// the geometries of these points are unknown
			ts->getTrajectories()->set<int32>(frameIndex, patchIndex, TOP_DEPTH, -1);
			// Data point assignment
			vector<card32> indices; indices.clear();
			ts->getDataPoints()->setIndices( frameIndex, patchIndex, indices );
		}
		psSource = pg2->getTrajectorySet()->copySingleFrame( frameIndex - frameOffset );
		for ( unsigned int patchIndex = patchOffset; patchIndex < ts->getNumPatches(); patchIndex++ )
		{
			// // Position
			// const Vector3f pos = psSource->get3f( 0, patchIndex - patchOffset, POSITION );
			// ts->getTrajectories()->set3f( frameIndex, patchIndex, POSITION, pos );
			ts->getTrajectories()->setVertex( frameIndex, patchIndex, psSource->getDataPointer( 0, patchIndex - patchOffset) );
			/// the geometries of these points are unknown
			ts->getTrajectories()->set<int32>(frameIndex, patchIndex, TOP_DEPTH, 0);

			// Data point assignment
			vector<card32> indices = pg2->getTrajectorySet()->getDataPoints()->getIndices( frameIndex - frameOffset, patchIndex - patchOffset );
			ts->getDataPoints()->setIndices( frameIndex, patchIndex, indices );
		}
		delete psSource; psSource = NULL;

		///Create DataPoints

	}

	// Copy known topology from pg1 and pg2
	tg->clearAndSetup( pg1->getTopology()->getEdges()->getDescr() );
	// Copy topology from pg1
	for (unsigned int patchIndex = 0; 
		patchIndex < pg1->getTrajectorySet()->getNumPatches(); 
		patchIndex++ )
	{
		for (unsigned int vertexIndex = 0; 
			vertexIndex < pg1->getTopology()->getNumAdjacentVertices( patchIndex ); 
			vertexIndex++ )
		{
			tg->addEdge(patchIndex, pg1->getTopology()->getAdjacentVertex( patchIndex, vertexIndex ) );
		}
	}
	// Copy topology from pg2
	for(unsigned int patchIndex = 0; 
		patchIndex < pg2->getTrajectorySet()->getNumPatches(); 
		patchIndex++ )
	{
		for (unsigned int vertexIndex = 0; 
			vertexIndex < pg2->getTopology()->getNumAdjacentVertices(patchIndex); 
			vertexIndex++ )
		{
			tg->addEdge(patchIndex + patchOffset, pg2->getTopology()->getAdjacentVertex( patchIndex, vertexIndex ) + patchOffset );
		}
	}
}
