//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphMergeCandidateSelect.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "ClassMethods.h"
#include "PatchGraph.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphMergeCandidateSelect, 4 )
{
	BEGIN_CLASS_INIT( PatchGraphMergeCandidateSelect )
	ADD_INT32_PROP( cand1, 0 )
	ADD_INT32_PROP( cand2, 0 )
	ADD_FLOAT32_PROP( omegaDistance, 3 )
	ADD_FLOAT32_PROP( omegaNumPatches, 3 )
	ADD_FLOAT32_PROP( omegaNumFrames, 3 )
	ADD_FLOAT32_PROP( omegaSize, 4 )
	ADD_NOARGS_METHOD( select )
}

/**
 * Constructor
 */
PatchGraphMergeCandidateSelect::PatchGraphMergeCandidateSelect() 
{
	cand1 = 0;
	cand2 = 2;
	omegaDistance = 0.5f;
	omegaNumFrames = 0.2f;
	omegaNumPatches = 0.5f;
	omegaSize = 1.0f;
}

/**
 * Destructor
 */
PatchGraphMergeCandidateSelect::~PatchGraphMergeCandidateSelect() 
{ 
}

/** 
 * Find next candidates
 */
void PatchGraphMergeCandidateSelect::select()
{
	if ( context == NULL ) return;
	if ( context->getMergeCandSelectionMode() == SELECTION_MODE_AUTOMATIC ){
		findNextMergeCandidates();
	} else if ( context->getMergeCandSelectionMode() == SELECTION_MODE_MANUAL ){
	} else {
		debugOutput << "Invalid selection mode.\n"; debugOutput.flush();
		cand1 = cand2 = -1;
	}
}

/** 
* Estimate the score of a merge operation
*/
float32 PatchGraphMergeCandidateSelect::estimateMergeScore( PatchGraph* pg1, PatchGraph* pg2, PatchGraphContainer* pgContainer )
{
	if ( ( pg1 == NULL ) || ( pg2 == NULL ) ){
		return -1;
	}

	/// Heuristic (1) - Check number of patches in pg1 with close neighbor (< epsilon) in pg2
	PointSet *psLeft = NULL, *psRight = NULL;
	AAT PS_LEFT_POSITION, PS_RIGHT_POSITION;
	// Choose TS with less patches as basis
	if ( pg1->getTrajectorySet()->getNumPatches() < pg2->getTrajectorySet()->getNumPatches() ){
		psLeft = pg1->getTrajectorySet()->copySingleFrame( pg1->getTrajectorySet()->getNumFrames() - 1 );
		PS_LEFT_POSITION = psLeft->getAAT( "position" );
		psRight = pg2->getTrajectorySet()->copySingleFrame( pg2->getTrajectorySet()->getNumFrames() - 1 );
		PS_RIGHT_POSITION = psRight->getAAT( "position" );
	} else {
		psLeft = pg2->getTrajectorySet()->copySingleFrame( pg2->getTrajectorySet()->getNumFrames() - 1 );
		PS_LEFT_POSITION = psLeft->getAAT( "position" );
		psRight = pg1->getTrajectorySet()->copySingleFrame( pg1->getTrajectorySet()->getNumFrames() - 1 );
		PS_RIGHT_POSITION = psRight->getAAT( "position" );
	}
	AttachedIndexedOctree *octree = PointSetKNNQuery::createOctree( psRight );
	PointSetKNNQuery knn( psRight, octree, NULL_VECTOR3F, psRight->getAAT("position") );
	card32 psRightIndex;
	bool found;
	card32 numClosePatches = 0;
	for ( unsigned int pgIndex = 0; pgIndex < psLeft->getNumEntries(); pgIndex++ ){
		Vector3f pos = psLeft->get3f( pgIndex, PS_LEFT_POSITION );
		knn.setTargetPoint( pos );
		knn.getNextNearestPointIndex( psRightIndex, found );
		if ( found ){	
			if ( norm( psRight->get3f( psRightIndex, PS_RIGHT_POSITION ) - pos ) < context->getSamplingSpacing() ){
				numClosePatches++;
			}
		}
	}
	delete octree;
	float32 distanceScore = omegaDistance * (float32)numClosePatches / float32( psLeft->getNumEntries() );

	/// Heuristic (2) - Higher score for patch graphs with similar number of patches
	float32 larger, smaller;
	if ( pg1->getTrajectorySet()->getNumPatches() > pg2->getTrajectorySet()->getNumPatches() ){
		larger = pg1->getTrajectorySet()->getNumPatches();
		smaller = pg2->getTrajectorySet()->getNumPatches();
	} else {
		larger = pg2->getTrajectorySet()->getNumPatches();
		smaller = pg1->getTrajectorySet()->getNumPatches();
	}
	float32 numPatchesScore = omegaNumPatches * ( 1.0f - float32( larger - smaller ) / float32( larger ) );

	/// Heuristic (3) - Higher score for patch graphs with similar number of frames
	if ( pg1->getTrajectorySet()->getNumFrames() > pg2->getTrajectorySet()->getNumFrames() ){
		larger = pg1->getTrajectorySet()->getNumFrames();
		smaller = pg2->getTrajectorySet()->getNumFrames();
	} else {
		larger = pg2->getTrajectorySet()->getNumFrames();
		smaller = pg1->getTrajectorySet()->getNumFrames();
	}
	float32 numFramesScore = omegaNumFrames * ( 1.0f - float32( larger - smaller ) / float32( larger ) );

	/// Heuristic (4) - favor patch graphs with few frames
	card32 numFrames = pg1->getTrajectorySet()->getNumFrames();
	numFrames = min( numFrames, pg2->getTrajectorySet()->getNumFrames() );
	// Determine maxNumFrames in container
	card32 maxNumFrames = 0;
	for ( unsigned int n = 0; n < pgContainer->getNumPatchGraphs(); n++ ){
		maxNumFrames = max( maxNumFrames, pgContainer->getPatchGraph( n )->getTrajectorySet()->getNumFrames() );		
	}
	float32 sizeScore = omegaSize * ( 1.0f - float32( numFrames ) / float32( maxNumFrames ) );

	// Combine scores
	return distanceScore + numPatchesScore + numFramesScore + sizeScore;
}

/**
* Find next best merge candidates
*/
void PatchGraphMergeCandidateSelect::findNextMergeCandidates()
{
	// Find candidates (look in involved frames only)
	float32 bestScore = -1;
	cand1 = cand2 = -1;
	for ( unsigned int pgIndex1 = 0; pgIndex1 < context->getPGContainer()->getNumPatchGraphs(); pgIndex1++ ){
		for ( unsigned int pgIndex2 = 0; pgIndex2 < context->getPGContainer()->getNumPatchGraphs(); pgIndex2++ ){
			PatchGraph* pg1 = context->getPGContainer()->getPatchGraph( pgIndex1 );
			PatchGraph* pg2 = context->getPGContainer()->getPatchGraph( pgIndex2 );
			card32 frameStart1 = pg1->getStartFrameIndex();
			card32 frameEnd1 = pg1->getStartFrameIndex() + pg1->getTrajectorySet()->getNumFrames() - 1;
			card32 frameStart2 = pg2->getStartFrameIndex();
			card32 frameEnd2 = pg2->getStartFrameIndex() + pg2->getTrajectorySet()->getNumFrames() - 1;
			if ( frameEnd1 + 1 == frameStart2 ){
				float32 score = estimateMergeScore( pg1, pg2, context->getPGContainer() );
				if ( score > bestScore ){
					bestScore = score;
					cand1 = pgIndex1;
					cand2 = pgIndex2;
				}
			}
		}
	}
}