//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphOpticalFlow.h"
#include "PointSetKNNQuery.h"
#include "PGDataPointIndices.h"
#include "SplatPullPushInterpolation.h"
#include "RGBAImage.h"
#include "ImageDisplayer.h"
#include "SeparatorClassProperty.h"
#include "Monomials2ndOrder.h"
#include "Monomial2ndOrderDisplayer.h"
#include "ProgressWindow.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphOpticalFlow, 0 )
{
	BEGIN_CLASS_INIT( PatchGraphOpticalFlow )
	ADD_OBJECT_PROP( splatInterpolation, 0, SplatPullPushInterpolation::getClass(), true )
	ADD_FLOAT32_PROP( influenceScale, 0 );
	ADD_SEPARATOR( "Visualization" )
	ADD_BOOLEAN_PROP( visualize, 0 )
}

/** 
 * Constructor
 */
PatchGraphOpticalFlow::PatchGraphOpticalFlow()
{
	splatInterpolation = new SplatPullPushInterpolation();
	influenceScale = 3.0f;
	imageSize = 50;
	visualize = true;
	comparisonWindow = NULL;
	errorImage = NULL;
	setImageSize( imageSize );
}

/** 
 * Destructor
 */
PatchGraphOpticalFlow::~PatchGraphOpticalFlow()
{
	delete splatInterpolation;
	if ( comparisonWindow != NULL ) delete comparisonWindow;
	if ( errorImage != NULL ) delete errorImage;
}

void PatchGraphOpticalFlow::setImageSize( const card32 imageSize )
{
	this->imageSize = imageSize;
	if ( comparisonWindow != NULL ) delete comparisonWindow;
	comparisonWindow = new TemplatedImage<Vector3f>( imageSize/2, imageSize/2 );
	if ( errorImage != NULL ) delete errorImage;
	errorImage = new TemplatedImage<float32>( imageSize - comparisonWindow->getSizeX(), imageSize - comparisonWindow->getSizeY() );	
}

/** 
 * Compute the gradient of the optical flow weighted by the corresponding eigenValues
 *  localFrameIndex and localFrameIndex+1 for all patches, return true on success
*/
bool PatchGraphOpticalFlow::computeAll( 
									vector<PatchGraphOpticalFlowGradient>& gradients, 
									PatchGraph* pg, const card32 localFrameIndex, 
									PointCloudAnimationSequence* animSeq, 
									const float32 sampling )
{
	if ( ( pg == NULL ) || ( pg->getTrajectorySet()->getNumFrames() <= localFrameIndex+1 ) ){
		debugOutput << "computeAll: invalid parameters.\n"; debugOutput.flush();
		return false;
	}

	// Compute all gradients
	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	gradients.resize( numPatches );
	for ( unsigned int patchIndex = 0; patchIndex < numPatches; patchIndex++ ){
		if ( !compute( gradients[patchIndex], pg, localFrameIndex, patchIndex, animSeq, sampling ) ){
			return false;
		}
	}
	return true;
}

/** 
 *  Compute the gradient of the optical flow weighted by the corresponding eigenValues
 *  localFrameIndex and localFrameIndex+1 at patch with number patchIndex, return true on success
 */
bool PatchGraphOpticalFlow::compute( 
		PatchGraphOpticalFlowGradient& gradient,
		PatchGraph* pg, const card32 localFrameIndex, const card32 patchIndex, 
		PointCloudAnimationSequence* animSeq, 
		const float32 sampling )
{
	if ( ( pg == NULL ) || ( pg->getTrajectorySet()->getNumFrames() <= localFrameIndex+1 ) || ( pg->getTrajectorySet()->getNumPatches() <= patchIndex ) ){
		debugOutput << "compute: invalid parameters.\n"; debugOutput.flush();
		return false;
	}
	
	PointSet* traj = pg->getTrajectorySet()->getTrajectories();
	AAT POSITION = traj->getAAT( "position" );
	AAT NORMAL = traj->getAAT( "normal" );
	AAT TANGENT_U = traj->getAAT( "tangent_u" );
	card32 globalFrameIndex = localFrameIndex + pg->getStartFrameIndex(); 
	card32 nDataIndex;
	bool found;
	Matrix3f T;
	T[0] = traj->get3f( localFrameIndex, patchIndex, TANGENT_U );
	T[2] = traj->get3f( localFrameIndex, patchIndex, NORMAL );
	T[1] = T[0].crossProduct( T[2] );
	T = T.transpose();
	vector<SplatPullPushInterpolation::TWorkingVertex> splats[2]; splats[0].clear(); splats[1].clear();
	SplatPullPushInterpolation::TWorkingVertex splat;
	AAT DATA_POSITION = animSeq->getFrame( 0 )->getAAT( "position" );
	AAT DATA_COLOR = animSeq->getFrame( 0 )->getAAT( "color" );

	Vector2f extremaU = makeVector2f( 1000000, -1000000 );
	Vector2f extremaV = makeVector2f( 1000000, -1000000 );
	for ( unsigned int imageIndex = 0; imageIndex < 2; imageIndex++ ){
		Vector3f pos = traj->get3f( localFrameIndex + imageIndex, patchIndex, POSITION );
		AttachedIndexedOctree* dataOctree = animSeq->fetchOctree( globalFrameIndex + imageIndex );
		PointSet* dataFrame = animSeq->getFrame( globalFrameIndex + imageIndex );
		PointSetKNNQuery knn( dataFrame, dataOctree, NULL_VECTOR3F, DATA_POSITION );
		knn.setTargetPoint( pos );
		knn.getNextNearestPointIndex( nDataIndex, found );
		Vector3f nDataPos = dataFrame->get3f( nDataIndex, DATA_POSITION );
		float32 d = norm( nDataPos - pos );
		while( found && ( d < sampling * influenceScale ) ){
			Vector3f p3D = T * ( nDataPos - pos );
			Vector2f uv = makeVector2f( p3D[0], p3D[1] );
			splat.paramPos = uv;
			if ( uv[0] < extremaU[0] ) extremaU[0] = uv[0];
			if ( uv[0] > extremaU[1] ) extremaU[1] = uv[0];
			if ( uv[1] < extremaV[0] ) extremaV[0] = uv[1];
			if ( uv[1] > extremaV[1] ) extremaV[1] = uv[1];
			Vector3f color = dataFrame->get3f( nDataIndex, DATA_COLOR );
			splat.attribute = makeVector4f( color[0], color[1], color[2], 0.0f );
			splats[imageIndex].push_back( splat );
			knn.getNextNearestPointIndex( nDataIndex, found );
			nDataPos = dataFrame->get3f( nDataIndex, DATA_POSITION );
			d = norm( nDataPos - pos );
		}
	}

	// Scale parameters to [0;1]
	float32 scaleU = 1.0f / ( 2.0f * max( fabs( extremaU[0] ), fabs( extremaU[1] ) ) );
	float32 scaleV = 1.0f / ( 2.0f * max( fabs( extremaV[0] ), fabs( extremaV[1] ) ) );
	TemplatedImage<Vector3f>* image[2];
	for ( unsigned int imageIndex = 0; imageIndex < 2; imageIndex++ ){
		for ( unsigned int n = 0; n < splats[imageIndex].size(); n++ ){
			Vector2f& param = splats[imageIndex][n].paramPos;
			splats[imageIndex][n].paramPos = makeVector2f( param[0] * scaleU + 0.5f, param[1] * scaleV + 0.5f );
		}
		image[imageIndex] = splatInterpolation->createImage3f( imageSize, splats[imageIndex], false );
	} 

	// Create comparison window
	for ( unsigned int x = 0; x < comparisonWindow->getSizeX(); x++ ){
		for ( unsigned int y = 0; y < comparisonWindow->getSizeY(); y++ ){
			comparisonWindow->setPixel( x, y, image[0]->getPixel( x+imageSize/4, y+imageSize/4 ) );
		}
	}
	card32 numComparionPixels = comparisonWindow->getSizeX() * comparisonWindow->getSizeY();

	// Compute error image
	float32 maxError = 0;
	for ( unsigned int i = 0; i < errorImage->getSizeX(); i++ ){
		for ( unsigned int j = 0; j < errorImage->getSizeY(); j++ ){
			float32 error = 0;
			for ( unsigned int x = 0; x < comparisonWindow->getSizeX(); x++ ){
				for ( unsigned int y = 0; y < comparisonWindow->getSizeY(); y++ ){
					error += norm( comparisonWindow->getPixel( x, y ) - image[1]->getPixel( x+i, y+j ) );
				}
			}
			error /= float32( numComparionPixels );
			if ( error > maxError ) maxError = error; 
			errorImage->setPixel( i, j, error );
		}
	}

	// Compute ellipsoid
	vector<Vector3f> heighfield;
	for ( unsigned int x = 0; x < errorImage->getSizeX(); x++ ){
		for ( unsigned int y = 0; y < errorImage->getSizeY(); y++ ){
			Vector3f hfPoint = makeVector3f( 
				x / float32( errorImage->getSizeX() ), 
				y / float32( errorImage->getSizeY() ), 
				errorImage->getPixel(x,y) / maxError );
			heighfield.push_back( hfPoint );
		}
	}
	Monomials2ndOrder::computeCoeffs( heighfield, gradient.coeffs );
	StaticMatrix<float32,2,2> M, eigenVectors;
	M[0][0] = gradient.coeffs[3];			//uu
	M[0][1] = M[1][0] = gradient.coeffs[4];	//uv
	M[1][1] = gradient.coeffs[5];			//vv
	Vector2f eigenValues;
	M.computeEigenStructure( eigenValues, eigenVectors );
	T = T.transpose();

	// Compute saddle point
	M[0][0] = 2.0f * gradient.coeffs[3];
	M[1][0] = M[0][1] = gradient.coeffs[4];
	M[1][1] = 2.0f * gradient.coeffs[5];
	bool hasSaddlePoint = false;
	Vector2f saddleCoords;
	if( fabs( M[0][0] * M[1][1] - M[1][0] * M[1][1] ) > 1e-8 ){
		saddleCoords = invertMatrix( M ) * makeVector2f( -gradient.coeffs[1], -gradient.coeffs[2] );
		if ( ( saddleCoords[0] >= 0 ) && ( saddleCoords[0] <= 1 ) && ( saddleCoords[1] >= 0 ) && ( saddleCoords[1] <= 1 ) ){
			hasSaddlePoint = true;
		}
	} 

	// Compute gradient
	if ( hasSaddlePoint ){
		gradient.g = makeVector3f( ( saddleCoords[0] - 0.5f )/ scaleU, ( saddleCoords[1] - 0.5f ) / scaleV, 0 );
		gradient.g = T * gradient.g;
	} else {
		gradient.g = makeVector3f( 0,0,0 );
	}

	if ( visualize ){
		// Show input images in event displayer
		for ( unsigned int imageIndex = 0; imageIndex < 2; imageIndex++ ){
			static RGBAFloat32Image* rgbaImage = new RGBAFloat32Image( image[imageIndex]->getSizeX(),image[imageIndex]->getSizeY() );
			Vector3f* src = image[imageIndex]->getPixels();
			Vector4f* dst = rgbaImage->getPixels();
			for( card32 i = 0; i < image[imageIndex]->getSizeX() * image[imageIndex]->getSizeY(); i++ ){
				dst[i] = expand3To4( src[i] );
			}
			ImageDisplayer::DisplayRGBAFloat32Image ev( rgbaImage );
			QString name = "Optical Flow - image " + QString::number( imageIndex );
			debugOutput.displayEvent( ImageDisplayer::getClass(), name.toStdString(), &ev );
		}

		// Show errorImage in event displayer
		static RGBAFloat32Image* rgbaImage = NULL;
		if( rgbaImage == NULL ) rgbaImage = new RGBAFloat32Image( errorImage->getSizeX(),errorImage->getSizeY() );
		float32* src = errorImage->getPixels();
		Vector4f* dst = rgbaImage->getPixels();
		card32 numPixels = errorImage->getSizeX() * errorImage->getSizeY();
		for( card32 i = 0; i < numPixels; i++ ){
			dst[i] = makeVector4f( src[i], src[i], src[i], 1.0f );
		}
		ImageDisplayer::DisplayRGBAFloat32Image ev( rgbaImage );
		debugOutput.displayEvent( ImageDisplayer::getClass(), "Error image", &ev );

		// Show monomial in event displayer
		Monomial2ndOrderDisplayer* m2dDisplayer = new Monomial2ndOrderDisplayer();
		m2dDisplayer->setup( gradient.coeffs, makeVector2f( 0,1 ), makeVector2f( 0, 1 ), 10 );
		m2dDisplayer->setPoints( heighfield );
		float32 evScale = 0.3f / max( fabs( eigenValues[0] ), fabs( eigenValues[1] ) );
		m2dDisplayer->setEigenVectors( 
			makeVector3f( eigenVectors[0][0], eigenVectors[0][1], 0 ) * eigenValues[0] * evScale,
			makeVector3f( eigenVectors[1][0], eigenVectors[1][1], 0 ) * eigenValues[1] * evScale );
		m2dDisplayer->setSaddle( hasSaddlePoint, saddleCoords );
		GLEventDisplayer::SetNewGLEventObject ev3D( m2dDisplayer );
		debugOutput.displayEvent( GLEventDisplayer::getClass(), "Error height field", &ev3D );
	}

	// Free allocated mem
	delete image[0];
	delete image[1];

	return true;
}