//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphPruning.h"
#include "PointSetKNNQuery.h"
#include "PGDataPointIndices.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphPruning, 0 )
{
	BEGIN_CLASS_INIT( PatchGraphPruning )
	ADD_FLOAT32_PROP( matchingThreshold, 0 )
	ADD_FLOAT32_PROP( sizeThreshold, 0 )
	ADD_FLOAT32_PROP( prunePatchGraphThreshold, 0 );
}

/** 
 * Constructor
 */
PatchGraphPruning::PatchGraphPruning()
{
	context = NULL;
	matchingThreshold = 0.5f;
	sizeThreshold = 3.0f;
	prunePatchGraphThreshold = 0.5f;
}

/** 
 * Destructor
 */
PatchGraphPruning::~PatchGraphPruning()
{
}

/** 
 * Prune patch graphs which are too small
 */
void PatchGraphPruning::pruneSmallPatchGraphs( bool verbose )
{
	// Prune invalid patch graphs (no data assigned)
	card32 numPruned = 0;
	for( unsigned int patchGraphIndex = 0; patchGraphIndex < context->getPGContainer()->getNumPatchGraphs(); patchGraphIndex++ ){
		PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
		if ( pg->getTrajectorySet()->getNumPatches() < context->getMinPatchGraphSize() ){
			// Check for too small patch graphs
			//debugOutput << "   Pruning patch graph (too small).\n"; debugOutput.flush();
			context->getPGContainer()->deletePatchGraph( patchGraphIndex );
			patchGraphIndex--;
			numPruned++;
		}
	}
	if ( verbose ) { if ( numPruned > 0 ) { debugOutput << "   Pruned " << numPruned << " patch graphs (too small).\n"; debugOutput.flush(); } }
}

/**
 * Prune patch graphs which are not supported by data points
 */
void PatchGraphPruning::pruneNoDataPatchGraphs( bool verbose )
{
	// Prune invalid patch graphs (no data assigned)
	for( unsigned int patchGraphIndex = 0; patchGraphIndex < context->getPGContainer()->getNumPatchGraphs(); patchGraphIndex++ ){
		PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
		AAT POSITION = pg->getTrajectorySet()->getTrajectories()->getAAT( "position" );
		float32 notSupported = 0;
		for( unsigned int frameIndex = 0; frameIndex < pg->getTrajectorySet()->getNumFrames(); frameIndex++ ){
			card32 globalFrameIndex = pg->getStartFrameIndex() + frameIndex;
			PointSet* ps = context->getPGContainer()->getAnimSeq()->getFrame( globalFrameIndex );
			AttachedIndexedOctree* dataOctree = context->getPGContainer()->getAnimSeq()->fetchOctree( globalFrameIndex );
			bool found;
			card32 nIndex;
			AAT DATA_POSITION = ps->getAAT( "position" );
			PointSetKNNQuery knn( ps, dataOctree, NULL_VECTOR3F, DATA_POSITION );	
			for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
				Vector3f pos = pg->getTrajectorySet()->getTrajectories()->get3f( frameIndex, patchIndex, POSITION );
				// Check a data point is close
				knn.setTargetPoint( pos );
				knn.getNextNearestPointIndex( nIndex, found );
				Vector3f nPos = ps->get3f( nIndex, DATA_POSITION );
				if ( norm( pos - nPos ) > context->getSamplingSpacing() ){
					notSupported += 1.0f;
				}
			}
		}
		notSupported /= float32( pg->getTrajectorySet()->getNumFrames() * pg->getTrajectorySet()->getNumPatches() );
		if ( notSupported > prunePatchGraphThreshold ){
			if ( verbose ) { debugOutput << "   Pruning patch graph (" << ( notSupported * 100.0f ) << "% of patches not sufficiently supported by data).\n"; debugOutput.flush(); }
			context->getPGContainer()->deletePatchGraph( patchGraphIndex );
			patchGraphIndex--;
		}
	}
}


/** 
 * Prune patch graphs which represent data contained in other patch graphs
 */
void PatchGraphPruning::pruneIncludedPatchGraphs( bool verbose )
{
	// Detect patch graphs which represent the same data ->merge
	for ( unsigned int i = 0; i < context->getPGContainer()->getNumPatchGraphs(); i++ ){
		for ( unsigned int j = i+1; j < context->getPGContainer()->getNumPatchGraphs(); j++ ){
			PatchGraph* pg1 = context->getPGContainer()->getPatchGraph( i );
			PatchGraph* pg2 = context->getPGContainer()->getPatchGraph( j );
			int32 pg1Start = pg1->getStartFrameIndex();
			int32 pg1End = pg1->getStartFrameIndex() + pg1->getTrajectorySet()->getNumFrames();
			int32 pg2Start = pg2->getStartFrameIndex();
			int32 pg2End = pg2->getStartFrameIndex() + pg2->getTrajectorySet()->getNumFrames();
			card32 pg1NumPatches = pg1->getTrajectorySet()->getNumPatches();
			card32 pg2NumPatches = pg2->getTrajectorySet()->getNumPatches();
			if ( ( pg1Start >= pg2Start ) && ( pg1End <= pg2End ) && ( sizeThreshold * pg1NumPatches < pg2NumPatches ) ){
				// Check if pg1 fits into pg2
				if ( isContainedFirstInSecond( pg1, pg2 ) ){
					context->getPGContainer()->deletePatchGraph( i );
					i--;
					j--;
					if ( verbose ) { 
						debugOutput << "   Removed patch graph which was no longer needed.\n"; debugOutput.flush();
					}
				}
			} else if ( ( pg2Start >= pg1Start ) && ( pg2End <= pg1End ) && ( sizeThreshold * pg2NumPatches < pg1NumPatches ) ){
				// Check if pg2 fits into pg1
				if ( isContainedFirstInSecond( pg2, pg1 ) ){
					context->getPGContainer()->deletePatchGraph( j );
					j--;
					if ( verbose ) { 
						debugOutput << "   Removed patch graph which was no longer needed.\n"; debugOutput.flush();	
					}
				}
			}
		}
	}
}


/**  
 * DEPRECATED: NOT REQUIRED ANY LONGER
 * Use all data points associated with pg1 to update pg2
 *
void PatchGraphPruning::reassignDataPointsFromFirstToSecond( const PatchGraph* pg1, PatchGraph* pg2 )
{
	AAT DATA_POSITION = context->getPGContainer()->getAnimSeq()->getFrame( 0 )->getAAT( "position" );
	for ( unsigned pg1FrameIndex = 0; pg1FrameIndex < pg1->getTrajectorySet()->getNumFrames(); pg1FrameIndex++ ){
		card32 pg2FrameIndex = pg1FrameIndex + pg1->getStartFrameIndex() - pg2->getStartFrameIndex();
		card32 globalFrameIndex = pg1FrameIndex + pg1->getStartFrameIndex();
		PointSet* ps = pg2->getTrajectorySet()->copySingleFrame( pg2FrameIndex );
		AAT POSITION = ps->getAAT( "position" );
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps );
		PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );	
		card32 nIndex;
		bool found;

		for ( unsigned int patchIndex = 0; patchIndex < pg1->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			// Allocate all data points in pg1 to pg2
			vector<card32> indices = pg1->getTrajectorySet()->getDataPoints()->getIndices( pg1FrameIndex, patchIndex );
			for ( unsigned int n = 0; n < indices.size(); n++ ){
				card32 dataIndex = indices[n];
				Vector3f dataPos = context->getPGContainer()->getAnimSeq()->getFrame( globalFrameIndex )->get3f( dataIndex, DATA_POSITION );
				// Find closest patch in pg2;
				knn.setTargetPoint( dataPos );
				knn.getNextNearestPointIndex( nIndex, found );
				Vector3f pos = ps->get3f( nIndex, POSITION );
				// Add data point if it lies close enough
				if ( norm ( pos - dataPos ) < context->getSamplingSpacing() * context->getDataPointsInfluence() ){
					pg2->getTrajectorySet()->getDataPoints()->addIndex( pg2FrameIndex, nIndex, dataIndex );
				}
			}
		}

		delete octree;
		delete ps;
	}
}
*/
	
/** 
 * Check if first patch graph is contained in second patch graph
 */
bool PatchGraphPruning::isContainedFirstInSecond( const PatchGraph* pg1, const PatchGraph* pg2 ) const
{
	card32 startFrameOffset = pg1->getStartFrameIndex() - pg2->getStartFrameIndex();
	AAT POSITION = pg1->getTrajectorySet()->getTrajectories()->getAAT( "position" );
	card32 nIndex;
	bool found;

	// Check if patches in all frames of pg1 fit with patches in corresponding pg2-frames
	float32 matching = 0;
	for ( unsigned int frameIndex = 0; frameIndex < pg1->getTrajectorySet()->getNumFrames(); frameIndex++ ){
		card32 pg2FrameIndex = startFrameOffset + frameIndex;
		PointSet* ps = pg2->getTrajectorySet()->copySingleFrame( pg2FrameIndex );
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps );
		PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );	
		for ( unsigned int patchIndex = 0; patchIndex < pg1->getTrajectorySet()->getNumPatches(); patchIndex++ ){
			Vector3f pos = pg1->getTrajectorySet()->getTrajectories()->get3f( frameIndex, patchIndex, POSITION );
			// Find closest patch in pg2;
			knn.setTargetPoint( pos );
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = ps->get3f( nIndex, POSITION );
			if ( norm( pos - nPos ) < context->getSamplingSpacing() ){
				matching += 1.0f;
			}
		}
		delete octree;
		delete ps;
	}
	matching /= float32( pg1->getTrajectorySet()->getNumFrames() * pg1->getTrajectorySet()->getNumPatches() );
	//debugOutput << " matching score: " << matching << ".\n"; debugOutput.flush();
	return ( matching > matchingThreshold );
}