//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphResample.h"
//---------------------------------------------------------------------------
#include "PatchGraphContainer.h"
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "PGDataPointIndices.h"
#include "PointSetKNNQuery.h"
#include <time.h>
#include <stdio.h>
#include <dos.h>
#include "PointSet_templCode.h"
#include "TGNeighborhoodKNNSearch.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PatchGraphResample, 0 )
{
	BEGIN_CLASS_INIT( PatchGraphResample )
}

/**
 * Constructor
 */
PatchGraphResample::PatchGraphResample() 
{
	context = NULL;
}

/**
 * Destructor
 */
PatchGraphResample::~PatchGraphResample() 
{ 
}

/** 
 * Object assignment
 */
void PatchGraphResample::assign( const Object* obj, CopyContext *context ) 
{
	copyObjectProperties( obj, this );
}

/** 
 * Update topology graph after merge operation
 */
bool PatchGraphResample::resample( PatchGraph* pg )
{
	PointSet* oldTraj = pg->getTrajectorySet()->getTrajectories();
	InCorePCTopologyGraph* oldTG = pg->getTopology();

	if ( !oldTraj->providesAttribute( "flags" ) ||
		!oldTraj->providesAttribute( "position" ) ){
			debugOutput << "!!resample: attributes position, flags required!!\n";
			debugOutput.flush();
			return false;
	}
	AAT POSITION = oldTraj->getAAT( "position" );
	AAT FLAGS = oldTraj->getAAT( "flags" );

	// Reset flags
	card32 oldNumPatches = pg->getTrajectorySet()->getNumPatches();
	for ( unsigned int pointIndex = 0; pointIndex < oldNumPatches; pointIndex++ ){
		oldTraj->set1i( 0, pointIndex, FLAGS, oldTraj->get1i( 0, pointIndex, FLAGS ) & !PF_FLAG_MODIFIED );
	}

	// Detect remove points
	PointSet* oldPS = pg->getTrajectorySet()->copySingleFrame( 0 );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( oldPS );
	PointSetKNNQuery knn( oldPS, octree, NULL_VECTOR3F, POSITION );
	card32 nIndex = 0;
	bool found;
	vector<card32> newPatches; newPatches.clear();  // Container for the new patch indices in point set
	vector<int32> map; map.clear();					// Mapping to new indices in point set
	card32 newPointIndex = 0;
	for ( unsigned int pointIndex = 0; pointIndex < oldPS->getNumEntries(); pointIndex++ ){
		if ( !( oldPS->get1i( pointIndex, FLAGS ) & PF_FLAG_MODIFIED ) ){
			// Point gets kept
			map.push_back( newPointIndex );
			Vector3f pos = oldPS->get3f( pointIndex, POSITION );
			newPatches.push_back( pointIndex );
			knn.setTargetPoint( pos );
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = oldPS->get3f( nIndex, POSITION );
			float32 dist = norm( nPos - pos );
			while( dist < context->getSamplingSpacing() ){
				// Check if point can be deleted
				if ( nIndex > pointIndex ){
					// Mark as delete
					StaticVector<int32,1> v; v[0] = oldPS->get1i( nIndex, FLAGS ) | PF_FLAG_MODIFIED;
					oldPS->set1i( nIndex, FLAGS, v );
				}
				// Next
				knn.getNextNearestPointIndex( nIndex, found );
				Vector3f nPos = oldPS->get3f( nIndex, POSITION );
				dist = norm( nPos - pos );
			}
			newPointIndex++;
		} else {
			// Point gets removed
			map.push_back( -1 );
		}
	}
	delete oldPS;
	delete octree;

	// Create new point set
	TrajectorySet* newTS = new TrajectorySet();
	PointSet* newTraj = new PointSet();
	newTraj->clearAndSetup( pg->getTrajectorySet()->getNumFrames(), newPatches.size(), oldTraj->getDescr() );
	newTS->setStartFrameIndex( pg->getStartFrameIndex() );
	newTS->clearAndSetup( newTraj );
	InCorePCTopologyGraph* newTG = new InCorePCTopologyGraph();
	VertexDescriptor vd;
	vd.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( VAD( "color", 4, VAD::DATA_FORMAT_FLOAT32 ) );
	newTG->clearAndSetup( &vd );
	PGDataPointIndices* dataPointIndices = new PGDataPointIndices();
	dataPointIndices->setup( pg->getTrajectorySet()->getNumFrames(), (card32)newPatches.size() );
	newTS->clearAndSetup( newTraj, dataPointIndices );

	// Set vertices
	for ( unsigned int frameIndex = 0; frameIndex < pg->getTrajectorySet()->getNumFrames(); frameIndex++ ){
		for( unsigned int newPatchIndex = 0; newPatchIndex < newPatches.size(); newPatchIndex++ ){
			newTraj->setVertex( frameIndex, newPatchIndex, oldTraj->getDataPointer( frameIndex, newPatches[newPatchIndex] ) );
			newTraj->set<int>( frameIndex, newPatchIndex, FLAGS, newTraj->get<int>( frameIndex, newPatchIndex, FLAGS) & (0xFFFFFFFF ^ PF_FLAG_MODIFIED ) );
		}
	}

   //if (oldTraj->getNumEntries() == 0) {
   //   warning("PatchGraphResample::resample() - no points found.");
   //   return false;
   //}

	// Set topology topology
	TGNeighborhoodKNNSearch oldTknn( oldTraj, oldTG, 0, 0 );
	card32 numEdges = 0;
	for ( unsigned int patchIndex = 0; patchIndex < newPatches.size(); patchIndex++ ){
		oldTknn.reset( 0, newPatches[patchIndex] );
		numEdges = 0;
		while( !oldTknn.atEnd() && ( numEdges < 8 ) ){
			card32 oldNPatchIndex = oldTknn.getCurrentIndex();
			int32 newNPatchIndex = map[oldNPatchIndex];
			if ( ( newNPatchIndex >= 0 ) && ( newNPatchIndex != patchIndex ) ){
				newTG->addEdge( patchIndex, newNPatchIndex );
				numEdges++;
			}
			oldTknn.next();
		}
		if ( numEdges < 8 ){
			debugOutput << "resample: wasn't able to assign 8 topological connections.\n"; debugOutput.flush();			
		}
	}

	pg->clearAndSetup( newTS, newTG, false );
	return true;
}