//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphTopologicalResampling.h"
//---------------------------------------------------------------------------
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "PGMultiResolution.h"
#include "TGNeighborhoodKNNSearch.h"
#include "BSROptPotential.h"
#include "PointSetKNNQuery.h"
#include "PGDataPointIndices.h"
#include "PointSet_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_ABSTRACT_CLASS( PatchGraphTopologicalResampling ,0)
{
	BEGIN_CLASS_INIT( PatchGraphTopologicalResampling );
}

/// ----

PatchGraphTopologicalResampling::PatchGraphTopologicalResampling() {
   useAverageDistance = true;
   singleFrame = 0;
}

void PatchGraphTopologicalResampling::setup(bool useAverageDistance, card32 frame) {
   this->useAverageDistance = useAverageDistance;
   this->singleFrame = frame;
}

struct PQEntry {
	card32 index;
	float32 dist;
	bool operator<(const PQEntry &other) const {
		return dist < other.dist;
	}
};

inline float32 PatchGraphTopologicalResampling::compDistance(const PointSet *ps, const PointSet *urshape, card32 p1, card32 p2, const AAT &posAAT) {
   if (useAverageDistance) {
      float32 d = 0;
      const card32 n = ps->getWidth();
      for (card32 i=0; i<n; i++) {
         d += norm(ps->get<float32, 3>(i, p1, posAAT) - ps->get<float32, 3>(i, p2, posAAT));
      }
      return d/n;
   } else {
      if (singleFrame != BSROPT_FRAME_URSHAPE) {
         return norm(ps->get<float32, 3>(singleFrame, p1, posAAT) - ps->get<float32, 3>(singleFrame, p2, posAAT));
      } else {
         return norm(urshape->get<float32, 3>(singleFrame, p1, posAAT) - urshape->get<float32, 3>(singleFrame, p2, posAAT));
      }
   }
}


PatchGraph *PatchGraphTopologicalResampling::resamplePatches
  (const PatchGraph* fullResolution, 
	float32 maxDist,
	vector<bool> *removable)
{
	const PointSet *ps = fullResolution->getTrajectorySet()->getTrajectories();
	const PointSet *urshape = fullResolution->getUrshape();
	const card32 numPoints = ps->getHeight();
	const InCorePCTopologyGraph *tg = fullResolution->getTopology();

   // --

	vector<card32> overlap;
	overlap.resize(numPoints, 0);
	vector< vector<card32> > overlapPoints;
	overlapPoints.resize(numPoints);
	vector<bool> deletePtn;
	deletePtn.resize(numPoints, false);

	AAT POS = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);

   PGNeighborhoodKNNSearch knn(fullResolution, singleFrame, 0, "position", useAverageDistance);
	
	priority_queue<PQEntry> removalCandidates;

	for (card32 i=0; i<numPoints; i++) {
		knn.reset(singleFrame, i);
		card32 index = 0;
		bool found = !knn.atEnd();
		pAssert(found);
	   index = knn.getCurrentIndex();
	   knn.next();
		float32 minDistance = 1E25f;
		while (found && compDistance(ps, urshape, index, i, POS) <= maxDist) {
			overlap[i]++;
			overlapPoints[i].push_back(index);
         found = !knn.atEnd();
			if (found) {
			   index = knn.getCurrentIndex();
			   knn.next();
				float32 distance = compDistance(ps, urshape, index, i, POS);
				if (distance < minDistance) {
					minDistance = distance;
				}
			}
		}
		if (minDistance < 1E25) {
			PQEntry e;
			e.dist = -minDistance;
			e.index = i;
			removalCandidates.push(e);
		}
	}
	
	card32 newNumPoints = numPoints;
	
	while (!removalCandidates.empty()) {
		PQEntry e = removalCandidates.top();
		removalCandidates.pop();
		if (overlap[e.index] > 1) {
			bool othersOk = true;
			const size_t otherCount = overlapPoints[e.index].size();
			for (card32 i=0; i<otherCount; i++) {
				if (overlap[overlapPoints[e.index][i]] <= 1) {
					othersOk = false;
				}
			}
			// check for constrained points
			if (removable) {
				othersOk = othersOk && ((*removable)[e.index]);
			}
			if (othersOk) {
				overlap[e.index]--;
				for (card32 i=0; i<otherCount; i++) {
					overlap[overlapPoints[e.index][i]]--;
				}
				deletePtn[e.index] = true;
				pAssert(newNumPoints > 0);
            newNumPoints--;
			}
		}	
	}
	
	TrajectorySet *resultTs = new TrajectorySet();
	PointSet *resultTraj = new PointSet();
	resultTraj->clearAndSetup(
		fullResolution->getTrajectorySet()->getTrajectories()->getWidth(), 
		fullResolution->getTrajectorySet()->getTrajectories()->getHeight(), 
		fullResolution->getTrajectorySet()->getTrajectories()->getDescr()
	);
	PGDataPointIndices *resultDp = new PGDataPointIndices();
	resultDp->setup(
		fullResolution->getTrajectorySet()->getNumFrames(), 
		fullResolution->getTrajectorySet()->getNumPatches()
	);
	
	resultTs->clearAndSetup(resultTraj, resultDp);
	InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();
	resultTg->getEdges()->clearAndSetup(
		resultTg->getEdges()->getWidth(), 
		resultTg->getEdges()->getHeight(), 
		fullResolution->getTopology()->getEdges()->getDescr()
	);
	PatchGraph *result = new PatchGraph();	
	result->clearAndSetup(resultTs, resultTg, fullResolution->getIn2d());

	result->getTrajectorySet()->getTrajectories()->clearAndSetup(ps->getWidth(), newNumPoints, ps->getDescr());
	if (result->getUrshape()) {
      result->getUrshape()->clearAndSetup(ps->getWidth(), newNumPoints, fullResolution->getUrshape()->getDescr());
	}
   bool writeOriginalIndex = false;
   if (!result->getTrajectorySet()->getTrajectories()->providesAttribute("originalIndex")) {
      result->createPatchAttribute(mVAD("originalIndex", 1, VAD::DATA_FORMAT_INT32));
      writeOriginalIndex = true;
   }
   const AAT ORIGIND = result->getTrajectorySet()->getTrajectories()->getAAT("originalIndex");
	
	result->getTrajectorySet()->getDataPoints()->setup(ps->getWidth(), newNumPoints);
	
	card32 targetPos = 0;
	for (card32 i=0; i<numPoints; i++) {
		if (!deletePtn[i]) {
			pAssert(targetPos < newNumPoints);
			for (card32 frame=0; frame<ps->getWidth(); frame++) {
   			result->getTrajectorySet()->getTrajectories()->setVertex(frame, targetPos, ps->getDataPointer(frame, i));
			   if (writeOriginalIndex) {
			      result->getTrajectorySet()->getTrajectories()->set<int32>(frame, targetPos, ORIGIND, (int32)i);
			   }
   	   }
			if (result->getUrshape()) {
			   result->getUrshape()->setVertex(targetPos, urshape->getDataPointer(i));
			   if (writeOriginalIndex) {
			      result->getUrshape()->set<int32>(targetPos, ORIGIND, (int32)i);
			   }
			}
			targetPos++;
		}
	}
	result->computeBBox();
	result->getTrajectorySet()->setStartFrameIndex( fullResolution->getTrajectorySet()->getStartFrameIndex() );
	return result;   
}

PatchGraph *PatchGraphTopologicalResampling::fullResampling
  (const PatchGraph* fullResolution, 
   float32 maxDist,
   vector< vector<card32> > *clusters,
   vector<bool> *removable) 
{
	vector< vector<card32> > tmpClusters;
	if (!clusters) {
		clusters = &tmpClusters;
	}
	PatchGraph *result = resamplePatches(fullResolution, maxDist, removable);
	topologicalClustering(result, fullResolution, clusters);
	reassignDataPoints(result, fullResolution, *clusters);
	return result;
}

struct Edge {
 private:
	card32 v1;
	card32 v2;
 public:
	inline Edge() {}
	inline void setup(card32 v1, card32 v2) {
		this->v1 = min(v1,v2);
		this->v2 = max(v1,v2);
	}
	inline Edge(card32 v1, card32 v2) {
		setup(v1,v2);
	}
	card32 getV1() {return v1;}
	card32 getV2() {return v2;}
};

struct PatchEntry {
	card32 patchIndex;
	card32 nextVertexIndex;
	float32 dist;
	
	inline PatchEntry() {}
	inline PatchEntry(card32 p, card32 nv, float32 d) {
		patchIndex = p;
		nextVertexIndex = nv;
		dist = d;
	}
	inline bool operator<(const PatchEntry &other) const {
		return dist < other.dist;
	}
};


void PatchGraphTopologicalResampling::topologicalClustering
  (PatchGraph *reduced, 
	const PatchGraph *fullResolution,
   vector< vector<card32> > *clusters)
{
	const card32 reducedNumPatches = reduced->getTrajectorySet()->getNumPatches();
	const card32 fullNumPatches = fullResolution->getTrajectorySet()->getNumPatches();
	const PointSet *full_ps = fullResolution->getTrajectorySet()->getTrajectories();
	const PointSet *full_urshape = fullResolution->getUrshape();
   const AAT REDUCED_ORIGIND = reduced->getTrajectorySet()->getTrajectories()->getAAT("originalIndex");
   const AAT FULL_POS = full_ps->getAAT("position");
	const InCorePCTopologyGraph *full_tg = fullResolution->getTopology();
	InCorePCTopologyGraph *reduced_tg = reduced->getTopology();
	//const PGDataPointIndices *full_dp = fullResolution->getTrajectorySet()->getDataPoints();
	//PGDataPointIndices *reduced_dp = reduced->getTrajectorySet()->getDataPoints();
	const card32 numFrames = full_ps->getWidth();
	pAssert(numFrames == reduced->getTrajectorySet()->getTrajectories()->getWidth());

	if (clusters) {
		clusters->clear();
		clusters->resize(reducedNumPatches);
	}
	
	// --- initialization: put each reduced vertex in queue (in region and in next-vertex-per-region queue)
	priority_queue<PatchEntry> patchesQueue;
	for (card32 i=0; i<reducedNumPatches; i++) {
		const card32 originalIndex = reduced->getTrajectorySet()->getTrajectories()->get<int>(0, i, REDUCED_ORIGIND);
		patchesQueue.push(PatchEntry(i, originalIndex, 0));
	}	
	
	// --- simultaneous region growing
	// topology: connectivity of regions (keep symmetric during region growing)
	vector< set<card32> > topology(reducedNumPatches);
	// processed vertices, pair is [vertex, takenBy]
	map<card32, card32> processedVertices;
	while (!patchesQueue.empty()) {
		// get next region
		PatchEntry region = patchesQueue.top();
		patchesQueue.pop();
			
		// check whether the vertex has already been taken
		map<card32, card32>::iterator processedVertexFindPos = processedVertices.find(region.nextVertexIndex);
		if (processedVertexFindPos == processedVertices.end()) {
			// vertex can be taken by region.patchIndex
			processedVertices.insert(pair<card32, card32>(region.nextVertexIndex, region.patchIndex));
			// update clustering
			if (clusters) {
				(*clusters)[region.patchIndex].push_back(region.nextVertexIndex);
			}
			// // update data points
			// for (card32 frameNum=0; frameNum<numFrames; frameNum++) {
			// 	reduced_dp->addIndices(frameNum, region.patchIndex, full_dp->getIndices(frameNum, region.nextVertexIndex) );
			// }
			// add neighboring vertices to queue
			const card32 numNeighbors = full_tg->getNumAdjacentVertices(region.nextVertexIndex);
			for (card32 i=0; i<numNeighbors; i++) {
				const card32 j = full_tg->getAdjacentVertex(region.nextVertexIndex, i);
				// not yet taken?
				map<card32, card32>::iterator newEdgeFind = processedVertices.find(j);
				if (newEdgeFind == processedVertices.end()) {
					const card32 originalIndex = reduced->getTrajectorySet()->getTrajectories()->get<int>(0, region.patchIndex, REDUCED_ORIGIND);
					float32 distance = compDistance(full_ps, full_urshape, originalIndex, j, FULL_POS);
					// put on this region's queue
					patchesQueue.push(PatchEntry(region.patchIndex, j, -distance));
				} else {
					topology[region.patchIndex].insert(newEdgeFind->second);
					topology[newEdgeFind->second].insert(region.patchIndex);
				}
			}
		} else {
			// hit neighboring region - upate topology (symmetric)
			topology[region.patchIndex].insert(processedVertexFindPos->second);
			topology[processedVertexFindPos->second].insert(region.patchIndex);
		}
	}
	// copy topology into "tg" container in "reduced"
	reduced_tg->clearAndSetup(full_tg->getEdges()->getDescr());
	const size_t numTop = topology.size();
	for (unsigned topIndex = 0; topIndex < numTop; topIndex++) {
		set<card32>::iterator pos = topology[topIndex].begin();
		set<card32>::iterator end = topology[topIndex].end();
		while (pos != end) {
			reduced_tg->addEdge(topIndex, *pos);
			++pos;
		}
	}	
}

void PatchGraphTopologicalResampling::collectTopEpsNeighbors(
	const PointSet *ps, 
	const PointSet *urshape, 
	const InCorePCTopologyGraph *tg,
	const card32 patchIndex,
	const float32 radius, 
	const AAT &posAAT,
	vector<card32> &result)
{
	vector<card32> candidates;
	candidates.push_back(patchIndex);
	set<card32> visited;
	while (!candidates.empty()) {
		card32 currentPos = candidates.back();
		candidates.pop_back();
		visited.insert(currentPos);
		float32 dist = compDistance(ps, urshape, patchIndex, currentPos, posAAT);
		if (dist <= radius) {
			result.push_back(currentPos);
			for (card32 i=0; i<tg->getNumAdjacentVertices(currentPos); i++) {
				card32 j = tg->getAdjacentVertex(currentPos, i);
				if (visited.find(j) == visited.end()) {
					visited.insert(j);
					candidates.push_back(j);
				}
			}
		}		
	}
}

void PatchGraphTopologicalResampling::computeInterpolationMap(
	const PatchGraph *reduced, const PatchGraph *fullResolution,
	const vector< vector<card32> > &clusters, const float32 maxDist,
	const AAT &reducedPosAAT,
	const AAT &fullResPosAAT,
	const AAT &reducedOrigIndexAAT,
	vector< vector< pair<card32, float32> > > &weights)
{
	weights.resize(fullResolution->getTrajectorySet()->getNumPatches());
	const float32 clampDist = maxDist*2;
	for (unsigned c=0; c<clusters.size(); c++) {
		vector<card32> neighbors;
		collectTopEpsNeighbors(
			reduced->getTrajectorySet()->getTrajectories(),
			reduced->getUrshape(),
			reduced->getTopology(),
			c,
			clampDist+maxDist,
			reducedPosAAT,
			neighbors
		);
		for (unsigned pi=0; pi<clusters[c].size(); pi++) {
			card32 fullResPoint = clusters[c][pi];
			for (unsigned ni=0; ni<neighbors.size(); ni++) {
				int32 origIndex = reduced->getTrajectorySet()->getTrajectories()->get<int>(0, neighbors[ni], reducedOrigIndexAAT);
				float32 dist = compDistance(
					fullResolution->getTrajectorySet()->getTrajectories(),
					fullResolution->getUrshape(),
					origIndex,
					fullResPoint,
					fullResPosAAT
				);
				float32 w = exp(-sqr(dist / maxDist))-exp(-4.0f);
				if (w > 0) {
					weights[fullResPoint].push_back(pair<card32,float32>(neighbors[ni], w));
				}
			}
			float32 wsum = 0.0f;
			for (unsigned entry=0; entry<weights[fullResPoint].size(); entry++) {
				wsum += weights[fullResPoint][entry].second;
			}
			if (wsum > 0) {
				for (unsigned entry=0; entry<weights[fullResPoint].size(); entry++) {
					weights[fullResPoint][entry].second /= wsum;
				}
			}
		}
	}
}

void PatchGraphTopologicalResampling::reassignDataPoints	//xxx doppelt
  (PatchGraph *reduced, const PatchGraph *fullResolution,
   const vector< vector<card32> > &clusters)
{
	const PGDataPointIndices *pgd_full = fullResolution->getTrajectorySet()->getDataPoints();
	PGDataPointIndices *pgd_reduced = reduced->getTrajectorySet()->getDataPoints();
	card32 numFrames = reduced->getTrajectorySet()->getNumFrames();
	pAssert(numFrames == fullResolution->getTrajectorySet()->getNumFrames());
	size_t numClusters = clusters.size();
	for (unsigned c=0; c<numClusters; c++) {
		size_t numPointsInCluster = clusters[c].size();
		for (unsigned ci=0; ci<numPointsInCluster; ci++) {
			for (card32 f=0; f<numFrames; f++) {
				pgd_reduced->addIndices(f, c, pgd_full->getIndices(f,	clusters[c][ci]) );
			}
		}		
	}
}


PatchGraphTopologicalResampling::~PatchGraphTopologicalResampling() {}


struct edge_TG{
public:
	float32	weight;
	card32	indexMin;
	card32	indexMax;
public:
	edge_TG(){}
	inline edge_TG(card32 v1, card32 v2, float32 w)
	{
		pAssert(v1 != v2);
		indexMin		= min(v1, v2);
		indexMax	= max(v1, v2);
		this->weight = w;
	}
	inline const bool operator < (const edge_TG &e) const
	{
		if(weight < e.weight)
		{
			return true;
		}
		else if(weight > e.weight)
		{
			return false;
		}
		else if(indexMax < e.indexMax)
		{
			return true;
		}
		else if(indexMax > e.indexMax)
		{
			return false;
		}
		else
		{
			return indexMin < e.indexMin;
		}
	}
};

static void edge_TG_Collapse(vector<vector<Vector3f> > *m_position,
							 vector<vector<Vector3f> > *m_normal,
							 vector<set<card32> > *edges,
							 const		card32	width,
							 card32 v1, card32 v2)
{
	pAssert(edges->size() == m_position->size()
		&& m_normal->size()== m_position->size());

	size_t	num1	= (*edges)[v1].size();
	size_t	num2	= (*edges)[v2].size();
	//first compute the geometry of the newvertex
	vector<Vector3f>		buf_pos;				buf_pos.resize(width);
	vector<Vector3f>		buf_nor;				buf_nor.resize(width);
	
	set<card32>::iterator	iter_set;
	/// Compute the geometry of the newvertex
	for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
	{
		buf_pos[frameIndex] = ((*m_position)[v1][frameIndex] + (*m_position)[v2][frameIndex])/2;
		Vector3f	new_nor =  ((*m_normal)[v1][frameIndex] + (*m_normal)[v2][frameIndex])/2;
		new_nor.normalize();
		buf_nor[frameIndex] = new_nor;
	}
	
	set<card32>					neighborsOfNewVertex;
	card32						newVertex_id = (card32)edges->size();
	for(iter_set = (*edges)[v1].begin();iter_set !=(*edges)[v1].end(); iter_set++)
	{
		if(*iter_set != v2)
		{
			neighborsOfNewVertex.insert(*iter_set);
			(*edges)[*iter_set].erase((*edges)[*iter_set].find(v1));
			(*edges)[*iter_set].insert(newVertex_id);
		}
	}
	(*edges)[v1].clear();
	for(iter_set = (*edges)[v2].begin();	iter_set !=(*edges)[v2].end();	iter_set++)
	{
		if(*iter_set != v1)
		{
			neighborsOfNewVertex.insert(*iter_set);
			(*edges)[*iter_set].erase((*edges)[*iter_set].find(v2));
			(*edges)[*iter_set].insert(newVertex_id);
		}
	}
	(*edges)[v2].clear();

	///Update
	(*m_position).push_back(buf_pos);
	(*m_normal).push_back(buf_nor);
	edges->push_back(neighborsOfNewVertex);
	//remove old topology
}

static void edge_TG_Split(vector<vector<Vector3f> > *m_position,
						 vector<vector<Vector3f> > *m_normal,
						 vector<set<card32> > *edges, 
						 const	card32	width,
						 card32 v1, card32 v2)
{
	pAssert(edges->size() == m_position->size()
		&& m_normal->size()== m_position->size());

	//first compute the geometry of the newvertex
	vector<Vector3f>		buf_pos;		buf_pos.resize(width);
	vector<Vector3f>		buf_nor;		buf_nor.resize(width);
	for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
	{
		buf_pos[frameIndex] = ((*m_position)[v1][frameIndex] + (*m_position)[v2][frameIndex])/2;
		Vector3f	new_nor =  ((*m_normal)[v1][frameIndex] + (*m_normal)[v2][frameIndex])/2;
		new_nor.normalize();
		buf_nor[frameIndex] = new_nor;
	}
	//
	set<card32>		neighborsOfNewVertex;
	neighborsOfNewVertex.insert(v1);
	neighborsOfNewVertex.insert(v2);
	set<card32>::iterator	iter_set;
	for(iter_set = (*edges)[v1].begin();
		iter_set !=(*edges)[v1].end();
		iter_set++)
	{
		if(*iter_set != v2)
		{
			set<card32>::iterator	iter_set1 = (*edges)[*iter_set].find(v2);
			if(iter_set1 != (*edges)[*iter_set].end())
			{
				neighborsOfNewVertex.insert(*iter_set);
			}
		}
	}
	if(neighborsOfNewVertex.size() >= 3)
	{
		mpcard newVertex_id = (mpcard)edges->size();
		iter_set = (*edges)[v1].find(v2);	
		(*edges)[v1].erase(iter_set);	
		iter_set = (*edges)[v2].find(v1);	
		(*edges)[v2].erase(iter_set);		
		(*m_position).push_back(buf_pos);
		(*m_normal).push_back(buf_nor);
		edges->push_back(neighborsOfNewVertex);
		for(iter_set = neighborsOfNewVertex.begin();
			iter_set != neighborsOfNewVertex.end();
			iter_set++)
		{
			(*edges)[*iter_set].insert(newVertex_id);
		}
	}
}

static  void	write_Buffers(const PatchGraph *fullResolution,
									  vector<vector<Vector3f> > *m_position,
									  vector<vector<Vector3f> > *m_normal,
									  vector<set<card32> > *edges,
									  bool		keepOldTopology)
{
	const	PointSet		*ps = fullResolution->getTrajectorySet()->getTrajectories();
	const	InCorePCTopologyGraph	*topology = fullResolution->getTopology();
	
	card32			width			= ps->getWidth(),
							height			= ps->getHeight();

	AAT					POSITION	= ps->getAAT("position");
	AAT					NORMAL		= ps->getAAT("normal");
	AAT					REF_U			= ps->getAAT("tangent_u");

	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		vector<Vector3f>		trajectory_pos;
		vector<Vector3f>		trajectory_nor;
		for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
		{
			Vector3f		pos = ps->get3f(frameIndex, surfelIndex, POSITION);
			Vector3f		nor = ps->get3f(frameIndex, surfelIndex, NORMAL);
			trajectory_pos.push_back(pos);
			trajectory_nor.push_back(nor);
		}
		(*m_position).push_back(trajectory_pos);
		(*m_normal).push_back(trajectory_nor);
		if(keepOldTopology)
		{
			card32			numOfEntries = topology->getNumAdjacentVertices(surfelIndex);
			set<card32>	IndicesOfNeighbors;
			for(card32 ni = 0; ni < numOfEntries; ni++)
			{
					IndicesOfNeighbors.insert(topology->getAdjacentVertex(surfelIndex, ni));
			}
			edges->push_back(IndicesOfNeighbors);
		}
	}
}

static	PatchGraph*	generate_New_PatchGraph(const PatchGraph *fullResolution,
													vector<vector<Vector3f> > *m_position,
													vector<vector<Vector3f> > *m_normal,
													vector<set<card32> > *edges)
{
	vector<int32>		index_transfer;		index_transfer.resize(edges->size());
	card32					height = 0;
	for(card32 ni = 0; ni < edges->size(); ni++)
	{
		if((*edges)[ni].size() > 0)
		{
			index_transfer[ni] = height;
			height++;
		}
		else
		{
			index_transfer[ni] = -1;
		}
	}

	TrajectorySet* ts = new TrajectorySet();
	PointSet* trajectories = new PointSet();
	const	PointSet		*ps = fullResolution->getTrajectorySet()->getTrajectories();
	trajectories->clearAndSetup(ps->getWidth(), height, ps->getDescr() );
	ts->setStartFrameIndex(fullResolution->getStartFrameIndex());
	ts->clearAndSetup( trajectories );
	InCorePCTopologyGraph* tg = new InCorePCTopologyGraph();
	tg->clearAndSetup(fullResolution->getTopology()->getEdges()->getDescr() );
	PatchGraph* newPG = new PatchGraph();
	newPG->clearAndSetup( ts, tg, false );
	
	newPG->createPatchAttribute(mVAD("topology_depth", 1,	VAD::DATA_FORMAT_INT32));
	trajectories = newPG->getTrajectorySet()->getTrajectories();
	//copy positions, normals and ref_u
	AAT					POSITION	= trajectories->getAAT("position"),
							NORMAL		= trajectories->getAAT("normal"),
							REF_U			= trajectories->getAAT("tangent_u"),
							TOP_DEPTH=  trajectories->getAAT("topology_depth");
	
	for(card32 frameIndex = 0; frameIndex < ps->getWidth(); frameIndex++)
	{
		vector<card32> indices;
		
		for(card32 surfelIndex = 0; surfelIndex < ps->getHeight(); surfelIndex++)
		{
			vector<card32>	localIndices = fullResolution->getTrajectorySet()->getDataPoints()->getIndices(frameIndex, surfelIndex);
			for(card32 i = 0; i < localIndices.size(); i++)
			{
				indices.push_back(localIndices[i]);
			}
		}
		ts->getDataPoints()->setIndices(frameIndex, 0, indices);
	}

	for(card32 ni = 0; ni < edges->size(); ni++)
	{
		if(index_transfer[ni] < 0)
			continue;
		
		card32 surfelIndex = index_transfer[ni];
		for(card32 frameIndex = 0; frameIndex < ps->getWidth(); frameIndex++)
		{
			trajectories->set3f(frameIndex, surfelIndex, POSITION, (*m_position)[ni][frameIndex]);	
			Vector3f		nik = (*m_normal)[ni][frameIndex];
			trajectories->set3f(frameIndex, surfelIndex, NORMAL, nik);
			Vector3f	Eik1;
			if(fabs(nik[0]) > fabs(nik[1]))
			{
				if(fabs(nik[0]) > fabs(nik[2])){	Eik1[0] = 1.0; Eik1[1] = 0.0; Eik1[2] = 0.0;}
				else	{	Eik1[0] = 0.0; Eik1[1] = 0.0; Eik1[2] = 1.0;}
			}
			else
			{
				if(fabs(nik[1]) > fabs(nik[2])){Eik1[0] = 0.0; Eik1[1] = 1.0; Eik1[2] = 0.0;}
				else{Eik1[0] = 0.0; Eik1[1] = 0.0; Eik1[2] = 1.0;}
			}
			Eik1 -= nik*(Eik1*nik);	Eik1.normalize();
			trajectories->set3f(frameIndex, surfelIndex, REF_U, Eik1);
			trajectories->set<int32>(frameIndex, surfelIndex, TOP_DEPTH, 0);
		}
		//copy topology
		for(set<card32>::iterator	iter_card32 = (*edges)[ni].begin();
			iter_card32 != (*edges)[ni].end(); iter_card32++)
		{
			if(index_transfer[*iter_card32] >= 0)
			{
				card32 surfelIndex1 = index_transfer[*iter_card32];
				if(surfelIndex < surfelIndex1)
					tg->addEdge(surfelIndex, surfelIndex1);
				else
					tg->addEdge(surfelIndex1, surfelIndex);
			}
		}
	}
	
	return newPG;
}

static	void	resampling_Initialization(PatchGraph* fullResolution,
										 vector<vector<Vector3f> > *m_position,
										 vector<vector<Vector3f> > *m_normal,
										 vector<set<card32> > *edges,
										 const	card32	maxNumOfNeighbors, 
										 const	card32	minNumOfNeighbors,
										 const float32	high_sqrDist,
										 const float32	maxStrech,	
										 const float32  pruneAngle1,
										 const float32	pruneangle2)
{
	PointSet								*ps = fullResolution->getTrajectorySet()->getTrajectories();

	AAT										POSITION		= ps->getAAT("position");

	card32									width = ps->getWidth(),
											height = ps->getHeight();

	edges->resize(height);

	PointSetKNNQuery **pg_knn = new PointSetKNNQuery*[width];
	PointSet **pg_knn_ps = new PointSet*[width];
	
	card32 neighbor_id;
	for (card32 frameIndex=0; frameIndex < width; frameIndex++) 
	{
		AttachedIndexedOctree *octree = fullResolution->getTrajectorySet()->fetchOctree(frameIndex);
		pg_knn_ps[frameIndex] = fullResolution->getTrajectorySet()->copySingleFrame(frameIndex);
		pg_knn[frameIndex] = new PointSetKNNQuery(pg_knn_ps[frameIndex], octree, NULL_VECTOR3F, POSITION);

		for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
		{
			pg_knn[frameIndex]->setTargetPoint((*m_position)[surfelIndex][frameIndex]);
			bool found;
			card32 numOfNeighbors = 0;
			do 
			{
				pg_knn[frameIndex]->getNextNearestPointIndex(neighbor_id, found);
				if(neighbor_id == surfelIndex)
					continue;

				(*edges)[surfelIndex].insert(neighbor_id);
				(*edges)[neighbor_id].insert(surfelIndex);
				numOfNeighbors++;
			} while(numOfNeighbors < maxNumOfNeighbors);
		}
	}

	set<edge_TG>				edge_queue;
	set<card32>::iterator	iter_set;
	for(card32  surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		for(iter_set = (*edges)[surfelIndex].begin(); iter_set != (*edges)[surfelIndex].end(); iter_set++)
		{
			card32 neighbor_id = *iter_set;
			bool	good_edge = true;
			float	asqv = 0.0, min_sqr = 1e10, max_sqr = -1e10;
			for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
			{
				Vector3f rijk = (*m_position)[surfelIndex][frameIndex] - (*m_position)[neighbor_id][frameIndex];
				asqv += rijk.getSqrNorm();
				if(rijk.getSqrNorm() > max_sqr)
				{
					max_sqr = rijk.getSqrNorm();
				}
				if(rijk.getSqrNorm() < min_sqr)
				{
					min_sqr = rijk.getSqrNorm();
				}
			}
			if(max_sqr < maxStrech*maxStrech*min_sqr)
			{
				edge_queue.insert(edge_TG(surfelIndex, neighbor_id, asqv/width));
			}
		}
	}
	edges->clear();
	edges->resize(height);
	do 
	{
		set<edge_TG>::iterator	iter_edge= edge_queue.begin();
		card32	surfelIndex  = iter_edge->indexMin;
		card32	neighbor_id = iter_edge->indexMax;
		float32	asqv = iter_edge->weight;
		edge_queue.erase(iter_edge);

		vector<card32>		neighborsOfNewVertex;
		size_t				num1 = (*edges)[surfelIndex].size(),
							num2 = (*edges)[neighbor_id].size();

		if(asqv > high_sqrDist)
		{
			if(min(num1, num2) < 4 || max(num1, num2) < 6)
			{
				(*edges)[surfelIndex].insert(neighbor_id);
				(*edges)[neighbor_id].insert(surfelIndex);
			}
			continue;
		}
		bool good_edge = true;
		if(good_edge)
		{
			(*edges)[surfelIndex].insert(neighbor_id);
			(*edges)[neighbor_id].insert(surfelIndex);
		}
	} while(edge_queue.size()>0);
	edge_queue.clear();
	for (card32 frameIndex = 0; frameIndex < width; frameIndex++) 
	{
		delete pg_knn[frameIndex];
		delete pg_knn_ps[frameIndex];
	}
	delete []pg_knn;
	delete []pg_knn_ps;

	//remove those points that have less than 2 neighboring points
	if(minNumOfNeighbors > 1)
	{
		vector<bool>	m_flag;	m_flag.resize(edges->size());
		for(card32 surfelIndex = 0; surfelIndex < edges->size(); surfelIndex++)
		{
			if((*edges)[surfelIndex].size() < 2)
			{
				m_flag[surfelIndex] = false;
			}
			else
			{
				m_flag[surfelIndex] = true;
			}
		}
		for(card32 surfelIndex = 0; surfelIndex < edges->size(); surfelIndex++)
		{
			if(m_flag[surfelIndex] == false)
			{
				for(iter_set = (*edges)[surfelIndex].begin();
					iter_set != (*edges)[surfelIndex].end();
					iter_set++)
				{
					set<card32>::iterator	iter_set1 = (*edges)[*iter_set].find(surfelIndex);
					(*edges)[*iter_set].erase(iter_set1);
				}
				(*edges)[surfelIndex].clear();
			}
		}
	}
}

static void	resampling_Edge_Collapse(vector<vector<Vector3f> >	*m_position,
												 vector<vector<Vector3f> > *m_normal,
												 vector<set<card32> >	*edges,
												 const	card32	width,
												 const	float32	low_sqrDis,
												 const float32	high_sqrDis)
{
	set<edge_TG>	edge_queue;
	set<card32>::iterator	iter_set;
	//compute edge_queue;
	size_t	height = edges->size();
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		for(iter_set = (*edges)[surfelIndex].begin();	iter_set != (*edges)[surfelIndex].end();	iter_set++)
		{
			if(*iter_set > surfelIndex)
			{
				card32 neighbor_id = *iter_set;
				float32	asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f	rijk = (*m_position)[surfelIndex][frameIndex] - (*m_position)[neighbor_id][frameIndex];
					asqv += rijk.getSqrNorm();
				}
				asqv /= width;
				edge_queue.insert(edge_TG(surfelIndex, neighbor_id, asqv));
			}
		}
	}
	do 
	{
		set<edge_TG>::iterator	iter_edge= edge_queue.begin();
		card32	surfelIndex  = iter_edge->indexMin;
		card32	neighbor_id = iter_edge->indexMax;
		float32	asqv = iter_edge->weight;
		if(asqv > low_sqrDis)
			break;

		edge_queue.erase(iter_edge);

		bool boundary_edge = false;
		for(iter_set = (*edges)[surfelIndex].begin();
			iter_set != (*edges)[surfelIndex].end();
			iter_set++)
		{
			if(*iter_set != neighbor_id)
			{
				set<card32>::iterator	iter_set1 = (*edges)[neighbor_id].find(*iter_set);
				if(iter_set1 != (*edges)[neighbor_id].end())
				{
					if((*edges)[*iter_set].size() == 2)
					{
						boundary_edge = true;
						break;
					}
				}
			}
		}
		if(boundary_edge)
			continue;

		set<card32>	neighborsOfNewVertex;
		for(iter_set = (*edges)[surfelIndex].begin();
			iter_set != (*edges)[surfelIndex].end();
			iter_set++)
		{
			if(*iter_set != neighbor_id)
			{
				neighborsOfNewVertex.insert(*iter_set);
			}
		}
		for(iter_set = (*edges)[neighbor_id].begin();
			iter_set != (*edges)[neighbor_id].end();
			iter_set++)
		{
			if(*iter_set != surfelIndex)
			{
				neighborsOfNewVertex.insert(*iter_set);
			}
		}

		size_t num = neighborsOfNewVertex.size(); card32 ni = 0;
		vector<float32>	asqv_buf_v;	asqv_buf_v.resize(num);

		for(iter_set = neighborsOfNewVertex.begin();iter_set != neighborsOfNewVertex.end();iter_set++)
		{
			for(card32 frameIndex = 0 ; frameIndex < width; frameIndex++)
			{
				Vector3f rijk = (*m_position)[*iter_set][frameIndex] - 
					((*m_position)[surfelIndex][frameIndex] +(*m_position)[neighbor_id][frameIndex])/2;
				asqv_buf_v[ni] += rijk.getSqrNorm();
			}
			ni++;
		}

		bool		need_collapse = true;
		for(ni = 0; ni < num; ni++)
		{
			if(asqv_buf_v[ni] > width*high_sqrDis)
			{
				need_collapse = false;
				break;
			}
		}
		if(!need_collapse)
			continue;

		for(iter_set = (*edges)[surfelIndex].begin();	iter_set != (*edges)[surfelIndex].end();	iter_set++)
		{
			if(*iter_set != neighbor_id)
			{
				float32 asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f rijk = (*m_position)[surfelIndex][frameIndex] - (*m_position)[*iter_set][frameIndex];
					asqv += rijk.getSqrNorm();
				}
				set<edge_TG>::iterator	pos = edge_queue.find(edge_TG(surfelIndex, *iter_set, asqv/width));
				if(pos != edge_queue.end())
					edge_queue.erase(pos);
			}
		}
		for(iter_set = (*edges)[neighbor_id].begin();	iter_set != (*edges)[neighbor_id].end();	iter_set++)
		{
			if(*iter_set != surfelIndex)
			{
				float32 asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f rijk = (*m_position)[neighbor_id][frameIndex] - (*m_position)[*iter_set][frameIndex];
					asqv += rijk.getSqrNorm();
				}
				set<edge_TG>::iterator	pos = edge_queue.find(edge_TG(neighbor_id, *iter_set, asqv/width));
				if(pos != edge_queue.end())
					edge_queue.erase(pos);
			}
		}
		//Now we could delete this edge now
		edge_TG_Collapse(m_position, m_normal, edges, width, surfelIndex, neighbor_id);
		card32	newVertex_id = (card32)edges->size()-1;
		for(iter_set = (*edges)[newVertex_id].begin();
			iter_set != (*edges)[newVertex_id].end();
			iter_set++)
		{
			if(*iter_set != newVertex_id)
			{
				float32 asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f rijk = (*m_position)[newVertex_id][frameIndex] - (*m_position)[*iter_set][frameIndex];
					asqv += rijk.getSqrNorm();
				}
				asqv /= width;
				edge_queue.insert(edge_TG(newVertex_id, *iter_set, asqv));
			}
		}
	} while(edge_queue.size() > 0);
	edge_queue.clear();
}


static void	resampling_Edge_Split(vector<vector<Vector3f> >	*m_position,
												 vector<vector<Vector3f> > *m_normal,
												 vector<set<card32> >	*edges,
												 const	card32	width,
												 const float32	high_sqrDis)
{
	
	set<edge_TG>	edge_queue;
	set<card32>::iterator	iter_set;
	//compute edge_queue;
	size_t	height = edges->size();
	for(card32 surfelIndex1 = 0; surfelIndex1 < height; surfelIndex1++)
	{
		for(iter_set = (*edges)[surfelIndex1].begin();	iter_set != (*edges)[surfelIndex1].end();	iter_set++)
		{
			if(*iter_set > surfelIndex1)
			{
				card32 neighbor_id = *iter_set;
				float32	asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f	rijk = (*m_position)[surfelIndex1][frameIndex] - (*m_position)[neighbor_id][frameIndex];
					asqv -= rijk.getSqrNorm();
				}
				asqv /= width;
				edge_queue.insert(edge_TG(surfelIndex1, neighbor_id, asqv));
			}
		}
	}
	do 
	{
		
		set<edge_TG>::iterator	iter_edge= edge_queue.begin();
		card32	surfelIndex  = iter_edge->indexMin;
		card32	neighbor_id = iter_edge->indexMax;
		float32	asqv = -iter_edge->weight;
		if(asqv < high_sqrDis)
			break;

		edge_queue.erase(iter_edge);
		//Now we could delete this edge now
		edge_TG_Split(m_position, m_normal, edges, width, surfelIndex, neighbor_id);
	
		card32	newVertex_id = (card32)edges->size()-1;
		for(iter_set = (*edges)[newVertex_id].begin();
			iter_set != (*edges)[newVertex_id].end();
			iter_set++)
		{
			if(*iter_set != newVertex_id)
			{
				float32 asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f rijk = (*m_position)[newVertex_id][frameIndex] - (*m_position)[*iter_set][frameIndex];
					asqv += rijk.getSqrNorm();
				}
				asqv /= width;
				edge_queue.insert(edge_TG(newVertex_id, *iter_set, -asqv));
			}
		}
	} while(edge_queue.size() > 0);
	edge_queue.clear();
}

static	void	balance_Topology(vector<vector<Vector3f> >	*m_position,
										  vector<vector<Vector3f> > *m_normal,
										  vector<set<card32> >	*edges,
										  const	card32	width)
{
	set<edge_TG>	edge_queue;
	set<card32>::iterator	iter_set;
	//compute edge_queue;
	size_t	height = edges->size();
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		for(iter_set = (*edges)[surfelIndex].begin();	iter_set != (*edges)[surfelIndex].end();	iter_set++)
		{
			if(*iter_set > surfelIndex)
			{
				card32 neighbor_id = *iter_set;
				float32	asqv = 0.0;
				for(card32 frameIndex = 0; frameIndex < width; frameIndex++)
				{
					Vector3f	rijk = (*m_position)[surfelIndex][frameIndex] - (*m_position)[neighbor_id][frameIndex];
					asqv -= rijk.getSqrNorm();
				}
				asqv /= width;
				edge_queue.insert(edge_TG(surfelIndex, neighbor_id, asqv));
			}
		}
	}
	edges->clear();
	edges->resize(m_position->size());
	/// PatchGraph
	do 
	{
		set<edge_TG>::iterator	iter_edge = edge_queue.begin();
		card32	surfelIndex		= iter_edge->indexMin;
		card32	neighbor_id	=	iter_edge->indexMax;
		edge_queue.erase(iter_edge);
		/// 
		size_t	num1 = (*edges)[surfelIndex].size();
		size_t	num2 = (*edges)[neighbor_id].size();
		if(min(num1, num2) < 2 || max(num1, num2) < 4)
		{
			(*edges)[surfelIndex].insert(neighbor_id);
			(*edges)[neighbor_id].insert(surfelIndex);
			continue;
		}
		/// check whether this edge is redundant
		vector<card32>	commonNeighbors;
		for(iter_set = (*edges)[surfelIndex].begin();
			iter_set != (*edges)[surfelIndex].end();
			iter_set++)
		{
			set<card32>::iterator	iter_set1 = (*edges)[neighbor_id].find(*iter_set);
			if(iter_set1 != (*edges)[neighbor_id].end())
			{
				commonNeighbors.push_back(*iter_set);
			}
		}
		bool good_edge = true;
		for(card32 ni = 0; ni < commonNeighbors.size(); ni++)
		{
			for(card32 ni1 = ni+1; ni1 < commonNeighbors.size(); ni1++)
			{
				set<card32>::iterator	iter_set1 = (*edges)[commonNeighbors[ni]].find(commonNeighbors[ni1]);
				if(iter_set1 != (*edges)[commonNeighbors[ni]].end())
				{
					good_edge = false;
					break;
				}
			}
			if(!good_edge)
			{
				break;
			}
		}
		if(good_edge)
		{
			(*edges)[surfelIndex].insert(neighbor_id);
			(*edges)[neighbor_id].insert(surfelIndex);
		}
	} while(edge_queue.size() > 0);
}

/// Peter's new code for topology and geometry resampling
PatchGraph *PatchGraphTopologicalResampling::fullResampling_Modified
(PatchGraph* fullResolution, 
 float32 maxAveragedDist,
 float32 maxStrech,
 const card32	averageNumberOfNeighbors,
 const card32	minNumberOfNeighbors,
 vector< vector<card32> > *clusters,
 vector<bool> *removable) 
{
	vector<vector<Vector3f>> m_position;
	vector<vector<Vector3f>> m_normal;
	vector<set<card32>>		edges;

	//Assign m_position, m_normal and edges respectively;
	write_Buffers(fullResolution, &m_position, &m_normal, &edges, false);

	resampling_Initialization(fullResolution, &m_position, &m_normal, &edges, averageNumberOfNeighbors, minNumberOfNeighbors,
		maxAveragedDist*maxAveragedDist, maxStrech, 0.2f, 1.2f);

	resampling_Edge_Collapse(&m_position, &m_normal, &edges,
		fullResolution->getTrajectorySet()->getTrajectories()->getWidth(), 
		0.13*maxAveragedDist*maxAveragedDist, 0.6*maxAveragedDist*maxAveragedDist);
	resampling_Edge_Split(&m_position, &m_normal, &edges, 
		fullResolution->getTrajectorySet()->getTrajectories()->getWidth(),
		0.4*maxAveragedDist*maxAveragedDist);
	resampling_Edge_Collapse(&m_position, &m_normal, &edges,
		fullResolution->getTrajectorySet()->getTrajectories()->getWidth(),
		0.1f*maxAveragedDist*maxAveragedDist, 
		0.6f*maxAveragedDist*maxAveragedDist);
//	balance_Topology(&m_position, &m_normal, &edges,
//		fullResolution->getTrajectorySet()->getTrajectories()->getWidth());
	return	generate_New_PatchGraph(fullResolution, &m_position, &m_normal, &edges);
	
}

static	void	topology_Filtering(vector<vector<Vector3f> >	*m_position,
											 vector<vector<Vector3f> > *m_normal,
											 vector<set<card32> >	*edges,
											 const	card32	width,
											 float32 angle_threshold1, float32 angle_threshold2)
{
	//remove bad edges and vertices
	set<card32>::iterator	iter_set, iter_set1;
	//add new edges to quads

/*	for(card32	surfelIndex = 0; surfelIndex < edges->size(); surfelIndex++)
	{
		for(iter_set = (*edges)[surfelIndex].begin();iter_set != (*edges)[surfelIndex].end();iter_set++)
		{
			card32	neighbor_id1 =  *iter_set;
			for(iter_set1 = iter_set; iter_set1 != (*edges)[surfelIndex].end(); iter_set1++)
			{
				card32	neighbor_id2 = *iter_set1;
				if(neighbor_id1 != neighbor_id2)
				{
					//check that they are connected
					set<card32>::iterator	iter_neighbor = (*edges)[neighbor_id1].find(neighbor_id2);
					if(iter_neighbor != )
				}
			}
		}
	}*/

	//down-sampling
	size_t	numOfVertices = edges->size();
	set<edge_TG>		edge_queue;
	for(card32	surfelIndex = 0; surfelIndex < edges->size(); surfelIndex++)
	{
		for(iter_set = (*edges)[surfelIndex].begin(); iter_set != (*edges)[surfelIndex].end(); iter_set++)
		{
			card32	neighbor_id = (*iter_set);
			if(surfelIndex < neighbor_id)
			{
				edge_queue.insert(edge_TG(surfelIndex, neighbor_id, 0.0f));
			}
		}
	}
	do 
	{
		set<edge_TG>::iterator	iter_edge = edge_queue.begin();
		card32	surfelIndex		= iter_edge->indexMin;
		card32	neighbor_id	= iter_edge->indexMax;
		edge_queue.erase(iter_edge);
		set<card32>		neighborsOfNewVertex;
		neighborsOfNewVertex.insert(surfelIndex);
		neighborsOfNewVertex.insert(neighbor_id);
		iter_set = (*edges)[surfelIndex].find(neighbor_id);	(*edges)[surfelIndex].erase(iter_set);
		iter_set = (*edges)[neighbor_id].find(surfelIndex);	(*edges)[neighbor_id].erase(iter_set);
		(*edges)[surfelIndex].insert((card32)edges->size());
		(*edges)[neighbor_id].insert((card32)edges->size());
		edges->push_back(neighborsOfNewVertex);

		vector<Vector3f>		buf_pos;		buf_pos.resize(width);
		vector<Vector3f>		buf_nor;		buf_nor.resize(width);
		for(card32 frameIndex= 0; frameIndex < width; frameIndex++)
		{
			buf_pos[frameIndex] = ((*m_position)[surfelIndex][frameIndex] + (*m_position)[neighbor_id][frameIndex])/2;
			buf_nor[frameIndex] = ((*m_normal)[surfelIndex][frameIndex] + (*m_normal)[neighbor_id][frameIndex])/2;
			buf_nor[frameIndex].normalize();
		}
		(*m_position).push_back(buf_pos);
		(*m_normal).push_back(buf_nor);
	} while(edge_queue.size() > 0);
	
	//remove the out-dated geometry from m_position and m_normal
	for(mpint surfelIndex = (mpint)edges->size()-1; surfelIndex >= 0; surfelIndex--)
	{
		if((*edges)[surfelIndex].size() == 0)
		{
			vector<vector<Vector3f>>::iterator	iter_array = m_position->begin() + surfelIndex;
			m_position->erase(iter_array);
			iter_array = m_normal->begin() + surfelIndex;
			m_normal->erase(iter_array);
		}
	}
}
PatchGraph	*PatchGraphTopologicalResampling::fullDownsampling(
	geoxl::PatchGraph *fullResolution, 
	geoxl::float32 maxDist, 
	const geoxl::card32 averageNumberOfNeighbors, 
	std::vector<vector<card32> > *clusters, 
	std::vector<bool> *removable)
{
	vector<vector<Vector3f>> m_position;
	vector<vector<Vector3f>> m_normal;
	vector<set<card32>>			 edges;

	//Assign m_position, m_normal and edges respectively;
	write_Buffers(fullResolution, &m_position, &m_normal, &edges, true);
	topology_Filtering(&m_position, &m_normal, &edges, 
		fullResolution->getTrajectorySet()->getTrajectories()->getWidth(),
		0.2f, 1.2f);

	PatchGraph	*geometrically_down_sampled = generate_New_PatchGraph(fullResolution, &m_position, &m_normal, &edges);
	m_position.clear(),	m_normal.clear();	edges.clear();

	PatchGraph	*topologically_down_sampled = fullResampling_Modified(geometrically_down_sampled, 2*maxDist, 2.0, averageNumberOfNeighbors, 1);

	delete	geometrically_down_sampled;

	return	topologically_down_sampled;
}

PatchGraph	*PatchGraphTopologicalResampling::fullTimesubdivision(
	PatchGraph *fullResolution,
	const	card32	numOfInbetweenFrames)
{
	if(numOfInbetweenFrames == 0)
		return NULL;
	if(numOfInbetweenFrames >= 10)
		return NULL;

	PointSet								*ps	= fullResolution->getTrajectorySet()->getTrajectories();
	card32							width	= ps->getWidth(),
											height	= ps->getHeight();

	card32	widthOfNewPatchGraph = width*(1+numOfInbetweenFrames);
	card32	heightOfNewPatchGraph = height;

	AAT						POSITION	= ps->getAAT("position");
	AAT						NORMAL		=	ps->getAAT("normal");
	AAT						REF_U			= ps->getAAT("tangent_u");
	AAT						ROTATION	= ps->getAAT("rotation_matrix");

	TrajectorySet				*ts	= new TrajectorySet();
	PointSet			*trajectories	= new PointSet();

	trajectories->clearAndSetup(widthOfNewPatchGraph, heightOfNewPatchGraph, ps->getDescr());
	ts->setStartFrameIndex(fullResolution->getStartFrameIndex());
	ts->clearAndSetup( trajectories );
	InCorePCTopologyGraph* tg = new InCorePCTopologyGraph();
	tg->clearAndSetup(fullResolution->getTopology()->getEdges()->getDescr() );
	PatchGraph* newPG = new PatchGraph();
	newPG->clearAndSetup( ts, tg, false );

	newPG->createPatchAttribute(mVAD("topology_depth", 1,	VAD::DATA_FORMAT_INT32));
	
	PointSet	*newPS = newPG->getTrajectorySet()->getTrajectories();

	card32	newFrameIndex = 0;

	for(card32 oldFrameIndex  = 0; oldFrameIndex < width; oldFrameIndex++)
	{
		for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
		{
			newFrameIndex = oldFrameIndex*(1+numOfInbetweenFrames);
			Vector3f		pik		= ps->get3f(oldFrameIndex, surfelIndex, POSITION);
			Vector3f		nik		= ps->get3f(oldFrameIndex, surfelIndex, NORMAL);
			Vector3f		Eik1		= ps->get3f(oldFrameIndex, surfelIndex, REF_U);
			char			*dataS	= ps->getDataPointer(oldFrameIndex, surfelIndex);
			float32	*quatS	= (float*)(dataS+ROTATION.getOffset());

			newPS->set3f(newFrameIndex, surfelIndex, POSITION, pik);
			newPS->set3f(newFrameIndex, surfelIndex, NORMAL, nik);
			newPS->set3f(newFrameIndex, surfelIndex, REF_U, Eik1);

			Vector3f		axis;		
			axis[0] = quatS[1];			axis[1] = quatS[2];			axis[2] = quatS[3];			axis.normalize();		
			float32	halfangle		= acos(quatS[0]);		
			float32	cosa = cos(halfangle/(1 + numOfInbetweenFrames));
			float32	sina = sin(halfangle/(1 + numOfInbetweenFrames));

			char			*dataC	= ps->getDataPointer(newFrameIndex, surfelIndex);
			float32	*quatC	= (float*)(dataS+ROTATION.getOffset());
			quatC[0] = cosa;	
			for(card32 k = 0; k < 3; k++)
			{
				quatC[k + 1] = axis[k]*sina;
			}
			if(oldFrameIndex < width - 1)
			{
				Vector3f		pjk		= ps->get3f(oldFrameIndex, surfelIndex, POSITION);
				Vector3f		njk		= ps->get3f(oldFrameIndex, surfelIndex, NORMAL);

				for(card32 in_between_index = 0; in_between_index < numOfInbetweenFrames; in_between_index++)
				{
					newFrameIndex		= oldFrameIndex*(1+numOfInbetweenFrames) + 1 + in_between_index;
					/// Let us set rotation matrix first
					char			*dataN	= ps->getDataPointer(newFrameIndex, surfelIndex);
					float32	*quatN	= (float*)(dataS+ROTATION.getOffset());
					for(card32 k = 0; k < 4; k++)
					{
						quatN[k] = quatC[k];
					}
					float32	t = (1.0f + in_between_index)/(1.0f + numOfInbetweenFrames);
					Vector3f		pos	= pik*(1-t)		+	pjk*t;
					Vector3f		nor	= njk*(1-t)	+	njk*t;
					nor.normalize();
					ps->set3f(newFrameIndex, surfelIndex, POSITION, pos);
					ps->set3f(newFrameIndex, surfelIndex, NORMAL, nor);
					Vector3f	E1;
					if(fabs(nor[0]) > fabs(nor[1]))
					{
						if(fabs(nor[0]) > fabs(nor[2]))
						{
							E1[0] = 1.0; E1[1] = 0.0; E1[2] = 0.0;
						}
						else
						{
							E1[0] = 0.0; E1[1] = 0.0; E1[2] = 1.0;
						}
					}
					else
					{
						if(fabs(nor[1]) > fabs(nor[2]))
						{
							E1[0] = 0.0; E1[1] = 1.0; E1[2] = 0.0;
						}
						else
						{
							E1[0] = 0.0; E1[1] = 0.0; E1[2] = 1.0;
						}
					}
					E1 -= nor*(E1*nor);	

					E1.normalize();
					ps->set3f(newFrameIndex, surfelIndex, REF_U, E1);
				}
			}
		}
	}
	return newPG;
}



