//---------------------------------------------------------------------------
#ifndef PatchGraphTopologicalResamplingH
#define PatchGraphTopologicalResamplingH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PointFlags.h"
//---------------------------------------------------------------------------

namespace geoxl {

class PatchGraph;
class PGMultiResolutionLevel;
class InCorePCTopologyGraph;
class PGDataPointIndices;



/**
	same as patch graph resampling, but works on topological neighbors only
	addtionally, it can create a multi-resolution indexing pyramid (see PGMultiResolution.h)
	
	// xxx todo: datapoints are not correctly resampled
	// xxx everything is untested!
*/
class ANIMATIONRECONSTRUCTION_API PatchGraphTopologicalResampling : public Persistent {
	GEOXL_ABSTRACT_CLASS( PatchGraphTopologicalResampling )
 private:
   struct Edge {
      card32 v1;
      card32 v2;
      float32 w;
      inline bool operator<(const Edge &e) const {return w<e.w;}
   };
	bool useAverageDistance;
	card32 singleFrame;
 
   inline float32 compDistance(const PointSet *ps, const PointSet *urshape, card32 p1, card32 p2, const AAT &posAAT);
	void collectTopEpsNeighbors(const PointSet *ps, const PointSet *urshape, const InCorePCTopologyGraph *tg, const card32 patchIndex,
									 	 const float32 radius, const AAT &posAAT, vector<card32> &result);

 public:
	PatchGraphTopologicalResampling();
	
	/// change mode of operation: average/single frame distance. frame can also refer to the urshape.
	void setup(bool useAverageDistance = true, card32 frame = 0);

	/// all in one method: resamples points and resamples topology and data point association
   PatchGraph *fullResampling(const PatchGraph* fullResolution, float32 maxDist,
                              vector< vector<card32> > *clusters = NULL,
                              vector<bool> *removable = NULL);

   /// create a new patch graph with fewer patches
   ///  * maxDistance is the maximum Euklidian distance between topologically neighboring points
   ///  * original indices:
   ///    The new patch graph will contain a channel named "originalIndex"
   ///    after resampling, this channel provides the index of each patch in "fullResolution"
   ///    this can be used to compute a mapping between simplified patch graphs and the original
   ///  * topology: topology and data points will not be resampled
   ///    use topologicalClustering or fullResampling to create resampled topology
   /// *  if parameter removable is not NULL, only those points with removable[pointindex] == true 
   ///    will be considered for removal
   PatchGraph *resamplePatches(const PatchGraph* fullResolution, float32 maxDist, vector<bool> *removable = NULL);

   /// given two resolutions, this method recreates the topology and resamples the data points
   /// by region growing on the surface. if "clusters" is not NULL, the method will store the
   /// indices of patches subsumed by the resampled patches in this container (i.e. store 
   /// full resolution poinst at the reduces points (at their indices) which are closest in 
   /// topological distance.
   void topologicalClustering(PatchGraph *reduced, const PatchGraph *fullResolution,
                              vector< vector<card32> > *clusters = NULL);
	
	/// assign data points to reduced points according to the cluster analysis
	/// (i.e.: data points in each cluster go to the new point)
	void reassignDataPoints(PatchGraph *reduced, const PatchGraph *fullResolution,
                           const vector< vector<card32> > &clusters);

	/// computes interpolation weights using Gaussian radial basis functions (topological distance)
	/// output is a vector of lists of pairs (reducedIndex, weight); the primary index is the full resolution index
	void computeInterpolationMap(
		const PatchGraph *reduced, const PatchGraph *fullResolution,
		const vector< vector<card32> > &clusters, const float32 maxDist,
		const AAT &reducedPosAAT,
		const AAT &fullResPosAAT,
		const AAT &reducedOrigIndexAAT,
		vector< vector< pair<card32, float32> > > &weights
	);

	/// all in one method: resamples points and resamples topology and data point association
	PatchGraph *fullResampling_Modified(PatchGraph* fullResolution, 
		float32	maxAveragedDist,
		float32 maxStrech, 
		const card32	averageNumberOfNeighbors,
		const card32	minNumberOfNeighbors,
		vector< vector<card32> > *clusters = NULL,
		vector<bool> *removable = NULL);

	/// a down-sampler method: increase the number of vertices by a factor of user-input
	PatchGraph *fullDownsampling(PatchGraph *fullResolution, float32 maxDist, const card32 averageNumberOfNeighbors,
		vector<vector<card32> >	*clusters = NULL,
		vector<bool>	 *removable = NULL);

	/// a subdivision method to add in-between frames
	PatchGraph *fullTimesubdivision(PatchGraph *fullResolution, 
		const card32 numOfInbetweenFrames);

	~PatchGraphTopologicalResampling();
};


}

#endif