//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphTriangleMesh.h"
#include "ClassMethods.h"
#include "SGObjectNode.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "BSROptPotential.h"
#include "BSRFitting.h"
#include "StandardChannels.h"
#include "PointSetKNNQuery.h"
#include "Monomials2ndOrder.h"
#include "SGSimpleAnimationNode.h"
#include "PointSet_templCode.h"
#include "StandardSceneGraphIterator.h"
#include "TriangleMeshTools.h"
#include "SeparatorClassProperty.h"
#include "ProgressWindow.h"
#include "PointSetKNNQuery.h"
#include "TGNeighborhoodKNNSearch.h" 
#include "GLShaderMaterial.h"
#include "TGNeighborhoodRangeSearch.h"
#include "StringHelper.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

string getExtrapolateDescr( const card32 i )
{
	if ( i == EXTRAPOLATE_COLOR_DATA ){ 
		return "From data";
	} else if ( i == EXTRAPOLATE_COLOR_REFERENCE ){
		return "From reference";
	} else if ( i == EXTRAPOLATE_COLOR_SET ){
		return "Reset color";
	} else {
		return "Undefined";
	}							 
}

string getExtrapolatorModeDescr( const card32 index )
{
	if ( index == EXTRAPOLATE_MODE_PHILIPP ){
		return "Philipp";
	} else if ( index == EXTRAPOLATE_MODE_MICHAEL ){
		return "Michael";
	} else {
		return "Undefined";
	}
}

string getMeshingModeDescr( const card32 index )
{
	if ( index == MESHING_MODE_PHILIPP ){
		return "Philipp";
	} else if ( index == MESHING_MODE_MICHAEL ){
		return "Michael";
	} else {
		return "Undefined";
	}
}

IMPLEMENT_GEOXL_CLASS( PatchGraphTriangleMesh, 12 )
{
	BEGIN_CLASS_INIT( PatchGraphTriangleMesh )
	ADD_SEPARATOR( "Parameters" )
	ADD_CARD32_PROP( frameIndex, 0 )
	ADD_CARD32_PROP( patchGraphIndex, 0 )
	ADD_CARD32_PROP( resolution, 2 )
	ADD_FLOAT32_PROP( evalSigma, 8 )
	ADD_FLOAT32_PROP( colorSigma, 9 )
	ADD_FLOAT32_PROP( emptyCell, 5 )
	ADD_CARD32_PROP( numSmoothings, 9 )
	ADD_CARD32_PROP_UPDATE_DESCR( extrapolatorMode, 12, NULL, getExtrapolatorModeDescr, 2 )
	ADD_CARD32_PROP_UPDATE_DESCR( meshingMode, 12, NULL, getMeshingModeDescr, 2 )
	ADD_SEPARATOR( "Implicit Mesher" )
	ADD_BOOLEAN_PROP( useImplicitMesher, 8 );
	ADD_OBJECT_PROP( implicitMesher, 8, ImplicitPatchGraphMeshing::getClass(), true );
	ADD_SEPARATOR( "Color" ) 
	ADD_BOOLEAN_PROP( setMaterialIndex, 3 )
	ADD_CARD32_PROP( materialIndex, 11 )
	ADD_BOOLEAN_PROP( colorFromData, 7 )
	ADD_VECTOR3F_PROP( meshColor, 7 )
	ADD_SEPARATOR( "Extra-/Interpolation" )
	ADD_CARD32_PROP( interplatedFrames, 7 )
	ADD_CARD32_PROP_UPDATE_DESCR( extrapolateColorSource, 7, NULL, getExtrapolateDescr, 3 )
	ADD_SEPARATOR( "Border detection" )
	ADD_CARD32_PROP( borderNumNeighborDirs, 7 )
	ADD_BOOLEAN_PROP( pruneSmallBorders, 7 )
	ADD_CARD32_PROP( minBorderSize, 7 )
	ADD_BOOLEAN_PROP( pruneBorderDeadEnds, 7 )
	ADD_CARD32_PROP( borderTopologyDepth, 7 )
	ADD_SEPARATOR( "Rendering" )
	ADD_BOOLEAN_PROP( showBBox, 5 )
	ADD_BOOLEAN_PROP( showGrid, 5 )
	ADD_BOOLEAN_PROP( showBorders, 7 )
	ADD_NOARGS_METHOD( computeRefAndHF )
	ADD_NOARGS_METHOD( createMesh )
	ADD_NOARGS_METHOD( extrapolate )
	ADD_NOARGS_METHOD( smoothNormals )
	ADD_NOARGS_METHOD( detectBorderPoints )
	ADD_NOARGS_METHOD( clipAtBorder )
}

/** 
 * Constructor
 */
PatchGraphTriangleMesh::PatchGraphTriangleMesh()
{	
	// Parameters
	frameIndex = 0;
	patchGraphIndex = 0;
	resolution = 50;
	emptyCell = 3.0f;
	evalSigma = 1.0f;
	colorSigma = 1.5f;
	individualMeshes = false;
	numSmoothings = 0;
	extrapolatorMode = EXTRAPOLATE_MODE_MICHAEL;
	meshingMode = MESHING_MODE_MICHAEL;

	// Implicit mesher
	implicitMesher = new ImplicitPatchGraphMeshing();
	useImplicitMesher = false;

	// Color
	setMaterialIndex = true;
	materialIndex = 4;
	colorFromData = false;
	extrapolateColorSource = EXTRAPOLATE_COLOR_SET;
	meshColor = makeVector3f( 0.4f, 0.4f, 0.4f );
	meshColor = makeVector3f( 0.051f, 0.2f, 0.43f );

	context = NULL;
	scene = NULL;

	// Interpolation
	interplatedFrames = 0;

	// Rendering
	rt = new RenderingTool();
	rt->init();
	showBBox = false;
	showGrid = false;
	
	normalUnifier = new UnifyNormals();
	mc = new MarchingCubes();

	// Border
	borderNumNeighborDirs = 5;
	borderTG = NULL;
	showBorders = true;
	pruneBorderDeadEnds = false;
	pruneSmallBorders = true;
	minBorderSize = 20;
	borderTopologyDepth = 2;
}

/** 
 * Destructor
 */
PatchGraphTriangleMesh::~PatchGraphTriangleMesh()
{
	delete normalUnifier;
	delete mc;
	delete rt;
}

/***
 * Setter: scene
 */
void PatchGraphTriangleMesh::setScene( Scene* scene )
{ 
	this->scene = scene; 
}

/** 
* Compute triangle mesh for a single frame and single patch graph
*/
void PatchGraphTriangleMesh::createMesh()
{
	if ( !checkParameters() ) return;
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );

	Grid grid;
	computeGrid( grid );

	UnstructuredInCoreTriangleMesh* mesh = NULL;
	if ( useImplicitMesher ){
		mesh = implicitMesher->createMesh( pg, frameIndex );
	} else {
		mesh = computeTriangleMesh( pg, frameIndex );
		mesh->snapMesh( grid.gridMaxLength/10.0f );
		mesh->closeMesh();
		mesh->clearGarbage();
	}
	if ( mesh != NULL ){
		// Compute normals
		setNormalsFromPatchGraph( mesh, pg, frameIndex, evalSigma * grid.gridMaxLength );
		for ( unsigned int n = 0; n < numSmoothings; n++ ) smoothTriangleMeshNormals( mesh, evalSigma * grid.gridMaxLength );

		// Set color / parameterization
		if ( setMaterialIndex ) mesh->setMaterialIndex( materialIndex );
		if ( colorFromData ) setColorFromData( mesh, frameIndex + pg->getStartFrameIndex(), context->getPGContainer()->getAnimSeq() );
		parameterizeTriangleMesh( mesh, MESH_PARAMETERIZATION_MODE_PLANE, 5, true, 0, 1 );

		QString nodeName = "pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( frameIndex );
		bool replaced = false;
		StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
		it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
		while ( !it->atEnd() ){
			if ( it->getNameWithPath() == string("root/") + qString2STLString(nodeName) ){
				SGObjectNode* sgon = dynamic_cast<SGObjectNode*>( it->getNode() );
				sgon->clearAndSetup( mesh );
				replaced = true;
			}
			it->next();
		}
		delete it;
		if ( !replaced ){
			SGObjectNode* triangleMeshNode = new SGObjectNode();
			triangleMeshNode->clearAndSetup( mesh );
			triangleMeshNode->setName( qString2STLString(nodeName) );
			scene->getRootNode()->addChildNode( triangleMeshNode );
		}
	} else {
		debugOutput <<  "createMesh: NULL mesh created.\n"; debugOutput.flush();
	}
}

/**
* Must only be called after octree was created
*/
float PatchGraphTriangleMesh::evalPhilipp( 
	float x, float y, float z, 
	const PointSet* ps, 
	AttachedIndexedOctree* octree, 
	Grid& grid )
{	
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT TANGENT_U = ps->getAAT( "tangent_u" );
	AAT COEFF = ps->getAAT( "hfieldCoeffs" );
	card32 nIndex;
	bool found;
	PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );	
	Vector3f pos = makeVector3f( x, y, z );
	float32 summedWeight = 0;
	float32 avgDist = 0;

	// Find first neighbor
	knn.setTargetPoint( pos );
	knn.getNextNearestPointIndex( nIndex, found );
	Vector3f nPos = ps->get3f( nIndex, POSITION );

	float32 d = norm( nPos - pos );
	card32 numPoints = 0;
	while( ( numPoints < 1 ) || ( found && ( d < evalSigma * grid.gridMaxLength ) ) ){

		// Process neighbor
		Vector3f nNormal = ps->get3f( nIndex, NORMAL );
		Vector3f nTangent_u = ps->get3f( nIndex, TANGENT_U );
		StaticVector<float,6> nCoeff = ps->get<float32,6>( nIndex, COEFF );
		Matrix3f nT;
		nT[0] = nTangent_u;
		nT[2] = nNormal;
		nT[1] = nT[2].crossProduct( nT[0] ); // tangent_v = normal x tangent_u
		Vector3f nPosTrans = nT.transpose() * ( pos - nPos );
		float32 hPatch = Monomials2ndOrder::evaluate( nCoeff, makeVector2f( nPosTrans[0], nPosTrans[1] ) );
		float32 hPos = nPosTrans[2];
		float32 weight = exp( - sqr(d) / sqr( evalSigma * grid.gridMaxLength ) );
		avgDist += ( hPos - hPatch ) * weight;
		summedWeight += weight;

		// Next neighbor
		knn.getNextNearestPointIndex( nIndex, found );
		nPos = ps->get3f( nIndex, POSITION );
		d = norm( nPos - pos );
		numPoints++;
	}
	return avgDist / summedWeight;
}

/**
* Must only be called after octree was created
*/
float PatchGraphTriangleMesh::evalMichael( float x, float y, float z, const PointSet* ps, const PatchGraph *pg, card32 frameIndex, AttachedIndexedOctree* octree, Grid& grid )
{	
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	card32 nIndex;
	bool found;
	PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );	
	Vector3f pos = makeVector3f( x, y, z );
	float32 summedWeight = 0;
	float32 avgDist = 0;
	
	// Find first neighbor
	knn.setTargetPoint( pos );
	knn.getNextNearestPointIndex( nIndex, found );
	if (!found) return 0;
	Vector3f nPos = ps->get3f( nIndex, POSITION );
	Vector3f firstNormal = ps->get3f( nIndex, NORMAL );
	float32 firstOrient = firstNormal*(pos-nPos);

   //PGNeighborhoodSphereSearch topRange;
   //topRange.setup(pg, frameIndex, nIndex, pos, evalSigma * grid.gridMaxLength*2.5, POSITION);
	
	float32 d = norm( nPos - pos );
	card32 numPoints = 0;
	while( ( numPoints < 1 ) || ( found && ( d < evalSigma * grid.gridMaxLength*2.5 ) ) ){

		// Process neighbor
		Vector3f nNormal = ps->get3f( nIndex, NORMAL );
		float32 newOrient = (pos-nPos)*nNormal;
		if (!((newOrient<0 && firstOrient>0) || (newOrient>0 && firstOrient<0))) {
		   float32 distance = (pos-nPos)*normalize(nNormal);
		   float32 weight = exp( - sqr(d) / sqr( evalSigma * grid.gridMaxLength ) );
		   avgDist += distance * weight;
		   summedWeight += weight;
      }
      		
		// Next neighbor
		knn.getNextNearestPointIndex( nIndex, found );
		nPos = ps->get3f( nIndex, POSITION );
		d = norm( nPos - pos );
		numPoints++;
	}
	return avgDist / summedWeight;
}

/**
 * Check if the parameters are valid
 */
bool PatchGraphTriangleMesh::checkParameters()
{
	if ( context == NULL ){
		return false;
	}
	if ( scene == NULL ){
		return false;
	}
	if ( patchGraphIndex >= context->getPGContainer()->getNumPatchGraphs() ){
		return false;
	}
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	if ( frameIndex >= pg->getTrajectorySet()->getNumFrames() ){
		return false;
	}
	return true;
}

/** 
 * Set normal in mesh based on associated patch graph
 */
void PatchGraphTriangleMesh::setNormalsFromPatchGraph( UnstructuredInCoreTriangleMesh* mesh, const PatchGraph* pg, const card32 frameIndex, float32 sigma )
{
	PointSet* vertices = mesh->getPointSet();
	if( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "setNormalsFromPatchGraph: attribute missing in mesh.\n"; debugOutput.flush();
		return;
	}
	AAT MESH_POSITION = vertices->getAAT( "position" );
	AAT MESH_NORMAL = vertices->getAAT( "normal" );
	PointSet* pgPS = pg->getTrajectorySet()->copySingleFrame( frameIndex );
	if ( !pgPS->providesAttribute( "position" ) || !pgPS->providesAttribute( "normal" ) ){
		debugOutput << "setNormalsFromPatchGraph: attribute missing in patch graph.\n"; debugOutput.flush();
		delete pgPS;
		return;
	}
	AAT PG_POSITION = pgPS->getAAT( "position" );
	AAT PG_NORMAL = pgPS->getAAT( "normal" );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( pgPS );
	PointSetKNNQuery knn( pgPS, octree, NULL_VECTOR3F, PG_POSITION );
	card32 nIndex;
	bool found;

	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, MESH_POSITION );
		knn.setTargetPoint( pos );
		// Average over 5 neighbors
		float32 summedWeight = 0;
		Vector3f avgNormal = makeVector3f( 0,0,0 );

		knn.getNextNearestPointIndex( nIndex, found );
		Vector3f nPos = pgPS->get3f( nIndex, PG_POSITION );
		Vector3f nNormal = pgPS->get3f( nIndex, PG_NORMAL );
		float32 d = norm( nPos - pos );
		
		card32 numFound = 0;
		while( ( numFound < 1 ) || ( found && ( d < sigma ) ) ){
			float32 weight = exp( - sqr( d ) / sqr( sigma ) );
			avgNormal += nNormal * weight;
			summedWeight += weight;

			knn.getNextNearestPointIndex( nIndex, found );
			nPos = pgPS->get3f( nIndex, PG_POSITION );
			nNormal = pgPS->get3f( nIndex, PG_NORMAL );
			d = norm( nPos - pos );
			numFound++;
		}
		avgNormal /= summedWeight;
		if ( norm( avgNormal ) < 1e-8 ){
			avgNormal = makeVector3f( 0,1,0 );
			debugOutput << "invalid normal.\n"; debugOutput.flush();
		} else {
			avgNormal = normalize( avgNormal );
		}
		vertices->set3f( pointIndex, MESH_NORMAL, avgNormal );
	}
	delete pgPS; 
	delete octree;
}

/** 
 * Debug visualization
 */
void PatchGraphTriangleMesh::drawGL()
{
	if ( showBBox ){
		Grid grid;
		computeGrid( grid );
		rt->disableLighting();
		rt->setRenderColor( 1,1,0 );
		rt->renderBox( grid.gridBBox );
		rt->enableLighting();
	}

	
	if ( showGrid ){
		Grid grid;
		computeGrid( grid );
		card32 numGridCells = ( grid.gridDim[0]-1) * ( grid.gridDim[1]-1 ) * ( grid.gridDim[2]-1 );
		if ( numGridCells <= grid.filledCells.size() ){
			BoundingBox3f bbox;
			rt->setRenderColor( 0,1,1 );
			for ( int32 i = 0; i < grid.gridDim[0]-1; i++ ){
				for ( int32 j = 0; j < grid.gridDim[1]-1; j++ ){
					for ( int32 k = 0; k < grid.gridDim[2]-1; k++ ){
						card32 index = ( k * ( grid.gridDim[1]-1) + j ) * ( grid.gridDim[0]-1 ) + i;
						if ( grid.filledCells[index] ){
							float32 xMin = grid.gridBBox.lowerCorner[0] + i * grid.delta[0];
							float32 yMin = grid.gridBBox.lowerCorner[1] + j * grid.delta[1];
							float32 zMin = grid.gridBBox.lowerCorner[2] + k * grid.delta[2];
							float32 xMax = grid.gridBBox.lowerCorner[0] + (i+1) * grid.delta[0];
							float32 yMax = grid.gridBBox.lowerCorner[1] + (j+1) * grid.delta[1];
							float32 zMax = grid.gridBBox.lowerCorner[2] + (k+1) * grid.delta[2];
							bbox.lowerCorner = makeVector3f( xMin, yMin, zMin );
							bbox.upperCorner = makeVector3f( xMax, yMax, zMax );
							rt->renderBox( bbox );
						}
					}
				}
				progressWindow->progress( (float)(i+1)/float(grid.gridDim[0])*100.0f );
			}
		}
	}

	if ( showBorders ){
		if ( checkParameters() && ( borderTG != NULL ) ){
			PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
			PointSet* traj = pg->getTrajectorySet()->getTrajectories();
			AAT POSITION = traj->getAAT( "position" );
			AAT FLAGS = traj->getAAT( "flags" );
			card32 numPatches = pg->getTrajectorySet()->getNumPatches();

			rt->setRenderColor( makeVector3f( 1,1,0 ) );
			rt->disableLighting();
			for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
				for ( unsigned int n = 0; n < borderTG->getNumAdjacentVertices( pointIndex ); n++ ){
					card32 nIndex = borderTG->getAdjacentVertex( pointIndex, n );
					if ( nIndex > pointIndex ){			
						Vector3f start = traj->get3f( frameIndex, pointIndex, POSITION );
						Vector3f end = traj->get3f( frameIndex, nIndex, POSITION );
						rt->renderLine( start, end );
					}
				}
			}
			rt->enableLighting();

			/// Render border points
			rt->setRenderColor( 0.6f, 0.6f, 0.6f );
			Plane plane;
			for ( unsigned int borderPointIndex = 0; borderPointIndex < borderPoints.size(); borderPointIndex++ ){
				BorderEntry borderEntry = borderPoints[borderPointIndex];
				card32 index = borderEntry.pointIndex;
				Vector3f pos = traj->get3f( frameIndex, index, POSITION );
				rt->renderSphere( pos, 0.01f );
				plane.setPoint( pos );
				plane.setNormal( borderEntry.extDir );
				rt->renderPlane( plane, 0.03f );
			}
		}
	}
}

/** 
* Compute a triangle mesh from a patch graph (single frame)
*/
UnstructuredInCoreTriangleMesh* PatchGraphTriangleMesh::computeTriangleMesh( PatchGraph* pg, const card32 frameIndex )
{
	UnstructuredInCoreTriangleMesh *result = NULL;
	//Own mesher
	progressWindow->pushStep( "Triangulation" );
	
	PointSet* ps = pg->getTrajectorySet()->copySingleFrame( frameIndex );
	AAT POSITION = ps->getAAT( "position" );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps );
	PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );
	card32 nIndex;
	bool found;

	Grid grid;
	computeGrid( grid );
	float32	emptyCellDist = sqrt( 2.0f ) * grid.gridMaxLength / 2.0f * emptyCell;
	mc->clearMesh();
	grid.filledCells.resize( (grid.gridDim[0]-1)*(grid.gridDim[1]-1)*(grid.gridDim[2]-1));
	for ( int32 i = 0; i < grid.gridDim[0]-1; i++ ){
		for ( int32 j = 0; j < grid.gridDim[1]-1; j++ ){
			for ( int32 k = 0; k < grid.gridDim[2]-1; k++ ){
				card32 cellIndex = ( k * (grid.gridDim[1]-1) + j ) * (grid.gridDim[0]-1) + i;
				float32 xCenter = grid.gridBBox.lowerCorner[0] + (i+0.5f) * grid.delta[0];
				float32 yCenter = grid.gridBBox.lowerCorner[1] + (j+0.5f) * grid.delta[1];
				float32 zCenter = grid.gridBBox.lowerCorner[2] + (k+0.5f) * grid.delta[2];
				float32 xMin = grid.gridBBox.lowerCorner[0] + i * grid.delta[0];
				float32 yMin = grid.gridBBox.lowerCorner[1] + j * grid.delta[1];
				float32 zMin = grid.gridBBox.lowerCorner[2] + k * grid.delta[2];
				float32 xMax = grid.gridBBox.lowerCorner[0] + (i+1) * grid.delta[0];
				float32 yMax = grid.gridBBox.lowerCorner[1] + (j+1) * grid.delta[1];
				float32 zMax = grid.gridBBox.lowerCorner[2] + (k+1) * grid.delta[2];
				BoundingBox3f bbox;
				bbox.lowerCorner = makeVector3f( xMin, yMin, zMin );
				bbox.upperCorner = makeVector3f( xMax, yMax, zMax );
				knn.setTargetPoint( makeVector3f( xCenter, yCenter, zCenter ) );
				grid.filledCells[cellIndex] = false;
				knn.getNextNearestPointIndex( nIndex, found );
				if ( found ){
					Vector3f nPos = ps->get3f( nIndex, POSITION );
					if ( norm( nPos - makeVector3f( xCenter, yCenter, zCenter ) ) < emptyCellDist ){
						grid.filledCells[cellIndex] = true;
					}
				}
				if ( grid.filledCells[cellIndex] ){ 
					// Create triangles
					Vector3i dim = makeVector3i( 2,2,2 );
					float data[8];
					for ( unsigned int xl = 0; xl < 2; xl++ ){
						for ( unsigned int yl = 0; yl < 2; yl++ ){
							for ( unsigned int zl = 0; zl < 2; zl++ ){
								if ( meshingMode == MESHING_MODE_MICHAEL ){
									data[(zl*2 + yl)*2+xl] = evalMichael( ( xl == 0 )? xMin : xMax, ( yl == 0 )? yMin : yMax, ( zl == 0 )? zMin : zMax, ps, pg, frameIndex, octree, grid );
								} else if ( meshingMode == MESHING_MODE_PHILIPP ){
									data[(zl*2 + yl)*2+xl] = evalPhilipp( ( xl == 0 )? xMin : xMax, ( yl == 0 )? yMin : yMax, ( zl == 0 )? zMin : zMax, ps, octree, grid );
								} else {
									data[(zl*2 + yl)*2+xl] = -1;
								}
							}
						}
					}
					mc->calculateVolume( data, bbox, dim, 0, false );
				}
			}
		}
		progressWindow->progress( (float)(i+1)/float(grid.gridDim[0])*100.0f );
	}
	progressWindow->popSteps( );

	result = new UnstructuredInCoreTriangleMesh();
	VertexDescriptor vd;
	vd.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( VAD( "texcoord", 2, VAD::DATA_FORMAT_FLOAT32 ) );
	result->clearAndSetup( &vd, mc->getNumVertices(), &INDEX_TRI_DESCR, mc->getNumTriangles() );
	PointSet *vertices = result->getPointSet();
	PointSet *triangles = result->getTriangles();
	AAT V_POS = vertices->getAAT( "position" );
	AAT V_NORMAL = vertices->getAAT( "normal" );
	AAT V_COLOR = vertices->getAAT( "color" );
	AAT T_INDEX = triangles->getAAT( "index" );
	const card32 numVertices = vertices->getNumEntries();
	for (card32 i = 0; i < numVertices; i++ ) {
		vertices->set3f( i, V_POS, mc->getVertex( i ) );
		vertices->set3f( i, V_COLOR, meshColor );
		vertices->set3f( i, V_NORMAL, normalize( mc->getVertex( i ) ) );
	}
	const card32 numTriangles = triangles->getNumEntries();
	for ( card32 i = 0; i < numTriangles; i++ ) {
		triangles->set<int32, 3>( i, T_INDEX, mc->getTriangle( i ) );
	}
	// Free allocated mem
	delete ps;
	delete octree;
	return result;
}

/** 
 * Detect border points
 */
void PatchGraphTriangleMesh::detectBorderPoints()
{
	checkParameters();
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	PointSet* traj = pg->getTrajectorySet()->getTrajectories();
	if ( !traj->providesAttribute( "position" ) || !traj->providesAttribute( "normal" ) || !traj->providesAttribute( "tangent_u" ) || !traj->providesAttribute( "color" ) ){
		debugOutput << "detectBorderPoints: patch graph does not provide all required attribute channels.\n"; debugOutput.flush();
		return;
	}

	InCorePCTopologyGraph* tg = pg->getTopology();
	AAT POSITION = traj->getAAT( "position" );
	AAT NORMAL = traj->getAAT( "normal" );
	AAT TANGENT_U = traj->getAAT( "tangent_u" );
	AAT COLOR = traj->getAAT( "color" );
	AAT FLAGS = traj->getAAT( "flags" );

	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	Vector3f* dirs = new Vector3f[borderNumNeighborDirs];
	bool* valids = new bool[borderNumNeighborDirs];
	Vector3f* neighbors = new Vector3f[borderNumNeighborDirs];
	Plane rp;
	Vector3f nPosProj, dir;
	float32 angleWidth = cos( float32( M_PI ) / float32( borderNumNeighborDirs ) );
	set<card32> borderPointsSet; borderPointsSet.clear();

	for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
		Vector3f pos = traj->get3f( frameIndex, pointIndex, POSITION );
		Vector3f normal = traj->get3f( frameIndex, pointIndex, NORMAL );
		rp.setNormal( normal );
		rp.setPoint( pos );
		// Find neighbor candidates
		set<card32> neighborCandidates; neighborCandidates.clear();
		for ( unsigned int n = 0; n < tg->getNumAdjacentVertices( pointIndex ); n++ ){
			card32 nIndex = tg->getAdjacentVertex( pointIndex, n );
			for ( unsigned int m = 0; m < tg->getNumAdjacentVertices( nIndex ); m++ ){
				neighborCandidates.insert( tg->getAdjacentVertex( nIndex, m ) );
			}
			neighborCandidates.insert( tg->getAdjacentVertex( pointIndex, n ) );
		}


		bool isFirst = true;
		card32 numNeighborsFound = 0;
		for ( set<card32>::iterator it = neighborCandidates.begin(); it != neighborCandidates.end(); it++ ){
			card32 nIndex = card32(*it);
			Vector3f nPos = traj->get3f( frameIndex, nIndex, POSITION );
			if ( isFirst ){
				projectPointPlane( rp, nPos, nPosProj );
				dir = nPosProj - pos;
				float normDir = norm( dir );
				if ( normDir > 1e-8 ){
					neighbors[0] = nPos;
					dirs[0] = dir / normDir;
					valids[0] = true;
					numNeighborsFound++;
					float32 rotAngle = 2 * float32( M_PI ) / float32( borderNumNeighborDirs );
					Matrix3f rot = makeRotVector3f( rp.getNormal(), rotAngle );
					for ( unsigned int n = 1; n < borderNumNeighborDirs; n++ ) {
						dirs[n] = rot * dirs[n-1];
						valids[n] = false;
					}
					isFirst = false;
				}
			} else {
				projectPointPlane( rp, nPos, nPosProj );
				dir = nPosProj - pos;
				float normDir = norm( dir );
				if ( normDir > 1e-8 ){
					dir = dir / normDir;
					// Test for neighbors
					for ( unsigned int n = 1; n < borderNumNeighborDirs; n++ ){
						if ( !valids[n] ){
							if ( dir * dirs[n] > angleWidth ){
								neighbors[n] = nPos;
								valids[n] = true;
								numNeighborsFound++;
							}
						}
					}
				}
			}
		}
		if ( numNeighborsFound < borderNumNeighborDirs ){
			borderPointsSet.insert( pointIndex );
		} 
	}

	// Create border topology graph
	if ( borderTG != NULL ) delete borderTG;
	borderTG = new InCorePCTopologyGraph();
	borderTG->clearAndSetup( tg->getEdges()->getDescr() );
	for ( set<card32>::iterator it = borderPointsSet.begin(); it != borderPointsSet.end(); it++ ){
		card32 borderPoint = (card32)(*it);
		for ( unsigned int n = 0; n < tg->getNumAdjacentVertices( borderPoint ); n++ ){
			card32 nIndex = tg->getAdjacentVertex( borderPoint, n );				
			set<card32>::iterator it = borderPointsSet.find( nIndex );
			if ( it != borderPointsSet.end() ){
				if ( nIndex > borderPoint ){
					borderTG->addEdge( borderPoint, nIndex );	
				}
			}
			if ( borderTopologyDepth > 1 ){
				// Consider neighbors of 2nd degree
				for ( unsigned int m = 0; m < tg->getNumAdjacentVertices( nIndex ); m++ ){
					card32 nnIndex = tg->getAdjacentVertex( nIndex, m );
					set<card32>::iterator it = borderPointsSet.find( nnIndex );
					if ( it != borderPointsSet.end() ){
						if ( nnIndex > borderPoint ){
							borderTG->addEdge( borderPoint, nnIndex );
						}
					}
					if ( borderTopologyDepth > 2 ){
						// Consider neighbors of 3nd degree
						for ( unsigned int k = 0; k < tg->getNumAdjacentVertices( nnIndex ); k++ ){
							card32 nnnIndex = tg->getAdjacentVertex( nnIndex, k );
							set<card32>::iterator it = borderPointsSet.find( nnnIndex );
							if ( it != borderPointsSet.end() ){
								if ( nnnIndex > borderPoint ){
									borderTG->addEdge( borderPoint, nnnIndex );
								}
							}
						}
					}
				}
			}
		}
	}

	// Prune dead end borders
	if ( pruneBorderDeadEnds ){
		/// Iteratively remove all border points with no more than one edge
		card32 removedEdges = 0;
		bool converged = false;
		while( !converged ){
			converged = true;
			for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
				if ( borderTG->getNumAdjacentVertices( pointIndex ) == 1 ){
					card32 nIndex = borderTG->getAdjacentVertex( pointIndex, 0 );
					// Delete edge in point index container
					borderTG->deleteEdge( pointIndex, 0 );
					// delete edge in nIndex container
					for ( unsigned int n = 0; n < borderTG->getNumAdjacentVertices( nIndex ); n++ ){
						if ( borderTG->getAdjacentVertex( nIndex, n ) == pointIndex ){
							borderTG->deleteEdge( nIndex, n );
						}
					}
					converged = false;
					removedEdges++;
				}
			}
		}
		debugOutput << "Pruned " << removedEdges << " dead-end edges.\n"; debugOutput.flush();
	}

	if ( pruneSmallBorders ){
		card32 numPruned = 0;
		for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
			if ( borderTG->getNumAdjacentVertices( pointIndex ) > 0 ){
				// Region growing on border topology graph
				stack<card32> stack;
				set<card32> regionPoints; regionPoints.clear();
				// Init
				stack.push( pointIndex );
				regionPoints.insert( pointIndex );
				// Grow
				while( !stack.empty() ){
					card32 index = stack.top();
					stack.pop();
					for ( unsigned int n = 0; n < borderTG->getNumAdjacentVertices( index ); n++ ){
						card32 nIndex = borderTG->getAdjacentVertex( index, n );
						if ( regionPoints.find( nIndex ) == regionPoints.end() ){
							stack.push( nIndex );
							regionPoints.insert( nIndex );
						}
					}
				}
				if ( regionPoints.size() < minBorderSize ){
					borderTG->clearVertex( pointIndex );
					numPruned++;
				}
			}
		}
		debugOutput << "Pruned " << numPruned << " vertices from too small border regions.\n"; debugOutput.flush();
	}

	// Detect and mark all border points
	borderPoints.clear();
	for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
		if ( borderTG->getNumAdjacentVertices( pointIndex ) > 0 ){
			BorderEntry borderEntry;
			borderEntry.pointIndex = pointIndex;
			borderEntry.extDir = makeVector3f( 0,1,0 );
			borderPoints.push_back( borderEntry );
			// Border
			traj->set1i( frameIndex, pointIndex, FLAGS, traj->get1i( frameIndex, pointIndex, FLAGS ) | IS_BORDER_FLAG );
		} else {
			// Not border
			traj->set1i( frameIndex, pointIndex, FLAGS, traj->get1i( frameIndex, pointIndex, FLAGS ) & !IS_BORDER_FLAG );
		}
	}

	// Compute exterior direction for all border points
	TGNeighborhoodKNNSearch topoKNN( traj, tg, frameIndex, 0 );
	for ( unsigned int borderIndex = 0; borderIndex < borderPoints.size(); borderIndex++ ){
		card32 index = borderPoints[borderIndex].pointIndex;
		Vector3f pos = traj->get3f( frameIndex, index, POSITION );
		set<card32> neighbors;
		neighbors.insert( index );
		for ( unsigned int i = 0; i < borderTG->getNumAdjacentVertices( index ); i++ ){
			card32 nIndex = borderTG->getAdjacentVertex( index, i );
			neighbors.insert( nIndex );
			for ( unsigned int j = 0; j < borderTG->getNumAdjacentVertices( nIndex ); j++ ){
				card32 nnIndex = borderTG->getAdjacentVertex( nIndex, j );
				neighbors.insert( nnIndex );
			}
		}
		PCA<float32,3> pca; pca.clear();
		Matrix3f eigenVectors;
		Vector3f eigenValues, centroid;;
		for ( set<card32>::iterator it = neighbors.begin(); it != neighbors.end(); it++ ){
			pca.addPoint( traj->get3f( frameIndex, *it, POSITION ) );	
		}
		pca.analyze( eigenValues, eigenVectors, centroid );
		borderPoints[borderIndex].extDir = traj->get3f( frameIndex, index, NORMAL ).crossProduct( eigenVectors[0] );

		// Adjust orientation
		centroid = makeVector3f( 0,0,0 );
		topoKNN.reset( frameIndex, index );
		card32 numNonBorderNeighbors = 0;
		while( ( numNonBorderNeighbors < 10 ) && ( !topoKNN.atEnd() ) ){
			topoKNN.next();
			card32 nIndex = topoKNN.getCurrentIndex();
			if ( !( traj->get1i( frameIndex, nIndex, FLAGS ) & IS_BORDER_FLAG ) ){
				centroid = ( centroid * numNonBorderNeighbors + traj->get3f( frameIndex, nIndex, POSITION ) ) / float32( numNonBorderNeighbors+1 );
				numNonBorderNeighbors++;
			}
		}
		if ( ( centroid - pos ) * borderPoints[borderIndex].extDir > 0 ) borderPoints[borderIndex].extDir = -borderPoints[borderIndex].extDir;
	}
		


	delete [] neighbors;
	delete [] valids;
	delete [] dirs;
}

/** 
 * Set color from data
 */
void PatchGraphTriangleMesh::setColorFromData( UnstructuredInCoreTriangleMesh* mesh, const card32 globalFrameIndex, PointCloudAnimationSequence* animSeq )
{
	if( !checkParameters() ){
		return;
	}
	Grid grid;
	computeGrid( grid );
	PointSet* vertices = mesh->getPointSet();
	AAT COLOR = vertices->getAAT( "color" );
	AAT POSITION = vertices->getAAT( "position" );
	PointSet* dataFrame = animSeq->getFrame( globalFrameIndex );
	AAT DATA_POSITION = dataFrame->getAAT( "position" );
	AAT DATA_COLOR = dataFrame->getAAT( "color" );
	AttachedIndexedOctree* dataOctree = animSeq->fetchOctree( globalFrameIndex );
	PointSetKNNQuery knn( dataFrame, dataOctree, NULL_VECTOR3F, DATA_POSITION );	
	card32 nDataIndex;
	bool found;
	card32 numVertices = vertices->getNumEntries();
	for ( unsigned int pointIndex = 0; pointIndex < numVertices; pointIndex++ ){
		Vector3f pos = vertices->get3f( frameIndex, pointIndex, POSITION );
		knn.setTargetPoint( pos );
		float32 summedWeights = 0;
		Vector3f avgColor = makeVector3f( 0,0,0 );
		card32 numFound = 0;
		// Find first data point
		knn.getNextNearestPointIndex( nDataIndex, found );
		Vector3f nDataPos = dataFrame->get3f( nDataIndex, DATA_POSITION );
		Vector3f nDataColor = dataFrame->get3f( nDataIndex, DATA_COLOR );
		float32 d = norm( pos - nDataPos );
		while( ( numFound < 1 ) || ( found && ( d < colorSigma * grid.gridMaxLength ) ) ){
			// Process data point
			float32 w = exp( - sqr( d ) / sqr( colorSigma * grid.gridMaxLength ) );
			summedWeights += w;
			avgColor += dataFrame->get3f( nDataIndex, DATA_COLOR ) * w;
			// Next data point
			knn.getNextNearestPointIndex( nDataIndex, found );
			nDataPos = dataFrame->get3f( nDataIndex, DATA_POSITION );
			nDataColor = dataFrame->get3f( nDataIndex, DATA_COLOR );
			d = norm( pos - nDataPos );
			numFound++;
		}
		vertices->set3f( frameIndex, pointIndex, COLOR, avgColor/summedWeights );
	}
}

/** 
* Compute grid parameters based on current settings
*/
void PatchGraphTriangleMesh::computeGrid( Grid& grid )
{

	if ( !checkParameters() ){
		debugOutput << "computeGrid: invalid parameters.\n"; debugOutput.flush();
		return;
	}
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	PointSet* traj = pg->getTrajectorySet()->getTrajectories(); 
	AAT POSITION = traj->getAAT( "position" );
	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	// Compute bounding box of patch graph frame
	bool first = true;
	for ( unsigned int pointIndex = 0; pointIndex < numPatches; pointIndex++ ){
		Vector3f pos = traj->get3f( frameIndex, pointIndex, POSITION );
		if ( first ){
			first = false;
			grid.gridBBox = BoundingBox3f( pos );
		} else {
			grid.gridBBox.addPoint( pos );
		}
	}
	/// Compute optimal dimensions and bounding box (extended by 10%)
	grid.delta[0] = ( grid.gridBBox.getMaxSideLength() * 0.9f / resolution ); 
	grid.gridDim[0] = card32( ( grid.gridBBox.upperCorner[0] - grid.gridBBox.lowerCorner[0] ) / grid.delta[0] + 1 );
	grid.gridDim[1] = card32( ( grid.gridBBox.upperCorner[1] - grid.gridBBox.lowerCorner[1] ) / grid.delta[0] + 1 );
	grid.gridDim[2] = card32( ( grid.gridBBox.upperCorner[2] - grid.gridBBox.lowerCorner[2] ) / grid.delta[0] + 1 );
	Vector3f center = grid.gridBBox.getCenter();
	grid.gridBBox.lowerCorner[0] = center[0] - grid.gridDim[0] / 2.0f * grid.delta[0];
	grid.gridBBox.upperCorner[0] = center[0] + grid.gridDim[0] / 2.0f * grid.delta[0];
	grid.gridBBox.lowerCorner[1] = center[1] - grid.gridDim[1] / 2.0f * grid.delta[0];
	grid.gridBBox.upperCorner[1] = center[1] + grid.gridDim[1] / 2.0f * grid.delta[0];
	grid.gridBBox.lowerCorner[2] = center[2] - grid.gridDim[2] / 2.0f * grid.delta[0];
	grid.gridBBox.upperCorner[2] = center[2] + grid.gridDim[2] / 2.0f * grid.delta[0];
	// Compute triangle mesh
	for ( unsigned int n = 0; n < 3; n++ ){
		grid.delta[n] = ( grid.gridBBox.upperCorner[n]- grid.gridBBox.lowerCorner[n] ) / float32( grid.gridDim[n] -1 );
	}
	grid.gridMaxLength = max( grid.delta[0], grid.delta[1] );
	grid.gridMaxLength = max( grid.gridMaxLength, grid.delta[2] );
}

/** 
 * Clip triangle mesh at border points
 */
void PatchGraphTriangleMesh::clipAtBorder()
{
	if ( !checkParameters() ){
		debugOutput << "clipAtBorder: invalid parameters.\n"; debugOutput.flush();
		return;
	}
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	PointSet* ps = pg->getTrajectorySet()->copySingleFrame( frameIndex );
	AAT PG_POSITION = ps->getAAT( "position" );
	AAT PG_FLAGS = ps->getAAT( "flags" );
	AttachedIndexedOctree* pgOctree = PointSetKNNQuery::createOctree( ps );
	PointSetKNNQuery knn( ps, pgOctree, NULL_VECTOR3F, PG_POSITION );
	card32 nPGIndex;
	bool found;
	// Get Triangle mesh
	QString name = "root/pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( frameIndex );
	UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, qString2STLString(name) ) );
	if ( mesh == NULL ){
		debugOutput << "clipAtBorder: cannot find mesh.\n"; debugOutput.flush();
		return;
	}

	Grid grid;
	computeGrid( grid );

	float32 maxDistFromCellCenter = sqrt( 2.0f ) * grid.gridMaxLength * emptyCell * 1.1f;

	// Detect border candidates
	PointSet* vertices = mesh->getPointSet();
	AAT VERTEX_POSITION = vertices->getAAT( "position" );
	AAT VERTEX_COLOR = vertices->getAAT( "color" );
	card32 numVertices = vertices->getNumEntries();
	for( unsigned int vertexIndex = 0; vertexIndex < numVertices; vertexIndex++ ){
		Vector3f vertexPos = vertices->get3f( vertexIndex, VERTEX_POSITION );
		// Check for closest 
		knn.setTargetPoint( vertexPos );
		knn.getNextNearestPointIndex( nPGIndex, found );
		Vector3f nPGPos = ps->get3f( nPGIndex, PG_POSITION );
		float32 d = norm( vertexPos - nPGPos );
		bool borderCand = false;
		while( found && ( d < maxDistFromCellCenter ) ){
			if ( ps->get1i( nPGIndex, PG_FLAGS ) & IS_BORDER_FLAG ){
				// Candidate
				borderCand = true;
				break;
			}
			knn.getNextNearestPointIndex( nPGIndex, found );
			nPGPos = ps->get3f( nPGIndex, PG_POSITION );
			d = norm( vertexPos - nPGPos );
		}
		if ( borderCand ){
			for ( unsigned int n = 0; n < borderPoints.size(); n++ ){
				if ( borderPoints[n].pointIndex == nPGIndex ){
					if ( ( vertexPos - nPGPos ) * borderPoints[n].extDir > 0 ){
						//vertices->set3f( vertexIndex, VERTEX_COLOR, makeVector3f( 1,0,0 ) );
						Plane plane;
						plane.setPoint( nPGPos );
						plane.setNormal( borderPoints[n].extDir );
						Vector3f newPos;
						projectPointPlane( plane, vertexPos, newPos );
						vertices->set3f( vertexIndex, VERTEX_POSITION, newPos );
					}
				}
			}
			
		}
	} 

	// Rebuild render objects
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}

	delete pgOctree;
	delete ps;
}


struct InterpEntry {
   card32 pgIndex;
   float32 weight;
};

/** 
* Compute triangle mesh for a complete patch graph (consistent with selected frame)
*/
void PatchGraphTriangleMesh::extrapolate()
{
	if ( extrapolatorMode == EXTRAPOLATE_MODE_PHILIPP ) {
		doOriginalExtrapolate();
		return;
	} else if ( extrapolatorMode == EXTRAPOLATE_MODE_MICHAEL ){
		if ( !checkParameters() ) return;
		PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
		// Find reference frame
		QString name = "root/pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( frameIndex );
		UnstructuredInCoreTriangleMesh* referenceMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, qString2STLString(name) ) );
		if ( referenceMesh == NULL ){
			debugOutput << "extrapolate: cannot extract reference mesh.\n"; debugOutput.flush();
			return;
		}

		PointSet* meshPS = referenceMesh->getPointSet();
		AAT MESH_POSITION = meshPS->getAAT( "position" );

		PointSet* pgPS = pg->getTrajectorySet()->copySingleFrame( frameIndex );
		AAT PG_POSITION = pgPS->getAAT( "position" );
		PGNeighborhoodSphereSearch topRange;
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( pgPS );
		PointSetKNNQuery knn( pgPS, octree, NULL_VECTOR3F, PG_POSITION );	
		card32 nIndex;
		bool found;
		float32 sigma = 0;
		vector< vector<InterpEntry> > interps(meshPS->getNumEntries());
		for ( unsigned int pointIndex = 0; pointIndex < meshPS->getNumEntries(); pointIndex++ ){
			// Determine closest points
			Vector3f meshPos = meshPS->get3f( pointIndex, MESH_POSITION );
			knn.setTargetPoint( meshPos );
			knn.getNextNearestPointIndex( nIndex, found );
			if (!found) continue;
			topRange.setup(pg, frameIndex, nIndex, meshPos, context->getSamplingSpacing()*evalSigma*2.5f, PG_POSITION);
			float32 weightSum = 0;
			while (!topRange.atEnd()) {
				card32 pIndex = topRange.getCurrentIndex();
				Vector3f posNPG = pgPS->get3f(0, pIndex, PG_POSITION);
				float32 w = exp(-(norm(posNPG - meshPos)/(context->getSamplingSpacing()*evalSigma)));
				InterpEntry e;
				e.pgIndex = pIndex;
				e.weight = w;
				weightSum += w;
				interps[pointIndex].push_back(e);
				topRange.next();
			}

			if (weightSum > 0) {
				for ( unsigned int n = 0; n < interps[pointIndex].size(); n++ ){
					interps[pointIndex][n].weight /= weightSum;
				}
			}
		}

		delete pgPS;
		delete octree;

		// Create all meshes relative to master frame
		SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
		QString rootNodeName = "pg" + QString::number( patchGraphIndex );
		animNode->setName( qString2STLString(rootNodeName) );
		UnstructuredInCorePointCloud* lastMesh = NULL;
		for ( unsigned int fIndex = 0; fIndex < pg->getTrajectorySet()->getNumFrames(); fIndex++ ){
			UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( referenceMesh->copy() );
			// Copy master frame; adjust positions
			PointSet* imeshPS = mesh->getPointSet();
			AAT MESH_POSITION = meshPS->getAAT( "position" );
			AAT MESH_NORMAL = meshPS->getAAT( "normal" );
			AAT MESH_COLOR = meshPS->getAAT( "color" );
			AAT MESH_TEX = meshPS->getAAT( "texcoord" );
			PointSet* traj = pg->getTrajectorySet()->getTrajectories();
			AAT PG_POSITION = traj->getAAT( "position" );
			AAT PG_NORMAL = traj->getAAT( "normal" );

			for (unsigned pointIndex = 0; pointIndex < meshPS->getNumEntries(); pointIndex++) {
				Vector3f pos = NULL_VECTOR3F;
				Vector3f normal = NULL_VECTOR3F;
				for (unsigned i=0; i<interps[pointIndex].size(); i++) {
					card32 pgIndex = interps[pointIndex][i].pgIndex;
					pos    += traj->get3f(fIndex, pgIndex, PG_POSITION) * interps[pointIndex][i].weight;
					normal += traj->get3f(fIndex, pgIndex, PG_NORMAL)   * interps[pointIndex][i].weight;
				}
				imeshPS->set3f( pointIndex, MESH_POSITION, pos );
				imeshPS->set3f( pointIndex, MESH_NORMAL, normal );
				imeshPS->set3f( pointIndex, MESH_COLOR, meshPS->get3f(pointIndex, MESH_COLOR) );
				imeshPS->set2f( pointIndex, MESH_TEX, meshPS->get2f(pointIndex, MESH_TEX) );
			}

			// Set material
			if ( setMaterialIndex ) mesh->setMaterialIndex( materialIndex );

			SGObjectNode* frameNode = new SGObjectNode();
			frameNode->clearAndSetup( mesh );
			QString frameNodeName = "pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( fIndex );
			frameNode->setName( qString2STLString(frameNodeName) );
			animNode->addNode( frameNode );
			lastMesh = mesh;
		}
		scene->getRootNode()->addChildNode( animNode );
	}
}

/** 
 * Smooth out normals
 */
void PatchGraphTriangleMesh::smoothNormals()
{
	/// Find mesh
	Grid grid;
	computeGrid( grid );
	QString nodeName = "pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( frameIndex );
	StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
	it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
	bool found = false;
	while ( !it->atEnd() ){
		if ( it->getNameWithPath() == qString2STLString( "root/" + nodeName )){
			SGObjectNode* sgon = dynamic_cast<SGObjectNode*>( it->getNode() );
			UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( sgon->getSceneObject() );
			if ( mesh != NULL ){
				smoothTriangleMeshNormals( mesh, evalSigma * grid.gridMaxLength );
				found = true;
			}
		}
		it->next();
	}
	delete it;
	if ( !found ){
		debugOutput << "smoothNormals: cannot find triangle mesh.\n"; debugOutput.flush();
	} else {
		// Rebuild render objects
		for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
			GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
			if ( mat != NULL ){
				mat->increaseVersion();
			}
		}
	}
}

/** 
 * Compute height field an reference frames
 */
void PatchGraphTriangleMesh::computeRefAndHF()
{
	// Check for required attributes
	if ( !checkParameters() ) return;
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	BSROptVariableIndexSet indices;
	BSROptPointIndex index;
	card32 numFrames = pg->getTrajectorySet()->getNumFrames();
	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	for ( unsigned int fIndex = 0; fIndex < numFrames; fIndex++ ){
		for ( unsigned int pIndex = 0; pIndex < numPatches; pIndex++ ){
			index.frame = fIndex;
			index.point = pIndex;
			indices.push_back( index );
		}
	}
	if ( !pg->getTrajectorySet()->getTrajectories()->providesAttribute( "normal" ) ) pg->createPatchAttribute( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	if ( !pg->getTrajectorySet()->getTrajectories()->providesAttribute( "tangent_u" ) ) pg->createPatchAttribute( mVAD( "tangent_u", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	if ( !pg->getTrajectorySet()->getTrajectories()->providesAttribute( "sampleSpacing" ) ) pg->createPatchAttribute( mVAD( "sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32 ) );
	if ( !pg->getTrajectorySet()->getTrajectories()->providesAttribute( "hfieldCoeffs" ) ) pg->createPatchAttribute( mVAD( "hfieldCoeffs", 6, VAD::DATA_FORMAT_FLOAT32 ) );
	BSROptContext* optContext = new BSROptContext();
	BSRFitting::estSampleSpacing( optContext, pg, indices );
	BSRFitting::fitRefFrames( optContext, pg, indices );
	normalUnifier->unifyNormals( pg );
	BSRFitting::fit2ndOrderHeightf( optContext, pg, indices );
	delete optContext;
}



/** 
* Compute triangle mesh for a complete patch graph (consistent with selected frame)
*/
#define NUM_NEIGHBORS 5
void PatchGraphTriangleMesh::doOriginalExtrapolate()
{
	if ( !checkParameters() ) return;
	PatchGraph* pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
	// Find reference frame
	QString name = "root/pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( frameIndex );
	UnstructuredInCoreTriangleMesh* referenceMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, qString2STLString(name) ) );
	if ( referenceMesh == NULL ){
		debugOutput << "extrapolate: cannot extract reference mesh.\n"; debugOutput.flush();
		return;
	}
	// Grid
	Grid grid;
	computeGrid( grid );
	// Compute master mesh frame coefficients
	PointSet* meshPS = referenceMesh->getPointSet();
	AAT MESH_POSITION = meshPS->getAAT( "position" );
	vector< StaticVector<card32,NUM_NEIGHBORS> > neighbors; neighbors.resize( referenceMesh->getPointSet()->getNumEntries() );
	vector< StaticVector<Vector3f,NUM_NEIGHBORS> > offsets; offsets.resize( referenceMesh->getPointSet()->getNumEntries() );
	PointSet* pgPS = pg->getTrajectorySet()->copySingleFrame( frameIndex );
	AAT PG_POSITION = pgPS->getAAT( "position" );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( pgPS );
	PointSetKNNQuery knn( pgPS, octree, NULL_VECTOR3F, PG_POSITION );	
	card32 nIndex;
	bool found;
	float32 sigma = 0;
	for ( unsigned int pointIndex = 0; pointIndex < meshPS->getNumEntries(); pointIndex++ ){
		// Determine closest points
		Vector3f meshPos = meshPS->get3f( pointIndex, MESH_POSITION );
		knn.setTargetPoint( meshPos );
		for ( unsigned int n = 0; n < NUM_NEIGHBORS; n++ ){
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = pgPS->get3f( nIndex, PG_POSITION );
			neighbors[pointIndex][n] = nIndex;
			offsets[pointIndex][n] = ( meshPos - nPos );
			sigma += norm( offsets[pointIndex][n] );
		}
	}
	sigma /= float32( meshPS->getNumEntries() * NUM_NEIGHBORS );
	delete pgPS;
	delete octree;

	// Create all meshes relative to master frame
	SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
	QString rootNodeName = "pg" + QString::number( patchGraphIndex );
	animNode->setName( qString2STLString(rootNodeName) );
	UnstructuredInCorePointCloud* lastMesh = NULL;
	for ( unsigned int fIndex = 0; fIndex < pg->getTrajectorySet()->getNumFrames(); fIndex++ ){
		UnstructuredInCoreTriangleMesh* mesh = NULL;
		if ( individualMeshes ){
			mesh = computeTriangleMesh( pg, fIndex );
			mesh->snapMesh( grid.gridMaxLength/10.0f );
			mesh->closeMesh();
			mesh->clearGarbage();
		} else {
			mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( referenceMesh->copy() );
			// Copy master frame; adjust positions
			PointSet* meshPS = mesh->getPointSet();
			AAT MESH_POSITION = meshPS->getAAT( "position" );
			AAT MESH_COLOR = meshPS->getAAT( "color" );
			AAT MESH_NORMAL = meshPS->getAAT( "normal" );
			AAT MESH_TEX_COORD = meshPS->getAAT( "texcoord" );
			PointSet* traj = pg->getTrajectorySet()->getTrajectories();
			AAT PG_POSITION = traj->getAAT( "position" );
			for ( unsigned int pointIndex = 0; pointIndex < meshPS->getNumEntries(); pointIndex++ ){
				Vector3f pos = makeVector3f( 0,0,0 );
				float32 summedWeight = 0;
				for ( unsigned int n = 0; n < NUM_NEIGHBORS; n++ ){
					card32 nIndex = neighbors[pointIndex][n];
					Vector3f offset = offsets[pointIndex][n];
					Vector3f nPos = traj->get3f( fIndex, nIndex, PG_POSITION );
					float32 weight = exp( - sqr( norm( offset ) ) / sqr( sigma ) );
					summedWeight += weight;
					pos += ( nPos + offset ) * weight;
				}
				pos /= summedWeight;
				meshPS->set3f( pointIndex, MESH_POSITION, pos );
				if ( extrapolateColorSource == EXTRAPOLATE_COLOR_SET ){
					meshPS->set3f( pointIndex, MESH_COLOR, meshColor );
				}
			}
		}
		if ( extrapolateColorSource == EXTRAPOLATE_COLOR_DATA ){
			setColorFromData( mesh, fIndex + pg->getStartFrameIndex(), context->getPGContainer()->getAnimSeq() );
		}
		// Update normals
		setNormalsFromPatchGraph( mesh, pg, fIndex, evalSigma * grid.gridMaxLength );
		for ( unsigned int n = 0; n < numSmoothings; n++ ) smoothTriangleMeshNormals( mesh, evalSigma * grid.gridMaxLength );
		// Set material
		if ( setMaterialIndex ) mesh->setMaterialIndex( materialIndex );
		

		// Create interpolated frames
		for( unsigned int interpolatedFrame = 0; interpolatedFrame < interplatedFrames; interpolatedFrame++ ){
			AAT MESH_POSITION = meshPS->getAAT( "position" );
			AAT MESH_COLOR = meshPS->getAAT( "color" );
			AAT MESH_NORMAL = meshPS->getAAT( "normal" );
			AAT MESH_TEX_COORD = meshPS->getAAT( "texcoord" );
			if ( ( fIndex > 0 ) && ( fIndex < pg->getTrajectorySet()->getNumFrames() ) ){
				UnstructuredInCorePointCloud* nextMesh = mesh;
				if ( ( lastMesh != NULL ) && ( nextMesh != NULL ) ){
					UnstructuredInCoreTriangleMesh* interMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( mesh->copy() );
					float32 t = float32( interpolatedFrame + 1 ) / float32( interplatedFrames + 1 );
					for ( unsigned int vertexIndex = 0; vertexIndex < interMesh->getPointSet()->getNumEntries(); vertexIndex++ ){
						// Position
						Vector3f lastPos = lastMesh->getPointSet()->get3f( vertexIndex, MESH_POSITION );
						Vector3f nextPos = nextMesh->getPointSet()->get3f( vertexIndex, MESH_POSITION );
						Vector3f pos = lastPos * (1-t) + nextPos * t;
						interMesh->getPointSet()->set3f( vertexIndex, MESH_POSITION, pos );
						// Color
						Vector3f lastColor = lastMesh->getPointSet()->get3f( vertexIndex, MESH_COLOR );
						Vector3f nextColor = nextMesh->getPointSet()->get3f( vertexIndex, MESH_COLOR );
						Vector3f color = lastColor * (1-t) + nextColor * t;
						interMesh->getPointSet()->set3f( vertexIndex, MESH_COLOR, color );
						// Normal
						Vector3f lastNormal = lastMesh->getPointSet()->get3f( vertexIndex, MESH_NORMAL );
						Vector3f nextNormal = nextMesh->getPointSet()->get3f( vertexIndex, MESH_NORMAL );
						Vector3f normal = normalize( lastNormal * (1-t) + nextNormal * t );
						interMesh->getPointSet()->set3f( vertexIndex, MESH_NORMAL, normal );
						// tex_coord
					   Vector2f lastTexCoord = lastMesh->getPointSet()->get2f( vertexIndex, MESH_TEX_COORD );
					   Vector2f nextTexCoord = nextMesh->getPointSet()->get2f( vertexIndex, MESH_TEX_COORD );
					   Vector2f texCoord = lastTexCoord * (1-t) + nextTexCoord * t;
					   interMesh->getPointSet()->set2f( vertexIndex, MESH_TEX_COORD, texCoord );
					}
					SGObjectNode* frameNode = new SGObjectNode();
					frameNode->clearAndSetup( interMesh );
					QString frameNodeName = "interpolated" + QString::number( interpolatedFrame );
					frameNode->setName( qString2STLString(frameNodeName) );
					animNode->addNode( frameNode );
				}
			} 
		}   
	
		SGObjectNode* frameNode = new SGObjectNode();
		frameNode->clearAndSetup( mesh );
		QString frameNodeName = "pg" + QString::number( patchGraphIndex ) + "Frame" + QString::number( fIndex );
		frameNode->setName( qString2STLString(frameNodeName) );
		animNode->addNode( frameNode );
		lastMesh = mesh;
	}
	scene->getRootNode()->addChildNode( animNode );
}
