//---------------------------------------------------------------------------
#ifndef PatchGraphTriangleMeshH
#define PatchGraphTriangleMeshH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
#include "Persistent.h"
#include "PatchGraph.h"
#include "AnimationBSRContext.h"
#include "Scene.h"
#include "BSROptContext.h"
#include "UnifyNormals.h"
#include "MarchingCubes.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "TriangleMeshTexture.h"
#include "RenderingTool.h"
#include "ImplicitPatchGraphMeshing.h"
#include "AttachedIndexedOctree.h"
//---------------------------------------------------------------------------

namespace geoxl {

struct Border : public vector<Vector2i>
{
	Vector3f color;
};

#define EXTRAPOLATE_COLOR_DATA 0
#define EXTRAPOLATE_COLOR_REFERENCE 1
#define EXTRAPOLATE_COLOR_SET 2
#define IS_BORDER_FLAG 1024

#define EXTRAPOLATE_MODE_PHILIPP 0
#define EXTRAPOLATE_MODE_MICHAEL 1

#define MESHING_MODE_PHILIPP 0
#define MESHING_MODE_MICHAEL 1

struct BorderEntry
{
	/// Index of point in current frame in current patch graph
	card32 pointIndex; 
	/// Direction pointing into outward direction
	Vector3f extDir;
};

/// Marching cubes grid
struct Grid
{
	BoundingBox3f gridBBox;
	Vector3i gridDim;
	float32 delta[3];
	vector<bool> filledCells;  /// Cells containing triangles
	float32 gridMaxLength;
};

/// Post-processing of patch graph results: triangle mesh transformation
class ANIMATIONRECONSTRUCTION_API PatchGraphTriangleMesh : public Persistent {
	GEOXL_CLASS( PatchGraphTriangleMesh )

	private:
		/*** IMPORTANT PARAMETERS ***/
		/// Scaling factor to determine empty cells
		float32 emptyCell;
		/// Epsilon environment used in all interpolations (distance function, color, normals); scaled with gridDelta
		float32 evalSigma;
		/// Same parameter for color
		float32 colorSigma;
		/// Index of frame to be triangulated
		card32 frameIndex;
		/// Index of patch graph in container to be meshed
		card32 patchGraphIndex;
		/// Grid resolution relative to longest bounding box size
		card32 resolution;
		/// Do not extrapolate single mesh but compute individual mesh for each frame
		bool individualMeshes;
		/// 
		card32 numSmoothings;

		card32 extrapolatorMode;
		card32 meshingMode;

		/*** COLOR ***/
		/// Mesh color
		Vector3f meshColor;
		/// Set material index
		card32 materialIndex;
		bool setMaterialIndex;
		/// Extract the color value from data
		bool colorFromData;
		/// Source of color value in extrapolated mesh
		card32 extrapolateColorSource;

		/// Context
		AnimationBSRContext* context;
		/// Scene
		Scene* scene;
		/// Normal unifier
		UnifyNormals* normalUnifier;
		/// Marching cubes object
		MarchingCubes* mc;

		ImplicitPatchGraphMeshing* implicitMesher;
		bool useImplicitMesher;

		/// Interpolation
		card32 interplatedFrames;

		/// Border
		card32 borderNumNeighborDirs;
		InCorePCTopologyGraph* borderTG;
		bool showBorders;
		bool pruneBorderDeadEnds;
		bool pruneSmallBorders;
		card32 minBorderSize;
		vector<BorderEntry> borderPoints;
		card32 borderTopologyDepth;

		/// Rendering
		RenderingTool* rt;
		bool showBBox;
		bool showGrid;

		/*** MEMBER FUNCTIONS ***/
		/// Compute a triangle mesh from a patch graph
		UnstructuredInCoreTriangleMesh* computeTriangleMesh( PatchGraph* pg, const card32 frameIndex );
		/// Evaluate implicit function (distance to patch graph)
		float evalMichael( 
			float x, float y, float z, 
			const PointSet* ps, 
			const PatchGraph *pg, 
			card32 frameIndex, 
			AttachedIndexedOctree* octree, 
			Grid& grid );
		float evalPhilipp( 
			float x, float y, float z, 
			const PointSet* ps, 
			AttachedIndexedOctree* octree, 
			Grid& grid );
		/// Check if the parameters are valid
		bool checkParameters();
		/// Set normal in mesh based on associated patch graph
		void setNormalsFromPatchGraph( UnstructuredInCoreTriangleMesh* mesh, const PatchGraph* pg, const card32 frameIndex, float32 sigma );
		/// Set color from data
		void setColorFromData( UnstructuredInCoreTriangleMesh* mesh, const card32 globalFrameIndex, PointCloudAnimationSequence* animSeq );
		/// Compute grid parameters based on current settings
		void computeGrid( Grid& grid );
		/// Compute height field an reference frames
		void computeRefAndHF();
		
	public:
		/// Constructor
		PatchGraphTriangleMesh();
		/// Destructor
		~PatchGraphTriangleMesh();

		/// Compute triangle mesh for a single frame and single patch graph
		void createMesh();
		/// Extrapolate mesh from single frame to complete patch graph (assumes existing single frame patchGraphIndex/frameIndex)
		void extrapolate();
		void doOriginalExtrapolate();
		/// Detect border points
		void detectBorderPoints();
		/// Clip triangle mesh at border points
		void clipAtBorder();
		/// Smooth out normals
		void smoothNormals();

		/// Debug visualization
		void drawGL();

		/*** MEMBER ACCESS ***/
		inline void setContext( AnimationBSRContext* context ){ this->context = context; }
		void setScene( Scene* scene );

};
}

#endif