//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TGNeighborhoodKNNSearch.h"
//---------------------------------------------------------------------------
#include "InCorePCTopologyGraph.h"
#include "PatchGraph.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------


TGNeighborhoodKNNSearch::TGNeighborhoodKNNSearch(const PointSet *ps, const InCorePCTopologyGraph *tg, card32 frame, card32 startIndex, const string channel3f) {
   this->tg = tg;
   this->ps = ps;
   channelToken = ps->getAAT(channel3f);
   reset(frame, startIndex);
}

void TGNeighborhoodKNNSearch::reset(card32 frame, card32 startIndex) {
   this->frame = frame;
   pos = ps->get<float32, 3>(frame, startIndex, channelToken);
	neighbors = priority_queue<QueueEntry>();
	visited.clear();
   QueueEntry q(0, startIndex);
	neighbors.push(q);
   visited.insert(startIndex);
}

bool TGNeighborhoodKNNSearch::atEnd() {
   return neighbors.empty();
}

void TGNeighborhoodKNNSearch::next() {
   pAssert(!neighbors.empty());
   card32 index = neighbors.top().index;
   neighbors.pop();
   const card32 n = tg->getNumAdjacentVertices(index);
   for (card32 i=0; i<n; i++) {
      card32 j = tg->getAdjacentVertex(index, i);
      if (visited.find(j) == visited.end()) {
		   Vector3f npos = ps->get<float32, 3>(frame, j, channelToken);
		   QueueEntry q(norm(pos-npos), j);
         neighbors.push(q);
         visited.insert(j);
      }
   }
}

card32 TGNeighborhoodKNNSearch::getCurrentIndex() {
   pAssert(!neighbors.empty());
   return neighbors.top().index;
}


// ----


GeneralPGNeighborhoodKNNSearch::GeneralPGNeighborhoodKNNSearch(const PatchGraph *pg, card32 frame, card32 startIndex, const string channel3f) {
   this->pg = pg;
   channelToken = pg->getTrajectorySet()->getTrajectories()->getAAT(channel3f);
   reset(frame, startIndex);
}

void GeneralPGNeighborhoodKNNSearch::reset(card32 frame, card32 startIndex) {
	this->frame = frame;
   this->startIndex = startIndex;
	neighbors = priority_queue<QueueEntry>();
	visited.clear();
   QueueEntry q(0, startIndex);
	neighbors.push(q);
   visited.insert(startIndex);
}

bool GeneralPGNeighborhoodKNNSearch::atEnd() {
   return neighbors.empty();
}

void GeneralPGNeighborhoodKNNSearch::next() {
   pAssert(!neighbors.empty());
   card32 index = neighbors.top().index;
   neighbors.pop();
   const card32 n = pg->getTopology()->getNumAdjacentVertices(index);
   for (card32 i=0; i<n; i++) {
      card32 j = pg->getTopology()->getAdjacentVertex(index, i);
      if (visited.find(j) == visited.end()) {

		   //Vector3f npos = pg->getTrajectorySet()->getTrajectories()->get<float32, 3>(frame, j, channelToken);
		   float32 dist = calcDistance(pg, frame, startIndex, j, channelToken);
		   QueueEntry q(dist, j);
         neighbors.push(q);
         visited.insert(j);
      }
   }
}

card32 GeneralPGNeighborhoodKNNSearch::getCurrentIndex() {
   pAssert(!neighbors.empty());
   return neighbors.top().index;
}


// --


PGNeighborhoodKNNSearch::PGNeighborhoodKNNSearch(const PatchGraph *pg, card32 frame, card32 startIndex, const string channel3f, bool averageDist) 
: GeneralPGNeighborhoodKNNSearch(pg, frame, startIndex, channel3f)
{
   this->averageDist = averageDist;
}

float32 PGNeighborhoodKNNSearch::calcDistance(const PatchGraph *pg, card32 frame, card32 patch1, card32 patch2, const AAT &channelToken) {
   const PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   if (!averageDist) {
      return norm(ps->get<float32, 3>(frame, patch1, channelToken) - ps->get<float32, 3>(frame, patch2, channelToken));
   } else {
      float32 dist = 0;
      const card32 numFrames = ps->getWidth();
      for (card32 i=0; i<numFrames; i++) {
         dist += norm(ps->get<float32, 3>(i, patch1, channelToken) - ps->get<float32, 3>(i, patch2, channelToken));
      }
      return dist / numFrames;
   }
}