//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TriangleMeshTexture.h"
#include "ClassMethods.h"
#include "qtgui/QFileDialog.h"
#include "PointSetKNNQuery.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include <stack>
#include "GLShaderMaterial.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

static string describeType( card32 q ) {
	if ( q == REGION_GROW ) {
		return "Region growing";
	} else if ( q == MANUAL_SELECTION ) {
		return "Manual selection";
	} else if ( q == DISTANCE ) {
		return "Distance";
	} else {
		return "UNDEFINED";
	}
}

IMPLEMENT_GEOXL_CLASS( TriangleMeshTexture, 0 )
{
	BEGIN_CLASS_INIT( TriangleMeshTexture )
	ADD_CARD32_PROP_UPDATE_DESCR( usePoints, 0, NULL, &describeType, 3 )
	ADD_STRING_PROP( textureFilename, 0 )
	ADD_STRING_PROP( meshNode, 0 )
	ADD_CARD32_PROP( seedPointIndex, 0 )
	ADD_VECTOR3F_PROP( up, 0 )
	ADD_NOARGS_METHOD( pickTexture )
	ADD_NOARGS_METHOD( applyTexture )
}

/** 
 * Constructor
 */
TriangleMeshTexture::TriangleMeshTexture()
{
	textureFilename = "D:/LocalProject/data/Pics/siggraphLogo.bmp";
	meshNode = "root/pg0Frame0";
	seedPointIndex = 0;
	up = makeVector3f( 1,0,0 );
	usePoints = MANUAL_SELECTION;
}

/** 
 * Destructor
 */
TriangleMeshTexture::~TriangleMeshTexture()
{
}

/** 
 * Load Texture to image; return true on success
 */
bool TriangleMeshTexture::getTexture( QImage& image )
{
	QString fname = QString::fromStdString( textureFilename );
	QFileInfo info( fname );
	if ( !info.exists() ){
		return false;
	}
	return image.load( fname );
}

/** 
 * Interpolate the color from the texture image
 */
Vector3f TriangleMeshTexture::getColor( Vector2f p, const QImage& texture )
{
	if ( p[0] < 0 ) p[0] = 0;
	if ( p[0] > 1 ) p[0] = 1;
	p[0] = 1-p[0];
	if ( p[1] < 0 ) p[1] = 0;
	if ( p[1] > 1 ) p[1] = 1;
	QRgb rgb = texture.pixel( p[0] * ( texture.width() - 1 ), p[1] * ( texture.height() - 1 ) );
	return makeVector3f( qRed( rgb ) / 255.0f, qGreen( rgb ) / 255.0f, qBlue( rgb ) / 255.0f );
}

/** 
 * Pick a texture file
 */
void TriangleMeshTexture::pickTexture()
{
	QString filename = QFileDialog::getOpenFileName( NULL, "Choose a file", "/home", "Images (*.png *.xpm *.jpg)" );
	if( !filename.isNull() ){
		textureFilename = filename.toStdString();
	}
}

/***
 * Select points used in texturizing process
 */
bool TriangleMeshTexture::selectPoints( PointSet* vertices, vector<card32>& points )
{
	AAT POSITION = vertices->getAAT( "position" );
	AAT FLAGS = vertices->getAAT( "flags" );
	AAT NORMAL = vertices->getAAT( "normal" );
	if ( usePoints == REGION_GROW ){
		points.clear();
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( vertices );
		PointSetKNNQuery knn( vertices, octree, NULL_VECTOR3F, POSITION );
		// Reset all point marks
		for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
			vertices->set1i( pointIndex, FLAGS, 0 );
		}
		// Find points that will be used in texture mapping
		card32 nIndex;
		bool found;
		stack<card32> stack;
		stack.push( seedPointIndex );
		points.push_back( seedPointIndex );
		Vector3f referenceNormal = vertices->get3f( seedPointIndex, NORMAL );
		while( !stack.empty() ){
			// Find nearest neighbors
			card32 index = stack.top();
			stack.pop();
			Vector3f pos = vertices->get3f( index, POSITION );
			knn.setTargetPoint( pos );
			for ( unsigned int n = 0; n < 5; n++ ){
				knn.getNextNearestPointIndex( nIndex, found );
				// Test if nIndex has been visited already
				if ( vertices->get1i( nIndex, FLAGS ) == 0 ){
					// Test if nIndex fulfills normal criterion
					if ( vertices->get3f( nIndex, NORMAL ) * referenceNormal > 0.8f ){
						// TODO: Test if nIndex fulfills distance criterion
						stack.push( nIndex );
						vertices->set1i( nIndex, FLAGS, 1 );
						points.push_back( nIndex );
					}
				}
			}
		}
		// Reset flags
		for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
			vertices->set1i( pointIndex, FLAGS, 0 );
		}
		delete octree;
		return true;
	} else if ( usePoints == MANUAL_SELECTION ){
		points.clear();
		for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
			if ( vertices->get1i( pointIndex, FLAGS ) != 0 ){
				points.push_back( pointIndex );
			}
		}
		return true;
	} else {
		debugOutput << "selectPoints: not implemented yet.\n"; debugOutput.flush();
		return false;
	}
	return false;
}

/** 
 * Apply texture
 */
void TriangleMeshTexture::applyTexture()
{
	// Find triangle mesh
	UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, meshNode ) );
	if ( mesh == NULL ){
		debugOutput << "applyTexture: can't find triangle mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* vertices = mesh->getPointSet();
	AAT POSITION = vertices->getAAT( "position" );
	AAT COLOR = vertices->getAAT( "color" );
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT FLAGS = vertices->getAAT( "flags" );

	if ( seedPointIndex >= vertices->getNumEntries() ){
		debugOutput << "applyTexture: invalid seed point index.\n"; debugOutput.flush();
		return;
	}
	
	Vector3f referenceNormal = vertices->get3f( seedPointIndex, NORMAL );

	if ( referenceNormal * normalize( up ) > 0.9f ){
		debugOutput << "applyTexture: invalid up vector.\n"; debugOutput.flush();
		return;
	}

	// Select points used for texture mapping
	vector<card32> points;
	selectPoints( vertices, points );

	// Compute parametrization in plane defined by seed points
	Matrix3f T;
	T[0] = normalize( up );
	T[2] = normalize( referenceNormal );
	T[1] = T[2].crossProduct( T[0] );
	T[0] = T[1].crossProduct( T[2] );
	
	BoundingBox<float32,2> bbox;
	bool first = true;
	Vector3f t = vertices->get3f( seedPointIndex, POSITION );
	vector<Vector2f> points2D; points2D.resize( points.size() );
	for ( unsigned int n = 0; n < points.size(); n++ ){
		card32 index = points[n];
		Vector3f p3D = T.transpose() * ( vertices->get3f( index, POSITION ) - t );
		Vector2f p2D = makeVector2f( p3D[0], p3D[1] );
		points2D[n] = p2D;
		if ( first ){ bbox = BoundingBox<float32,2>( p2D ); first = false; }
		else bbox.addPoint( p2D );
	}

	// Transformation to [0;1]^2
	float32 scale = 1.0f / bbox.getMaxSideLength();
	Vector2f trans = bbox.lowerCorner;

	QImage texture;
	if ( !getTexture( texture ) ){
		debugOutput << "applyTexture: cannot load texture.\n"; debugOutput.flush();
	} else {
		// Mark selected points
		for ( unsigned int n = 0; n < points.size(); n++){
			card32 index = points[n];
			Vector2f p2D = ( points2D[n] - trans ) * scale;
			Vector3f color = getColor( p2D, texture );
			vertices->set3f( index, COLOR, color );
		}
	}

	// Rebuild render objects
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
}