//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "UnifyNormals.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "VertexDescriptor.h"
#include "TrajectorySet.h"
#include "PatchTypes.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
#include <set>
#include <queue>
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS(UnifyNormals, 0)
{
	BEGIN_CLASS_INIT(UnifyNormals)
}


// --


UnifyNormals::UnifyNormals() {
}

void UnifyNormals::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

struct Entry {
	card32 i;
	float32 w;
	
	Entry() {}
	inline Entry(card32 ni, float32 nw) {i=ni; w=nw;}
	bool operator<(const Entry &e) const {return w < e.w;}
};

void UnifyNormals::unifyNormals(PatchGraph *pg, const BSROptVariableIndexSet *indices) 
{
	set<BSROptPointIndex> activeIndices;
	if (indices) {
		for (unsigned i=0; i<indices->size(); i++) {
			activeIndices.insert((*indices)[i]);
		}
	}
	TrajectorySet *ts = pg->getTrajectorySet();
	InCorePCTopologyGraph *tg = pg->getTopology();
	PointSet* traj = pg->getTrajectorySet()->getTrajectories();
	const card32 numFrames = ts->getNumFrames();
	const card32 numPatches = ts->getNumPatches();
	const AAT NORMAL = traj->getAAT( "normal" );
	for (card32 frame=0; frame<numFrames; frame++) {
		set<card32> unprocessedPatches;
		for (card32 patch=0; patch<numPatches; patch++) {
			bool insert = true;
			if (indices) {
				BSROptPointIndex e(frame, patch);
				if (activeIndices.find(e) == activeIndices.end()) {
					insert = false;
				}
			}
			if (insert) {
				unprocessedPatches.insert(patch);
			}
		}
		while (!unprocessedPatches.empty()) {
			priority_queue<Entry> patchesQueue;
			patchesQueue.push(Entry(*unprocessedPatches.begin(), 0));
			unprocessedPatches.erase(unprocessedPatches.begin());
			while (!patchesQueue.empty()) {
				card32 nextPatch = patchesQueue.top().i;
				patchesQueue.pop();

				const Vector3f thisNormal = traj->get<float32, 3>(frame, nextPatch, NORMAL);

				const card32 numNeighbors = tg->getNumAdjacentVertices(nextPatch);
				for (card32 neighbor=0;	neighbor<numNeighbors; neighbor++) {
					const card32 neighborIndex = tg->getAdjacentVertex(nextPatch, neighbor);
					if (unprocessedPatches.find(neighborIndex) != unprocessedPatches.end()) {
						const Vector3f otherNormal = traj->get<float32, 3>(frame, neighborIndex, NORMAL);
						if (thisNormal * otherNormal < 0) {
							traj->set3f(frame, neighborIndex, NORMAL, -otherNormal);
						}
						unprocessedPatches.erase(neighborIndex);
						patchesQueue.push(Entry(neighborIndex, 1.0f - fabs(thisNormal * otherNormal)));
					}
				}
			}
		}
		/// Find patch with smallest x-value for global outside orientation
		PointSet* ps = pg->getTrajectorySet()->copySingleFrame( frame );
		
		AAT POSITION = ps->getAAT( "position" );
		AAT NORMAL = ps->getAAT( "normal" );

		if ( frame == 0 ){
			AttachedIndexedOctree *octree = PointSetKNNQuery::createOctree( ps );
			PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, POSITION );
			card32 nIndex;
			bool found;
			knn.setTargetPoint( makeVector3f( -10000000, 0, 0 ) );
			knn.getNextNearestPointIndex( nIndex, found );
			if ( found ){
				Vector3f normal = ps->get3f( nIndex, NORMAL );
				Vector3f pos = ps->get3f( nIndex, POSITION );
				if ( normal * makeVector3f( -1, 0, 0 ) < 0 ){	
					// Flip all normals in frame
					for (card32 patch=0; patch<numPatches; patch++) {
						traj->set3f( frame, patch, NORMAL, -traj->get3f( frame, patch, NORMAL ) );
					}
				}
			} else {
				debugOutput << "unifyNormals: cannot unify normals.\n"; debugOutput.flush();
			}
			delete ps;
			delete octree;
		} else {
			if ( traj->get3f( frame, 0, NORMAL ) * traj->get3f( frame-1, 0, NORMAL ) < 0 ){
				for (card32 patch=0; patch<numPatches; patch++) {
					traj->set3f( frame, patch, NORMAL, -traj->get3f( frame, patch, NORMAL ) );
				}
			}

		}
	}
}


void UnifyNormals::flipNormals(PatchGraph *pg) {
	PointSet *traj = pg->getTrajectorySet()->getTrajectories();
	const card32 n = traj->getNumEntries();
	const AAT NORMAL = traj->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
	for (card32 i=0; i<n; i++) {
		traj->set3f(i, NORMAL, -traj->get3f(i, NORMAL));		
	}
}

UnifyNormals::~UnifyNormals() {
}

