//---------------------------------------------------------------------------
#ifndef BSROptContextH
#define BSROptContextH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {
   using namespace std;

class PGTopologicalNeighborhood;
class PGTopologicalNeighborhoodDataPts;


class ANIMATIONRECONSTRUCTION_API BSROptContext : public Persistent {
	GEOXL_CLASS( BSROptContext );

 public:
	static	const	card32	MODE_MOVE_POINTS						= 0;
	static	const	card32	MODE_FIT_COEFFICIENTS					= 1;
	
	static	const	card32	SOLVER_GRADIENT_DESCENT					= 0;
	static	const	card32	SOLVER_NEWTON_CG						= 1;
	static	const	card32	SOLVER_GAUSS_NEWTON_CG					= 2;
	static	const	card32	SOLVER_QUASI_NEWTON_CG					= 3;
	static	const	card32	SOLVER_NONLINEAR_CG						= 4;

   /// stop after a fixed number of steps (numIterations)
   static const card32 STOP_NUM_ITERATIONS = 0;
   /// stop after convergence, but no more than numIterations
   static const card32 STOP_CONV_GRAD_INFNORM = 1;

	// multigrid
	static const card32 MG_NO_MULTIGRID = 0;
	static const card32 MG_INTERPOLATION = 1;
	static const card32 MG_LONG_EDGES = 2;


   //  *** NUMERICS ***
   /// mode of operation: what to optimize - coefficients or point positions?
	card32						optMode;
	/// numerical solver
	card32						solver;
	/// when to stop iterating
	card32						stoppingCriterion;
	/// Step size in the gradient descent method
	float32						stepSize;
	/// Number of iterations in the optimization procedure
	card32						numIterations;
	/// threshold for convergence (depends on stoppingCriterion)
	float32						convergenceEps;
	/// update frequency for coefficients (corr, hfield), sample spacing 
	card32						coeffsUpdateFreq;
	/// update frequency for local coordinate systems
	card32						coordsUpdateFreq;
	/// use multi-grid solver?
	card32                  multiGrid;
	/// sample spacing for low resolution model in multi-grid technique
	float32						lowresSampleSpacing;
	/// add identity matrix with this weight to improve stability
	float32						newtonRegularization;
	

   // *** WEIGHTS ***
	/// Weight of the likelihood term
	float32						weightLikelihood;
	/// Weight of attraction to nearby points
	float32						weightICPLikelihood;
	/// Weight of smooth trajectories in time
	float32						weightSimpleSpaceTime;
	/// Weight of attraction towards center of nearby points
	float32						weightSimpleCenterLikelihood;
	/// Weight of the smoothness term
	float32						weightSmooth;
	/// Penalty of local monimial curvature (only for non-simpleSmoothness)
	float32						weightSmoothCurvature;
	/// Weight of the spatio-temporal prior
	float32						weightSTSmoothness;
   /// tries to make correspondence field locally linear
	float32						weightSTAngleSmoothness;
   /// tries to make correspondence field locally linear
	float32						weightCorrSmooth;
	/// Weight of the density prior
	float32						weightDensity;
	/// Weight of low resolution solution in multi-grid solver
	float32						weightMultires;
	/// exponent for stLengthWeighting
	float32                 edgeLengthWeightExponent;

	/************************************************************************/
	/* The following codes are added by peter for testing                                                          */
	/************************************************************************/
	/// Parameters for nonlinear conjugate gradient with Newton-Raphson and Fletcher-Reeves
	card32	maxIterations_CG_NCG_NRFR;
	card32	maxIterations_NR_NCG_NRFR;
	float32	errorTolerance_CG_NCG_NRFR;
	float32	errorTolerance_NR_NCG_NRFR;

	/// Parameters for Gauss-Newton method
	card32	maxIterations_Gauss_Newton;
	float32	errorTolerance_Gauss_Newton;
	
	/// Parameters related to features
	bool							usingFeatureValues;
	card32					numOfFeatureValues;
	vector<float32>	featureWeights;

	/// Parameters for Newton method
	card32	maxIterations_Newton;
	float32	errorTolerance_Newton;

	///The lambda parameter used in optimization
	float32	lambda;
	///The prunedist used in determing need smoothness flag
	float32	pruneDist_smoothness;
	/// The trade-off used in squared distance potential
	float32	tradeoff_sdf;
	///the percentage of overlapping region
	float32	overlapping_rate;
	///weight of squared distance term
	float32	weightSquaredDistance;
	/// weight of marker error potential
	float32	weightMarkerError;
	///weight of planar smoothness term
	float32	weightPlanarSmoothness;
	///weight of normal smoothness term
	float32	weightNormalSmoothness;
	///weight of circular smoothness term
	float32	weightCircularSmoothness;
	///weight of laplacian smoothness term
	float32	weightLaplacianSmoothness;
	///weight of position rigidity term
	float32	weightPositionRigidity;
	///weight of normal rigidity term
	float32	weightNormalRigidity;
	/// weight of position spatial temporal coherence
	float32	weightPositionCoherence;
	/// weight of normal spatial temporal coherence
	float32	weightNormalCoherence;
	/// weight of rotation matrix coherence
	float32	weightRotationCoherence;

	/// Parameters of line search
	float32	minimumEnergyDecay;
	float32	minimumStepsize;

	/************************************************************************/
	/* The above codes are added by peter for testing                                                             */
	/************************************************************************/
   // *** FLAGS ***
	/// attraction to just one data point or weighted normal direction attraction?
	bool simpleLikelihood;
   /// fix points to data (just write, no gradients)
	bool hardLikelihoodConstraint;
	/// simple spatial smoothness using projection to polynomial
	bool simpleSmoothness;
   /// direction of the spatio-temporal smoothness priors in between frames
	bool stSmoothnessForward;
   /// direction of the spatio-temporal smoothness priors in between frames
	bool stSmoothnessBackward;
   /// direction of the spatio-temporal smoothness priors between urshape and frames
	bool stSmoothnessUrshapeToFrames;
   /// direction of the spatio-temporal smoothness priors between urshape and frames
	bool stSmoothnessFramesToUrshape;
	/// linear correspondence field smoothing
	bool stCorrLinPrjSmoothing;
	/// preserve angles in mesh
	bool stPreserveAngles;
	/// alter length preservation constraint weights by edge length
	bool stLengthWeighting;
	/// apply spatial smoothing to patches (frames)
	bool smoothPatchGraph;
	/// apply spatial smoothing to urshape (latent rest state)
	bool smoothUrshape;
	/// unify normals after each coordinate system estimation (more expensive)
	bool unifyNormals;
	
	// *** NEIGHBORHOOD ***
	/// points to be considered for the data fitting potential
	PGTopologicalNeighborhoodDataPts *likelihoodNeighborhood;
	/// points to be considered for smoothing
	PGTopologicalNeighborhood *priorsNeighborhood;

   // *** linear system solver ***
   float32 cgStopRes;
   card32 cgMaxNumIterations;
   bool cgDebugging;

   // *** MISC ***
   bool verbose;

	/// Constructor
	BSROptContext();

   virtual void assign(const Object* obj, CopyContext *context = NULL);
	
	~BSROptContext();
		
};


}

#endif

