//---------------------------------------------------------------------------
#ifndef BSROptPotentialH
#define BSROptPotentialH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "DynamicLinearAlgebra.h"
#include "SparseLinearAlgebra.h"
#include "PatchGraph.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {
   using namespace std;

class BSROptContext;
class PointCloudAnimationSequence;
class PointSet;


// -------------------------------------------------------------------------------                                  
// -- helper classes -------------------------------------------------------------                                  
// -------------------------------------------------------------------------------                                  


/// this special frame number refers to the urshape
const int32 BSROPT_FRAME_URSHAPE = -1;


/// helper data structure - 2d point index (frame, point)   
struct ANIMATIONRECONSTRUCTION_API BSROptPointIndex {
   int32 frame;
   card32 point;
   
   inline BSROptPointIndex() {
      #ifdef _DEBUG
      frame = point = 0xEFEFEFEF;
      #endif
   }
   inline BSROptPointIndex(card32 frame, card32 point) {
      this->frame = frame; this->point = point;
   }
   bool operator<(const BSROptPointIndex &other) const {
		if (frame < other.frame) {
			return true;
		} else if (frame > other.frame) {
			return false;
		} else {
			return point < other.point;
		}
	}
};


/// set of variables to be optimized
typedef vector<BSROptPointIndex> BSROptVariableIndexSet;


// -------------------------------------------------------------------------------                                  
// -- main class -----------------------------------------------------------------                                  
// -------------------------------------------------------------------------------                                  

                                  
/// ---
/// --- Potential Function for Gradient Descent (Likelihood or Prior)
/// ---
class ANIMATIONRECONSTRUCTION_API BSROptPotential : public Persistent {
   GEOXL_ABSTRACT_CLASS( BSROptPotential );

 public:
   static const card32 FLAG_PATCH_SAMPLE_SPACING				= 1;
   static const card32 FLAG_PATCH_REF_FRAMES							= 2;
	static const card32 FLAG_NEED_PATCH_2ND_ORDER_HFIELD	= 4;
	static const card32 FLAG_NEED_ROTATION_MATRIX				= 8;
   
   // what to prepare...
   virtual card32 needPrecomputedChannels(BSROptContext *context) {return 0;}

   /// initialize the computation, clear internal caches if necessary
   virtual void initialize() {}
   /// this event notifies the potential function that the coordinate system has changed since the last iteration
   virtual void coordsChanged() {}
   
   /// compute objective function and gradient for all points in index set
   /// gradient returns the gradient vector (the indexing can be found in channel varIndex in the patch graph)
   /// potential returns the overall statistical potential (energy, log-likelihood,...)
   /// if perPointPotentials is not NULL, the potentials are returned for every single point
   /// all quantities are additive, new values are added to those present in the variables/vectors (dimension must have been set properly)
   ///
   /// --- quadApprox, linApprox:
   /// --- quadratic approximation: can be used instead of gradient descent 
   /// --- computes local quadratic approximation to the objective function for Newton-style optimization
   ///
   /// matrix "quad" stores the 2nd order terms, "lin" stores the linear terms, constants are neglected
   /// the resulting objective function is given by 
   ///
   ///    fun(x) = 0.5 * (x^t * QUAD * x)  +  (x^t * lin) + c
   ///
   /// the new estimate for the minimum (one Newton step) is given by
   ///
   ///    QUAD * x_min = - lin,  i.e.  x_min = -QUAD^(-1) * lin
   ///
   /// the current potential can be obtained by evaluating the quadratic approximation at x = x_current 
   /// (the model should be chosen so that this is exact, e.g. by a 2nd order Taylor approximation of the non-linear function)
   /// again, all quantities are additive, new values are added to those present in the matrices/vectors (dimension must have been set properly)
   /// set pointer to NULL if the quadratic model does not need to be computed
   virtual void computeGradient(
		const BSROptContext *context, 
		const BSROptVariableIndexSet *indices,
      const PointCloudAnimationSequence *data, 
		PatchGraph *patchGraph,
      DVectorF &gradient, 
		float32 &potential, 
		DVectorF *perPointPotentials, 
		SparseMatrixF *quadApprox, 
		DVectorF *linApprox) = 0;

	/// The following method is added by peter
	/// It is similar with computeGradient
	virtual	void		peter_UpdateGradientandHessian(
		const BSROptContext *context, 
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors, 
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox, 
		DVectorF *linApprox) = 0;

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) = 0;
   /// the computation has been completed, all internal caches can be disposed
   virtual void finished() {}
};


// -------------------------------------------------------------------------------                                  
// -- child classes: different variants of BSROptPotential -----------------------                                  
// -------------------------------------------------------------------------------                                  

// -- Point-based potentials (gradients of points)

/// ---
/// --- Gaussian Likelihood Function
/// ---
class ANIMATIONRECONSTRUCTION_API BOPGaussianPrjLikelihood : public BSROptPotential {
   GEOXL_CLASS( BOPGaussianPrjLikelihood )

 public:
   virtual card32 needPrecomputedChannels(BSROptContext *context);

   virtual void computeGradient(
      const BSROptContext *context,
      const BSROptVariableIndexSet *indices,
      const PointCloudAnimationSequence *data,
      PatchGraph *patchGraph,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
   );

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) {}
};


/// ---
/// ---  Point-to-Plane ICP Likelihood Function (very simple, just a test function)
/// ---
class ANIMATIONRECONSTRUCTION_API BOPSimpleICPLikelihood : public BSROptPotential {
   GEOXL_CLASS( BOPSimpleICPLikelihood )

 public:
   virtual card32 needPrecomputedChannels(BSROptContext *context);

   virtual void computeGradient(
      const BSROptContext *context,
      const BSROptVariableIndexSet *indices,
      const PointCloudAnimationSequence *data,
      PatchGraph *patchGraph,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
   );

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) {}
};

/// ---
/// ---  Simple likelihood-term; center of associated data points
/// ---
class ANIMATIONRECONSTRUCTION_API BOPSimpleCenterLikelihood : public BSROptPotential {
	GEOXL_CLASS( BOPSimpleCenterLikelihood )

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);
	virtual void computeGradient( const BSROptContext *context, const BSROptVariableIndexSet *indices, const PointCloudAnimationSequence *data, PatchGraph *patchGraph, DVectorF &gradient, float32 &potential, DVectorF *perPointPotentials, SparseMatrixF *quadApprox, DVectorF *linApprox );
	virtual	void peter_UpdateGradientandHessian(const BSROptContext *context, const	PointCloudAnimationSequence *data, PatchGraph *patchGraph, const	card32	numOfNeighbors, const	int	*neighbor_buffer, SparseMatrixF *quadApprox, DVectorF *linApprox){}
	/// Compute energies
	virtual	void peter_ComputeEnergy(const BSROptContext *context, const PointCloudAnimationSequence *data, PatchGraph *patchGraph, const card32 numOfNeighbors, const int *neighbor_buffer, const DVectorF deviatioon, float &energy, bool m_flag ) {}
};

/// ---
/// ---  Simple space-time-prior; required for testing
/// ---
class ANIMATIONRECONSTRUCTION_API BOPSimpleSpaceTime : public BSROptPotential {
	GEOXL_CLASS( BOPSimpleSpaceTime )

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);
	virtual void computeGradient( const BSROptContext *context, const BSROptVariableIndexSet *indices, const PointCloudAnimationSequence *data, PatchGraph *patchGraph, DVectorF &gradient, float32 &potential, DVectorF *perPointPotentials, SparseMatrixF *quadApprox, DVectorF *linApprox );
	virtual	void peter_UpdateGradientandHessian(const BSROptContext *context, const	PointCloudAnimationSequence *data, PatchGraph *patchGraph, const	card32	numOfNeighbors, const	int	*neighbor_buffer, SparseMatrixF *quadApprox, DVectorF *linApprox){}
	/// Compute energies
	virtual	void peter_ComputeEnergy(const BSROptContext *context, const PointCloudAnimationSequence *data, PatchGraph *patchGraph, const card32 numOfNeighbors, const int *neighbor_buffer, const DVectorF deviatioon, float &energy, bool m_flag ) {}
};



/// ---
/// --- Spatial smoothness function
/// ---
class ANIMATIONRECONSTRUCTION_API BOPPrjSmoothness : public BSROptPotential {
	GEOXL_CLASS( BOPPrjSmoothness )

 private:
   // used for both patchGraph and urshape
   inline void computeSinglePotential(
      const BSROptContext *context,
      PointSet *ps,
      InCorePCTopologyGraph *tg,
      const card32 frameIndex,
      const card32 patchIndex,
      const AAT &POSITION,
      const AAT &VARINDEX, const AAT &COEFF,
      const AAT &REF_U, const AAT &REF_N,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
   );

 public:
   virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
	);

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) {}
};


/// ---
/// --- spatial-temporal rigidity prior, based on simple spring model
/// ---
class ANIMATIONRECONSTRUCTION_API BOPSpringBasedRigidity : public BSROptPotential {
	GEOXL_CLASS( BOPSpringBasedRigidity )

 private:
   // used for both patchGraph and urshape
   inline void computeSinglePotential(
      const BSROptContext *context,
      const AAT &REF_U,
	   const AAT &REF_N,
	   const AAT &POSITION,
	   const AAT &VARINDEX,
	   PointSet *sourcePS,
	   PointSet *targetPS,
	   InCorePCTopologyGraph *tg,
	   const card32 sourceFrame,
	   const card32 targetFrame,
	   const card32 patchIndex,	
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox,
      float32 generalLinkWeight = 1.0f
   );

 public:
   virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
	);

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) {}
};

class ANIMATIONRECONSTRUCTION_API BOPSquaredDistancePotential : public BSROptPotential {
	GEOXL_CLASS( BOPSquaredDistancePotential )

private:
	
public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPMarkerErrorPotential : public BSROptPotential {
	GEOXL_CLASS( BOPMarkerErrorPotential )

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPPlanarSmoothnessPotential : public BSROptPotential {
	GEOXL_CLASS( BOPPlanarSmoothnessPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPNormalSmoothnessPotential : public BSROptPotential {
	GEOXL_CLASS( BOPNormalSmoothnessPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energies
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPCircularSmoothnessPotential : public BSROptPotential {
	GEOXL_CLASS( BOPCircularSmoothnessPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPLaplacianSmoothnessPotential : public BSROptPotential {
	GEOXL_CLASS( BOPLaplacianSmoothnessPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPPositionRigidityPotential : public BSROptPotential {
	GEOXL_CLASS( BOPPositionRigidityPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);
};

class ANIMATIONRECONSTRUCTION_API BOPNormalRigidityPotential : public BSROptPotential {
	GEOXL_CLASS( BOPNormalRigidityPotential )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);
	
	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox);

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag);

};

class ANIMATIONRECONSTRUCTION_API BOPSTPositionCoherence : public BSROptPotential {
	GEOXL_CLASS( BOPSTPositionCoherence )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag){}
};

class ANIMATIONRECONSTRUCTION_API BOPSTNormalCoherence : public BSROptPotential {
	GEOXL_CLASS( BOPSTNormalCoherence )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag){}
};

class ANIMATIONRECONSTRUCTION_API BOPSTRotationCoherence : public BSROptPotential {
	GEOXL_CLASS( BOPSTRotationCoherence )

private:

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);

	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		){}

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag){}
};


class ANIMATIONRECONSTRUCTION_API BOPMultiResPotential : public BSROptPotential {
	GEOXL_CLASS( BOPMultiResPotential )

public:
	virtual card32 needPrecomputedChannels(BSROptContext *context);
	
	virtual void computeGradient(
		const BSROptContext *context,
		const BSROptVariableIndexSet *indices,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		DVectorF &gradient,
		float32 &potential,
		DVectorF *perPointPotentials,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox
		);

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox) {}

	/// Compute energy
	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag) {}
};

// -- Patch-based potentials (gradients of coefficients)


/// ---
/// --- Least-Squares Fitting of Data Points (Likelihood)
/// ---
class ANIMATIONRECONSTRUCTION_API BOPDataLeastSquares : public BSROptPotential {
   GEOXL_CLASS( BOPDataLeastSquares )
 private:               
   struct LocalLSSystem {
      StaticMatrix<float32, 6, 6> mat;
      StaticVector<float32, 6> rhs;
   };
   vector< LocalLSSystem > cachedLSS;
   
   void computeLSS(
      const BSROptContext *context,
      const BSROptVariableIndexSet *indices,
      const PointCloudAnimationSequence *data,
      const PatchGraph *patchGraph
   );

   
 public:
   virtual void initialize();
   virtual void coordsChanged();

   virtual card32 needPrecomputedChannels(BSROptContext *context);

   virtual void computeGradient(
      const BSROptContext *context,
      const BSROptVariableIndexSet *indices,
      const PointCloudAnimationSequence *data,
      PatchGraph *patchGraph,
      DVectorF &gradient,
      float32 &potential,
      DVectorF *perPointPotentials,
      SparseMatrixF *quadApprox,
      DVectorF *linApprox
   );

	virtual	void		peter_UpdateGradientandHessian(const BSROptContext *context,
		const	PointCloudAnimationSequence *data,
		PatchGraph *patchGraph,
		const	card32	numOfNeighbors,
		const	int	*neighbor_buffer,
		SparseMatrixF *quadApprox,
		DVectorF *linApprox){}

	virtual	void		peter_ComputeEnergy(const BSROptContext *context,
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	card32		numOfNeighbors,
		const	int				*neighbor_buffer,
		const	DVectorF	deviatioon,
		float						&energy,
		bool							m_flag){}

   virtual void finished();
};



// -- more helper classes --------------------------------------------------------


/// ---
/// --- write index channel (indexing gradient/hessian variables)
/// --- read and write back channels by index
/// ---
class ANIMATIONRECONSTRUCTION_API BOPIndexingChannel {
 public:
   /// write index in data vector to channel "varIndex"
   /// for each entry in indices, the index is incremented by _one_
	static void writeIndexingChannel(
		const BSROptVariableIndexSet *indices,
		PatchGraph *patchGraph
	);

	/************************************************************************/
	/* The following codes are added by Peter                                                                         */
	/************************************************************************/
	static	void peter_writeIndexingChannel(
		const card32 offset, const std::string name, 
		const BSROptVariableIndexSet *indices,
		PatchGraph	*patchGraph
		);

	static void peter_readChannel(
		const BSROptVariableIndexSet *indices,
		const PatchGraph		*patchGraph,
		const	card32			offset,
		const string				channel,
		const card32				channelDim,
		DVectorF					&result
		);

	/// copy DVectorF to channel in the order of "indices"
	static void peter_writeChannel(
		const BSROptVariableIndexSet *indices,
		PatchGraph			*patchGraph,
		const card32			offset,
		const string			channel,
		const card32			channelDim,
		const DVectorF		&data
		);
	/************************************************************************/
	/* The above codes are added by Peter*/
	/************************************************************************/
	/// copy channel to DVectorF in the order of "indices"
	static void readChannel(
		const BSROptVariableIndexSet *indices,
		const PatchGraph *patchGraph,
		const string channel,
		const card32 channelDim,
		DVectorF &result
	);

	/// copy DVectorF to channel in the order of "indices"
	static void writeChannel(
		const BSROptVariableIndexSet *indices,
		PatchGraph *patchGraph,
		const string channel,
		const card32 channelDim,
		const DVectorF &data
	);
};




}

#endif