//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiResolutionPGOpt.h"
//---------------------------------------------------------------------------
#include "PatchGraphTopologicalResampling.h"
#include "PGLaplacianDiffusion.h"
#include "Timer.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( MultiResolutionPGOpt ,0)
{
	BEGIN_CLASS_INIT( MultiResolutionPGOpt );
}


// --


MultiResolutionPGOpt::MultiResolutionPGOpt() {
	data = NULL;
	pg = NULL;
	lowres = NULL;
	resampler = new PatchGraphTopologicalResampling();
	resampler->setup(true, 0);
	debugVis = true;
	afterEachIteration = NULL;
	afterEachIterationContext = NULL;
	indices = NULL;
}

MultiResolutionPGOpt::~MultiResolutionPGOpt() {
	delete resampler;
}

void MultiResolutionPGOpt::afterEachIterationCallback(void *context) {
	//MultiResolutionPGOpt *cthis = (MultiResolutionPGOpt*)context;
	//if (cthis->debugVis) {
	//	cthis->writeHighres();
	//	cthis->afterEachIteration(cthis->afterEachIterationContext);	
	//}
}

void MultiResolutionPGOpt::extrapolateHighres(bool directWrite) {
	const PointSet *ps_low = lowres->getTrajectorySet()->getTrajectories();
	const PointSet *urshape_low = lowres->getUrshape();
	PointSet *ps_full = pg->getTrajectorySet()->getTrajectories();
	PointSet *urshape_full = pg->getUrshape();
	
	const AAT posAAT_low = ps_low->getAAT("position");
	const AAT posAAT_full = ps_full->getAAT("position");
	const AAT posInterpolateAAT_full = ps_full->getAAT("lowresInterpPos");
	
	for (unsigned i=0; i<indices->size(); i++) {
		int32 f = (*indices)[i].frame;
		card32 p = (*indices)[i].point;
		Vector3f iPos = NULL_VECTOR3F;
		const PointSet *activePS_low = ps_low;
		PointSet *activePS_full = ps_full;
		int32 activeFrame = f;
		if (f == BSROPT_FRAME_URSHAPE) {
			activePS_low = urshape_low; 
			activePS_full = urshape_full; 
			activeFrame = 0;
		}
		for (card32 wi=0; wi<interpolationWeights[p].size(); wi++) {
			iPos +=   activePS_low->get<float32,3>(activeFrame, interpolationWeights[p][wi].first, posAAT_low)
			        * interpolationWeights[p][wi].second;
		}			  
		if (directWrite) {
			activePS_full->set<float32,3>(activeFrame, p, posAAT_full, iPos);
		} else {
			activePS_full->set<float32,3>(activeFrame, p, posInterpolateAAT_full, iPos);
		}
	}
}


void MultiResolutionPGOpt::setupAndBuildLowresTopology
  (PointCloudAnimationSequence *data,
   PatchGraph *patchGraph,
   float32 lowresDist,
   vector<bool> *lowResRemovablePatches,
   bool debugVis)
{
	Timer t;
	debugOutput << "Multigrid solver: compute lowres approx.";
	debugOutput.flush();
	clusters.clear();
	this->data = data;
	this->pg = patchGraph;
	this->debugVis = debugVis;
	this->lowres = resampler->fullResampling(patchGraph, lowresDist, &clusters, lowResRemovablePatches);
	inverseClusterMapping.resize(patchGraph->getTrajectorySet()->getNumPatches(), -1);
	for (card32 c=0; c<clusters.size(); c++) {
		for (card32 ni=0; ni<clusters[c].size(); ni++) {
			card32 ind = clusters[c][ni];
			pAssert(ind < inverseClusterMapping.size());
			inverseClusterMapping[ind] = c;
		}
	}
	debugOutput << " --- " << t.getDeltaValue() << " ms\n";
	debugOutput << "Multigrid solver: computeInterpolationMap.";
	debugOutput.flush();
	resampler->computeInterpolationMap(
		lowres, patchGraph, clusters, lowresDist,
		lowres->getTrajectorySet()->getTrajectories()->getAAT("position"),
		patchGraph->getTrajectorySet()->getTrajectories()->getAAT("position"),
		lowres->getTrajectorySet()->getTrajectories()->getAAT("originalIndex"),
		interpolationWeights
	);
	//if (!directInterpolation) {
	pg->createPatchAttribute(mVAD("lowresInterpPos", 3, VAD::DATA_FORMAT_FLOAT32));
	//}
	debugOutput << " --- " << t.getDeltaValue() << " ms\n\n";
	debugOutput.flush();
}

void MultiResolutionPGOpt::calcLowresIndices(const BSROptVariableIndexSet *indices, BSROptVariableIndexSet &lowresIndices) {
	set< BSROptPointIndex > selected;
	for (unsigned i=0; i<indices->size(); i++) {
		int32 f = (*indices)[i].frame;
		card32 p = (*indices)[i].point;
		pAssert(p<inverseClusterMapping.size());
		card32 reduced = inverseClusterMapping[p];
		if (reduced != -1) {
			selected.insert(BSROptPointIndex(f, reduced));
		}
	}
	lowresIndices.clear();
	lowresIndices.reserve(selected.size());
	set< BSROptPointIndex >::iterator pos = selected.begin();
	set< BSROptPointIndex >::iterator end = selected.end();
	while (pos != end) {
		lowresIndices.push_back(*pos);	
		++pos;
	}
}

void MultiResolutionPGOpt::twoResolutionOpt
  (PatchGraphBSROpt *optimizer, 
   const BSROptVariableIndexSet *indices,
   PatchGraphBSROpt::FeedbackCallback afterEachIteration,
   void *callbackContext)
{
	this->indices = indices;
	this->afterEachIteration = afterEachIteration;
	this->afterEachIterationContext = callbackContext;
	Timer t;
	debugOutput << "Multigrid solver: compute lowres solution.";
	BSROptVariableIndexSet lowresIndices;
	calcLowresIndices(indices, lowresIndices);
	optimizer->compute(&lowresIndices, data, lowres, afterEachIterationCallback, this);
	debugOutput << " --- " << t.getDeltaValue() << " ms\n";
	debugOutput << "extrapolate to full res.";
	debugOutput.flush();
	extrapolateHighres(false);
	debugOutput << " --- " << t.getDeltaValue() << " ms\n";
	debugOutput << "solve full res regularized by interpolated lowres solution.";
	debugOutput.flush();
	optimizer->addPotential(BOPMultiResPotential::getClass());
	optimizer->compute(indices, data, pg, afterEachIteration, callbackContext);
	debugOutput << " --- " << t.getDeltaValue() << " ms\n";
	//debugOutput << "solve full res.";
	//debugOutput.flush();
	optimizer->removePotential(BOPMultiResPotential::getClass());
	//optimizer->compute(indices, data, pg, afterEachIteration, callbackContext);
	//debugOutput << " --- " << t.getDeltaValue() << " ms\n\n";
	//debugOutput.flush();
	this->indices = NULL;
}


//void MultiResolutionPGOpt::longEdgesOpt
//  (PatchGraphBSROpt *optimizer, 
//   const BSROptVariableIndexSet *indices,
//   PatchGraphBSROpt::FeedbackCallback afterEachIteration,
//   void *callbackContext)
//{
//	this->indices = indices;
//	this->afterEachIteration = afterEachIteration;
//	this->afterEachIterationContext = callbackContext;
//
//	InCorePCTopologyGraph *highResExtraEdges = new InCorePCTopologyGraph();
//	highResExtraEdges->clearAndSetup(lowres->getTopology()->getEdges()->getDescr());
//	const PointSet *psLow = lowres->getTrajectorySet()->getTrajectories();
//	AAT origIndexAAT = psLow->getAAT("originalIndex");
//	unsigned n = lowres->getTopology()->getNumVertices();
//	for (unsigned i=0; i<n; i++) {
//		card32 numEdges = lowres->getTopology()->getNumAdjacentVertices(i);
//		card32 origIndex1 = psLow->get<int>(0,i, origIndexAAT);
//		for (card32 e=0; e<numEdges; e++) {
//			card32 adjVertex = lowres->getTopology()->getAdjacentVertex(i, e);
//			card32 origIndex2 = psLow->get<int>(0, adjVertex, origIndexAAT); 
//			highResExtraEdges->addEdge(origIndex1, origIndex2);
//		}
//	}
//
//	pg->setMultiResTopology(highResExtraEdges);
//	
//	Timer t;
//	debugOutput << "solving...";
//	optimizer->compute(indices, data, pg, afterEachIteration, callbackContext);
//	debugOutput << " --- " << t.getDeltaValue() << " ms\n";
//	this->indices = NULL;
//}
