//---------------------------------------------------------------------------
#ifndef NonLinearOptimizationH
#define NonLinearOptimizationH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
#include "DynamicLinearAlgebra.h"
#include "SparseLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl {


// ---- function types


// - INTERFACE CLASS -
/// a scalar function depending on multiple input variables
template <typename FloatType, typename IntType = int32>
class MultiVariateScalarFunction {
 public:
   /// return dimension of input vector
   virtual IntType getDimension() const = 0;
   /// evaluate function at position "pos"
   virtual FloatType calc(const DynamicVector<FloatType> &pos) const = 0;
};


// - INTERFACE CLASS -
/// interface that adds the capability of computing gradients of a mvsf
template <typename FloatType>
class MVSFGradient {
 public:
   virtual void calcGradient(const DynamicVector<FloatType> &pos, DynamicVector<FloatType> &gradient) const = 0;
};


// - INTERFACE CLASS -
/// interface that adds the capability of computing hessians of a mvsf
template <typename FloatType>
class MVSFHessian { //
 public:
   virtual void calcHessian(const DynamicVector<FloatType> &pos, SparseMatrix<FloatType> &hessian) const = 0;
};


// - INTERFACE CLASS -
/// interface that allows to determine the (potential) interdependence of input variables for computing the hessian matrix
/// returns for each input variable a list of all variables for which the partial derivative depends on the value of the input variable
template <typename FloatType, typename IntType = int32>
class MVSFSparseHessianDependency {
 public:
   /// input: variable index
   /// output: list of all variables for which the partial derivatives depend on the value of the input variable
   ///         the result is written in vector depVars, for efficiency reasons, the size of the vector is not
   ///         necessarily reduced if beeing larger than needed (increasing size if necessary is guaranteed);
   ///         the number of valid variables is returned in numDepVars
   virtual void getDependentPartialDerivVariables(IntType variableIndex, vector<IntType> &depVars, IntType &numDepVars) const = 0;
};


// - compound interfaces


// - INTERFACE CLASS -
/// compound interface - multi variante scalar function that supports evaluation of gradients
template <typename FloatType, typename IntType = int32>
class MVSFunctionWithGradient : public MultiVariateScalarFunction<FloatType, IntType>,
                                public MVSFGradient<FloatType>
{};

// - INTERFACE CLASS -
/// compound interface - multi variante scalar function that supports evaluation of gradients and hessians
template <typename FloatType, typename IntType = int32>
class MVSFunctionWithHessian : public MVSFunctionWithGradient<FloatType, IntType>,
                               public MVSFHessian<FloatType>
{};


// ---- function adaptors


/// converts a simple scalar function in a scalar function with gradient.
/// uses finite differences for gradient computiations.
template <typename FloatType, typename IntType = int32>
class MVSFunctionWithNumericalGradientAdaptor : public MVSFunctionWithGradient<FloatType, IntType> {
 private:
   FloatType delta;
   MultiVariateScalarFunction<FloatType, IntType> *mvsf;

 public:
   MVSFunctionWithNumericalGradientAdaptor(MultiVariateScalarFunction<FloatType, IntType> * mvsf, FloatType delta = 0.001) {
      pAssert(mvsf);
      this->mvsf = mvsf;
      this->delta = delta;
   }


   inline FloatType getDelta() const {
      return delta;
   }
   inline void setDelta(FloatType delta) {
      pAssert(delta > (FloatType)0);
      this->delta = delta;
   }

   virtual IntType getDimension() const {
      pAssert(mvsf);
      return mvsf->getDimension();
   }
   virtual FloatType calc(const DynamicVector<FloatType> &pos) const {
      pAssert(mvsf);
      return mvsf->calc(pos);
   }
   virtual void calcGradient(const DynamicVector<FloatType> &pos, DynamicVector<FloatType> &gradient) const {
      DynamicVector<FloatType> modPos = pos;
      pAssert(mvsf);
      FloatType centerValue = mvsf->calc(pos);
      IntType dim = mvsf->getDimension();
      gradient.setDim(dim);
      for (IntType i=0; i<dim; i++) {
         modPos[i] = pos[i] + delta;
         gradient[i] = (mvsf->calc(modPos) - centerValue) / delta;
         modPos[i] = pos[i];
      }
   }
};


/// converts a scalar function with gradient in a scalar function with gradient and (sparse) hessian.
/// needs a "sparse dependencies" interface to compute a sparse hessian with limited computational effort.
/// uses finite differences for gradient computiations.
// template <typename FloatType, typename IntType = int32>
// class MVSFunctionWithNumericalHessianAdaptor : public MVSFunctionWithHessian<FloatType, IntType> {
//  private:
//    FloatType delta;
//    MVSFunctionWithGradient<FloatType, IntType> *mvsfg;
//    MVSFSparseHessianDependency<FloatType, IntType> *dep;
// 
//  public:
//    MVSFunctionWithNumericalGradientAdaptor(
//       MVSFunctionWithGradient<FloatType, IntType> *mvsfg,
//       MVSFSparseHessianDependency<FloatType, IntType> *dep,
//       FloatType delta = 0.001
//    );
// 
//    inline FloatType getDelta() const;
//    inline void setDelta(FloatType delta);
// 
//    virtual IntType getDimension() const;
//    virtual FloatType calc(const DynamicVector<FloatType> &pos) const;
//    virtual void calcGradient(const DynamicVector<FloatType> &pos, DynamicVector<FloatType> &gradient) const;
//    virtual void calcHessian(const DynamicVector<FloatType> &pos, SparseMatrix<FloatType> &hessian) const;
// 
// };



// ---- solvers

class NLNullClass {};


/// base class of all non linear optimizatino procedures. setup of the function to be optimized in defined in base classes.
template <typename FloatType, typename IntType = int32, class BaseClass = NLNullClass>
class NonLinearOptimizer : public BaseClass {
 public:
   /// query current (best) guess for the (local) optimum. uses reference parameters for efficiency reasons
   /// (avoid allocation of new vectors if one of matching size is already handed in)
   virtual void getCurrentEstimate(DynamicVector<FloatType> &est) const = 0;
   /// performs another step of optimization (what a "step" is depends entirely on the employed technique)
   virtual void doStep() = 0;
   /// returns true if (the algorithm would guess that) the iteration has converged to a local optimum
   virtual bool converged() const = 0;
   /// reset solver, forget all internal computations (such as cg direction)
   virtual void reset() = 0;
};

/// optimization of scalar functions without knowing any derivative information (or the similar)
template <typename FloatType, typename IntType = int32, class BaseClass = NLNullClass>
class NLScalerFunctionOptimizer : public NonLinearOptimizer<FloatType, IntType, BaseClass> {
 public:
   virtual void setup(const MultiVariateScalarFunction<FloatType, IntType> *function, const DynamicVector<FloatType> &startEstimate) = 0;
};


/// optimization of scalar functions where gradients are known/computable (use coresponding num. grad. apdaptor if you don't know grad's)
template <typename FloatType, typename IntType = int32, class BaseClass = NLNullClass>
class NLScalerFunctionWithGradientOptimizer : public NonLinearOptimizer<FloatType, IntType, BaseClass> {
 public:
   virtual void setup(const MVSFunctionWithGradient<FloatType, IntType> *function, const DynamicVector<FloatType> &startEstimate) = 0;
};


/// optimization of scalar functions where gradients and the hessians are known/computable (use coresponding apdaptor if you don't know the 1st or 2nd order derivatives)
template <typename FloatType, typename IntType = int32, class BaseClass = NLNullClass>
class NLScalerFunctionWithHessianOptimizer : public NonLinearOptimizer<FloatType, IntType, BaseClass> {
 public:
   virtual void setup(const MVSFunctionWithGradient<FloatType, IntType> *function, const DynamicVector<FloatType> &startEstimate) = 0;
};


}

#endif
