//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphBSROpt.h"
//---------------------------------------------------------------------------

#include "BSRFitting.h"
#include "BSROptContext.h"
#include "BSROptPotential.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers_templCode.h"
//#include "LineSearchConjGradDescent.h"
#include "MgcEigen.h"
#include "Monomials2ndOrder.h"
#include "PatchGraph.h"
#include "PatchTypes.h"
#include "PGTopologicalNeighborhood.h"
#include "PointCloudAnimationSequence.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "PointSetKNNQuery.h"
#include "SeparatorClassProperty.h"
#include "SparseLinearAlgebra.h"
#include "SparseLinearAlgebra_templCode.h"
#include "Timer.h"
#include "UnifyNormals.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------


IMPLEMENT_GEOXL_CLASS( PatchGraphBSROpt, 0 )
{
	BEGIN_CLASS_INIT( PatchGraphBSROpt )
		ADD_OBJECT_PROP(context, 0, BSROptContext::getClass(), true)
		ADD_VAR_ARRAY_PROP( potentials, 0, 
		new ObjectClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		BSROptPotential::getClass(), true, NULL )
		)
}


PatchGraphBSROpt::PatchGraphBSROpt() {
	context = new BSROptContext();
	/************************************************************************/
	/* Added by peter for testing                                                                     */
	/************************************************************************/
	edge_weight_rigidity			= NULL;
	edge_weight_smoothness	= NULL;
}

void PatchGraphBSROpt::addPotential(BSROptPotential *pot) {
	potentials.push_back(pot);
}

void PatchGraphBSROpt::addPotential(const MetaClass *potClass) {
	if (!potClass->inheritsFrom(BSROptPotential::getClass())) {
		throw PException("PatchGraphBSROpt::addPotential() - wrong type.");
	}
	potentials.push_back((BSROptPotential*)potClass->newInstance());
}

void PatchGraphBSROpt::removePotential(const MetaClass *potClass) {
	bool found = false;
	do {
		found = false;
		for (unsigned i=0; i<potentials.size(); i++) {
			if (potentials[i]->getInstanceClass()->inheritsFrom(potClass)) {
				delete potentials[i];
				vector<BSROptPotential*>::iterator pos = potentials.begin();
				pos += i;
				potentials.erase(pos);
				found = true;
				break;
			}
		}
	} while (found);
}

void PatchGraphBSROpt::setContext(const BSROptContext* ctxt) {
	context->assign(ctxt);
}

void PatchGraphBSROpt::clearPotentials() {
	for (unsigned i=0; i<potentials.size(); i++) {
		delete potentials[i];
	}
	potentials.clear();
}

void PatchGraphBSROpt::setPotentialSet(PotentialSet potentials) {
	clearPotentials();
	for (unsigned i=0; i<potentials.size(); i++) {
		BSROptPotential *newPot = dynamic_cast<BSROptPotential*>(potentials[i]->newInstance());
		if (newPot != NULL) {
			addPotential(newPot);
		} else {
			pAssert(false);
		}
	}
}


void PatchGraphBSROpt::compute( const BSROptVariableIndexSet *indices,
										 PointCloudAnimationSequence *data,
										 PatchGraph *patchGraph,
										 FeedbackCallback afterEachIteration,
										 void *callbackContext)
{
	bool needSampleSpacing = false;
	bool needRefFrames = false;
	bool needHFieldCoeffs = false;
	for (unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++) 
	{
		card32 flags = potentials[potentialIndex]->needPrecomputedChannels(context);
		needSampleSpacing |= (flags & BSROptPotential::FLAG_PATCH_SAMPLE_SPACING) != 0;
		needRefFrames     |= (flags & BSROptPotential::FLAG_PATCH_REF_FRAMES) != 0;
		needHFieldCoeffs  |= (flags & BSROptPotential::FLAG_NEED_PATCH_2ND_ORDER_HFIELD) != 0;
	}	
	if (needHFieldCoeffs) 
	{
		needRefFrames = true;
	}

	patchGraph->createPatchAttribute( mVAD("varIndex", 1, VAD::DATA_FORMAT_INT32) );
	if (needRefFrames) 
	{
		patchGraph->createPatchAttribute( NORMAL_PATCH_VAD );
		patchGraph->createPatchAttribute( TANGENT_U_PATCH_VAD );
	}
	if (needSampleSpacing) 
		patchGraph->createPatchAttribute( mVAD("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32) );
	if (needHFieldCoeffs)  
		patchGraph->createPatchAttribute( createHeightfieldCoeffVAD(6) );

	PointSet *ps = patchGraph->getTrajectorySet()->getTrajectories();
	/// get AATs
	AAT POSITION = ps->getAAT( "position" );
	AAT HF_COEFFS;
	if (context->optMode == BSROptContext::MODE_FIT_COEFFICIENTS) 
	{
		HF_COEFFS = ps->getAAT( HEIGHTFIELD_COEFFS_NAME );
	}

	card32 coeffUpdateCounter = 0;

	/// reset potentials
	for (unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++) 
	{
		potentials[potentialIndex]->initialize();
		potentials[potentialIndex]->coordsChanged();
	}

	card32 numVarsPerVar = 0;
	string channelName;
	switch (context->optMode) {
		case BSROptContext::MODE_MOVE_POINTS: 
		{
			numVarsPerVar = 3;
			channelName = "position";
			break;
		}
		case BSROptContext::MODE_FIT_COEFFICIENTS: 
		{
			numVarsPerVar = 6;
			channelName = HEIGHTFIELD_COEFFS_NAME;
			break;
		}
	}
	DVectorF vars, gradient;
	SparseMatrixF quadApprox;
	DVectorF linApprox;
	gradient.setDim((card32)indices->size()*numVarsPerVar);
	vars.setDim((card32)indices->size()*numVarsPerVar);
	if (context->solver != BSROptContext::SOLVER_GRADIENT_DESCENT) 
	{
		quadApprox.setRows((card32)indices->size()*numVarsPerVar);
		linApprox.setDim((card32)indices->size()*numVarsPerVar);
	}

	BOPIndexingChannel::writeIndexingChannel(indices,patchGraph);
	vector<float32> potVals(potentials.size());
	if (context->verbose) 
	{
		debugOutput << "Optimizing...\n";
		debugOutput << "time\tgrad inf norm:\tpotential";
		for (card32 pot=0; pot<potentials.size(); pot++) 
		{
			debugOutput << "\t" << potentials[pot]->getInstanceClass()->getClassName();
		}
		debugOutput << "\n";
	}
		
	for ( unsigned int step = 0; step < context->numIterations; step++ ) 
	{
		Timer timer;
		if (coeffUpdateCounter % context->coeffsUpdateFreq == 0) 
		{
			if (needSampleSpacing) 
			{
				BSRFitting::estSampleSpacing(context, patchGraph, *indices);
			}
		}
		if (coeffUpdateCounter % context->coordsUpdateFreq == 0) 
		{
			if (needRefFrames) 
			{
				BSRFitting::fitRefFrames(context, patchGraph, *indices);
				if (context->unifyNormals) 
				{
					UnifyNormals un;
					un.unifyNormals(patchGraph, indices);
				}
				for (unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++) 
				{
					potentials[potentialIndex]->coordsChanged();
				}
			}
		}
		if (coeffUpdateCounter % context->coeffsUpdateFreq == 0) 
		{
			if (needHFieldCoeffs) 
			{
				BSRFitting::fit2ndOrderHeightf(context, patchGraph, *indices);
			}
		}            
		coeffUpdateCounter++;

		// Compute gradients/hessians using potentials
		const size_t numIndices = indices->size();
		for (card32 i=0; i<numVarsPerVar*numIndices; i++) 
		{
			gradient[i] = 0.0f;
		}

		BOPIndexingChannel::readChannel(indices, patchGraph, channelName, numVarsPerVar, vars);
		float32 potential = 0;
		float32 maxGradient = 0;

		if (context->solver == BSROptContext::SOLVER_GRADIENT_DESCENT) 
		{
			// gradient descent
			for ( unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++ ) 
			{
				potVals[potentialIndex] = potential;
				potentials[potentialIndex]->computeGradient(context, indices, data, patchGraph, gradient, potential, NULL, NULL, NULL);
				potVals[potentialIndex] = potential - potVals[potentialIndex];
			}
			maxGradient = max(fabs(gradient.maxComponent()), fabs(gradient.minComponent()));
			gradient *= context->stepSize;
			vars -= gradient;
		} 
		else 
		{	
			// newton step
			quadApprox.setZero();
			linApprox.setZero();
			for ( unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++ ) 
			{
				potVals[potentialIndex] = potential;
				potentials[potentialIndex]->computeGradient(context, indices, data, patchGraph, gradient, potential, NULL, &quadApprox, &linApprox);
				potVals[potentialIndex] = potential - potVals[potentialIndex];
			}
			maxGradient = max(fabs(gradient.maxComponent()), fabs(gradient.minComponent()));
			linApprox *= -0.5; // convert to right hand side
			if (context->newtonRegularization > 0.0f) {
				SparseMatrixF id(quadApprox.getRows());
				for (card32 i=0; i<quadApprox.getRows(); i++) {
					id[i].pushBackEntry(i, context->newtonRegularization);
				}
				quadApprox += id;
				linApprox += vars*context->newtonRegularization;
			}
			card32 numIterations = context->cgMaxNumIterations;
			conjugateGradientsSolve<float32>(quadApprox, vars, linApprox, numIterations, context->cgStopRes, context->cgDebugging);
		}
		BOPIndexingChannel::writeChannel(indices, patchGraph, channelName, numVarsPerVar, vars);

		if (afterEachIteration) afterEachIteration(callbackContext);
		if (context->verbose) 
		{
			debugOutput << timer.getDeltaValue() << "\t" << maxGradient << "\t" << potential;
			for (card32 pot=0; pot<potentials.size(); pot++) 
			{
				debugOutput << "\t" << potVals[pot];
			}
			debugOutput << "\n";
		}
		if (context->stoppingCriterion == BSROptContext::STOP_CONV_GRAD_INFNORM && maxGradient < context->convergenceEps) break;
		
		for ( unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++ )
		{
			potentials[potentialIndex]->finished();
		}
	}
}

PatchGraphBSROpt::~PatchGraphBSROpt() 
{
	delete context;
	for (unsigned i=0; i<potentials.size(); i++) 
	{
		delete potentials[i];
	}
	/************************************************************************/
	/* Added by peter for testing                                                                     */
	/************************************************************************/
	if(edge_weight_rigidity != NULL)
		delete []edge_weight_rigidity;

	if(edge_weight_smoothness != NULL)
		delete []edge_weight_smoothness;

	potential_flags.clear();
}
