//---------------------------------------------------------------------------
#ifndef PatchGraphBSROptH
#define PatchGraphBSROptH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "DynamicLinearAlgebra.h"
#include "BSROptPotential.h"
//#include "NonLinearOptimization.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------

namespace geoxl {
   using namespace std;

class BSROptContext;
class PlainSimpleGradientDescentF32I32;
class LineSearchConjGradDescentF32I32;
class PointCloudAnimationSequence;
class PatchGraph;
class PointSet;
class UnifyNormals;

   
/// universal non-linear fitting of patch graphs to multiple frames
class ANIMATIONRECONSTRUCTION_API PatchGraphBSROpt : public Persistent {
	GEOXL_CLASS( PatchGraphBSROpt )
 private:
   /// parameters   
	BSROptContext *context;
	/// list of statistical potentials
	vector<BSROptPotential*> potentials;
	
	IMPLEMENT_VAR_ARRAY_METHODS_STL(potentials)
	
 public:
   /// callback type, e.g. for rendering steps of the optimization
   typedef void (*FeedbackCallback)(void*);
   /// set of meta classes of BSROptPotentials
   typedef vector<const MetaClass *> PotentialSet;

   /// constructor
	PatchGraphBSROpt();
	
	/// ownership transfered to this object
	void					addPotential(BSROptPotential *pot);
	/// class must be of type BSROptPotential
	void					addPotential(const MetaClass *potClass);
	/// remove all potentials of the given type (or subtypes)
	void					removePotential(const MetaClass *potClass);
	/// remove all potentials
	void					clearPotentials();
	/// clear potentials and set a new set of potentials
	void					setPotentialSet(PotentialSet potentials);

   /// get reference to parameter object(read/write)
	inline	BSROptContext* getContext() const { return context; }
	/// replace current parameters object with a new one
	void					setContext(const BSROptContext* ctxt);

   /// main method - performs the optimization
	virtual		void		compute(
      const BSROptVariableIndexSet *indices,
      PointCloudAnimationSequence *data,
      PatchGraph *patchGraph,
      FeedbackCallback afterEachIteration = NULL,
      void *callbackContext = NULL
   );

	/************************************************************************/
	/* The following functions are added by peter for testing               */
	/************************************************************************/

	/// typeOfSDFPotential = 0, we don't want to use data points for optimization(Partial Matching)
	/// typeOfSDFPotential = 1, data points are guaranteed to have correspondences(global assembly)
	/// typeOfSDFPotential = 2, only part of the data points have correspondences(pair-wise matching)

	/// the final version of peter's optimization
	// 0 for standard pair-wise matching, 
	// 1 for correspondence based pai-wise matching
	// 3 for partial-matching
	// 4 for global assembly
	virtual	void	optimization(
		const BSROptVariableIndexSet	*indices_pos,			//	the indices of positions to be optimized
		const BSROptVariableIndexSet	*indices_nor,			//	the indices of normals to be optimized
		const BSROptVariableIndexSet	*indices_rot,			//	the indices of rotation matrices to be optimized
		const	vector<card32>			*indices_data,			//  the indices of data scans to be used in optimization
		PointCloudAnimationSequence		*data,					//	these are actually the data points used in this optimization
		PatchGraph						*patchGraph,			//	the patch graph
		SparseMatrixF					*topologyOfTargetFrame,	//  the topology used in pairwise matching
		card32							typeOfOptimization, 
		FeedbackCallback afterEachIteration = NULL,				//  User interface I
		void *callbackContext = NULL							//  User interface II
		);

	vector<bool>		potential_flags;

	/************************************************************************/
	/* private members variable added by peter                                                                     */
	/************************************************************************/
private:

	/************************************************************************/
	/************************************************************************/
	/************************************************************************/

	/************************************************************************/
	/* We do three things within preprocessing
	/* 
	/************************************************************************/
	
	/// This method is only used in partial matching
	virtual	void	initialize_Geometry_For_Partial_Matching(
		const BSROptVariableIndexSet *indices_pos,			//the indices of positions to be optimized
		const BSROptVariableIndexSet *indices_nor,       //the indices of normals to be optimized
		const BSROptVariableIndexSet	*indices_rot,
		PatchGraph		*patchGraph
		);
	
	virtual	void	preprocessing_For_Optimization(
		const BSROptVariableIndexSet *indices_pos,			//the indices of positions to be optimized
		const BSROptVariableIndexSet *indices_nor,       //the indices of normals to be optimized
		const BSROptVariableIndexSet *indices_rot,       //the indices of rotation matrices to be optimized
		PointCloudAnimationSequence  *data,						//the raw data for optimization
		const	vector<card32>		 *indices_data,
		card32	typeOfOptimization,
		PatchGraph *patchGraph);											//the patch graph

	///Compute initial rotation matrix for active surfels
	virtual	void	compute_initial_rotation_matrices(
		const	BSROptVariableIndexSet	*indices_rot,			//the indices of rotation matrix to be optimized
		PatchGraph	*patchGraph											//the patch graph
		);


	void			 Construct_Edge_Weight_Arrays(PatchGraph *pg);
	/// This array stores the weight of each edge for smoothness potential
	float32	*edge_weight_smoothness;

	/// This array stores the weight of each edge for rigidity potential
	float32	*edge_weight_rigidity;

	/************************************************************************/
	/************************************************************************/
	/************************************************************************/
	
	/* Various optimization techniques										                                                     */
	
	/************************************************************************/
	/************************************************************************/
	/************************************************************************/
	
	virtual	void	gauss_Newton_Method(
		const BSROptVariableIndexSet	*indices_pos,				//the indices of positions to be optimized
		const BSROptVariableIndexSet	*indices_nor,				//the indices of normals to be optimized
		const BSROptVariableIndexSet	*indices_rot,				//the indices of rotation matrices to be optimized
		const	vector<card32>			*indices_data,				// 
		PointCloudAnimationSequence		*data,						//only related to feature values now and maybe changed later
		PatchGraph						*patchGraph,				//the patch graph
		SparseMatrixF					*topologyOfTargetFrame,		//topology of the target frame
		card32							typeOfOptimization,			//
		FeedbackCallback				afterEachIteration = NULL,	//User interface I
		void							*callbackContext = NULL		//User interface II
		);


	virtual	void	initialize_Sparse_Matrix(
		const BSROptVariableIndexSet	*indices_pos,			//the indices of positions to be optimized
		const BSROptVariableIndexSet	*indices_nor,				//the indices of normals to be optimized
		const BSROptVariableIndexSet	*indices_rot,				//the indices of rotation matrices to be optimized
		PatchGraph		*patchGraph,
		SparseMatrixF	*quadApprox);
	
	///Used particularly for partial optimization
	virtual	void	compute_Related_Variable_Indices(
		BSROptVariableIndexSet				*indices_related,
		PatchGraph									*patchGraph
		);

	/// Used in index transfer for marker error potential
	vector<int32>	index_transfer;
	virtual	void	compute_GNM_Gradient_and_Hessian( 
		const	BSROptVariableIndexSet				*indices_related,		/// only these indices are related to this optimization
		const	PointCloudAnimationSequence	*data,							/// the raw data
		PatchGraph		*patchGraph,										/// the patch graph to be optimized
		DVectorF			*gradient,										/// the first order derivative
		SparseMatrixF	*quadApprox											/// the second order derivative														
		);

	virtual	void	compute_GNM_Gradient_and_Hessian_SDF_Global(
		const	vector<card32>			*indices_data,						/// only these indices are related to this optimization
		PointCloudAnimationSequence	*data,									/// contains the data points used in this optimization
		PatchGraph		*patchGraph,										/// the patch graph to be optimized
		DVectorF			*gradient,										/// the first order derivative
		SparseMatrixF	*quadApprox,										/// the second order derivative
		float32				tradeoff										/// the tradeoff between point-point distance
																			/// point-plane distance												
		);

	virtual	float32	compute_GNM_Gradient_and_Hessian_SDF_Pairwise(
		const	vector<card32>			*indices_data,						/// only these indices are related to this optimization
		PointCloudAnimationSequence	*data,									/// the raw data
		PatchGraph		*patchGraph,										/// the patch graph to be optimized
		SparseMatrixF	*topologyOfTargetFrame,								/// the topology of the target frame
		DVectorF		*gradient,											/// the first order derivative
		SparseMatrixF	*quadApprox,										/// the second order derivative
		float32			&previous_median_d2,								/// used to compute descent_flag 
		const	float32	tradeoff_sdf
		);

	/// Nonlinear Conjugate Gradients with Newton-Raphson and Fletcher-Reeves

	virtual void nonlinear_Conjugate_Gradients_NRFR(
		const	BSROptVariableIndexSet				*indices_related,
		PointCloudAnimationSequence *data,									//the raw data for optimization
		PatchGraph *patchGraph,												//the patch graph
		const	card32	m_dimension,										//the dimension of this problem
		FeedbackCallback afterEachIteration = NULL,							//User interface I
		void *callbackContext = NULL										//User interface II
		);

	virtual	void	compute_NCG_NRFR_1st_and_2st_Terms( 
		const PointCloudAnimationSequence *data,							/// the raw data
		PatchGraph		*patchGraph,										/// the patch graph to be optimized
		float32				&term_1st,										/// the first order derivative
		float32				&term_2st										/// the second order derivative
		);


	virtual	void	compute_NCG_NRFR_Gradient(
		const PointCloudAnimationSequence *data,							/// the raw data
		PatchGraph		*patchGraph,										/// the patch graph to be optimized
		DVectorF			*gradient										/// the first order derivative
		);


	virtual	void	setSmoothnessFlag(
		const	card32				TypeOfOptimization,
		PointCloudAnimationSequence	*data,
		PatchGraph					*patchGraph
		);

	/// Preconditioned Nonlinear-Conjugate Gradients with Secant and Polak-Ribiere

	/************************************************************************/
	/* Line search method and projection onto manifold                                                             */
	/************************************************************************/
		
	virtual	void	projectionOntoManifold(
		const	BSROptVariableIndexSet				*indices_related,			///Only the surfels specified in this index array are to be optimized
		PatchGraph			*patchGraph,
		const	DVectorF	*displacement);

	virtual	void	lineSearch_Armijo(
		PointCloudAnimationSequence	*data,							/// the raw data for optimization
		PatchGraph									*patchGraph,			/// the patch graph
		DVectorF									&direction,					/// the direction to be searched
		float32										&cEnergy,					/// current potential energies
		float32											&nEnergy,					/// next		potential energies
		float												&stepsize					/// the stepsize after line search
		);

	virtual	void	ComputeEnergy(
		const PointCloudAnimationSequence *data,
		PatchGraph			*patchGraph,
		const	DVectorF	&perturb,
		float						&energy,
		bool							m_flag
		);

	/************************************************************************/
	/*  The above codes are added by peter for testing                                                                    */
	/************************************************************************/

public:
	/// destructor, deletes all potentials
	~PatchGraphBSROpt();
};
   
}

#endif
