//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphBSROpt.h"
//---------------------------------------------------------------------------

#include "BSRFitting.h"
#include "BSROptContext.h"
#include "BSROptPotential.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "InCorePCTopologyGraph.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers_templCode.h"
//#include "LineSearchConjGradDescent.h"
#include "MgcEigen.h"
#include "Monomials2ndOrder.h"
#include "PatchGraph.h"
#include "PatchTypes.h"
#include "PGTopologicalNeighborhood.h"
#include "PointCloudAnimationSequence.h"
#include "PGDataPointIndices.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "PointSetKNNQuery.h"
#include "SeparatorClassProperty.h"
#include "SparseLinearAlgebra.h"
#include "SparseLinearAlgebra_templCode.h"
#include "Timer.h"
#include "UnifyNormals.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

/************************************************************************/
/*PETER: The following codes are just for testing                                                                     */
/************************************************************************/

void		Quaternion_To_Matrix3f(const float	*q, Matrix3f &rotation)
{
	pAssert(q != NULL);
	rotation[0][0] = 2*q[1]*q[1] + (2*q[0]*q[0]-1);
	rotation[1][0] = 2*q[1]*q[2] - 2*q[0]*q[3];
	rotation[2][0] = 2*q[1]*q[3] + 2*q[0]*q[2];
	rotation[0][1] = 2*q[2]*q[1] + 2*q[0]*q[3];
	rotation[1][1] = 2*q[2]*q[2] + (2*q[0]*q[0]-1);
	rotation[2][2] = 2*q[2]*q[3] - 2*q[0]*q[1];
	rotation[0][2] = 2*q[3]*q[1] - 2*q[0]*q[2];
	rotation[1][2] = 2*q[3]*q[2] + 2*q[0]*q[1];
	rotation[2][2] = 2*q[3]*q[3] + (2*q[0]*q[0]-1);
}

void		PatchGraphBSROpt::optimization(
	const BSROptVariableIndexSet	*indices_pos, 
	const BSROptVariableIndexSet	*indices_nor, 
	const BSROptVariableIndexSet	*indices_rot, 
	const vector<card32>			*indices_data, 
	PointCloudAnimationSequence		*data,
	PatchGraph						*patchGraph, 
	SparseMatrixF					*topologyOfTargetFrame,
	card32							typeOfOptimization,
	FeedbackCallback				afterEachIteration, 
	void *callbackContext)
{
	/// preprocessing for optimization

	preprocessing_For_Optimization(indices_pos, indices_nor, indices_rot, 
			data, indices_data, typeOfOptimization, patchGraph);
	
	/// use gauss-newton optimizer for optimization
	gauss_Newton_Method(indices_pos, indices_nor, indices_rot, indices_data,
		data, patchGraph, topologyOfTargetFrame, typeOfOptimization, afterEachIteration, callbackContext);

	/// Dont't know whether we indeed need this
	for ( unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++ )
	{
		potentials[potentialIndex]->finished();
	}
}

void		PatchGraphBSROpt::Construct_Edge_Weight_Arrays(PatchGraph *pg)
{
	if(edge_weight_rigidity != NULL && edge_weight_smoothness != NULL)
		return;

	if(edge_weight_rigidity != NULL)
		delete []edge_weight_rigidity;
	if(edge_weight_smoothness != NULL)
		delete []edge_weight_smoothness;

	card32	width		= pg->getTrajectorySet()->getNumFrames();
	card32	height		= pg->getTrajectorySet()->getNumPatches();

	card32 size = 0;
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		size += pg->getTopology()->getNumAdjacentVertices(surfelIndex);
	}
	size *= width;

	edge_weight_rigidity			= new float32[size];
	edge_weight_smoothness	= new float32[size];

	for(card32 i = 0; i < size; i++)
	{
		edge_weight_rigidity[i] = edge_weight_smoothness[i] = 1.0f;
	}
}

void		PatchGraphBSROpt::compute_initial_rotation_matrices(
	const BSROptVariableIndexSet *indices_rot, 
	PatchGraph *patchGraph)
{
	/// Construct the eigen-system for computing rotation matrix
	/// The method is based on the following paper
	/// Horn, B.K.P. (1987) "Closed form solution of absolute orientation 
	/// using unit quaternions", JOSA A, Vol. 4, Np. 4, April.

//	pAssert(edge_weight_rigidity != NULL);
	/// check whether this problem is ill-posed
	

	bool			flag_pos_r	= potential_flags[6],
					flag_nor_r	= potential_flags[7],
					flag_pos_c	= potential_flags[8],
					flag_nor_c	= potential_flags[9];

	if(!flag_pos_r&&!flag_nor_r&&(!flag_pos_c||!flag_nor_c))
		return;
											
	float32			w_pos_r		= context->weightPositionRigidity, 
					w_nor_r		= context->weightNormalRigidity,
					w_pos_c		= context->weightPositionCoherence,
					w_nor_c		= context->weightNormalCoherence;

	PointSet		*ps			= patchGraph->getTrajectorySet()->getTrajectories();
	PointSet		*urshape	= patchGraph->getUrshape();
	InCorePCTopologyGraph	*tg	= patchGraph->getTopology();

	AAT					REF_N		= ps->getAAT( "normal" );
	AAT					POSITION	= ps->getAAT( "position" );
	//take care of this, if there is not "varIndex_rot" channel, we should create it
	AAT					ROTATION	= ps->getAAT( "rotation_matrix");
	AAT					TOP_DEPTH	= ps->getAAT( "topology_depth");
	
	card32				offset		= 0;
	geoxl::Eigen<float32>			*eigenSolver= new geoxl::Eigen<float32>(4);
	
	Vector3f			pik, pikn, pikp, pjk, pjkn,
							nik, nikn, nikp, njk, njkn,
							rijk, rijkn, 
							nijk, nijkn, nikpk, nikkn;

	int32				ik_depth, jk_depth, ikn_depth, jkn_depth, ikp_depth; 

	DMatrixF			M;	M.setDimension(3, 3, false);

	card32				width = ps->getWidth(),
						height = ps->getHeight();
	//start computation
	
	for(card32 id = 0; id < indices_rot->size(); id++)
	{
		card32	frameIndex	= (*indices_rot)[id].frame;
		card32	surfelIndex	= (*indices_rot)[id].point;

		pik				= ps->get3f(frameIndex,		surfelIndex,	POSITION);
		pikn				= ps->get3f((frameIndex+1)%width,	surfelIndex,	POSITION);
		nik				= ps->get3f(frameIndex,		surfelIndex,	REF_N);
		nikn				= ps->get3f((frameIndex+1)%width,	surfelIndex,	REF_N);
		ik_depth		= ps->get<int32>(frameIndex, surfelIndex, TOP_DEPTH);
		ikn_depth  = ps->get<int32>((frameIndex+1)%width, surfelIndex, TOP_DEPTH);
		if(ik_depth >= 2 || ikn_depth >= 2)
			continue;
			
		M.setZero();
		
		if(frameIndex > 0 && frameIndex < ps->getWidth()-1)
		{
			pikp			= ps->get3f(frameIndex-1,	surfelIndex,	POSITION);
			nikp			= ps->get3f(frameIndex-1,	surfelIndex,	REF_N);
			ikp_depth = ps->get<int32>(frameIndex-1, surfelIndex, TOP_DEPTH);
			if(ikp_depth <= 1)
			{
				nikpk			= nikp -  nik;
				nikkn			= nik  -  nikn;

				if(flag_nor_c)
				{
					for(card32 i = 0; i < 3; i++)
					{
						for(card32 j = 0; j < 3; j++)
						{
							M[i][j] += w_nor_c*nikpk[i]*nikkn[j];
						}
					}
				}
			}
		}

		card32		numOfEntries= tg->getNumAdjacentVertices(surfelIndex);
		for (card32 ni = 0; ni< numOfEntries; ni++) 
		{
			card32	neighbor_id = tg->getAdjacentVertex(surfelIndex, ni);
			pjk					= ps->get3f(frameIndex,	 neighbor_id, POSITION);
			pjkn				= ps->get3f((frameIndex+1)%width,neighbor_id, POSITION);
			njk					= ps->get3f(frameIndex,	 neighbor_id, REF_N);
			njkn				= ps->get3f((frameIndex+1)%width,neighbor_id, REF_N);
			jk_depth			= ps->get<int32>(frameIndex, neighbor_id, TOP_DEPTH);
			jkn_depth			= ps->get<int32>((frameIndex+1)%width, neighbor_id, TOP_DEPTH);
			if(jk_depth >= 2 || jkn_depth >= 2)
				continue;
			rijk				= pik  - pjk;
			rijkn				= pikn - pjkn;
			nijk				= nik  - njk;
			nijkn				= nikn - njkn;
			float	weight		= 1.0f/numOfEntries;
			if(flag_pos_r)
			{
				for(card32 i = 0; i < 3; i++)
				{
					for(card32 j = 0; j < 3; j++)
					{
						M[i][j] += weight*w_pos_r*rijk[i]*rijkn[j]; 
					}
				}
			}
			if(flag_nor_r)
			{
				for(card32 i = 0; i < 3; i++)
				{
					for(card32 j = 0; j < 3; j++)
					{
						M[i][j] += weight*w_nor_r*nijk[i]*nijkn[j];
					}
				}
			}
			eigenSolver->Matrix(0, 0) =		M[0][0] + M[1][1] + M[2][2];
			eigenSolver->Matrix(1, 1) =		M[0][0] - M[1][1] - M[2][2];
			eigenSolver->Matrix(2, 2) =	-	M[0][0] + M[1][1] - M[2][2];				
			eigenSolver->Matrix(3, 3) =	-	M[0][0] - M[1][1] + M[2][2];
			eigenSolver->Matrix(0, 1) =		M[1][2] - M[2][1];	eigenSolver->Matrix(1, 0) =		M[1][2] - M[2][1];
			eigenSolver->Matrix(0, 2) =		M[2][0] - M[0][2];	eigenSolver->Matrix(2, 0) =		M[2][0] - M[0][2];
			eigenSolver->Matrix(0, 3) =		M[0][1] - M[1][0];	eigenSolver->Matrix(3, 0) =		M[0][1] - M[1][0];
			eigenSolver->Matrix(1, 2) =		M[0][1] + M[1][0];	eigenSolver->Matrix(2, 1) =		M[0][1] + M[1][0];
			eigenSolver->Matrix(1, 3) =		M[0][2] + M[2][0];	eigenSolver->Matrix(3, 1) =		M[0][2] + M[2][0];
			eigenSolver->Matrix(2, 3) =		M[1][2] + M[2][1];	eigenSolver->Matrix(3, 2) =		M[1][2] + M[2][1];

			eigenSolver->DecrSortEigenStuff4();
			//  Destruct the eigen-system used for computing rotation matrix
			char			*dataC = ps->getDataPointer(frameIndex, surfelIndex);
			float32	*dataF = (float*)(dataC+ROTATION.getOffset());
			for(card32 id = 0; id < 4; id++)
			{
				dataF[ id ] = eigenSolver->GetEigenvector(id, 0);
			}
		}
		offset	+= numOfEntries;
	}
	delete	eigenSolver;
}

/************************************************************************/
/* Gauss-Newton Method															                                                  */
/************************************************************************/

void	PatchGraphBSROpt::initialize_Geometry_For_Partial_Matching(
	const BSROptVariableIndexSet *indices_pos,
	const BSROptVariableIndexSet *indices_nor, 
	const BSROptVariableIndexSet *indices_rot,
	PatchGraph *patchGraph 
	)
{
	PointSet								*ps = patchGraph->getTrajectorySet()->getTrajectories();
	InCorePCTopologyGraph *tg	= patchGraph->getTopology();
	card32			width = ps->getWidth(),
							height = ps->getHeight();


	AAT				POSITION		= ps->getAAT("position"),				
						NORMAL			= ps->getAAT("normal"),
						REF_U				= ps->getAAT("tangent_u"),			
						POS_INDEX	= ps->getAAT("varIndex_pos"),		
						NOR_INDEX	= ps->getAAT("varIndex_nor"),
						ROT_INDEX	= ps->getAAT("varIndex_rot"),
						ROTATION		= ps->getAAT("rotation_matrix"),
						TOP_DEPTH	=	ps->getAAT("topology_depth");

	//FILE *f = fopen("D:\\temp.txt", "w");
	for(card32 id = 0; id < indices_pos->size(); id++)
	{
		card32 frameIndex	= (*indices_pos)[id].frame;
		card32 surfelIndex	= (*indices_pos)[id].point;
		int32		ik_depth  = ps->get<int32>(frameIndex, surfelIndex, TOP_DEPTH);
		if(ik_depth == 1)
		{
			Vector3f		pik;	pik[0] = pik[1] = pik[2] = 0.0f;
			int32			pos_id		= ps->get<int32>(frameIndex, surfelIndex, POS_INDEX);
			if(pos_id < 0)
				continue;
			card32			numOfEntries = tg->getNumAdjacentVertices(surfelIndex);
			card32			numOfNeighbors = 0;
			for(card32 ni = 0; ni < numOfEntries; ni++)
			{
				/// the index of this neighbor
				card32 neighbor_id	= tg->getAdjacentVertex(surfelIndex, ni);
				if(neighbor_id == surfelIndex)
				{
					continue;
				}
				int	jk_depth		=	ps->get<int32>(frameIndex, neighbor_id, TOP_DEPTH);
				if(jk_depth >= 1)
					continue;

				Vector3f pjk = ps->get3f(frameIndex, neighbor_id, POSITION);
				pik += pjk;				numOfNeighbors++;
			}
			pik /= numOfNeighbors;
			ps->set3f(frameIndex, surfelIndex, POSITION, pik);
		}
	}	

	for(card32 id = 0; id < indices_nor->size(); id++)
	{
		card32 frameIndex	= (*indices_nor)[id].frame;
		card32 surfelIndex	= (*indices_nor)[id].point;
		int32		pik_depth  = ps->get<int32>(frameIndex, surfelIndex, TOP_DEPTH);
		if(pik_depth == 1)
		{
			Vector3f		nik; nik[0] = nik[1] = nik[2] = 0.0f;
			int32			nor_id		= ps->get<int32>(frameIndex, surfelIndex, NOR_INDEX);
			if(nor_id < 0)
				continue;
			card32			numOfEntries = tg->getNumAdjacentVertices(surfelIndex);
			for(card32 ni = 0; ni < numOfEntries; ni++)
			{
				/// the index of this neighbor
				card32 neighbor_id	= tg->getAdjacentVertex(surfelIndex, ni);
				if(neighbor_id == surfelIndex)
				{
					continue;
				}
				int	jk_depth		=	ps->get<int32>(frameIndex, neighbor_id, TOP_DEPTH);
				if(jk_depth >= 1)
					continue;

				nik += ps->get3f(frameIndex, neighbor_id, NORMAL);
			}
			nik.normalize();
			ps->set3f(frameIndex, surfelIndex, NORMAL, nik);
			Vector3f	Eik1;
			if(fabs(nik[0]) > fabs(nik[1]))
			{
				if(fabs(nik[0]) > fabs(nik[2]))
				{
					Eik1[0] = 1.0; Eik1[1] = 0.0; Eik1[2] = 0.0;
				}
				else
				{
					Eik1[0] = 0.0; Eik1[1] = 0.0; Eik1[2] = 1.0;
				}
			}
			else
			{
				if(fabs(nik[1]) > fabs(nik[2]))
				{
					Eik1[0] = 0.0; Eik1[1] = 1.0; Eik1[2] = 0.0;
				}
				else
				{
					Eik1[0] = 0.0; Eik1[1] = 0.0; Eik1[2] = 1.0;
				}
			}
			Eik1 -= nik*(Eik1*nik);	

			Eik1.normalize();
			ps->set3f(frameIndex, surfelIndex, REF_U, Eik1);
		}
	}
	for(card32 id = 0; id < indices_rot->size(); id++)
	{
		card32 frameIndex	= (*indices_rot)[id].frame;
		card32 surfelIndex	= (*indices_rot)[id].point;
		int32		ik_depth  = ps->get<int32>(frameIndex, surfelIndex, TOP_DEPTH);
		int32			rot_id		= ps->get<int32>(frameIndex, surfelIndex, ROT_INDEX);
		if(rot_id < 0)
			continue;

		char					*dataC = ps->getDataPointer(frameIndex, surfelIndex);
		float32			*quatC = (float32*)(dataC+ROTATION.getOffset());

		quatC[0] = 1.0f;	quatC[1] = 0.0f;	quatC[2] = 0.0f;	quatC[3] = 0.0f;
		/*
		card32			numOfEntries = tg->getNumAdjacentVertices(surfelIndex);
		for(card32 ni = 0; ni < numOfEntries; ni++)
		{
			/// the index of this neighbor
			card32 neighbor_id	= tg->getAdjacentVertex(surfelIndex, ni);
			if(neighbor_id == surfelIndex)
			{
				continue;
			}
			int	jk_depth		=	ps->get<int32>(frameIndex, neighbor_id, TOP_DEPTH);
			int   jk_depth1  = ps->get<int32>(frameIndex+1, neighbor_id, TOP_DEPTH);
			if(jk_depth == 0 && jk_depth1 == 0)
			{
				char					*dataS = ps->getDataPointer(frameIndex, neighbor_id);
				float32			*quatS = (float32*)(dataC+ROTATION.getOffset());

				for(card32	k = 0; k < 4; k++)
				{
					quatC[k] = quatS[k];
				}
				break;
			}
		}*/
	}	
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		Vector3f		pik = ps->get3f(2, surfelIndex, POSITION);
		Vector3f		nik = ps->get3f(2, surfelIndex, NORMAL);
		Vector3f		Eik1 = ps->get3f(2, surfelIndex, REF_U);
		int32		depth= ps->get<int32>(2, surfelIndex, TOP_DEPTH);
		//fprintf(f, "depth: %d pos: %f %f %f, nor: %f %f %f, ref_u: %f %f %f\n",
		//	depth, pik[0], pik[1], pik[2], nik[0], nik[1], nik[2], Eik1[0], Eik1[1], Eik1[2]);
		//fflush(f);
	}
	//fclose(f);
}

void	PatchGraphBSROpt::preprocessing_For_Optimization(
	const BSROptVariableIndexSet *indices_pos, 
	const BSROptVariableIndexSet *indices_nor, 
	const BSROptVariableIndexSet *indices_rot,
	PointCloudAnimationSequence	*data,
	const vector<card32>		*indices_data,
	card32	typeOfOptimization,
	PatchGraph *patchGraph)
{
/*	/// First of all, we check whether all the channels are available or not
	bool		needSampleSpacing = false;		/// Used for constructing the weights
	bool		needRefFrames			= false;		/// Used when we want to optimize surface normals
	bool		needHFieldCoeffs		= false;		/// Maybe used when Squared distance potential is applied
	bool		needRotationMatrix	= false;		/// Used when we want to optimize rotation matrices 

	for (unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++) 
	{
		card32 flags = potentials[potentialIndex]->needPrecomputedChannels(context);
		needSampleSpacing		|=	(flags & BSROptPotential::FLAG_PATCH_SAMPLE_SPACING)				!= 0;
		needRefFrames				|=	(flags & BSROptPotential::FLAG_PATCH_REF_FRAMES)							!= 0;
		needHFieldCoeffs			|=	(flags & BSROptPotential::FLAG_NEED_PATCH_2ND_ORDER_HFIELD)	!= 0;
		needRotationMatrix		|=	(flags & BSROptPotential::FLAG_NEED_ROTATION_MATRIX)				!= 0;
	}	
	if (needHFieldCoeffs) 
		needRefFrames = true;

	patchGraph->createPatchAttribute( mVAD("varIndex_pos",	1,		VAD::DATA_FORMAT_INT32));
	patchGraph->createPatchAttribute( mVAD("varIndex_nor",		1,		VAD::DATA_FORMAT_INT32));
	patchGraph->createPatchAttribute( mVAD("varIndex_rot",		1,		VAD::DATA_FORMAT_INT32));
	
	patchGraph->createPatchAttribute( NORMAL_PATCH_VAD );
	patchGraph->createPatchAttribute( TANGENT_U_PATCH_VAD );
	patchGraph->createPatchAttribute( ROTATION_MATRIX_PATCH_VAD );

	/// write varIndex_pos, varIndex_nor, varIndex_rot channels
	BOPIndexingChannel::peter_writeIndexingChannel(0, "varIndex_pos",indices_pos, patchGraph);
	BOPIndexingChannel::peter_writeIndexingChannel(3*indices_pos->size(),	"varIndex_nor",indices_nor, patchGraph);
	BOPIndexingChannel::peter_writeIndexingChannel(3*indices_pos->size() + 2*indices_nor->size(), "varIndex_rot",indices_rot, patchGraph);

	if(typeOfOptimization == 2)
	{
//		initialize_Geometry_For_Partial_Matching(indices_pos, indices_nor, indices_rot, patchGraph);
	}

	if (needRefFrames&& typeOfOptimization != 2) 
	{
		BSROptVariableIndexSet		indices_all;
		PointSet	*ps = patchGraph->getTrajectorySet()->getTrajectories();
		for(card32 frameIndex = 0; frameIndex < ps->getWidth(); frameIndex++)
		{
			for(card32 surfelIndex = 0; surfelIndex < ps->getHeight(); surfelIndex++)
			{
				BSROptPointIndex		e;
				e.frame	= frameIndex;
				e.point	=	surfelIndex;
				indices_all.push_back(e);
			}
		}
		BSRFitting::fitRefFrames(context, patchGraph, indices_all);
		//And then orient them
		UnifyNormals	normalunifier;		normalunifier.unifyNormals(patchGraph, &indices_all);
	}

	/// Compute initial rotation matrices
	if(typeOfOptimization != 2)
		compute_initial_rotation_matrices(indices_rot, patchGraph);

	///compute indices_transfer;
	card32	width = patchGraph->getTrajectorySet()->getTrajectories()->getWidth();
	index_transfer.resize(width);
	for(card32 i = 0 ; i < width; i++)
	{
		index_transfer[i] = -1;
	}
	for(card32 i = 0 ; i < indices_data->size(); i++)
	{
		index_transfer[(*indices_data)[i]] = i;
	}
	setSmoothnessFlag(typeOfOptimization, data, patchGraph);*/
		/// First of all, we check whether all the channels are available or not
	bool		needSampleSpacing = false;		/// Used for constructing the weights
	bool		needRefFrames			= false;		/// Used when we want to optimize surface normals
	bool		needHFieldCoeffs		= false;		/// Maybe used when Squared distance potential is applied
	bool		needRotationMatrix	= false;		/// Used when we want to optimize rotation matrices 

	for (unsigned int potentialIndex = 0; potentialIndex < potentials.size(); potentialIndex++) 
	{
		card32 flags = potentials[potentialIndex]->needPrecomputedChannels(context);
		needSampleSpacing		|=	(flags & BSROptPotential::FLAG_PATCH_SAMPLE_SPACING)				!= 0;
		needRefFrames				|=	(flags & BSROptPotential::FLAG_PATCH_REF_FRAMES)							!= 0;
		needHFieldCoeffs			|=	(flags & BSROptPotential::FLAG_NEED_PATCH_2ND_ORDER_HFIELD)	!= 0;
		needRotationMatrix		|=	(flags & BSROptPotential::FLAG_NEED_ROTATION_MATRIX)				!= 0;
	}	
	if (needHFieldCoeffs) 
		needRefFrames = true;

	patchGraph->createPatchAttribute( mVAD("varIndex_pos",	1,		VAD::DATA_FORMAT_INT32));
	patchGraph->createPatchAttribute( mVAD("varIndex_nor",		1,		VAD::DATA_FORMAT_INT32));
	patchGraph->createPatchAttribute( mVAD("varIndex_rot",		1,		VAD::DATA_FORMAT_INT32));
	
	patchGraph->createPatchAttribute( NORMAL_PATCH_VAD );
	patchGraph->createPatchAttribute( TANGENT_U_PATCH_VAD );
	patchGraph->createPatchAttribute( ROTATION_MATRIX_PATCH_VAD );

	/// write varIndex_pos, varIndex_nor, varIndex_rot channels
	BOPIndexingChannel::peter_writeIndexingChannel(0, "varIndex_pos",indices_pos, patchGraph);
	BOPIndexingChannel::peter_writeIndexingChannel(3*(card32)indices_pos->size(),	"varIndex_nor",indices_nor, patchGraph);
	BOPIndexingChannel::peter_writeIndexingChannel(3*(card32)indices_pos->size() + 2*(card32)indices_nor->size(), "varIndex_rot",indices_rot, patchGraph);

	if(typeOfOptimization == 2)
	{
//		initialize_Geometry_For_Partial_Matching(indices_pos, indices_nor, indices_rot, patchGraph);
	}

	if (needRefFrames&& typeOfOptimization != 2) 
	{
		BSROptVariableIndexSet		indices_all;
		PointSet	*ps = patchGraph->getTrajectorySet()->getTrajectories();
		for(card32 frameIndex = 0; frameIndex < ps->getWidth(); frameIndex++)
		{
			for(card32 surfelIndex = 0; surfelIndex < ps->getHeight(); surfelIndex++)
			{
				BSROptPointIndex		e;
				e.frame	= frameIndex;
				e.point	=	surfelIndex;
				indices_all.push_back(e);
			}
		}
		BSRFitting::fitRefFrames(context, patchGraph, indices_all);
		//And then orient them
		UnifyNormals	normalunifier;		normalunifier.unifyNormals(patchGraph, &indices_all);
	}

	/// Compute initial rotation matrices
	if(typeOfOptimization != 2)
		compute_initial_rotation_matrices(indices_rot, patchGraph);

	///compute indices_transfer;
	card32	width = patchGraph->getTrajectorySet()->getTrajectories()->getWidth();
	index_transfer.resize(width);
	for(card32 i = 0 ; i < width; i++)
	{
		index_transfer[i] = -1;
	}
	for(card32 i = 0 ; i < indices_data->size(); i++)
	{
		index_transfer[(*indices_data)[i]] = i;
	}
	setSmoothnessFlag(typeOfOptimization, data, patchGraph);
}

void		PatchGraphBSROpt::projectionOntoManifold(
	const	BSROptVariableIndexSet				*indices_related,
	PatchGraph			*pg,
	const	DVectorF	*displacement)
{
	PointSet	*ps			= pg->getTrajectorySet()->getTrajectories();

	card32		width		= ps->getWidth(),
				height		= ps->getHeight();

	AAT			POSITION	= ps->getAAT("position");
	AAT			NORMAL		= ps->getAAT("normal");
	AAT			REF_U		= ps->getAAT("tangent_u");
	AAT			ROTATION	= ps->getAAT("rotation_matrix");
	AAT			POS_VARID	= ps->getAAT("varIndex_pos");
	AAT			NOR_VARID	= ps->getAAT("varIndex_nor");
	AAT			ROT_VARID	= ps->getAAT("varIndex_rot");

	//Variables
	Vector3f				pik, nik, Eik1, Eik2;
	int						ik_posid, ik_norid, ik_rotid;
	char					*dataC;
	float32					*quatC;

	AAT			TOP_DEPTH	= ps->getAAT("topology_depth");


	for(card32 id = 0; id < indices_related->size(); id++)
	{
		card32	frameIndex	= (*indices_related)[id].frame;
		card32	surfelIndex = (*indices_related)[id].point;
		pik			= ps->get3f(frameIndex, surfelIndex,	POSITION);
		nik			= ps->get3f(frameIndex, surfelIndex,	NORMAL);
		Eik1		= ps->get3f(frameIndex, surfelIndex,	REF_U);
		Eik2		= nik.crossProduct(Eik1);
		ik_posid	= ps->get<int32>(frameIndex, surfelIndex, POS_VARID);
		ik_norid	= ps->get<int32>(frameIndex, surfelIndex, NOR_VARID);
		ik_rotid	= ps->get<int32>(frameIndex, surfelIndex, ROT_VARID);
		if(ik_posid >= 0)
		{
			for(card32 k = 0; k < 3; k++)
				pik[k] += (*displacement)[ik_posid + k];
			ps->set3f(frameIndex, surfelIndex, POSITION, pik);
		}
		if(ik_norid >= 0)
		{
			nik								+=	Eik1*(*displacement)[ik_norid]	+	Eik2*(*displacement)[ik_norid + 1];
			nik.normalize();
			ps->set3f(frameIndex, surfelIndex, NORMAL, nik);
			/// Update Ref_U
			Eik1		-=	nik*(Eik1*nik);
			Eik1.normalize();
			ps->set3f(frameIndex, surfelIndex, REF_U, Eik1);
		}
		if(ik_rotid >= 0)
		{
			//get current rotation matrix
			dataC	= ps->getDataPointer(frameIndex, surfelIndex);
			quatC	= (float32*)(dataC+ROTATION.getOffset());

			float32		theta		= sqrt((*displacement)[ik_rotid]*(*displacement)[ik_rotid] 
			+ (*displacement)[ik_rotid+1]*(*displacement)[ik_rotid+1] 
			+ (*displacement)[ik_rotid+2]*(*displacement)[ik_rotid+2]);

			float32	quat[4], quatn[4], sina;
			sina								= sin(theta/2);
			quat[0]						= cos(theta/2);
			if(theta > 1e-6)
			{
				for(card32	k = 0; k < 3; k++)
				{
					quat[k + 1]		= sina*(*displacement)[ik_rotid + k]/theta;
				}
			}
			else
			{
				quat[1] = sina;			quat[2] = 0.0f;			quat[3] = 0.0f;
			}
			quatn[0] = quat[0]*quatC[0] - quat[1]*quatC[1] - quat[2]*quatC[2] - quat[3]*quatC[3];
			quatn[1] = quat[0]*quatC[1] + quat[1]*quatC[0] + quat[2]*quatC[3] - quat[3]*quatC[2];
			quatn[2] = quat[0]*quatC[2] + quat[2]*quatC[0] + quat[3]*quatC[1] - quat[1]*quatC[2];
			quatn[3] = quat[0]*quatC[3] + quat[3]*quatC[0] + quat[1]*quatC[2] - quat[2]*quatC[1];
			theta = sqrt(quatn[0]*quatn[0]+quatn[1]*quatn[1]+quatn[2]*quatn[2]+quatn[3]*quatn[3]);
			quatn[0] /= theta;	quatn[1] /= theta;	quatn[2] /= theta;	quatn[3] /= theta;

			for(card32	k = 0; k < 4; k++)
			{
				quatC[k] = quatn[k];
			}
		}
	}
}

void		PatchGraphBSROpt::ComputeEnergy(
	const PointCloudAnimationSequence *data,
	PatchGraph *patchGraph,
	const DVectorF &perturb, 
	float &energy, 
	bool m_flag )
{
	bool		flag_SDFPotential		= potential_flags[0],
		flag_MarPotential		= potential_flags[1],
		flag_PlaSPotential		=	potential_flags[2],
		flag_NorSPotential	= potential_flags[3],
		flag_CirSPotential		= potential_flags[4],
		flag_LapSPotential	= potential_flags[5],
		flag_PosRPotential	= potential_flags[6],
		flag_NorRPotential	= potential_flags[7];


	PointSet								*ps					= patchGraph->getTrajectorySet()->getTrajectories();
	PointSet								*urshape		= patchGraph->getUrshape();
	InCorePCTopologyGraph *tg					= patchGraph->getTopology();
	PGDataPointIndices			*dataPoints	= patchGraph->getTrajectorySet()->getDataPoints();

	card32								width				=	ps->getWidth(),
		height				= ps->getHeight();

	AAT										POSITION		= ps->getAAT("position");
	AAT										NORMAL			= ps->getAAT("normal");
	AAT										REF_U				= ps->getAAT("tangent_u");
	AAT										ROTATION		= ps->getAAT("rotation_matrix");
	AAT										POS_INDEX	= ps->getAAT("varIndex_pos");
	AAT										NOR_INDEX	= ps->getAAT("varIndex_nor");
	AAT										ROT_INDEX	= ps->getAAT("varIndex_rot");
	AAT										DATA_POS_AAT	= data->getFrame(0)->getAAT( "position", 3, VAD::DATA_FORMAT_FLOAT32 );


	/// Current term rotation matrix
	Matrix3f								Aik;

	card32								offset = 0, k, numOfEdges = 0;
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		numOfEdges += tg->getNumAdjacentVertices(surfelIndex);
	}
	for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
	{
		card32		numOfEntries = tg->getNumAdjacentVertices(surfelIndex);
		for(card32 frameIndex = 0; frameIndex < width-1; frameIndex++)
		{
			Vector3f	pik	=	ps->get3f(frameIndex, surfelIndex,	POSITION);
			Vector3f	nik	= ps->get3f(frameIndex, surfelIndex,  NORMAL); 
			Vector3f	pikn	=	ps->get3f(frameIndex+1, surfelIndex,	POSITION);
			Vector3f	nikn	= ps->get3f(frameIndex+1, surfelIndex,  NORMAL); 

			int	ik_posid	= ps->get<int32>(frameIndex, surfelIndex, POS_INDEX);
			int	ik_norid	= ps->get<int32>(frameIndex, surfelIndex, NOR_INDEX);
			int	ikn_posid= ps->get<int32>(frameIndex+1, surfelIndex, POS_INDEX);
			int	ikn_norid	= ps->get<int32>(frameIndex+1, surfelIndex, NOR_INDEX);
			int	ik_rotid	= ps->get<int32>(frameIndex, surfelIndex, ROT_INDEX);

			//get current rotation matrix
			char			*dataC	= ps->getDataPointer(frameIndex, surfelIndex);
			float32	*quatc	= (float*)(dataC+ROTATION.getOffset());

			if(m_flag)
			{
				for(k = 0; k < 3; k++)
				{
					pik[k]		+= perturb[ik_posid+k];
					pikn[k]	+= perturb[ikn_posid+k];
				}
				Vector3f	Eik1 = ps->get3f(frameIndex, surfelIndex, REF_U);
				Vector3f	Eik2 = nik.crossProduct(Eik1);
				nik += Eik1*perturb[ik_norid] + Eik2*perturb[ik_norid+1];
				nik.normalize();
				Vector3f	Eikn1 = ps->get3f(frameIndex+1, surfelIndex, REF_U);
				Vector3f	Eikn2 = nik.crossProduct(Eikn1);
				nikn += Eikn1*perturb[ikn_norid] + Eikn2*perturb[ikn_norid+1];
				nikn.normalize();

				float32	theta	= sqrt(perturb[ik_rotid]*perturb[ik_rotid] 
				+ perturb[ik_rotid + 1]*perturb[ik_rotid + 1] 
				+ perturb[ik_rotid + 2]*perturb[ik_rotid + 2]);

				float32			quat[4], quatn[4], sina;
				sina						= sin(theta/2);
				quat[0]				= cos(theta/2);
				if(theta > 1e-6)
				{
					for(k = 0; k < 3; k++)
					{
						quat[k + 1]	= sina*perturb[ik_rotid + k]/theta;
					}
				}
				else
				{
					quat[1] = sina;			quat[2] = 0.0f;			quat[3] = 0.0f;
				}
				quatn[0] = quat[0]*quatc[0] - quat[1]*quatc[1] - quat[2]*quatc[2] - quat[3]*quatc[3];
				quatn[1] = quat[0]*quatc[1] + quat[1]*quatc[0] + quat[2]*quatc[3] - quat[3]*quatc[2];
				quatn[2] = quat[0]*quatc[2] + quat[2]*quatc[0] + quat[3]*quatc[1] - quat[1]*quatc[2];
				quatn[3] = quat[0]*quatc[3] + quat[3]*quatc[0] + quat[1]*quatc[2] - quat[2]*quatc[1];
				theta = sqrt(quatn[0]*quatn[0]+quatn[1]*quatn[1]+quatn[2]*quatn[2]+quatn[3]*quatn[3]);
				quatn[0] /= theta;	quatn[1] /= theta;	quatn[2] /= theta;	quatn[3] /= theta;
				Quaternion_To_Matrix3f(quatn, Aik);
			}
			else
			{
				Quaternion_To_Matrix3f(quatc, Aik);
			}
			if(flag_MarPotential)
			{
				card32 numDataPoints = dataPoints->getNumDataPoints(frameIndex, surfelIndex);
				if (numDataPoints > 0) 
				{
					card32 dataIndexInData = dataPoints->getIndex(frameIndex, surfelIndex, 0);
					Vector3f devik = pik - data->getFrame(frameIndex)->get3f(dataIndexInData, DATA_POS_AAT);
					energy	+=	context->weightMarkerError*devik.getSqrNorm();
				}
				if(frameIndex == width - 2)
				{
					numDataPoints = dataPoints->getNumDataPoints(frameIndex+1, surfelIndex);
					if (numDataPoints > 0) 
					{
						card32 dataIndexInData = dataPoints->getIndex(frameIndex+1, surfelIndex, 0);
						Vector3f devikn = pikn - data->getFrame(frameIndex+1)->get3f(dataIndexInData, DATA_POS_AAT);
						energy	+=	context->weightMarkerError*devikn.getSqrNorm();
					}
				}
			}
			for (card32 ni = 0; ni< numOfEntries; ni++) 
			{
				card32		neighbor_id =		tg->getAdjacentVertex(surfelIndex, ni);
				Vector3f		pjk	=	ps->get3f(frameIndex, neighbor_id, POSITION);
				Vector3f		njk	= ps->get3f(frameIndex, neighbor_id,  NORMAL);
				Vector3f		pjkn	=	ps->get3f(frameIndex+1, neighbor_id, POSITION);
				Vector3f		njkn	= ps->get3f(frameIndex+1, neighbor_id,  NORMAL);

				int		jk_posid	= ps->get<int32>(frameIndex, neighbor_id, POS_INDEX);
				int		jk_norid	= ps->get<int32>(frameIndex, neighbor_id, NOR_INDEX);
				int		jkn_posid= ps->get<int32>(frameIndex+1, neighbor_id, POS_INDEX);
				int		jkn_norid	= ps->get<int32>(frameIndex+1, neighbor_id, NOR_INDEX);

				if(m_flag)
				{
					for(k = 0; k < 3; k++)
					{
						pjk[k]		+= perturb[jk_posid+k];
						pjkn[k]	+= perturb[jkn_posid+k];
					}
					Vector3f	Ejk1 = ps->get3f(frameIndex, neighbor_id, REF_U);
					Vector3f	Ejk2 = njk.crossProduct(Ejk1);
					njk += Ejk1*perturb[jk_norid] + Ejk2*perturb[jk_norid+1];
					njk.normalize();
					Vector3f	Ejkn1 = ps->get3f(frameIndex+1, neighbor_id, REF_U);
					Vector3f	Ejkn2 = njk.crossProduct(Ejkn1);
					njkn += Ejkn1*perturb[jkn_norid] + Ejkn2*perturb[jkn_norid+1];
					njkn.normalize();
				}
				Vector3f		rijk	= pik	- pjk;
				Vector3f		rijkn	=	pikn	- pjkn;
				Vector3f		nijk	=	nik	- njk;
				Vector3f		nijkn= nikn	- njkn;
				Vector3f		rijkc	= (Aik.transpose())*rijk;
				Vector3f		nijkc=	(Aik.transpose())*nijk;
				float32		weight;
				float32		rweightijk	= edge_weight_rigidity[offset+ni],
					sweightijk	= edge_weight_smoothness[offset+ni],
					sweightijkn	=	edge_weight_rigidity[offset+numOfEdges+ni];
				if(flag_NorRPotential)
				{
					weight = context->weightNormalRigidity*rweightijk/numOfEntries;
					Vector3f	devijk = nijkc - nijkn;
					energy += weight*devijk.getSqrNorm();
				}
				if(flag_PosRPotential)
				{
					weight = context->weightPositionRigidity*rweightijk/numOfEntries;
					Vector3f	devijk = rijkc - rijkn;
					energy += weight*devijk.getSqrNorm();
				}
				float32	weightijk		= exp(context->lambda*rijk.getSqrNorm());
				float32	weightijkn;
				if(frameIndex == width-2)
				{
					weightijkn = exp(context->lambda*rijkn.getSqrNorm());
				}
				if(flag_LapSPotential)
				{

				}
				if(flag_CirSPotential)
				{
					weight = context->weightCircularSmoothness*sweightijk*weightijk/numOfEntries;
					float32	dis = (nik + njk)*rijk;
					energy	+=	weight*dis*dis;
					if(frameIndex == width - 2)
					{
						weight = context->weightCircularSmoothness*sweightijkn*weightijkn/numOfEntries;
						dis = (nikn + njkn)*rijkn;
						energy	+=	weight*dis*dis;
					}
				}
				if(flag_NorSPotential)
				{
					weight = context->weightNormalSmoothness*sweightijk*weightijk/numOfEntries;
					energy	+=	weight*nijk.getSqrNorm();
					if(frameIndex == width - 2)
					{
						weight = context->weightNormalSmoothness*sweightijkn*weightijkn/numOfEntries;
						energy	+=	weight*nijkn.getSqrNorm();
					}
				}
				if(flag_PlaSPotential)
				{
					weight = context->weightPlanarSmoothness*sweightijk*weightijk/numOfEntries;
					float32	dis = rijk*nik;
					energy	+= weight*dis*dis;
					if(frameIndex == width - 2)
					{
						weight = context->weightPlanarSmoothness*sweightijkn*weightijkn/numOfEntries;
						dis = rijkn*nikn;
						energy	+=	weight*dis*dis;
					}
				}
			}
			offset += numOfEntries;
		}
	}
}
void		PatchGraphBSROpt::lineSearch_Armijo(
	PointCloudAnimationSequence *data, 
	PatchGraph *patchGraph, 
	DVectorF	&direction, 
	float32	&cEnergy, 
	float32	&nEnergy, 
	float	&stepsize)
{
	ComputeEnergy(data, patchGraph, direction, cEnergy, false);
	stepsize		=	2.0f;		direction	*= 2.0f;
	do 
	{
		direction		/= 2;		stepsize		/= 2.0f;
		ComputeEnergy(data, patchGraph, direction, nEnergy, true);
	} while(nEnergy > (1.0 - context->minimumEnergyDecay)*cEnergy && stepsize > context->minimumStepsize);
}
