//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "AnimationBSRRenderer.h"
#include "StandardSceneGraphIterator.h"
#include "PointSetKNNQuery.h"
#include "Feature.h"
#include "PointSet_templCode.h"
#include "FeatureCorrespondences.h"
#include "CorrespondenceList.h"
#include "Monomials2ndOrder.h"
#include "PatchGraphOpticalFlow.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( AnimationBSRRenderer, 8 )
{
	BEGIN_CLASS_INIT( AnimationBSRRenderer )
	ADD_INT32_PROP( showPGIndex, 5 )
	ADD_INT32_PROP( monitorPoint, 8 )
	ADD_FLOAT32_PROP( renderScale, 8 )

	ADD_OBJECT_PROP( patchGraphRenderer, 0, PatchGraphRenderer::getClass(), NULL )
	
	ADD_BOOLEAN_PROP( showTensorVotingSigma, 0 )
	ADD_BOOLEAN_PROP( showMergeCandidates, 0 )
	ADD_BOOLEAN_PROP( showSamplingSpacing, 2 )
	ADD_BOOLEAN_PROP( showNormals, 1 )
	ADD_BOOLEAN_PROP( showHeightFieldMonomial, 7 )
	ADD_BOOLEAN_PROP( showOpticalFlow, 8 )	
	ADD_BOOLEAN_PROP( showFeaturePoints, 3 )
	ADD_BOOLEAN_PROP( showAllPatchGraphs, 4 )
	ADD_BOOLEAN_PROP( drawFeatureCorrespondences, 6 )
}

/**
 * Constructor
 */
AnimationBSRRenderer::AnimationBSRRenderer()
{
	context = NULL;
	selecter = NULL;
	merger = NULL;
	rt = new RenderingTool();
	rt->init();
	
	monitorPoint = 150;
	currentFrame = -1;
	renderScale = 1.0f;
	showPGIndex = -1;
	
	showTensorVotingSigma = false;
	showMergeCandidates = true;
	showNormals = false;
	showSamplingSpacing = false;
	showFeaturePoints = true;
	showAllPatchGraphs = false;
	drawFeatureCorrespondences = true;
	showHeightFieldMonomial = false;
	showOpticalFlow = false;

	patchGraphRenderer = new StandardPatchGraphRenderer();
	patchGraphRenderer->setShowPoints( true );
	patchGraphRenderer->setShowTopology( true );
	patchGraphRenderer->setShowCoords( false );
	patchGraphRenderer->setShowDataPoints( false );
	patchGraphRenderer->setUseDL( false );
	patchGraphRenderer->setShowAllFrames( false );
	patchGraphRenderer->setShowCorrespondences( false );
	patchGraphRenderer->setShowUrshape( false );
}

/**
* Destructor
*/
AnimationBSRRenderer::~AnimationBSRRenderer()
{
	delete patchGraphRenderer;
	delete rt;
}

/**
* Call rendering functionality
*/
void AnimationBSRRenderer::render()
{
	if ( context == NULL ) return;

	rt->enableLighting();
	
	card32 globalFrame = getCurrentFrameIndex();

	if ( showOpticalFlow ){
		renderGLOpticalFlow();
	}
	
	// Render patch graphs in current frame
	card32 currentFrame = getCurrentFrameIndex();
	int32 visibleIndex = 0;
	PatchGraphContainer* pgContainer = context->getPGContainer();
	if ( pgContainer != NULL ){
		card32 pgCounter = 0;
		for ( card32 patchGraphIndex = 0; patchGraphIndex < pgContainer->getNumPatchGraphs(); patchGraphIndex++ ){
			PatchGraph* pg = pgContainer->getPatchGraph( patchGraphIndex );
			AAT POSITION = pg->getTrajectorySet()->getTrajectories()->getAAT( "position" );
			AAT NORMAL = pg->getTrajectorySet()->getTrajectories()->getAAT( "normal" );
			int32 localFrame = globalFrame - pg->getStartFrameIndex();
			PointCloudAnimationSequence* animSeq = pgContainer->getAnimSeq();
			if ( pg != NULL ){
				PointSet* ps = pg->getTrajectorySet()->getTrajectories();
				if ( pg->containsFrame( localFrame ) || showAllPatchGraphs 
					|| (drawFeatureCorrespondences && pg->containsFrame( localFrame+1 ))) {
					if ( ( ( showPGIndex < 0 ) || ( showPGIndex == visibleIndex ) ) || showAllPatchGraphs 
						|| (drawFeatureCorrespondences && pg->containsFrame( localFrame+1 ))){
						patchGraphRenderer->renderGL( pg, animSeq, currentFrame );
						pgCounter++;

						// Height field monomial
						if ( showHeightFieldMonomial ){
							// Extract coefficients
							if ( ps->providesAttribute( "normal" ) && ps->providesAttribute( "tangent_u" ) && ps->providesAttribute( "hfieldCoeffs" ) ){
								for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
									if ( monitorPoint == patchIndex ){
										Vector3f pos = ps->get3f( localFrame, patchIndex, POSITION );
										Vector3f normal = ps->get3f( localFrame, patchIndex, NORMAL );
										AAT REF_U = ps->getAAT( "tangent_u" );
										AAT REF_N = ps->getAAT( "normal" );
										AAT COEFF = ps->getAAT( "hfieldCoeffs" );

										StaticVector<float32,6> coeff;
										Matrix3f T;
										Vector3f posT;
										coeff = ps->get<float32,6>( localFrame, patchIndex, COEFF );
										T[0] = ps->get<float32,3>( localFrame, patchIndex, REF_U );
										T[2] = ps->get<float32,3>( localFrame, patchIndex, REF_N );
										T[1] = T[2].crossProduct(T[0]); // tangent_v = normal x tangent_u
										for ( int32 i = -5; i < 5; i++ ){
											for ( int32 j = -5; j < 5; j++ ){
												float32 x = float32( i ) / 30.0f;
												float32 y = float32( j ) / 30.0f;
												float32 n = Monomials2ndOrder::evaluate( coeff, makeVector2f( x,y ) );
												posT = T * makeVector3f( x,y,n ) + pos;
												rt->setRenderColor( 1,1,0 );
												rt->renderSphere( posT, 0.003f );
											}
										}
									}
								}
							}
						}

						// Normals
						if ( showNormals ){
							rt->disableLighting();
							rt->setRenderColor( 1,0,0 );
							for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
								Vector3f pos = pg->getTrajectorySet()->getTrajectories()->get3f( localFrame, patchIndex, POSITION );
								Vector3f normal = pg->getTrajectorySet()->getTrajectories()->get3f( localFrame, patchIndex, NORMAL );
								rt->renderLine( pos, pos + normal * 0.05f * context->getRenderScale() );
							}
							rt->enableLighting();
						}

						// Feature points
						if( showFeaturePoints ) {
							//rt->setRenderColor(1,0,0);
							for( card32 f=0;f<pg->getNumFeatureLists();f++ )
							{
								if( pg->getNumFeatureLists() <= f )
									break;
								if( pg->getStartFrameIndex() + f != currentFrame &&
									pg->getStartFrameIndex() + f != currentFrame+1 && drawFeatureCorrespondences)
									continue;

								FeatureList * features = pg->getFeatures(f);
								Vector3f translate = (pg->getStartFrameIndex() + f == currentFrame+1 && drawFeatureCorrespondences) ? patchGraphRenderer->getTranslationVector() : NULL_VECTOR3F;
								for( card32 i=0;i<features->size();i++ )
								{
									Feature *f = features->get(i);
									card32 col = f->getColor();
									rt->setRenderColor( col & 0xff, (col >> 8) & 0xff, (col >> 16) & 0xff );
									rt->renderSphere( f->getPosition() + translate, context->getRenderScale() * 0.1f );

								}
							}
						}
					}
					visibleIndex++;
				}
			}
		}
		if( drawFeatureCorrespondences )
		{
			glPushAttrib(GL_ALL_ATTRIB_BITS);
			glDisable(GL_LIGHTING);
			glDepthMask(0);
//			glLineWidth(1.0f);
			glEnable(GL_LINE_SMOOTH);
			glBegin( GL_LINES );
			//glColor4f(1.0f,0.1f,0.1f,0.5f );
			if( pgContainer->getFeatureCorrespondences()->getNumFrames() > currentFrame )
			{
				if( showAllPatchGraphs )
				{
					for( int fi = 0;fi < (int)pgContainer->getFeatureCorrespondences()->getNumFrames();fi++ )
					{
						CorrespondenceList * corr = pgContainer->getFeatureCorrespondences()->getFrameCorrespondences(fi);
						for( card32 i=0;i<corr->getNumCorrespondences();i++ )
						{
							//Vector3f col = corr->get(i)->getColor();
							//glColor3fv( &col[0] );
							if( corr->get(i)->getFlags() & Correspondence::CF_Initial )
								continue;

							/*if( corr->get(i)->getFlags() & Correspondence::CF_Guessed )
								glColor3f(1,0,1);
							else*/
								glColor3f(0.85f,0.85f,0);
							Vector3f p1 = corr->get(i)->getPosition1() + patchGraphRenderer->getTranslationVector()*(fi - (int)currentFrame);
							Vector3f p2 = corr->get(i)->getPosition2() + patchGraphRenderer->getTranslationVector()*(fi - (int)currentFrame +1);
							glVertex3fv( &p1[0] );
							glVertex3fv( &p2[0] );
						}
					}
				}
				else
				{
					CorrespondenceList * corr = pgContainer->getFeatureCorrespondences()->getFrameCorrespondences(currentFrame);
					for( card32 i=0;i<corr->getNumCorrespondences();i++ )
					{
						//Vector3f col = corr->get(i)->getColor();
						//glColor3fv( &col[0] );
						if( corr->get(i)->getFlags() & Correspondence::CF_Initial )
							continue;

						/*if( corr->get(i)->getFlags() & Correspondence::CF_Guessed )
							glColor3f(1,0,1);
						else*/
							glColor3f(1,1,0);
						Vector3f p1 = corr->get(i)->getPosition1();
						Vector3f p2 = corr->get(i)->getPosition2() + patchGraphRenderer->getTranslationVector();
						glVertex3fv( &p1[0] );
						glVertex3fv( &p2[0] );
					}
				}
			}
			glEnd();
			glPopAttrib();
		}

	}

	// Show merge candidates
	if ( showMergeCandidates && ( pgContainer != NULL ) && ( selecter != NULL ) ){	
		PatchGraph* pg1 = pgContainer->getPatchGraph( selecter->getCand1() );
		PatchGraph* pg2 = pgContainer->getPatchGraph( selecter->getCand2() );	
		if ( ( pg1 != NULL ) && ( pg2 != NULL ) ){
			rt->setRenderColor( makeVector3f( rand()%100/100.0f, rand()%100/100.0f, rand()%100/100.0f ) );
			if (!showAllPatchGraphs) {
				if ( pg1->containsFrame( currentFrame - pg1->getStartFrameIndex() ) ){
					rt->renderBox( pg1->getBoundingBox() );
				}
				if ( pg2->containsFrame( currentFrame - pg2->getStartFrameIndex() ) ){
					rt->renderBox( pg2->getBoundingBox() );
				}
			} else {
				glMatrixMode(GL_MODELVIEW);
				glPushMatrix();
				float32 timeShift = (int32)pg1->getStartFrameIndex() - (int32)currentFrame;
				Matrix4f M = makeTranslation4f(patchGraphRenderer->getInterFrameTranslation()*timeShift);
				glMultMatrixf(M.data());
				rt->renderBox( pg1->getBoundingBox() );
				glPopMatrix();
				glPushMatrix();
				timeShift = (int32)pg2->getStartFrameIndex() - (int32)currentFrame;
				M = makeTranslation4f(patchGraphRenderer->getInterFrameTranslation()*timeShift);
				glMultMatrixf(M.data());
				rt->renderBox( pg2->getBoundingBox() );
				glPopMatrix();
			
			}
			
		}
	}

	// Create octree
	UnstructuredInCorePointCloud* pc = getDataFrame( getCurrentFrameIndex() );
	if ( ( pc != NULL ) && ( showTensorVotingSigma || showSamplingSpacing ) ){
		PointSet* ps = pc->getPointSet();
		AAT POSITION = ps->getAAT( "position" );
		AttachedIndexedOctree *octree = NULL;
		PointSetKNNQuery* knn = NULL;
		card32 nIndex;
		bool found;
		if ( ps != NULL ){
			octree = PointSetKNNQuery::createOctree(ps);
			knn = new PointSetKNNQuery( ps, octree, NULL_VECTOR3F, ps->getAAT("position") );
		}
			
		if ( ( ps != NULL ) && showTensorVotingSigma ){
			rt->setRenderColor( 0.8f, 0.8f, 0.2f );
			Vector3f pos = ps->get3f( monitorPoint, POSITION );
			knn->setTargetPoint( pos );
			knn->getNextNearestPointIndex( nIndex, found );
			float32 d = norm( pos - ps->get3f( nIndex, POSITION ) );
			while( found && ( d < context->getTVSigma() ) ){
				Vector3f nPos = ps->get3f( nIndex, POSITION );
				rt->renderSphere( nPos, 0.01f * context->getRenderScale() );
				knn->getNextNearestPointIndex( nIndex, found );
				d = norm( ps->get3f( nIndex, POSITION ) - pos );
			}
		}
		if ( ( ps != NULL ) && showSamplingSpacing ){
			rt->setRenderColor( 0.8f, 0.2f, 0.9f );
			Vector3f pos = ps->get3f( monitorPoint, POSITION );
			knn->setTargetPoint( pos );
			knn->getNextNearestPointIndex( nIndex, found );
			float32 d = norm( pos - ps->get3f( nIndex, POSITION ) );
			while( found && ( d < context->getSamplingSpacing() ) ){
				Vector3f nPos = ps->get3f( nIndex, POSITION );
				rt->renderSphere( nPos, 0.01f * context->getRenderScale() );
				knn->getNextNearestPointIndex( nIndex, found );
				d = norm( ps->get3f( nIndex, POSITION ) - pos );
			}
		}

		if ( octree != NULL ) delete octree;
		if ( knn != NULL ) delete knn;
	}
}

/** 
* Get number of frames
*/
card32 AnimationBSRRenderer::getNumFrames() const
{
	SGSimpleAnimationNode* animDataNode = getDataAnimationNode();
	if ( animDataNode == NULL ){
		return 0;
	}
	return animDataNode->getNumChildNodes( scene->getRootState() );
}

/**
* Get the data animation node
*/
SGSimpleAnimationNode* AnimationBSRRenderer::getDataAnimationNode() const
{
	// Find data node
	StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
	SGSimpleAnimationNode* animDataNode = NULL;
	it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
	while ( !it->atEnd() ){
		if ( it->getNameWithPath() == "root/data" ){
			animDataNode = dynamic_cast<SGSimpleAnimationNode*>( it->getNode() );
			if ( animDataNode != NULL ){
				break;            
			}
		}
		it->next();
	}
	delete it;
	return animDataNode;
}

/**
* Return index of current frame
*/
int32 AnimationBSRRenderer::getCurrentFrameIndex() const
{
	return scene->getRootState()->dynamicState->time;
}

/**
* Get data point cloud for frame index
*/
UnstructuredInCorePointCloud* AnimationBSRRenderer::getDataFrame( card32 index )
{
	return dynamic_cast<UnstructuredInCorePointCloud*>( getPointCloud( scene, getDataFrameName( index ) ) ); 
}

/** 
 * Debug visualization of the optical flow
 */
void AnimationBSRRenderer::renderGLOpticalFlow()
{
	// Find patch graph
	card32 frameIndex = getCurrentFrameIndex();
	PatchGraph* pg = NULL;
	for ( unsigned int patchGraphIndex = 0; patchGraphIndex < context->getPGContainer()->getNumPatchGraphs(); patchGraphIndex++ ){
		pg = context->getPGContainer()->getPatchGraph( patchGraphIndex );
		if ( ( frameIndex >= pg->getStartFrameIndex() ) && ( frameIndex < pg->getStartFrameIndex() + pg->getTrajectorySet()->getNumFrames() ) ){
			break;
		} else {
			pg = NULL;
		}
	}

	if ( ( pg == NULL ) || ( merger == NULL ) ){
		debugOutput << "renderGLOpticalFlow: invalid settings.\n"; debugOutput.flush();
		return;
	}
	PointSet* traj = pg->getTrajectorySet()->getTrajectories();
	AAT POSITION = traj->getAAT( "position" );
	card32 numPatches = pg->getTrajectorySet()->getNumPatches();
	if ( monitorPoint >= int32( numPatches ) ){
		debugOutput << "renderGLOpticalFlow: invalid monitor point.\n"; debugOutput.flush();
		return;
	}

	/*
	/// Optical flow
	if ( monitorPoint < 0 ){
		// Compute all
		vector<PatchGraphOpticalFlowGradient> gradients;
		if ( merger->getOpticalFlow()->computeAll( gradients, pg, frameIndex, context->getPGContainer()->getAnimSeq(), context->getSamplingSpacing() ) ){
			for ( unsigned int pointIndex = 0; pointIndex < pg->getTrajectorySet()->getNumPatches(); pointIndex++ ){
				Vector3f pos = traj->get3f( frameIndex, pointIndex, POSITION );
				rt->setRenderColor( 1,0,0 );
				rt->renderArrow( pos, pos + gradients[pointIndex].g * renderScale, 1.0f );
			}
		}
	} else {
		// Compute single gradient
		PatchGraphOpticalFlowGradient gradient;
		if ( merger->getOpticalFlow()->compute( gradient, pg, frameIndex, monitorPoint, context->getPGContainer()->getAnimSeq(), context->getSamplingSpacing() ) ){
			Vector3f pos = traj->get3f( frameIndex, monitorPoint, POSITION );
			rt->setRenderColor( 1,0,0 );
			rt->renderArrow( pos, pos + gradient.g * renderScale, 1.0f );
		}
	}
	*/
}