//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Monomial2ndOrderDisplayer.h"
//---------------------------------------------------------------------------
#include "GLEventDisplayerWidget.h"
#include "DebugOutput.h"
#include "GLHelper.h"
#include "NumericalClassProperty.h"
#include "FixedArrayClassProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( Monomial2ndOrderDisplayer, 0 ) 
{
	BEGIN_CLASS_INIT( Monomial2ndOrderDisplayer )
	ADD_VECTOR2F_PROP( domainU, 0 )
	ADD_VECTOR2F_PROP( domainV, 0 )
	ADD_CARD32_PROP( resolution, 0 )
}

Monomial2ndOrderDisplayer::Monomial2ndOrderDisplayer() 
{
	domainU = makeVector2f( 0,1 );
	domainV = makeVector2f( 0,1 );
	resolution = 10;
	for ( unsigned int n = 0; n < 6; n++ ){
		coeffs[n] = 0;
	}
	points.clear();
	hasSaddlePoint = false;
	saddleCoords = makeVector2f( 0,0 );
}

BoundingBox3f Monomial2ndOrderDisplayer::getBoundingBox() 
{
	return BoundingBox3f( makeVector3f( domainU[0], domainV[0], 0 ), makeVector3f( domainU[1], domainV[1], 1 ) );
}

/** 
 * Setup
 */
void Monomial2ndOrderDisplayer::setup( StaticVector<float32,6> coeffs, Vector2f domainU, Vector2f domainV, card32 resolution )
{
	this->domainU = domainU;
	this->domainV = domainV;
	this->resolution = resolution;
	this->coeffs = coeffs;
}

void Monomial2ndOrderDisplayer::drawGL( QGLWidget *widget, Camera &cam ) 
{
	rt.disableLighting();

	/*
	glDisable( GL_LIGHTING ) ;
	glBegin( GL_LINES );
	glColor3f( 1,1,0 );
	glVertex3f( -1,-1,-1 );
	glVertex3f(1,1,1) ;
	glEnd();
	glEnable( GL_LIGHTING );
	*/

	// Show additional points
	rt.setRenderColor( 0,0,1 );
	for ( unsigned int n = 0; n < points.size(); n++ ){
		rt.renderPoint( points[n] );		
	}

	// Show tangents
	rt.setRenderColor( 0,1,1 );
	Vector3f origin = makeVector3f( 0.5f, 0.5f, Monomials2ndOrder::evaluate( coeffs, makeVector2f( 0.5f, 0.5f ) ) );
	rt.renderLine( origin, origin + t_u );		
	rt.renderLine( origin, origin + t_v );		

	// Show saddle point
	if ( hasSaddlePoint ){
		rt.setRenderColor( 1,0,1 );
		rt.renderLine( makeVector3f( saddleCoords[0], saddleCoords[1], 0 ), makeVector3f( saddleCoords[0], saddleCoords[1], 0 ) + makeVector3f( 0,0,1 ) );
	}

	rt.setRenderColor( 1,1,0 );
	float32 dx = ( domainU[1] - domainU[0] ) / float32( resolution );
	float32 dy = ( domainV[1] - domainV[0] ) / float32( resolution );
	for ( unsigned int i = 0; i <= resolution; i++ ){
		float32 x = domainU[0] + dx * i;
		for ( unsigned int j = 0; j < resolution; j++ ){
			float32 y0 = domainV[0] + dy * j;
			float32 y1 = domainV[0] + dy * (j+1);
			float32 h0 = Monomials2ndOrder::evaluate( coeffs, makeVector2f( x,y0 ) );
			float32 h1 = Monomials2ndOrder::evaluate( coeffs, makeVector2f( x,y1 ) );
			rt.renderLine( makeVector3f( x, y0, h0 ), makeVector3f( x, y1, h1 ) );
		}
	}
	for ( unsigned int i = 0; i <= resolution; i++ ){
		float32 y = domainV[0] + dy * i;
		for ( unsigned int j = 0; j < resolution; j++ ){
			float32 x0 = domainU[0] + dx * j;
			float32 x1 = domainU[0] + dx * (j+1);
			float32 h0 = Monomials2ndOrder::evaluate( coeffs, makeVector2f( x0,y ) );
			float32 h1 = Monomials2ndOrder::evaluate( coeffs, makeVector2f( x1,y ) );
			rt.renderLine( makeVector3f( x0, y, h0 ), makeVector3f( x1, y, h1 ) );
		}
	}

	float32 x0 = domainU[0]; 
	float32 x1 = domainU[1]; 
	float32 y0 = domainV[0];
	float32 y1 = domainV[1];
	rt.setRenderColor( 0,1,0 );
	rt.renderLine( makeVector3f( x0, y0, 0 ), makeVector3f( x0, y0, 1 ) );
	rt.renderLine( makeVector3f( x0, y1, 0 ), makeVector3f( x0, y1, 1 ) );
	rt.renderLine( makeVector3f( x1, y0, 0 ), makeVector3f( x1, y0, 1 ) );
	rt.renderLine( makeVector3f( x1, y1, 0 ), makeVector3f( x1, y1, 1 ) );
	rt.renderLine( makeVector3f( x0, y0, 0 ), makeVector3f( x0, y1, 0 ) );
	rt.renderLine( makeVector3f( x0, y1, 0 ), makeVector3f( x1, y1, 0 ) );
	rt.renderLine( makeVector3f( x1, y1, 0 ), makeVector3f( x1, y0, 0 ) );
	rt.renderLine( makeVector3f( x1, y0, 0 ), makeVector3f( x0, y0, 0 ) );

}

void Monomial2ndOrderDisplayer::updateData( GLEventObject *other ) 
{
	Monomial2ndOrderDisplayer *cOther = dynamic_cast<Monomial2ndOrderDisplayer*>(other);
}


Monomial2ndOrderDisplayer::~Monomial2ndOrderDisplayer() 
{
}

GLEventConfiguration *Monomial2ndOrderDisplayer::createConfigWindow(QWidget *parent) 
{
	return new Monomial2ndOrderDisplayerConfig( parent, this );
}

Monomial2ndOrderDisplayerConfig::Monomial2ndOrderDisplayerConfig( QWidget *parent, GLEventObject *obj ) : GLEventConfiguration( parent, obj )
{
}	

#include "Monomial2ndOrderDisplayer_moc.cpp"