//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphRenderer.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "TrajectorySet.h"
#include "VertexDescriptor.h"
#include "PatchTypes.h"
#include "NumericalClassProperty.h"
#include "GLHelper.h"
#include "SeparatorClassProperty.h"
#include "InCorePCTopologyGraph.h"
#include "PointCloudAnimationSequence.h"
#include "HFBFBivariateMonomials.h"
#include "ClassMethods.h"
#include "PGDataPointIndices.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_ABSTRACT_CLASS(PatchGraphRenderer, 0)
{
	BEGIN_CLASS_INIT(PatchGraphRenderer);
}

static string describeCoordsDrawStyle(card32 v) {
	if (v==0) return "COORDS_SIMPLE";
	if (v==1) return "COORDS_ARROWS";
	if (v==2) return "COORDS_XYZ";
	return "undefined";
};

static string describeInterpKernel(card32 v) {
	if (v==0) return "PATCH_INTERP_LIN";
	if (v==1) return "PATCH_INTERP_GAUSSIAN";
	if (v==2) return "PATCH_INTERP_COS";
	return "undefined";
};
											 
static string describePtsColorMode(card32 v) {
	if (v==0) return "CM_PATCH_GRAPH_COLOR";
	if (v==1) return "CM_COLOR_CHANNEL_COLOR";
	if (v==2) return "CM_FIXED_COLOR";
	if (v==3) return "CM_CLUSTER_COLOR";
	if (v==4) return "CM_POTENTIALS";
	return "undefined";
};

static string describeColorMode(card32 v) {
	if (v==0) return "CM_PATCH_GRAPH_COLOR";
	if (v==1) return "CM_COLOR_CHANNEL_COLOR";
	if (v==2) return "CM_FIXED_COLOR";
	if( v==CM_CLUSTER_COLOR ) return "CM_CLUSTER_COLOR";
	return "undefined";
};


IMPLEMENT_GEOXL_CLASS(StandardPatchGraphRenderer, 11)
{
	BEGIN_CLASS_INIT(StandardPatchGraphRenderer);

	ADD_SEPARATOR("options")
	ADD_BOOLEAN_PROP(showPoints, 0)
	ADD_BOOLEAN_PROP(showCoords, 0)
	ADD_BOOLEAN_PROP(showCorrDataPoints, 0)
	ADD_BOOLEAN_PROP(showAllDataPoints, 6)
	ADD_BOOLEAN_PROP(showLocalPatches, 0)
	ADD_BOOLEAN_PROP(showInterpolatedPatches, 0)
	ADD_BOOLEAN_PROP(showTopology, 0)
	ADD_BOOLEAN_PROP(showSampleSpacing, 0)
	ADD_BOOLEAN_PROP(showCorrespondences, 4)
	ADD_SEPARATOR("range")
	ADD_BOOLEAN_PROP(showAllFrames, 5)
	ADD_VECTOR3F_PROP(afInterFrameTransl, 5);
	ADD_BOOLEAN_PROP(showUrshape, 5)
	ADD_VECTOR3F_PROP(urshapeTransl, 5)
	ADD_BOOLEAN_PROP(globalTimeTranslation, 8)

	ADD_SEPARATOR("parameters")
	ADD_BOOLEAN_PROP(useDList, 3)
	ADD_FLOAT32_PROP(pointSize, 1)
	ADD_CARD32_PROP_UPDATE_DESCR(coordsDrawStyle, 1, NULL, describeCoordsDrawStyle, 3)
	ADD_FLOAT32_PROP(coordsSize, 1)
	ADD_FLOAT32_PROP(patchSize, 2)
	ADD_CARD32_PROP(patchSubdiv, 2)
	ADD_BOOLEAN_PROP(patchLighting, 2)
	ADD_CARD32_PROP_UPDATE_DESCR(interpKernel, 1, NULL, describeInterpKernel, 3)
	ADD_FLOAT32_PROP(kernelWidthFactor, 3)
	
	ADD_SEPARATOR("colors")
	ADD_CARD32_PROP_UPDATE_DESCR(pointsColorMode,		  6, NULL, describePtsColorMode, 5)
	ADD_CARD32_PROP_UPDATE_DESCR(topologyColorMode,		  6, NULL, describeColorMode, 3)
	ADD_CARD32_PROP_UPDATE_DESCR(dataPointsColorMode,	  6, NULL, describeColorMode, 3)
	ADD_CARD32_PROP_UPDATE_DESCR(allDataPointsColorMode, 6, NULL, describeColorMode, 3)
	
	ADD_VECTOR4F_PROP(pointsColor, 6)
	ADD_VECTOR4F_PROP(corrDataPointsColor, 6)
	ADD_VECTOR4F_PROP(otherDataPointsColor, 6)
	ADD_VECTOR4F_PROP(topologyColor, 6)
	ADD_VECTOR4F_PROP(urshapePointColor, 6)
	ADD_VECTOR4F_PROP(urshapeTopologyColor, 6)
	
	ADD_BOOLEAN_PROP(extraAlpha, 9)
	
	ADD_FLOAT32_PROP(minPotValue, 10)
	ADD_FLOAT32_PROP(maxPotValue, 10)
	ADD_VECTOR3F_PROP(potMask, 10)
	ADD_CARD32_PROP(correspondencesOffset, 11)
	
	ADD_NOARGS_METHOD(rebuildDList)
}


// ----


void PatchGraphRenderer::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);	
}


// --


StandardPatchGraphRenderer::StandardPatchGraphRenderer() {
	showPoints = true;
	showCoords = true;
	showCorrDataPoints = true;
	showAllDataPoints = true;
	showLocalPatches = false;
	showInterpolatedPatches = false;
	showTopology = false;
	showSampleSpacing = false;

	showCorrespondences = true;
	showAllFrames = true;
	showUrshape = true;

	afInterFrameTransl = makeVector3f(1.2f,0,0);
   
	urshapeTransl = makeVector3f(0,1,0);
	globalTimeTranslation = true;
	
	useDList = true;
	needsUpdate = true;
	pointSize = 1.0f;
	coordsDrawStyle = COORDS_ARROWS;	
	coordsSize = 0.05f;
	patchSize = 0.10f;
	patchSubdiv = 8;
	patchLighting = true;
	dlist = -42;
	interpKernel = PATCH_INTERP_LIN;
	kernelWidthFactor = 1.0f;

	pointsColorMode = CM_FIXED_COLOR;
	topologyColorMode = CM_FIXED_COLOR;
	dataPointsColorMode = CM_FIXED_COLOR;
	allDataPointsColorMode = CM_FIXED_COLOR;
	
	pointsColor = makeVector4f(1, 1, 1, 1);
	corrDataPointsColor = makeVector4f(1, 1, 1, 1);
	//otherDataPointsColor = makeVector4f(0.3f, 0.3f, 0.4f, 1);
	//otherDataPointsColor = makeVector4f(0.9f, 0.3f, 0.4f, 1);
	//otherDataPointsColor = makeVector4f(1.0f, 1.0f, 1.0f, 0.1f);
	otherDataPointsColor = makeVector4f(0.4f, 0.2f, 1.0f, 0.2f);
	topologyColor = makeVector4f(1, 0.6f, 0.4f, 0.1f);
	urshapePointColor = makeVector4f(0.8f, 0.8f, 0.8f, 1.0f);
	urshapeTopologyColor = makeVector4f(0.4f, 0.35f, 0.7f, 0.1f);
	
	extraAlpha = true;
	minPotValue = 0.0f;
	maxPotValue = 1.0f;
	potMask = makeVector3f(1,1,1);
	correspondencesOffset = 1;
}

inline float32 StandardPatchGraphRenderer::calcWeight(const Vector3f &neighborPos, const Vector3f &pos, float32 kernelSize) {
	float32 relDist = norm(neighborPos - pos) / kernelSize;
	if (interpKernel == PATCH_INTERP_LIN) {
		float32 r = 1-relDist;
		if (r>=0) 
			return r;
		else
			return 0;
	} else if (interpKernel == PATCH_INTERP_GAUSSIAN) {
		return exp(-relDist*relDist);
	} else if (interpKernel == PATCH_INTERP_COS) {
		relDist *= (float32)M_PI;
		if (relDist > M_PI) return 0;
		return 0.5f*(1+cos(relDist));
	}
	return 1.0f;
}


/**
 * Global time used (use startFrameIndex)
 */
void StandardPatchGraphRenderer::renderGL(PatchGraph *pg, PointCloudAnimationSequence *dataPoints, card32 globalTime) {
	int32 localTime = globalTime - pg->getStartFrameIndex();
	glPointSize(pointSize);
	if (!(useDList && dlist != -42) || needsUpdate) {
		needsUpdate = false;
		if (dlist == -42) {
			dlist = glGenLists(1);
		}
		if (useDList) glNewList(dlist, GL_COMPILE_AND_EXECUTE);
		if( pg ) {
		   card32 numFrames = pg->getTrajectorySet()->getNumFrames();
		   for ( card32 t = 0; t < numFrames; t++ ) {
				if ( !showAllFrames && t != localTime ) continue;
				if ( showAllFrames ) {
					if (correspondencesOffset > 1) {
						if (t != localTime && t!= localTime+correspondencesOffset) {
							continue;
						}
					}
					glMatrixMode(GL_MODELVIEW);
					glPushMatrix();
					float32 timeShift;
					if (globalTimeTranslation) {
						timeShift = (-(float32)localTime)+((float32)t);
					} else {
						timeShift = t;
					}
					if (correspondencesOffset > 1 && t!=localTime) timeShift = 1;
					Matrix4f M = makeTranslation4f(afInterFrameTransl*timeShift);
					glMultMatrixf(M.data());
				}
				// if (t < pg->getTrajectorySet()->getNumFrames()) {
				{
				
					PointSet *ps = pg->getTrajectorySet()->copySingleFrame(t);
					PointSet *psNext = NULL;
					if (showCorrespondences && t + correspondencesOffset < pg->getTrajectorySet()->getNumFrames()) {
						psNext = pg->getTrajectorySet()->copySingleFrame( t + correspondencesOffset );
					}
					InCorePCTopologyGraph *tg = pg->getTopology();
					if (ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
						AAT TS_POS = ps->getAAT("position");
						AAT TS_COL;
						bool providesColor = false;
						if (ps->providesAttribute("color", 3, VAD::DATA_FORMAT_FLOAT32)) {
							TS_COL = ps->getAAT("color");
							providesColor = true;
						}
						
						bool providesClusterID = false;
						AAT clusterAAT;
						if( ps->providesAttribute("clusterID", 1, VAD::DATA_FORMAT_INT32 ) )
						{
							providesClusterID = true;
							clusterAAT = ps->getAAT( "clusterID" );
						}
						card32 clusterColor[8];
						clusterColor[0] = 0xFF0000;
						clusterColor[1] = 0xFFFF00;
						clusterColor[2] = 0x00FFFF;
						clusterColor[3] = 0x0000FF;
						clusterColor[4] = 0xFF00FF;
						clusterColor[5] = 0xFFFFFF;
						clusterColor[6] = 0x000000;
						clusterColor[7] = 0x00FF00;
					//clusterColor[0] = 0x002200;
					//clusterColor[1] = 0x004400;
					//clusterColor[2] = 0x006600;
					//clusterColor[3] = 0x008800;
					//clusterColor[4] = 0x00AA00;
					//clusterColor[5] = 0x00CC00;
					//clusterColor[6] = 0x00FF00;
					//clusterColor[7] = 0x000000;
						
						// Show points	
						if ( showPoints ) {	
						   bool potAATsAvailable = ps->providesAttribute("sdfPot", 1, VAD::DATA_FORMAT_FLOAT32)
					                     	   && ps->providesAttribute("smoothPot", 1, VAD::DATA_FORMAT_FLOAT32)
               						         && ps->providesAttribute("rigidityPot", 1, VAD::DATA_FORMAT_FLOAT32);

                     AAT sdfAAT = NULL_AAT;
                     AAT rigidAAT = NULL_AAT;
                     AAT smoothAAT = NULL_AAT;

                     if (potAATsAvailable) {
                        sdfAAT = ps->getAAT("sdfPot");
                        rigidAAT = ps->getAAT("rigidityPot");
                        smoothAAT = ps->getAAT("smoothPot");
                     }                     
						
							glPushAttrib(GL_ALL_ATTRIB_BITS);
							glDisable(GL_LIGHTING);
							glPointSize(pointSize);
							glBegin(GL_POINTS);
							for( card32 i = 0; i < ps->getNumEntries(); i++ ){
								if (pointsColorMode == CM_PATCH_GRAPH_COLOR) {
									glColor3f( pg->getPointColor()[0], pg->getPointColor()[1], pg->getPointColor()[2] );
								} else if (pointsColorMode == CM_COLOR_CHANNEL_COLOR && providesColor) {
									glColor3fv( ps->get3f(i, TS_COL).data());
								} else if ( pointsColorMode == CM_CLUSTER_COLOR && providesClusterID )
								{
									glColor3ubv( (card8*)&clusterColor[ (ps->get1i(i, clusterAAT )) % 8 ] );
								}	else {
								   if (pointsColorMode == CM_POTENTIALS && potAATsAvailable && minPotValue < maxPotValue) {
								      float32 s = 1.0f/(maxPotValue-minPotValue);
								      float32 r = s*(ps->get1f(i, sdfAAT)-minPotValue) * potMask[0];
								      float32 g = s*(ps->get1f(i, rigidAAT)-minPotValue) * potMask[1];
								      float32 b = s*(ps->get1f(i, smoothAAT)-minPotValue) * potMask[2];
								      glColor3f(r,g,b);
								   } else {
   									glColor4fv(pointsColor.data());
								   }
								}
								glVertex3fv( ps->get3f( i, TS_POS ).data());
							}
							glEnd();
							glPopAttrib();
						}

						// Show correspondences
						if ( showCorrespondences && psNext ) {	
							glPushAttrib(GL_ALL_ATTRIB_BITS);
							glDisable(GL_LIGHTING);
							glEnable(GL_BLEND);
							glDepthMask(0);
							glBlendFunc(GL_SRC_ALPHA, GL_ONE);
							glColor4f(1, 1, 1, 0.05f);
							glBegin(GL_LINES);
							for( card32 i = 0; i < ps->getNumEntries(); i++ ){
								glVertex3fv( ps->get3f( i, TS_POS ).data());
								Vector3f nextPos = psNext->get3f( i, TS_POS ) + afInterFrameTransl;
								glVertex3fv( nextPos.data());
							}
							glEnd();
							glPopAttrib();
						}

						// Show topology					
						if( showTopology ) {
							glPushAttrib(GL_ALL_ATTRIB_BITS);
							glDisable(GL_LIGHTING);
							glEnable(GL_BLEND);
							glDepthMask(0);
							glBlendFunc(GL_SRC_ALPHA, GL_ONE);
							AAT TOP_COL;
							bool providesColor = false;
							if (tg->getEdges()->providesAttribute("color", 4, VAD::DATA_FORMAT_FLOAT32)) {
								TOP_COL = tg->getEdges()->getAAT("color");
								providesColor = true;
							}
							glBegin(GL_LINES);
							for( card32 v = 0; v < ps->getNumEntries(); v++ ) {
								InCorePCTopologyGraph::VertexIndex ne = tg->getNumAdjacentVertices(v);
								for (InCorePCTopologyGraph::VertexIndex i=0; i<ne; i++) {
									if (topologyColorMode == CM_PATCH_GRAPH_COLOR) {
										glColor4f( pg->getTopologyColor()[0], pg->getTopologyColor()[1], pg->getTopologyColor()[2], 0.1f );
									} else if (topologyColorMode == CM_COLOR_CHANNEL_COLOR && providesColor) {
										glColor4fv( tg->getEdges()->get<float32,4>(i, TOP_COL).data());
									//} else if (topologyColorMode == CM_FIXED_COLOR && providesColor) {
									//	glColor4f( 0.3f, 0.3f, 1.0f, 0.3f );
									} else {
										glColor4fv( topologyColor.data() );
									
									}
									int32 v2 = tg->getAdjacentVertex( v, i );
									Vector3f pos1 = ps->get3f( v, TS_POS );
									Vector3f pos2 = ps->get3f( v2, TS_POS );
									glVertex3fv( pos1.data() );
									glVertex3fv( pos2.data() );
								}
							}
							glEnd();
							glPopAttrib();
						}

						if ( showAllDataPoints ) {
							if ( dataPoints && ( dataPoints->getNumFrames() > ( t + pg->getStartFrameIndex() ) ) ) {
								PointSet* dataPS = dataPoints->getFrame( t + pg->getStartFrameIndex() );
								if( dataPS->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32) ) {
									AAT DATA_POS = dataPS->getAAT("position");
									AAT DP_COL;
									bool providesColor = false;
									if ( dataPS->providesAttribute("color", 3, VAD::DATA_FORMAT_FLOAT32) ) {
										DP_COL = dataPS->getAAT("color");
										providesColor = true;
									}
									glPushAttrib(GL_ALL_ATTRIB_BITS);
									glDisable(GL_LIGHTING);
									
									if (extraAlpha) {
							         glEnable(GL_BLEND);
							         glDepthMask(0);
							         glBlendFunc(GL_SRC_ALPHA, GL_ONE);
									}
									
									glPointSize(pointSize);
									glBegin(GL_POINTS);
									for( card32 i = 0; i < dataPS->getNumEntries(); i++ ){
									   Vector4f dpcol;
										if (pointsColorMode == CM_COLOR_CHANNEL_COLOR && providesColor) {
											dpcol = expand3To4(dataPS->get3f(i, DP_COL));
											dpcol[3] = 0.1f;											
										} else {
											dpcol = otherDataPointsColor;
										}
										glColor4fv(dpcol.data());
										glVertex3fv( dataPS->get3f( i, DATA_POS ).data() );
									}
									glEnd();
									glPopAttrib();
								}
							}
						}

						// Show associated data points
						if ( showCorrDataPoints ){
							if ( dataPoints && ( dataPoints->getNumFrames() > ( t + pg->getStartFrameIndex() ) ) ) {
								PointSet* dataPS = dataPoints->getFrame( t + pg->getStartFrameIndex() );
								if( dataPS->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32) ) {
									AAT DATA_POS = dataPS->getAAT("position");
									AAT DP_COL;
									bool providesColor = false;
									if (dataPS->providesAttribute("color", 3, VAD::DATA_FORMAT_FLOAT32)) {
										DP_COL = dataPS->getAAT("color");
										providesColor = true;
									}	
									// Draw points
									set<card32> corrDataPointIndices; corrDataPointIndices.clear();
									// Extract set of data points associated with current patch graph
									for ( unsigned int patchIndex = 0; patchIndex < pg->getTrajectorySet()->getNumPatches(); patchIndex++ ){
										vector<card32> indices = pg->getTrajectorySet()->getDataPoints()->getIndices( t, patchIndex );
										for ( unsigned int n = 0; n < indices.size(); n++ ){
											corrDataPointIndices.insert( indices[n] );
										}		
									}
									glPushAttrib(GL_ALL_ATTRIB_BITS);
									glDisable(GL_LIGHTING);
									glPointSize(pointSize);	
									glBegin(GL_POINTS);
									for( set<card32>::iterator it = corrDataPointIndices.begin(); it != corrDataPointIndices.end(); it++ ) {
										card32 index = card32(*it);
										if ( dataPointsColorMode == CM_COLOR_CHANNEL_COLOR && providesColor ) {
											glColor3fv( dataPS->get3f(index, DP_COL).data() );
										} else if ( ( dataPointsColorMode == CM_FIXED_COLOR ) && providesColor ) {
											glColor3fv( otherDataPointsColor.data() );
										}
										glVertex3fv( dataPS->get3f( index, DATA_POS ).data() );
									}
									glEnd();
									// Draw links
									glEnable( GL_BLEND );
									glDepthMask( 0 );
									glBlendFunc( GL_SRC_ALPHA, GL_ONE );
									glBegin( GL_LINES );
									glColor4f( 0.5f, 0.5f, 0.5f, 0.2f );
									PGDataPointIndices* dataPointIndices = pg->getTrajectorySet()->getDataPoints();
									for ( unsigned int patchIndex = 0; patchIndex < ps->getNumEntries(); patchIndex++ ){
										Vector3f patchPos = ps->get3f( patchIndex, TS_POS );
										for ( unsigned int dataPointNumber = 0; dataPointNumber < dataPointIndices->getNumDataPoints( t, patchIndex ); dataPointNumber++ ){
											card32 dataPointIndex = dataPointIndices->getIndex( t, patchIndex, dataPointNumber );
											Vector3f dataPos = dataPS->get3f( dataPointIndex, DATA_POS );
											glVertex3f( patchPos[0], patchPos[1], patchPos[2] );
											glVertex3f( dataPos[0], dataPos[1], dataPos[2] );
										}
									}
									glEnd();
									glDisable( GL_BLEND );
									glPopAttrib();
								}
							}
						}

						// Show coords
						if ( showCoords ){
							if( ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32 ) &&
								ps->getDescr()->providesAttribute(NORMAL_PATCH_VAD) &&
								ps->getDescr()->providesAttribute(TANGENT_U_PATCH_VAD) ) {
								const AAT POS = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
								const AAT NORMAL = ps->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
								const AAT TANGENT_U = ps->getAAT("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32);	
								glPushAttrib(GL_ALL_ATTRIB_BITS);
								if (extraAlpha) {
						         glEnable(GL_BLEND);
						         glDepthMask(0);
						         glBlendFunc(GL_SRC_ALPHA, GL_ONE);
								}
								glDisable(GL_LIGHTING);
								glMatrixMode(GL_MODELVIEW);
								const card32 n = ps->getNumEntries();
								for (card32 i=0; i<n; i++) {
									Vector3f pos = ps->get3f(i, POS);
									ReferenceFramePatch ref;
									ref.get(ps, i, NORMAL, TANGENT_U);
									Vector3f coords_v = ref.computeTangentV();
									glPushMatrix();
									glTranslatef(pos[0], pos[1], pos[2]);
									Matrix4f coordsSys = IDENTITY4F;
									coordsSys[0][0] = ref.tangent_u[0];
									coordsSys[0][1] = ref.tangent_u[1];
									coordsSys[0][2] = ref.tangent_u[2];
									coordsSys[1][0] = ref.normal[0];
									coordsSys[1][1] = ref.normal[1];
									coordsSys[1][2] = ref.normal[2];
									coordsSys[2][0] = coords_v[0];
									coordsSys[2][1] = coords_v[1];
									coordsSys[2][2] = coords_v[2];
									//coordsSys = coordsSys.transpose();
									glMultMatrixf(coordsSys.data());
									glScalef(coordsSize, coordsSize, coordsSize);
									glBegin(GL_LINES);
									switch (coordsDrawStyle) {
										case COORDS_SIMPLE: {
											GLHelper::drawCoordsSendVerticesNoXYZNoArrows();
											break;
										}
										case COORDS_ARROWS: {
											GLHelper::drawCoordsSendVerticesNoXYZ();
											break;
										}
										default:
										case COORDS_XYZ: {
											GLHelper::drawCoordsSendVertices();
											break;
										}
									}
									glEnd();
									glPopMatrix();
								}
								glPopAttrib();	
							}
						}

						// ---- showLocalPatches, showInterpolatedPatches
						if (showLocalPatches || showInterpolatedPatches) {
							if (ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
								if (ps->providesAttribute(HEIGHTFIELD_COEFFS_NAME)
									 && ps->getDescr()->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)
									 && ps->getDescr()->providesAttribute("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32) 
									 && ps->getDescr()->getAttribute(HEIGHTFIELD_COEFFS_NAME).dataFormat == VAD::DATA_FORMAT_FLOAT32
									 && ps->getDescr()->providesAttribute("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32)) {
									const card32 numComponents = ps->getDescr()->getAttribute(HEIGHTFIELD_COEFFS_NAME).numComponents;
									
									const AAT POS = ps->getAAT("position");
									const AAT NORMAL = ps->getAAT("normal");
									const AAT TANGENT_U = ps->getAAT("tangent_u");
									const AAT PATCH_HF = ps->getAAT(HEIGHTFIELD_COEFFS_NAME);
									const AAT PATCH_SAMPLE_SPACING = ps->getAAT("sampleSpacing");
									
									glPushAttrib(GL_ALL_ATTRIB_BITS);
									
									if (patchLighting) {
										glEnable(GL_LIGHTING);
										glEnable(GL_NORMALIZE);
										glEnable(GL_LIGHT0);
										glEnable(GL_DEPTH_TEST);
										glDisable(GL_CULL_FACE);
										glDepthMask(1);
										glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 10.0f);
										glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, makeVector4f(0.1f, 0.1f, 0.1f, 1.0f).data());
										if (showInterpolatedPatches) {
											glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, makeVector4f(0.8f, 0.3f, 0.6f, 1.0f).data()); 
										} else {
											glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, makeVector4f(0.6f, 0.3f, 0.8f, 1.0f).data()); 
										}
										glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, makeVector4f(0.5f, 0.5f, 0.5f, 1.0f).data());
										glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
									} else {
										glDisable(GL_LIGHTING);
										if (!showInterpolatedPatches) {
											glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
										}
										glColor3f(0.6f, 0.3f, 0.8f);
									}
									
									HFBFBivariateMonomials bf;
									bf.setup(bf.getDegreeFromNumBF(numComponents));
									const card32 numBF = bf.getNumBasisFunctions();
									pAssert(numBF <= numComponents);						
									
									CoefficientSet coeff(numComponents);
									CoefficientSet neighborCoeff(numComponents);
									
									vector<Vector3f> patch((patchSubdiv+1)*(patchSubdiv+1));
									vector<Vector3f> normal(patchSubdiv*patchSubdiv);
									
									const card32 n = ps->getNumEntries();
									for (card32 i=0; i<n; i++) {
										Vector3f pos = ps->get3f(i, POS);
										ReferenceFramePatch refFrame;
										refFrame.get(ps, i, NORMAL, TANGENT_U);
										Vector3f coords_v = refFrame.computeTangentV();
										coeff.get(ps, i, PATCH_HF);

										for (card32 v=0; v<=patchSubdiv; v++) {
											for (card32 u=0; u<=patchSubdiv; u++) {
												float32 effectivePatchSize = patchSize;
												if (showInterpolatedPatches) {
													effectivePatchSize = ps->get<float32>(i, PATCH_SAMPLE_SPACING);
												}
												if (!showInterpolatedPatches) {
													Vector3f p;
													p[0] = ((float32)u/(float32)(patchSubdiv-1) - 0.5f) * effectivePatchSize;
													p[1] = ((float32)v/(float32)(patchSubdiv-1) - 0.5f) * effectivePatchSize;
													p[2] = 0.0f;
													for (card32 bfn=0; bfn<numBF; bfn++) {
														p[2] += coeff[bfn]*bf.evaluate(bfn, makeVector2f(p[0], p[1]));
													}
													patch[u+(patchSubdiv+1)*v] = refFrame.tangent_u*p[0] + coords_v*p[1] + refFrame.normal*p[2] + pos;
												} else {
													card32 numNeighbors = tg->getNumAdjacentVertices(i);
													patch[u+(patchSubdiv+1)*v] = NULL_VECTOR3F;
													float32 weightSum = 0;
													Matrix3f refFrameMatrix;
													refFrameMatrix[0] = refFrame.tangent_u;
													refFrameMatrix[1] = coords_v;
													refFrameMatrix[2] = refFrame.normal;
													
													for (card32 j=0; j<=numNeighbors; j++) {
														card32 neighborIndex;
														if (j==numNeighbors) {
															neighborIndex = i;
														} else {
															neighborIndex = tg->getAdjacentVertex(i, j);
														}
														Vector3f neighborPos = ps->get3f(neighborIndex, POS);
														ReferenceFramePatch neighborRefFrame;
														neighborRefFrame.get(ps, neighborIndex, NORMAL, TANGENT_U);
														Vector3f neighborCoords_v = neighborRefFrame.computeTangentV();
														neighborCoeff.get(ps, neighborIndex, PATCH_HF);
														Vector3f p;
														p[0] = ((float32)u/(float32)(patchSubdiv-1) - 0.5f) * effectivePatchSize;
														p[1] = ((float32)v/(float32)(patchSubdiv-1) - 0.5f) * effectivePatchSize;
														p[2] = 0.0f;
														
														Matrix3f neighborRefFrameMatrix;
														neighborRefFrameMatrix[0] = neighborRefFrame.tangent_u;
														neighborRefFrameMatrix[1] = neighborCoords_v;
														neighborRefFrameMatrix[2] = neighborRefFrame.normal;
														Matrix3f neighborRefFrameMatrixInv = neighborRefFrameMatrix.transpose();
														
														Vector3f worldSpaceCoords = refFrameMatrix * p + pos;
														float32 weight = calcWeight(worldSpaceCoords, pos, ps->get<float32>(neighborIndex, PATCH_SAMPLE_SPACING)*kernelWidthFactor);
														
														Vector3f pn = neighborRefFrameMatrixInv * (worldSpaceCoords - neighborPos);
														pn[2] = 0;
														
														for (card32 bfn=0; bfn<numBF; bfn++) {
															pn[2] += neighborCoeff[bfn]*bf.evaluate(bfn, makeVector2f(pn[0], pn[1]));
														}
														Vector3f basePos = neighborRefFrame.tangent_u*pn[0] + neighborCoords_v*pn[1];
														Vector3f heightPos = neighborRefFrame.normal*pn[2];
														patch[u+(patchSubdiv+1)*v] += (basePos + heightPos + neighborPos) * weight;

														weightSum += weight;
													}
													patch[u+(patchSubdiv+1)*v] /= weightSum;
												}
											}
										}
										
										for (card32 v=0; v<patchSubdiv; v++) {
											for (card32 u=0; u<patchSubdiv; u++) {
												normal[u+patchSubdiv*v] = (patch[u+1+(patchSubdiv+1)*v] - patch[u+(patchSubdiv+1)*v]).crossProduct(
																						(patch[u+(patchSubdiv+1)*v] - patch[u+(patchSubdiv+1)*(v+1)]));
											}
										}
										
										if (!showInterpolatedPatches) {
											for (card32 v=0; v<patchSubdiv-1; v++) {
												glBegin(GL_QUAD_STRIP);
												for (card32 u=0; u<patchSubdiv-1; u++) {
													glNormal3fv(normal[u+v*patchSubdiv].data());
													glVertex3fv(patch[u+v*(patchSubdiv+1)].data());
													glNormal3fv(normal[u+(v+1)*patchSubdiv].data());
													glVertex3fv(patch[u+(v+1)*(patchSubdiv+1)].data());
													glNormal3fv(normal[u+1+v*patchSubdiv].data());
													glVertex3fv(patch[u+1+v*(patchSubdiv+1)].data());
													glNormal3fv(normal[u+1+(v+1)*patchSubdiv].data());
													glVertex3fv(patch[u+1+(v+1)*(patchSubdiv+1)].data());
												}
												glEnd();
											}
										} else {
											glBegin(GL_POINTS);
											for (card32 v=0; v<patchSubdiv-1; v++) {
												for (card32 u=0; u<patchSubdiv-1; u++) {
													if (patchLighting) {
														glNormal3fv(normal[u+v*patchSubdiv].data());
													} else {
														glColor3fv(normalize(normal[u+v*patchSubdiv]).data());
													}
													glVertex3fv(patch[u+v*(patchSubdiv+1)].data());
												}
											}
											glEnd();
										}
									}
									glPopAttrib();
								}	
							}					
						}
						
						// ---- showSampleSpacing
						if (showSampleSpacing) {
						
							if (  ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)
							&& ps->getDescr()->providesAttribute("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32)
								 && ps->getDescr()->providesAttribute(NORMAL_PATCH_VAD)
								 && ps->getDescr()->providesAttribute(TANGENT_U_PATCH_VAD)) {
								const AAT POS = ps->getAAT("position");
								const AAT SPAC = ps->getAAT("sampleSpacing");
								const AAT NORMAL = ps->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
								const AAT TANGENT_U = ps->getAAT("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32);
								
								glPushAttrib(GL_ALL_ATTRIB_BITS);
								
								glDisable(GL_LIGHTING);
								
								const card32 n = ps->getNumEntries();
								for (card32 i=0; i<n; i++) {
									float32 blue = ((101*i) %13)/13.0f;
   								glColor3f(1.0f, 1.0f-blue, blue);
									float32 spac = ps->get<float32>(i, SPAC);
									Vector3f u = ps->get3f(i, TANGENT_U);
									Vector3f n = ps->get3f(i, NORMAL);
									Vector3f v = u.crossProduct(n);
   								glBegin(GL_LINE_LOOP);
   								for (card32 k=0; k<24; k++) {
   									Vector3f p = ps->get3f(i, POS);
   									p += u*cos(k/12.0f*M_PI)*spac;
   									p += v*sin(k/12.0f*M_PI)*spac;
   									glVertex3fv(p.data());
   								}
   								glEnd();
								}
								
								glPopAttrib();
							}	
						}

						// - old version -
						// // Show correspondences	
						//if ( showCorrespondences ) {	
						//	if (ps->getDescr()->providesAttribute("corrCoeffs", 6, VAD::DATA_FORMAT_FLOAT32)) {
						//		AAT CCOEFF = ps->getAAT("corrCoeffs");
						//		glPushAttrib(GL_ALL_ATTRIB_BITS);
						//		glDisable(GL_LIGHTING);
						//		glColor3f( 0.4f, 0.6f, 0.8f );
						//		glPointSize(pointSize+2);
						//		glBegin(GL_POINTS);
						//		for( card32 i = 0; i < ps->getNumEntries(); i++ ){
						//			StaticVector<float32, 6> w = ps->get<float32, 6>( i, CCOEFF );
						//			Vector3f pos = ps->get3f( i, TS_POS );
						//			Vector3f v = pos+makeVector3f(w[0], w[1], w[2]);
						//			glVertex3fv( v.data());
						//		}
						//		glEnd();
						//		glPopAttrib();
						//	}
						//}
						delete ps;			
						delete psNext;
					}
				}
            if (showAllFrames) {
               glMatrixMode(GL_MODELVIEW);
               glPopMatrix();
            }
			}

			PointSet *urshape = pg->getUrshape();
			InCorePCTopologyGraph *tg = pg->getTopology();
			if (urshape && showUrshape) {
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
				float32 timeShift;
				if (globalTimeTranslation) {
					timeShift = -(float32)localTime;
				} else {
					timeShift = 0;
				}
            Matrix4f M = makeTranslation4f(urshapeTransl+afInterFrameTransl*timeShift);
            glMultMatrixf(M.data());
			   if (urshape->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
				   AAT POS = urshape->getAAT("position");
					if ( showPoints ) {	
				      glPushAttrib(GL_ALL_ATTRIB_BITS);
				      glDisable(GL_LIGHTING);
				      glColor3fv( urshapePointColor.data() );
				      glPointSize(pointSize);
				      glBegin(GL_POINTS);
				      for( card32 i = 0; i < urshape->getNumEntries(); i++ ){
					      glVertex3fv( urshape->get3f( i, POS ).data());
				      }
				      glEnd();
   				   glPopAttrib();
				   }

					if( showTopology && tg ) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glDisable(GL_LIGHTING);
						glEnable(GL_BLEND);
						glDepthMask(0);
						glBlendFunc(GL_SRC_ALPHA, GL_ONE);
						glColor4fv( urshapeTopologyColor.data() );
						glBegin(GL_LINES);
						for( card32 v = 0; v < urshape->getNumEntries(); v++ ) {
							InCorePCTopologyGraph::VertexIndex ne = tg->getNumAdjacentVertices(v);
							for (InCorePCTopologyGraph::VertexIndex i=0; i<ne; i++) {
								int32 v2 = tg->getAdjacentVertex( v, i );
								Vector3f pos1 = urshape->get3f( v, POS );
								Vector3f pos2 = urshape->get3f( v2, POS );
								glVertex3fv( pos1.data() );
								glVertex3fv( pos2.data() );
							}
						}
						glEnd();
						glPopAttrib();
					}
		      }
            glMatrixMode(GL_MODELVIEW);
            glPopMatrix();
		   }

		}
		if (useDList) glEndList();
	} else {
		glCallList(dlist);
	}
}

void StandardPatchGraphRenderer::rebuildDList() {
	needsUpdate = true;
}

void StandardPatchGraphRenderer::invalidate() {
	rebuildDList();
}

StandardPatchGraphRenderer::~StandardPatchGraphRenderer() {}

