//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "HoleFilling2.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "PatchGraphOpt2.h"
#include "AnimationBSRContext.h"
#include "PatchTypes.h"
#include "InCorePCTopologyGraph.h"
#include "PGOpt2Config.h"
#include "PGOpt2GraphCreator.h"
#include "PGOpt2FramePairVars.h"
#include "PGOpt2FramePair.h"
#include "PGOpt2Frame.h"
#include "OptimizationGraph.h"
#include "PatchTypes.h"
#include "MathTools.h"
#include "PGOpt2Debugging.h"
#include "PGOpt2FramePair.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( HoleFilling2 ,0) {
	BEGIN_CLASS_INIT( HoleFilling2 );
}

// ----

HoleFilling2::HoleFilling2() {
}

void HoleFilling2::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

HoleFilling2::~HoleFilling2() {
}




void HoleFilling2::fillHoles( 
	AnimationBSRContext *context,
	PatchGraph* pg, 
   PointCloudAnimationSequence *data,
	DebugRenderer *debugRenderer,
	OptimizationRecord2 *optRecord)
{
	PGOpt2Debugging::outputValidityCheck(pg);
   vector<OptPatch> uninitNodes;
   do {
      uninitNodes.clear();
      computeNextOptPatches(context, pg, uninitNodes);
      if (uninitNodes.empty()) break;
      predict(context, pg, uninitNodes);
      partialOptimization(context, pg, data, uninitNodes, debugRenderer, optRecord);
      markAsInitialized(pg, uninitNodes);
   } while (!uninitNodes.empty());
   checkInitEverything(pg);   
}

void HoleFilling2::computeNextOptPatches(AnimationBSRContext *context, const PatchGraph *pg, vector<OptPatch> &result) 
{
   const PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   const InCorePCTopologyGraph *tg = pg->getTopology();
   const card32 width = ps->getWidth();
   const card32 height = ps->getHeight();
   const AAT flagsAAT = ps->getAAT("flags");
   for (card32 patch=0; patch<height; patch++) {
      for (card32 frame=0; frame<width; frame++) {
         int32 flags = ps->get<int>(frame, patch, flagsAAT);
         if ((flags & FLAG_PATCH_INITIALIZED) == 0) {
            bool pickit = false;
            if (frame>0) {
               if (ps->get<int>(frame-1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED) {
                  pickit = true;
               }
            }
            if (frame+1<width) {
               if (ps->get<int>(frame+1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED) {
                  pickit = true;
               }
            }
            if (pickit) {
               OptPatch op;
               op.frame = frame;
               op.patch = patch;
               result.push_back(op);
            }
         }
      }
   }
}

void HoleFilling2::predict(AnimationBSRContext *context, PatchGraph *pg, const vector<OptPatch> &indices) 
{
   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   const VertexDescriptor *descr = ps->getDescr();
   const InCorePCTopologyGraph *tg = pg->getTopology();
   const AAT flagsAAT = ps->getAAT("flags");
   const AAT positionAAT = ps->getAAT("position");
   const AAT normalAAT = ps->getAAT("normal"); 
   const AAT tangentUAAT = ps->getAAT("tangent_u");
   const AAT topologyDepthAAT = ps->getAAT("topology_depth");
   const size_t numIndices = indices.size();
   const card32 numFrames = ps->getWidth();
   if (numIndices == 0) return;
   for (card32 i=0; i<numIndices; i++) {
      const card32 frame = indices[i].frame;
      const card32 patch = indices[i].patch;

		card32 predFrame;
		bool found = false;
		if (frame > 0) {
			if (ps->get<int32>(frame-1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED) {
				predFrame = frame-1; 
				found = true;
			}
		}
		if (!found && frame < numFrames-1) {
			if (ps->get<int32>(frame+1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED) {
				predFrame = frame+1; 
				found = true;
			}
		}
		
		if (!found) {
			warning("HoleFilling2::predict() - did not find patch in adjacent frame.");
			continue;
		}

      Vector3f predPos = ps->get3f(predFrame, patch, positionAAT);
      Vector3f predNormal = ps->get3f(predFrame, patch, normalAAT);
      Vector3f predTangentU = ps->get3f(predFrame, patch, tangentUAAT);

      descr->zero(ps->getDataPointer(frame, patch));
      ps->set3f(frame, patch,    positionAAT,      predPos);
      ps->set3f(frame, patch,    normalAAT,        predNormal);
      ps->set3f(frame, patch,    tangentUAAT,      predTangentU);
      ps->set<int>(frame, patch, topologyDepthAAT, 1);
      ps->set<int>(frame, patch, flagsAAT,         FLAG_PATCH_INITIALIZED);
   }
   pg->getTrajectorySet()->invalidateAllAttachments();
}

void HoleFilling2::markAsInitialized(PatchGraph *pg, const vector<OptPatch> &indices) 
{
   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   const AAT flagsAAT = ps->getAAT("flags");
   const AAT topologyDepthAAT = ps->getAAT("topology_depth");
   const size_t numIndices = indices.size();
   if (numIndices == 0) return;
   for (card32 i=0; i<numIndices; i++) {
      const card32 frame = indices[i].frame;
      const card32 patch = indices[i].patch;
      ps->set<int>(frame, patch, flagsAAT, ps->get<int>(frame, patch, flagsAAT) | FLAG_PATCH_INITIALIZED);
      ps->set<int>(frame, patch, topologyDepthAAT, 0);
   }
}

void HoleFilling2::checkInitEverything(PatchGraph *pg) 
{
   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   const AAT flagsAAT = ps->getAAT("flags");
   const AAT topologyDepthAAT = ps->getAAT("topology_depth");
   const card32 numEntries = ps->getNumEntries();
   for (card32 i=0; i<numEntries; i++) {
      pAssert(ps->get<int>(i, flagsAAT) & FLAG_PATCH_INITIALIZED);
      ps->set<int>(i, topologyDepthAAT, 0);
   }
}
   


void HoleFilling2::partialOptimization(
   AnimationBSRContext *context, 
   PatchGraph *pg, 
   PointCloudAnimationSequence *data,
   const vector<OptPatch> &indices, 
   DebugRenderer *debugRenderer, 
   OptimizationRecord2 *optRecord) 
{
   PatchGraphOpt2 opt;
   context->getHoleFillingConfig()->potentials.configure(&opt);

	const card32 numFrames = pg->getTrajectorySet()->getNumFrames();
   const card32 numPatches = pg->getTrajectorySet()->getNumPatches();
   const PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   AAT flagsAAT = ps->getAAT("flags");
   
   vector< vector<card32> > frameVarIndices(numFrames);
   vector< vector<bool> > edgeFlags(numFrames);
   for (card32 i=0; i<numFrames; i++) {
		edgeFlags[i].resize(numPatches);
   }
   const size_t numIndices = indices.size();
   for (card32 i=0; i<numIndices; i++) {
		const card32 frame = indices[i].frame;
		const card32 patch = indices[i].patch;
		frameVarIndices[frame].push_back(patch);
      bool linkToLeft = true;
      if (frame+1<numFrames) {
         linkToLeft = (ps->get<int>(frame+1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED);
      }
		edgeFlags[frame][patch] = linkToLeft;
		if (frame > 0) {
			edgeFlags[frame-1][patch] = (ps->get<int>(frame-1, patch, flagsAAT) & FLAG_PATCH_INITIALIZED);
		}
   }

   vector< vector<card32> > edgeVarIndices(numFrames);
   for (card32 f=0; f<numFrames; f++) {
		for (card32 p=0; p<numPatches; p++) {
			if (edgeFlags[f][p]) {
				edgeVarIndices[f].push_back(p);
			}
		}
   }

   PGOpt2GraphCreator gc;
   PGOpt2FramePairVars *pairVars = gc.createPairVars(pg, false, false, false, false, true);
   OptimizationGraph *optGraph = gc.createOptGraph(
      pg, data, pairVars, false, false, false, false, true, 
      &frameVarIndices, &edgeVarIndices, false, false, 
      context->getTopology2MinimumVertexDegree()
   );
	opt.optimize(context->getHoleFillingConfig()->params, context, optGraph, pg, data, debugRenderer, optRecord);
   
/*
   PGOpt2GraphCreator gc;
   PGOpt2FramePairVars *pairVars = gc.createPairVars(pg, false, false, false, false, false);
   OptimizationGraph *optGraph = gc.createOptGraph(pg, data, pairVars, false, false, false, false, false, &frameVarIndices, &edgeVarIndices);
   PGOpt2Parameters paramsCopy = *(context->getHoleFillingConfig()->params);
   card32 maxIt = paramsCopy.numIterations;
   paramsCopy.numIterations = 1;
   card32 numIt = 0;
   bool converged = false;

   while (numIt < maxIt && !converged) {
   
      opt.optimize(&paramsCopy, optGraph, pg, data, debugRenderer, optRecord);

      card32 numEdges = optGraph->getNumEdges();
      for (card32 i=0; i<numEdges; i++) {
         optGraph->getEdge(i)->swapFrames();
      }

      opt.optimize(&paramsCopy, optGraph, pg, data, debugRenderer, optRecord);
      converged = optRecord->converged;
      
      for (card32 i=0; i<numEdges; i++) {
         optGraph->getEdge(i)->swapFrames();
      }
      numIt++;
   }  */
   
   delete pairVars;
   delete optGraph;
}









//void HoleFilling2::computeNextOptPatches(AnimationBSRContext *context, const PatchGraph *pg, vector<OptPatch> &result) 
//{
//   const PointSet *ps = pg->getTrajectorySet()->getTrajectories();
//   const InCorePCTopologyGraph *tg = pg->getTopology();
//   const card32 width = ps->getWidth();
//   const card32 height = ps->getHeight();
//   const AAT flagsAAT = ps->getAAT("flags");
//   for (card32 patch=0; patch<height; patch++) {
//      for (card32 frame=0; frame<width; frame++) {
//         int32 flags = ps->get<int>(frame, patch, flagsAAT);
//         if ((flags & FLAG_PATCH_INITIALIZED) == 0) {
//            card32 numNeighbors = tg->getNumAdjacentVertices(patch);
//            bool pickit = false;
//            for (card32 i=0; i<numNeighbors; i++) {
//               card32 nIndex = tg->getAdjacentVertex(patch, i);
//               int32 nFlags = ps->get<int>(frame, nIndex, flagsAAT);
//               if (nFlags & FLAG_PATCH_INITIALIZED) {
//						pickit = true;
//						break;
//               }
//            }
//            if (pickit) {
//               OptPatch op;
//               op.frame = frame;
//               op.patch = patch;
//               result.push_back(op);
//            }
//         }
//      }
//   }
//}
//
//void HoleFilling2::predict(AnimationBSRContext *context, PatchGraph *pg, const vector<OptPatch> &indices) 
//{
//   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
//   const VertexDescriptor *descr = ps->getDescr();
//   const InCorePCTopologyGraph *tg = pg->getTopology();
//   const AAT flagsAAT = ps->getAAT("flags");
//   const AAT positionAAT = ps->getAAT("position");
//   const AAT normalAAT = ps->getAAT("normal"); 
//   const AAT tangentUAAT = ps->getAAT("tangent_u");
//   const AAT topologyDepthAAT = ps->getAAT("topology_depth");
//   const card32 numIndices = indices.size();
//   if (numIndices == 0) return;
//   for (card32 i=0; i<numIndices; i++) {
//      const card32 frame = indices[i].frame;
//      const card32 patch = indices[i].patch;
//      card32 numNeighbors = tg->getNumAdjacentVertices(patch);
//      Vector3f predPos = NULL_VECTOR3F;
//      Vector3f predNormal = NULL_VECTOR3F;
//      Vector3f anyNormal = NULL_VECTOR3F;
//      Vector3f predTangentU = NULL_VECTOR3F;
//      int32 numPred = 0;
//      for (card32 i=0; i<numNeighbors; i++) {
//         card32 nIndex = tg->getAdjacentVertex(patch, i);
//         int32 nFlags = ps->get<int>(frame, nIndex, flagsAAT);
//         if (nFlags & FLAG_PATCH_INITIALIZED) {
//            predPos      += ps->get<float32, 3>(frame, nIndex, positionAAT);
//            anyNormal     = ps->get<float32, 3>(frame, nIndex, normalAAT);
//            predNormal   += anyNormal;
//            predTangentU += ps->get<float32, 3>(frame, nIndex, tangentUAAT);
//            numPred++;
//         }
//      }
//      if (numPred == 0) {
//         pAssert(false);
//         continue;
//      }
//
//      predPos /= numPred;
//      if (norm(predNormal) < 0.01) {
//         predNormal = anyNormal;
//      } else {
//         predNormal = normalize(predNormal);
//      }
//      predTangentU /= numPred;
//      if (norm(predTangentU) < 0.01) {
//         predTangentU = makeOrthonormalVector(predNormal);
//      } else {
//         predTangentU = normalize(orthogonalize(predNormal, predTangentU));
//      }
//      
//      descr->zero(ps->getDataPointer(frame, patch));
//      ps->set3f(frame, patch,    positionAAT,      predPos);
//      ps->set3f(frame, patch,    normalAAT,        predNormal);
//      ps->set3f(frame, patch,    tangentUAAT,      predTangentU);
//      ps->set<int>(frame, patch, topologyDepthAAT, 1);
//      ps->set<int>(frame, patch, flagsAAT,         FLAG_PATCH_INITIALIZED);
//   }
//   pg->getTrajectorySet()->invalidateAllAttachments();
//}
