//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PGOpt2RemoveLowDegreeVertices.h"
//---------------------------------------------------------------------------
#include "PatchGraph.h"
#include "InCorePCTopologyGraph.h"
#include "TrajectorySet.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "AnimationBSRContext.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


card32 PGOpt2RemoveLowDegreeVertices::processStep(const AnimationBSRContext *context, PatchGraph *pg) 
{
	const card32 minDegree = context->getTopology2MinimumVertexDegree();
   if (minDegree == 0) return 0;
   InCorePCTopologyGraph *tg = pg->getTopology();
   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   PointSet *urshape = pg->getUrshape();
   const card32 numVertices = ps->getHeight();//tg->getNumVertices();
   vector<bool> keep(numVertices);
   card32 numRemove = 0;
   for (card32 i=0; i<numVertices; i++) {
      const card32 degree = tg->getNumAdjacentVertices(i);
      bool keepThis = (degree>=minDegree);
      keep[i] = keepThis;
      if (!keepThis) numRemove++;
   }
   if (numRemove == 0) return 0;
   card32 writePos = 0;
   vector<int32> remap(numVertices);
   //pAssert(numVertices == ps->getHeight());
   const card32 numFrames = ps->getWidth();
   InCorePCTopologyGraph *newTG = new InCorePCTopologyGraph();
   newTG->clearAndSetup(tg->getEdges()->getDescr());
   // Remove vertices with too small valence 
   for (card32 p=0; p<numVertices; p++) {
      if (keep[p]) {
         for (card32 f=0; f<numFrames; f++) {
            ps->setVertex(f, writePos, ps->getDataPointer(f, p));
         }
         if (urshape) {
            urshape->setVertex(writePos, urshape->getDataPointer( p));
         }
         remap[p] = writePos;
         writePos++;
      } else {
         remap[p] = -1;
      }
   }            
   // Create new topology graph
   for (card32 p=0; p<numVertices; p++) {
      const card32 numAdjacent = tg->getNumAdjacentVertices(p);
      if (remap[p] != -1) {
         pAssert(remap[p] < (int)writePos);
         for (card32 ni=0; ni<numAdjacent; ni++) {
            const card32 otherP = tg->getAdjacentVertex(p, ni);
            if (remap[otherP] != -1) {
               pAssert(remap[otherP] < (int)writePos);
               newTG->addEdge(remap[p], remap[otherP]);
            }
         }
      }
   }
   ps->changeHeight(writePos);
   if (urshape) urshape->changeHeight(writePos);
   pg->setTopology(newTG);

   if (context->getVerbose()) debugOutput << "PGOpt2RemoveLowDegreeVertices: removed " << numRemove << " vertices.\n"; debugOutput.flush();
   
   return numRemove;
}


void PGOpt2RemoveLowDegreeVertices::process(const AnimationBSRContext *context, PatchGraph *pg) {
   int n = 0;
   do {
      n = processStep(context, pg);
   } while (n>0);
}

void PGOpt2RemoveLowDegreeVertices::removeUnsupported(const AnimationBSRContext *context, const card32 numDataFrames, PatchGraph *pg, PointCloudAnimationSequence *data) {
   InCorePCTopologyGraph *tg = pg->getTopology();
   PointSet *ps = pg->getTrajectorySet()->getTrajectories();
   PointSet *urshape = pg->getUrshape();
   const card32 numVertices = ps->getHeight();//tg->getNumVertices();
   vector<bool> keep(numVertices);
   card32 numRemove = 0;

   vector<PointSetKNNQuery*> queries(ps->getWidth());
   AAT DATA_POS = data->getFrame(0)->getAAT("position");
   for (card32 i=0; i<ps->getWidth(); i++) {
      const card32 frame = i + pg->getStartFrameIndex();
      PointSet *ps = data->getFrame(frame);
      PointSetKNNQuery *q = new PointSetKNNQuery(ps, data->fetchOctree(frame), NULL_VECTOR3F, DATA_POS);
      queries[i] = q;
   }

   AAT PS_POS = ps->getAAT("position");

   for (card32 i=0; i<numVertices; i++) {
      card32 count = 0;
      for (card32 frame=0; frame<ps->getWidth(); frame++) {
         const Vector3f pos = ps->get3f(frame, i, PS_POS);
         queries[frame]->setTargetPoint(pos);
         card32 index;
         bool found;
         queries[frame]->getNextNearestPointIndex(index, found);
         if (found) {
            const Vector3f npos = data->getFrame(frame)->get3f(index, DATA_POS);
            if (norm(npos-pos) <= context->getSamplingSpacing() * context->getTopology2MaxDistanceRelSampleSpacing()) {
               count++;
            }
         }
      }      

      bool keepThis = (count>=numDataFrames);
      keep[i] = keepThis;
      if (!keepThis) numRemove++;
   }

   for (card32 i=0; i<pg->getTrajectorySet()->getNumFrames(); i++) {
      delete queries[i];
   }
   queries.clear();

   if (numRemove == 0) return;
   card32 writePos = 0;
   vector<int32> remap(numVertices);
   //pAssert(numVertices == ps->getHeight());
   const card32 numFrames = ps->getWidth();
   InCorePCTopologyGraph *newTG = new InCorePCTopologyGraph();
   newTG->clearAndSetup(tg->getEdges()->getDescr());
   // Remove vertices with too small valence 
   for (card32 p=0; p<numVertices; p++) {
      if (keep[p]) {
         for (card32 f=0; f<numFrames; f++) {
            ps->setVertex(f, writePos, ps->getDataPointer(f, p));
         }
         if (urshape) {
            urshape->setVertex( writePos, urshape->getDataPointer( p));
         }
         remap[p] = writePos;
         writePos++;
      } else {
         remap[p] = -1;
      }
   }            
   // Create new topology graph
   for (card32 p=0; p<numVertices; p++) {
      const card32 numAdjacent = tg->getNumAdjacentVertices(p);
      if (remap[p] != -1) {
         pAssert(remap[p] < (int)writePos);
         for (card32 ni=0; ni<numAdjacent; ni++) {
            const card32 otherP = tg->getAdjacentVertex(p, ni);
            if (remap[otherP] != -1) {
               pAssert(remap[otherP] < (int)writePos);
               newTG->addEdge(remap[p], remap[otherP]);
            }
         }
      }
   }
   ps->changeHeight(writePos);
   if (urshape) urshape->changeHeight(writePos);
   pg->setTopology(newTG);

   if (context->getVerbose()) debugOutput << "PGOpt2RemoveLowDegreeVertices::removeUnsupported() -- removed " << numRemove << " vertices.\n"; debugOutput.flush();
}
