//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraphMerge2.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "Timer.h"
#include "PatchGraphJoin2.h"
#include "PairwiseAlignment2.h"
#include "TopologyMerging2.h"
#include "HoleFilling2.h"
#include "PatchGraphResampler2.h"
#include "GlobalOptimization2.h"
#include "AnimationBSRContext.h"
#include "FeatureCorrespondences.h"
#include "PointSetKNNQuery.h"
#include "PatchGraphOpt2.h"
#include "PGOpt2Debugging.h"
#include "RecomputeTopology2.h"
#include "PGOpt2RemoveLowDegreeVertices.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


static string descrStopEarly(card32 i) {
   if (i==0) return "perform all steps";
   if (i==1) return "stop after merge aligned";
   if (i==2) return "stop after merge reversely aligned";
   if (i==3) return "stop after fill holes";
   if (i==4) return "stop after resample";
   return "unknown";
}


IMPLEMENT_GEOXL_CLASS( PatchGraphMerge2, 1 ) {
	BEGIN_CLASS_INIT( PatchGraphMerge2 );
   ADD_OBJECT_PROP(joinParts, 0, PatchGraphJoin2::getClass(), true)
   ADD_OBJECT_PROP(alignParts, 0, PairwiseAlignment2::getClass(), true)
   ADD_OBJECT_PROP(mergeAlignedRegion, 0, TopologyMerging2::getClass(), true)
   ADD_OBJECT_PROP(fillHoles, 0, HoleFilling2::getClass(), true)
   ADD_OBJECT_PROP(resampler, 0, PatchGraphResampler2::getClass(), true)
   ADD_OBJECT_PROP(globalOpt, 0, GlobalOptimization2::getClass(), true)
   ADD_CARD32_PROP_UPDATE_DESCR(debugStopEarly, 1, NULL, descrStopEarly, 5)
}

// ----

PatchGraphMerge2::PatchGraphMerge2() {
   joinParts = new PatchGraphJoin2();
   alignParts = new PairwiseAlignment2();
   mergeAlignedRegion = new TopologyMerging2();
   fillHoles = new HoleFilling2();
   resampler = new PatchGraphResampler2();
   globalOpt = new GlobalOptimization2();
   debugStopEarly = 0;
}

void PatchGraphMerge2::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

PatchGraphMerge2::~PatchGraphMerge2() {}



/// helper function to extract feature correspondences
static void buildFeatureCorrespondences(
	map<card32, Vector3f> &featureCorrespondencesLeft,
	map<card32, Vector3f> &featureCorrespondencesRight,
	AnimationBSRContext* context, 
	const int32 sourceFrame, 
	PatchGraph *left,
	PatchGraph *right)
{
	PatchGraphContainer *pgc = context->getPGContainer();
	FeatureCorrespondences *corr = pgc->getFeatureCorrespondences();
	
	if (!corr) return;
	if (corr->getNumFrames() <= 0) return;

	TrajectorySet *leftTS = left->getTrajectorySet();
	TrajectorySet *rightTS = right->getTrajectorySet();
	leftTS->invalidateAllAttachments();
	rightTS->invalidateAllAttachments();

	AttachedIndexedOctree *oLeft = leftTS->fetchOctree(leftTS->getNumFrames()-1);		
	AttachedIndexedOctree *oRight = rightTS->fetchOctree(0);		
	
	PointSet *leftPSCopy = leftTS->copySingleFrame(leftTS->getNumFrames()-1);
	PointSet *rightPSCopy = rightTS->copySingleFrame(0);

	PointSetKNNQuery knnLeft(leftPSCopy, oLeft, NULL_VECTOR3F, leftPSCopy->getAAT("position"));
	PointSetKNNQuery knnRight(rightPSCopy, oRight, NULL_VECTOR3F, rightPSCopy->getAAT("position"));
	
	const card32 numCorr = corr->getNumCorrespondenciesInFrame(sourceFrame);
	for (card32 i=0; i<numCorr; i++) {
		if (corr->getCorrespondence(sourceFrame, i)->getFlags() == Correspondence::CF_Initial) continue;
		Vector3f leftPos = corr->getCorrespondence(sourceFrame, i)->getPosition1();
		Vector3f rightPos = corr->getCorrespondence(sourceFrame, i)->getPosition2();
		knnLeft.setTargetPoint(leftPos);
		knnRight.setTargetPoint(rightPos);
		
		card32 leftIndex;
		card32 rightIndex;
		
		bool found;
		knnLeft.getNextNearestPointIndex(leftIndex, found);
		if (!found) {
			warning("buildFeatureCorrespondences - something's wrong");
			continue;
		}
		knnRight.getNextNearestPointIndex(rightIndex, found);
		if (!found) {
			warning("buildFeatureCorrespondences - something's wrong");
			continue;
		}
		
		pair<card32, Vector3f> vl(leftIndex, rightPos);
		featureCorrespondencesLeft.insert(vl);
		pair<card32, Vector3f> vr(rightIndex, leftPos);
		featureCorrespondencesRight.insert(vr);
	}
	
	delete leftPSCopy;
	delete rightPSCopy;
}


static void debugOutputDegreeHistogram(string caption, const InCorePCTopologyGraph *tg) {
   debugOutput << "---\n";
   debugOutput << caption << "\n";
   vector<card32> h;
   PGOpt2Debugging::computeTopDegHistogram(tg, h);
   PGOpt2Debugging::ouputHistogram(h);
   debugOutput << "---\n";
}

static void debugOutputDegree(string caption, const InCorePCTopologyGraph *tg) {
   debugOutput << "---\n";
   debugOutput << caption << "\n";

   for (card32 i=0; i<tg->getNumVertices(); i++) {
      debugOutput << i << "\t" << tg->getNumAdjacentVertices(i) << "\n";
   }

   debugOutput << "---\n";
}

static void debugOutputAverageDegree(string caption, const InCorePCTopologyGraph *tg) {
   debugOutput << caption << " average node degree: ";

   double av = 0.0;

   for (card32 i=0; i<tg->getNumVertices(); i++) {
      av += tg->getNumAdjacentVertices(i);
   }
   av /= tg->getNumVertices();
   
   debugOutput << av << "\n";
}


void PatchGraphMerge2::mergeTwoFrameRanges(
	AnimationBSRContext *context, 
	const int32 pgIndex1, 
	const int32 pgIndex2, 
   PatchGraphContainer *container,
	DebugRenderer *debugRenderer,
	AnimationBSRRenderer* renderer)
{
   Timer timer;
	card32 mergeStartTime = timer.getCurrentValue();

	// --- retrieve parts and check ---
	
	// Test for valid index
	PatchGraph* pg1 = context->getPGContainer()->getPatchGraph( pgIndex1 );
	PointCloudAnimationSequence* animSeq = context->getPGContainer()->getAnimSeq();
	PatchGraph* pg2 = context->getPGContainer()->getPatchGraph( pgIndex2 );
	if (pg1 == NULL || pg2 == NULL)	{
		debugOutput << "merge: invalid indices!\n"; debugOutput.flush();
		return;
	}
   if (context->getVerbose()) {
	   debugOutput << "\n\nMerging patches "
		   << pg1->getStartFrameIndex() << ".." << ( pg1->getStartFrameIndex() + pg1->getTrajectorySet()->getNumFrames() - 1 )
		   << " with " << pg2->getStartFrameIndex() << ".." << ( pg2->getStartFrameIndex() + pg2->getTrajectorySet()->getNumFrames() - 1 ) << ".\n"; 
	   debugOutput.flush();
	   if ( pg1->getTrajectorySet()->getNumFrames() + pg1->getStartFrameIndex() != pg2->getStartFrameIndex() ){
		   debugOutput << "merge: can only merge adjacent patch graphs.\n"; 
		   debugOutput.flush();
		   return;
	   }
   }
   
   PGOpt2RemoveLowDegreeVertices::process(context, pg1);
   PGOpt2RemoveLowDegreeVertices::process(context, pg2);

	pg1->createPatchAttribute(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
	pg1->createPatchAttribute(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
	pg1->createPatchAttribute(mVAD("varIndex_pos", 1, VAD::DATA_FORMAT_INT32));
	pg1->createPatchAttribute(mVAD("varIndex_nor", 1, VAD::DATA_FORMAT_INT32));

	pg2->createPatchAttribute(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
	pg2->createPatchAttribute(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
	pg2->createPatchAttribute(mVAD("varIndex_nor", 1, VAD::DATA_FORMAT_INT32));
	pg2->createPatchAttribute(mVAD("varIndex_pos", 1, VAD::DATA_FORMAT_INT32));

   animSeq->createDataAttribute(mVAD("flag_inlier", 1, VAD::DATA_FORMAT_INT32));
   
   debugOutputAverageDegree("pg1", pg1->getTopology());
   debugOutputAverageDegree("pg2", pg2->getTopology());
  
  
	// --- create combined patch graph ---

   if (context->getVerbose()) {
      debugOutput << "pg1: " << pg1->getTrajectorySet()->getNumFrames() << " frames, " << pg1->getTrajectorySet()->getNumPatches() << " patches.\n";
      debugOutput << "pg2: " << pg2->getTrajectorySet()->getNumFrames() << " frames, " << pg2->getTrajectorySet()->getNumPatches() << " patches.\n";
      debugOutput << "input complexity:\n pg1: " << pg1->getTrajectorySet()->getNumFrames() << " frames x " << pg1->getTrajectorySet()->getNumPatches() << " patches.\n\n";      
      debugOutput << "input complexity:\n pg2: " << pg2->getTrajectorySet()->getNumFrames() << " frames x " << pg2->getTrajectorySet()->getNumPatches() << " patches.\n\n";      
      debugOutput << "\nassembling combinedPG...\n"; 
      debugOutput.flush();
   }
   PatchGraph *combinedPG = joinParts->createCombinedPatchGraph(pg1, pg2);

   if (context->getVerbose()) {debugOutput << "setup time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}


	// --- pairwise alignment ---

	// Correspondences
	map<card32, Vector3f> featureCorrespondences12;
	map<card32, Vector3f> featureCorrespondences21;
	buildFeatureCorrespondences(
		featureCorrespondences12, 
		featureCorrespondences21, 
		context, 
		pg2->getStartFrameIndex()-1, 
		pg1,
		pg2
	);	

	// Extrapolate information from pg1 into frames in pg2 and vice versa
	card32 alignStartTime = timer.getCurrentValue();
	OptimizationRecord2 optRecord12;
	OptimizationRecord2 optRecord21;
   if (context->getVerbose()) {debugOutput << "\npairwise alignment12...\n"; debugOutput.flush();}
	PointSet *aligned12 = alignParts->alignParts(context, pg1, pg2, pg1->getTrajectorySet()->getNumFrames()-1, 0, featureCorrespondences12, debugRenderer, &optRecord12);
   if (context->getVerbose()) {debugOutput << "\npairwise alignment21...\n"; debugOutput.flush();}
	PointSet *aligned21 = alignParts->alignParts(context, pg2, pg1, 0, pg1->getTrajectorySet()->getNumFrames()-1, featureCorrespondences21, debugRenderer, &optRecord21);
   
   if (context->getVerbose()) {debugOutput << "pairwise alignment time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}


	// --- merge topology I ---

   if (context->getVerbose()) {debugOutput << "\ntopology merging (I) - merged aligned pieces...\n"; debugOutput.flush();}
   card32 whichPartHasBeenMerged = mergeAlignedRegion->merge(context, combinedPG, pg1, pg2, aligned12, &optRecord12, aligned21, &optRecord21);
   if (context->getVerbose()) {debugOutputAverageDegree("combinedPG", combinedPG->getTopology());}
   if (context->getVerbose()) {debugOutput << "topology merging time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   if (debugStopEarly == 1) goto earlyDebugExit;

	// --- reverse pairwise alignment ---

   if (context->getVerbose()) {
      debugOutput << "\nreverse pairwise alignment - " << ((whichPartHasBeenMerged==0)?"deform left is already there, now deforming right...\n":"deform right is already there, now deforming left...\n");
      debugOutput.flush();
   }
   alignParts->reverseAlignment(
      context, 
      combinedPG, 
      whichPartHasBeenMerged,
      pg1, pg2, 
      (whichPartHasBeenMerged == TopologyMerging2::MERGED_LEFT) ? featureCorrespondences21 : featureCorrespondences12,
      debugRenderer
   );

   if (context->getVerbose()) {debugOutput << "reverse pairwise alignment time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   //PGOpt2Debugging::searchLongEdges(context, combinedPG);

	// --- merge topology ---

   if (context->getVerbose()) {debugOutput << "\ntopology merging (II) - merged reversly aligned part...\n"; debugOutput.flush();}
   combinedPG->getTrajectorySet()->invalidateAllAttachments();
   mergeAlignedRegion->recomputeTopology(context, combinedPG);
   if (context->getVerbose()) {debugOutputAverageDegree("combinedPG", combinedPG->getTopology());}
   if (context->getVerbose()) {debugOutput << "topology merging time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   //PGOpt2Debugging::searchLongEdges(context, combinedPG);

   if (debugStopEarly == 2) goto earlyDebugExit;

	// --- hole / new-trajectory interpolation ---

   if (context->getVerbose()) {debugOutput << "\nhole filling...\n"; debugOutput.flush();}

	{
	   OptimizationRecord2 optRecord;
	   fillHoles->fillHoles(context, combinedPG, context->getPGContainer()->getAnimSeq(), debugRenderer, &optRecord);
   }

   if (context->getVerbose()) {debugOutput << "trajectories/hole filling time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   if (debugStopEarly == 3) goto earlyDebugExit;

	// --- recompute topology / resample ---

   //if (context->getVerbose()) {debugOutput << "\nrecompute topology (merging part 2 - after reverse alignment)...\n"; debugOutput.flush();}
   //card32 topologyMergingTime2 = timer.getCurrentValue();
   //{
   //   RecomputeTopology2 rec;
   //   rec.setupStandardFilterConfig(RecomputeTopology2::FILTER_INIT_LENGTH_DEV);
   //   InCorePCTopologyGraph *tg = rec.recomputeTopology(context, combinedPG);
   //   combinedPG->setTopology(tg);
   //   //PGOpt2RemoveLowDegreeVertices::process(context, combinedPG);
   //}
   //if (context->getVerbose()) {debugOutputAverageDegree("combinedPG", combinedPG->getTopology());}
   //if (context->getVerbose()) {debugOutput << "topology merging 2 time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   //PGOpt2Debugging::searchLongEdges(context, combinedPG);

   if (context->getVerbose()) {debugOutput << "\nresampling...\n"; debugOutput.flush();}
   resampler->resample(context, combinedPG);
   if (context->getVerbose()) {debugOutput << "resample time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   //PGOpt2Debugging::searchLongEdges(context, combinedPG);

   //PGOpt2Debugging::outputGraph(combinedPG->getTopology());
   //PGOpt2Debugging::outputPoints(combinedPG);

   PGOpt2RemoveLowDegreeVertices::process(context, combinedPG);

   //PGOpt2Debugging::outputGraph(combinedPG->getTopology());
   //PGOpt2Debugging::outputPoints(combinedPG);

   //PGOpt2Debugging::searchLongEdges(context, combinedPG);

   if (debugStopEarly == 4) goto earlyDebugExit;

	// --- global optimization ---

   if (context->getVerbose()) {debugOutput << "\nglobal optimization...\n"; debugOutput.flush();}
   
   {
   	OptimizationRecord2 optRecord;
      globalOpt->optimize(context, combinedPG, context->getPGContainer()->getAnimSeq(), debugRenderer, &optRecord);
   }
   
   if (context->getVerbose()) {debugOutput << "global opt time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}


	// // --- resampling ---

   // if (context->getVerbose()) {debugOutput << "\nresampling...\n"; debugOutput.flush();}
   // resampler->resample(context, combinedPG);
   // if (context->getVerbose()) {debugOutput << "resample time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}

   // nasty direct jump debug hack ;-)
   earlyDebugExit:

	// Update bounding box
	combinedPG->computeBBox();

	// Delete merged patch graphs
	context->getPGContainer()->deletePatchGraph( pgIndex1 );
	context->getPGContainer()->deletePatchGraph( ( pgIndex1 < pgIndex2 )? pgIndex2 - 1 : pgIndex2 );
	context->getPGContainer()->addPatchGraph( combinedPG );

	card32 endTime = timer.getCurrentValue();

   if (context->getVerbose()) {
      debugOutput << "overall merge time: " << (endTime - mergeStartTime)/1000.0 << " seconds.\n\n"; 
      debugOutput << "output complexity: " << combinedPG->getTrajectorySet()->getNumFrames() << " frames x " << combinedPG->getTrajectorySet()->getNumPatches() << " patches.\n\n";      
      debugOutput.flush();
   }
}


void PatchGraphMerge2::performGlobalOptOnly(
	AnimationBSRContext *context, 
	PatchGraph *pg,
	PointCloudAnimationSequence *data,
	DebugRenderer *debugRenderer,
	AnimationBSRRenderer* renderer)
{
	Timer timer;
   if (context->getVerbose()) {debugOutput << "\nglobal optimization...\n"; debugOutput.flush();}

	pg->createPatchAttribute(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
	pg->createPatchAttribute(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
	pg->createPatchAttribute(mVAD("varIndex_pos", 1, VAD::DATA_FORMAT_INT32));
	pg->createPatchAttribute(mVAD("varIndex_nor", 1, VAD::DATA_FORMAT_INT32));
	pg->createPatchAttribute(mVAD("topology_depth", 1, VAD::DATA_FORMAT_INT32));	
	data->createDataAttribute(mVAD("flag_inlier", 1, VAD::DATA_FORMAT_INT32));

	OptimizationRecord2 optRecord;
   globalOpt->optimize(context, pg, data, debugRenderer, &optRecord);
   if (context->getVerbose()) {debugOutput << "global opt time: " << timer.getDeltaValue()/1000.0 << " seconds.\n";}	
}

void PatchGraphMerge2::recomputeTopologyOnly(
   AnimationBSRContext *context, PatchGraph *pg) 
{
   pg->getTrajectorySet()->invalidateAllAttachments();
   mergeAlignedRegion->recomputeTopology(context, pg);
}

void PatchGraphMerge2::resampleOnly( 
	AnimationBSRContext *context,
	PatchGraph *pg)
{
	pg->getTrajectorySet()->invalidateAllAttachments();
	resampler->resample(context, pg);
}
