//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RecomputeTopology2.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "PointSetKNNQuery.h"
#include "InCorePCTopologyGraph.h"
#include "AnimationBSRContext.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( RecomputeTopology2 ,0) {
	BEGIN_CLASS_INIT( RecomputeTopology2 );
}

// ----

RecomputeTopology2::RecomputeTopology2() {
}

void RecomputeTopology2::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

RecomputeTopology2::~RecomputeTopology2() {
   clearFilters();
}




void RecomputeTopology2::clearFilters() 
{
   for (unsigned i=0; i<vertexFilters.size(); i++) {
      delete vertexFilters[i];
   }
   vertexFilters.clear();
   for (unsigned i=0; i<edgeFilters.size(); i++) {
      delete edgeFilters[i];
   }
   edgeFilters.clear();
}

void RecomputeTopology2::addVertexFilter(RecTop2VertexFilter *filter) 
{
   vertexFilters.push_back(filter);
}

void RecomputeTopology2::addEdgeFilter(RecTop2EdgeFilter *filter) 
{
   edgeFilters.push_back(filter);
}

void RecomputeTopology2::setupStandardFilterConfig(StandardFilterConfiguration config) {
   clearFilters();
   switch(config) {
      case FILTER_INIT: {
         addVertexFilter(new RT2VertexFlagsFilter());
         addEdgeFilter(new RT2EdgeInitFilter());
         break;
      }
      case FILTER_INIT_LENGTH_DEV: {
         addVertexFilter(new RT2VertexFlagsFilter());
         addEdgeFilter(new RT2EdgeInitFilter());
         addEdgeFilter(new RT2EdgeLengthChangeFilter());
         break;
      }
      case NO_FILTER:
      default:
         break;
   
   };
}



inline bool RecomputeTopology2::useVertex(const RecTop2PGVertex v) {
   size_t n=vertexFilters.size();
   for (unsigned i=0; i<n;i++) {
      if (!vertexFilters[i]->filterVertex(v)) return false;
   }
   return true;
}

inline bool RecomputeTopology2::useEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2) {
   size_t n=edgeFilters.size();
   for (unsigned i=0; i<n;i++) {
      if (!edgeFilters[i]->filterEdge(v1, v2)) return false;
   }
   return true;
}

InCorePCTopologyGraph *RecomputeTopology2::recomputeTopology(const AnimationBSRContext *context, const PatchGraph *pg)
{
   const TrajectorySet *traj = pg->getTrajectorySet();
   const card32 numFrames = traj->getNumFrames();
   const card32 numPatches = traj->getNumPatches();
   const PointSet *ps = traj->getTrajectories();
   const AAT psPosAAT = ps->getAAT("position");
   
   for (unsigned i=0; i<vertexFilters.size(); i++) {
      vertexFilters[i]->begin(context, pg, ps);
   }
   for (unsigned i=0; i<edgeFilters.size(); i++) {
      edgeFilters[i]->begin(context, pg, ps);
   }
   
   InCorePCTopologyGraph *result = new InCorePCTopologyGraph();
   result->clearAndSetup(pg->getTopology()->getEdges()->getDescr());

   card32 rejectedE = 0;
   card32 rejectedV = 0;
   card32 candidateE = 0;
   card32 candidateV = 0;

   /// new topology
   for (card32 frame=0; frame<numFrames; frame++) {
      PointSet *singleFrame = traj->copySingleFrame(frame);
      //const_cast<TrajectorySet*>(traj)->invalidateAllAttachments();
      AttachedIndexedOctree *octree = const_cast<TrajectorySet*>(traj)->fetchOctree(frame);

      PointSetKNNQuery knn(singleFrame, octree, NULL_VECTOR3F, singleFrame->getAAT("position"));

      for (card32 patch=0; patch<numPatches; patch++) {
         Vector3f targetPoint = ps->get<float32, 3>(frame, patch, psPosAAT);
         knn.setTargetPoint(targetPoint);
         candidateV++;
         if (useVertex(RecTop2PGVertex(frame, patch))) {
            card32 k=0;
            while (k <= context->getTopology2NumNearestNeighbors()) {
               card32 neighborIndex;
               bool found;
               knn.getNextNearestPointIndex(neighborIndex, found);
               if (neighborIndex == patch) {candidateV--; continue;}
               if (!found) break;
               Vector3f nextPos = ps->get<float32, 3>(frame, neighborIndex, psPosAAT);
               if (norm(nextPos-targetPoint) > context->getTopology2MaxDistanceRelSampleSpacing()*context->getSamplingSpacing()) break;
               
               candidateE++;
               if (useEdge(RecTop2PGVertex(frame, patch), RecTop2PGVertex(frame, neighborIndex))) {
                  result->addEdge(patch, neighborIndex);
               } else {
                  rejectedE++;
               }
               
               k++;
            }   
         } else {
            rejectedV++;
         }
      }      
      delete singleFrame;
   }
   
   if (context->getVerbose()) {
      debugOutput << "recomputeTopology\nrejected vertices: " << rejectedV << "/" << candidateV << "\nrejected edges: " << rejectedE << "/" << candidateE <<"\n\n";
   }
   
   for (unsigned i=0; i<vertexFilters.size(); i++) {
      vertexFilters[i]->end(context, pg, ps);
   }
   for (unsigned i=0; i<edgeFilters.size(); i++) {
      edgeFilters[i]->end(context, pg, ps);
   }

   return result;
}



RT2VertexFlagsFilter::RT2VertexFlagsFilter(int32 requiredMask, int32 excludeMask)
{
   context = NULL;
   pg = NULL;
   trajectories = NULL;
   this->requiredMask = requiredMask;
   this->excludeMask = excludeMask;
   flagsAAT = NULL_AAT;
}

void RT2VertexFlagsFilter::begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) 
{
   this->context = context;
   this->pg = pg;
   this->trajectories = trajectories;
   flagsAAT = trajectories->getAAT("flags");
}

bool RT2VertexFlagsFilter::filterVertex(const RecTop2PGVertex v) 
{
   int32 f = trajectories->get<int>(v.frame, v.patch, flagsAAT);
   return ((f & requiredMask) != 0) && ((f & excludeMask) == 0);
}

void RT2VertexFlagsFilter::end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) {}


// ----


RT2EdgeLengthChangeFilter::RT2EdgeLengthChangeFilter() 
{
   context = NULL;
   pg = NULL;
   trajectories = NULL;
}

void RT2EdgeLengthChangeFilter::begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) 
{
   this->context = context;
   this->pg = pg;
   this->trajectories = trajectories;
   positionAAT = trajectories->getAAT("position");
   flagsAAT = trajectories->getAAT("flags");
}

bool RT2EdgeLengthChangeFilter::filterEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2) 
{
   float32 minLength = 1e35f;
   float32 maxLength = -1e35f;
   card32 w = trajectories->getWidth();
   if (w == 0) {
      pAssert(false);
      return false;
   }
   bool found = false;
   for (card32 i=0; i<w; i++) {
      int32 flags1 = trajectories->get<int>(i, v1.patch, flagsAAT);
      int32 flags2 = trajectories->get<int>(i, v2.patch, flagsAAT);
      if (((flags1 & FLAG_PATCH_INITIALIZED )!= 0) && ((flags2 & FLAG_PATCH_INITIALIZED )!= 0)) {
         float32 l = norm(trajectories->get<float32, 3>(i, v1.patch, positionAAT) - trajectories->get<float32, 3>(i, v2.patch, positionAAT));
         if (minLength > l) minLength = l;
         if (maxLength < l) maxLength = l;
         found = true;
      }
   }
   if (!found) {
      return false;
   }
   if (maxLength-minLength > context->getTopology2MaxAdditiveEdgeLengthDev()) {
      return false;
   }
   
   if (maxLength > context->getTopology2MinMultRejectLengthFraction() * context->getSamplingSpacing()) {
      if (context->getTopology2MaxMultiplicativeEdgeLengthDev() > 0) {
         if (minLength/maxLength < 1.0f/context->getTopology2MaxMultiplicativeEdgeLengthDev()) {
            return false;
         }
      }
   }
   return true;
}

void RT2EdgeLengthChangeFilter::end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) {}




RT2EdgeInitFilter::RT2EdgeInitFilter() 
{
   context = NULL;
   pg = NULL;
   trajectories = NULL;
}

void RT2EdgeInitFilter::begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) 
{
   this->context = context;
   this->pg = pg;
   this->trajectories = trajectories;
   flagsAAT = trajectories->getAAT("flags");
}

bool RT2EdgeInitFilter::filterEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2) 
{
   pAssert(v1.frame == v2.frame);
   int32 flags1 = trajectories->get<int>(v1.frame, v1.patch, flagsAAT);
   int32 flags2 = trajectories->get<int>(v2.frame, v2.patch, flagsAAT);
   return (((flags1 & FLAG_PATCH_INITIALIZED )!= 0) && ((flags2 & FLAG_PATCH_INITIALIZED )!= 0));
}

void RT2EdgeInitFilter::end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) {}
