//---------------------------------------------------------------------------
#ifndef RecomputeTopology2H
#define RecomputeTopology2H
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PatchTypes.h"
//---------------------------------------------------------------------------

namespace geoxl {

class AnimationBSRContext;
class PatchGraph;
class InCorePCTopologyGraph;


/// identifies one patch graph patch (not in urshape)
struct RecTop2PGVertex {
   card32 frame;
   card32 patch;
   
   inline RecTop2PGVertex() {}
   inline RecTop2PGVertex(card32 f, card32 p) {frame=f; patch=p;}
};

/// decide whether vertices should be considered for inclusion in the topology
class RecTop2VertexFilter : public Persistent {
 public:
   virtual void begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) = 0;
   virtual bool filterVertex(const RecTop2PGVertex v) = 0;
   virtual void end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) = 0;
};

/// decide whether edges should be kept in the topology
class RecTop2EdgeFilter : public Persistent {
 public:
   virtual void begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) = 0;
   virtual bool filterEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2) = 0;
   virtual void end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories) = 0;
};


///
///   this class recomputes the topology of a patch graph
///   a set of filters can be specified to control the behaviour
///   unless specified otherwise by filters, all kNN points in eps distance are considered
///   the corresponding parameters (k, eps) are taken from the AnimationBSRContext
///
class ANIMATIONRECONSTRUCTION_API RecomputeTopology2 : public Persistent {
	GEOXL_CLASS( RecomputeTopology2 )
 private:
   vector<RecTop2VertexFilter*> vertexFilters;
   vector<RecTop2EdgeFilter*> edgeFilters;
  
   inline bool useVertex(const RecTop2PGVertex v);
   inline bool useEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2);
   
 public:
	RecomputeTopology2();

	virtual void assign(const Object* obj, CopyContext *context = NULL);

   void clearFilters();
   /// this object takes ownership
   void addVertexFilter(RecTop2VertexFilter *filter);
   /// this object takes ownership
   void addEdgeFilter(RecTop2EdgeFilter *filter);
   
   enum StandardFilterConfiguration {NO_FILTER, FILTER_INIT, FILTER_INIT_LENGTH_DEV};
   /// setup a standard configuration
   void setupStandardFilterConfig(StandardFilterConfiguration config);

   /// recomputes topology, returns the new topology (input pg remains unchanged)
   InCorePCTopologyGraph *recomputeTopology(const AnimationBSRContext *context, const PatchGraph *pg);

   ~RecomputeTopology2();
};


// ---

/// removes vertices that do not provide certain flags (typically: being initilized)
class RT2VertexFlagsFilter : public RecTop2VertexFilter {
 private:
   const AnimationBSRContext *context;
   const PatchGraph *pg;
   const PointSet *trajectories;
   int32 requiredMask;
   int32 excludeMask;
   AAT flagsAAT;

 public:
   RT2VertexFlagsFilter(int32 requiredMask = FLAG_PATCH_INITIALIZED, int32 excludeMask = 0);
   virtual void begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
   virtual bool filterVertex(const RecTop2PGVertex v);
   virtual void end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
};

/// removes edges to uninitialized patches
class RT2EdgeInitFilter : public RecTop2EdgeFilter {
 private:
   const AnimationBSRContext *context;
   const PatchGraph *pg;
   const PointSet *trajectories;
   AAT flagsAAT;

 public:
   RT2EdgeInitFilter();
   virtual void begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
   virtual bool filterEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2);
   virtual void end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
};

/// removes edges that vary too much in length over time
class RT2EdgeLengthChangeFilter : public RecTop2EdgeFilter {
 private:
   const AnimationBSRContext *context;
   const PatchGraph *pg;
   const PointSet *trajectories;
   AAT positionAAT;
   AAT flagsAAT;

 public:
   RT2EdgeLengthChangeFilter();
   virtual void begin(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
   virtual bool filterEdge(const RecTop2PGVertex v1, const RecTop2PGVertex v2);
   virtual void end(const AnimationBSRContext *context, const PatchGraph *pg, const PointSet *trajectories);
};

}

#endif