// Magic Software, Inc.
// http://www.magic-software.com
// Copyright (c) 2000, 2001.  All Rights Reserved
//
// Source code from Magic Software is supplied under the terms of a license
// agreement and may not be copied or disclosed except in accordance with the
// terms of that agreement.  The various license agreements may be found at
// the Magic Software web site.  This file is subject to the license
//
// FREE SOURCE CODE
// http://www.magic-software.com/License/free.pdf

#ifndef MGCLINEARSYSTEM_H
#define MGCLINEARSYSTEM_H



namespace geoxl {

template<class FloatType>
class LinearSystem
{
public:
    // 2x2 and 3x3 systems (avoids overhead of Gaussian elimination)
    inline FloatType& Tolerance ();
    inline bool Solve2 (const FloatType aafA[2][2], const FloatType afB[2],
        FloatType afX[2]);
    inline bool Solve3 (const FloatType aafA[3][3], const FloatType afB[3],
        FloatType afX[3]);

    // convenience for allocation, memory is zeroed out
    inline FloatType** NewMatrix (int iSize);
    inline void DeleteMatrix (int iSize, FloatType** aafA);
    inline FloatType* NewVector (int iSize);
    inline void DeleteVector (FloatType* afB);

    // Input:
    //     A[iSize][iSize], entries are A[row][col]
    // Output:
    //     return value is TRUE if successful, FALSE if pivoting failed
    //     A[iSize][iSize], inverse matrix
    inline bool Inverse (int iSize, FloatType** aafA);

    // Input:
    //     A[iSize][iSize] coefficient matrix, entries are A[row][col]
    //     B[iSize] vector, entries are B[row]
    // Output:
    //     return value is TRUE if successful, FALSE if pivoting failed
    //     A[iSize][iSize] is inverse matrix
    //     B[iSize] is solution x to Ax = B
    inline bool Solve (int iSize, FloatType** aafA, FloatType* afB);

    // Input:
    //     Matrix is tridiagonal.
    //     Lower diagonal A[iSize-1]
    //     Main  diagonal B[iSize]
    //     Upper diagonal C[iSize-1]
    //     Right-hand side R[iSize]
    // Output:
    //     return value is TRUE if successful, FALSE if pivoting failed
    //     U[iSize] is solution
    inline bool SolveTri (int iSize, FloatType* afA, FloatType* afB, FloatType* afC,
        FloatType* afR, FloatType* afU);

    // Input:
    //     Matrix is tridiagonal.
    //     Lower diagonal is constant, A
    //     Main  diagonal is constant, B
    //     Upper diagonal is constant, C
    //     Right-hand side Rr[iSize]
    // Output:
    //     return value is TRUE if successful, FALSE if pivoting failed
    //     U[iSize] is solution
    inline bool SolveConstTri (int iSize, FloatType fA, FloatType fB, FloatType fC,
        FloatType* afR, FloatType* afU);

    // Input:
    //     A[iSize][iSize] symmetric matrix, entries are A[row][col]
    //     B[iSize] vector, entries are B[row]
    // Output:
    //     return value is TRUE if successful, FALSE if (nearly) singular
    //     decomposition A = L D L^t (diagonal terms of L are all 1)
    //         A[i][i] = entries of diagonal D
    //         A[i][j] for i > j = lower triangular part of L
    //     B[iSize] is solution to x to Ax = B
    inline bool SolveSymmetric (int iSize, FloatType** aafA, FloatType* afB);

    // Input:
    //     A[iSize][iSize], entries are A[row][col]
    // Output:
    //     return value is TRUE if successful, FALSE if algorithm failed
    //     AInv[iSize][iSize], inverse matrix
    inline bool SymmetricInverse (int iSize, FloatType** aafA, FloatType** aafAInv);

public:
    // tolerance for 2x2 and 3x3 system solving
    FloatType ms_fTolerance;
};


} // namespace Mgc


#include "MgcLinearSystem_templCode.h"

#endif
