//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Monomials2ndOrder.h"
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools_templCode.h"
#include "MathTools.h"
#include "MathTools_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

const float32 INV_PIVOT_THRESHOLD = 1e-7f;

IMPLEMENT_GEOXL_CLASS( Monomials1stOrder, 0 ) {
	BEGIN_CLASS_INIT( Monomials1stOrder )
}

IMPLEMENT_GEOXL_CLASS( Monomials2ndOrder, 0 ) {
	BEGIN_CLASS_INIT( Monomials2ndOrder )
}


// ----


// -- 1nd order


void Monomials1stOrder::computeCoeffs(
	const vector<Vector3f> &points,
	StaticVector<float32, 3> &coeffs,
	vector< StaticVector<float32, 3> > &coeffsDerivs,
	float32 gaussianWindowVariance )
{
	const size_t NUM_NEIGHBORS = points.size();
	coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f;
	coeffsDerivs.resize( NUM_NEIGHBORS );
	StaticVector<float32, 3> cd;
	StaticMatrix<float32, 3,3> dm;
	for (card32 i=0; i<3;i++) {
		for (card32 j=0; j<3;j++) {
			dm[i][j] = 0.0f;
		}
	}

	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 u = points[pIndex][0];
		const float32 v = points[pIndex][1];
		const float32 n = points[pIndex][2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(u*u+v*v)/gaussianWindowVariance)) : 1) / (float32)NUM_NEIGHBORS;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = u   * w;
		cd[2] = v   * w;
		coeffsDerivs[pIndex] = cd;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;  
		dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;
		dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;
	}

	StaticMatrix<float32, 3,3> invDm;
	bool success;
	invDm = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
	if (!success) {
		for (card32 i=0; i<3; i++) {
			for (card32 j=0; j<3; j++) {
				invDm[i][j] = 0.0f;
			}
		}
	}
	coeffs = invDm * coeffs;
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		coeffsDerivs[pIndex] = invDm*coeffsDerivs[pIndex];
	}
}

void Monomials1stOrder::computeCoeffs(
	const vector<Vector3f> &points,
	StaticVector<float32, 3> &coeffs,
	float32 gaussianWindowVariance )
{
	const size_t NUM_NEIGHBORS = points.size();
	coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f;
	StaticVector<float32, 3> cd;
	StaticMatrix<float32, 3,3> dm;
	for (card32 i=0; i<3;i++) {
		for (card32 j=0; j<3;j++) {
			dm[i][j] = 0.0f;
		}
	}

	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 u = points[pIndex][0];
		const float32 v = points[pIndex][1];
		const float32 n = points[pIndex][2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(u*u+v*v)/gaussianWindowVariance)) : 1) / (float32)NUM_NEIGHBORS;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = u   * w;
		cd[2] = v   * w;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;  
		dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;
		dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;
	}

	StaticMatrix<float32, 3,3> invDm;
	bool success;
	invDm = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
	if (!success) {
		for (card32 i=0; i<3; i++) {
			for (card32 j=0; j<3; j++) {
				invDm[i][j] = 0.0f;
			}
		}
	}
	coeffs = invDm * coeffs;
}

float32 Monomials1stOrder::evaluate(
   const StaticVector<float32, 3> &coeffs,
   const Vector2f &position)
{
   const float32 u = position[0];
   const float32 v = position[1];
   return   coeffs[0]
          + coeffs[1] * u
          + coeffs[2] * v;
}


// -- 1st order 3d -> 3d


void AffineMapping3dFit::computeCoeffs(
   const vector<MapPoint> &points,
   AffineMap &leastSquareFit,
   vector< AffineMap > &coeffsDerivs,
   float32 inversionSingularValThreshold,
   float32 gaussianWindowVariance )
{
   vector< Vector4f > coeffsDerivs_x, coeffsDerivs_y, coeffsDerivs_z;
   Vector4f coeffs_x = NULL_VECTOR4F;
   Vector4f coeffs_y = NULL_VECTOR4F;
   Vector4f coeffs_z = NULL_VECTOR4F;
	const size_t NUM_NEIGHBORS = points.size();
	leastSquareFit.b = NULL_VECTOR3F;
	leastSquareFit.M[0] = NULL_VECTOR3F;
	leastSquareFit.M[1] = NULL_VECTOR3F;
	leastSquareFit.M[2] = NULL_VECTOR3F;
	StaticVector<float32, 4> cd;
	StaticMatrix<float32, 4,4> dm;
	for (card32 i=0; i<4;i++) {
		for (card32 j=0; j<4;j++) {
			dm[i][j] = 0.0f;
		}
	}
   coeffsDerivs_x.resize(NUM_NEIGHBORS, NULL_VECTOR4F);
   coeffsDerivs_y.resize(NUM_NEIGHBORS, NULL_VECTOR4F);
   coeffsDerivs_z.resize(NUM_NEIGHBORS, NULL_VECTOR4F);
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 us = points[pIndex].from[0];
		const float32 vs = points[pIndex].from[1];
		const float32 ns = points[pIndex].from[2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(us*us+vs*vs+ns*ns)/gaussianWindowVariance)) : 1) / (float32)NUM_NEIGHBORS;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = us  * w;
		cd[2] = vs  * w;
		cd[3] = ns  * w;
		coeffsDerivs_x[pIndex] = cd;
		coeffsDerivs_y[pIndex] = cd;
		coeffsDerivs_z[pIndex] = cd;

		coeffs_x += cd * points[pIndex].to[0];
		coeffs_y += cd * points[pIndex].to[1];
		coeffs_z += cd * points[pIndex].to[2];

		dm[0][0] += w;      dm[0][1] += us*w;      dm[0][2] += vs*w;       dm[0][3] += ns*w;  
		dm[1][0] += w*us;   dm[1][1] += us*w*us;   dm[1][2] += vs*w*us;    dm[1][3] += ns*w*us;
		dm[2][0] += w*vs;   dm[2][1] += us*w*vs;   dm[2][2] += vs*w*vs;    dm[2][3] += ns*w*vs;
		dm[3][0] += w*ns;   dm[3][1] += us*w*ns;   dm[3][2] += vs*w*ns;    dm[3][3] += ns*w*ns;
	}

	StaticMatrix<float32, 4,4> invDm;
	try {

		DMatrixF invDmD(4,4);
		convertMatrixType(dm, invDmD, 4, 4);

      DMatrixF u;
      DVectorF singularValues;
      DMatrixF v;
		
		DynamicPCA<float32>::computeSVD(invDmD, u, singularValues, v, true, false); /// SVD: A = U^t * D * V (therefore, transpose_u is set to true)
		
		singularValues[0] = (fabs(singularValues[0])>inversionSingularValThreshold) ? 1.0f/singularValues[0] : 0;
		singularValues[1] = (fabs(singularValues[1])>inversionSingularValThreshold) ? 1.0f/singularValues[1] : 0;
		singularValues[2] = (fabs(singularValues[2])>inversionSingularValThreshold) ? 1.0f/singularValues[2] : 0;
		singularValues[3] = (fabs(singularValues[3])>inversionSingularValThreshold) ? 1.0f/singularValues[3] : 0;

		v = v.transpose();
		v[0] *= singularValues[0];
		v[1] *= singularValues[1];
		v[2] *= singularValues[2];
		v[3] *= singularValues[3];
		
		u = u.transpose();
		
		invDmD = v*u;

		convertMatrixType(invDmD, invDm, 4, 4);

	} catch (...) {
		for (card32 i=0; i<3; i++) {
			for (card32 j=0; j<3; j++) {
				invDm[i][j] = 0.0f;
			}
		}
	}
	coeffs_x = invDm * coeffs_x;
	coeffs_y = invDm * coeffs_y;
	coeffs_z = invDm * coeffs_z;
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		coeffsDerivs_x[pIndex] = invDm*coeffsDerivs_x[pIndex];
		coeffsDerivs_y[pIndex] = invDm*coeffsDerivs_y[pIndex];
		coeffsDerivs_z[pIndex] = invDm*coeffsDerivs_z[pIndex];
	}
   leastSquareFit.M = makeMatrix3f(
      coeffs_x[1], coeffs_x[2], coeffs_x[3],
      coeffs_y[1], coeffs_y[2], coeffs_y[3],
      coeffs_z[1], coeffs_z[2], coeffs_z[3]
   );
   leastSquareFit.b = makeVector3f(coeffs_x[0], coeffs_y[0], coeffs_z[0]);

	coeffsDerivs.resize(NUM_NEIGHBORS);
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++) {
      coeffsDerivs[pIndex].M = makeMatrix3f(
         coeffsDerivs_x[pIndex][1], coeffsDerivs_x[pIndex][2], coeffsDerivs_x[pIndex][3],
         coeffsDerivs_y[pIndex][1], coeffsDerivs_y[pIndex][2], coeffsDerivs_y[pIndex][3],
         coeffsDerivs_z[pIndex][1], coeffsDerivs_z[pIndex][2], coeffsDerivs_z[pIndex][3]
      );
      coeffsDerivs[pIndex].b = makeVector3f(coeffsDerivs_x[pIndex][0], coeffsDerivs_y[pIndex][0], coeffsDerivs_z[pIndex][0]);
	}
	/// result is:
	/// 
	/// coeffsDerivs[i].b = ({d}pi.x c1, {d}pi.y c1, {d}pi.z c1,)
	///
	///                     ({d}pi.x cx, {d}pi.x cy, {d}pi.x cz)
	/// coeffsDerivs[i].M = ({d}pi.y cx, {d}pi.y cy, {d}pi.y cz)
	///                     ({d}pi.z cx, {d}pi.z cy, {d}pi.z cz)
	
}



void AffineMapping3dFit::computeCoeffs(
   const vector<MapPoint> &points,
   AffineMap &leastSquareFit,
   float32 inversionSingularValThreshold,
   float32 gaussianWindowVariance )
{
   Vector4f coeffs_x = NULL_VECTOR4F;
   Vector4f coeffs_y = NULL_VECTOR4F;
   Vector4f coeffs_z = NULL_VECTOR4F;
	const size_t NUM_NEIGHBORS = points.size();
	leastSquareFit.b = NULL_VECTOR3F;
	leastSquareFit.M[0] = NULL_VECTOR3F;
	leastSquareFit.M[1] = NULL_VECTOR3F;
	leastSquareFit.M[2] = NULL_VECTOR3F;
	StaticVector<float32, 4> cd;
	StaticMatrix<float32, 4,4> dm;
	for (card32 i=0; i<4;i++) {
		for (card32 j=0; j<4;j++) {
			dm[i][j] = 0.0f;
		}
	}
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 us = points[pIndex].from[0];
		const float32 vs = points[pIndex].from[1];
		const float32 ns = points[pIndex].from[2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(us*us+vs*vs+ns*ns)/gaussianWindowVariance)) : 1) / (float32)NUM_NEIGHBORS;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = us  * w;
		cd[2] = vs  * w;
		cd[3] = ns  * w;

		coeffs_x += cd * points[pIndex].to[0];
		coeffs_y += cd * points[pIndex].to[1];
		coeffs_z += cd * points[pIndex].to[2];

		dm[0][0] += w;      dm[0][1] += us*w;      dm[0][2] += vs*w;       dm[0][3] += ns*w;  
		dm[1][0] += w*us;   dm[1][1] += us*w*us;   dm[1][2] += vs*w*us;    dm[1][3] += ns*w*us;
		dm[2][0] += w*vs;   dm[2][1] += us*w*vs;   dm[2][2] += vs*w*vs;    dm[2][3] += ns*w*vs;
		dm[3][0] += w*ns;   dm[3][1] += us*w*ns;   dm[3][2] += vs*w*ns;    dm[3][3] += ns*w*ns;
	}

	StaticMatrix<float32, 4,4> invDm;
	try {

		DMatrixF invDmD(4,4);
		convertMatrixType(dm, invDmD, 4, 4);

      DMatrixF u;
      DVectorF singularValues;
      DMatrixF v;
		
		DynamicPCA<float32>::computeSVD(invDmD, u, singularValues, v, true, false); /// SVD: A = U^t * D * V (therefore, transpose_u is set to true)
		
		singularValues[0] = (fabs(singularValues[0])>inversionSingularValThreshold) ? 1.0f/singularValues[0] : 0;
		singularValues[1] = (fabs(singularValues[1])>inversionSingularValThreshold) ? 1.0f/singularValues[1] : 0;
		singularValues[2] = (fabs(singularValues[2])>inversionSingularValThreshold) ? 1.0f/singularValues[2] : 0;
		singularValues[3] = (fabs(singularValues[3])>inversionSingularValThreshold) ? 1.0f/singularValues[3] : 0;

		v = v.transpose();
		v[0] *= singularValues[0];
		v[1] *= singularValues[1];
		v[2] *= singularValues[2];
		v[3] *= singularValues[3];
		
		u = u.transpose();
		
		invDmD = v*u;

		convertMatrixType(invDmD, invDm, 4, 4);

	} catch (...) {
		for (card32 i=0; i<3; i++) {
			for (card32 j=0; j<3; j++) {
				invDm[i][j] = 0.0f;
			}
		}
	}
	coeffs_x = invDm * coeffs_x;
	coeffs_y = invDm * coeffs_y;
	coeffs_z = invDm * coeffs_z;
   leastSquareFit.M = makeMatrix3f(
      coeffs_x[1], coeffs_x[2], coeffs_x[3],
      coeffs_y[1], coeffs_y[2], coeffs_y[3],
      coeffs_z[1], coeffs_z[2], coeffs_z[3]
   );
   leastSquareFit.b = makeVector3f(coeffs_x[0], coeffs_y[0], coeffs_z[0]);
}


void AffineMapping3dFit::orthoghonalize(AffineMap &map) {
	DMatrixF M(3,3);
	convertMatrixType(map.M, M, 3, 3);

   static DMatrixF u(4,4);
   static DVectorF singularValues(4);
   static DMatrixF v(4,4);
	
	DynamicPCA<float32>::computeSVD(M, u, singularValues, v, true, false); /// SVD: A = U^t * D * V (therefore, transpose_u is set to true)

   Matrix3f us, vs;
   	
	convertMatrixType(u, us, 3, 3);
	convertMatrixType(v, vs, 3, 3);
	
	map.M = us.transpose()*vs;
}



// -- 2nd order


void Monomials2ndOrder::computeCoeffs(
	const vector<Vector3f> &points,
	StaticVector<float32, 6> &coeffs,
	vector< StaticVector<float32, 6> > &coeffsDerivs,
	float32 gaussianWindowVariance )
{
	const size_t NUM_NEIGHBORS = points.size();
	coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f; coeffs[3] = 0.0f; coeffs[4] = 0.0f; coeffs[5] = 0.0f;
	coeffsDerivs.resize( NUM_NEIGHBORS );
	StaticVector<float32, 6> cd;
	StaticMatrix<float32, 6,6> dm;
	for (card32 i=0; i<6;i++) {
		for (card32 j=0; j<6;j++) {
			dm[i][j] = 0.0f;
		}
	}

	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 u = points[pIndex][0];
		const float32 v = points[pIndex][1];
		const float32 n = points[pIndex][2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(u*u+v*v)/gaussianWindowVariance)) : 1) / (float32)NUM_NEIGHBORS;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = u   * w;
		cd[2] = v   * w;
		cd[3] = u*u * w;
		cd[4] = u*v * w;
		cd[5] = v*v * w;
		coeffsDerivs[pIndex] = cd;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;
		coeffs[3] += cd[3] * n;
		coeffs[4] += cd[4] * n;
		coeffs[5] += cd[5] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;     dm[0][3] += u*u*w;     dm[0][4] += u*v*w;     dm[0][5] += v*v*w;
		dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;   dm[1][3] += u*u*w*u;   dm[1][4] += u*v*w*u;   dm[1][5] += v*v*w*u;
		dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;   dm[2][3] += u*u*w*v;   dm[2][4] += u*v*w*v;   dm[2][5] += v*v*w*v;
		dm[3][0] += w*u*u; dm[3][1] += u*w*u*u; dm[3][2] += v*w*u*u; dm[3][3] += u*u*w*u*u; dm[3][4] += u*v*w*u*u; dm[3][5] += v*v*w*u*u;
		dm[4][0] += w*u*v; dm[4][1] += u*w*u*v; dm[4][2] += v*w*u*v; dm[4][3] += u*u*w*u*v; dm[4][4] += u*v*w*u*v; dm[4][5] += v*v*w*u*v;
		dm[5][0] += w*v*v; dm[5][1] += u*w*v*v; dm[5][2] += v*w*v*v; dm[5][3] += u*u*w*v*v; dm[5][4] += u*v*w*v*v; dm[5][5] += v*v*w*v*v;
	}

	StaticMatrix<float32, 6,6> invDm;
	bool success;
	invDm = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
	if (!success) {
		for (card32 i=0; i<6; i++) {
			for (card32 j=0; j<6; j++) {
				invDm[i][j] = 0.0f;
			}
		}
	}
	coeffs = invDm * coeffs;
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		coeffsDerivs[pIndex] = invDm*coeffsDerivs[pIndex];
	}
}

void Monomials2ndOrder::computeCoeffs(
	const vector<Vector3f> &points,
	StaticVector<float32, 6> &coeffs,
	float32 gaussianWindowVariance )
{
	computeCoeffs( &points[0], (card32)points.size(), coeffs, gaussianWindowVariance );
}

void Monomials2ndOrder::computeCoeffs( const Vector3f * points, card32 numPoints, StaticVector<float32, 6> &coeffs,
float32 gaussianWindowVariance)

{
	const card32 NUM_NEIGHBORS = numPoints;
	coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f; coeffs[3] = 0.0f; coeffs[4] = 0.0f; coeffs[5] = 0.0f;
	StaticVector<float32, 6> cd;
	StaticMatrix<float32, 6,6> dm;
	for (card32 i=0; i<6;i++) {
		for (card32 j=0; j<6;j++) {
			dm[i][j] = 0.0f;
		}
	}

	float invGaussianWindowVariance = 1.0f / gaussianWindowVariance;
	float invNumNeighbors = 1.0f / NUM_NEIGHBORS;

	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		const float32 u = points[pIndex][0];
		const float32 v = points[pIndex][1];
		const float32 n = points[pIndex][2];
		const float32 w = ( (gaussianWindowVariance>0) ? sqr(exp(-(u*u+v*v)*invGaussianWindowVariance)) : 1) * invNumNeighbors;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
		cd[0] =       w;
		cd[1] = u   * w;
		cd[2] = v   * w;
		cd[3] = u*u * w;
		cd[4] = u*v * w;
		cd[5] = v*v * w;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;
		coeffs[3] += cd[3] * n;
		coeffs[4] += cd[4] * n;
		coeffs[5] += cd[5] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;     dm[0][3] += u*u*w;     dm[0][4] += u*v*w;     dm[0][5] += v*v*w;
		dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;   dm[1][3] += u*u*w*u;   dm[1][4] += u*v*w*u;   dm[1][5] += v*v*w*u;
		dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;   dm[2][3] += u*u*w*v;   dm[2][4] += u*v*w*v;   dm[2][5] += v*v*w*v;
		dm[3][0] += w*u*u; dm[3][1] += u*w*u*u; dm[3][2] += v*w*u*u; dm[3][3] += u*u*w*u*u; dm[3][4] += u*v*w*u*u; dm[3][5] += v*v*w*u*u;
		dm[4][0] += w*u*v; dm[4][1] += u*w*u*v; dm[4][2] += v*w*u*v; dm[4][3] += u*u*w*u*v; dm[4][4] += u*v*w*u*v; dm[4][5] += v*v*w*u*v;
		dm[5][0] += w*v*v; dm[5][1] += u*w*v*v; dm[5][2] += v*w*v*v; dm[5][3] += u*u*w*v*v; dm[5][4] += u*v*w*v*v; dm[5][5] += v*v*w*v*v;
	}

	StaticMatrix<float32, 6,6> invDm;
	bool success;
	invDm = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
	if (success) {
		coeffs = invDm * coeffs;
	} else {
		coeffs[0] = 0; coeffs[1] = 0; coeffs[2] = 0;
		coeffs[3] = 0; coeffs[4] = 0; coeffs[5] = 0;
	}
}

void Monomials2ndOrder::computeCoeffs( const Vector3f * points, const float * weightPtr, card32 numPoints, StaticVector<float32, 6> &coeffs )
{
	const card32 NUM_NEIGHBORS = numPoints;
	coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f; coeffs[3] = 0.0f; coeffs[4] = 0.0f; coeffs[5] = 0.0f;
	StaticVector<float32, 6> cd;
	StaticMatrix<float32, 6,6> dm;
	for (card32 i=0; i<6;i++) {
		for (card32 j=0; j<6;j++) {
			dm[i][j] = 0.0f;
		}
	}

	float32 * tmpWeightPtr = (float*)weightPtr;
	float weightMultiplier = 1.0f / (float32)NUM_NEIGHBORS;
	float32 * pointsPtr = (float*)points;
	for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
		/*const float32 u = points[pIndex][0];
		const float32 v = points[pIndex][1];
		const float32 n = points[pIndex][2];
		const float32 w = weightPtr[pIndex] / (float32)NUM_NEIGHBORS;
		*/

		const float32 u = *pointsPtr;
		pointsPtr++;
		const float32 v = *pointsPtr;
		pointsPtr++;
		const float32 n = *pointsPtr;
		pointsPtr++;
		const float32 w = *weightPtr * weightMultiplier;
		weightPtr++;

		// c[j] += b[j](uv(i)) * n(i) * w(uv(i));


		const float32 uw = u * w;
		const float32 uuw = u * uw;
		const float32 vw = v * w;
		const float32 vvw = v * vw;
		const float32 uvw = u * vw;
		const float32 uuuw = u * uuw;
		const float32 uuvw = v * uuw;
		const float32 uvvw = u * vvw;
		const float32 vvvw = v * vvw;
		const float32 uuuuw = u * uuuw;
		const float32 uuuvw = v * uuuw;
		const float32 uuvvw = u * uvvw;
		const float32 uvvvw = u * vvvw;
		const float32 vvvvw = v * vvvw;

		cd[0] =       w;
		cd[1] = uw;
		cd[2] = vw;
		cd[3] = uuw;
		cd[4] = uvw;
		cd[5] = vvw;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;
		coeffs[3] += cd[3] * n;
		coeffs[4] += cd[4] * n;
		coeffs[5] += cd[5] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += uw;     dm[0][2] += vw;     dm[0][3] += uuw;     dm[0][4] += uvw;     dm[0][5] += vvw;
		/*dm[1][0] += uw;*/   dm[1][1] += uuw;   dm[1][2] += uvw;   dm[1][3] += uuuw;   dm[1][4] += uuvw;   dm[1][5] += uvvw;
		/*dm[2][0] += vw;   dm[2][1] += uvw;*/   dm[2][2] += vvw;   dm[2][3] += uuvw;   dm[2][4] += uvvw;   dm[2][5] += vvvw;
		/*dm[3][0] += uuw; dm[3][1] += uuuw; dm[3][2] += uuvw;*/ dm[3][3] += uuuuw; dm[3][4] += uuuvw; dm[3][5] += uuvvw;
		/*dm[4][0] += uvw; dm[4][1] += u*w*u*v; dm[4][2] += v*w*u*v; dm[4][3] += u*u*w*u*v;*/ dm[4][4] += uuvvw; dm[4][5] += uvvvw;
		/*dm[5][0] += vvw; dm[5][1] += u*w*v*v; dm[5][2] += v*w*v*v; dm[5][3] += u*u*w*v*v; dm[5][4] += u*v*w*v*v;*/ dm[5][5] += vvvvw;
		
/*		cd[0] =       w;
		cd[1] = u   * w;
		cd[2] = v   * w;
		cd[3] = u*u * w;
		cd[4] = u*v * w;
		cd[5] = v*v * w;

		coeffs[0] += cd[0] * n;
		coeffs[1] += cd[1] * n;
		coeffs[2] += cd[2] * n;
		coeffs[3] += cd[3] * n;
		coeffs[4] += cd[4] * n;
		coeffs[5] += cd[5] * n;

		// 1 u v uu uv vv

		dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;     dm[0][3] += u*u*w;     dm[0][4] += u*v*w;     dm[0][5] += v*v*w;
		dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;   dm[1][3] += u*u*w*u;   dm[1][4] += u*v*w*u;   dm[1][5] += v*v*w*u;
		dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;   dm[2][3] += u*u*w*v;   dm[2][4] += u*v*w*v;   dm[2][5] += v*v*w*v;
		dm[3][0] += w*u*u; dm[3][1] += u*w*u*u; dm[3][2] += v*w*u*u; dm[3][3] += u*u*w*u*u; dm[3][4] += u*v*w*u*u; dm[3][5] += v*v*w*u*u;
		dm[4][0] += w*u*v; dm[4][1] += u*w*u*v; dm[4][2] += v*w*u*v; dm[4][3] += u*u*w*u*v; dm[4][4] += u*v*w*u*v; dm[4][5] += v*v*w*u*v;
		dm[5][0] += w*v*v; dm[5][1] += u*w*v*v; dm[5][2] += v*w*v*v; dm[5][3] += u*u*w*v*v; dm[5][4] += u*v*w*v*v; dm[5][5] += v*v*w*v*v;*/



	}

	dm[1][0] = dm[0][1];
	dm[2][0] = dm[0][2];	dm[2][1] = dm[1][2];
	dm[3][0] = dm[0][3];	dm[3][1] = dm[1][3]; dm[3][2] = dm[2][3];
	dm[4][0] = dm[0][4];	dm[4][1] = dm[1][4]; dm[4][2] = dm[2][4]; dm[4][3] = dm[3][4];
	dm[5][0] = dm[0][5];	dm[5][1] = dm[1][5]; dm[5][2] = dm[2][5]; dm[5][3] = dm[3][5]; dm[5][4] = dm[4][5];


	StaticMatrix<float32, 6,6> invDm;
	bool success;
	invDm = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
	if (success) {
		coeffs = invDm * coeffs;
	} else {
		coeffs[0] = 0; coeffs[1] = 0; coeffs[2] = 0;
		coeffs[3] = 0; coeffs[4] = 0; coeffs[5] = 0;
	}
}


float32 Monomials2ndOrder::evaluate(
   const StaticVector<float32, 6> &coeffs,
   const Vector2f &position)
{
   const float32 u = position[0];
   const float32 v = position[1];
   return   coeffs[0]
          + coeffs[1] * u
          + coeffs[2] * v
          + coeffs[3] * u*u
          + coeffs[4] * u*v
          + coeffs[5] * v*v;
}


Vector3f Monomials2ndOrder::evaluateNormal( const StaticVector<float32,6> &coeffs, 
															const Vector2f &position ) 
{
	const float u = position[0];
	const float v = position[1];
	const float derivateU = 2.0f * coeffs[3] * u + coeffs[4] * v + coeffs[1];
	const float derivateV = 2.0f * coeffs[5] * v + coeffs[4] * u + coeffs[2];
	return normalize( makeVector3f(1,0, derivateU).crossProduct( makeVector3f(0,1,derivateV) ) );
}

bool Monomials2ndOrder::evaluateCriticalPoint( const StaticVector<float32,6> &coeffs, Vector3f &criticalPoint ) 
{
	// estimate maxima/minima
	// see "Tessellation of Quadratic Elements" Dillard et al.
	const float A = coeffs[5];
	const float B = coeffs[4];
	const float C = coeffs[3];
	const float D = coeffs[2];
	const float E = coeffs[1];
	float detH = 4.0f * A * C - B*B;
	if( detH > 0 )
	{
		float cu = (-2.0f * C * D + B * E) / detH;
		float cv = (-2.0f * A * E + B * D) / detH;
		criticalPoint = makeVector3f(cu,cv,Monomials2ndOrder::evaluate( coeffs, makeVector2f(cu,cv)));
		return true;
	}
	else
	{
		return false;
	}

}


void geoxl::Monomials2ndOrder_unitTest() {
/*   Matrix3f A = makeMatrix3f(
      1,0,0,
      0,1,0,
      0,0,1);
   Vector3f b = makeVector3f(0,0,0);	*/

   Matrix3f A = makeMatrix3f(
      1,2,3,
      4,5,6,
      7,8,9);
   Vector3f b = makeVector3f(10,11,12);

   vector<AffineMapping3dFit::MapPoint> testData;
   for (card32 i=0; i<100; i++) {
		float32 a = rnd01();
      Vector3f testPoint = makeVector3f(a, a, 0);
      AffineMapping3dFit::MapPoint m;
      m.from = testPoint;
      m.to = A*testPoint + b;
      testData.push_back(m);
   }
      
   AffineMapping3dFit::AffineMap leastSquareFit;
   vector< AffineMapping3dFit::AffineMap > coeffsDerivs;
   AffineMapping3dFit::computeCoeffs( 
      testData,
      leastSquareFit,
      coeffsDerivs
   );
   debugOutput << leastSquareFit.M.toString() << "\n"
               << leastSquareFit.b.toString() << "\n";
}




//void Monomials2ndOrder3D::computeInvDesignMatrix( 
//   const vector<Vector4f> &points,
//   StaticMatrix<float32, 10, 10> &invDM)
//{
//	const card32 numPts = points.size();
//	StaticMatrix<float32, 10,10> dm;
//	for (card32 i=0; i<10;i++) {
//		for (card32 j=0; j<10;j++) {
//			dm[i][j] = 0.0f;
//		}
//	}
//
//	for (card32 pIndex=0; pIndex<numPts; pIndex++){
//		const float32 u = points[pIndex][0];
//		const float32 v = points[pIndex][1];
//		const float32 w = points[pIndex][2];
//
//		// 1 u v uu uv vv
//
//		dm[0][0] += 1;     dm[0][1] += u;       dm[0][2] += v;       dm[0][3] += u*u;       dm[0][4] += u*v;       dm[0][5] += v*v;       dm[0][6] += w;       dm[0][7] += w*u;       dm[0][8] += w*v;        dm[0][9] += w*w; 
//		dm[1][0] +=   u;   dm[1][1] += u * u;   dm[1][2] += v * u;   dm[1][3] += u*u * u;   dm[1][4] += u*v * u;   dm[1][5] += v*v * u;   dm[1][6] += w * u;   dm[1][7] += w*u * u;   dm[1][8] += w*v * u;    dm[1][9] += w*w * u;
//		dm[2][0] +=   v;   dm[2][1] += u * v;   dm[2][2] += v * v;   dm[2][3] += u*u * v;   dm[2][4] += u*v * v;   dm[2][5] += v*v * v;   dm[2][6] += w * v;   dm[2][7] += w*u * v;   dm[2][8] += w*v * v;    dm[2][9] += w*w * v;
//		dm[3][0] +=   u*u; dm[3][1] += u * u*u; dm[3][2] += v * u*u; dm[3][3] += u*u * u*u; dm[3][4] += u*v * u*u; dm[3][5] += v*v * u*u; dm[3][6] += w * u*u; dm[3][7] += w*u * u*u; dm[3][8] += w*v * u*u;  dm[3][9] += w*w * u*u;
//		dm[4][0] +=   u*v; dm[4][1] += u * u*v; dm[4][2] += v * u*v; dm[4][3] += u*u * u*v; dm[4][4] += u*v * u*v; dm[4][5] += v*v * u*v; dm[4][6] += w * u*v; dm[4][7] += w*u * u*v; dm[4][8] += w*v * u*v;  dm[4][9] += w*w * u*v;
//		dm[5][0] +=   v*v; dm[5][1] += u * v*v; dm[5][2] += v * v*v; dm[5][3] += u*u * v*v; dm[5][4] += u*v * v*v; dm[5][5] += v*v * v*v; dm[5][6] += w * v*v; dm[5][7] += w*u * v*v; dm[5][8] += w*v * v*v;  dm[5][9] += w*w * v*v;
//		dm[6][0] +=   w;   dm[6][1] += u * w;   dm[6][2] += v * w;   dm[6][3] += u*u * w;   dm[6][4] += u*v * w;   dm[6][5] += v*v * w;   dm[6][6] += w * w;   dm[6][7] += w*u * w;   dm[6][8] += w*v * w;    dm[6][9] += w*w * w;  
//		dm[7][0] +=   w*u; dm[7][1] += u * w*u; dm[7][2] += v * w*u; dm[7][3] += u*u * w*u; dm[7][4] += u*v * w*u; dm[7][5] += v*v * w*u; dm[7][6] += w * w*u; dm[7][7] += w*u * w*u; dm[7][8] += w*v * w*u;  dm[7][9] += w*w * w*u;
//		dm[8][0] +=   w*v; dm[8][1] += u * w*v; dm[8][2] += v * w*v; dm[8][3] += u*u * w*v; dm[8][4] += u*v * w*v; dm[8][5] += v*v * w*v; dm[8][6] += w * w*v; dm[8][7] += w*u * w*v; dm[8][8] += w*v * w*v;  dm[8][9] += w*w * w*v;
//		dm[9][0] +=   w*w; dm[9][1] += u * w*w; dm[9][2] += v * w*w; dm[9][3] += u*u * w*w; dm[9][4] += u*v * w*w; dm[9][5] += v*v * w*w; dm[9][6] += w * w*w; dm[9][7] += w*u * w*w; dm[9][8] += w*v * w*w;  dm[9][9] += w*w * w*w;
//	}
//
//   bool success;
//	invDM = invertMatrix(dm, &success, INV_PIVOT_THRESHOLD);
//}

//void Monomials2ndOrder3D::computeCoeffs(
//	const vector<Vector4f> &points,
//	const StaticMatrix<float32, 10, 10> &invDM,
//	StaticVector<float32, 10> &coeffs)
//{
//	const card32 numPts = points.size();
//	coeffs[0] = 0.0f;
//	coeffs[1] = 0.0f;
//	coeffs[2] = 0.0f;
//	coeffs[3] = 0.0f;
//	coeffs[4] = 0.0f;
//	coeffs[5] = 0.0f;
//	coeffs[6] = 0.0f;
//	coeffs[7] = 0.0f;
//	coeffs[8] = 0.0f;
//	coeffs[9] = 0.0f;
//	StaticVector<float32, 10> cd;
//
//	for (card32 pIndex=0; pIndex<numPts; pIndex++){
//		const float32 u = points[pIndex][0];
//		const float32 v = points[pIndex][1];
//		const float32 w = points[pIndex][2];
//		const float32 n = points[pIndex][3];
//
//		// c[j] += b[j](uv(i)) * n(i)w);
//		cd[0] = 1  ;
//		cd[1] = u  ;
//		cd[2] = v  ;
//		cd[3] = u*u;
//		cd[4] = u*v;
//		cd[5] = v*v;
//		cd[6] = w;
//		cd[7] = w*u;
//		cd[8] = w*v;
//		cd[9] = w*w;
//
//		coeffs[0] += cd[0] * n;
//		coeffs[1] += cd[1] * n;
//		coeffs[2] += cd[2] * n;
//		coeffs[3] += cd[3] * n;
//		coeffs[4] += cd[4] * n;
//		coeffs[5] += cd[5] * n;
//		coeffs[6] += cd[6] * n;
//		coeffs[7] += cd[7] * n;
//		coeffs[8] += cd[8] * n;
//		coeffs[9] += cd[9] * n;
//
//	}
//
//	coeffs = invDM * coeffs;
//}
//
//float32 Monomials2ndOrder3D::evaluate(
//   const StaticVector<float32, 10> &coeffs,
//   const Vector3f &position)
//{
//   const float32 u = position[0];
//   const float32 v = position[1];
//   const float32 w = position[2];
//   return   coeffs[0]
//          + coeffs[1] * u
//          + coeffs[2] * v
//          + coeffs[3] * u*u
//          + coeffs[4] * u*v
//          + coeffs[5] * v*v
//          + coeffs[6] * w
//          + coeffs[7] * w*u
//          + coeffs[8] * w*v
//          + coeffs[9] * w*w;
//}


void geoxl::Monomials2ndOrder::computeCurvatures( const StaticVector<float32,6> &coeff, float &k1, float &k2, Vector2f &k1_dir, Vector2f &k2_dir )
{
	const float f_x = coeff[1];
	const float f_y = coeff[2];
	const float f_xy = coeff[4];
	const float f_xx = 2.0f * coeff[3];
	const float f_yy = 2.0f * coeff[5];

// 	float gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
// 		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );
// 
// 	float meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
// 		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;

	// compute principal frame from patch
	Matrix2f M = makeMatrix2f(f_xx, f_xy, f_xy, f_yy);
	Matrix2f eigenVectors2f;
	Vector2f eigenValues2f;
	M.computeEigenStructure(eigenValues2f,eigenVectors2f);
	eigenVectors2f = eigenVectors2f;
	k1_dir = eigenVectors2f[0];
	k2_dir = eigenVectors2f[1];
	k1 = eigenValues2f[0];
	k2 = eigenValues2f[1];
	if( fabs(k2) > fabs(k1) )
	{
		swap(k1_dir,k2_dir);
		swap(k2,k1);
	}
}

geoxl::Vector3f geoxl::Monomials2ndOrder1D::computeCoeffs( Vector2f * dataPoints , float * weights, const unsigned numPoints )
{
	// compute gradient of squared difference function & solve grad = 0
	Matrix3f mat = IDENTITY3F * 0;
	Vector3f v = NULL_VECTOR3F;

	for( unsigned i=0;i<numPoints;i++ )
	{
		const float x = dataPoints[i][0];
		const float y = dataPoints[i][1];
		const float weight = weights[i];

		const float x2 = x*x;
		const float x3 = x2*x;
		const float x4 = x3*x;
		const float x5 = x4*x;


		v[0] += y * weight;
		v[1] += y * weight * x;
		v[2] += y * weight * x2;


		mat[0][0] += weight;	mat[0][1] += weight*x ;		mat[0][2] += weight*x2;		
		mat[1][0] += weight*x;	mat[1][1] += weight*x2;		mat[1][2] += weight*x3;		
		mat[2][0] += weight*x2;	mat[2][1] += weight*x3;		mat[2][2] += weight*x4;		

	}

	Matrix3f invDm;
	bool success;
	invDm = invertMatrix(mat, &success, 1e-7f );
	if (!success) 
	{
		return makeVector3f(0,0,0);
	}
	return invDm * v;
}