#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SpectralCorrespondenceValidation.h"
//---------------------------------------------------------------------------
#include "SparseLinearAlgebra_templCode.h"
#include "DynamicLinearAlgebra.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( SpectralCorrespondenceValidation ,3)
{
	BEGIN_CLASS_INIT( SpectralCorrespondenceValidation );

	ADD_FLOAT32_PROP(m_ScaleDiff,0);
	ADD_FLOAT32_PROP(m_AddDiff,0);
	ADD_FLOAT32_PROP(m_Sigma,0);
	ADD_FLOAT32_PROP(m_MaxDistance,0);
	ADD_FLOAT32_PROP(m_UseEuclidRadius,0);
	ADD_FLOAT32_PROP(m_UseGeodeticRadius,0);
	ADD_FLOAT32_PROP(m_DistanceSimilarityFactor,0);
	ADD_FLOAT32_PROP(m_MinConfidence,0);
	ADD_CARD32_PROP(m_NumPowerIterations,0);
	ADD_FLOAT32_PROP(m_MaxNormalAngleDiff,1);
	ADD_FLOAT32_PROP(m_NormalAngleSigma,1);
	ADD_BOOLEAN_PROP(m_Deformable,2);
	ADD_FLOAT32_PROP(m_DistanceTolerance,3);
}

SpectralCorrespondenceValidation::SpectralCorrespondenceValidation(void)
{
	m_MaxNormalAngleDiff = 15.0f;
	m_NormalAngleSigma = 5.0f;
	m_ScaleDiff = 0.0f;
	m_AddDiff = 0.1f;
	m_Sigma = 0.1f;
	m_MaxDistance = 1e10f;
//	m_RegistrationContext = NULL;
	m_UseEuclidRadius = 0.2f;
	m_UseGeodeticRadius = 0.4f;
	m_DistanceSimilarityFactor = 3.0f;
	m_MinConfidence = 1e-10f;
	m_NumPowerIterations = 1;
	m_Deformable = false;
	m_DistanceTolerance = 0.1f;
}

SpectralCorrespondenceValidation::~SpectralCorrespondenceValidation(void)
{
}

void SpectralCorrespondenceValidation::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const SpectralCorrespondenceValidation * o = dynamic_cast<const SpectralCorrespondenceValidation*>(obj);
	pAssert( o != NULL );

	m_Matrix = o->m_Matrix;
	m_ScaleDiff = o->m_ScaleDiff;
	m_AddDiff = o->m_AddDiff;
	m_DistanceMatrix1 = o->m_DistanceMatrix1;
	m_DistanceMatrix2 = o->m_DistanceMatrix2;
	m_EuclidDistanceMatrix1 = o->m_EuclidDistanceMatrix1;
	m_EuclidDistanceMatrix2 = o->m_EuclidDistanceMatrix2;
	m_Sigma = o->m_Sigma;
	m_MaxDistance = o->m_MaxDistance;
}


void SpectralCorrespondenceValidation::setup( card32 numberOfPairs, card32 numberOfFeatures1, card32 numberOfFeatures2, card32 numPowerIterations ) 
{
	m_Matrix.clear();
	m_Matrix.setRows( numberOfPairs );
	m_DistanceMatrix1.setDimension( numberOfFeatures1, numberOfFeatures1,false);
	m_DistanceMatrix1.setZero();
	m_DistanceMatrix2.setDimension( numberOfFeatures2, numberOfFeatures2,false);
	m_DistanceMatrix2.setZero();
	m_NormalAngle1.setDimension( numberOfFeatures1, numberOfFeatures1,false);
	m_NormalAngle1.setZero();
	m_NormalAngle2.setDimension( numberOfFeatures2, numberOfFeatures2,false);
	m_NormalAngle2.setZero();
	m_EuclidDistanceMatrix1 = m_DistanceMatrix1;
	m_EuclidDistanceMatrix2 = m_DistanceMatrix2;
	m_NumPowerIterations = numPowerIterations;
	TFeaturePair p;
	p.i1 = 0;
	p.i2 = 0;
	m_Pairs.resize(numberOfPairs, p);
}

void SpectralCorrespondenceValidation::setMatrixEntry( card32 x, card32 y, float value ) 
{
	if( value != 0.0f )
	{
		if( x >= m_Pairs.size() || y >= m_Pairs.size() )
			throw "muh!";
		m_Matrix[x].setEntry( y, value );
		m_Matrix[y].setEntry( x, value );
	}
}


void SpectralCorrespondenceValidation::setDistanceEntry1( card32 index1, card32 index2, float geo_distance, float euclid_distance  ) 
{
	m_DistanceMatrix1[index1][index2] = geo_distance;
	m_DistanceMatrix1[index2][index1] = geo_distance;
	m_EuclidDistanceMatrix1[index1][index2] = euclid_distance;
	m_EuclidDistanceMatrix1[index2][index1] = euclid_distance;
}

void SpectralCorrespondenceValidation::setDistanceEntry2( card32 index1, card32 index2, float geo_distance, float euclid_distance  ) 
{
	m_DistanceMatrix2[index1][index2] = geo_distance;
	m_DistanceMatrix2[index2][index1] = geo_distance;
	m_EuclidDistanceMatrix2[index1][index2] = euclid_distance;
	m_EuclidDistanceMatrix2[index2][index1] = euclid_distance;
}

float32 SpectralCorrespondenceValidation::calculate() 
{
	m_Solution.clear();
	list<int> L;
	vector<int> solutionIndices;
	for( card32 i=0;i<m_Pairs.size();i++ )
		if( m_Pairs[i].valid )
		{
			m_Solution.push_back( m_Pairs[i] );
			solutionIndices.push_back( i );
		}

	for( card32 i=0;i<m_Pairs.size();i++ )
		if( !m_Pairs[i].valid )
		{
			bool conflicting = false;
			for( card32 j=0;j<solutionIndices.size();j++ )
			{
				conflicting |= conflict( solutionIndices[j], i );
				if( conflicting )
					break;
			}
			if( !conflicting )
				L.push_back( i );
			else
				m_Pairs[i].conflicting = true;

		}

	if( m_Pairs.size() < 2 )
		return 0;
	// 

	// calculate confidence-matrix
	float maxNormalDiffAngle = cos(m_MaxNormalAngleDiff / 180.0f * M_PI);
	float sigma_normalAngle = cos(m_NormalAngleSigma / 180.0f * M_PI);
	float sigma_2 = m_Sigma * m_Sigma;
	card32 numNonZeroEntries = 0;
	for( card32 i=0;i<m_Pairs.size();i++ )
	{
		//setMatrixEntry(i,i, m_Pairs[i].confidence );
		for( card32 j=i+1;j<m_Pairs.size();j++ )
		{
			if( m_Pairs[i].conflicting || m_Pairs[j].conflicting )
				continue;

			float distance1, distance2;
			getDistances( i,j, distance1, distance2 );

			float normalAngle1 = m_NormalAngle1[m_Pairs[i].i1][m_Pairs[j].i1];
			float normalAngle2 = m_NormalAngle2[m_Pairs[i].i2][m_Pairs[j].i2];

			float conf = 0.0f;
			float sigma_d_2 = sigma_2;
			if( m_Deformable )
			{
				if( distance1 < m_MaxDistance && distance2 < m_MaxDistance )
				{
					distance1 = max(0.0f,distance1 - m_DistanceTolerance);
					distance2 = max(0.0f,distance2 - m_DistanceTolerance);
					float diff = fabs(distance1 - distance2);

					conf +=  max<float>(0.0f,4.5f - pow(diff,2.0f) / 2.0f / sigma_2);
				}

			}
			else
			{
				if( distance1 < m_MaxDistance && distance2 < m_MaxDistance 
					&& fabs(normalAngle1 - normalAngle2) < maxNormalDiffAngle )
				{
					float diff = fabs(distance1 - distance2);


					conf +=  max<float>(0.0f,4.5f - pow(diff,2.0f) / 2.0f / sigma_2)
						* max<float>(0.0f,4.5f - pow(normalAngle1 - normalAngle2,2.0f) / 2.0f / sigma_normalAngle / sigma_normalAngle);
				}
			}

			if( conf > 1e-15f )	// keep matrix sparse
			{
				numNonZeroEntries++;
				setMatrixEntry(i,j, conf );
			}
		}
	}

	//size_t matDim = m_Pairs.size();

	// calculate first eigenvector
	DynamicVector<float> eigenVector;
	float eigenValue = 0;
	vector<DynamicVector<float> > excludedDirections;
	
	for( card32 i=0;i<m_NumPowerIterations;i++ )
	{
		DynamicVector<float> tmpEigenVector;
		float tmpEigenValue;
		powerIteration<float, SparseMatrixF>( m_Matrix, excludedDirections, tmpEigenVector, tmpEigenValue,1.0001f, 1000 );
		if( fabs( tmpEigenValue ) > fabs(eigenValue) || i == 0 )
		{
			eigenVector = tmpEigenVector;
			eigenValue = tmpEigenValue;
		}
	}

	debugOutput << "eigenVector:" << eigenVector << "\n";

	if( eigenVector[0] < 0 )
		eigenVector = -eigenVector;


	while(true)
	{
		if( L.empty() )
			break; // finished

		// find argmax(eigenVector[0][a]) with a out of L and delete a from L
		float maxV = -1;
		card32 a = 0;
		list<int>::iterator iterA = L.begin();
		for( list<int>::iterator iter = L.begin(); iter != L.end(); ++iter )
		{
			if( maxV < eigenVector[*iter])
			{
				iterA = iter;
				maxV = eigenVector[*iter];
				a = *iter;
			}
		}
		if( maxV <= m_MinConfidence )
			break; // finished 

		L.erase( iterA );
		m_Solution.push_back( m_Pairs[a] );

		// remove every pair in conflict with pair a
		for( list<int>::iterator iter = L.begin(); iter != L.end(); )
		{
			if( conflict(*iter,a) )
				iter = L.erase( iter );
			else
				++iter;
		}
		

	}

	return eigenValue;
}

void SpectralCorrespondenceValidation::setCorrespondencePair( card32 index, card32 i1, card32 i2, card32 orgI1, card32 orgI2, float confidence,const Vector3f &pos1, const Vector3f &pos2, bool valid) 
{
	pAssert( index < m_Pairs.size());
	m_Pairs[index].i1 = i1;
	m_Pairs[index].i2 = i2;
	m_Pairs[index].orgIndex1 = orgI1;
	m_Pairs[index].orgIndex2 = orgI2;
	m_Pairs[index].confidence = confidence;
	m_Pairs[index].p1 = pos1;
	m_Pairs[index].p2 = pos2;
	m_Pairs[index].valid = valid;
	m_Pairs[index].conflicting =false;
}

bool SpectralCorrespondenceValidation::conflict( card32 pair1, card32 pair2 ) 
{
	if( m_Pairs[pair1].i1 == m_Pairs[pair2].i1 || m_Pairs[pair1].i2 == m_Pairs[pair2].i2 )
		return true;

	
/*	Vector3f dir1 = m_Pairs[pair1].p2 - m_Pairs[pair1].p1;
	Vector3f dir2 = m_Pairs[pair2].p2 - m_Pairs[pair2].p1;

	float minMovement = 0.05f;
	if( norm(dir1) > minMovement && norm(dir2) > minMovement )
	{
		if( norm( m_Pairs[pair1].p1 - m_Pairs[pair2].p1) < 0.1f || 
			norm( m_Pairs[pair1].p2 - m_Pairs[pair2].p2) < 0.1f)
			if( normalize( dir1 ) * normalize( dir2 ) < cos( 60.0f / 180.0f * M_PI ) )
				return true;
	}
*/
	
	float d1, d2;
	getDistances( pair1, pair2, d1,d2);

	if( d1 >= m_MaxDistance || d2 >= m_MaxDistance )
	{
		return false;//fabs(d1-d2) > m_AddDiff && min(d1,d2) < m_Sigma * 2.0f;
	}

	if( !m_Deformable)
	{
		float normalAngle1 = m_NormalAngle1[m_Pairs[pair1].i1][m_Pairs[pair2].i1];
		float normalAngle2 = m_NormalAngle2[m_Pairs[pair1].i2][m_Pairs[pair2].i2];

		float maxNormalDiffAngle = cos(m_MaxNormalAngleDiff / 180.0f * M_PI);
		if( fabs(normalAngle1-normalAngle2) > maxNormalDiffAngle )
			return true;
	}

	return fabs(d1-d2) > min(d1,d2) * m_ScaleDiff + m_AddDiff;
}

void SpectralCorrespondenceValidation::getDistances( card32 pair1, card32 pair2, float &distance1, float &distance2) 
{
	float geo_dist1 = m_DistanceMatrix1[m_Pairs[pair1].i1][m_Pairs[pair2].i1];
	float geo_dist2 = m_DistanceMatrix2[m_Pairs[pair1].i2][m_Pairs[pair2].i2];

	float euclid_dist1 = m_EuclidDistanceMatrix1[m_Pairs[pair1].i1][m_Pairs[pair2].i1];
	float euclid_dist2 = m_EuclidDistanceMatrix2[m_Pairs[pair1].i2][m_Pairs[pair2].i2];

	distance1 = geo_dist1;
	distance2 = geo_dist2;

	if( geo_dist1 > m_MaxDistance || geo_dist2 > m_MaxDistance)
		return;

	if( euclid_dist1 * m_DistanceSimilarityFactor >= geo_dist1 && 
		euclid_dist2 * m_DistanceSimilarityFactor >= geo_dist2)
	{
		if( euclid_dist1 < m_UseEuclidRadius )
			distance1 = euclid_dist1;
		else
			if( euclid_dist1 < m_UseGeodeticRadius )
			{
				float t = (euclid_dist1 - m_UseEuclidRadius) / (m_UseGeodeticRadius-m_UseEuclidRadius);
				distance1 = geo_dist1 * t + euclid_dist1 * (1.0f - t);
			}
		if( euclid_dist2 < m_UseEuclidRadius )
			distance2 = euclid_dist2;
		else
			if( euclid_dist2 < m_UseGeodeticRadius )
			{
				float t = (euclid_dist2 - m_UseEuclidRadius) / (m_UseGeodeticRadius-m_UseEuclidRadius);
				distance2 = geo_dist2 * t + euclid_dist2 * (2.0f - t);
			}

	}

}

struct TRemainingElement
{
	card32 index;
	card32 numConflicts;
};

void SpectralCorrespondenceValidation::forwardSearch() 
{
	m_Solution.clear();
	vector<card32> validatedIndices,solutionIndices;
	list<int> L;
	for( card32 i=0;i<m_Pairs.size();i++ )
	{
		if( m_Pairs[i].valid )
		{
			m_Solution.push_back(m_Pairs[i]);
			validatedIndices.push_back(i);
		}
		else
			L.push_back(i);
	}

	while(true)
	{
		if( L.empty() )
			break; // finished

		// find argmax(confidence) with a out of L and delete a from L
		float maxV = -1;
		card32 a = 0;
		list<int>::iterator iterA = L.begin();
		for( list<int>::iterator iter = L.begin(); iter != L.end(); ++iter )
		{
			if( maxV < m_Pairs[*iter].confidence )
			{
				iterA = iter;
				maxV = m_Pairs[*iter].confidence ;
				a = *iter;
			}
		}
		if( maxV == 0.0f )
			break; // finished 

		L.erase( iterA );

		// test for conflict with other solution pairs
		bool goodCorr = true;
		for( card32 i=0;i<validatedIndices.size();i++ )
		{
			if( conflict( validatedIndices[i], a ) )
			{
				goodCorr = false;
				break;
			}


		}
		if( goodCorr )
		{
			//m_Solution.push_back( m_Pairs[a] );
			solutionIndices.push_back( a );
		}


	}


	// count conflicts
	vector<int> numConflicts( solutionIndices.size() ,0 );
	list<TRemainingElement> remainingList;
	for( card32 i=0;i<solutionIndices.size();i++ )
	{
		for( card32 j=i+1;j<solutionIndices.size();j++ )
		{
			if( conflict( solutionIndices[i],solutionIndices[j] ) )
			{
				numConflicts[ i ]++;
				numConflicts[ j ]++;
			}
		}
		if( numConflicts[i] == 0 )
			m_Solution.push_back( m_Pairs[solutionIndices[i]] );
		else
		{
			TRemainingElement el;
			el.index = solutionIndices[i];
			el.numConflicts = numConflicts[i];
			remainingList.push_back( el );
		}
	}

	while( !remainingList.empty() )
	{
		int minConflicts = 100000;
		list<TRemainingElement>::iterator bestIter = remainingList.begin();

		for( list<TRemainingElement>::iterator iter = remainingList.begin(); iter != remainingList.end(); ++iter )
		{
			if( (int)iter->numConflicts < minConflicts )
			{
				minConflicts = iter->numConflicts;
				bestIter = iter;
			}
		}

		card32 bestIndex = bestIter->index;
		m_Solution.push_back( m_Pairs[bestIndex] );
		remainingList.erase( bestIter );

		// remove every pair in conflict with selected pair 
		for( list<TRemainingElement>::iterator iter = remainingList.begin(); iter != remainingList.end();  )
		{
			if( conflict(iter->index,bestIndex) )
				iter = remainingList.erase( iter );
			else
				++iter;
		}



	}

		
}

void SpectralCorrespondenceValidation::setNormalAngle1( card32 index1, card32 index2, float diff_angle) 
{
	m_NormalAngle1[index1][index2] = diff_angle;
	m_NormalAngle1[index2][index1] = diff_angle;
}

void SpectralCorrespondenceValidation::setNormalAngle2( card32 index1, card32 index2, float diff_angle) 
{
	m_NormalAngle2[index1][index2] = diff_angle;
	m_NormalAngle2[index2][index1] = diff_angle;
}
