#include "stdafx.h"
//---------------------------------------------------------------------------
#include "TensorVoting.h"
#include "DebugOutput.h"
#include "MathTools.h"
#include "HierarchicalKNNIterator.h"
#include "ProgressWindow.h"
#include "PointSetKNNQuery.h"
#include "PointSet_templCode.h"
#include <time.h>
#include "NumericalClassProperty.h"
#include "CachedRangeQuerry.h"
#include "Timer.h"
#include "StringHelper.h"
#include "Random.h"
#include "GridBasedSphericalRangeQuery.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( TensorVoting, 0 )
{
	BEGIN_CLASS_INIT( TensorVoting )
}

TensorVoting::TensorVoting()
{
}		

TensorVoting::~TensorVoting()
{
}

Matrix3f TensorVoting::computeTensorVote( 
	const Vector3f& p, 
	const Vector3f& np,
	const float32 epsilon )
{
	const Vector3f l = p - np;
	const float32 norm_l = norm( l );
	if ( norm_l < 1e-7 ) return makeMatrix3f( 0,0,0,0,0,0,0,0,0 );
	const Vector3f t = l / norm_l;
	const float32 weight = exp( - sqr( norm_l ) / sqr( epsilon ) );
	return makeMatrix3f( 
		1-sqr(t[0]), -t[0]*t[1],  -t[0]*t[2], 
		-t[0]*t[1],  1-sqr(t[1]), -t[1]*t[2],
		-t[0]*t[2],  -t[1]*t[2],  1-sqr(t[2])
		) * weight;
}

bool TensorVoting::computeTensorVote(
	Matrix3f& vote,
	const Matrix3f& receiverEigenVectors,
	const Vector3f& receiverPos,
	const Matrix3f& senderTensor,
	const Vector3f& senderPos,
	const float32 epsilon )
{
	vote = makeMatrix3f( 0,0,0, 0,0,0, 0,0,0 );
	Matrix3f senderEigenVectors;
	Vector3f senderEigenValues;
	senderTensor.computeEigenStructure( senderEigenValues, senderEigenVectors );

	Vector3f l = senderPos - receiverPos;
	float32 norm_l = norm( l );
	if ( norm_l < 1e-8 ) return false;
	l /= norm_l;
	float32 theta = receiverEigenVectors[0] * l;
	theta = acos( ( theta < 0 )? -theta : theta );
	theta = M_PI / 2.0f - theta;
	float32 theta_deg = theta * 180.0f / M_PI;
	if ( theta_deg > 45 ) return false;
	
	float32 s = norm_l * theta / sin( theta );
	float32 kappa = 2.0f * sin( theta ) / norm_l;
	float32 c = 1;
	float32 weight = exp( - ( sqr( s ) + c * sqr( kappa ) ) / sqr( epsilon ) );

	weight = 1;
	
	vote = receiverEigenVectors.transpose() * senderEigenVectors * senderTensor * weight;
	return true;
}

void TensorVoting::tensorVote( 
	bool initTensors,
	PointCloud* pc, 
	const float32 epsilon,
	const bool saveTensors,
	const card32 numKNN,
	const float32 timesSamplingSpacing,
	const card32 mode )
{
	// ---------------------------
	// MB - introduced debug flag
	bool debugOutputFlush = false;	// flush debugoutput every 5%

	if ( pc == NULL ) {
		debugOutput << "initStickTensors: invalid point cloud.\n"; debugOutput.flush();
		return;
	}
	if ( !pc->providesAttribute( "position" ) ){
		debugOutput << "initStickTensors: 'position' channel required.\n"; debugOutput.flush();
		return;
	}
	if ( !pc->providesAttribute( "surfaceness" ) ){
		debugOutput << "initStickTensors: 'surfaceness' channel required.\n"; debugOutput.flush();
		return;
	}
	AAT POSITION = pc->getAAT( "position" );
	AAT SURFACENESS = pc->getAAT( "surfaceness" );
	AAT TENSOR_0, TENSOR_1;
	AAT SAMPLING_SPACING;

	if ( saveTensors || !initTensors ){
		if ( !pc->providesAttribute( "tensor_0" ) || !pc->providesAttribute( "tensor_1" ) ){
			debugOutput << "initStickTensors: 'tensor_0/tensor_1' channel required.\n"; debugOutput.flush();
			return;
		}
		TENSOR_0 = pc->getAAT( "tensor_0" );
		TENSOR_1 = pc->getAAT( "tensor_1" );
	}

	if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ){
		if ( !pc->providesAttribute( "samplingSpacing" ) ){
			debugOutput << "initStickTensors: channel 'samplingSpacing' required.\n"; debugOutput.flush();
			return;
		}
		SAMPLING_SPACING = pc->getAAT( "samplingSpacing" );
	}

	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
	if ( modIt == NULL ){
		debugOutput << "Cannot create modifyable iterator.\n"; debugOutput.flush();
		return;
	}
	HierarchicalKNNIterator* knnIt = new HierarchicalKNNIterator( pc, 32, NULL );
	float32 sqrEpsilon = sqr( epsilon );
	Matrix3f eigenVectors;
	Vector3f eigenValues;

	float32 startTime = clock();

	progressWindow->pushStep( true, "Tensor Voting" );

	basicIt->reset();
	long pointIndex = 0;
	long numOutliers = 0;
	float32 avgNumNeighbors = 0;
	card32 nextPercentage = 5;
	while( !basicIt->atEnd() ){
		const Vector3f& p = basicIt->get3f( POSITION );
		knnIt->setSeekPointAndReset( p );
		Vector3f np = knnIt->get3f( POSITION );
		float32 sqrD = ( np - p ) * ( np - p );
		Matrix3f tensor = makeMatrix3f( 0,0,0,0,0,0,0,0,0 );
		Matrix3f receiverEigenVectors;

		// Extract orientation of tensor at point
		if ( !initTensors ){
			Vector3f receiver_tensor_0 = basicIt->get3f( TENSOR_0 );
			Vector3f receiver_tensor_1 = basicIt->get3f( TENSOR_1 );
			Matrix3f receiverTensor;
			receiverTensor[0][0] = receiver_tensor_0[0];
			receiverTensor[0][1] = receiver_tensor_0[1];
			receiverTensor[0][2] = receiver_tensor_0[2];
			receiverTensor[1][1] = receiver_tensor_1[0];
			receiverTensor[1][2] = receiver_tensor_1[1];
			receiverTensor[2][2] = receiver_tensor_1[2];
			receiverTensor[1][0] = receiverTensor[0][1];
			receiverTensor[2][0] = receiverTensor[0][2];
			receiverTensor[2][1] = receiverTensor[1][2];
			Vector3f receiverEigenValues;
			receiverTensor.computeEigenStructure( receiverEigenValues, receiverEigenVectors );
		}

		if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
			float32 samplingSpacing = basicIt->get1f( SAMPLING_SPACING );
			sqrEpsilon = sqr( timesSamplingSpacing * samplingSpacing );
		}

		card32 numNeighbors = 0;
		bool converged = false;
		while( !converged ){

			if ( initTensors ){
				tensor += computeTensorVote( p, np, epsilon );
				numNeighbors++;
			} else {
				// Extract senderTensor
				Vector3f sender_tensor_0 = knnIt->get3f( TENSOR_0 );
				Vector3f sender_tensor_1 = knnIt->get3f( TENSOR_1 );
				Matrix3f senderTensor;
				senderTensor[0][0] = sender_tensor_0[0];
				senderTensor[0][1] = sender_tensor_0[1];
				senderTensor[0][2] = sender_tensor_0[2];
				senderTensor[1][1] = sender_tensor_1[0];
				senderTensor[1][2] = sender_tensor_1[1];
				senderTensor[2][2] = sender_tensor_1[2];
				senderTensor[1][0] = senderTensor[0][1];
				senderTensor[2][0] = senderTensor[0][2];
				senderTensor[2][1] = senderTensor[1][2];
				Matrix3f vote;
				if ( computeTensorVote( vote, receiverEigenVectors, p, senderTensor, np, epsilon ) ){
					tensor += vote;
					numNeighbors++;
				}
			}

			np = knnIt->get3f( POSITION );
			sqrD = ( np - p ) * ( np - p );
			knnIt->next();
			

			if ( mode == ENVIRONMENT_KNN ){ 
				if ( numNeighbors >= numKNN ) 
					converged = true;
			} else 
				if ( sqrD > sqrEpsilon ) 
					converged = true;
		}
		//if ( numNeighbors > 0 ) tensor /= float32( numNeighbors );
		avgNumNeighbors = ( avgNumNeighbors * pointIndex + numNeighbors ) / float32( pointIndex + 1 );
		tensor.computeEigenStructure( eigenValues, eigenVectors );
		const float32 surfaceness = eigenValues[0] - eigenValues[1];
		modIt->set1f( SURFACENESS, surfaceness );
		if ( saveTensors ){
			Vector3f tv0, tv1;
			tv0[0] = tensor[0][0];
			tv0[1] = tensor[0][1];
			tv0[2] = tensor[0][2];
			tv1[0] = tensor[1][1];
			tv1[1] = tensor[1][2];
			tv1[2] = tensor[2][2];
			modIt->set3f( TENSOR_0, tv0 );
			modIt->set3f( TENSOR_1, tv1 );
		}
		basicIt->next();
		pointIndex++;
		card32 percentage = card32( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 );
		if( pointIndex % 1000 == 0)
			progressWindow->progress( percentage );

		if( debugOutputFlush )
		{
			if ( percentage >= nextPercentage ){
				debugOutput << "Tensor Voting: " << percentage << "%.\n"; debugOutput.flush();
				nextPercentage += 5;
			}	
		}
	}
	progressWindow->popSteps();
	delete it;
	delete knnIt;

	debugOutput << "Time required for tensor voting: " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush();
	debugOutput << avgNumNeighbors << " neighbors visited on average.\n"; debugOutput.flush();
}

void TensorVoting::tensorVote( 
	bool initTensors,
	PointSet* ps, 
	const float32 epsilon,
	const bool saveTensors,
	const card32 numKNN,
	const float32 timesSamplingSpacing,
	const card32 mode,
	bool useCachedRangeQuerries)
{
	if ( ps == NULL ) {
		debugOutput << "initStickTensors: invalid point cloud.\n"; debugOutput.flush();
		return;
	}
	if ( !ps->providesAttribute( "position" ) ){
		debugOutput << "initStickTensors: 'position' channel required.\n"; debugOutput.flush();
		return;
	}
	if ( !ps->providesAttribute( "surfaceness" ) ){
		debugOutput << "initStickTensors: 'surfaceness' channel required.\n"; debugOutput.flush();
		return;
	}
	if ( !ps->providesAttribute( "curveness" ) ){
		debugOutput << "initStickTensors: 'curveness' channel required.\n"; debugOutput.flush();
		return;
	}

	AAT POSITION = ps->getAAT( "position" );
	AAT SURFACENESS = ps->getAAT( "surfaceness" );
	AAT CURVENESS = ps->getAAT( "curveness" );
	AAT TENSOR_0, TENSOR_1;
	AAT SAMPLING_SPACING;

	if ( saveTensors || !initTensors ){
		if ( !ps->providesAttribute( "tensor_0" ) || !ps->providesAttribute( "tensor_1" ) ){
			debugOutput << "initStickTensors: 'tensor_0/tensor_1' channel required.\n"; debugOutput.flush();
			return;
		}
		TENSOR_0 = ps->getAAT( "tensor_0" );
		TENSOR_1 = ps->getAAT( "tensor_1" );
	}

	if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ){
		if ( !ps->providesAttribute( "samplingSpacing" ) ){
			debugOutput << "initStickTensors: channel 'samplingSpacing' required.\n"; debugOutput.flush();
			return;
		}
		SAMPLING_SPACING = ps->getAAT( "samplingSpacing" );
	}

	float32 sqrEpsilon = sqr( epsilon );
	Matrix3f eigenVectors;
	Vector3f eigenValues;

	float32 startTime = clock();
	progressWindow->pushStep( true, "Tensor Voting" );
	Timer timer;
	Timer T; card32 vt;
	long numOutliers = 0;
	card32 nIndex = 0; 
	if( !useCachedRangeQuerries )
	{
		T.getDeltaValue();
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
		PointSetKNNQuery knn( ps, octree, makeVector3f( 0,0,0 ), POSITION );
		//PointSetRangeQuery annQuery( ps, sqrEpsilon );
		vt = T.getDeltaValue();

		debugOutput << "Build search structure: " <<  vt << " ms\n";debugOutput.flush();
		for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
			const Vector3f& p = ps->get3f( pointIndex, POSITION );

			knn.setTargetPoint( p );
			//annQuery.setTargetPoint(p);
			
			Vector3f np = p;
			float32 sqrD = ( np - p ) * ( np - p );
			Matrix3f tensor = makeMatrix3f( 0,0,0,0,0,0,0,0,0 );
			bool found = true;
			nIndex = pointIndex;
			Matrix3f receiverEigenVectors;

			// Extract orientation of tensor at point
			if ( !initTensors ){
				Vector3f receiver_tensor_0 = ps->get3f( pointIndex, TENSOR_0 );
				Vector3f receiver_tensor_1 = ps->get3f( pointIndex, TENSOR_1 );
				Matrix3f receiverTensor;
				receiverTensor[0][0] = receiver_tensor_0[0];
				receiverTensor[0][1] = receiver_tensor_0[1];
				receiverTensor[0][2] = receiver_tensor_0[2];
				receiverTensor[1][1] = receiver_tensor_1[0];
				receiverTensor[1][2] = receiver_tensor_1[1];
				receiverTensor[2][2] = receiver_tensor_1[2];
				receiverTensor[1][0] = receiverTensor[0][1];
				receiverTensor[2][0] = receiverTensor[0][2];
				receiverTensor[2][1] = receiverTensor[1][2];
				Vector3f receiverEigenValues;
				receiverTensor.computeEigenStructure( receiverEigenValues, receiverEigenVectors );
			}

			if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
				float32 samplingSpacing = ps->get1f( pointIndex, SAMPLING_SPACING );
				sqrEpsilon = sqr( timesSamplingSpacing * samplingSpacing );
			}

			card32 numNeighbors = 0;
			bool converged = false;
			while( !converged ){
				if ( initTensors ){
					tensor += computeTensorVote( p, np, epsilon );
					numNeighbors++;
				} else {
					// Extract sender tensor
					Vector3f sender_tensor_0 = ps->get3f( nIndex, TENSOR_0 );
					Vector3f sender_tensor_1 = ps->get3f( nIndex, TENSOR_1 );
					Matrix3f senderTensor;
					senderTensor[0][0] = sender_tensor_0[0];
					senderTensor[0][1] = sender_tensor_0[1];
					senderTensor[0][2] = sender_tensor_0[2];
					senderTensor[1][1] = sender_tensor_1[0];
					senderTensor[1][2] = sender_tensor_1[1];
					senderTensor[2][2] = sender_tensor_1[2];
					senderTensor[1][0] = senderTensor[0][1];
					senderTensor[2][0] = senderTensor[0][2];
					senderTensor[2][1] = senderTensor[1][2];
					Matrix3f vote;
					if ( computeTensorVote( vote, receiverEigenVectors, p, senderTensor, np, epsilon ) ){
						tensor += vote;
						numNeighbors++;
					}
				}

				knn.getNextNearestPointIndex( nIndex, found );
				//annQuery.getNextNearestPointIndex(nIndex, found);
				if (found)
				{
					np = ps->get3f( nIndex, POSITION );
					sqrD = ( np - p ) * ( np - p );
				}

				if ( mode == ENVIRONMENT_KNN ){ 
					if ( numNeighbors >= numKNN ) 
						converged = true;
				} else 
					if ( sqrD > sqrEpsilon || !found ) 
						converged = true;
			}
			//if ( numNeighbors > 0 ) tensor /= float32( numNeighbors );
			tensor.computeEigenStructure( eigenValues, eigenVectors );
			const float32 surfaceness = eigenValues[0] - eigenValues[1];
			const float32 curveness = eigenValues[2];
			ps->set1f( pointIndex, SURFACENESS, surfaceness );
			ps->set1f( pointIndex, CURVENESS, curveness );
			if ( saveTensors ){
				Vector3f tv0, tv1;
				tv0[0] = tensor[0][0];
				tv0[1] = tensor[0][1];
				tv0[2] = tensor[0][2];
				tv1[0] = tensor[1][1];
				tv1[1] = tensor[1][2];
				tv1[2] = tensor[2][2];
				ps->set3f( pointIndex, TENSOR_0, tv0 );
				ps->set3f( pointIndex, TENSOR_1, tv1 );
			}
			if( pointIndex % 1000 == 0)
				progressWindow->progress( card32( float32( pointIndex ) / float32( ps->getNumEntries() ) * 100 ) );
		}
		delete octree;
	}
	else
	{
		// ----------------------------
		// MB - tv using grid range querries

		if( mode == ENVIRONMENT_KNN )
			throw "TensorVoting - KNN-mode not compatible with grid range querries!";

		float radius = sqrt(sqrEpsilon);
		GridBasedSphericalRangeQuery gquerry(ps,radius);
		

		for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
			const Vector3f& p = ps->get3f( pointIndex, POSITION );
			vector<int32> indices;
			gquerry.getIndices(p, indices);
			Matrix3f tensor = makeMatrix3f( 0,0,0,0,0,0,0,0,0 );
			bool found = true;
			Matrix3f receiverEigenVectors;

			// Extract orientation of tensor at point
			if ( !initTensors ){
				Vector3f receiver_tensor_0 = ps->get3f( pointIndex, TENSOR_0 );
				Vector3f receiver_tensor_1 = ps->get3f( pointIndex, TENSOR_1 );
				Matrix3f receiverTensor;
				receiverTensor[0][0] = receiver_tensor_0[0];
				receiverTensor[0][1] = receiver_tensor_0[1];
				receiverTensor[0][2] = receiver_tensor_0[2];
				receiverTensor[1][1] = receiver_tensor_1[0];
				receiverTensor[1][2] = receiver_tensor_1[1];
				receiverTensor[2][2] = receiver_tensor_1[2];
				receiverTensor[1][0] = receiverTensor[0][1];
				receiverTensor[2][0] = receiverTensor[0][2];
				receiverTensor[2][1] = receiverTensor[1][2];
				Vector3f receiverEigenValues;
				receiverTensor.computeEigenStructure( receiverEigenValues, receiverEigenVectors );
			}

			if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
				float32 samplingSpacing = ps->get1f( pointIndex, SAMPLING_SPACING );
				sqrEpsilon = sqr( timesSamplingSpacing * samplingSpacing );
			}

			for( unsigned ci=0;ci<indices.size();ci++)
			{
				nIndex = indices[ci];
				Vector3f np = ps->get3f(nIndex, POSITION);//(*positions)[ci];

				if ( initTensors ){
					tensor += computeTensorVote( p, np, epsilon );
				} else {
					// Extract sender tensor
					Vector3f sender_tensor_0 = ps->get3f( nIndex, TENSOR_0 );
					Vector3f sender_tensor_1 = ps->get3f( nIndex, TENSOR_1 );
					Matrix3f senderTensor;
					senderTensor[0][0] = sender_tensor_0[0];
					senderTensor[0][1] = sender_tensor_0[1];
					senderTensor[0][2] = sender_tensor_0[2];
					senderTensor[1][1] = sender_tensor_1[0];
					senderTensor[1][2] = sender_tensor_1[1];
					senderTensor[2][2] = sender_tensor_1[2];
					senderTensor[1][0] = senderTensor[0][1];
					senderTensor[2][0] = senderTensor[0][2];
					senderTensor[2][1] = senderTensor[1][2];
					Matrix3f vote;
					if ( computeTensorVote( vote, receiverEigenVectors, p, senderTensor, np, epsilon ) ){
						tensor += vote;
					}
				}

			}
			//if ( numNeighbors > 0 ) tensor /= float32( numNeighbors );
			tensor.computeEigenStructure( eigenValues, eigenVectors );
			const float32 surfaceness = eigenValues[0] - eigenValues[1];
			ps->set1f( pointIndex, SURFACENESS, surfaceness );
			if ( saveTensors ){
				Vector3f tv0, tv1;
				tv0[0] = tensor[0][0];
				tv0[1] = tensor[0][1];
				tv0[2] = tensor[0][2];
				tv1[0] = tensor[1][1];
				tv1[1] = tensor[1][2];
				tv1[2] = tensor[2][2];
				ps->set3f( pointIndex, TENSOR_0, tv0 );
				ps->set3f( pointIndex, TENSOR_1, tv1 );
			}
			if( pointIndex % 1000 == 0)
				progressWindow->progress( card32( float32( pointIndex ) / float32( ps->getNumEntries() ) * 100 ) );
		}
	}
	debugOutput << "tensorVorting: " << convertTimeToString(timer.getDeltaValue()) << "\n";

	progressWindow->popSteps( );
}