//---------------------------------------------------------------------------
#ifndef PGOpt2PotentialH
#define PGOpt2PotentialH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "SparseLinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl {

class PGOpt2Parameters;
class InCorePCTopologyGraph;
class OptimizationGraph;
class AnimationBSRContext;

///
///  A single potential function for variational patch graph optimization
///
class ANIMATIONRECONSTRUCTION_API PGOpt2Potential : public Persistent {
	GEOXL_ABSTRACT_CLASS( PGOpt2Potential )
 public:
	PGOpt2Potential();

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   /// main method: quadraticApprox - computes a quadratic approximation of the objective function
   /// input is: optimization graph, topology, and parameters
   /// output is: value at current pos (read pos from frames), its gradient and its Hessian
   /// frame pair attributes can be stored in framePair (use PointSet that represents the patchPairs or derive class for additional attributes)
   /// per-frame attributes can be stored in frame1, frame2 (derive class)
   virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
   ) = 0;

	~PGOpt2Potential();
};



class ANIMATIONRECONSTRUCTION_API PGOpt2SquaredDistancePotential_Global : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2SquaredDistancePotential_Global )

public:

	 virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};

class ANIMATIONRECONSTRUCTION_API PGOpt2SquaredDistancePotential_Pairwise : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2SquaredDistancePotential_Pairwise )

public:

	virtual void quadraticApprox(
		// -- input
		const PGOpt2Parameters     *parameters,
		const AnimationBSRContext   *recParameters,
		const InCorePCTopologyGraph *topology,
		// -- input & output
		OptimizationGraph           *optGraph,
		// -- output
		float32                     &potential,
		DVectorF                    *gradient,
		SparseMatrixF               *quadApprox
		);
};

class ANIMATIONRECONSTRUCTION_API PGOpt2MarkerErrorPotential : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2MarkerErrorPotential )

public:
	 virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};

class ANIMATIONRECONSTRUCTION_API PGOpt2SmoothnessPotential : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2SmoothnessPotential )

public:
	 virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};


class ANIMATIONRECONSTRUCTION_API PGOpt2RigidityPotential : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2RigidityPotential )

public:
	 virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};

class ANIMATIONRECONSTRUCTION_API PGOpt2TemporalCoherence : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2TemporalCoherence )

public:
	 virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};


class ANIMATIONRECONSTRUCTION_API PGOpt2STColorSimilarity : public PGOpt2Potential {
	GEOXL_CLASS( PGOpt2STColorSimilarity )

public:
	virtual void quadraticApprox(
      // -- input
	   const PGOpt2Parameters     *parameters,
	   const AnimationBSRContext   *recParameters,
      const InCorePCTopologyGraph *topology,
      // -- input & output
      OptimizationGraph           *optGraph,
      // -- output
      float32                     &potential,
      DVectorF                    *gradient,
      SparseMatrixF               *quadApprox
	  );
};

}

#endif