//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PatchGraph.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ObjectClassProperty.h"
#include "BooleanClassProperty.h"
#include "TrajectorySet.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.h"
#include "ObjectListProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( PatchGraph, 5 )
{
	BEGIN_CLASS_INIT( PatchGraph );
	ADD_OBJECT_PROP( ts, 0, TrajectorySet::getClass(), true )
	ADD_OBJECT_PROP( urshape, 4, PointSet::getClass(), true )
	ADD_OBJECT_PROP( urshapeAttachments, 5, AttachedDataContainer::getClass(), true );
	ADD_OBJECT_PROP( tg, 0, InCorePCTopologyGraph::getClass(), true )
	ADD_BOOLEAN_PROP( in2d, 0 )
	ADD_OBJECT_LIST_PROP( FeatureList, 3, FeatureList::getClass() )
}


// ----


PatchGraph::PatchGraph() {
	ts = NULL;
	tg = NULL;
	urshape = NULL;
	urshapeAttachments = new AttachedDataContainer;
	topologyColor = makeVector3f( rand()%100/100.0f, rand()%100/100.0f, rand()%100/100.0f );
	pointColor = makeVector3f( rand()%100/100.0f, rand()%100/100.0f, rand()%100/100.0f );
}

void PatchGraph::clearAndSetup(TrajectorySet *ts, InCorePCTopologyGraph *tg, bool in2d) {
	this->in2d = in2d;
	if (this->ts != ts) {
		delete this->ts;
		this->ts = NULL;
	}
	if (this->tg != tg) {
		delete this->tg;
		this->tg = NULL;
	}
	this->ts = ts;
	this->tg = tg;
	for( card32 i=0;i<features.size();i++ )
		delete features[i];
	features.clear();
}

PatchGraph *PatchGraph::compatibleClone(card32 numFrames, card32 numPatches) const {
   PointSet *ps = new PointSet();
   ps->clearAndSetup(numFrames, numPatches, ts->getTrajectories()->getDescr());
	TrajectorySet *newTs = new TrajectorySet();
	newTs->clearAndSetup(ps);
	InCorePCTopologyGraph *newTg = new InCorePCTopologyGraph;
	newTg->clearAndSetup(tg->getEdges()->getDescr());
   PatchGraph *result = new PatchGraph();
   result->clearAndSetup(newTs, newTg, in2d);
   if (urshape) {
      result->createAndClearUrshape();
   }
   return result;
}


void PatchGraph::assign(const Object* obj, CopyContext *context) {
	const PatchGraph *other = dynamic_cast<const PatchGraph*>(obj);
	if (other == NULL) {
		throw PException("PatchGraph::assign() - wrong parameter type.");
	}
	copyObjectProperties(other, this);
	
	this->topologyColor = other->topologyColor;
	this->pointColor = other->pointColor;
	this->bbox = other->bbox;
}

PatchGraph::~PatchGraph() {
	delete ts;
	delete tg;
	delete urshape;
	delete urshapeAttachments;

	for( card32 i=0;i<features.size();i++ )
		delete features[i];
}

/**
 * Test if patch graph contains a frame 'index'
 */
bool PatchGraph::containsFrame( const card32 index ) const
{
	return ( index < ts->getNumFrames() );
}

void PatchGraph::createPatchAttribute(VAD attribute) {
	ts->createPatchAttribute(attribute);
	if (urshape != NULL) {
		if (!urshape->providesAttribute(attribute.name, attribute.numComponents, attribute.dataFormat)) {
			if (urshape->providesAttribute(attribute.name)) {
				throw PException("PatchGraph::createPatchAttribute() - channel exists already in URSHAPE with different parameters (numComponents/dataFormat)");
			}
			VertexDescriptor newAttribs;
			newAttribs.pushAttrib(attribute);
			PointSet *newUS = urshape->appendFields(&newAttribs);
			delete urshape;
			urshape = newUS;		
		}
	}
}

void PatchGraph::createAndClearUrshape(bool initializeWithFrame, card32 initFrameIndex) {
	delete urshape;
	urshape = NULL;
	if (initializeWithFrame) {
	   pAssert(initFrameIndex < ts->getNumFrames());
      urshape = ts->copySingleFrame(initFrameIndex);
	} else {
	   urshape = new PointSet();
	   urshape->clearAndSetup(1, ts->getNumPatches(), ts->getTrajectories()->getDescr());
   }
   invalidateUrshapeAttachments();
}

/** 
 * Compute bounding box of patch graph; (based on first frame)
 */
void PatchGraph::computeBBox()
{
	AAT POSITION = ts->getTrajectories()->getAAT( "position" );
	bbox = BoundingBox3f( ts->getTrajectories()->get3f( 0, 0, POSITION ) );
	for ( unsigned int pointIndex = 0; pointIndex < ts->getTrajectories()->getNumEntries(); pointIndex++ ){
		bbox.addPoint( ts->getTrajectories()->get3f( pointIndex, POSITION ) );
	}
}
void PatchGraph::addFeatureList( Object * obj ) 
{
	features.push_back( (FeatureList*)obj );
}

void PatchGraph::deleteFeatureLists( card32 begin, card32 end ) 
{
	for( card32 i=begin;i<end;i++ )
		delete features[i];
	features.erase( features.begin() + begin, features.begin() + end );
}

Object *PatchGraph::getFeatureList( card32 index ) 
{
	return features[index];
}

void PatchGraph::moveFeatureList( card32 oldPos, card32 newPos ) 
{
	FeatureList * list = features[oldPos];
	features.erase( features.begin() + oldPos );
	features.insert( features.begin() + newPos, list );
}

void PatchGraph::invalidateUrshapeAttachments()
{
	urshapeAttachments->deleteDelOnWriteData();
}

AttachedIndexedOctree *PatchGraph::fetchUrshapeOctree(card32 maxPointsPerNode)
{
   AttachedIndexedOctree *octree = dynamic_cast<AttachedIndexedOctree*>(urshapeAttachments->getData("SimpleOctree"));
   if (octree == NULL) {
		octree = new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
		octree->octreesetup(urshape, maxPointsPerNode);
		urshapeAttachments->attachData(octree);
   }
   return octree;
}

