//---------------------------------------------------------------------------
#ifndef PatchGraphH
#define PatchGraphH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------
#include "TrajectorySet.h"
#include "Feature.h"
#include "InCorePCTopologyGraph.h"
#include "AttachedIndexedOctree.h"
//---------------------------------------------------------------------------


namespace geoxl {

//class InCorePCTopologyGraph;

/**
	a graph of patches (i.e. a point set with topology graph)
	supports animations (one frame for 3d only case)
	optionally, a set of original data points can be attached that are associated with each patch
	(also) optionally, an urshape can be associated with the patch graph; initially this object is NULL
*/
class ANIMATIONRECONSTRUCTION_API PatchGraph : public Persistent {
	GEOXL_CLASS( PatchGraph )
 private:
	TrajectorySet *ts;
	InCorePCTopologyGraph *tg;
	PointSet *urshape;
	AttachedDataContainer *urshapeAttachments;
	bool in2d;
	vector<FeatureList *> features;

   // -- rendering
	Vector3f topologyColor;			// Used for rendering via PatchGraphRenderer (topology)
	Vector3f pointColor;			// Used for rendering via PatchGraphRenderer (points)
	BoundingBox3f bbox;

 public:
   // --- basics: create and destroy
	PatchGraph();
	void clearAndSetup(TrajectorySet *ts, InCorePCTopologyGraph *tg, bool in2d = false);
   /// creates an empty (no values set) patch graph of same vertex descriptor
	PatchGraph *compatibleClone(card32 numFrames=1, card32 numPatches=0) const;
	virtual void assign(const Object* obj, CopyContext *context = NULL);
	virtual ~PatchGraph();
	
	// --- member access
	inline bool getIn2d() const {return in2d;}
	inline		 TrajectorySet *getTrajectorySet()		  {return ts;}
	inline const TrajectorySet *getTrajectorySet() const {return ts;}
	inline		 InCorePCTopologyGraph *getTopology()	     {return tg;}
	inline const InCorePCTopologyGraph *getTopology() const {return tg;}
	inline void setTopology( InCorePCTopologyGraph* newTG ){ delete tg; tg = newTG; }
	inline void detachTopology() {tg=NULL;}
	inline void setStartFrameIndex( const card32 index ){ ts->setStartFrameIndex( index ); }
	inline card32 getStartFrameIndex() const { return ts->getStartFrameIndex(); }
	/// Test if patch graph contains a frame 'index'
	bool containsFrame( const card32 index ) const;
	/// check wether the given attribute channel is present, create if not
	/// throws an exception if channel name exists with different parameters
	void createPatchAttribute(VAD attribute);

	// --- urshapes
	/// creates an urshape (if it not yet exists), sets height to trajectoryset height
	/// and vertex attributes to the trajectory set's current attribs
	void createAndClearUrshape(bool initializeWithFrame = false, card32 initFrameIndex = 0);
	inline PointSet *getUrshape() {return urshape;}
	inline const PointSet *getUrshape() const {return urshape;}
	AttachedDataContainer *getUrshapeAttachments() {return urshapeAttachments;}
	void invalidateUrshapeAttachments();
	/// helper method - fetch or create an index octree for the urshape
	AttachedIndexedOctree *fetchUrshapeOctree(card32 maxPointsPerNode = 20);	

   // --- features
	void addFeatureList( Object * obj );
	void deleteFeatureLists( card32 begin, card32 end );
   Object * getFeatureList( card32 index );
   void moveFeatureList( card32 oldPos, card32 newPos );
	mpcard getNumFeatureLists() {return features.size();}
	string getFeatureListName(card32 index ) {return "list";}
	FeatureList * getFeatures( card32 index ) {return features[index];}

   // --- rendering related stuff
	// Compute bounding box of patch graph
	void computeBBox();
	inline Vector3f getTopologyColor() const { return topologyColor; }
	inline void setTopologyColor( const Vector3f& color ){ topologyColor = color; }
	inline Vector3f getPointColor() const { return pointColor; }
	inline void setPointColor( const Vector3f& color ){ pointColor = color; }
	inline BoundingBox3f getBoundingBox(){ return bbox; }
};

}

#endif