//---------------------------------------------------------------------------
#ifndef PatchTypesDH
#define PatchTypesDH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
#include "LinearAlgebra.h"
#include "VertexDescriptor.h"
#include "PointFlags.h"
//---------------------------------------------------------------------------

namespace geoxl {


// ---- identifiers for attribute channels (e.g. in point sets)

// position (3f)
const string PATCH_CHANNEL_NAME_POSITION						= "position";
// reference frame is normal (3f) and tangent_u (3f)
const string PATCH_CHANNEL_NAME_NORMAL							= "normal";
const string PATCH_CHANNEL_NAME_TANGENT_U						= "tangent_u";
// rotation matrix is represented as a quaternion (4f)
const string PATCH_CHANNEL_NAME_ROTATION_MATRIX					= "rotation_matrix";

// coefficients are stored as variably sized float array
const string HEIGHTFIELD_COEFFS_NAME							= "hfieldCoeffs";
// and so are correspondence coefficients
const string CORRESPONDENCE_COEFFS_NAME							= "corrCoeffs";

extern ANIMATIONRECONSTRUCTION_API const VAD POSITION_PATCH_VAD;
extern ANIMATIONRECONSTRUCTION_API const VAD NORMAL_PATCH_VAD;
extern ANIMATIONRECONSTRUCTION_API const VAD TANGENT_U_PATCH_VAD;
extern ANIMATIONRECONSTRUCTION_API const VAD ROTATION_MATRIX_PATCH_VAD;

VAD		createHeightfieldCoeffVAD(card32 numCoeff);
VAD		createCorrespondenceCoeffVAD(card32 numCoeff);


// ---- patches themselves


struct ANIMATIONRECONSTRUCTION_API ReferenceFramePatch {
	Vector3f normal;
	/// tangent_v is normal x coords_u
	Vector3f tangent_u;
	
	inline void get(const PointSet *ps, card32 i, AAT normal_aat, const AAT tangent_u_aat) {
		normal = ps->get<float32, 3>(i, normal_aat);
		tangent_u = ps->get<float32, 3>(i, tangent_u_aat);
	}
	inline void get(const PointSet *ps, card32 x, card32 y, AAT normal_aat, const AAT tangent_u_aat) {
		normal = ps->get<float32, 3>(x, y, normal_aat);
		tangent_u = ps->get<float32, 3>(x, y, tangent_u_aat);
	}
	inline void set(PointSet *ps, card32 i, const AAT normal_aat, const AAT tangent_u_aat) {
		ps->set<float32, 3>(i, normal_aat, normal);
		ps->set<float32, 3>(i, tangent_u_aat, tangent_u);
	}
	inline void set(PointSet *ps, card32 x, card32 y, const AAT normal_aat, const AAT tangent_u_aat) {
		ps->set<float32, 3>(x, y, normal_aat, normal);
		ps->set<float32, 3>(x, y, tangent_u_aat, tangent_u);
	}
	inline Vector3f computeTangentV() const {
		return normal.crossProduct(tangent_u);
	}
};

struct CoefficientSet {
	DynamicVector<float32> coeffs;
	
	inline CoefficientSet(card32 numCoeffs) {coeffs.setDim(numCoeffs);}
	inline CoefficientSet(const DynamicVector<float32> &coeffs) {this->coeffs = coeffs;}

	void get(const PointSet *ps, card32 i, AAT aat);
	void set(PointSet *ps, card32 i, AAT aat);

	inline float32 operator[](card32 i) const {return coeffs[i];}
	inline float32 &operator[](card32 i) {return coeffs[i];}
};


// ---- flags used in animation reconstruction to tag patches in patch graphs


/// the point has been initialized before (not just plain mem)
const int32 FLAG_PATCH_INITIALIZED = max(256, 1<<MAX_NUM_POINT_FLAGS);
/// the point is just a guess, not a reliable interpolated or fitted value
const int32 FLAG_PATCH_GUESS       = max(512, 2<<MAX_NUM_POINT_FLAGS);
/// big hole in patch graph - do not smooth
const int32 FLAG_PATCH_NOSMOOTH    = max(1024, 4<<MAX_NUM_POINT_FLAGS);

}

#endif