//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PointCloudAnimationSequence.h"
//---------------------------------------------------------------------------
#include "VariableArrayClassProperty.h"
#include "ObjectClassProperty.h"
#include "PointSetKNNQuery.h"
#include "MgcEigen.h"
//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS(PointCloudAnimationSequence, 1)
{
	BEGIN_CLASS_INIT( PointCloudAnimationSequence );
	ADD_VAR_ARRAY_PROP( frames, 0, 
		new ObjectClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		                         PointSet::getClass(), true, NULL )
	)
	ADD_VAR_ARRAY_PROP( frameAttachments, 1, 
		new ObjectClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		                         AttachedDataContainer::getClass(), true, NULL )
	)
}


//----


PointCloudAnimationSequence::PointCloudAnimationSequence() {
}

void PointCloudAnimationSequence::addFrame(const PointSet *frame, InsertWhere where) {
	PointSet *ps = (PointSet*)(frame->copy());
	if (where == INSERT_FRONT) {
		frames.push_front(ps);
		frameAttachments.push_front(new AttachedDataContainer());
	} else {
		frames.push_back(ps);
		frameAttachments.push_back(new AttachedDataContainer());
	}
	
}

PointSet *PointCloudAnimationSequence::getFrame(card32 time) {
	if (time > frames.size()) {
		throw ERangeCheck("PointCloudAnimationSequence::getFrame() - index out of range()");
	}
	return frames[time];
}

const PointSet *PointCloudAnimationSequence::getFrame(card32 time) const {
	if (time > frames.size()) {
		throw ERangeCheck("PointCloudAnimationSequence::getFrame() - index out of range()");
	}
	return frames[time];
}

/** 
 * Remove all frames
 */
void PointCloudAnimationSequence::clear()
{
	for ( unsigned int frameIndex = 0; frameIndex < frames.size(); frameIndex++ ){
		delete frames[frameIndex];
	}
	frames.clear();

	for (unsigned i=0; i<frameAttachments.size(); i++) {
		delete frameAttachments[i];
	}
	frameAttachments.clear();
}

void PointCloudAnimationSequence::invalidateAttachments(card32 frame) {
	pAssert(frame < frameAttachments.size());
	frameAttachments[frame]->deleteDelOnWriteData();
}

void PointCloudAnimationSequence::invalidateAllAttachments() {
	for (unsigned i=0; i<frameAttachments.size(); i++) {
		invalidateAttachments(i);
	}
}

AttachedIndexedOctree *PointCloudAnimationSequence::fetchOctree( card32 frame, card32 maxPointsPerNode ) 
{
   AttachedDataContainer *attachments = getAttachments(frame);
   AttachedIndexedOctree *octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));
   if (octree == NULL) {
		PointSet *ps = frames[frame];
		octree = new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
		octree->octreesetup(ps, maxPointsPerNode);
		attachments->attachData(octree);
		//debugOutput << "  Building octree for animation sequence frame.\n"; debugOutput.flush();
   }
   return octree;
}

AttachedDataContainer *PointCloudAnimationSequence::getAttachments(card32 frame) {
	pAssert(frameAttachments.size() == 0 || frameAttachments.size() == frames.size());
	if (frameAttachments.size() != frames.size()) {
		frameAttachments.resize(frames.size(), NULL);
		for (unsigned i=0; i<frames.size(); i++) {
			if (frameAttachments[i] == NULL) {
				frameAttachments[i] = new AttachedDataContainer();
			}
		}
	}
	return frameAttachments[frame];
}



/**
 * Destructor
 */
PointCloudAnimationSequence::~PointCloudAnimationSequence() {
	for (unsigned i=0; i<frames.size(); i++) {
		delete frames[i];
	}
	for (unsigned i=0; i<frameAttachments.size(); i++) {
		delete frameAttachments[i];
	}
}

void	PointCloudAnimationSequence::computeGradientOfColors(const card32 numOfNeighbors)
{
	createDataAttribute(mVAD("gradient_color_r", 3, VAD::DATA_FORMAT_FLOAT32));
	createDataAttribute(mVAD("gradient_color_g", 3, VAD::DATA_FORMAT_FLOAT32));
	createDataAttribute(mVAD("gradient_color_b", 3, VAD::DATA_FORMAT_FLOAT32));
	geoxl::Eigen<float32>	*eigenSolver= new geoxl::Eigen<float32>(3);

	for(card32	frameIndex = 0; frameIndex < frames.size(); frameIndex++)
	{
		PointSet		*pc = frames[frameIndex];

		AAT		POSITION	= pc->getAAT("position");
		AAT		COLOR			= pc->getAAT("color");
		AAT		GRA_COLOR_R = pc->getAAT("gradient_color_r");
		AAT		GRA_COLOR_G = pc->getAAT("gradient_color_g");
		AAT		GRA_COLOR_B = pc->getAAT("gradient_color_b");

		card32	height	= pc->getHeight();
		AttachedIndexedOctree *dataOctree = fetchOctree(frameIndex);
		PointSetKNNQuery data_knn(pc, dataOctree, NULL_VECTOR3F, POSITION);

		float32		A[2][2], InvA[2][2];
		Vector3f		b_color[2];
		float32	fsumXX[3][3], fsumX[3];

		for(card32 surfelIndex = 0; surfelIndex < height; surfelIndex++)
		{
			Vector3f		dik = pc->get3f(0, surfelIndex, POSITION);

			for(card32 k = 0; k < 2; k++)
			{
				for(card32 j = 0; j < 2; j++)
				{
					A[k][j] = 0.0f;
				}
				b_color[k][0] = b_color[k][1] = b_color[k][2] = 0.0f;
			}
			//Used in computing local coordinate system
			for(card32 k = 0; k < 3; k++)
			{
				for(card32 j = 0; j < 3; j++)
				{
					fsumXX[k][j] = 0.0f;
				}
				fsumX[k] = 0.0f;
			}

			//Collect neighbors of this data point
			data_knn.setTargetPoint(dik);
			vector<card32>	neighbors_id;
			card32		neighbor_id, numOfFoundNeighbors = 0;
			do 
			{
				bool found;
				data_knn.getNextNearestPointIndex(neighbor_id, found);
				if(found && neighbor_id != surfelIndex)
				{
					neighbors_id.push_back(neighbor_id);
				}
				numOfFoundNeighbors++;
			} while(numOfFoundNeighbors < numOfNeighbors);

			/// Compute local coordinate system
			for(card32 ni = 0; ni < neighbors_id.size(); ni++)
			{
				neighbor_id = neighbors_id[ni];
				Vector3f	pjk = pc->get3f(0, neighbor_id, POSITION);
				for(card32 k = 0; k < 3; k++)
				{
					for(card32 j = 0; j < 3; j++)
					{
						fsumXX[k][j] += pjk[k]*pjk[j];
					}
					fsumX[k] += pjk[k];
				}
			}
			for(card32 k = 0; k < 3; k++)
			{
				for(card32 j = 0; j < 3; j++)
				{
					fsumXX[k][j] /= neighbors_id.size();
				}
				fsumX[k] /= neighbors_id.size();
			}
			for(card32 k = 0; k < 3; k++)
			{
				for(card32 j = 0; j < 3; j++)
				{
					eigenSolver->Matrix(k, j) = 	fsumXX[k][j] - fsumX[k]*fsumX[j];
				}
			}
			//solve the eigen-system
			eigenSolver->DecrSortEigenStuff3();
			Vector3f	nor, E1, E2;
			for(card32 k = 0; k < 3; k++)
			{
				nor[k] = eigenSolver->GetEigenvector(k, 2);
				E1[k] = eigenSolver->GetEigenvector(k, 0);
				E2[k] = eigenSolver->GetEigenvector(k, 1);
			}
			Vector3f		ci = pc->get3f(0, surfelIndex, COLOR);
			for(card32 ni = 0; ni < neighbors_id.size(); ni++)
			{
				neighbor_id	= neighbors_id[ni];
				Vector3f	rij		= dik - pc->get3f(0, neighbor_id, POSITION);
				Vector3f	cij	= ci - pc->get3f(0, neighbor_id, COLOR);
				float32	tp[2];  tp[0]= rij*E1,	tp[1] = rij*E2;
				for(card32 k = 0; k < 2; k++)
				{
					for(card32 j = 0; j < 2; j++)
					{
						A[k][j] += tp[k]*tp[j];
					}
					b_color[k] += cij*tp[k];
				}
			}
			float32 det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
			InvA[0][0] = A[1][1]/det;		InvA[0][1] = -A[0][1]/det;
			InvA[1][0] = -A[1][0]/det;		InvA[1][1] = A[0][0]/det;

			Vector3f	coef1, coef2;
			coef1 = b_color[0]*InvA[0][0] + b_color[1]*InvA[0][1];
			coef2 = b_color[0]*InvA[1][0] + b_color[1]*InvA[1][1];
			Vector3f	grad_r = E1*coef1[0] + E2*coef2[0];
			pc->set3f(0, surfelIndex, GRA_COLOR_R, grad_r);
			Vector3f	grad_g = E1*coef1[1] + E2*coef2[1];
			pc->set3f(0, surfelIndex, GRA_COLOR_G, grad_g);
			Vector3f	grad_b = E1*coef1[2] + E2*coef2[2];
			pc->set3f(0, surfelIndex, GRA_COLOR_B, grad_b);
		}
	}

	delete eigenSolver;
}

void	PointCloudAnimationSequence::createDataAttribute(VAD attribute)
{
   invalidateAllAttachments();
   size_t numFrames = frames.size();
   VertexDescriptor vd;
   vd.pushAttrib(attribute);
   for (unsigned i=0; i<numFrames; i++) {
      PointSet *newPS = frames[i]->appendFields(&vd);
      delete frames[i];
      frames[i] = newPS;
   }
}