//---------------------------------------------------------------------------
#ifndef QuadraticObjectiveFunctionH
#define QuadraticObjectiveFunctionH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
#include "NonLinearOptimization.h"
#include "DynamicLinearAlgebra.h"
#include "SparseLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl {

/// quadratic objective function
/// represents quadratic function 0.5*(x^t)*[quadMat]*x + [linVect]*x + [constVect]
/// 
/// derived from multi-variate scalar function with gradient and hessian
template <typename FloatType, typename IntType = int32>
class QuadraticObjectiveFunction : public MVSFunctionWithHessian<FloatType, IntType> {
 private:
	SparseMatrix<FloatType> m_quadMat;
	DynamicVector<FloatType> m_linVect;
	FloatType m_constVal;
	
	/// throws an exception if lin and quad dimensions do not match
	void checkDim() const;
	
 public:
	/// creates zero function
   QuadraticObjectiveFunction(IntType dim=0);
	QuadraticObjectiveFunction(const SparseMatrix<FloatType> &quadMat, const DynamicVector<FloatType> &linVect, const FloatType &constVal);

   inline       SparseMatrix<FloatType> &quadMat()						{return quadMat;}
   inline const SparseMatrix<FloatType> &quadMat() const				{return quadMat;}
   inline void	setQuadMat(const SparseMatrix<FloatType> &quadMat) {m_quadMat = quadMat;}

   inline       DynamicVector<FloatType> &linVect()					 {return linVect;}
   inline const DynamicVector<FloatType> &linVect() const			 {return linVect;}
   inline void	setLinVect(const DynamicVector<FloatType> &linVect) {m_linVect = linVect;}

   inline       FloatType &constVal()			  {return constVect;}
   inline const FloatType &constVal() const	  {return constVect;}
   inline void setConstVal(FloatType constVal) {m_constVal = constVal;}
   
   // --- inherited
   virtual IntType getDimension() const;
	/// returns 0.5 * (pos^t)[quadMat](pos) + (pos^t)(linVect) + constVal
   virtual FloatType calc(const DynamicVector<FloatType> &pos) const;
   virtual void calcGradient(const DynamicVector<FloatType> &pos, DynamicVector<FloatType> &gradient) const;
   virtual void calcHessian(const DynamicVector<FloatType> &pos, SparseMatrix<FloatType> &hessian) const;   
};


}

#endif
