//---------------------------------------------------------------------------
#include "SparseScatterOperations.h"
//---------------------------------------------------------------------------
#include "SparseLinearAlgebra_templCode.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "Exceptions.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



template<typename FloatType, typename IntType>
void geoxl::scatterAndAddMatrix(
	const DynamicMatrix<FloatType> &dense, 
	const DynamicVector<IntType> &indicesMapping, 
	SparseMatrix<FloatType> &dest,
	bool monotonousAscIndices) 
{
	if (!dense.getColumns() == dense.getRows()) {
		throw ERangeCheck("scatterAndAddMatrix() - currently only square matrices are supported.");
	}
	if (!dense.getColumns() == indicesMapping.getDim()) {
		throw ERangeCheck("scatterAndAddMatrix() - indices vector does not match matrix dimension.");
	}
	if (indicesMapping.getDim() == 0) {
		throw ERangeCheck("scatterAndAddMatrix() - zero size input is not possible.");
	}
	const IntType n = indicesMapping.getDim();
	for (IntType row=0; row<n; row++) {
		SparseVector<FloatType> rowVec;
		if (monotonousAscIndices) {
			for (IntType column=0; column<n; column++) {
				rowVec.pushBackEntry(indicesMapping[column], dense[column][row]);
			}
		} else {
			for (IntType column=0; column<n; column++) {
				rowVec.setEntry(indicesMapping[column], dense[column][row]);
			}
		}
		dest[indicesMapping[row]] += rowVec;
	}
}


template<typename FloatType, typename IntType>
void geoxl::scatterAndAddDenseMatrixByName(SparseMatrix<FloatType> &dest, const DynamicMatrix<FloatType> &dense, const VAR_AAT &vaat, IntType elementIndex) {
	const card32 n = dense.getRows();
   DynamicVector<IntType> indicesMapping(n);
   pAssert(n <= vaat.getNumElements());
   for (card32 i=0; i<n; i++) {
      indicesMapping[i] = vaat.getOverallNumElements() * elementIndex + vaat.getIndexOffset() + i;
   }
   scatterAndAddMatrix(dense, indicesMapping, dest, true);
}

template<typename FloatType, typename IntType>
void geoxl::scatterAndAddVectorByName(DynamicVector<FloatType> &dest, const DynamicVector<FloatType> &source, const VAR_AAT &vaat, IntType elementIndex) {
	const card32 n = source.getDim();
   if (n > vaat.getNumElements()) {
      throw ERangeCheck("scatterAndAddVectorByName: source is larger than dest channel (numElements).");
   }
   if (elementIndex*vaat.getOverallNumElements() + vaat.getIndexOffset() + n >= dest.getDim()+1 ) {
      throw ERangeCheck("scatterAndAddVectorByName: dest is smaller than largest target index.");
   }
   for (card32 i=0; i<n; i++) {
      dest[vaat.getOverallNumElements() * elementIndex + vaat.getIndexOffset() + i] += source[i];
   }
}

template <typename FloatType>
void geoxl::convertDenseToSparse(const DynamicMatrix<FloatType> &source, SparseMatrix<FloatType> &dest) {
	const card32 rows = source.getRows();
	const card32 columns = source.getColumns();
	dest.clear();
	dest.setRows(rows);
	for (card32 r=0; r<rows; r++) {
		for (card32 c=0; c<columns; c++) {
			dest[r].pushBackEntry(c, source[c][r]);
		}
	}
}


