//---------------------------------------------------------------------------
#ifndef VariablesNamingH
#define VariablesNamingH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
#include "SparseLinearAlgebra.h"
//---------------------------------------------------------------------------
#include <vector>
#include <string>
//---------------------------------------------------------------------------


// ---- this module allows to access variables in vectors (StaticVector, DynamicVector, SparseVector)
// ---- and matrices (StaticMatrix, DynamicMatrix, SparseMatrix) by symbolic names
//
// the layout of a variable stream with variables named a and b of length 3 and 2 looks like this:
// a[0][0] a[0][1] a[0][2] b[0][0] b[0][1]   a[1][0] a[1][1] a[1][2] b[1][0] b[1][1]   a[2][0] a[2][...
// all values are packed into the vector, occupying a <FloatType> each
// in contrast to VertexDescriptor, no varying attribute types are supported (for obvious reasons)


namespace geoxl {
	using namespace std;

class VariablesNamingScheme;


/// symbolic name for a part of a variable
class ANIMATIONRECONSTRUCTION_API VariablesAttributeDescriptor {
 public:
	string name;
	mpcard numElements;
 VariablesAttributeDescriptor() {}
 VariablesAttributeDescriptor(string n, mpcard dim) {name = n; numElements = dim;}
};
typedef VariablesAttributeDescriptor VAR_AD;


/// attribute access token, describes offset (and length) of a subvariable in a stream of values
class ANIMATIONRECONSTRUCTION_API VariablesAAT {
 private:
	/// offset of the attribute in one variable set
	mpcard indexOffset;
	/// num elements in this attribute
	mpcard numElements;
	/// overall number of elements in the variable set
	mpcard overallNumElements;
	
 public:
	inline mpcard getIndexOffset() const {return indexOffset;}
	inline mpcard getNumElements() const {return numElements;}
	inline mpcard getOverallNumElements() const {return overallNumElements;}
 friend VariablesNamingScheme;
};

typedef VariablesAAT VAR_AAT;


/// this is the main class in this module - it describes a stream of a reoccuring set of multi-dimensional variables
/// access to subvalues is provided by AATs (similar to VertexDescriptors for point sets)
class ANIMATIONRECONSTRUCTION_API VariablesNamingScheme {
 private:
	vector<VariablesAttributeDescriptor> attributes;
	vector<mpcard> offsets;
	
	/// returns -1 if not found
	mpint getIndex(string name) const;
	
 public:
	/// add an attribute
	void			 addAttrib(VariablesAttributeDescriptor attrib);
	/// get a VAR_AAT by name
	VariablesAAT getAAT(string name) const;
};




/// access dynamic vector components by names and indices
/// aat - component (variable) name
/// varIndex - index of the specific variable (eg. position[2<--this index])
/// streamIndex - index of record (which of the subsequent variable sets, similar to point index in point sets)
template <class FloatType>
class NamedDynamicVector {
 private:
	DynamicVector<FloatType> &op;
 public:
	NamedDynamicVector(DynamicVector<FloatType> &init_op);
	inline		 FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned streamIndex);
	inline const FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned streamIndex) const;
};

/// access sparse vector components by names and indices
/// --- attention: access by names is possibly not efficient (O(n) for n entries in sparse vector)
/// aat - component (variable) name
/// varIndex - index of the specific variable (eg. position[2<--this index])
/// streamIndex - index of record (which of the subsequent variable sets, similar to point index in point sets)
template <class FloatType>
class NamedSparseVector {
 private:
	SparseVector<FloatType> &op;
 public:
	NamedSparseVector(SparseVector<FloatType> &init_op);
	inline		 FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned index);
	inline const FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned index) const;
};


/// access dynamic matrix components by names and indices
/// aat - component (variable) name
/// varIndex - index of the specific variable (eg. position[2<--this index])
/// streamIndex - index of record (which of the subsequent variable sets, similar to point index in point sets)
template <class FloatType>
class NamedDynamicMatrix {
 private:
	DynamicMatrix<FloatType> &op;
 public:
	NamedDynamicMatrix(DynamicMatrix<FloatType> &init_op);
	inline		 DynamicVector<FloatType> &getRow(VAR_AAT aat, unsigned varIndex, unsigned row);
	inline const DynamicVector<FloatType> &getRow(VAR_AAT aat, unsigned varIndex, unsigned row) const;
	inline		 FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned row, unsigned column);
	inline const FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned row, unsigned column) const;
};


/// access sparse vector components by names and indices
/// --- attention: access by names via "get" is possibly not efficient (O(n) for n entries in sparse vector)
///                getRow(), however, is O(1)
/// aat - component (variable) name
/// varIndex - index of the specific variable (eg. position[2<--this index])
/// streamIndex - index of record (which of the subsequent variable sets, similar to point index in point sets)
template <class FloatType>
class NamedSparseMatrix {
 private:
	SparseMatrix<FloatType> &op;
 public:
	NamedSparseMatrix(SparseMatrix<FloatType> &init_op);
	inline		 SparseVector<FloatType> &getRow(VAR_AAT aat, unsigned varIndex, unsigned row);
	inline const SparseVector<FloatType> &getRow(VAR_AAT aat, unsigned varIndex, unsigned row) const;
	inline		 FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned row, unsigned column);
	inline const FloatType &get(VAR_AAT aat, unsigned varIndex, unsigned row, unsigned column) const;
};


}

#endif
