#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "CentroidDescriptor.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointCloudTools.h"
#include "PCA.h"
#include "opencv/cv.h"
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( CentroidDescriptor ,0)
{
	BEGIN_CLASS_INIT( CentroidDescriptor );

	ADD_CARD32_PROP( m_NumScales, 0 );
	ADD_CARD32_PROP( m_NumSamplesPerScale, 0 );
	ADD_FLOAT32_PROP( m_StartSigma, 0 );
	ADD_FLOAT32_PROP( m_BaseSigma, 0 );
	ADD_FLOAT32_PROP( m_WindowSize, 0 );
	ADD_FLOAT32_PROP( m_WindowSigma, 0 );

	ADD_STRING_VAR_ARRAY_PROP( m_Attribute, 0 );
}

CentroidDescriptor::CentroidDescriptor(void)
{
	m_NumScales = 1;
	m_NumSamplesPerScale = 4;
	m_StartSigma = 0.05f;
	m_BaseSigma = 2;

	m_WindowSize = 10.0f;
	m_WindowSigma = 2.0f;

	m_Attribute.push_back( "curvature" );

	m_hIter = NULL;
}

CentroidDescriptor::~CentroidDescriptor(void)
{
	delete m_hIter;
}

void CentroidDescriptor::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const CentroidDescriptor * o = dynamic_cast<const CentroidDescriptor*>(obj);
	pAssert( o != NULL );

	copyObjectProperties( obj, this );
}

static bool testRad = true;

CentroidSignature *CentroidDescriptor::createSignature(const Vector3f &pos) 
{
	if( m_hIter == NULL )
		throw "Use beginSignatureCalculations() before creating signatures!";

	CentroidSignature * sig = new CentroidSignature;

	sig->setupZero(m_NumScales,m_NumSamplesPerScale * (card32)m_Attribute.size());
	m_hIter->setSeekPointAndReset(pos);
	Vector3f normal = m_hIter->get3f( m_NormalAAT );
	vector<float> attributes(m_Attribute.size());
	float maxRadius = m_WindowSize * m_CurrentMedianPointDistance;
	float spatialSigma = m_WindowSigma * m_CurrentMedianPointDistance;
	float inv_spatialSigma_2 = 1.0f / spatialSigma / spatialSigma;
	float maxRad_2 = maxRadius * maxRadius;

	if( testRad )
	{
		debugRenderer->addSphere(pos, maxRadius,makeVector3f(0,0.6f,1.0f));
		testRad = false;
	}


	PCA<float,3> pca;
	sig->m_SummedAreaWeight = 0;
	while( !m_hIter->atEnd() )
	{
		Vector3f sample_pos = m_hIter->get3f(m_PosAAT);
		if( norm( sample_pos - pos) > maxRadius )
			break;
		pca.addPoint(sample_pos);

		/*for(card32 i=0;i<attributes.size();i++)
		{
			attributes[i] = m_hIter->get1f( m_AttributeAAT[i] );
		}*/
		Vector4f tmp = m_hIter->get4f( m_AttributeAAT[0] );
		attributes[0] = tmp[2];
		//attributes[1] = tmp[1];

		sample_pos = (sample_pos - pos);
		float pos_weight = exp(-max(0.0f,normQuad(sample_pos)-maxRad_2) * inv_spatialSigma_2);
		sig->m_SummedAreaWeight += 1.0f;

		for( card32 i=0;i<m_NumScales;i++ )
		{
			float sigma = m_StartSigma * pow( m_BaseSigma, -(float)i);
			float inv_sigma_2 = 1.0f / pow( sigma,2.0f);
			for( card32 j=0;j<m_NumSamplesPerScale;j++ )
			{
				for( card32 k=0;k<attributes.size();k++ )
				{
					float a0 = ((float)j+0.5f)/(float)m_NumSamplesPerScale;
					float attribute_weight = exp( -pow(attributes[k] - a0,2.0f) *inv_sigma_2);

					card32 index = (i * m_NumSamplesPerScale + j)* (card32)attributes.size() + k;
					sig->m_Centroids[index] += sample_pos * attribute_weight * pos_weight;
					sig->m_CentroidWeights[index] += attribute_weight * pos_weight;
				}
			}
		}

		m_hIter->next();
	}

	for( card32 i=0;i<sig->m_Centroids.size();i++ )
	{
		if( sig->m_CentroidWeights[i] > 0 )
			sig->m_Centroids[i] = sig->m_Centroids[i] / sig->m_CentroidWeights[i] + pos;
		else
			sig->m_Centroids[i] = pos;
	}

	// pca normal
	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues, eigenVectors, centroid );
	Vector3f pcaNormal = normalize(eigenVectors[0].crossProduct(eigenVectors[1])) * maxRadius / 0.5f;
	if( pcaNormal * normal < 0)
		pcaNormal = -pcaNormal;
	sig->m_PCANormal = pcaNormal;
	sig->m_FeaturePos = pos;

	return sig;
}

void CentroidDescriptor::beginSignatureCalculations(PointCloud * pc) 
{
	testRad = true;
	delete m_hIter;
	m_hIter = new HierarchicalKNNIterator(pc,32,NULL);
	m_AttributeAAT.clear();

	for( card32 i=0;i<m_Attribute.size();i++ )
	{
		if( !pc->providesAttribute( m_Attribute[i] ) )
			throw "Source Attribute isn't provided!";

		m_AttributeAAT.push_back( pc->getAAT( m_Attribute[i] ) );
	}
	m_PosAAT = pc->getAAT("position");
	m_NormalAAT = pc->getAAT("normal");
	m_CurrentMedianPointDistance = getMedianPointDistance(pc);
}

void CentroidDescriptor::endSignatureCalculations() 
{
	delete m_hIter;
	m_hIter = NULL;
}


// ******************************************************************
//						CentroidSignature
// ******************************************************************

IMPLEMENT_GEOXL_CLASS( CentroidSignature ,0)
{
	BEGIN_CLASS_INIT( CentroidSignature );
}

CentroidSignature::CentroidSignature(void)
{
}

CentroidSignature::~CentroidSignature(void)
{
}

void CentroidSignature::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const CentroidDescriptor * o = dynamic_cast<const CentroidDescriptor*>(obj);
	pAssert( o != NULL );

	copyObjectProperties( obj, this );
}


void CentroidSignature::setupZero(card32 numScales, card32 numSamplesPerScale)
{
	m_Centroids = vector<Vector3f>(numScales*numSamplesPerScale,makeVector3f(0,0,0));
	m_CentroidWeights = vector<float>(numScales*numSamplesPerScale,0);
}


Vector4f CentroidSignature::calculateStabilityInformation() 
{
	PCA<float,3> pca;
	for( card32 i=0;i<m_Centroids.size();i++ )
	{
		pca.addPoint( (m_Centroids[i] - m_FeaturePos)*m_CentroidWeights[i]  );
	}
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	pca.analyze( eigenValues, eigenVectors, centroid );

	Vector4f ret = makeVector4f(
		eigenValues[0]*eigenValues[1],
		eigenValues[0],
		eigenValues[0]*eigenValues[1]*eigenValues[2],
		0
		);

	return ret;
}

Matrix4f CentroidSignature::calculateTransformation(CentroidSignature * sourceSignature, bool usePCANormal)
{
	debugRenderer->beginRenderJob("test");
	for( unsigned i=0;i<m_Centroids.size()-1;i++)
	{
		Vector3f col = makeVector3f(rnd01(),rnd01(),rnd01());
		debugRenderer->addLine(m_Centroids[i],m_Centroids[i+1], col, col, 2);
		debugRenderer->addLine(sourceSignature->m_Centroids[i],sourceSignature->m_Centroids[i+1], col, col, 2);
	}
	debugRenderer->endRenderJob();
	vector<Vector4f> centroidsA, centroidsB;
	for( card32 i=0;i<m_Centroids.size();i++ )
	{
		Vector4f c1 = expand3To4(m_Centroids[i] );
		c1[3] = m_CentroidWeights[i];
		centroidsA.push_back( c1 );
		Vector4f c2 = expand3To4(sourceSignature->m_Centroids[i] );
		c2[3] = sourceSignature->m_CentroidWeights[i];
		centroidsB.push_back( c2 );
	}

	if( usePCANormal )
	{
		for( card32 i=0;i<m_Centroids.size();i++ )
		{
			Vector4f c1 = expand3To4(m_Centroids[i]+m_PCANormal );
			c1[3] = m_CentroidWeights[i];
			centroidsA.push_back( c1 );
			Vector4f c2 = expand3To4(sourceSignature->m_Centroids[i]+sourceSignature->m_PCANormal );
			c2[3] = sourceSignature->m_CentroidWeights[i];
			centroidsB.push_back( c2 );
		}

	}

	// estimate centroids of centroids :-)
	Vector3f centroidA = NULL_VECTOR3F, centroidB = NULL_VECTOR3F;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		centroidA += shrink4To3(centroidsA[i]);
		centroidB += shrink4To3(centroidsB[i]);
	}
	centroidA /= centroidsA.size();
	centroidB /= centroidsB.size();
	
	// calculate cross-covariance matrix
	Matrix3f C;
	C *= 0;
	float summedWeight = 0;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				if( min(centroidsA[i][3],centroidsB[i][3]) > 10.0f && min(centroidsA[i][3],centroidsB[i][3])/max(centroidsA[i][3],centroidsB[i][3])>0.9f)
				{
					summedWeight += 1;
					C[j][k] += (centroidsA[i][j] - centroidA[j])
						* (centroidsB[i][k] - centroidB[k]);
				}
			}
		}
	}
	C /=summedWeight;

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f ret = makeTranslation4f(centroidA) * expand3To4(rotation) * makeTranslation4f(-centroidB);

	return ret;
}

float CentroidSignature::calculateDistance( CentroidSignature * sig)
{
	float ret = 0;
	for( card32 i=0;i<m_Centroids.size()-1;i++)
	{
		for( card32 j=i+1;j<m_Centroids.size();j++)
		{
			ret += pow( norm(m_Centroids[i]-m_Centroids[j]) - norm(sig->m_Centroids[i]-sig->m_Centroids[j]),2.0f ) 
				* (m_CentroidWeights[i]+m_CentroidWeights[j]);
		}

	}

	return sqrt(ret) / (m_SummedAreaWeight+sig->m_SummedAreaWeight);
}

Vector4f CentroidSignature::calculateDistanceVector( CentroidSignature * sig)
{
	Vector4f ret = NULL_VECTOR4F;
	for( card32 i=0;i<m_Centroids.size()-1;i++)
	{
		for( card32 j=i+1;j<m_Centroids.size()-1;j++)
		{
			ret[0] += pow( norm(m_Centroids[i]-m_Centroids[j]) - norm(sig->m_Centroids[i]-sig->m_Centroids[j]),2.0f ) 
				* max(m_CentroidWeights[i]+m_CentroidWeights[j],sig->m_CentroidWeights[i]+sig->m_CentroidWeights[j]);
			ret[1] += fabs( norm(m_Centroids[i]-m_Centroids[j]) - norm(sig->m_Centroids[i]-sig->m_Centroids[j]) ) 
				* (m_CentroidWeights[i]+m_CentroidWeights[j])*(sig->m_CentroidWeights[i]+sig->m_CentroidWeights[j]);
		}
		ret[2] += fabs(m_CentroidWeights[i] - sig->m_CentroidWeights[i]);
		ret[3] += sqr(m_CentroidWeights[i] - sig->m_CentroidWeights[i]);

	}

	ret[0] = sqrt(ret[0]);
	ret[3] = sqrt(ret[3]);
	return ret;
}

Matrix4f CentroidSignature::calculateTransformation( vector<CentroidSignature*> sourceSignatures, vector<CentroidSignature*> destSignatures, bool usePCANormal,float descSize )
{
	vector<Vector4f> centroidsA, centroidsB;
	for( unsigned s=0;s<sourceSignatures.size();s++ )
	{
		for( card32 i=0;i<destSignatures[s]->m_Centroids.size();i++ )
		{
			Vector4f c1 = expand3To4(destSignatures[s]->m_Centroids[i] );
			c1[3] = destSignatures[s]->m_CentroidWeights[i];
			centroidsA.push_back( c1 );
			Vector4f c2 = expand3To4(sourceSignatures[s]->m_Centroids[i] );
			c2[3] = sourceSignatures[s]->m_CentroidWeights[i];
			centroidsB.push_back( c2 );
		}
		if( usePCANormal )
		{
			for( card32 i=0;i<destSignatures[s]->m_Centroids.size();i++ )
			{
				Vector4f c1 = expand3To4(destSignatures[s]->m_Centroids[i]+destSignatures[s]->m_PCANormal*descSize );
				c1[3] = destSignatures[s]->m_CentroidWeights[i];
				centroidsA.push_back( c1 );
				Vector4f c2 = expand3To4(sourceSignatures[s]->m_Centroids[i]+sourceSignatures[s]->m_PCANormal*descSize );
				c2[3] = sourceSignatures[s]->m_CentroidWeights[i];
				centroidsB.push_back( c2 );
			}

		}
	}

	// estimate centroids of centroids :-)
	Vector3f centroidA = NULL_VECTOR3F, centroidB = NULL_VECTOR3F;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		centroidA += shrink4To3(centroidsA[i]);
		centroidB += shrink4To3(centroidsB[i]);
	}
	centroidA /= centroidsA.size();
	centroidB /= centroidsB.size();

	// calculate cross-covariance matrix
	Matrix3f C;
	C *= 0;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (centroidsA[i][j] - centroidA[j])
					* (centroidsB[i][k] - centroidB[k]) * min(centroidsA[i][3],centroidsB[i][3]);
			}
		}
	}
	C /= centroidsA.size();

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f ret = makeTranslation4f(centroidA) * expand3To4(rotation) * makeTranslation4f(-centroidB);

	return ret;

}

float CentroidSignature::computeHistogramDistance( CentroidSignature * sig )
{
	float dist = 0;
	for( unsigned i=0;i<m_CentroidWeights.size();i++ )
	{
		dist += pow((m_CentroidWeights[i]-sig->m_CentroidWeights[i])/m_SummedAreaWeight,2.0f);
	}
	dist = sqrt(dist);
	return dist;
}