#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "CmdObjPointCloudDistortion.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "BooleanClassProperty.h"
#include "StringClassProperty.h"
#include "NormalDistribution.h"
#include "PointCloudTools.h"
#include "FixedArrayClassProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( CmdObjPointCloudDistortion ,1)
{
	BEGIN_CLASS_INIT( CmdObjPointCloudDistortion );

	ADD_CARD32_PROP(m_DistortionMode,0);
	ADD_FLOAT32_PROP(m_Strength,0);
	ADD_CARD32_PROP(m_RandomSeed,0);
	ADD_CARD32_PROP(m_NumGravityBalls,0);

	ADD_VECTOR3F_PROP(m_RotationCenter,1);
	ADD_VECTOR3F_PROP(m_RotationAxis,1);

	ADD_FLOAT32_FIXED_ARRAY_PROP(m_PointCloudBB,1,6,3);
}

CmdObjPointCloudDistortion::CmdObjPointCloudDistortion(void)
{
	m_DistortionMode = DM_Noise;
	m_Strength = 0.01f;
	m_RandomSeed = 15091980;
	m_NumGravityBalls = 10;

	setSelectionMode( SM_Selected );
	setSourceChannel( "ret_position" );
	setDestChannel( "position" );
	setCreateHierarchicalIterator( false );
	setInvalidateAttachments( true );

	m_PointCloudBB.lowerCorner = makeVector3f(-1,-1,-1);
	m_PointCloudBB.upperCorner = makeVector3f(1,1,1);

	m_RotationCenter = makeVector3f(0,0,0);
	m_RotationAxis = makeVector3f(0,0,1);

}

CmdObjPointCloudDistortion::~CmdObjPointCloudDistortion(void)
{
}

void CmdObjPointCloudDistortion::assign(const Object* obj, CopyContext *context)
{
	ProcessPointsCmdObj::assign(obj,context);
	const CmdObjPointCloudDistortion * o = dynamic_cast<const CmdObjPointCloudDistortion*>(obj);
	pAssert( o != NULL );

	m_DistortionMode = o->m_DistortionMode;
	m_Strength = o->m_Strength;
	m_RandomSeed = o->m_RandomSeed;

	m_PointCloudBB = o->m_PointCloudBB;
	m_RotationCenter = o->m_RotationCenter;
	m_RotationAxis = o->m_RotationAxis;

}

string CmdObjPointCloudDistortion::getDescription() 
{
	string ret = "PC Distortion";
	switch( m_DistortionMode )
	{
	case DM_Noise:
		ret += " - Noise";
		break;
	case DM_Twist:
		ret += " - Twist";
		break;
	case DM_GravityBalls:
		ret += " - GravityBalls";
		break;
	case DM_Rotate:
		ret += " - Rotate";
		break;
	}
	return ret;
}

struct TDistortionContext
{
	BoundingBox3f	bb;
};

void CmdObjPointCloudDistortion::doAction(Object *context) 
{
	Scene *scene = dynamic_cast<Scene*>(context); pAssert(scene);

	// define required channels
	m_RequiredChannels.clear();
	m_RequiredChannels.push_back( VAD( m_SourceChannel, 3, VAD::DATA_FORMAT_FLOAT32 ) );
	m_RequiredChannels.push_back( VAD( m_DestChannel, 3, VAD::DATA_FORMAT_FLOAT32 ) );

	TDistortionContext dc;
	dc.bb = m_PointCloudBB;

	processAllPoints( scene, &dc, (ProcessPointsCmdObj::ProcessPointMethod)&CmdObjPointCloudDistortion::prepareDistortion,
		(ProcessPointsCmdObj::ProcessPointMethod)&CmdObjPointCloudDistortion::processDistortion );
}

void CmdObjPointCloudDistortion::prepareDistortion(PointInfo* info, void *context) 
{
	TDistortionContext * c = (TDistortionContext*)context;

	srand( m_RandomSeed );
	if( m_DistortionMode == DM_GravityBalls )
	{
		m_GravityBalls.resize( m_NumGravityBalls );
		for( card32 i=0;i<m_GravityBalls.size();i++ )
		{
			m_GravityBalls[i].pos = c->bb.lowerCorner + (makeVector3f( rand(), rand(), rand() ) 
				/ (float)RAND_MAX).componentProduct(c->bb.upperCorner - c->bb.lowerCorner);
			m_GravityBalls[i].strength = ((float)rand()/(float)RAND_MAX + 1.0f);
			if( rand() % 2 == 0 )
				m_GravityBalls[i].strength = -m_GravityBalls[i].strength;
			m_GravityBalls[i].strength *= m_Strength;
		}
	}

}

void CmdObjPointCloudDistortion::processDistortion(PointInfo* info, void *context) 
{
	TDistortionContext * c = (TDistortionContext*)context;
	switch( m_DistortionMode )
	{
	case DM_Noise:
		{
			Vector3f attr = info->bIt->get3f( info->sourceAAT );
			for( card32 i=0;i<3;i++ )
			{
				float ampl = m_Strength * norm( c->bb.upperCorner - c->bb.lowerCorner );
				if( rand() % 2 == 0 )
					ampl = -ampl;
				attr[i] += ampl * log( (float)rand() / RAND_MAX );
			}
			info->mIt->set3f( info->destAAT, attr );
		}
		break;
	case DM_Twist:
		{
			Vector3f attr = info->bIt->get3f( info->sourceAAT );
			int rot_axis = 1;
			int u_axis = 0;
			int v_axis = 2;
			Vector3f rot_vector = makeVector3f(0,1,0);
			
			float angle = (attr[rot_axis] - c->bb.lowerCorner[rot_axis]) / 
				c->bb.getSideLength( rot_axis ) * m_Strength;

			Vector3f center = m_RotationCenter;
			Matrix3f rotAxisMatrix = makeRotVector3f( m_RotationAxis, angle );
			attr -= center;
			attr = rotAxisMatrix * attr;
			attr += center;
			info->mIt->set3f( info->destAAT, attr );
		}
		break;
	case DM_GravityBalls:
		{
			Vector3f attr = info->bIt->get3f( info->sourceAAT );
			Vector3f dir = makeVector3f(0,0,0);
			for( card32 i=0;i<m_GravityBalls.size();i++ )
			{
				Vector3f distanceVec = m_GravityBalls[i].pos - attr;
				float dist = norm(distanceVec);

				float mult = exp( - dist/norm( c->bb.upperCorner - c->bb.lowerCorner ) );


				distanceVec.normalize();
				distanceVec *= mult * m_GravityBalls[i].strength / norm( c->bb.upperCorner - c->bb.lowerCorner );
				dir += distanceVec;
			}

			attr += dir;
			info->mIt->set3f( info->destAAT, attr );

		}
		break;
	case DM_Rotate:
		{
			Vector3f attr = info->bIt->get3f( info->sourceAAT );
			Vector3f normal = info->bIt->get3f( info->normalAAT );
			
			float angle = m_Strength;

			Vector3f center = m_RotationCenter;
			Matrix3f rotAxisMatrix = makeRotY3f( angle );
			attr -= center;
			attr = rotAxisMatrix * attr;
			attr += center;
			normal = rotAxisMatrix * normal;
			info->mIt->set3f( info->destAAT, attr );
			info->mIt->set3f( info->normalAAT, normal );
		}
		break;
	}
}
