#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DescriptorClustering.h"
//---------------------------------------------------------------------------
#include "Timer.h"
#include "KMeansClustering.h"
#include "KMeansClustering_templCode.h"
#include "FDAttribute.h"
#include "CopyObjectProperties.h"
#include "PGDescriptorFactory.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( DescriptorClustering ,0)
{
	BEGIN_CLASS_INIT( DescriptorClustering );

	ADD_OBJECT_PROP( m_DescriptorFactory, 0, PGDescriptorFactory::getClass(),true);
	ADD_CARD32_PROP( m_NumClusters, 0);
	ADD_CARD32_PROP( m_NumCells, 0);
}

DescriptorClustering::DescriptorClustering(void)
{
	m_DescriptorFactory = new PGDescriptorFactory;
	m_DescriptorFactory->setupHistogramCurvature();
	m_NumClusters = 8;
	m_NumCells = 32;
}

DescriptorClustering::~DescriptorClustering(void)
{
	delete m_DescriptorFactory;
}

void DescriptorClustering::assign(const Object* obj, CopyContext *context)
{
	FeatureDetector::assign(obj,context);
	const DescriptorClustering * o = dynamic_cast<const DescriptorClustering*>(obj);
	pAssert( o != NULL );

	copyObjectProperties(obj,this);

}
vector<FeatureList*> DescriptorClustering::detectFeatures( vector<PointCloud*> pcs, float radius ) 
{
	// build compressed descriptors
	m_DescriptorFactory->setDescriptorSize(radius);
	m_DescriptorFactory->computeDescriptors(pcs, "color", 3);


	// cluster descriptors in descriptor-space, store result in clusterID
	Timer timer;
	string descChan = "color";
	KMeansClustering<float> kmeans;
	card32 numClusters = m_NumClusters;
	card32 num_cells = m_NumCells;
	map<card32,card32> histogram;
	vector<FeatureList*> ret;

	// estimate boundingbox in descriptor-space
	BoundingBox3f bb;
	for( card32 i=0;i<pcs.size();i++ )
	{
		UnstructuredInCorePointCloud * pc = (UnstructuredInCorePointCloud*)pcs[i];
		PointSet * ps = pc->getPointSet();
		AAT descAAT = ps->getAAT( descChan );
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f desc = ps->get3f(j, descAAT );
			if( i == 0 && j == 0 )
				bb.lowerCorner = bb.upperCorner = desc;
			else
				bb.addPoint( desc );
		}
	}

	bb.makeBoundingCube();

	
	// preclustering
	for( card32 i=0;i<pcs.size();i++ )
	{
		UnstructuredInCorePointCloud * pc = (UnstructuredInCorePointCloud*)pcs[i];
		checkAttribute( pc,"clusterID", 1, VAD::DATA_FORMAT_INT32);
		checkAttribute( pc,"clusterColor", 3, VAD::DATA_FORMAT_FLOAT32);
		PointSet * ps = pc->getPointSet();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		if( !ps->providesAttribute( descChan ) )
		{
			message( "calculate descriptors before using descriptor clustering" );
			return ret;
		}

		AAT descAAT = ps->getAAT( descChan );
		
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f desc = ps->get3f(j, descAAT );
			desc -= bb.lowerCorner;
			desc /= bb.getSideLength(0);
			ps->set3f(j, descAAT, desc );
			desc = desc * num_cells;
			Vector3i desc_i = makeVector3i( min(num_cells-1,desc[0]), 
				min(num_cells-1,desc[1]),
				min(num_cells-1,desc[2]) );
			card32 index = (desc_i[0] * num_cells + desc_i[1]) * num_cells + desc_i[2];
			histogram[index]++;
		}


	}

	for( map<card32,card32>::iterator iter = histogram.begin(); iter != histogram.end(); ++iter )
	{
		card32 index = iter->first;
		card32 num = iter->second;
		Vector3f cell_midpoint = makeVector3f( index / num_cells / num_cells, (index / num_cells) % num_cells, index % num_cells ) + 
			makeVector3f(0.5f, 0.5f, 0.5f);
		cell_midpoint = cell_midpoint / num_cells;
		DVectorF vec;
		vec.setDim(3);
		vec[0] = cell_midpoint[0];
		vec[1] = cell_midpoint[1];
		vec[2] = cell_midpoint[2];
		kmeans.add( vec, num );
	}

	vector< DVectorF > codebook;
	vector< card32 > indices;
	kmeans.calcEx( numClusters, codebook, indices, 0.00001f, 10 );

	card32 index = 0;
	for( card32 i=0;i<pcs.size();i++ )
	{
		UnstructuredInCorePointCloud * pc = (UnstructuredInCorePointCloud*)pcs[i];
		PointSet * ps = pc->getPointSet();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		if( !ps->providesAttribute( descChan ) )
		{
			message( "calculate descriptors before using descriptor clustering" );
			return ret;
		}

		AAT descAAT = ps->getAAT( descChan );
		AAT clusterColorAAT = ps->getAAT( "clusterColor" );
		card32 clusterColor[8];
		clusterColor[0] = 0x002200;
		clusterColor[1] = 0x004400;
		clusterColor[2] = 0x006600;
		clusterColor[3] = 0x008800;
		clusterColor[4] = 0x00AA00;
		clusterColor[5] = 0x00CC00;
		clusterColor[6] = 0x00FF00;
		clusterColor[7] = 0x000000;
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f desc = ps->get3f( j, descAAT );
			int32 cluster = 0;
			DVectorF vec;
			vec.setDim(3);
			vec[0] = desc[0];
			vec[1] = desc[1];
			vec[2] = desc[2];
			float minDist = 1e20f;
			for( card32 k=0;k<codebook.size();k++ )
			{
				float tmpDist = norm( codebook[k] - vec );
				if( tmpDist < minDist )
				{
					minDist = tmpDist;
					cluster = k;
				}
			}

			index++;
			ps->set1i( j, clusterAAT, cluster );
			Vector3f col = makeVector3f( clusterColor[cluster] & 0xff,
				(clusterColor[cluster] >> 8) & 0xff,
				(clusterColor[cluster] >> 16) & 0xff ) / 255;
			ps->set3f( j, clusterColorAAT, col );
		}
	}

	// cluster descriptors 
	for( card32 i=0;i<pcs.size();i++ )
	{
		UnstructuredInCorePointCloud * pc = (UnstructuredInCorePointCloud*)pcs[i];
		FixedKNNTopologyIterator topologyIter( pc->getPointSet(), pc->getAttachments(), pc, 8 );

		//InCorePCTopologyGraph * top = pgc->getPatchGraph(i)->getTopology();
		PointSet * ps = pc->getPointSet();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		AAT posAAT = ps->getAAT( "position" );
		AAT descAAT = ps->getAAT( descChan );
		HierarchicalKNNIterator hIt( pc, 32, NULL );

		FeatureList * features = new FeatureList;

		vector<int> visited( ps->getNumEntries(), 0 );
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			if( visited[j] != 0 )
				continue;

			int32 clusterID = ps->get1i( j, clusterAAT );

			// start clustering
			list<int> queue;
			vector<int> clusterIndices;
			vector<Vector3f> clusterPoints;
			queue.push_back(j);
			visited[j] = 1;
			clusterIndices.push_back( j );
			clusterPoints.push_back( ps->get3f(j, posAAT ) );
			Vector3f mean_desc = ps->get3f(j, descAAT );

			while( !queue.empty() )
			{
				int node = queue.front();
				queue.pop_front();

				topologyIter.goToIndex( node );
				for( card32 n=0;n<topologyIter.getNumNeighbours();n++ )
				{
					int32 neighbour = topologyIter.getNeighbour(n);
					if( ps->get1i(neighbour, clusterAAT ) != clusterID )
						continue;

					if( visited[neighbour] == 0 )
					{
						queue.push_back( neighbour );
						visited[neighbour] = 1;
						clusterIndices.push_back( neighbour );
						clusterPoints.push_back( ps->get3f(neighbour, posAAT ) );
						mean_desc += ps->get3f(neighbour, descAAT );
					}

				}

			}

			if( clusterIndices.size() < 4 )
				continue;

			mean_desc /= clusterIndices.size();

			
			/*vector<DVectorF> points;
			for( card32 k=0;k<clusterPoints.size();k++ )
			{
				DVectorF vec;
				vec.setDim(3);
				for( card32 l=0;l<3;l++ )
					vec[l] = clusterPoints[k][l];
				points.push_back( vec );
			}
			DynamicPCA<float32> pca( &points[0], points.size() );
			DynamicVector<float32> mean = pca.calcMean();
			Vector3f centroid = makeVector3f( mean[0],mean[1],mean[2] );*/
			PCA<float,3> pca;
			for( card32 k=0;k<clusterPoints.size();k++ )
			{
				pca.addPoint( clusterPoints[k] );
			}
			Vector3f centroid, eigenValues;
			Matrix3f eigenVectors;
			pca.analyze( eigenValues, eigenVectors, centroid );
			
			//DynamicVector<float32> eigenValues = pca.calcEigenValues();
			vector<float> signature;
			signature.push_back( fabs( eigenValues[0] ) );
			signature.push_back( fabs( eigenValues[1] ) );
			signature.push_back( mean_desc[0] );
			signature.push_back( mean_desc[1] );
			signature.push_back( mean_desc[2] );
			signature.push_back( clusterID );
			//norm( eigenValues ), sqrt( absf( eigenValues[1] / max(eigenValues[2],1e-10f) )));
			FDAttribute * fa = new FDAttribute;
			fa->setNumComponents( (card32)signature.size() );
			fa->setupDescriptorData( &signature[0], (card32)signature.size() );

			/*// check if centroid is near the surface
			hIt.setSeekPointAndReset( centroid );
			if( norm( centroid - hIt.get3f( posAAT ) ) >  averagePointDistance * 1.5f )
				continue;*/

			Feature * f = new Feature;
			f->setPosition( centroid );
			f->setPointID(-1);
			f->addDescriptor( fa );
			f->setSize( fabs( eigenValues[0] ) );
			vector<float> fdesc;
			fdesc.push_back( clusterID );
			card32 clusterColor[8];
			/*clusterColor[0] = 0xFF0000;
			clusterColor[1] = 0xFFFF00;
			clusterColor[2] = 0x00FFFF;
			clusterColor[3] = 0x0000FF;
			clusterColor[4] = 0xFF00FF;
			clusterColor[5] = 0xFFFFFF;
			clusterColor[6] = 0x000000;
			clusterColor[7] = 0x00FF00;*/
			
			clusterColor[0] = 0x002200;
			clusterColor[1] = 0x004400;
			clusterColor[2] = 0x006600;
			clusterColor[3] = 0x008800;
			clusterColor[4] = 0x00AA00;
			clusterColor[5] = 0x00CC00;
			clusterColor[6] = 0x00FF00;
			clusterColor[7] = 0x000000;
			f->setColor( clusterColor[clusterID] );
			features->push_back( f );

		}
		ret.push_back(features);
	}

	return ret;
}
