#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDAverage.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "StringHelper.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FDAverage ,0)
{
	BEGIN_CLASS_INIT( FDAverage );
	ADD_CARD32_PROP( m_NumRings, 0 );
	ADD_STRING_PROP( m_Attribute, 0 );
	ADD_CARD32_PROP( m_NumComponents,0);
	ADD_VAR_ARRAY_PROP( DescriptorData, 0, new Float32ClassProperty("element",MetaClass::FIRST_VERSION+0,NULL));
}

FDAverage::FDAverage(void)
{
	m_Attribute = "curvature";
	m_NumRings = 5;
	m_NumComponents = 1;
}

FDAverage::~FDAverage(void)
{
}

void FDAverage::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const FDAverage * o = dynamic_cast<const FDAverage*>(obj);
	pAssert( o != NULL );

	copyObjectProperties( obj, this );
}

template<int dim>
void buildDescriptor( const Vector3f &point, float size, HierarchicalKNNIterator *hIt, AAT posAAT, AAT srcAAT,
	card32 numRings, vector<float> &data )
{
	vector< StaticVector<float,dim> > centroids;
	vector<float> centroids_summedweight(numRings,0);;
	StaticVector<float,dim> nullVector;
	for( card32 i=0;i<dim;i++ )
		nullVector[i] = 0;
	for( card32 i=0;i<numRings;i++ )
		centroids.push_back( nullVector );

	vector<float> ringWeight;
	float summedRingWeight = 0;
	for( card32 i=0;i<numRings;i++ )
	{
		float radius = (float)(i+1) / (float)numRings * size;
		float radius0 = (float)(i) / (float)numRings * size;
		ringWeight.push_back( sqr(radius) - sqr(radius0));
		summedRingWeight += ringWeight.back();
	}
	for( card32 i=0;i<numRings;i++ )
	{
		ringWeight[i] /= summedRingWeight;
	}

	hIt->setSeekPointAndReset(point);
	while( !hIt->atEnd() )
	{
		float dist = norm( point - hIt->get3f(posAAT));
		if( dist > size )
			break;

		StaticVector<float,dim> attribute = hIt->genericGet<float,dim>( srcAAT );

		for( card32 i=0;i<numRings;i++ )
		{
			if( dist < size * (float)(i+1) / (float)(numRings) )
			{
				centroids[i] += attribute;
				centroids_summedweight[i] += 1.0f;
			}
		}

		hIt->next();
	}

	for( card32 i=0;i<numRings;i++ )
	{
		if( centroids_summedweight[i] > 1e-10f )
			centroids[i] /= centroids_summedweight[i];
		else
			centroids[i] = nullVector;
		centroids[i] *= ringWeight[i];

		for( card32 j=0;j<dim;j++ )
			data[i*dim+j] = centroids[i][j];
	}

}

void FDAverage::createDescriptor( const Vector3f &point, float size, PointCloud * pc, TDescriptorContext * context ) 
{
	m_DescriptorData.resize(getNumDescriptorValues(),0);
	HierarchicalKNNIterator hIt(pc, 32, NULL );
	AAT posAAT = pc->getAAT( "position" );
	AAT srcAAT = pc->getAAT( m_Attribute );

	if( m_NumComponents == 0 || m_NumComponents > 4 )
		debugOutput << "FDAverage::m_NumComponents out of range!";


	if( m_NumComponents == 1 )
		buildDescriptor<1>( point, size, &hIt, posAAT, srcAAT, m_NumRings, m_DescriptorData );
		
	if( m_NumComponents == 2 )
		buildDescriptor<2>( point, size, &hIt, posAAT, srcAAT, m_NumRings, m_DescriptorData );

	if( m_NumComponents == 3 )
		buildDescriptor<2>( point, size, &hIt, posAAT, srcAAT, m_NumRings, m_DescriptorData );

	if( m_NumComponents == 4 )
		buildDescriptor<4>( point, size, &hIt, posAAT, srcAAT, m_NumRings, m_DescriptorData );

}


void FDAverage::setNumDescriptorDatas(card32 size) 
{
	m_DescriptorData.resize(size);
}

void FDAverage::deleteDescriptorData( card32 index ) 
{
	m_DescriptorData.erase( m_DescriptorData.begin() + index );
}

char *FDAverage::getDescriptorData( card32 index ) 
{
	return (char*)&m_DescriptorData[index];
}

mpcard FDAverage::getNumDescriptorDatas() 
{
	return m_DescriptorData.size();
}

string FDAverage::getDescriptorDataName(card32 index ) 
{
	return intToStr(index);
}
