#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDClusteredHistogram.h"
//---------------------------------------------------------------------------
#include "KMeansClustering_templCode.h"
#include "PointCloudAnimationSequence.h"
#include "TrajectorySet.h"
#include "PatchGraph.h"
#include "PGDataPointIndices.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FDClusteredHistogram ,0)
{
	BEGIN_CLASS_INIT( FDClusteredHistogram );
	ADD_VAR_ARRAY_PROP( m_DescriptorData, 0, new Float32ClassProperty("element",MetaClass::FIRST_VERSION+0,NULL));
	ADD_CARD32_PROP( m_NumBins, 0 );
	ADD_CARD32_PROP( m_NumHistograms, 0 );
	ADD_STRING_PROP( m_Attribute, 0 );
	ADD_CARD32_PROP( m_NumComponents,0 );
}

FDClusteredHistogram::FDClusteredHistogram(void)
{
	m_Attribute = "color";
	m_NumComponents = 3;
	m_NumHistograms = 4;
	m_NumBins = 16;
}

FDClusteredHistogram::~FDClusteredHistogram(void)
{
}

void FDClusteredHistogram::assign(const Object* obj, CopyContext *context)
{
	FeatureDescriptor::assign(obj,context);
	const FDClusteredHistogram * o = dynamic_cast<const FDClusteredHistogram*>(obj);
	pAssert( o != NULL );

	m_NumBins = o->m_NumBins;
	m_NumHistograms = o->m_NumHistograms;
	m_Attribute = o->m_Attribute;
	m_NumComponents = o->m_NumComponents;

	m_Clusters = o->m_Clusters;
}

void FDClusteredHistogram::setupClusters( const vector<PatchGraph*> &pgs, const PointCloudAnimationSequence *data ) 
{
	KMeansClustering<float> kmeans;
	if( m_NumComponents != 3 )
		throw "ToDo";

	for( card32 i=0;i<pgs.size();i++ )
	{
		const PatchGraph * pg = pgs[i];
		card32 frameIndex = pg->getStartFrameIndex();
		AAT attrAAT = data->getFrame(frameIndex)->getAAT( m_Attribute );
		const PointSet * pg_ps = pg->getTrajectorySet()->getTrajectories();
		for( card32 j=0;j<pg_ps->getNumEntries();j++ )
		{

			vector<card32> dps = pg->getTrajectorySet()->getDataPoints()->getIndices(0, j);
			const size_t numDPoints = dps.size();
			Vector3f c = NULL_VECTOR3F;
	      
			for (card32 k=0; k<numDPoints; k++) {
				c += data->getFrame(frameIndex)->get<float32, 3>(dps[k], attrAAT);
			}
			c /= numDPoints;


			DVectorF dvec;
			dvec.setDim(m_NumComponents);
			dvec[0] = c[0];
			dvec[1] = c[1];
			dvec[2] = c[2];
			kmeans.add( dvec );
		}

	}

	vector< DVectorF> result;
	kmeans.calc( m_NumBins, result, 0.01f, 1 );
	m_Clusters.clear();

	for( card32 i=0;i<result.size();i++ )
	{
		for( card32 j=0;j<result[i].size();j++ )
		{
			m_Clusters.push_back( result[i][j] );			
		}
	}
}

void FDClusteredHistogram::createDescriptor( const Vector3f &point, float size, PointCloud * pc, TDescriptorContext * context ) 
{
	pAssert( pc->providesAttribute( m_Attribute ));
	float h_2 = - log(0.01f) / pow(size,2.0f);

	AAT attrAAT = pc->getAAT( m_Attribute );
	AAT posAAT = pc->getAAT( "position" );
	VAD attrVAD = pc->getDescr()->getAttribute( m_Attribute );
	pAssert( attrVAD.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc,32,NULL);

	m_DescriptorData = vector<float>( m_NumBins * m_NumHistograms, 0.0f );
	hIt->setSeekPointAndReset( point );
	float scale = 1.0f;
	
	if( m_NumComponents != 3 )
		throw "FDClusteredHistogram::createDescriptor() - Implemented for m_NumComponents=3 only!";

	while( !hIt->atEnd() )
	{
		Vector3f pos = hIt->get3f( posAAT );
		float dist = norm( point - pos );
		if( dist > size )
			break;

		Vector3f attr = hIt->get3f( attrAAT );

		float magnitude = exp( -pow(dist,2.0f) / h_2 ) * scale;
		
		Vector3f * clusters = (Vector3f*)&m_Clusters[0];

		card32 binID = 0;
		float distToBin = 1e20f;
		for( card32 i=0;i<m_NumBins;i++ )
		{
			float tmpDist = norm( attr - clusters[i] );
			if( tmpDist < distToBin )
			{
				distToBin = tmpDist;
				binID = i;
			}
		}

		// use linear interpolation in radial direction 
		float hCoord = dist / (float)size * (m_NumHistograms) - 0.5f;
		if( hCoord <= 0.0f )
		{
			m_DescriptorData[binID] += magnitude;			
		}
		else
		{
			int h1 = floor(hCoord);
			int h2 = ceil(hCoord);
			float hInterp = hCoord - floor(hCoord);
			if( h1 < (int)m_NumHistograms )
			{
				m_DescriptorData[ h1 * m_NumBins + binID] += magnitude * (1.0f - hInterp);
			}
			if( h2 < (int)m_NumHistograms )
			{
				m_DescriptorData[ h2 * m_NumBins + binID] += magnitude * (hInterp);
			}
		}
		hIt->next();
	}

	delete hIt;

}
