#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDCurvature.h"
//---------------------------------------------------------------------------
#include "copyObjectProperties.h"
#include "TopologicalKNNIterator.h"
#include "LinearTransformPriors.h"
#include "HierarchicalKNNIterator.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FDCurvature ,1)
{
	BEGIN_CLASS_INIT( FDCurvature );
	ADD_BOOLEAN_PROP( m_UseMean, 0 );
	ADD_BOOLEAN_PROP( m_UsePrincipal, 0 );
	ADD_VAR_ARRAY_PROP( m_DescriptorData, 0, new Float32ClassProperty("element",MetaClass::FIRST_VERSION+0,NULL));
	ADD_BOOLEAN_PROP( m_UseTopology, 1);
}

FDCurvature::FDCurvature(void)
{
	m_UseMean = true;
	m_UsePrincipal = true;
	m_UseTopology = false;
}

FDCurvature::~FDCurvature(void)
{
}

void FDCurvature::assign(const Object* obj, CopyContext *context)
{
	FeatureDescriptor::assign(obj,context);
	const FDCurvature * o = dynamic_cast<const FDCurvature*>(obj);
	pAssert( o != NULL );

	//copyObjectProperties( obj, this );
	m_UseMean = o->m_UseMean;
	m_UsePrincipal = o->m_UsePrincipal;
	m_UseTopology = o->m_UseTopology;

}
void FDCurvature::createDescriptor( const Vector3f &point, float size, PointCloud * pc, TDescriptorContext * context ) 
{
	HierarchicalKNNIterator hIt( pc, 32, NULL );
	hIt.setSeekPointAndReset( point );
	context->tIt->setStartPointAndReset( hIt.getPointSetPointNumber() );
	vector<Vector3f> positions;
	AAT posAAT = pc->getAAT( "position" );
	float initialDist = norm( context->tIt->get3f(posAAT) - point);
	Vector3f centroid = NULL_VECTOR3F;

	// gather points
	if( m_UseTopology )
	{
		while( !context->tIt->atEnd() )
		{
			positions.push_back( context->tIt->get3f( posAAT ) );
			centroid += positions.back();
		
			float dist = norm( positions.back() - point );
			if( dist - initialDist > size )
				break;

			context->tIt->next();
		}
	}
	else
	{
		HierarchicalKNNIterator hIt(pc, 32, NULL );
		hIt.setSeekPointAndReset( point );
		while( !hIt.atEnd() )
		{
			positions.push_back( hIt.get3f( posAAT ) );
			centroid += positions.back();
		
			float dist = norm( positions.back() - point );
			if( dist - initialDist > size )
				break;

			hIt.next();
		}

	}
	// calculate pca and a local coordinate frame
	Vector3f eigenValues;
	Matrix3f eigenVectors;

	centroid /= max(1,positions.size());
	Matrix3f C;
	for( card32 k=0;k<positions.size();k++ )
	{
		for ( unsigned int i = 0; i < 3; i++ ){
			for ( unsigned int j = 0; j <= i; j++ ){
				C[i][j] += ( positions[k][i]-centroid[i]) * (positions[k][j]-centroid[j]);
				C[j][i] = C[i][j];
			}
		}
	}
	C /= max(1,positions.size());
   C.computeEigenStructure( eigenValues, eigenVectors );

	Vector3f pca_normal = normalize( eigenVectors[0].crossProduct( eigenVectors[1] ) );
	Matrix3f localToWorld = calcTangentSystem( pca_normal );
	Matrix3f worldToLocal = localToWorld.transpose();

	// transform points into local coord frame and fitt a quadratic function 

	float scale = 1.0f / size;
	for( card32 i=0;i<positions.size();i++ )
	{
		positions[i] = worldToLocal * (positions[i] - point) * scale;
	}

	StaticVector<float32, 6> coeff;
	vector< StaticVector<float32, 6> > coeffsDerivs;
	float sigma = sqrt( - size*size / log( 0.01f ) );
	GLTMonomials2ndOrder::computeCoeffs( positions, coeff, coeffsDerivs, sigma );

   float32 uu = coeff[3];
   float32 uv = coeff[4];
   float32 vv = coeff[5];
   
	m_DescriptorData.resize( getNumDescriptorValues() );
	card32 offset = 0;

	if( m_UseMean )
		m_DescriptorData[offset++] = atan2(2.0f*uu + 2.0f*vv + uv,1.0f);

	if( m_UsePrincipal )
	{
		float32 lambda1, lambda2;
		eigenValuesSym2f(uu, uv, vv, lambda1, lambda2);

		m_DescriptorData[offset++] = atan2(lambda1,1.0f);
		m_DescriptorData[offset++] = atan2(lambda2,1.0f);
	}


}
