#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDHistogram.h"
//---------------------------------------------------------------------------
#include "HierarchicalKNNIterator.h"
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FDHistogram ,1)
{
	BEGIN_CLASS_INIT( FDHistogram );

	ADD_CARD32_PROP( m_NumBins, 0 );
	ADD_CARD32_PROP( m_NumHistograms, 0 );
	ADD_STRING_PROP( m_Attribute, 0 );
	ADD_FLOAT32_PROP( m_MaxValue, 0 );
	ADD_FLOAT32_PROP( m_MinValue, 0 );
	ADD_VAR_ARRAY_PROP( m_DescriptorData, 0, new Float32ClassProperty("element",MetaClass::FIRST_VERSION+0,NULL));
	ADD_CARD32_PROP( m_AttributeCompNumber, 1 );
}

FDHistogram::FDHistogram(void)
{
	m_NumBins = 8;
	m_NumHistograms = 4;
	m_Attribute = "curvature";
	m_MaxValue = 1.0f;
	m_MinValue = 0;
	m_AttributeCompNumber = 0;
}

FDHistogram::~FDHistogram(void)
{
}

void FDHistogram::assign(const Object* obj, CopyContext *context)
{
	FeatureDescriptor::assign(obj,context);
	const FDHistogram * o = dynamic_cast<const FDHistogram*>(obj);
	pAssert( o != NULL );
	m_NumBins = o->m_NumBins;
	m_NumHistograms = o->m_NumHistograms;
	m_Attribute = o->m_Attribute;
	m_MaxValue = o->m_MaxValue;
	m_MinValue = o->m_MinValue;
}
void FDHistogram::createDescriptor( const Vector3f &point, float size, PointCloud * pc, TDescriptorContext * context ) 
{
	pAssert( pc->providesAttribute( m_Attribute ));
	float h_2 = - log(0.01f) / pow(size,2.0f);

	AAT attrAAT = pc->getAAT( m_Attribute );
	AAT posAAT = pc->getAAT( "position" );
	VAD attrVAD = pc->getDescr()->getAttribute( m_Attribute );
	pAssert( attrVAD.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc,32,NULL);

	m_DescriptorData = vector<float>( m_NumBins * m_NumHistograms, 0.0f );
	hIt->setSeekPointAndReset( point );
	vector<float> ringWeight;
	for( card32 i=0;i<m_NumHistograms;i++ )
	{
		float radius = (float)(i+1) / (float)m_NumHistograms * size;
		float radius0 = (float)(i) / (float)m_NumHistograms * size;
		ringWeight.push_back( sqr(radius) - sqr(radius0) );
	}
	

	float scale = 1.0f;
	float summed_magnitude = 0;
	while( !hIt->atEnd() )
	{
		Vector3f pos = hIt->get3f( posAAT );
		float dist = norm( point - pos );
		if( dist > size )
			break;
		float attr = 0;
		switch( attrVAD.numComponents )
		{
		case 1:
			attr = hIt->get1f( attrAAT );
			break;
		case 2:
			attr = hIt->get2f( attrAAT )[m_AttributeCompNumber];
			break;
		case 3:
			attr = hIt->get3f( attrAAT )[m_AttributeCompNumber];
			break;
		case 4:
			attr = hIt->get4f( attrAAT )[m_AttributeCompNumber];
			break;
		}

		float magnitude = 1.0f;//exp( -pow(dist,2.0f) / h_2 ) * scale;
		summed_magnitude += magnitude;
		float binID = (attr - m_MinValue) / (m_MaxValue - m_MinValue) * (float)m_NumBins;
		int bin1 = floor(binID);
		int bin2 = ceil(binID);
		float binInterp = binID - floor(binID);
		
		// use linear interpolation in radial direction 
		float hCoord = dist / (float)size * (m_NumHistograms) - 0.5f;
		if( hCoord <= 0.0f )
		{
			if( bin1 >= 0 && bin1 < (int)m_NumBins )
				m_DescriptorData[ bin1 ] += magnitude * (1.0f - binInterp);
			if( bin2 >= 0 && bin2 < (int)m_NumBins )
				m_DescriptorData[ bin2 ] += magnitude * binInterp;
			//ringWeights[0] += magnitude;
		}
		else
		{
			int h1 = floor(hCoord);
			int h2 = ceil(hCoord);
			float hInterp = hCoord - floor(hCoord);
			if( h1 < (int)m_NumHistograms )
			{
				if( bin1 >= 0 && bin1 < (int)m_NumBins )
					m_DescriptorData[ h1 * m_NumBins + bin1] += magnitude * (1.0f - binInterp) * (1.0f - hInterp);
				if( bin2 >= 0 && bin2 < (int)m_NumBins )
					m_DescriptorData[ h1 * m_NumBins + bin2] += magnitude * binInterp * (1.0f - hInterp);
				//ringWeights[h1] += magnitude * (1.0f - hInterp);
			}
			if( h2 < (int)m_NumHistograms )
			{
				if( bin1 >= 0 && bin1 < (int)m_NumBins )
					m_DescriptorData[ h2 * m_NumBins + bin1] += magnitude * (1.0f - binInterp) * hInterp;
				if( bin2 >= 0 && bin2 < (int)m_NumBins )
					m_DescriptorData[ h2 * m_NumBins + bin2] += magnitude * binInterp * hInterp;
				//ringWeights[h2] += magnitude * hInterp;
			}
		}
		hIt->next();
	}

	/*for( card32 i=0;i<ringWeights.size();i++ )
	{

	}*/
	float inv_summed_magnitude = 1.0f;
	if( summed_magnitude > 1e-10f )
		inv_summed_magnitude = 1.0f / summed_magnitude;
	for( card32 i=0;i<m_DescriptorData.size();i++ )
		m_DescriptorData[i] *= inv_summed_magnitude;

	/*summed_magnitude = 0;
	for( card32 i=0;i<m_NumHistograms;i++ )
	{
		for( card32 j=0;j<m_NumBins;j++ )
		{
			m_DescriptorData[i*m_NumBins+j] *= ringWeight[i];
			summed_magnitude += m_DescriptorData[i*m_NumBins+j];
		}
	}

	inv_summed_magnitude = 1.0f;
	if( summed_magnitude > 1e-10f )
		inv_summed_magnitude = 1.0f / summed_magnitude;
	for( card32 i=0;i<m_DescriptorData.size();i++ )
		m_DescriptorData[i] *= inv_summed_magnitude;*/

	delete hIt;

}

void FDHistogram::preparePointCloud( PointCloud * pc )
{
}
