#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDSIFT.h"
//---------------------------------------------------------------------------
#include "RegistrationTools.h"
#include "RGBAImage.h"
#include "ImageDisplayer.h"
#include "MathTools.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FDSIFT ,1)
{
	BEGIN_CLASS_INIT( FDSIFT );
	
	ADD_STRING_PROP(m_GradientAttribute,1);
	ADD_FLOAT32_PROP(m_OrientationMagnitudeThreshold,1);
	ADD_FLOAT32_PROP(m_WindowSmoothPercentage,1);
	ADD_CARD32_PROP(m_NumOrientationBins,1);
	ADD_CARD32_PROP(m_NumDesriptorBins,1);
	ADD_CARD32_PROP(m_NumSampleRegions,1);
}

FDSIFT::FDSIFT(void)
{
	m_GradientAttribute = "gradient";
	m_OrientationMagnitudeThreshold = 0.8f;
	m_WindowSmoothPercentage = 0.1f;
	m_NumOrientationBins = 36;
	m_NumDesriptorBins = 8;
	m_NumSampleRegions = 2;
}

FDSIFT::~FDSIFT(void)
{
}

void FDSIFT::assign(const Object* obj, CopyContext *context)
{
	FeatureDescriptor::assign(obj,context);
	const FDSIFT * o = dynamic_cast<const FDSIFT*>(obj);
	pAssert( o != NULL );

	m_GradientAttribute = o->m_GradientAttribute;
	m_OrientationMagnitudeThreshold = o->m_OrientationMagnitudeThreshold;
	m_WindowSmoothPercentage = o->m_WindowSmoothPercentage;
	m_NumOrientationBins = o->m_NumOrientationBins;
	m_NumDesriptorBins = o->m_NumDesriptorBins;
	m_NumSampleRegions = o->m_NumSampleRegions;
	
	m_Descriptors = o->m_Descriptors;
	m_Point = o->m_Point;
	m_OrientationHistogram = o->m_OrientationHistogram;
}

struct TGradient
{
	float angle, magnitude;
	Vector3f rel_pos;
};

void FDSIFT::createDescriptor( const Vector3f &point, float size, PointCloud * pc, TDescriptorContext * context ) 
{
	pAssert( pc->providesAttribute( m_GradientAttribute ));
	m_Point = point;
	float h_2 = - log(0.01f) / pow(size,2.0f);

	AAT gradAAT = pc->getAAT( m_GradientAttribute );
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT( "normal" );

	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc,32,NULL);

	float maxDistance = size*(1.0f + m_WindowSmoothPercentage);
	float windowSmoothStart = size*(1.0f - m_WindowSmoothPercentage);
	float smoothDist = maxDistance - windowSmoothStart;

	// calculate normal via pca
	hIt->setSeekPointAndReset( point );
	PCA<float,3> pca;
	while(!hIt->atEnd())
	{
		Vector3f pos = hIt->get3f(posAAT);
		float distance = norm( pos - point );
		if( distance > size )
			break;

		pca.addPoint(pos);
		hIt->next();
	}

	Vector3f eigenValues, centroid;
	Matrix3f eigenVectors;
	pca.analyze( eigenValues, eigenVectors, centroid );

	Vector3f normal = normalize( eigenVectors[0].crossProduct( eigenVectors[1] ) );
	
	// -------------------------------------------------
	// create orientation histogram
	Matrix3f tangentSystem = calcTangentSystem( normal );
	Matrix3f invTangentSystem = tangentSystem.transpose();
	vector<float> bins(m_NumOrientationBins,0.0f);
	vector<TGradient> gradients;
	hIt->setSeekPointAndReset( point );
	if( normal * hIt->get3f(normalAAT) < 0.0f )
		normal = -normal;
	while(!hIt->atEnd())
	{
		Vector3f pos = hIt->get3f(posAAT);
		float distance = norm( pos - point );
		if( distance > size )
			break;


		Vector3f gradient = hIt->get3f( gradAAT );
		float gradient_length = norm( gradient );
		gradient_length *= exp( -pow(distance,2.0f) / h_2 );
		
		// transform position and gradient into local coordinate frame
		gradient = invTangentSystem * gradient;
		Vector3f rel_pos = invTangentSystem * (pos - point);


		// estimate orientation angle
		if( fabs(gradient[0]) > 1e-20f && fabs(gradient[1]) > 1e-20f )
		{
			float angle = atan2( gradient[1],gradient[0] );
			TGradient grad;
			grad.angle = angle;
			grad.magnitude = gradient_length;
			grad.rel_pos = rel_pos;
			gradients.push_back( grad );

			angle *=  (float)(bins.size()) / M_PI  / 2.0f;
			if( angle < 0.0f )
				angle += (float)(bins.size());
			int binID1 = ((int)floor(angle)) % bins.size();
			int binID2 = ((int)ceil(angle)) % bins.size();
			float interp = angle - floor(angle);
			bins[ binID1 ] += gradient_length * (1.0f - interp);
			bins[ binID2 ] += gradient_length * (interp);
		}
		hIt->next();
	}

	m_OrientationHistogram = bins;
	// apply triangle filter
	for( card32 i=0;i<m_OrientationHistogram.size();i++ )
	{
		bins[i] = m_OrientationHistogram[i] * 0.6f +
			m_OrientationHistogram[(i+1)%m_OrientationHistogram.size()] * 0.2f
			+ m_OrientationHistogram[(i+m_OrientationHistogram.size()-1)%m_OrientationHistogram.size()] * 0.2f;
	}

	m_OrientationHistogram = bins;
	

	// estimate biggest orientantion magnitude
	float maxMagnitude = 0.0f;
	card32 maxBin = 0;
	for( card32 i=0;i<bins.size();i++ )
	{
		if( bins[i] > maxMagnitude )
		{
			maxMagnitude = bins[i];
			maxBin = i;
		}
	}

	// count number of main orientantions
	card32 numMainOrientations = 0;
	for( card32 i=0;i<bins.size();i++ )
	{
		if( bins[i] > maxMagnitude * m_OrientationMagnitudeThreshold )
		{
			numMainOrientations++;
		}
	}

	if( numMainOrientations > bins.size() * 0.25f )
		return;

	// create descriptors
	for( card32 i=0;i<bins.size();i++ )
	{
		if( bins[i] <= maxMagnitude * m_OrientationMagnitudeThreshold )
			continue;

		TDesc desc;

		// calc orientantion vector
		float angle = (float)i / (float)(bins.size()) * 2.0f * M_PI;
		desc.orientation = makeVector3f( cos(angle),sin(angle),0) * bins[i];
		desc.orientation = normalize( tangentSystem * desc.orientation ) * size;

		// ---------------------------------------------------------
		// generate keypoint descriptors

		// init keypoint descriptor
		for( card32 x=0;x<m_NumSampleRegions;x++ )
		{
			for( card32 y=0;y<m_NumSampleRegions;y++ )
			{
				vector<float> keypoints(m_NumDesriptorBins,0);
				desc.keypoints.push_back( keypoints );
			}
		}

		// build descriptor
		float descSize = size;
		Matrix3f localToKeypoint = makeRotZ3f( -angle );
		for( card32 g = 0;g<gradients.size();g++ )
		{
			Vector3f relPos = localToKeypoint * gradients[g].rel_pos;
			Vector2f r = (makeVector2f(relPos[0],relPos[1]) + makeVector2f(descSize,descSize) )
				/ 2.0f / descSize * (float)m_NumSampleRegions - makeVector2f(0.5f,0.5f);
			
			// bilinear interpolation of regions
			float xf = r[0] - floor(r[0]);
			float yf = r[1] - floor(r[1]);
			int rx1 = max(0, min( m_NumSampleRegions-1, floor(r[0])));
			int rx2 = max(0, min( m_NumSampleRegions-1, ceil(r[0])));
			int ry1 = max(0, min( m_NumSampleRegions-1, floor(r[1])));
			int ry2 = max(0, min( m_NumSampleRegions-1, ceil(r[1])));

			float tmp_angle =  (gradients[g].angle - angle) * (float)(m_NumDesriptorBins) / M_PI  / 2.0f;
			while( tmp_angle < 0.0f )
				tmp_angle += (float)(m_NumDesriptorBins);

			int binID1 = ((int)floor(tmp_angle)) % m_NumDesriptorBins;
			int binID2 = ((int)ceil(tmp_angle)) % m_NumDesriptorBins;

			// linear interpolation of orientation bin
			float mInterp = tmp_angle - floor(tmp_angle);
			float magnitude1 = gradients[g].magnitude * (1.0f - mInterp);
			float magnitude2 = gradients[g].magnitude * mInterp;

			desc.keypoints[rx1 * m_NumSampleRegions + ry1][binID1] += magnitude1 * (1.0f-xf) * (1.0f - yf);
			desc.keypoints[rx1 * m_NumSampleRegions + ry1][binID2] += magnitude2 * (1.0f-xf) * (1.0f - yf);

			desc.keypoints[rx2 * m_NumSampleRegions + ry1][binID1] += magnitude1 * xf * (1.0f - yf);
			desc.keypoints[rx2 * m_NumSampleRegions + ry1][binID2] += magnitude2 * xf * (1.0f - yf);

			desc.keypoints[rx1 * m_NumSampleRegions + ry2][binID1] += magnitude1 * (1.0f-xf) * yf;
			desc.keypoints[rx1 * m_NumSampleRegions + ry2][binID2] += magnitude2 * (1.0f-xf) * yf;

			desc.keypoints[rx2 * m_NumSampleRegions + ry2][binID1] += magnitude1 * xf * yf;
			desc.keypoints[rx2 * m_NumSampleRegions + ry2][binID2] += magnitude2 * xf * yf;
		}

		m_Descriptors.push_back( desc );
	}
	
	delete hIt;
}

float FDSIFT::compare( const FeatureDescriptor * other ) 
{
	const FDSIFT * sift = dynamic_cast<const FDSIFT*>(other);
	pAssert( sift != NULL );
	pAssert( sift->m_NumSampleRegions == m_NumSampleRegions );
	pAssert( sift->m_NumDesriptorBins == m_NumDesriptorBins );

	float dist = 1e20f;
	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		for( card32 j=0;j<sift->m_Descriptors.size();j++ )
		{
			float tmpDist = 0.0f;
			for( card32 x=0;x<sift->m_NumSampleRegions * sift->m_NumSampleRegions;x++ )
			{
				for( card32 y=0;y<sift->m_NumDesriptorBins;y++ )
				{
					tmpDist += pow( sift->m_Descriptors[j].keypoints[x][y] - m_Descriptors[i].keypoints[x][y], 2.0f );
				}
			}
			tmpDist = sqrt( tmpDist );
			if( tmpDist < dist )
				dist = tmpDist;
		}
	}
	return dist;
}

void FDSIFT::drawOrientations(float scale) 
{
	glBegin( GL_LINES );
	if(m_Descriptors.size() > 1 )
		glColor3f(1,0,0);
	else
		glColor3f(1,1,0);
	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		glVertex3fv( &m_Point[0] );
		Vector3f p2 = m_Point + m_Descriptors[i].orientation;
		glVertex3fv( &p2[0] );
	}
	glEnd();
}

void FDSIFT::preparePointCloud( PointCloud * pc ) 
{
	checkAttribute(pc, "gradient", 3, VAD::DATA_FORMAT_FLOAT32 );
	calculateGradients( pc, 20, "gradient", "color" );
}

void FDSIFT::debugReport() 
{
	debugOutput << "SIFT Feature:\nNum Main Orientations: " << (mpcard) m_Descriptors.size() << 
		"\nHistogram: ";
	float maxV = 0;
	for( card32 i=0;i<m_OrientationHistogram.size();i++ )
	{
		if( i % 8 == 0 )
			debugOutput << "\n";

		debugOutput << "(" << i << "," << m_OrientationHistogram[i] << ")";
		maxV = max(maxV,m_OrientationHistogram[i]);
	}
	debugOutput << "\n";

	static RGBAFloat32Image * image = NULL;
	float imageHeight = 20.0f;
	if( image == NULL )
		image = new RGBAFloat32Image( (card32)m_OrientationHistogram.size(), imageHeight);

	image->clear( makeVector4f(0,0,0,0));
	for( card32 i=0;i<image->getSizeX();i++ )
	{
		for( card32 j=0;j<m_OrientationHistogram[i] / maxV * imageHeight;j++ )
			image->setPixel(i,imageHeight-1-j, makeVector4f(1,1,1,1));
	}

	ImageDisplayer::DisplayRGBAFloat32Image ev(image);
	debugOutput.displayEvent( ImageDisplayer::getClass(), "muh", &ev );

}
