#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FastFeatures.h"
//---------------------------------------------------------------------------
#include "MultiResolutionPC.h"
#include "HeighfieldParameterization.h"
#include "VertexArray.h"
#include "Monomials2ndOrder.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FastFeatures ,0)
{
	BEGIN_CLASS_INIT( FastFeatures );
}

FastFeatures::FastFeatures(void)
{
}

FastFeatures::~FastFeatures(void)
{
}

void FastFeatures::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const FastFeatures * o = dynamic_cast<const FastFeatures*>(obj);
	pAssert( o != NULL );

}

vector<Vector3f> FastFeatures::detectFeatures( MultiResolutionPC * pc )
{
	vector<Vector3f> ret;
	float averagePointDistance = pc->getPCAveragePointDistance(0);

	float h1 = averagePointDistance * 6.0f;
	float h2 = h1 * 2.25f;
	float h1_2 = h1*h1;
	float h2_2 = h2*h2;

	float minSamplingSpace = h2 * 0.5f;
	float radius = sqrt(-log(0.05f) * pow(h2,2.0f));
	
	PointCloud * s_pc = pc->getClosestPointCloud(minSamplingSpace);
	SceneObjectIterator * iter = s_pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	AAT posAAT = s_pc->getAAT("position");
	HeighfieldParameterization hparam;
	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f(posAAT);
		hparam.calculate(pc->getPointCloud(0), pos, radius);
		PointSet * ps = hparam.getParametricPoints();

		StaticVector<float32, 6> coeffs;
		const card32 NUM_NEIGHBORS = ps->getNumEntries();
		coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f; coeffs[3] = 0.0f; coeffs[4] = 0.0f; coeffs[5] = 0.0f;
		StaticVector<float32, 6> cd;
		StaticMatrix<float32, 6,6> dm;
		for (card32 i=0; i<6;i++) {
			for (card32 j=0; j<6;j++) {
				dm[i][j] = 0.0f;
			}
		}

		AAT paramPosAAT = ps->getAAT( "param_pos" );
		AAT heightAAT = ps->getAAT( "height" );

		for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++)
		{
			Vector2f paramPos = ps->get2f( pIndex, paramPosAAT );
			const float32 u = paramPos[0];
			const float32 v = paramPos[1];
			const float32 n = ps->get1f( pIndex, heightAAT);
			const float32 w = (exp(-(u*u+v*v)/h2_2) - exp(-(u*u+v*v)/h1_2)) / (float32)NUM_NEIGHBORS;

			// c[j] += b[j](uv(i)) * n(i) * w(uv(i));
			cd[0] =       w;
			cd[1] = u   * w;
			cd[2] = v   * w;
			cd[3] = u*u * w;
			cd[4] = u*v * w;
			cd[5] = v*v * w;

			coeffs[0] += cd[0] * n;
			coeffs[1] += cd[1] * n;
			coeffs[2] += cd[2] * n;
			coeffs[3] += cd[3] * n;
			coeffs[4] += cd[4] * n;
			coeffs[5] += cd[5] * n;

			// 1 u v uu uv vv

			dm[0][0] += w;     dm[0][1] += u*w;     dm[0][2] += v*w;     dm[0][3] += u*u*w;     dm[0][4] += u*v*w;     dm[0][5] += v*v*w;
			dm[1][0] += w*u;   dm[1][1] += u*w*u;   dm[1][2] += v*w*u;   dm[1][3] += u*u*w*u;   dm[1][4] += u*v*w*u;   dm[1][5] += v*v*w*u;
			dm[2][0] += w*v;   dm[2][1] += u*w*v;   dm[2][2] += v*w*v;   dm[2][3] += u*u*w*v;   dm[2][4] += u*v*w*v;   dm[2][5] += v*v*w*v;
			dm[3][0] += w*u*u; dm[3][1] += u*w*u*u; dm[3][2] += v*w*u*u; dm[3][3] += u*u*w*u*u; dm[3][4] += u*v*w*u*u; dm[3][5] += v*v*w*u*u;
			dm[4][0] += w*u*v; dm[4][1] += u*w*u*v; dm[4][2] += v*w*u*v; dm[4][3] += u*u*w*u*v; dm[4][4] += u*v*w*u*v; dm[4][5] += v*v*w*u*v;
			dm[5][0] += w*v*v; dm[5][1] += u*w*v*v; dm[5][2] += v*w*v*v; dm[5][3] += u*u*w*v*v; dm[5][4] += u*v*w*v*v; dm[5][5] += v*v*w*v*v;
		}

		StaticMatrix<float32, 6,6> invDm;
		bool success;
		invDm = invertMatrix(dm, &success, 1e-7f);
		if (success) {
			coeffs = invDm * coeffs;
		} else {
			coeffs[0] = 0; coeffs[1] = 0; coeffs[2] = 0;
			coeffs[3] = 0; coeffs[4] = 0; coeffs[5] = 0;
		}

		// estimate maxima/minima
		// see "Tessellation of Quadratic Elements" Dillard et al.
		const float A = coeffs[5];
		const float B = coeffs[4];
		const float C = coeffs[3];
		const float D = coeffs[2];
		const float E = coeffs[1];
		float detH = 4.0f * A * C - B*B;
		if( detH > 0 )
		{
			float cu = (-2.0f * C * D + B * E) / detH;
			float cv = (-2.0f * A * E + B * D) / detH;
			Vector3f featurePoint = makeVector3f(cu,cv,Monomials2ndOrder::evaluate( coeffs, makeVector2f(cu,cv)));
			featurePoint = shrink4To3( hparam.getParametricToWorld4f() * expand3To4(featurePoint));
			ret.push_back( featurePoint );
		}

		bIt->next();
	}

	return ret;
}