#include "StdAfx.h"
#include "../externalIncludes/opencv/cxcore.h"
//---------------------------------------------------------------------------
#include "FeatureRegistration.h"
//---------------------------------------------------------------------------
#include "FeatureDescriptor.h"
#include "FeatureDetector.h"
#include "MLSMultiScaleFeatureDetector.h"
#include "FDSift.h"
#include "MultiResolutionPC.h"
#include "MultiScaleGenerator.h"
#include "ProgressWindow.h"
#include "RegistrationTools.h"
#include "SpectralCorrespondenceValidation.h"
#include "FDHistogram.h"
#include "LocalExtremaDetector.h"
#include "Timer.h"
#include "SplatPullPushInterpolation.h"
#include "HeighfieldParameterization.h"
#include "AnimationBSRContext.h"
#include "PatchGraphContainer.h"
#include "ClassMethods.h"
#include "UnstructuredInCorePointCloud.h"
#include "RegistrationContext.h"
#include "ProgressWindow.h"
#include "IncorePCTopologyGraph.h"
#include "FeatureCorrespondences.h"
#include "CorrespondenceList.h"
#include "PGDataPointIndices.h"
#include "PointCloudAnimationSequence.h"
#include "StringHelper.h"
#include "UniformPointCloudResampling.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "DynamicLinearAlgebraTools.h"
//#include "DynamicLinearAlgebraTools_templCode.h"
#include "FDAttribute.h"
#include "KMeansClustering_templCode.h"
#include "MultiFrameCorrespondenceValidation.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( FeatureRegistration ,2)
{
	BEGIN_CLASS_INIT( FeatureRegistration );

	//ADD_OBJECT_PROP( m_FeatureDescriptor, 0, FeatureDescriptor::getClass(),true );
	//ADD_OBJECT_PROP( m_FeatureDetector, 0, FeatureDetector::getClass(),true);

	//ADD_BOOLEAN_PROP( m_UseGeodeticDistances, 1 );

	ADD_CARD32_PROP( m_StartFrame,2);
	ADD_CARD32_PROP( m_EndFrame,2);

	ADD_NOARGS_METHOD_DESC( calculateAllFeatures, "calculate features for every frame" );
	ADD_NOARGS_METHOD_DESC( calculateFeatures_OneFrame, "calculate features for one frame only" );
	ADD_NOARGS_METHOD_DESC( buildFeatureCorrespondences, "build feature correspondences between frames" );
	ADD_NOARGS_METHOD_DESC( buildFeatureCorrespondences_OneFrame, "build feature correspondences between two frames" );
	ADD_NOARGS_METHOD_DESC( calculateAllFeatures_Clustering, "calculate features using descriptor clustering" );
	ADD_NOARGS_METHOD_DESC( setCorrespondencesValid, "copies current correspondences into validated container (and mark guessed corr. as valid)" );
	ADD_NOARGS_METHOD_DESC( deletePatchGraphFrames, "deletes frames [m_StartFrame,m_EndFrame]" );

}

FeatureRegistration::FeatureRegistration(void)
{
	m_FeatureDescriptor = NULL;//new FDHistogram;
	m_FeatureDetector = NULL;//new MLSMultiScaleFeatureDetector;

	m_FeatureRadius = 5.0f;
	m_AveragePointDistance = 0.0f;
	m_UseGeodeticDistances = false;
	m_StartFrame = 0;
	m_EndFrame = 100;

	m_SpectralSigma = 2.0f;
	m_SpectralAddDiff = 0.5f;
	m_SpectralScaleDiff = 0.0f;


}

FeatureRegistration::~FeatureRegistration(void)
{
//	delete m_FeatureDetector;
//	delete m_FeatureDescriptor;
}

void FeatureRegistration::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const FeatureRegistration * o = dynamic_cast<const FeatureRegistration*>(obj);
	pAssert( o != NULL );

	m_FeatureDetector = o->m_FeatureDetector;
	m_FeatureDescriptor = o->m_FeatureDescriptor;
	m_FeatureRadius = o->m_FeatureRadius;
	m_AveragePointDistance = o->m_AveragePointDistance;
	m_UseGeodeticDistances = o->m_UseGeodeticDistances;
}

void FeatureRegistration::setup( FeatureDetector * det, FeatureDescriptor * desc ) 
{
	//delete m_FeatureDetector;
	//delete m_FeatureDescriptor;
	m_FeatureDetector = det;
	m_FeatureDescriptor = desc;
}

void FeatureRegistration::setupSIFT( string gradientAttribute, card32 numberOfScales, float startSigma,
												float F, float averagePointDistance, card32 numSampleRegions, card32 numOrientationBins) 
{
	delete m_FeatureDetector;
	delete m_FeatureDescriptor;

	MLSMultiScaleFeatureDetector * mlsDet = new MLSMultiScaleFeatureDetector;
	m_FeatureDetector = mlsDet;
	mlsDet->setAveragePointDistance(averagePointDistance);
	mlsDet->setF( F );
	mlsDet->setH0( startSigma );
	mlsDet->setNumberOfScales( numberOfScales );

	FDSIFT * sift = new FDSIFT;
	m_FeatureDescriptor = sift;
	sift->setGradientAttribute( gradientAttribute );
	m_AveragePointDistance = averagePointDistance;
}


void FeatureRegistration::setupHistogram( string sourceAttribute, card32 numberOfScales, float startSigma,
													  float F, float averagePointDistance,card32 numRadialHistograms, 
													  card32 numBins ) 
{
	//delete m_FeatureDetector;
	//delete m_FeatureDescriptor;

	MLSMultiScaleFeatureDetector * mlsDet = new MLSMultiScaleFeatureDetector;
	m_FeatureDetector = mlsDet;
	mlsDet->setAveragePointDistance(averagePointDistance);
	mlsDet->setF( F );
	mlsDet->setH0( startSigma );
	mlsDet->setNumberOfScales( numberOfScales );

	FDHistogram * fd = new FDHistogram;
	m_FeatureDescriptor = fd;
	fd->setAttribute( sourceAttribute );
	fd->setNumHistograms( numRadialHistograms );
	fd->setNumBins( numBins );
	
	if( sourceAttribute == "curvature" )
	{
		fd->setMaxValue( 1.0f );
		fd->setMinValue( -1.0f );
	}
	else
	{
		fd->setMaxValue( 1.0f );
		fd->setMinValue( 0 );
	}
	m_AveragePointDistance = averagePointDistance;
}


vector<FeaturePair>	FeatureRegistration::estimateFeaturePairs( PointCloud * pc1, PointCloud * pc2, EDetectionMode mode  ) 
{
	vector<FeaturePair> ret;
	return ret;
}

FeatureList*		FeatureRegistration::calculateFeatures( PointCloud * pc, EDetectionMode mode  ) 
{
	float averagePointDistance = m_AveragePointDistance;
	if( averagePointDistance == 0.0f)
		averagePointDistance = geoxl::getAveragePointDistance( pc );

	checkAttribute(pc, "gradient", 3, VAD::DATA_FORMAT_FLOAT32 );
	if( mode != DETECT_COLOR )
		checkAttribute(pc, "curvature", 4, VAD::DATA_FORMAT_FLOAT32 );
	else
		createGrayChannel( pc );
	//calculateCurvature(pc, 30, "curvature", averagePointDistance );

	card32 numberOfScales = 5;
	MLSMultiScaleFeatureDetector * mlsDet = dynamic_cast<MLSMultiScaleFeatureDetector*>(m_FeatureDetector);
	if( mlsDet )
	{
		numberOfScales = mlsDet->getNumberOfScales();
		if( mode == DETECT_COLOR )
			mlsDet->setAttribute( "gray" );
	}

	Timer timer;

	MultiResolutionPC mres;
	mres.setup( pc, 100, 2.0f, 1.0f );

	MultiScaleGenerator gen;
	gen.setMinNumberOfSamples(50);
	gen.setNumberOfScales(numberOfScales);
	if( mlsDet )
	{
		gen.setF( mlsDet->getF() );
		gen.setH0( mlsDet->getH0() );
	}

	switch( mode )
	{
	case DETECT_POS:
			gen.setupPosition();
			
			break;
	case DETECT_COLOR:
			gen.setupColor();
			break;
	case DETECT_POS_AND_COLOR:
			gen.setupPositionAndColor();
			throw "not implemented yet!";
			break;
	}
	
	PointCloud * mscale = gen.create( pc, &mres );

	FeatureList* ret = m_FeatureDetector->detectFeatures( mscale, m_FeatureRadius*averagePointDistance );

	//if( mode != DETECT_COLOR )
	//	calculateCurvature( pc, 15, "curvature", averagePointDistance );


	TDescriptorContext dc;
	dc.interpolation = new SplatPullPushInterpolation;
	dc.parameterization = new HeighfieldParameterization;
	dc.parametricImageSize = 64;
	dc.pc = pc;
	dc.tIt = NULL;
	dc.averagePointDistance = averagePointDistance;

	for( card32 i=0;i<ret->size();i++ )
	{
		FeatureDescriptor * fd = (FeatureDescriptor *)m_FeatureDescriptor->copy(); 
		fd->createDescriptor( ret->get(i)->getPosition(), m_FeatureRadius * averagePointDistance, dc.pc, &dc );
		fd->normalizeDescriptor();
		ret->get(i)->addDescriptor( fd );
	}

	card32 dtime = timer.getDeltaValue();
	//debugOutput << "Feature Detection Time: " << dtime/1000 << "s " <<  dtime % 1000 << "ms\n";

	//pc->assign( mscale );
	delete mscale;
	return ret;
}

bool testPairsFunc( const FeaturePair &p1,const FeaturePair &p2 )
{
	if( p1.m_ValidatedCorrespondence && !p2.m_ValidatedCorrespondence )
		return true;

	if( !p1.m_ValidatedCorrespondence && p2.m_ValidatedCorrespondence )
		return false;

	return p1.m_DescriptorDistance < p2.m_DescriptorDistance;
}

vector<FeaturePair>	FeatureRegistration::buildFeaturePairs( FeatureList *features1,FeatureList *features2, float maxDistance, int maxNumberOfPairs) 
{
	progressWindow->pushStep( "build pairs");
	vector<FeaturePair> ret;
	float minDist = 1e20f;
	float maxDist = -1e20f;
	for( card32 i=0;i<features1->size();i++ )
	{
		progressWindow->progress( (float)i / (float)features1->size() * 100.0f );
		for( card32 j=0;j<features2->size();j++ )
		{
			if( features1->get(i)->compareDetectorSignature( features2->get(j) ) > 0.1f )
				continue;

			float dist = ((FeatureDescriptor*)features1->get(i)->getDescriptor(0))->compare(((FeatureDescriptor*)features2->get(j)->getDescriptor(0)) );
			if( dist < maxDistance )
			{
				FeaturePair pair;
				pair.m_DescriptorDistance = dist;
				pair.m_Feature1 = features1->get(i);
				pair.m_Feature2 = features2->get(j);
				pair.m_Index1 = i;
				pair.m_Index2 = j;
				ret.push_back( pair );
			}
			if( dist < 1e19f )
			{
				minDist = min( minDist, dist );
				maxDist = max( maxDist, dist );
			}
		}
	}
	progressWindow->popSteps();

	debugOutput << "buildFeaturePairs: MinDist = " << minDist << ", MaxDist = " << maxDist << "\n";

	sort( ret.begin(), ret.end(), testPairsFunc );

	/*for( card32 i=0;i<ret.size()-1;i++ )
	{
		int newIndex = i + rand() % (ret.size() - i);
		FeaturePair pair = ret[i];
		ret[i] = ret[newIndex];
		ret[newIndex] = pair;
	}*/

	if( (int)ret.size() > maxNumberOfPairs )
	{
		ret.erase( ret.begin() + maxNumberOfPairs, ret.end() );
	}
	return ret;
}

vector<FeaturePair>	FeatureRegistration::validateFeaturePairs( const vector<FeaturePair> pairs, PointCloud *p1, PointCloud *p2 ) 
{
	card32 maxNumberOfPairs = 3000;
	SpectralCorrespondenceValidation spectralCV;
	spectralCV.setMinConfidence( m_AnimationBSRContext->getRegistrationContext()->getMinConfidence() );
//	spectralCV.setRegistrationContext( m_AnimationBSRContext->getRegistrationContext() );
	card32 numberOfPairs = min((card32)pairs.size(),(card32)maxNumberOfPairs);

	// reduce 
	map<card32,card32> idToEntry1, idToEntry2;
	map<card32,card32> entryToID1,entryToID2;
	vector<Vector3f> points1, points2;
	for( card32 i=0;i<numberOfPairs;i++ )
	{
		card32 index1 = pairs[i].m_Index1;
		card32 index2 = pairs[i].m_Index2;

		if( idToEntry1.find( index1 ) == idToEntry1.end() )
		{
			idToEntry1[index1] = (card32)points1.size();
			points1.push_back( pairs[i].m_Feature1->getPosition() );
		}
		if( idToEntry2.find( index2 ) == idToEntry2.end() )
		{
			idToEntry2[index2] = (card32)points2.size();
			points2.push_back( pairs[i].m_Feature2->getPosition() );
		}
	}

	spectralCV.setup(numberOfPairs, (card32)points1.size(), (card32)points2.size() );

	for( card32 i=0;i<numberOfPairs; )
	{
		FeaturePair pair = pairs[i];
		spectralCV.setCorrespondencePair(i, idToEntry1[pair.m_Index1], idToEntry2[pair.m_Index2], 
			pair.m_Index1, pair.m_Index2, 0.0f, pair.m_Feature1->getPosition(), pair.m_Feature2->getPosition() );
		i++;
	}

	float averagePointDistance = m_AveragePointDistance;
	if( averagePointDistance == 0.0f)
		averagePointDistance = geoxl::getAveragePointDistance(p1);

	float sigma = m_SpectralSigma * averagePointDistance;
	spectralCV.setSigma(sigma);
	spectralCV.setAddDiff( m_SpectralAddDiff * averagePointDistance );
	spectralCV.setScaleDiff( m_SpectralScaleDiff );

	// calculate distances
	if( m_UseGeodeticDistances )
	{
		// TODO: use existing topology!
		UnstructuredInCorePointCloud * upc1 = dynamic_cast<UnstructuredInCorePointCloud*>(p1);
		FixedKNNTopologyIterator topologyIter1( upc1->getPointSet(), p1->getAttachments(), p1, 8 );
		UnstructuredInCorePointCloud * upc2 = dynamic_cast<UnstructuredInCorePointCloud*>(p2);
		FixedKNNTopologyIterator topologyIter2( upc2->getPointSet(), p2->getAttachments(), p2, 8 );

		progressWindow->pushStep(true, "geo dist" );
		float32 scaleProgress = 100.0f / (float)(points1.size() + points2.size());
		for( card32 i=0;i<points1.size();i++ )
		{
			vector<float> dists = calculateGeodeticDistances( p1, &topologyIter1, points1[i], points1 ); 
			for( card32 j=0;j<dists.size();j++ )
			{
				spectralCV.setDistanceEntry1( i,j, dists[j], norm(points1[i]- points1[j]) );
			}
			progressWindow->progress( (float)i * scaleProgress );

		}
		for( card32 i=0;i<points2.size();i++ )
		{
			vector<float> dists = calculateGeodeticDistances( p2, &topologyIter2, points2[i], points2 ); 
			for( card32 j=0;j<dists.size();j++ )
			{
				spectralCV.setDistanceEntry2( i,j, dists[j], norm(points2[i]- points2[j]) );
			}
			progressWindow->progress( (float)(i+points1.size()) * scaleProgress );

		}
		progressWindow->popSteps();
	}
	else
	{
		for( card32 i=0;i<points1.size();i++ )
		{
			for( card32 j=i+1;j<points1.size();j++ )
			{
				spectralCV.setDistanceEntry1( i,j, norm( points1[i]-points1[j] ), norm( points1[i]-points1[j] ));
			}
		}
		for( card32 i=0;i<points2.size();i++ )
		{
			for( card32 j=i+1;j<points2.size();j++ )
			{
				spectralCV.setDistanceEntry2( i,j, norm( points2[i]-points2[j] ), norm( points2[i]-points2[j] ));
			}
		}
	}
	

	//spectralCV.setAddDiff(averagePointDistance*10.0f);
	//spectralCV.setScaleDiff( 0.2f );
	spectralCV.calculate();

	vector<FeaturePair> ret;
	vector<SpectralCorrespondenceValidation::TFeaturePair> tmpPairs;
	tmpPairs = spectralCV.getValidPairs();
	for( card32 i=0;i<tmpPairs.size();i++ )
	{
		for( card32 j=0;j<pairs.size();j++ )
		{
			if( tmpPairs[i].orgIndex1 == pairs[j].m_Index1 && 
				tmpPairs[i].orgIndex2 == pairs[j].m_Index2 )
			{
				ret.push_back( pairs[j] );
				break;
			}
		}
	}

	return ret;
}

void FeatureRegistration::setupFunkhouser2D( string sourceAttribute, card32 componentNumber, card32 numberOfScales, 
									float startSigma, float F, float averagePointDistance, 
								card32 numRadialSamples, card32 numCircularSamples ) 
{
	delete m_FeatureDetector;
	delete m_FeatureDescriptor;

	MLSMultiScaleFeatureDetector * mlsDet = new MLSMultiScaleFeatureDetector;
	m_FeatureDetector = mlsDet;
	mlsDet->setAveragePointDistance(averagePointDistance);
	mlsDet->setF( F );
	mlsDet->setH0( startSigma );
	mlsDet->setNumberOfScales( numberOfScales );
	
	m_AveragePointDistance = averagePointDistance;
}

void FeatureRegistration::calculateAllFeatures() 
{
	setupFromContext();

	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	if( m_AveragePointDistance == 0.0f )
	{
		UnstructuredInCorePointCloud * newPC = new UnstructuredInCorePointCloud;
		newPC->clear();
		newPC->getPointSet()->assign( pgc->getPatchGraph(0)->getTrajectorySet()->getTrajectories() );
		m_AveragePointDistance = geoxl::getAveragePointDistance(newPC);
		delete newPC;
	}

	progressWindow->pushStep( true,"features" );
	for( card32 i=0;i<pgc->getNumPatchGraphs();i++ )
	{
		calculateFeaturesPG( pgc->getPatchGraph(i) );
		if( !progressWindow->progress( (float)i/(float)pgc->getNumPatchGraphs() * 100.0f ) )
			break;
	}
	progressWindow->popSteps();
}

void FeatureRegistration::calculateFeatures_OneFrame() 
{
	setupFromContext();


	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	for( card32 i=0;i<pgc->getNumPatchGraphs();i++ )
	{
		if( pgc->getPatchGraph(i)->getStartFrameIndex() == 0 )
			calculateFeaturesPG( pgc->getPatchGraph(i) );
	}
	
}

struct TVertex_FRG
{
	Vector3f pos, normal, color;
};

void FeatureRegistration::calculateFeaturesPG( PatchGraph * frame ) 
{
	FeatureList * features = NULL;
	PointSet * ps = frame->getTrajectorySet()->getTrajectories();

	UnstructuredInCorePointCloud * newPC = new UnstructuredInCorePointCloud;
#if 1
	newPC->clear();
	newPC->getPointSet()->assign( ps );
#else
	VertexDescriptor * vd = new VertexDescriptor;
	vd->pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd->pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd->pushAttrib( VAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	newPC->clearAndSetup( vd, 0 );

	AAT patch_posAAT = ps->getAAT( "position" );
	AAT patch_normalAAT = ps->getAAT( "normal" );
	AAT patch_tangentuAAT = ps->getAAT( "tangent_u" );
	AAT patch_coeffsAAT = ps->getAAT( "hfieldCoeffs" );

	PGDataPointIndices * ind = frame->getTrajectorySet()->getDataPoints();
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	PointCloudAnimationSequence * seq = pgc->getAnimSeq();
	PointSet * dataPoints = seq->getFrame( frame->getStartFrameIndex() );
	AAT posAAT = dataPoints->getAAT( "position" );
	AAT colorAAT = dataPoints->getAAT( "color" );


	// project data points on patch surface
	for( card32 i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f patch_pos = ps->get3f( i,patch_posAAT );
		Vector3f patch_normal = ps->get3f( i,patch_normalAAT );
		Vector3f patch_tangentU = ps->get3f( i,patch_tangentuAAT );
		StaticVector<float,6> patch_coeffs = ps->get<float,6>( i,patch_coeffsAAT );

		Matrix3f worldToPatch;
		worldToPatch[0] = patch_tangentU;
		worldToPatch[1] = patch_normal.crossProduct( patch_tangentU	 );
		worldToPatch[2] = patch_normal;
		Matrix3f patchToWorld = worldToPatch;
		worldToPatch = worldToPatch.transpose();
		
		for( card32 j=0;j<ind->getNumDataPoints(0, i );j++ )
		{
			card32 index = ind->getIndex(0, i, j);
			Vector3f pos = dataPoints->get3f( index, posAAT );
			Vector3f color = dataPoints->get3f( index, colorAAT );

			if( norm( pos - patch_pos) > m_AnimationBSRContext->getSamplingSpacing() * 0.5 )
				continue;
			
			Vector3f localPos = worldToPatch * (pos - patch_pos);
			float u = localPos[0];
			float v = localPos[1];
			localPos[2] =	pow( u, 2.0f ) * patch_coeffs[3]
							+	pow( v, 2.0f ) * patch_coeffs[5]
							+	u * v * patch_coeffs[4]
							+	u * patch_coeffs[1]
							+	v * patch_coeffs[2]
							+ patch_coeffs[0];

			float derivate_u = patch_coeffs[1] + 2.0f * u * patch_coeffs[3] + v * patch_coeffs[4];
			float derivate_v = patch_coeffs[2] + 2.0f * v * patch_coeffs[5] + u * patch_coeffs[4];

			Vector3f normal = -makeVector3f(1,0, derivate_u).crossProduct(makeVector3f(0,1, derivate_v));
			
			TVertex_FRG vertex;
			vertex.pos = patchToWorld * localPos + patch_pos;
			vertex.normal = patchToWorld * normal;
			vertex.color = color;
			newPC->getPointSet()->addPoint( (char*)&vertex );
		}
				
	}

	UnstructuredInCorePointCloud *tmpPC = new UnstructuredInCorePointCloud;
	tmpPC->clear();
	tmpPC->getPointSet()->assign( ps );
	float tmppc_averagePointDistance = geoxl::getAveragePointDistance( tmpPC );
	delete tmpPC;

	UniformPointCloudResampling resampler;
	resampler.setup( tmppc_averagePointDistance * 0.25f );
	PointSet * resampledPS = resampler.resamplePointSet( newPC->getPointSet());
	newPC->getPointSet()->assign( resampledPS );
	delete resampledPS;

	m_AveragePointDistance = geoxl::getAveragePointDistance( newPC );

	if( frame->getStartFrameIndex() == 0 )
	{
		static int numPC = 1;
		string pcName = string("debugPC") + intToStr(numPC++);		
		addPointCloud(debugOnly_Scene, newPC, pcName );
	}
	//return;
#endif

	features = calculateFeatures( newPC, DETECT_POS );
	bool useGeodetic = m_AnimationBSRContext->getRegistrationContext()->getUseGeodeticDistances();
	vector<float> distances( features->size() * features->size() );
	if( useGeodetic )
	{
		UnstructuredInCorePointCloud *tmpPC = new UnstructuredInCorePointCloud;
		tmpPC->clear();
		tmpPC->getPointSet()->assign( ps );


		InCorePCTopologyGraphPointSetIterator * topIter = new InCorePCTopologyGraphPointSetIterator( ps, NULL, 
			frame->getTopology() );
		vector<Vector3f> points;
		for( card32 i=0;i<features->size();i++ )
		{
			points.push_back( features->get(i)->getPosition() );
		}
		for( card32 i=0;i<points.size();i++ )
		{
			vector<float> tmpDist = calculateGeodeticDistances( tmpPC, topIter, points[i], points );
			for( card32 j=0;j<tmpDist.size();j++ )
			{
				distances[i*features->size()+j] = tmpDist[j];
			}

		}
		delete topIter;
		delete tmpPC;
	}
	else
	{
		for( card32 i=0;i<features->size();i++ )
		{
			for( card32 j=i;j<features->size();j++ )
			{
				float dist = norm( features->get(i)->getPosition() - features->get(j)->getPosition() );
				distances[i*features->size()+j] = dist;
				distances[j*features->size()+i] = dist;
			}
		}
	}
	features->setupFeatureDistances( distances );

	if( frame->getNumFeatureLists() > 0 )
		frame->deleteFeatureLists(0, frame->getNumFeatureLists() );
	
	frame->addFeatureList( features );	
	delete newPC;
}

void FeatureRegistration::setupFromContext() 
{
	RegistrationContext * rc = m_AnimationBSRContext->getRegistrationContext();
	m_AveragePointDistance = rc->getAveragePointDistance();
	//m_FeatureDescriptor->assign( rc->getFeatureDescriptor() );
	//m_FeatureDetector->assign( rc->getFeatureDetector() );
	/*delete m_FeatureDescriptor;
	delete m_FeatureDetector;
	m_FeatureDescriptor = (FeatureDescriptor*)rc->getFeatureDescriptor()->copy();
	m_FeatureDetector = (FeatureDetector*)rc->getFeatureDetector()->copy();*/
	m_FeatureDescriptor = rc->getFeatureDescriptor();
	m_FeatureDetector = rc->getFeatureDetector();
	
	m_FeatureRadius = rc->getFeatureRadius();
	m_UseGeodeticDistances = rc->getUseGeodeticDistances();
}

void FeatureRegistration::buildFeatureCorrespondences() 
{
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer(); 
	if( m_AveragePointDistance == 0.0f )
	{
		UnstructuredInCorePointCloud * newPC = new UnstructuredInCorePointCloud;
		newPC->clear();
		newPC->getPointSet()->assign( pgc->getPatchGraph(0)->getTrajectorySet()->getTrajectories() );
		m_AveragePointDistance = geoxl::getAveragePointDistance(newPC);
		delete newPC;
	}
	pgc->getFeatureCorrespondences()->setup( pgc->getAnimSeq()->getNumFrames() );
	if( pgc->getTempCorrespondences()->getNumFrames() != pgc->getAnimSeq()->getNumFrames() )
		pgc->getTempCorrespondences()->setup(pgc->getAnimSeq()->getNumFrames() );

	vector<PatchGraph*> frame1,frame2;
	card32 frameID = 0;
	for( card32 pgID=0;pgID<pgc->getNumPatchGraphs();pgID++ )
	{
		PatchGraph * pg = pgc->getPatchGraph( pgID );
		if( pg->getStartFrameIndex() == frameID )
		{
			frame2.push_back( pg );
		}
		else
		{
			if( frameID != 0 )
			{
				buildCorresponcences( frame1, frame2 );
			}
			frame1 = frame2;
			frame2.clear();
			frame2.push_back( pg );
			frameID++;
			//if( frameID >= 2 )
			//	break;

		}

	}
	buildCorresponcences( frame1, frame2 );

	MultiFrameCorrespondenceValidation validation;
	//validation.validateCorrspondences( m_AnimationBSRContext->getPGContainer()->getFeatureCorrespondences() );

}

struct TTmpFeature
{
	card32 featureIndex;
	card32 featurePG;
	Feature * f;

};

void FeatureRegistration::buildCorresponcences( vector<PatchGraph*> frame1, vector<PatchGraph*> frame2 ) 
{
	// -------------------------------------------------
	// collect features
	vector<TTmpFeature> f1,f2;
	for( card32 i=0;i<frame1.size();i++ )
	{
		for( card32 j=0;j<frame1[i]->getFeatures(0)->size();j++ )
		{
			TTmpFeature tmpF;
			tmpF.f = frame1[i]->getFeatures(0)->get(j);
			tmpF.featureIndex = j;
			tmpF.featurePG = i;
			f1.push_back( tmpF );
		}
	}

	for( card32 i=0;i<frame2.size();i++ )
	{
		for( card32 j=0;j<frame2[i]->getFeatures(0)->size();j++ )
		{
			TTmpFeature tmpF;
			tmpF.f = frame2[i]->getFeatures(0)->get(j);
			tmpF.featureIndex = j;
			tmpF.featurePG = i;
			f2.push_back( tmpF );
		}
	}

	// -------------------------------------------------
	// build initial pairs
	float threshold = m_AnimationBSRContext->getRegistrationContext()->getPruningThreshold();
	float maxDist = m_AnimationBSRContext->getRegistrationContext()->getMaxMovement();
	CorrespondenceList * oldCorrList = m_AnimationBSRContext->getPGContainer()->getTempCorrespondences()->getFrameCorrespondences( frame1[0]->getStartFrameIndex() );
	bool useExistingCorrespondenceAsValid = m_AnimationBSRContext->getRegistrationContext()->getUseExistingCorrespondenceAsValid() && oldCorrList != NULL;

	vector<FeaturePair> pairs;
	if( useExistingCorrespondenceAsValid )
	{
		// prune conflicting feature with respect to existing correspondences
		for( card32 c=0;c<oldCorrList->getNumCorrespondences();c++ )
		{
			Correspondence * corr = oldCorrList->get(c);
			if( corr->getFlags() == Correspondence::CF_Initial )
				continue;

			FeaturePair pair;
			pair.m_ValidatedCorrespondence = true;
			pair.m_DescriptorDistance = 0;
			pair.m_Index1 = (card32)f1.size();
			pair.m_Index2 = (card32)f2.size();
			for( card32 i=0;i<f1.size();i++ )
			{
				if( norm( f1[i].f->getPosition() - corr->getPosition1() ) < 1e-10f )
				{
					pair.m_Feature1 = f1[i].f;
					pair.m_Index1 = i;
					break;
				}
			}
			for( card32 i=0;i<f2.size();i++ )
			{
				if( norm( f2[i].f->getPosition() - corr->getPosition2() ) < 1e-10f )
				{
					pair.m_Feature2 = f2[i].f;
					pair.m_Index2 = i;
					break;
				}
			}
			if( pair.m_Index1 == f1.size() || pair.m_Index2 == f2.size() )
			{
				debugOutput << "BuildCorr - Couldn't find features from validated correspondences!\n";
			}
			else
				pairs.push_back( pair );
		}
	}
	for( card32 i=0;i<f1.size();i++ )
	{
		for( card32 j=0;j<f2.size();j++ )
		{
			float dist = f1[i].f->getFeatureDescriptor(0)->compare( f2[j].f->getFeatureDescriptor(0) );
			if( dist < threshold && (norm( f1[i].f->getPosition() - f2[j].f->getPosition() ) < maxDist || maxDist == 0.0f))
			{
				bool isExistingCorr = false;
				bool conflictingCorr = false;
				if( useExistingCorrespondenceAsValid )
				{
					// prune conflicting feature with respect to existing correspondences
					for( card32 c=0;c<oldCorrList->getNumCorrespondences();c++ )
					{
						Correspondence * corr = oldCorrList->get(c);
						if( corr->getFlags() != 0 )
							continue;

						if( norm( corr->getPosition1() - f1[i].f->getPosition() ) < 1e-10f &&
							norm( corr->getPosition2() - f2[j].f->getPosition() ) < 1e-10f )
							isExistingCorr = true;
						else
						{
							if( norm( corr->getPosition1() - f1[i].f->getPosition() ) < 1e-10f )
								conflictingCorr = true;

							if( norm( corr->getPosition2() - f2[j].f->getPosition() ) < 1e-10f )
								conflictingCorr = true;

						}

						if( conflictingCorr || isExistingCorr)
							break;
					}
					if( conflictingCorr || isExistingCorr)
						continue;
				}


				FeaturePair pair;
				pair.m_DescriptorDistance = dist;
				pair.m_Feature1 = f1[i].f;
				pair.m_Feature2 = f2[j].f;
				pair.m_Index1 = i;
				pair.m_Index2 = j;
				pair.m_ValidatedCorrespondence = false;
				pairs.push_back( pair );
			}
		}
	}
	if( pairs.size() == 0 )
		return;

	sort( pairs.begin(), pairs.end(), testPairsFunc );

//	debugOutput << "initial pairs: " << pairs.size() << "\n";

	card32 maxNumberOfPairs = 3000;
	SpectralCorrespondenceValidation spectralCV;
	spectralCV.setMinConfidence( m_AnimationBSRContext->getRegistrationContext()->getMinConfidence() );
	//spectralCV.setRegistrationContext( m_AnimationBSRContext->getRegistrationContext() );
	card32 numberOfPairs = min((card32)pairs.size(),(card32)maxNumberOfPairs);

	// reduce 
	map<card32,card32> idToEntry1, idToEntry2;
	map<card32,card32> entryToID1,entryToID2;
	card32 points1size = 0;
	card32 points2size = 0;
	for( card32 i=0;i<numberOfPairs;i++ )
	{
		card32 index1 = pairs[i].m_Index1;
		card32 index2 = pairs[i].m_Index2;

		if( idToEntry1.find( index1 ) == idToEntry1.end() )
		{
			idToEntry1[index1] = points1size;
			points1size++;
		}
		if( idToEntry2.find( index2 ) == idToEntry2.end() )
		{
			idToEntry2[index2] = points2size;
			points2size++;
		}
	}

	spectralCV.setup(numberOfPairs, points1size, points2size );

	float sigma = m_AnimationBSRContext->getRegistrationContext()->getSelfConfidenceSigma();
	float sigma_2 = pow(sigma * m_AveragePointDistance,2.0f);
	for( card32 i=0;i<numberOfPairs; )
	{
		FeaturePair pair = pairs[i];
		//float confidence = m_AnimationBSRContext->getRegistrationContext()->getSelfConfidenceScale() *
		//	exp( -normQuad(pair.m_Feature1->getPosition() - pair.m_Feature2->getPosition() ) / sigma_2 );
		float confidence = max(0,m_AnimationBSRContext->getRegistrationContext()->getSelfConfidenceScale() *
			(1.0f - pair.m_DescriptorDistance / m_AnimationBSRContext->getRegistrationContext()->getPruningThreshold()));
		if( pair.m_ValidatedCorrespondence )
			confidence = m_AnimationBSRContext->getRegistrationContext()->getSelfConfidenceScale();
		spectralCV.setCorrespondencePair(i, idToEntry1[pair.m_Index1], idToEntry2[pair.m_Index2], 
			pair.m_Index1, pair.m_Index2, confidence, pair.m_Feature1->getPosition(), pair.m_Feature2->getPosition(),
			pair.m_ValidatedCorrespondence);
		i++;
	}

	float averagePointDistance = m_AveragePointDistance;
	if( averagePointDistance == 0.0f )
		throw "precalculate average distance!";

	for( map<card32,card32>::iterator iter = idToEntry1.begin(); iter != idToEntry1.end(); ++iter )
	{
		for(map<card32,card32>::iterator iter2 = idToEntry1.begin(); iter2 != idToEntry1.end(); ++iter2 )
		{
			card32 orgI1 = iter->first;
			card32 orgI2 = iter2->first;
			card32 sI1 = iter->second;
			card32 sI2 = iter2->second;
			TTmpFeature tmpF1 = f1[orgI1];
			TTmpFeature tmpF2 = f1[orgI2];
			float dist = 1e20f;
			if( tmpF1.featurePG == tmpF2.featurePG )
			{
				PatchGraph * pg = frame1[tmpF1.featurePG];
				FeatureList * fl = pg->getFeatures(0);
				dist = fl->dist( tmpF1.featureIndex, tmpF2.featureIndex );
			}
			spectralCV.setDistanceEntry1( sI1, sI2, dist, norm( tmpF1.f->getPosition() - tmpF2.f->getPosition() ) );
		}
		
	}
	for( map<card32,card32>::iterator iter = idToEntry2.begin(); iter != idToEntry2.end(); ++iter )
	{
		for(map<card32,card32>::iterator iter2 = idToEntry2.begin(); iter2 != idToEntry2.end(); ++iter2 )
		{
			card32 orgI1 = iter->first;
			card32 orgI2 = iter2->first;
			card32 sI1 = iter->second;
			card32 sI2 = iter2->second;
			TTmpFeature tmpF1 = f2[orgI1];
			TTmpFeature tmpF2 = f2[orgI2];
			float dist = 1e20f;
			if( tmpF1.featurePG == tmpF2.featurePG )
				dist = frame2[tmpF1.featurePG]->getFeatures(0)->dist( tmpF1.featureIndex, tmpF2.featureIndex );
			spectralCV.setDistanceEntry2( sI1, sI2, dist, norm( tmpF1.f->getPosition() - tmpF2.f->getPosition() ) );
		}
		
	}

	vector<Correspondence*> correspondences;
	
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	FeatureCorrespondences * fc = pgc->getFeatureCorrespondences();
	CorrespondenceList * corrlist = fc->getFrameCorrespondences( frame1[0]->getStartFrameIndex() ); 

	if( m_AnimationBSRContext->getRegistrationContext()->getValidateCorrespondences() )
	{

		spectralCV.setAddDiff(averagePointDistance*m_AnimationBSRContext->getRegistrationContext()->getValidate_ConflictAdd());
		spectralCV.setScaleDiff( m_AnimationBSRContext->getRegistrationContext()->getValidate_ConflictScale() );
		spectralCV.setSigma(averagePointDistance* m_AnimationBSRContext->getRegistrationContext()->getValidate_ConfSigma() );
		//if( !useExistingCorrespondenceAsValid )
			spectralCV.calculate();
		//else
		//	spectralCV.forwardSearch();

		vector<FeaturePair> ret;
		vector<SpectralCorrespondenceValidation::TFeaturePair> tmpPairs;
		tmpPairs = spectralCV.getValidPairs();


		
		for( card32 i=0;i<tmpPairs.size();i++ )
		{
			Correspondence * c = new Correspondence;
			c->setPosition1( f1[tmpPairs[i].orgIndex1].f->getPosition() );
			c->setPosition2( f2[tmpPairs[i].orgIndex2].f->getPosition() );
			c->setStartFrame( frame1[0]->getStartFrameIndex() );
			if( tmpPairs[i].valid || !useExistingCorrespondenceAsValid)
				c->setFlags(0);
			else
				c->setFlags( Correspondence::CF_Guessed);

			correspondences.push_back( c );
		}

		for( card32 i=0;i<pairs.size();i++ )
		{
			bool valid = false;
			for( card32 j=0;j<tmpPairs.size();j++ )
			{
				if( norm( pairs[i].m_Feature1->getPosition() - f1[tmpPairs[j].orgIndex1].f->getPosition() ) < 1e-10f 
					&& norm( pairs[i].m_Feature2->getPosition() - f2[tmpPairs[j].orgIndex2].f->getPosition() ) < 1e-10f )
				{
					valid = true;
					break;
				}
			}

			if( !valid )
			{
				Correspondence * c = new Correspondence;
				c->setPosition1( pairs[i].m_Feature1->getPosition() );
				c->setPosition2( pairs[i].m_Feature2->getPosition() );
				c->setStartFrame( frame1[0]->getStartFrameIndex() );
				c->setFlags( Correspondence::CF_Initial );
				correspondences.push_back( c );
			}
		}
	}
	else
	{	
	
		for( card32 i=0;i<pairs.size();i++ )
		{
			Correspondence * c = new Correspondence;
			c->setPosition1( pairs[i].m_Feature1->getPosition() );
			c->setPosition2( pairs[i].m_Feature2->getPosition() );
			c->setStartFrame( frame1[0]->getStartFrameIndex() );
			if( pairs[i].m_ValidatedCorrespondence || !useExistingCorrespondenceAsValid)
				c->setFlags(0);
			else
				c->setFlags( Correspondence::CF_Guessed);
			correspondences.push_back( c );
		}
	}
	corrlist->setup( correspondences );

}

struct TGridCell
{
	float weight;
	Vector3f midPoint;
	bool operator < ( const TGridCell &other ) {return weight < other.weight;}
};


void FeatureRegistration::calculateAllFeatures_Clustering() 
{
	// cluster descriptors in descriptor-space, store result in clusterID
	Timer timer;
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	string descChan = "color";
	KMeansClustering<float> kmeans;
	card32 numClusters = m_AnimationBSRContext->getRegistrationContext()->getNumDescriptorClusters();
	card32 num_cells = m_AnimationBSRContext->getRegistrationContext()->getNumDescriptorSpaceCells();
	map<card32,card32> histogram;

	for( card32 i=0;i<pgc->getNumPatchGraphs();i++ )
	{
		pgc->getPatchGraph(i)->createPatchAttribute( VAD( "clusterID", 1, VAD::DATA_FORMAT_INT32 ) );
		PointSet * ps = pgc->getPatchGraph(i)->getTrajectorySet()->getTrajectories();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		if( !ps->providesAttribute( descChan ) )
		{
			message( "calculate descriptors before using descriptor clustering" );
			return;
		}

		AAT descAAT = ps->getAAT( descChan );
		
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f desc = ps->get3f(j, descAAT );
			/*DVectorF vec;
			vec.setDim(3);
			vec[0] = desc[0];
			vec[1] = desc[1];
			vec[2] = desc[2];
			kmeans.add( vec );*/
			desc = desc * num_cells;
			Vector3i desc_i = makeVector3i( min(num_cells-1,desc[0]), 
				min(num_cells-1,desc[1]),
				min(num_cells-1,desc[2]) );
			card32 index = (desc_i[0] * num_cells + desc_i[1]) * num_cells + desc_i[2];
			histogram[index]++;
		}

		if( i > 4 )
			break;

	}

	for( map<card32,card32>::iterator iter = histogram.begin(); iter != histogram.end(); ++iter )
	{
		card32 index = iter->first;
		card32 num = iter->second;
		Vector3f cell_midpoint = makeVector3f( index / num_cells / num_cells, (index / num_cells) % num_cells, index % num_cells ) + 
			makeVector3f(0.5f, 0.5f, 0.5f);
		cell_midpoint = cell_midpoint / num_cells;
		DVectorF vec;
		vec.setDim(3);
		vec[0] = cell_midpoint[0];
		vec[1] = cell_midpoint[1];
		vec[2] = cell_midpoint[2];
		kmeans.add( vec, num );
	}

	vector< DVectorF > codebook;
	vector< card32 > indices;
	kmeans.calcEx( numClusters, codebook, indices, 0.001f, 10 );

	card32 index = 0;
	for( card32 i=0;i<pgc->getNumPatchGraphs();i++ )
	{
		//pgc->getPatchGraph(i)->createPatchAttribute( VAD( "clusterID", 1, VAD::DATA_FORMAT_INT32 ) );
		PointSet * ps = pgc->getPatchGraph(i)->getTrajectorySet()->getTrajectories();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		if( !ps->providesAttribute( descChan ) )
		{
			message( "calculate descriptors before using descriptor clustering" );
			return;
		}

		AAT descAAT = ps->getAAT( descChan );
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f desc = ps->get3f( j, descAAT );
			int32 cluster = 0;
			/*for( card32 d=0;d<3;d++ )
			{
				cluster |= (desc[d] < 0.5f ? 0:1) << d;
			}*/
			//cluster = indices[index];
			DVectorF vec;
			vec.setDim(3);
			vec[0] = desc[0];
			vec[1] = desc[1];
			vec[2] = desc[2];
			float minDist = 1e20f;
			for( card32 k=0;k<codebook.size();k++ )
			{
				float tmpDist = norm( codebook[k] - vec );
				if( tmpDist < minDist )
				{
					minDist = tmpDist;
					cluster = k;
				}
			}
			index++;
			ps->set1i( j, clusterAAT, cluster );
		}
	}

	// cluster descriptors 
	for( card32 i=0;i<pgc->getNumPatchGraphs();i++ )
	{
		InCorePCTopologyGraph * top = pgc->getPatchGraph(i)->getTopology();
		PointSet * ps = pgc->getPatchGraph(i)->getTrajectorySet()->getTrajectories();
		AAT clusterAAT = ps->getAAT( "clusterID" );
		AAT posAAT = ps->getAAT( "position" );
		AAT descAAT = ps->getAAT( descChan );
		UnstructuredInCorePointCloud tmpCopyPC;
		tmpCopyPC.getPointSet()->assign( ps );
		HierarchicalKNNIterator hIt( &tmpCopyPC, 32, NULL );
		float averagePointDistance = geoxl::getAveragePointDistance(&tmpCopyPC,3);

		FeatureList * features = new FeatureList;

		vector<int> visited( ps->getNumEntries(), 0 );
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			if( visited[j] != 0 )
				continue;

			int32 clusterID = ps->get1i( j, clusterAAT );

			// start clustering
			list<int> queue;
			vector<int> clusterIndices;
			vector<Vector3f> clusterPoints;
			queue.push_back(j);
			visited[j] = 1;
			clusterIndices.push_back( j );
			clusterPoints.push_back( ps->get3f(j, posAAT ) );
			Vector3f mean_desc = ps->get3f(j, descAAT );

			while( !queue.empty() )
			{
				int node = queue.front();
				queue.pop_front();

				for( card32 n=0;n<top->getNumAdjacentVertices( node );n++ )
				{
					int32 neighbour = top->getAdjacentVertex( node, n );
					if( ps->get1i(neighbour, clusterAAT ) != clusterID )
						continue;

					if( visited[neighbour] == 0 )
					{
						queue.push_back( neighbour );
						visited[neighbour] = 1;
						clusterIndices.push_back( neighbour );
						clusterPoints.push_back( ps->get3f(neighbour, posAAT ) );
						mean_desc += ps->get3f(neighbour, descAAT );
					}

				}

			}

			if( clusterIndices.size() < 4 )
				continue;

			mean_desc /= clusterIndices.size();

			
			vector<DVectorF> points;
			for( card32 k=0;k<clusterPoints.size();k++ )
			{
				DVectorF vec;
				vec.setDim(3);
				for( card32 l=0;l<3;l++ )
					vec[l] = clusterPoints[k][l];
				points.push_back( vec );
			}
			DynamicPCA<float32> pca( &points[0], (card32)points.size() );
			DynamicVector<float32> mean = pca.calcMean();
			Vector3f centroid = makeVector3f( mean[0],mean[1],mean[2] );
			
			DynamicVector<float32> eigenValues = pca.calcEigenValues();
			vector<float> signature;
			signature.push_back( fabs( eigenValues[0] ) );
			signature.push_back( fabs( eigenValues[1] ) );
			//signature.push_back( mean_desc[0] );
			//signature.push_back( mean_desc[1] );
			//signature.push_back( mean_desc[2] );
			signature.push_back( clusterID );
			//norm( eigenValues ), sqrt( absf( eigenValues[1] / max(eigenValues[2],1e-10f) )));
			FDAttribute * fa = new FDAttribute;
			fa->setNumComponents( (card32)signature.size() );
			fa->setupDescriptorData( &signature[0], (card32)signature.size() );

			/*// check if centroid is near the surface
			hIt.setSeekPointAndReset( centroid );
			if( norm( centroid - hIt.get3f( posAAT ) ) >  averagePointDistance * 1.5f )
				continue;*/

			Feature * f = new Feature;
			f->setPosition( centroid );
			f->setPointID(-1);
			f->addDescriptor( fa );
			card32 clusterColor[8];
			clusterColor[0] = 0xFF0000;
			clusterColor[1] = 0xFFFF00;
			clusterColor[2] = 0x00FFFF;
			clusterColor[3] = 0x0000FF;
			clusterColor[4] = 0xFF00FF;
			clusterColor[5] = 0xFFFFFF;
			clusterColor[6] = 0x000000;
			clusterColor[7] = 0x00FF00;
			
			//clusterColor[0] = 0x002200;
			//clusterColor[1] = 0x004400;
			//clusterColor[2] = 0x006600;
			//clusterColor[3] = 0x008800;
			//clusterColor[4] = 0x00AA00;
			//clusterColor[5] = 0x00CC00;
			//clusterColor[6] = 0x00FF00;
			//clusterColor[7] = 0x000000;
			f->setColor( clusterColor[clusterID] );
			features->push_back( f );

		}

		// setup distances

		bool useGeodetic = m_AnimationBSRContext->getRegistrationContext()->getUseGeodeticDistances();
		vector<float> distances( features->size() * features->size() );
		if( useGeodetic )
		{
			UnstructuredInCorePointCloud *tmpPC = new UnstructuredInCorePointCloud;
			tmpPC->clear();
			tmpPC->getPointSet()->assign( ps );


			InCorePCTopologyGraphPointSetIterator * topIter = new InCorePCTopologyGraphPointSetIterator( ps, NULL, 
				top );
			vector<Vector3f> points;
			for( card32 i=0;i<features->size();i++ )
			{
				points.push_back( features->get(i)->getPosition() );
			}
			for( card32 i=0;i<points.size();i++ )
			{
				vector<float> tmpDist = calculateGeodeticDistances( tmpPC, topIter, points[i], points );
				for( card32 j=0;j<tmpDist.size();j++ )
				{
					distances[i*features->size()+j] = tmpDist[j];
				}

			}
			delete topIter;
			delete tmpPC;
		}
		else
		{
			for( card32 i=0;i<features->size();i++ )
			{
				for( card32 j=i;j<features->size();j++ )
				{
					float dist = norm( features->get(i)->getPosition() - features->get(j)->getPosition() );
					distances[i*features->size()+j] = dist;
					distances[j*features->size()+i] = dist;
				}
			}
		}
		features->setupFeatureDistances( distances );

		if( pgc->getPatchGraph(i)->getNumFeatureLists() > 0 )
			pgc->getPatchGraph(i)->deleteFeatureLists(0, pgc->getPatchGraph(i)->getNumFeatureLists() );
	
		pgc->getPatchGraph(i)->addFeatureList( features );	


	}

	debugOutput << "Feature Clustering :" << timer.getDeltaValue() / 1000 << "s\n";
}

void FeatureRegistration::setCorrespondencesValid() 
{
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	pgc->getTempCorrespondences()->assign( pgc->getFeatureCorrespondences() );
	FeatureCorrespondences * fc = pgc->getTempCorrespondences();
	for( card32 i=0;i<fc->getNumFrames();i++ )
	{
		for( card32 j=0;j<fc->getFrameCorrespondences(i)->getNumCorrespondences();j++ )
			if( fc->getFrameCorrespondences(i)->get(j)->getFlags() != Correspondence::CF_Initial )
				fc->getFrameCorrespondences(i)->get(j)->setFlags(0);
	}
}

void FeatureRegistration::deletePatchGraphFrames() 
{
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer();
	for( int i=(int)m_EndFrame;i>=(int)m_StartFrame;i-- )
	{
		if( i >= (int)pgc->getNumPatchGraphs() )
			continue;

		if( i < 0 )
			break;

		pgc->deletePatchGraph(i);
	}
}

void FeatureRegistration::buildFeatureCorrespondences_OneFrame() 
{
	PatchGraphContainer * pgc = m_AnimationBSRContext->getPGContainer(); 
	if( m_AveragePointDistance == 0.0f )
	{
		UnstructuredInCorePointCloud * newPC = new UnstructuredInCorePointCloud;
		newPC->clear();
		newPC->getPointSet()->assign( pgc->getPatchGraph(0)->getTrajectorySet()->getTrajectories() );
		m_AveragePointDistance = geoxl::getAveragePointDistance(newPC);
		delete newPC;
	}
	//pgc->getFeatureCorrespondences()->setup( pgc->getAnimSeq()->getNumFrames() );
	if( pgc->getFeatureCorrespondences()->getNumFrames() != pgc->getAnimSeq()->getNumFrames() )
		pgc->getFeatureCorrespondences()->setup( pgc->getAnimSeq()->getNumFrames() );
	
	CorrespondenceList * newList = new CorrespondenceList;
	pgc->getFeatureCorrespondences()->getFrameCorrespondences( m_StartFrame )->assign( newList );

	if( pgc->getTempCorrespondences()->getNumFrames() != pgc->getAnimSeq()->getNumFrames() )
		pgc->getTempCorrespondences()->setup(pgc->getAnimSeq()->getNumFrames() );

	vector<PatchGraph*> frame1,frame2;
	card32 frameID = 0;
	for( card32 pgID=0;pgID<pgc->getNumPatchGraphs();pgID++ )
	{
		PatchGraph * pg = pgc->getPatchGraph( pgID );
		if( pg->getStartFrameIndex() == m_StartFrame )
		{
			frame1.push_back( pg );
		}
		if( pg->getStartFrameIndex() == m_StartFrame+1 )
		{
			frame2.push_back( pg );
		}
	}
	buildCorresponcences( frame1, frame2 );

	MultiFrameCorrespondenceValidation validation;
	//validation.validateCorrspondences( m_AnimationBSRContext->getPGContainer()->getFeatureCorrespondences() );
}

Matrix4f FeatureRegistration::calculateTransformation( const vector<FeaturePair> &pairs ) 
{
	// estimate centroids of features
	Vector3f centroidA = NULL_VECTOR3F, centroidB = NULL_VECTOR3F;
	for( card32 i=0;i<pairs.size();i++ )
	{
		centroidA += pairs[i].m_Feature1->getPosition() / (float)pairs.size();
		centroidB += pairs[i].m_Feature2->getPosition() / (float)pairs.size();
	}

	// calculate cross-covariance matrix
	Matrix3f C;
	C *= 0;
	for( card32 i=0;i<pairs.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (pairs[i].m_Feature1->getPosition()[j] - centroidA[j])
					* (pairs[i].m_Feature2->getPosition()[k] - centroidB[k]);
			}
		}
	}

   // Prepare OpenCV Datasructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
   CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
   cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

   // Do SVD
   cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f ret = makeTranslation4f(centroidA) * expand3To4(rotation) * makeTranslation4f(-centroidB);

	return ret;
}
