#ifndef FeatureRegistrationH
#define FeatureRegistrationH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class FeatureDescriptor;
class Feature;
class FeatureDetector;
class FeatureList;
class AnimationBSRContext;
class PatchGraph;
class Scene;

class FeaturePair
{
public:
	Feature	*m_Feature1,*m_Feature2;
	card32	m_Index1, m_Index2;
	float		m_DescriptorDistance;
	bool		m_ValidatedCorrespondence;
};

/**
	main class for registration with features
*/
class ANIMATIONRECONSTRUCTION_API FeatureRegistration : public Persistent
{
	GEOXL_CLASS( FeatureRegistration )

private:

	void setupFromContext();

	FeatureDescriptor *	m_FeatureDescriptor;
	FeatureDetector	*	m_FeatureDetector;
	float32					m_FeatureRadius;		// relative to AveragePointDistance
	float32					m_AveragePointDistance;
	BoundingBox3f			m_SceneBB;
	bool						m_UseGeodeticDistances;
	AnimationBSRContext * m_AnimationBSRContext;
	card32					m_StartFrame, m_EndFrame;

	float32					m_SpectralSigma;
	float32					m_SpectralAddDiff;
	float32					m_SpectralScaleDiff;


	


public:
	FeatureRegistration(void);
	~FeatureRegistration(void);

   virtual void assign(const Object* obj, CopyContext *context = NULL);
	
	void setup( FeatureDetector * det, FeatureDescriptor * desc );

	/// setup for SIFT feature detection + descriptors
	/// gradientAttribute		-		source attribute for gradients (position,color)
	/// numberOfScale				-		number of scales in multi-scale representation
	/// startSigma					-		sigma for a scale: sigma[i] = startSigma * (F ^ i) * averagePointDistance
	/// F								-		see sigma
	/// averagePointDistance	-		will be calculated if zero
	/// numSampleRegions			-		SIFT descriptor consists of NxN keypointdescriptors, N is numSampleRegions
	/// numOrientationBins		-		number of bins for a SIFT keypointdescriptor
	void setupSIFT( string gradientAttribute, card32 numberOfScales = 5, float startSigma = 1.0f,
		float F = 2.25f, float averagePointDistance = 0.0f, card32 numSampleRegions=2, card32 numOrientationBins=8 );

	/// setup for multiscale feature detection + histogram descriptor
	/// sourceAttribute			-		builds histogram from this attribute (first component)
	/// numberOfScale				-		number of scales in multi-scale representation
	/// startSigma					-		sigma for a scale: sigma[i] = startSigma * (F ^ i) * averagePointDistance
	/// F								-		see sigma
	/// averagePointDistance	-		will be calculated if zero
	/// numRadialHistograms		-		descriptor consists of numRadialHistograms histograms representing an interval(radius)
	/// numBins						-		number of bins for each histogram
	void setupHistogram( string sourceAttribute, card32 numberOfScales, float startSigma = 1.0f,
		float F = 2.25f, float averagePointDistance = 0.0f,card32 numRadialHistograms=3, card32 numBins=8 );

	/// setup for multiscale feature detection + funkhouser descriptor
	/// sourceAttribute			-		calculates spectrum from attribute[componentNumber]
	/// componentNumber			-		^^
	/// numberOfScale				-		number of scales in multi-scale representation
	/// startSigma					-		sigma for a scale: sigma[i] = startSigma * (F ^ i) * averagePointDistance
	/// F								-		see sigma
	/// averagePointDistance	-		will be calculated if zero
	/// numRadialSamples			-		descriptor consists of numRadialSamples circular spectrums representing an interval(radius)
	/// numCircularSamples		-		
	void setupFunkhouser2D( string sourceAttribute, card32 componentNumber, card32 numberOfScales, 
		float startSigma = 1.0f, float F = 2.25f, float averagePointDistance = 0.0f,
		card32 numRadialSamples=4, card32 numCircularSamples=8 );

	
	enum EDetectionMode
	{
		DETECT_POS = 1,
		DETECT_COLOR = 2,
		DETECT_POS_AND_COLOR = 3
	};

	/// returns valid feature pairs "the black box function"
	/// mode may be DETECT_POS, DETECT_COLOR, DETECT_POS_AND_COLOR
	vector<FeaturePair>	estimateFeaturePairs( PointCloud * pc1, PointCloud * pc2, EDetectionMode mode );

	/// detects features using the specified feature detector and
	/// calculates a featuredescriptor for each feature
	/// mode may be DETECT_POS, DETECT_COLOR, DETECT_POS_AND_COLOR
	FeatureList *		calculateFeatures( PointCloud * pc, EDetectionMode mode );

	/// builds initial feature pairs based on descriptor distances
	/// returns best pairs (sorted by descriptor difference) limited by maxNumberOfPairs
	vector<FeaturePair>	buildFeaturePairs( FeatureList *features1,FeatureList *features2, float maxDistance, int maxNumberOfPairs=1000);

	/// spectral feature validation
	vector<FeaturePair>	validateFeaturePairs( const vector<FeaturePair> pairs, PointCloud *p1, PointCloud *p2);

	Matrix4f calculateTransformation( const vector<FeaturePair> &pairs );

	float32 getFeatureRadius() const {return m_FeatureRadius;}
	void setFeatureRadius(float32 v) {m_FeatureRadius=v;}

	float32 getAveragePointDistance() const {return m_AveragePointDistance;}
	void setAveragePointDistance(float32 v) {m_AveragePointDistance=v;}

	BoundingBox3f getSceneBB() const {return m_SceneBB;}
	void setSceneBB(BoundingBox3f v) {m_SceneBB=v;}

	bool getUseGeodeticDistances() const {return m_UseGeodeticDistances;}
	void setUseGeodeticDistances(bool v) {m_UseGeodeticDistances=v;}

	AnimationBSRContext * getAnimationBSRContext() const {return m_AnimationBSRContext;}
	void setAnimationBSRContext(AnimationBSRContext * v) {m_AnimationBSRContext=v;}

	void calculateFeaturesPG( PatchGraph * frame );

	void buildCorresponcences( vector<PatchGraph*> frame1, vector<PatchGraph*> frame2 );

	// --------------------------------------------------------------------------
	// public methods (memberbrowser)

	/// calculates corner points for every frame in AnimationBSRContext
	void calculateAllFeatures();

	/// calculates corner points frame[0] in AnimationBSRContext
	void calculateFeatures_OneFrame();

	/// builds correspondences between every frame + validation
	void buildFeatureCorrespondences();

	void buildFeatureCorrespondences_OneFrame();

	void calculateAllFeatures_Clustering();

	void setCorrespondencesValid();

	void deletePatchGraphFrames();

	Scene * debugOnly_Scene;

	float32 getSpectralSigma() const {return m_SpectralSigma;}
	void setSpectralSigma(float32 v) {m_SpectralSigma=v;}

	float32 getSpectralAddDiff() const {return m_SpectralAddDiff;}
	void setSpectralAddDiff(float32 v) {m_SpectralAddDiff=v;}

	float32 getSpectralScaleDiff() const {return m_SpectralScaleDiff;}
	void setSpectralScaleDiff(float32 v) {m_SpectralScaleDiff=v;}


};

}

#endif