#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MLSMultiScaleFeatureDetector.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "PointCloudTools.h"
#include "VertexDescriptor.h"
#include "StringHelper.h"
#include "ProgressWindow.h"
#include "HierarchicalKNNIterator.h"
#include "VertexArray.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( MLSMultiScaleFeatureDetector ,0)
{
	BEGIN_CLASS_INIT( MLSMultiScaleFeatureDetector );
	ADD_CARD32_PROP(m_NumberOfScales,0);
	ADD_FLOAT32_PROP( m_H0, 0 );
	ADD_FLOAT32_PROP( m_F, 0 );
	ADD_FLOAT32_PROP( m_AveragePointDistance, 0 );
}

MLSMultiScaleFeatureDetector::MLSMultiScaleFeatureDetector(void)
{
	m_NumberOfScales = 4;
	m_H0 = 1.5;
	m_F = 2.25f;
	m_AveragePointDistance = 0.0f;

}

MLSMultiScaleFeatureDetector::~MLSMultiScaleFeatureDetector(void)
{
}

void MLSMultiScaleFeatureDetector::assign(const Object* obj, CopyContext *context)
{
	FeatureDetector::assign(obj,context);
	const MLSMultiScaleFeatureDetector * o = dynamic_cast<const MLSMultiScaleFeatureDetector*>(obj);
	pAssert( o != NULL );

	m_NumberOfScales = o->m_NumberOfScales;
	m_H0 = o->m_H0;
	m_F = o->m_F;
	m_AveragePointDistance = o->m_AveragePointDistance;
}


FeatureList * MLSMultiScaleFeatureDetector::detectFeatures( PointCloud * pc, float radius ) 
{
	// check input point cloud
	if( m_Attribute.length() == 0 )
	{
		if( !pc->providesAttribute( "normal") )
			throw "Bad point cloud!";

		for( card32 i=0;i<m_NumberOfScales;i++ )
		{
			string iStr = intToStr(i);
			if( !pc->providesAttribute( string("pos") + iStr) 
				|| !pc->providesAttribute( string("normal") + iStr))
				throw "Bad point cloud!";
		}
	}
	else
	{
		if( !pc->providesAttribute( m_Attribute ) )
			throw "Bad point cloud!";

		for( card32 i=0;i<m_NumberOfScales;i++ )
		{
			string iStr = intToStr(i);
			if( !pc->providesAttribute( m_Attribute + iStr) )
				throw "Bad point cloud!";
		}
	}

	// calculate average point distance (if not defined)
	float averagePointDistance;
	if( m_AveragePointDistance == 0.0f )
		averagePointDistance = geoxl::getAveragePointDistance( pc );
	else
		averagePointDistance = m_AveragePointDistance;

	// estimate h for each scale
	vector<float> h;
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		h.push_back( m_H0 * pow( m_F, (float)i ) );
	}


	// ------------- search for maximas/minimas
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC
		| SceneObjectIteratorType::CAP_MODIFYABLE_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc,32,NULL);

	vector<AAT> mlsPosAAT,normalAAT;
	vector<AAT> attrAAT;
	if( m_Attribute.length() == 0 )
	{
		for( card32 i=0;i<m_NumberOfScales;i++ )
		{
			mlsPosAAT.push_back( pc->getAAT( string("pos") + intToStr(i) ) );
			normalAAT.push_back( pc->getAAT( string("normal") + intToStr(i) ) );
		}
	}
	else
	{
		for( card32 i=0;i<m_NumberOfScales;i++ )
		{
			attrAAT.push_back( pc->getAAT( m_Attribute + intToStr(i) ) );
		}
	}
	AAT posAAT = pc->getAAT( "position" );

	FeatureList * features = new FeatureList;
	//progressWindow->pushStep(true, "min/max");
	card32 index = 0;
	while( !bIt->atEnd() )
	{
		// get current data
		vector<Vector3f> mlsPosition(m_NumberOfScales);
		vector<Vector3f> mlsNormal(m_NumberOfScales);
		vector<float> attributeData(m_NumberOfScales);
		vector<Vector3f> colorData(m_NumberOfScales);
		Vector3f pos = bIt->get3f(posAAT);
		if( m_Attribute.length() == 0 )
		{
			for( card32 j=0;j<m_NumberOfScales;j++ )
			{
				mlsPosition[j] = bIt->get3f( mlsPosAAT[j] );
				mlsNormal[j] = bIt->get3f( normalAAT[j] );
			}
		}
		else
		{
			for( card32 j=0;j<m_NumberOfScales;j++ )
			{
				attributeData[j] = bIt->get1f( attrAAT[j] );
			}
		}

		// estimate radi
		float C = 0.5f;
		vector<float> radius(m_NumberOfScales+2);
		for( card32 j=1;j<m_NumberOfScales;j++ )
		{
			radius[j+1] = C * h[j] * averagePointDistance;
		}
		radius[0] = 0;
		radius.back() = radius[m_NumberOfScales-1];

		// calculate normal difference d or simple attribute difference
		// add additional two scales with 0
		vector<float> d(m_NumberOfScales+1);
		if( m_Attribute.length() == 0 )
		{
			for( card32 j=0;j<m_NumberOfScales-1;j++ )
			{
				d[j+1] = mlsNormal[j+1] * (mlsPosition[j+1] - mlsPosition[j]);
			}
		}
		else
		{
			for( card32 j=0;j<m_NumberOfScales-1;j++ )
			{
				d[j+1] = attributeData[j+1] - attributeData[j];
			}
		}
		d[0] = 0;
		d.back() = 0;

		// inital minima/maxima test
		card32 numberOfExtremePoints = 0;
		vector<bool> isMaxima(m_NumberOfScales-1,false),isMinima(m_NumberOfScales-1,false);
		for( card32 j=1;j<m_NumberOfScales;j++ )
		{
			if( d[j] > d[j-1] && d[j] > d[j+1] )
			{
				numberOfExtremePoints++;
				isMaxima[j-1] = true;				
			}
			if( d[j] < d[j-1] && d[j] < d[j+1] )
			{
				numberOfExtremePoints++;
				isMinima[j-1] = true;				
			}
		}

		// test local enviroment
		hIt->setSeekPointAndReset( pos );
		hIt->next();
		int n=0;
		while(numberOfExtremePoints>0 && !hIt->atEnd() )
		{
			vector<Vector3f> nPos(m_NumberOfScales);
			vector<Vector3f> nNormal(m_NumberOfScales);
			vector<float>	  nAttributeData(m_NumberOfScales);
			vector<float> d_n(m_NumberOfScales+1);
			Vector3f nPosOrg = hIt->get3f( posAAT );
			float dist = norm( pos - nPosOrg );
			if( dist > radius.back() )
				break;

			// get neighbour data and calculate difference
			if( m_Attribute.length() == 0)
			{
				for( card32 j=0;j<m_NumberOfScales;j++ )
				{
					nPos[j] = hIt->get3f( mlsPosAAT[j] );
					nNormal[j] = hIt->get3f( normalAAT[j] );
				}
				for( card32 j=0;j<m_NumberOfScales-1;j++ )
				{
					d_n[j+1] = nNormal[j+1] * (nPos[j+1] - nPos[j]);
				}
			}
			else
			{
				for( card32 j=0;j<m_NumberOfScales;j++ )
				{
					nAttributeData[j] = hIt->get1f( attrAAT[j] );
				}
				for( card32 j=0;j<m_NumberOfScales-1;j++ )
				{
					d_n[j+1] = nAttributeData[j+1] - nAttributeData[j];
				}
			}

			d_n[0] = 0;
			d_n.back() = 0;


			// calculate normal difference d

			for( card32 j=1;j<m_NumberOfScales;j++ )
			{
				if( isMaxima[j-1] )
				{
					if( (d[j] <= d_n[j-1] && dist < radius[j]) || 
						 (d[j] <= d_n[j] && dist < radius[j+1]) ||  
						 (d[j] <= d_n[j+1] && dist < radius[j+2]) )
					{
						numberOfExtremePoints--;
						isMaxima[j-1] = false;				
					}
				}
				if( isMinima[j-1] )
				{
					if( (d[j] >= d_n[j-1] && dist < radius[j]) || 
						 (d[j] >= d_n[j] && dist < radius[j+1])  ||  
						 (d[j] >= d_n[j+1] && dist < radius[j+2]) )
					{
						numberOfExtremePoints--;
						isMinima[j-1] = false;				
					}
				}
			}

			n++;
			hIt->next();
		}

		card32 colors[20];
		colors[0] = 0xffFF0000;
		colors[1] = 0xffFF2200;
		colors[2] = 0xffFF4400;
		colors[3] = 0xffFF8800;
		colors[4] = 0xffFFCC00;
		colors[5] = 0xffFFFF00;
		colors[6] = 0xff000000;
		colors[7] = 0xff000000;
		colors[8] = 0xff000000;
		
		if( numberOfExtremePoints > 0 )
		{
			Feature *f = new Feature;
			f->m_Magnitude = numberOfExtremePoints;
			for( card32 i=0;i<isMaxima.size();i++ )
				if( isMaxima[i] || isMinima[i] )
				{
					f->setSize( radius[i+1] );
					f->m_Position = mlsPosition[i+1];
					vector<float> sign;
					sign.push_back(i);
					sign.push_back(isMaxima[i] ? 1 : 0 );
					f->setDetectorSignature(sign);
					if( isMaxima[i] )
						f->setColor( colors[i] );
					else
						f->setColor( colors[i] | 0x000000ff );
					break;
				}
			f->setPointID( index );
			features->push_back( f );
		}

		/*
		if( !progressWindow->progress( (float)index++/(float)pc->getNumPoints() * 100.0f ) )
			break;
			*/
		bIt->next();
	}
//	bool aborted;
//	progressWindow->popSteps(aborted);
	//debugOutput << "Detected " << features->size() << " features\n";


	delete iter;
	delete hIt;

	return features;	
}
