#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiScaleGenerator.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
#include "MultiResolutionPC.h"
#include "ProgressWindow.h"
#include "CachedRangeQuerry.h"
#include "StringHelper.h"
#include "LinearTransformPriors.h"
#include "Timer.h"
#include <float.h>
#include "VertexArray.h"
#include "HeighfieldParameterization.h"
#include "Monomials2ndOrder.h"
#include "ProgressWindow.h"
#include "Timer.h"
#include "PointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "FastFeature.h"
#include "FeatureAttachment.h"
#include "DebugRenderer.h"
#include "PointSet_templCode.h"
#include "PointCloudTools.h"
#ifdef _OPENMP
#include <omp.h>
#endif
//---------------------------------------------------------------------------
using namespace geoxl;

MultiScaleGenerator::TFastFeatureInfo MultiScaleGenerator::ffInfo;

IMPLEMENT_GEOXL_CLASS( MultiScaleGenerator ,0)
{
	BEGIN_CLASS_INIT( MultiScaleGenerator );
}

MultiScaleGenerator::MultiScaleGenerator(bool mlsSurface)
{
	m_NumberOfScales = 6;
	m_ScaleVD = new VertexDescriptor;
	if( mlsSurface )
	{
		m_ScaleVD->pushAttrib( VAD( "pos", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		m_ScaleVD->pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	}
	m_AdditionalScaleVD = new VertexDescriptor;
	m_H0 = 1.0f;
	m_F = 2.25f;
	m_MinNumberOfSamples = 50;
}

MultiScaleGenerator::~MultiScaleGenerator(void)
{
	delete m_ScaleVD;
	delete m_AdditionalScaleVD;
}

void MultiScaleGenerator::assign(const Object* obj, CopyContext *context)
{
	//Persistent::assign(obj,context);
	const MultiScaleGenerator * o = dynamic_cast<const MultiScaleGenerator*>(obj);
	pAssert( o != NULL );

	m_NumberOfScales = o->m_NumberOfScales;
	m_ScaleVD->assign( o->m_ScaleVD );
	m_AdditionalScaleVD->assign( o->m_AdditionalScaleVD );
	m_H0 = o->m_H0;
	m_F = o->m_F;
	m_MinNumberOfSamples = o->m_MinNumberOfSamples;

}

struct TAttribute
{
	string name, srcName;
	card32 numComponents;
	AAT	 srcAAT,destAAT;
	card32 scale;
	bool	 normalize;
};

template<class FloatType,int dim> 
StaticVector<FloatType,dim> smoothAttribute( PointSet * ps, const Vector3f &p0,
														  AAT posAAT, AAT attrAAT, float inv_h_sqr,
														  bool do_normalize)
{
	float summed_weight = 0.0f;
	StaticVector<FloatType,dim> ret;
	for( card32 i=0;i<dim;i++ )
		ret[i] = 0;

	for( card32 i=0;i<ps->getNumEntries();i++ )
	{
		StaticVector<FloatType,dim> attr = ps->get<FloatType,dim>(i, attrAAT);
		Vector3f pos = ps->get3f( i,posAAT );
		float weight = exp( -normQuad(pos-p0) * inv_h_sqr );
		ret += (attr) * weight;
		summed_weight += weight;
	}

	if( summed_weight > 0 )
		if( do_normalize )
		{
			return normalize( ret );
		}
		else
			return ret / summed_weight;

	return ret;	// should never happen
}

struct TMultiScaleContext
{
	CachedRangeQuerry * querry;
	AAT posAAT, normalAAT;
	float h;
	Vector3f pos;
	Vector3f normal;
	float minRadius;
};

TMultiScaleContext MLSContext;

float mlsFunc( float t )
{
	Vector3f x = MLSContext.pos + MLSContext.normal*t;
	float h=MLSContext.h;
	float h2 = h*h;
	float ret = 0;

	float radius = max(MLSContext.minRadius,sqrt( -h2 / log(0.001f) ));
	PointSet * ps = MLSContext.querry->getSpherePoints( x, radius, 0.1f );
	if( ps->getNumEntries() == 0 )
	{
		static int numZ = 0;
		numZ++;
		if( numZ % 100 == 0 )
			debugOutput << "numZ: " << numZ << "\n";
		return 1e20f;
	}
	
	float summed_weight = 0;
	for( card32 i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f pos = ps->get3f( i, MLSContext.posAAT );
		Vector3f normal = ps->get3f( i, MLSContext.normalAAT );
		float weight = exp( - normQuad( x - pos) / h2 );
		summed_weight += weight;
		ret += weight * pow( normal * (x-pos),2.0f );
	}

	if( summed_weight>0.0f )
		return ret / summed_weight;
	return ret;
}

PointCloud *MultiScaleGenerator::create( PointCloud * pc, MultiResolutionPC * mresPC, float pcAveragePointDistance ) 
{
	float maxRad = 20.0f;
	if( pcAveragePointDistance == 0.0f )
		pcAveragePointDistance = getAveragePointDistance(pc);
	// ----------------------------------
	// create point cloud
	VertexDescriptor * vd = new VertexDescriptor;
	vector<TAttribute>	smoothingAttributes;
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		for( card32 j=0;j<m_ScaleVD->getNumAttributes();j++ )
		{
			VAD vad = m_ScaleVD->getAttribute(j);
			string name = vad.name + intToStr(i);
			vd->pushAttrib( VAD( name, vad.numComponents, vad.dataFormat ) );
			if( vad.name == "pos" )
				continue;
			TAttribute attr;
			attr.name = name;
			attr.srcName = vad.name;
			attr.numComponents = vad.numComponents;
			attr.scale = i;
			attr.normalize = (vad.name == "normal");
			smoothingAttributes.push_back(attr);
		}
		for( card32 j=0;j<m_AdditionalScaleVD->getNumAttributes();j++ )
		{
			VAD vad = m_AdditionalScaleVD->getAttribute(j);
			string name = vad.name + intToStr(i);
			vd->pushAttrib( VAD( name, vad.numComponents, vad.dataFormat ) );
		}
	}
	PointCloud * retPC = pc->appendFields(vd);

	// ----------------------------------
	// init iterator and AATs
	SceneObjectIterator * iter = retPC->createIterator( SceneObjectIteratorType::CAP_BASIC_PC
		| SceneObjectIteratorType::CAP_MODIFYABLE_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	for( card32 i=0;i<smoothingAttributes.size();i++ )
	{
		smoothingAttributes[i].destAAT = retPC->getAAT( smoothingAttributes[i].name );
		smoothingAttributes[i].srcAAT = pc->getAAT( smoothingAttributes[i].srcName );
	}

	// calc H for each scale and create cached querries
	vector<CachedRangeQuerry*> querries;
	vector<AAT> mlsPosAATs, normalAATs;
	vector<float> H;
	//debugOutput << "m_MinNumberOfSamples = " << m_MinNumberOfSamples << "\n";
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		float h = m_H0 * pow( m_F,(float)i);
		//if( h > 10.0f )
		//	h = 10.0f;
		H.push_back( h );
		float kernelRadius = min( pcAveragePointDistance * maxRad, sqrt( -log(0.05f) * pow(h * pcAveragePointDistance,2.0f) ));
		float requiredAveragePointDistance = 2.0f * kernelRadius / sqrt(m_MinNumberOfSamples);
		//debugOutput << "scale[" << i << "]: kernelRadius = " << kernelRadius << " reqAv = " << requiredAveragePointDistance
		//	<< " first res dist = " << mresPC->getPCAveragePointDistance(0) << "\n";
		PointCloud * resPC;
		if( requiredAveragePointDistance < mresPC->getPCAveragePointDistance(0) )
			resPC = pc;
		else
			resPC = mresPC->getClosestPointCloud( requiredAveragePointDistance );
		querries.push_back( new CachedRangeQuerry(resPC) );
	}

	// ------------------------------------------------------------
	// smooth attributes
	card32 index = 0;
	progressWindow->pushStep(true, "smoothing");
	bool aborted = false;
	AAT posAAT = retPC->getAAT( "position" );
	Timer timer;
	card32 startSmoothingTime = timer.getCurrentValue();
	vector<float> numSmoothingScamples(smoothingAttributes.size(),0);
	while( !bIt->atEnd() && !aborted)
	{
		Vector3f pos = bIt->get3f(posAAT);

		for( card32 i=0;i<smoothingAttributes.size();i++ )
		{
			TAttribute a = smoothingAttributes[i];
			float radius = min(pcAveragePointDistance * maxRad,  sqrt( -pow(H[a.scale] * pcAveragePointDistance,2.0f) * log( 0.05f ) ));
			PointSet * querriedPoints = querries[ a.scale ]->getSpherePoints( pos, radius,0 );
			numSmoothingScamples[i] += (float)querriedPoints->getNumEntries() / (float)retPC->getNumPoints();
			float inv_h_sqr = 1.0f / pow(H[a.scale] * pcAveragePointDistance,2.0f);
			
			if( a.numComponents == 1 )
			{
				StaticVector<float,1> attr = smoothAttribute<float,1>( querriedPoints, pos, posAAT, a.srcAAT, inv_h_sqr, a.normalize );
				mIt->set1f( a.destAAT, attr[0] );
			}
			if( a.numComponents == 2 )
			{
				StaticVector<float,2> attr = smoothAttribute<float,2>( querriedPoints, pos, posAAT, a.srcAAT, inv_h_sqr, a.normalize );
				mIt->set2f( a.destAAT, attr );
			}
			if( a.numComponents == 3 )
			{
				StaticVector<float,3> attr = smoothAttribute<float,3>( querriedPoints, pos, posAAT, a.srcAAT, inv_h_sqr, a.normalize );
				mIt->set3f( a.destAAT, attr );
			}
			if( a.numComponents == 4 )
			{
				StaticVector<float,4> attr = smoothAttribute<float,4>( querriedPoints, pos, posAAT, a.srcAAT, inv_h_sqr, a.normalize );
				mIt->set4f( a.destAAT, attr );
			}
		}

		index++;
		if( index % 100 == 1 )
			aborted = !progressWindow->progress( (float)index / (float)retPC->getNumPoints() * 100.0f );
		bIt->next();
	}
	progressWindow->popSteps(aborted);
	card32 deltaSmoothingTime = timer.getCurrentValue() - startSmoothingTime;
	//debugOutput << "Smoothing:" << deltaSmoothingTime / 1000 << "s " << deltaSmoothingTime % 1000 << "ms\n";

	for( card32 i=0;i<numSmoothingScamples.size();i++ )
	{
		//debugOutput << "num samples[" << i << "] = " << numSmoothingScamples[i] << "\n";
	}

	if( !m_ScaleVD->providesAttribute( "pos" ) )
		return retPC;

	// ------------------------------------------------------------
	// create a temporary point cloud for each scale by merging multires points & multiscale normals
	vector<PointCloud*> tempPC;
	vector<CachedRangeQuerry*> mlsQuerries;
	vector<AAT> projNormalAAT, destProjNormalAAT, projPosAAT, destProjPosAAT;
	vector<HierarchicalKNNIterator *> hIterators(m_NumberOfScales);
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		string destPosName = string("pos") + intToStr(i);
		string sourcePosName = "position";
		string projNormalsName = string("normal") + intToStr(i);
		destProjPosAAT.push_back( retPC->getAAT(destPosName) ); 
		destProjNormalAAT.push_back( retPC->getAAT(projNormalsName) );

		// calc required resolution
		float kernelRadius = min( pcAveragePointDistance * maxRad, sqrt( -pow(H[i] * pcAveragePointDistance,2.0f) * log(0.05f)));
		float requiredAveragePointDistance = 2.0f * kernelRadius / sqrt(m_MinNumberOfSamples) * 0.5f;
		if( requiredAveragePointDistance < mresPC->getPCAveragePointDistance(0) )
		{
			tempPC.push_back(NULL);
			hIterators[i] = new HierarchicalKNNIterator(retPC,32,NULL);
			mlsQuerries.push_back( new CachedRangeQuerry( retPC ) );
			projNormalAAT.push_back( retPC->getAAT( projNormalsName) );
			projPosAAT.push_back( retPC->getAAT( sourcePosName ) );
			continue;
		}

		VertexDescriptor * vd = new VertexDescriptor;
		vd->pushAttrib( VAD( projNormalsName, 3, VAD::DATA_FORMAT_FLOAT32 ) );
		PointCloud * tmppc = mresPC->getClosestPointCloud( requiredAveragePointDistance )
			->appendFields( vd );

		projNormalAAT.push_back( tmppc->getAAT( projNormalsName) );
		projPosAAT.push_back( tmppc->getAAT( sourcePosName ) );

		HierarchicalKNNIterator hIt(retPC,32,NULL);
		SceneObjectIterator * iter = tmppc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC
			| SceneObjectIteratorType::CAP_MODIFYABLE_PC );
		pAssert( iter != NULL );
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

		AAT posAAT = tmppc->getAAT( "position" );
		AAT sourceNormalAAT = retPC->getAAT( projNormalsName );

		while( !bIt->atEnd() )
		{
			Vector3f pos = bIt->get3f( posAAT );
			hIt.setSeekPointAndReset( pos );
			mIt->set3f( projNormalAAT.back(), hIt.get3f(sourceNormalAAT) );
			bIt->next();
		}

		tempPC.push_back( tmppc );
		mlsQuerries.push_back( new CachedRangeQuerry( tmppc ) );
		hIterators[i] = new HierarchicalKNNIterator(tmppc,32,NULL);

		delete iter;
	}


	// ------------------------------------------------------------
	// project position
	card32 startProjectionTime = timer.getCurrentValue();
	index = 0;
	bIt->reset();
	progressWindow->pushStep(true, "projecting");
	vector<AAT> testNormalAAT;
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		testNormalAAT.push_back( retPC->getAAT( string("normal") + intToStr(i)));
	}
	while( !bIt->atEnd() && !aborted)
	{
		Vector3f pos = bIt->get3f(posAAT);

		for( card32 i=0;i<m_NumberOfScales;i++ )
		{

			Vector3f normal = bIt->get3f(testNormalAAT[i]);
			float radius = min(pcAveragePointDistance * maxRad,sqrt( -pow(H[i] * pcAveragePointDistance,2.0f) * log( 0.05f ) ));
			//Vector3f normal = bIt->get3f( projNormalAAT[i] );
			Matrix3f tangentSystem = calcTangentSystem( normal );
			Matrix3f invTangentSystem = tangentSystem.transpose();


			vector<Vector3f> posBuffer;
			HierarchicalKNNIterator * hIt = hIterators[i];
			hIt->setSeekPointAndReset( pos );
			while( !hIt->atEnd() )
			{
				Vector3f p = hIt->get3f(posAAT);
				Vector3f n = hIt->get3f(projNormalAAT[i]);
				if( normal * n > 0.0f )
					posBuffer.push_back( invTangentSystem * (p-pos));
				if( norm(p-pos) > radius && posBuffer.size() > 0)
					break;
				hIt->next();
			}

/*			PointSet * querriedPoints = mlsQuerries[ i ]->getSpherePoints( pos, radius );

			for( card32 j=0;j<querriedPoints->getNumEntries();j++ )
			{
				Vector3f p = querriedPoints->get3f(j,posAAT);
				Vector3f n = querriedPoints->get3f(j,projNormalAAT[i]);
//				if( norm(p-pos) > radius )
//					break;
				if( normal * n > 0.0f )
					posBuffer.push_back( invTangentSystem * (p-pos));
			}*/
			
			StaticVector<float32, 6> coeff;
			vector< StaticVector<float32, 6> > coeffsDerivs;
			if( posBuffer.size() > 5 )
				GLTMonomials2ndOrder::computeCoeffs( posBuffer, coeff, coeffsDerivs, H[i] * pcAveragePointDistance );
			else
				coeff[0] = 0;

			Vector3f mlsPos = pos + normal * coeff[0];
			
			mIt->set3f(destProjPosAAT[i], mlsPos);

		}

		index++;
		if( index % 100 == 1 )
			aborted = !progressWindow->progress( (float)index / (float)retPC->getNumPoints() * 100.0f );
		bIt->next();
	}
	progressWindow->popSteps();
	card32 deltaProjectionTime = timer.getCurrentValue() - startProjectionTime;
	//debugOutput << "Projection:" << deltaProjectionTime / 1000 << "s " << deltaProjectionTime % 1000 << "ms\n";

	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		delete tempPC[i];
		delete querries[i];
		delete mlsQuerries[i];
		delete hIterators[i];
	}

	return retPC;
}

void MultiScaleGenerator::addVertexAttributef( string name, int numComponents ) 
{
	m_ScaleVD->pushAttrib( VAD( name, numComponents, VAD::DATA_FORMAT_FLOAT32 ) );
}


void MultiScaleGenerator::addAdditionalVertexAttributef( string name, int numComponents ) 
{
	m_AdditionalScaleVD->pushAttrib( VAD( name, numComponents, VAD::DATA_FORMAT_FLOAT32 ) );
}

void MultiScaleGenerator::setupPosition() 
{
	m_ScaleVD->clear();
	m_ScaleVD->pushAttrib( VAD( "pos", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	m_ScaleVD->pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	m_AdditionalScaleVD->clear();
}

void MultiScaleGenerator::setupColor() 
{
	m_ScaleVD->clear();
	m_ScaleVD->pushAttrib( VAD( "gray", 1, VAD::DATA_FORMAT_FLOAT32 ) );
	m_AdditionalScaleVD->clear();
}

void MultiScaleGenerator::setupPositionAndColor() 
{
	m_ScaleVD->clear();
	m_ScaleVD->pushAttrib( VAD( "pos", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	m_ScaleVD->pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	m_ScaleVD->pushAttrib( VAD( "gray", 1, VAD::DATA_FORMAT_FLOAT32 ) );
	m_AdditionalScaleVD->clear();
}

PointCloud *MultiScaleGenerator::smoothPC( PointCloud * pc, MultiResolutionPC * mresPC, float sigma, float patchSamplingSpace, float sourceSamplingSpace, float pcAveragePointDistance) 
{
	if( pcAveragePointDistance == 0.0f )
		pcAveragePointDistance = getMedianPointDistance( pc );
	patchSamplingSpace *= pcAveragePointDistance;
	sourceSamplingSpace *= pcAveragePointDistance;

	// create output point cloud and set all positions to (0,0,0)
	UnstructuredInCorePointCloud * ret = new UnstructuredInCorePointCloud;
	ret->assign( pc );
	VertexArray retVA( ret->getPointSet() );
	for( card32 i=0;i<retVA.getNumElements();i++)
	{
		retVA.setPosition3f(i, makeVector3f(0,0,0));
		retVA.setNormal3f(i, makeVector3f(0,0,0));
	}
	
	vector<float> weight( pc->getNumPoints(), 0);
	UnstructuredInCorePointCloud * samplePC = (UnstructuredInCorePointCloud *)mresPC->getClosestPointCloud( patchSamplingSpace );
	VertexArray var( samplePC->getPointSet() );
	PointCloud * sourcePC = mresPC->getClosestPointCloud(sourceSamplingSpace);
	HeighfieldParameterization hparam;
	
	

	
	// calculate patch for each point in sample pc and "splat" the projected points into ret pc
	float radius = sqrt( -log( 0.01f ) * sqr(sigma*pcAveragePointDistance));
	for( card32 i=0;i<var.getNumElements();i++)
	{
		// parameterize neighbour points
		Vector3f samplePos = var.getPosition3f(i);
		hparam.calculate(pc, samplePos, radius);
		PointSet * ps = hparam.getParametricPoints();	
		vector<Vector3f> points;
		AAT paramposAAT = ps->getAAT( "param_pos" );
		AAT heightAAT = ps->getAAT( "height" );
		AAT normalAAT = ps->getAAT( "normal");
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector2f ppos = ps->get2f(j,paramposAAT);
			float height = ps->get1f(j, heightAAT);
			points.push_back( makeVector3f(ppos[0], ppos[1], height)/radius );
		}

		// evaluate patch coefficients
		StaticVector<float,6> coeffs;
		Monomials2ndOrder::computeCoeffs( points, coeffs);

		// project points onto patch
		AAT posAAT = ps->getAAT( "position" );
		AAT pointIndexAAT = ps->getAAT( "point_index" );
		Vector3f patchNormal = normalize(shrink4To3( hparam.getParametricToWorld4f()[2]));
		if( var.getNormal3f(i) * patchNormal < 0 )
			patchNormal *= -1;
		float scaleGaussian = sqr(sigma * pcAveragePointDistance);
		Matrix3f paramToWorld3f = shrink4To3( hparam.getParametricToWorld4f());
		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			Vector3f npos = ps->get3f(j, posAAT);
			int index = ps->get1i(j,pointIndexAAT);

			float pointWeight = exp( -normQuad(npos - samplePos) / scaleGaussian); // hier stimmt was nscht!
			npos = npos + patchNormal * (patchNormal * (samplePos - npos));
			float u = points[j][0];
			float v = points[j][1];
			npos += patchNormal * Monomials2ndOrder::evaluate(coeffs, makeVector2f(u,v)) * radius;
			float derivateU = 2.0f * coeffs[3] * u + coeffs[4] * v + coeffs[1];
			float derivateV = 2.0f * coeffs[5] * v + coeffs[4] * u + coeffs[2];
			Vector3f normal = -normalize( makeVector3f(1,0, derivateU).crossProduct( makeVector3f(0,1,derivateV) ) );
			normal = paramToWorld3f * normal;
			if( normal * patchNormal < 0)
				normal *= -1;
			retVA.setPosition3f( index, retVA.getPosition3f(index) + npos * pointWeight);
			retVA.setNormal3f( index, retVA.getNormal3f(index) + normal * pointWeight);
			weight[index] += pointWeight;
		}
		
	}

	card32 invalidPoints = 0;
	for( card32 i=0;i<retVA.getNumElements();i++ )
	{
		if( weight[i] > 0)
		{
			retVA.setPosition3f(i, retVA.getPosition3f(i) / weight[i]);
			retVA.setNormal3f(i, retVA.getNormal3f(i)/weight[i]);
		}
		else
			invalidPoints++;
	}
	if( invalidPoints > 0)
		debugOutput << "Warning! " << invalidPoints << " invalid points in points cloud (MultiScaleGenerator)\n";

	return ret;
}


struct TResLevelCache
{
	vector<card32> index;
	vector<Vector3f> position;
	Vector3f summedPosition;
	vector<float> distance;
	card32 numUsed;
	card32 currentIndex;
	HierarchicalKNNIterator *hIt;
	SceneObjectIterator * iter;
	BasicPointCloudIterator *bIt;
	ModifyablePointCloudIterator *mIt;
	RandomAccessPointCloudIterator *rIt;
	AAT posAAT;

	TResLevelCache(PointCloud * pc)
	{
		numUsed=0;
		iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC |  
			SceneObjectIteratorType::CAP_MODIFYABLE_PC|
			SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
		bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
		rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);
		hIt = new HierarchicalKNNIterator(pc,32,NULL);
		currentIndex = 0x7fffffff;
		posAAT = pc->getAAT("position");
		summedPosition = makeVector3f(0,0,0);
	}
	~TResLevelCache()
	{
		delete iter;
		delete hIt;
	}

	void clear() {numUsed = 0;summedPosition = makeVector3f(0,0,0);}
	void add( const card32 ind, const Vector3f pos, const float dist ) 
	{
		if( numUsed >= index.size() )
		{
			index.resize(max(1,index.size() * 2));
			position.resize(index.size());
			distance.resize(index.size());
		}
		index[numUsed] = ind;
		position[numUsed] = pos;
		summedPosition += pos;
		distance[numUsed] = dist;
		numUsed++;
	}
	
};

void MultiScaleGenerator::projectPointOntoSurface( const Vector3f &pos, const Matrix3f &paramToWorld, const Matrix3f &worldToParam, const StaticVector<float,6> &coeffs,
								 const float scale, const Vector3f &point,
								 Vector3f &ret_pos, Vector3f &ret_normal)
{
	Vector3f tmp = worldToParam * (point - pos) / scale;
	float u = tmp[0];
	float v = tmp[1];
	tmp[2] = Monomials2ndOrder::evaluate( coeffs, makeVector2f(u,v));
	ret_pos = paramToWorld * (tmp*scale) + pos;

	float derivateU = 2.0f * coeffs[3] * u + coeffs[4] * v + coeffs[1];
	float derivateV = 2.0f * coeffs[5] * v + coeffs[4] * u + coeffs[2];
	ret_normal = paramToWorld * normalize( makeVector3f(1,0, derivateU).crossProduct( makeVector3f(0,1,derivateV) ) );
}

static StaticVector<float,6> lastCoeffs;
static Matrix3f lastMatrix;
static float lastScale;
static vector<Vector3f> lastPosBuffer;


void estimateSmoothPoint( const Vector3f &srcPoint, const Vector3f &srcNormal, Vector3f &smoothPoint,
						 Vector3f &smoothNormal, const vector<MultiScaleGenerator::TScaleInfo> &info, int scale )
{
	HeighfieldParameterization hparam;
	hparam.calculate( info[scale].sourcePC, srcPoint, info[scale].maxRadius );
	PointSet * ps = hparam.getParametricPoints();
	AAT pposAAT = ps->getAAT("param_pos");
	AAT heightAAT = ps->getAAT("height");
	vector<Vector3f> posbuffer( ps->getNumEntries() );
	float scalePos = info[scale].maxRadius;
	lastScale = scalePos;
	for( card32 i=0;i<ps->getNumEntries();i++ )
	{
		Vector2f ppos = ps->get2f(i, pposAAT );
		posbuffer[i] =  makeVector3f( ppos[0],ppos[1],ps->get1f( i, heightAAT ) ) / scalePos ;
	}
	StaticVector<float,6> coeffs;
	Monomials2ndOrder::computeCoeffs( posbuffer, coeffs, pow( info[scale].sigma / info[scale].maxRadius,2.0f ) );
	lastCoeffs = coeffs;
	lastPosBuffer = posbuffer;


	Matrix3f paramToWorld = shrink4To3(hparam.getParametricToWorld4f());
	lastMatrix = paramToWorld;
	smoothPoint = paramToWorld * (makeVector3f(0,0, Monomials2ndOrder::evaluate( coeffs, makeVector2f(0,0)))*scalePos) + srcPoint;
	smoothNormal = paramToWorld * Monomials2ndOrder::evaluateNormal( coeffs, makeVector2f(0,0));
	if( smoothNormal * srcNormal < 0)
		smoothNormal = -smoothNormal;
}

void refineFeature( FastFeature * f, const vector<MultiScaleGenerator::TScaleInfo> &info )
{
	float radius = info[ f->m_Level ].requiredSourceSamplingSpace * 1.5f;
	card32 numSamples = 60;
	card32 numIter = 1;
	vector<Vector3f> samples(numSamples);
	vector<Vector3f> smoothPos(numSamples);
	vector<Vector3f> smoothNormals(numSamples);
	//debugRenderer->clearDebugData( "featureRefine");
	debugRenderer->beginRenderJob( "featureRefine");
	Vector3f startPoint = f->m_Pos;
	Vector3f originalPos = f->m_Pos;

	for( card32 i=0;i<numIter;i++ )
	{
		// start with feature point
		Vector3f tmpPos, tmpNormal;
		estimateSmoothPoint( startPoint, f->m_Normal, tmpPos, tmpNormal, info, f->m_Level );
		estimateSmoothPoint( startPoint, f->m_Normal, smoothPos[0], smoothNormals[0], info, f->m_Level+1 );
		float difference = smoothNormals[0] * (smoothPos[0] - tmpPos);
		samples[0] = makeVector3f(0,0, difference ) / radius;
		float bestDifference = difference;
		int bestSample = 0;

		//debugRenderer->addLine( tmpPos, smoothPos[0], makeVector3f(1,0,0), makeVector3f(1,1,0),2);

		// sample some random points near the feature point
		for( card32 j=1;j<numSamples;j++ )
		{
			float angle = rnd01() * 2.0f * M_PI;
			float r = acos( rnd01() ) / M_PI * radius;
			//samples[j] = makeVector3f(sin(angle),cos(angle),0) * r;
			samples[j] = (makeVector3f(rnd01(),rnd01(),0) * 2 - makeVector3f(1,1,0));
			Vector3f wPos = f->m_TangentFrame * samples[j]*radius + startPoint;

			estimateSmoothPoint( wPos, f->m_Normal, tmpPos, tmpNormal, info, f->m_Level );
			estimateSmoothPoint( wPos, f->m_Normal, smoothPos[j], smoothNormals[j], info, f->m_Level+1 );
			float difference = smoothNormals[j] * (smoothPos[j] - tmpPos);
			samples[j][2] = difference / radius;
			Vector3f dpos = tmpPos + smoothNormals[j] * (smoothNormals[j] * (smoothPos[j] - tmpPos));
			debugRenderer->addLine( tmpPos, dpos, makeVector3f(1,0,0), makeVector3f(1,1,0),2);

/*			static int tmp = 1;
			if( fabs(difference) > 0.4f)
			{
				debugRenderer->addPatch( wPos, lastMatrix, lastCoeffs, 10, 10, makeVector3f(1,1,1), lastScale,info[f->m_Level].sigma * 0.5f );
				debugRenderer->addSphere( wPos, info[f->m_Level].sigma * 0.05f, makeVector3f(1,0,0));

				tmp++;
			}*/

			if( f->m_IsMin )
			{
				if( difference < bestDifference )
				{
					bestSample = j;
					bestDifference = difference;
				}
			}
			else
			{
				if( difference > bestDifference )
				{
					bestSample = j;
					bestDifference = difference;
				}
			}
		}

		StaticVector<float,6> diffCoeffs;
		Monomials2ndOrder::computeCoeffs(samples, diffCoeffs );

		Vector3f criticalPoint, orgPos = f->m_Pos;
		if( Monomials2ndOrder::evaluateCriticalPoint( diffCoeffs, criticalPoint ) )
		{
			criticalPoint[2] = 0;
			criticalPoint *= radius;
			Vector3f wPos = f->m_TangentFrame * criticalPoint + f->m_Pos;
			estimateSmoothPoint( wPos, f->m_Pos, f->m_Normal, f->m_Normal , info, f->m_Level+1 );
		}
		else
		{
			f->m_Pos = smoothPos[ bestSample ];
			f->m_Normal = smoothNormals[ bestSample ];
		}
		f->m_Movement = f->m_Pos - originalPos;

		radius /= 3.0f;

	}
	debugRenderer->endRenderJob();
		
}

void MultiScaleGenerator::createFast( PointCloud * pc, float medianPointDistance) 
{
	if( medianPointDistance == 0.0f )
		medianPointDistance = getMedianPointDistance( pc );


	// ----------------------------------
	// create point cloud
	VertexDescriptor * vd = new VertexDescriptor;
	vector<TAttribute>	smoothingAttributes;
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		string posName = string("position") + intToStr(i);
		string normalName = string("normal") + intToStr(i);
		vd->pushAttrib( VAD( posName, 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd->pushAttrib( VAD( normalName, 3, VAD::DATA_FORMAT_FLOAT32 ) );
	}

	PointCloud * retPC = pc->appendFields(vd);

	// ----------------------------------
	// setup multi resolution point cloud
	checkAttribute( pc, "orgPointIndex", 1, VAD::DATA_FORMAT_INT32);
	VertexArray var( pc );
	AAT indexAAT = pc->getAAT("orgPointIndex");
	StaticVector<int,1> tmpVec1i;
	for( card32 i=0;i<var.getNumElements();i++ )
	{
		tmpVec1i[0] = i;
		var.set<int,1>( i, indexAAT, tmpVec1i );
	}

	MultiResolutionPC mresPC;
	mresPC.setup( pc, 100, 1.5f );
	vector<int> pointLevel( pc->getNumPoints(),0);
	for( card32 i=0;i<mresPC.getNumberOfPointClouds();i++ )
	{
		VertexArray va_mres_i( mresPC.getPointCloud(i));
		for( card32 j=0;j<va_mres_i.getNumElements();j++ )
		{	
			int index = va_mres_i.get<int,1>(j, indexAAT)[0];
			pointLevel[index] = i+1;
		}	
	}

	// ----------------------------------
	// project points on scales
	vector< TScaleInfo > info(m_NumberOfScales);
	vector<float> H;
	vector<UnstructuredInCorePointCloud*> multiResMultiScale;
	vector<UnstructuredInCorePointCloud*> differenceClouds;
	vector<card32> multiResMultiScale_numUsedPoints( m_NumberOfScales,0 );
	for( card32 i=0;i<m_NumberOfScales;i++ )
	{
		float h = m_H0 * pow( m_F,(float)i);
		H.push_back( h );
		float kernelRadius = sqrt( -log(0.05f) * pow(h * medianPointDistance,2.0f) );
		float requiredAveragePointDistance = sqrt( -log(0.85f) * pow(h * medianPointDistance,2.0f) );
		float requiredSourceSamplingSpace = kernelRadius * 0.75f / 5.0f;
		if( requiredSourceSamplingSpace < mresPC.getPCAveragePointDistance(0))
		{
			info[i].sourcePC = pc;
			info[i].sourceLevel = 0;
		}
		else
		{
			for( card32 j=0;j<mresPC.getNumberOfPointClouds();j++ )
			{
				card32 index = mresPC.getNumberOfPointClouds() - j - 1;
				if( mresPC.getPCAveragePointDistance(index) < requiredSourceSamplingSpace )
				{
					info[i].sourcePC = mresPC.getPointCloud(index);
					info[i].sourceLevel = index+1;
					break;
				}
			}
				
		}

		if( i+1 < m_NumberOfScales )
		{
			differenceClouds.push_back( (UnstructuredInCorePointCloud*)info[i].sourcePC->copy() );
			checkAttribute( differenceClouds.back(), "difference", 1, VAD::DATA_FORMAT_FLOAT32 );
			checkAttribute( differenceClouds.back(), "projPos", 3, VAD::DATA_FORMAT_FLOAT32 );
			checkAttribute( differenceClouds.back(), "normal", 3, VAD::DATA_FORMAT_FLOAT32 );
		}
		


		info[i].minRes = 0;
		info[i].patchPC = pc;
		card32 numPoints_MR = pc->getNumPoints();
		for( card32 j=0;j<mresPC.getNumberOfPointClouds();j++ )
		{
			if( mresPC.getPCAveragePointDistance(j) > requiredAveragePointDistance )
				break;
			numPoints_MR = mresPC.getPointCloud(j)->getNumPoints();
			info[i].minRes = j+1;
			info[i].patchPC = mresPC.getPointCloud(j);
		}

		UnstructuredInCorePointCloud * scalePC = new UnstructuredInCorePointCloud;
		VertexDescriptor vd;
		vd.pushAttrib( VAD("position",3,VAD::DATA_FORMAT_FLOAT32) );
		vd.pushAttrib( VAD("localFrame",9,VAD::DATA_FORMAT_FLOAT32) );
		vd.pushAttrib( VAD("coeff",6,VAD::DATA_FORMAT_FLOAT32) );
		vd.pushAttrib(  VAD("scalePatch",1,VAD::DATA_FORMAT_FLOAT32) );
		
		scalePC->clearAndSetup( &vd, numPoints_MR);
		multiResMultiScale.push_back( scalePC );

		info[i].sigma = h * medianPointDistance;
		info[i].inv_sigma_2 = 1.0f / pow( info[i].sigma, 2.0f );
		info[i].posAAT = retPC->getAAT(string("position") + intToStr(i));
		info[i].normalAAT = retPC->getAAT(string("normal") + intToStr(i));
		info[i].maxRadius = kernelRadius;
		info[i].requiredSourceSamplingSpace = requiredSourceSamplingSpace;
	}

	/*vector<HierarchicalKNNIterator*> retpc_hIt;
	for( card32 i=0;i<multiResMultiScale.size();i++ )
		retpc_hIt.push_back( new HierarchicalKNNIterator(multiResMultiScale[i],32,NULL));*/

	VertexArray va_ret( retPC );
	for( card32 i=0;i<va_ret.getNumElements();i++ )
	{
		for( card32 j=0;j<info.size();j++ )
		{
			va_ret.set<float,3>(i, info[j].posAAT, makeVector3f(0,0,0));
			va_ret.set<float,3>(i, info[j].normalAAT, makeVector3f(0,0,0));
		}
	}
	
	//HeighfieldParameterization hparam;
	vector<TResLevelCache*> cache(mresPC.getNumberOfPointClouds()+1,NULL);
	Timer timer;
/*	for( int i=0;i<va_ret.getNumElements();i++ )
	{
		
		Vector3f pos = va_ret.getPosition3f(i);
		Vector3f pointNormal = va_ret.getNormal3f(i);
		int plevel = pointLevel[i];

		for( card32 j=0;j<info.size();j++ )
		{
			if(	plevel < info[j].minRes )
				break;*/

	for( card32 j=0;j<m_NumberOfScales;j++ )
	{
		VertexArray currentVA(info[j].patchPC);
		
#ifdef _OPENMP
//#pragma omp parallel for firstprivate (cache) shared(weight,va_ret)
#endif
		for( card32 i=0;i<currentVA.getNumElements();i++ )
		{
			Vector3f pos = currentVA.getPosition3f(i);
			Vector3f pointNormal = currentVA.getNormal3f(i);

			if( cache[info[j].sourceLevel] == NULL )
				cache[info[j].sourceLevel] = new TResLevelCache(info[j].sourcePC);
			
			TResLevelCache * rcache = cache[info[j].sourceLevel];

			// collect neighbours
			//if( rcache->currentIndex != i )
			//{
				rcache->clear();
				rcache->currentIndex = i;
				rcache->hIt->setSeekPointAndReset( pos );
			//}

			while( !rcache->hIt->atEnd())
			{
				Vector3f spos = rcache->hIt->get3f( rcache->posAAT );
				float dist = norm( pos - spos);
				if( dist > info[j].maxRadius && rcache->numUsed > 20)
					break;

				rcache->add( rcache->hIt->getPointSetPointNumber(), spos, dist );

				rcache->hIt->next();
			}

			

			Vector3f centroid = rcache->summedPosition / rcache->numUsed;
			Matrix3f covariance = IDENTITY3F;
			covariance *= 0;
			for( card32 k=0;k<rcache->numUsed;k++ )
			{
				Vector3f diff = rcache->position[k] - centroid;
				for( card32 d1=0;d1<3;d1++)
				{
					for( card32 d2=0;d2<=d1;d2++)
					{
						covariance[d1][d2] += diff[d1] * diff[d2];
						covariance[d2][d1] = covariance[d1][d2];
					}
				}
			}
			Matrix3f eigenVectors;
			Vector3f eigenValues;
			covariance.computeEigenStructure(eigenValues, eigenVectors);
			
			Matrix3f paramToWorld3f = eigenVectors;
			paramToWorld3f[2] = normalize( paramToWorld3f[0].crossProduct( paramToWorld3f[1]) );
			if( paramToWorld3f[2] * pointNormal < 0)
				paramToWorld3f[2] = -paramToWorld3f[2];
			Matrix3f worldToParam3f = paramToWorld3f.transpose();
			vector<Vector3f> points;
			for( card32 k=0;k<rcache->numUsed;k++ )
			{
				points.push_back( worldToParam3f * (rcache->position[k] - pos)/info[j].maxRadius );
			}


			// evaluate patch coefficients
			StaticVector<float,6> coeffs;
			Monomials2ndOrder::computeCoeffs( points, coeffs, pow(info[j].sigma/info[j].maxRadius,2.0f));

			// store patch
			UnstructuredInCorePointCloud * rpc = multiResMultiScale[j];
			card32 r_posOffset = rpc->getDescr()->getAttributeOffset(rpc->getDescr()->getIndex( "position" ));
			card32 r_localFrameOffset = rpc->getDescr()->getAttributeOffset(rpc->getDescr()->getIndex( "localFrame"));
			card32 r_coeffOffset = rpc->getDescr()->getAttributeOffset(rpc->getDescr()->getIndex(  "coeff" ));
			card32 r_scalePatchOffset = rpc->getDescr()->getAttributeOffset(rpc->getDescr()->getIndex(  "scalePatch" ));
							

			card32 store_index = i;//multiResMultiScale_numUsedPoints[j];
			//multiResMultiScale_numUsedPoints[j]++;
			char * dataPtr = rpc->getPointSet()->getDataPointer(store_index);

			*((Vector3f*)(dataPtr + r_posOffset)) = pos;
			*((Matrix3f*)(dataPtr + r_localFrameOffset)) = paramToWorld3f;
			*((StaticVector<float,6>*)(dataPtr + r_coeffOffset)) = coeffs;
			*((float32*)(dataPtr + r_scalePatchOffset)) = info[j].maxRadius;

//			rpc->getPointSet()->set3f(store_index, r_posAAT, pos);
//			rpc->getPointSet()->set<float,9>(store_index, r_localFrameAAT, *((StaticVector<float,9>*)paramToWorld3f.data()) );
//			rpc->getPointSet()->set<float,6>(store_index, r_coeffAAT, coeffs );


			// project points onto patch
			/*AAT posAAT = info[j].sourcePC->getAAT( "position" );
			Vector3f patchNormal = paramToWorld3f[2];
			if( var.getNormal3f(i) * patchNormal < 0 )
				patchNormal *= -1;
			float scaleGaussian = sqr(info[j].sigma);
//			Matrix3f paramToWorld3f = shrink4To3( hparam.getParametricToWorld4f());
			for( card32 k=0;k<ps->getNumEntries();k++ )
			{
				Vector3f npos = ps->get3f(k, posAAT);
				int index = ps->get1i(k,pointIndexAAT);

				float pointWeight = exp( -normQuad(npos -pos) / scaleGaussian);
				npos = npos + patchNormal * (patchNormal * (pos - npos));
				float u = points[k][0];
				float v = points[k][1];
				npos += patchNormal * Monomials2ndOrder::evaluate(coeffs, makeVector2f(u,v)) * info[j].maxRadius;
				float derivateU = 2.0f * coeffs[3] * u + coeffs[4] * v + coeffs[1];
				float derivateV = 2.0f * coeffs[5] * v + coeffs[4] * u + coeffs[2];
				Vector3f normal = -normalize( makeVector3f(1,0, derivateU).crossProduct( makeVector3f(0,1,derivateV) ) );
				normal = paramToWorld3f * normal;
				if( normal * patchNormal < 0)
					normal *= -1;
				va_ret.set( index, info[j].posAAT, va_ret.get<float,3>(index,info[j].posAAT) + npos * pointWeight);
				va_ret.set( index, info[j].normalAAT, va_ret.get<float,3>(index,info[j].normalAAT) + normal * pointWeight);
				weight[j][index] += pointWeight;
			}*/
		}

	}

	/*for( card32 i=0;i<va_ret.getNumElements();i++ )
	{
		for( card32 j=0;j<m_NumberOfScales;j++ )
		{
			va_ret.set( i, info[j].posAAT, va_ret.get<float,3>(i,info[j].posAAT)/weight[j][i]);
			va_ret.set( i, info[j].normalAAT, va_ret.get<float,3>(i,info[j].normalAAT)/weight[j][i]);
		}
	}*/

	for( card32 i=0;i<m_NumberOfScales-1;i++ )
	{
		UnstructuredInCorePointCloud * diffCloud = differenceClouds[i];
		AAT diffAAT = diffCloud->getAAT( "difference");
		AAT projPosAAT = diffCloud->getAAT("projPos");
		AAT posAAT = diffCloud->getAAT( "position");
		PointSet * ps = diffCloud->getPointSet();
	
		PointCloud * level1Cloud = multiResMultiScale[i];
		PointCloud * level2Cloud = multiResMultiScale[i+1];

		HierarchicalKNNIterator hIt1(level1Cloud,32,NULL);
		HierarchicalKNNIterator hIt2(level2Cloud,32,NULL);

		card32 r_posOffset = level1Cloud->getDescr()->getAttributeOffset(level1Cloud->getDescr()->getIndex( "position" ));
		card32 r_localFrameOffset = level1Cloud->getDescr()->getAttributeOffset(level1Cloud->getDescr()->getIndex( "localFrame"));
		card32 r_coeffOffset = level1Cloud->getDescr()->getAttributeOffset(level1Cloud->getDescr()->getIndex(  "coeff" ));
		card32 r_scalePatchOffset = level1Cloud->getDescr()->getAttributeOffset(level1Cloud->getDescr()->getIndex(  "scalePatch" ));

		vector<char> vbuffer( level1Cloud->getDescr()->getSize() );
		Vector3f *posPtr = (Vector3f*)(&vbuffer[0] + r_posOffset);
		Matrix3f *framePtr = (Matrix3f*)(&vbuffer[0] + r_localFrameOffset);
		StaticVector<float,6> *coeffPtr = (StaticVector<float,6>*)(&vbuffer[0] + r_coeffOffset);
		float *scalePtr = (float*)(&vbuffer[0] + r_scalePatchOffset);

		float inv_sigma_sqr_level1 = info[i].inv_sigma_2;
		float inv_sigma_sqr_level2 = info[i+1].inv_sigma_2;

		for( card32 j=0;j<diffCloud->getNumPoints();j++)
		{
			Vector3f currentPos = ps->get3f(j,posAAT);
			
			// project level i
			Vector3f pos0, normal0;
			pos0 = normal0 = makeVector3f(0,0,0);
			hIt1.setSeekPointAndReset( currentPos );
			float summed_weight = 0;
			card32 numSamples = 0;
			while( !hIt1.atEnd() )
			{
				hIt1.getVertex(&vbuffer[0]);

				float dist_2 = normQuad(*posPtr - currentPos);
				float weight = exp( -dist_2 * inv_sigma_sqr_level1 );

				Vector3f tmpPos, tmpNormal;
				projectPointOntoSurface( *posPtr, *framePtr, framePtr->transpose(), *coeffPtr, *scalePtr, currentPos, tmpPos, tmpNormal);

				pos0 += tmpPos * weight;
				normal0 += tmpNormal * weight;
				summed_weight += weight;


				if( weight < 0.01f && numSamples++ < 3)
					break;
				hIt1.next();
			}

			pos0 /= summed_weight;
			normal0 /= summed_weight;
			ps->set3f( j, projPosAAT, pos0);
			ps->set3f( j, posAAT, pos0 );

			// project level i+1
			Vector3f pos1, normal1;
			pos1 = normal1 = makeVector3f(0,0,0);
			hIt2.setSeekPointAndReset( currentPos );
			summed_weight = 0;
			numSamples = 0;
			while( !hIt2.atEnd() )
			{
				hIt2.getVertex(&vbuffer[0]);

				float dist_2 = normQuad(*posPtr - currentPos);
				float weight = exp( -dist_2 * inv_sigma_sqr_level2 );

				Vector3f tmpPos, tmpNormal;
				projectPointOntoSurface( *posPtr, *framePtr, framePtr->transpose(), *coeffPtr, *scalePtr, currentPos, tmpPos, tmpNormal);

				pos1 += tmpPos * weight;
				normal1 += tmpNormal * weight;
				summed_weight += weight;

				if( weight < 0.01f && numSamples++ < 3)
					break;
				hIt2.next();
			}
			pos1 /= summed_weight;
			normal1 /= summed_weight;

			// calculate and store difference
			float difference = normal1 * (pos1 - pos0);
			ps->set1f(j, diffAAT, difference);
		}

	}

	// search local minima/maxima
	vector<FastFeature*> features;
	vector<HierarchicalKNNIterator*> diffhIt;
	diffhIt.push_back(new HierarchicalKNNIterator( differenceClouds[0],32,NULL));
	AAT posAAT = differenceClouds[0]->getAAT("position");
	AAT diffAAT = differenceClouds[0]->getAAT("difference");
	for( card32 i=1;i<differenceClouds.size();i++ )
	{
		diffhIt.push_back(new HierarchicalKNNIterator( differenceClouds[i],32,NULL));
		HierarchicalKNNIterator * hIt = diffhIt.back();
		AAT projposAAT = differenceClouds[i]->getAAT("projPos");
		
		float C = 0.5f;
		float maxRad = info[i].sigma * C;
		VertexArray va_diff( differenceClouds[i]);
		for( card32 j=0;j<va_diff.getNumElements();j++ )
		{
			Vector3f pos = va_diff.getPosition3f(j);
			hIt->setSeekPointAndReset(pos);
			float currentDifference = hIt->get1f( diffAAT );
			hIt->next();
			float nextDifference = hIt->get1f( diffAAT );
			bool isMin = nextDifference > currentDifference;
			bool isMax = nextDifference < currentDifference;
			while( !hIt->atEnd() && (isMin ^ isMax))
			{
				Vector3f tmpPos = hIt->get3f(posAAT);
				float dist = norm( tmpPos - pos);
				if( dist > maxRad )
					break;

				float tmpDifference = hIt->get1f( diffAAT );
				if( isMin )
					isMin = tmpDifference > currentDifference;
				else
					isMax = tmpDifference < currentDifference;


				hIt->next();
			}

			if( isMin ^ isMax )
			{
				FastFeature *f = new FastFeature;
				f->m_Pos = va_diff.get<float,3>(j, projposAAT);
				f->m_SourcePoint = va_diff.getPosition3f(j);
				f->m_IsMin = isMin;
				f->m_Sigma = info[i].sigma;
				f->m_Level = i;
				f->m_Radius = maxRad;
				f->m_Difference = currentDifference;
				f->m_Normal = va_diff.getNormal3f(j);
				features.push_back(f);
			}
		}
	}

	vector<FastFeature*> featureCandidates;
	for( card32 i=0;i<features.size();i++ )
	{
		card32 level = features[i]->m_Level;
		HierarchicalKNNIterator * hIt = diffhIt[ level -1 ];

		float C = 0.5f;
		float maxRad = info[level-1].sigma * C;

		Vector3f fpos = features[i]->m_Pos;
		hIt->setSeekPointAndReset( fpos );
		float currentDifference = features[i]->m_Difference;
		bool isFeature = true;
		bool isMin = features[i]->m_IsMin;
		while( true )
		{
			Vector3f tmpPos = hIt->get3f( posAAT );
			if( norm(tmpPos - fpos) > maxRad)
				break;

			float tmpDifference = hIt->get1f( diffAAT );
			if( isMin )
			{
				if( tmpDifference < currentDifference )
				{
					isFeature = false;
					break;
				}
			}
			else // isMax = true
			{	
				if( tmpDifference > currentDifference )
				{
					isFeature = false;
					break;
				}

			}

			hIt->next();
		}

		maxRad = info[level+1].sigma * C;
		if( level < differenceClouds.size() - 1 && isFeature )
		{
			HierarchicalKNNIterator * hIt = diffhIt[ level +1 ];
			hIt->setSeekPointAndReset( fpos );
			float currentDifference = features[i]->m_Difference;
			bool isMin = features[i]->m_IsMin;
			while( true )
			{
				Vector3f tmpPos = hIt->get3f( posAAT );
				if( norm(tmpPos - fpos) > maxRad)
					break;

				float tmpDifference = hIt->get1f( diffAAT );
				if( isMin )
				{
					if( tmpDifference < currentDifference )
					{
						isFeature = false;
						break;
					}
				}
				else // isMax = true
				{	
					if( tmpDifference > currentDifference )
					{
						isFeature = false;
						break;
					}

				}

				hIt->next();
			}
		}

		if( isFeature )
			featureCandidates.push_back( features[i] );
		else
			delete features[i];


	}

	FeatureAttachment * featureAttachment = (FeatureAttachment*)pc->getAttachments()->getData( FeatureAttachment::attachmentName() );
	if( featureAttachment )
	{
		featureAttachment->clearFeatures();
	}
	else
	{
		featureAttachment = new FeatureAttachment;
		featureAttachment->setup( FeatureAttachment::attachmentName(), AttachedData::ADF_PERSISTENT );
		pc->getAttachments()->attachData( featureAttachment );
	}

	// drop border features
	HierarchicalKNNIterator sourcehIt = HierarchicalKNNIterator( pc, 32, NULL);
	AAT sourcePosAAt = pc->getAAT( "position" );
	card32 r_localFrameOffset = multiResMultiScale[0]->getDescr()->getAttributeOffset(multiResMultiScale[0]->getDescr()->getIndex( "localFrame"));
	vector<char> vbuffer( multiResMultiScale[0]->getDescr()->getSize() );
	Matrix3f * matrixPtr = (Matrix3f*)(&vbuffer[0] + r_localFrameOffset);
	float featureRadius = 5.0f * medianPointDistance;
	for( card32 i=0;i<featureCandidates.size();i++ )
	{
		FastFeature * f = featureCandidates[i];
		HierarchicalKNNIterator hIt( multiResMultiScale[f->m_Level],32,NULL );
		hIt.setSeekPointAndReset( f->m_Pos );
		hIt.getVertex( &vbuffer[0]);
		f->m_TangentFrame = *matrixPtr;
		Matrix3f worldToParam = matrixPtr->transpose();

		sourcehIt.setSeekPointAndReset(f->m_Pos);
		card32 numSlices = 6;
		vector<float> histogram( numSlices,0);
		while( true )
		{
			Vector3f pos = sourcehIt.get3f(sourcePosAAt);

			if( norm( pos - f->m_Pos) > featureRadius )
				break;

			// transform into local frame and compute angle
			pos = worldToParam * (pos - f->m_Pos);
			float angle = (atan2(pos[0],pos[1]) + M_PI) / 2.0f / M_PI * (float)numSlices;
			int slice = (angle + numSlices);
			slice = slice % numSlices;
			histogram[slice]++;

			sourcehIt.next();
		}
		sort( histogram.begin(), histogram.end() );
		f->m_Valid = histogram[0] > histogram.back() * 0.25f;
		if( f->m_Valid )
		{
			refineFeature( f, info );
 			featureAttachment->addFeature( f );
		}
		else
			delete f;

	}

	ffInfo.diffClouds = differenceClouds;
	ffInfo.info = info;
	ffInfo.multiScaleClouds = multiResMultiScale;

	card32 dtime = timer.getDeltaValue();
	debugOutput << "Time: " << dtime / 1000 << "." << dtime % 1000 << "\n";

	//return retPC;
//	return multiResMultiScale;
}
