#ifndef MultiScaleGeneratorH
#define MultiScaleGeneratorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class MultiResolutionPC;

/**
	creates a multi scale point cloud (mls)
	ScaleVD describes a scale-vertex (every attribute will be n-times in the
	resulting point cloud). Every attribute(float) in ScaleVD is smoothed (except for "pos")
	AdditionalScaleVD adds attributes to every scale vertex, but without any initialization
	(no smoothing, e.g. gradients or pos for mls).
	Default for ScaleVD: "pos", "normal", AdditionalScaleVD is empty by default.
*/
class ANIMATIONRECONSTRUCTION_API MultiScaleGenerator : public Persistent
{
	GEOXL_CLASS( MultiScaleGenerator )

private:


	card32					m_NumberOfScales;
	VertexDescriptor	*	m_ScaleVD;
	float32					m_H0;		// sigma of first scale
	float32					m_F;		// sigma[n] = H0 * F^n
	VertexDescriptor	*	m_AdditionalScaleVD;
	float32					m_MinNumberOfSamples;



public:
	MultiScaleGenerator(bool mlsSurface=true);
	~MultiScaleGenerator(void);

   virtual void assign(const Object* obj, CopyContext *context = NULL);

    /// setup for position multi scale pointcloud
    void setupPosition();

	/// setup for pure color multi scale pointcloud
	void setupColor();

    /// setup for position + color multi scale pointcloud
	void setupPositionAndColor();

   
    /// builds multi scale point cloud from point cloud pc
	/// mresPC is a multi resolution point cloud from pc
	/// pcAveragePointDistance will be calculated if not specified
	PointCloud * create( PointCloud * pc, MultiResolutionPC * mresPC, float pcAveragePointDistance = 0.0f );

	PointCloud * smoothPC( PointCloud * pc, MultiResolutionPC * mresPC, float sigma, float patchSamplingSpace, float sourceSamplingSpace, float pcAveragePointDistance = 0.0f );

	void createFast( PointCloud * pc, float medianPointDistance = 0.0f);

	card32 getNumberOfScales() const {return m_NumberOfScales;}
	void setNumberOfScales(card32 v) {m_NumberOfScales=v;}

	/// adds a n-f attribute
	void addVertexAttributef( string name, int numComponents );

	/// adds a n-f attrbiute to additional attributes
	void addAdditionalVertexAttributef( string name, int numComponents );

	float32 getH0() const {return m_H0;}
	void setH0(float32 v) {m_H0=v;}

	float32 getF() const {return m_F;}
	void setF(float32 v) {m_F=v;}

	float32 getMinNumberOfSamples() const {return m_MinNumberOfSamples;}
	void setMinNumberOfSamples(float32 v) {m_MinNumberOfSamples=v;}

	struct TScaleInfo
	{
		float maxRadius;
		int minRes;
		int sourceLevel;
		PointCloud * sourcePC;
		PointCloud * patchPC;
		float sigma;
		float inv_sigma_2;
		AAT normalAAT;
		AAT posAAT;
		float requiredSourceSamplingSpace;
	};

	struct TFastFeatureInfo
	{
		vector<UnstructuredInCorePointCloud*> diffClouds;
		vector<UnstructuredInCorePointCloud*> multiScaleClouds;
		vector<TScaleInfo> info;
	};
	
	static TFastFeatureInfo ffInfo;

static void projectPointOntoSurface( const Vector3f &pos, const Matrix3f &paramToWorld, const Matrix3f &worldToParam, const StaticVector<float,6> &coeffs,
		const float scale, const Vector3f &point,
		Vector3f &ret_pos, Vector3f &ret_normal);

};

}

#endif