#ifndef PCIFeatureTestH
#define PCIFeatureTestH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PCInteractionTool.h"
//---------------------------------------------------------------------------
#include "Feature.h"
#include "FeatureRegistration.h"

namespace geoxl
{

class FeatureDetector;
class FeatureDescriptor;
class PCIFeatureTestEditorWidget;
class FeatureRegistration;
class RenderingTool;

/**
	test tool to benchmark different feature detection algorithms and feature desciptors
*/
class ANIMATIONRECONSTRUCTION_API PCIFeatureTest : public PCInteractionTool
{
	GEOXL_CLASS( PCIFeatureTest )

private:

	void preparePointClouds();
	bool getPointFromCursor( int x, int y, Vector3f &retPoint, string &retPC );
	FeatureRegistration * setupRegistration(float averagePointDistance);

	float32	m_DistortionStrength;
	card32	m_DistortionMode;
	card32	m_DistortionRandomSeed;
	card32	m_DistortionNumGravityBalls;
	card32	m_ScreenWidth, m_ScreenHeight;

	FeatureDetector *		m_FeatureDetector;
	FeatureDescriptor *	m_FeatureDescriptor;

	friend PCIFeatureTestEditorWidget;

	FeatureList*	m_Features;
	FeatureList*	m_FeaturesA, *m_FeaturesB;
	string			m_PointCloudA_Name,m_PointCloudB_Name;
	PointCloud*		m_PointCloudA, *m_PointCloudB;

	BoundingBox3f		m_SceneBB;
	vector<FeaturePair>	m_Pairs, m_InitialPairs;
	bool						m_HideInitialPairs;
	float						m_ShowPairsPercentage;

	bool						m_DrawGradients;
	float						m_GradientLength;
	bool						m_DrawFeatureScale;

	float32					m_SpectralSigma;
	float32					m_SpectralAddDiff;
	float32					m_SpectralScaleDiff;



	vector<FeatureDescriptor*>	m_Descriptors;
	float						m_RelativeFeatureSize;

	// debuginfo for geodetic distances
	vector< vector<Vector3f> >	m_Paths;

	float						m_DescClusterRadius;
	struct TGlobalRegistrationCloud
	{
		string name;
		PointCloud * pc;
		FeatureList* features;
	};
	vector<TGlobalRegistrationCloud> m_Clouds;
	RenderingTool * m_RenderTool;


public:
	PCIFeatureTest(void);
	~PCIFeatureTest(void);

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   virtual void mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void mouseMoved(int32 x, int32 y);
   virtual void mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void areaResize(card32 width, card32 height);

   virtual void keyDown(GeneralKey key);
   virtual void keyUp(GeneralKey key);

   virtual void connectToSceneImpl(Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget, QWidget* toolWidget);
   virtual void glDrawTool(GLContext *glContext);

	void distortPointClouds();
	
	// test detector by applying increasing distortions (from 0 to _DistortionStrength_ in _numberOfTests_ steps)
	void testDetector( int numberOfTests );

	/// test feature descriptor similar to testDetector
	void testDescriptor( int numberOfTests, int numberOfSamples, float radius );

	/// test registration
	void testRegistration( int maxNumberOfPairs, float featureRadius, float pruningThreshold);

	/// calculate features for two point clouds
	void calculateFeatures( float featureRadius );

	/// calculate descriptors for existing features
	void recalculateFeatureDescriptors( float featureRadius );

	/// build features pairs from featuresA and featuresB
	void buildFeaturePairs( int maxNumberOfPairs, float pruningThreshold );

	/// validate feature pairs from initialPairs
	void validateFeaturePairs();

	/// test for geodetic distances
	void testGeodeticDistances();

	/// creates a multi scale point cloud and adds every scale as a point cloud to the scene
	void createMultiScalePointClouds();

	float32 getDistortionStrength() const {return m_DistortionStrength;}
	void setDistortionStrength(float32 v) {m_DistortionStrength=v;}

	card32 getDistortionMode() const {return m_DistortionMode;}
	void setDistortionMode(card32 v) {m_DistortionMode=v;}

	card32 getDistortionRandomSeed() const {return m_DistortionRandomSeed;}
	void setDistortionRandomSeed(card32 v) {m_DistortionRandomSeed=v;}

	card32 getDistortionNumGravityBalls() const {return m_DistortionNumGravityBalls;}
	void setDistortionNumGravityBalls(card32 v) {m_DistortionNumGravityBalls=v;}

	FeatureDetector * getFeatureDetector() const {return m_FeatureDetector;}
	void setFeatureDetector(FeatureDetector * v) {m_FeatureDetector=v;}

	FeatureDescriptor * getFeatureDescriptor() const {return m_FeatureDescriptor;}
	void setFeatureDescriptor(FeatureDescriptor * v) {m_FeatureDescriptor=v;}

	bool getHideInitialPairs() const {return m_HideInitialPairs;}
	void setHideInitialPairs(bool v) {m_HideInitialPairs=v;}

	float getShowPairsPercentage() const {return m_ShowPairsPercentage;}
	void setShowPairsPercentage(float v) {m_ShowPairsPercentage=v;}

	bool getDrawGradients() const {return m_DrawGradients;}
	void setDrawGradients(bool v) {m_DrawGradients=v;}

	float getGradientLength() const {return m_GradientLength;}
	void setGradientLength(float v) {m_GradientLength=v;}

	float getRelativeFeatureSize() const {return m_RelativeFeatureSize;}
	void setRelativeFeatureSize(float v) {m_RelativeFeatureSize=v;}

	bool getDrawFeatureScale() const {return m_DrawFeatureScale;}
	void setDrawFeatureScale(bool v) {m_DrawFeatureScale=v;}

	float32 getSpectralSigma() const {return m_SpectralSigma;}
	void setSpectralSigma(float32 v) {m_SpectralSigma=v;}

	float32 getSpectralAddDiff() const {return m_SpectralAddDiff;}
	void setSpectralAddDiff(float32 v) {m_SpectralAddDiff=v;}

	float32 getSpectralScaleDiff() const {return m_SpectralScaleDiff;}
	void setSpectralScaleDiff(float32 v) {m_SpectralScaleDiff=v;}

	float getDescClusterRadius() const {return m_DescClusterRadius;}
	void setDescClusterRadius(float v) {m_DescClusterRadius=v;}

	void calculateCurvatureForAlleClouds( float sigma, float normalSigma, bool median);

	void calculateCurvaturePC( vector<PointCloud *> clouds, float sigma, float normalSigma, bool median);

	void testDetectorStability_Rigid();

	void testDetectorStability_Rigid( PointCloud * pc, int noiseLevel, float noise, int numTests );

	struct TDescTest
	{
		TDescTest( FeatureDescriptor * e_fd, string e_rowName, float size ) 
			{fd = e_fd; rowName = e_rowName;descSize=size; }
		FeatureDescriptor * fd;
		string rowName;
		float descSize;
		vector< DVectorF > pcatest_descriptors;
	};
	
	void testDescriptorStability();
	
	void testDescriptorStability(FeatureList * features, PointCloud * pc, vector<TDescTest> &test, int noiseLevel, float noise, int numTests);

	PointCloud * getNoisyPointCloud( PointCloud * pc, float noise, float bb_diag, Matrix3f &randRot, Vector3f &randTransl );

	void calculateAllFeatures(float featureRadius);

	void registerAllClouds(float featureRadius, int maxNumberOfPairs, float pruningThreshold);

	float testOverlap( PointCloud * pcA, PointCloud *pcB, Matrix4f transformBtoA );

	FeatureList * removeBorderFeatures( FeatureList * fl, float radius, PointCloud * pc );

	void testValidation();

	void copyTransformationsToClipboard();
	void pasteTransformationsFromClipboard();

};

}

#endif