#ifndef PCIGlobalRegistrationH
#define PCIGlobalRegistrationH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PCInteractionTool.h"
#include "SpectralCorrespondenceValidation.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class UnstructuredInCoreTriangleMesh;
class FastFeature;
class FeatureDescriptor;
class FeatureAttachment;
class CentroidSignature;

/**
	tool for global registration
*/
class ANIMATIONRECONSTRUCTION_API PCIGlobalRegistration : public PCInteractionTool
{
	GEOXL_CLASS( PCIGlobalRegistration )

private:

	UnstructuredInCoreTriangleMesh * createPatchMesh( PointCloud * pc, float patchSize );
	void drawFeature( FastFeature * f, const Vector3f &color );

	void benchStability( PointCloud * pc, float attribute_noise, float feature_pos_noise, int cd_numScales, 
		int cd_numSamples, float startSigma, float baseSigma, float cd_radius, int row, int column );

	float32		m_Sigma;
	float32		m_PatchSamplingSpace;
	float32		m_SourceSamplingSpace;
	card32		m_CurrentScale;

	card32		m_MaxNumInitialCorrespondences;

	bool			m_ShowDroppedFeatures;
	int			m_ShowFeatureMode;

	card32		m_NumberOfScales;
	float32		m_StartScale;
	float32		m_Scale_F;

	card32		m_FS_NumTests;
	float32		m_FS_NoiseAmplitude;
	float32		m_FS_ResamplingPercentage;

	float32		m_FeatureRadius;
	float32		m_CurvatureSigma;
	FeatureDescriptor * m_Descriptor;

	float32		m_TestMinDifferenceStrength;
	float32		m_CD_WindowSize;
	float32		m_CD_StartSigma;

	card32		m_InteractionMode;
	bool			m_ShowFeatureRadius;

	float32		m_FDistance;
	float32		m_FDistanceDeviation;
	float32		m_FMaxDistance;

	// tensor voting parameters
	float32		m_TVSigma;
	float32		m_TVCurvSigma;
	float32		m_TVMaxAngle;
	float32		m_TVLP_MinLineLength;
	float32		m_TVLP_SamplesPercent;
	float32		m_TVLP_MaxInitialDistance;
	card32		m_TVLP_MinNumLineSamples;
	float32		m_TVLP_LineSmoothingSigma;
	float32		m_TVLP_MaxLineJoinDistance;
	float32		m_TVLP_MaxLineJoinAngle;
	float32		m_TVLP_MaxLineDistance;
	float32		m_TVLP_MaxLineToLineDistance;
	float32		m_TVLP_MinLineLengthForIntersection;
	float32		m_TVCP_MinEigenProduct;

	enum ETVDefaultSettings
	{
		TVD_Dragon = 0,
		TVD_Armadillo = 1
	};
	card32		m_TVDefaultSettings;

	void tvSettingsChanged();
	
	enum EInteractionMode
	{
		IM_DrawSignature = 0,
		IM_Registration = 1,
		IM_AnalyseLocalNeighbourhood = 2,
		IM_DrawSimilarity = 3,
		IM_SelectFeaturePoints = 4,
		IM_SelectManualCorrespondence = 5
	};

	// members
	card32		m_ScreenWidth;
	card32		m_ScreenHeight;
	vector<CentroidSignature*> debugSignatures;
	vector<Vector2f> m_FeatureCurvPoints;
	vector<float> m_FeatureCurvRadius;
	
	vector<Vector3f> m_ManualCorrespondence;


	struct TCorrespondence
	{
		bool operator < ( const TCorrespondence &o) const {return distance < o.distance;}
		FastFeature * f1, *f2;
		float		distance;
		card32		i1,i2;
	};

	struct TClickRegisterTest
	{
		CentroidSignature* sig;
		PointCloud * pc;
		string pcName;
	};
	vector<TClickRegisterTest>	m_ClickSignatures;

	struct TLinePoint
	{
		Vector3f pos;
		Vector3f lineDir;
		float lineLength;
		float movement;
	};

	vector<TLinePoint> m_LinePoints;
	struct TCornerPoint
	{
		Vector3f pos;
		float movement;
	};
	vector<TCornerPoint> m_CornerPoints;

	Vector3f computePCANormal( PointCloud * pc, const Vector3f &pos, float sigma);


public:
	PCIGlobalRegistration(void);
	~PCIGlobalRegistration(void);

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   virtual void mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void mouseMoved(int32 x, int32 y);
   virtual void mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void areaResize(card32 width, card32 height);

   virtual void keyDown(GeneralKey key);
   virtual void keyUp(GeneralKey key);

   virtual void glDrawTool(GLContext *glContext);	


   void testScale();
	
   bool getPointFromCursor( int x, int y, Vector3f &retPoint, string &retPC );

	void testFeatureStability();
	void buildScalePointClouds();
	void calculateFeatureSignatures();
	void removeBoringFeatures();

	vector<TCorrespondence> buildInitialSet( PointCloud * pc1, PointCloud * pc2);
	vector<TCorrespondence> buildInitialSet( FeatureAttachment * fa1, FeatureAttachment * fa2 );	
	vector<TCorrespondence> validateInitialSet( const vector<TCorrespondence>& initialset, float medianPointDistance, Matrix4f &transf );

	void calculateFeatures( PointCloud * pc, float medianDistance);
	void registerTwoClouds();
	void registerAllClouds();
	void calculateFeatures();
	Matrix4f calculateTransformation( const vector<SpectralCorrespondenceValidation::TFeaturePair> &pairs );

	void calculateCentroidDescriptors();
	void estimateMaxima();
	void resetTransformations();

	void bench();
	void descriptorClustering();
	void showFeatureLines();
	void ransacSymmetryPlaneEstimation();
	void distanceTransform();
	void tensorVoting();
	void initLinePoints();
	void stepLinePoints();
	void estimateLines();
	void initCornerPoints();
	void stepCornerPoints();
	void multiStepLinePoints();
	void drawDistributions();
	void ransacPlaneExtraction();
	void goodOldMSFeatures();

};

}

#endif