//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PGDescriptorChannels.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PatchGraph.h"
#include "TrajectorySet.h"
#include "PointCloudAnimationSequence.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools_templCode.h"
#include "MathTools.h"
#include "MathTools_templCode.h"
#include "PatchTypes.h"
#include "PGDataPointIndices.h"
#include "ProgressWindow.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_ABSTRACT_CLASS( PGDescriptorChannels ,0)
{
	BEGIN_CLASS_INIT( PGDescriptorChannels );
}

IMPLEMENT_GEOXL_CLASS( PGDescriptorChannelsTest ,0)
{
	BEGIN_CLASS_INIT( PGDescriptorChannelsTest );
}


// ----


PGDescriptorChannels::PGDescriptorChannels()
{
}

void PGDescriptorChannels::assign(const Object* obj, CopyContext *context)
{
   copyObjectProperties(obj, this);   
}

PGDescriptorChannels::~PGDescriptorChannels()
{
}


void PGDescriptorChannels::computeDescriptors
  (vector<PatchGraph*> pgs,
   const PointCloudAnimationSequence *data,
   const string channelName,
   const card32 channelPCADim)
{
   vector<Component> components;
   startComputeDescriptors(pgs, data, components);
   const size_t numComponents = components.size();
   if (numComponents == 0) {
      throw PException("startComputeDescriptors failed.");
   }

   /// compute descriptors
   vector<DVectorF> descriptors;
   DVectorF result;
	progressWindow->pushStep( true, "desc" );
   for (card32 pgIndex=0; pgIndex<pgs.size(); pgIndex++) {
      PatchGraph *pg = pgs[pgIndex];
      PointSet *ps = pg->getTrajectorySet()->getTrajectories();
      const card32 numFrames = ps->getWidth();
      const card32 numPatches = ps->getHeight();
		for (card32 f=0; f<numFrames; f++) {
	      for (card32 p=0; p<numPatches; p++) {
				computeDescriptor(pg, data, f, p, result);
            descriptors.push_back(result);
				progressWindow->progress( ((float)pgIndex+(float)p/(float)numPatches) / (float)pgs.size()* 100.0f  );
         }
			//progressWindow->progress( (float)p/(float)numPatches * 100.0f  );
      }

   }
	progressWindow->popSteps();
	const size_t numDescriptors = descriptors.size();

	/// normalize categories
	for (card32 c=0; c<numComponents; c++) {
      const card32 compOffset = components[c].offset;
      const card32 compSize = components[c].size;
      DVectorF comp(compSize);
      DVectorF av = nullDVector<float32>(compSize);
      for (card32 i=0; i<numDescriptors; i++) {
         for (card32 vi=0; vi<compSize; vi++) {
            comp[vi] = descriptors[i][vi+compOffset];
         }
         av += comp;
      }
      av /= numDescriptors;
      float32 avNorm = 0;
      for (card32 i=0; i<numDescriptors; i++) {
         for (card32 vi=0; vi<compSize; vi++) {
            comp[vi] = descriptors[i][vi+compOffset];
         }
         avNorm += normQuad(comp-av);
      }
      avNorm /= numDescriptors;
      avNorm = sqrt(fabs(avNorm));

      if (avNorm > 1e-5) {
         for (card32 i=0; i<numDescriptors; i++) {
            for (card32 vi=0; vi<compSize; vi++) {
               descriptors[i][vi+compOffset] /= avNorm;
            }
         }
      }
   }
   
   /// compress by PCA and store in channel
   DynamicPCA<float32> pca(&(descriptors[0]), (card32)descriptors.size());
   pca.calcMean();
	pca.calcEigenValues(false);
	DynamicMatrix<float32> evs = pca.getEigenVectors();
	
   card32 pgElementIndex = 0;
   const card32 maxDim = min(channelPCADim, evs.getColumns());

   for (card32 pgIndex=0; pgIndex<pgs.size(); pgIndex++) {
      PatchGraph *pg = pgs[pgIndex];
   	pg->createPatchAttribute(mVAD(channelName, channelPCADim, VAD::DATA_FORMAT_FLOAT32));
   	const AAT channelAAT = pg->getTrajectorySet()->getTrajectories()->getAAT(channelName);
      PointSet *ps = pg->getTrajectorySet()->getTrajectories();
      const card32 numFrames = ps->getWidth();
      const card32 numPatches = ps->getHeight();

		for (card32 p=0; p<numPatches; p++) {
         for (card32 f=0; f<numFrames; f++) {
            float32 *dataPos = (float32*)(ps->getDataPointer(f,p) + channelAAT.getOffset());
            for (card32 d=0; d<maxDim; d++) {
               *dataPos = evs[d]*descriptors[pgElementIndex];
               dataPos++;
            }
            pgElementIndex++;
         }
      }   
   }   
   endComputeDescriptors(pgs, data);
}



// --


PGDescriptorChannelsTest::PGDescriptorChannelsTest()
{
   useColor = false;
   useCurvature = false;
   numComponents = 0;
   lastPG = NULL;
}


void PGDescriptorChannelsTest::startComputeDescriptors(
   const vector<PatchGraph*> &pgs,
   const PointCloudAnimationSequence *data,
   vector<Component> &components)
{
   pAssert(!pgs.empty());
   numComponents = 0;
   if (data->getFrame(0)->getDescr()->providesAttribute("color", 3, VAD::DATA_FORMAT_FLOAT32)) {
      useColor = true;
      numComponents += 3;
      dataColorAAT = data->getFrame(0)->getAAT("color");
      Component c;
      c.offset = 0;
      c.size = 3;
      components.push_back(c);
   }
   if (pgs[0]->getTrajectorySet()->getTrajectories()->getDescr()->providesAttribute(HEIGHTFIELD_COEFFS_NAME, 6, VAD::DATA_FORMAT_FLOAT32)) {
      useCurvature = true;
      numComponents += 3;
      patchCurvatureAAT = pgs[0]->getTrajectorySet()->getTrajectories()->getAAT(HEIGHTFIELD_COEFFS_NAME);
      Component c;
      c.offset = useColor?3:0;
      c.size = 3;
      components.push_back(c);
   }
   if (!useColor && !useCurvature) {
      throw PException("PGDescriptorChannelsTest::startComputeDescriptors() - cannot find data color or local monomials.");
   }
   lastPG = pgs[0];
}

void PGDescriptorChannelsTest::computeDescriptor(
   const PatchGraph *pg,
   const PointCloudAnimationSequence *data,
   const int32 frameIndex, const card32 patchIndex,
   DVectorF &result)
{
   if (pg != lastPG) {
      if (pg->getTrajectorySet()->getTrajectories()->getDescr()->providesAttribute(HEIGHTFIELD_COEFFS_NAME, 6, VAD::DATA_FORMAT_FLOAT32)) {
         patchCurvatureAAT = pg->getTrajectorySet()->getTrajectories()->getAAT(HEIGHTFIELD_COEFFS_NAME);
      }
      lastPG = pg;
   }
   result.setDim(numComponents);
   if (useColor) {
      vector<card32> dps = pg->getTrajectorySet()->getDataPoints()->getIndices(frameIndex, patchIndex);
      const size_t numDPoints = dps.size();
      Vector3f c = NULL_VECTOR3F;
      
      for (card32 i=0; i<numDPoints; i++) {
         c += data->getFrame(frameIndex + pg->getStartFrameIndex())->get<float32, 3>(dps[i], dataColorAAT);
      }
      c /= numDPoints;
      result[0] = c[0];
      result[1] = c[1];
      result[2] = c[2];
   }
   if (useCurvature) {
      const PointSet *ps = pg->getTrajectorySet()->getTrajectories();
      StaticVector<float32, 6> coeffs = ps->get<float32, 6>(frameIndex, patchIndex, patchCurvatureAAT);
      float32 uu = coeffs[3];
      float32 uv = coeffs[4];
      float32 vv = coeffs[5];
      
      float32 lambda1, lambda2;
      eigenValuesSym2f(uu, uv, vv, lambda1, lambda2);
      
      card32 offset = useColor?3:0;      
      result[0+offset] = 2*uu + 2*vv + uv;
      result[1+offset] = lambda1;
      result[2+offset] = lambda2;            
   }
}

void PGDescriptorChannels::computeDescriptors( 
		vector<PointCloud*> pointclouds,  const string channelName,  const card32 channelPCADim )
{
   vector<Component> components;
   startComputeDescriptors(pointclouds, components);
   const size_t numComponents = components.size();
   if (numComponents == 0) {
      throw PException("startComputeDescriptors failed.");
   }

   /// compute descriptors
   vector<DVectorF> descriptors;
   DVectorF result;
	progressWindow->pushStep( true, "desc" );
   for (card32 p=0; p<pointclouds.size(); p++) {
		PointCloud * pc = pointclouds[p];
		for( card32 i=0;i<pc->getNumPoints();i++ )
		{
			computeDescriptor(pc, i, result);
         descriptors.push_back(result);
			if( i % 300 == 1 )
				progressWindow->progress( ((float)p+(float)i/(float)pc->getNumPoints()) / (float)pointclouds.size()* 100.0f  );
      }

   }
	progressWindow->popSteps();
   const size_t numDescriptors = descriptors.size();

   /// normalize categories
   for (card32 c=0; c<numComponents; c++) {
      const card32 compOffset = components[c].offset;
      const card32 compSize = components[c].size;
      DVectorF comp(compSize);
      DVectorF av = nullDVector<float32>(compSize);
      for (card32 i=0; i<numDescriptors; i++) {
         for (card32 vi=0; vi<compSize; vi++) {
            comp[vi] = descriptors[i][vi+compOffset];
         }
         av += comp;
      }
      av /= numDescriptors;
      float32 avNorm = 0;
      for (card32 i=0; i<numDescriptors; i++) {
         for (card32 vi=0; vi<compSize; vi++) {
            comp[vi] = descriptors[i][vi+compOffset];
         }
         avNorm += normQuad(comp-av);
      }
      avNorm /= numDescriptors;
      avNorm = sqrt(fabs(avNorm));

      if (avNorm > 1e-5) {
         for (card32 i=0; i<numDescriptors; i++) {
            for (card32 vi=0; vi<compSize; vi++) {
               descriptors[i][vi+compOffset] /= avNorm;
            }
         }
      }
   }
   
   /// compress by PCA and store in channel
   DynamicPCA<float32> pca(&(descriptors[0]), (card32)descriptors.size());
   pca.calcMean();
	pca.calcEigenValues(false);
	DynamicMatrix<float32> evs = pca.getEigenVectors();
	
   card32 descIndex = 0;
   const card32 maxDim = min(channelPCADim, evs.getColumns());

   for (card32 p=0; p<pointclouds.size(); p++) 
	{
      PointCloud * pc = pointclouds[p];
		checkAttribute( pc, channelName, channelPCADim, VAD::DATA_FORMAT_FLOAT32 );
   	const AAT channelAAT = pc->getAAT(channelName);
		PointSet *ps = ((UnstructuredInCorePointCloud*)pc)->getPointSet();

		for( card32 i=0;i<ps->getNumEntries();i++ )	
		{
			float32 *dataPos = (float32*)(ps->getDataPointer(i) + channelAAT.getOffset());
            for (card32 d=0; d<maxDim; d++) 
				{
               *dataPos = evs[d]*descriptors[descIndex];
               dataPos++;
            }
            descIndex++;
      }
		pc->clearAttachments();
   }   
   endComputeDescriptors(pointclouds);
}

void PGDescriptorChannels::computeDescriptor( PointCloud * pc, card32 pointIndex, DVectorF &result ) 
{
}
