//---------------------------------------------------------------------------
#ifndef PGDescriptorChannelsH
#define PGDescriptorChannelsH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {
   using namespace std;

class PatchGraph;
class PointCloudAnimationSequence;


/**
	This class computes descriptors for all patches of a patch graphs and compresses them using PCA.
	The result is stored in a user specified channel
*/
class ANIMATIONRECONSTRUCTION_API PGDescriptorChannels : public Persistent {
	GEOXL_ABSTRACT_CLASS( PGDescriptorChannels )

 protected:
   struct Component {
      card32 offset;
      card32 size;
   };

   /// start computing descriptors
   /// components should return a list of start indices of descriptor values
   /// for example: for a descriptor (color7f|curvature2f|sift40f) it should return ([0,7],[7,2],[9,40])
   virtual void startComputeDescriptors(
      const vector<PatchGraph*> &pgs,
      const PointCloudAnimationSequence *data,
      vector<Component> &components) = 0;

   /// compute the descriptors for a single patch
   /// return result as concatenated values
   virtual void computeDescriptor(
      const PatchGraph *pg,
      const PointCloudAnimationSequence *data,
      const int32 frameIndex, const card32 patchIndex,
      DVectorF &result) = 0;

   /// cleanup (if necessary)
   virtual void endComputeDescriptors(
      const vector<PatchGraph*> &pgs,
      const PointCloudAnimationSequence *data) {}

	// same methods but for simple point clouds
	virtual void startComputeDescriptors(vector<PointCloud*> pcs, vector<Component> &components)=0;
	virtual void computeDescriptor( PointCloud * pc, card32 pointIndex, DVectorF &result );
	virtual void endComputeDescriptors(vector<PointCloud*> pcs) {}


public:
	PGDescriptorChannels();
   
   virtual void assign(const Object* obj, CopyContext *context = NULL);

   /// compute descriptors and store in channel
   /// normalizes descriptors in each "component" (see startComputeDescriptors) over all patches
   /// then applies PCA over all descriptor vectors to remove correlations / reduce dimension
   virtual void computeDescriptors(
      vector<PatchGraph*> pg,
      const PointCloudAnimationSequence *data,
      const string channelName = "descriptor",
      const card32 channelPCADim = 3
   );

	virtual void computeDescriptors(
		vector<PointCloud*> pointclouds,
      const string channelName = "descriptor",
      const card32 channelPCADim = 3 );
   
	~PGDescriptorChannels();
};



/// simple test class
/// uses average data points color + min/max/integral curvature as 6dim descriptor
class ANIMATIONRECONSTRUCTION_API PGDescriptorChannelsTest : public PGDescriptorChannels {
	GEOXL_CLASS( PGDescriptorChannelsTest )

 private:
   bool useColor;
   bool useCurvature;
   card32 numComponents;
   AAT dataColorAAT;
   AAT patchCurvatureAAT;
   const PatchGraph *lastPG;
   
 protected:
   virtual void startComputeDescriptors(
      const vector<PatchGraph*> &pg,
      const PointCloudAnimationSequence *data,
      vector<Component> &components);

   virtual void computeDescriptor(
      const PatchGraph *pg,
      const PointCloudAnimationSequence *data,
      const int32 frameIndex, const card32 patchIndex,
      DVectorF &result);

	virtual void startComputeDescriptors(vector<PointCloud*> pcs, vector<Component> &components){}

 public:
	PGDescriptorChannelsTest();
};


}

#endif