#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PGDescriptorFactory.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "FeatureDescriptor.h"
#include "ObjectListProperty.h"
#include "StringHelper.h"
#include "UnstructuredInCorePointCloud.h"
#include "TrajectorySet.h"
#include "PatchGraph.h"
#include "VertexArray.h"
#include "SplatPullPushInterpolation.h"
#include "HeighfieldParameterization.h"
#include "PGDataPointIndices.h"
#include "PointCloudAnimationSequence.h"
#include "UniformPointCloudResampling.h"
#include "ProgressWindow.h"
#include "Timer.h"
#include "ClassMethods.h"
#include "FDHistogram.h"
#include "FDAttribute.h"
#include "FDClusteredHistogram.h"
#include "TopologicalKNNIterator.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( PGDescriptorFactory ,0)
{
	BEGIN_CLASS_INIT( PGDescriptorFactory );

	ADD_OBJECT_LIST_PROP( Descriptor, 0, FeatureDescriptor::getClass() );
	ADD_BOOLEAN_PROP( m_UseDataPoints , 0 );
	ADD_FLOAT32_PROP( m_DescriptorSize, 0 );

	ADD_NOARGS_METHOD( setupFunkhouserChrom );
	ADD_NOARGS_METHOD( setupFunkhouserRGB );
	ADD_NOARGS_METHOD( setupHistogramChrom );
	ADD_NOARGS_METHOD( setupHistogramRGB );
	ADD_NOARGS_METHOD( setupHistogramChrom_Curvature );
}

PGDescriptorFactory::PGDescriptorFactory(void)
{
	m_UseDataPoints = false;
	m_CurrentFramePointCloud = NULL;
	m_DescriptorSize = 0.25f;

	m_DescriptorContext.interpolation = new SplatPullPushInterpolation;
	m_DescriptorContext.parameterization = new HeighfieldParameterization;
	m_DescriptorContext.parametricImageSize = 32;
	m_DescriptorContext.tIt = NULL;
	m_TopologyIterator = NULL;
	m_CurrentPC = NULL;
}

PGDescriptorFactory::~PGDescriptorFactory(void)
{
	clearData();
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	delete m_DescriptorContext.tIt;
	delete m_TopologyIterator;
}

void PGDescriptorFactory::assign(const Object* obj, CopyContext *context)
{
	PGDescriptorChannels::assign(obj,context);
	const PGDescriptorFactory * o = dynamic_cast<const PGDescriptorFactory*>(obj);
	pAssert( o != NULL );

	copyObjectProperties(obj, this );
	clearData();

}
void PGDescriptorFactory::startComputeDescriptors( const vector<PatchGraph*> &pgs, const PointCloudAnimationSequence *data, vector<Component> &components) 
{
	card32 offset = 0;
	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		Component comp;
		comp.size = m_Descriptors[i]->getNumDescriptorValues();
		comp.offset = offset;
		offset += comp.size;
		components.push_back( comp );
	}
	m_Components = components;
	m_NumDescriptorComponents = offset;
	m_CurrentFrameIndex = -1;
	m_CurrentPG = NULL;

	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		FDClusteredHistogram * fdc = dynamic_cast<FDClusteredHistogram*> (m_Descriptors[i]);
		if( fdc )
			fdc->setupClusters( pgs, data );
	}
}

void PGDescriptorFactory::computeDescriptor( const PatchGraph *pg, const PointCloudAnimationSequence *data, const int32 frameIndex, const card32 patchIndex, DVectorF &result) 
{
	if( m_CurrentFrameIndex != frameIndex || m_CurrentPG != pg)
	{
		m_CurrentFrameIndex = frameIndex;
		m_CurrentPG = (PatchGraph*)pg;
		buildCurrentPointCloud( pg, data, frameIndex );
	}

	result.setDim( m_NumDescriptorComponents );

	Vector3f position = m_CurrentFramePointCloud->getPointSet()->get3f( patchIndex, m_PosAAT );

	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		m_Descriptors[i]->createDescriptor( position, m_DescriptorSize, m_CurrentFramePointCloud, &m_DescriptorContext );
		for( card32 j=0;j<m_Descriptors[i]->getNumDescriptorValues();j++ )
		{
			result[ m_Components[i].offset + j] = (*m_Descriptors[i])[j];
		}
	}

}

void PGDescriptorFactory::addDescriptor( Object * obj ) 
{
	m_Descriptors.push_back( (FeatureDescriptor*)obj );
}

void PGDescriptorFactory::deleteDescriptors( card32 begin, card32 end ) 
{
	for( card32 i=begin;i<end;i++ )
		delete m_Descriptors[i];
	m_Descriptors.erase( m_Descriptors.begin() + begin, m_Descriptors.begin() + end );
}

Object *PGDescriptorFactory::getDescriptor( card32 index ) 
{
	return m_Descriptors[index];
}

void PGDescriptorFactory::moveDescriptor( card32 oldPos, card32 newPos ) 
{
	FeatureDescriptor * desc = m_Descriptors[oldPos];
	m_Descriptors.erase( m_Descriptors.begin() + oldPos );
	m_Descriptors.insert( m_Descriptors.begin() + newPos, desc );
}

mpcard PGDescriptorFactory::getNumDescriptors() 
{
	return m_Descriptors.size();
}

string PGDescriptorFactory::getDescriptorName(card32 index ) 
{
	return "desc" + intToStr(index);
}

void PGDescriptorFactory::buildCurrentPointCloud(const PatchGraph *pg, const PointCloudAnimationSequence *data, const int32 frameIndex) 
{
	clearData();

	if( m_UseDataPoints )
	{
		PointSet * ps = (PointSet*)pg->getTrajectorySet()->getTrajectories();
		m_CurrentFramePointCloud = new UnstructuredInCorePointCloud;
		VertexDescriptor * vd = new VertexDescriptor;
		vd->pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd->pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd->pushAttrib( VAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd->pushAttrib( VAD( "curv3f", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd->pushAttrib( VAD( "color_yuv", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		m_CurrentFramePointCloud->clearAndSetup( vd, 0 );

		AAT patch_posAAT = ps->getAAT( "position" );
		AAT patch_normalAAT = ps->getAAT( "normal" );
		AAT patch_tangentuAAT = ps->getAAT( "tangent_u" );
		AAT patch_coeffsAAT = ps->getAAT( "hfieldCoeffs" );

		PGDataPointIndices * ind = pg->getTrajectorySet()->getDataPoints();
		PointCloudAnimationSequence * seq = (PointCloudAnimationSequence*)data;
		PointSet * dataPoints = seq->getFrame( pg->getStartFrameIndex() );
		AAT posAAT = dataPoints->getAAT( "position" );
		AAT colorAAT = dataPoints->getAAT( "color" );
		card32 destPosOffset = m_CurrentFramePointCloud->getDescr()->getAttributeOffset( m_CurrentFramePointCloud->getDescr()->getIndex( "position"));
		card32 destNormalOffset = m_CurrentFramePointCloud->getDescr()->getAttributeOffset( m_CurrentFramePointCloud->getDescr()->getIndex( "normal" ));
		card32 destColorOffset = m_CurrentFramePointCloud->getDescr()->getAttributeOffset( m_CurrentFramePointCloud->getDescr()->getIndex( "color" ));
		card32 destCurv3fOffset = m_CurrentFramePointCloud->getDescr()->getAttributeOffset( m_CurrentFramePointCloud->getDescr()->getIndex( "curv3f" ));
		card32 destColorYUVOffset = m_CurrentFramePointCloud->getDescr()->getAttributeOffset( m_CurrentFramePointCloud->getDescr()->getIndex( "color_yuv" ));

		vector<char> buffer(m_CurrentFramePointCloud->getDescr()->getSize(),0);
		Vector3f *posPtr = (Vector3f*) ( ((char*)&buffer[0]) + destPosOffset  );
		Vector3f *colorPtr = (Vector3f*) ( ((char*)&buffer[0]) + destColorOffset  );
		Vector3f *normalPtr = (Vector3f*) ( ((char*)&buffer[0]) + destNormalOffset  );
		Vector3f *curv3fPtr = (Vector3f*) ( ((char*)&buffer[0]) + destCurv3fOffset  );
		Vector3f *colorYUVPtr = (Vector3f*) ( ((char*)&buffer[0]) + destColorYUVOffset  );

		UnstructuredInCorePointCloud *tmpPC = new UnstructuredInCorePointCloud;
		tmpPC->clear();
		tmpPC->getPointSet()->assign( ps );
		HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator( tmpPC, 32, NULL );


		// project data points on patch surface
		for( card32 i=0;i<ps->getNumEntries();i++ )
		{
			Vector3f patch_pos = ps->get3f( i,patch_posAAT );
			Vector3f patch_normal = ps->get3f( i,patch_normalAAT );
			Vector3f patch_tangentU = ps->get3f( i,patch_tangentuAAT );
			StaticVector<float,6> patch_coeffs = ps->get<float,6>( i,patch_coeffsAAT );

			// setup coordinate systems
			Matrix3f worldToPatch;
			worldToPatch[0] = patch_tangentU;
			worldToPatch[1] = patch_normal.crossProduct( patch_tangentU	 );
			worldToPatch[2] = patch_normal;
			Matrix3f patchToWorld = worldToPatch;
			worldToPatch = worldToPatch.transpose();

			// calculate patch curvature
			float32 uu = patch_coeffs[3];
			float32 uv = patch_coeffs[4];
			float32 vv = patch_coeffs[5];
			float32 lambda1, lambda2;
			eigenValuesSym2f(uu, uv, vv, lambda1, lambda2);

			// get direct neighbours (for discarding data points near other patches)
			vector<Vector3f> neighbours;
			hIt->setSeekPointAndReset( patch_pos );
			for( card32 j=0;j<10;j++ )
			{
				if( hIt->atEnd() )
					break;
				hIt->next();
				neighbours.push_back( hIt->get3f( patch_posAAT ) );
			}

			for( card32 j=0;j<ind->getNumDataPoints(0, i );j++ )
			{
				card32 index = ind->getIndex(0, i, j);
				Vector3f pos = dataPoints->get3f( index, posAAT );
				Vector3f color = dataPoints->get3f( index, colorAAT );

				float dist = norm( pos - patch_pos );
				bool nextPoint = false;
				for( card32 k=0;k<neighbours.size();k++ )
				{
					if( norm( pos - neighbours[k] ) < dist )
					{
						nextPoint = true;
						break;
					}
				}
				if( nextPoint )
					continue;

				Vector3f localPos = worldToPatch * (pos - patch_pos);
				float u = localPos[0];
				float v = localPos[1];
				localPos[2] =	pow( u, 2.0f ) * patch_coeffs[3]
				+	pow( v, 2.0f ) * patch_coeffs[5]
				+	u * v * patch_coeffs[4]
				+	u * patch_coeffs[1]
				+	v * patch_coeffs[2]
				+ patch_coeffs[0];

				float derivate_u = patch_coeffs[1] + 2.0f * u * patch_coeffs[3] + v * patch_coeffs[4];
				float derivate_v = patch_coeffs[2] + 2.0f * v * patch_coeffs[5] + u * patch_coeffs[4];

				Vector3f normal = -makeVector3f(1,0, derivate_u).crossProduct(makeVector3f(0,1, derivate_v));

				*curv3fPtr = makeVector3f(2*uu + 2*vv + uv, lambda1, lambda2 );

				*posPtr = patchToWorld * localPos + patch_pos;
				*normalPtr = patchToWorld * normal;
				
				
				Vector3f colorYUV;
				colorYUV[0] = 0.299f * color[0] + 0.587f * color[1] + 0.114f * color[2];
				colorYUV[1] = (color[2] - colorYUV[0]) * 0.493f;
				colorYUV[2] = (color[1] - colorYUV[0]) * 0.877f;
				
				*colorYUVPtr = colorYUV;
				*colorPtr = color;

				m_CurrentFramePointCloud->getPointSet()->addPoint( (char*)&buffer[0] );
			}

		}


		//float tmppc_averagePointDistance = geoxl::getAveragePointDistance( tmpPC );
		delete hIt;
		delete tmpPC;


	}
	else
	{
		m_CurrentFramePointCloud = new UnstructuredInCorePointCloud;
		m_CurrentFramePointCloud->getPointSet()->assign(pg->getTrajectorySet()->getTrajectories() );
		checkAttribute( m_CurrentFramePointCloud, "color", 3, VAD::DATA_FORMAT_FLOAT32 );
		checkAttribute( m_CurrentFramePointCloud, "curv3f", 3, VAD::DATA_FORMAT_FLOAT32 );
		AAT colorAAT = m_CurrentFramePointCloud->getAAT( "color" );
		AAT curv3fAAT = m_CurrentFramePointCloud->getAAT( "curv3f" );
		UnstructuredInCorePointCloud tmpPC;
		tmpPC.clear();
		tmpPC.getPointSet()->assign( data->getFrame(frameIndex + pg->getStartFrameIndex()) );
		HierarchicalKNNIterator hIt( &tmpPC, 32, NULL);
		AAT posAAT = tmpPC.getAAT( "position" );

		// estimate color form data points
		VertexArray va( m_CurrentFramePointCloud->getPointSet() );
		AAT srcColorAAT = data->getFrame(frameIndex + pg->getStartFrameIndex())->getAAT( "color" );
		AAT srcPatchAAT = m_CurrentFramePointCloud->getAAT( "hfieldCoeffs" );
		float maxRadius = getMedianPointDistance( m_CurrentFramePointCloud, 5, 0 ) * 0.8f;
		for( card32 i=0;i<va.getNumElements();i++ )
		{
			Vector3f c = NULL_VECTOR3F;
	      
			// calculate patch curvature
			StaticVector<float,6> patch_coeffs = va.get<float,6>( i,srcPatchAAT );
			float32 uu = patch_coeffs[3];
			float32 uv = patch_coeffs[4];
			float32 vv = patch_coeffs[5];
			float32 lambda1, lambda2;
			eigenValuesSym2f(uu, uv, vv, lambda1, lambda2);
			float mean = 2.0f*uu + 2.0f*vv + uv;
			Vector3f curv3f = makeVector3f( mean, lambda1, lambda2 );
			va.set<float,3>( i, curv3fAAT, curv3f );

			Vector3f patchPos = va.getPosition3f(i);
			hIt.setSeekPointAndReset( patchPos );
			card32 numNeighbours = 0;
			while( !hIt.atEnd() )
			{
				c += hIt.get3f( srcColorAAT );
				numNeighbours++;
				Vector3f pos = hIt.get3f( posAAT );
				if( norm( pos - patchPos ) > maxRadius )
					break;
				hIt.next();
			}
			c /= numNeighbours;

			va.setColor3f( i, c );
		}

		delete m_DescriptorContext.tIt;
		delete m_TopologyIterator;
		m_TopologyIterator = new InCorePCTopologyGraphPointSetIterator( m_CurrentFramePointCloud->getPointSet(),
			m_CurrentFramePointCloud->getAttachments(), (InCorePCTopologyGraph *)pg->getTopology() );
		m_DescriptorContext.tIt = new TopologicalKNNIterator(m_CurrentFramePointCloud, m_TopologyIterator );

	}


	m_PosAAT = m_CurrentFramePointCloud->getAAT( "position" );
	m_DescriptorContext.averagePointDistance = getAveragePointDistance( m_CurrentFramePointCloud );
	m_DescriptorContext.pc = m_CurrentFramePointCloud;
}

void PGDescriptorFactory::clearData() 
{
	delete m_CurrentFramePointCloud;
	m_CurrentFramePointCloud = NULL;
}

void PGDescriptorFactory::setupFunkhouserChrom() 
{
}

void PGDescriptorFactory::setupFunkhouserRGB() 
{
}

void PGDescriptorFactory::setupHistogramChrom() 
{
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	FDHistogram * fd = new FDHistogram;
	fd->setAttribute( "color_yuv" );
	fd->setMinValue( -1 );
	fd->setMaxValue( 1 );
	fd->setNumHistograms( 3 );
	fd->setNumBins( 8 );
	fd->setAttributeCompNumber(1);
	addDescriptor( fd );
	fd = (FDHistogram*) fd->copy();
	fd->setAttributeCompNumber(2);
	addDescriptor( fd );
}

void PGDescriptorFactory::setupHistogramRGB() 
{
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	FDHistogram * fd = new FDHistogram;
	fd->setAttribute( "color" );
	fd->setMinValue( 0 );
	fd->setMaxValue( 1 );
	fd->setNumHistograms( 3 );
	fd->setNumBins( 8 );
	fd->setAttributeCompNumber(0);
	addDescriptor( fd );
	fd = (FDHistogram*) fd->copy();
	fd->setAttributeCompNumber(1);
	addDescriptor( fd );
	fd = (FDHistogram*) fd->copy();
	fd->setAttributeCompNumber(2);
	addDescriptor( fd );
}

void PGDescriptorFactory::setupHistogramChrom_Curvature() 
{
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	FDHistogram * fd = new FDHistogram;
	fd->setAttribute( "color_yuv" );
	fd->setMinValue( -1 );
	fd->setMaxValue( 1 );
	fd->setNumHistograms( 3 );
	fd->setNumBins( 8 );
	fd->setAttributeCompNumber(1);
	addDescriptor( fd );
	fd = (FDHistogram*) fd->copy();
	fd->setAttributeCompNumber(2);
	addDescriptor( fd );
	FDAttribute * fa = new FDAttribute;
	fa->setAttributeName( "curv3f" );
	fa->setFirstComponent(0);
	fa->setNumComponents(3);
	addDescriptor( fa );
}

void PGDescriptorFactory::startComputeDescriptors(vector<PointCloud*> pcs, vector<Component> &components) 
{
	card32 offset = 0;
	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		Component comp;
		comp.size = m_Descriptors[i]->getNumDescriptorValues();
		comp.offset = offset;
		offset += comp.size;
		components.push_back( comp );
	}
	m_Components = components;
	m_NumDescriptorComponents = offset;
	m_CurrentFrameIndex = -1;
	m_CurrentPG = NULL;
	m_CurrentPC = NULL;

/*	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		FDClusteredHistogram * fdc = dynamic_cast<FDClusteredHistogram*> (m_Descriptors[i]);
		if( fdc )
			fdc->setupClusters( pgs, data );
	}*/
}

void PGDescriptorFactory::computeDescriptor( PointCloud * pc, card32 pointIndex, DVectorF &result ) 
{
	if( m_CurrentPC != pc )
	{
		m_CurrentPC = (UnstructuredInCorePointCloud*)pc;
		m_PosAAT = pc->getAAT( "position" );
		m_DescriptorContext.averagePointDistance = getAveragePointDistance( m_CurrentPC );
		m_DescriptorContext.pc = m_CurrentPC;
	}
	result.setDim( m_NumDescriptorComponents );

	Vector3f position = m_CurrentPC->getPointSet()->get3f( pointIndex, m_PosAAT );

	for( card32 i=0;i<m_Descriptors.size();i++ )
	{
		m_Descriptors[i]->createDescriptor( position, m_DescriptorSize, m_CurrentPC, &m_DescriptorContext );
		for( card32 j=0;j<m_Descriptors[i]->getNumDescriptorValues();j++ )
		{
			result[ m_Components[i].offset + j] = (*m_Descriptors[i])[j];
		}
	}
}

void PGDescriptorFactory::setupHistogramCurvature() 
{
	deleteDescriptors(0, (card32)m_Descriptors.size() );
	FDHistogram * fd = new FDHistogram;
	fd->setAttribute( "curvature" );
	fd->setMinValue( 0 );
	fd->setMaxValue( 1 );
	fd->setNumHistograms( 4 );
	fd->setNumBins( 8 );
	fd->setAttributeCompNumber(0);
	addDescriptor( fd );
}
