#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RegistrationTools.h"
//---------------------------------------------------------------------------
#include "PointCloudTools.h"
#include "ProgressWindow.h"
#include "LinearTransformPriors.h"
#include "PointCloudITerators.h"
#include "NormalDistribution.h"
#include "VertexArray.h"
#include "Monomials2ndOrder.h"
//---------------------------------------------------------------------------
using namespace geoxl;

void geoxl::calculateCurvature( PointCloud * pc, int numNeighbours, string destAttribute,
									  float averagePointDistance)
{
	checkAttribute( pc, destAttribute, 2, VAD::DATA_FORMAT_FLOAT32 );
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT( "normal" );
	AAT destAAT = pc->getAAT( destAttribute );
	if( averagePointDistance == 0.0f )
		averagePointDistance = getAveragePointDistance(pc);

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc, 32, NULL );

	vector<float> meanCurv( pc->getNumPoints()), gaussianCurv( pc->getNumPoints());

	card32 index = 0;
	while( !bIt->atEnd() )
	{
		static vector<Vector3f> posBuffer;
		posBuffer.resize( numNeighbours );

		Vector3f normal = bIt->get3f( normalAAT );
		Matrix3f invTangentSystem = calcTangentSystem( normal ).transpose();

		Vector3f pos = bIt->get3f( posAAT );
		hIt->setSeekPointAndReset( pos );

		posBuffer[0] = makeVector3f(0,0,0);

		int actPoint = 1;
		for( card32 i=0;i<1000;i++ )
		{
			hIt->next();

			Vector3f tmp_normal = hIt->get3f( normalAAT );
			//if( tmp_normal * normal > 0 )
			{
				posBuffer[actPoint] = invTangentSystem * (hIt->get3f( posAAT ) - pos);
				actPoint++;
			}
			if( actPoint >= numNeighbours )
				break;
		}


		// scale positions
		for( card32 i=0;i<(card32)actPoint;i++ )
		{
			posBuffer[i] /= sqrt( (float)numNeighbours ) * averagePointDistance;
		}

		StaticVector<float32, 6> coeff;
		vector< StaticVector<float32, 6> > coeffsDerivs;
		if( actPoint == numNeighbours )
		{
			GLTMonomials2ndOrder::computeCoeffs( posBuffer, coeff, coeffsDerivs );
		}
		else
		{
			vector<Vector3f> tmpBuffer;
			for( int i=0;i<actPoint;i++ )
				tmpBuffer.push_back( posBuffer[i] );
			GLTMonomials2ndOrder::computeCoeffs( tmpBuffer, coeff, coeffsDerivs );

			debugOutput << "Couldn't find enough points for curvature calculation!(num=" << actPoint << ")\n";

		}

		float meanCurvature =	pow( coeff[3] , 2.0f ) + 
										2.0f * pow( coeff[4] , 2.0f ) + 
										2.0f * pow( coeff[5] , 2.0f );

		float extrema = 0.0f;

		float f_x = coeff[1];
		float f_y = coeff[2];
		float f_xy = coeff[3];
		float f_xx = coeff[4];
		float f_yy = coeff[5];

		float K = (f_xx*f_yy - pow( f_xy, 2.0f ))
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

		float H = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;


		Vector2f res_vec;
		float sign = H<0 ? -1:1;
		res_vec[ 0 ] = sqrt(meanCurvature);
		res_vec[ 1 ] = K;
		mIt->set2f( destAAT, res_vec );
		meanCurv[index] = res_vec[0];
		gaussianCurv[index] = res_vec[1];
		index++;

		bIt->next();
	}

	sort( meanCurv.begin(), meanCurv.end());
	sort( gaussianCurv.begin(), gaussianCurv.end());

	Vector2f meanBounds = makeVector2f( meanCurv[meanCurv.size() * 0.05f], meanCurv[meanCurv.size() * 0.95f]);
	Vector2f gaussianBounds = makeVector2f( gaussianCurv[gaussianCurv.size() * 0.05f], gaussianCurv[gaussianCurv.size() * 0.95f]);

	bIt->reset();
	while( !bIt->atEnd())
	{
		Vector2f curv = bIt->get2f( destAAT );
		curv[0] = (curv[0] - meanBounds[0]) * (meanBounds[1]-meanBounds[0]);
		curv[1] = (curv[1] - gaussianBounds[0]) * (gaussianBounds[1]-gaussianBounds[0]);
		mIt->set2f( destAAT, curv);

		bIt->next();
	}
	
	delete hIt;
	delete iter;

}

void geoxl::calculateGradients( PointCloud * pc, int numNeighbours, string destAttribute, 
														 string sourceAttribute,card32 sourceCompIndex,
														 float averagePointDistance,string positionAttribute) 
{
	pAssert( pc->providesAttribute(destAttribute) );
	AAT posAAT = pc->getAAT( positionAttribute );
	AAT normalAAT = pc->getAAT( "normal" );
	AAT destAAT = pc->getAAT( destAttribute );
	if( averagePointDistance == 0.0f )
		averagePointDistance = getAveragePointDistance(pc);

	AAT sourceAAT;
	VAD sourceVAD;
	if( sourceAttribute!="")
	{
		sourceAAT = pc->getAAT(sourceAttribute);
		sourceVAD = pc->getDescr()->getAttribute( sourceAttribute );
	}

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc, 32, NULL );
	
	progressWindow->pushStep( true, "gradients" );
	card32 index = 0;
	while( !bIt->atEnd() )
	{
		if( !progressWindow->progress( (float)index / (float)pc->getNumPoints() * 100.0f ) )
			break;
		index++;
		static vector<Vector3f> posBuffer;
		posBuffer.resize( numNeighbours );

		Vector3f normal = bIt->get3f( normalAAT );
		Matrix3f tangentSystem = calcTangentSystem( normal );
		Matrix3f invTangentSystem = tangentSystem.transpose();

		Vector3f pos = bIt->get3f( posAAT );
		hIt->setSeekPointAndReset( pos );

		int actPoint = 0;
		while( !hIt->atEnd() )
		{
			Vector3f tmp_normal = hIt->get3f( normalAAT );
			//if( tmp_normal * normal > 0 )
			{
				Vector3f samplePos = hIt->get3f( posAAT );
				posBuffer[actPoint] = invTangentSystem * (samplePos - pos);
				if( sourceAttribute.length() )
				{
					float height;
					switch( sourceVAD.numComponents )
					{
					case 1:
						height = hIt->get1f( sourceAAT);
						break;
					case 2:
						height = hIt->get2f( sourceAAT)[sourceCompIndex];
						break;
					case 3:
						height = hIt->get3f( sourceAAT)[sourceCompIndex];
						break;
					case 4:
						height = hIt->get4f( sourceAAT)[sourceCompIndex];
						break;
					}
					posBuffer[actPoint][2] = height;
				}
				actPoint++;
			}
			if( actPoint >= numNeighbours )
				break;
			hIt->next();
		}


		// scale positions
		for( card32 i=0;i<(card32)actPoint;i++ )
		{
			posBuffer[i] /= sqrt( (float)numNeighbours ) * averagePointDistance;
		}

		StaticVector<float32, 6> coeff;
		vector< StaticVector<float32, 6> > coeffsDerivs;
		if( actPoint == numNeighbours )
		{
			GLTMonomials2ndOrder::computeCoeffs( posBuffer, coeff, coeffsDerivs );
		}
		else
		{
			debugOutput << "asdf!\n";
		}

		Vector3f gradient = makeVector3f( coeff[1], coeff[2], 0 );
		float length = norm( gradient );
		gradient = normalize(gradient) * atan2(length,1.0f);
		mIt->set3f( destAAT, tangentSystem * gradient );

		/*RegressionPlane rp;
		rp.clear();
		for( card32 i=0;i<posBuffer.size();i++ )
		{
			rp.addPoint( posBuffer[i] );
		}
		Vector3f plane_normal = rp.getNormal();
		Vector3f gradient = makeVector3f( plane_normal[0] / plane_normal[2],
					plane_normal[1] / plane_normal[2],0 );
		//if( gradient.getSqrNorm() > 2.0f )
		//	gradient = normalize( gradient ) * 2.0f;
		gradient = tangentSystem * gradient;
		mIt->set3f( destAAT, tangentSystem * gradient );*/

		bIt->next();
	}

	progressWindow->popSteps();
	
	delete hIt;
	delete iter;

//	smoothAttribute3f( pc, "gradient", "gradient", 4.0f,false,averagePointDistance );
}

vector<float> geoxl::calculateGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
															 const Vector3f startPoint, const vector<Vector3f> points,
															 vector< vector<Vector3f> > * paths) 
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	pAssert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	// -----------------------------------------
	// distance transformation for the complete point cloud
	vector<float> distanceField( pc->getNumPoints(), 1e30f );
	HierarchicalKNNIterator hIt(pc,32,NULL);
	hIt.setSeekPointAndReset( startPoint );
	queue<PointIndex>	indexQueue;
	PointIndex startIndex;
	for( card32 i=0;i<6 && !hIt.atEnd();i++ )
	{
		startIndex = hIt.getPointSetPointNumber();
		distanceField[startIndex] = norm( hIt.get3f( posAAT ) - startPoint);
		indexQueue.push( startIndex );
		hIt.next();
	}

	while( !indexQueue.empty() )
	{
		PointIndex index = indexQueue.front();
		indexQueue.pop();	// remove first element

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		float currentDistance = distanceField[index];
		for( card32 i=0;i<tIter->getNumNeighbours();i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			neighbour_rIt->goToIndex(neighbour);
			Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

			float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
			if( distanceField[neighbour] > distance )
			{
				distanceField[neighbour] = distance;
				indexQueue.push( neighbour );
			}
			
		}
	}

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	vector<float> distances( points.size() );
	for( card32 i=0;i<points.size();i++ )
	{
		hIt.setSeekPointAndReset( points[i] );
		PointIndex index = hIt.getPointSetPointNumber();
		for( card32 j=0;j<6 && !hIt.atEnd();j++ )
		{
			float tmpDist = distanceField[index] + norm( points[i] - hIt.get3f(posAAT));
			if( j == 0 || tmpDist < distances[i] )
			{
				distances[i] = tmpDist;
				index = hIt.getPointSetPointNumber();
			}
			hIt.next();
		}

		if( points[i] == startPoint )
			distances[i] = 0.0f;

		if( paths != NULL )
		{
			vector<Vector3f> path;
			float lastDistance = 1e30f;
			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path.push_back( bIt->get3f(posAAT) );
				if( path.size() > 50000 )
					break;

				float minDist = 1e30f;
				PointIndex bestNeighbour = tIter->getNeighbour(0);
				for( card32 j=0;j<tIter->getNumNeighbours();j++ )
				{
					PointIndex neighbour = tIter->getNeighbour(j);
					if( distanceField[neighbour] < minDist )
					{
						minDist = distanceField[neighbour];
						bestNeighbour = neighbour;
					}
				}
				if( minDist > lastDistance )
				{
					break;
				}
				lastDistance = minDist;

				index = bestNeighbour;
			}
			paths->push_back( path );

		}
	}


	delete iter;
	return distances;
}

void geoxl::createGrayChannel( PointCloud * pc ) 
{
	checkAttribute( pc, "gray", 1, VAD::DATA_FORMAT_FLOAT32 );
	AAT colorAAT = pc->getAAT( "color" );
	AAT grayAAT = pc->getAAT( "gray" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_MODIFYABLE_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	while( !bIt->atEnd() )
	{
		Vector3f color = bIt->get3f( colorAAT );
		float gray = (color[0] + color[1] + color[2]) / 3.0f;
		mIt->set1f( grayAAT, gray );
		bIt->next();
	}

}

PointCloud *geoxl::createNoisyPointCloud( PointCloud * pc, float stdDeviation, float maxAmplitude ) 
{
	UnstructuredInCorePointCloud * pcCopy = (UnstructuredInCorePointCloud*)pc->copy();
	VertexArray va( pcCopy->getPointSet() );
	for( card32 i=0;i<va.getNumElements();i++ )
	{
		Vector3f noisyPos = va.getPosition3f(i) + 
			va.getNormal3f(i) * max( min( NormalDistribution::randomNumber( 0, stdDeviation ), maxAmplitude), - maxAmplitude);
		va.setPosition3f(i, noisyPos );
	}

	return pcCopy;
}

void geoxl::calculateCurvature( PointCloud * pc, float32 sigma, string destAttribute ) 
{
	vector<float> meanCurv, gaussianCurv;
	card32 numNANs = 0;
	checkAttribute( pc, destAttribute, 4 , VAD::DATA_FORMAT_FLOAT32 );
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT( "normal" );
	AAT destAAT = pc->getAAT( destAttribute );

	float sigma_2 = pow( sigma, 2.0f );
	float inv_sigma_2 = 1.0f / sigma_2;
	float maxRad = sqrt( -log(0.05f) * sigma_2 );

	HierarchicalKNNIterator hIt( pc, 32, NULL);
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	pAssert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);


	card32 actPoint = 0;
	card32 currentIndex = 0;
//	progressWindow->pushStep(true,"curv");
	while( !bIt->atEnd() )
	{
		vector<Vector3f> posBuffer,posBufferNormal;

		Vector3f pos = bIt->get3f( posAAT );
		Vector3f org_normal = bIt->get3f( normalAAT);
		hIt.setSeekPointAndReset( pos );
		PCA<float,3> pca;

		posBuffer.push_back( makeVector3f(0,0,0) );

		int actPoint = 1;
		while(!hIt.atEnd() )
		{
			hIt.next();

			Vector3f tmpPos = hIt.get3f( posAAT );

			float dist = norm( tmpPos - pos);

			if( dist > maxRad && actPoint >= 10)
				break;

			if( dist <= maxRad || actPoint < 5 )
			{
				posBuffer.push_back( tmpPos );
				pca.addPoint( tmpPos );
			}

			actPoint++;
		}

		Vector3f eigenValues, centroid;
		Matrix3f eigenVectors;
		pca.analyze( eigenValues, eigenVectors, centroid );

		Vector3f normal = normalize( eigenVectors[0].crossProduct( eigenVectors[1] ) );
		if( normal * org_normal < 0)
			normal = -normal;
		Matrix3f tangentSystem = calcTangentSystem( normal );
		Matrix3f invTangentSystem = tangentSystem.transpose();



		// transform points into local coordinate frame
		for( card32 i=0;i<posBuffer.size();i++ )
		{
			posBuffer[i] = invTangentSystem * ( posBuffer[i] - pos) / sigma;
		}

		StaticVector<float32, 6> coeff;
		vector< StaticVector<float32, 6> > coeffsDerivs;
		Monomials2ndOrder::computeCoeffs( posBuffer, coeff, 1 );

		float meanCurvature =	pow( coeff[3] , 2.0f ) + 
			2.0f * pow( coeff[4] , 2.0f ) + 
			2.0f * pow( coeff[5] , 2.0f );

		float extrema = 0.0f;

		float f_x = coeff[1];
		float f_y = coeff[2];
		float f_xy = coeff[3];
		float f_xx = 2.0f * coeff[4];
		float f_yy = 2.0f * coeff[5];

		float K = (f_xx*f_yy - pow( f_xy, 2.0f ))
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

		float H = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;


		Vector4f res_vec;
		float signH = H<0 ? -1:1;
		H = sqrt( fabs(H) ) * signH;
		float signK = K<0 ? -1 : 1;
		K = sqrt( fabs(K) ) * signK;
		res_vec[ 0 ] = H;
		res_vec[ 1 ] = K;
		mIt->set4f( destAAT, res_vec );
		//curvatures.push_back(res_vec);
		meanCurv.push_back( H );
		gaussianCurv.push_back( K );



		bIt->next();

//			if( currentIndex++ % 300 == 1 )
//				if( !progressWindow->progress( 100.0f * (float)currentIndex / pc->getNumPoints() ) )
//					break;
	}
//	bool aborted = false;
//	progressWindow->popSteps(aborted);


	if( numNANs )
		debugOutput << "num nans: " << numNANs << "\n";

	sort( meanCurv.begin(), meanCurv.end() );
	sort( gaussianCurv.begin(), gaussianCurv.end() );


	float minH = meanCurv[0];
	float maxH = meanCurv.back();
	float minK = gaussianCurv[0];
	float maxK = gaussianCurv.back();
	minH = meanCurv[ (float)meanCurv.size() * 0.05f ];
	maxH = meanCurv[ (float)meanCurv.size() * 0.95f ];
	minK = gaussianCurv[ (float)gaussianCurv.size() * 0.05f ];
	maxK = gaussianCurv[ (float)gaussianCurv.size() * 0.95f ];

	bIt->reset();
	while( !bIt->atEnd() )
	{
		Vector4f curv = bIt->get4f( destAAT );
		float tmpMean = min(1.0f, max(0.0f,(curv[0] - minH) / (maxH - minH ) ));
		curv[0] = ( fabs(tmpMean) );
		float tmpGaussian = min(1.0f, max(0.0f,(curv[1] - minK) / (maxK - minK ) ));
		curv[1] = ( fabs(tmpGaussian) );
		mIt->set4f( destAAT, curv );
		bIt->next();
	}

	delete iter;
}
