#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARAPDeformationContext.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ASAPDeformationSolver.h"
#include "ASAPDeformationEvaluator.h"
#include "ARSettings.h"
#include "DeformationTools.h"
#include "InCorePCTopologyGraph.h"
#include "PointSetKNNQuery.h"
#include "UnstructuredInCorePointCloud.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers_templCode.h"
#include "GeometricTools.h"
#include "DebugOutputImages.h"
#include "TopologicalKNNIterator.h"
#include "Misc/TriangleMeshTools.h"
//#include "fastSparseMathIntelMKL/FastSparseMatrix.h"
#include "private/fastSparseMathIntelMKL/FastSparseMatrix.h"
#include "StringHelper.h"
#include "Timer.h"
#include "CopyObjectProperties.h"
#include "PointCloudTools.h"
#include "SVD.h"
#include "VertexArray.h"
#include "ColorTools.h"
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( ARAPDeformationContext ,0)
{
	BEGIN_CLASS_INIT( ARAPDeformationContext );
}

ARAPDeformationContext::ARAPDeformationContext(void)
{
	A_fac = NULL;
	alwaysComputePerPointRotations = true;
}

ARAPDeformationContext::~ARAPDeformationContext(void)
{
}

void ARAPDeformationContext::assign(const Object* obj, CopyContext *context)
{
	ASAPDeformationContext::assign(obj,context);
	const ARAPDeformationContext * o = dynamic_cast<const ARAPDeformationContext*>(obj);
	pAssert( o != NULL );

	copyObjectProperties(obj,this);
}

void geoxl::ARAPDeformationContext::initializeFromShape( const SceneObject *obj, Vector2f timeInterval, const ARSettings *settings )
{
	
	this->settings = settings;
	weightChannel = false;
	
	if (timeInterval[0] != timeInterval[1])
	{
		warning("ASAPDeformationContext::initializeFromShape - animations are currently not supported.");
	}
	tg = getTopology((PointCloud*)obj);
	if (!tg) {warning("need topology."); return;}
	urshape = const_cast<UnstructuredInCorePointCloud*>( dynamic_cast<const UnstructuredInCorePointCloud*>(obj) );
	if (!urshape) {warning("need unstructured in core pc or descendant (e.g. mesh)."); return;}

}



void geoxl::ARAPDeformationContext::beginConstraints()
{
	if(!octree)
		octree = PointSetKNNQuery::createOctree(urshape->getPointSet());
	positionAAT = urshape->getPointSet()->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	if(!query)
		query = new PointSetKNNQuery(urshape->getPointSet(), octree, NULL_VECTOR3F, positionAAT);
	
	

	currentConstraintID = 0;
	allConstraints.clear();
}

void geoxl::ARAPDeformationContext::solve()
{
	if(urshape->getAttachments()->getData("SimpleOctree"))
		urshape->deleteAttachment("SimpleOctree"); // octree is invalid after deformation!

	
	writeLinearSystem();
	for (mpcard i = 0; i < allConstraints.size(); i++)
	{
		ASAPDeformationContext::addPositionConstraint(allConstraints[i]);
		delete allConstraints[i];
	}
	allConstraints.clear();
	
	
	Timer t;
	if (!A_fac){
		A_fac = new FastSparseMatrixFloatPrecision;
		A_fac->initUpperHalfMatrixFromFullSymmetricSparseMatrixF(linSystem_A);
		A_fac->computeFactorization();
		debugOutput << "factor A took: " << t.getDeltaValue() << "\n";
	}


	A_fac->solveUsingFactorization(linSystem_x, linSystem_b);
	debugOutput << "solve x = A-1 b took: " << t.getDeltaValue() << "\n";
	return;	

	if ( true )
	{
		debugOutput << "using direct solver\n";
		FastSparseMatrix linSystem_A_fast;
		linSystem_A_fast.initUpperHalfMatrixFromFullSymmetricSparseMatrixF(linSystem_A);
		linSystem_A_fast.solveWithIntelDirectSolverFloatPrescicion(linSystem_x, linSystem_b);
		debugOutput << "\n-----------------\nsolveWithIntelDirectSolver: "  << t.getDeltaValue() << "\n";
	}
}

DeformationSolver * geoxl::ARAPDeformationContext::createSolver()
{
	return NULL;
}

DeformationEvaluator * geoxl::ARAPDeformationContext::createEvaluator()
{
	return NULL;
}

void geoxl::ARAPDeformationContext::addPositionConstraint( const PositionARConstraint *c )
{

	pAssert(lastConstraintAnchors.size() >= currentConstraintID);
	
	if (lastConstraintAnchors.size() == currentConstraintID)
	{
		delete A_fac;
		A_fac = NULL;
		lastConstraintAnchors.push_back(c->getAnchor().pos);
	}
	else{
		if (lastConstraintAnchors[currentConstraintID] != c->getAnchor().pos)
		{
			delete A_fac;
			A_fac = NULL;
			lastConstraintAnchors[currentConstraintID] = c->getAnchor().pos;
		}
	}

	allConstraints.push_back((PositionARConstraint *)c->copy());
	currentConstraintID++;
	

}

void geoxl::ARAPDeformationContext::writeLinearSystem()
{
	Timer t;
	const AAT positionAAT = urshape->getAAT("position");
	const mpcard numPts = urshape->getPointSet()->getNumEntries();
	const mpcard dim = numPts*3;

	AAT weightAAT = NULL_AAT;

	bool providesAtanWeights = tg->getEdges()->providesAttribute("weight", 1, VAD::DATA_FORMAT_FLOAT32);
	if (settings->asapUseCotangentWeights && providesAtanWeights) {
		weightAAT = tg->getDescr()->getToken("weight");
		if (!weightChannel)
		{
			delete A_fac;
			A_fac = NULL;
		}
		weightChannel = true;
	}

	if (linSystem_A.getNumRows() != dim) {
		linSystem_A = SparseMatrixF(dim);
	}
	//linSystem_A.setZero();
	// set all entries to zero
	if (!A_fac)
		for( mpcard i=0; i < linSystem_A.getNumRows(); i++ )
		{
			SparseVectorF &v = linSystem_A[i];
			const mpcard numEntries = v.getNumNonZeroEntries();
			for( mpcard j = 0; j < numEntries; j++ )
				v.setListEntryValue(j,0);
		}

		debugOutput << "sysFillInVariables: linSystem_A.setZero() took: " << t.getDeltaValue() <<"\n";

		if (linSystem_b.getDim() != dim) 
			linSystem_b.setDim(dim);
		linSystem_b.setZero();


		if (linSystem_x.getDim() !=dim) {
			linSystem_x.setDim(dim);
			// reset solution only if system changed    
			for (mpcard i=0; i<numPts; i++) 
				for (mpcard d=0; d<3; d++) 
					linSystem_x[3*i+d] = urshape->getPointSet()->get3f(i, positionAAT)[d];
		}
		else{
			debugOutput << "->sysFillInVariables() using last x result to init x\n";
		}

		computePerPointRotations();
		t.getDeltaValue();

		pAssert(linSystem_A.getNumRows() == dim && linSystem_b.getDim() == dim && linSystem_x.getDim() == dim && tg->getNumVertices() == numPts);



		for (mpcard i=0; i<numPts; i++) {
			const mpcard numEdges = tg->getNumAdjacentVertices(i);
			for (mpcard j=0; j<numEdges; j++) {
				PointCloudTopologyGraph::EdgeIndex edgeAttrIndex;
				const mpcard k = tg->getAdjacentVertex(i, j, &edgeAttrIndex);         
				float32 weight = 1;
				if (weightChannel) {
					weight = tg->getEdges()->get1f(edgeAttrIndex, weightAAT);
					if( fabs(weight) < 1e-15f )
						continue;			
				}
				Vector3f target = ((perPointRotations[i]+perPointRotations[k])*0.5)*(urshape->getPointSet()->get3f(i, positionAAT) - urshape->getPointSet()->get3f(k, positionAAT));
				// 				if (settings->asapL1Norm) {
				// 					Vector3f targetCurrent = makeVector3f(linSystem_x[3*i], linSystem_x[3*i+1], linSystem_x[3*i+2])
				// 						- makeVector3f(linSystem_x[3*k], linSystem_x[3*k+1], linSystem_x[3*k+2]);
				// 					float32 res = norm(target-targetCurrent);
				// 					weight /= res+settings->asapL1NormSmoothing;
				// 				}


				for (mpcard d=0; d<3; d++) {
					// update the target vector on the right hand side...
					linSystem_b[3*i+d] += 2*weight*target[d] ;
				}

				if (!A_fac)
					for (mpcard d=0; d<3; d++) {
						// Laplacian matrix
						linSystem_A[3*i+d].addEntryBinary(3*i+d, 2* weight);
						linSystem_A[3*i+d].addEntryBinary(3*k+d,-1 *weight);
						linSystem_A[3*k+d].addEntryBinary(3*i+d,-1 *weight);
					}
			}
		}
}

geoxl::Matrix3f geoxl::ARAPDeformationContext::getSpatialGradient( const ARConstraintAnchor &at )
{
	mpint index = at.topologicalAnchor;
	if (index == ARConstraintAnchor::NO_TOP_ANCHOR) {
#ifdef GEOXL_DEBUG
		warning("ASAPDeformationContext::getPosition() - position is approximate");
#endif
		query->setTargetPoint(at.pos);
		bool found;
		card32 i;
		query->getNextNearestPointIndex(i, found);
		index = i;
		pAssert(found);
	}
	
	return perPointRotations[index];
}


IMPLEMENT_GEOXL_CLASS( ARAPDeformationSolver ,0)
{
	BEGIN_CLASS_INIT( ARAPDeformationSolver );
}
