//---------------------------------------------------------------------------
#ifndef GeneralDeformationH
#define GeneralDeformationH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "RegularData.h"
//---------------------------------------------------------------------------

// DifferentialDeformationConstraints - mw 9/27/2007

namespace geoxl {


/// this class describes an achor point for a deformation/reconstruction constraint
/// this is the point at which the constraint should be enforced
/// a "topological anchor" is also provided to restrict the effect 
/// to a geodesic neighborhood on the shape
class ConstraintAnchor : public Persistent {
	GEOXL_CLASS( ConstraintAnchor )

	public:
	static const mpint NO_TOP_ANCHOR = -1;

	/// spatial position at which a constraint should be enforced
	Vector3f pos;
	/// index of a point in the reference domain (urshape) which is
	/// geodesically close to pos (its neighborhood will be used in
	/// the constraint)
	/// set to NO_TOP_ANCHOR if point is (temporarly) not attached to a specific surface point
	mpint topologicalAnchor;	
};


/// this class describes a deformation constraint in general
/// the only thing common to all constraints is an anchor, so the base class
/// only has a single member
class GeneralConstraint : public Persistent {
	GEOXL_CLASS( GeneralConstraint )
	private:
	/// this is the anchor / position of the constraint
	ConstraintAnchor anchor;
	
	public:

	GeneralConstraint() {}
	
	void setup(
		ConstraintAnchor anchor
	);

	inline const ConstraintAnchor &getAnchor() {return anchor;}	
}; 


// -- Position Constraint(3f). Has an extra toPoint attribute and 
// Covariance Matrix, which penalizes deformations in different directions differently
class PositionConstraint : public GeneralConstraint {
	GEOXL_CLASS( PositionConstraint )

 private:
	Vector3f		toPoint;
	Matrix3f		covariance;
	float32		time;
 
 public:
  
	PositionConstraint() {}
	
	void setup(		
	   ConstraintAnchor anchor,
		Vector3f toPoint,
		Matrix3f covariance,
		float32 time = 0
	);
	
	inline const Vector3f &getToPoint() {return toPoint;}	
	inline const Matrix3f &getCovariance() {return covariance;}	
	inline const float32 &getTime() {return time;}	
};

class VelocityVectorConstraint : public GeneralConstraint {
	GEOXL_CLASS( VelocityVectorConstraint )
 private:
   
	float32		fromTime;
	float32		toTime;
	Vector3f		velocity;
	Matrix3f		covariance;
 
 public:
	VelocityVectorConstraint() {}
	
	void setup(		
		ConstraintAnchor anchor,
		float32 fromTime,
		float32 toTime,
		Vector3f velocity,
		Matrix3f covariance
	);
	
	inline const Vector3f &getVelocity() {return velocity;}	
	inline const Matrix3f &getCovariance() {return covariance;}	
	inline const float32 &getFromTime() {return fromTime;}	
	inline const float32 &getToTime() {return toTime;}	
};


class InstantaneousVelocityConstraint : public GeneralConstraint {
	GEOXL_CLASS( InstantaneousVelocityConstraint )
 private:

	float32		time;
	Vector3f		instVelocity;
	Matrix3f		covariance;
 
 public:
	InstantaneousVelocityConstraint() {}
	
	void setup(		
		ConstraintAnchor anchor,
		float32 time,
		Vector3f instVelocity,
		Matrix3f covariance
	);
	
	inline const Vector3f &getVelocity()	{return instVelocity;}	
	inline const Matrix3f &getCovariance() {return covariance;}	
	inline const float32	&getTime() {return time;}
};



}

#endif