//---------------------------------------------------------------------------
#ifndef ARConstraintsH
#define ARConstraintsH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "RegularData.h"
//---------------------------------------------------------------------------

// DifferentialDeformationConstraints - mw 9/27/2007

namespace geoxl {


/// this class describes an anchor point for a deformation/reconstruction constraint
/// this is the point at which the constraint should be enforced
/// a "topological anchor" is also provided to restrict the effect 
/// to a geodesic neighborhood on the shape
class DEFORMATION_API ARConstraintAnchor : public Persistent {
	GEOXL_CLASS( ARConstraintAnchor )

 public:
	static const mpint NO_TOP_ANCHOR = -1;

	/// spatial position at which a constraint should be enforced
	Vector3f pos;

	/// index of a point in the reference domain (urshape) which is
	/// geodesically close to pos (its neighborhood will be used in
	/// the constraint)
	/// set to NO_TOP_ANCHOR if point is (temporarly) not attached to a specific surface point
	mpint topologicalAnchor;	


	ARConstraintAnchor() {}
	
	ARConstraintAnchor(Vector3f pos, mpint topologicalAnchor = NO_TOP_ANCHOR) {
		this->pos = pos;
		this->topologicalAnchor = topologicalAnchor;
	}

	void setup(Vector3f pos, mpint topologicalAnchor = NO_TOP_ANCHOR) {
		this->pos = pos;
		this->topologicalAnchor = topologicalAnchor;
	}

	mpint getTopologicalAnchor() const {return topologicalAnchor;}
	void setTopologicalAnchor(mpint v) {topologicalAnchor=v;}
	Vector3f getPos() const {return pos;}
	void setPos(Vector3f v) {pos=v;}

};


/// this class describes a deformation constraint in general
/// the only thing common to all constraints is an anchor, so the base class
/// only has a single member
class DEFORMATION_API GeneralARConstraint : public Persistent {
	GEOXL_CLASS( GeneralARConstraint )	
 public:

	/// this is the anchor / position of the constraint
	ARConstraintAnchor anchor;

	enum ConstraintTypesHint {
		POSITION_CONSTRAINT, 
		VELOCITY_CONSTRAINT, 
		INST_VELOCITY_CONSTRAINT, 
		UNKNOWN_CONSTRAINT,
		HALF_SPACE_CONSTRAINT,
		SPATIAL_DERIVATIVE_CONSTRAINT,
		SIMPLE_SPATIAL_DERIVATIVE_CONSTRAINT };
 

 	GeneralARConstraint() {}
	
	void setup(ARConstraintAnchor anchor) {
		this->anchor = anchor;
	}

	inline ARConstraintAnchor &getAnchor() {return anchor;}	
	inline const ARConstraintAnchor &getAnchor() const {return anchor;}	

	virtual ConstraintTypesHint getConstraintTypesHint() const {return UNKNOWN_CONSTRAINT;}
}; 

class VectorValuedARConstraint : public GeneralARConstraint {
	GEOXL_ABSTRACT_CLASS( VectorValuedARConstraint )	
 public:
	virtual Vector3f getConstrVector() const = 0;
	virtual void setConstrVector(const Vector3f &v) = 0;
};


/// PositionARConstraint: Has an extra toPoint attribute and 
/// a covariance matrix, which penalizes deformations in different directions differently
/// in case of static (surface) reconstruction, time might be ignored
class DEFORMATION_API PositionARConstraint : public VectorValuedARConstraint {
	GEOXL_CLASS( PositionARConstraint )

  public:
	Vector3f		toPoint;
	Matrix3f		covariance;
	float32		time;
  
	PositionARConstraint() {}
	
	void setup(		
		ARConstraintAnchor anchor,
		Vector3f toPoint,
		Matrix3f covariance,
		float32 time ) 
	{
		GeneralARConstraint::setup( anchor );
		this->toPoint = toPoint;
		this->covariance = covariance;
		this->time = time;
	}

	inline const Vector3f &getToPoint() const { return toPoint; }
	void setToPoint(const Vector3f &newToPoint) { toPoint = newToPoint; }
	inline const Matrix3f &getCovariance() const {return covariance;}	
	inline const float32   getTime() const {return time;}		

	virtual ConstraintTypesHint getConstraintTypesHint() const {return POSITION_CONSTRAINT;}
	virtual Vector3f getConstrVector() const {return toPoint;}
	virtual void setConstrVector(const Vector3f &v) {toPoint = v;}

};

/// this class describes a constraint that specifies the position differences
/// at two time instances
class VelocityVectorARConstraint : public VectorValuedARConstraint {
	GEOXL_CLASS( VelocityVectorARConstraint )
 public:
	float32		fromTime;
	float32		toTime;
	Vector3f		velocity;
	Matrix3f		covariance;

	VelocityVectorARConstraint() {}
	
	void setup(		
		ARConstraintAnchor anchor,
		float32 fromTime,
		float32 toTime,
		Vector3f velocity,
		Matrix3f covariance
	) {
		this->anchor = anchor;
		this->fromTime = fromTime;
		this->toTime = toTime;
		this->velocity = velocity;
		this->covariance = covariance;
	}
	
	inline const Vector3f &getVelocity() const {return velocity;}	
	inline const Matrix3f &getCovariance() const {return covariance;}	
	inline const float32 getFromTime() const {return fromTime;}	
	inline const float32 getToTime() const {return toTime;}	

	virtual ConstraintTypesHint getConstraintTypesHint() const {return VELOCITY_CONSTRAINT;}
	virtual Vector3f getConstrVector() const {return velocity;}
	virtual void setConstrVector(const Vector3f &v) {velocity = v;}
};


/// this constraint specifies the velocity at a given time
class InstantaneousVelocityARConstraint : public VectorValuedARConstraint {
	GEOXL_CLASS( InstantaneousVelocityARConstraint )
 public:
	float32		time;
	Vector3f		instVelocity;
	Matrix3f		covariance;
 
	InstantaneousVelocityARConstraint() {}
	
	void setup(		
		ARConstraintAnchor anchor,
		float32 time,
		Vector3f instVelocity,
		Matrix3f covariance
	) {
		this->anchor = anchor;
		this->time = time;
		this->instVelocity = instVelocity;
		this->covariance = covariance;
	}
	
	inline const Vector3f &getVelocity() const {return instVelocity;}	
	inline const Matrix3f &getCovariance() const {return covariance;}	
	inline const float32	&getTime() const {return time;}

	virtual ConstraintTypesHint getConstraintTypesHint() const {return INST_VELOCITY_CONSTRAINT;}
	virtual Vector3f getConstrVector() const {return instVelocity;}
	virtual void setConstrVector(const Vector3f &v) {instVelocity = v;}
};

/// a point should be located within a half space
class ARHalfSpaceConstraint : public VectorValuedARConstraint {
	GEOXL_CLASS( ARHalfSpaceConstraint )
 public:
   /// time at which the constraint is active
   float32 time;
   /// footpoint of the plane that separates the half spaces
   Vector3f halfSpaceFootPoint;
   /// normal of the plane that separates the half spaces
   /// the normal is oriented outward, into the forbidden region (walk in normal direction --> increased penalty)
   Vector3f halfSpaceNormal;
   /// quadratic weight that is applied when entering the half space: 
   /// penalty = w*[(x-fp)*n]^2, if (x-fp)*n>0, zero otherwise
   float32 penaltyWeight;

	void setup(		
		ARConstraintAnchor anchor,
      float32 time,
      Vector3f halfSpaceFootPoint,
      Vector3f halfSpaceNormal,
      float32 penaltyWeight)
   {
		this->anchor = anchor;
		this->time = time;
		this->halfSpaceFootPoint = halfSpaceFootPoint;
		this->halfSpaceNormal = halfSpaceNormal;
		this->penaltyWeight = penaltyWeight;
	}

	virtual ConstraintTypesHint getConstraintTypesHint() const { return HALF_SPACE_CONSTRAINT; }
	float32 getTime() const {return time;}
	void setTime(float32 v) {time=v;}
	Vector3f getHalfSpaceFootPoint() const {return halfSpaceFootPoint;}
	void setHalfSpaceFootPoint(Vector3f v) {halfSpaceFootPoint=v;}
	Vector3f getHalfSpaceNormal() const {return halfSpaceNormal;}
	void setHalfSpaceNormal(Vector3f v) {halfSpaceNormal=v;}
	float32 getPenaltyWeight() const {return penaltyWeight;}
	void setPenaltyWeight(float32 v) {penaltyWeight=v;}
   
	Vector3f getConstrVector() const { return makeVector3f( -1,-1,-1 ); };
	void setConstrVector(const Vector3f &v) {};
 
};


class MatrixValuedARConstraint : public GeneralARConstraint {
	GEOXL_ABSTRACT_CLASS( MatrixValuedARConstraint )	
 public:
	virtual Matrix3f getConstrMatrix() const = 0;
	virtual void setConstrMatrix(const Matrix3f &m) = 0;
};


/// Prescribe deformation gradients
class DEFORMATION_API SpatialDerivativeARConstraint : public MatrixValuedARConstraint {
	GEOXL_CLASS( SpatialDerivativeARConstraint )

  public:
	Matrix3f		               targetDerivative;
	StaticMatrix<float32, 9,9>	covariance;
	float32		               time;
  
	SpatialDerivativeARConstraint() {}
	
	void setup(		
		ARConstraintAnchor anchor,
		Matrix3f targetDerivative,
		StaticMatrix<float32,9,9> covariance,
		float32 time ) 
	{
		GeneralARConstraint::setup( anchor );
		this->targetDerivative= targetDerivative;
		this->covariance = covariance;
		this->time = time;
	}

	inline const Matrix3f &getTargetDerivative() const { return targetDerivative; }	
	inline const StaticMatrix<float32, 9,9> &getCovariance() const {return covariance;}	
	inline const float32   getTime() const {return time;}		

	virtual ConstraintTypesHint getConstraintTypesHint() const {return SPATIAL_DERIVATIVE_CONSTRAINT;}
	virtual Matrix3f getConstrMatrix() const {return targetDerivative;}
	virtual void setConstrMatrix(const Matrix3f &m) {targetDerivative = m;}

};

/// Prescribe deformation gradients (simplified)
class DEFORMATION_API SimpleSpatialDerivativeARConstraint : public MatrixValuedARConstraint {
	GEOXL_CLASS( SimpleSpatialDerivativeARConstraint )

  public:
   /// dx_f(), dy_f(), dz_f() in columns
	Matrix3f	targetDerivative;
   Vector3f columnWeights;
	float32	time;
  
	SimpleSpatialDerivativeARConstraint() {}
	
	void setup(		
		ARConstraintAnchor anchor,
		Matrix3f targetDerivative,
		Vector3f columnWeights,
		float32 time ) 
	{
		GeneralARConstraint::setup( anchor );
		this->targetDerivative= targetDerivative;
		this->columnWeights = columnWeights;
		this->time = time;
	}

	inline const Matrix3f &getTargetDerivative() const { return targetDerivative; }	
	inline const Vector3f &getColumnWeights() const {return columnWeights;}	
	inline const float32   getTime() const {return time;}		

	virtual ConstraintTypesHint getConstraintTypesHint() const {return SIMPLE_SPATIAL_DERIVATIVE_CONSTRAINT;}
	virtual Matrix3f getConstrMatrix() const {return targetDerivative;}
	virtual void setConstrMatrix(const Matrix3f &m) {targetDerivative = m;}

};


}

#endif