//---------------------------------------------------------------------------
#ifndef ARSettingsH
#define ARSettingsH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl {

class LocalICP;

const unsigned int SHAPE_SOLVER_STEEPEST_DESCENT = 0;	// Steepest descent
const unsigned int SHAPE_SOLVER_NEWTON = 1;				// Newton's method
const unsigned int SHAPE_SOLVER_DIRECT = 2;				// Solve local 3x3 systems

const unsigned int CONSISTENCY_MODE_SINGLE_NEIGHBOR = 0;
const unsigned int CONSISTENCY_MODE_INTEGRATION = 1;
const unsigned int CONSISTENCY_MODE_NONE = 2;
const unsigned int CONSISTENCY_MODE_NORMALS_PLANE = 3;

const card32 TS_MODEL_LAPLACE = 0;
const card32 TS_MODEL_THINPLATESPLINES = 1;
const card32 TS_MODEL_MIXED_MODE = 2;
const card32 TS_MODEL_THINPLATESPLINES_2D = 3;

const card32 TSS_COSINE = 0;
const card32 TSS_GAUSSIAN = 1;
const card32 TSS_POLYSPLINE = 2;

const card32 ASAP_LIN_SOLVER_CG = 0;
const card32 ASAP_LIN_SOLVER_GS = 1;
const card32 ASAP_LIN_SOLVER_DPCG = 2;
const card32 ASAP_LIN_SOLVER_CGS = 3;
const card32 ASAP_LIN_SOLVER_DPCGS = 4;
const card32 ASAP_LIN_SOLVER_DIRECT_INTEL = 5;

///
///   settings for animation reconstruction pipeline
///
class DEFORMATION_API ARSettings : public Persistent {
	GEOXL_CLASS( ARSettings )
 public:
   // --- Topology computation parameters
	float32 topSurfelSpacing;				// Spacing in surfel graph (when creating)
	float32 topResamplingSurfelSpacing;	// Spacing in surfel graph (when deleting superflous points)
   float32 topRelRadius;               // maximum distance at which topological connections are made, relative to topSurfelSpacing (multiply the two constants)
   card32 topMaxNumNeighbors;          // maximum number of kNN used for building the topology (per point, might be larger overall due to graph symmetry [undirectioned])
	float32 topMaxRelAddLengthChange;   // maximum additive length change of topology edges (relative to topSurfelSpacing)
	float32 topMaxMultLengthChange;     // maximum multiplicative length change of topology edges (relative to topSurfelSpacing)
	float32 topMinRelLengthRelCheck;    // minimum maximum length of an edge to apply the relative check (again, measures relative to topSurfelSpacing)
	float32 topCutMaxStressFactor;      // remove vertices if largest eigenvalue in stress tensor exceeds this value
	float32 topCutMinRelativeLength;      // remove vertices if smallest eigenvalue in stress tensor falls below this value
	card32 topCutBorder;                // boundary around deleted vertices that are also removed
	bool topCheckWholeAnimation;        // go through the whole animation in order to check topology?
   bool resampleInitialUrshapes;       // should the initial urshapes be resampled or taken as-is?
   bool resampleMergedUrshapes;        // should the urshapes be resampled after merging?
	card32 outlier_minNumberOfPoints;	// remove patches of topological connected points with less than minNumberOfPoints points
	bool outlier_removeSmallPatches;		// activates outlier (patches) removal
   
	// --- Numerics
	card32 maxNumICPIterations;
	card32 maxNumShapeOptIterations;
	card32 numNewtonIterations;
	card32 numCGIterations;
	float32 convThreshICP;
	float32 convThreshNewton;
	float32 convThreshCG;
	float32 newtonRegularization;
	bool d2aLinesearch;
	static const card32 MMS_BFGS = 0;
	static const card32 MMS_NON_LINEAR_CG = 1;	
	static const card32 MMS_LBFGS = 2;	
	card32 mmSolverMethod;

	// --- Reconstruction parameters
	float32 pjDeprecated01;					// Deprecared
	float32 pjDeprecated02;					// Deprecared
	card32  pjDeprecated03;					// Deprecared
	float32 pjDeprecated04;					// Deprecated

	// --- Shape solver parameters
	float32 weightSmoothNormal;				// Normals of adjacent pathes match
	float32 weightSmoothPlane;				// Adjacent point lie in same plane
	float32 weightSmoothLaplacian;			// Laplacian potential
	card32 shapeSolveConsistencyMethod;		// CONSISTENCY_MODE_SINGLE_NEIGHBOR vs. CONSISTENCY_MODE_INTEGRATION
	card32 shapeSolveSolver;				// Steepest descent vs. Newton method
	float32 shapeSolveStepSize;				// Step size with steepest descent method
	card32 shapeSolveMaxNumIterations;		// #Newton iterations
	float32 shapeSolveConvergence;			// Convergence criterion
	bool shapeSolveLocallyWeightedICP;		// Locally weighted ICP constraint (see, shape reconstruction pdf)

	// --- MMSolver parameters
	float32 nodalRadius;
	float32 nodalOffset;
	
	float32 volumeParam;					// volume preservation
	float32 rigidityParam;					// rigidity
	float32 correspondenceParam;			// correspondences
	float32 accelerationParam;				// temporal acceleration between frames
	float32 velocityParam;					// penalize velocity
	float32 halfSpaceParam;					// penalize halfspace penetrations
	card32 windowSize;
	
	card32 gridSize; // size for grid sampler 
	float32 topReach; // size for grid sampler 
	
	const MetaClass *mmNodesSampler;

	bool nearestBaseIndexHack; // if set to true, we will look for the nearest base point and use its nodal neighbors
	                           // if we want to deform a point which doesn't have a proper basepoint.

   // --- ICP constraints parameters
   card32 icpConstraintsMode;          // point-to-point/plane
   bool icpPercentileCulling;          // enable distance percentile culling in long-dinstance icp
	float32 icpOutlierPercentile;       // percentile of farest data points to throw away in ICP matching
	bool icpNormalCulling;              // enable normal culling in long-distance icp
	float32 icpNormalCosAngleCutOff;    // maximum cosine normal/tovector angle
	bool icpNearFieldCulling;           // enable distance culling in close-distance icp
	float32 icpNearFieldRelativeCutOff; // distance (x sample spacing) at which ICP constraints are cut-off in final optimization
   float32 icpSkipPercentage;          // randomly skip points to speed up computations (typically a debugging option)
   bool icpUseDataErrorQuads;          // try to use data error matrices if present
   float32 icpRigidityInitialFactor;   // we can start at a higher rigidity level in registration when this is turned on (use with the next two params)
   float32 icpRigidityMultiPassFactor; // if != 1, perform a second icp registration pass with different (typ. reduced) rigidity weight
   card32 icpNumRPasses;               // number of icp passes with different rigidity
   bool icpTwoWayMatch;                // if false, only A-->B matches are performed

	// --- Patch ICP contraints parameters
	LocalICP * patchICP_localICP;			// settings object for patch icp
	bool patchICP_active;
	float32 patchICP_weight;
	float32 patchICP_maxDist;				// urshape resampling sample spacing

	// --- Residual parameters
	float32 residualSupportSigma;			// sigma = residualSupportSigma * topSurfelSpacing;
													// residual is computed by averaging signed distances weighted by exp(-x/sigma)
	float32 residualUpperLimit;
	float32 residualLowerLimit;
	float32 residualRefittingRigidity;

	// --- Unsupported Points Culling
	card32  cullStartNumFrames;			// number of frames to gather before parts that are only weakly supported by data are thrown away
	card32  cullBelowNumFrames;			// minimum number of frames required to be kept
	float32 cullBelowPercentage;        // minimum percentage required to be kept (or'ed with previous criterion)
	//card32 cullNumConsecutiveFrames;  // if the data point is supported by this number of consecutive frames, it is kept no matter what

   // --- Final output parameters
   card32 outFramesPerFrame;           // number of frames created per frame (>1 means in-between frames)
   card32 outPointsPerPoint;           // number of points created per point (upsampling the geometry)

   // --- debugging
   bool mmDebugOutput;                 // enable this to output extended debug info for mmsolver
   bool mergeDebugOutput;              // enable this to enable debug outputs for ARMergeSequence
   bool surfaceRecDebugOutput;				// enable this to enable debug outputs for surface reconstruction
   bool singleStepDebugVis;
   card32 dataToUsNumIterations;
   bool debugOutputStressTensors;      // enable output of stress tensors in topology check

   // --- Constraints OOC settings
   bool   useConstraintsOOC;
   card32 maxConstraintsInMem;
   string constraintsTempFileDir;

 private:
	void initDefaultValues();
	
 public:
	ARSettings();

	void setPresets();
	virtual void assign(const Object* obj, CopyContext *context = NULL);

	~ARSettings();
};

}

#endif
