//---------------------------------------------------------------------------
#ifndef ICPConstraintsH
#define ICPConstraintsH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
#include "Persistent.h"
#include "ARConstraintsInterface.h"
#include "AROptToolWithFeedback.h"
#include "Delegate.h"
//---------------------------------------------------------------------------

namespace geoxl {

class ARSettings;
class PointCloud;
class ARConstraintsInterface;
class PointSetANNQuery;
class PrecomputedDistanceField;

const unsigned int ICP_CONSTRAINT_POINT_TO_POINT = 0;
const unsigned int ICP_CONSTRAINT_POINT_TO_PLANE = 1;

const unsigned int ICP_OUTLIERS_NONE = 0;
//const unsigned int ICP_OUTLIERS_NORMALS = 1; ---> into seperate flag to combine normals & distance
const unsigned int ICP_OUTLIERS_DISTANCE = 1;
const unsigned int ICP_OUTLIERS_PERCENTILE = 2;

///
///   this class builds ICP-like constraints from data points
///   
///
class DEFORMATION_API ICPConstraints : public Persistent, public AROptToolWithFeedback
{
   GEOXL_CLASS(ICPConstraints)

 private:
	card32 constraintMode;		// Point2Point vs. Point2Plane
	card32 distanceOutlierMode;			// _NONE, _DISTANCE, _PERCENTILE
	bool cullNormalOutliers;

	// _NORMAL: times sampling spacing
	// e.g.: 0.7: pruned if fabs( normal * ( data - surfel ).normalize ) < 0.7
	float32 outliersNormals;	
	
	// _DISTANCE: times sampling spacing
	// points within this distances are NEVER CULLED
	// e.g.: 3.0: pruned if dist( data - surfel ) > 3.0f * samplingSpacing
	float32 outliersDistanceFactor;	
	
	// _PERCENTILE: percentage of furthest point pruned; 
	// e.g.: 0.3 -> 30% furthest points pruned
	float32 outliersPercentile;	
   
   /// will be multiplied with constraint quadrics, default = 1
   float32 constraintsWeight;
   
   struct VisInfo {
      Vector3f first, second;
      float32 res;
      VisInfo(Vector3f f, Vector3f s, float32 r) {first = f; second = s; res = r;}
   };
	
	vector < VisInfo > corrVis;

   card32 statistics_numOutliers;
   float64 statistics_sumOfSquaredNormalDistances;

 public:
	ICPConstraints();
	~ICPConstraints();

	virtual void assign(const Object* obj, CopyContext *context = NULL);

	virtual void glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings);

	/// builds constraints that move points from referenceShape
	/// to match matchingShape
	void buildICPConstraints(
		const ARSettings *settings,
		ARConstraintsInterface *receiver,
		/// parametrization (urshape)
		PointCloud *referenceShape,
		/// urshape as in its current form ("best guess"), used for correspondences
		PointCloud *deformedReference,
		/// data points
		PointCloud *matchingShape,
		/// the constraints are specified for this time
		const float32 time,
		/// redraw scene, calling 
		Delegate callForRepaint = NULL,
		/// show correspondences
		bool showCorrVis = false,
		/// if this is more than 0, the corresponding amount of constraint points is skipped using a fixed random pattern (srand(42))
		const float32 skipPercentage = 0,
		/// if this option is set, the channel "numDataPts" in the reference shape is incremented for each corresponding point found
		/// the channel must exist; if it does not exist, the option is ignored
		bool countCorrespondences = false,
		/// use debug renderer to visualize constraints
		bool useDebugRenderer=false
	);
   
   /// builds reverse constraints: deformed us is attracted to data (nearest neighbor for each def us point, not the other way round)
	void buildReverseICPConstraints(
		const ARSettings *settings,
		ARConstraintsInterface *receiver,
		/// parametrization (urshape)
		PointCloud *referenceShape,
		/// urshape as in its current form ("best guess"), used for correspondences
		PointCloud *deformedReference,
		/// data points
		PointCloud *matchingShape,
		/// the constraints are specified for this time
		const float32 time,
		/// if this is more than 0, the corresponding amount of constraint points is skipped using a fixed random pattern (srand(42))
		const float32 skipPercentage = 0,
		/// use debug renderer to visualize constraints
		bool useDebugRenderer=false
	);
	
	/// builds reverse constraints: deformed us is attracted to data (nearest neighbor for each def us point, not the other way round)
	void buildConditionalReverseICPConstraints(
		const ARSettings *settings,
		ARConstraintsInterface *receiver,
		/// parametrization (urshape)
		PointCloud *referenceShape,
		/// urshape as in its current form ("best guess"), used for correspondences
		PointCloud *deformedReference,
		/// data points
		PointCloud *matchingShape,
		/// cached query for matchingShape
		const PointSetANNQuery* annQuery,
		/// the constraints are specified for this time
		const float32 time,
		/// if this is more than 0, the corresponding amount of constraint points is skipped using a fixed random pattern (srand(42))
		const float32 skipPercentage = 0,
		/// use debug renderer to visualize constraints
		bool useDebugRenderer=false,
		/// use precomputed distance field (ignores ANN querry)
		PrecomputedDistanceField * df = NULL
		);




	/// strongly simplified point-to-point version for debugging
	void buildSimpleICPConstraints(
		const ARSettings *settings,
		ARConstraintsInterface *receiver,
		/// parametrization (urshape)
		PointCloud *referenceShape,
		/// urshape as in its current form ("best guess"), used for correspondences
		PointCloud *deformedReference,
		/// data points
		PointCloud *matchingShape,
		/// the constraints are specified for this time
		const float32 time,
		/// redraw scene, calling 
		Delegate callForRepaint = NULL,
		/// show correspondences
		bool showCorrVis = false
	);

	/// debugging only - builds 1:1 constraints by index		
	void buildPerfectConstraints(
		const ARSettings *settings,
		ARConstraintsInterface *receiver,
		/// parametrization (urshape)
		PointCloud *referenceShape,
		/// urshape as in its current form ("best guess"), used for correspondences
		PointCloud *deformedReference,
		/// data points
		PointCloud *matchingShape,
		/// the constraints are specified for this time
		const float32 time
	);

	card32 getConstraintMode() const { return constraintMode; }
	void setConstraintMode(card32 v) { constraintMode = v; }

	card32 getDistanceOutliersMode() const {return distanceOutlierMode;}
	void setDistanceOutliersMode(card32 v) {distanceOutlierMode=v;}

	bool getCullNormalOutliers() const {return cullNormalOutliers;}
   void setCullNormalOutliers(bool v) {cullNormalOutliers=v;}
	
	float32 getOutliersDistanceFactor() const {return outliersDistanceFactor;}
	void setOutliersDistanceFactor( float32 d ) { outliersDistanceFactor = d; }
	
	float32 getOutlierNormals() const { return outliersNormals;}
	void setOutlierNormals(float32 v) {outliersNormals=v;}
	
	float32 getOutliersPercentile() const {return outliersPercentile;}
	void setOutliersPercentile(float32 v) {outliersPercentile=v;}		

   float32 getConstraintsWeight() const {return constraintsWeight;}
   void setConstraintsWeight(float32 w) {constraintsWeight = w;}

   void clearStatistics();
   card32 statisticsGetNumOutliers() {return statistics_numOutliers;}
   float64 statisticsGetSumOfSquaredNormalDistances() {return statistics_sumOfSquaredNormalDistances;}

};

}

#endif