//---------------------------------------------------------------------------
#ifndef MMDeformationContextH
#define MMDeformationContextH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "ARDeformationOptimizationContext.h"
//---------------------------------------------------------------------------
#include "MMDeformations.h"
#include "MMNodes.h"
#include "MMConstraints.h"
#include "UnstructuredInCorePointCloud.h"
#include <hash_map>

namespace geoxl {

///
///   this class stores all the data to describe a deformation as computed by the FDDeformationSolver/Evaluator
///
class DEFORMATION_API MMDeformationContext : public ARDeformationOptimizationContext {
	GEOXL_CLASS( MMDeformationContext )
 private:
    // only for getBaseResolution()
    float32 baseResolution;
	 // cached neighbors and shape functions
	 stdext::hash_map<mpint, MMNeighborContainer*> cached_neighbors;

	 std::vector<MMNeighborContainer*> no_id_neighbors;

	 // original geometry
	 // PointCloud *urshape;
     
	 // original nodes which should be sampled from geometry
	 MMNodes *nodes;
	 // deformations for the nodes for each frame (unknowns we're solving for)
	 vector<MMDeformations *> deformations;
	 
	 vector<MMDeformations *> ddeformations;
	 
	 
	 //vector<bool> framesActive;
	 
	 double alpha;

	 // constraints for each frame
	 vector<MMConstraints*> constraints;
	 
 	 //std::vector<ARHalfSpaceConstraint*> halfSpaceConstraints;

	 // the real number of unknowns (some nodes might be inactive during some
	 // frames.
	 card32 realNbUnknowns;

	 bool haveTopology;

	 bool nearestBaseIndexHack;

	 bool identity;

    IMPLEMENT_VAR_ARRAY_METHODS_STL(deformations)
	//IMPLEMENT_VAR_ARRAY_METHODS_STL(ddeformations)


	 MMDeformations * createDeformationStorageObject(PointCloud *pc, const ARSettings *settings);

	 void checkAndAddTopology(UnstructuredInCorePointCloud * pc);

	 void getNodes(const UnstructuredInCorePointCloud *pc, const float radius, std::vector<Vector3f> &nodes);

	 void setRealNbUnknowns(card32 nb) { realNbUnknowns = nb; }

 public:
	MMDeformationContext();
	MMDeformationContext(MMDeformationContext &copy);

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

	void reinitialize(MMDeformationContext &copy);
	
	void createDDeformations();

	virtual void initializeFromShape(const SceneObject *obj, Vector2f timeInterval, const ARSettings *settings);
	
	virtual float32 getDuration();
	virtual void visualize(Scene * scene);

	inline MMNodes* getNodes() { return nodes; }
	inline mpcard nbDeformations() { return deformations.size(); }
	inline mpcard nbFrames() { return deformations.size(); }
	inline mpcard nbConstraints() { return constraints.size(); }
	inline card32 nbNodes() { return getNodes()->nbNodes(); }
	inline card32 getOffset(card32 index) {
		return nbNodes()*index;
	}
	inline MMDeformations* getDeformations(card32 index) { if( index >= deformations.size() )
		throw PException("MMDeformations::getDeformations() - index out of range!");return deformations[index]; }
	inline MMDeformations* getDDeformations(card32 index) { return ddeformations[index]; }
	inline MMConstraints* getConstraints(card32 index) { return constraints[index]; }

	//inline void addHalfSpaceConstraint(ARHalfSpaceConstraint *c) { halfSpaceConstraints.push_back(c); }
	//inline card32 nbHalfSpaceConstraints() const { return halfSpaceConstraints.size(); }
	//inline ARHalfSpaceConstraint* getHalfSpaceConstraint(const card32 index) { return halfSpaceConstraints[index]; }
	inline bool isIdentity() const { return identity; }
	inline void setIdentity() { identity = true; }

	void clearConstraints();

	void getDeformationAndGradient(const Vector3f &pos, mpint id, Vector3f &deformation, Matrix3f &gradient, bool computeGradient=false, card32 frame=0);
	Vector3f getDeformation(const Vector3f &pos, mpint id, card32 frame=0);
	Vector3f getDeformation(const Vector3f &pos, MMNeighborContainer &neighbors, card32 frame=0);
	Matrix3f getDeformationGradient(const Vector3f &pos, const Vector3f &myDeformation, MMNeighborContainer &neighbors, card32 frame=0);

	// Get the deformation for node index in frame frame.
	Vector3f getDeformation(card32 index, card32 frame=0);
	// Get a point cloud containing the displaced  nodes for frame frame.
	UnstructuredInCorePointCloud* getDisplacedNodes(card32 frame);
	
	// Get the nodal neighbors and their shape values for a certain
	// anchor id. If no neighbors exist yet, the position pos is used to compute them.
	MMNeighborContainer* getNeighbors(mpint id, const Vector3f &pos, bool &owner);


	// compute the nodes real deformations (for all frames)
   void computeRealDeformations();

   // set a certain deformation for a node in a particular frame to be active
   // or not.
   void setActive(const card32 frame, const card32 node, const bool active);
    // Compute the indices of the deformation unknowns.
    // Because some might be inactive, the unknowns are not just
    // indexed in increasing order
	void computeRealIndices();
	
	void setWindow(card32 startFrame, card32 endFrame);
	
	//bool getFrameActive(card32 i);
	
	// This gives you the real number of unknowns as computed
	// above (only the active nodal deformations are counted)
	// Note that in 3D, you will have 3*getRealNbUnknowns()
	// scalar unknowns to solve for, not 1*getRealNbUnknowns().
	card32 getRealNbUnknowns() const { return realNbUnknowns; }
	
	void setAlpha(double alpha){
		this->alpha = alpha;
	}
	
	double getAlpha(){
		return this->alpha;
	}

	virtual void writeNodesVisToSceneGraph(Scene *scene, string path, string subgraphName);

	virtual DeformationSolver *createSolver();
	virtual DeformationEvaluator *createEvaluator();
	virtual float32 getBaseResolution();

   ~MMDeformationContext();
};

}

#endif