#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "LocalICP.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "HierarchicalKNNIterator.h"
#include "DebugRenderer.h"
#include "MathTools.h"
#include "PointCloudTools.h"
#include "ObjectViewsTable.h"
#include <externalIncludes\opencv\cxcore.h>
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( LocalICP ,0)
{
	BEGIN_CLASS_INIT( LocalICP );
	ADD_FLOAT32_PROP(m_Sigma,0);
	SET_USER_EDIT(false);
	ADD_CARD32_PROP(m_MaxNumIterations,0);
	ADD_FLOAT32_PROP(m_MaxResidual,0);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_ConvergeDistance,0);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_Slippage_Spacing,0);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_OutlierDistance,0);
	ADD_FLOAT32_PROP(m_Slippage_MinFirstEigen,0);
	SET_USER_EDIT(false);
}

LocalICP::LocalICP(void)
{
	m_Sigma = 0.03f;
	m_MaxNumIterations = 10;
	m_MaxResidual = 0.1f;
	m_ConvergeDistance = 0.01f;
	m_Slippage_Spacing = 0.05f;
	m_OutlierDistance = 3.0f;
	m_Slippage_MinFirstEigen = 0.01f;

	m_Reference = NULL;
	m_Destination = NULL;
	m_hIt_ref = NULL;
	m_hIt_dest = NULL;

	m_RedrawMethod = NULL;
}

LocalICP::~LocalICP(void)
{
	delete m_hIt_ref;
	delete m_hIt_dest;
}

void LocalICP::assign(const Object* obj, CopyContext *context)
{
	const LocalICP * o = dynamic_cast<const LocalICP*>(obj);
	pAssert( o != NULL );

	copyObjectProperties(obj,this);
}

void geoxl::LocalICP::beginICP( PointCloud * reference, PointCloud * destination )
{
	m_Reference = reference;
	m_Destination = destination;
	m_hIt_ref = new HierarchicalKNNIterator(reference,32,NULL);
	m_hIt_dest = new HierarchicalKNNIterator(destination,32,NULL);
	m_Ref_PosAAT = reference->getAAT("position");
	m_Ref_NormalAAT = reference->getAAT("normal");
	m_Dest_PosAAT = destination->getAAT("position");
	m_Dest_NormalAAT = destination->getAAT("normal");
	m_MedPointDistance = getMedianPointDistance(reference);
}


bool geoxl::LocalICP::computeICP( const Vector3f &refPoint, Vector3f &destPoint, Matrix3f &rotation )
{
	debugRenderer->addPoint(refPoint, makeVector3f(0,1,0));
	debugRenderer->addPoint(destPoint, makeVector3f(0,1,0));

	// collect surfels in reference
	vector<Surfel> refSurfels;
	float summed_weight = 0;
	m_hIt_ref->setSeekPointAndReset(refPoint);
	float32 radius_2 = pow(m_Sigma * 2.0f,2.0f);
	float32 inv_sigma_2 = 1.0f / pow(m_Sigma,2.0f);
	while( !m_hIt_ref->atEnd() )
	{
		Vector3f pos = m_hIt_ref->get3f(m_Ref_PosAAT);
		float dist_2 = normQuad(pos - refPoint);
		if( dist_2 > radius_2 )
			break;

		Surfel surfel;
		surfel.pos = pos;
		surfel.normal = m_hIt_ref->get3f(m_Ref_NormalAAT);
		surfel.weight = 1.0f;//exp( - dist_2 * inv_sigma_2 );
		refSurfels.push_back(surfel);
		summed_weight += surfel.weight;

		m_hIt_ref->next();
	}

	if( refSurfels.size() < 10 )
		return false;

	float residual = 0;
	float movement = 0;
	vector<Vector3f> destPoints(refSurfels.size());
	for( unsigned pass=0;pass<m_MaxNumIterations;pass++)
	{
		residual = 0;
		summed_weight = 0;
		Vector3f centroid2 = NULL_VECTOR3F;
		Vector3f centroid1 = NULL_VECTOR3F;

		// calculate corresponding points
		for( unsigned i=0;i<refSurfels.size();i++)
		{
			// transform surfel by current transformation
			Surfel surfelTransformed;
			surfelTransformed.pos = rotation * (refSurfels[i].pos - refPoint)  + destPoint;
			surfelTransformed.normal = rotation * refSurfels[i].normal;

			// seek closest point
			m_hIt_dest->setSeekPointAndReset(surfelTransformed.pos);
			Vector3f closestPoint = m_hIt_dest->get3f(m_Dest_PosAAT);
			Vector3f closestPointNormal =  m_hIt_dest->get3f(m_Dest_NormalAAT);
			refSurfels[i].dynamicWeight = pow(closestPointNormal * surfelTransformed.normal,2.0f) * 
				refSurfels[i].weight;

			// project onto surfel
			Vector3f projectedPoint = surfelTransformed.pos - closestPointNormal * 
				(closestPointNormal * (surfelTransformed.pos - closestPoint));
			float projectedDistance = norm(surfelTransformed.pos - projectedPoint);
			destPoints[i] = projectedPoint;
			refSurfels[i].dynamicWeight *= (projectedDistance > m_OutlierDistance*m_MedPointDistance) ? 0.0f:1.0f;

			// current residual
			residual += projectedDistance * refSurfels[i].dynamicWeight;

			// centroids
			centroid1 += refSurfels[i].pos * refSurfels[i].dynamicWeight;
			centroid2 += projectedPoint * refSurfels[i].dynamicWeight;
			summed_weight +=  refSurfels[i].dynamicWeight;

			Vector3f d = makeVector3f(1,0,0) * ((pass+0.0f) * m_Sigma * 3.0f);

		}

		if( summed_weight < 1e-20f )
			return false;

		centroid1 /= summed_weight;
		centroid2 /= summed_weight;
		residual /= summed_weight;


		// estimate new transformation
		Matrix3f crossCovariance = IDENTITY3F * 0;
		for( unsigned i=0;i<refSurfels.size();i++ )
		{
			crossCovariance += outerProduct(refSurfels[i].pos-centroid1,destPoints[i]-centroid2) * refSurfels[i].dynamicWeight;
		}

		// Prepare OpenCV Datasructure CvMat
		Matrix3f U,V;
		Vector3f eigenValues;
		CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
		cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, crossCovariance.data());
		cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
		cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
		cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

		// Do SVD
		cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

		rotation = U*V;
		Vector3f oldPoint = destPoint;
		destPoint = rotation * (refPoint-centroid1) + centroid2;
		movement = norm(destPoint - oldPoint);
		residual /= m_Sigma;
		movement /= m_Sigma;
		for( unsigned j=0;j<refSurfels.size();j++ )
		{
			Surfel s = refSurfels[j];
			Vector3f p = rotation * (s.pos-centroid1) + centroid2 + makeVector3f(1,0,0) * ((pass+1.0f) * m_Sigma * 3.0f);
			//debugRenderer->addPoint(p,makeVector3f(1,0,0));
		}
		debugRenderer->addLine( destPoint,oldPoint,makeVector3f(1,1,0),makeVector3f(1,0,0),3 );
		if( movement < m_ConvergeDistance )
		{
			//debugOutput << "converged\n";
			return residual < m_MaxResidual;
		}

		//debugOutput << " residual: " << residual << 
		//	movement << " movement: " << movement << "\n";
		

	}
	debugOutput << "finished\n";
	return false;
}

float geoxl::LocalICP::computeICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat )
{
	float residual = 0;
	float lastResidual = 0;
	float movement = 0;
	const mpcard numSurfels = surfels->getNumEntries();
	AAT posAAT = surfels->getAAT("position");
	AAT normalAAT = surfels->getAAT("normal");
	vector<Vector3f> destPoints(numSurfels);
	vector<float> surfelWeights(numSurfels);
	for( unsigned pass=0;pass<m_MaxNumIterations;pass++)
	{
		lastResidual = residual;
		Matrix4f refToDest = invertFrame(destMat) * refMat;
		Matrix3f rotation = shrink4To3(refToDest);
		Vector3f translation = shrink4To3(refToDest[3]);

		residual = 0;
		float summed_weight = 0;
		Vector3f centroid2 = NULL_VECTOR3F;
		Vector3f centroid1 = NULL_VECTOR3F;

		// calculate corresponding points
		for( unsigned i=0;i<numSurfels;i++)
		{
			Vector3f spos = surfels->get3f(i,posAAT);
			Vector3f snormal = surfels->get3f(i,normalAAT);

			// transform surfel by current transformation
			Surfel surfelTransformed;
			surfelTransformed.pos = rotation * spos  + translation;
			surfelTransformed.normal = rotation * snormal;

			// seek closest point
			m_hIt_dest->setSeekPointAndReset(surfelTransformed.pos);
			Vector3f closestPoint = m_hIt_dest->get3f(m_Dest_PosAAT);
			Vector3f closestPointNormal =  m_hIt_dest->get3f(m_Dest_NormalAAT);
			surfelWeights[i] = 1.0f;//pow(closestPointNormal * surfelTransformed.normal,2.0f);

			// project onto surfel
			Vector3f projectedPoint = surfelTransformed.pos - closestPointNormal * 
				(closestPointNormal * (surfelTransformed.pos - closestPoint));
			float projectedDistance = norm(surfelTransformed.pos - projectedPoint);
			destPoints[i] = projectedPoint;
			surfelWeights[i] *= (projectedDistance > m_OutlierDistance*m_MedPointDistance) ? 0.0f:1.0f;

			// current residual
			residual += projectedDistance * surfelWeights[i];

			// centroids
			centroid1 += spos * surfelWeights[i];
			centroid2 += projectedPoint * surfelWeights[i];
			summed_weight +=  surfelWeights[i];

			Vector3f d = makeVector3f(1,0,0) * ((pass+0.0f) * m_Sigma * 3.0f);

			//debugRenderer->addLine(transformVector3f(refMat,spos),transformVector3f(destMat,projectedPoint));

		}

		if( summed_weight < 1e-20f )
			return 1e20f;

		centroid1 /= summed_weight;
		centroid2 /= summed_weight;
		residual /= summed_weight;

		// transform centroids to world space
		centroid1 = transformVector3f(refMat, centroid1);
		centroid2 = transformVector3f(destMat, centroid2);


		// estimate new transformation
		Matrix3f crossCovariance = IDENTITY3F * 0;
		for( unsigned i=0;i<numSurfels;i++ )
		{
			Vector3f x1 = transformVector3f(refMat, surfels->get3f(i,posAAT));
			Vector3f x2 = transformVector3f(destMat, destPoints[i]);
			crossCovariance += outerProduct(x1-centroid1,x2-centroid2) * surfelWeights[i];
		}

		// Prepare OpenCV Datasructure CvMat
		Matrix3f U,V;
		Vector3f eigenValues;
		CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
		cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, crossCovariance.data());
		cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
		cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
		cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

		// Do SVD
		cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

		rotation = (U*V).transpose();
		Vector3f oldTranslation = translation;
		translation = rotation * ( - centroid1) + centroid2;
		movement = norm(translation - oldTranslation);
		residual /= m_Sigma;
		movement /= m_Sigma;

		refToDest = makeTranslation4f(translation)*expand3To4(rotation);
		refMat = refToDest * refMat;

		// visualization
		if( m_RedrawMethod )//&& (pass % 10) == 1 )
		{
			debugRenderer->clearDebugData("asdf");
			debugRenderer->beginRenderJob("asdf");
			Matrix4f refToLocal = refMat;
			Matrix4f destToLocal = destMat;
			float cloudSize = 0;
			for( unsigned i=0;i<surfels->getNumEntries();i++ )
			{
				Vector3f pos = surfels->get3f(i, posAAT);
				Vector3f normal = surfels->get3f(i,normalAAT);
				pos = transformVector3f(refToLocal,pos);
				normal = shrink4To3(refToLocal)*normal;
				Matrix3f tangentFrame = calcTangentSystem(normal);
				float f = 0.5f;
				float ca = cos(45.0f/180.0f*M_PI);
				float light = max(0.0f, (fabs(normal[2]) - ca) / (1.0f - ca));
				Vector3f color = makeVector3f(1,0.7f,0) * light;
				debugRenderer->addQuad( pos + tangentFrame[0] * m_MedPointDistance * f,
					pos + tangentFrame[1] * m_MedPointDistance*f,
					pos - tangentFrame[0] * m_MedPointDistance*f,
					pos - tangentFrame[1] * m_MedPointDistance*f, color);
				cloudSize = max( cloudSize, norm(pos));
			}

			Vector3f destPoint = shrink4To3(invertFrame(destMat)[3]);
			m_hIt_dest->setSeekPointAndReset(destPoint);
			while( !m_hIt_dest->atEnd() )
			{
				Vector3f pos = m_hIt_dest->get3f(posAAT);
				Vector3f normal = m_hIt_dest->get3f(normalAAT);
				if( norm( pos - destPoint) > cloudSize )
					break;

				pos = transformVector3f(destToLocal, pos);
				normal = shrink4To3(destToLocal) * normal;
				Matrix3f tangentFrame = calcTangentSystem(normal);
				float f = 1.0f;
				float ca = cos(45.0f/180.0f*M_PI);
				float light = max(0.0f, (fabs(normal[2]) - ca) / (1.0f - ca));
				Vector3f color = makeVector3f(0,0.5f,1) * light;
				debugRenderer->addQuad( pos + tangentFrame[0] * m_MedPointDistance * f,
					pos + tangentFrame[1] * m_MedPointDistance*f,
					pos - tangentFrame[0] * m_MedPointDistance*f,
					pos - tangentFrame[1] * m_MedPointDistance*f, color);

				m_hIt_dest->next();
			}

			debugRenderer->endRenderJob();
			(*m_RedrawMethod)();
			debugRenderer->clearDebugData("asdf");

		}

		/*if( movement < m_ConvergeDistance )
		{
			debugOutput << "converged\n";
			return residual;
		}*/

		if( pass > 0 )
		{
			//debugOutput << "residual change: " << lastResidual - residual << "\n";
		}

	}
	debugOutput << " residual: " << residual << 
		"\n";
	return residual;
}

void geoxl::LocalICP::endICP()
{
	delete m_hIt_ref;
	m_hIt_ref = NULL;
	delete m_hIt_dest;
	m_hIt_dest = NULL;
}

float geoxl::LocalICP::computeResidual( const Vector3f &refPoint, const Vector3f &refNormal, const Vector3f &destPoint,const Matrix3f &rotation )
{
	float residual = 0;
	float summed_weight = 0;
	m_hIt_ref->setSeekPointAndReset(refPoint);
	float32 radius_2 = pow(m_Sigma * 2.0f,2.0f);
	float32 inv_sigma_2 = 1.0f / pow(m_Sigma,2.0f);
	vector<Vector3f> samples;
	vector<float> weights;
	Vector3f average = NULL_VECTOR3F;
	while( !m_hIt_ref->atEnd() )
	{
		Vector3f pos = m_hIt_ref->get3f(m_Ref_PosAAT);
		float dist_2 = normQuad(pos - refPoint);
		if( dist_2 > radius_2 )
			break;

		Vector3f normal = m_hIt_ref->get3f(m_Ref_NormalAAT);

		// transform surfel by current transformation
		Surfel surfelTransformed;
		surfelTransformed.pos = rotation * (pos - refPoint)  + destPoint;
		surfelTransformed.normal = rotation * normal;

		// seek closest point
		m_hIt_dest->setSeekPointAndReset(surfelTransformed.pos);
		Vector3f closestPoint = m_hIt_dest->get3f(m_Dest_PosAAT);
		Vector3f closestPointNormal =  m_hIt_dest->get3f(m_Dest_NormalAAT);
		float weight = 1.0f;//pow(closestPointNormal * surfelTransformed.normal,2.0f);
		summed_weight += weight;

		// project onto surfel
		Vector3f projectedPoint = surfelTransformed.pos - closestPointNormal * 
			(closestPointNormal * (surfelTransformed.pos - closestPoint));

		samples.push_back(surfelTransformed.pos - projectedPoint);
		weights.push_back(weight);
		average += samples.back() * weights.back();

		m_hIt_ref->next();
	}

	// remove global translation from samples and compute resiual
	average /= summed_weight;
	average = refNormal * (average * refNormal);
	for( unsigned i=0;i<samples.size();i++ )
	{
		residual += norm(samples[i] - average) * weights[i];
	}

	residual /= summed_weight;
	return residual;
}

geoxl::Matrix2f geoxl::LocalICP::computeSlippage( const Vector3f &refPoint, const Vector3f &refNormal, const Vector3f &destPoint,const Matrix3f &rotation )
{
	Matrix3f tangentMatrix = calcTangentSystem(refNormal);
	float delta = m_Sigma * m_Slippage_Spacing;
	const Vector3f tangent_u = tangentMatrix[0] * delta;
	const Vector3f tangent_v = tangentMatrix[1] * delta;

/* function values:

		   fv1   
   fu1    0    fu2
	      fv2   fuv
*/

	float32 fu1 = computeResidual( refPoint, refNormal, refPoint - tangent_u,rotation);
	float32 fu2 = computeResidual( refPoint, refNormal, refPoint + tangent_u,rotation);
	float32 fv1 = computeResidual( refPoint, refNormal, refPoint - tangent_v,rotation);
	float32 fv2 = computeResidual( refPoint, refNormal, refPoint + tangent_v,rotation);
	float32 fuv = computeResidual( refPoint, refNormal, refPoint + tangent_v + tangent_u,rotation);

	float32 delta_2 = pow(delta,2.0f);
	float32 du2 = (fu1 + fu2) / delta_2;
	float32 dv2 = (fv1 + fv2) / delta_2;

	float32 dudv = ((fu2 - 0) - (fuv - fv2)) / delta_2;

	Matrix2f hessian = makeMatrix2f(du2, dudv,dudv,dv2);
	return hessian;
}

bool geoxl::LocalICP::computeSlippageCovariance( const Matrix2f &h, const Vector3f &refNormal, Matrix3f &covariance )
{
	// compute eigenvalues
	Vector2f eigenValues;
	Matrix2f eigenVectors;
	h.computeEigenStructure(eigenValues, eigenVectors);
	if( eigenValues[0] < m_Slippage_MinFirstEigen)
	{
		debugOutput << "eigenValues: " << eigenValues << " <--- dropped \n";
		return false;
	}

	// rotate eigenvectors
	Matrix3f tangentMatrix = calcTangentSystem(refNormal);
	Vector3f eigenVector_u = tangentMatrix * makeVector3f(eigenVectors[0][0],eigenVectors[0][1],0);
	Vector3f eigenVector_v = tangentMatrix * makeVector3f(eigenVectors[1][0],eigenVectors[1][1],0);

	covariance = outerProduct(eigenVector_u,eigenVector_u) * eigenValues[0];
	covariance += outerProduct(eigenVector_v,eigenVector_v) * eigenValues[1];

	return true;	
}

void geoxl::LocalICP::computeSlippage6d( const Vector3f &refPoint,  Matrix6f &covariance, Vector6f &eigenValues )
{
	m_hIt_ref->setSeekPointAndReset(refPoint);
	m_hIt_ref->next();
	float32 radius_2 = pow(m_Sigma * 2.0f,2.0f);
	float32 inv_sigma_2 = 1.0f / pow(m_Sigma,2.0f);

	Matrix3f posMatrix, normalMatrix, correlationMatrix;
	posMatrix = normalMatrix = correlationMatrix = IDENTITY3F*0;

	float summedWeight = 0;
	while( !m_hIt_ref->atEnd())
	{
		Vector3f pos = m_hIt_ref->get3f(m_Ref_PosAAT);
		Vector3f normal = m_hIt_ref->get3f(m_Ref_NormalAAT);
		Vector3f p = pos - refPoint;
		float dist_2 = normQuad(p);
		if( dist_2 > radius_2 )
			break;

		p /= m_Sigma;
		float weight = exp(-dist_2 * inv_sigma_2);
		summedWeight += weight;
		Vector3f c = p.crossProduct(normal);

		posMatrix += outerProduct(c,c) * weight;
		normalMatrix += outerProduct(normal,normal) * weight;
		correlationMatrix += outerProduct(normal,c) * weight;

		m_hIt_ref->next();
	}

	for( unsigned i=0;i<3;i++ )
	{
		for( unsigned j=0;j<3;j++ )
		{
			covariance[i][j] = posMatrix[i][j];
			covariance[i+3][j+3] = normalMatrix[i][j];
			covariance[i+3][j] = correlationMatrix[i][j];
			covariance[i][j+3] = correlationMatrix[j][i];
		}
	}

	Matrix6f eigenVectors;
	if( summedWeight > 0.0f)
		covariance /= summedWeight;
	covariance.computeEigenStructure(eigenValues, eigenVectors);
	eigenValues *= m_Slippage_MinFirstEigen;

	//debugOutput << "eigenvalues: " << eigenValues << "\n";
}

float geoxl::LocalICP::computeFastICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat, float maxMovement )
{
	float residual = 0;
	float lastResidual = 1e20f;
	float movement = 0;
	const mpcard numSurfels = surfels->getNumEntries();
	AAT posAAT = surfels->getAAT("position");
	AAT normalAAT = surfels->getAAT("normal");
	vector<Vector3f> destPoints(numSurfels);
	vector<float> surfelWeights(numSurfels);
	Vector3f startPointRef = shrink4To3(invertFrame(refMat)[3]);
	for( unsigned pass=0;pass<m_MaxNumIterations;pass++)
	{
		lastResidual = residual;
		Matrix4f refToDest = invertFrame(destMat) * refMat;
		Matrix3f rotation = shrink4To3(refToDest);
		Vector3f translation = shrink4To3(refToDest[3]);

		// visualization
		if( m_RedrawMethod )//&& (pass % 10) == 1 )
		{
			debugRenderer->clearDebugData("asdf");
			debugRenderer->beginRenderJob("asdf");
			Matrix4f refToLocal = refMat;
			Matrix4f destToLocal = destMat;
			BoundingBox3f bb;
			for( unsigned i=0;i<surfels->getNumEntries();i++ )
			{
				Vector3f pos = surfels->get3f(i, posAAT);
				Vector3f normal = surfels->get3f(i,normalAAT);
				pos = transformVector3f(refToLocal,pos);
				normal = shrink4To3(refToLocal)*normal;
				Matrix3f tangentFrame = calcTangentSystem(normal);
				float f = 0.5f;
				float ca = cos(45.0f/180.0f*M_PI);
				float light = max(0.0f, (fabs(normal[2]) - ca) / (1.0f - ca));
				Vector3f color = makeVector3f(1,0.7f,0) * light;
				debugRenderer->addQuad( pos + tangentFrame[0] * m_MedPointDistance * f,
					pos + tangentFrame[1] * m_MedPointDistance*f,
					pos - tangentFrame[0] * m_MedPointDistance*f,
					pos - tangentFrame[1] * m_MedPointDistance*f, color);

				if( i == 0)
					bb.lowerCorner = bb.upperCorner = pos;
				else
					bb.addPoint(pos);
			}
			float cloudSize = bb.getMaxSideLength()/2.0f;

			Vector3f destPoint = shrink4To3(invertFrame(destMat)[3]);
			m_hIt_dest->setSeekPointAndReset(destPoint);
			while( !m_hIt_dest->atEnd() )
			{
				Vector3f pos = m_hIt_dest->get3f(posAAT);
				Vector3f normal = m_hIt_dest->get3f(normalAAT);
				if( norm( pos - destPoint) > cloudSize )
					break;

				pos = transformVector3f(destToLocal, pos);
				normal = shrink4To3(destToLocal) * normal;
				Matrix3f tangentFrame = calcTangentSystem(normal);
				float f = 1.0f;
				float ca = cos(45.0f/180.0f*M_PI);
				float light = max(0.0f, (fabs(normal[2]) - ca) / (1.0f - ca));
				Vector3f color = makeVector3f(0,0.5f,1) * light;
				debugRenderer->addQuad( pos + tangentFrame[0] * m_MedPointDistance * f,
					pos + tangentFrame[1] * m_MedPointDistance*f,
					pos - tangentFrame[0] * m_MedPointDistance*f,
					pos - tangentFrame[1] * m_MedPointDistance*f, color);

				m_hIt_dest->next();
			}

			debugRenderer->endRenderJob();
			(*m_RedrawMethod)();
			debugRenderer->clearDebugData("asdf");

		}

		residual = 0;
		float summed_weight = 0;
		Matrix3f P = IDENTITY3F * 0;
		Matrix3f S = IDENTITY3F * 0;
		Matrix3f R = IDENTITY3F * 0;
		Vector6f resB = NULL_VECTOR6F;

		// calculate corresponding points
		for( unsigned i=0;i<numSurfels;i++)
		{
			Vector3f spos = surfels->get3f(i,posAAT);
			Vector3f snormal = surfels->get3f(i,normalAAT);

			// transform surfel by current transformation
			Surfel surfelTransformed;
			surfelTransformed.pos = rotation * spos  + translation;
			surfelTransformed.normal = rotation * snormal;

			// seek closest point
			m_hIt_dest->setSeekPointAndReset(surfelTransformed.pos);
			Vector3f closestPoint = m_hIt_dest->get3f(m_Dest_PosAAT);
			Vector3f closestPointNormal =  m_hIt_dest->get3f(m_Dest_NormalAAT);
			surfelWeights[i] = norm(closestPoint-surfelTransformed.pos) > m_MedPointDistance * m_OutlierDistance ? 0 : 1;
				
			if( surfelWeights[i] < 1e-20f )
				continue;

			// compute local quadratic approximation
			// M =  n n^t
			// F(x) = (x-p)^tM(x-p) = x^tMx - 2x^tMp + ...
			Vector3f relP = surfelTransformed.pos;
			Matrix3f matM = outerProduct(closestPointNormal, closestPointNormal);
			Vector3f b = matM * closestPoint * (-2.0f);
			// F(x) = Ax + Bxy + Cy + Dxz + Eyz + Fz + Gx + Hy + Iz (Mitra et al 2004)
			float A = matM[0][0];
			float B = matM[1][0] + matM[0][1];
			float C = matM[1][1];
			float D = matM[2][0] + matM[0][2];
			float E = matM[2][1] + matM[1][2];
			float F = matM[2][2];
			float G = b[0];
			float H = b[1];
			float I = b[2];

			// compute derivates
			float x = relP[0];
			float y = relP[1];
			float z = relP[2];
			float J = 2.0f * (C*x*x - B*x*y + A*y*y);
			float K = 2.0f * (F*x*x - D*x*z + A*z*z);
			float L = 2.0f * (F*y*y - E*y*z + C*z*z);
			float M = -E*x*x + D*x*y + B*x*z - 2.0f*A*y*z;
			float N = -D*y*y + E*x*y -2.0f*C*x*z + B*y*z;
			float T = -B*z*z -2.0f*F*x*y + E*x*z + D*y*z;
			float U = B*(x*x-y*y) + 2.0f*(C-A)*x*y + E*x*z - D*y*z + H*x - G*y;
			float V = D*(z*z-x*x) - E*x*y + 2.0f*(A-F)*x*z + B*y*z -I*x + G*z;
			float W = E*(y*y-z*z) + D*x*y - B*x*z + 2.0f*(F-C)*y*z + I*y - H*z;

			// setup linear system
			P += makeMatrix3f(J, M, N, M, K, T, N, T, L) * surfelWeights[i];
			S += makeMatrix3f(B*x - A*y*2.0f, C*x*2.0f - B*y, E*x - D*y,
									-D*x + A*z*2.0f, -E*x + B*z, -2.0f*F*x + D*z,
									D*y - B*z, E*y - 2.0f*C*z, 2.0f*F*y - E*z) * surfelWeights[i];
			R += makeMatrix3f(2.0f*A, B, D, 
									B, 2.0f*C, E, 
									D, E, 2.0f*F) * surfelWeights[i];
			resB -= makeVector6f(U,V,W,
				2.0f*A*x + B*y + D*z + G,
				B*x + 2.0f*C*y + E*z + H,
				D*x + E*y + 2.0f*F*z + I) * surfelWeights[i];

			
			// project onto surfel
			Vector3f projectedPoint = surfelTransformed.pos - closestPointNormal * 
				(closestPointNormal * (surfelTransformed.pos - closestPoint));
			float projectedDistance = norm(surfelTransformed.pos - projectedPoint);
			destPoints[i] = projectedPoint;

			// current residual
			residual += projectedDistance * surfelWeights[i];

			summed_weight +=  surfelWeights[i];

		}

		if( summed_weight < 1e-20f )
			return 1e20f;

		residual /= summed_weight;

		// setup linear system
		Matrix3f S_T = S.transpose();
		Matrix6f matM;
		for( unsigned a=0;a<3;a++ )
		{
			for( unsigned b=0;b<3;b++ )
			{
				matM[a][b] = P[a][b];
				matM[a+3][b] = S[a][b];
				matM[a][b+3] = S_T[a][b];
				matM[a+3][b+3] = R[a][b];
			}
		}

		CvMat cvMatInput, cvMatOutput;
		Matrix6f matM_inv;
		cvInitMatHeader( &cvMatInput, 6, 6, CV_32FC1, matM.data());
		cvInitMatHeader( &cvMatOutput, 6, 6, CV_32FC1, matM_inv.data());
		cvInvert(&cvMatInput, &cvMatOutput,CV_SVD_SYM);

		Vector6f coeffs = matM_inv * resB;

		float cos_a = cos(coeffs[0]);
		float sin_a = sin(coeffs[0]);
		float cos_b = cos(coeffs[1]);
		float sin_b = sin(coeffs[1]);
		float cos_c = cos(coeffs[2]);
		float sin_c = sin(coeffs[2]);
		Matrix3f delta_rotation = 
			makeMatrix3f(cos_a, -sin_a,0,
							 sin_a, cos_a, 0,
							 0,0,1) * 
			makeMatrix3f(cos_b, 0, sin_b,
							 0,1,0,
							 -sin_b, 0, cos_b) *
			makeMatrix3f(1, 0, 0,
			             0, cos_c, -sin_c,
							 0, sin_c, cos_c );

		Matrix4f deltaTransformation = makeTranslation4f(coeffs[3],coeffs[4],coeffs[5]) * expand3To4(delta_rotation);
		residual /= m_Sigma;

		rotation = shrink4To3(refToDest);
		translation = shrink4To3(refToDest[3]);
		refMat = (destMat)*deltaTransformation *invertFrame(destMat)* refMat;

		float movement = norm( shrink4To3(invertFrame(refMat)[3]) - startPointRef);
		if( movement > maxMovement)
			return 1e20f;

		if( fabs(lastResidual - residual)/lastResidual < 0.05f )
			return residual;

		lastResidual = residual;

	}
	//debugOutput << " residual: " << residual << 
	//	"\n";
	return residual;

}

float geoxl::LocalICP::computeFastICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat, float maxMovement, Scene*m_scene )
{

	
	float residual = 0;
	float lastResidual = 0;
	float movement = 0;
	const mpcard numSurfels = surfels->getNumEntries();
	AAT posAAT = surfels->getAAT("position");
	AAT normalAAT = surfels->getAAT("normal");
	vector<Vector3f> destPoints(numSurfels);
	vector<float> surfelWeights(numSurfels);
	Vector3f startPointRef = shrink4To3(invertFrame(refMat)[3]);
	for( unsigned pass=0;pass<m_MaxNumIterations;pass++)
	{
		
		string myid = "icp";
		debugRenderer->clearDebugData( myid );	// clear data from last call
		debugRenderer->beginRenderJob( myid );	// specify job name 
		
		lastResidual = residual;
		Matrix4f refToDest = invertFrame(destMat) * refMat;
		Matrix3f rotation = shrink4To3(refToDest);
		Vector3f translation = shrink4To3(refToDest[3]);

		residual = 0;
		float summed_weight = 0;
		Matrix3f P = IDENTITY3F * 0;
		Matrix3f S = IDENTITY3F * 0;
		Matrix3f R = IDENTITY3F * 0;
		Vector6f resB = NULL_VECTOR6F;

		// calculate corresponding points
		for( unsigned i=0;i<numSurfels;i++)
		{
			Vector3f spos = surfels->get3f(i,posAAT);
			Vector3f snormal = surfels->get3f(i,normalAAT);

			// transform surfel by current transformation
			Surfel surfelTransformed;
			surfelTransformed.pos = rotation * spos  + translation;
			surfelTransformed.normal = rotation * snormal;

			// seek closest point
			m_hIt_dest->setSeekPointAndReset(surfelTransformed.pos);
			Vector3f closestPoint = m_hIt_dest->get3f(m_Dest_PosAAT);
			Vector3f closestPointNormal =  m_hIt_dest->get3f(m_Dest_NormalAAT);
			surfelWeights[i] = norm(closestPoint-surfelTransformed.pos) > m_MedPointDistance * m_OutlierDistance ? 0 : 1;

			// compute local quadratic approximation
			// M =  n n^t
			// F(x) = (x-p)^tM(x-p) = x^tMx - 2x^tMp + ...
			Vector3f relP = surfelTransformed.pos;
			Matrix3f matM = outerProduct(closestPointNormal, closestPointNormal);
			Vector3f b = matM * closestPoint * (-2.0f);
			// F(x) = Ax + Bxy + Cy + Dxz + Eyz + Fz + Gx + Hy + Iz (Mitra et al 2004)
			float A = matM[0][0];
			float B = matM[1][0] + matM[0][1];
			float C = matM[1][1];
			float D = matM[2][0] + matM[0][2];
			float E = matM[2][1] + matM[1][2];
			float F = matM[2][2];
			float G = b[0];
			float H = b[1];
			float I = b[2];

			// compute derivates
			float x = relP[0];
			float y = relP[1];
			float z = relP[2];
			float J = 2.0f * (C*x*x - B*x*y + A*y*y);
			float K = 2.0f * (F*x*x - D*x*z + A*z*z);
			float L = 2.0f * (F*y*y - E*y*z + C*z*z);
			float M = -E*x*x + D*x*y + B*x*z - 2.0f*A*y*z;
			float N = -D*y*y + E*x*y -2.0f*C*x*z + B*y*z;
			float T = -B*z*z -2.0f*F*x*y + E*x*z + D*y*z;
			float U = B*(x*x-y*y) + 2.0f*(C-A)*x*y + E*x*z - D*y*z + H*x - G*y;
			float V = D*(z*z-x*x) - E*x*y + 2.0f*(A-F)*x*z + B*y*z -I*x + G*z;
			float W = E*(y*y-z*z) + D*x*y - B*x*z + 2.0f*(F-C)*y*z + I*y - H*z;

			// setup linear system
			P += makeMatrix3f(J, M, N, M, K, T, N, T, L) * surfelWeights[i];
			S += makeMatrix3f(B*x - A*y*2.0f, C*x*2.0f - B*y, E*x - D*y,
				-D*x + A*z*2.0f, -E*x + B*z, -2.0f*F*x + D*z,
				D*y - B*z, E*y - 2.0f*C*z, 2.0f*F*y - E*z) * surfelWeights[i];
			R += makeMatrix3f(2.0f*A, B, D, 
				B, 2.0f*C, E, 
				D, E, 2.0f*F) * surfelWeights[i];
			resB -= makeVector6f(U,V,W,
				2.0f*A*x + B*y + D*z + G,
				B*x + 2.0f*C*y + E*z + H,
				D*x + E*y + 2.0f*F*z + I) * surfelWeights[i];


			// project onto surfel
			Vector3f projectedPoint = surfelTransformed.pos - closestPointNormal * 
				(closestPointNormal * (surfelTransformed.pos - closestPoint));
			float projectedDistance = norm(surfelTransformed.pos - projectedPoint);
			destPoints[i] = projectedPoint;

			// current residual
			residual += projectedDistance * surfelWeights[i];

			debugRenderer->addLine(surfelTransformed.pos, closestPoint, makeVector3f(1,0,0),makeVector3f(0,1,0),2);
			debugRenderer->addSphere(surfelTransformed.pos,0.2f,makeVector3f(1,1,0));

			summed_weight +=  surfelWeights[i];

		}

		debugRenderer->endRenderJob();
		ObjectViewsTable::update(m_scene);

		if( summed_weight < 1e-20f )
			return 1e20f;

		residual /= summed_weight;

		// setup linear system
		Matrix3f S_T = S.transpose();
		Matrix6f matM;
		for( unsigned a=0;a<3;a++ )
		{
			for( unsigned b=0;b<3;b++ )
			{
				matM[a][b] = P[a][b];
				matM[a+3][b] = S[a][b];
				matM[a][b+3] = S_T[a][b];
				matM[a+3][b+3] = R[a][b];
			}
		}

		CvMat cvMatInput, cvMatOutput;
		Matrix6f matM_inv;
		cvInitMatHeader( &cvMatInput, 6, 6, CV_32FC1, matM.data());
		cvInitMatHeader( &cvMatOutput, 6, 6, CV_32FC1, matM_inv.data());
		cvInvert(&cvMatInput, &cvMatOutput);

		Vector6f coeffs = matM_inv * resB;

		float cos_a = cos(coeffs[0]);
		float sin_a = sin(coeffs[0]);
		float cos_b = cos(coeffs[1]);
		float sin_b = sin(coeffs[1]);
		float cos_c = cos(coeffs[2]);
		float sin_c = sin(coeffs[2]);
		Matrix3f delta_rotation = 
			makeMatrix3f(cos_a, -sin_a,0,
			sin_a, cos_a, 0,
			0,0,1) * 
			makeMatrix3f(cos_b, 0, sin_b,
			0,1,0,
			-sin_b, 0, cos_b) *
			makeMatrix3f(1, 0, 0,
			0, cos_c, -sin_c,
			0, sin_c, cos_c );

		Matrix4f deltaTransformation = makeTranslation4f(coeffs[3],coeffs[4],coeffs[5]) * expand3To4(delta_rotation);
		residual /= m_Sigma;
		//debugOutput << "residual["<<pass<<"]=" << residual << " coeff:" << coeffs << "\n";

		//refToDest = deltaTransformation * makeTranslation4f(translation)*expand3To4(rotation);
		rotation = shrink4To3(refToDest);
		translation = shrink4To3(refToDest[3]);
		refMat = (destMat)*deltaTransformation *invertFrame(destMat)* refMat;

		float movement = norm( shrink4To3(invertFrame(refMat)[3]) - startPointRef);
		if( movement > maxMovement)
			return 1e20f;


	}
	debugOutput << " residual: " << residual << 
		"\n";
	return residual;
}