#ifndef LocalICPH
#define LocalICPH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl
{
class PointCloud;
class HierarchicalKNNIterator;
typedef void (*LocalICP_redrawFunctionPtr)();
class Scene;

/**
	computes iterative closest point (ICP), weighted with a gaussian
	filter kernel. uses point-to-plane
*/
class DEFORMATION_API LocalICP : public Persistent
{
	GEOXL_CLASS( LocalICP )

private:

	// properties
	float32		m_Sigma;
	card32		m_MaxNumIterations;
	float32		m_MaxResidual;
	float32		m_ConvergeDistance;
	float32		m_Slippage_Spacing;
	float32		m_OutlierDistance;
	float32		m_Slippage_MinFirstEigen;

	// members
	PointCloud * m_Reference;
	PointCloud * m_Destination;
	HierarchicalKNNIterator * m_hIt_ref;
	HierarchicalKNNIterator * m_hIt_dest;
	AAT			m_Ref_PosAAT;
	AAT			m_Dest_PosAAT;
	AAT			m_Ref_NormalAAT;
	AAT			m_Dest_NormalAAT;
	float32		m_MedPointDistance;

	LocalICP_redrawFunctionPtr m_RedrawMethod;

	struct Surfel
	{
		Vector3f pos;
		Vector3f normal;
		float weight;
		float dynamicWeight;
	};


public:
	LocalICP(void);
	~LocalICP(void);

   virtual void assign(const Object* obj, CopyContext *context = NULL);

	void beginICP(PointCloud * reference, PointCloud * destination);

	/// computes corresponding point in destination point cloud
	/// destPoint and rotation have to be initialized with a rough estimation
	bool computeICP( const Vector3f &refPoint, Vector3f &destPoint, Matrix3f &rotation );

	/// computes point-to-plane residual
	float computeResidual( const Vector3f &refPoint, const Vector3f &refNormal, const Vector3f &destPoint,const Matrix3f &rotation );

	/// compute 2d hessian
	Matrix2f computeSlippage( const Vector3f &refPoint, const Vector3f &refNormal, const Vector3f &destPoint,const Matrix3f &rotation );

	/// compute slippage covariance, returns false when slippage is too high
	bool computeSlippageCovariance( const Matrix2f &h,const Vector3f &refNormal,  Matrix3f &covariance);

	/// compute 6d slippage
	void computeSlippage6d( const Vector3f &refPoint, Matrix6f &covariance, Vector6f &eigenValues );

	/// compute point-to-point icp (hack) for a set of surfels to the destination point cloud
	/// returns the residual
	float computeICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat );

	/// compute point-to-plane icp (energy minimizing)  for a set of surfels to the destination point cloud
	/// returns the residual
	float computeFastICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat, float maxMovement );
	float computeFastICP( PointSet * surfels, Matrix4f &refMat, const Matrix4f &destMat, float maxMovement, Scene*m_scene );

	void endICP();


	// property interface
	void setSigma( float s ){m_Sigma = s;}

	void setRedrawMethod( LocalICP_redrawFunctionPtr ptr ) {m_RedrawMethod = ptr;}


};

}

#endif