//---------------------------------------------------------------------------
#ifndef LocalSurfaceFunctionH
#define LocalSurfaceFunctionH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------

namespace geoxl {


template<unsigned NUM_BASIS_FUNCTIONS>
class LocalSurfaceFunction
{
	public:
		// Evaluate basis function monomial
		static float32 f_i( const float32& u, const float32& v, const card32 bfIndex );
		// Evaluate function at position u, v
		static float32 f( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v );
		// Return normal at position u, v
		static Vector3f getNormal( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v );
		// Derivative of surface function in direction u
		static float32 f_u( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v );
		// Derivative of surface function in direction v
		static float32 f_v( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v );
		/// Return the number of basis functions
		static inline card32 getNumBasisFunctions() { return NUM_BASIS_FUNCTIONS; }
};

template<unsigned NUM_BASIS_FUNCTIONS>
float32 LocalSurfaceFunction<NUM_BASIS_FUNCTIONS>::f_i( const float32& u, const float32& v, const card32 bfIndex )
{
	if ( NUM_BASIS_FUNCTIONS == 3 ){
		// 1st order polynomial
		if ( bfIndex == 0 ) return 1;
		else if ( bfIndex == 1 ) return u;
		else if ( bfIndex == 2 ) return v;
		else return 0;
	} else if ( NUM_BASIS_FUNCTIONS == 6 ){
		// 2nd order polynomial
		if ( bfIndex == 0 ) return 1;
		else if ( bfIndex == 1 ) return u;
		else if ( bfIndex == 2 ) return v;
		else if ( bfIndex == 3 ) return u * v;
		else if ( bfIndex == 4 ) return u * u;
		else if ( bfIndex == 5 ) return v * v;
		else return 0;
	} else if ( NUM_BASIS_FUNCTIONS == 10 ){
		// 2rd order polynomial
		if ( bfIndex == 0 ) return 1;
		else if ( bfIndex == 1 ) return u;
		else if ( bfIndex == 2 ) return v;
		else if ( bfIndex == 3 ) return u * v;
		else if ( bfIndex == 4 ) return u * u;
		else if ( bfIndex == 5 ) return v * v;
		else if ( bfIndex == 6 ) return u * u * v;
		else if ( bfIndex == 7 ) return v * v * u;
		else if ( bfIndex == 8 ) return u * u * u;
		else if ( bfIndex == 9 ) return v * v * v;
		else return 0;
	} else {
		return 0;
	}
}

template<unsigned NUM_BASIS_FUNCTIONS>
float32 LocalSurfaceFunction<NUM_BASIS_FUNCTIONS>::f( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v ) 
{
	float32 value = 0;
	for ( unsigned int bfIndex = 0; bfIndex < NUM_BASIS_FUNCTIONS; bfIndex++ ){
		value += coeffs[bfIndex] * f_i( u, v, bfIndex );
	}
	return value;
}

template<unsigned NUM_BASIS_FUNCTIONS>
Vector3f LocalSurfaceFunction<NUM_BASIS_FUNCTIONS>::getNormal( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v )
{ 
	Vector3f normal = makeVector3f( -f_u( coeffs, u, v ), -f_v( coeffs, u,v ), 1 );  
	normal.normalize(); 
	return normal; 
}

template<unsigned NUM_BASIS_FUNCTIONS>
float32 LocalSurfaceFunction<NUM_BASIS_FUNCTIONS>::f_u( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v )
{
	if ( NUM_BASIS_FUNCTIONS == 3 ){
		return coeffs[1];	
	} else if ( NUM_BASIS_FUNCTIONS == 6 ){
		return coeffs[1] + coeffs[3] * v + 2 * coeffs[4] * u;
	} else if ( NUM_BASIS_FUNCTIONS == 10 ){
		return 0; // TODO
	}
}

template<unsigned NUM_BASIS_FUNCTIONS>
float32 LocalSurfaceFunction<NUM_BASIS_FUNCTIONS>::f_v( const StaticVector<float32,NUM_BASIS_FUNCTIONS>& coeffs, const float32& u, const float32& v )
{
	if ( NUM_BASIS_FUNCTIONS == 3 ){
		return coeffs[2];	
	} else if ( NUM_BASIS_FUNCTIONS == 6 ){
		return coeffs[2] + coeffs[3] * u + 2 * coeffs[5] * v;
	} else if ( NUM_BASIS_FUNCTIONS == 10 ){
		return 0; // TODO
	}
}

}

#endif