#include "stdafx.h"
//---------------------------------------------------------------------------
#include "TangentsEstimator.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "ProgressWindow.h"
#include "PointSetKNNQuery.h"
#include "PointSet_templCode.h"
#include <time.h>
#include "NumericalClassProperty.h"
#include "PCA.h"
#include "MathTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

void TangentsEstimator::computeTangents( const Vector3f normal, Vector3f& tangent_u, Vector3f& tangent_v )
{
	tangent_u = ( fabs( normal * XAXIS_VECTOR3F ) > 0.85 )? normal.crossProduct( YAXIS_VECTOR3F ) : normal.crossProduct( XAXIS_VECTOR3F );
	tangent_v = normal.crossProduct( tangent_u );
	tangent_u.normalize();
	tangent_v.normalize();
}

void TangentsEstimator::estimateTangents( PointSet* ps )
{
	if ( ps == NULL ) {
		debugOutput << "estimateTangents: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !ps->providesAttribute( "normal" ) || !ps->providesAttribute( "tangent_u" ) || 
		 !ps->providesAttribute( "tangent_v" ) ){
		debugOutput << "estimateTangents: channel missing\n"; debugOutput.flush();
		return;
	}

	AAT NORMAL = ps->getAAT( "normal" );
	AAT TANGENT_U = ps->getAAT( "tangent_u" );
	AAT TANGENT_V = ps->getAAT( "tangent_v" );

	progressWindow->pushStep( true, "Tangents" );

	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		const Vector3f& normal = ps->get3f( pIndex, NORMAL );
		Vector3f tangent_u, tangent_v;
		computeTangents( normal, tangent_u, tangent_v );
		ps->set3f( pIndex, TANGENT_U, tangent_u );
		ps->set3f( pIndex, TANGENT_V, tangent_v );

		progressWindow->progress( float32( pIndex * 100 ) / float32( ps->getNumEntries() ) );
	}

	progressWindow->popSteps();
}

void TangentsEstimator::estimateTangents( PointCloud* pc )
{
	if ( pc == NULL ) {
		debugOutput << "estimateTangents: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !pc->providesAttribute( "normal" ) || !pc->providesAttribute( "tangent_u" ) || 
		 !pc->providesAttribute( "tangent_v" ) ){
		debugOutput << "estimateTangents: channel missing\n"; debugOutput.flush();
		return;
	}

	AAT NORMAL = pc->getAAT( "normal" );
	AAT TANGENT_U = pc->getAAT( "tangent_u" );
	AAT TANGENT_V = pc->getAAT( "tangent_v" );

	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );

	progressWindow->pushStep( true, "Tangents" );

	basicIt->reset();
	long pIndex = 0;
	while( !basicIt->atEnd() ){
		const Vector3f& normal = basicIt->get3f( NORMAL );
		Vector3f& tangent_u = ( fabs( normal * XAXIS_VECTOR3F ) > 0.85 )? normal.crossProduct( YAXIS_VECTOR3F ) : normal.crossProduct( XAXIS_VECTOR3F );
		Vector3f& tangent_v = normal.crossProduct( tangent_u );
		tangent_u.normalize();
		tangent_v.normalize();
		modIt->set3f( TANGENT_U, tangent_u );
		modIt->set3f( TANGENT_V, tangent_v );

		progressWindow->progress( float32( pIndex * 100 ) / float32( pc->getNumPoints() ) );
		pIndex++;
		basicIt->next();
	}

	progressWindow->popSteps();

	delete it;
}