//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SurfelGraphInit2.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ARSettings.h"
#include "SGEShape.h"
#include "InCorePCTopologyGraph.h"
#include "PointSetKNNQuery.h"
#include "TopologyAttachment.h"
#include "SGESurfelGraph.h"
#include "SGEPointCloud.h"
#include "ARTopologicalResampling.h"
#include "DeformationTools.h"
#include "Resampling.h"
#include "time.h"
#include "PointSetANNQuery.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( SurfelGraphInit2 , 0 ) 
{
	BEGIN_CLASS_INIT( SurfelGraphInit2 );
}

SurfelGraphInit2::SurfelGraphInit2() 
{
}

SurfelGraphInit2::~SurfelGraphInit2() 
{
}

void SurfelGraphInit2::initialize( SGESurfelGraph* surfels, SGEPointCloud* data, ARSettings *settings )
{
	createSurfelGraph(surfels, settings);
	createSurfels(surfels, data, settings);
	if (settings->resampleInitialUrshapes) {
		resample(surfels, settings);
		//Resampling::downsample( surfels->getPointSet(), surfels->getTopology(), settings->topSurfelSpacing );	
	}
	if( settings->outlier_removeSmallPatches ){
		removeSmallComponents( surfels->getPointSet(), surfels->getTopology(), settings->outlier_minNumberOfPoints, settings );
	}
	initChannels( surfels, settings );
}

void SurfelGraphInit2::createSurfelGraph( SGESurfelGraph* surfels, ARSettings* settings )
{
	float32 startTime = clock();
	// create initial empty point cloud
	PointCloud* surfelPC = getPointCloud( surfels->getScene(), surfels->getPath() );
	if ( surfelPC == NULL ){
		PointSet* ps = NULL;
		UnstructuredInCorePointCloud *upc = new UnstructuredInCorePointCloud();
		surfelPC = upc;
		ps = new PointSet();
		VertexDescriptor vd;
		vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
		vd.pushAttrib( mVAD( "tangent_u", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "tangent_v", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "coeffs", SURFEL_NUM_BASIS_FUNCTIONS, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "numDataPts", 1, VAD::DATA_FORMAT_INT32 ) );
		vd.pushAttrib( mVAD( "propOrigin", 2, VAD::DATA_FORMAT_INT32 ) ); // used for urshape propagation: index of propagating frame (-1 if not set), index of original left-side point
		vd.pushAttrib( mVAD( "defRefitChannel", 2, VAD::DATA_FORMAT_INT32 ) ); // used for urshape propagation: index of propagating frame (-1 if not set), index of original left-side point
		ps->clearAndSetup( 1,1, &vd );
		upc->setPointSet( ps );
		addPointCloud( surfels->getScene(), surfelPC, surfels->getPath() );
	} else {
		throw PException("SurfelGraphInit2::createSurfelGraph() - cannot create surfel graph, another object is in the way.");
	}
	// initialize empty topology... (used later)
	AttachedDataContainer *attachments = surfelPC->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if ( topAtt == NULL ){
		topAtt = new TopologyAttachment();
		topAtt->setup( "TopologyAtt", AttachedData::ADF_PERSISTENT );
		attachments->attachData( topAtt );
		InCorePCTopologyGraph* topology = new InCorePCTopologyGraph();
		VertexDescriptor topotVD;
		topotVD.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
		topology->clearAndSetup( &topotVD );
		topAtt->setTopology( topology );
	}
	if ( settings->surfaceRecDebugOutput ){
		debugOutput << "SurfelGraphInit2::createSurfelGraph time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; 
		debugOutput.flush();
	}
}

void SurfelGraphInit2::createSurfels( SGESurfelGraph* surfels, SGEPointCloud* data, ARSettings *settings )
{
	float32 startTime = clock();
	string msg;
	if ( !data->check( msg ) ) {
		throw PException(string("SurfelGraphInit2::createSurfels() - error checking scene graph structure [data]; msg: ")+msg);
	}
	PointCloud* dataPC = data->getPointCloud();
	
	SceneObjectIterator *it = data->createIterator(SOIT::CAP_BASIC_PC);
	BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
	if (!it || !bit) {	
		delete it;
		throw PException("SurfelGraphInit2::createSurfels() - cannot create iterator on data.");
	}
	const PointSet* dataPS = bit->lockBlockRead();
	AAT DATA_POSITION = dataPS->getAAT( "position" );

   const card32 numSurfacePoints = dataPS->getNumEntries();

	if ( !surfels->check( msg ) ){
		throw PException(string("SurfelGraphInit2::createSurfels() - error checking scene graph structure [surfels]; msg: ")+msg);
	}

	PointSet* surfelPS = surfels->getPointSet();
	surfelPS->changeHeight( numSurfacePoints );
	InCorePCTopologyGraph* topology = surfels->getTopology();
	topology->clearAndSetup( topology->getEdges()->getDescr() );
	AAT SURFEL_POSITION = surfelPS->getAAT( "position" );
	card32 surfaceIndex = 0;
	for ( unsigned int pIndex = 0; pIndex < dataPS->getNumEntries(); pIndex++ ){
		const Vector3f dataP = dataPS->get3f( pIndex, DATA_POSITION );
		surfelPS->set3f( surfaceIndex, SURFEL_POSITION, dataP );
		surfaceIndex++;
	}
	if ( settings->surfaceRecDebugOutput ) {
		debugOutput << "Created " << surfelPS->getNumEntries() << " surfels from " << dataPS->getNumEntries() << " data points.\n"; debugOutput.flush();
	}

	// Create initial topology graph

   // -- old version
	//AttachedIndexedOctree* surfelOctree = PointSetKNNQuery::createOctree( surfelPS, 32 );
	//PointSetKNNQuery knn( surfelPS, surfelOctree, makeVector3f( 0,0,0 ), SURFEL_POSITION );
	//for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
	//	const Vector3f& p = surfelPS->get3f( pIndex, SURFEL_POSITION );
	//	knn.setTargetPoint( p );
	//	bool found = true;
	//	card32 nIndex = pIndex;
	//	card32 numNeighbors = 0;
	//	while( ( numNeighbors < settings->topMaxNumNeighbors ) && found ){
	//		if ( nIndex != pIndex ) {
	//			Vector3f nPos = surfelPS->get3f(nIndex, SURFEL_POSITION);
	//			if (norm(nPos - p) < settings->topRelRadius*settings->topSurfelSpacing) {
	//				topology->addEdge( pIndex, nIndex );
	//			}
 //           numNeighbors++;
	//		}
	//		knn.getNextNearestPointIndex( nIndex, found );
	//	}
	//}
	//delete surfelOctree;
	//bit->unlock();
	//delete it;
	
	// -- new version, with ANN  (mw - Sept/07/2008)
   PointSetANNQuery knn(surfelPS, settings->topMaxNumNeighbors+1);
   for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
      const Vector3f& p = surfelPS->get3f( pIndex, SURFEL_POSITION );
      knn.setTargetPoint( p );
      bool found = true;
      card32 nIndex = pIndex;
      card32 numNeighbors = 0;
      while( ( numNeighbors < settings->topMaxNumNeighbors ) && found ){
         knn.getNextNearestPointIndex( nIndex, found );
         if ( nIndex != pIndex ) {
            Vector3f nPos = surfelPS->get3f(nIndex, SURFEL_POSITION);
            if (norm(nPos - p) < settings->topRelRadius*settings->topSurfelSpacing) {
               topology->addEdge( pIndex, nIndex );
            }
            numNeighbors++;
         }
      }
   }
   bit->unlock();
   delete it;
	



	if ( settings->surfaceRecDebugOutput ){
		debugOutput << "SurfelGraphInit2::createSurfels time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; 
		debugOutput.flush();
	}
}

void SurfelGraphInit2::resample( SGESurfelGraph* surfels, ARSettings *settings )
{
	float32 startTime = clock();
	string msg;
	if ( !surfels->check( msg ) ){
		throw PException(string("SurfelGraphInit2::resample() - error checking scene graph structure [data]; msg: ")+msg);
 	}
	UnstructuredInCorePointCloud *pc = surfels->getPointCloud();
	
	PointCloud *resampled = ARTopologicalResampling::resample(
		pc, 
      settings->topSurfelSpacing,
      settings->topRelRadius,
      settings->topMaxNumNeighbors
   );
   
   pc->assign(resampled);
   delete resampled;

   if ( settings->surfaceRecDebugOutput ){
	   debugOutput << "SurfelGraphInit2::resample time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; 
	   debugOutput.flush();
   }
}


void SurfelGraphInit2::initChannels( SGESurfelGraph* surfels, ARSettings* settings )
{
	float32 startTime = clock();
	string msg;
	if ( !surfels->check( msg ) ){
		throw PException(string("SurfelGraphInit2::initChannels() - error checking scene graph structure [surfels]; msg: ")+msg);
	}

	SceneObjectIterator *it = surfels->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
	ModifyablePointCloudIterator *mit = dynamic_cast<ModifyablePointCloudIterator*>(it);
	if (!it || !bit || !mit) {	
		delete it;
		throw PException("SurfelGraphInit2::initChannels() - cannot create iterator on surfels.");
	}
	PointSet* surfelPS = mit->lockBlockReadWrite();
	AAT PROP_ORIGIN = surfelPS->getAAT("propOrigin");
	AAT DEF_REFIT_CHANNEL = surfelPS->getAAT("defRefitChannel");
	for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
		surfelPS->set2i( pIndex, PROP_ORIGIN, makeVector2i(-1,-1) );
		surfelPS->set2i( pIndex, DEF_REFIT_CHANNEL, makeVector2i(-1,-1) );
	}
	
	// Init channels
	recomputeNormalsAndFrames( surfels->getPointCloud(), settings );
	
	bit->unlock();
	
	delete it;
	if ( settings->surfaceRecDebugOutput ){
		debugOutput << "SurfelGraphInit2::initChannels time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; 
		debugOutput.flush();
	}
}


void geoxl::SurfelGraphInit2::removeSmallComponents( PointSet * ps, InCorePCTopologyGraph * topology, card32 minConnectedPoints, ARSettings* settings )
{
	float32 startTime = clock();
	/// -1 = unassigned points
	vector<int> regionID(ps->getNumEntries(),-1);
	vector<unsigned> regionPointCount;

	list<unsigned> queue;
	int currentRegionID = -1;
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		// already assigned to a region?
		if( regionID[i] != -1 )
			continue;

		// new region
		currentRegionID++;
		regionPointCount.push_back(0);

		// start region growing
		queue.push_back(i);
		while( !queue.empty())
		{
			unsigned currentIndex = queue.front();
			queue.pop_front();

			if( regionID[currentIndex] != -1 )	// already assigned
				continue;

			// assign to current region
			regionPointCount[currentRegionID]++;
			regionID[currentIndex] = currentRegionID;

			// push neighbours into queue
			for( unsigned j=0;j<topology->getNumAdjacentVertices(currentIndex);j++ )
			{
				queue.push_back(topology->getAdjacentVertex(currentIndex,j));
			}

		}
	}


	// copy valid points & edge to new point set/topology
	// start with copying valid points into newPS and store
	// new point index 
	InCorePCTopologyGraph * newGraph = new InCorePCTopologyGraph;
	PointSet * newPS = new PointSet;
	newGraph->clearAndSetup( topology->getEdges()->getDescr() );
	newPS->clearAndSetup(1,0, ps->getDescr());
	vector<char> buffer(ps->getDescr()->getSize());
	char * bufferPtr = &buffer[0];
	vector<unsigned> mapPoint(ps->getNumEntries(),ps->getNumEntries());
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		if( regionPointCount[ regionID[i]] > minConnectedPoints )
		{
			mapPoint[i] = newPS->getNumEntries();
			ps->getVertex(i, bufferPtr);
			newPS->addPoint(bufferPtr);
		}
	}

	// copy topology information from valid points
	for(  unsigned i=0;i<ps->getNumEntries();i++ )
	{
		if( mapPoint[i] < ps->getNumEntries())
		{
			unsigned newPointIndex = mapPoint[i];
			for( unsigned j=0;j<topology->getNumAdjacentVertices(i);j++ )
			{
				PointCloudTopologyGraph::VertexIndex n = topology->getAdjacentVertex(i,j);
				newGraph->addEdge(newPointIndex,mapPoint[n]);
			}

		}
	}

	topology->assign(newGraph);
	delete newGraph;
	ps->assign(newPS);
	delete newPS;
	ps->invalidateLastNumFlagged();

	if ( ( settings != NULL ) && settings->surfaceRecDebugOutput ){
		debugOutput << "SurfelGraphInit2::removeSmallComponents time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; 
		debugOutput.flush();
	}
}

void SurfelGraphInit2::assign(const Object* obj, CopyContext *context) 
{
	copyObjectProperties(obj, this);
}