//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SurfelGraphTriangulator.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ClassMethods.h"
#include "SeparatorClassProperty.h"
#include "SGEShape.h"
#include "SGESurfelGraph.h"
#include "PointSet_templCode.h"
#include "ARConstraints.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "HoppeMesher.h"
#include "NormalUnifier.h"
#include <time.h>
#include "ProgressWindow.h"
#include "MarchingCubes3D.h"
#include "LocalSurfaceFunction.h"
#include "TopologyRangeSearch.h"
#include "SGSimpleAnimationNode.h"
#include "SGObjectNode.h"
#include "StandardSceneGraphIterator.h"
#include "GLShaderMaterial.h"
#include "TriangleMeshTools.h"
#include "DeformationTools.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "FifoStack.h"
#include "ARDeformationOptimizationContext.h"
#include "AREvaluateDeformation.h"
#include "DeformationEvaluator.h"
#include "TopologicalKNNIterator.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( SurfelGraphTriangulator, 0 ) 
{
	BEGIN_CLASS_INIT( SurfelGraphTriangulator )
}

SurfelGraphTriangulator::SurfelGraphTriangulator() 
{
	triangulationSettings = NULL;
}

SurfelGraphTriangulator::~SurfelGraphTriangulator() 
{
}

SGSimpleAnimationNode* SurfelGraphTriangulator::triangulate( 
	card32 startFrame, 
	card32 endFrame,
	UnstructuredInCorePointCloud* surfelPC, 
	const float32 samplingSpacing,
	ARSequence* seq, 
	DeformationEvaluator* evaluator,
	Scene* scene )
{
	if ( surfelPC == NULL ) return NULL;
	PointSet* surfelPS = surfelPC->getPointSet();
	if ( surfelPS == NULL ) return NULL;
	UnstructuredInCoreTriangleMesh* urShapeMesh = NULL;
	InCorePCTopologyGraph* surfelTopology = getTopology( surfelPC );
	AAT POSITION = surfelPS->getAAT( "position" );
	AAT TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT NORMAL = surfelPS->getAAT( "normal" );
	AAT COEFFS = surfelPS->getAAT( "coeffs" );
	AttachedIndexedOctree* surfelOctree = PointSetKNNQuery::createOctree( surfelPS, 32 );
	PointSetKNNQuery surfelKNN( surfelPS, surfelOctree, makeVector3f( 0, 0, 0 ), POSITION );
	PointSet* borderPS = NULL;
	InCorePCTopologyGraph* borderTopology = NULL;
	AttachedIndexedOctree* borderOctree = NULL;
	UnstructuredInCorePointCloud* borderPC = NULL;
	if ( triangulationSettings->doClipBorder ) {
		borderPC = detectBorderSurfels( surfelPC, samplingSpacing );
		borderPS = borderPC->getPointSet();
		borderTopology = getTopology( borderPC );
		borderOctree = PointSetKNNQuery::createOctree( borderPS, 32 );
	}
	if ( triangulationSettings->doMarchingCubes ){
		// Grid subdivision
		BoundingBox3f bbox = getMCDomain( surfelPS );
		Vector3f corners[8];
		float32 values[8];
		vector<Vector3f> verts;
		card32 nIndex;
		bool found;
		vector<Vector3f> vertices; vertices.clear();

		const float32 mcCubeSize = getMCubeSize( bbox, triangulationSettings->mcOctreeLevel );
		const float32 clipInfluence = getClipInfluence( mcCubeSize, samplingSpacing );
		const card32 resolution = power( 2, triangulationSettings->mcOctreeLevel );

		debugOutput << "  Meshing: resolution=" << resolution << "; tau=" << mcCubeSize << "; box=" << bbox << ".\n"; debugOutput.flush();

		if ( triangulationSettings->mcMode == MC_REGULAR ){
			progressWindow->pushStep( true, "MC" );
			// Regular grid version
			float32 startTime = clock();
			for ( unsigned int k = 0; k < resolution; k++ ){
				for ( unsigned int j = 0; j < resolution; j++ ){
					for ( unsigned int i = 0; i < resolution; i++ ){
						corners[0] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     j*mcCubeSize,     k*mcCubeSize );
						corners[1] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, j*mcCubeSize,     k*mcCubeSize );
						corners[2] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, (j+1)*mcCubeSize, k*mcCubeSize );
						corners[3] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     (j+1)*mcCubeSize, k*mcCubeSize );
						corners[4] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     j*mcCubeSize,     (k+1)*mcCubeSize );
						corners[5] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, j*mcCubeSize,     (k+1)*mcCubeSize );
						corners[6] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, (j+1)*mcCubeSize, (k+1)*mcCubeSize );
						corners[7] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     (j+1)*mcCubeSize, (k+1)*mcCubeSize );
						if ( triangulateCube( corners, surfelPS, surfelKNN, samplingSpacing ) ){
							for ( unsigned int n = 0; n < 8; n++ ){
								values[n] = getValue( corners[n], surfelKNN, surfelPS, surfelTopology, samplingSpacing );
							}
							MarchingCubes3D::triangulate( corners, values, 0, verts );
							for ( unsigned int n = 0; n < verts.size(); n++ ){
								vertices.push_back( verts[n] );
							}
						}
					}
				}
				progressWindow->progress( float32( k ) * 100.0f / float32( resolution ) );
			}
			progressWindow->popSteps();
			debugOutput << "  Meshing (regular): time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; debugOutput.flush();
		} else if ( triangulationSettings->mcMode == MC_HIERARCHICAL ) {
			// Hierarchical version
			float32 startTime = clock();
			vertices = hierarchicalTriangulate( bbox, 0, surfelPS, surfelKNN, surfelTopology, samplingSpacing );
			debugOutput << "  Meshing (hierarchical): time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; debugOutput.flush();
		}

		if ( triangulationSettings->doClipBorder ){
			vertices = clipBorder( vertices, clipInfluence, borderPS, borderTopology, borderOctree, samplingSpacing );
		}

		// Create mesh
		urShapeMesh = new UnstructuredInCoreTriangleMesh();
		VertexDescriptor vdesc;
		vdesc.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vdesc.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vdesc.pushAttrib( VAD( "texCoord", 2, VAD::DATA_FORMAT_FLOAT32 ) );
		VertexDescriptor idesc;
		idesc.pushAttrib( VAD( "index", 3, VAD::DATA_FORMAT_INT32 ) );
		urShapeMesh->clearAndSetup( &vdesc, vertices.size(), &idesc, vertices.size()/3 );
		urShapeMesh->setMaterialIndex( triangulationSettings->materialIndex );
		AAT VERTEX_POSITION = urShapeMesh->getPointSet()->getAAT( "position" );
		AAT VERTEX_NORMAL = urShapeMesh->getPointSet()->getAAT( "normal" );
		AAT VERTEX_TEXCOORD = urShapeMesh->getPointSet()->getAAT( "texCoord" );
		AAT TRIANGLE_INDICES = urShapeMesh->getTriangles()->getAAT( "index" );
		// Set mesh content
		for( card32 i=0; i < vertices.size(); i++ ){
			urShapeMesh->getPointSet()->set3f( i, VERTEX_POSITION, vertices[i] );
		}
		for( card32 i=0; i < vertices.size()/3; i++ ){
			urShapeMesh->getTriangles()->set3i( i, TRIANGLE_INDICES, makeVector3i( i*3, i*3+1, i*3+2 ) );
		}
		// Finalize mesh
		urShapeMesh->snapMesh( mcCubeSize * 0.1f );
		urShapeMesh->closeMesh();
		urShapeMesh->calculateVertexNormals();
		// Compute normals from urshape
		for ( unsigned int vIndex = 0; vIndex < urShapeMesh->getPointSet()->getNumEntries(); vIndex++ ){
			const Vector3f v = urShapeMesh->getPointSet()->get3f( vIndex, VERTEX_POSITION );
			Vector3f normal = makeVector3f( 0,0,0 );
			surfelKNN.setTargetPoint( v );
			surfelKNN.getNextNearestPointIndex( nIndex, found );
			const Vector3f seedPoint = surfelPS->get3f( nIndex, POSITION );
			float32 sqrD = 0;
			float32 summedWeight = 0;
			while( found && sqrD < sqr( samplingSpacing * triangulationSettings->normalInfluence ) ){
				const Vector3f np = surfelPS->get3f( nIndex, POSITION );
				sqrD = ( np - seedPoint ) * ( np - seedPoint );
				float32 weight = exp( - sqrD / sqr( samplingSpacing * triangulationSettings->normalInfluence ) );
				summedWeight += weight;
				normal += surfelPS->get3f( nIndex, NORMAL ) * weight;
				surfelKNN.getNextNearestPointIndex( nIndex, found );
			}
			normal /= summedWeight;
			urShapeMesh->getPointSet()->set3f( vIndex, VERTEX_NORMAL, normal );
		}
	}

	if ( triangulationSettings->doParameterization ){
		parameterize( urShapeMesh );
	}

	// Create animation node 
	SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
	animNode->setName( "meshes" );
	animNode->setVisible( true );
	if ( triangulationSettings->doExtrapolation ){
		if ( urShapeMesh == NULL ){
			SceneGraphNode* node = getSceneGraphNodeByName( scene, "root/backend/meshes/urShapeMesh" );
			SGObjectNode* objNode = dynamic_cast<SGObjectNode*>( node );
			urShapeMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( objNode->getSceneObject() );
		}
		extrapolate( startFrame, endFrame, animNode, urShapeMesh, samplingSpacing, seq, evaluator );
	}

	// Save border point cloud
	if ( borderPC != NULL ){
		SGObjectNode* borderNode = new SGObjectNode();
		borderNode->clearAndSetup( borderPC );
		borderNode->setName( "border" );
		animNode->addNode( borderNode );
	}

	// Save urshape mesh
	if ( urShapeMesh != NULL ){
		SGObjectNode* urShapeNode = new SGObjectNode();
		urShapeNode->clearAndSetup( urShapeMesh );
		urShapeNode->setName( "urShapeMesh" );
		animNode->addNode( urShapeNode );
	}

	// Cleanup
	if ( surfelOctree != NULL ){
		delete surfelOctree;
		surfelOctree = NULL; 
	}
	if ( borderOctree != NULL ){
		delete borderOctree;
		borderOctree = NULL; 
	}

	return animNode;
}

vector<Vector3f> SurfelGraphTriangulator::hierarchicalTriangulate( 
	BoundingBox3f bbox, 
	const card32 level, 
	PointSet* surfelPS, 
	PointSetKNNQuery& surfelKNN, 
	InCorePCTopologyGraph* surfelTopology,
	const float32 samplingSpacing )
{
	vector<Vector3f> result; result.clear();
	Vector3f corners[8];
	float32 values[8];
	corners[0] = makeVector3f( bbox.lowerCorner[0], bbox.lowerCorner[1], bbox.lowerCorner[2] );
	corners[1] = makeVector3f( bbox.upperCorner[0], bbox.lowerCorner[1], bbox.lowerCorner[2] );
	corners[2] = makeVector3f( bbox.upperCorner[0], bbox.upperCorner[1], bbox.lowerCorner[2] );
	corners[3] = makeVector3f( bbox.lowerCorner[0], bbox.upperCorner[1], bbox.lowerCorner[2] );
	corners[4] = makeVector3f( bbox.lowerCorner[0], bbox.lowerCorner[1], bbox.upperCorner[2] );
	corners[5] = makeVector3f( bbox.upperCorner[0], bbox.lowerCorner[1], bbox.upperCorner[2] );
	corners[6] = makeVector3f( bbox.upperCorner[0], bbox.upperCorner[1], bbox.upperCorner[2] );
	corners[7] = makeVector3f( bbox.lowerCorner[0], bbox.upperCorner[1], bbox.upperCorner[2] );
	if ( triangulateCube( corners, surfelPS, surfelKNN, samplingSpacing ) ){
		if ( level == triangulationSettings->mcOctreeLevel ){
			for ( unsigned int n = 0; n < 8; n++ ){
				values[n] = getValue( corners[n], surfelKNN, surfelPS, surfelTopology, samplingSpacing );
			}
			MarchingCubes3D::triangulate( corners, values, 0, result );
		} else {
			// Subdivide
			float32 tau_subdiv = bbox.getMaxSideLength() / 2;
			BoundingBox3f bbox_subdiv;
			for ( unsigned int i = 0; i < 2; i++ ){
				bbox_subdiv.lowerCorner[0] = bbox.lowerCorner[0] + i*tau_subdiv;
				bbox_subdiv.upperCorner[0] = bbox_subdiv.lowerCorner[0] + tau_subdiv;
				for ( unsigned int j = 0; j < 2; j++ ){
					bbox_subdiv.lowerCorner[1] = bbox.lowerCorner[1] + j*tau_subdiv;
					bbox_subdiv.upperCorner[1] = bbox_subdiv.lowerCorner[1] + tau_subdiv;
					for ( unsigned int k = 0; k < 2; k++ ){
						bbox_subdiv.lowerCorner[2] = bbox.lowerCorner[2] + k*tau_subdiv;
						bbox_subdiv.upperCorner[2] = bbox_subdiv.lowerCorner[2] + tau_subdiv;
						vector<Vector3f> verts = hierarchicalTriangulate( bbox_subdiv, level+1, surfelPS, surfelKNN, surfelTopology, samplingSpacing );
						for ( unsigned int n = 0; n < verts.size(); n++ ){
							result.push_back( verts[n] );
						}
					}
				}
			}
		}
	}
	return result;
}


BoundingBox3f SurfelGraphTriangulator::getMCDomain( PointSet* ps )
{
	BoundingBox3f bbox;
	bbox = ps->getBoundingBox();
	bbox.enlarge( 1.05f );
	const float32 d = bbox.getMaxSideLength()/2.0f;
	const Vector3f center = bbox.getCenter();
	bbox.lowerCorner = center - makeVector3f( d, d, d );
	bbox.upperCorner = center + makeVector3f( d, d, d );
	return bbox;
}

bool SurfelGraphTriangulator::triangulateCube( 
	Vector3f* corners, 
	PointSet* surfelPS, 
	PointSetKNNQuery& knn, 
	const float32 samplingSpacing ) const
{
	BoundingBox3f extendedBox;
	const Vector3f center = ( corners[0] + corners[6] ) / 2.0f;
	float32 side_length = fabs( corners[6][0] - corners[0][0] ) + samplingSpacing * 3;
	extendedBox.lowerCorner = center - makeVector3f( side_length, side_length, side_length ) * 0.5f;
	extendedBox.upperCorner = center + makeVector3f( side_length, side_length, side_length ) * 0.5f;
	AAT POSITION = surfelPS->getAAT( "position" );
	card32 nIndex;
	bool found;
	for ( unsigned int n = 0; n < 8; n++ ){
		knn.setTargetPoint( corners[n] );
		knn.getNextNearestPointIndex( nIndex, found );
		if ( found ){
			const Vector3f closestPoint = surfelPS->get3f( nIndex, POSITION );
			if ( extendedBox.inBoundingBox( closestPoint ) ) return true;
		}
	}
	return false;
}

float32 SurfelGraphTriangulator::getValue( 
	const Vector3f x, 
	PointSetKNNQuery& knn, 
	PointSet* ps,
	InCorePCTopologyGraph* topology, 
	const float32 samplingSpacing ) const 
{
	if ( ps == NULL ) return -1;
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT TANGENT_U = ps->getAAT( "tangent_u" );
	AAT TANGENT_V = ps->getAAT( "tangent_v" );
	AAT COEFFS = ps->getAAT( "coeffs" );
	card32 nIndex;
	bool found;
	knn.setTargetPoint( x );
	knn.getNextNearestPointIndex( nIndex, found );
	const Vector3f closestPoint = ps->get3f( nIndex, POSITION );
	const Vector3f closestNormal = ps->get3f( nIndex, NORMAL );
	Vector3f seedPoint = x;
	projectPointPlane( Plane( closestNormal, closestPoint ), x, seedPoint );

	vector<card32> indices; indices.clear();
	TopNeighborhoodSphereSearch topoSphereSearch;
	topoSphereSearch.setup( 
		topology, 
		ps, 
		nIndex, 
		seedPoint, 
		samplingSpacing * triangulationSettings->influence,
		POSITION );
	while( !topoSphereSearch.atEnd() ){
		indices.push_back( topoSphereSearch.getCurrentIndex() );
		topoSphereSearch.next();
	}
	if ( indices.size() < 1 ) indices.push_back( nIndex );
	float32 summedWeight = 0;
	float32 value = 0;
	const float32 sqrSamplingSpacing = sqr( samplingSpacing );
	const float32 sqrInfluence = sqr( samplingSpacing * triangulationSettings->influence );
	for ( unsigned int index = 0; index < indices.size(); index++ ){
		nIndex = indices[index];
		const Vector3f p = ps->get3f( nIndex, POSITION );
		const Vector3f tangent_u = ps->get3f( nIndex, TANGENT_U );
		const Vector3f tangent_v = ps->get3f( nIndex, TANGENT_V );
		const Vector3f normal = ps->get3f( nIndex, NORMAL );
		float32 dPlane = x * normal - p * normal;  	
		float32 sqrD = ( seedPoint - p ) * ( seedPoint - p );
		float32 weight = exp( - sqrD / sqrSamplingSpacing );
		//float32 weight = exp( - sqrD / sqrInfluence );
		summedWeight += weight;
		value += dPlane * weight;
	}
	return value / summedWeight;
}

vector<Vector3f> SurfelGraphTriangulator::clipBorder( 
	vector<Vector3f>& originalVertices, 
	const float32 maxDist, 
	PointSet* borderPS, 
	InCorePCTopologyGraph* borderTopology, 
	AttachedIndexedOctree* borderOctree,
	const float32 samplingSpacing )
{
	card32 borderSeedIndex;
	bool found;

	AAT BORDER_POSITION = borderPS->getAAT( "position" );
	vector<Vector3f> clippedVertices; clippedVertices.clear();
	bool invalidTriangle = true;
	for ( unsigned int tIndex = 0; tIndex < originalVertices.size()/3; tIndex++ ){
		Vector3f triangleCenter = ( originalVertices[tIndex*3] + originalVertices[tIndex*3+1] + originalVertices[tIndex*3+2] ) / 3.0f;
		
		/*
		PointSetKNNQuery borderKNN( borderPS, borderOctree, triangleCenter, BORDER_POSITION );
		borderKNN.getNextNearestPointIndex( borderSeedIndex, found );
		*/

		for ( unsigned int vIndex = 0; vIndex < 3; vIndex++ ){
			Vector3f vertex = originalVertices[tIndex*3+vIndex];

			PointSetKNNQuery borderKNN( borderPS, borderOctree, vertex, BORDER_POSITION );
			borderKNN.getNextNearestPointIndex( borderSeedIndex, found );
			if ( found ){

				Plane clippingPlane;
				if ( getClippingPlane( vertex, borderSeedIndex, maxDist*1.2f, clippingPlane, borderPS, borderTopology, borderOctree, samplingSpacing ) ){
					const float32 d = clippingPlane.getDistanceFromPlane( vertex );
					if ( d > 0 ) projectPointPlane( clippingPlane, vertex, originalVertices[tIndex*3+vIndex] );
					else invalidTriangle = false;
				} else {
					invalidTriangle = false;
				}
			} else {
				invalidTriangle = false;
				break;
			}
		}
		if ( !invalidTriangle ){
			for ( unsigned int vIndex = 0; vIndex < 3; vIndex++ ){
				clippedVertices.push_back( originalVertices[tIndex*3+vIndex] );
			}
		}
	}
	return clippedVertices;
}

void SurfelGraphTriangulator::parameterize( UnstructuredInCoreTriangleMesh* mesh )
{	
	if ( mesh == NULL ) return;
	parameterizeTriangleMesh( mesh, triangulationSettings->paramerizationMode, triangulationSettings->texCoordScale );
	mesh->setMaterialIndex( triangulationSettings->materialIndex );	
}

UnstructuredInCorePointCloud* SurfelGraphTriangulator::detectBorderSurfels( 
	UnstructuredInCorePointCloud* urShapePC, 
	const float32 samplingSpacing )
{
	if ( urShapePC == NULL ) return NULL;
	PointSet* surfelPS = urShapePC->getPointSet();
	if ( surfelPS == NULL ) return NULL;
	InCorePCTopologyGraph* surfelTopology = getTopology( urShapePC );
	if ( surfelTopology == NULL ) return NULL;
	AAT SURFEL_POSITION = surfelPS->getAAT( "position" );
	AAT SURFEL_TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT SURFEL_TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT SURFEL_NORMAL = surfelPS->getAAT( "normal" );
	AAT SURFEL_FLAGS = surfelPS->getAAT( "flags" );
	AttachedIndexedOctree* surfelOctree = PointSetKNNQuery::createOctree( surfelPS, 32 );
	PointSetKNNQuery surfelKNN( surfelPS, surfelOctree, makeVector3f( 0,0,0 ), SURFEL_POSITION );
	TopNeighborhoodSphereSearch surfelTopoSphereSearch, borderTopoSphereSearch;
	Matrix3f T;
	card32 nBorderIndex, nSurfelIndex;
	bool found;

	// Find surfel border candidates
	vector<card32> borderSurfels; borderSurfels.clear();
	for ( unsigned int sIndex = 0; sIndex < surfelPS->getNumEntries(); sIndex++ ){
		const Vector3f surfelP = surfelPS->get3f( sIndex, SURFEL_POSITION );
		T[0] = surfelPS->get3f( sIndex, SURFEL_TANGENT_U );
		T[1] = surfelPS->get3f( sIndex, SURFEL_TANGENT_V );
		T[2] = surfelPS->get3f( sIndex, SURFEL_NORMAL );
		vector<card32> neighbors; neighbors.clear();
		surfelTopoSphereSearch.setup( 
			surfelTopology, 
			surfelPS, 
			sIndex, 
			surfelP, 
			samplingSpacing * triangulationSettings->borderInfluence,
			SURFEL_POSITION );
		while( !surfelTopoSphereSearch.atEnd() ){
			const card32 nSurfelIndex = surfelTopoSphereSearch.getCurrentIndex(); 
			if ( nSurfelIndex != sIndex ) neighbors.push_back( nSurfelIndex );
			surfelTopoSphereSearch.next();
		}
		bool *cones = new bool[triangulationSettings->borderNumCones]; 
		for ( unsigned int cIndex = 0; cIndex < 6; cIndex++ ) cones[cIndex] = false;
		card32 numSpatialNeighbors = 0;
		for ( unsigned int n = 0; n < neighbors.size(); n++ ){
			const Vector3f np = surfelPS->get3f( neighbors[n], SURFEL_POSITION );
			const Vector3f np_local = T.transpose() * ( np - surfelP );
			const float32 alpha = getAngle( makeVector2f( np_local[0], np_local[1] ) );
			const card32 alpha_index = card32( ( triangulationSettings->borderNumCones * alpha ) / ( M_PI * 2.0f ) );
			if ( !cones[alpha_index] ){
				cones[alpha_index] = true;
				numSpatialNeighbors++;
			}
		}
		delete [] cones;
		if ( numSpatialNeighbors < triangulationSettings->borderConesThreshold ){
			borderSurfels.push_back( sIndex );
		}
	}

	// Create border point set
	PointSet* borderPS = new PointSet();
	VertexDescriptor vd;
	vd.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( VAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	borderPS->clearAndSetup( 1, borderSurfels.size(), &vd );
	UnstructuredInCorePointCloud* borderPC = new UnstructuredInCorePointCloud();
	InCorePCTopologyGraph* borderTopology = new InCorePCTopologyGraph();
	VertexDescriptor topotVD;
	topotVD.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
	borderTopology->clearAndSetup( &topotVD );
	AttachedDataContainer *attachments = borderPC->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>( attachments->getData("TopologyAtt"));
	if ( topAtt == NULL ){
		topAtt = new TopologyAttachment();
		topAtt->setup( "TopologyAtt", AttachedData::ADF_PERSISTENT );
		attachments->attachData( topAtt );
	}
	topAtt->setTopology( borderTopology );
	borderPC->setPointSet( borderPS );
	AAT BORDER_POSITION = borderPS->getAAT( "position" );
	AAT BORDER_NORMAL = borderPS->getAAT( "normal" );
	for ( unsigned int n = 0; n < borderSurfels.size(); n++ ){
		borderPS->set3f( n, BORDER_POSITION, surfelPS->get3f( borderSurfels[n], SURFEL_POSITION ) );
		borderPS->set3f( n, BORDER_NORMAL, surfelPS->get3f( borderSurfels[n], SURFEL_NORMAL ) );
	}

	// Connect border points topologically
	const float32 sqrTopoDist = sqr( samplingSpacing * triangulationSettings->borderTopoInfluence );
	AttachedIndexedOctree* borderOctree = PointSetKNNQuery::createOctree( borderPS, 32 );
	PointSetKNNQuery borderKNN( borderPS, borderOctree, makeVector3f( 0,0,0 ), BORDER_POSITION );
	for ( unsigned int bIndex = 0; bIndex < borderPS->getNumEntries(); bIndex++ ){
		const Vector3f borderP = borderPS->get3f( bIndex, BORDER_POSITION );
		borderKNN.setTargetPoint( borderP );
		
		// Init
		float32 sqrD = 0;
		found = true;
		nBorderIndex = bIndex;
		Vector3f nBorderP = borderPS->get3f( nBorderIndex, BORDER_POSITION );

		while( found && ( sqrD < sqrTopoDist ) ){
			if ( nBorderIndex != bIndex ){
				borderTopology->addEdge( nBorderIndex, bIndex );
			}
			// Next point
			borderKNN.getNextNearestPointIndex( nBorderIndex, found );
			nBorderP = borderPS->get3f( nBorderIndex, BORDER_POSITION );
			sqrD = ( nBorderP - borderP ) * ( nBorderP - borderP );
		}
		// Determine border normal
		surfelKNN.setTargetPoint( borderP );		
		surfelKNN.getNextNearestPointIndex( nSurfelIndex, found );
		const Vector3f nSurfel = surfelPS->get3f( nSurfelIndex, SURFEL_POSITION );
		surfelTopoSphereSearch.setup( 
			surfelTopology, 
			surfelPS, 
			nSurfelIndex, 
			nSurfel, 
			samplingSpacing * triangulationSettings->borderInfluence,
			SURFEL_POSITION );
		Vector3f c = makeVector3f( 0,0,0 );
		card32 numSurfelNeighbors = 0;
		while( !surfelTopoSphereSearch.atEnd() ){
			nSurfelIndex = surfelTopoSphereSearch.getCurrentIndex(); 
			c += surfelPS->get3f( nSurfelIndex, SURFEL_POSITION );
			surfelTopoSphereSearch.next();
			numSurfelNeighbors++;
		}
		c /=  float32( numSurfelNeighbors );
		Vector3f borderNormal = c - borderP;
		borderNormal.normalize();
		borderPS->set3f( bIndex, BORDER_NORMAL, -borderNormal );
	}

	// Cleanup
	delete surfelOctree;
	delete borderOctree;
	return pruneSmallBorderComponents( borderPC );
}

UnstructuredInCorePointCloud* SurfelGraphTriangulator::pruneSmallBorderComponents( UnstructuredInCorePointCloud* oldBorderPC )
{
	if ( oldBorderPC == NULL ) return NULL;
	InCorePCTopologyGraph* oldBorderTopology = getTopology( oldBorderPC );
	PointSet* oldBorderPS = oldBorderPC->getPointSet();

	PointSet* newBorderPS = new PointSet();
	newBorderPS->clearAndSetup( 1, 0 , oldBorderPS->getDescr() );
	UnstructuredInCorePointCloud* newBorderPC = new UnstructuredInCorePointCloud();
	InCorePCTopologyGraph* newBorderTopology = new InCorePCTopologyGraph();
	VertexDescriptor topotVD;
	topotVD.pushAttrib( VAD( "vindex", 2, VAD::DATA_FORMAT_INT32 ) );
	newBorderTopology->clearAndSetup( &topotVD );
	AttachedDataContainer *attachments = newBorderPC->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>( attachments->getData( "TopologyAtt" ) );
	if ( topAtt == NULL ){
		topAtt = new TopologyAttachment();
		topAtt->setup( "TopologyAtt", AttachedData::ADF_PERSISTENT );
		attachments->attachData( topAtt );
	}
	topAtt->setTopology( newBorderTopology );
	newBorderPC->setPointSet( newBorderPS );
	AAT BORDER_POSITION = newBorderPS->getAAT( "position" );
	AAT BORDER_NORMAL = newBorderPS->getAAT( "normal" );
	AAT BORDER_FLAGS = newBorderPS->getAAT( "flags" );
	// flags == 0: not visited
	// flags == 1: valid connection
	// flags == 2: invalid connection
	for ( unsigned int bIndex = 0; bIndex < oldBorderPS->getNumEntries(); bIndex++ ){
		oldBorderPS->set1i( bIndex, BORDER_FLAGS, 0 );
	}
	for ( unsigned int bIndex = 0; bIndex < oldBorderPS->getNumEntries(); bIndex++ ){
		if ( oldBorderPS->get1i( bIndex, BORDER_FLAGS ) == 0 ){
			FifoStack<card32> stack; stack.clear();
			vector<card32> component; component.clear();

			// Init
			stack.push( bIndex );
			oldBorderPS->set1i( bIndex, BORDER_FLAGS, 1 );
			component.push_back( bIndex );

			while( !stack.empty() ){
				const card32 index = stack.pop();
				for ( unsigned int n = 0; n < oldBorderTopology->getNumAdjacentVertices( index ); n++ ){
					const card32 nIndex = oldBorderTopology->getAdjacentVertex( index, n );
					if ( oldBorderPS->get1i( nIndex, BORDER_FLAGS ) == 0 ){
						stack.push( nIndex );
						component.push_back( nIndex );
						oldBorderPS->set1i( nIndex, BORDER_FLAGS, 1 );
					}
				}
			}

			if ( component.size() < triangulationSettings->connectedComponentSize ){
				for ( unsigned int n = 0; n < component.size(); n++ ){
					oldBorderPS->set1i( component[n], BORDER_FLAGS, 2 );
				}
			} else {
				for ( unsigned int n = 0; n < component.size(); n++ ){
					oldBorderPS->set1i( component[n], BORDER_FLAGS, 1 );
				}
			}
		}
	}

	vector<card32> mapOld2New; mapOld2New.resize( oldBorderPS->getNumEntries() );

	// Remove loosely connected components
	card32 newBIndex = 0;
	for ( unsigned int bIndex = 0; bIndex < oldBorderPS->getNumEntries(); bIndex++ ){
		mapOld2New[bIndex] = newBIndex;
		if ( oldBorderPS->get1i( bIndex, BORDER_FLAGS ) == 1 ){
			newBorderPS->changeHeight( newBIndex+1 );
			newBorderPS->set3f( newBIndex, BORDER_POSITION, oldBorderPS->get3f( bIndex, BORDER_POSITION ) );
			newBorderPS->set3f( newBIndex, BORDER_NORMAL, oldBorderPS->get3f( bIndex, BORDER_NORMAL ) );
			for ( unsigned int n = 0; n < oldBorderTopology->getNumAdjacentVertices( bIndex ); n++ ){
				const card32 nIndex = oldBorderTopology->getAdjacentVertex( bIndex, n );
				if ( nIndex < bIndex ){
					newBorderTopology->addEdge( newBIndex, mapOld2New[nIndex] );
				}
			}
			newBIndex++;
		}
	}

	delete oldBorderPC;
	return newBorderPC;
}

float32 SurfelGraphTriangulator::getAngle( const Vector2f v )
{
	Vector2f v_ = v / norm( v );
	const float32 alpha = acos( v_ * makeVector2f( 1,0 ) );
	const float32 b = v_ * makeVector2f( 0,1 );
	return ( b > 0 )? alpha : 2.0f * M_PI - alpha;
}

bool SurfelGraphTriangulator::getClippingPlane( 
	const Vector3f p, 
	const card32 borderSeedIndex,
	const float32 maxDist,
	Plane& clippingPlane, 
	PointSet* borderPS,
	InCorePCTopologyGraph* borderTopology,
	AttachedIndexedOctree* borderOctree, 
	const float32 samplingSpacing )
{
	if ( borderPS == NULL ) return false;
	if ( borderOctree == NULL ) return false;
	AAT BORDER_POSITION = borderPS->getAAT( "position" );
	AAT BORDER_NORMAL = borderPS->getAAT( "normal" );
	const Vector3f nBorder = borderPS->get3f( borderSeedIndex, BORDER_POSITION );
	if ( norm( nBorder - p ) > maxDist ) return false;
	TopNeighborhoodSphereSearch borderTopoSphereSearch;
	borderTopoSphereSearch.setup( 
		borderTopology, 
		borderPS, 
		borderSeedIndex, 
		nBorder, 
		samplingSpacing * triangulationSettings->borderInfluence,
		BORDER_POSITION );
	vector<card32> borderIndices; borderIndices.clear();
	while ( !borderTopoSphereSearch.atEnd() ){
		borderIndices.push_back( borderTopoSphereSearch.getCurrentIndex() );
		borderTopoSphereSearch.next();
	}
	if ( borderIndices.size() < 1 ) return false;
	Vector3f p_ = makeVector3f( 0,0,0 );
	Vector3f normal_ = makeVector3f( 0,0,0 );
	float32 summedWeight = 0;
	for ( unsigned int n = 0; n < borderIndices.size(); n++ ){
		const card32 borderIndex = borderIndices[n];
		const Vector3f borderP = borderPS->get3f( borderIndex, BORDER_POSITION );
		const float32 sqrD = ( borderP - p ) * ( borderP - p );
		const float32 weight = exp( - sqrD / sqr( samplingSpacing * triangulationSettings->borderInfluence ) );
		summedWeight += weight;
		p_ += borderP * weight;
		normal_ += borderPS->get3f( borderIndex, BORDER_NORMAL );
	}
	p_ /= summedWeight;
	normal_.normalize();
	clippingPlane = Plane( normal_, p_ );
	return true; 
}

void SurfelGraphTriangulator::extrapolate( 
	card32 startFrame,
	card32 endFrame,
	SGSimpleAnimationNode* animNode, 
	UnstructuredInCoreTriangleMesh* urShapeMesh,
	const float32 samplingSpacing, 
	ARSequence* seq, 
	DeformationEvaluator* evaluator )
{
	if ( animNode == NULL ) return;
	if ( urShapeMesh == NULL ) return;
	if ( seq == NULL ) return;

	ARDeformationOptimizationContext *context = seq->getDeformation();
	UnstructuredInCorePointCloud *urshape = dynamic_cast<UnstructuredInCorePointCloud*>( seq->getUrshape() );
	if ( urshape == NULL ) return;
	card32 findex = 0;
	PointSet* ps = urshape->getPointSet();
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT TANGENT_U = ps->getAAT( "tangent_u" );
	AAT TANGENT_V = ps->getAAT( "tangent_v" );
	//TopNeighborhoodKNNSearch topoKNN( urshape->getPointSet(), getTopology( urshape ), 0, POSITION );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 0,0,0 ), POSITION );
	card32 nIndex;
	bool found;
	const float32 sqrInfluence = sqr( samplingSpacing * triangulationSettings->influence );
	progressWindow->pushStep( true, "Extrapolation" );
	for ( card32 time = startFrame; time <= endFrame; time++ ) {
		UnstructuredInCoreTriangleMesh* frameMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( urShapeMesh->copy( NULL ) );
		AAT MESH_POSITION = frameMesh->getAAT( "position" );
		AAT MESH_NORMAL = frameMesh->getAAT( "normal" );
		evaluator->beginQuery( time-startFrame );
		for ( unsigned int vIndex = 0; vIndex < frameMesh->getPointSet()->getNumEntries(); vIndex++ ){
			const Vector3f vertexP = frameMesh->getPointSet()->get3f( vIndex, MESH_POSITION );
			knn.setTargetPoint( vertexP );
			knn.getNextNearestPointIndex( nIndex, found );
			//topoKNN.reset( nIndex );
			float32 summedWeight = 0;
			Vector3f deformedPos = makeVector3f( 0,0,0 );
			Vector3f deformedNormal = makeVector3f( 0,0,0 );
			for ( unsigned int n = 0; n < 5; n++ ){
				const Vector3f np = ps->get3f( nIndex, POSITION );
				Matrix3f T_urshape, T_deformed;
				T_urshape[0] = ps->get3f( nIndex, TANGENT_U );
				T_urshape[1] = ps->get3f( nIndex, TANGENT_V );
				T_urshape[2] = ps->get3f( nIndex, NORMAL );
				ARConstraintAnchor at;
				at.topologicalAnchor = nIndex;
				at.pos = np;
				float32 weight = exp( - ( np - vertexP ) * ( np - vertexP ) / sqrInfluence  );
				summedWeight += weight;
				Matrix3f jac = evaluator->getSpatialGradient( at );
				const Matrix3f transform = invertMatrix( jac ).transpose();
				T_deformed[0] = jac * ps->get3f( nIndex, TANGENT_U );
				T_deformed[1] = jac * ps->get3f( nIndex, TANGENT_V );
				T_deformed[2] = transform * ps->get3f( nIndex, NORMAL );
				Vector3f offset = vertexP - np;
				offset = T_deformed * T_urshape.transpose() * offset;
				deformedPos += ( evaluator->getPosition( at ) + offset ) * weight;
				deformedNormal += T_deformed[2] * weight; 
				//topoKNN.next();
				//nIndex = topoKNN.getCurrentIndex();
				knn.getNextNearestPointIndex( nIndex, found );
			}
			deformedPos /= summedWeight;
			deformedNormal.normalize();
			frameMesh->getPointSet()->set3f( vIndex, MESH_POSITION, deformedPos );
			frameMesh->getPointSet()->set3f( vIndex, MESH_NORMAL, deformedNormal );
		}
		evaluator->endQuery();
		SGObjectNode* frameNode = new SGObjectNode();
		char tID[256];
		itoa( time, tID, 10 );
		frameNode->setName( "mesh_" + string( tID ) );
		if ( triangulationSettings->saveMeshOOC ){
			GenericSwapAllOutOfCorePointCloud *newGSA = new GenericSwapAllOutOfCorePointCloud();
			string m_oocSwapFileNamePath = "C:/Temp/";
			newGSA->clearAndSetup( frameMesh, triangulationSettings->oocFilename + frameNode->getName()+"_mesh.swap.object", true );
			newGSA->setMaterialIndex( triangulationSettings->materialIndex );
			frameNode->clearAndSetup( newGSA );
			delete frameMesh;	
		} else {
			frameNode->clearAndSetup( frameMesh );
		}
		animNode->addNode( frameNode );
		progressWindow->progress( float32 ( time - startFrame ) * 100.0f / float32( endFrame - startFrame ) );
	}
	progressWindow->popSteps();

	delete octree;
}