//---------------------------------------------------------------------------
#ifndef SurfelGraphTriangulatorH
#define SurfelGraphTriangulatorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "Scene.h"
#include "SGESurfelGraph.h"
#include "ARSettings.h"
#include "SGSimpleAnimationNode.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "ARSequence.h"
#include "BackendSettings.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------

namespace geoxl {

///
///   Triangulation functionality for a surfel graph
///
class DEFORMATION_API SurfelGraphTriangulator: public Persistent
{
	GEOXL_CLASS( SurfelGraphTriangulator )

	private:
		BackendSettings* triangulationSettings;

		/* HELPER METHODS */
		/// Helper method: compute orientation of vector in plane
		float32 getAngle( const Vector2f v );
		// Parameterize mesh
		void parameterize( UnstructuredInCoreTriangleMesh* mesh );
		// Detect border points
		UnstructuredInCorePointCloud* detectBorderSurfels( UnstructuredInCorePointCloud* urShapePC, const float32 samplingSpacing );
		// Compute the extrapolation of the urshape to the entire animation
		void extrapolate( 
			card32 startFrame,
			card32 endFrame,
			SGSimpleAnimationNode* animNode, 
			UnstructuredInCoreTriangleMesh* urShapeMesh, 
			const float32 samplingSpacing,
			ARSequence* seq, 
			DeformationEvaluator *evaluator );
		// Prune loosely connected border components
		UnstructuredInCorePointCloud* pruneSmallBorderComponents( UnstructuredInCorePointCloud* oldBorderPC );
		// Clip vertices, prune completely clipped triangles
		vector<Vector3f> clipBorder( 
			vector<Vector3f>& originalVertices, 
			const float32 maxDist, 
			PointSet* borderPS, 
			InCorePCTopologyGraph* borderTopology, 
			AttachedIndexedOctree* borderOctree,
			const float32 samplingSpacing );
		vector<Vector3f> hierarchicalTriangulate( 
			BoundingBox3f bbox, 
			const card32 level,
			PointSet* surfelPS, 
			PointSetKNNQuery& knn, 
			InCorePCTopologyGraph* surfelTopology,
			const float32 samplingSpacing );

	public:
		SurfelGraphTriangulator();
		~SurfelGraphTriangulator();

		/***----------------- MADE PUBLIC FOR DEBUGGING PURPOSES ---------------------------------***/

		// Return a clipping plane at position p (if required) for a dataset with border
		bool getClippingPlane( 
			const Vector3f x, 
			const card32 borderSeedIndex,
			const float32 maxDist,
			Plane& clippingPlane, 
			PointSet* borderPS,
			InCorePCTopologyGraph* borderTopology,
			AttachedIndexedOctree* borderOctree, 
			const float32 samplingSpacing );
		// Return the triangulation domain
		BoundingBox3f getMCDomain( PointSet* ps );
		// Check if cube is close enough to surface to be triangulated
		bool triangulateCube( 
			Vector3f* corners, 
			PointSet* surfelPS, 
			PointSetKNNQuery& knn, 
			const float32 samplingSpacing ) const;
		// Return implicit function value at position p
		float32 getValue( 
			const Vector3f p, 
			PointSetKNNQuery& knn,
			PointSet* ps,
			InCorePCTopologyGraph* topology,
			const float32 samplingSpacing ) const;
		// Return the cube size
		float32 getMCubeSize( BoundingBox3f bbox, const card32 mcLevel ){
			return bbox.getMaxSideLength() / power( 2, mcLevel );
		}
		inline float32 SurfelGraphTriangulator::getClipInfluence( const float32 mcCubeSize, const float32 samplingSpacing ) const
		{
			return triangulationSettings->clipInfluence * sqrt( 3.0f ) * ( mcCubeSize + samplingSpacing * 2 );
		}

		/***----------------- END MADE PUBLIC FOR DEBUGGING PURPOSES ---------------------------------***/


		// Compute a triangulation for the surfelGraph
		SGSimpleAnimationNode* triangulate( 
			card32 startFrame, 
			card32 endFrame,
			UnstructuredInCorePointCloud* urShapePC, 
			const float32 samplingSpacing,
			ARSequence* seq = NULL, 
			DeformationEvaluator *evaluator = NULL, 
			Scene* scene = NULL );

		/* MEMBER ACCESS */
		inline string getMeshNodeName() const { return triangulationSettings->meshNodeName; }
		inline void setBackendSettings( BackendSettings* triangulationSettings ){ this->triangulationSettings = triangulationSettings; }
};

}

#endif