//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SGEMeshlessUrshape.h"
#include "SGListNode.h"
#include "SGobjectNode.h"
#include "DeformationTools.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "UnstructuredInCorePointCloud.h"
#include "InCorePCTopologyGraph.h"
#include "AttachedNodeInformation.h"
#include "PointSetSphericalQuery.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( SGEMeshlessUrshape , 0 ) {
	BEGIN_CLASS_INIT( SGEMeshlessUrshape );
}

// ----

SGEMeshlessUrshape::SGEMeshlessUrshape() {
	geometryChannels      = NULL;
	geometryGraphChannels = NULL;
	mmNodesChannels       = NULL;
	mmNodesGraphChannels  = NULL;
}

SGEMeshlessUrshape::~SGEMeshlessUrshape() {
	delete geometryChannels;
	delete geometryGraphChannels;
	delete mmNodesChannels;
	delete mmNodesGraphChannels;
}


// ---


void SGEMeshlessUrshape::setup(
	VertexDescriptor *geometryChannels,
	VertexDescriptor *geometryGraphChannels,
	VertexDescriptor *mmNodesChannels,
	VertexDescriptor *mmNodesGraphChannels)
{
	this->geometryChannels      = geometryChannels;
	this->geometryGraphChannels = geometryGraphChannels;
	this->mmNodesChannels       = mmNodesChannels;
	this->mmNodesGraphChannels  = mmNodesGraphChannels;
}


bool SGEMeshlessUrshape::check(string &msg, bool checkMMNodes) const
{
	if (!SceneGraphEntity::check(msg)) {
		return false;
	}
	
	Matrix4f forget;
	SceneGraphNode *node;
   getSGNodeTransformation(getScene(), getPath(), forget, node);
	SGListNode *list = dynamic_cast<SGListNode*>(node);
	if (!list) {
		msg = "path is not pointing to a list node (check scene graph structure)";
		return false;
	}
	if (checkMMNodes) {
		if (!(list->getNumChildNodes(NULL) == 2)) {
			msg = "list must contain two objects for geometry and nodes (check scene graph structure)";
			return false;		
		}
	} else {
		if (!(list->getNumChildNodes(NULL) >= 1)) {
			msg = "list must contain at least a geometry object (check scene graph structure)";
			return false;		
		}
	}
	SGObjectNode *geometryNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 0));
	SGObjectNode *mmNodesNode = NULL;		
	if (checkMMNodes) {
		mmNodesNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 1));
	}
	if (!geometryNode) {
		msg = "first node in list is not an SGObjectNode (check scene graph structure)";
		return false;		
	}
	if (checkMMNodes && !mmNodesNode) {
		msg = "second node in list is not an SGObjectNode (check scene graph structure)";
		return false;		
	}
	if (geometryNode->getName() != "geometry") {
		msg = "first node in list must be named \"geometry\"";
		return false;		
	}
	if (checkMMNodes && mmNodesNode->getName() != "mmNodes") {
		msg = "first node in list must be named \"mmNodes\"";
		return false;		
	}

	UnstructuredInCorePointCloud *upcGeometry = dynamic_cast<UnstructuredInCorePointCloud*>(geometryNode->getSceneObject());
	UnstructuredInCorePointCloud *upcMMNodes = NULL;
	if (checkMMNodes) {
		upcMMNodes = dynamic_cast<UnstructuredInCorePointCloud*>(mmNodesNode->getSceneObject());
	}
	if (!upcGeometry) {
		msg = "geometry scene object is not an UnstructureInCorePointCloud (or descendant)";
		return false;
	}	
	if (checkMMNodes && !upcMMNodes) {
		msg = "mmNodes scene object is not an UnstructureInCorePointCloud (or descendant)";
		return false;
	}	
	if (!compareDescr(upcGeometry->getDescr(), geometryChannels)) {
		msg = "attribute channels of geometry point cloud do not match the requirements.";
		return false;
	} 
	if (checkMMNodes && !compareDescr(upcMMNodes->getDescr(), mmNodesChannels)) {
		msg = "attribute channels of mmNodes point cloud do not match the requirements.";
		return false;
	} 
	TopologyAttachment *topAttGeometry = dynamic_cast<TopologyAttachment*>(upcGeometry->getAttachments()->getData("TopologyAtt"));
	TopologyAttachment *topAttMMNodes = NULL;
	if (checkMMNodes) {
		topAttMMNodes = dynamic_cast<TopologyAttachment*>(upcMMNodes->getAttachments()->getData("TopologyAtt"));
	}
	if (!topAttGeometry) {
		msg = "topology for geometry is missing";
		return false;
	}
	if (checkMMNodes && !topAttMMNodes) {
		msg = "topology for mmNodes is missing";
		return false;
	}

	InCorePCTopologyGraph *topGeometry = topAttGeometry->getTopology();
	if (topGeometry == NULL) {
		msg = "InCorePCTopologyGraph for geometry is NULL";
		return false;	
	}
	InCorePCTopologyGraph *topMMNodes = NULL;
	if (checkMMNodes) {
		topMMNodes = topAttMMNodes->getTopology();
		if (topMMNodes == NULL) {
			msg = "InCorePCTopologyGraph for mmNodes is NULL";
			return false;	
		}
	}
	
	if (!compareDescr(topGeometry->getEdges()->getDescr(), geometryGraphChannels)) {
		msg = "attribute channels of geometry topology graph do not match the requirements.";
		return false;
	} 	
	if (checkMMNodes && !compareDescr(topMMNodes->getEdges()->getDescr(), mmNodesGraphChannels)) {
		msg = "attribute channels of mmNodes topology graph do not match the requirements.";
		return false;
	} 	

	return true;
}


UnstructuredInCorePointCloud *SGEMeshlessUrshape::getGeometryPointCloud() const {
	Matrix4f forget;
	SceneGraphNode *node;
   getSGNodeTransformation(getScene(), getPath(), forget, node);
	checkIfSomethingWentWrong(node);
	
	SGListNode *list = dynamic_cast<SGListNode*>(node);
	checkIfSomethingWentWrong(list);
	
	SGObjectNode *geometryNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 0));
	checkIfSomethingWentWrong(geometryNode);

	UnstructuredInCorePointCloud *upcGeometry = dynamic_cast<UnstructuredInCorePointCloud*>(geometryNode->getSceneObject());
	checkIfSomethingWentWrong(upcGeometry);

	return upcGeometry;
}

PointSet *SGEMeshlessUrshape::getGeometryPointSet() const {
	UnstructuredInCorePointCloud *upc = getGeometryPointCloud();
	return upc->getPointSet();
}

SceneObjectIterator *SGEMeshlessUrshape::createGeometryIterator(SceneObjectIteratorType type) const {
	UnstructuredInCorePointCloud *upc = getGeometryPointCloud();
	return upc->createIterator(type);
}

InCorePCTopologyGraph *SGEMeshlessUrshape::getGeometryTopology() const {
	UnstructuredInCorePointCloud *upc = getGeometryPointCloud();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData("TopologyAtt"));
	checkIfSomethingWentWrong(topAtt);
	return topAtt->getTopology();
}

AttachedIndexedOctree* SGEMeshlessUrshape::getGeometryOctree(card32 createNumPtsPerLeaf) const {
   UnstructuredInCorePointCloud *upc = getGeometryPointCloud();
   AttachedDataContainer *attachments = upc->getAttachments();
   AttachedIndexedOctree *octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));
   if (octree == NULL) {
		PointSet *ps = upc->getPointSet();
		octree = new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
		octree->octreesetup(ps, createNumPtsPerLeaf);
		attachments->attachData(octree);
   }
   return octree;
}


void SGEMeshlessUrshape::createNodeInformation() {
	return;
	/*
	// Create the AttachedNodeInformation
	cout << "Creating nodeinformation attachment" << endl;
	AttachedNodeInformation * ani = new AttachedNodeInformation();
	cout << "Setting up attachment" << endl;
	ani->setup("NodeInformationAtt", AttachedData::ADF_DELETEONUPDATE,  NULL);
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
    AttachedDataContainer *attachments = upc->getAttachments();
	cout << "Adding attachment" << endl;
	attachments->attachData(ani);

	// Compute node neighbors
	cout << "Computing neighbors for nodes" << endl;
	computeNodeNeighbors();
	// Now for each node, compute the shape functions (normal and gradient)
	cout << "Evaluate nodal shape functions" << endl;
	evaluateNodeShapeFunctions();
	*/
}

#if 0

void SGEMeshlessUrshape::computeNodeNeighbors() {
	// node positions
	UnstructuredInCorePointCloud * pc = getMMNodesPointCloud();
	PointSet * ps = getMMNodesPointSet();
	// node radii
	// this->mmNodeChannels
	AAT radiusAAT = pc->getAAT("radius");
	AAT positionAAT = pc->getAAT("position");

	// total number of nodes
	const card32 nbnodes = pc->getNumPoints();

	// First compute the maximal possible radius.
	// This radius will be used for range querying (we have to
	// be conservative when nodes have varying radii.
	float32 maxRadius = 0;
	for (card32 i=0; i<nbnodes; i++) {
		const float32 radius = ps->get1f(i, radiusAAT);
		//cout << "Radius of point " << i << " : " << radius << endl;
		if (radius>maxRadius) maxRadius = radius;
	}
	//int bla; cin >> bla;
	
	// octree for neighborhood queries
	AttachedIndexedOctree * octree = getMMNodesOctree();

	// TODO: we should keep this, for the correspondences.
	PointSetSphericalQuery query(ps, octree, Vector3f(), maxRadius);

	// Get the AttachedNodeInformation
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
    AttachedDataContainer *attachments = upc->getAttachments();
	AttachedNodeInformation * ani = dynamic_cast<AttachedNodeInformation*>(attachments->getData("NodeInformationAtt"));

	// For each node, create the NodeInformation, query for the
	// neighbors and store it in there
	for (card32 i=0; i<nbnodes; i++) {
		const Vector3f pos = ps->get3f(i, positionAAT);
		const float32 radius = ps->get1f(i, radiusAAT);
		query.setTargetSphere(pos, maxRadius);

		cout << "querying with radius: " << maxRadius << endl;

		NodeInformation *ni = new NodeInformation();
		ani->addNodeInformation(ni);
		bool found;
		card32 neighbor_index; // neighbor index
		query.getNextPointIndex(neighbor_index, found);
		while (found) {	
			
			const Vector3f neighbor_pos = ps->get3f(neighbor_index, positionAAT);
			//cout << neighbor_pos[0] <<"  " << neighbor_pos[1] << " " << neighbor_pos[2] << endl;
			const float32 neighbor_sqr_dist = (pos-neighbor_pos).getSqrNorm(); 			
			const float32 neighbor_radius = ps->get1f(neighbor_index, radiusAAT); // radius of neighbor
			
			if (neighbor_radius*neighbor_radius>=neighbor_sqr_dist) {
				const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
				float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
				ni->addNeighbor(neighbor_index, neighbor_dist, weight);
			}
			query.getNextPointIndex(neighbor_index, found);
		}
		cout << "Number of neighbors : " << ni->nbNeighbors() << endl;
		// TODO: normalize weights so that they sum to 1. Might be more stable for later computations.
	}
	cout << "Using radius: " << maxRadius << endl;
	/*for (card32 i=0; i<nbnodes; i++) {
		NodeInformation * ni = new NodeInformation();
		ani->addNodeInformation(ni);
		const Vector3f posi = ps->get3f(i, positionAAT);
		for (card32 j=0; j<nbnodes; j++) {
			const Vector3f posj = ps->get3f(j, positionAAT);
			const float32 radius = ps->get1f(j, radiusAAT);				
			const float32 neighbor_sqr_dist = (posj-posi).getSqrNorm();
			//cout << "Testing: " << radius << " > " << sqrt(neighbor_sqr_dist) << "?" << endl;
			if (radius*radius>=neighbor_sqr_dist) {
				const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
				float32 weight = 1.0f - neighbor_sqr_dist/(radius*radius); weight = weight*weight*weight;
				ni->addNeighbor(j, neighbor_dist, weight);
			}		
		}
		//cout << "Number of neighbors : " << ni->nbNeighbors() << endl;
		// TODO: normalize weights so that they sum to 1. Might be more stable for later computations.
	}*/
	cout << "Using radius: " << maxRadius << endl;
}

void SGEMeshlessUrshape::evaluateNodeShapeFunctions() {
	// node positions
	UnstructuredInCorePointCloud * pc = getMMNodesPointCloud();
	PointSet * ps = getMMNodesPointSet();
	// Get the AttachedNodeInformation
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
    AttachedDataContainer *attachments = upc->getAttachments();
	AttachedNodeInformation * ani = dynamic_cast<AttachedNodeInformation*>(attachments->getData("NodeInformationAtt"));

	AAT positionAAT = pc->getAAT("position");


	// total number of nodes
	const card32 nbnodes = pc->getNumPoints();

	for (card32 i = 0; i<nbnodes; i++) {
		cout << "Processing : " << i << " " << nbnodes << endl;
		const Vector3f &pos = ps->get3f(i, positionAAT);
		NodeInformation* ni = ani->getNodeInformation(i);	
		// normal shape functions
		evaluateNodeShapeFunctions(pos, ni);	
		// gradient shape functions
		evaluateNodeGradientShapeFunctions(pos, ni);	
	}
}

void SGEMeshlessUrshape::evaluateNodeShapeFunctions(const Vector3f &x, NodeInformation *ni) {
	UnstructuredInCorePointCloud * pc = getMMNodesPointCloud();
	PointSet * ps = getMMNodesPointSet();
	AAT positionAAT = pc->getAAT("position");
	AAT radiusAAT = pc->getAAT("radius");


	
	const card32 nbneighbors = ni->nbNeighbors();
	// First compute the moment matrix (it's a 4x4 matrix)
	Matrix4d M; M.setZero();
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = ni->getWeight(i);
		const card32 index = ni->getNeighborIndex(i);
		const Vector3f &pos = ps->get3f(index, positionAAT);
		const float32 radius = ps->get1f(index, radiusAAT);
		const float32 inv_radius = 1.0f/radius;

		// We translate and scale. This improves the condition number
		// of the moment matrix (see for example Fries and Matthies).
		const Vector3f diff = (pos - x)*inv_radius;

		M[0][0] += w*diff[2]*diff[2];
		M[0][1] += w*diff[1]*diff[2];
		M[0][2] += w*diff[0]*diff[2];
		M[0][3] += w*diff[2];

		M[1][0] += w*diff[2]*diff[1];
		M[1][1] += w*diff[1]*diff[1];
		M[1][2] += w*diff[0]*diff[1];
		M[1][3] += w*diff[1];

		M[2][0] += w*diff[2]*diff[0];
		M[2][1] += w*diff[1]*diff[0];
		M[2][2] += w*diff[0]*diff[0];
		M[2][3] += w*diff[0];

		M[3][0] += w*diff[2];
		M[3][1] += w*diff[1];
		M[3][2] += w*diff[0];
		M[3][3] += w;
	}
	// TODO: use SVD or gauss for safe/faster inversion
	Matrix4d invM = invertMatrix(M);

	// Now, finally compute for each neighbor the shape function value.
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = ni->getWeight(i);
		const card32 index = ni->getNeighborIndex(i);
		const Vector3f &pos = ps->get3f(index, positionAAT);
		const float32 radius = ps->get1f(index, radiusAAT);
		const float32 inv_radius = 1.0f/radius;
		const Vector3f diff = (pos - x)*inv_radius;

		// Difficult to follow what I'm doing here.
		// See Fries and Matthies, for the optimized shape function
		// evaluation (basically we translate and scale to improve
		// the conditioning number to compute invM.
		float result = invM[3][3]*1;
		for (card32 k=0; k<3; k++) result+= invM[3][k]*diff[2-k];		
		result*=w;
		ni->setShapeValue(i, result);
	}
	// Check if shapevalues sum up to 1.
	float32 test = 0;
	for (card32 i=0; i<nbneighbors; i++) {
		test+=ni->getShapeValue(i);
	}
	if (fabs(test-1.0f)>0.3) {
		cout << "Error: Shape functions do not sum to 1, but to: " << test << endl;
	}
}

void SGEMeshlessUrshape::evaluateNodeGradientShapeFunctions(const Vector3f &x, NodeInformation *ni) {
	UnstructuredInCorePointCloud * pc = getMMNodesPointCloud();
	PointSet * ps = getMMNodesPointSet();
	AAT positionAAT = pc->getAAT("position");
	
	const card32 nbneighbors = ni->nbNeighbors();

	// First compute the moment matrix (it's a 3x3 matrix for the gradient)
	Matrix3d M; M.setZero();
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = ni->getWeight(i);
		const card32 index = ni->getNeighborIndex(i);
		const Vector3f &pos = ps->get3f(index, positionAAT);

		// We translate and scale. This improves the condition number
		// of the moment matrix (see for example Fries and Matthies).
		const Vector3f diff = (pos - x);

		M[0][0] += w*diff[0]*diff[0];
		M[0][1] += w*diff[0]*diff[1];
		M[0][2] += w*diff[0]*diff[2];

		M[1][0] += w*diff[1]*diff[0];
		M[1][1] += w*diff[1]*diff[1];
		M[1][2] += w*diff[1]*diff[2];

		M[2][0] += w*diff[2]*diff[0];
		M[2][1] += w*diff[2]*diff[1];
		M[2][2] += w*diff[2]*diff[2];				
	}
	// TODO: use SVD or gauss for safe/faster inversion
	Matrix3d invM = invertMatrix(M);

	// Now, finally compute for each neighbor the shape function value.
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = ni->getWeight(i);
		const card32 index = ni->getNeighborIndex(i);
		const Vector3f &pos = ps->get3f(index, positionAAT);

		const Vector3f diff = (pos - x);
		Vector3f result; result.setZero();
		// compute invM*diff
		// TODO: is there a helper method for doing this?
		for (card32 j=0; j<3; j++) {
			for (card32 k=0; k<3; k++) {
				result[j] += invM[j][k]*diff[k];
			}
		}
		result*=w;
		ni->setGradientShapeValue(i, result);
	}	
}
#endif


AttachedNodeInformation* SGEMeshlessUrshape::getNodeInformation() const {
	UnstructuredInCorePointCloud *upc = getGeometryPointCloud();
    AttachedDataContainer *attachments = upc->getAttachments();
	AttachedNodeInformation *result = dynamic_cast<AttachedNodeInformation*>(attachments->getData("NodeInformationAtt"));
	checkIfSomethingWentWrong(result);
	return result;
}


UnstructuredInCorePointCloud *SGEMeshlessUrshape::getMMNodesPointCloud() const {
	Matrix4f forget;
	SceneGraphNode *node;
    getSGNodeTransformation(getScene(), getPath(), forget, node);

   	checkIfSomethingWentWrong(node);

		
	SGListNode *list = dynamic_cast<SGListNode*>(node);
	checkIfSomethingWentWrong(list);

		
	SGObjectNode *mmNodesNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 1));
	checkIfSomethingWentWrong(mmNodesNode);

	

	UnstructuredInCorePointCloud *upcMMNodes = dynamic_cast<UnstructuredInCorePointCloud*>(mmNodesNode->getSceneObject());
	checkIfSomethingWentWrong(upcMMNodes);
	return upcMMNodes;
}

PointSet *SGEMeshlessUrshape::getMMNodesPointSet() const {
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
	return upc->getPointSet();
}

SceneObjectIterator *SGEMeshlessUrshape::createMMNodesIterator(SceneObjectIteratorType type) const {
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
	return upc->createIterator(type);
}

InCorePCTopologyGraph *SGEMeshlessUrshape::getMMNodesTopology() const {
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData("TopologyAtt"));
	checkIfSomethingWentWrong(topAtt);
	return topAtt->getTopology();
}

AttachedIndexedOctree* SGEMeshlessUrshape::getMMNodesOctree(card32 createNumPtsPerLeaf) const {
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
   AttachedDataContainer *attachments = upc->getAttachments();
   AttachedIndexedOctree *octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));
   if (octree == NULL) {
		PointSet *ps = upc->getPointSet();
		octree = new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
		octree->octreesetup(ps, createNumPtsPerLeaf);
		attachments->attachData(octree);
   }
   return octree;
}



void SGEMeshlessUrshape::checkIfSomethingWentWrong(void *mustNotBeNULL) const 
{
	if (!mustNotBeNULL) {
		throw EInvalidState("SGEMeshlessUrshape::checkIfSomethingWentWrong() - cannot peform requested operation, please call check() before and make sure it returns true.");
	}
}



/*
	Matrix4f forget;
	SceneGraphNode *node;
   getSGNodeTransformation(getScene(), getPath(), forget, node);
	checkIfSomethingWentWrong(node);
	
	SGListNode *list = dynamic_cast<SGListNode*>(node);
	checkIfSomethingWentWrong(list);
	
	SGObjectNode *geometryNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 0));
	SGObjectNode *mmNodesNode = dynamic_cast<SGObjectNode*>(list->getChildNode(NULL, 1));
	checkIfSomethingWentWrong(geometryNode);
	checkIfSomethingWentWrong(mmNodesNode);

	UnstructuredInCorePointCloud *upcGeometry = dynamic_cast<UnstructuredInCorePointCloud*>(geometryNode->getSceneObject());
	UnstructuredInCorePointCloud *upcMMNodes = dynamic_cast<UnstructuredInCorePointCloud*>(mmNodesNode->getSceneObject());
	checkIfSomethingWentWrong(upcGeometry);
	checkIfSomethingWentWrong(upcMMNodes);
	
	TopologyAttachment *topAttGeometry = dynamic_cast<TopologyAttachment*>(upcGeometry->getAttachments()->getData("TopologyAtt"));
	TopologyAttachment *topAttMMNodes = dynamic_cast<TopologyAttachment*>(upcMMNodes->getAttachments()->getData("TopologyAtt"));
	checkIfSomethingWentWrong(topAttGeometry);
	checkIfSomethingWentWrong(topAttMMNodes);

	InCorePCTopologyGraph *topGeometry = topAttGeometry->getTopology();
	InCorePCTopologyGraph *topMMNodes = topAttMMNodes->getTopology();
	checkIfSomethingWentWrong(topGeometry);
	checkIfSomethingWentWrong(topMMNodes);
*/
SceneObject *SGEMeshlessUrshape::getSceneObject() const 
{
	return getGeometryPointCloud();
}

SceneObjectIterator *SGEMeshlessUrshape::createIterator(SceneObjectIteratorType type) const 
{
	return createGeometryIterator(type);
}


void SGEMeshlessUrshape::tempTest() {
	cout << "Running tempTest()" << endl;
	UnstructuredInCorePointCloud *pc = getMMNodesPointCloud();
	AAT radiusAAT = pc->getAAT("radius");
	PointSet * ps = pc->getPointSet();
	// add some points
	int index = 0;
	for (float32 x=0; x<=1; x+=0.1f) {
		for (float32 y=0; y<=1; y+=0.1f) {
			for (float32 z=0; z<=1; z+=0.1f) {
				Vector3f point; 
				point[0]=x;
				point[1]=y;
				point[2]=z;
				ps->addPoint((char*)&point);
				//int index = pc->getNumPoints()-1;
				ps->set1f(index, radiusAAT, 0.2f);
				index++;
			}
	    }
	}
	createNodeInformation();

	// Test the gradient shape functions
	// Take an arbitray function in 3D
	// f(x,y,z) = x^2 + x*y - 4*x*z + x + y
	// df/dx = 2x+y-4*z+1
	// df/dy = x+1
	// df/dz = -4*x

	// node positions
	// Get the AttachedNodeInformation
	UnstructuredInCorePointCloud *upc = getMMNodesPointCloud();
    AttachedDataContainer *attachments = upc->getAttachments();
	AttachedNodeInformation * ani = dynamic_cast<AttachedNodeInformation*>(attachments->getData("NodeInformationAtt"));

	AAT positionAAT = pc->getAAT("position");


	// total number of nodes
	const card32 nbnodes = pc->getNumPoints();

	for (card32 i = 0; i<nbnodes; i++) {
		const Vector3f &pos = ps->get3f(i, positionAAT);
		NodeInformation* ni = ani->getNodeInformation(i);
		// Exact
		float32 dfdx = 2*pos[0]+pos[1]-4*pos[2]+1;
		float32 dfdy = pos[0]+1;
		float32 dfdz = -4*pos[0];
		float32 fi = pos[0]*pos[0] + pos[0]*pos[1] - 4*pos[0]*pos[2] + pos[0] + pos[1];
		// Approximationg
		Vector3f df; df.setZero();
		for (card32 k=0; k<ni->nbNeighbors(); k++) {
			const Vector3f &posj = ps->get3f(ni->getNeighborIndex(k), positionAAT);
			float32 fj = posj[0]*posj[0] + posj[0]*posj[1] - 4*posj[0]*posj[2] + posj[0] + posj[1];
			df+=ni->getGradientShapeValues(k)*(fj-fi);
		}
		// These should be close together
		cout << dfdx << " " << df[0] << endl;
		cout << dfdy << " " << df[1] << endl;
		cout << dfdz << " " << df[2] << endl;
	}
}