//---------------------------------------------------------------------------
#ifndef SGEMeshlessUrshapeH
#define SGEMeshlessUrshapeH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "SGEShape.h"
//---------------------------------------------------------------------------
//#include "AttachedNodeInformation.h"
//#include "NodeInformation.h"

namespace geoxl {

class InCorePCTopologyGraph;
class VertexDescriptor;
class AttachedIndexedOctree;
class AttachedNodeInformation;
class NodeInformation;

///
///   mw - 11/20/2007
///
///   encapsulates scene graph representation of a meshless-deformation urshape 
///
///   it consists of two parts - geometry and a nodal representation, which are expected to reside in an SGListNode
///   the path must point to the list, which has to contain two objects named "geometry" and "mmNodes"
///   the SGE also checks for channels of geometry and mmNodes urshape
///   the geometry is a surfel graph (UnstructuredInCorePointCloud with topology attachment)
///   the nodal representation is a resampled version of the geometry that forms the parametrization
///   domain of the deformation. It is also a surfel graph, with an additional attachment storing
///   precomputed data for the meshless solver
///
class DEFORMATION_API SGEMeshlessUrshape : public SGEShape {
	GEOXL_CLASS( SGEMeshlessUrshape )
 private:
	/// attribute channels the geometry surfels are expected to have
	VertexDescriptor *geometryChannels;
	/// attribute channels the geometry's topology graph is expected to have
	VertexDescriptor *geometryGraphChannels;
	/// attribute channels the deformation nodes' surfels are expected to have
	VertexDescriptor *mmNodesChannels;
	/// attribute channels the deformation nodes' topology graph is expected to have
	VertexDescriptor *mmNodesGraphChannels;

	/// throws an exception if checked assetions are not valid during operations
	void checkIfSomethingWentWrong(void *mustNotBeNULL) const;

	// Compute node neighbors
	void computeNodeNeighbors();
	// For each node, compute the shape functions (normal and gradient)
	void evaluateNodeShapeFunctions();
	// Normal shape functions for the point x
	void evaluateNodeShapeFunctions(const Vector3f &x, NodeInformation *ni);
	// Gradient shape functions for the ponit x
	void evaluateNodeGradientShapeFunctions(const Vector3f &x, NodeInformation *ni);
	
 public:
	SGEMeshlessUrshape();

	/// setup channels to be checked
	void setup(
		VertexDescriptor *geometryChannels,
		VertexDescriptor *geometryGraphChannels,
		VertexDescriptor *mmNodesChannels,
		VertexDescriptor *mmNodesGraphChannels
	);

	/// property access
	inline const VertexDescriptor *getGeometryChannels() const      {return geometryChannels;}
	/// property access
   inline const VertexDescriptor *getGeometryGraphChannels() const {return geometryGraphChannels;}
	/// property access
	inline const VertexDescriptor *getMMNodesChannels() const       {return mmNodesChannels;}
	/// property access
   inline const VertexDescriptor *getMMNodesGraphChannels() const  {return mmNodesGraphChannels;}
		
	/// is the scene graph part setup correctly?
	/// the method only checks if the geometry part is valid (as required by the superclass)
	/// to check the availability of the mmNodes part, the parameter checkMMNodes must be set to 
	/// true explicitly (which is only possible knowing we are dealing with an SGEMeshlessUrshape)
	virtual bool check(string &msg, bool checkMMNodes = false) const;

	/// get point cloud object for the geometry part
	UnstructuredInCorePointCloud *getGeometryPointCloud() const;
	/// directly get the point set of the geometry part
	PointSet *getGeometryPointSet() const;
	/// get iterator to the geometry part
   SceneObjectIterator *createGeometryIterator(SceneObjectIteratorType type) const;
	/// directly get the topology graph of the geometry part
	InCorePCTopologyGraph *getGeometryTopology() const;
	/// Get octree attachment of the geometry part (needed for range queries)
	/// creates the octree if needed, using the specified parameter
	AttachedIndexedOctree* getGeometryOctree(card32 createNumPtsPerLeaf = 20) const;

	/// get point cloud object for the mmNodes part
	UnstructuredInCorePointCloud *getMMNodesPointCloud() const;
	/// directly get the point set of the mmNodes part
	PointSet *getMMNodesPointSet() const;
	/// get iterator to the mmNodes part
   SceneObjectIterator *createMMNodesIterator(SceneObjectIteratorType type) const;
	/// directly get the topology graph of the mmNodes part
	InCorePCTopologyGraph *getMMNodesTopology() const;
	/// Get octree attachment of the mmNodes part (needed for range queries)
	/// creates the octree if needed, using the specified parameter
	AttachedIndexedOctree* getMMNodesOctree(card32 createNumPtsPerLeaf = 20) const;
	
	/// inherited method works on geometry
	virtual SceneObject *getSceneObject() const;
	/// inherited method works on geometry
   virtual SceneObjectIterator *createIterator(SceneObjectIteratorType type) const;


   // Create the attached node information: compute node neighbors, evaluate shape functions, etc.
   void createNodeInformation();
   // Get the cached node information (which holds node neighbors, evaluated shape functions, etc.)
   AttachedNodeInformation* getNodeInformation() const;

   // a temporary test function to test the node stuff
   void tempTest();
	
   ~SGEMeshlessUrshape();
};

}

#endif