#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARComputeGlobalResidual.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ARSettings.h"
#include "ARDeformationOptimizationContext.h"
#include "DeformationEvaluator.h"
#include "PointCloud.h"
#include "TopologyAttachment.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyRangeSearch.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;



void geoxl::ARComputeGlobalResidual::computeGlobalResidual( PointCloud * urshape, ARDeformationOptimizationContext *deformation,
																			 DeformationEvaluator *evaluator, vector<PointCloud*> data, ARSettings * settings )
{
	// setting up constants
	bool outputWeirdThings = true;	// debugging
	float sigma = settings->residualSupportSigma * settings->topSurfelSpacing;
	float inv_sigma_2 = 1.0f / pow(sigma,2.0f);
	float maxRad_2 = pow( sigma*2.0f,2.0f);

	// prepare urshape
	checkAttribute(urshape, "residual",2,VAD::DATA_FORMAT_FLOAT32);

	// create iterator and get AATs
	SceneObjectIterator * iterUrshape = urshape->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iterUrshape);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iterUrshape);
	AAT posAAT = urshape->getAAT("position");
	AAT residualAAT = urshape->getAAT("residual");
	AAT normalAAT = urshape->getAAT("normal");

	// set residual + residualweight to zero	
	while(!bIt->atEnd())
	{
		mIt->set2f(residualAAT, makeVector2f(0,0));
		bIt->next();
	}
	
	// iterate over all frames
	for( unsigned frame=0;frame<data.size();frame++ )
	{
		// setup frame
		PointCloud * frame_pc = data[frame];
		HierarchicalKNNIterator frame_hIt(frame_pc,32,NULL);
		AAT framePosAAT = frame_pc->getAAT("position");

		// init evaluator
		float frameTime = frame;
		evaluator->beginQuery(frameTime);

		// iterator over all urshape points
		unsigned urshape_index = 0;
		bIt->reset();
		while(!bIt->atEnd())
		{
			Vector3f urshape_pos = bIt->get3f(posAAT);
			Vector3f urshape_normal = bIt->get3f(normalAAT);

			// deform urshape point + normal
			ARConstraintAnchor anchor(urshape_pos, urshape_index);
			Vector3f deformed_pos = evaluator->getPosition(anchor);
			Matrix3f gradient = evaluator->getSpatialGradient(anchor);
			bool success;
			Matrix3f invGradient = invertMatrix(gradient, &success, 1e-15f);
			Vector3f deformed_normal;
			if( !success && outputWeirdThings )
			{
				debugOutput << "Couldn't invert spatial gradient! Using A*n instead of (A^-1)^T n\n";
				deformed_normal = normalize(gradient * urshape_normal);
			}
			else
			{
				deformed_normal = normalize( invGradient.transpose() * urshape_normal );
			}

			// compute signed distances
			frame_hIt.setSeekPointAndReset(deformed_pos);
			float summed_weight = 0;
			float summed_signed_dist = 0;
			while(!frame_hIt.atEnd())
			{
				// check distance to urshape point
				Vector3f frame_pos = frame_hIt.get3f(framePosAAT);
				float dist_2 = normQuad(frame_pos - deformed_pos);
				if( dist_2 > maxRad_2 )
					break; // too far away

				// compute weight and distance
				float weight = exp( -dist_2 * inv_sigma_2 );
				summed_weight += weight;
				float signed_dist = (frame_pos - deformed_pos) * deformed_normal;
				summed_signed_dist += signed_dist * weight;

				frame_hIt.next();
			}


			// add residual to existing one
			Vector2f residual = bIt->get2f(residualAAT);
			residual += makeVector2f( fabs(summed_signed_dist), summed_weight );
			mIt->set2f(residualAAT,residual);

			bIt->next();
			urshape_index++;
		}
		evaluator->endQuery();
	}


	delete iterUrshape;
}

struct TMedianResidualSample
{
	card32 index;
	float32 res;
	bool operator < (const TMedianResidualSample &o) const {return res < o.res;}
};

PointCloud *  ARComputeGlobalResidual::adaptRigidityToResiduals(
 PointCloud * urshape, ARSettings * settings, float32 medianFilterRadius
 )
{
	// check residual channel and rigidity channel
	if( !urshape->providesAttribute("residual"))
		throw PException("ARComputeGlobalResidual::adaptRigidityToResiduals - no residual channel found!");
	if( urshape->getNumPoints() == 0)
		throw PException("ARComputeGlobalResidual::adaptRigidityToResiduals - empty urshape!");


	// create iterator and get AATs
	SceneObjectIterator * iterUrshape = urshape->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iterUrshape);
	AAT posAAT = urshape->getAAT("position");
	AAT residualAAT = urshape->getAAT("residual");

	// get residuals
	vector<float> residual(urshape->getNumPoints());
	unsigned index=0;
	while(!bIt->atEnd())
	{
		Vector2f res = bIt->get2f(residualAAT);
		if( res[1] > 0.0f)
			residual[index] = res[0]/res[1];
		else
		{
			debugOutput << "found invalid residual!\n";
			residual[index] = 0;
		}
		
		index++;
		bIt->next();
	}

	// clamp outliers and scale residuals
	vector<float> residualCopy = residual;
	sort(residualCopy.begin(),residualCopy.end());
	float highResidual = residualCopy[min( residualCopy.size()-1, (float)residualCopy.size() * settings->residualUpperLimit)];
	float lowResidual = residualCopy[(float)residualCopy.size() * settings->residualLowerLimit];
	for( unsigned i=0;i<residual.size();i++ )
	{
		residual[i] = max(0.0f,min( 1.0f, (residual[i] - lowResidual) / (highResidual - lowResidual) ));
		if(residual[i]>0.0f && residual[i]<0.5f )
			residual[i] = 0.5f;
		else
			if(residual[i]>0.5f )
				residual[i] = 0.25f;
			else
				residual[i] = 1.0f;
	}

	// median filter
	vector<float32> rigidityWeights(residual.size());
	AttachedDataContainer *attachments = urshape->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	InCorePCTopologyGraph *topGeometry = topAtt->getTopology();
	const PointSet * ps = bIt->lockBlockRead();
	TopNeighborhoodSphereSearch querry;
	int flushTime = urshape->getNumPoints() * 0.1f;
	for( unsigned i=0;i<urshape->getNumPoints();i++ )
	{
		Vector3f pos = ps->get3f(i,posAAT);
		querry.setup(topGeometry,ps,i, pos, medianFilterRadius, posAAT );

		vector<TMedianResidualSample> samples;
		while( !querry.atEnd())
		{
			TMedianResidualSample s;
			s.index = querry.getCurrentIndex();
			s.res = residual[s.index];
			samples.push_back(s);

			querry.next();
		}
		sort(samples.begin(),samples.end());
		rigidityWeights[i] = samples[ samples.size()/2].res;

		if( i % flushTime == 0 )
		{
			debugOutput << "medianfilter(rigidityWeights) " << i * 100.0f / urshape->getNumPoints() << "%\n";
			debugOutput.flush();
		}
	}
	
	// grow regions
	vector<bool> visited(residual.size(),false);
	for( int level=2;level>=1;level--)
	{
		float growRadius = pow( 2.0f, -(float)level + 2.0f) / (float)settings->gridSize;
		float resValue = (level == 2) ? 0.25f : 0.5f;

		for( unsigned i=0;i<ps->getNumEntries();i++ )
		{
			if( visited[i] )
				continue;

			if( rigidityWeights[i] == resValue )
			{
				list<unsigned> queue;
				queue.push_back(i);
				Vector3f startPos = ps->get3f(i,posAAT);
				
				while(!queue.empty())
				{
					unsigned index = queue.front();
					queue.pop_front();

					if( visited[index] )
						continue;
					
					visited[index] = true;
					rigidityWeights[index] = resValue;

					for( unsigned n=0;n<topGeometry->getNumAdjacentVertices(index);n++ )
					{
						unsigned nIndex = topGeometry->getAdjacentVertex(index,n);
						if( visited[nIndex] )
							continue;
						if( rigidityWeights[nIndex] <= resValue )
							continue;
						if( norm(ps->get3f(nIndex,posAAT) - startPos ) > growRadius )
							continue;
						queue.push_back(nIndex);
					}

				}
			}
		}
	}
	bIt->unlock();


		
	PointCloud * newUrshape = (PointCloud*)urshape->copy();	
	checkAttribute(newUrshape, "rigidityWeight",1, VAD::DATA_FORMAT_FLOAT32);
	AAT rigidityAAT = newUrshape->getAAT("rigidityWeight");
	SceneObjectIterator * iterNewUrshape = newUrshape->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bNewIt = dynamic_cast<BasicPointCloudIterator*>(iterNewUrshape);
	ModifyablePointCloudIterator * mNewIt = dynamic_cast<ModifyablePointCloudIterator*>(iterNewUrshape);

	
	// set rigidity weights 
	bNewIt->reset();
	index = 0;
	while(!bNewIt->atEnd())
	{
		mNewIt->set1f(rigidityAAT,rigidityWeights[index]);
		index++;
		bNewIt->next();
	}
	
	return newUrshape;
}