//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARGlobalOptimization.h"
//---------------------------------------------------------------------------
#include "ARSequence.h"
#include "ARSettings.h"
#include "AROptimizationContext.h"
#include "ARShapeOptimizationContext.h"
#include "ARTransfConstraintsToUrshape.h"
#include "ICPConstraints.h"
#include "DeformationEvaluator.h"
#include "AREvaluateDeformation.h"
#include "DeformationSolver.h"
#include "ARDeformationOptimizationContext.h"
#include "SGListNode.h"
#include "SGObjectNode.h"
#include "ShapeReconstruction.h"
#include "ShapeEvaluator.h"
#include "InCorePCTopologyGraph.h"
#include "DeformationTools.h"
#include "Timer.h"
#include "ARGlobalDeformationOptimization.h"
#include "ARUrshapeOptimization.h"
#include "ARUrshapeDeformationOpt.h"
#include "ARChangeUrshape.h"
#include "InterfaceGlobals.h"
#include "ARPruneUnsupportedOutliers.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

#define IFDBO if (settings->mergeDebugOutput)


ARGlobalOptimization::ARGlobalOptimization() {
   visTool = NULL;
   myVisPC = NULL;
   myVisColor = makeVector4f(1,0,1,1);
}


ARGlobalOptimization::~ARGlobalOptimization() {}



void ARGlobalOptimization::performGlobalOpt(
   ARSequence *seq, 
   DeformationSolver *deformationSolver,
   ShapeReconstruction *shapeSolver, 
   Scene *scene,
   string dataPath,
   ARSettings *settings,
   card32 startFrame,
   card32 endFrame,
   Delegate visualFeedback,
   bool optimizeUrshapeDeformation,   
   bool optimizeShape,
   bool optimizeDeformation,
   bool removeUnsupportedPoints)
{
   if (!seq->getDuration() == endFrame - startFrame) {
      throw ERangeCheck("ARGlobalOptimization::performGlobalOpt() - sequence does not match specified range.");
   }
   vector<PointCloud*> dataPoints = getDataPoints(scene, dataPath, startFrame, endFrame);
   Timer timer;

   PointCloud* originalUrshape = dynamic_cast<PointCloud*>( seq->getUrshape()->copy() );
   UnstructuredInCorePointCloud* originalUrshapePC = dynamic_cast<UnstructuredInCorePointCloud*>( originalUrshape );

   //
   // --- urshape geometry optimization
   //
   if (optimizeShape) {
      IFDBO debugOutput << "   optimizing urshape geometry...\n";
      IFDBO debugOutput.flush();
      recomputeNormalsAndFrames(seq->getUrshape(), settings);
      ARUrshapeOptimization shapeOpt;
      DeformationEvaluator *eval = deformationSolver->createEvaluator();
      eval->setup(seq->getDeformation(), settings);
      PointCloud *optUrshape = shapeOpt.optimizeUrshape(seq, shapeSolver, eval, dataPoints, settings);
      delete eval; eval = NULL;
      assignTopology(seq->getUrshape(), optUrshape);
      seq->getUrshape()->assign(optUrshape);
      delete optUrshape;
      IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
      IFDBO debugOutput.flush();
      show(seq->getUrshape(), makeVector4f(0.7f,1.0f,1.0f,1), visualFeedback, settings);
   }

   //
   // --- urshape deformation optimization
   //
   if (optimizeUrshapeDeformation) {
      IFDBO debugOutput << "   deforming urshape to best fit the data...\n";
      IFDBO debugOutput.flush();
      ARUrshapeDeformationOpt opt;
      PointCloud *defUrshape = opt.optimizeUrshape(seq, deformationSolver, dataPoints, settings);
      assignTopology(seq->getUrshape(), defUrshape);
      seq->getUrshape()->assign(defUrshape);
      delete defUrshape;
      IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
      IFDBO debugOutput.flush();
      show(seq->getUrshape(), makeVector4f(0.2f, 0.8f, 0.8f,1), visualFeedback, settings);
   }
   
   ARDeformationOptimizationContext *toDelete = NULL;
   //
   // --- change urshape (compute new context)
   //
   if (optimizeShape || optimizeUrshapeDeformation) {
      IFDBO debugOutput << "   recomputing deformation for changed urshape...\n";
      IFDBO debugOutput.flush();
      ARDeformationOptimizationContext *resampledUrshapeDeformation = ARChangeUrshape::recomputeDeformation(
         seq->getUrshape(),
		 originalUrshapePC,
         seq->getDeformation(), 
         deformationSolver, 
         settings);
      
      toDelete = seq->getDeformation();
      seq->setDeformation(resampledUrshapeDeformation);
      IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
      IFDBO debugOutput.flush();
   }

   // Cleanup saved urshape copy
   if ( originalUrshape != NULL ){
	   delete originalUrshape;
	   originalUrshape = NULL;
   }
   originalUrshapePC = NULL;
   
   //                                  
   // --- global deformation optimization
   //
   if (optimizeDeformation) {
      IFDBO debugOutput << "   global deformation optimization (fit deformation field to data)...\n";
      IFDBO debugOutput.flush();
      ARPruneUnsupportedOutliers puo(settings, (card32)(seq->getDuration()+1));
      ARGlobalDeformationOptimization globalDefOpt;
      globalDefOpt.globalOpt(
         seq,
         deformationSolver,
         dataPoints, settings, 
         removeUnsupportedPoints?&puo:NULL
      );
      IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
      IFDBO debugOutput.flush();

      if (removeUnsupportedPoints) {
         IFDBO debugOutput << "   culling urshape points not supported by sufficient data...\n";
         IFDBO debugOutput.flush();
         vector<bool> prunePts;
         puo.getResult(prunePts);
         prunePointsAndReplace(seq, prunePts);
         IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
         IFDBO debugOutput.flush();
      }

      show(seq->getUrshape(), makeVector4f(0.7f, 0.7f, 0.7f,1), visualFeedback, settings);
   }
   delete toDelete;
}


vector<PointCloud*> ARGlobalOptimization::getDataPoints(Scene *scene, string dataPath, card32 startFrame, card32 endFrame) 
{
   SGListNode *dataPointsList = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene, dataPath));
   if (!dataPointsList) {
      throw PException("ARGlobalOptimization::getDataPoints() - dataPath does not point to a list node.");
   }
   vector<PointCloud*> result;
   for (card32 i=startFrame; i<=endFrame; i++) {
      SGObjectNode *sg = dynamic_cast<SGObjectNode*>(dataPointsList->getChildNode(NULL, i));
      if (!sg) {
         throw PException("ARGlobalOptimization::getDataPoints() - one of the data nodes is not an SGObjectNode.");
      }
      PointCloud *pcData = dynamic_cast<PointCloud*>(sg->getSceneObject());
      if (!pcData) {
         throw PException("ARGlobalOptimization::getDataPoints() - no point cloud found in data node.");
      }
      result.push_back(pcData);
   }
   return result;
}

void ARGlobalOptimization::show(PointCloud *showPC, Vector4f showColor, Delegate repaint, ARSettings *settings) {
   try {
      myVisColor = showColor;
      myVisPC = showPC;
      visTool = NULL;
      repaint.callIfNotNull();
      if (settings->singleStepDebugVis) {
         static bool callOnce = false;
         if (!callOnce) {
            callOnce = true;
            if (QMessageBox::question(
               InterfaceGlobals::getMainWindow(), 
               "Single Step Visualization", "Stay in single step mode?", 
               QMessageBox::Yes | QMessageBox::No, 
               QMessageBox::Yes) == QMessageBox::No) 
            {
               settings->singleStepDebugVis = false;
            }
            callOnce = false;
         }
      }
   } catch (...) {
      myVisPC = NULL;
      myVisColor = makeVector4f(1,0,1,1);
      throw;
   }
   myVisPC = NULL;
   myVisColor = makeVector4f(1,0,1,1);
}

void ARGlobalOptimization::glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings) 
{
   if (visTool) {
      visTool->glDrawFeedback(glContext, visSettings);
   } else if (myVisPC) {
		SceneObjectIterator *it = myVisPC->createIterator(SOIT::CAP_BASIC_PC);
		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {delete it; return;}
		AAT positionAAT = myVisPC->getAAT("position");
		AAT normalAAT = myVisPC->getAAT("normal");
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glColor4fv(myVisColor.data());
		glBegin(GL_POINTS);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			glVertex3fv(pos.data());
			bit->next();
		}
		glEnd();
      bit->reset();
		glBegin(GL_LINES);
		glColor4f(0.8f,1.0f,0.7f,0.3f);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			Vector3f normal = bit->get3f(normalAAT);
			glVertex3fv(pos.data());
			glVertex3fv((pos+normal*0.03f).data());
			bit->next();
		}
		glEnd();
		delete it;
      InCorePCTopologyGraph *tg = getTopology(myVisPC);
      if (tg) {
         UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(myVisPC);
         if (upc) {
			   glDisable(GL_DEPTH_TEST);
			   glEnable(GL_BLEND);
			   glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      		//glColor4f(1.0f, 1.0f, 1.0f, 0.02f);
      		Vector4f newCol = myVisColor;
      		newCol[3] = 0.02f;
      		glColor4fv(newCol.data());
      		glBegin(GL_LINES);
            PointSet *ps = upc->getPointSet();
            for (card32 i=0; i<ps->getNumEntries(); i++) {
               card32 numAdj = tg->getNumAdjacentVertices(i);
               for (card32 j=0; j<numAdj; j++) {
                  glVertex3fv(ps->get3f(i, positionAAT).data());
                  glVertex3fv(ps->get3f(tg->getAdjacentVertex(i, j), positionAAT).data());
               }
            }
            glEnd();
         }
      }
      glPopAttrib();
   }
}
