#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARPairwiseMatching.h"
//---------------------------------------------------------------------------
#include "ARSurfaceDeformation.h"
#include "ICPConstraints.h"
#include "ARVisualizationSettings.h"
#include "DeformationSolver.h"
#include "DeformationEvaluator.h"
#include "AREvaluateDeformation.h"
#include "ARDeformationOptimizationContext.h"
#include "ARSettings.h"
#include "ARMergeSettings.h"
#include "DeformationTools.h"
#include "LocalICP.h"
#include "SparseGrid.h"
#include "DebugRenderer.h"
#include "VertexArray.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


ARPairwiseMatching::ARPairwiseMatching() {
	visPC1 = NULL;
	visPC2 = NULL;
	subTool = NULL;
	visAlternateColor = false;
}

ARPairwiseMatching::~ARPairwiseMatching() {
	if (!visPC1 == NULL || !visPC2 == NULL) {
		warning("ARPairwiseMatching - cleanup incorrect.");
	}
}

static void gridResamplingPatchICP(PointSet *ps, float32 cellSize) {
	PointSet *result = new PointSet();
	result->clearAndSetup(1,0, ps->getDescr());
	SparseGrid<3, bool, Object> grid;

	const card32 n = ps->getNumEntries();
	AAT positionAAT = ps->getAAT("position");
	for (card32 i=0; i<n; i++) {
		Vector3f pos = ps->get3f(i, positionAAT);
		Vector3i quant = makeVector3i(
			(int)((pos[0]+12345)/cellSize+0.5f),  // +12345 avoids artifacts at zero crossing
			(int)((pos[1]+12345)/cellSize+0.5f),
			(int)((pos[2]+12345)/cellSize+0.5f)
			);
		bool value;
		bool found = grid.getObject(quant, value);
		bool itsthere = value && found;
		if (!itsthere) {
			grid.setObject(quant, true);
			result->addPoint(ps->getDataPointer(i));
		}
	}
	ps->assign(result);
	delete result;
}


void ARPairwiseMatching::matchGeometry(
		Scene* scene,
		card32 frameNo,
   PointCloud *referenceShape, 
   PointCloud *matchingShape, 
   DeformationSolver *solver,
   const ARMergeSettings *settings,
   PointCloud *&deformedGeometry,
   ARDeformationOptimizationContext *&deformation,
   card32 &numIterationsNeeded,
   Delegate callForRepaint,
   bool alternateColor)
{
   visAlternateColor = alternateColor;
	deformation = (ARDeformationOptimizationContext*)solver->createOptContext();
	deformation->initializeFromShape(referenceShape, makeVector2f(0,0), settings->optimizationSettings);

	// compute patch icp constraints
	vector<PositionARConstraint> patchicpContraints;
	if( settings->optimizationSettings->patchICP_active )
	{
		HierarchicalKNNIterator hIt(referenceShape,32,NULL);
		bool visualizePatchICP = true;
		LocalICP * icp = settings->optimizationSettings->patchICP_localICP;

		// create resampled reference shape
		UnstructuredInCorePointCloud * resRef = (UnstructuredInCorePointCloud*)referenceShape->copy();
		gridResamplingPatchICP(resRef->getPointSet(), settings->optimizationSettings->patchICP_maxDist);
		VertexArray arr(resRef->getPointSet());

		static int t=0;
		if( (t++) % 2 == 0)
			debugRenderer->clearDebugData("patchicp");
		debugRenderer->beginRenderJob("patchicp");

		debugOutput << "Testing " << (unsigned int)resRef->getNumPoints() << " patches\n";
		debugOutput.flush();
		for( unsigned i=0;i<arr.getNumElements();i++ )		
		{
			Vector3f pos = arr.getPosition3f(i);
			Vector3f normal = arr.getNormal3f(i);

			// compute slippage
			icp->beginICP(referenceShape,matchingShape);
			Matrix2f h = icp->computeSlippage(pos,normal,pos,IDENTITY3F );
			Matrix3f covariance;
			if( icp->computeSlippageCovariance(h, normal, covariance) )
			{
				debugRenderer->addPoint(pos, makeVector3f(1,0,0));
				covariance += outerProduct(normal,normal);
				Matrix3f rotation = IDENTITY3F;
				Vector3f destpos = pos;
				if( icp->computeICP(pos,destpos,rotation) )
				{
					hIt.setSeekPointAndReset(pos);
					card32 index = hIt.getCurrentPointIndex();
					PositionARConstraint c;
					c.covariance = IDENTITY3F;//rotation*covariance;
					c.anchor = ARConstraintAnchor(pos, index);
					c.time = 0;					
					c.toPoint = destpos;
					patchicpContraints.push_back(c);
					debugRenderer->addLine( pos, destpos, makeVector3f(0.5,0,0),makeVector3f(1,1,0),2);
				}

			}
			else
			{
				debugRenderer->addPoint(pos, makeVector3f(0,1,1));
			}


		}

		debugRenderer->endRenderJob();


		delete resRef;
	}

	
	deformedGeometry = (PointCloud*)referenceShape->copy();
	visPC1 = deformedGeometry;
	visPC2 = matchingShape;
	callForRepaint.callIfNotNull();
   visPC1 = NULL;
   visPC2 = NULL;
	solver->setup(deformation, settings->optimizationSettings);
	card32 numIterations = 0;
	card32 numPasses = (settings->optimizationSettings->icpRigidityMultiPassFactor!=1.0f)
	                   ? settings->optimizationSettings->icpNumRPasses:1;
	float32 originalRigidity = settings->optimizationSettings->rigidityParam;
	settings->optimizationSettings->rigidityParam *= settings->optimizationSettings->icpRigidityInitialFactor;
	for (card32 pass=0; pass<numPasses; pass++) {
      if (settings->optimizationSettings->mergeDebugOutput && numPasses > 1) {
   	   debugOutput << "   pass " << pass << " rigidity " << settings->optimizationSettings->rigidityParam; debugOutput.flush();
	   }
	   card32 i;
      for (i=0; i<settings->optimizationSettings->maxNumICPIterations; i++) {

		   if (i+numIterations > settings->optimizationSettings->maxNumICPIterations) break;

		   ICPConstraints icp;
         icp.setConstraintMode(settings->optimizationSettings->icpConstraintsMode);
         icp.setDistanceOutliersMode(settings->optimizationSettings->icpPercentileCulling?ICP_OUTLIERS_PERCENTILE:ICP_OUTLIERS_NONE);
         // this value is also necessary to avoid culling in the near field
         icp.setOutliersDistanceFactor(settings->optimizationSettings->icpNearFieldRelativeCutOff);
         icp.setOutliersPercentile(settings->optimizationSettings->icpOutlierPercentile);
         icp.setCullNormalOutliers(settings->optimizationSettings->icpNormalCulling);
         icp.setOutlierNormals(settings->optimizationSettings->icpNormalCosAngleCutOff);

		   solver->beginConstraints();
			
  		   icp.buildICPConstraints(
  		      settings->optimizationSettings,
  		      solver, 
  		      referenceShape, 
  		      deformedGeometry, 
  		      matchingShape,
  		      0,
  		      callForRepaint,
  		      settings->visualizationSettings->showConstraints,
		      settings->optimizationSettings->icpSkipPercentage);
			
			// patch icp contraints
			for( unsigned ci=0;ci<patchicpContraints.size();ci++ )
			{
				solver->addPositionConstraint(&patchicpContraints[ci]);
			}
		   solver->solve();

   	   
		   DeformationEvaluator *evaluator = solver->createEvaluator();
		   DVectorF lastPositions = convertPCToVector(deformedGeometry);
		   delete deformedGeometry;
		   deformedGeometry = AREvaluateDeformation::evaluatePC(referenceShape, deformation, evaluator, 0, settings->optimizationSettings);
		   delete evaluator;
		   DVectorF newPositions = convertPCToVector(deformedGeometry);
   		
		   visPC1 = deformedGeometry;
		   visPC2 = matchingShape;
		   subTool = &icp;
		   callForRepaint.callIfNotNull();
		   visPC1 = NULL;
		   visPC2 = NULL;
	      subTool = NULL;

         float32 movedBy = normInfinity(newPositions-lastPositions);

		   if (movedBy <= settings->optimizationSettings->convThreshICP) {
		      break;
		   }
      }
      if (settings->optimizationSettings->mergeDebugOutput && numPasses > 1) {
   	   debugOutput << " it " <<i << "\n"; debugOutput.flush();
   	}
      numIterations += i+1;
      settings->optimizationSettings->rigidityParam *= settings->optimizationSettings->icpRigidityMultiPassFactor;
   }
	settings->optimizationSettings->rigidityParam = originalRigidity;
   numIterationsNeeded = numIterations;
   debugOutput << "    number of iterations: " << numIterations+1 << "\n";
}

void ARPairwiseMatching::glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings) {
	if (visPC1 && visSettings->showVisualizations) {
		SceneObjectIterator *it = visPC1->createIterator(SOIT::CAP_BASIC_PC);
		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {delete it; return;}
		AAT positionAAT = visPC1->getAAT("position");
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		if (visAlternateColor) {
   		glColor4f(1,0.5f,0,1);
		} else {
   		glColor4f(1,1,0,1);                                 
		}
		glBegin(GL_POINTS);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			glVertex3fv(pos.data());
			bit->next();
		}
		glEnd();
		delete it;

		it = visPC2->createIterator(SOIT::CAP_BASIC_PC);
		bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {delete it; return;}
		positionAAT = visPC2->getAAT("position");
		if (visAlternateColor) {
   		glColor4f(0.65f,0.65f,0.65f,1);
		} else {
   		glColor4f(1, 1, 1,1);
		}
		glBegin(GL_POINTS);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			glVertex3fv(pos.data());
			bit->next();
		}
		glEnd();
		glPopAttrib();
		
		delete it;

	}
	if (subTool) subTool->glDrawFeedback(glContext, visSettings);
}
