//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARTopologicalResampling.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "TopologyAttachment.h"
#include "TopologyRangeSearch.h"
#include "InCorePCTopologyGraph.h"
#include "ARSettings.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


struct PQEntry {
	card32 index;
	float32 dist;
	bool operator<(const PQEntry &other) const {
		return dist < other.dist;
	}
};


UnstructuredInCorePointCloud *ARTopologicalResampling::resample(
   UnstructuredInCorePointCloud* pc, 
   float32 maxDist, 
   float32 reconnectTopologyRelativeRadius,
   card32  reconnectTopologyMaxNumNeighbors,
   vector<bool>* deletePtn_User)
{
	const PointSet *ps = pc->getPointSet();
	const card32 numPoints = ps->getNumEntries();
	AttachedDataContainer *attachments = pc->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (!topAtt) {
	   throw PException("ARTopologicalResampling::resample() - input point cloud does not have a topology attachment.");
	}
	const InCorePCTopologyGraph *tg = topAtt->getTopology();

   // --

	vector<card32> overlap;
	overlap.resize(numPoints, 0);
	vector< vector<card32> > overlapPoints;
	overlapPoints.resize(numPoints);

	vector<bool> deletePtn_;
	vector<bool>& deletePtn(*(deletePtn_User?deletePtn_User:&deletePtn_));
	deletePtn.resize(numPoints, false);
	std::fill(deletePtn.begin(), deletePtn.end(), false);

	AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);

   TopNeighborhoodKNNSearch knn(ps, tg, 0, positionAAT);
	
	priority_queue<PQEntry> removalCandidates;

	for (card32 i=0; i<numPoints; i++) {
		knn.reset(i);
		card32 index = 0;
		bool found = !knn.atEnd();
		pAssert(found);
	   index = knn.getCurrentIndex();
	   Vector3f pos_i = ps->get3f(i, positionAAT);
	   Vector3f pos_index = ps->get3f(index, positionAAT);
	   knn.next();
		float32 minDistance = 1E25f;
		while (found && norm(pos_i - pos_index) <= maxDist) {
			overlap[i]++;
			overlapPoints[i].push_back(index);
         found = !knn.atEnd();
			if (found) {
			   index = knn.getCurrentIndex();
      	   pos_index = ps->get3f(index, positionAAT);
			   knn.next();
				float32 distance = norm(pos_i - pos_index);
				if (distance < minDistance) {
					minDistance = distance;
				}
			}
		}
		if (minDistance < 1E25f) {
			PQEntry e;
			e.dist = -minDistance;
			e.index = i;
			removalCandidates.push(e);
		}
	}
	
	card32 newNumPoints = numPoints;
	
	while (!removalCandidates.empty()) {
		PQEntry e = removalCandidates.top();
		removalCandidates.pop();
		if (overlap[e.index] > 1) {
			bool othersOk = true;
			const size_t otherCount = overlapPoints[e.index].size();
			for (card32 i=0; i<otherCount; i++) {
				if (overlap[overlapPoints[e.index][i]] <= 1) {
					othersOk = false;
				}
			}
			// check for constrained points
			if (othersOk) {
				overlap[e.index]--;
				for (card32 i=0; i<otherCount; i++) {
				   if (overlapPoints[e.index][i] != e.index) {
   					overlap[overlapPoints[e.index][i]]--;
					}
				}
				deletePtn[e.index] = true;
				pAssert(newNumPoints > 0);
				newNumPoints--;
			}
		}	
	}

   

	UnstructuredInCorePointCloud *result = new UnstructuredInCorePointCloud();
	result->clearAndSetup(pc->getDescr(), newNumPoints);
	PointSet *resultPS = result->getPointSet();

   /// contains list of new indices for each original point
   /// -1 means point has been removed
   vector<int32> newIndices(numPoints, -1);
   /// contains list of original indices for each new point
   vector<int32> oldIndices(newNumPoints, -1);
   
	card32 targetPos = 0;
	for (card32 oldIndex=0; oldIndex<numPoints; oldIndex++) {
		if (!deletePtn[oldIndex]) {
			pAssert(targetPos < newNumPoints);
			resultPS->setVertex(targetPos, ps->getDataPointer(oldIndex));
			newIndices[oldIndex] = targetPos;
			oldIndices[targetPos] = oldIndex;
			targetPos++;
		}
	}

	InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();
	resultTg->clearAndSetup(tg->getEdges()->getDescr());
	
	for (card32 newIndex=0; newIndex<newNumPoints; newIndex++) {
	   int32 searchCenter = oldIndices[newIndex];
	   pAssert(searchCenter >= 0);
      Vector3f searchCenterPos = ps->get3f(searchCenter, positionAAT);
	   knn.reset(searchCenter);
	   card32 numEdges = 0;
      while (!knn.atEnd()) {
         int32 currentOldPoint = knn.getCurrentIndex();
         if (currentOldPoint != searchCenter) {
            Vector3f currentOldPointPos = ps->get3f(currentOldPoint, positionAAT);
            float32 distance = norm(currentOldPointPos-searchCenterPos);
            if (distance > reconnectTopologyRelativeRadius * maxDist) {
               break;
            }
            if (newIndices[currentOldPoint] != -1) {
               //float32 newDist = norm(resultPS->get3f(newIndex, positionAAT) - resultPS->get3f(newIndices[currentOldPoint], positionAAT));
               //if (newDist >0.2) {
               //   float q = numEdges;
               //}
               resultTg->addEdge(newIndex, newIndices[currentOldPoint]);
               numEdges++;
               if (numEdges >= reconnectTopologyMaxNumNeighbors) {
                  break;
               }
            }
         
         }
         knn.next();
      }
	}
	                        
	TopologyAttachment *newTopAtt = new TopologyAttachment();
   newTopAtt->setup("TopologyAtt", AttachedData::ADF_PERSISTENT);
	newTopAtt->setTopology(resultTg);
	result->getAttachments()->attachData(newTopAtt);	

	return result;   
}

void ARTopologicalResampling::resampleInPlace(PointCloud *pc, ARSettings *settings) 
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
   if (!upc) {
      notImplemented();
   }
   UnstructuredInCorePointCloud *res = resample(
      upc, 
      settings->topResamplingSurfelSpacing,
      settings->topRelRadius,
      settings->topMaxNumNeighbors);
   upc->assign(res);
   delete res;
}







void geoxl::ARTopologicalResampling::resampleMeshFindToDeleteVertices(
	UnstructuredInCorePointCloud* pc, 
	float32 maxDist, 
	//	float32 reconnectTopologyRelativeRadius,
	//	card32  reconnectTopologyMaxNumNeighbors,
	vector<bool>& toDelete )
{
	const PointSet *ps = pc->getPointSet();
	const card32 numPoints = ps->getNumEntries();
	AttachedDataContainer *attachments = pc->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (!topAtt) {
		throw PException("ARTopologicalResampling::resample() - input point cloud does not have a topology attachment.");
	}
	const InCorePCTopologyGraph *tg = topAtt->getTopology();
	AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	TopNeighborhoodKNNSearch knn(ps, tg, 0, positionAAT);


	// --

	vector<card32> overlap;
	overlap.resize(numPoints, 0);
	vector< vector<card32> > overlapPoints;
	overlapPoints.resize(numPoints);

	vector<bool>& deletePtn(toDelete);
	deletePtn.resize(numPoints, false);
	std::fill(deletePtn.begin(), deletePtn.end(), false);


	priority_queue<PQEntry> removalCandidates;

	for (card32 i=0; i<numPoints; i++) {
		knn.reset(i);
		card32 index = 0;
		bool found = !knn.atEnd();
		pAssert(found);
		index = knn.getCurrentIndex();
		Vector3f pos_i = ps->get3f(i, positionAAT);
		Vector3f pos_index = ps->get3f(index, positionAAT);
		knn.next();
		float32 minDistance = 1E25f;
		while (found && norm(pos_i - pos_index) <= maxDist) {
			overlap[i]++;
			overlapPoints[i].push_back(index);
			found = !knn.atEnd();
			if (found) {
				index = knn.getCurrentIndex();
				pos_index = ps->get3f(index, positionAAT);
				knn.next();
				float32 distance = norm(pos_i - pos_index);
				if (distance < minDistance) {
					minDistance = distance;
				}
			}
		}
		if (minDistance < 1E25f) {
			PQEntry e;
			e.dist = -minDistance;
			e.index = i;
			removalCandidates.push(e);
		}
	}

	card32 newNumPoints = numPoints;

	while (!removalCandidates.empty()) {
		PQEntry e = removalCandidates.top();
		removalCandidates.pop();
		if (overlap[e.index] > 1) {
			bool othersOk = true;
			const size_t otherCount = overlapPoints[e.index].size();
			for (card32 i=0; i<otherCount; i++) {
				if (overlap[overlapPoints[e.index][i]] <= 1) {
					othersOk = false;
				}
			}
			// check for constrained points
			if (othersOk) {
				overlap[e.index]--;
				for (card32 i=0; i<otherCount; i++) {
					if (overlapPoints[e.index][i] != e.index) {
						overlap[overlapPoints[e.index][i]]--;
					}
				}
				deletePtn[e.index] = true;
				pAssert(newNumPoints > 0);
				newNumPoints--;
			}
		}	
	}
}

UnstructuredInCorePointCloud* geoxl::ARTopologicalResampling::resampleMeshAndTopology(
	UnstructuredInCorePointCloud* pc, 
	float32 maxDist, 
	float32 reconnectTopologyRelativeRadius,
	card32  reconnectTopologyMaxNumNeighbors,
	bool reconnectTopologyInsertBridgeEdges,
	const vector<bool>& toDelete) {
		const PointSet *ps = pc->getPointSet();
		const card32 numPoints = ps->getNumEntries();
		AttachedDataContainer *attachments = pc->getAttachments();
		TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
		if (!topAtt) {
			throw PException("ARTopologicalResampling::resample() - input point cloud does not have a topology attachment.");
		}
		const InCorePCTopologyGraph *tgOld = topAtt->getTopology();
		card32 newNumPoints = std::count(toDelete.begin(), toDelete.end(), false );
		AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
		TopNeighborhoodKNNSearch knn(ps, tgOld, 0, positionAAT);



		UnstructuredInCorePointCloud *result = new UnstructuredInCorePointCloud();
		result->clearAndSetup(pc->getDescr(), newNumPoints);
		PointSet *resultPS = result->getPointSet();

		/// contains list of new indices for each original point
		/// -1 means point has been removed
		vector<int32> newIndices(numPoints, -1);
		/// contains list of original indices for each new point
		vector<int32> oldIndices(newNumPoints, -1);

		card32 targetPos = 0;
		for (card32 oldIndex=0; oldIndex<numPoints; oldIndex++) {
			if (!toDelete[oldIndex]) {
				pAssert(targetPos < newNumPoints);
				resultPS->setVertex(targetPos, ps->getDataPointer(oldIndex));
				newIndices[oldIndex] = targetPos;
				oldIndices[targetPos] = oldIndex;
				targetPos++;
			}
		}

		const InCorePCTopologyGraph *tg = NULL;
		if( reconnectTopologyInsertBridgeEdges ) {
			InCorePCTopologyGraph *tg_ = dynamic_cast<InCorePCTopologyGraph*>( tgOld->copy() );

			for( int i1 = 0; i1 < (int)toDelete.size(); ++ i1 ) {
				if( !toDelete[i1] )
					continue;

				int replace = -1;
				float bestDist;
				Vector3f p1( ps->get3f( i1, positionAAT ) );
				for( int i2 = 0; i2 < (int)tg_->getNumAdjacentVertices(i1); ++ i2 ) {
					int cur = tg_->getAdjacentVertex(i1,i2);
					Vector3f p2( ps->get3f( cur, positionAAT ) );
					float dist = (p1-p2).getSqrNorm();
					if( replace == -1 || dist < bestDist ) {
						bestDist = dist;
						replace = cur;
					}
				}
				for( int i2 = 0; i2 < (int)tg_->getNumAdjacentVertices(i1); ++ i2 ) {
					int cur = tg_->getAdjacentVertex(i1,i2);
					tg_->addEdge(replace, cur);
				}
//				tg_->clearVertex(i1); // takes too much time
			}
			tg = tg_;
		} else {
			tg = tgOld;
		}

		InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();


		resultTg->clearAndSetup(tg->getEdges()->getDescr());

		for (card32 newIndex=0; newIndex<newNumPoints; newIndex++) {
			int32 searchCenter = oldIndices[newIndex];
			pAssert(searchCenter >= 0);
			Vector3f searchCenterPos = ps->get3f(searchCenter, positionAAT);
			knn.reset(searchCenter);
			card32 numEdges = 0;
			while (!knn.atEnd()) {
				int32 currentOldPoint = knn.getCurrentIndex();
				if (currentOldPoint != searchCenter) {
					Vector3f currentOldPointPos = ps->get3f(currentOldPoint, positionAAT);
					float32 distance = norm(currentOldPointPos-searchCenterPos);
					if (distance > reconnectTopologyRelativeRadius * maxDist) {
						break;
					}
					if (newIndices[currentOldPoint] != -1) {
						//float32 newDist = norm(resultPS->get3f(newIndex, positionAAT) - resultPS->get3f(newIndices[currentOldPoint], positionAAT));
						//if (newDist >0.2) {
						//   float q = numEdges;
						//}
						resultTg->addEdge(newIndex, newIndices[currentOldPoint]);
						numEdges++;
						if (numEdges >= reconnectTopologyMaxNumNeighbors) {
							break;
						}
					}

				}
				knn.next();
			}
		}

		TopologyAttachment *newTopAtt = new TopologyAttachment();
		newTopAtt->setup("TopologyAtt", AttachedData::ADF_PERSISTENT);
		newTopAtt->setTopology(resultTg);
		result->getAttachments()->attachData(newTopAtt);	

		return result;   
}
