//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARUrshapeOptimization.h"
//---------------------------------------------------------------------------
#include "ARVisualizationSettings.h"
#include "ARSequence.h"
#include "ARSettings.h"
#include "ShapeReconstruction.h"
#include "AROptimizationContext.h"
#include "ARShapeOptimizationContext.h"
#include "ARTransfConstraintsToUrshape.h"
#include "ICPConstraints.h"
#include "DeformationEvaluator.h"
#include "AREvaluateDeformation.h"
#include "ShapeEvaluator.h"
#include "ARDeformationOptimizationContext.h"
#include "TopologyAttachment.h"
#include "SurfelOptimizationContext.h"
//#include "LocalSurfaceFunction.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



ARUrshapeOptimization::ARUrshapeOptimization() {
   visPC = NULL;
}

void ARUrshapeOptimization::glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings) {
   if (visPC) {
		SceneObjectIterator *it = visPC->createIterator(SOIT::CAP_BASIC_PC);
		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {delete it; return;}
		AAT positionAAT = visPC->getAAT("position");
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glColor4f(1,1,0,1);
		glBegin(GL_POINTS);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			glVertex3fv(pos.data());
			bit->next();
		}
		glEnd();
		glPopAttrib();
   }
   forget(glContext);
}

PointCloud *ARUrshapeOptimization::optimizeUrshape(
   ARSequence *seq, 
   ShapeReconstruction *shapeOpt, 
   DeformationEvaluator *deformEval,
   const vector<PointCloud*> &dataPoints,
   ARSettings *settings) 
{
   ARShapeOptimizationContext *recContext = (ARShapeOptimizationContext*)shapeOpt->createOptContext();
   recContext->initializeFromShape(seq->getUrshape(), settings);
   shapeOpt->setup(recContext, settings);
   shapeOpt->beginConstraints();

   const card32 numFrames = seq->getDuration()+1;
   
   for (card32 i=0; i<numFrames; i++) {
      ICPConstraints icp;
      icp.setConstraintMode(settings->icpConstraintsMode);
      icp.setDistanceOutliersMode(settings->icpNearFieldCulling?ICP_OUTLIERS_DISTANCE:ICP_OUTLIERS_NONE);
      icp.setOutliersDistanceFactor(settings->icpNearFieldRelativeCutOff);
      icp.setCullNormalOutliers(false);

      deformEval->setup(seq->getDeformation(), settings);
      
      PointCloud *deformedReference = AREvaluateDeformation::evaluatePC(
         seq->getUrshape(), 
         seq->getDeformation(), 
         deformEval,
         i,
         settings);

      deformEval->beginQuery(i);
      ARTransfConstraintsToUrshape adaptor(deformEval, shapeOpt);
      
	   icp.buildICPConstraints(
		   settings, &adaptor, seq->getUrshape(),
		   deformedReference,
		   dataPoints[i],
		   0,
		   NULL, false,
		   settings->icpSkipPercentage
	   );
	   delete deformedReference;
      deformEval->endQuery();
   }
   
   shapeOpt->solve();
   
   ShapeEvaluator *shapeEval = shapeOpt->createEvaluator();
   shapeEval->setup(recContext, settings);
   PointCloud *result = shapeEval->evaluate();

	AttachedDataContainer *attachments = seq->getUrshape()->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt != NULL) {
      result->getAttachments()->attachData((TopologyAttachment*)topAtt->copy());
	}

   delete recContext;
   delete shapeEval;
   return result;
}


PointCloud *ARUrshapeOptimization::optimizeSingleFrame(
   PointCloud *pc, 
   ShapeReconstruction *shapeOpt, 
   PointCloud *dataPoints,
   ARSettings *settings,
   Delegate callForRedraw)
{
   ARShapeOptimizationContext *recContext = (ARShapeOptimizationContext*)shapeOpt->createOptContext();

   ICPConstraints icp;
   icp.setConstraintMode(settings->icpConstraintsMode);
   icp.setOutliersDistanceFactor(settings->icpNearFieldRelativeCutOff);
   icp.setDistanceOutliersMode(settings->icpNearFieldCulling?ICP_OUTLIERS_DISTANCE:ICP_OUTLIERS_NONE);
   icp.setCullNormalOutliers(false);

   PointCloud *result = (PointCloud*)pc->copy();   
   visPC = result;
   callForRedraw.callIfNotNull();
   visPC = NULL;
   ShapeEvaluator *shapeEval = shapeOpt->createEvaluator();
   
   for (card32 i=0; i<settings->maxNumShapeOptIterations; i++) {
      recContext->initializeFromShape(result, settings);
      shapeOpt->setup(recContext, settings);
      shapeOpt->beginConstraints();
      icp.buildICPConstraints(settings, shapeOpt, result, result, dataPoints, 0, NULL, false, settings->icpSkipPercentage);
      shapeOpt->solve();

      shapeEval->setup(recContext, settings);
      PointCloud *oldResult = result;
      result = shapeEval->evaluate();
      delete oldResult;

	   AttachedDataContainer *attachments = pc->getAttachments();
	   TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	   if (topAtt != NULL) {
         result->getAttachments()->attachData((TopologyAttachment*)topAtt->copy());
	   }

      visPC = result;
      callForRedraw.callIfNotNull();
      visPC = NULL;
   }
   
   delete shapeEval;
   delete recContext;
   return result;
}
