//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PCCBoundaryDetector.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSetANNQuery.h"
#include "PCA.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS(PCCBoundaryDetector, 1)
{
	BEGIN_CLASS_INIT(PCCBoundaryDetector)
   ADD_CARD32_PROP(numNeighbors,0)
   ADD_FLOAT32_PROP(sigmaRel, 0)
   ADD_FLOAT32_PROP(quantizeThreshold, 1)
   ADD_BOOLEAN_PROP(quantize, 1)
   ADD_BOOLEAN_PROP(boundaryTangents, 1)
}

//---------------------------------------------------------------------------

PCCBoundaryDetector::PCCBoundaryDetector() {
   numNeighbors = 20;
   sigmaRel = 1;
   quantizeThreshold = 0.75f;
   quantize = true;
   boundaryTangents = true;
}

PCCBoundaryDetector::~PCCBoundaryDetector() {}

void PCCBoundaryDetector::assign(const Object* obj, CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------

void PCCBoundaryDetector::detectBoundaries(PointCloud *pc) {
   if (!pc->providesAttribute("position")) {
      warning("no position channel - selected pc ignored.");
      return;
   }
	if (!pc->providesAttribute("normal")) {
		warning("no normal channel - selected pc ignored.");
		return;
	}
   checkAttribute(pc, "boundary", 1, VAD::DATA_FORMAT_FLOAT32);
   if (boundaryTangents) {
      checkAttribute(pc, "boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);
   }
	SceneObjectIterator * it = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	if (!it) return;
	BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
	ModifyablePointCloudIterator *mit = dynamic_cast<ModifyablePointCloudIterator*>(it);
	PointSet *ps = mit->lockBlockReadWrite();
	AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	AAT normalAAT = ps->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
	AAT boundaryAAT = ps->getAAT("boundary", 1, VAD::DATA_FORMAT_FLOAT32);
	PointSetANNQuery query(ps, numNeighbors);
	mpcard numPts = ps->getNumEntries();
	for (mpcard i=0; i<numPts; i++) {
	   Vector3f pos = ps->get3f(i, positionAAT);
	   //Vector3f tangent_u = ps->get3f(i, tangent_uAAT);
	   //Vector3f tangent_v = ps->get3f(i, tangent_vAAT);
		Vector3f normal = ps->get3f(i, normalAAT);
		Matrix3f tangentFrame = calcTangentSystem(normal);
		const Vector3f tangent_u = tangentFrame[0];
		const Vector3f tangent_v = tangentFrame[1];
	   query.setTargetPoint(pos);
	   card32 k;
	   int32 *buf = query.getAllIndices(k);
	   vector<Vector2f> buf2D(k);
	   for (card32 j=0; j<k; j++) {
	      Vector3f d = pos - ps->get3f(buf[j], positionAAT);
	      buf2D[j] = makeVector2f(d*tangent_u, d*tangent_v);
	   }
	   
	   float32 avDist = 0;
	   for (card32 j=0; j<k; j++) {
	      avDist += norm(buf2D[j]);
	   }
	   avDist /= k;
	   static const card32 NUM_DISCR = 24;
	   float32 minMaxR = 1e20f;
	   for (card32 r=0; r<NUM_DISCR; r++) {
	      float32 alpha = (float32)r/(float32)NUM_DISCR*M_PI*2;
	      Vector2f rv = makeVector2f(sin(alpha), cos(alpha));
	      float32 maxR = -1e20f;
		   for (card32 j=0; j<k; j++) {
		      float32 rd = rv*buf2D[j];
		      if (maxR < rd) maxR = rd;
		   }
		   if (maxR < minMaxR) minMaxR = maxR;
	   }
	   float32 value = exp(-minMaxR*minMaxR/(avDist*avDist*sigmaRel*sigmaRel));
	   ps->set1f(i, boundaryAAT, value);
	}
   if (boundaryTangents) {
		AAT boundaryTangentsAAT = ps->getAAT("boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);
		for (mpcard i=0; i<numPts; i++) {
	      Vector3f pos = ps->get3f(i, positionAAT);
			Vector3f normal = ps->get3f(i, normalAAT);
			Matrix3f tangentFrame = calcTangentSystem(normal);
			const Vector3f tangent_u = tangentFrame[0];
			const Vector3f tangent_v = tangentFrame[1];
	      query.setTargetPoint(pos);
	      card32 k;
	      int32 *buf = query.getAllIndices(k);
	      PCA<float32, 2> pca;
	      card32 ptsInPCA = 0;
	      for (card32 j=0; j<k; j++) {
	         Vector3f d = pos - ps->get3f(buf[j], positionAAT);
	         Vector2f d2 = makeVector2f(d*tangent_u, d*tangent_v);
	         float32 b = ps->get1f(buf[j], boundaryAAT);
	         pca.addPoint(d2, b); 
	         if (b > 0.01) ptsInPCA++;
	      }
	      Vector2f center, values;
	      Matrix2f vects;
	      if (ptsInPCA >= 2) {
	         pca.analyze(values, vects, center);
	         ps->set3f(i, boundaryTangentsAAT, tangent_u*vects[0][0] + tangent_v*vects[0][1]);
	      } else {
	         ps->set3f(i, boundaryTangentsAAT, NULL_VECTOR3F);
	      }
      }
   }		
   if (quantize) {
		for (mpcard i=0; i<numPts; i++) {
		   float32 value = ps->get1f(i, boundaryAAT);
         if (value > quantizeThreshold) {
            value = 1;
         } else {
            value = 0;
         }
	      ps->set1f(i, boundaryAAT, value);
	   }
   }
	bit->unlock();
	delete it;
}

void PCCBoundaryDetector::execute(Scene *scene) 
{
   vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m=0; m<selection.size(); m++) {
      PointCloud *pc = getPointCloud(scene, selection[m]);
      detectBoundaries(pc);
   }
}
