//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCDataResampling.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "SGListNode.h"
#include "SGObjectNode.h"
#include "PointCloud.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "DeformationTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "UniformPointCloudResampling.h"
#include "ProgressWindow.h"
#include "SparseGrid.h"
#include "SparseGrid_templCode.h"
#include "SceneObjectIterator.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


string describeMethod(card32 i) {
	if (i == 0) return "METHOD_POINT_REMOVAL";
	if (i == 1) return "METHOD_GRID";
	if (i == 2) return "METHOD_HYBRID";
	return "undefined";
}


IMPLEMENT_GEOXL_CLASS( PCCDataResampling , 2 ) {
	BEGIN_CLASS_INIT( PCCDataResampling );
	ADD_STRING_PROP(sourceFolder, 0)
	ADD_STRING_PROP(targetFolder, 0)
	ADD_STRING_PROP(filenamePrefix, 1)
	ADD_BOOLEAN_PROP(avoidZeroCrossing, 1)
	
   ADD_FLOAT32_PROP(maxDist, 0)
   ADD_CARD32_PROP(frameResampling, 0)
   ADD_CARD32_PROP_UPDATE_DESCR(method, 0, NULL, describeMethod, 3)
}

// ----

PCCDataResampling::PCCDataResampling() {
	sourceFolder = "root/data";
	targetFolder = "root/lowresData";
	filenamePrefix = "resampled_";
   maxDist = 0.025f;
   frameResampling = 1;
   avoidZeroCrossing = false;

   method = METHOD_GRID;
}

void PCCDataResampling::setup(
	string sourceFolder,
	string targetFolder,
	float32 maxDist,
	card32 frameResampling)
{
	this->sourceFolder = sourceFolder;
	this->targetFolder = targetFolder;
	this->maxDist = maxDist;
	this->frameResampling = frameResampling;
}

void PCCDataResampling::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

PCCDataResampling::~PCCDataResampling() {
}


static void gridResampling(PointSet *ps, float32 cellSize, bool avoidZeroCrossing) {
	PointSet *result = new PointSet();
	result->clearAndSetup(1,0, ps->getDescr());
	SparseGrid<3, bool, Object> grid;

	const card32 n = ps->getNumEntries();
	AAT positionAAT = ps->getAAT("position");
	for (card32 i=0; i<n; i++) {
		Vector3f pos = ps->get3f(i, positionAAT);
		Vector3i quant = makeVector3i(
			(int)((pos[0]+avoidZeroCrossing?12345:0)/cellSize+0.5f),  // +12345 avoids artifacts at zero crossing
			(int)((pos[1]+avoidZeroCrossing?12345:0)/cellSize+0.5f),
			(int)((pos[2]+avoidZeroCrossing?12345:0)/cellSize+0.5f)
		);
		bool value;
		bool found = grid.getObject(quant, value);
		bool itsthere = value && found;
		if (!itsthere) {
			grid.setObject(quant, true);
			result->addPoint(ps->getDataPointer(i));
		}
	}
	ps->assign(result);
	delete result;
}



void PCCDataResampling::execute(Scene *scene) 
{
	SGListNode *sourceNode = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene, sourceFolder));
	if (!sourceNode) {
		error("PCCDataResampling::execute() - sourceNode not found (or not an SGListNode).");
		return;
	}																															  
	SGListNode *targetNode = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene, targetFolder));
	if (!targetNode) {
		error("PCCDataResampling::execute() - targetNode not found (or not an SGListNode).");
		return;
	}
	const card32 n = sourceNode->getNumChildNodes(NULL);
	progressWindow->pushStep();
	for (card32 i=0; i<n; i++) {
		if (frameResampling > 1) {
			if ((i % frameResampling) != 0) continue;
		}
		SGObjectNode *sgo = dynamic_cast<SGObjectNode*>(sourceNode->getChildNode(NULL, i));
		if (!sgo) {
			warning("PCCDataResampling::execute() - non object node found, ignored.");
			continue;
		}
		PointCloud *pc = dynamic_cast<PointCloud*>(sgo->getSceneObject());
		if (!pc) {
			warning("PCCDataResampling::execute() - empty object (or no point cloud).");
			continue;
		}
		SceneObjectIterator *it = pc->createIterator(SOIT::CAP_BASIC_PC);
		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {
			delete it;
			warning("PCCDataResampling::execute() - cannot create basic iterator.");			
			continue;
		}

		PointSet *ps = (PointSet*)(bit->lockBlockRead()->copy());
		bit->unlock();
		delete it;
		card32 numPointsInput = ps->getNumEntries();
		
		if (method == METHOD_GRID) {
			gridResampling(ps, maxDist/(2.0f*sqrt(3.0f)), avoidZeroCrossing);
		} else if (method == METHOD_HYBRID) {
			// fine resampling first
			gridResampling(ps, maxDist/(4.0f*sqrt(3.0f)), avoidZeroCrossing);	
		}
		if (method == METHOD_POINT_REMOVAL || method == METHOD_HYBRID) {
			UniformPointCloudResampling resampler;
			resampler.setup(maxDist);
			PointSet *resPS = resampler.resamplePointSet(ps);
			ps->assign(resPS);
			delete resPS;
		}

		card32 numPointsOutput = ps->getNumEntries();
		float32 ratioOutIn = (float32)numPointsOutput / (float32)numPointsInput;
		debugOutput << "  resampled to " << ratioOutIn*100 << "% of the original size\n";
		
		if (ps->providesAttribute("errorQuad", 6, VAD::DATA_FORMAT_FLOAT32)) {
		   debugOutput << "  rescaling erorr quadrics accordingly...\n";
		   AAT errorQuadAAT = ps->getAAT("errorQuad");
		   for (card32 p=0; p<numPointsOutput; p++) {
		      ps->set<float32, 6>(p, errorQuadAAT, ps->get<float32, 6>(p, errorQuadAAT));
		   }
		}
		
		PointCloud *resultPC = (PointCloud*)pc->createOOCAwareCopy(filenamePrefix, true, "");
		
		it = resultPC->createIterator(SOIT::CAP_MODIFYABLE_PC);
		bit = dynamic_cast<BasicPointCloudIterator*>(it);
		ModifyablePointCloudIterator *mit = dynamic_cast<ModifyablePointCloudIterator*>(it);
		if (!(mit && bit)) {
			delete it;
			delete ps;
			warning("PCCDataResampling::execute() - cannot write to output pc.");
			continue;
		}
		
		PointSet *targetPS = mit->lockBlockReadWrite();
		
		targetPS->assign(ps);
		
		bit->unlock();		
		delete it;
				
		targetNode->addChildNode(new SGObjectNode(resultPC));
		progressWindow->progress((card32)(100.0*(double)i/(double)n));
	}
	progressWindow->popSteps();
}
