//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCSetErrorQuadrics.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "SGListNode.h"
#include "PointCloudTools.h"
#include "PointSetKNNQuery.h"
#include "PointSet.h"
#include "VertexDescriptor.h"
#include "Random.h"
#include "SGObjectNode.h"
#include "PointSet_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS( PCCSetErrorQuadrics , 0 ) {
	BEGIN_CLASS_INIT( PCCSetErrorQuadrics );
	ADD_STRING_PROP(m_Path,0);
	ADD_MATRIX3F_PROP(m_Quadric,0);
	ADD_BOOLEAN_PROP(m_Rescale,0);
	ADD_BOOLEAN_PROP(m_UseOldQuadric,0);
	ADD_FLOAT32_PROP(m_GlobalScale,0);
	ADD_CARD32_PROP(m_NumRandomSamplesPerFrame,0);

}

// ----

PCCSetErrorQuadrics::PCCSetErrorQuadrics() {
	m_Path;
	m_Quadric = IDENTITY3F;
	m_Rescale = true;		
	m_GlobalScale = 1.0f;
	m_NumRandomSamplesPerFrame = 5000;
	m_UseOldQuadric = false;
}

void PCCSetErrorQuadrics::assign(const Object* obj, CopyContext *context) {
	copyObjectProperties(obj, this);
}

PCCSetErrorQuadrics::~PCCSetErrorQuadrics() {
}




void PCCSetErrorQuadrics::execute(Scene *scene) 
{
	SGListNode *list = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene,m_Path));
	if( !list ){
		debugOutput << "!PCCSetErrorQuadrics::bad path!\n";
		throw "bad path!";
	}

	float globalScale = m_GlobalScale;
	
	if( m_Rescale )
		globalScale *= computeMedianPointDistance(list);

	Vector6f globalQuad = makeVector6f(m_Quadric[0][0],m_Quadric[1][1],m_Quadric[2][2],
		m_Quadric[0][1],m_Quadric[1][2],m_Quadric[0][2]) * globalScale;

	SceneGraphState *state = new SceneGraphState;
	state->staticState->defaultTraversalMode = SGStaticState::tmAllNodes;
	vector<float> distanceArray;
	for( unsigned i=0;i<list->getNumChildNodes(state);i++ )
	{
		// -------------------------
		// get point cloud
		SGObjectNode * objnode = dynamic_cast<SGObjectNode*>(list->getChildNode(state,i));
		if( !objnode )
			continue; // exception or more robust?

		SceneObject *so= objnode->getSceneObject();
		if( !so )
			continue;

      PointCloud *pc = dynamic_cast<PointCloud*>(so);
      if (!pc)
         continue;
        
		checkAttribute(pc, "errorQuad", 6, VAD::DATA_FORMAT_FLOAT32);

		SceneObjectIterator * iter = so->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
		if( !iter ) {
		   delete iter;
			continue;
		}

		ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		PointSet * ps = mIt->lockBlockReadWrite();

		// -------------------------
		// compute distances between a few random points
		AAT errQuadAAT = ps->getDescr()->getToken("errorQuad");
		const unsigned numPoints = ps->getNumEntries();
		for(unsigned j=0;j<numPoints;j++ )
		{
			Vector6f quad = ps->get<float,6>(j,errQuadAAT);
			if( !m_UseOldQuadric )
				quad = globalQuad;
			else
				quad *= globalScale;
			ps->set(j,errQuadAAT,quad);
		}

		bIt->unlock();
		delete iter;
	}
	delete state;

}

float geoxl::PCCSetErrorQuadrics::computeMedianPointDistance( SGListNode* list )
{
	debugOutput << "  PCCSetErrorQuadrics: started computeMedianPointDistance\n";
	SceneGraphState *state = new SceneGraphState;
	state->staticState->defaultTraversalMode = SGStaticState::tmAllNodes;
	vector<float> distanceArray;

	debugOutput << "  looking in " << list->getNumChildNodes(state) <<" point clouds\n"; debugOutput.flush();
	for( unsigned i=0;i<list->getNumChildNodes(state);i++ )
	{
		// -------------------------
		// get point cloud
		SGObjectNode * objnode = dynamic_cast<SGObjectNode*>(list->getChildNode(state,i));
		if( !objnode )
			continue; // exception or more robust?

		SceneObject *so= objnode->getSceneObject();
		if( !so )
			continue;

		SceneObjectIterator * iter = so->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
		if( !iter ) {
		   delete iter;
			continue;
	   }


		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		const PointSet * ps = bIt->lockBlockRead();

		// -------------------------
		// compute distances between a few random points
		AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
		PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, ps->getAAT("position") );
		AAT posAAT = ps->getDescr()->getToken("position");
		card32 numNeighbours = 8;
		for(unsigned p=0;p<m_NumRandomSamplesPerFrame;p++ )
		{
			card32 index = random32bit(ps->getNumEntries());
			Vector3f pos = ps->get3f(index,posAAT);
			knn.setTargetPoint( pos );
			for( unsigned j=0;j<numNeighbours;j++ )
			{
				bool found;
				card32 index2;
				knn.getNextNearestPointIndex(index2,found);	// ignore point itself
				knn.getNextNearestPointIndex(index2,found);
				if( found )
				{
					distanceArray.push_back( norm(pos-ps->get3f(index2,posAAT)) );
				}
			}
		}

		bIt->unlock();
		delete iter;
		delete octree;

		if (i%100==0)
		{
			debugOutput << "  pc " << i <<" currently " <<distanceArray.size()<< " samples\n"; debugOutput.flush();
		}
	}
	delete state;

	if( distanceArray.size() == 0)
		throw "no pointcloud found!";

	debugOutput << "  sorting " << distanceArray.size()<< " samples\n"; debugOutput.flush();
	sort( distanceArray.begin(),distanceArray.end());
//	debugOutput << distanceArray[0] << " | " << distanceArray.back() << " med: " << 
//		distanceArray[ distanceArray.size()/2] << "\n";

	debugOutput << "  PCCSetErrorQuadrics: finished computeMedianPointDistance\n";debugOutput.flush();
	return distanceArray[ distanceArray.size()/2];
}