//---------------------------------------------------------------------------
#ifndef CmdObjScriptProgramH
#define CmdObjScriptProgramH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <vector>
#include <map>
#include "CommandObject.h"
#include "ScriptEngine.h"
#include "PropertyParser.h"
#include "Persistent.h"
#include "ObjectDictionary.h"
#include "ScriptDictionaryAdapter.h"
#include "Scene.h"
//---------------------------------------------------------------------------

namespace geoxl {

class CmdObjScriptProgram;
class CmdObjScriptEngine;

/// Symbol table for access to the point cloud scene
class EDITOR_API PointCloudObjSymTab : public ObjectSymbolTable {
  private:
    CmdObjScriptEngine* cmdScript;

  public:
    PointCloudObjSymTab() { cmdScript = NULL; }
    void init(CmdObjScriptEngine* cmdScript) { this->cmdScript = cmdScript; }
    virtual bool getValue(const std::string name, Object* &value);
    virtual bool setValue(const std::string name, Object* value) { return false; };
    virtual bool removeSymbol(const std::string name) { return false; }
};

/// EXECUTE command: Allows to execute command objects
class EDITOR_API ExecuteCommand : public ScriptCommand {
  protected:
    CmdObjScriptEngine* cmdScript;

  public:
    virtual std::string supportedCommands() { return "execute"; };
    void setCmdScriptEngine(CmdObjScriptEngine* cmdScript) { this->cmdScript = cmdScript; }

    virtual void execute(std::string parameters, bool syntaxCheck);
};

/// PCLOUD NUMBER/SELECT/DELETE/GETNAME/SETNAME command: Allows access to point clouds
class EDITOR_API PCloudCommand : public ScriptCommand {
  protected:
    CmdObjScriptEngine* cmdScript;

  public:
    virtual std::string supportedCommands() { return "pcloud"; };
    void setCmdScriptEngine(CmdObjScriptEngine* cmdScript) { this->cmdScript = cmdScript; }

    virtual void execute(std::string parameters, bool syntaxCheck);
};

/// Special version of GEOXLscript that is used in the point cloud editor
class EDITOR_API CmdObjScriptEngine : public ScriptEngine {
  protected:
    Scene* scene;

    ExecuteCommand executeCommand;
    PCloudCommand pCloudCommand;
    CmdObjScriptProgram* cmdObjScriptProg;

    /// Access to the command objects in the dictionary
    ObjDicSymbolTable dictionarySymbols;
    /// Access to the scene
    PointCloudObjSymTab pointCloudSymbols;

    virtual void initVars();

  public:
    CmdObjScriptEngine();
    /// set the scene that this script engine works on
    void setScene(Scene* scene) { this->scene = scene; }
    /// For use by commands: returns the connected scene
    Scene* getScene() { return scene; }
    /// Sets the command object that this engine is in
    void setCmdObjScript(CmdObjScriptProgram* cmdObjScriptProg);
    /// For use by commands: Returns the command object that this engine is in
    CmdObjScriptProgram* getCmdObjScriptProg() const { return cmdObjScriptProg; }
};

/// Command object that contains a command object version of GEOXLscript
class EDITOR_API CmdObjScriptProgram : public CommandObject {
  private:
    static MetaClass *CmdObjScriptProgramMetaClass;
    static Object* newInstance();

  protected:
    CmdObjScriptEngine scriptEngine;

    ObjectDictionary cmdObjects;

  public:
    CmdObjScriptProgram();
    void initVars();

    virtual const MetaClass *getInstanceClass() const;
    static const MetaClass *getClass();
    static void init(const MetaClass *parentClass);
    static void shutdown();
    void write(OutputObjectStream *s) const;
    void read(InputObjectStream *s);

    virtual void doAction(Object *context);
    virtual void undoAction(Object *context);
    virtual bool canUndo();
    virtual std::string getDescription();
    virtual void assign(const Object* obj, CopyContext *context = NULL);

    /// Copies the command object given into the script program
    void addCommand(const CommandObject* command);
    /// Returns the command object with the name given
    bool getCmdObj(const std::string name, CommandObject*& cmdObj) const;
    /// Returns the command object dictionary
    ObjectDictionary* getDictionary() { return &cmdObjects; }
    /// Returns the script engine
    CmdObjScriptEngine* getScript() { return &scriptEngine; }

    ~CmdObjScriptProgram();
};

}

#endif