#ifndef FIFO_STACK_H
#define FIFO_STACK_H

// System
#include <assert.h>

template <class ItemType>
struct FifoStackElement
{
	/// Item
	ItemType item;
	/// Stack management: lower and upper item in the stack
	FifoStackElement<ItemType> *upper, *lower;
};

/**
 * Provides a stack of arbitrary items
 */
template <class ItemType>
class FifoStack
{
	private:
		/// Pointer towards start and end of list
		FifoStackElement<ItemType> *top, *bottom;
		/// Free all allocated memory
		void freeMemory();
		/// Number of items on the stack
		int numItems;

	public:
		/// Constructor
		FifoStack();
		/// Destructor
		~FifoStack();
		/// Empty stack
		void clear();
		/// Put new item in top of stack
		void push( ItemType item );
		/// Pop lowest item from stack
		ItemType pop();
		/// Test if stack is empty
		inline bool empty() const { return bottom == NULL; }
		/// Debugging purpose
		void printDebugInfo() const;
		/// Return number of items on stack
		inline int getNumItems() const { return numItems; }
		/// Return first item in stack without pop'ing
		ItemType next();

};

template<class ItemType>
FifoStack<ItemType>::FifoStack()
{
	// Init empty stack
	top = bottom = NULL;
   numItems = 0;
}

template<class ItemType>
FifoStack<ItemType>::~FifoStack()
{
	freeMemory();
}

template<class ItemType>
void FifoStack<ItemType>::freeMemory()
{
	// Free mem for each item in stacl
	FifoStackElement<ItemType> *e = top, *next;
	while ( e != NULL ){
		next = e->lower;
		delete e;
		e = next;
	}
	top = bottom = NULL;
   numItems = 0;
}

template<class ItemType>
void FifoStack<ItemType>::clear()
{
   freeMemory();
}

template<class ItemType>
void FifoStack<ItemType>::push( ItemType item )
{
	// Create new item
	FifoStackElement<ItemType>* e = new FifoStackElement<ItemType>;
	e->item = item;
	// Update data structure
	if ( top == NULL ){
		// First and only item in the stack
		bottom = top = e;
		e->upper = NULL;
		e->lower = NULL;
	} else {
		// At least one element already in stack
		top->upper = e;
		e->lower = top;
		e->upper = NULL;
		top = e;
	}
   numItems++;
}

template<class ItemType>
ItemType FifoStack<ItemType>::pop()
{
	// Never pop an empty stack!
	assert( bottom != NULL );
	// Temp save bottom element
	FifoStackElement<ItemType>* e = bottom;
	ItemType item = e->item;
	// Move bottom pointer up
	bottom = bottom->upper;
	if ( bottom == NULL ){
		// No more elements in the stack
		top = NULL;
	} else {
		bottom->lower = NULL;
	}
	numItems--;
	delete e;
	return item;
}

template<class ItemType>
ItemType FifoStack<ItemType>::next()
{
	return bottom->item;
}

template<class ItemType>
void FifoStack<ItemType>::printDebugInfo() const
{	
	cout << "* FifoStack state * " << endl;
	cout << "  Top element:    " << top << endl;
	cout << "  Bottom element: " << bottom << endl;
}

#endif