//---------------------------------------------------------------------------
#ifndef PCInteractionToolH
#define PCInteractionToolH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "KeyboardController.h"
#include "MouseController.h"
#include "Tools.h"
#include "OnSceneChangeCallback.h"
#include <qtgui/qwidget.h>
#include "LinearAlgebra.h"
#include "PointCloud.h"
//---------------------------------------------------------------------------


namespace geoxl {

class GLContext;
class Scene;

class EDITOR_API PCInteractionTool : public Persistent, public KeyboardController, public MouseController 
{
	GEOXL_ABSTRACT_CLASS( PCInteractionTool )	
	private:

	public:
		

	protected:
		Scene *scene;
		OnSceneChangeCallback *onChange;
		QWidget *sceneEditorWidget;
		card32 screenSizeX,screenSizeY;

	   /// call to notify editor concerning changes
	   virtual void sceneChanged(PCInteractionToolChanges whatChange) {if (onChange != NULL) {onChange->changed(whatChange);}}
	   /// call to cause immediate redraw
	   virtual void redrawScene() {if (onChange != NULL) {onChange->changed(PCI_CHANGE_VIEW);}}
	   /// call to write a frame as bitmap to disc (if activated in main editor, otherwise nothing happens)
	   virtual void writeFrameToDisk() {if (onChange != NULL) {onChange->changed(PCI_WRITE_FRAME);}}

		/// returns true and the corresponding point, point cloud if cursor points on point cloud
		bool getIntersectionFromCursor( int x, int y, Vector3f &retPoint, string &pcName, PointCloud ** pc );

		/// returns a vector of point cloud from the scene
		vector<PointCloud*> getAllPointCloudsFromScene();

		/// returns all selected point clouds from the scene
		vector<PointCloud*> getSelectedPointCloudsFromScene();

      /// pass the mouse event on to the camera controller
      virtual void mouseDownPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
      /// pass the mouse event on to the camera controller
      virtual void mouseMovedPassToCam(int32 x, int32 y);
      /// pass the mouse event on to the camera controller
      virtual void mouseUpPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
      /// pass the mouse event on to the camera controller
      virtual void areaResizePassToCam(card32 width, card32 height);
      /// pass the mouse event on to the camera controller
      virtual void mouseWheelRotatedPassToCam(int32 rotatedDelta, Modifiers modifiersState);

      /// pass the keyboard event on to the camera controller
      virtual void keyDownPassToCam(GeneralKey key);
      /// pass the keyboard event on to the camera controller
      virtual void keyUpPassToCam(GeneralKey key);
   
	public:
		PCInteractionTool();

 

	   // --- tool definition methods

	   /// called once to connect tool with scene and editor
	   /// scene: point cloud scene
	   /// onChange: callback to notify editor of changes
	   /// doRedraw: callback to cause redraw of current editor view of scene
	   /// sceneEditorWidget: top level widget that contains the editor (used to attach child windows to the right parents, do not cast to SceneEditorWidget to access it directly)
	   virtual void connectScene( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget );
	   /// tool has been choosen (again) by user, re-initialize tool
	   virtual void activateTool() {}
	   /// tool has been deactivated by user (other tool has been choosen)
	   virtual void deactivateTool() {}
	   /// override this method to provide user feedback, e.g. handles / bounding boxes etc...
	   virtual void glDrawTool(GLContext *glContext) {forget(glContext);}
	   /// determines whether the main application should draw a 3d view of the scene (before glDrawTool)
	   /// default: true - if false, all visualization must be performed in glDrawTool
	   virtual bool drawGLSceneRendering() {return true;}
      /// override this method and return true if the tool should operate in camera mode by default
      virtual bool forceCamByDefault();

	   // ---

	   inline Scene *getScene() {return scene;}

		/// resize pci-intern screensize
		void areaResize_base(card32 w, card32 h) {screenSizeX=w;screenSizeY=h;}

	   ~PCInteractionTool();

	   virtual void mouseWheelRotated(int32 rotatedDelta, Modifiers modifiersState ){};
};

}

#endif
