//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "PointCloudExporter.h"
#include "PointCloudIterators.h"
#include "LinearAlgebra_templCode.h"
//---------------------------------------------------------------------------
#include <fstream>
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


PointCloudExporter::PointCloudExporter(PointCloud* pc, ofstream* outf)
{
  this->pc = pc;
  this->outf = outf;
}

void PointCloudExporter::exportPointCloud()
{
  *outf << "#PC: ";// << pc->getName() << endl;
  const VertexDescriptor* vertexDesc = pc->getDescr();

  *outf << "#Attributes:" << endl;
  vector<AttributeWriter> attribs;

  for (card32 attribNum = 0; attribNum < vertexDesc->getNumAttribs(); ++attribNum) {
    VAD attribDesc = vertexDesc->getAttribute(attribNum);
    *outf << attribDesc.name << " ";
    *outf << attribDesc.numComponents << " ";

    bool doNotExport = false;
    switch (attribDesc.dataFormat) {
      case VAD::DATA_FORMAT_CARD8:
        *outf << "card8" << endl;
        break;
      case VAD::DATA_FORMAT_FLOAT32:
        *outf << "float32" << endl;
        break;
      case VAD::DATA_FORMAT_INT32:
        *outf << "int32" << endl;
        break;
      default:
        string message = "Could not export attribute channel " + attribDesc.name;
        warning(message);
        doNotExport = true;
    }

    if (!doNotExport) {
      AAT attribToken = vertexDesc->getToken(attribNum);
      AttributeWriter attribW(attribDesc, attribToken);
      attribs.push_back(attribW);
    }
  }

  SceneObjectIterator* readIt = pc->createIterator(SOIT(SOIT::CAP_BASIC_PC));
  BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(readIt);

  basicIt->reset();
  *outf << "#Data:" << endl;
  while (!basicIt->atEnd()) {
    for (card32 i = 0; i != attribs.size(); ++i) {
      attribs[i].writeAttrib(*outf, basicIt);
    }
    *outf << endl;

    basicIt->next();
  }

  delete readIt;
}

AttributeWriter::AttributeWriter(VAD& attribDesc, AAT& attribToken)
{
  this->attribDesc = attribDesc;
  this->attribToken = attribToken;
}

void AttributeWriter::writeAttrib(std::ofstream& outf, BasicPointCloudIterator* basicIt)
{
  switch (attribDesc.numComponents) {
    case 1:
      switch (attribDesc.dataFormat) {
        case VAD::DATA_FORMAT_CARD8: {
          card8 vCard8 = basicIt->get1ub(attribToken);
          outf << " " << vCard8;
        } break;
        case VAD::DATA_FORMAT_FLOAT32: {
          float32 vFloat32 = basicIt->get1f(attribToken);
          outf << " " << vFloat32;
        } break;
        case VAD::DATA_FORMAT_INT32: {
          int32 vInt32 = basicIt->get1i(attribToken);
          outf << " " << vInt32;
        } break;
        default:
          pAssert(false);
      }
    break;

    case 2:
      switch (attribDesc.dataFormat) {
        case VAD::DATA_FORMAT_CARD8: {
          Vector2ub vec2ub = basicIt->get2ub(attribToken);
          outf << " " << vec2ub[0] << " " << vec2ub[1];
        } break;
        case VAD::DATA_FORMAT_FLOAT32: {
          Vector2f vec2f = basicIt->get2f(attribToken);
          outf << " " << vec2f[0] << " " << vec2f[1];
        } break;
        case VAD::DATA_FORMAT_INT32: {
          Vector2i vec2i = basicIt->get2i(attribToken);
          outf << " " << vec2i[0] << " " << vec2i[1];
        } break;
        default:
          pAssert(false);
      }
    break;

    case 3:
      switch (attribDesc.dataFormat) {
        case VAD::DATA_FORMAT_CARD8: {
          Vector3ub vec3ub = basicIt->get3ub(attribToken);
          outf << " " << vec3ub[0] << " " << vec3ub[1] << " " << vec3ub[2];
        } break;
        case VAD::DATA_FORMAT_FLOAT32: {
          Vector3f vec3f = basicIt->get3f(attribToken);
          outf << " " << vec3f[0] << " " << vec3f[1] << " " << vec3f[2];
        } break;
        case VAD::DATA_FORMAT_INT32: {
          Vector3i vec3i = basicIt->get3i(attribToken);
          outf << " " << vec3i[0] << " " << vec3i[1] << " " << vec3i[2];
        } break;
        default:
          pAssert(false);
      }
    break;

    case 4:
      switch (attribDesc.dataFormat) {
        case VAD::DATA_FORMAT_CARD8: {
          Vector4ub vec4ub = basicIt->get4ub(attribToken);
          outf << " " << vec4ub[0] << " " << vec4ub[1] << " " << vec4ub[2] << " " << vec4ub[3];
        } break;
        case VAD::DATA_FORMAT_FLOAT32: {
          Vector4f vec4f = basicIt->get4f(attribToken);
          outf << " " << vec4f[0] << " " << vec4f[1] << " " << vec4f[2] << " " << vec4f[3];
        } break;
        case VAD::DATA_FORMAT_INT32: {
          Vector4i vec4i = basicIt->get4i(attribToken);
          outf << " " << vec4i[0] << " " << vec4i[1] << " " << vec4i[2] << " " << vec4i[3];
        } break;
        default:
          pAssert(false);
      }
    break;
  }
}

