//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "PCCAttributeModifier.h"
#include "FixedArrayClassProperty.h"
#include "StringClassProperty.h"
#include "BoundingBox.h"
#include "BoundingBox_templCode.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra_templCode.h"
#include "PointCloudIterators.h"
#include "PointCloudTools.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "GlobalMacros.h" 
#include "ClassMethods.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

/// Describe modification type enum
static string describeModificationType( card32 q )
{
	if ( q == OPERATION_MULTIPLICATION ) return "Multiplication";
	else if ( q == OPERATION_ADDITION ) return "Add vector";
	else if ( q == OPERATION_SET ) return "Set";
	else if ( q == OPERATION_COPY ) return "Copy";
	else return "Undefined";
}

IMPLEMENT_GEOXL_CLASS( PCCAttributeModifier, 2 )
{
	BEGIN_CLASS_INIT( PCCAttributeModifier )
	ADD_STRING_PROP( attributeID, 0 )
	ADD_CARD32_PROP( copyToCoordinate, 0 )
	ADD_CARD32_PROP_UPDATE_DESCR( modificationType, 0 , NULL, describeModificationType, 4 );
	ADD_MATRIX4F_PROP( muliplicationMatrix, 0 )
	ADD_BOOLEAN_PROP( homogenious, 0 )
	ADD_VECTOR4F_PROP( additionVector, 0 )
	ADD_STRING_PROP( copyFromAttribute, 0 )
	ADD_CARD32_PROP( copyFromCoordinate, 0 )
	ADD_FLOAT32_PROP( alpha, 2 )
	ADD_NOARGS_METHOD( rotateX )
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( rotateY )
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( rotateZ )
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD_DESC( identifyTransformation, "Applies transformation of a point cloud to all points and reset transformation to identity");
	SET_METHOD_AS_CMD_OBJ( true )
}

/// Constructor
PCCAttributeModifier::PCCAttributeModifier()
{
   muliplicationMatrix = makeMatrix4f( 1.1f, 0.0f, 0.0f, 0.0f,
                                       0.0f, 1.1f, 0.0f, 0.0f,
                                       0.0f, 0.0f, 1.1f, 0.0f,
                                       0.0f, 0.0f, 0.0f, 1.0f );
   additionVector = NULL_VECTOR4F;
   attributeID = "position";
   homogenious = false;
   modificationType = OPERATION_MULTIPLICATION;
   copyFromAttribute = "color";
   copyFromCoordinate = 2;
   copyToCoordinate = 2;
   alpha = 20.0f;
}

PCCAttributeModifier::~PCCAttributeModifier()
{

}

void PCCAttributeModifier::rotate( const Matrix3f& R )
{ 
	vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;

	for ( unsigned int m = 0; m < selection.size(); m++ ){
		pc = getPointCloud( scene, selection[m] );
		if ( pc != NULL ){
			AAT POSITION = pc->getAAT( "position" );
			SceneObjectIterator *newIt = pc->createIterator( SOIT ( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC )  );
			if (newIt == NULL) continue;
			BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
			ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);
			basicIt->reset();
			while( !basicIt->atEnd() ){
				const Vector3f& p = basicIt->get3f( POSITION );
				modIt->set3f( POSITION, R * p );
				basicIt->next();
			}
			delete newIt;
		}
	}
}

void PCCAttributeModifier::rotateX()
{
	float32 a = alpha * M_PI / 180.0f;
	rotate( 
		makeMatrix3f( 
			1, 0,         0, 
		    0, cos( a ),  sin( a ),
			0, -sin( a ), cos( a ) ) );
}

void PCCAttributeModifier::rotateY()
{
	float32 a = alpha * M_PI / 180.0f;
	rotate( 
		makeMatrix3f(
			cos( a ), 0, -sin( a ),
			0,		  1, 0,
			sin( a ), 0, cos( a ) ) );
}

void PCCAttributeModifier::rotateZ()
{
	float32 a = alpha * M_PI / 180.0f;
	rotate( 
		makeMatrix3f(
		cos( a ),  sin( a ), 0,
		-sin( a ), cos( a ), 0,
		0,		   0,        1 ) );
}

/// Execute command
void PCCAttributeModifier::execute(Scene *scene)
{
   unsigned int m; 
   vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
   PointCloud* pc;

   for ( m = 0; m < selection.size(); m++ ){
      pc = getPointCloud( scene, selection[m] );

      /// Create iterators
      SceneObjectIterator *newIt = pc->createIterator( SOIT ( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC )  );
      if (newIt == NULL) continue;
      BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
      ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);

      /// Check if attribute is available
      if (!pc->providesAttribute(attributeID)) {
         //Application->MessageBox("Attribute is not available!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
		 WARNINGMESSAGE("Attribute is not available!");

         return;
      }

      VertexAttribDescriptor attribDesc = pc->getDescr()->getAttribute( attributeID );
      VertexAttribDescriptor copyFromAttribDesc;
      if ( modificationType == OPERATION_COPY )
         copyFromAttribDesc = pc->getDescr()->getAttribute( copyFromAttribute );

      /// Copy: check if copyFromAttribute is available and if coordinates are valid
      float32 copyFromValue = 0;
      if ( modificationType == OPERATION_COPY ){
         if ( !pc->providesAttribute( copyFromAttribute ) ){
            //Application->MessageBox("CopyFrom Attribute is invalid!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
			 WARNINGMESSAGE("CopyFrom Attribute is invalid!");
            return;
         } else {
            if ( ( copyFromAttribDesc.numComponents <= copyFromCoordinate ) ||
                 ( attribDesc.numComponents <= copyToCoordinate ) ){
               //Application->MessageBox("Copy coordinates are invalid!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				WARNINGMESSAGE("Copy coordinates are invalid!");
               return;
            }
         }
      }
      
      /// Go through all vertices
      if ( basicIt && modIt ){
         while(!basicIt->atEnd()) {

            /// Check if attribute exists in point cloud
            VertexAttribDescriptor attribDesc = pc->getDescr()->getAttribute( attributeID );

            /// Helper vector for computations
            Vector4f vec = homogenious ? makeVector4f(0,0,0,1) : makeVector4f(0,0,0,0);

            /// Extract copyFromValue if required
            if ( modificationType == OPERATION_COPY ){
               if ( copyFromAttribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_CARD8 ) {
                  /// CARD8
                  if ( copyFromAttribDesc.numComponents == 1 ){
                     copyFromValue = basicIt->get1ub( pc->getAAT( copyFromAttribute ) );
                  } else if ( copyFromAttribDesc.numComponents == 2 ){
                      Vector2ub v = basicIt->get2ub( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 3 ){
                      Vector3ub v = basicIt->get3ub( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 4 ){
                      Vector4ub v = basicIt->get4ub( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  }
               } else if ( copyFromAttribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_INT32 ) {
                  /// INT32
                  if ( copyFromAttribDesc.numComponents == 1 ){
                     copyFromValue = basicIt->get1i( pc->getAAT( copyFromAttribute ) );
                  } else if ( copyFromAttribDesc.numComponents == 2 ){
                      Vector2i v = basicIt->get2i( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 3 ){
                      Vector3i v = basicIt->get3i( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 4 ){
                      Vector4i v = basicIt->get4i( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  }
               } else if ( copyFromAttribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_FLOAT32 ) {
                  /// FLOAT32
                  if ( copyFromAttribDesc.numComponents == 1 ){
                     copyFromValue = basicIt->get1f( pc->getAAT( copyFromAttribute ) );
                  } else if ( copyFromAttribDesc.numComponents == 2 ){
                      Vector2f v = basicIt->get2f( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 3 ){
                      Vector3f v = basicIt->get3f( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  } else if ( copyFromAttribDesc.numComponents == 4 ){
                      Vector4f v = basicIt->get4f( pc->getAAT( copyFromAttribute ) );
                      copyFromValue = v[copyFromCoordinate];
                  }
               } else {
                  //Application->MessageBox("Copy failed!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				   WARNINGMESSAGE("Copy failed!");
                  return;
               }
            }

            /// Apply modification
            if ( attribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_CARD8 ) {
               /// CARD8
               if ( attribDesc.numComponents == 1 ){
                  card8 cardVec = basicIt->get1ub( pc->getAAT(attributeID) );
                  vec[0] = cardVec;
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  cardVec = vec[0];
                  modIt->set1ub( pc->getAAT(attributeID), cardVec );
               } else if ( attribDesc.numComponents == 2 ){
                  Vector2ub cardVec = basicIt->get2ub( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 2; i++ )
                     vec[i] = cardVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					 WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  for ( int i = 0; i < 2; i++ )
                     cardVec[i] = vec[i];
                  modIt->set2ub( pc->getAAT(attributeID), cardVec );
               } else if ( attribDesc.numComponents == 3 ){
                  Vector3ub cardVec = basicIt->get3ub( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 3; i++ )
                     vec[i] = cardVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  for ( int i = 0; i < 3; i++ )
                     cardVec[i] = vec[i];
                  modIt->set3ub( pc->getAAT(attributeID), cardVec );
               } else if ( attribDesc.numComponents == 4 ){
                  Vector4ub cardVec = basicIt->get4ub( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 4; i++ )
                     vec[i] = cardVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  for ( int i = 0; i < 4; i++ )
                     cardVec[i] = vec[i];
                  modIt->set4ub( pc->getAAT(attributeID), cardVec );
               } else {
                  //Application->MessageBox("Attribute has invalid dimension!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				  WARNINGMESSAGE("Attribute has invalid dimension!");
                  return;
               }
            } else if ( attribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_INT32 ){
               /// INT32
               if ( attribDesc.numComponents == 1 ){
                  int32 intVec = basicIt->get1i( pc->getAAT(attributeID) );
                  vec[0] = intVec;
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  intVec = vec[0];
                  modIt->set1i( pc->getAAT(attributeID), intVec );
               } else if ( attribDesc.numComponents == 2 ){
                  Vector2i intVec = basicIt->get2i( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 2; i++ )
                     vec[i] = intVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
                     //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
                     return;
                  }
                  for ( int i = 0; i < 2; i++ )
                     intVec[i] = vec[i];
                  modIt->set2i( pc->getAAT(attributeID), intVec );
               } else if ( attribDesc.numComponents == 3 ){
                  Vector3i intVec = basicIt->get3i( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 3; i++ )
                     vec[i] = intVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  for ( int i = 0; i < 3; i++ )
                     intVec[i] = vec[i];
                  modIt->set3i( pc->getAAT(attributeID), intVec );
               } else if ( attribDesc.numComponents == 4 ){
                  Vector4i intVec = basicIt->get4i( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 4; i++ )
                     vec[i] = intVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  for ( int i = 0; i < 4; i++ )
                     intVec[i] = vec[i];
                  modIt->set4i( pc->getAAT(attributeID), intVec );
               } else {
                  //Application->MessageBox("Attribute has invalid dimension!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				   WARNINGMESSAGE("Attribute has invalid dimension!");
                  return;
               }
            } else if ( attribDesc.dataFormat == VertexAttribDescriptor::DATA_FORMAT_FLOAT32 ){
               /// FLOAT32
               if ( attribDesc.numComponents == 1 ){
                  float32 floatVec = basicIt->get1f( pc->getAAT(attributeID) );
                  vec[0] = floatVec;
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  floatVec = vec[0];
                  modIt->set1f( pc->getAAT(attributeID), floatVec );
               } else if ( attribDesc.numComponents == 2 ){
                  Vector2f floatVec = basicIt->get2f( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 2; i++ )
                     vec[i] = floatVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  for ( int i = 0; i < 2; i++ )
                     floatVec[i] = vec[i];
                  modIt->set2f( pc->getAAT(attributeID), floatVec );
               } else if ( attribDesc.numComponents == 3 ){
                  Vector3f floatVec = basicIt->get3f( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 3; i++ )
                     vec[i] = floatVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                     if ( homogenious ) vec /= vec[3];
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  for ( int i = 0; i < 3; i++ )
                     floatVec[i] = vec[i];
                  modIt->set3f( pc->getAAT(attributeID), floatVec );
               } else if ( attribDesc.numComponents == 4 ){
                  Vector4f floatVec = basicIt->get4f( pc->getAAT(attributeID) );
                  for ( int i = 0; i < 4; i++ )
                     vec[i] = floatVec[i];
                  if ( modificationType == OPERATION_MULTIPLICATION ){
                     /// Multiplication
                     vec = muliplicationMatrix * vec;
                  } else if ( modificationType == OPERATION_ADDITION ){
                     /// Addition
                     vec += additionVector;
                  } else if ( modificationType == OPERATION_SET ){
                     /// Set
                     vec = additionVector;
                  } else if ( modificationType == OPERATION_COPY ){
                     /// Copy
                     vec[copyToCoordinate] = copyFromValue;
                  } else {
                     /// Undefined
					  //Application->MessageBox("Invalid operation!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
					  WARNINGMESSAGE("Invalid operation!");
					  return;
                  }
                  for ( int i = 0; i < 4; i++ )
                     floatVec[i] = vec[i];
                  modIt->set4f( pc->getAAT(attributeID), floatVec );
               } else {
                  //Application->MessageBox("Attribute has invalid dimension!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				  WARNINGMESSAGE("Attribute has invalid dimension!");
                  return;
               }
            } else {
               /// Invalid data type of attribute
               //Application->MessageBox("Attribute has invalid type!", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
				WARNINGMESSAGE("Attribute has invalid type!");
               return;
            }

            /// Next vertex
            basicIt->next();
         }
      }

      /// Free mem
      delete newIt;
   }
}

void geoxl::PCCAttributeModifier::identifyTransformation()
{
	vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;

	for ( unsigned int m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( scene, selection[m] );
		if ( pc != NULL )
		{
			SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
			Matrix4f T = pc->getTransformation();
			Matrix3f T_N = shrink4To3(T);
			AAT posAAT = pc->getAAT("position");
			AAT normalAAT;
			bool hasNormal = pc->providesAttribute("normal");
			if( hasNormal )
				normalAAT = pc->getAAT("normal");

			while( !bIt->atEnd())
			{
				Vector3f p = bIt->get3f(posAAT);
				mIt->set3f(posAAT,transformVector3f(T,p));
				if( hasNormal )
				{
					Vector3f n = normalize( T_N * bIt->get3f(normalAAT));
					mIt->set3f(normalAAT,n);
				}
				bIt->next();
			}
			pc->setTransformation(IDENTITY4F);
		}
	}
}