//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCEstimateSampleSpacing.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "StandardSceneGraphIterator.h"
#include "LinearAlgebra_templCode.h"
#include "ProgressWindow.h"
#include "Timer.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



IMPLEMENT_GEOXL_CLASS(PCCEstimateSampleSpacing, 0)
{
	BEGIN_CLASS_INIT(PCCEstimateSampleSpacing);
	ADD_CARD32_PROP(numNeighbors, 0)
}

PCCEstimateSampleSpacing::PCCEstimateSampleSpacing()
{
   numNeighbors = 9;
}

PCCEstimateSampleSpacing::~PCCEstimateSampleSpacing() {}

void PCCEstimateSampleSpacing::assign(const Object* obj, CopyContext *context)
{
	copyObjectProperties(obj, this);
}



void PCCEstimateSampleSpacing::execute(Scene *scene) 
{
   if (numNeighbors <=0) {
      throw PException("PCCEstimateSampleSpacing - numN must be > 0");
   }
   /// Find selected point clouds
   vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
   /// Use all point clouds in selection
   for (unsigned m=0; m<selection.size(); m++) {
      /// Check for required attributes
      Timer timer;
      checkAttribute( scene, selection[m], "position", 3, VAD::DATA_FORMAT_FLOAT32 );
      checkAttribute( scene, selection[m], "sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32 );
      debugOutput << "add channel: " << timer.getDeltaValue() << " ms\n";

      PointCloud* pc = getPointCloud( scene, selection[m] );

	   if ( pc == NULL ){
		   warning("PCCEstimateSampleSpacing: Invalid (NULL) point cloud!\n");
		   continue;
	   }

	   // Check for required attributes
	   if ( !pc->providesAttribute( "position" ) ||
            !pc->providesAttribute( "sampleSpacing" ) ) {
		   warning("PCCEstimateSampleSpacing: Attribute missing (position, sampleSpacing required)!\n");
		   continue;
	   }

	   // Create iterators
	   SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC ) );
	   BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
	   ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);
	   HierarchicalKNNIterator *pcNNIt = new HierarchicalKNNIterator( pc, 32, NULL );
	   AAT sampleSpacingAAT = pc->getAAT( "sampleSpacing" );
	   AAT positionAAT = pc->getAAT( "position" );

	   /// Estimate normals
	   progressWindow->pushStep( true, "est. sample spacing" );
	   basicIt->reset();
	   card32 pointIndex = 0;
	   const float32 normFact = sqrt((float32)numNeighbors);
	   while ( !basicIt->atEnd() ){
		   Vector3f pos = basicIt->get3f( positionAAT );
         pcNNIt->setSeekPointAndReset( pos );
         card32 numN = 0;
         float32 distAv = 0.0f;
         while (numN <= numNeighbors && !pcNNIt->atEnd()) {
            Vector3f nPos = pcNNIt->get3f(positionAAT);
            pcNNIt->next();
            if (nPos == pos) {
               continue;
            } else {
               distAv += norm(pos-nPos);
               numN++;
            }
         }
         if (numN > 0) {
            distAv /= numN;
         }
		   modIt->set1f( sampleSpacingAAT, distAv/normFact );

		   basicIt->next();
		   pointIndex++;
		   if ((pointIndex % 10000) == 0) {
   		   if (!progressWindow->progress( card32( pointIndex * 100.0f / pc->getNumPoints() ) )) break;
   		}
	   }

	   bool dummy;
	   progressWindow->popSteps( dummy );

	   // Free allocated memory
	   delete newIt;
	   delete pcNNIt;
   }
}
