#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCImplicitMesher.h"
//---------------------------------------------------------------------------
#include "MarchingCubes.h"
#include "HierarchicalKNNIterator.h"
#include "Scene.h"
#include "PointCloud.h"
#include "NumericalClassProperty.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "PointCloudIterators.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "VertexArray.h"
#include "RegularData.h"
#include "RegularData_templCode.h"
#include "AttachedIndexedOctree.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_CLASS( PCCImplicitMesher ,2)
{
	BEGIN_CLASS_INIT( PCCImplicitMesher );
	ADD_CARD32_PROP( m_Resolution, 0);
	ADD_FLOAT32_PROP( m_Sigma, 0 );
	ADD_FLOAT32_PROP( m_GridSnapTolerance, 0);
	ADD_BOOLEAN_PROP (smoothNormals,1)
	ADD_FLOAT32_PROP(smoothNormals_Sigma,2)
}

PCCImplicitMesher::PCCImplicitMesher(void)
{
	m_Sigma = 1.0f;
	m_Resolution = 30;
	m_GridSnapTolerance = 0.1f;
	smoothNormals=true;
	smoothNormals_Sigma=1.0f;
}

PCCImplicitMesher::~PCCImplicitMesher(void)
{
}

void PCCImplicitMesher::assign(const Object* obj, CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCImplicitMesher * o = dynamic_cast<const PCCImplicitMesher*>(obj);
	pAssert( o != NULL );

	m_Sigma = o->m_Sigma;
	m_Resolution = o->m_Resolution;
}

struct TMesherContext
{
	AttachedIndexedOctree * octree;
	HierarchicalKNNIterator *	hIt;
	VertexArray * vertexArray;
	float		sigma;
	float		averagePointDistance;
	float		cubeSize;
	AAT		posAAT;
	AAT		normalAAT;
};

TMesherContext * sdContext = NULL;

float signedDistanceFunction( void *context, float x, float y, float z )
{
	Vector3f pos = makeVector3f(x,y,z);
	sdContext->hIt->setSeekPointAndReset(pos);
	float summedWeight = 0.0f;
	card32 numSamples = 0;
	float signedDistance = 0.0f;
	float scale = 1.0f / sdContext->averagePointDistance;
	while(!sdContext->hIt->atEnd())
	{
		Vector3f samplePos = sdContext->hIt->get3f( sdContext->posAAT );
		Vector3f normal = sdContext->hIt->get3f( sdContext->normalAAT );
		float distance = norm( samplePos - pos );

		if( distance > sdContext->averagePointDistance * 3.0f && numSamples == 0)
		{
			Vector3f dir = -normalize(samplePos - pos);
			Vector3f testPos = samplePos + dir*0.5f*sdContext->averagePointDistance;
			float ret = signedDistanceFunction( context, testPos[0],testPos[1],testPos[2] );
			if( ret < 0 )
				return -distance;
			else
				return distance;
		}

		distance *=  scale;
		sdContext->hIt->next();

		float weight = //0.5f / M_PI / sdContext->sigma * 
			exp( -0.5f * pow(distance/sdContext->sigma,2.0f) );

		signedDistance += (samplePos*normal - pos*normal) * weight;

		summedWeight += weight;
		

		if( numSamples++ > 100 || weight * 100<summedWeight || weight < 0.0001f)
			if( numSamples > 1 )
				break;
			else
				return (samplePos*normal - pos*normal);
	}

	if( summedWeight > 0.1e-25f )
		return signedDistance / summedWeight;
	else
		return 1.0f;
}

bool isBoundingBoxEmpty( const BoundingBox3f &bb, AIOctreeNode * node )
{
   if( node->isLeaf() )
   {
      for( card32 i=0;i<node->getContent().m_index.size();i++ )
      {
         card32 index = node->getContent().m_index[i];
			if( bb.inBoundingBox( sdContext->vertexArray->getPosition3f( index ) ) )
            return false;
      }
      return true;
   }
   else
   {
      if( node->getContent().m_bb.intersectBoundingBox( bb ) )
      {
          for( card32 i=0;i<8;i++ )
          {
            if( !isBoundingBoxEmpty( bb, node->getChild(i) ) )
               return false;
          }
          return true;
      }
   }
   return true;
}

bool trianglulate_isBoundingBoxEmpty( void *context, const BoundingBox3f &bb )
{
	BoundingBox3f checkBB = bb;
	checkBB.addBorder( max(sdContext->cubeSize * 0.55f,sdContext->averagePointDistance*1.5f) );
	return isBoundingBoxEmpty( checkBB, sdContext->octree->getOctree() );
}

void PCCImplicitMesher::execute(Scene *scene) 
{
   vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();

	if( selection.size() == 0 )
	{
		// collect all point clouds
		StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
		it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
		while ( !it->atEnd() )
		{
			SGObjectNode* sgon = dynamic_cast<SGObjectNode*>( it->getNode() );
			if( sgon )
			{
				PointCloud * pc = dynamic_cast<PointCloud*>( sgon->getSceneObject() );
				UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( sgon->getSceneObject() );
				if( pc && !mesh)
				{
					selection.push_back( it->getNameWithPath() );
				}
			}
			it->next();
		}
	}


	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		triangulatePointCloud( scene, selection[m] );
	}
	
}

void PCCImplicitMesher::triangulatePointCloud( Scene * scene, string pcName ) 
{
	PointCloud * pc =  getPointCloud( scene, pcName );
	UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	pAssert( pc->providesAttribute( "position" ) );
	pAssert( pc->providesAttribute( "normal" ) );
	
	// prepare context
	TMesherContext context;
	context.hIt = new HierarchicalKNNIterator(pc, 32, NULL );
	context.posAAT = pc->getAAT( "position" );
	context.normalAAT = pc->getAAT( "normal" );
	context.sigma = m_Sigma;
	context.octree = new AttachedIndexedOctree;
	context.octree->octreesetup( upc->getPointSet(), 32 );
	context.vertexArray = new VertexArray( upc->getPointSet() );

	sdContext = &context;

	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_HIERARCHICAL | SOIT::CAP_BASIC_PC ) );
   HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator*>(newIt);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(newIt);

	// calculate average point distance
	context.averagePointDistance = getAveragePointDistance( pc );

	// calc boundingbox and enlarge it a little bit ( 10% each dim + average point distance*2)
   BoundingBox3f bb = hIt->getBoundingBox();
   for( card32 i=0;i<3;i++ )
   {
      float delta = max(bb.upperCorner[i] - bb.lowerCorner[i],0.0001f);
      bb.lowerCorner[i] -= delta * 0.05f + context.averagePointDistance*2.0f;
      bb.upperCorner[i] += delta * 0.05f + context.averagePointDistance*2.0f;
   }
	bb.makeBoundingCube();
	context.cubeSize = bb.getSideLength(0) / (float)m_Resolution;

	/*// calc debug output
	RegularDataF result( makeVector3i(1,1,1) * m_Resolution+makeVector3i(1,1,1) ); 
	float scale = 1.0f / (float)m_Resolution * bb.getMaxSideLength();
	for( card32 x=0;x<=m_Resolution;x++ )
	{
		for( card32 y=0;y<=m_Resolution;y++ )
		{
			for( card32 z=0;z<=m_Resolution;z++ )
			{
				float fx = bb.lowerCorner[0] + (float)x*scale;
				float fy = bb.lowerCorner[1] + (float)y*scale;
				float fz = bb.lowerCorner[2] + (float)z*scale;
				float value = signedDistanceFunction( fx,fy,fz );
				result.setElement( makeVector3i(x,y,z), value );
			}
		}
	}
	GLVolumeDisplayer *voldisp = new GLVolumeDisplayer(&result, "x", "y", "z");
	GLEventDisplayer::SetNewGLEventObject ev(voldisp);

	debugOutput.displayEvent(GLEventDisplayer::getClass(), "tach", &ev);*/


	// calculate triangulation
	MarchingCubes mc;
	Vector3i dim = makeVector3i(1,1,1) * m_Resolution;
	mc.setGridSnappingTolerance( m_GridSnapTolerance );
	mc.calculateHierarchical( &signedDistanceFunction, NULL, &trianglulate_isBoundingBoxEmpty, bb, dim, 0.0f );

	// create output point cloud
	UnstructuredInCoreTriangleMesh * mesh = new UnstructuredInCoreTriangleMesh;
   mesh->setMaterialIndex( 3 );

	VertexDescriptor vdesc;
   vdesc.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
   vdesc.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );

	VertexDescriptor idesc;
   idesc.pushAttrib( VAD( "index", 3, VAD::DATA_FORMAT_INT32 ) );

	mesh->clearAndSetup( &vdesc, mc.getNumVertices(), &idesc, mc.getNumTriangles() );
	VertexArray vertices( mesh->getPointSet() );
	VertexArray indices( mesh->getTriangles() );
	for( card32 i=0;i<mc.getNumVertices();i++ )
	{
		vertices.setPosition3f( i, mc.getVertex(i) );
	}
	for( card32 i=0;i<mc.getNumTriangles();i++ )
	{
		indices.setIndex3i( i, mc.getTriangle(i) );
	}

	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
	mesh->closeMesh();
	mesh->calculateVertexNormals();


	// output mesh
	string outputName = pcName + "_mesh";
	if( getPointCloud(scene, outputName ) )
		replacePointCloud( scene, outputName, mesh );
	else
		addPointCloud( scene, mesh, outputName );

	delete newIt;
	delete context.hIt;
	delete context.octree;
	delete context.vertexArray;


	if(smoothNormals){
		checkAttribute( mesh, "smoothNormal", 3, VAD::DATA_FORMAT_FLOAT32 ) ;
		smoothAttribute3f( mesh, "normal", "smoothNormal", smoothNormals_Sigma, true);	
	}

}
