//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "GlobalMacros.h"
//---------------------------------------------------------------------------
#include "PCCImporter.h"
//---------------------------------------------------------------------------
#include "PointCloudImporterAdder.h"
#include "PointCloudImporterReader.h"
#include "PointCloudImporterReaderCustomASCII.h"
#include "BinaryObjectStreams.h"
#include "StringHelper.h"
#include "SGListNode.h"
#include "UnstructuredInCoreTriangleMesh.h"
//---------------------------------------------------------------------------
#include "SGObjectNode.h"
#include "SGAllocatorList.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


MetaClass* PCCImporter::PCCImporterMetaClass = NULL;

const MetaClass* PCCImporter::getInstanceClass() const {
   return getClass();
}

Object* PCCImporter::newInstance() {
   return new PCCImporter;
}

const MetaClass* PCCImporter::getClass() {
   if (PCCImporterMetaClass == NULL) {
      PCCImporterMetaClass = new MetaClass("PCCImporter", "no description.",
                                       newInstance, sizeof(PCCImporter));
   }
   return PCCImporterMetaClass;
}

void PCCImporter::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 0);
   PCCImporterMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT( PCCImporter )
	ADD_STRING_PROP(m_InputFileName, 0)
	ADD_OBJECT_PROP(m_ImporterReader, 0, PointCloudImporterReader::getClass(), true );
	ADD_STRING_PROP(m_TargetName, 0)
	ADD_BOOLEAN_PROP(paramWriteByCoords, 0)
	ADD_STRING_PROP(paramCoordsChannel, 0)
   ADD_FLOAT64_PROP(scalePCoordsX, 0)
   ADD_FLOAT64_PROP(scalePCoordsY, 0)
   ADD_FLOAT64_PROP(offsetPCoordsX, 0)
   ADD_FLOAT64_PROP(offsetPCoordsY, 0)
}

void PCCImporter::shutdown() {
   delete PCCImporterMetaClass;
}

PCCImporter::PCCImporter() {
   m_InputFileName = "";
   m_TargetName = "root/pc1";
   m_ImporterReader = new PointCloudImporterReaderCustomASCII();
   paramWriteByCoords = false;
   paramCoordsChannel = "param_coords";
   scalePCoordsX = 1;
   scalePCoordsY = 1;
   offsetPCoordsX = 0;
   offsetPCoordsY = 0;
}

PCCImporter::~PCCImporter() 
{
   delete m_ImporterReader;
}


void PCCImporter::execute(Scene *scene) {

	SceneGraphNode* SGNode = getSceneGraphNodeByName(scene, m_TargetName);
	if (SGNode==NULL) {
		WARNINGMESSAGE("PCCImporter::execute()\nTarget \"" + m_TargetName + "\" not found in scene graph!");
		return;
	};

	if (strUCase(extractFileExt(m_InputFileName))==".OBJECT") {
		// execute geoxl .object importer
		SGListNode *lnode = dynamic_cast<SGListNode*>(SGNode);
		if (lnode == NULL) {
			WARNINGMESSAGE("PCCImporter::execute()\nTarget \"" + m_TargetName + "\" is not a list node, but this is needed for GEOXL .object import due to some reasons!\nPlease set a list node (e.g. root) as target.");
			return;
		}
		execGEOXLObjectImport(lnode);

	} else {
		// execute import by reader/adder
		PointCloud *targetPC = getPointCloud(SGNode);
		if (targetPC==NULL) {
			WARNINGMESSAGE("PCCImporter::execute()\nTarget \"" + m_TargetName + "\" is not a point cloud!");
			return;
		};
		execImporter(targetPC);
	};

}


void PCCImporter::execImporter(PointCloud* targetPC) {

   pAssert(targetPC != NULL);
	if (!geoxl::FileExists(m_InputFileName.c_str())) {
		WARNINGMESSAGE("PCCImporter::execImporter()\nFile \"" + m_InputFileName + "\" not found.");
		return;
	};

	if ((targetPC->supportsIteratorType(SOIT::CAP_TRIANGLE_MESH)) && (!m_ImporterReader->supportsTriangleImport())) {
		WARNINGMESSAGE("PCCImporter::execImporter()\nTarget is a triangle mesh type but chosen import-reader doesn't support triangle import!");
		return;
	}

	// begin import
	bool success = m_ImporterReader->beginImport(m_InputFileName, targetPC->supportsIteratorType(SOIT::CAP_TRIANGLE_MESH));
		if (!success) {
			WARNINGMESSAGE("PCCImporter::execImporter()\nImport preparations failed!");
			return;
		}
		PointCloudImporterAdder* adder = NULL;

		if (targetPC->supportsIteratorType(SOIT::CAP_BASIC_PC | SOIT::CAP_DYNAMIC_PC | SOIT::CAP_TRIANGLE_MESH)) {
			// mesh adder
			adder = new PointCloudImporterAdderMesh();
			UnstructuredInCoreTriangleMesh* test = dynamic_cast<UnstructuredInCoreTriangleMesh*>(targetPC);
			if (test==NULL) {
				// just UnstructuredInCoreTriangleMesh are supported at the moment
				// if mesh types are extended please change in PointCloudImporterAdderMesh also!
				WARNINGMESSAGE("PCCImporter::execImporter()\nMesh Importer just supports UnstructuredInCoreTriangleMesh at the moment!");
				pAssert(false);
				return;
			}

		} else if (targetPC->supportsIteratorType(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PARAMETRIC_PC)) {
			// parametric dynamic adder
			adder = new PointCloudImporterAdderParametricDynamic(paramWriteByCoords, paramCoordsChannel, scalePCoordsX, scalePCoordsY, offsetPCoordsX, offsetPCoordsY);
		} else if (targetPC->supportsIteratorType(SOIT::CAP_BASIC_PC | SOIT::CAP_DYNAMIC_PC)) {
			// dynamic adder
			adder = new PointCloudImporterAdderDynamic();
		} else {
			WARNINGMESSAGE("PCCImporter::execImporter()\nCannot handle this point cloud type! Need point clouds which are dynamic-parametric, dynamic or triangle mesh!");
			return;
		};

		adder->doImport(m_ImporterReader, targetPC);
		delete adder;

	m_ImporterReader->endImport();

};


void PCCImporter::execGEOXLObjectImport(SGListNode *listnode) {
	if (listnode==NULL) return;

	if (!geoxl::FileExists(m_InputFileName.c_str())) {
		WARNINGMESSAGE("PCCImporter::execGEOXLObjectImport()\nFile \"" + m_InputFileName + "\" not found.");
		return;
	};

	BinaryInputObjectStream in(m_InputFileName.c_str());
	Persistent *obj = NULL;
	in.readObject(obj);
	if (dynamic_cast<PointSet*>(obj) != NULL) {
		PointSet *ps = dynamic_cast<PointSet*>(obj);
		UnstructuredInCorePointCloud *pc = new UnstructuredInCorePointCloud;
		pc->setPointSet(ps);
		SGObjectNode *n = new SGObjectNode();
		n->clearAndSetup(pc);
		n->setName("importedPointSet");
		listnode->addNode(n);
	} else if (dynamic_cast<PointCloud*>(obj) != NULL) {
		PointCloud *pc = dynamic_cast<PointCloud*>(obj);
		SGObjectNode *n = new SGObjectNode();
		n->clearAndSetup(pc);
		n->setName("importedPointCloud");
		listnode->addNode(n);
	} else if (dynamic_cast<Scene*>(obj) != NULL) {
		Scene *s = dynamic_cast<Scene*>(obj);
		s->getRootNode()->setName("importedScene");
		listnode->addNode(s->getRootNode());
	
		// delete references to allocators before delting the scene
		s->getRootState()->staticState->allocators->hackRemoveAllocatorReferences();

		delete s;
	} else {
		WARNINGMESSAGE("PCCImporter::execGEOXLObjectImport()\nObject class not recognized. Supported types are: PointSet, PointCloud, Scene. Actual type is: " + obj->getInstanceClass()->getClassName());
		return;
	}

};


void PCCImporter::setImporterReader(PointCloudImporterReader* newImp) {
	if (m_ImporterReader->getInstanceClass()==newImp->getInstanceClass()) return;
	delete m_ImporterReader;
	m_ImporterReader = newImp;
};
