//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCPointCloudCreator.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCorePointCloud.h"
#include "ParametricInCorePointCloud.h"
#include "HierarchicalInCorePointCloud.h"
#include "HierarchicalOutOfCorePointCloud.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "BlockedMultiObjectStream.h"
#include "MHSettings.h"
#include "SGAllocatorList.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



MetaClass* PCCPointCloudCreator::PCCPointCloudCreatorMetaClass = NULL;

const MetaClass* PCCPointCloudCreator::getInstanceClass() const {
   return getClass();
}

Object* PCCPointCloudCreator::newInstance() {
   return new PCCPointCloudCreator;
}

const MetaClass* PCCPointCloudCreator::getClass() {
   if (PCCPointCloudCreatorMetaClass==NULL) {
      PCCPointCloudCreatorMetaClass = new MetaClass("PCCPointCloudCreator", "no description.",
                                  newInstance, sizeof(PCCPointCloudCreator));
   }
   return PCCPointCloudCreatorMetaClass;
}


static string describeType(card32 q) {
   if (q == 0) {
      return "Unstructured";
   } else if (q == 1) {
      return "Parametric";
   } else if (q == 2) {
      return "Hierachy In-Core";
   } else if (q == 3) {
      return "Hierachy Out-Of-Core";
   } else if (q == 4) {
      return "Unstructured In-Core Triangle-Mesh";
   } else {
      return "UNDEFINED";
   }
}


void PCCPointCloudCreator::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION+3);
   PCCPointCloudCreatorMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT( PCCPointCloudCreator )
	ADD_STRING_PROP(m_pcloud_name,0)
	ADD_OBJECT_PROP(m_vd,0,VertexDescriptor::getClass(),true)
	ADD_CARD32_PROP_UPDATE_DESCR(m_type,0,NULL,describeType, 5)
	ADD_CARD32_PROP(allocatorIndex,3)
	ADD_STRING_PROP(ooc_filename_depreciated,0)
   SET_USER_EDIT(false)
	ADD_CARD32_PROP(maxNodeInMem,0)
	ADD_CARD32_PROP(blockSize_depreciated,0)
   SET_USER_EDIT(false)
	ADD_MAX_PLATFORM_CARD_PROP(maxNumObjectsPerLeaf,0)
	ADD_CARD32_PROP(multiResGridSize,0)
}

void PCCPointCloudCreator::shutdown() {
   delete PCCPointCloudCreatorMetaClass;
}


// ----

PCCPointCloudCreator::PCCPointCloudCreator():
    m_pcloud_name("root/pc1")
{
   m_vd = new VertexDescriptor;
   m_vd->pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   m_vd->pushAttrib(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
   m_vd->pushAttrib(mVAD("color", 3, VAD::DATA_FORMAT_FLOAT32));
   m_vd->pushAttrib(mVAD("flags", 1, VAD::DATA_FORMAT_INT32));

   m_type = 0;	// begin with unstructured

	allocatorIndex = 0;
   ooc_filename_depreciated= "rename_me.swap";
   maxNodeInMem=128;
   blockSize_depreciated=524288;

   MHStandardSettings stdSettings;
   maxNumObjectsPerLeaf = stdSettings.maxNumObjectsPerLeaf;
   multiResGridSize = stdSettings.multiResGridSize;
}

PCCPointCloudCreator::~PCCPointCloudCreator()
{
   delete m_vd;
}

//#include <qtCore/qdatetime>
//#include <qtCore/qCoreApplication>

void PCCPointCloudCreator::execute(Scene *scene) 
{
/*	debugOutput << "start";
	QTime time;
	time.start();
	int tmp=0;
	while( time.elapsed() < 3 * 1000 )
	{
		if( time.elapsed() > tmp * 1000 )
		{
			tmp++;
			debugOutput << "asd" << tmp << "\n";
			QCoreApplication::flush();
		}
	}*/
	executeCreator(scene);
}


void PCCPointCloudCreator::executeCreator(Scene *scene) {

   PointCloud* pc;

   //for out-of-core
   MultiObjectStream *multiObj = NULL;   
   MHOutOfCorePointSetOctreeNodeBase::NodeAllocator *nAlloc = NULL;

   MHStandardSettings settings;

   switch (m_type)
   {
      case 0:
         pc = new UnstructuredInCorePointCloud();
         ((UnstructuredInCorePointCloud*)pc)->clearAndSetup(m_vd, 0);
         break;
      case 1:
         pc = new ParametricInCorePointCloud();
         ((ParametricInCorePointCloud*)pc)->clearAndSetup(0, 0, m_vd);
         break;
      case 2:
         pc = new HierarchicalInCorePointCloud();
         ((HierarchicalInCorePointCloud*)pc)->clearAndSetup(m_vd);

         ((HierarchicalInCorePointCloud*)pc)->getOctree()->getSettings(settings);
         settings.maxNumObjectsPerLeaf=  maxNumObjectsPerLeaf;
         settings.multiResGridSize= multiResGridSize ;
         ((HierarchicalInCorePointCloud*)pc)->getOctree()->setSettings(settings);
         break;
      case 3:
         pc = new HierarchicalOutOfCorePointCloud();

			if (allocatorIndex >= scene->getRootState()->staticState->allocators->getNumAllocators()) {
				throw ECommandFailed("PCCPointCloudCreator::executeCreator() - allocator index out of range.");
			}
			
			multiObj = scene->getRootState()->staticState->allocators->getAllocator(allocatorIndex);
         nAlloc = new MHOutOfCorePointSetOctreeNode::NodeAllocator(
				MHOutOfCorePointSetOctreeNode::getClass(),
				multiObj,
				maxNodeInMem  // maxNumNodesInMem
			);
         // mw - 10/18/07 - moved in constructor - multiObj->addGlobalObjectPointer(nAlloc, 1);

         ((HierarchicalOutOfCorePointCloud*)pc)->clearAndSetup(m_vd, nAlloc);

         ((HierarchicalOutOfCorePointCloud*)pc)->getOctree()->getSettings(settings);
         settings.maxNumObjectsPerLeaf=  maxNumObjectsPerLeaf;
         settings.multiResGridSize= multiResGridSize ;
         ((HierarchicalOutOfCorePointCloud*)pc)->getOctree()->setSettings(settings);
         break;
		case 4: {
			pc = new UnstructuredInCoreTriangleMesh();
			VertexDescriptor faceDescr;
			faceDescr.pushAttrib(mVAD("index", 3, VAD::DATA_FORMAT_INT32));
			((UnstructuredInCoreTriangleMesh*)pc)->clearAndSetup(m_vd, 0, &faceDescr, 0);
			}
         break;
		// mh - 04/10/2006 - removed AnimatedPointCloud
      default: return;
   };


	addPointCloud(scene, pc, m_pcloud_name);

}

