//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCResampler.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGAnimationNode.h"
#include "ProgressWindow.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( PCCResampler, 1 )
{
	BEGIN_CLASS_INIT( PCCResampler )
	ADD_FLOAT32_PROP( spacing, 1 )
	ADD_SEPARATOR( "Below: deprecated" )
	ADD_BOOLEAN_PROP( selection, 0 );
	ADD_STRING_PROP( source, 0 );
	ADD_CARD32_PROP( numPoints, 0 );
	ADD_BOOLEAN_PROP( animated, 0 );
}

PCCResampler::PCCResampler()
{
	spacing = 0.01f;

	/*** DEPRECATED ***/
	source = "root/data";
	numPoints = 20000;
	animated = true;
	selection = false;
}

PCCResampler::~PCCResampler()
{
}

void PCCResampler::resamplePC(PointCloud* pc, float spacing)
{
   if (pc == NULL) return;
   
   // TODO: check if point cloud is hierarchical

   if ( !pc->providesAttribute( "position" ) ){
      debugOutput << "Resampler: point cloud does not provide attribute 'position'\n"; debugOutput.flush();
      return;
   }

   if ( !pc->providesAttribute( "flags" ) ){
      debugOutput << "Resampler: point cloud does not provide attribute 'flags'\n"; debugOutput.flush();
      return;
   }

   AAT POSITION = pc->getAAT( "position" );
   //AAT COLOR = pc->getAAT( "color" );
   AAT FLAGS = pc->getAAT( "flags" );

   // Create iterators
   SceneObjectIterator* it = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC ) );
   BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
   ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
   DynamicPointCloudIterator* dynIt = dynamic_cast<DynamicPointCloudIterator*>( it );
   ModifyableKNNIterator* knnIt = new ModifyableKNNIterator( pc, 32, NULL );

   if ( basicIt == NULL ){ 
      debugOutput << "Cannot create basic iterator.\n"; debugOutput.flush();
      return;
   }

   if ( modIt == NULL ){ 
      debugOutput << "Cannot create modifyable iterator.\n"; debugOutput.flush();
      return;
   }

   if ( dynIt == NULL ){ 
      debugOutput << "Cannot create dynamic iterator.\n"; debugOutput.flush();
      return;
   }

   // Reset all flags
   debugOutput << "Resampler: resetting flags ...\n"; debugOutput.flush();
   progressWindow->pushStep( true, "Reset" );
   basicIt->reset();
   unsigned long pointIndex = 0;
   while( !basicIt->atEnd() ){
      const Vector3f& p = basicIt->get3f( POSITION );
      if ( _isnan( p[0] ) || _isnan( p[1] ) || _isnan( p[2] ) ){
         debugOutput << "Resampler: point cloud contains NAN point ... deleting point\n"; debugOutput.flush();
         dynIt->deleteCurrent();
      }
      if ( !_finite( p[0] ) || !_finite( p[1] ) || !_finite( p[2] ) ){
         debugOutput << "Resampler: point cloud contains INF point ... deleting point\n"; debugOutput.flush();
         dynIt->deleteCurrent();
      }

      modIt->set1i( FLAGS, 0 );
      basicIt->next();
      pointIndex++;
      progressWindow->progress( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 );
   }
   bool aborted;
   progressWindow->popSteps( aborted );

   card32 numPointStart = pc->getNumPoints();

   // Mark all delete points
   debugOutput << "Resampler: tagging points ...\n"; debugOutput.flush();
   progressWindow->pushStep( true, "Tagging" );
   float32 sqrSpacing = sqr( spacing );
   basicIt->reset();
   pointIndex = 0;
   while( !basicIt->atEnd() ){
      const Vector3f& p = basicIt->get3f( POSITION );
      const int32 flag = basicIt->get1i( FLAGS );
      if ( flag == 0 ){
         knnIt->setSeekPointAndReset( p );
         knnIt->next();			// Skip self
         while( !knnIt->atEnd() ){
			 const Vector3f& np = knnIt->get3f( POSITION );
			 float sqrD = ( p - np ) * ( p - np );
			 if( sqrD > sqrSpacing )
				 break;
            knnIt->set1i( FLAGS, 1 );
            knnIt->next();
         }
      }
      basicIt->next();
      pointIndex++;
      progressWindow->progress( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 );
   }
   progressWindow->popSteps( aborted );
   delete knnIt;

   // Delete marked points
   progressWindow->pushStep( true, "Deleting" );
   debugOutput << "Resampler: deleting points ...\n"; debugOutput.flush();
   basicIt->reset();
   pointIndex = 0;
   card32 numDeleted = 0;
   while( !basicIt->atEnd() ){
      if ( basicIt->get1i( FLAGS ) == 1 ){
         //modIt->set3f( COLOR, makeVector3f( 1,1,0 ) );
         dynIt->deleteCurrent();
         numDeleted++;
      }
      else
         basicIt->next();
      progressWindow->progress( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 );
   }
   progressWindow->popSteps( aborted );

   debugOutput << "deleted: " << numDeleted << " points\n";
   debugOutput << "Resampled  from " << numPointStart << " points to " << card32( pc->getNumPoints() ) << " points.\n"; debugOutput.flush();

   delete it;

}

void PCCResampler::execute( Scene *scene )
{
	vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
	float32 sqrSpacing = sqr( spacing );
	for ( unsigned int m = 0; m < selection.size(); m++ ){

		// Get point cloud
		PointCloud* pc = getPointCloud( scene, selection[m] );
		resamplePC(pc, spacing);
	}


	/*
	if ( animated ){
		StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
		SGAnimationNode* animNode = NULL;
		it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
		while ( !it->atEnd() ){
			if ( it->getNameWithPath() == source ){
				animNode = dynamic_cast<SGAnimationNode*>( it->getNode() );
				if ( animNode != NULL ){
					break;
				}
			}
			it->next();
		}
		delete it;
		if ( animNode == NULL ){
			debugOutput << "PCCResampler::execute: cannot find animation node!\n";
			return;
		}
		// Check number of frames
		card32 numFrames = animNode->getNumChildNodes( scene->getRootState() );
		for ( unsigned int n = 0; n < numFrames; n++ ){
			SceneGraphNode* child = animNode->getChildNode( scene->getRootState(), n );
			SGObjectNode* sgo = dynamic_cast<SGObjectNode*>( child );
			if (sgo != NULL) {
				PointCloud* pc = dynamic_cast<PointCloud*>( sgo->getSceneObject() );
				UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
				if ( upc != NULL ){
					PointSet* newPS = downSample( upc->getPointSet(), numPoints );
					upc->setPointSet( newPS );
					upc->clearAttachments();
				}
			}	
		}
	} else {
        UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( getPointCloud( scene, source ) );
        if ( upc == NULL ){
			debugOutput << "PCCResampler::execute: can't find source point cloud!\n";
            return;
        }
		PointSet* newPS = downSample( upc->getPointSet(), numPoints );
		upc->setPointSet( newPS );
		upc->clearAttachments();
	}
	*/
}

/*
PointSet* PCCResampler::downSample( PointSet* ps, const card32 numPoints )
{
if ( ps->getNumEntries() <= numPoints ){
return ps;
}
PointSet* newPS = new PointSet;
newPS->clearAndSetup( 1, numPoints, ps->getDescr() );
float32 inc = float32( ps->getNumEntries() ) / float32( numPoints );
card32 newPointIndex = 0;
char* buffer = new char[ ps->getDescr()->getSize() ];
for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
ps->getVertex( pointIndex, buffer );
newPS->setVertex( newPointIndex, buffer ); 
newPointIndex++;
while( pointIndex < newPointIndex * inc ){
pointIndex++;
}
}
delete [] buffer;
return newPS;
}
*/