//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCCTriangleMeshSampler.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "TriangleMeshPointCloud.h"
#include "VertexDescriptor.h"
#include "NumericalClassProperty.h"
#include "StringClassProperty.h"
#include "PointSet.h"
#include "Random.h"
#include "SGObjectNode.h"
#include "MathTools.h"
#include "ProgressWindow.h"
#ifndef _DEBUG
#include "PointSet_templCode.h"
#endif
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


MetaClass* PCCTriangleMeshSampler::PCCTriangleMeshSamplerMetaClass = NULL;

const MetaClass* PCCTriangleMeshSampler::getInstanceClass() const {
   return getClass();
}

Object* PCCTriangleMeshSampler::newInstance() {
   return new PCCTriangleMeshSampler;
}

const MetaClass* PCCTriangleMeshSampler::getClass() {
   if (PCCTriangleMeshSamplerMetaClass==NULL) {
      PCCTriangleMeshSamplerMetaClass = new MetaClass("PCCTriangleMeshSampler", "no description.",
                                  newInstance, sizeof(PCCTriangleMeshSampler));
   }
   return PCCTriangleMeshSamplerMetaClass;
}

void PCCTriangleMeshSampler::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 2);
   PCCTriangleMeshSamplerMetaClass->copyParentProperties();

	BEGIN_CLASS_INIT( PCCTriangleMeshSampler )
	ADD_STRING_PROP(sourceNode,0)
	ADD_STRING_PROP(destNode,0)
	ADD_CARD64_PROP(expNumPoints,0)
	ADD_BOOLEAN_PROP(createTangentDiscs,0)
	ADD_FLOAT32_PROP(tangentDiscOversampling,0)
	ADD_BOOLEAN_PROP(surfaceNormalSmoothing,2);
}

void PCCTriangleMeshSampler::shutdown() {
   delete PCCTriangleMeshSamplerMetaClass;
}


// ----


void geoxl::sampleMeshRandomly(Scene * scene, string destID, SceneObject *sourceMesh, SceneObject *destPointCloud, card64 numPoints, bool createTangentDiscs, 
										float32 tangentDiscOversampling, bool surfaceNormalSmoothing) {

   // scene object types...
   TriangleMeshPointCloud *sourceAsMesh = dynamic_cast<TriangleMeshPointCloud*>(sourceMesh);
   if (!sourceAsMesh) {
      throw PException("geoxl::sampleMeshRandomly() - source is no TriangleMeshPointCloud.");
   }
   PointCloud *sourceAsPC = dynamic_cast<PointCloud*>(sourceMesh);

	if(!sourceAsMesh) {
      throw PException("geoxl::sampleMeshRandomly() - source is no PointCloud.");
   }
	
	pAssert(sourceAsPC);
   PointCloud *destAsPC = dynamic_cast<PointCloud*>(destPointCloud);
   if (!destAsPC) {
      throw PException("geoxl::sampleMeshRandomly() - dest is no PointCloud.");
   }

	// append necessary fields to destPointCloud
	if( !destAsPC->getDescr()->equal( sourceAsPC->getDescr() ) )
	{
		const VertexDescriptor * vd = sourceAsPC->getDescr();
		for( card32 i=0;i<vd->getNumAttributes();i++ )
		{
			checkAttribute( scene, destID, vd->getAttribute(i).name, vd->getAttribute(i).numComponents, vd->getAttribute(i).dataFormat );
		}

	   SceneGraphNode *destN = getSceneGraphNodeByName(scene, destID);
	   SGObjectNode *destNO = dynamic_cast<SGObjectNode*>(destN);
		destPointCloud = destNO->getSceneObject();
		destAsPC = dynamic_cast<PointCloud*>(destNO->getSceneObject());
	}


   // iterators...
   SceneObjectIterator *sourceIt = sourceMesh->createIterator(SOIT::CAP_BASIC_PC);
   BasicPointCloudIterator *sourceIt_basic = dynamic_cast<BasicPointCloudIterator*>(sourceIt);
   if (sourceIt == NULL || sourceIt_basic == NULL) {
      delete sourceIt;
      throw PException("geoxl::sampleMeshRandomly() - source is no basic point cloud.");
   }

   SceneObjectIterator *sourceMeshIt = sourceMesh->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_TRIANGLE_MESH);
   BasicPointCloudIterator *sourceMeshIt_basic = dynamic_cast<BasicPointCloudIterator*>(sourceMeshIt);
   if (sourceMeshIt == NULL) {
      delete sourceIt; delete sourceMeshIt;
      throw PException("geoxl::sampleMeshRandomly() - source is no basic mesh.");
   }

   SceneObjectIterator *destIt = destPointCloud->createIterator(SOIT::CAP_DYNAMIC_PC);
   DynamicPointCloudIterator *destIt_dyn = dynamic_cast<DynamicPointCloudIterator*>(destIt);
   if (destIt == NULL) {
      delete destIt; delete sourceMeshIt; delete sourceIt;
      throw PException("geoxl::sampleMeshRandomly() - dest is not a dynamic point cloud.");
   }



   // vertex descriptors, AATS...
   const VertexDescriptor *sourceVertexDescr = sourceAsPC->getDescr();
   const VertexDescriptor *destVertexDescr = destAsPC->getDescr();

   if (!sourceVertexDescr->equal(destVertexDescr)) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw PException("geoxl::sampleMeshRandomly() - source and dest VertexDescriptors do not match.");
   }
   if (!sourceAsPC->providesAttribute("position")) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw PException("geoxl::sampleMeshRandomly() - source does not have position channel.");
   }
   AAT SOURCE_POS = sourceAsPC->getAAT("position");
   AAT TANGENT_U;
   if (destAsPC->providesAttribute("tangent_u")) {
      TANGENT_U = destAsPC->getAAT("tangent_u");
   } else {
      if (createTangentDiscs) throw PException("geoxl::sampleMeshRandomly() - cannot create tangent discs, attribute channel is missing.");
   }
   AAT TANGENT_V;
   if (destAsPC->providesAttribute("tangent_v")) {
      TANGENT_V = destAsPC->getAAT("tangent_v");
   } else {
      if (createTangentDiscs) throw PException("geoxl::sampleMeshRandomly() - cannot create tangent discs, attribute channel is missing.");
   }
	AAT normalAAT;
	if( surfaceNormalSmoothing )
	{
		normalAAT = sourceAsPC->getAAT("normal");
	}

   if (!sourceAsMesh->getTriangleDescr()->providesAttribute("index")) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw PException("geoxl::sampleMeshRandomly() - source mesh pc does not provide index channel.");
   }
   AAT SOURCE_INDEX = sourceAsMesh->getTriangleDescr()->getToken("index");

   // lock vertices...
   const PointSet *vertices = sourceIt_basic->lockBlockRead();

   
   // setup progress window
	ProgressWindow::WeightList steps;
	steps.push_back( new ProgressWindow::WeightElem( true, 0.5f, "area calc" ) );
	steps.push_back( new ProgressWindow::WeightElem( true, 0.5f, "sampling" ) );
	progressWindow->pushSteps( steps );
	
	// step 1: calculate overall area...
   float64 area = 0;
	card32 numTriangles = sourceAsMesh->getNumTriangles();
	card32 actTriangle = 0;
	bool aborted = false;
   while (!sourceMeshIt_basic->atEnd() && !aborted) {
      Vector3i indices = sourceMeshIt_basic->get3i(SOURCE_INDEX);
      Vector3f pos[3];
      pos[0] = vertices->get3f(indices[0], SOURCE_POS);
      pos[1] = vertices->get3f(indices[1], SOURCE_POS);
      pos[2] = vertices->get3f(indices[2], SOURCE_POS);

      float32 thisarea = 0.5f*norm((pos[1]-pos[0]).crossProduct(pos[2]-pos[0]));

      area += (float64)thisarea;

      sourceMeshIt_basic->next();

		// progress
		aborted = !progressWindow->progress( (float)actTriangle++ / (float)numTriangles * 100.0f );

   }
   sourceMeshIt_basic->reset();
	progressWindow->nextStep( aborted );

   float64 areaPerPoint = area / numPoints;
   float32 tangentSize = sqrt(areaPerPoint * tangentDiscOversampling);


	// step 2: sampling...
   char *destBuffer = new char[destVertexDescr->getSize()];
   //typedef const char *constCharPtr;
   const char **vbuff = new const char*[3];
	card32 actPoint = 0;
   while (!sourceMeshIt_basic->atEnd() && !aborted) {
      Vector3i indices = sourceMeshIt_basic->get3i(SOURCE_INDEX);
      Vector3f pos[3];
      pos[0] = vertices->get3f(indices[0], SOURCE_POS);
      pos[1] = vertices->get3f(indices[1], SOURCE_POS);
      pos[2] = vertices->get3f(indices[2], SOURCE_POS);

      Vector3f tNormal = (pos[1]-pos[0]).crossProduct(pos[2]-pos[0]);
      float64 thisarea = 0.5f*norm(tNormal);

      float64 numPoints = thisarea / areaPerPoint;

      card64 numPointsInt = floor(numPoints);
      float32 remPts = (float32)(numPoints - numPointsInt);

      if (remPts > rnd01()) numPointsInt++;

      for (card64 pnum = 0; pnum < numPointsInt; pnum++) {

         vbuff[0] = vertices->getDataPointer(indices[0]);
         vbuff[1] = vertices->getDataPointer(indices[1]);
         vbuff[2] = vertices->getDataPointer(indices[2]);


         float32 u = rnd01();
         float32 v = rnd01();
         if (u+v > 1) {
            float32 nu = 1.0f - v;
            v = 1.0f - u;
            u = nu;
         }

         float32 factors[3];
         factors[0] = u;
         factors[1] = v;
         factors[2] = 1.0f - u - v;

         sourceVertexDescr->interpolate(vbuff, factors, 3, destBuffer);

			if( surfaceNormalSmoothing )
			{
				Vector3f * posPtr = (Vector3f*)destBuffer;
				Vector3f * normalPtr = (Vector3f*)(&destBuffer[normalAAT.getOffset()]);
				Vector3f p[3];
				p[0] = *((Vector3f*)vbuff[0]);
				p[1] = *((Vector3f*)vbuff[1]);
				p[2] = *((Vector3f*)vbuff[2]);

				float dist1 = norm(p[0] - *posPtr);
				float dist2 = norm(p[1] - *posPtr);
				float dist3 = norm(p[2] - *posPtr);

				Vector3f normal[3];
				normal[0] = *((Vector3f*)(vbuff[0]+normalAAT.getOffset()));
				normal[1] = *((Vector3f*)(vbuff[1]+normalAAT.getOffset()));
				normal[2] = *((Vector3f*)(vbuff[2]+normalAAT.getOffset()));

				// implementation of PN curved triangles [Vlachos et al.]
				Vector3f b300 = p[0];
				Vector3f b030 = p[1];
				Vector3f b003 = p[2];
				float weight[3][3];
				for( unsigned i=0;i<3;i++ )
				{
					for( unsigned j=0;j<3;j++ )
					{
						weight[i][j] = (p[j]-p[i])*normal[i];
					}
				}
				Vector3f b210 = (p[0]*2.0f + p[1] - normal[0]*weight[0][1])/3.0f;
				Vector3f b120 = (p[1]*2.0f + p[0] - normal[1]*weight[1][0])/3.0f;
				Vector3f b021 = (p[1]*2.0f + p[2] - normal[1]*weight[1][2])/3.0f;
				Vector3f b012 = (p[2]*2.0f + p[1] - normal[2]*weight[2][1])/3.0f;

				Vector3f b102 = (p[2]*2.0f + p[0] - normal[2]*weight[2][0])/3.0f;
				Vector3f b201 = (p[0]*2.0f + p[2] - normal[0]*weight[0][2])/3.0f;

				Vector3f E = (b210 + b120 + b021 + b012 + b102 + b201) / 6.0f;
				Vector3f V = (p[0]+p[1]+p[2])/3.0f;

				Vector3f b111 = E + (E-V)/2.0f;

				float w = factors[2];
				float u_2 = u*u; float u_3 = u_2 * u;
				float v_2 = v*v; float v_3 = v_2 * v;
				float w_2 = w*w; float w_3 = w_2 * w;
				*posPtr = b300 * w_3 + b030 * u_3 + b003 * v_3
					+ b210*3.0f*w_2*u + b120*3.0f*w*u_2 + b201*3.0f*w_2*v
					+ b021*3.0f*u_2*v + b102*3.0f*w*v_2 + b012*3.0f*u*v_2
					+ b111*6.0f*w*u*v;

				Vector3f middleNormals[3];
				for( unsigned i=0;i<3;i++ )
				{
					unsigned j=(i+1)%3;
					float nweight = (p[j]-p[i])*(normal[i]+normal[j])*2.0f
						/ normQuad(p[j]-p[i]);
					middleNormals[i] = normal[i] + normal[j] - (p[j]-p[i])*nweight;
					normalize( middleNormals[i]);
				}

				*normalPtr = normal[0]*w_2 + normal[1]*u_2 + normal[0]*v_2
					+middleNormals[0]*w*u + middleNormals[1]*u*v + middleNormals[2]*w*v;
			}

         if (createTangentDiscs) {
            Vector3f normTNormal = tNormal / (thisarea * 2.0f);
            Vector3f tangent_u = makeOrthonormalVector(normTNormal);
            Vector3f tangent_v = tangent_u.crossProduct(normTNormal);
            tangent_u *= tangentSize;
            tangent_v *= tangentSize;
            TANGENT_U.setAttrib<float32,3>(destBuffer, tangent_u);
            TANGENT_V.setAttrib<float32,3>(destBuffer, tangent_v);
         }

         destIt_dyn->add(destBuffer);

      }

      sourceMeshIt_basic->next();

	}
   delete[] vbuff;
   delete[] destBuffer;

   sourceIt_basic->unlock();
	
	progressWindow->popSteps( aborted );
}


PCCTriangleMeshSampler::PCCTriangleMeshSampler() {
   expNumPoints = 100;
   createTangentDiscs = true;
   tangentDiscOversampling = 10;
	surfaceNormalSmoothing = false;
	sourceNode="root/pc1";
	destNode="root/pc2";
}


void PCCTriangleMeshSampler::execute(Scene *scene) {
   SceneGraphNode *sourceN = getSceneGraphNodeByName(scene, sourceNode);
   SceneGraphNode *destN = getSceneGraphNodeByName(scene, destNode);
   SGObjectNode *sourceNO = dynamic_cast<SGObjectNode*>(sourceN);
   SGObjectNode *destNO = dynamic_cast<SGObjectNode*>(destN);
   if (!sourceNO) throw PException("PCCTriangleMeshSampler::execute() - sourceNode not found.");
   if (!destNO) throw PException("PCCTriangleMeshSampler::execute() - destNode not found.");;

   SceneObject *sourceSO = sourceNO->getSceneObject();
   SceneObject *destSO = destNO->getSceneObject();
   if (sourceSO == NULL || destSO == NULL) return;
   sampleMeshRandomly(scene, destNode, sourceSO, destSO, expNumPoints, createTangentDiscs, tangentDiscOversampling,surfaceNormalSmoothing);
}

