#include "StdAfx.h"
#include "MacroScriptEditorWidget.h"
#include "Timer.h"
#include "DebugOutput.h"
#include "Configuration.h"
#include "FileDialogs.h"
#include "BinaryObjectStreams.h"
#include "ExceptionHandling.h"
#include "ObjectDictionaryEditorWidget.h"
#include "ScriptEngineDebugWidget.h"
#include "CmdObjScriptProgram.h"
#include "CommandObjMacro.h"
#include "qtgui/QLayout.h"
#include "SceneEditorWidget.h"
#include "AttachedDataTypes.h"
//---------------------------------------------------------------------------

using namespace geoxl;

bool MacroScriptEditorWidget::scriptRunning = false;
SceneEditorWidget *MacroScriptEditorWidget::activeEditorWidget = NULL;

/*MacroScriptEditorWidget *geoxl::macroScriptEditorWidget = NULL;*/


MacroScriptEditorWidget::MacroScriptEditorWidget(Scene *scene, SceneEditorWidget *editorWidget)
:QMainWindow()
{
	//call ui.setupUi(this) to set up the layout from QtDesigner
	ui.setupUi(this);
	// used Qt auto-connect feature for slots

	this->scene = scene;
	this->editorWidget = editorWidget;
	
   // Create dictionary attachment if it does not exist yet
   AttachedDataContainer* attachments = scene->getAttachments();
   if (attachments->getData("dictionary") == NULL) {
      AttachedDictionaryObject* attachedDic = new AttachedDictionaryObject();
      attachedDic->setup("dictionary", AttachedData::ADF_PERSISTENT);
      attachments->attachData(attachedDic);
   }

   AttachedDictionaryObject* attachedDic =
     (AttachedDictionaryObject*)scene->getAttachments()->getData("dictionary");
   dictionary = attachedDic->getObjectDictionary();

   dictionaryEditor = new ObjectDictionaryEditor;
   dictionaryEditorWidget = (ObjectDictionaryEditorWidget*)dictionaryEditor->makeEditor(ui.frame, dictionary);
   QHBoxLayout* editorLayout = new QHBoxLayout(ui.frame);
   editorLayout->addWidget(dictionaryEditorWidget);

   dictionaryEditor->update();
}

void MacroScriptEditorWidget::update() {
   dictionaryEditor->update();
}

MacroScriptEditorWidget::~MacroScriptEditorWidget(void)
{
   delete dictionaryEditor;
}

void MacroScriptEditorWidget::on_action_Load_triggered(bool checked)
{
   dictionaryEditorWidget->load();
}

void MacroScriptEditorWidget::on_action_Save_triggered(bool checked)
{
   dictionaryEditorWidget->save();
}

void MacroScriptEditorWidget::on_action_Update_triggered(bool checked)
{
   dictionaryEditor->update();
}

void MacroScriptEditorWidget::on_BTConvertToScript_clicked()
{
   Object* obj = dictionaryEditorWidget->getSelectedObject();
   if (obj == NULL) {
      QMessageBox::critical(this, "Error", "No object selected");
      return;
   }

   CommandObjMacro* macro = dynamic_cast<CommandObjMacro*>(obj);
   if (macro == NULL) {
      QMessageBox::critical(this, "Error", "Only a macro can be converted to a script");
      return;
   }

   CmdObjScriptProgram* scriptProgram = new CmdObjScriptProgram();
   for (unsigned i = 0; i < macro->numCommands(); ++i)
      scriptProgram->addCommand(macro->getCommand(i));

   dictionary->addObject(scriptProgram);
   dictionaryEditor->update();   
}

void MacroScriptEditorWidget::on_BTExecute_clicked()
{
   Object* obj = dictionaryEditorWidget->getSelectedObject();
   if (obj == NULL) {
      QMessageBox::critical(this, "Error", "No object selected");
      return;
   }

   CommandObject* cmdObj = dynamic_cast<CommandObject*>(obj);
   if (cmdObj == NULL) {
      CommandObjMacro* macro = dynamic_cast<CommandObjMacro*>(obj);
      if (macro == NULL) {
         QMessageBox::critical(this, "Wrong object type", "Only command objects and macros can be executed");
         return;
      } else {
         STD_GLOBAL_EXCEPTION_HANDLER_START

            scriptRunning = true;
            activeEditorWidget = editorWidget;
            for (card32 i = 0; i < macro->numCommands(); ++i) {
               CommandObject* cmdObj = const_cast<CommandObject*>(macro->getCommand(i));
               scene->getCommandHistory()->executeCommand(scene, cmdObj);
            }
            scriptRunning = false;
            activeEditorWidget = NULL;

         STD_GLOBAL_EXCEPTION_HANDLER_END
      }
   } else {
      CmdObjScriptProgram* script = dynamic_cast<CmdObjScriptProgram*>(cmdObj);
      if (script != NULL) {
         try {
            scriptRunning = true;
            activeEditorWidget = editorWidget;
            scene->getCommandHistory()->executeCommand(scene, cmdObj);
            scriptRunning = false;
            activeEditorWidget = NULL;
         } catch(PException &e) {
            std::string msg = "Error in line " + intToStr(script->getScript()->getLine() + 1) + ": \n" + e.getMsg();
            QMessageBox::critical( this, "GEOXLscript error", QString(msg.c_str()));
         }
      } else {
         STD_GLOBAL_EXCEPTION_HANDLER_START
            scriptRunning = true;
            activeEditorWidget = editorWidget;
            scene->getCommandHistory()->executeCommand(scene, cmdObj);
            scriptRunning = false;
            activeEditorWidget = NULL;
         STD_GLOBAL_EXCEPTION_HANDLER_END
      }
   }
}

void MacroScriptEditorWidget::on_BTDebug_clicked()
{
   Object* obj = dictionaryEditorWidget->getSelectedObject();
   if (obj == NULL) {
      QMessageBox::critical(this, "Error", "No object selected");
      return;
   }

   CmdObjScriptProgram* script = dynamic_cast<CmdObjScriptProgram*>(obj);
   if (script != NULL) {
      STD_GLOBAL_EXCEPTION_HANDLER_START
         CmdObjScriptEngine* scriptEngine = script->getScript();
         scriptEngine->setScene(scene);
         ScriptEngineDebugWidget* debugger = new ScriptEngineDebugWidget(scriptEngine);
         debugger->start();
      STD_GLOBAL_EXCEPTION_HANDLER_END
   } else {
      QMessageBox::critical(this, "Error", "Debugger only works with GEOXLscript command objects (CmdObjScriptEngine)");
      return;
   }
}


bool MacroScriptEditorWidget::getScriptRunning() 
{
   return scriptRunning;
}

void MacroScriptEditorWidget::repaintInScript() 
{
   if (activeEditorWidget != NULL) {
      activeEditorWidget->repaintGLArea();
   	//QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
   } else {
      pAssert(false);
   }
}


#include "MacroScriptEditorWidget_moc.cpp"
