//---------------------------------------------------------------------------
#ifndef SceneEditorWidgetH 
#define SceneEditorWidgetH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "ClassEditor.h"
#include "SceneEditorForm_ui.h"
#include "Renderer.h"
#include "ClassInstanceSet.h"
#include "GLContext.h"
#include "GeoXLGLWidget.h"
#include "AnimationBarWidget.h"
//---------------------------------------------------------------------------

namespace geoxl {
	using namespace std;


class Scene;
class PCInteractionTool;
class PCCommand;
class CommandHistoryEditorWidget;
class MovieStudioWidget;
class MacroScriptEditorWidget;
class SceneInterface;
class AddToToolbarWidget;
class GEOXLEvent;
class SceneToolbarElement;
class SceneEditorOptions;
class VerticalScrollView;
class DebugRendererEditorWidget;


class EDITOR_API SceneEditorWidget : public QMainWindow {
	Q_OBJECT

 private:

	enum EImageFormat
	{
		IF_BMP,
		IF_PNG,
		IF_JPG
	};

	/// save current gl viewport to disk
	void saveScreenShot( string filename, EImageFormat format, int quality = 80 );

	/// reads current frame buffer + zbuffer and creates a "range scan"
	void rangeScan( string pcName );
	
	/// save frame to disk, uses parameters from movieStudio
	void saveFrame();

	string	m_ScreenShotBaseName;
	int		m_ScreenShotID;
	EImageFormat	m_ScreenShotImageFormat;
	bool		m_RecordHighQualityScreenShot;

	// member variables for tiled screenshots
	bool		m_SetupTiledImage;
	Vector2f	m_TiledImage_X0;
	Vector2f m_TiledImage_X1;

	Ui_SceneEditorForm ui;

	Scene	*	scene;

   // scene editor panel
   ClassEditor *sceneEditor;
   QWidget *sceneEditorControl;

   // tools editor panel
   ClassEditor *toolsEditor;
   QWidget *toolsEditorControl;
   PCInteractionTool *currentTool;

   // commands editor panel
   ClassEditor *cmdsEditor;
   QWidget *cmdsEditorControl;
   PCCommand *currentCommand;

   // rendering panel
   ClassEditor *rendererEditor;
   QWidget *rendererEditorControl;
   Renderer *currentRenderer;

   // scene graph state
   ClassEditor *staticStateEditor;
   QWidget *staticStateEditorWidget;
   ClassEditor *dynamicStateEditor;
   QWidget *dynamicStateEditorWidget;

	ClassInstanceSet oldTools;

	bool force_cam;
	PCInteractionTool *camTool;
	
	GLContext *glContext;
	GeoXLGLWidget* mainGLArea;
	static DebugRendererEditorWidget *debugRendererEditorWidget;

	QLabel* statusLabel;

	CommandHistoryEditorWidget* cmdHistoryForm;

   MovieStudioWidget * movieStudio;
   MacroScriptEditorWidget * macroScriptEditor;

	AddToToolbarWidget	*	addToToolbarDlg;
	QToolBar *animationBarToolbar;
	AnimationBarWidget *animationBar;

	void initLocalVars();
	void changeGlWindowResolution(int32 x, int32 y);

	// toolbar
	vector<QToolBar*> toolBars;
	card32 m_InterfaceVersion;
	map<QAction*,GEOXLEvent*>	toolBarEvents;

	// editor options
	SceneEditorOptions *	editorOptions;
	QDialog					*	editorOptionsDlg;

	// scroll areas for tabs
	VerticalScrollView * editSceneView;
	VerticalScrollView * toolView;
	VerticalScrollView * commandView;
	VerticalScrollView * rendererView;

	// timer id for redraw timer
	int	timerID;
	int	redrawDelay;
	
	class CallMeBack : public Object {
	 public:
	   SceneEditorWidget *me;
	   
	   CallMeBack(SceneEditorWidget *w) {me = w;}
   	// create command objects for clicked methods
	   bool cmdMethodCalled(string name);
   	bool toolMethodCalled(string name);
   };
   CallMeBack methodClickCallback;
   bool cmdMethodCalled(string name);
	bool toolMethodCalled(string name);

 public:
	SceneEditorWidget(QWidget * parent = NULL, Qt::WFlags f=0);
	~SceneEditorWidget(void);

	void setup( Scene * obj );

	/// update editor content with object data
	void updateEditor();

	AnimationBarWidget* getAnimationBar();
	void recordMovieNow(const char* rawFileName, float32 minTime, float32 maxTime, float32 timeStep, bool setSGTime,
		bool doRecord);
	void saveFrameBMP( const QString& filename );
	bool inContinuousRenderingLoop;

	void clearToolBarsVec();
	void addToolBarVec( QToolBar * tb );
	void clearToolBarEvents() {toolBarEvents.clear();}
	void addEvent( QAction * action, GEOXLEvent * e ) {toolBarEvents[action]=e;}

	void processEvent( GEOXLEvent * e );
	map<int, vector< SceneToolbarElement*> >	toolBarShortCuts;

	void updateSceneEditor( PCInteractionToolChanges whatChange );

	PCInteractionTool* getCamTool() const { return camTool; }
   GLContext *getGLContext() {return glContext;}
	
	void repaintGLArea();
	
	/// advises the SceneEditorWidget to immediately execute the command of the given class
	/// if the class is not found or of wrong type, an exception is thrown
	void executeCommandClass(const MetaClass *cmd);

   QPoint mapGLAreaToGlobal(const QPoint &p);

   /// pass the mouse event on to the camera controller (internal use)
   virtual void mouseDownPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   /// pass the mouse event on to the camera controller (internal use)
   virtual void mouseMovedPassToCam(int32 x, int32 y);
   /// pass the mouse event on to the camera controller (internal use)
   virtual void mouseUpPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   /// pass the mouse event on to the camera controller (internal use)
   virtual void areaResizePassToCam(card32 width, card32 height);
   /// pass the mouse event on to the camera controller (internal use)
   virtual void mouseWheelRotatedPassToCam(int32 rotatedDelta, Modifiers modifiersState);
   /// pass the keyboard event on to the camera controller (internal use)
   virtual void keyDownPassToCam(GeneralKey key);
   /// pass the keyboard event on to the camera controller (internal use)
   virtual void keyUpPassToCam(GeneralKey key);

 private:
	 void updateMenu();
	 SceneInterface * getSceneInterface();
	 void buildToolBar();
	 void addToToolbar( string className, bool et_object, bool et_command );
	 bool processToolbarKeyEvents( QKeyEvent * e ); // returns true if key is accepted

	 void changeCommand( const MetaClass * mc, PCCommand * obj=NULL);
	 void changeTool( const MetaClass * mc, PCInteractionTool * obj=NULL);

	 void saveTiledScreenshot(int numTiles, int numOverSampling);

 public slots:
	void sltGLResize(const QSize);
	void sltGLRepaint(void);
	void sltGLMouseMove(QMouseEvent *event);
	void sltGLMousePress(QMouseEvent *event);
	void sltGLMouseRelease(QMouseEvent *event);
	void sltGLWheelEvent(QWheelEvent * event);
	void sltGLDropEvent( QMimeData * d, QPoint pos );
	void sltScreenShot();


	void on_actionSave_as_triggered(bool checked);
	void on_actionImport_triggered(bool checked);
	void on_actionImport_Animation_triggered(bool checked);
	void on_actionExport_triggered(bool checked);
	void on_actionExport_Mesh_triggered(bool checked);
	void on_actionExport_Animation_triggered(bool checked);
	void on_actionE_xit_triggered(bool checked);
	
	void on_actionCommandHistory_triggered(bool checked);
	void on_actionUndo_triggered(bool checked);
	void on_actionRedo_triggered(bool checked);
	void on_actionRepeat_triggered(bool checked);
	void on_actionDelete_triggered(bool checked);
	void on_actionCopy_triggered(bool checked);
	void on_actionSplit_triggered(bool checked);
	void on_actionJoin_Clouds_triggered(bool checked);
	void on_actionDuplicate_triggered(bool checked);
	void on_actionSplit_to_single_cloud_triggered(bool checked);
	
	void on_actionMovieStudio_triggered(bool checked);
	void on_actionMacrosAndScripts_triggered(bool checked);
	void on_action256_x_256_triggered(bool checked);
	void on_action400_x_300_triggered(bool checked);
	void on_action512_x_512_triggered(bool checked);
	void on_action640_x_480_triggered(bool checked);
	void on_action1024_x_768_triggered(bool checked);
	void on_action1280_x_960_triggered(bool checked);
	void on_action_Auto_Layout_triggered(bool checked);
	
	
	void on_actionContext_Help_triggered(bool checked);
	void on_actionExport_Bitmap_triggered(bool checked);
	void on_actionRangescan_triggered(bool checked);
	void on_actionSet_GSA_OOC_Cache_Size_triggered(bool checked);
	void on_actionDebugRenderer_triggered(bool checked);

 private slots:

	void sltToolChanged( const QString & newToolName );
	void sltCommandChanged( const QString & newToolName );
	void sltSceneEditChanged( const QString & newToolName );
	void sltRendererChanged( const QString & newToolName );
	void sltExecCommand(void);

	void sltToolbarHandler( QAction * action );
	void sltToolbarCheckBoxHandler( bool checked, char * data );

	//void on_viewButton_clicked();
	//void on_selectButton_clicked();
	//void on_drawButton_clicked();
	//void on_transButton_clicked();
	void sltAnimationBarFrameChanged(int);
	void sltAnimationBarTraversalModeChanged(SGStaticState::TraversalModeType);
	void sltSwitchCameraMode();

	void on_pbAddTool_clicked();
	void on_pbAddCommand_clicked();
	
	void on_actionAdd_Quick_Cam_Checkbox_triggered();
	void on_actionEditor_Options_triggered();

	void on_actionScreenshot_triggered();
	
	void on_actionScreenshot_4000x4000_triggered();
	void on_actionScreenshot_Tiles_triggered();

	void on_actionBackgroundBlack_triggered();
	void on_actionBackgroundDark_triggered();
	void on_actionBackgroundBright_triggered();
	void on_actionBackgroundWhite_triggered();
	
	void sltStateDockShowing();
	


 protected:
	virtual void keyPressEvent ( QKeyEvent * event );
	virtual void keyReleaseEvent  ( QKeyEvent * event );

	virtual void focusInEvent( QFocusEvent * e );
	virtual void timerEvent( QTimerEvent * e);
};

MAKE_CLASS_EDITOR( EDITOR_API , Scene )


}

#endif
