//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCITransfEdit.h"
//---------------------------------------------------------------------------
#include "Scene.h"
#include "PointCloud.h"
#include "NumericalClassProperty.h"
#include "NumericalClassProperty_templCode.h"
#include "FixedArrayClassProperty.h"
#include "SceneGraphState.h"
#include "PointCloudIterators.h"
#include "IteratorAdaptors.h"
#include "LinearAlgebra_templCode.h"
#include "SelectionIteratorAdaptor.h"
#include "CmdObjTransfEdit.h"
#include "CommandHistory.h"
#include "DebugOutput.h"
#include "SGObjectNode.h"
#include "SelectionTools.h"
#include "PointCloudTools.h"
#include "ObjectViewsTable.h"
#include "StandardSceneGraphIterator.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



MetaClass* PCITransfEdit::PCITransfEditMetaClass = NULL;

const MetaClass* PCITransfEdit::getInstanceClass() const {
   return getClass();
}

Object* PCITransfEdit::newInstance() {
   return new PCITransfEdit;
}

const MetaClass* PCITransfEdit::getClass() {
   if (PCITransfEditMetaClass==NULL) {
      PCITransfEditMetaClass = new MetaClass("PCITransfEdit", "no description.",
                                  newInstance, sizeof(PCITransfEdit));
   }
   return PCITransfEditMetaClass;
}

void PCITransfEdit::init( const MetaClass *parentClass ) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 0);
   PCITransfEditMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT( PCITransfEdit )
   ADD_CARD32_PROP(TransfMode,0)
   ADD_VECTOR3F_PROP(rotScaleCenter, 0)
   ADD_BOOLEAN_PROP(moveRotScaleCent, 0)
   ADD_BOOLEAN_PROP(alignWorld, 0)
   ADD_BOOLEAN_PROP(transPoints, 0)
   ADD_BOOLEAN_PROP(lockRotScaleCent, 0)
   ADD_BOOLEAN_PROP(showWorldCS, 0)
   ADD_CARD32_PROP(showPivotCS, 0)
   ADD_FLOAT32_PROP(CSWorldScale, 0)
   ADD_FLOAT32_PROP(CSPivotScale, 0)
   ADD_BOOLEAN_PROP(SteppedRotation, 0)
   ADD_FLOAT32_PROP(SteppedRotationDegrees, 0)
   ADD_BOOLEAN_PROP(SteppedMovement, 0)
   ADD_FLOAT32_PROP(SteppedMovementValue, 0)
}

void PCITransfEdit::shutdown() {
   delete PCITransfEditMetaClass;
}


PCITransfEdit::PCITransfEdit() {
   lastX = 0;
   lastY = 0;
   leftDown = false;
   rightDown = false;
   shiftDown = false;
   ctrlDown = false;

   SteppedRotation = false;
   SteppedRotationDegrees = 15.0f;
   SteppedMovement = false;
   SteppedMovementValue = 1.0f;
   shiftDistance = 0;
   shiftDistanceToggleValue = 100.0f;

   width = 100;
   height = 100;
   WidthToHeight = 1.0f;
   
   alignWorld = false;
   transPoints = false;

   ReferencePoint = makeVector3f(0.0f,0.0f,0.0f);
   RefPointZ = 0.0f;
   RefPointFound = false;

   rotScaleCenter = makeVector3f(0.0f,0.0f,0.0f);
   moveRotScaleCent = false;
   lockRotScaleCent = false;

   showWorldCS = false;
   showPivotCS = 1; // 0 = no, 1 = just if mouse button pressed, 2 = yes
   CSWorldScale = 1.0f;
   CSPivotScale = 1.0f;
   TransfMode = 0; // modes: 0=move, 1=rotate, 2=scale

   mergeCommandObject = false;

}


void PCITransfEdit::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {

   mergeCommandObject = false;
   if (shiftDown || SteppedRotation || SteppedMovement) shiftDistance = 0.0f;

   if (buttonsState.getLeft()) {
      leftDown = true;
      lastX = x;
      lastY = y;
   } else if (buttonsState.getRight()) {
      rightDown = true;
      lastX = x;
      lastY = y;
   }

   if (leftDown || rightDown) {

      RefPointFound = false;

      if (moveRotScaleCent) {
         // move rotScaleCenter
         Vector3f tmp = rotScaleCenter;
         tmp -= (scene->getRootState()->staticState->camera).getPosition();
         RefPointZ = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection() * tmp;

      } else if (TransfMode==0) {
         // move clouds/points
         Vector3f PointPos;
         string CloudName;

         // get reference point which was clicked on from selected clouds or selected points
         if (transPoints) {
            if (!getClosestPointPositionFromSelectedPoints(x, y, width, height, 5, &(scene->getRootState()->staticState->camera), &(scene->getRootState()->staticState->viewFrustum), scene, PointPos, CloudName)) return;
         } else {
            if (!getClosestPointPositionFromSelectedClouds(x, y, width, height, 5, &(scene->getRootState()->staticState->camera), &(scene->getRootState()->staticState->viewFrustum), scene, PointPos, CloudName)) return;
         }

         PointCloud* PC = getPointCloud(scene, CloudName);
         if (PC==NULL) return;
         ReferencePoint = transformVector3f(PC->getTransformation(), PointPos);
         ReferencePoint -= (scene->getRootState()->staticState->camera).getPosition();
         RefPointZ = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection() * ReferencePoint;
         RefPointFound = true;

      }; // end else if moveRotScaleCent

   }; // end if left or right click

   forget(modifiersState);
}



void PCITransfEdit::mouseMoved(int32 x, int32 y) {
   if ((width <= 0) || (height <= 0)) return;

   if (leftDown || rightDown) {

      // if stepwise enabled, update mouse movement distance
      if (shiftDown || SteppedRotation || SteppedMovement) shiftDistance += ((x-lastX)*(x-lastX)+(y-lastY)*(y-lastY));

      float32 RightAtZ = 0;
      float32 UpAtZ = 0;
      float32 DirAtZ = 0;

      // move just if clicked on a point or move rotScaleCenter
      if ((TransfMode==0) || moveRotScaleCent) {

         float32 PixelSize = (tan((scene->getRootState()->staticState->viewFrustum).getVerticalFieldOfView()*0.5f*(M_PI/180.0f))) / (height*0.5f);
         float32 Click3D_X = (x-width*0.5f) * PixelSize;
         float32 Click3D_Y = (y-height*0.5f) * PixelSize;

         if (!RefPointFound && !moveRotScaleCent) return;

         if (leftDown) {
            if ((shiftDown || SteppedMovement)) {
               if (shiftDistance>shiftDistanceToggleValue) {
                  // if stepwise enabled and mouse moved far enough, move a fixed distance in x or y
                  shiftDistance = 0.0f;
                  RightAtZ = ((x-lastX)*PixelSize) / sqrt(1 + Click3D_X*Click3D_X);
                  UpAtZ = ((y-lastY)*PixelSize) / sqrt(1 + Click3D_Y*Click3D_Y);
                  RightAtZ *= -RefPointZ;
                  UpAtZ *= -RefPointZ;
                  int32 RS = 1;
                  int32 US = 1;
                  if (RightAtZ<0) RS = -1;
                  if (UpAtZ<0) US = -1;
                  if ((RightAtZ*RightAtZ)>(UpAtZ*UpAtZ)) { RightAtZ = RS*SteppedMovementValue; UpAtZ = 0.0f; }
                     else { UpAtZ = US*SteppedMovementValue; RightAtZ = 0.0f; };
               };
            } else {
               // calculate standard x-y-movement
               RightAtZ = ((x-lastX)*PixelSize) / sqrt(1 + Click3D_X*Click3D_X);
               UpAtZ = ((y-lastY)*PixelSize) / sqrt(1 + Click3D_Y*Click3D_Y);
               RightAtZ *= -RefPointZ;
               UpAtZ *= -RefPointZ;
            };
         }
         else if (rightDown) {
            if ((shiftDown || SteppedMovement)) {
               if (shiftDistance>shiftDistanceToggleValue) {
                  // if stepwise enabled and mouse moved far enough, move a fixed distance in z
                  shiftDistance = 0.0f;
                  DirAtZ = ((y-lastY)*PixelSize) / sqrt(1 + Click3D_Y*Click3D_Y);
                  DirAtZ *= -RefPointZ;
                  if (DirAtZ>0) DirAtZ = SteppedMovementValue;
                     else DirAtZ = -SteppedMovementValue;
               };
            } else {
               // calculate standard z-movement
               DirAtZ = ((y-lastY)*PixelSize) / sqrt(1 + Click3D_Y*Click3D_Y);
               DirAtZ *= -RefPointZ;
            };
         };

         if (moveRotScaleCent) {
            // move Pivot
            Matrix4f t = makeMoveMatrix(RightAtZ, UpAtZ, DirAtZ);
            rotScaleCenter = transformVector3f(t, rotScaleCenter);
            lastX = x;
            lastY = y;
            return;
         };

      // rotate and zoom, just dependent of mouse movement not of scene
      } else {

         if (leftDown) {
            if ((TransfMode==1) && (shiftDown || SteppedRotation)) {
               if (shiftDistance>shiftDistanceToggleValue) {
                  // if stepwise enabled and mouse moved far enough, rotate fixed angle around x- or y-axis
                  shiftDistance = 0.0f;
                  float32 DEG2RAD = 3.141592654f/180.0f;
                  int32 distX = (x-lastX);
                  int32 distY = (y-lastY);
                  if ((distX*distX)>(distY*distY)) { RightAtZ = SteppedRotationDegrees*DEG2RAD; UpAtZ = 0.0f; }
                     else { UpAtZ = SteppedRotationDegrees*DEG2RAD; RightAtZ = 0.0f; };
                  if (distX<0) RightAtZ = -RightAtZ;
                  if (distY<0) UpAtZ = -UpAtZ;
               };
            } else {
               // calculate standard x-y-rotation or scaling
               RightAtZ = (x-lastX)*0.01f;
               UpAtZ = (y-lastY)*0.01f;
            };
         }
         else if (rightDown) {
            if ((TransfMode==1) && (shiftDown || SteppedRotation)) {
               if (shiftDistance>shiftDistanceToggleValue) {
                  // if stepwise enabled and mouse moved far enough, rotate fixed angle around z-axis
                  shiftDistance = 0.0f;
                  float32 DEG2RAD = 3.141592654f/180.0f;
                  DirAtZ = SteppedRotationDegrees*DEG2RAD;
                  if (y-lastY<0) DirAtZ = -DirAtZ;
               };
            } else {
               // calculate standard z-rotation or scaling
               DirAtZ = (y-lastY)*0.01f;
            };
         };

      };

      // get appropriate transformation matrix to the x-y-z-values calculated above
      Matrix4f t;
      switch (TransfMode) {
         case 0: t = makeMoveMatrix(RightAtZ, UpAtZ, DirAtZ); if (lockRotScaleCent) rotScaleCenter = transformVector3f(t, rotScaleCenter); break;
         case 1: t = makeRotateMatrix(RightAtZ, UpAtZ, DirAtZ); break;
         case 2: t = makeScaleMatrix(-RightAtZ, -UpAtZ, -DirAtZ, true); break;
         default: t = IDENTITY4F;
      };

      // create command object and apply transformation by executing it
      CmdObjTransfEdit *newCmd = new CmdObjTransfEdit();
      newCmd->setup(t,transPoints,mergeCommandObject);
      scene->getCommandHistory()->executeCommand(scene, newCmd, true);
      mergeCommandObject = true;

   }; // end if leftDown||rightDown

   lastX = x;
   lastY = y;
}



void PCITransfEdit::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
   leftDown = false;
   rightDown = false;
   mergeCommandObject = false;
   if (shiftDown || SteppedRotation || SteppedMovement) shiftDistance = 0.0f;
   RefPointFound = false;
   forget(x+y);
   forget(buttonsState);
   forget(modifiersState);
}



void PCITransfEdit::areaResize(card32 width, card32 height) {
   this->width = width;
   this->height = height;
   if (width==0) { WidthToHeight=0.0f; return; };
   WidthToHeight = ((float)height)/((float)width);
}



void PCITransfEdit::keyDown(GeneralKey key) {

   if (key.getKey() == keyShift) {
      if (!shiftDown) shiftDistance = 0.0f;
      shiftDown = true;
      SteppedMovement = true;
      SteppedRotation = true;
      ObjectViewsTable::update(this);
   };

   if (key.getKey() == keyControl) ctrlDown = true;
}

void PCITransfEdit::keyUp(GeneralKey key) {

   if (key.getKey() == keyShift) {
      shiftDown = false;
      SteppedMovement = false;
      SteppedRotation = false;
      shiftDistance = 0.0f;
      ObjectViewsTable::update(this);
   };

   if (key.getKey() == keyControl) ctrlDown = false;
}


Matrix4f PCITransfEdit::makeMoveMatrix(float32 mX, float32 mY, float32 mZ) {
   Matrix4f t;
   if (!alignWorld) {
      // move in CameraCoords
      Vector3f r = (scene->getRootState()->staticState->camera).getOrthoNormRightDirection() * mX;
      Vector3f u = (scene->getRootState()->staticState->camera).getOrthoNormUpDirection() * mY;
      Vector3f v = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection() * mZ;
      t = makeTranslation4f(r+u+v);
   } else {
      // move in WorldCoords
      t = makeTranslation4f(mX, mZ, mY) ;
   };
   return t;
};


Matrix4f PCITransfEdit::makeRotateMatrix(float32 rX, float32 rY, float32 rZ) {
   Matrix4f t;
   if (!alignWorld) {
      // rotate in CameraCoords
      Vector3f u = (scene->getRootState()->staticState->camera).getOrthoNormRightDirection();
      Vector3f r = (scene->getRootState()->staticState->camera).getOrthoNormUpDirection();
      Vector3f v = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection();
      t = makeTranslation4f(rotScaleCenter) * makeRotVector4f(r, rX) * makeRotVector4f(u, rY) * makeRotVector4f(v, rZ) * makeTranslation4f(-rotScaleCenter);
   } else {
      // rotate in WorldCoords
      t = makeTranslation4f(rotScaleCenter) * makeRotX4f(rX) * makeRotY4f(rY) * makeRotZ4f(rZ) * makeTranslation4f(-rotScaleCenter);
   };
   return t;
};


Matrix4f PCITransfEdit::makeScaleMatrix(float32 sX, float32 sY, float32 sZ, bool e) {
   Matrix4f t;
   if (ctrlDown) {
      if (!alignWorld) {
         // non-uniform scale in CameraCoords
         Matrix3f CamSystem2;
         CamSystem2[0] = (scene->getRootState()->staticState->camera).getOrthoNormRightDirection();
         CamSystem2[1] = (scene->getRootState()->staticState->camera).getOrthoNormUpDirection();
         CamSystem2[2] = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection();
         CamSystem2 = CamSystem2.transpose();
         Matrix4f CamSystemT = expand3To4(CamSystem2);
         Matrix4f CamSystem = invertMatrix(CamSystemT);
         if (e) {
            sX = exp(-sX); sY = exp(sY); sZ = exp(sZ);
         };
         t = makeTranslation4f(rotScaleCenter) * CamSystem * makeScale4f(sX, sY, sZ) * CamSystemT * makeTranslation4f(-rotScaleCenter);
      } else {
         // non-uniform scale in WorldCoords
         if (e) {
            sX = exp(-sX); sY = exp(-sY); sZ = exp(-sZ);
         };
         t = makeTranslation4f(rotScaleCenter) * makeScale4f(sX, sY, sZ) * makeTranslation4f(-rotScaleCenter);
      }
   } else {
      // uniform scale
      float32 d;
      if (e) {
         sX = exp(sX); sY = exp(sY); sZ = exp(sZ);
      };
      if (rightDown) d = sZ;
                else d = sY;
      t = makeTranslation4f(rotScaleCenter) * makeScale4f(d,d,d) * makeTranslation4f(-rotScaleCenter);
   }
   return t;
};


void PCITransfEdit::centerRotScaleCentToSelClouds() {

   Vector3f mean = makeVector3f(0.0f, 0.0f, 0.0f);
   int32 count = 0;

   StandardSceneGraphIterator* sceneIterator = new StandardSceneGraphIterator();
   sceneIterator->clearAndSetup(scene->getRootNode(), scene->getRootState());

   while (!sceneIterator->atEnd()) {

      // selected clouds only
      if (!testPointCloudIsSelected(scene, sceneIterator->getNameWithPath())) { sceneIterator->next(); continue; }

      SGObjectNode* SGN = dynamic_cast<SGObjectNode*>(sceneIterator->getNode());
      if (SGN==NULL) { sceneIterator->next(); continue; }
      PointCloud* PC = dynamic_cast<PointCloud*>(SGN->getSceneObject());
      if (PC==NULL) { sceneIterator->next(); continue; }
      if (!PC->providesAttribute("position")) { sceneIterator->next(); continue; }
      AAT POSITION = PC->getAAT("position");

      SceneObjectIterator *newItSO = PC->createIterator(SOIT(SOIT::CAP_BASIC_PC));
      PointCloudIterator *newIt = dynamic_cast<PointCloudIterator*>(newItSO);
      if ((newIt==NULL)||(newItSO==NULL)) { delete newItSO; sceneIterator->next(); continue; };
      BasicPointCloudIterator *basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
      if (basicIt == NULL) {delete newItSO; sceneIterator->next(); continue;}

      Matrix4f curtransmatr = sceneIterator->getState()->dynamicState->ctm;

      while(!basicIt->atEnd()){
        //calculate mean Position
         Vector3f tmpPos = transformVector3f(curtransmatr, basicIt->get3f(POSITION));
         mean[0] += tmpPos[0];
         mean[1] += tmpPos[1];
         mean[2] += tmpPos[2];
         count++;
         basicIt->next();
      }

      delete newItSO;
      sceneIterator->next();
   }; // end while sceneIterator

   delete sceneIterator;

   if (count==0) return;
   rotScaleCenter = mean / count;

   redrawScene();
};


void PCITransfEdit::centerRotScaleCentToSelPoints() {

   Vector3f mean = makeVector3f(0.0f, 0.0f, 0.0f);
   int32 count = 0;

   StandardSceneGraphIterator* sceneIterator = new StandardSceneGraphIterator();
   sceneIterator->clearAndSetup(scene->getRootNode(), scene->getRootState());

   while (!sceneIterator->atEnd()) {

      SGObjectNode* SGN = dynamic_cast<SGObjectNode*>(sceneIterator->getNode());
      if (SGN==NULL) { sceneIterator->next(); continue; }
      PointCloud* PC = dynamic_cast<PointCloud*>(SGN->getSceneObject());
      if (PC==NULL) { sceneIterator->next(); continue; }
      if (!PC->providesAttribute("position")) { sceneIterator->next(); continue; }
      AAT POSITION = PC->getAAT("position");
      if (!PC->providesAttribute("flags")) { sceneIterator->next(); continue; }
      AAT FLAGS = PC->getAAT("flags");

      SceneObjectIterator *newItSO = PC->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_HIERARCHICAL));
      PointCloudIterator *newIt = dynamic_cast<PointCloudIterator*>(newItSO);
      if ((newIt==NULL)||(newItSO==NULL)) { delete newItSO; sceneIterator->next(); continue; };

      // selected points only
      IteratorAdaptor *selIt = getSelectionIteratorAdaptor(PC, newIt);
      if (selIt == NULL) { delete newItSO; sceneIterator->next(); continue; };

      Matrix4f curtransmatr = sceneIterator->getState()->dynamicState->ctm;

      while(!selIt->atEnd()){
        //calculate mean Position
         Vector3f tmpPos = transformVector3f(curtransmatr, selIt->get3f(POSITION));
         mean[0] += tmpPos[0];
         mean[1] += tmpPos[1];
         mean[2] += tmpPos[2];
         count++;
         selIt->next();
      }

      delete selIt;
      delete newItSO;
      sceneIterator->next();
   }; // end while sceneIterator

   delete sceneIterator;

   if (count==0) return;
   rotScaleCenter = mean / count;

   redrawScene();
};


void PCITransfEdit::executeXYZ(float32 eX, float32 eY, float32 eZ) {

   if (TransfMode==1) {
      // for rotation: convert values to degrees
      float32 DEG2RAD = 3.141592654f/180.0f;
      eX *= DEG2RAD;
      eY *= DEG2RAD;
      eZ *= DEG2RAD;
   };

   if (moveRotScaleCent) {
      // Pivot can't be rotated or scaled
      card32 oldTM = TransfMode;
      TransfMode=0;
      // move Pivot
      Matrix4f t = makeMoveMatrix(eX, eY, eZ);
      rotScaleCenter = transformVector3f(t, rotScaleCenter);
      TransfMode = oldTM;

   } else {
      // to allow non-uniform scaling
      bool ctrlPressed2 = ctrlDown;
      ctrlDown = true;

      // get appropriate transformation matrix to the x-y-z-values calculated above
      Matrix4f t;
      switch (TransfMode) {
         case 0: t = makeMoveMatrix(eX, eY, eZ); break;
         case 1: t = makeRotateMatrix(eX, eY, eZ); break;
         case 2: t = makeScaleMatrix(eX, eY, eZ, false); break;
         default: t = IDENTITY4F;
      };

      ctrlDown = ctrlPressed2;

      // create command object and apply transformation by executing it
      mergeCommandObject = false;
      CmdObjTransfEdit *newCmd = new CmdObjTransfEdit();
      newCmd->setup(t,transPoints,mergeCommandObject);
      scene->getCommandHistory()->executeCommand(scene, newCmd, true);

   };


   redrawScene();
};


void PCITransfEdit::glDrawTool(GLContext *glContext) {
   forget(glContext);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
      glMultMatrixf(scene->getRootState()->dynamicState->ctm.data());

      glPushAttrib(GL_ALL_ATTRIB_BITS);
         glDisable(GL_LIGHTING);

            // allways display pivot point
            glPointSize(3.0f);
            glColor4f(0.5f,0.1f,1.0f,1.0f);
            glBegin(GL_POINTS);
               glVertex3fv(rotScaleCenter.data());
            glEnd();

            glLineWidth(1.5f);
            // display pivot coordinate system
            if ((showPivotCS==2) || ((showPivotCS==1) && (leftDown || rightDown))) {
               if (alignWorld) {
                  glBegin(GL_LINES);
                     glVertex3fv((rotScaleCenter-makeVector3f(CSPivotScale*0.5f, 0.0f, 0.0f)).data());
                     glVertex3fv((rotScaleCenter+makeVector3f(CSPivotScale, 0.0f, 0.0f)).data());
                     glVertex3fv((rotScaleCenter-makeVector3f(0.0f, CSPivotScale*0.5f, 0.0f)).data());
                     glVertex3fv((rotScaleCenter+makeVector3f(0.0f, CSPivotScale, 0.0f)).data());
                     glVertex3fv((rotScaleCenter-makeVector3f(0.0f, 0.0f, CSPivotScale*0.5f)).data());
                     glVertex3fv((rotScaleCenter+makeVector3f(0.0f, 0.0f, CSPivotScale)).data());
                  glEnd();
               } else {
                  glBegin(GL_LINES);
                     glVertex3fv((rotScaleCenter-(scene->getRootState()->staticState->camera).getOrthoNormRightDirection()*CSPivotScale*0.5f).data());
                     glVertex3fv((rotScaleCenter+(scene->getRootState()->staticState->camera).getOrthoNormRightDirection()*CSPivotScale).data());
                     glVertex3fv((rotScaleCenter-(scene->getRootState()->staticState->camera).getOrthoNormUpDirection()*CSPivotScale*0.5f).data());
                     glVertex3fv((rotScaleCenter+(scene->getRootState()->staticState->camera).getOrthoNormUpDirection()*CSPivotScale).data());
                     glVertex3fv((rotScaleCenter-(scene->getRootState()->staticState->camera).getOrthoNormViewDirection()*CSPivotScale*0.5f).data());
                     glVertex3fv((rotScaleCenter+(scene->getRootState()->staticState->camera).getOrthoNormViewDirection()*CSPivotScale).data());
                  glEnd();
               }
            };

            // display world coordinate system
            if (showWorldCS) {
               glPointSize(3.0f);
               glColor4f(0.0f, 1.0f, 0.3f, 1.0f);
               glBegin(GL_POINTS);
                  glVertex3f(0.0f, 0.0f, 0.0f);
               glEnd();
               glBegin(GL_LINES);
                  glVertex3f(-CSWorldScale*0.5f, 0.0f, 0.0f);
                  glVertex3f(CSWorldScale, 0.0f, 0.0f);
                  glVertex3f(0.0f, -CSWorldScale*0.5f, 0.0f);
                  glVertex3f(0.0f, CSWorldScale, 0.0f);
                  glVertex3f(0.0f, 0.0f, -CSWorldScale*0.5f);
                  glVertex3f(0.0f, 0.0f, CSWorldScale);
               glEnd();
            };

      glPopAttrib();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

};
