//---------------------------------------------------------------------------
#ifndef PCITransfEditH
#define PCITransfEditH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PCInteractionTool.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl {


class EDITOR_API PCITransfEdit : public PCInteractionTool {
 private:
   static MetaClass *PCITransfEditMetaClass;
   static Object* newInstance();

   // properties
   /// center point (Pivot) used for rotation/scaling
   Vector3f rotScaleCenter;
   /// if true, the Pivot is moved instead of clouds or points
   bool moveRotScaleCent;
   /// mode set in the custom editor: 0=move, 1=rotate, 2=scale
   card32 TransfMode;
   /// if true, world axes are used. if false camera axes are used.
   bool alignWorld;
   /// if true then points are affected by transformation, else clouds
   bool transPoints;
   /// if true then the Pivot is moved together with clouds or points
   bool lockRotScaleCent;
   /// if true, a world coordinate system is displayed in scene-window
   bool showWorldCS;
   /// used to display the actual coordinate system in the Pivot point. 0=don't show, 1=only show if mouse button pressed, 2=show allways
   card32 showPivotCS;
   /// scaling-factor for display of the world coordinate system. Defines the size of the lines.
   float32 CSWorldScale;
   /// scaling-factor for display of the Pivot coordinate system. Defines the size of the lines.
   float32 CSPivotScale;
   /// if true, movement will apply only in steps of defined size.
   bool SteppedMovement;
   /// defines the size of one movement step
   float32 SteppedMovementValue;
   /// if true, rotation will apply only in steps of defined size.
   bool SteppedRotation;
   /// defines the size of one rotation step in degrees
   float32 SteppedRotationDegrees;

   // internal tmp
   /// temp var for last x and y value
   int32 lastX, lastY;
   /// temp var: true while left mouse button pressed
   bool leftDown;
   /// temp var: true while right mouse button pressed
   bool rightDown;
   /// temp var: width and height of window
   int32 width, height;
   /// temp var: factor of width to height
   float32 WidthToHeight;
   /// temp var: true while ctrl key is pressed
   bool ctrlDown;
   /// temp var: true while shift key is pressed
   bool shiftDown;
   /// temp var: defines if the following command object may be merged with the actual one. Used to avoid multiple command objects if you move the mouse.
   bool mergeCommandObject;
   /// temp var: if stepwise transformation (SteppedMovement/SteppedRotation) is activated, the distance the mouse cursor has travelled is saved here. Used for better mouse interaction purpose.
   float32 shiftDistance;
   /// temp var: defines the distance the mouse cursor has to travel until next step in stepwise mode. Says how far you must move the mouse in stepwise mode before next step is performed. Hardcoded in the constructor at the moment.
   float32 shiftDistanceToggleValue;
   /// temp var: reference point of mouse click
   Vector3f ReferencePoint;
   /// temp var: reference point of mouse click
   bool RefPointFound;
   /// temp var: reference point of mouse click
   float32 RefPointZ;

   /// calculates a movement matrix of the three given directions
   Matrix4f makeMoveMatrix(float32 mX, float32 mY, float32 mZ);
   /// calculates a rotation matrix of the three given angles
   Matrix4f makeRotateMatrix(float32 rX, float32 rY, float32 rZ);
   /// calculates a scaling matrix of the three given factors
   Matrix4f makeScaleMatrix(float32 sX, float32 sY, float32 sZ, bool e);


 public:
   PCITransfEdit();

   inline void setMode(card32 m) { TransfMode = m; };
   inline card32 getMode() { return TransfMode; };
   inline void setAlignWorld(bool aw) { alignWorld = aw; };
   inline bool getAlignWorld() { return alignWorld; };
   inline void setTransPoints(bool tp) { transPoints = tp; };
   inline bool getTransPoints() { return transPoints; };
   inline void setMoveRotScaleCent(bool mrsc) { moveRotScaleCent = mrsc; };
   inline bool getMoveRotScaleCent() { return moveRotScaleCent; };
   inline void setLockRotScaleCent(bool lrsc) { lockRotScaleCent = lrsc; };
   inline bool getLockRotScaleCent() { return lockRotScaleCent; };
   inline void setSteppedRotation(bool step) { SteppedRotation = step; shiftDistance = 0.0f; };
   inline bool getSteppedRotation() { return SteppedRotation; };
   inline void setSteppedRotationDegrees(float deg) { SteppedRotationDegrees = deg; };
   inline float32 getSteppedRotationDegrees() { return SteppedRotationDegrees; };
   inline void setSteppedMovement(bool step) { SteppedMovement = step; shiftDistance = 0.0f; };
   inline bool getSteppedMovement() { return SteppedMovement; };
   inline void setSteppedMovementValue(float val) { SteppedMovementValue = val; };
   inline float32 getSteppedMovementValue() { return SteppedMovementValue; };

   inline void setShowWorldCoordinateSystem(bool wcs) { showWorldCS = wcs; };
   inline bool getShowWorldCoordinateSystem() { return showWorldCS; };
   inline void setShowPivotCoordinateSystem(card32 pcs) { showPivotCS = pcs; };
   inline card32 getShowPivotCoordinateSystem() { return showPivotCS; };

   inline void setCSWorldScale(float32 ws) { CSWorldScale = ws; };
   inline float32 getCSWorldScale() { return CSWorldScale; };
   inline void setCSPivotScale(float32 ps) { CSPivotScale = ps; };
   inline float32 getCSPivotScale() { return CSPivotScale; };

   /// centers the Pivot in the middle of selected clouds
   void centerRotScaleCentToSelClouds();
   /// centers the Pivot in the middle of selected points
   void centerRotScaleCentToSelPoints();
   /// used for transformation if exact values are entered in custom editor
   void executeXYZ(float32 eX, float32 eY, float32 eZ);

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init( const MetaClass *parentClass );
   static void shutdown();

   virtual void mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void mouseMoved(int32 x, int32 y);
   virtual void mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void areaResize(card32 width, card32 height);

   virtual void keyDown(GeneralKey key);
   virtual void keyUp(GeneralKey key);

   virtual void glDrawTool(GLContext *glContext);

};


}

#endif
