//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "ProgressWindow.h"
#include "Exceptions.h"
#include "DebugOutput.h"
#include "InterfaceGlobals.h"
//---------------------------------------------------------------------------
#include "ProgressWindowWidget.h"
//---------------------------------------------------------------------------

using namespace geoxl;

ProgressWindow* geoxl::progressWindow = NULL;
MetaClass* ProgressWindow::ProgressWindowMetaClass = NULL;

const MetaClass* ProgressWindow::getInstanceClass() const {
  return getClass();
}

Object* ProgressWindow::newInstance() {
  return new ProgressWindow;
}

const MetaClass* ProgressWindow::getClass() {
  if (ProgressWindowMetaClass == NULL) {
    ProgressWindowMetaClass = new MetaClass("ProgressWindow", "no description.",
                                    newInstance, sizeof(ProgressWindow));
  }
  return ProgressWindowMetaClass;
}

void ProgressWindow::init(const MetaClass *parentClass) {
  ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION);
  ProgressWindowMetaClass->copyParentProperties();
}

void ProgressWindow::shutdown() {
  delete ProgressWindowMetaClass;
}

// ----

/****************************************************************************
 * ProgressWindow - constructor
 ****************************************************************************/
ProgressWindow::ProgressWindow()
{
//  ProgressWindowForm = new TProgressWindowForm(NULL);
	progressWindowWidget= new ProgressWindowWidget(InterfaceGlobals::getMainWindow());
	
}

ProgressWindow::~ProgressWindow()
{
//  delete ProgressWindowForm;
//	delete progressWindowWidget;
}

/****************************************************************************/
//  ProgressWindow - pushSteps
//  Pushes a list of steps onto the operations stack
/// If the stack was empty, the progress window will be displayed. If not,
/// the step list will be pushed onto the stack and the total progress for
/// this part of the stack will be the weight of the active step in the level
/// below.
/****************************************************************************/
void ProgressWindow::pushSteps(WeightList steps)
{
  if (steps.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: Pushed empty step list");
    #endif
    return;
  }

  if (weightStack.empty()) {
    // First call to the progress window; initialize it for first display
    startTime = timer.getCurrentValue();
    lastProgressTime = 0;
    progressWindowWidget->setOverallProgress(0);
    progressWindowWidget->setToolMessage("");
    if (steps.size() == 1 && !steps.begin()->progressAvailable)
      overallProgressAvailable = false;
    else
      overallProgressAvailable = true;
    progressWindowWidget->setOverallProgressAvailable(overallProgressAvailable);

    isVisible = false;
    isAborted = false;
    while (!percentageDoneStack.empty()) percentageDoneStack.pop();

    // Clear keyboard buffer for ESC and "q" key
    while ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1);
    while ((GetAsyncKeyState(0x51) & 1) == 1);
  }

  // Normalize weights
  float32 totalWeight = 0.0f;
  for (WeightList::iterator it = steps.begin(); it != steps.end(); ++it)
    totalWeight += it->weight;
  if (!weightStack.empty())
    totalWeight /= weightStack.top().begin()->weight;
  for (WeightList::iterator it = steps.begin(); it != steps.end(); ++it)
    it->weight /= totalWeight;

  weightStack.push(steps);

  if (!updateIntervalStack.empty())
    updateIntervalStack.push(updateIntervalStack.top());
  else
    updateIntervalStack.push(90);

  if (percentageDoneStack.empty()) percentageDoneStack.push(0.0f);
    else percentageDoneStack.push(percentageDoneStack.top());

  initForm();
}

/****************************************************************************/
//  ProgressWindow - pushStep
//  Pushes a single step onto the operations stack
/// Works just like pushSteps, but instead of pushing a list of steps onto
/// the stack, it just pushes a single step (for convenience).
/****************************************************************************/
void ProgressWindow::pushStep(bool progressAvailable, std::string stepName)
{
  WeightElem elem;
  elem.stepName = stepName;
  elem.weight = 1.0f;
  elem.progressAvailable = progressAvailable;

  WeightList wList;
  wList.push_back(elem);
  pushSteps(wList);
}

/****************************************************************************/
//  ProgressWindow - nextStep
//  Moves to the next step in the list
/// For the overall progress, it is assumed that the step before was
/// completed, no matter what the last progress value was. If the whole
/// process is done, nextStep can not be called. Use popSteps instead.
/// \param aborted Contains true if the progress was aborted at some point.
///                In that case, the tool should not continue with what it's
///                doing.
/****************************************************************************/
void ProgressWindow::nextStep(bool& aborted)
{
  aborted = false;

  if (isAborted) {
    aborted = true;
    return;
  }

  if (weightStack.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: nextStep() called with empty stack");
    #endif
    return;
  }
  if (weightStack.top().empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: nextStep() called with empty step list");
    #endif
    return;
  }

  progressWindowWidget->setToolProgress(0);
  if (overallProgressAvailable) {
    percentageDoneStack.top() += weightStack.top().begin()->weight * 100.0f;
  }

  weightStack.top().pop_front();
  #ifdef _DEBUG
  if (weightStack.top().empty()) {
    warning("ProgressWindow: nextStep() called for last step in list. Use popSteps() instead.");
    return;
  }
  #endif

  initForm();

  if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < updateIntervalStack.top()) return;
  lastProgressTime = timer.getCurrentValue();
  progressWindowWidget->drawUpdates();
}

/****************************************************************************/
//  ProgressWindow - popSteps
//  Pops the step list on the top of the stack.
/// The step list on the top of the stack will be popped and the algorithm
/// returns to the next higher level in the progress stack. If the last
/// level is popped, the progress window will be closed.
/// \param aborted Contains true if the progress was aborted at some point.
///                In that case, the tool should not continue with what it's
///                doing.
/****************************************************************************/
void ProgressWindow::popSteps(bool& aborted)
{
  aborted = false;

  if (isAborted) {
    aborted = true;
    return;
  }

  if (weightStack.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: Trying to pop empty stack");
    #endif
    return;
  }

  while (!weightStack.top().empty()) {
    weightStack.top().pop_front();
  }
  weightStack.pop();
  percentageDoneStack.pop();
  updateIntervalStack.pop();

  if (!weightStack.empty()) {
    initForm();
    if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < updateIntervalStack.top()) return;
    lastProgressTime = timer.getCurrentValue();
    progressWindowWidget->drawUpdates();
  } else {
    progressWindowWidget->closeForm();
  }
}

/****************************************************************************/
//  ProgressWindow - initForm [PRIVATE]
//  Updates the form of the progress window
/// The form of the progress window will be updated with the values from the
/// top of the stack, and everything will be initialized for the next step.
/****************************************************************************/
void ProgressWindow::initForm()
{
  WeightElem* firstStep = &(*weightStack.top().begin());

  toolProgressAvailable = firstStep->progressAvailable;
  progressWindowWidget->setToolName(firstStep->stepName);
  progressWindowWidget->setToolProgressAvailable(toolProgressAvailable);
  progressWindowWidget->setToolProgress(0);
  progressWindowWidget->setOverallProgressAvailable(overallProgressAvailable);
  if( overallProgressAvailable )
	progressWindowWidget->setOverallProgress( percentageDoneStack.top() );

}

/****************************************************************************/
//  ProgressWindow - showMessage
//  Displays the tool message given
/****************************************************************************/
void ProgressWindow::showMessage(std::string msg)
{
  progressWindowWidget->setToolMessage(msg);
  progressWindowWidget->drawUpdates();
	
}

/****************************************************************************/
//  ProgressWindow - progress
//  Updates the progress window and returns false if ESC was pressed.
/// This method has to be called frequently during the processing in order
/// to update the progress form and check whether the processing was
/// cancelled. If the user requested to cancel the process, false will be
/// returned, otherwise true.
/****************************************************************************/
bool ProgressWindow::progress(card32 position)
{
  if (isAborted) return false;
  if (weightStack.empty()) return true;

  if (!isVisible) {
    // Show window if it isn't visible already
    if (timer.getDeltaValue(startTime) > 1000) {
      isVisible = true;
      progressWindowWidget->showForm();
    }
  }

  if (isVisible) {

    // Dont update progress window if last call was less then <updateInterval> ms ago
    if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < updateIntervalStack.top()) return true;
    lastProgressTime = timer.getCurrentValue();

    if (toolProgressAvailable) {
      progressWindowWidget->setToolProgress(position);
      if (overallProgressAvailable)
        progressWindowWidget->setOverallProgress(percentageDoneStack.top() + weightStack.top().begin()->weight * float32(position));
    }

	 progressWindowWidget->drawUpdates();

    if ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1 && (GetAsyncKeyState(0x51) & 1) == 1) {
      bool dummy;
      while (!weightStack.empty()) popSteps(dummy);
      isAborted = true;
      return false;
    } else {
      return true;
    }
  } else return true;
	return true;
}

/****************************************************************************/
//  ProgressWindow - setUpdateInterval
//  Sets the minimum interval between GUI-updates for this level in the stack
/// This method sets the update interval for the active level in the stack
/// (and all sublevels, if they don't call it again).
/****************************************************************************/
void ProgressWindow::setUpdateInterval(card32 value)
{
  if (updateIntervalStack.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: setUpdateInterval() called with empty stack");
    #endif
    return;
  }

  updateIntervalStack.pop();
  updateIntervalStack.push(value);
}
