#include "StdAfx.h"
#include "ScriptEngineDebugWidget.h"
#include "Timer.h"
#include "DebugOutput.h"
#include "Configuration.h"
#include "ExceptionHandling.h"
#include "ProgressWindow.h"
#include "ObjectBrowserWidget.h"
#include "ScriptEngine.h"
#include "StringHelper.h"
#include "ScriptDictionaryAdapter.h"
//---------------------------------------------------------------------------

using namespace geoxl;

/*ScriptEngineDebugWidget *geoxl::scriptEngineDebugWidget = NULL;*/


ScriptEngineDebugWidget::ScriptEngineDebugWidget(ScriptEngine *script)
:QWidget()
{
	//call ui.setupUi(this) to set up the layout from QtDesigner
	ui.setupUi(this);
	// used Qt auto-connect feature for slots

	this->script = script;
	this->scrDicAdapter = NULL;
	
	this->setAttribute(Qt::WA_DeleteOnClose);
	
	QWidget::connect(ui.RBNumber, SIGNAL(toggled(bool)), this, SLOT(handleVarTypeToggle(bool)));
	QWidget::connect(ui.RBString, SIGNAL(toggled(bool)), this, SLOT(handleVarTypeToggle(bool)));
	QWidget::connect(ui.RBObject, SIGNAL(toggled(bool)), this, SLOT(handleVarTypeToggle(bool)));
}

void ScriptEngineDebugWidget::update() {
   // TODO?
}

ScriptEngineDebugWidget::~ScriptEngineDebugWidget(void)
{
   if (scrDicAdapter != NULL) {
      scrDicAdapter->done();
      delete scrDicAdapter;
   }
}

void ScriptEngineDebugWidget::start()
{
   QString textContent;
   for (size_t i = 0; i < script->code.size(); ++i) {
      textContent += script->code[i].c_str();
      textContent += '\n';
   } 

   ui.CodeEdit->setPlainText(textContent);

   ui.CBReadOnly->setChecked(true);
   ui.CodeEdit->setReadOnly(true);
   ui.BTEditContinue->setEnabled(false);

   breakpoints.clear();

   script->debugStart();
   progressWindow->pushStep(false, "GEOXLscript Debugger");
   updateCodeMarkers(true);

   this->show();
}

void ScriptEngineDebugWidget::registerScrDicAdapter(ScriptDictionaryAdapter* adapter)
{
   this->scrDicAdapter = adapter;
}


void ScriptEngineDebugWidget::handleVarTypeToggle(bool b)
{
  ui.BTObjDetail->setEnabled(false);
  ui.BTObjDetail->setVisible(ui.RBObject->isChecked());
  ui.LEVarContent->setVisible(!ui.RBObject->isChecked());
  ui.BTModify->setVisible(!ui.RBObject->isChecked());
  debugObject = NULL;
}

void ScriptEngineDebugWidget::on_LEVarName_returnPressed()
{
   string varName = ui.LEVarName->text().toLatin1().constData();

   if (ui.RBNumber->isChecked()) {
      ExpressionNum num;
      if (!script->getNumTables()->getValue(varName, num)) {
         QMessageBox::critical(this, "Invalid variable", ui.LEVarName->text() + " is not a valid number variable.");
         return;
      }
      ui.LEVarContent->setText(numToStr(num).c_str());
      return;
   }

   if (ui.RBString->isChecked()) {
      string s;
      if (!script->getStringTables()->getValue(varName, s)) {
         QMessageBox::critical(this, "Invalid variable", ui.LEVarName->text() + " is not a valid string variable.");
         return;
      }
      ui.LEVarContent->setText(s.c_str());
      return;
   }

   if (ui.RBObject->isChecked()) {
      Object* obj;
      if (!script->getObjectTables()->getValue(varName, obj)) {
         QMessageBox::critical(this, "Invalid variable", ui.LEVarName->text() + " is not a valid object variable.");
         obj = NULL;
      }

      if (obj != NULL) {
         ui.BTObjDetail->setEnabled(true);
         string name = obj->getInstanceClass()->getClassName() + "...";
         ui.BTObjDetail->setText(name.c_str());
         debugObject = obj;
      } else {
         ui.BTObjDetail->setEnabled(false);
         ui.BTObjDetail->setText("NULL");
         debugObject = NULL;
      }
      return;
   }
}

void ScriptEngineDebugWidget::on_BTModify_clicked()
{
   string varName = ui.LEVarName->text().toLatin1().constData();

   if (ui.RBNumber->isChecked()) {
      ExpressionNum num;
      if (!strToNum(ui.LEVarContent->text().toLatin1().constData(), num)) {
         QMessageBox::critical(this, "Invalid format", "'" + ui.LEVarContent->text() + "' is not a valid number.");
         return;
      }

      if (!script->getNumTables()->setValue(varName, num)) {
         QMessageBox::critical(this, "Error", "Could not set " + ui.LEVarName->text() + " to '" + ui.LEVarContent->text() + "'.");
         return;
      }
      return;
   }

   if (ui.RBString->isChecked()) {
      string s = ui.LEVarContent->text().toLatin1().constData();
      if (!script->getStringTables()->setValue(varName, s)) {
         QMessageBox::critical(this, "Error", "Could not set " + ui.LEVarName->text() + " to '" + ui.LEVarContent->text() + "'.");
         return;
      }
      return;
   }

   if (ui.RBObject->isChecked()) {
      QMessageBox::information(this, "Not necessary",
         "Changing values in the object editor immediately modifies it.\n"
         "Clicking Modify is not necessary.");
      return;
   }
}

void ScriptEngineDebugWidget::on_BTObjDetail_clicked()
{
   if (debugObject == NULL) {
      ui.BTObjDetail->setEnabled(false);
      return;
   }

	const GUIDescriptor * gui = debugObject->getInstanceClass()->getGUIDescriptor();
	ClassEditor * ce = (ClassEditor*)gui->getDefaultClassEditor()->newInstance();
   ObjectBrowserWidget *newBrowser = new ObjectBrowserWidget(debugObject, ce, this);
   QWidget * w = ce->makeEditor( newBrowser, debugObject );
   newBrowser->setupEditor(w);
   newBrowser->show();
   ObjectViewsTable::bindView( debugObject, ce );
}

void ScriptEngineDebugWidget::updateCodeMarkers(bool scroll)
{
   QList<QTextEdit::ExtraSelection> sels;

   // Color current execution line
   QTextEdit::ExtraSelection* sel = new QTextEdit::ExtraSelection();
   sel->cursor = QTextCursor(ui.CodeEdit->document());
   sel->cursor.movePosition(QTextCursor::Down, QTextCursor::MoveAnchor, script->getLine());
   sel->cursor.select(QTextCursor::LineUnderCursor);

   sel->format = ui.CodeEdit->currentCharFormat();
   sel->format.setBackground(QBrush(QColor(0,0,128)));
   sel->format.setForeground(QBrush(QColor(255,255,255)));
   sel->format.setFontWeight(QFont::Bold);
   sels.push_back(*sel);
   delete sel;
   
   // Color breakpoints
   for (size_t i = 0; i < breakpoints.size(); ++i) {
      sel = new QTextEdit::ExtraSelection();
      sel->cursor = QTextCursor(ui.CodeEdit->document());
      sel->cursor.movePosition(QTextCursor::Down, QTextCursor::MoveAnchor, breakpoints[i]);
      sel->cursor.select(QTextCursor::LineUnderCursor);

      sel->format = ui.CodeEdit->currentCharFormat();
      sel->format.setBackground(QBrush(QColor(255,0,0)));
      sel->format.setForeground(QBrush(QColor(255,255,255)));
      sel->format.setFontWeight(QFont::Bold);
      sels.push_back(*sel);
   }

   ui.CodeEdit->setExtraSelections(sels);
}

void ScriptEngineDebugWidget::on_BTSingleStep_clicked()
{
   if (!progressWindow->progress(0)) {
      on_BTExit_clicked();
      return;
   }

   do {
      try {
         if (!script->debugNextCommand()) {
            QMessageBox::information(this, "Information", "Script is done.");
            this->close();
            bool dummy;
            progressWindow->popSteps(dummy);
            return;
         }
      } catch (EParserOperandException& e) {
         QMessageBox::critical(this, "Operand error", e.getMsg().c_str());
         break;
      } catch (EParserException& e) {
         QMessageBox::critical(this, "Parser error", e.getMsg().c_str());
         break;
      } catch (EPropertyParserException& e) {
         QMessageBox::critical(this, "Property parser error", e.getMsg().c_str());
         break;
      }
   } while (!script->isActive());
   updateCodeMarkers(true);
   this->raise();
}

void ScriptEngineDebugWidget::on_BTRun_clicked()
{
   hide();
   bool doExecute = true;
   while (doExecute) {
      if (!progressWindow->progress(0)) {
         doExecute = false;
         progressWindow->pushStep(false, "GEOXLscript Debugger");
         updateCodeMarkers(true);
      } else {
         try {
            if (!script->debugNextCommand()) {
               QMessageBox::information(this, "Information", "Script is done.");
               close();
               bool dummy;
               progressWindow->popSteps(dummy);
               return;
            } else {
               if (script->isActive() && std::find(breakpoints.begin(), breakpoints.end(), script->getLine()) != breakpoints.end()) {
                  doExecute = false;
                  updateCodeMarkers(true);
               }
            }
         } catch (EParserOperandException& e) {
            QMessageBox::critical(this, "Operand error", e.getMsg().c_str());
            doExecute = false;
            updateCodeMarkers(true);
            break;
         } catch (EParserException& e) {
            QMessageBox::critical(this, "Parser error", e.getMsg().c_str());
            doExecute = false;
            updateCodeMarkers(true);
            break;
         } catch (EPropertyParserException& e) {
            QMessageBox::critical(this, "Property parser error", e.getMsg().c_str());
            doExecute = false;
            updateCodeMarkers(true);
            break;
         }
      }
   }

   show();
}


void ScriptEngineDebugWidget::on_BTBreakpoint_clicked()
{
   int pos = ui.CodeEdit->textCursor().blockNumber();
   vector<int>::iterator it = std::find(breakpoints.begin(), breakpoints.end(), pos);
   if (it != breakpoints.end()) {
      breakpoints.erase(it);
   } else {
      breakpoints.push_back(pos);
   }
   updateCodeMarkers(false);
}

void ScriptEngineDebugWidget::on_BTEditContinue_clicked()
{
   bool changed = false;
   if (ui.CodeEdit->document()->isModified()) {
      script->code.clear();

      string code = ui.CodeEdit->document()->toPlainText().toLatin1().constData();
      string token;
      do {
         token = extractToChar(code, '\n');
         if (token != "") script->code.push_back(token);
      } while (token != "");
      
      changed = true;
      ui.CodeEdit->document()->setModified(false);
  }

  int lineNumber = ui.CodeEdit->textCursor().blockNumber();
  if (lineNumber != (int)script->getLine()) {
     changed = true;
     if (QMessageBox::question(this, "Edit and Continue", "Update code position as well?",
                               QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes) {
        script->setLine(lineNumber);
        updateCodeMarkers(false);
     }
   }

   if (!changed)
      QMessageBox::information(this, "Information", "Nothing changed.");
}

void ScriptEngineDebugWidget::on_BTExit_clicked()
{
  bool dummy;
  progressWindow->popSteps(dummy);
  script->debugEnd();
  close();
}

void ScriptEngineDebugWidget::on_CBReadOnly_clicked()
{
   if (ui.CBReadOnly->checkState() == Qt::Checked) {
      if (ui.CodeEdit->document()->isModified()) {
         // Changed to read-only -> ask to undo changes
         if (QMessageBox::question(this, "Undo changes", "This will dismiss all changes to the code. Do you want to continue?",
                                   QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes) {
            QString textContent;
            for (size_t i = 0; i < script->code.size(); ++i) {
               textContent += script->code[i].c_str();
               textContent += '\n';
            } 

            ui.CodeEdit->setPlainText(textContent);
            updateCodeMarkers(true);
         } else return;
      }
   } else ui.CodeEdit->document()->setModified(false);
   
   ui.CodeEdit->setReadOnly(ui.CBReadOnly->checkState() == Qt::Checked);
   ui.BTEditContinue->setEnabled(ui.CBReadOnly->checkState() != Qt::Checked);
}

#include "ScriptEngineDebugWidget_moc.cpp"