//---------------------------------------------------------------------------
#include "KMeansClustering.h"
#include "DynamicLinearAlgebra_templCode.h"
#include "DynamicLinearAlgebraTools_templCode.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

template<typename VectorType>
void nullVectorGeneric(unsigned dim, VectorType &result) {
   result = VectorType(dim);
   for (unsigned i=0; i<dim; i++) {
      result[i] = 0;
   }
}

template<typename FloatType, unsigned dim>
void nullVectorGeneric(unsigned d, StaticVector<FloatType, dim> &result) {
   for (unsigned i=0; i<dim; i++) {
      result[i] = 0;
   }
}

template <typename FloatType, typename VectorType>
void calcBoundingBoxGeneric(const vector< VectorType > &data, VectorType &lower, VectorType &upper) {
	size_t n = data.size();
	pAssert(n > 0);
	if (n == 0) return;
	lower = data[0];
	upper = data[0];
	card32 dim = data[0].getDim();
	for (unsigned i=1; i<n; i++) {
		pAssert(data[i].getDim() == dim);
		for (card32 d=0; d<dim; d++) {
			if (data[i][d] < lower[d]) lower[d] = data[i][d];
			if (data[i][d] > upper[d]) upper[d] = data[i][d];
		}
	}   
}

template <typename VectorType>
void randomDVectorGeneric(VectorType &v, card32 dim) {
	v.setDim(dim);
	for (card32 i=0; i<dim; i++) {
		v[i] = rnd01();
	}
}

template <typename FloatType, unsigned dim>
void randomDVectorGeneric(StaticVector<FloatType, dim> &v, card32 dim) {
	for (card32 i=0; i<dim; i++) {
		v[i] = rnd01();
	}
}

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::add(const VectorType &v, float weight) {
	data.push_back(v);
	weights.push_back(weight);
}

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::setup(bool initWithData=false) {
	this->initWithData = initWithData;
}

const card32 MAX_NUM_ITERATIONS = 100;

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::chooseNearest(
	vector<unsigned> &indices, 
	const vector< VectorType > &data, 
	const vector< VectorType > &result)
{
	size_t n = data.size();
	indices.resize(n);
	for (unsigned ind = 0; ind<n; ind++) {
		VectorType pos = data[ind];
		FloatType diff = norm(pos - result[0]);
		indices[ind] = 0;
		const size_t numClusters = result.size();
		for (unsigned c=1; c<numClusters; c++) {
			FloatType newDiff = norm(pos - result[c]);
			if (diff > newDiff) {
				diff = newDiff;
				indices[ind] = c;
			}
		}
	}
}

template <typename FloatType, typename VectorType>
FloatType KMeansClusteringTempl<FloatType, VectorType>::updateResult(
	const vector<unsigned> &indices, 
	const vector< VectorType > &data, 
	vector< VectorType > &result) 
{
	FloatType delta = 0;
	size_t n = data.size();
	const size_t numClusters = result.size();
	for (unsigned c=0; c<numClusters; c++) {
		VectorType centr;
		nullVectorGeneric(data[0].getDim(), centr);
		card32 numVals = 0;
		float summed_weight = 0.0f;
		for (unsigned i=0; i<n; i++) {
			if (indices[i] == c) {
				centr += data[i] * weights[i];
				summed_weight += weights[i];
				numVals++;
			}
		}
		if (summed_weight > 1e-5) {
		   centr /= summed_weight;
		}
		FloatType diff = norm(centr - result[c]);
		result[c] = centr;
		if (diff > delta) delta = diff;
	}
	return delta;
}



template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::calc(card32 numClusters, vector< VectorType > &result, FloatType relConvThreshold, card32 numTrials) {
   if (data.empty()) {
      pAssert(false);
      return;
   }
	vector<unsigned> indices;
	calcEx(numClusters, result, indices, relConvThreshold, numTrials);
}


template <typename FloatType, typename VectorType>
FloatType KMeansClusteringTempl<FloatType, VectorType>::calcQuality(
	const vector<unsigned> &indices, 
	const vector< VectorType > &data, 
	vector< VectorType > &result)
{
	FloatType resultQ = 0;
	for (unsigned i=0; i<data.size(); i++) {
		resultQ += normQuad(data[i] - result[indices[i]]) * weights[i];
	}
	return resultQ;
}

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::calcIntraClusterVariances(
	const vector<unsigned> &indices, 	
	const vector< VectorType > &result,
	vector< FloatType > &variances)
{
	vector< FloatType > sumsWeights;
	variances.resize(result.size());
	sumsWeights.resize(result.size());
	for (unsigned i=0; i<result.size(); i++)
	{
		variances[i] = 0.0;
		sumsWeights[i] = 0;
	}
	
	for (unsigned i=0; i<data.size(); i++) 
	{
		unsigned c = indices[i];
		variances[c] += normQuad(data[i] - result[c]) * weights[i];
		sumsWeights[c] += weights[i];
	}
	for (unsigned i=0; i<result.size(); i++)
	{
		if (sumsWeights[i]>1e-9f)
		{
			variances[i] /= sumsWeights[i];
		}
	}
}

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::calcEx(card32 numClusters, vector< VectorType > &result, vector<unsigned> &indices, FloatType relConvThreshold, card32 numTrials) {
	vector< VectorType > tmpResult;
	vector<unsigned> tmpIndices;
	calcSingleTry(numClusters, result, indices, relConvThreshold);
	FloatType quality = calcQuality(indices, data, result);
	for (card32 i=1; i<numTrials; i++) {
		calcSingleTry(numClusters, tmpResult, tmpIndices, relConvThreshold);
		FloatType q = calcQuality(tmpIndices, data, tmpResult);
		if (q < quality) {
			quality = q;
			result = tmpResult;
			indices = tmpIndices;
		}
	}
}

template <typename FloatType, typename VectorType>
void KMeansClusteringTempl<FloatType, VectorType>::calcSingleTry(card32 numClusters, vector< VectorType > &result, vector<unsigned> &indices, FloatType relConvThreshold) {
	pAssert(data.size() != 0);
	vector< VectorType > centroids;
	VectorType lowerBB;
	VectorType upperBB;
	calcBoundingBoxGeneric<FloatType, VectorType>(data, lowerBB, upperBB);
	VectorType diagonal = upperBB - lowerBB;
	FloatType threshold = relConvThreshold * sqrt(diagonal*diagonal);
	
	card32 dim = data[0].getDim();
	result.clear();
	if (!initWithData) // default
	{
		VectorType r;
		for (card32 k=0; k<numClusters; k++) {
			randomDVectorGeneric(r, dim);
			result.push_back(lowerBB + r.componentProduct(upperBB - lowerBB));
		}
	}
	else // initWithData
	{		
		pAssert(data.size() > numClusters);
		// choose index into data vector randomly
		vector<int> inits;
		inits.resize(numClusters);		
		for (card32 k=0; k<numClusters; k++)
		{
			// do not use one sample more than once
			bool coincident = true;
			while (coincident) 
			{
				inits[k] = randomNumber((card32)data.size());
				coincident = false;
				for (card32 i=0; i<k; i++)
				{
					if (inits[i] == inits[k]) coincident = true;
				}
			}
		}
		for (card32 k=0; k<numClusters; k++)
		{
			result.push_back(data[(inits[k])]);
		}
	}
	int i = 0;
	FloatType delta = 0;
	do {
		chooseNearest(indices, data, result);
		delta = updateResult(indices, data, result);
		i++;
	} while (fabs(delta) > threshold && i < MAX_NUM_ITERATIONS);
	//#ifdef _DEBUG
	//	debugOutput << "KMeans: Num iterations = " << i << "\n\n";
	//#endif
}
