//---------------------------------------------------------------------------
#ifndef MultiDimensionalScalingH
#define MultiDimensionalScalingH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------

namespace geoxl {



/**
	computes classic multi-dimensional scaling:
	least-quares optimal low dimensional embedding of elements given a positive semidefinite distance matrix
	The distance matrix contains all pairwise distances squared, embeddingDimension gives the number of eigenvectors 
	(emb. dimensions) to be computed. embVects contains the resulting directions (already rescaled) and embEigenvalues the
	eigenvalues obtained during computation. Set allEigenvalues to true to output all (more than embeddingDimension) eigenvalues.
*/
template <typename FloatType>
void computeClassicMDS( const DynamicMatrix< FloatType > &squaredDistanceMatrix,
                        card32 embeddingDimension,
                        vector< DynamicVector<FloatType> > &embVects,
                        vector< FloatType > &embEigenvalues,
                        bool allEigenvalues = false );


/**
	computes classic multi-dimensional scaling:
	this function assumes that the pairwise scalar product matrix is already known (i.e. data has been centered)
	use computeClassicMDS for general distance matrix. Set allEigenvalues to true to output all (more than embeddingDimension) eigenvalues. 
*/
template <typename FloatType>
void computeClassicMDSByDotProds( const DynamicMatrix< FloatType > &pairwiseDotProducts,
											 card32 embeddingDimension,
											 vector< DynamicVector<FloatType> > &embVects,
											 vector< FloatType > &embEigenvalues,
											 bool allEigenvalues = false );



/**
	Helper class - add a set of examples and provide a distance metric, then compute MDS embedding.
	The template assumes a member function 
		
		MDSObjectWDist::dist(const MDSObjectWDist &) 
		
	to compute distances between the examples
*/
template <typename FloatType, class MDSObject>
class ClassicMDS {
 private:
	vector<MDSObject> examples;

 public:
	ClassicMDS();
	
	void calcEmb( card32 embeddingDimension,
                 vector< DynamicVector<FloatType> > &embVects,
                 vector< FloatType > &embEigenvalues );
};



}

#endif