//---------------------------------------------------------------------------
#include "MultiDimensionalScaling.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebraTools.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


template <typename FloatType>
void geoxl::computeClassicMDS( 
	const DynamicMatrix< FloatType > &squaredDistanceMatrix,
   card32 embeddingDimension,
   vector< DynamicVector<FloatType> > &embVects,
   vector< FloatType > &embEigenvalues,
   bool allEigenvalues )
{
	const card32 cols = squaredDistanceMatrix.getColumns();
	const card32 rows = squaredDistanceMatrix.getRows();
	pAssert( cols == rows );
	DynamicMatrix< FloatType > H(cols, rows);
	for (card32 y=0; y<rows; y++) {
		for (card32 x=0; x<cols; x++) {
			H[x][y] = - (FloatType)1/(FloatType)cols;
			if (x == y) H[x][y] += 1.0f;
		}
	}
	DynamicMatrix< FloatType > centered = H * squaredDistanceMatrix * H;
	centered *= (FloatType)(-0.5);
	computeClassicMDSByDotProds<FloatType>(centered, embeddingDimension, embVects, embEigenvalues, allEigenvalues);
}

template <typename FloatType>
void geoxl::computeClassicMDSByDotProds( 
	const DynamicMatrix< FloatType > &pairwiseDotProducts,
	card32 embeddingDimension,
	vector< DynamicVector<FloatType> > &embVects,
	vector< FloatType > &embEigenvalues,
	bool allEigenvalues )
{
	if (pairwiseDotProducts.getColumns() != pairwiseDotProducts.getRows()) {
		throw PException("computeClassicMDSByDotProds() - 'pairwiseDotProducts' must be a square matrix.");
	}
	if (embeddingDimension > pairwiseDotProducts.getColumns()) {
		embeddingDimension = pairwiseDotProducts.getColumns();
	}
	const card32 dim = pairwiseDotProducts.getColumns();
	DynamicMatrix<FloatType> u;
   DynamicVector<FloatType> singularValues;
   DynamicMatrix<FloatType> v;
   
   DynamicPCA<FloatType>::computeSVD(pairwiseDotProducts, u, singularValues, v, true);

	for (card32 i=0; i<embeddingDimension; i++) {
		u[i] /= norm(u[i]);
		u[i] *= sqrt(fabs(singularValues[i]));
	}
   
   embEigenvalues.resize((allEigenvalues?singularValues.getDim():embeddingDimension));
   embVects.resize(dim);
   for (card32 i=0; i<(allEigenvalues?singularValues.getDim():embeddingDimension); i++) {
		embEigenvalues[i] = singularValues[i];
   }
   for (card32 i=0; i<dim; i++) {
		DynamicVector<FloatType> ev(embeddingDimension);
		for (card32 j=0; j<embeddingDimension; j++) {
			ev[j] = u[j][i];	
		}
		embVects[i] = ev;
	}
}
