//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "NonLinearIterativeSolvers.h"
#include "NonLinearIterativeSolvers_templCode.h"
#include "Random.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------
#pragma warning(disable: 4101)

class NonLinearTest0 : public ObjectiveFunction2ndOrderDirDerivF {
 public:
	//	     ( 1	 0	)	
	//	 x^t (      ) x  
	//	     ( 0	 1 )	
	virtual float32 eval(const DVectorF &pos) const {
		return pos[0]*pos[0]+pos[1]*pos[1];
	}
	virtual void gradient(const DVectorF &pos, DVectorF &grad) const {

		grad[0] = 2*pos[0];
		grad[1] = 2*pos[1];
	}
	virtual float32 gradDirSecondDerivative(const DVectorF &pos, const DVectorF &normalizedGradient) const {
		return 2;
	}
};


class NonLinearTest1 : public ObjectiveFunctionAuto2ndOrderDirDerivF {
 public:
	//	     ( 3	 1	)		 ( 1 )^t
	//	 x^t (      ) x  + (	  )	  +3
	//	     ( 1	 3 )		 (	5 )
	float32 eval(const DVectorF &pos) const {
		return 3*pos[0]*pos[0]+2*pos[0]*pos[1]+3*pos[1]*pos[1] + pos[0] + 5*pos[1] + 3;
	}
	//	    ( 3  1 )      ( 1 )
	//	 2* (      ) x  + (   )
	//	    ( 1  3 )		( 5 )
	void gradient(const DVectorF &pos, DVectorF &grad) const {
		grad[0] = 6*pos[0] + 2*pos[1] + 1;
		grad[1] = 2*pos[0] + 6*pos[1] + 5;
	}
};

class NonLinearTest2 : public ObjectiveFunctionAuto2ndOrderDirDerivF {
 public:
	float32 eval(const DVectorF &pos) const {
		return sin(pos[0])*cos(pos[1]);
	}
	void gradient(const DVectorF &pos, DVectorF &grad) const {
		grad[0] = cos(pos[0])*cos(pos[1]);
		grad[1] = -sin(pos[0])*sin(pos[1]);
	}
	virtual float32 gradDirSecondDerivative(const DVectorF &pos, const DVectorF &normalizedGradient) const {
		Matrix2f hessian;
		hessian[0][0] = -sin(pos[0])*cos(pos[1]);		hessian[0][1] = -cos(pos[0])*sin(pos[1]);
		hessian[1][0] = -cos(pos[0])*sin(pos[1]);		hessian[1][1] = -sin(pos[0])*cos(pos[1]);
		Vector2f gn = makeVector2f(normalizedGradient[0], normalizedGradient[1]);
		return  (gn*(hessian*gn))*2;
	}
};

static const unsigned TEST3DIM = 12;

class NonLinearTest3 : public ObjectiveFunctionAuto2ndOrderDirDerivF {
 public:
	DMatrixF test;
	
	NonLinearTest3() {
		test.setDimension(TEST3DIM,TEST3DIM);
		for (unsigned j=0; j<TEST3DIM; j++) 
		for (unsigned i=0; i<=j; i++) {
			test[i][j] = rnd01();
			test[j][i] = test[i][j];
		}
		test = test*test;
	}
	float32 eval(const DVectorF &pos) const {
		return (test*pos)*pos;
	}
	void gradient(const DVectorF &pos, DVectorF &grad) const {
		grad = (test*pos)*2;
	}
	virtual float32 gradDirSecondDerivative(const DVectorF &pos, const DVectorF &normalizedGradient) const {
		return normalizedGradient*(test*normalizedGradient)*2;
	}
};

void geoxl::NonLinearIterativeSolvers_unitTest() {
	Matrix2f M = makeMatrix2f(6,2,
									  2,6);
	Matrix2f MI = invertMatrix(M);
	Vector2f solution = MI * makeVector2f(-1,-5);
	
	debugOutput << solution.toString() << "\n";
	
	DVectorF start(2);
	start[0] = 5;
	start[1] = 4;
	DVectorF start3(TEST3DIM);
	for (unsigned i=0; i<TEST3DIM; i++) {start3[i] = rnd01();}

	NonLinearTest0 fun0;
	NonLinearTest1 fun1;
	NonLinearTest2 fun2;
	NonLinearTest3 fun3;

	GradientDescentOptimizationF opt;
	ConjugateGradientDescentOptimizationF cgOpt;
	GradientDescentOptimizationF::Parameters params;
	ConjugateGradientDescentOptimizationF::Parameters cgParams;
	DVectorF result(2);

	debugOutput << "\n\ntest 0\n\n";
	opt.optimizeSimple(params, &fun0, start, result, true);
	debugOutput << result.toString() << "\n";
	opt.optimizeLineSearch(params, &fun0, start, result, true);
	debugOutput << result.toString() << "\n";
	cgOpt.optimize(cgParams, &fun0, start, result, true);
	debugOutput << result.toString() << "\n";

	debugOutput << "\n\ntest 1\n\n";
	opt.optimizeSimple(params, &fun1, start, result, true);
	debugOutput << result.toString() << "\n";
	opt.optimizeLineSearch(params, &fun1, start, result, true);
	debugOutput << result.toString() << "\n";
	cgOpt.optimize(cgParams, &fun1, start, result, true);
	debugOutput << result.toString() << "\n";

	start[0] = 0.3f;
	start[1] = 3.0f;
	debugOutput << "\n\ntest 2\n\n";
	opt.optimizeSimple(params, &fun2, start, result, true);
	debugOutput << result.toString() << "\n";
	opt.optimizeLineSearch(params, &fun2, start, result, true);
	debugOutput << result.toString() << "\n";
	cgOpt.optimize(cgParams, &fun2, start, result, true);
	debugOutput << result.toString() << "\n";

	start[0] = 5;
	start[1] = 4;
	debugOutput << "\n\ntest 3\n\n";
	opt.optimizeSimple(params, &fun3, start3, result, true);
	debugOutput << result.toString() << "\n";
	opt.optimizeLineSearch(params, &fun3, start3, result, true);
	debugOutput << result.toString() << "\n";	
	cgOpt.optimize(cgParams, &fun3, start3, result, true);
	debugOutput << result.toString() << "\n";
}


