//---------------------------------------------------------------------------
#ifndef NonLinearIterativeSolversH
#define NonLinearIterativeSolversH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl {


template <typename FloatType>
class ObjectiveFunction {
 public:
	virtual FloatType eval(const DynamicVector<FloatType> &pos) const = 0;
	virtual void gradient(const DynamicVector<FloatType> &pos, DynamicVector<FloatType> &gradient) const = 0;
};

typedef ObjectiveFunction<float32> ObjectiveFunctionF;
typedef ObjectiveFunction<float64> ObjectiveFunctionD;


template <typename FloatType>
class ObjectiveFunction2ndOrderDirDeriv : public ObjectiveFunction<FloatType> {
 public:
	virtual FloatType gradDirSecondDerivative(
		const DynamicVector<FloatType> &pos, 
		const DynamicVector<FloatType> &normalizedDir) const = 0;
};

typedef ObjectiveFunction2ndOrderDirDeriv<float32> ObjectiveFunction2ndOrderDirDerivF;
typedef ObjectiveFunction2ndOrderDirDeriv<float64> ObjectiveFunction2ndOrderDirDerivD;


template <typename FloatType>
class ObjectiveFunctionAuto2ndOrderDirDeriv : public ObjectiveFunction2ndOrderDirDeriv<FloatType> {
 private:
	FloatType numericalStepWidth;
 public:
	ObjectiveFunctionAuto2ndOrderDirDeriv() {numericalStepWidth = (FloatType)0.1;}
	void setup(FloatType numericalStepWidth) {this->numericalStepWidth = numericalStepWidth;}
	
	virtual FloatType gradDirSecondDerivative(
		const DynamicVector<FloatType> &pos, 
		const DynamicVector<FloatType> &normalizedDir) const;
};

typedef ObjectiveFunctionAuto2ndOrderDirDeriv<float32> ObjectiveFunctionAuto2ndOrderDirDerivF;
typedef ObjectiveFunctionAuto2ndOrderDirDeriv<float64> ObjectiveFunctionAuto2ndOrderDirDerivD;


/// old-school plainsimplestupidfixedstepwidth gd, just for comparison
template <typename FloatType>
class GradientDescentOptimization {
 public:
	struct Parameters {
   /// fixed step size
		FloatType stepSize;
		/// minimum step taken to announce convergence
		FloatType maxGradConv;
		/// max number of steps until convergence is declared no matter what
		card32 maxNumSteps;
		
		Parameters() {
			stepSize = 0.01f;
			maxGradConv = 1e-5f;
			maxNumSteps = 10000.0f;
		}
   };

   static void optimizeSimple(
		const Parameters &params,
		const ObjectiveFunction<FloatType> *function, 
		const DynamicVector<FloatType> &startEstimate,
		DynamicVector<FloatType> &result,
		bool printDebugOutput = false
	);
	static void optimizeLineSearch(
		const Parameters &params,
		const ObjectiveFunction2ndOrderDirDeriv<FloatType> *function, 
		const DynamicVector<FloatType> &startEstimate,
		DynamicVector<FloatType> &result,
		bool printDebugOutput = false
	);
};

typedef GradientDescentOptimization<float32> GradientDescentOptimizationF;
typedef GradientDescentOptimization<float64> GradientDescentOptimizationD;


/// non-linear cg solver
template <typename FloatType>
class ConjugateGradientDescentOptimization {
 public:
	struct Parameters {
		/// minimum step taken to announce convergence
		FloatType maxGradConv;
		/// max number of steps until convergence is declared no matter what
		card32 maxNumSteps;
		/// restart the conjugacy (set direction to gradient) after the specified number of iterations
		card32 restartIterations;
		/// use Polak-Ribire or Fletcher-Reeves conjugation (the first is the default)
		bool usePolakRibiere;
		
		Parameters() {
			maxGradConv = 1e-5f;
			maxNumSteps = 1000.0f;
			restartIterations = 50;
			usePolakRibiere = true;
		}
   };

   static void optimize(
		const Parameters &params,
		const ObjectiveFunction2ndOrderDirDeriv<FloatType> *function, 
		const DynamicVector<FloatType> &startEstimate,
		DynamicVector<FloatType> &result,
		bool printDebugOutput = false
	);
};

typedef ConjugateGradientDescentOptimization<float32> ConjugateGradientDescentOptimizationF;
typedef ConjugateGradientDescentOptimization<float64> ConjugateGradientDescentOptimizationD;




void NonLinearIterativeSolvers_unitTest();

}


#endif
