//---------------------------------------------------------------------------
#ifndef Polygon_templCodeH
#define Polygon_templCodeH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Polygon.h"
//---------------------------------------------------------------------------
#include "math_include.h"
#include "Exceptions.h"
#include <string>
//---------------------------------------------------------------------------

namespace geoxl {
   using namespace std;

template <class Vertex>
Polygon<Vertex>::Polygon()
{
}

template <class Vertex>
Polygon<Vertex>::Polygon( const Polygon &p )
{
   m_Points = p.m_Points;
}

template <class Vertex>
void Polygon<Vertex>::operator = ( const Polygon &p )
{
   m_Points = p.m_Points;
}

template <class Vertex>
void Polygon<Vertex>::addVertex( const Vertex &vec )
{
   m_Points.push_back( vec );
}

template <class Vertex>
inline card32 Polygon<Vertex>::getNumVertices() const
{
   return (card32)m_Points.size();
}

template <class Vertex>
inline Vertex & Polygon<Vertex>::operator [] ( card32 i )
{
   pAssert( i < m_Points.size());
   return m_Points[i];
}

template <class Vertex>
inline const Vertex & Polygon<Vertex>::operator [] ( card32 i ) const
{
   pAssert( i < m_Points.size());
   return m_Points[i];
}

template <class Vertex>
inline bool Polygon<Vertex>::isValid() const
{
   return m_Points.size() >= 3;
}

template <class Vertex>
EPlaneSide Polygon<Vertex>::classify( const Plane &plane ) const
{
  card32 numInfront = 0,
        numBehind = 0;
  for( card32 i=0;i<m_Points.size();i++ )
  {
     float distance = plane.calculateSignedDistance( m_Points[i].getPos3f() );
     numInfront += (distance > COINCIDING_EPSILON) ? 1 : 0;
     numBehind += (distance < -COINCIDING_EPSILON) ? 1 : 0;
  }

  if( numInfront > 0 )
     if( numBehind == 0 )
        return PS_INFRONT;
     else
        return PS_SPANNING;

  if( numBehind > 0 )
     return PS_BEHIND;

  return PS_COINCIDING;
}

template <class Vertex>
Vector3f Polygon<Vertex>::calculateNormal( bool &valid )
{
   if( m_Points.size() < 3 )
   {
      valid = false;
      return makeVector3f(0,0,0);
   }

   Vector3f tangent1 = m_Points[1].getPos3f() - m_Points[0].getPos3f();
   Vector3f tangent2 = m_Points[2].getPos3f() - m_Points[0].getPos3f();

   Vector3f ret = tangent2.crossProduct( tangent1 );
   float length = norm( ret );
   if( length < 1e-20f )
   {
      valid = false;
      return makeVector3f(0,0,0);
   }
   ret /= length;
   return ret;
}


template <class Vertex>
Plane Polygon<Vertex>::calculatePlane() const
{
  pAssert( m_Points.size() >= 3 );

  Plane ret;
  ret.point = m_Points[0].getPos3f();
  Vector3f tangent1 = m_Points[0].getPos3f() - m_Points[m_Points.size()-1].getPos3f();
  tangent1.normalize();
  for( card32 i=1;i<m_Points.size();i++ )
  {
     Vector3f tmpTangent = m_Points[i].getPos3f() - m_Points[i-1].getPos3f();
     tmpTangent.normalize();
     if( fabs(tmpTangent * tangent1) < 0.99f )
     {
        ret.normal = -tmpTangent.crossProduct( tangent1 );
        ret.normal.normalize();
        return ret;
     }
  }
  pAssert( false ); // degenerated polygon !
  return ret;
}

template <class Vertex>
Polygon<Vertex> Polygon<Vertex>::cut( const Plane &plane ) const
{
  Polygon ret( *this );
  if( m_Points.size() < 3 )
     return ret;

  vector<Vertex> tmp;
  float lastDistance = plane.calculateSignedDistance( ret.m_Points[0].getPos3f() );
  for( card32 i=0;i<ret.m_Points.size();i++ )
  {
     card32 j=(i+1) % ret.m_Points.size();
     float distance = plane.calculateSignedDistance( ret.m_Points[j].getPos3f() );

     if( lastDistance >= -COINCIDING_EPSILON )
     {
        tmp.push_back( ret.m_Points[i] );
     }

     bool coinciding = (fabs(lastDistance) < COINCIDING_EPSILON)
        && (fabs(distance) < COINCIDING_EPSILON);
     if( !coinciding && (lastDistance*distance < 0.0f) )
     {
        float t = plane.calculateLineIntersection( ret.m_Points[i].getPos3f(), ret.m_Points[j].getPos3f() );
        Vertex newVertex = ret.m_Points[i] + (ret.m_Points[j]-ret.m_Points[i]) * t;
        //float test = plane.calculateSignedDistance( newVertex.getPos3f() );
        tmp.push_back( newVertex );
     }
     lastDistance = distance;

  }

  ret.m_Points = tmp;
  return ret;
}

template <class Vertex>
bool Polygon<Vertex>::intersect( const MyType &poly )
{
  Plane this_plane = calculatePlane();

  if( poly.classify( this_plane ) != PS_SPANNING )
     return false;

  Plane poly_plane = poly.calculatePlane();
  if( classify( poly_plane ) != PS_SPANNING )
     return false;

  // calculate intersection of this_edges with poly_plane
  vector<Vector3f> tmp;
  float lastDistance = poly_plane.calculateSignedDistance( m_Points[0].getPos3f() );
  for( card32 i=0;i<m_Points.size();i++ )
  {
     card32 j=(i+1) % m_Points.size();
     float distance = poly_plane.calculateSignedDistance( m_Points[j].getPos3f() );

     bool coinciding = (fabs(lastDistance) < COINCIDING_EPSILON)
        && (fabs(distance) < COINCIDING_EPSILON);
     if( !coinciding && (lastDistance*distance < 0.0f) )
     {
        float t = poly_plane.calculateLineIntersection( m_Points[i].getPos3f(), m_Points[j].getPos3f() );
        Vector3f vec = m_Points[i].getPos3f() + (m_Points[j].getPos3f()-m_Points[i].getPos3f()) * t;
        tmp.push_back( vec );
     }
     lastDistance = distance;
     if( tmp.size() == 2 )
        break;

  }

  if( tmp.size() != 2 )
     return false;

  // cut line-segment at poly_edge_planes
  for( card32 i=0;i<poly.m_Points.size();i++ )
  {
     Plane edgePlane;
     edgePlane.point = poly.m_Points[i].getPos3f();
     edgePlane.normal = poly_plane.normal.crossProduct(
        poly.m_Points[(i+1) % poly.m_Points.size()].getPos3f() -
        poly.m_Points[i].getPos3f() );
     edgePlane.normal.normalize();

     float dist1 = edgePlane.calculateSignedDistance( tmp[0] );
     float dist2 = edgePlane.calculateSignedDistance( tmp[1] );

     if( dist1 < -COINCIDING_EPSILON && dist2 < -COINCIDING_EPSILON )
        return false;

     if( dist1 > COINCIDING_EPSILON && dist2 > COINCIDING_EPSILON )
        continue;

     if( dist1 < -COINCIDING_EPSILON && dist2 > -COINCIDING_EPSILON
        && dist2 < COINCIDING_EPSILON )
           return false;
     if( dist2 < -COINCIDING_EPSILON && dist1 > -COINCIDING_EPSILON
        && dist1 < COINCIDING_EPSILON )
           return false;

     if( dist1 < -COINCIDING_EPSILON || dist2 < -COINCIDING_EPSILON )
     {
        float t = edgePlane.calculateLineIntersection( tmp[0], tmp[1] );
        Vector3f tmpVec = tmp[0] + (tmp[1]-tmp[0])*t;
        if( dist1 < -COINCIDING_EPSILON )
           tmp[0] = tmpVec;
        else
           tmp[1] = tmpVec;
     }
  }

  return true;
}

}

#endif
