//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SVD.h"
//---------------------------------------------------------------------------
#include "cv_include.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

/// explicit instantiation
template class SVD<float32>;
template class SVD<float64>;



// ============================================================================
//         IMPLEMENTATION
// ============================================================================


template <typename FloatType>
SVD<FloatType>::SVD(){
}


template<>
const int SVD<float32>::getType(){
    return CV_32FC1;
}

template<>
const int SVD<float64>::getType(){
	return CV_64FC1;
}


// needed for OpenCV Datasructure
//xxx_open_cv_missing
typedef CvMat* cvMatrixPointer;



template <typename FloatType>
void SVD<FloatType>::computeSVD(const DynamicMatrix<FloatType>  &input,
												DynamicMatrix<FloatType>  &U,
												DynamicVector<FloatType>  &sv,
												DynamicMatrix<FloatType>  &Vt)
{
   int type =  getType();
	// decomposes matrix input into a product of a diagonal matrix and two orthogonal matrices:
    // A=U*S*Vt

   card32 rows = input.getRowsDim();
   card32 cols = input.getColsDim();
   //debugOutput << "computeSVD, rows: " << rows << "   columns " << cols << "\n"; debugOutput.flush();


   // Convert input to array (because OpenCv needs an array)
   FloatType *inputArray = new FloatType[rows*cols];
   MatrixToArray(input, inputArray);

   // allocate SVD result arrays
   FloatType *resArray_U = new FloatType[rows*rows];
   FloatType *resArray_Vt = new FloatType[cols*cols];
   sv.setDim(cols);

   // Prepare OpenCV Datasructure CvMat
   CvMat cvMatInput;
   CvMat cvMat_U;
   CvMat cvMat_S;
   CvMat cvMat_Vt;
   cvInitMatHeader( &cvMatInput, rows, cols, type, inputArray);
   cvInitMatHeader( &cvMat_S, cols, 1, type, sv.data());
   cvInitMatHeader( &cvMat_U, rows, rows, type, resArray_U);
   cvInitMatHeader( &cvMat_Vt, cols, cols, type, resArray_Vt);

   // Do SVD
	cvSVD(&cvMatInput, &cvMat_S, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_V_T);
   // CV_SVD_MODIFY_A enables modification of matrix A (inputArray) during the operation. It speeds up the processing.

   // convert result to DynamicMatrix
   U.setDimension(rows, rows);
   ArrayToMatrix(resArray_U,U);
   Vt.setDimension(cols, cols);
   ArrayToMatrix(resArray_Vt,Vt);

	delete[] inputArray;
   delete[] resArray_U;
   delete[] resArray_Vt;
}

template <typename FloatType>
void SVD<FloatType>::computeSVDPseudoInverseMatrix(const DynamicMatrix<FloatType> &input, DynamicMatrix<FloatType> &returnedPseudoInverse, float64 &returnedInversedConditionNumberOfA )
{
   int type =  getType();


   card32 rows = input.getRowsDim();
   card32 cols = input.getColsDim();
   //debugOutput << "computeSVDPseudoInverseMatrix, rows: " << rows << "   columns " << cols << "\n"; debugOutput.flush();


   // Convert input to array (because OpenCv needs an array)
   FloatType *inputArray = new FloatType[rows*cols];
   MatrixToArray(input, inputArray);

   // allocate SVD inv result array
   FloatType *resArray = new FloatType[rows*cols];

   // Prepare OpenCV Datasructure CvMat
   CvMat cvMatInput;
   CvMat cvMat_res;
   cvInitMatHeader( &cvMatInput, rows, cols, type, inputArray);
   cvInitMatHeader( &cvMat_res, cols, rows, type, resArray);
 

   // Do SVD invert
	returnedInversedConditionNumberOfA = cvInvert(&cvMatInput, &cvMat_res, CV_SVD);


   // convert result to DynamicMatrix
   returnedPseudoInverse.setDimension(cols, rows);
   ArrayToMatrix(resArray, returnedPseudoInverse);


	delete[] inputArray;
   delete[] resArray;

}

template <typename FloatType>
void SVD<FloatType>::computeSVDPseudoInverseMatrix(const Matrix3f &input, Matrix3f &returnedPseudoInverse, float64 &returnedInversedConditionNumberOfA )
{
   
	  int type =  getType();
	// Convert input to array (because OpenCv needs an array)
   FloatType inputArray[9];
   MatrixToArray(input, inputArray);

   FloatType resArray[9];
      CvMat cvMatInput;
   CvMat cvMat_res;
   cvInitMatHeader( &cvMatInput, 3, 3, type, inputArray);
   cvInitMatHeader( &cvMat_res, 3, 3, type, resArray);

   // Do SVD invert
   returnedInversedConditionNumberOfA = cvInvert(&cvMatInput, &cvMat_res, CV_SVD);

   ArrayToMatrix(resArray, returnedPseudoInverse);

}


template <typename FloatType>
void SVD<FloatType>::computeSVD(const Matrix3f  &input,
												Matrix3f &U,
												Vector3f &sv,
												Matrix3f  &Vt)
{
   int type =  getType();
	// decomposes matrix input into a product of a diagonal matrix and two orthogonal matrices:
    // A=U*S*Vt

   // debugOutput << "computeSVD for symmetric 3x3 matrix" << "\n"; debugOutput.flush();

   // Convert input to array (because OpenCv needs an array)
   FloatType inputArray[9];
   MatrixToArray(input, inputArray);

   // allocate SVD result arrays
   FloatType resArray_U[9];
   FloatType resArray_Vt[9];

   // Prepare OpenCV Datasructure CvMat
   CvMat cvMatInput;
   CvMat cvMat_U;
   CvMat cvMat_S;
   CvMat cvMat_Vt;
   cvInitMatHeader( &cvMatInput, 3, 3, type, inputArray);
   cvInitMatHeader( &cvMat_S, 3, 1, type, sv.data());
   cvInitMatHeader( &cvMat_U, 3, 3, type, resArray_U);
   cvInitMatHeader( &cvMat_Vt, 3, 3, type, resArray_Vt);

   // Do SVD
	cvSVD(&cvMatInput, &cvMat_S, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_V_T);
   // CV_SVD_MODIFY_A enables modification of matrix A (inputArray) during the operation. It speeds up the processing.

   // convert result to DynamicMatrix
   ArrayToMatrix(resArray_U,U);
   ArrayToMatrix(resArray_Vt,Vt);
}


template <typename FloatType>
inline void SVD<FloatType>::MatrixToArray(const DynamicMatrix<FloatType> &source, FloatType* target) {
	card32 columns = source.getColsDim();
	card32 rows = source.getRowsDim();
   for (unsigned j=0; j<columns; ++j)
      for (unsigned i=0; i<rows; ++i){
         target[i*columns+j] = source[j][i];
      }
}

template <typename FloatType>
inline void SVD<FloatType>::ArrayToMatrix(FloatType* source, DynamicMatrix<FloatType> &target){
	card32 columns = target.getColsDim();
	card32 rows = target.getRowsDim();
   for (unsigned j=0; j<columns; ++j)
      for (unsigned i=0; i<rows; ++i){
         target[j][i] = source[i*columns+j];
		}
}


template <typename FloatType>
inline void SVD<FloatType>::MatrixToArray(const vector<Vector3f> &source, FloatType* target) {
	size_t columns = source.size();
	for (size_t j=0; j<columns; ++j) {
		const Vector3f &tmp = source[j];
      target[j] = tmp[0];
      target[columns+j] = tmp[1];
      target[2*columns+j] = tmp[2];
	}
}

template <typename FloatType>
inline void SVD<FloatType>::MatrixToArray(const Matrix3f &source, FloatType* target) {
	const Vector3f &u = source[0];
	const Vector3f &v = source[1];
	const Vector3f &w = source[2];
	target[0] = u[0];
	target[3] = u[1];
	target[6] = u[2];
	target[1] = v[0];
	target[4] = v[1];
	target[7] = v[2];
	target[2] = w[0];
	target[5] = w[1];
	target[8] = w[2];
}

template <typename FloatType>
inline void SVD<FloatType>::ArrayToMatrix(FloatType* source, vector<Vector3f> &target){
	size_t columns = target.size();
	for (size_t j=0; j<columns; ++j) {
		Vector3f &tmp = target[j];
      tmp[0] = source[j];
      tmp[1] = source[columns+j];
      tmp[2] = source[2*columns+j];
	}
}

template <typename FloatType>
inline void SVD<FloatType>::ArrayToMatrix(FloatType* source, Matrix3f &target){
	Vector3f &u = target[0];
	Vector3f &v = target[1];
	Vector3f &w = target[2];
	u[0] = source[0];
	u[1] = source[3];
	u[2] = source[6];
	v[0] = source[1];
	v[1] = source[4];
	v[2] = source[7];
	w[0] = source[2];
	w[1] = source[5];
	w[2] = source[8];
}




template <typename FloatType>
void SVD<FloatType>::invertMatrix(const Matrix3f &input, Matrix3f &output, float32 eps) {

	// in = U S Vt  --> in * (V S^+ Ut) = I
	// out = in^+ = V S^+ Ut

	Matrix3f U,Vt;
	Vector3f sv;
	computeSVD(input, U, sv, Vt);
	
	for (card32 i=0; i < 3; ++i) {
		// if ( (sv[i] > 0 && sv[i] < eps) || (sv[i] < 0 && sv[i] > -eps) ) {
		if ( sv[i]*sv[i] < eps ) {
			sv[i] = 0;
		} else {
			sv[i] = 1.0f/sv[i];
		}
	}

	Vector3f &vt0 = Vt[0];
	Vector3f &vt1 = Vt[1];
	Vector3f &vt2 = Vt[2];

	Vector3f &u0 = U[0];
	Vector3f &u1 = U[1];
	Vector3f &u2 = U[2];

	Vector3f &o0 = output[0];
	Vector3f &o1 = output[1];
	Vector3f &o2 = output[2];

	o0[0] = vt0[0]*sv[0]*u0[0] + vt0[1]*sv[1]*u1[0] + vt0[2]*sv[2]*u2[0];
	o0[1] = vt1[0]*sv[0]*u0[0] + vt1[1]*sv[1]*u1[0] + vt1[2]*sv[2]*u2[0];
	o0[2] = vt2[0]*sv[0]*u0[0] + vt2[1]*sv[1]*u1[0] + vt2[2]*sv[2]*u2[0];

	o1[0] = vt0[0]*sv[0]*u0[1] + vt0[1]*sv[1]*u1[1] + vt0[2]*sv[2]*u2[1];
	o1[1] = vt1[0]*sv[0]*u0[1] + vt1[1]*sv[1]*u1[1] + vt1[2]*sv[2]*u2[1];
	o1[2] = vt2[0]*sv[0]*u0[1] + vt2[1]*sv[1]*u1[1] + vt2[2]*sv[2]*u2[1];

	o2[0] = vt0[0]*sv[0]*u0[2] + vt0[1]*sv[1]*u1[2] + vt0[2]*sv[2]*u2[2];
	o2[1] = vt1[0]*sv[0]*u0[2] + vt1[1]*sv[1]*u1[2] + vt1[2]*sv[2]*u2[2];
	o2[2] = vt2[0]*sv[0]*u0[2] + vt2[1]*sv[1]*u1[2] + vt2[2]*sv[2]*u2[2];
}



template <typename FloatType>
void SVD<FloatType>::solve(const Matrix3f &A, const Vector3f &b, Vector3f &x, float32 eps) {
	// solve Ax=b --> x = A^+ b
	
	Matrix3f invA;
	invertMatrix(A, invA, eps);
	x = invA*b;
}

template <typename FloatType>
void SVD<FloatType>::computeSVD_NoMatV(const DynamicMatrix<FloatType>  &input,
										DynamicMatrix<FloatType>  &U,
										DynamicVector<FloatType>  &sv)
{

	debugOutput << "IN !!!\n";
	debugOutput << input.getColsDim() << "  " << input.getRowsDim() << " " <<  U.getColsDim() << "  " << U.getRowsDim() << "   " << sv.getDim() << "IN !!!\n";
	debugOutput.flush();





	int type =  getType();
	// decomposes matrix input into a product of a diagonal matrix and two orthogonal matrices:
	// A=U*S*Vt

	card32 rows = input.getRowsDim();
	card32 cols = input.getColsDim();
	//debugOutput << "computeSVD, rows: " << rows << "   columns " << cols << "\n"; debugOutput.flush();


	// Convert input to array (because OpenCv needs an array)
	FloatType *inputArray = new FloatType[rows*cols];
	MatrixToArray(input, inputArray);

	// allocate SVD result arrays
	FloatType *resArray_U = new FloatType[rows*rows];
	sv.setDim(cols);

	// Prepare OpenCV Datasructure CvMat
	CvMat cvMatInput;
	CvMat cvMat_U;
	CvMat cvMat_S;
	cvInitMatHeader( &cvMatInput, rows, cols, type, inputArray);
	cvInitMatHeader( &cvMat_S, cols, 1, type, sv.data());
	cvInitMatHeader( &cvMat_U, rows, rows, type, resArray_U);

	debugOutput << "BEFORE !!!\n";
	debugOutput.flush();

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_S, &cvMat_U, NULL, CV_SVD_MODIFY_A);
	// CV_SVD_MODIFY_A enables modification of matrix A (inputArray) during the operation. It speeds up the processing.

	debugOutput << "AFTER !!!\n";
	debugOutput.flush();

	// convert result to DynamicMatrix
	U.setDimension(rows, rows);
	ArrayToMatrix(resArray_U,U);

	delete[] inputArray;
	delete[] resArray_U;

	debugOutput << "OUT !!!\n";
	debugOutput.flush();

}

template <typename FloatType>
void SVD<FloatType>::computeSVD_NoMatV(const Matrix3f  &input,
								Matrix3f &U,
								Vector3f &sv)
{
	int type =  getType();
	// decomposes matrix input into a product of a diagonal matrix and two orthogonal matrices:
	// A=U*S*Vt

	// debugOutput << "computeSVD for symmetric 3x3 matrix" << "\n"; debugOutput.flush();

	// Convert input to array (because OpenCv needs an array)
	FloatType inputArray[9];
	MatrixToArray(input, inputArray);

	// allocate SVD result arrays
	FloatType resArray_U[9];

	// Prepare OpenCV Datasructure CvMat
	CvMat cvMatInput;
	CvMat cvMat_U;
	CvMat cvMat_S;
	cvInitMatHeader( &cvMatInput, 3, 3, type, inputArray);
	cvInitMatHeader( &cvMat_S, 3, 1, type, sv.data());
	cvInitMatHeader( &cvMat_U, 3, 3, type, resArray_U);

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_S, &cvMat_U, NULL, CV_SVD_MODIFY_A);
	// CV_SVD_MODIFY_A enables modification of matrix A (inputArray) during the operation. It speeds up the processing.

	// convert result to DynamicMatrix
	ArrayToMatrix(resArray_U,U);

}
