//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "BlockedMultiObjectStream.h"
//---------------------------------------------------------------------------
#include "FileBlockAllocator.h"
#include "FileBlockChainAllocator.h"
#include "MemoryObjectStreams.h"
#include "Exceptions.h"
#include "StringClassProperty.h"
#include "NumericalClassProperty.h"
#include "NumericalClassProperty_templCode.h"
#include "FixedArrayClassProperty.h"
#include <qtCore/qfile.h>
#include "BooleanClassProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


MetaClass* BlockedMultiObjectStream::BlockedMultiObjectStreamMetaClass = NULL;

const MetaClass* BlockedMultiObjectStream::getInstanceClass() const {
   return getClass();
}

Object* BlockedMultiObjectStream::newInstance() {
   return new BlockedMultiObjectStream;
}

const MetaClass* BlockedMultiObjectStream::getClass() {
   if (BlockedMultiObjectStreamMetaClass==NULL) {
      BlockedMultiObjectStreamMetaClass = new MetaClass("BlockedMultiObjectStream", "no description.",
                                  newInstance, sizeof(BlockedMultiObjectStream));
   }
   return BlockedMultiObjectStreamMetaClass;
}

string BlockedMultiObjectStream::getFilename() const {
   return filename;
}

void BlockedMultiObjectStream::setFilename(string f) {
   if (status == statusConnected) {
      throw EIOException("BlockedMultiObjectStream::setFilename() - cannot change file name while being connected to a file.");
   }
   filename = f;
   createAllocators();
}


void BlockedMultiObjectStream::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   BlockedMultiObjectStreamMetaClass->copyParentProperties();
	
	BEGIN_CLASS_INIT( BlockedMultiObjectStream );


   BlockedMultiObjectStreamMetaClass->addProperty(
      new StringClassProperty("filename", 1, (GetStringPropMethod)&getFilename, (SetStringPropMethod)&setFilename)
   );

   BlockedMultiObjectStream * obj = NULL;
	Card64ClassProperty *bsp = new Card64ClassProperty("blockSize", 1, (pointer_diff)(&obj->blockSize-NULL));
   bsp->setEditReadonly(true);

   BlockedMultiObjectStreamMetaClass->addProperty(bsp);

	ADD_CARD8_FIXED_ARRAY_PROP( instanceID, 0, sizeof(GloballyUniqueID), sizeof(GloballyUniqueID) );
	ADD_CARD8_PROP( status, 0 );
	
	BlockedMultiObjectStreamMetaClass->getGUIDescriptor()->addClassCreator(BlockedMultiObjectStreamCreator::getClass());
}

void BlockedMultiObjectStream::shutdown() {
   delete BlockedMultiObjectStreamMetaClass;
}


// ----

IMPLEMENT_GEOXL_CLASS(BlockedMultiObjectStreamCreator, 0) 
{
	BEGIN_CLASS_INIT(BlockedMultiObjectStreamCreator)
	ADD_STRING_PROP(filename, 0)
	ADD_CARD64_PROP(blockSize, 0)
	ADD_BOOLEAN_PROP(showDebugEvents, 0)
}

// ----

BlockedMultiObjectStream::BlockedMultiObjectStream() {
   baseAlloc = NULL;
   alloc = NULL;
   filename = "";
   blockSize = 0;
   showDebugEvents = true;
   instanceID.createNewID();
   status = statusDisconnected;
}

void BlockedMultiObjectStream::createAllocators() {
   baseAlloc = new FileBlockAllocator();
	if( QFile::exists( filename.c_str() ) ) 
	{
      baseAlloc->openFile(filename.c_str());
      if (instanceID != baseAlloc->getFileGUID()) {
         status = statusDisconnected;
         baseAlloc->close();
         delete baseAlloc;
         baseAlloc = NULL;
         return;
      } else {
         status = statusConnected;
      }
   } else {
      baseAlloc->createNewFile(filename.c_str(), blockSize);
      instanceID = baseAlloc->getFileGUID();
      status = statusConnected;
   }
   alloc = new FileBlockChainAllocator(baseAlloc);
}

BlockedMultiObjectStream::BlockedMultiObjectStream(const char *filename, const card64 blockSize, bool showDebugEvents) {
   this->filename = filename;
   this->blockSize = blockSize;
   this->showDebugEvents = showDebugEvents;
   status = statusDisconnected;
   createAllocators();
}

MultiObjectStream *BlockedMultiObjectStream::createOOCAwareCopy(string idPrefix, bool insertOriginal, string idSuffix) {
	string newFilename = removeFilenameFromPath(filename) + idPrefix + (insertOriginal?removePathFromFilename(filename):string("")) + idSuffix;
	return new BlockedMultiObjectStream(newFilename.c_str(), blockSize, showDebugEvents);
}


BlockedMultiObjectStream::ObjectID BlockedMultiObjectStream::createObjectEntry() {
   pAssert(status == statusConnected);
   return alloc->allocBlock(0);
}

void BlockedMultiObjectStream::updateObject(ObjectID id, const MultiPartPersistent *obj) {
   pAssert(status == statusConnected);
   MemoryOutputObjectStream out(false, NULL, true);
   const card32 n = (card32)globalObjectPointers.size();
   for (card32 i=0; i<n; i++) {
      out.addGlobalObject(globalObjectPointers[i].obj, globalObjectPointers[i].index);
   }
   out.writeObject(obj);
   const vector<card8> *memblock = out.closeAndGetMemoryBuffer();
   alloc->reallocBlock(id, (max_platform_card)memblock->size());
   alloc->writeBuffer(id, &((*memblock)[0]), (max_platform_card)memblock->size());
}

MultiPartPersistent *BlockedMultiObjectStream::getObject(ObjectID id) {
   pAssert(status == statusConnected);
   max_platform_card memSize = alloc->getMaxSize(id);
   card8 *mem = new card8[memSize];
   alloc->readBuffer(id, mem, memSize);
   MemoryInputObjectStream in(mem, memSize, NULL);
   const card32 n = (card32)globalObjectPointers.size();
   for (card32 i=0; i<n; i++) {
      in.addGlobalObject(globalObjectPointers[i].obj, globalObjectPointers[i].index);
   }
   Persistent *obj;
   in.readObject(obj);
   delete[] mem;
   return (MultiPartPersistent*)obj;
}

void BlockedMultiObjectStream::deleteObject(ObjectID id) {
   pAssert(status == statusConnected);
   alloc->deleteBlock(id);
}

void BlockedMultiObjectStream::sync() {
   pAssert(status == statusConnected);
   baseAlloc->sync();
}

MultiObjectStream::Status BlockedMultiObjectStream::getStatus() {
   return status;
}


void BlockedMultiObjectStream::write(OutputObjectStream *s) const {
   MultiObjectStream::write(s);
   s->writeString(filename);
   s->write(blockSize);
   s->write(instanceID);
}

void BlockedMultiObjectStream::read(InputObjectStream *s) {
   MultiObjectStream::read(s);
   s->readString(filename);
   s->read(blockSize);
   s->read(instanceID);
   delete alloc;
   delete baseAlloc;
   createAllocators();
}


BlockedMultiObjectStream::~BlockedMultiObjectStream() {
   if (alloc != NULL) {
      delete alloc;
      baseAlloc->close();
      delete baseAlloc;
   }
}


// ---


BlockedMultiObjectStreamCreator::BlockedMultiObjectStreamCreator() {
	filename = "test.swap";
	blockSize = 1024*1024;
	showDebugEvents = true;
}
Object *BlockedMultiObjectStreamCreator::makeObject() {
	return new BlockedMultiObjectStream(filename.c_str(), blockSize, showDebugEvents);
}
