//---------------------------------------------------------------------------
#ifndef BlockedMultiObjectStreamH
#define BlockedMultiObjectStreamH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MultiObjectStream.h"
#include "Exceptions.h"
#include "GloballyUniqueID.h"
#include "ClassCreator.h"
//---------------------------------------------------------------------------

namespace geoxl {

class FileBlockAllocator;
class FileBlockChainAllocator;

/// Class to store multiple objects in one stream / file.
/// each single object must fit into main memory; however, the set of all objects may be larger (64 bit limited)
class FOUNDATION_API BlockedMultiObjectStream : public FileBasedMultiObjectStream {
 private:
   static MetaClass *BlockedMultiObjectStreamMetaClass;
   static Object* newInstance();

   FileBlockAllocator *baseAlloc;
   FileBlockChainAllocator *alloc;
   GloballyUniqueID instanceID;
   Status status;

   /// for serialization only
   string filename;
   /// for serialization only
   card64 blockSize;

   /// display performance statistics
   bool showDebugEvents;

   BlockedMultiObjectStream();
   void createAllocators();

 public:
   /// property access
   virtual string getFilename() const;
   
 private:
   /// property access
   void setFilename(string f);

 public:
   /// create new instance: if filename exists, the existing database is reopened. otherwise, a new one is created.
   /// (only) in the latter case, the blocksize is used for splitting objects into blocks.
   /// the database consists of two files: the one named "filename" and a second one named "filename.classtable"
   /// the latter is used to store class identifiers (it is updated at each sync).
   /// reopening the data base is only possible with valid ".classtable" file.
   /// --- REMARK: The .classtable mechanism is not yet implemented. ---
   BlockedMultiObjectStream(const char *filename, const card64 blockSize = 32768, bool showDebugEvents = true);

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual MultiObjectStream *createOOCAwareCopy(string idPrefix = "", bool insertOriginal = true, string idSuffix = "");

   virtual ObjectID createObjectEntry();
   virtual void updateObject(ObjectID id, const MultiPartPersistent *obj);
   virtual MultiPartPersistent *getObject(ObjectID id);
   virtual void deleteObject(ObjectID id);
   virtual void sync();
   virtual Status getStatus();

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   virtual ~BlockedMultiObjectStream();

};

/// helper class that creates new BlockedMultiObjectStreams (for easier interactive setup)
class BlockedMultiObjectStreamCreator : public ClassCreator {
 GEOXL_CLASS(BlockedMultiObjectStreamCreator)
private:
	string filename;
	card64 blockSize;
	bool showDebugEvents;
public:
	BlockedMultiObjectStreamCreator();
	virtual Object *makeObject();
};

}


#endif
