//---------------------------------------------------------------------------
#ifndef CommandObjectH
#define CommandObjectH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "Tools.h"
//---------------------------------------------------------------------------

namespace geoxl {

class CommandHistory;
class CommandObjMacro;


/// Objects that serve as context for CommandObjects must be derived (prob. indirectly)
/// from "Object". Additionally, they might implement this interface to make additional
/// functionality accessible.
class CommandObjectContextInterface { // -- INTERFACE CLASS
 public:
   virtual void clear() = 0;
   virtual CommandHistory *getCommandHistory() = 0;
   virtual void addMacro(CommandObjMacro* macro, string name) = 0;
};


/** Object describing a command performed on a point cloud scene. */
class FOUNDATION_API CommandObject : public Persistent {
 private:
   static MetaClass *CommandObjectMetaClass;
   static Object* newInstance();

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// standard action is "copyObjectProperties", not mem-copy
   virtual void assign(const Object* obj, CopyContext *context = NULL);

   /// perform command; context is the object that contains all objects in which the action can take place
   virtual void doAction(Object *context) = 0;
   /// undo the command (if possible); context is the object that contains all objects in which the action can take place
   virtual void undoAction(Object *context) = 0;
   /// test if the command is undoable
   virtual bool canUndo() = 0;

   /// return brief description for user
   virtual string getDescription() = 0;

   /// try to merge command with latest command of the same type.
   /// this == latest, param == new command, this is merged with param
   /// result == true if successful, false if no merge has been (could have been) performed.
   /// this option is useful for continous operations such as transformation, change of view point etc...
   virtual bool mergeCommand(const CommandObject *nextCmd) {return false;}
};


/// thrown if command cannot be performed
class FOUNDATION_API ECommandFailed : public PException {
 public:
   ECommandFailed(const string msg) : PException(msg) {};
   virtual string getTypeString() {return "ECommandFailed";}
};

/// thrown if command object does not provide undo but undo has been called
class FOUNDATION_API ECannotUndo : public ECommandFailed {
 public:
   ECannotUndo(const string msg) : ECommandFailed(msg) {};
   virtual string getTypeString() {return "ECannotUndo";}
};

}

#endif
