//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "CubeMap.h"
#include "CubeMap_templCode.h"
//---------------------------------------------------------------------------
#include "Exceptions.h"
#include "ObjectClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "ClassMethods.h"
#include "LinearAlgebra_templCode.h"
#include "TemplatedImage_templCode.h"
#include "FileDialogs.h"
#include "InterfaceGlobals.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


MetaClass* RGBACubeMap::RGBACubeMapMetaClass = NULL;

const MetaClass* RGBACubeMap::getInstanceClass() const {
   return getClass();
}

Object* RGBACubeMap::newInstance() {
   return new RGBACubeMap;
}

const MetaClass* RGBACubeMap::getClass() {
   if (RGBACubeMapMetaClass==NULL) {
      RGBACubeMapMetaClass = new MetaClass("RGBACubeMap", "no description.",
                                  newInstance, sizeof(RGBACubeMap));
   }
   return RGBACubeMapMetaClass;
}

void RGBACubeMap::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   RGBACubeMapMetaClass->copyParentProperties();

	BEGIN_CLASS_INIT( RGBACubeMap );
	ADD_OBJECT_FIXED_ARRAY_PROP( sides, 0, 6, RGBAImage::getClass(), true );

   RGBACubeMapMetaClass->addMethod(new NoArgsClassMethod(
      (NoArgsMethodPointer)&interactiveLoadCubeMapFromFile, "loadCubeMap...", 
	      "open file dialog and load cube map from file (cross format)", ":/foundation/Images/Buttons/open.png" )
   );
   RGBACubeMapMetaClass->addMethod(new NoArgsClassMethod(
      (NoArgsMethodPointer)&setupTestMap, "setupTestMap", "create a simple test cube map with colored sides")
   );
}

void RGBACubeMap::shutdown() {
   delete RGBACubeMapMetaClass;
}


// ----

RGBACubeMap::RGBACubeMap() {
   for (card32 i=0; i<6; i++) {
      delete sides[i]; // remove "TemplatedImage<Vector4ub>"
      sides[i] = new RGBAImage();
   }
}

void RGBACubeMap::assignSide(card32 index, const TemplatedImage<Vector4ub> *img) {
   pAssert(dynamic_cast<const RGBAImage*>(img) != NULL);
   CubeMap<Vector4ub>::assignSide(index, img);
}


void RGBACubeMap::loadFromCrossBMP(const char *filename) {

	QImage* img = new QImage(QImage(QString(filename)).convertToFormat(QImage::Format_RGB32));

   card32 w = img->width();
   card32 h = img->height();

   card32 sw = w/3;
   card32 sh = h/4;

   if (sw == 0 || sh == 0 || w%3 != 0 || h%4 != 0) {
      throw PException("RGBACubeMap::loadFromCrossBMP() - size does not match.");
   }

   // coordinates:
   //
   //        +---.
   //        |-y |
   //    +---+---+---.
   //    |-x |+z |+x |
   //    .---+---.---.
   //        |+y |
   //        +---.  + = reference point
   //        |-z |
   //        .---.

   const Vector2i refNX = makeVector2i(0,sh);
   const Vector2i refPX = makeVector2i(2*sw,sh);
   const Vector2i refNY = makeVector2i(sw,0);
   const Vector2i refPY = makeVector2i(sw,2*sh);
   const Vector2i refNZ = makeVector2i(sw,3*sh);
   const Vector2i refPZ = makeVector2i(sw,sh);

   ((RGBAImage*)(sides[CMS_POSITIVE_X]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_POSITIVE_X]))->copyFromBMP(img, refPX[0], refPX[1], sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_X]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_X]))->copyFromBMP(img, refNX[0], refNX[1], sw, sh);
   ((RGBAImage*)(sides[CMS_POSITIVE_Y]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_POSITIVE_Y]))->copyFromBMP(img, refPY[0], refPY[1], sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_Y]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_Y]))->copyFromBMP(img, refNY[0], refNY[1], sw, sh);
   ((RGBAImage*)(sides[CMS_POSITIVE_Z]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_POSITIVE_Z]))->copyFromBMP(img, refPZ[0], refPZ[1], sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_Z]))->resize(sw, sh);
   ((RGBAImage*)(sides[CMS_NEGATIVE_Z]))->copyFromBMP(img, refNZ[0], refNZ[1], sw, sh);

	delete img;
}

void RGBACubeMap::interactiveLoadCubeMapFromFile() {
	string filename = FileDialogs::getOpenFileName(InterfaceGlobals::getMainWindow(), "Open Cube Map Image (Cross Layout)",
	                                               "Bitmap Files (*.bmp; *.jpg; *.png)", NULL, "", FileDialogs::IMPORT_DIR);

   if (filename != "") {
      try {
         loadFromCrossBMP(filename.c_str());
      } catch (...) {
         error("error loading map.");
      }
   }
}


Vector4ub calcAngular( QImage *img, Vector3f &vec )
{
   float vAngle = acos( vec[1] ) / M_PI * 180.0f;
   Vector2f tmp = normalize( makeVector2f( vec[0], vec[2] ) );
   float hAngle = atan2( tmp[0], tmp[1] ) / M_PI * 180.0f;

   if( hAngle < 0 )
      hAngle += 360.0f;
   hAngle = max( 0, min( 360, hAngle ) );

   float x = min(img->width()-1, max( 0, hAngle / 360.0f * (float)img->width() ));
   float y = min(img->height()-1, max( 0, vAngle / 180.0f * (float)img->height() ));

	QRgb rgb = img->pixel(x,y);
   Vector4ub ret = makeVector4ub(qBlue(rgb), qGreen(rgb), qRed(rgb), 0xFF);
   return ret;
}


void RGBACubeMap::loadFromAngularBMP(const char * filename )
{
	QImage* img = new QImage(QImage(QString(filename)).convertToFormat(QImage::Format_RGB32));

   card32 w = img->width();
   card32 h = img->height();

   int size = 1;
   while( size < (int)w/2 && size < (int)h/2 )
      size *= 2;

   for( card32 i=0;i<6;i++ )
      sides[i]->resize(size,size);

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(0.5f,  -yf, -xf );
            vec = normalize( vec );

            sides[ CMS_POSITIVE_X ]->setPixel(x,y, calcAngular( img, vec) );
		  }

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(xf, -yf,0.5f );
				vec = normalize( vec );
            sides[ CMS_POSITIVE_Z ]->setPixel(x,y, calcAngular( img, vec) );
		  }

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(-0.5f, -yf,xf );
				vec = normalize( vec );
            sides[ CMS_NEGATIVE_X ]->setPixel(x,y, calcAngular( img, vec) );
		  }

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(xf, yf,-0.5f );
				vec = normalize( vec );
            sides[ CMS_NEGATIVE_Z ]->setPixel(x,y, calcAngular( img, vec) );
		  }

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(xf, -0.5f,-yf );
				vec = normalize( vec );
            sides[ CMS_POSITIVE_Y ]->setPixel(x,y, calcAngular( img, vec) );
		  }

	 for( int x=0;x<size;x++ )
		  for( int y=0;y<size;y++ )
		  {
				float xf = ((float)x+0.5f) / (float)size - 0.5f;
				float yf = ((float)y+0.5f) / (float)size - 0.5f;
				Vector3f vec = makeVector3f(xf, 0.5f,yf );
				vec = normalize( vec );
            sides[ CMS_NEGATIVE_Y ]->setPixel(x,y, calcAngular( img, vec) );
		  }

}


void RGBACubeMap::setupTestMap() {
   delete sides[0];
   sides[0] = new RGBAImage(64,64);
   sides[0]->clear(makeVector4ub(255,0,0,0));
   delete sides[1];
   sides[1] = new RGBAImage(64,64);
   sides[1]->clear(makeVector4ub(0,255,0,0));
   delete sides[2];
   sides[2] = new RGBAImage(64,64);
   sides[2]->clear(makeVector4ub(0,0,255,0));
   delete sides[3];
   sides[3] = new RGBAImage(64,64);
   sides[3]->clear(makeVector4ub(255,255,0,0));
   delete sides[4];
   sides[4] = new RGBAImage(64,64);
   sides[4]->clear(makeVector4ub(255,0,255,0));
   delete sides[5];
   sides[5] = new RGBAImage(64,64);
   sides[5]->clear(makeVector4ub(0,255,255,0));
}

void RGBACubeMap::prefilterDiffuse(card32 mapSize) {
   for (card32 i=0; i<6; i++) {
      RGBAImage *newSide = new RGBAImage(mapSize, mapSize);
      ((RGBAImage*)(sides[i]))->downsampleAA(sides[i]->getSizeX()/mapSize, newSide);
      delete sides[i];
      sides[i] = newSide;
   }

   for (card32 id=0; id<6; id++) {
      for (card32 xd=0; xd<mapSize; xd++) {
         for (card32 yd=0; yd<mapSize; yd++) {

            Vector4f col = NULL_VECTOR4F;
            float32 wSum = 0.0f;

            for (card32 is=0; is<6; is++) {
               for (card32 xs=0; xs<mapSize; xs++) {
                  for (card32 ys=0; ys<mapSize; ys++) {

                     Vector3f sourceVector = getPixelDirection(is,xs,ys);
                     Vector3f destVector = getPixelDirection(id,xd,yd);
                     float32 w = sourceVector * destVector;
                     if (w > 0) {
                        w = w*w*w*w*w;
                        wSum += w;
                        Vector4ub cs = lookupCubeMap(getPixelDirection(is,xs,ys));
                        col += makeVector4f(cs[0], cs[1], cs[2], cs[3])*w;
                     }
                  }
               }
            }

            col /= wSum;
            Vector4ub iCol = makeVector4ub((card8)(col[0]), (card8)(col[1]), (card8)(col[2]), (card8)(col[3]));
            sides[id]->setPixel(xd,yd, iCol);
         }
      }
   }

}



template<class PixelType>
CubeMap<PixelType>::CubeMap() {
   for (card32 i=0; i<6; i++) {
      sides[i] = new TemplatedImage<PixelType>();
   }
}

template<class PixelType>
void CubeMap<PixelType>::assign(const Object* obj, CopyContext *context) {
   if (this == obj) return;
   const CubeMap<PixelType> *arg = dynamic_cast<const CubeMap<PixelType>*>(obj);
   if (arg == NULL) {
      throw PException("CubeMap<PixelType>::assign - obj == NULL or incompatible type.");
   }
   for (card32 i=0; i<6; i++) {
      delete sides[i];
      sides[i] = (TemplatedImage<PixelType>*)(arg->sides[i]->copy(context));
   }
}


template<class PixelType>
void CubeMap<PixelType>::assignSide(card32 index, const TemplatedImage<PixelType> *img) {
   pAssert(index <=5);
   sides[index]->assign(img);
}

template<class PixelType>
CubeMap<PixelType>::~CubeMap() {
   for (card32 i=0; i<6; i++) {
      delete sides[i];
   }
}


template<class PixelType>
PixelType CubeMap<PixelType>::lookup(TemplatedImage<PixelType> *img, float32 x, float32 y) {
   const card32 sizeX = img->getSizeX();
   const card32 sizeY = img->getSizeY();

   x *= sizeX/2;
   y *= sizeY/2;
   x += sizeX/2;
   y += sizeY/2;

   int32 xi = (int32)x;
   int32 yi = (int32)y;

   if (xi<0) xi = 0;
   if (xi>=(int32)sizeX) xi = sizeX-1;
   if (yi<0) yi = 0;
   if (yi>=(int32)sizeY) yi = sizeY-1;

   return img->getPixel(xi, yi);
}

// "cross format: .. -y ..
//                -x +z +x
//                .. +y ..
//                .. -z ..

template<class PixelType>
inline float32 CubeMap<PixelType>::scale(int32 v) {
   const float32 size = (float32)sides[0]->getSizeX();
   const float32 sh = size/2.0f;
   return (((float32)v) - sh) / sh;
}

template<class PixelType>
Vector3f CubeMap<PixelType>::getPixelDirection(CubeMapSide index, int32 x, int32 y) {
   switch (index) {
    case CMS_POSITIVE_X: {
      return makeVector3f(1, scale(y), -scale(x));
    }
    case CMS_NEGATIVE_X: {
      return makeVector3f(-1, scale(y), scale(x));
    }
    case CMS_POSITIVE_Y: {
      return makeVector3f(scale(x), 1, -scale(y));
    }
    case CMS_NEGATIVE_Y: {
      return makeVector3f(scale(x), -1, scale(y));
    }
    case CMS_POSITIVE_Z: {
      return makeVector3f(scale(x), scale(y), 1);
    }
    case CMS_NEGATIVE_Z: {
      return makeVector3f(scale(x), -scale(y), -1);
    }
   }
   return NULL_VECTOR3F; // avoid warning
}


template<class PixelType>
PixelType CubeMap<PixelType>::lookupCubeMap(const Vector3f &dir) {
   float32 maxVal = fabs(dir[0]);
   card32 maxDir = 0;

   float32 comp = fabs(dir[1]);
   if (comp > maxVal) {
      maxVal = comp;
      maxDir = 1;
   }

   comp = fabs(dir[2]);
   if (comp > maxVal) {
      maxVal = comp;
      maxDir = 2;
   }

   bool positive;
   if (dir[maxDir] >= 0) {
      positive = true;
   } else {
      positive = false;
   }

   switch (maxDir) {
    case 0: {
      if (positive) {
         return lookup(sides[CMS_POSITIVE_X], -dir[2]/maxVal,  dir[1]/maxVal);
      } else {
         return lookup(sides[CMS_NEGATIVE_X],  dir[2]/maxVal,  dir[1]/maxVal);
      }
    }
    case 1: {
      if (positive) {
         return lookup(sides[CMS_POSITIVE_Y],  dir[0]/maxVal, -dir[2]/maxVal);
      } else {
         return lookup(sides[CMS_NEGATIVE_Y],  dir[0]/maxVal,  dir[2]/maxVal);
      }
    }
    case 2: {
      if (positive) {
         return lookup(sides[CMS_POSITIVE_Z],  dir[0]/maxVal,  dir[1]/maxVal);
      } else {
         return lookup(sides[CMS_NEGATIVE_Z],  dir[0]/maxVal, -dir[1]/maxVal);
      }
    }
   }
   return lookup(sides[CMS_POSITIVE_X],0,0); // avoid warning
}

template class FOUNDATION_API CubeMap< Vector3ub>;
template class FOUNDATION_API CubeMap< Vector4ub>;
template class FOUNDATION_API CubeMap< StaticVector<card8,1> >;
template class FOUNDATION_API CubeMap< Vector3f >;
template class FOUNDATION_API CubeMap< Vector4f >;
template class FOUNDATION_API CubeMap< StaticVector<float32,1> >;
