#include "stdafx.h"

//---------------------------------------------------------------------------
#include "ExpressionParser.h"
#include "StringHelper.h"
#include "PTypes.h"
//---------------------------------------------------------------------------
#include <iostream>
#include <cmath>
//---------------------------------------------------------------------------
#include "Exceptions.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

using namespace std;
using namespace geoxl;

/****************************************************************************
 * ExpressionParser - default constructor
 ****************************************************************************/
ExpressionParser::ExpressionParser()
{
  symbolTables = NULL;

  operatorChars = "+-*/()=!><&|";

  // defined like the precedence in C++
  // http://www.highscore.de/cpp/einfuehrung/operatoren.html
  operatorPrecedence["&&"] = 2;
  operatorPrecedence["||"] = 2;
  operatorPrecedence["=="] = 8;
  operatorPrecedence["!="] = 8;
  operatorPrecedence[">"] = 9;
  operatorPrecedence[">="] = 9;
  operatorPrecedence["<"] = 9;
  operatorPrecedence["<="] = 9;
  operatorPrecedence["+"] = 11;
  operatorPrecedence["-"] = 11;
  operatorPrecedence["*"] = 12;
  operatorPrecedence["/"] = 12;
  operatorPrecedence["%"] = 12;
  operatorPrecedence["!"] = 13;
  operatorPrecedence["sin"] = 13;
  operatorPrecedence["cos"] = 13;
  operatorPrecedence["sqrt"] = 13;

  // "(", ")" have the highest precedence and are handled separately
}

/****************************************************************************/
//  ExpressionParser - tokenIsOperator [PRIVATE]
//  Returns true if the token given is a valid operator and false otherwise
/****************************************************************************/
bool ExpressionParser::tokenIsOperator(const std::string& token) const
{
  return token == "(" || token == ")" ||
    operatorPrecedence.find(strLCase(token)) != operatorPrecedence.end();
}

/****************************************************************************/
//  ExpressionParser - getNextToken [PRIVATE]
//  Gets the next token from the expression given
/// The token will be removed from the expression and will be returned in
/// token.
/// \param expression (in) The expression where the token is taken from
/// \param isOperator (out) true if token is an operator, false if it is an operand
/// \param token (out) contains the token
/****************************************************************************/
void ExpressionParser::getNextToken(string& expression, bool& isOperator, string& token)
{
  if (expression.size() == 0) return;

  token.clear();
  expression = lTrimStr(expression);
  isOperator = operatorChars.find(expression[0], 0) != string::npos;

  bool checkNextChar = true;
  unsigned pos = 0;
  while (checkNextChar) {
    if (expression[pos] == ' ') break;

    if (operatorChars.find(expression[pos], 0) == string::npos) {
      if (isOperator)
        checkNextChar = false;
      else
        token += expression[pos];
    } else {
      if (isOperator) {
        if (!tokenIsOperator(token)) {
          // Token is not a valid operator yet
          token += expression[pos];
        } else {
          if (tokenIsOperator(token + expression[pos])) {
            // Token will still be a valid operator if the next character is
            // added
            token += expression[pos];
          } else checkNextChar = false;
        }
      } else {
        checkNextChar = false;
      }
    }

    ++pos;
    if (pos >= expression.size()) checkNextChar = false;
  }
  expression.erase(0, token.size());

  // Word operators ("sin", "cos", "sqrt")
  if (!isOperator) {
    if (tokenIsOperator(strLCase(token))) {
      token = strLCase(token);
      isOperator = true;
    }
  }
}

/****************************************************************************/
//  ExpressionParser - handleOperator [PRIVATE]
//  Handles the operator given
/***************************************************************************/
void ExpressionParser::handleOperator(std::string token)
{
  if (!tokenIsOperator(token))
    throw EParserException("Expression Syntax: Invalid operator " + token);

  if (token != "(" && token != ")") {
    while (!operatorStack.empty()) {
      if (operatorPrecedence[token] <= operatorPrecedence[operatorStack.top()])
        handleTopOperator();
      else
        break;
    }
  }

  operatorStack.push(token);
  if (token == ")") handleTopOperator();
}

/****************************************************************************/
//  ExpressionParser - popTopOperand [PRIVATE]
//  Reads the top operand on the operand stack, pops it and returns it
/// If the operand stack is empty, an exception is thrown with the calling
/// operator indicated
/****************************************************************************/
ExpressionNum ExpressionParser::popTopOperand(std::string callingOperator)
{
  if (operandStack.empty())
    throw EParserException("Expression Syntax: Missing operand (operator '"
                        + callingOperator + "')");

  ExpressionNum result = operandStack.top();
  operandStack.pop();
  return result;
}

/****************************************************************************/
//  ExpressionParser - handleTopOperator [PRIVATE]
//  Handles the operator at the top of the operator stack
/// Executes the operator at the top of the stack and pushes the result onto
/// the stack
/****************************************************************************/
bool ExpressionParser::handleTopOperator()
{
  if (operatorStack.empty()) return false;

  string op = operatorStack.top();
  operatorStack.pop();

  // BRACKETS
  if (op == "(") {
    // is handled by ')'
    throw EParserException("Expression Syntax: '(' without ')'");
  }

  if (op == ")") {
    if (operatorStack.empty())
       throw EParserException("Expression Syntax: ')' without '('");

    while (operatorStack.top() != "(") {
      handleTopOperator();

      if (operatorStack.empty()) {
        throw EParserException("Expression Syntax: ')' without '('");
      }

    }
    // remove "(" from operator stack
    operatorStack.pop();
    return true;
  }

  // ARITHMETIC OPERATORS
  if (op == "+") {
    ExpressionNum a,b;
    b = popTopOperand("+");
    a = popTopOperand("+");

    operandStack.push(a + b);
    return true;
  }

  if (op == "-") {
    ExpressionNum a,b;
    b = popTopOperand("-");
    a = popTopOperand("-");

    operandStack.push(a - b);
    return true;
  }

  if (op == "*") {
    ExpressionNum a,b;
    b = popTopOperand("*");
    a = popTopOperand("*");

    operandStack.push(a * b);
    return true;
  }

  if (op == "/") {
    ExpressionNum a,b;
    b = popTopOperand("/");
    a = popTopOperand("/");

    operandStack.push(a / b);
    return true;
  }

  if (op == "%") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("%");
    a = (max_platform_int)popTopOperand("%");

    operandStack.push(a % b);
    return true;
  }

  if (op == "sin") {
    ExpressionNum a = popTopOperand("sin");
    operandStack.push(sin(a));
    return true;
  }

  if (op == "cos") {
    ExpressionNum a = popTopOperand("cos");
    operandStack.push(cos(a));
    return true;
  }

  if (op == "sqrt") {
    ExpressionNum a = popTopOperand("sqrt");
    operandStack.push(sqrt(a));
    return true;
  }

  // BOOLEAN OPERATORS
  if (op == "<=") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("<=");
    a = (max_platform_int)popTopOperand("<=");

    operandStack.push(a <= b);
    return true;
  }

  if (op == "<") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("<");
    a = (max_platform_int)popTopOperand("<");

    operandStack.push(a < b);
    return true;
  }

  if (op == ">=") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand(">=");
    a = (max_platform_int)popTopOperand(">=");

    operandStack.push(a >= b);
    return true;
  }

  if (op == ">") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand(">");
    a = (max_platform_int)popTopOperand(">");

    operandStack.push(a > b);
    return true;
  }

  if (op == "!=") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("!=");
    a = (max_platform_int)popTopOperand("!=");

    operandStack.push(a != b);
    return true;
  }

  if (op == "==") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("==");
    a = (max_platform_int)popTopOperand("==");

    operandStack.push(a == b);
    return true;
  }

  if (op == "&&") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("&&");
    a = (max_platform_int)popTopOperand("&&");

    operandStack.push(a && b);
    return true;
  }

  if (op == "||") {
    max_platform_int a,b;
    b = (max_platform_int)popTopOperand("||");
    a = (max_platform_int)popTopOperand("||");

    operandStack.push(a || b);
    return true;
  }

  if (op == "!") {
    max_platform_int a;
    a = (max_platform_int)popTopOperand("!");

    operandStack.push(!a);
    return true;
  }

  throw EParserException("Expression Syntax: Invalid operator " + op);
}

/****************************************************************************/
//  ExpressionParser - evaluate
//  Evaluates the arithmetic expression and returns its result
/// Variables that are registered in the symbol table handler that is
/// connected to the Expression parser are also evaluated.
/****************************************************************************/
ExpressionNum ExpressionParser::evaluate(string expression)
{
  if (expression == "") {
    throw EParserException("Expression empty");
  }

  ExpressionNum result;
  if (strToNum(expression, result)) return result;

  while (!operandStack.empty()) operandStack.pop();
  while (!operatorStack.empty()) operatorStack.pop();

  bool lastOperator = true;   // was last token an operator?
  while (expression != "") {
    string token;
    bool isOperator;
    getNextToken(expression, isOperator, token);
    if (isOperator) {
      handleOperator(token);
    } else {
      if (!lastOperator)
        throw EParserException("Expression Syntax: Two operands without operators");

      ExpressionNum num;
      if (!strToNum(token, num)) {
        bool symbolFound = false;
        if (symbolTables != NULL)
          symbolFound = symbolTables->getValue(token, num);

        if (!symbolFound)
          throw EParserOperandException("Expression Syntax: Invalid operand " + token);
      }

      operandStack.push(num);
    }

    lastOperator = isOperator;
  }

  while (!operatorStack.empty()) handleTopOperator();

  if (operandStack.empty())
      throw EParserException("Expression Syntax: Not enough operands");
  result = operandStack.top();

  operandStack.pop();
  if (!operandStack.empty())
      throw EParserException("Expression Syntax: Too many operands");

  return result;
}

