//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "FileBlockAllocator.h"
//---------------------------------------------------------------------------
#include "PAssert.h"
#include "DebugOutput.h"
#include "LargeBinaryFile.h"
#include "Random.h"
//---------------------------------------------------------------------------
#ifdef _DEBUG
   #include "EventCounterDisplay.h"
#endif
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


#ifdef _DEBUG
struct CheckAllocsEntry {
   FileBlockAllocator *inst;
   string filename;
};

struct CheckAllocs {
   vector<CheckAllocsEntry> checkAllocs;

   void addAndCheck(FileBlockAllocator *inst, string f) {
      for (card32 i=0; i<checkAllocs.size(); i++) {
         if (checkAllocs[i].filename == f) {
            pAssert(false);
         }
         CheckAllocsEntry e;
         e.inst = inst;
         e.filename = f;
         checkAllocs.push_back(e);
      }
   }

   void remove(FileBlockAllocator *inst, string f) {
      int result = -1;
      for (card32 i=0; i<checkAllocs.size(); i++) {
         if (checkAllocs[i].filename == f) {
            if (checkAllocs[i].inst != inst) {
               pAssert(false);
            }
            if (result != -1) {
               pAssert(false);
            }
            result = i;
         }
         if (result != -1) {
            vector<CheckAllocsEntry>::iterator pos = checkAllocs.begin();
            pos += result;
            checkAllocs.erase(pos);
         }
      }
   }
};

static CheckAllocs checkAllocs;
#endif



const char *FileBlockAllocator::HEADER_STRING = "GEOXLFileBlockAllocator";
static const card64 ALLOCATED_BLOCK = 0xFFFFFFFFFFFFFFFF;

FileBlockAllocator::FileBlockAllocator() {
   file = NULL;
   filename = "";

   fileSize = 0;
   strcpy(rootBlock.headerString, HEADER_STRING);
   rootBlock.version = FILE_VERSION;
   rootBlock.firstFreeBlock = 0;
   rootBlock.blockSize = 0;
   rootBlock.allocs = 0;
   rootBlock.deletes = 0;
   rootBlock.numBlocksInUse = 0;
   rootBlock.status = 0;

   showDebugEvents = true;
}

void FileBlockAllocator::createNewFile(const char *filename, const card64 blockSize, bool showDebugEvents) {
   #ifdef _DEBUG
   checkAllocs.addAndCheck(this, filename);
   #endif

   pAssert(file == NULL);
   pAssert(blockSize >= sizeof(rootBlock));

   this->filename = filename;
   this->showDebugEvents = showDebugEvents;

   strcpy(rootBlock.headerString, HEADER_STRING);
   rootBlock.version = FILE_VERSION;
   rootBlock.firstFreeBlock = 0;
   rootBlock.blockSize = blockSize;
   rootBlock.allocs = 0;
   rootBlock.deletes = 0;
   rootBlock.numBlocksInUse = 0;
   rootBlock.status = STATUS_OPENED;
   rootBlock.id.createNewID();

   file = new LargeBinaryFile(filename, true, true);
   file->setFileSize(blockSize);
   fileSize = blockSize;

   writeRootBlock();
}

void FileBlockAllocator::openFile(const char *filename, bool showDebugEvents) {
   #ifdef _DEBUG
   checkAllocs.addAndCheck(this, filename);
   #endif
   pAssert(file == NULL);

   this->filename = filename;
   this->showDebugEvents = showDebugEvents;
   file = new LargeBinaryFile(filename, true, true);

   readRootBlock();

   fileSize = file->getFileSize();
   if (string(rootBlock.headerString) != string(HEADER_STRING) || rootBlock.version != FILE_VERSION) {
      throw EIOException("FileBlockAllocator::openFile - this is no valid geoxl block file.");
   }
   if (rootBlock.status != STATUS_CLOSED) {
      warning(string("FileBlockAllocator::openFile - rootBlock indicates that the file ")
              + string(filename)
              + string(" has not been shutdown properly."));
   }
}

GloballyUniqueID FileBlockAllocator::getFileGUID() {
   pAssert(file != NULL);
   return rootBlock.id;
}


void FileBlockAllocator::close() {
   pAssert(file != NULL);
   rootBlock.status = STATUS_CLOSED;
   writeRootBlock();
   file->sync();
   delete file;
   file = NULL;
   #ifdef _DEBUG
   checkAllocs.remove(this, filename);
   #endif
}


void FileBlockAllocator::readRootBlock() {
   pAssert(file != NULL);
   file->seek(0);
   file->readBuffer(&rootBlock, sizeof(rootBlock));
#ifdef _DEBUG
   if (showDebugEvents) {
      EventCounterDisplay::NewValueStepOnChange e;
      e.value = (int32)rootBlock.numBlocksInUse;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "numBlocksInUse", &e);
      e.value *= (int32)rootBlock.blockSize;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "allocatedBlockBytes", &e);
   }
#endif
}

void FileBlockAllocator::writeRootBlock() {
   pAssert(file != NULL);
   file->seek(0);
   file->writeBuffer(&rootBlock, sizeof(rootBlock));

#ifdef _DEBUG
   if (showDebugEvents) {
      EventCounterDisplay::NewValueStepOnChange e;
      e.value = (int32)rootBlock.numBlocksInUse;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "numBlocksInUse", &e);
      e.value *= (int32)rootBlock.blockSize;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "allocatedBlockBytes", &e);
   }
#endif
}

card64 FileBlockAllocator::allocBlock() {
   pAssert(file != NULL);
   card64 firstFree = rootBlock.firstFreeBlock;
   card64 result;
   if (firstFree == 0) {
      result = createNewBlock();
   } else {
      result = firstFree;
      pAssert(firstFree*rootBlock.blockSize < fileSize);
      file->seek(firstFree * rootBlock.blockSize);
      card64 nextFree = 0;
      file->readBuffer(&nextFree, sizeof(nextFree));
      pAssert(nextFree != ALLOCATED_BLOCK);
      rootBlock.firstFreeBlock = nextFree;
      file->seek(firstFree * rootBlock.blockSize);
      file->writeBuffer(&ALLOCATED_BLOCK, sizeof(ALLOCATED_BLOCK));
   }
   rootBlock.allocs++;
   rootBlock.numBlocksInUse++;

#ifdef _DEBUG
   if (showDebugEvents) {
      EventCounterDisplay::Increment e;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "allocs", &e);
   }
#endif

   writeRootBlock();
   pAssert(result != 0);
   return result;
}

card64 FileBlockAllocator::createNewBlock() {
   pAssert((fileSize % rootBlock.blockSize) == 0);
   card64 result = fileSize / rootBlock.blockSize;
   file->seek(fileSize);
   file->writeBuffer(&ALLOCATED_BLOCK, sizeof(ALLOCATED_BLOCK));
   file->setFileSize((fileSize/rootBlock.blockSize+1)*rootBlock.blockSize);
   fileSize = (fileSize/rootBlock.blockSize+1)*rootBlock.blockSize;
   return result;
}


void FileBlockAllocator::deleteBlock(card64 blockNo) {
   pAssert(file != NULL);
   file->seek(blockNo * rootBlock.blockSize);
   card64 blockStatus;
   file->readBuffer(&blockStatus, sizeof(blockStatus));
   if (blockStatus != ALLOCATED_BLOCK) {
      throw EIOException("BlockedMultiObjectStream::deleteBlock - block has not been not allocated.");
   }
   file->seek(blockNo * rootBlock.blockSize);
   file->writeBuffer(&rootBlock.firstFreeBlock, sizeof(rootBlock.firstFreeBlock));
   rootBlock.firstFreeBlock = blockNo;
   rootBlock.deletes++;
   rootBlock.numBlocksInUse--;
   writeRootBlock();

#ifdef _DEBUG
   if (showDebugEvents) {
      EventCounterDisplay::Increment e;
      debugOutput.displayEvent(EventCounterDisplay::getClass(), "deletes", &e);
   }
#endif
}


void FileBlockAllocator::writeBlock(card64 blockNo, const card8 *mem, max_platform_card size) {
   pAssert(file != NULL);
   pAssert((blockNo+1)*rootBlock.blockSize <= fileSize);

   #ifdef _DEBUG
    file->seek(blockNo*rootBlock.blockSize);
    card64 check;
    file->readBuffer(&check, sizeof(check));
    pAssert(check == ALLOCATED_BLOCK);
   #endif

   file->seek(blockNo*rootBlock.blockSize + (card64)BLOCK_OVERHEAD);
   if (size + BLOCK_OVERHEAD > rootBlock.blockSize) {
      throw EIOException("FileBlockAllocator::writeBlock() - block size exceeded.");
   }
   if (mem != NULL && size != 0) {
      file->writeBuffer(mem, size);
   }
}

void FileBlockAllocator::readBlock(card64 blockNo, card8 *mem, max_platform_card size) {
   pAssert(file != NULL);
   pAssert((blockNo+1)*rootBlock.blockSize <= fileSize);

   #ifdef _DEBUG
    file->seek(blockNo*rootBlock.blockSize);
    card64 check;
    file->readBuffer(&check, sizeof(check));
    pAssert(check == ALLOCATED_BLOCK);
   #endif

   file->seek(blockNo*rootBlock.blockSize + (card64)BLOCK_OVERHEAD);
   if (size + BLOCK_OVERHEAD > rootBlock.blockSize) {
      throw EIOException("FileBlockAllocator::writeBlock() - block size exceeded.");
   }
   if (mem != NULL && size != 0) {
      file->readBuffer(mem, size);
   }
}

void FileBlockAllocator::sync() {
   file->sync();
}


void FileBlockAllocator::dumpDebugStatistics() {
   pAssert(file != NULL);
   debugOutput << "rootBlock:"; debugOutput.cr();
   debugOutput << "----------"; debugOutput.cr();
   debugOutput << "rootBlock.headerString:   " << rootBlock.headerString; debugOutput.cr();
   debugOutput << "rootBlock.version:        " << (int)rootBlock.version; debugOutput.cr();
   debugOutput << "rootBlock.firstFreeBlock: " << (int)rootBlock.firstFreeBlock; debugOutput.cr();
   debugOutput << "rootBlock.blockSize:      " << (int)rootBlock.blockSize; debugOutput.cr();
   debugOutput << "rootBlock.allocs:         " << (int)rootBlock.allocs; debugOutput.cr();
   debugOutput << "rootBlock.deletes:        " << (int)rootBlock.deletes; debugOutput.cr();
   debugOutput << "rootBlock.numBlocksInUse: " << (int)rootBlock.numBlocksInUse; debugOutput.cr();
   debugOutput << "rootBlock.status:         " << (int)rootBlock.status; debugOutput.cr();
}


FileBlockAllocator::~FileBlockAllocator() {
   if (file != NULL) {
      close();
   }
}


bool FileBlockAllocator::classTest() {
   static const card32 TEST_SIZE = 100;
   static const card32 TESTS_NUMBER = 1000;
   static const card32 MIN_FILL = 100;
   static const card32 BLOCK_SIZE = 1024;

   int actions[TESTS_NUMBER];
   for (card32 i=0; i<TESTS_NUMBER; i++) {
      actions[i] = randomNumber(3);
   }
   int rndNumbers[TESTS_NUMBER];
   for (card32 i=0; i<TESTS_NUMBER; i++) {
      rndNumbers[i] = randomNumber(10000);
   }

   card32 numErrors = 0;

   card8 blockBuffer[BLOCK_SIZE];

   card64 blocks[TEST_SIZE];
   for (card32 i=0; i<TEST_SIZE; i++) {
      blocks[i] = 0;
   }
   card32 numBlocks = 0;

   FileBlockAllocator allocTest;
   allocTest.createNewFile("testFile.tmp", BLOCK_SIZE);
   static const card32 DATA_SIZE = (card32)allocTest.getDataSizePerBlock();

   for (card32 i=0; i<TESTS_NUMBER; i++) {
      card32 action = actions[i];
      if (i < MIN_FILL || i > TESTS_NUMBER - MIN_FILL) action = 0;
      switch (action) {
       case 0: { // create block
         if (numBlocks < TEST_SIZE) {
            blocks[numBlocks] = allocTest.allocBlock();
            pAssert(blocks[numBlocks] != 0);
            srand(numBlocks);
            for (card32 i=0; i<DATA_SIZE; i++) {
               blockBuffer[i] = (card8)randomNumber(256);
            }
            allocTest.writeBlock(blocks[numBlocks], blockBuffer, DATA_SIZE);
            debugOutput << "Write block No. " << numBlocks << " in file block " << (int)blocks[numBlocks] << ".";
            debugOutput.cr();
            numBlocks++;
         }
         break;
       }
       case 1: { // delete block
         if (numBlocks > 0) {
            allocTest.deleteBlock(blocks[numBlocks-1]);
            debugOutput << "Delete block No. " << (numBlocks - 1) << " (file block " << ((int)blocks[numBlocks - 1]) << ").";
            debugOutput.cr();
            numBlocks--;
         }
         break;
       }
       case 2: { // read and check block
         if (numBlocks < TEST_SIZE) {
            card32 testBlock = rndNumbers[i] % numBlocks;
            allocTest.readBlock(blocks[testBlock], blockBuffer, DATA_SIZE);
            srand(testBlock);
            for (card32 i=0; i<DATA_SIZE; i++) {
               if (blockBuffer[i] != (card8)randomNumber(256)) {
                  debugOutput << "test failed in block " << (testBlock) << ", file block " << ((int)blocks[testBlock]) << " byte number " << i;
                  debugOutput.cr();
                  numErrors++;
               }
            }
            debugOutput << "Test block No. " << (testBlock) << ", file block " << ((int)blocks[testBlock]) << ", succeded.";
            debugOutput.cr();
         }
         break;
       }
      }
   }

   allocTest.dumpDebugStatistics();

   allocTest.close();

   allocTest.openFile("testFile.tmp");
   pAssert(allocTest.getBlockSize() == BLOCK_SIZE);

   for (card32 i=0; i<numBlocks; i++) {
      allocTest.readBlock(blocks[i], blockBuffer, DATA_SIZE);
      srand(i);
      for (card32 j=0; j<DATA_SIZE; j++) {
         if (blockBuffer[j] != (card8)randomNumber(256)) {
            debugOutput << "test failed in block " << (j) << ", file block " << ((int)blocks[j]) << " byte number " << j;
            debugOutput.cr();
            numErrors++;
         }
      }
      debugOutput << "Test block No. " << (i) << ", file block " << ((int)blocks[i]) << ", succeded.";
      debugOutput.cr();
   }

   for (card32 i=0; i<numBlocks; i++) {
      allocTest.deleteBlock(blocks[i]);
   }
   debugOutput << "--------------------------------------";
   debugOutput.cr();
   debugOutput << "Test completed. Errors: " << numErrors;
   debugOutput.cr();
   debugOutput << "--------------------------------------";
   debugOutput.cr();
   allocTest.dumpDebugStatistics();
   allocTest.close();
   return numErrors == 0;
}
