//---------------------------------------------------------------------------
#ifndef FileBlockAllocatorH
#define FileBlockAllocatorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
#include "LargeBinaryFile.h"
#include "PAssert.h"
#include "GloballyUniqueID.h"
//---------------------------------------------------------------------------
#include <string>
//---------------------------------------------------------------------------

namespace geoxl {
   using namespace std;


class FileBlockAllocator {
 private:
   LargeBinaryFile *file;
   card64 fileSize;
   string filename;
   bool showDebugEvents;

   static const char *HEADER_STRING;
   static const card32 HEADER_STRING_SIZE = 22;
   static const card32 FILE_VERSION = 1;
   static const card32 STATUS_CLOSED = 0;
   static const card32 STATUS_OPENED = 1;
   static const max_platform_card BLOCK_OVERHEAD = 8; // <-- sizeof(card64)

   struct RootBlock;
   friend RootBlock;

   struct RootBlock {
      char headerString[HEADER_STRING_SIZE];
      card32 version;
      card64 firstFreeBlock;
      card64 blockSize;
      card64 allocs;
      card64 deletes;
      card64 numBlocksInUse;
      card32 status;
      GloballyUniqueID id;
   };

   RootBlock rootBlock;

   void readRootBlock();
   void writeRootBlock();
   card64 createNewBlock();

 public:
   FileBlockAllocator();

   void createNewFile(const char *filename, const card64 blockSize = 32768, bool showDebugEvents = true);
   void openFile(const char *filename, bool showDebugEvents = true);

   inline card64 getBlockSize() {return rootBlock.blockSize;}
   inline card64 getDataSizePerBlock() {return rootBlock.blockSize - sizeof(card64);}
   static inline card64 getPerBlockOverhead() {return sizeof(card64);}

   card64 allocBlock();
   void deleteBlock(card64 blockNo);

   void writeBlock(card64 blockNo, const card8 *mem, max_platform_card size);
   template <typename HeaderType>
   void writeBlock(card64 blockNo, HeaderType header, const card8 *mem, max_platform_card size);

   void readBlock(card64 blockNo, card8 *mem, max_platform_card size);
   template <typename HeaderType>
   void readBlock(card64 blockNo, HeaderType &header, card8 *mem, max_platform_card size);

   void dumpDebugStatistics();

   void sync();

   GloballyUniqueID getFileGUID();

   void close();

   ~FileBlockAllocator();

   /// performs functionality test (debugging)
   static bool classTest();
};


// ----


template <typename HeaderType>
void FileBlockAllocator::writeBlock(card64 blockNo, HeaderType header, const card8 *mem, max_platform_card size) {
   pAssert((blockNo+1)*rootBlock.blockSize <= fileSize);
   file->seek(blockNo*rootBlock.blockSize + (card64)BLOCK_OVERHEAD);
   if (sizeof(HeaderType) + size + BLOCK_OVERHEAD > rootBlock.blockSize) {
      throw EIOException("FileBlockAllocator::writeBlock() - block size exceeded.");
   }
   file->writeBuffer(&header, sizeof(HeaderType));
   if (mem != NULL && size != 0) {
      file->writeBuffer(mem, size);
   }
}

template <typename HeaderType>
void FileBlockAllocator::readBlock(card64 blockNo, HeaderType &header, card8 *mem, max_platform_card size) {
   pAssert((blockNo+1)*rootBlock.blockSize <= fileSize);
   file->seek(blockNo*rootBlock.blockSize + (card64)BLOCK_OVERHEAD);
   if (sizeof(HeaderType) + size + BLOCK_OVERHEAD > rootBlock.blockSize) {
      throw EIOException("FileBlockAllocator::writeBlock() - block size exceeded.");
   }
   file->readBuffer(&header, sizeof(HeaderType));
   if (mem != NULL && size != 0) {
      file->readBuffer(mem, size);
   }
}


}

#endif
