//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "FreeCameraController.h"
#include "FixedArrayClassProperty.h"
#include "NumericalClassProperty.h"
#include "NumericalClassProperty_templCode.h"
#include "ObjectViewsTable.h"
#include "DebugOutput.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( FreeCameraController, 1 )
{
	BEGIN_CLASS_INIT( FreeCameraController );
	ADD_FLOAT32_PROP( speed, 0 );
	ADD_VECTOR3F_PROP( eye, 1 );
	ADD_VECTOR3F_PROP( ref, 1 );
	ADD_VECTOR3F_PROP( up, 1 );
}

/**
 * Create new object
 */
FreeCameraController::FreeCameraController()
{
   state = neutral;
   speed = 0.01f;
   eye = makeVector3f( 0, 0, -5 );
   ref = makeVector3f( 0,0,0 );
   up = makeVector3f( 0,1,0 );
}

/**
 * Set camera according to controller state
 */
void FreeCameraController::calcCamera()
{
   camera->setup( eye, ref, up );
}

/**
 * Controller state has changed, update camera
 */
void FreeCameraController::changed()
{
   if (camera == NULL) return;
   ObjectViewsTable::update( camera );
   calcCamera();
   cameraChange();
}


/**
 * Key down event
 */
void FreeCameraController::keyDown( GeneralKey key )
{
	if (camera == NULL) return;
	if ( key.getKey() == keyAlphaNum ){
		float32 d = norm( camera->getPosition() - camera->getLookAt() );
		if ( key.getkeyAlphaNumKey() == 70 ){
			 // key = 'f', move ref to the left
			 Vector3f viewDir = camera->getPosition() - camera->getLookAt();
			 Vector3f sideDir = viewDir.crossProduct( camera->getUp() );
			 sideDir = normalize( sideDir );
			 ref = camera->getLookAt() + sideDir * d * -0.05f;
			 camera->setLookAt( ref );
		} else if ( key.getkeyAlphaNumKey() == 71 ){
			 // key = 'g', move ref to the right
			 Vector3f viewDir = camera->getPosition() - camera->getLookAt();
			 Vector3f sideDir = viewDir.crossProduct( camera->getUp() );
			 sideDir = normalize( sideDir );
			 ref = camera->getLookAt()+ sideDir * d * 0.05f;
			 camera->setLookAt( ref );
		} else if ( key.getkeyAlphaNumKey() == 84 ){
			// key = 't', move ref up
			ref = camera->getLookAt() + camera->getUp() * d * -0.05f;
			camera->setLookAt( ref );
		} else if ( key.getkeyAlphaNumKey() == 86 ){
			 // key = 'v', move ref down
			ref = camera->getLookAt() + camera->getUp() * d * 0.05f;
			 camera->setLookAt( ref );
		} else if ( key.getkeyAlphaNumKey() == 67 ){
			 // key = 'c', move ref closer
			 Vector3f dir = camera->getPosition() - camera->getLookAt();
			 dir = normalize( dir );
			 ref = camera->getLookAt() + dir * d * 0.05f;
			 camera->setLookAt( ref );
		} else if ( key.getkeyAlphaNumKey() == 90 ){
			 // key = 'z', move ref further away
			 Vector3f dir = camera->getPosition() - camera->getLookAt();
			 dir = normalize( dir );
			 ref = camera->getLookAt() + dir * d * -0.05f;
			 camera->setLookAt( ref );
		}
	} else if ( key.getKey() == keyPageUp ){
		// Zoom in
		float32 d = 0.95f;
		eye = camera->getLookAt() + (camera->getPosition() - camera->getLookAt() ) * d;
		camera->setPosition( eye );
		changed();
	} else if ( key.getKey() == keyPageDown ){
		// Zoom out
		float32 d = 1.05f;
		eye = camera->getLookAt() + ( camera->getPosition() - camera->getLookAt() ) * d;
		camera->setPosition( eye );
		changed();
	}
	/// Update camera
	changed();
}

/**
 * Key released event
 */
void FreeCameraController::keyUp( GeneralKey key )
{

}

/**
 * Set reference point in camera controller
 */
void FreeCameraController::setup( const Vector3f& position, const Vector3f& ref, const Vector3f& up )
{
	if (camera == NULL) return;
	this->eye = position;
	this->ref = ref;
	this->up = up;
	camera->setPosition( position );
	camera->setLookAt( ref );
	camera->setUp( up );
	changed();
}

/**
 * Mouse button pressed event
 */
void FreeCameraController::mouseDown( int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState )
{
   if (modifiersState.getShift()) {
      if (buttonsState.getLeft()) {
         state = dragDist;
         lastX = x;
         lastY = y;
      }
   } else {
      if (buttonsState.getLeft()) {
         state = dragAngles;
         lastX = x;
         lastY = y;
      }
   }
}

/**
 * Rotate the camera around axis ( ref ^ up )
 */
void FreeCameraController::rotateUp( float s )
{
	if (camera == NULL) return;
	up = normalize( camera->getUp() );
	camera->setUp( up );
	Vector3f view = camera->getLookAt() - camera->getPosition();
	view = normalize( view );
	Vector3f rotationAxis = camera->getUp().crossProduct( view );
	Matrix3f rotation = makeRotVector3f( rotationAxis, speed * s );
	eye = ( rotation * ( camera->getPosition() - camera->getLookAt() ) ) + camera->getLookAt();
	camera->setPosition( eye );
	up = view.crossProduct( rotationAxis );
	camera->setUp( up );
}

/**
 * Rotate the camera around axis up
 */
void FreeCameraController::rotateSide( float s )
{
	if (camera == NULL) return;
	up = normalize( camera->getUp() );
	camera->setUp( up );
	Matrix3f rotation = makeRotVector3f( camera->getUp(), -speed * s );
	eye = ( rotation * ( camera->getPosition() - camera->getLookAt() ) ) + camera->getLookAt();
	camera->setPosition( eye );
}


void FreeCameraController::setYRotationAngle( float32 alpha )
{
	float32 d = norm( ref - eye );
	up = makeVector3f( 0,1,0 );
	ref = makeVector3f( 0,0,0 );
	eye = makeVector3f( cos( alpha ), 0, sin( alpha ) ) * d;
	camera->setUp( up );
	camera->setPosition( eye );
	camera->setLookAt( ref );
}

/**
 * Mouse has moved
 */
void FreeCameraController::mouseMoved( int32 x, int32 y )
{
	int32 deltaX = x - lastX;
	int32 deltaY = y - lastY;
	lastX = x;
	lastY = y;
	if (state == dragAngles) {
		rotateUp( (float)deltaY );
		rotateSide( (float)deltaX );
		changed();
	} else if (state == dragDist) {
		float32 d = (float)deltaY / 100.0f;
		eye = camera->getPosition() + ( camera->getLookAt() - camera->getPosition() ) * d;
		camera->setPosition( eye );
		changed();
	}
}

/**
 * Mouse button released
 */
void FreeCameraController::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState)
{
   if ( buttonsState.getLeft() ) {
      state = neutral;
   }
}

void geoxl::FreeCameraController::mouseWheelRotated( int32 rotatedDelta, Modifiers modifiersState )
{
	
	if (rotatedDelta>0)
	{
		//Zoom in
		float32 d = 0.9f;
		eye = camera->getLookAt() + (camera->getPosition() - camera->getLookAt() ) * d;
		camera->setPosition( eye );
		changed();
	} else {
		// Zoom out
		float32 d = 1.1f;
		eye = camera->getLookAt() + ( camera->getPosition() - camera->getLookAt() ) * d;
		camera->setPosition( eye );
		changed();
	}
}

/**
 * Resize widget
 */
void FreeCameraController::areaResize( card32 width, card32 height )
{
}

/**
 * Update ?!?
 */
void FreeCameraController::update()
{
}

/**
 * Apply controller to camera
 */
void FreeCameraController::apply()
{
	if (camera == NULL) return;
   ObjectViewsTable::update( camera );
   calcCamera();
   cameraChange();
}

/**
 * Reset ?!?
 */
void FreeCameraController::setToCamera( Camera* cam, Vector3f newCenter )
{
}

/** 
 * Assign camera
 */
void FreeCameraController::bindCamera( Camera *cam )
{
	if ( cam != NULL ){
		cam->setPosition( eye );
		cam->setLookAt( ref );
		cam->setUp( up );
	}
	CameraController::bindCamera( cam );
}

