//---------------------------------------------------------------------------

#include "stdafx.h"


#include "ManagedMemory.h"
#include "Object.h"
#include "DebugOutput.h"
#include <algorithm>

//---------------------------------------------------------------------------



using namespace geoxl;

ManagedMemory *geoxl::managedMemory;

ManagedMemory::ManagedMemory(SIZE_T size)
{
  // Reserve memory
  basePtr = VirtualAlloc(NULL, size, MEM_RESERVE, PAGE_NOACCESS);
  if (basePtr == NULL) throw EManagedMemory("Could not reserve memory");
  nextFreePage = (DWORD)basePtr;

  // Get page size
  SYSTEM_INFO sSysInfo;
  GetSystemInfo(&sSysInfo);
  DWORD pageSize = sSysInfo.dwPageSize;
  pageModMask = pageSize-1;

  pageSizePow = 0;
  while (!(pageSize & 1)) {
    pageSize = pageSize >> 1;
    ++pageSizePow;
  }
  pAssertMsg(pageSize == 1, "System page size is not a power of two");

  totalPages = size >> pageSizePow;
  if ((size & pageModMask) > 0) ++totalPages;
  freePages = totalPages;
  totalSize = totalPages << pageSizePow;
  maxAllocationSize = totalSize;

  #ifdef _DEBUG
  defragCounter = 0;
  #endif
}

ManagedMemory::DataPtr ManagedMemory::allocate(SIZE_T size)
{
  pAssertMsg(size > 0, "Cannot allocate 0 bytes. Use free() instead.");
  DataPtr ptr;

  if ((freePages << pageSizePow) < size) throw EManagedMemory("Out of memory");
  if (maxAllocationSize < size)
    defragment();

  SIZE_T numPages = size >> pageSizePow;
  if ((size & pageModMask) > 0) ++numPages;

  DataListElem listElem;
  listElem.ptr = VirtualAlloc((LPVOID)nextFreePage, size, MEM_COMMIT, PAGE_READWRITE);
  listElem.numPages = numPages;
  dataPointers.push_back(listElem);

  nextFreePage += numPages << pageSizePow;
  freePages -= numPages;
  maxAllocationSize -= numPages << pageSizePow;

  ptr = dataPointers.end();
  --ptr;
  return ptr;
}

ManagedMemory::DataPtr ManagedMemory::reallocate(ManagedMemory::DataPtr dataPtr, SIZE_T size)
{
  pAssertMsg(size > 0, "Cannot allocate 0 bytes. Use free() instead.");
  DataPtr ptr;
  SIZE_T newNumPages = size >> pageSizePow;
  if ((size & pageModMask) > 0) ++newNumPages;

  // Is reallocation needed? Check if new size is still within the page borders
  if ((dataPtr->numPages << pageSizePow) >= size) {
    // New size is smaller -> free memory pages
    if (newNumPages < dataPtr->numPages) {
      VirtualFree((LPVOID)((DWORD)dataPtr->ptr + (newNumPages << pageSizePow)),
                  (dataPtr->numPages-newNumPages) << pageSizePow, MEM_DECOMMIT);
      freePages += dataPtr->numPages-newNumPages;

      // Is it the last pointer in the list? If yes, change nextFreePage accordingly
      ptr = dataPtr;
      ++ptr;
      if (ptr == dataPointers.end()) {
        nextFreePage -= (dataPtr->numPages-newNumPages) << pageSizePow;
        maxAllocationSize += (dataPtr->numPages-newNumPages) << pageSizePow;
      }

      dataPtr->numPages = newNumPages;
    }
    return dataPtr;
  }

  // Is the pointer that is to be reallocated at the end of our memory area?
  // If yes (and enough free pages at the end left), just allocate more pages
  ptr = dataPtr;
  ++ptr;
  if (ptr == dataPointers.end()) {
    if ((DWORD)dataPtr->ptr + (newNumPages << pageSizePow) <
        (DWORD)basePtr + totalSize) {
      // enough free pages, so just allocate more memory
      LPVOID newPtr = VirtualAlloc(dataPtr->ptr, newNumPages << pageSizePow, MEM_COMMIT, PAGE_READWRITE);
      pAssert(newPtr == dataPtr->ptr);
      freePages -= newNumPages - dataPtr->numPages;
      nextFreePage += (newNumPages - dataPtr->numPages) << pageSizePow;
      maxAllocationSize -= (newNumPages - dataPtr->numPages) << pageSizePow;
      dataPtr->numPages = newNumPages;
      return dataPtr;
    }
  }             

  // None of these cases, so move data to new address
  ptr = allocate(size);
  SIZE_T bytesToMove = min(size, dataPtr->numPages << pageSizePow);
  MoveMemory(ptr->ptr, dataPtr->ptr, bytesToMove);
  free(dataPtr);
  return ptr;
}

void ManagedMemory::free(ManagedMemory::DataPtr dataPtr)
{
  VirtualFree(dataPtr->ptr, dataPtr->numPages << pageSizePow, MEM_DECOMMIT);
  freePages += dataPtr->numPages;

  DataPtr p = dataPtr;
  ++p;
  if (p == dataPointers.end()) {
    if (dataPtr == dataPointers.begin()) {
      nextFreePage = (DWORD)basePtr;
      maxAllocationSize = totalSize;
    } else {
      p = dataPtr;
      --p;

      nextFreePage = (DWORD)p->ptr + (p->numPages << pageSizePow);
      maxAllocationSize = totalSize - (nextFreePage - (DWORD)basePtr);
    }
  }

  dataPointers.erase(dataPtr);
}

void ManagedMemory::defragment()
{
  /*checkConsistency();*/
  #ifdef _DEBUG
  ++defragCounter;
  #endif

  nextFreePage = (DWORD)basePtr;
  //VirtualAlloc(basePtr, (totalPages-freePages) << pageSizePow, MEM_COMMIT, PAGE_READWRITE);
  for (DataPtr p = dataPointers.begin(); p != dataPointers.end(); ++p) {
    if ((LPVOID)nextFreePage != p->ptr) {
      VirtualAlloc((LPVOID)nextFreePage, p->numPages << pageSizePow, MEM_COMMIT, PAGE_READWRITE);
      MoveMemory((LPVOID)nextFreePage, p->ptr, p->numPages << pageSizePow);
      VirtualFree(p->ptr, p->numPages << pageSizePow, MEM_DECOMMIT);
      p->ptr = (LPVOID)nextFreePage;
    }

    nextFreePage += p->numPages << pageSizePow;
  }
  maxAllocationSize = freePages << pageSizePow;
  //VirtualFree((LPVOID)nextFreePage, maxAllocationSize, MEM_DECOMMIT);

  for (list<MMPointerChangeIntf*>::iterator it = registeredObjects.begin(); it != registeredObjects.end(); ++it)
    (*it)->handlePointersChanged();
}

void ManagedMemory::registerObject(MMPointerChangeIntf* obj)
{
  registeredObjects.push_back(obj);
}

void ManagedMemory::unRegisterObject(MMPointerChangeIntf* obj)
{
  list<MMPointerChangeIntf*>::iterator elem;
  for (elem = registeredObjects.begin(); elem != registeredObjects.end(); ++elem)
    if (*elem == obj) break;
  #ifdef _DEBUG
  if (elem == registeredObjects.end())
    warning("ManagedMemory: Trying to unregister object that is not registered");
  #endif
  registeredObjects.erase(elem);
}


#ifdef _DEBUG
void ManagedMemory::checkConsistency()
{
  // Check value of freePages
  SIZE_T checkFreePages = totalPages;
  for (DataPtr it = dataPointers.begin(); it != dataPointers.end(); ++it) {
    checkFreePages -= it->numPages;
  }
  pAssertMsg(freePages == checkFreePages, "Value of freePages is wrong");

  // Check for overlap
  DWORD lastEnd = (DWORD)basePtr;
  DWORD lastStart = NULL;
  for (DataPtr it = dataPointers.begin(); it != dataPointers.end(); ++it) {
    pAssertMsg((DWORD)it->ptr >= lastEnd, "Data overlap");
    pAssertMsg((DWORD)it->ptr > lastStart, "Data not sorted");
    lastEnd = (DWORD)it->ptr + (it->numPages << pageSizePow);
    lastStart = (DWORD)it->ptr;
  }
  pAssertMsg(nextFreePage == lastEnd, "nextFreePage is wrong");

  // Check if maxAllocationSize is correct
  SIZE_T checkMaxAlloc = totalSize - (nextFreePage - (DWORD)basePtr);
  pAssertMsg(maxAllocationSize == checkMaxAlloc, "maxAllocationSize inconsistent with nextFreePage");
}
#endif

ManagedMemory::~ManagedMemory()
{
  VirtualFree(basePtr, 0, MEM_RELEASE);
}