//---------------------------------------------------------------------------
#ifndef MarchingCubesH
#define MarchingCubesH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <vector>
#include <map>
#include "BoundingBox.h"
#include "LinearAlgebra.h"
#include "pAssert.h"
#include "Object.h"
#include "Plane.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class FOUNDATION_API MarchingCubes
{
public:

   MarchingCubes();
//   typedef float (Object::* EvaluateFunctionMethod)(float x,float y,float z) const;
   typedef float (* EvaluateFunctionMethod)(void *context, float x,float y,float z);
   typedef bool (* IsBoundingBoxEmptyMethod)(void *context, const BoundingBox3f &bb);

private:

   card32 getVertexIndex( const Vector3i &coord );
   int getEdgeVertexIndex( card32 cellIndex, card32 edge );
   void setEdgeVertexIndex( card32 cellIndex, card32 edge, card32 index );
   void addTriangle( const Vector3i &cellCoord, card32 e1,card32 e2,card32 e3,float s );
   Vector3f calcEdgeVertex( const Vector3i &cellCoord, card32 edge,float s );
   void makeUnique(card32 &cellIndex,card32 &edge);
   Vector3f findSurface( EvaluateFunctionMethod evalMethod, void *context, Vector3f v1,
      Vector3f v2, float f1, float f2, float s, card32 maxIterations );

	void calcHierarchical( EvaluateFunctionMethod evalMethod, IsBoundingBoxEmptyMethod isEmptyMethod, void *context, const BoundingBox3f &bb, const Vector3i &dim, float s, float undefValue );

   float    *     m_VolumeData;
   BoundingBox3f  m_VolumeBB;
   Vector3i       m_VolumeDim;
   Vector3f       m_Scale;
	float				m_GridSnapTolerance;	// [0,0.5f), relative to samplegrid

	// hierarchical stuff
	BoundingBox3f	m_OriginalBB;
	card32			m_CurrentBox, m_NumBoxes;
	bool				m_Aborted;

   // output data
   vector<Vector3f>  m_Vertices;
   vector<Vector3i>  m_Triangles;

   // helper data
   struct TEdgePointer
   {
      int edge_vertex[12];
      TEdgePointer() {for( card32 i=0;i<12;i++ ) edge_vertex[i] = -1;}
   };
   map<card32,TEdgePointer>   m_EdgeMap;

public:

   struct FOUNDATION_API TMC_Vertex
   {
      Vector3f pos;
      TMC_Vertex operator + ( const TMC_Vertex &v ) const;
      TMC_Vertex operator - ( const TMC_Vertex &v ) const;
      TMC_Vertex operator * ( float f ) const;
      inline Vector3f getPos3f() const {return pos;}
   };

   /// Clear the mesh
   void clearMesh();

   /// calculates a triangle mesh for a given volume with levelset s
	/// data:	pointer to a float volume
	/// bb:		bounding box of the volume, represents 3d position of volume samples
	/// dim:		volume-size
	/// newMesh:	if true -> clear existing mesh from previous calculations
   /// volumeformat: index(x,y,z) = (z * dim[1] + y)*dim[0] + x
	/// output is stored internally, accessable with getVertex, getTriangle (see below)
	/// undefValue: if set to a value other than 0.0f, all cells with one or more data points of the given value are ignored (added mw, 10/10/06)
   void calculateVolume( float * data, BoundingBox3f bb, Vector3i dim, float s, bool newMesh = true, float undefValue = 0.0f);

   /// calculates a triangle mesh for a given function with levelset s
	/// evalMethod:	function pointer to evaluate a function value at a position (x,y,z)
	/// bb:		bounding box of the volume, represents 3d position of volume samples
	/// dim:		# of samples in each dimension
	/// newMesh:	if true -> clear existing mesh from previous calculations
	/// output is stored internally, accessable with getVertex, getTriangle (see below)
   void calculateVolume( EvaluateFunctionMethod evalMethod, void *context, BoundingBox3f bb, Vector3i dim, float s, bool newMesh = true );


	void calculateHierarchical( EvaluateFunctionMethod evalMethod, void *context, IsBoundingBoxEmptyMethod isEmptyMethod, 
		BoundingBox3f bb, Vector3i dim, float s, bool newMesh = true, float undefValue = 0.0f );

   // grid snapping tolerance
	void setGridSnappingTolerance( float s ) {m_GridSnapTolerance=max(0.0f,min(0.49999f,s));}
	float getGridSnappingTolerance() {return m_GridSnapTolerance;}
	
	// output access interface
   card32 getNumVertices() {return (card32)m_Vertices.size(); }
   card32 getNumTriangles() {return (card32)m_Triangles.size(); }
   Vector3f getVertex( card32 i ) {pAssert( i < m_Vertices.size() ); return m_Vertices[i]; }
   Vector3i getTriangle( card32 i ) {pAssert( i < m_Triangles.size() ); return m_Triangles[i]; }
   
// -------------------------------------------- 	
	
	/// calculate isosurface from single cube, clears existing output
   /// if maxIterations > 0 then a modified mc is used ( iterate-to-surface approx. )
   void calculateCube( EvaluateFunctionMethod evalMethod, void *context, const BoundingBox3f &bb, float s,
      card32 maxIterations = 0, bool newMesh = false, float undefValue = 0.0f);
//   void calculateCube( const Object * obj, EvaluateFunctionMethod evalMethod, const BoundingBox3f &bb, float s);

   /// calculates isosurface from single cube, clears existing output, v must point to 8 float values
   void calculateCube( float *v, const BoundingBox3f &bb, float s, bool newMesh = false );

   /// cuts every triangle at plane
   void cutTriangles( const Plane &p );

   /// cuts every triangle at every plane
   void cutTriangles( const vector<Plane> &planes, bool complex_case );

};

}

#endif
