//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiFileMultiObjectStream.h"
//---------------------------------------------------------------------------
#include "StringHelper.h"
#include "MathTools.h"
#include "BinaryObjectStreams.h"
#include "NumericalClassProperty.h"
#include "BooleanClassProperty.h"
#include "StringClassProperty.h"
#include "FileBlockAllocator.h"
#include "Tools.h"
#include "MemoryObjectStreams.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------
using namespace geoxl;

// --- init

IMPLEMENT_GEOXL_CLASS( MultiFileMultiObjectStreamIndex, 0 )
{
	BEGIN_CLASS_INIT( MultiFileMultiObjectStreamIndex );
}

// -

IMPLEMENT_GEOXL_CLASS( MultiFileMultiObjectStream, 1 )
{
	BEGIN_CLASS_INIT( MultiFileMultiObjectStream );
	ADD_CARD32_PROP(status, 0)
   ADD_STRING_PROP(baseFileName, 0)
   ADD_BOOLEAN_PROP(showDebugEvents, 0)
   ADD_BOOLEAN_PROP(relativePath, 1)

	MultiFileMultiObjectStreamMetaClass->getGUIDescriptor()->addClassCreator(MultiFileMultiObjectStreamCreator::getClass());
}

// -

IMPLEMENT_GEOXL_CLASS( MultiFileMultiObjectStreamCreator, 1 )
{
	BEGIN_CLASS_INIT( MultiFileMultiObjectStreamCreator );
	ADD_STRING_PROP(baseFileName, 0)
	ADD_BOOLEAN_PROP(showDebugEvents, 0)
   ADD_BOOLEAN_PROP(relativePath, 1)
}

const card32 INVALID_FILE = 0xFFFFFFFF;
const card32 EMPTY_SLOT   = 0xFFFFFFFE;
const card32 DUMMY_SLOT   = 0xFFFFFFFD;



// --- MultiFileMultiObjectStreamIndex


MultiFileMultiObjectStreamIndex::MultiFileMultiObjectStreamIndex() {
	Entry e;
	e.fileNumber = DUMMY_SLOT;
	e.inFileIndex = DUMMY_SLOT;
	e.dataSize = DUMMY_SLOT;
	// index zero must not be used (reserved for null pointer)
	entries.push_back(e);
}


card64 MultiFileMultiObjectStreamIndex::addNewEntry(card32 fileNumber, card64 inFileIndex, card64 dataSize) {
	Entry e;
	e.fileNumber = fileNumber;
	e.inFileIndex = inFileIndex;
	e.dataSize = dataSize;
	
	if (unusedEntries.empty()) {
		mpcard newID = entries.size();
		entries.push_back(e);
		return newID;
	} else {
		mpcard newID = unusedEntries.back();
		unusedEntries.pop_back();
		if (entries.size() <= newID || entries[newID].fileNumber != INVALID_FILE)  {
			throw PException("MultiFileMultiObjectStreamIndex::addNewEntry() - something went wrong (space already occupied).");
		}

		entries[newID] = e; 
		return newID;
	}
	
}

void MultiFileMultiObjectStreamIndex::getIndices(card64 objectID,	card32 &fileNumber, card64 &inFileIndex, card64 &dataSize) {
	mpcard id = (mpcard)objectID;
	if (id != objectID || id >= entries.size()) {
		throw PException("MultiFileMultiObjectStreamIndex::getIndices() - something went wrong (objectID out of range).");
	}
	Entry e = entries[id];
	fileNumber = e.fileNumber;
	inFileIndex = e.inFileIndex;
	dataSize = e.dataSize;
	if (fileNumber == INVALID_FILE) {
		throw PException("MultiFileMultiObjectStreamIndex::getIndices() - something went wrong (got objectID/pointer to free space).");
	}
}

void MultiFileMultiObjectStreamIndex::setIndices(card64 objectID,	card32 fileNumber, card64 inFileIndex, card64 dataSize) {
	mpcard id = (mpcard)objectID;
	if (id != objectID || id >= entries.size()) {
		throw PException("MultiFileMultiObjectStreamIndex::getIndices() - something went wrong (objectID out of range).");
	}
	if (entries[id].fileNumber == INVALID_FILE) {
		throw PException("MultiFileMultiObjectStreamIndex::getIndices() - something went wrong (got objectID/pointer to free space).");
	}
	entries[id].fileNumber = fileNumber;
	entries[id].inFileIndex = inFileIndex;
	entries[id].dataSize = dataSize;
}


void MultiFileMultiObjectStreamIndex::deleteEntry(card64 objectID) {
	mpcard id = (mpcard)objectID;
	if (id != objectID || id >= entries.size()) {	
		throw PException("MultiFileMultiObjectStreamIndex::deleteEntry() - something went wrong (objectID out of range).");
	}
	entries[id].fileNumber = INVALID_FILE;
	unusedEntries.push_back(id);
}

void MultiFileMultiObjectStreamIndex::write(OutputObjectStream *s) const {
	s->write(instanceID);
	mpcard entriesSize = entries.size();
	s->writeMaxPlatformCard(entriesSize);
	for (mpcard i=0; i<entriesSize; i++) {
		s->write(entries[i].fileNumber);
		s->write(entries[i].inFileIndex);
		s->write(entries[i].dataSize);
	}
}

void MultiFileMultiObjectStreamIndex::read(InputObjectStream *s) {	
	s->read(instanceID);
	mpcard entriesSize;
	s->readMaxPlatformCard(entriesSize);
	entries.resize(entriesSize);
	for (mpcard i=0; i<entriesSize; i++) {
		s->read(entries[i].fileNumber);
		s->read(entries[i].inFileIndex);
		s->read(entries[i].dataSize);
		if (entries[i].fileNumber == INVALID_FILE) {
			unusedEntries.push_back(i);
		} 
	}
}



// ---



MultiFileMultiObjectStream::MultiFileMultiObjectStream(string baseFileName, bool showDebugEvents, 
																	    const GloballyUniqueID *initialGUID, bool relativePath) {
	if (initialGUID == NULL) {
		globalId.createNewID();
	} else {
		globalId = *initialGUID;
	}

	this->baseFileName = baseFileName;
	this->showDebugEvents = showDebugEvents;
	this->relativePath = relativePath;
	index = NULL;
	status = statusDisconnected;
	if (baseFileName != "") { 
		bindAllocators();
	}
}

MultiObjectStream *MultiFileMultiObjectStream::createOOCAwareCopy(string idPrefix, bool insertOriginal, string idSuffix) {
	string newFilename = removeFilenameFromPath(baseFileName) + idPrefix + (insertOriginal?removePathFromFilename(baseFileName):string("")) + idSuffix;
	return new MultiFileMultiObjectStream(newFilename, showDebugEvents, NULL, relativePath);
}


void MultiFileMultiObjectStream::truncatePathRelativeTo(string basePath, string filename, string &result, bool &matches) {
	if (basePath.size() >= filename.size()) {
		matches = false;
	} else {
		for (unsigned i=0; i<basePath.size(); i++) {
			if (basePath[i] != filename[i]) {
				matches = false;
				return;
			}
		}
		string resultToReturn = filename.substr(basePath.size(), filename.size() - basePath.size());
		result = resultToReturn;
		matches = true;
	}
}

void MultiFileMultiObjectStream::write(OutputObjectStream *s) const {
	string saveFileName = baseFileName;
	bool matches = false;
	if (relativePath) {
		string result = allDirSeperatorsToDos(baseFileName);
		truncatePathRelativeTo(
			allDirSeperatorsToDos(removeFilenameFromPath(s->getFileName())),
			allDirSeperatorsToDos(baseFileName), 
			result, 
			matches
		);
		if (matches) {
			const_cast<MultiFileMultiObjectStream*>(this)->baseFileName = result;
		}
		#ifndef WIN32
		this needs to be fixed for unix (dos directory separators)
		#endif
	}
	MultiObjectStream::write(s);
	s->write(globalId);
	s->writeBool(matches);
	const_cast<MultiFileMultiObjectStream*>(this)->baseFileName = saveFileName;
}

void MultiFileMultiObjectStream::read(InputObjectStream *s) {
	MultiObjectStream::read(s);
	s->read(globalId);	

	bool matches = false;
	if (s->getClassVersion(MultiFileMultiObjectStream::getClass()) >= MetaClass::FIRST_VERSION + 1) {
		s->readBool(matches);
	}
	if (relativePath && matches) {
		baseFileName = removeFilenameFromPath(s->getFileName()) + baseFileName;
	}

	index = NULL;
	status = statusDisconnected;
	if (baseFileName != "") {
		bindAllocators();
	}
}


MultiFileMultiObjectStream::~MultiFileMultiObjectStream(void) {
	sync();
	for (mpcard i=0; i<files.size(); i++) {
		// closes and syncs files automatically
		delete files[i];
	}
}

void MultiFileMultiObjectStream::assign(const Object* obj, CopyContext *context) {
	throw PException("MultiFileMultiObjectStream::assign - not implemented.");
}


FileBlockAllocator *MultiFileMultiObjectStream::getAlloc(card32 logBlockSize) {
	pAssert(logBlockSize <= 64);
	if (logBlockSize >= files.size()) {
		files.resize(logBlockSize+1, NULL);
	}	
	if (files[logBlockSize] == NULL) {
		files[logBlockSize] = new FileBlockAllocator();
		string fileName = getBlockFileName(logBlockSize);
		if (FileExists(fileName.c_str())) {
			files[logBlockSize]->openFile(fileName.c_str(), showDebugEvents);
		} else {
			card64 blockSize = 1 << logBlockSize;
			files[logBlockSize]->createNewFile(fileName.c_str(), blockSize, showDebugEvents);
		}
	}
	return files[logBlockSize]; 
}

void MultiFileMultiObjectStream::bindAllocators() {
	// rebind existing
	if (FileExists(getIndexFileName().c_str())) {
		BinaryInputObjectStream in(getIndexFileName().c_str());
		Persistent *p;
		in.readObject(p);
		if (p == NULL || !p->getInstanceClass()->inheritsFrom(MultiFileMultiObjectStreamIndex::getClass())) {
			status = statusDisconnected;
			index = NULL;
			throw PException("MultiFileMultiObjectStream::bindAllocators() - cannot recognize existing index file; setup stopped and file disconnected.");
		}
		index = (MultiFileMultiObjectStreamIndex*)p;
		if (index->getInstanceID() != globalId) {
			status = statusDisconnected;
			throw PException("MultiFileMultiObjectStream::bindAllocators() - GUID of existing index file does not match the GUID of this object (i.e. trying to connect to the wrong set of files); setup stopped and file disconnected.");
		}
		status = statusConnected;
	} else {
		index = new MultiFileMultiObjectStreamIndex();
		index->setInstanceID(globalId);
		status = statusConnected;
	}
}

string MultiFileMultiObjectStream::getIndexFileName() const {
	return baseFileName + string("_index.mfSwap");
}

string MultiFileMultiObjectStream::getBlockFileName(mpcard logBlockSize) const {
	return baseFileName + string("_logSize") + intToStr(logBlockSize) + string(".mfSwap");
}

card32 MultiFileMultiObjectStream::getLogBlockSize(mpcard blockSize) {
	blockSize += FileBlockAllocator::getPerBlockOverhead();
   card32 result = 0;
   while (blockSize > 0) {
      blockSize = blockSize >> 1;			 
      result++;
   }
   return result;
}

void MultiFileMultiObjectStream::sync() {
	BinaryOutputObjectStream out(getIndexFileName().c_str());
	out.writeObject(index);
	for (mpcard i=0; i<files.size(); i++) {
		if (files[i] != NULL) files[i]->sync();
	}
}

MultiFileMultiObjectStream::ObjectID MultiFileMultiObjectStream::createObjectEntry() {
	return index->addNewEntry(EMPTY_SLOT, EMPTY_SLOT, EMPTY_SLOT);
}

void MultiFileMultiObjectStream::updateObject(ObjectID id, const MultiPartPersistent *obj) {
	pAssert(index && status == statusConnected);
	
	// --- create memory stream
	MemoryOutputObjectStream s;
   const card32 n = (card32)globalObjectPointers.size();
   for (card32 i=0; i<n; i++) {
      s.addGlobalObject(globalObjectPointers[i].obj, globalObjectPointers[i].index);
   }
	s.writeObject(obj);
	BinaryContainer buf;
	buf = *(s.closeAndGetMemoryBuffer());
	updateObject(id, buf);
}

void MultiFileMultiObjectStream::updateObject(ObjectID id, const BinaryContainer &buf) {

	// --- store to disc
	card64 newDataSize = buf.size();
	card32 newFileNumber = getLogBlockSize(buf.size());
	card32 fileNumber; card64 inFileIndex;	card64 dataSize;
	index->getIndices(id, fileNumber, inFileIndex, dataSize);

	if (fileNumber == EMPTY_SLOT) {
		// case 1: create new block from scratch
		fileNumber = newFileNumber; // just for safety
		FileBlockAllocator *alloc = getAlloc(newFileNumber);
		inFileIndex = alloc->allocBlock();
		index->setIndices(id, newFileNumber, inFileIndex, newDataSize);		
		alloc->writeBlock(inFileIndex, &(buf[0]), buf.size());
	} else {
		if (fileNumber == newFileNumber) {
			// case 2: same block size - just update
			FileBlockAllocator *alloc = getAlloc(newFileNumber);
			alloc->writeBlock(inFileIndex, &(buf[0]), buf.size());	
			// update size entry
			index->setIndices(id, newFileNumber, inFileIndex, newDataSize);							
		} else {
			// case 3: different size - move to another file
			FileBlockAllocator *newAlloc = getAlloc(newFileNumber);
			FileBlockAllocator *oldAlloc = getAlloc(fileNumber);
			oldAlloc->deleteBlock(inFileIndex);
			card64 newInFileIndex = newAlloc->allocBlock();
			newAlloc->writeBlock(newInFileIndex, &(buf[0]), buf.size());			
			index->setIndices(id, newFileNumber, newInFileIndex, newDataSize);
		}
	}
}

MultiPartPersistent *MultiFileMultiObjectStream::getObject(ObjectID id) {
	pAssert(index && status == statusConnected);

	card32 fileNumber;
	card64 inFileIndex;
	card64 dataSize;
	index->getIndices(id, fileNumber, inFileIndex, dataSize);

	if (fileNumber >= EMPTY_SLOT) {
		throw PException("MultiFileMultiObjectStream::getObject() - id points to invalid or empty slot.");
	}
	
	card8 *mem = new card8[dataSize];
	
	FileBlockAllocator *alloc = getAlloc(fileNumber);
	alloc->readBlock(inFileIndex, mem, dataSize);
	
	MemoryInputObjectStream s(mem, dataSize);
   const card32 n = (card32)globalObjectPointers.size();
   for (card32 i=0; i<n; i++) {
      s.addGlobalObject(globalObjectPointers[i].obj, globalObjectPointers[i].index);
   }
	Persistent *result;
	s.readObject(result);
	delete[] mem;
	
	return (MultiPartPersistent*)result;
}

void MultiFileMultiObjectStream::deleteObject(ObjectID id) {
	pAssert(index && status == statusConnected);

	card32 fileNumber;
	card64 inFileIndex;
	card64 dataSize;
	index->getIndices(id, fileNumber, inFileIndex, dataSize);
	if (fileNumber == INVALID_FILE) {
		throw PException("MultiFileMultiObjectStream::deleteObject() - id points to invalid location (free space). Possibly a pointer has been freed twice.");
	}
	if (fileNumber != EMPTY_SLOT) {
		FileBlockAllocator *alloc = getAlloc(fileNumber);
		alloc->deleteBlock(inFileIndex);
	}
	index->deleteEntry(id);	
}

MultiFileMultiObjectStream::Status MultiFileMultiObjectStream::getStatus() {
	return status;
}

string MultiFileMultiObjectStream::getFilename() const {
	return getIndexFileName();
}



// ---


MultiFileMultiObjectStreamCreator::MultiFileMultiObjectStreamCreator() {
	baseFileName = "C:\\temp\\test";
	showDebugEvents = true;	 
	relativePath = true;
}
Object *MultiFileMultiObjectStreamCreator::makeObject() {
	return new MultiFileMultiObjectStream(baseFileName, showDebugEvents, NULL, relativePath);
}
