//---------------------------------------------------------------------------
#ifndef MultiFileMultiObjectStreamH
#define MultiFileMultiObjectStreamH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MultiObjectStream.h"
#include "GloballyUniqueID.h"
#include "ClassCreator.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {
	using namespace std;

class FileBlockAllocator;

/** helper class for MultiFileMultiObjectStream (stores object indices and file indices) */
class	MultiFileMultiObjectStreamIndex : public Persistent {
	GEOXL_CLASS( MultiFileMultiObjectStreamIndex )

 private:
   GloballyUniqueID instanceID;
   
   struct Entry {
		card32 fileNumber;
		card64 inFileIndex;
		card64 dataSize;
   };
   
   /// stores fileNumber and index within file for each objectID 
   /// (which is the indexing type)
   /// the current implementation is thus resticted to O(2GB) blocks
   vector<Entry> entries;
   /// list of objectIDs in entries list that are available (again)
   vector<mpcard> unusedEntries;
   
 public:
	MultiFileMultiObjectStreamIndex();
	
   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);
   
	void setInstanceID(GloballyUniqueID &id)	{instanceID = id;}
	GloballyUniqueID getInstanceID()				{return instanceID;}

   card64 addNewEntry(card32 fileNumber, card64 inFileIndex, card64 dataSize);
   void setIndices(card64 objectID,	card32 fileNumber, card64 inFileIndex, card64 dataSize);
   void getIndices(card64 objectID,	card32 &fileNumber, card64 &inFileIndex, card64 &dataSize);
   void deleteEntry(card64 objectID);
};


/// friend class that is used in the multi-threaded version
class MFMTLoadStoreThread;

/**
	multi object stream that swapes objects to disc (and back)
	the objects are stored in different files, according to their size
	new files are created as required
*/
class FOUNDATION_API MultiFileMultiObjectStream : public FileBasedMultiObjectStream {
	GEOXL_CLASS( MultiFileMultiObjectStream )

 private:
	typedef vector<card8> BinaryContainer;

   // --- persistent properties
   Status status;
   /// base filename, suffixes will be added
   string baseFileName;
   /// display performance statistics
   bool showDebugEvents;
   /// this object's id to check whether beeing connected to the right set of files
   GloballyUniqueID globalId;
   /// if this flag is set, the path is stored relative to the location the object is beeing stored at
   /// and restored to the (possibly new) location the object is retrieved from
   /// activate this flag if you want to move the swapfiles around
   /// store the swapfiles in the same directory (or a subdirectory) of the object
   bool relativePath;

	// --- stuff on disc
	MultiFileMultiObjectStreamIndex *index;
	vector<FileBlockAllocator*> files;

	// --- internal methods
	void bindAllocators();
	string getIndexFileName() const;
	string getBlockFileName(mpcard logBlockSize) const;
	/// returns a number so that 1 << (this number) is never smaller than the input number plus per block overhead of the FileBlockAllocator
	card32 getLogBlockSize(mpcard blockSize);
	FileBlockAllocator *getAlloc(card32 logBlockSize);
	
	/// update object providing a serialized byte stream, needed for implementing the multi-threaded version (MultiFileMultiObjectStreamMT)
   virtual void updateObject(ObjectID id, const BinaryContainer &buf);

 public:
   /// create new instance: if filename + index suffix exists, the existing database is reopened. otherwise, a new one is created.
	MultiFileMultiObjectStream(string baseFileName = "", bool showDebugEvents = true, const GloballyUniqueID *initialGUID = NULL, bool relativePath = true);

   virtual MultiObjectStream *createOOCAwareCopy(string idPrefix = "", bool insertOriginal = true, string idSuffix = "");

	/// throws exception
   virtual void assign(const Object* obj, CopyContext *context = NULL);
   
   virtual ObjectID createObjectEntry();
   virtual void updateObject(ObjectID id, const MultiPartPersistent *obj);
   virtual MultiPartPersistent *getObject(ObjectID id);
   virtual void deleteObject(ObjectID id);
   virtual void sync();
   virtual Status getStatus();

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);
   
   virtual string getFilename() const;

	~MultiFileMultiObjectStream(void);

	/// helper function for relativePath processing
	static void truncatePathRelativeTo(string basePath, string filename, string &result, bool &matches);

 friend MFMTLoadStoreThread;
};

/// helper class that creates new MultiFileMultiObjectStream (for easier interactive setup)
class MultiFileMultiObjectStreamCreator : public ClassCreator {
 GEOXL_CLASS(MultiFileMultiObjectStreamCreator)
private:
	string baseFileName;
	bool showDebugEvents;
	bool relativePath;
	
public:
	MultiFileMultiObjectStreamCreator();
	virtual Object *makeObject();
};


}

#endif