//---------------------------------------------------------------------------
#ifndef MultiFileMultiObjectStreamMTH
#define MultiFileMultiObjectStreamMTH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MultiFileMultiObjectStream.h"
#include "ThreadCommandQueue.h"
//---------------------------------------------------------------------------
#include <QtCore/QThread>
//---------------------------------------------------------------------------
#include <deque>
//---------------------------------------------------------------------------


namespace geoxl {
	using namespace std;


class MFMTLoadStoreThread : public QThread {
	Q_OBJECT		

 public:
	typedef vector<card8> BinaryContainer;
	typedef MultiFileMultiObjectStream::ObjectID ObjectID;

 private:
	typedef list< pair<ObjectID, MultiPartPersistent*> > FetchBufferType;
	MultiFileMultiObjectStream *multiObj;
	FetchBufferType fetchBuffer;
	card32 maximumQueueLength;
	
	ThreadCommandQueue commands;
	
	class StopCommand : public ThreadCommandQueue::AnsweredMessage {};
	
	class AddUpdateJobCommand : public ThreadCommandQueue::Message {
	 public:
		ObjectID id;
		BinaryContainer *serializedData;
	};
	
	class FetchJobCommand : public ThreadCommandQueue::Message {
	 public:
		ObjectID id;
	};

	class GetObjectCommand : public ThreadCommandQueue::AnsweredMessage {
	 public:
		ObjectID id;
		MultiPartPersistent *obj_answer;
		bool error;
	};
	
	class CreateObjectEntryCommand : public ThreadCommandQueue::AnsweredMessage {
	 public:
		ObjectID id_answer;
		bool error;
	};
	
	class QueryStatusCommand : public ThreadCommandQueue::AnsweredMessage {
	 public:
		FileBasedMultiObjectStream::Status status_answer;
		bool error;
	};
	
	class AddGlobalObjectPointerCommand : public ThreadCommandQueue::Message {
	 public:
		Persistent *obj;
		card32 index;
	};
	class DeleteObjectCommand : public ThreadCommandQueue::Message {
	 public:
		ObjectID id;
	};
	class QueryObjectAvailabilityCommand : public ThreadCommandQueue::AnsweredMessage {
	 public:
		ObjectID id;
		bool available_answer;
		bool error;
	};
	
	class SyncCommand : public ThreadCommandQueue::Message {};
	
	
	void checkBufferSize();
	void queueStatistics();

 protected:
	virtual void run();
	
 public:
	MFMTLoadStoreThread(string baseFileName, card32 maximumQueueLength, GloballyUniqueID initialGUID);
	/// requests an update of the specified object on disc (deferred execution)
	void addUpdateJob(ObjectID id, BinaryContainer *serializedData);
	/// requests loading an object and unserializing it (deferred execution)
	void addFetchJob(ObjectID id);

	/// retrieve an object (executes immediately, blocks until object is retrieved)
	MultiPartPersistent *getObject(ObjectID id);
	/// creates a new object entry (executes immediately, blocks until entry is created)
	ObjectID createObjectEntry();	
	/// delete an object
	void deleteObject(ObjectID id);
	/// query multiObj status
	FileBasedMultiObjectStream::Status getStatus();
	/// write back dirty data (concurrently)
	void sync();
	/// keep global object table up-to-date
   void addGlobalObjectPointer(Persistent *obj, card32 index);
   /// see if object is in fetch buffer
   bool objectAvailable(ObjectID id);

	virtual ~MFMTLoadStoreThread();
};


///
/// This is a multi-threaded version of MultiFileMultiObjectStream that supports background prefetching
/// and background updates.
///
class FOUNDATION_API MultiFileMultiObjectStreamMT : public FileBasedMultiObjectStream {
	GEOXL_CLASS( MultiFileMultiObjectStreamMT )

 private:		
	typedef vector<card8> BinaryContainer;

	MFMTLoadStoreThread *loadStoreThread;
 
   // --- persistent properties
   /// base filename, suffixes will be added
   string baseFileName;
   /// maximum number of buffers used for incomming / outgoing object streams
   card32 maximumQueueLength; 
	/// identifier to avoid reopening unrelated swap files
   GloballyUniqueID globalID;
   /// alter baseFileName according to object storage location upon serialization (see MultiFileMultiObjectStream for details)
   bool relativePath;

	void createThread();

 public:
   /// create new instance: if filename + index suffix exists, the existing database is reopened. otherwise, a new one is created.
   /// maximum queue length specifies the maximum amount of unsatisfied request handled concurrently (in order to limit buffer usage)
	MultiFileMultiObjectStreamMT(string baseFileName = "", card32 maximumQueueLength = 50, bool relativePath = true);

	MultiObjectStream *createOOCAwareCopy(string idPrefix, bool insertOriginal, string idSuffix);

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);
   
   virtual void addGlobalObjectPointer(Persistent *obj, card32 index);

   virtual ObjectID createObjectEntry();
   virtual void updateObject(ObjectID id, const MultiPartPersistent *obj);
   virtual MultiPartPersistent *getObject(ObjectID id);
   virtual void deleteObject(ObjectID id);
   virtual void sync();
   virtual Status getStatus();
   virtual bool objectAvailable(ObjectID id);

   virtual void prefetchObject(ObjectID id);

   virtual string getFilename() const;

	~MultiFileMultiObjectStreamMT(void);

};

/// helper class that creates new MultiFileMultiObjectStreamMT (for easier interactive setup)
class MultiFileMultiObjectStreamMTCreator : public ClassCreator {
 GEOXL_CLASS(MultiFileMultiObjectStreamMTCreator)
private:
	string baseFileName;
	card32 maximumQueueLength;
	bool relativePath;
	
public:
	MultiFileMultiObjectStreamMTCreator();
	virtual Object *makeObject();
};

		
}

#endif