//---------------------------------------------------------------------------
#ifndef MultiObjectStreamH
#define MultiObjectStreamH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PTypes.h"
#include "ObjectStreams.h"
#include "Persistent.h"
#include "MultiPartPersistent.h"
#include "Tools.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {
	using namespace std;


/// Class to store multiple objects in one stream / file.
/// each single object must fit into main memory; however, the set of all objects may be larger (64 bit limited)
class FOUNDATION_API MultiObjectStream : public Persistent {
 private:
   static MetaClass *MultiObjectStreamMetaClass;

 public:
   enum Status {statusDisconnected = 0, statusConnected = 1, force32bit = 0x0FFFFFFF};

 protected:
   struct GOPEntry {
      Persistent* obj;
      card32 index;
   };
   vector<GOPEntry> globalObjectPointers;
   
   static vector<MultiObjectStream*> activeStreams;

 public:
   typedef card64 ObjectID;

   /// create new instance
   MultiObjectStream() {activeStreams.push_back(this);}

   /// these objects are not stored in each stream but referenced only
   /// the list of these objects is stored along with the MultiObjectStream
   /// index is a globally unique identifying number
   virtual void addGlobalObjectPointer(Persistent *obj, card32 index);
   virtual mpcard getNumGlobalObjects();
   virtual bool hasGlobalObjectPointer(Persistent *obj);
   virtual void removeGlobalObjectPointer(Persistent *obj);

   // inherited
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// creates a copy of the stream; for inCore versions, this is equivalent to ->copy(); this is also the default behavior
   /// for out-of-core objects, the identifier on secondary storage (filename) is changed to
   /// create a real copy, not overwriting the original object
   virtual MultiObjectStream *createOOCAwareCopy(string idPrefix = "", bool insertOriginal = true, string idSuffix = "");

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   /// perform all pending updates, the file may be closed afterwards without negative side effects
   virtual void sync() = 0;

   /// create new, empty slot (id) for storing the object
   virtual ObjectID createObjectEntry() = 0;
   /// write data from the specified object into the specified slot/id
   virtual void updateObject(ObjectID id, const MultiPartPersistent *obj) = 0;
   /// read data from the specified slot/id; create new object in memory with the corresponding data
   virtual MultiPartPersistent *getObject(ObjectID id) = 0;
   /// delete slot/id; the associated data is lost
   virtual void deleteObject(ObjectID id) = 0;
   
   /// indicate that the object will be needed soon and should be made available
   /// standard behavior is to just ignore this hint
   virtual void prefetchObject(ObjectID id) {}
   /// query if object is already prefetched; returns allways true if no prefetching is supported
   virtual bool objectAvailable(ObjectID id) {return true;}

   /// get status of MultiObjectStream. the result is "statusConnected" if the stream is ready for storing and loading objects.
   /// if the stream is not ready (e.g. because of connecting to a wrong storage file), the status is "statusDisconnected".
   virtual Status getStatus() = 0;

   /// close database again, sync is always called before closing
   virtual ~MultiObjectStream();

	/// get all multi object stream objects that are currently resident in memory (e.g. for proper shutdown at program exit)
	static vector<MultiObjectStream*> getActiveStreams() {return activeStreams;}
};

class FOUNDATION_API FileBasedMultiObjectStream : public MultiObjectStream {
 private:
   static MetaClass *FileBasedMultiObjectStreamMetaClass;
   
 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

	/// must be overridden
   virtual MultiObjectStream *createOOCAwareCopy(string idPrefix = "", bool insertOriginal = true, string idSuffix = "") {
		notImplemented();
		return NULL;
   }

   virtual string getFilename() const = 0; 
};


/// helper function: read ObjectID from any ObjectInputStream
inline void readObjectID(InputObjectStream *s, MultiObjectStream::ObjectID &id) {
   s->read<MultiObjectStream::ObjectID>(id);
}

/// helper function: write ObjectID into any ObjectOutputStream
inline void writeObjectID(OutputObjectStream *s, const MultiObjectStream::ObjectID id) {
   s->write<MultiObjectStream::ObjectID>(id);
}


}

#endif
