#ifndef PCA_templCode_H
#define PCA_templCode_H

#include "PCA.h"

// Xrt
#include "LinearAlgebra.h"
#include "LinearAlgebra_templCode.h"

// System
#include <vector>

namespace geoxl {

/**
 * Create new object
 */
template<class FloatType, unsigned Dim>
PCA<FloatType,Dim>::PCA()
{
   clear();
}

/**
 * Reset PCA
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::clear()
{
	points.clear();
	weights.clear();
	// Compute covariance matrix
	for ( unsigned int i = 0; i < Dim; i++ ){
		centroid[i] = 0;
		for ( unsigned int j = 0; j < Dim; j++ ){
			C[i][j] = 0;
		}
	}
}

/**
 * Add new point
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::addPoint( Point p )
{
   centroid = ( centroid * float32( points.size() ) + p ) / float( points.size() + 1 );
   points.push_back( p );
}

/**
* Add new point
*/
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::addPoint( Point p, FloatType w )
{
	weights.push_back(w);
	points.push_back( p );
}

/**
 * Solve eigen system
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::analyze( Point& eigenValues, Matrix& eigenVectors, Point& c )
{
	if( weights.size() > 0)
	{
		if( weights.size() != points.size() )
			throw "Don't mix weighted and unweigted addPoint() methods!";

		// estimate centroid
		float summedWeight = 0;
		for( unsigned i=0;i<points.size();i++ )
		{
			centroid += points[i]*weights[i];
			summedWeight += weights[i];
		}
		if( summedWeight > 1e-20f )
			centroid /= summedWeight;

		// compute weighted covariance
		for ( unsigned int n = 0; n < points.size(); n++ ){
			for ( unsigned int i = 0; i < Dim; i++ ){
				for ( unsigned int j = 0; j <= i; j++ ){
					C[i][j] += ( points[n][i]-centroid[i]) * (points[n][j]-centroid[j])*weights[n];
					C[j][i] = C[i][j];
				}
			}
		}
		if( summedWeight > 1e-20f )
			C /= summedWeight;

	}
	else
	{
		for ( unsigned int n = 0; n < points.size(); n++ ){
			for ( unsigned int i = 0; i < Dim; i++ ){
				for ( unsigned int j = 0; j <= i; j++ ){
					C[i][j] += ( points[n][i]-centroid[i]) * (points[n][j]-centroid[j]);
					C[j][i] = C[i][j];
				}
			}
		}
	}

	/// Normalize covariance matrix
	if ( points.size() > 0 ) C /= float32( points.size() );

	// Analyze eigenstructure
	C.computeEigenStructure( eigenValues, eigenVectors );
	c = centroid;
}

}

#endif
