//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "TemplatedImage.h"
#include "TemplatedImage_templCode.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


template <class PixelType>
TemplatedImage<PixelType>::TemplatedImage() {
   sizeX = 0;
   sizeY = 0;
   pixels = NULL;
}

template <class PixelType>
TemplatedImage<PixelType>::TemplatedImage(card32 sizeX, card32 sizeY) {
   this->sizeX = sizeX;
   this->sizeY = sizeY;
   pixels = new PixelType[sizeX*sizeY];
}

template <class PixelType>
void TemplatedImage<PixelType>::clear(PixelType col) {
   PixelType *pos = pixels;
   card32 numPixels = sizeX*sizeY;
   for (card32 c = 0; c<numPixels; c++) {
      *pos = col;
      ++pos;
   }
}

template <class PixelType>
void TemplatedImage<PixelType>::resize(card32 sizeX, card32 sizeY) {
   delete[] pixels;
   this->sizeX = sizeX;
   this->sizeY = sizeY;
   pixels = new PixelType[sizeX*sizeY];
}

template <class PixelType>
void TemplatedImage<PixelType>::assign(const Object* obj, CopyContext *context) {
   if (obj == NULL) {
      sizeX = sizeY = 0;
      delete[] pixels;
      pixels = NULL;
   }

   if (this == obj) return;

   const TemplatedImage<PixelType> *iObj = dynamic_cast<const TemplatedImage<PixelType>*>(obj);
   if (iObj == NULL) {
      throw PException("TemplatedImage<PixelType>::assign - wrong argument type");
   } else {
      delete[] pixels;
      this->sizeX = iObj->sizeX;
      this->sizeY = iObj->sizeY;
      if (iObj->pixels == NULL) {
         pixels = NULL;
      } else {
         const PixelType *pos = iObj->pixels;
         card32 numPixels = sizeX*sizeY;
         this->pixels = new PixelType[numPixels];
         PixelType *tpos = this->pixels;
         for (card32 c = 0; c<numPixels; c++) {
            *tpos = *pos;
            ++tpos;
            ++pos;
         }
      }
   }
}

template <class PixelType>
void TemplatedImage<PixelType>::write(OutputObjectStream *s) const {
   s->write(sizeX);
   s->write(sizeY);
   PixelType *pos = pixels;
   for (card32 i=0; i<sizeX*sizeY; i++) {
      s->write<PixelType>(*pos);
      ++pos;
   }
}

template <class PixelType>
void TemplatedImage<PixelType>::read(InputObjectStream *s) {
   s->read(sizeX);
   s->read(sizeY);
   delete[] pixels;
   pixels = new PixelType[sizeX*sizeY];
   PixelType *pos = pixels;
   for (card32 i=0; i<sizeX*sizeY; i++) {
      PixelType p;
      s->read<PixelType>(p);
      *pos = p;
      ++pos;
   }
}

template <class PixelType>
TemplatedImage<PixelType>::~TemplatedImage() {
   delete[] pixels;
}

template <class PixelType>
void TemplatedImage<PixelType>::downsample(card32 factor, TemplatedImage<PixelType> *dest) {
   card32 destWidth = sizeX/factor;
   card32 destHeight = sizeY/factor;
   pAssert(dest->sizeX >= destWidth && dest->sizeY >= destHeight);
   for (card32 yd = 0; yd<destHeight; yd++) {
      for (card32 xd = 0; xd<destWidth; xd++) {
         card32 sourceAddr = xd*factor + yd*factor*sizeX;
         pAssert(sourceAddr < sizeX*sizeY);
         card32 destAddr = xd+yd*dest->sizeX;
         pAssert(destAddr < dest->sizeX*dest->sizeY);
         dest->pixels[destAddr] = pixels[sourceAddr];
      }
   }
}

template <class PixelType>
inline PixelType TemplatedImage<PixelType>::getPixel(card32 x, card32 y) const {
   pAssert(x<sizeX && y<sizeY);
   return pixels[x+sizeX*y];
}

template <class PixelType>
inline void TemplatedImage<PixelType>::setPixel(card32 x, card32 y, PixelType pix) {
   pAssert(x<sizeX && y<sizeY);
   pixels[x+sizeX*y] = pix;
}

template <class PixelType>
PixelType TemplatedImage<PixelType>::getBilinearInterpolatedPixel( float x, float y) const
{
	x = max(0.0f,min(x,(float)sizeX-0.01f));
	y = max(0.0f,min(y,(float)sizeY-0.01f));

	card32 ix = x;
	card32 iy = y;
	float fx = x - ix;
	float fy = y - iy;

	if( ix + 1 < sizeX && iy + 1 < sizeY )
	{
		return (pixels[ix+sizeX*iy] * (1.0f - fx) + pixels[ix+1+sizeX*iy] * fx) * (1.0f - fy)
			+ (pixels[ix+sizeX*(iy+1)] * (1.0f - fx) + pixels[ix+1+sizeX*(iy+1)] * fx) * fy;
	}

	if( ix + 1 < sizeX )
	{
		return (pixels[ix+sizeX*iy] * (1.0f - fx) + pixels[ix+1+sizeX*iy] * fx);
	}

	if( iy + 1 < sizeY )
	{
		return pixels[ix+sizeX*iy] * (1.0f - fy) + pixels[ix+sizeX*(iy+1)] * fy;
	}
	return pixels[ix+sizeX*iy];

}

template <class PixelType>
void TemplatedImage<PixelType>::filter(float kernelSize, EImageFilterType filterType)
{
	// prepare filter mask
	int iKernelSize = ceil(kernelSize);
	int iSize = iKernelSize*2.0f+1.0f;
	vector<float> mask(iSize*iSize);
	float inv_gaussSigma_2 = 1.0f / pow(kernelSize / 2.0f,2.0f);
	for( int x=-iKernelSize;x<=iKernelSize;x++ )
	{
		for( int y=-iKernelSize;y<=iKernelSize;y++)
		{
			int index = (y+iKernelSize)*iSize + x + iKernelSize;
			float maskEntry = 1.0f; // box
			switch( filterType)
			{
			case FT_Gauss:
				maskEntry = exp(-(float)(x*x+y*y) * inv_gaussSigma_2);
				break;
			case FT_Triangle:
				maskEntry = max( 0.0f, 1.0f - sqrt( (float)(x*x+y*y) )/kernelSize);
				break;
			}
			mask[index] = maskEntry;
		}
	}

	filterMask(iSize,iSize, &mask[0]);
}

#pragma warning(push)
#pragma warning(disable:4700)
template <class PixelType>
void TemplatedImage<PixelType>::filterMask(unsigned xSize,unsigned ySize, float * mask, bool normalizeMask )
{
	int cx = (int)xSize / 2;
	int cy = (int)ySize / 2;
	PixelType * res = new PixelType[sizeX*sizeY];
	for( int x=0;x<(int)getSizeX();x++ )
	{
		for( int y=0;y<(int)getSizeY();y++ )
		{
			PixelType p;
			p *= 0;
			float summedMask = 0;
			int minCY = min(cy,y);
			int maxCY = min(cy,(int)getSizeY()-1-y);
			for( int mx = -cx;mx<=cx;mx++)
			{
				if( x + mx < 0 || x + mx >(int)getSizeX()-1)
					continue;
				for( int my = -minCY;my<=maxCY;my++)
				{
					int wIndex = (my+cy)*ySize+mx+cx;
					p += getPixel(x+mx,y+my) * mask[wIndex];
					summedMask += mask[wIndex];

				}

			}
			if( normalizeMask)
				p/= summedMask;
			//setPixel(x,y,p);
			res[y*sizeX + x] = p;
		}
	}
	delete[] pixels;
	pixels = res;
}

template <class PixelType>
void TemplatedImage<PixelType>::smoothGaussian(float sigma)
{
	// setup 1d gaussian
	const int kernelSize = ceil(sigma * 2.0f);
	vector<float> kernel(kernelSize*2+1);
	float summedWeight = 0;
	for( int x=-kernelSize;x<=kernelSize;x++ )
	{
		kernel[x+kernelSize] = exp(-(float)x*x / sigma / sigma);
		summedWeight += kernel[x+kernelSize];
	}

	for( unsigned i=0;i<kernel.size();i++ )
		kernel[i] /= summedWeight;

	// filter image horizontally
	PixelType * tmp = new PixelType[sizeX*sizeY];
	const int iSizeX = (int)getSizeX();
	const int iSizeY = (int)getSizeY();
	PixelType p0; p0*= 0;
	const PixelType zeroPixel = p0;
	for( int x=0;x<iSizeX;x++ )
	{
		int minI = -kernelSize;
		if( x+minI < 0 )
			minI = -x;
		int maxI = kernelSize;
		if( x+maxI >= iSizeX )
			maxI = iSizeX-x-1;
		for( int y=0;y<iSizeY;y++ )
		{
			PixelType p = zeroPixel;
			PixelType * ptr = &pixels[y*iSizeX+x+minI];
			float * kernelPtr = &kernel[minI + kernelSize];
			for( int i=minI;i<=maxI;i++ )
			{
				p += *ptr++ * *kernelPtr++;
			}
			tmp[y*iSizeX+x] = p;
		}
	}

	// filter image vertically
	for( int y=0;y<iSizeY;y++ )
	{
		int minI = -kernelSize;
		if( y+minI < 0 )
			minI = -y;
		int maxI = kernelSize;
		if( y+maxI >= iSizeY )
			maxI = iSizeY-y-1;
		for( int x=0;x<iSizeX;x++ )
		{
			PixelType p = zeroPixel;
			PixelType * ptr = &pixels[(y+minI)*iSizeX+x];
			float * kernelPtr = &kernel[minI + kernelSize];
			for( int i=minI;i<=maxI;i++ )
			{
				p += *ptr * *kernelPtr++;
				ptr += iSizeX;
			}
			setPixel(x,y,p);
		}
	}

	delete tmp;

}

template <class PixelType>
void TemplatedImage<PixelType>::drawRectangle( card32 x1,card32 y1, card32 x2, card32 y2, PixelType pix)
{
	// check bounds
	x2 = min(x2,sizeX);
	y2 = min(y2,sizeY);
	for( card32 x=x1;x<x2;x++ )
	{
		for( card32 y=y1;y<y2;y++ )
		{
			setPixel(x,y,pix);
		}
	}
}

template <class PixelType>
void TemplatedImage<PixelType>::resize_bilinear( card32 width, card32 height)
{
	TemplatedImage<PixelType> copy;
	copy.assign(this);

	resize( width, height );
	Vector2f scale = makeVector2f((float)copy.getSizeX()/(float)width,
		(float)copy.getSizeY()/(float)height);
	for( unsigned x=0;x<width;x++ )
	{
		for( unsigned y=0;y<height;y++ )
		{
			Vector2f coord = makeVector2f(x,y).componentProduct(scale);
			PixelType pix =   copy.getBilinearInterpolatedPixel( coord[0], coord[1]);
			setPixel(x,y,pix);
		}
	}

}


#pragma warning (pop)


template class FOUNDATION_API TemplatedImage< Vector4ub >;
template class FOUNDATION_API TemplatedImage< Vector3ub >;
template class FOUNDATION_API TemplatedImage< Vector3f >;
template class FOUNDATION_API TemplatedImage< Vector4f >;
template class FOUNDATION_API TemplatedImage< StaticVector<card8,1> >;
template class FOUNDATION_API TemplatedImage< StaticVector<float32,1> >;
template class FOUNDATION_API TemplatedImage< StaticVector<float32,2> >;
template class FOUNDATION_API TemplatedImage< float >;
template class FOUNDATION_API TemplatedImage< card8 >;