//---------------------------------------------------------------------------
#ifndef ThreadCommandQueueH
#define ThreadCommandQueueH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <QtCore/QMutex>
#include <QtCore/QSemaphore>
//---------------------------------------------------------------------------


namespace geoxl {
	using namespace std;
	
	
/// this class implements a simple queue of commmands
/// the commands are received by a server thread which processes them and possibly returns results
class ThreadCommandQueue {
 public:
	/// base class for messages in the queue
	class Message {
	 public:
		virtual ~Message() {}
	};
	/// message that wait for an answer
	class AnsweredMessage : public Message {};
 
 private:
	// --- concurrent state
	deque<Message*> messages;
	deque<Message*> answers;
	// --- syncs
	QMutex queueMutex;
	QSemaphore messagesAvailable;
	QSemaphore answersAvailable;
	QSemaphore queueSlotsAvailable;
	card32 maxNumMsgs;
 
 public:
	/// construct command queue. maxNumMsgs limits the number of messages in the queue; zero refers to no limit.
	ThreadCommandQueue(card32 maxNumMsgs = 0);

	// --- client methods
	/// send a message to the server thread, no result is send back, the server deletes the message object
	/// if immediately == true, the command is placed at the front end of the queue
	void sendAsyncMessage(Message *msg, bool immediately = false);
	/// send a message to the server thread and wait until it has been processed; the result is stored in the same message object
	/// the message object has to be deleted by the client
	/// if immediately == true, the command is placed at the front end of the queue
	void sendMessageAndReceiveReply(Message *msg, bool immediately = false);
	
	// --- server methods
	/// get the next message (on the server side); blocks until a message becomes available
	Message *getMessage();
	/// answers message. additional information may be written to the msg object that was send to the server
	/// answerable messages must be send using sendMessageAndReceiveReply and thus the client is blocked while the server answers
	void answerMessage();
	
	/// for debugging: read current queue length
	int32 getNumAvailableMessageSlots();
};
	
}

#endif