//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "Tools.h"
#include "PAssert.h"
#include "Exceptions.h"
#include "DebugOutput.h"
#include "ClassMethods.h"
#include "PropertyParser.h"
#include "ScriptEngine.h"
#include "StringHelper.h"
#include "VariableArrayClassProperty.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace geoxl;
using namespace std;
//---------------------------------------------------------------------------

MetaClass* ScriptEngine::ScriptEngineMetaClass = NULL;

/****************************************************************************/
//  LocalNumSymbol - getValue
/// Returns the value of the variable "name", or 0 if the variable does not
/// exist
/****************************************************************************/
bool LocalNumSymbol::getValue(const std::string name, ExpressionNum& value)
{
  if (name.size() == 0)
    return false;

  // Local number variables always start with %
  if (name[0] != '%')
    return false;

  if (variables.find(name) == variables.end()) {
    variables[name] = 0;
  }

  value = variables[name];
  return true;
}

/****************************************************************************/
//  LocalNumSymbol - setValue
/// sets the value of the local variable given
/****************************************************************************/
bool LocalNumSymbol::setValue(const std::string name, ExpressionNum value)
{
  if (name.size() == 0)
    return false;

  // Local number variables always start with %
  if (name[0] != '%')
    return false;

  variables[name] = value;
  return true;
}

/****************************************************************************/
//  LocalNumSymbol - clear
/// Clears the local variables
/****************************************************************************/
void LocalNumSymbol::clear()
{
  variables.clear();
}

/****************************************************************************/
//  LocalStringSymbol - setValue
/// sets the value of the local string given
/****************************************************************************/
bool LocalStringSymbol::setValue(const std::string name, string value)
{
  if (name.size() == 0)
    return false;

  // Local strings always start with $
  if (name[0] != '$')
    return false;

  variables[name] = value;

  return true;
}

/****************************************************************************
//  LocalStringSymbol - getValue
/// Returns the value of the string given in value. Method returns true if
/// successful and false otherwise.
/****************************************************************************/
bool LocalStringSymbol::getValue(const std::string strName, std::string& value)
{
  if (strName.size() == 0)
    return false;

  // Local strings always start with $
  if (strName[0] != '$')
    return false;

  value = variables[strName];

  return true;
}

/****************************************************************************/
//  LocalStringSymbol - clear
/// Clears the local strings
/****************************************************************************/
void LocalStringSymbol::clear()
{
  variables.clear();
}

/****************************************************************************
 * ScriptEngine - default constructor
 ****************************************************************************/
ScriptEngine::ScriptEngine()
{
  initVars();
}

/****************************************************************************/
//  ScriptEngine - initVars
//  Initializes the command registry, symbol tables and the expression parser
/****************************************************************************/
void ScriptEngine::initVars()
{
  lineChanged = false;

  parametersNum.init(this);
  parametersString.init(this);

  numTables.clearTables();
  numTables.addTable(&localNums);
  numTables.addTable(&parametersNum);
  expression.setSymbolTableHandler(&numTables);

  stringTables.clearTables();
  stringTables.addTable(&localStrings);
  stringTables.addTable(&parametersString);

  objectTables.clearTables();

  scriptCommands.clear();
  scriptCommands.registerCommand(&printCommand, this);
  scriptCommands.registerCommand(&setCommand, this);
  scriptCommands.registerCommand(&ifCommand, this);
  scriptCommands.registerCommand(&whileCommand, this);
  scriptCommands.registerCommand(&timerCommand, this);
  scriptCommands.registerCommand(&stringCommand, this);
  scriptCommands.registerCommand(&objectCommand, this);
  scriptCommands.registerCommand(&callMethodCommand, this);
  scriptCommands.registerCommand(&progressCommand, this);

  keepLocalVariables = false;
}

/****************************************************************************
 * ScriptEngine - Methods for framework
 ****************************************************************************/
const MetaClass* ScriptEngine::getInstanceClass() const
{
  return getClass();
}

Object* ScriptEngine::newInstance() {
   return new ScriptEngine;
}

const MetaClass* ScriptEngine::getClass()
{
   if (ScriptEngineMetaClass == NULL) {
      ScriptEngineMetaClass = new MetaClass("ScriptEngine", "no description.",
                                       newInstance, sizeof(ScriptEngine));
   }
   return ScriptEngineMetaClass;
}

void ScriptEngine::init(const MetaClass *parentClass)
{
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION);
   ScriptEngineMetaClass->copyParentProperties();

	// mh - 13.03.2006 - moved custom editor init to InitFoundation

	BEGIN_CLASS_INIT( ScriptEngine );
	ADD_STRING_VAR_ARRAY_PROP( Code, 0 );
}

void ScriptEngine::setNumCodes(card32 size)
{
	code.resize(size);
}

void ScriptEngine::deleteCode( card32 index )
{
	code.erase( code.begin() + index );
}

char * ScriptEngine::getCode( card32 index )
{
	return (char*)&code[index];
}

mpcard ScriptEngine::getNumCodes()
{
	return code.size();
}

string ScriptEngine::getCodeName(card32 index )
{
	return intToStr(index);
}


/*std::string ScriptEngine::getCode(card32 index) const
{
  pAssert(index < code.size());
  return code[index];
}

void ScriptEngine::setCode(card32 index, std::string elem)
{
  pAssert(index < code.size());
  code[index] = elem;
}

card32 ScriptEngine::getCodeSize() const
{
  return (card32)code.size();
}

void ScriptEngine::setCodeSize(card32 newSize)
{
  code.resize(newSize);
}

void ScriptEngine::pushCode()
{
  code.push_back("");
}

void ScriptEngine::popCode() {
  code.pop_back();
}
*/

void ScriptEngine::shutdown()
{
   delete ScriptEngineMetaClass;
}

void ScriptEngine::assign(const Object* obj, CopyContext *context)
{
  pAssert(obj->getInstanceClass() == getInstanceClass());
  ScriptEngine* other = (ScriptEngine*)obj;
  code = other->code;

  initVars();

  keepLocalVariables = other->keepLocalVariables;
  if (keepLocalVariables) {
    localNums = other->localNums;
    localStrings = other->localStrings;
  }

  forget(context);
}

/****************************************************************************/
//  ScriptEngine - preprocessLine
//  Returns the line given, fully preprocessed.
/// Removes the comments, leading or trailing spaces etc. from the line and
/// returns the resulting string
/****************************************************************************/
string ScriptEngine::preprocessLine(card32 line) const
{
  string s = code[line];

  // remove leading spaces
  s = trimStr(s);
  if (s == "") return "";

  // clear comment lines
  if (s[0] == ';') return "";

  // remove double spaces in commands
  s = removeDblSpace(s);

  return s;
}

/****************************************************************************/
//  ScriptEngine - onScriptStart
//  Initializes the variables on script start
/// Event that is executed right before the script execution is started.
/// Add all the necessary initializations for commands to this method.
/// If you override this method, don't forget to call the original method of
/// ScriptEngine by calling ScriptEngine::onScriptStart() in your derived
/// version of GEOXLscript.
/****************************************************************************/
void ScriptEngine::onScriptStart()
{
  timerCommand.reset();
  progressCommand.init();
}

/****************************************************************************/
//  ScriptEngine - onScriptEnd
//  Does all the necessary steps after script execution
/// This event is executed right after the execution of a script is finished.
/// Add all necessary clean-up calls (emptying vectors etc.) to this method.
/// If you override this method, don't forget to call the original method of
/// ScriptEngine by calling ScriptEngine::onScriptStart() in your derived
/// version of GEOXLscript.
/****************************************************************************/
void ScriptEngine::onScriptEnd()
{
  progressCommand.done();
}

/****************************************************************************/
//  ScriptEngine - execute
//  Execute the script (syntaxCheck == false) or do a syntaxCheck
/// Executes the script if syntaxCheck == false. Does not execute the script
/// if syntaxCheck == true, but exceptions for syntax errors in commands are
/// thrown.
/****************************************************************************/
void ScriptEngine::execute(bool syntaxCheck)
{
  line = 0;
  lineChanged = false;
  if (code.size() == 0) return;
  while (!scriptStack.empty())
    scriptStack.pop();
  if (!keepLocalVariables) clearLocalVariables();

  progressWindow->pushStep(false, "GEOXLscript");

  bool doExecute = true;
  bool userBreak = false;
  onScriptStart();
  while (doExecute) {
    string codeLine = preprocessLine(line);

    if (codeLine != "") {
      if (syntaxCheck) {
        // Ignore invalid symbol names if syntax check is active
        try {
          scriptCommands.execute(codeLine, syntaxCheck);
        } catch (EParserOperandException* e) {
			forget(e);
        }
      } else {
        scriptCommands.execute(codeLine, syntaxCheck);
      }
    }

    if (!lineChanged) {
      ++line;
    } else {
      lineChanged = false;
    }
    if (line >= code.size()) doExecute = false;

    if (!progressWindow->progress(progressPercentage)) {
      // User break
      while (!scriptStack.empty()) scriptStack.pop();
      doExecute = false;
      userBreak = true;
    }
  }

  if (!scriptStack.empty() && !userBreak) {
    line = scriptStack.top().line;
    throw EParserException("Closing statement for " + strUCase(scriptStack.top().command) +
      " command expected.");
  }

  if (!keepLocalVariables) clearLocalVariables();

  onScriptEnd();
  bool dummy;
  progressWindow->popSteps(dummy);
}

/****************************************************************************/
//  ScriptEngine - debugStart
//  Initializes the script engine for step-by-step execution
/// This method needs to be called in order to initialize the step-by-step
/// execution in debug mode. Returns TRUE if the script engine is ready
/// for step-by-step-execution and FALSE if the script is empty. (i.e.
/// there is nothing to execute)
/****************************************************************************/
bool ScriptEngine::debugStart()
{
  line = 0;
  lineChanged = false;
  if (code.size() == 0) return false;
  while (!scriptStack.empty())
    scriptStack.pop();
  if (!keepLocalVariables) clearLocalVariables();

  onScriptStart();

  while (preprocessLine(line) == "") ++line;

  return true;
}

/****************************************************************************/
//  ScriptEngine - debugNextCommand
//  Executes the command that line points to in debug mode
/// This method will execute just one command and increase the line number
/// afterwards to the next command (empty lines will be ignored). Returns
/// TRUE if there is more to execute and FALSE if the script execution is
/// done.
/****************************************************************************/
bool ScriptEngine::debugNextCommand()
{
  string codeLine = preprocessLine(line);

  if (lineChanged && line < code.size()) {
    while (preprocessLine(line) == "") {
      ++line;
      if (line >= code.size()) break;
    }
    lineChanged = false;
  }

  if (line < code.size()) {
    if (codeLine != "")
      scriptCommands.execute(codeLine, false);

    if (!lineChanged) {
      do {
        ++line;
        if (line >= code.size()) break;
      } while (preprocessLine(line) == "");
    } else {
      lineChanged = false;
    }
  }

  if (line >= code.size()) {
    // Script execution done
    if (!scriptStack.empty()) {
      line = scriptStack.top().line;
      throw EParserException("Closing statement for " + strUCase(scriptStack.top().command) +
        " command expected.");
    }

    if (!keepLocalVariables) clearLocalVariables();

    onScriptEnd();
    return false;
  }

  return true;
}

/****************************************************************************/
//  ScriptEngine - debugEnd
//  Interrupts script execution in debug mode
/// It is not necessary to call this method when the script ends normally.
/// In that case, debugNextCommand() will return false and clean up
/// automatically.
/****************************************************************************/
void ScriptEngine::debugEnd()
{
  if (!keepLocalVariables) clearLocalVariables();
  onScriptEnd();
}

/****************************************************************************/
//  ScriptEngine - isActive
//  For use by commands: Is the script active?
/// Returns true if the machine is active at the moment (= top element of
/// the stack is active or stack is empty). Commands can set the machine
/// inactive so that other commands do not get executed (for example in an
/// IF...ENDIF block if the condition is not fulfilled.
/****************************************************************************/
bool ScriptEngine::isActive() const
{
  if (scriptStack.empty()) return true;
  StackElement elem = scriptStack.top();

  return elem.active;
}

/****************************************************************************/
//  ScriptEngine - clearLocalVariables
//  Clears the local variables
/****************************************************************************/
void ScriptEngine::clearLocalVariables()
{
  localNums.clear();
  localStrings.clear();
}

/****************************************************************************/
//  ScriptEngine - evaluateArrayExpressions
//  For use by commands: Evaluate the array expression given (see detailed descr.)
/// Preprocessor for array expressions: Replaces all the numerical expressions
/// inside [] with the result. For example: "position[%i]" -> "position[0]"
/****************************************************************************/
std::string ScriptEngine::evaluateArrayExpressions(std::string propertyName)
{
  ExpressionParser expression;
  expression.setSymbolTableHandler(&numTables);

  // Evaluate expressions in []
  card32 strPos = 0;
  bool inBrackets = false;
  string expr;
  while (strPos < propertyName.size()) {
    if (propertyName[strPos] == '[')
      inBrackets = true;
    if (inBrackets && propertyName[strPos] != '[' && propertyName[strPos] != ']')
      expr += propertyName[strPos];
    if (propertyName[strPos] == ']') {
      inBrackets = false;
      // Evaluate expression
      ExpressionNum value = expression.evaluate(expr);
      string valueStr = numToStr(value);

      // Delete expression from string
      strPos -= (card32)expr.size();
      propertyName.erase(strPos, expr.size());

      // Add result of expression to string
      propertyName.insert(strPos, valueStr);
      strPos += (card32)valueStr.size() + 1;
      expr.clear();
    } else {
      ++strPos;
    }
  }
  if (inBrackets) {
    throw EParserException("EvaluateArrayExpressions: ] expected");
  }

  return propertyName;
}

/****************************************************************************/
//  ScriptCommandRegistry - registerCommand
//  Registers a script command
/// This method registers a command for the script and sets the script
/// pointer of the command to the script engine given.
/****************************************************************************/
void ScriptCommandRegistry::registerCommand(ScriptCommand* command, ScriptEngine* script)
{
  command->setScriptEngine(script);
  string words = command->supportedCommands();
  while (words.size() > 0) {
    string s;
    s = extractToChar(words, ';');
    commands[s] = command;
  }
}

/****************************************************************************/
//  ScriptCommandRegistry - execute
//  Executes the command given or performs a syntax check
/****************************************************************************/
void ScriptCommandRegistry::execute(std::string command, bool syntaxCheck)
{
  string commandWord = strLCase(extractToChar(command, ' '));

  if (commands.find(commandWord) != commands.end())
    commands[commandWord]->execute(commandWord + " " + command, syntaxCheck);
  else
    throw EParserException("Invalid command \"" + strUCase(commandWord) + "\"");
}

/****************************************************************************/
//  PrintCommand - execute
//  Prints the parameters to debugOutput
/****************************************************************************/
void PrintCommand::execute(std::string command, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  string parameters = command;
  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("PRINT: No parameters");

  string outStr;

  if (syntaxCheck) return;

  bool inString = false;
  string str, expr;
  for (card32 i = 0; i < parameters.size(); ++i) {
    if (parameters[i] == '\"') {  // begin or end of string
      if (inString) {
        outStr += str;
        str.clear();
      } else {
        expr = trimStr(expr);
        if (expr.size() > 0) {
          string strValue;
          if (!script->getStringTables()->getValue(expr, strValue)) {
            // should be arithmetic expression
            ExpressionNum num = script->getExpParser()->evaluate(expr);
            outStr += numToStr(num);
          } else {
            // is string
            outStr += strValue;
          }
          expr.clear();
        }
      }
      inString = !inString;
    } else {                      // within string or expression
      if (inString)
        str += parameters[i];
      else
        expr += parameters[i];
    }
  }

  if (inString) {
    outStr += str;
    str.clear();
  } else {
    expr = trimStr(expr);
    if (expr.size() > 0) {
      string strValue;
      if (!script->getStringTables()->getValue(expr, strValue)) {
        // should be arithmetic expression
        ExpressionNum num = script->getExpParser()->evaluate(expr);
        outStr += numToStr(num);
      } else {
        // is string
        outStr += strValue;
      }
      expr.clear();
    }
  }

  if (!syntaxCheck) {
    debugOutput << outStr;
    debugOutput.cr();
  }
}

/****************************************************************************/
//  SetCommand - execute
//  Sets the variable to the artihmetic expression
/****************************************************************************/
void SetCommand::execute(std::string command, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  string parameters = command;
  extractToChar(parameters, ' ');

  string variable = extractToChar(parameters, ' ');
  if (variable == "")
    throw EParserException("SET: Variable name expected");

  if (extractToChar(parameters, ' ') != "=")
    throw EParserException("SET: = expected");

  string targetValue = parameters;
  if (targetValue == "")
    throw EParserException("SET: Arithmetic expression expected");

  if (!syntaxCheck) {
    ExpressionNum num = script->getExpParser()->evaluate(targetValue);

  // Set variable
    if (!script->getNumTables()->setValue(variable, num))
      throw EParserException("SET: Invalid variable name or setting not possible");
  }
}

/****************************************************************************/
//  IfCommand - execute
//  Typical IF - ELSE - ENDIF block
/****************************************************************************/
void IfCommand::execute(std::string command, bool syntaxCheck)
{
  string parameters = command;
  command = extractToChar(parameters, ' ');

  if (command == "if") {
    ScriptEngine::StackElement elem;
    elem.command = "if";
    elem.line = script->getLine();

    if (!script->isActive()) {
      elem.active = false;
      script->stackPush(elem);
      return;
    } else {
      ExpressionNum num = script->getExpParser()->evaluate(parameters);
      elem.active = num;
      script->stackPush(elem);
      return;
    }
  }

  if (command == "else") {
    ScriptEngine::StackElement elem;
    if (script->stackEmpty()) {
      throw EParserException("ELSE without IF found");
    }
    elem = script->stackTop();

    if (elem.command != "if") {
      throw EParserException("ELSE without IF found");
    }

    elem.active = !elem.active;
    script->stackPop();
    script->stackPush(elem);
    return;
  }

  if (command == "endif") {
    ScriptEngine::StackElement elem;
    if (script->stackEmpty()) {
      throw EParserException("ENDIF without IF found");
    }
    elem = script->stackTop();

    if (elem.command != "if") {
      throw EParserException("ENDIF without IF found");
    }

    script->stackPop();
    return;
  }

  forget(syntaxCheck);
}

/****************************************************************************/
//  WhileCommand - execute
//  WHILE - ENDWHILE block
/****************************************************************************/
void WhileCommand::execute(std::string command, bool syntaxCheck)
{
  string parameters = command;
  command = extractToChar(parameters, ' ');

  if (command == "while") {
    ScriptEngine::StackElement elem;
    elem.command = "while";
    elem.line = script->getLine();

    if (!script->isActive()) {
      elem.active = false;
    } else {
      elem.active = script->getExpParser()->evaluate(parameters);
    }
    script->stackPush(elem);
    return;
  }

  if (command == "endwhile") {
    ScriptEngine::StackElement elem;
    if (script->stackEmpty())
      throw EParserException("ENDWHILE without WHILE found");

    elem = script->stackTop();
    if (elem.command != "while")
      throw EParserException("ENDWHILE without WHILE found");

    if (!elem.active) {
      script->stackPop();
    } else {
      script->stackPop();
      if (!syntaxCheck)
        script->setLine(elem.line);
    }
    return;
  }
}

/****************************************************************************
 * TimerCommand - constructor
 ****************************************************************************/
TimerCommand::TimerCommand() : ScriptCommand()
{
  isPaused = false;
  pauseTime = 0;
  initialTime = timer.getCurrentValue();
}

/****************************************************************************/
//  TimerCommand - reset
//  Resets the timer, i.e. sets its value to 0.
/****************************************************************************/
void TimerCommand::reset()
{
  initialTime = timer.getCurrentValue();
  pauseTime = 0;
}

/****************************************************************************/
//  TimerCommand - execute
//  TIMER RESET/GET/PAUSE/RESUME command
/****************************************************************************/
void TimerCommand::execute(std::string parameters, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("TIMER: No parameters");

  string subCommand = strLCase(extractToChar(parameters, ' '));

  if (subCommand == "reset") {
    if (!syntaxCheck) {
      reset();
    }
    return;
  }

  if (subCommand == "get") {
    if (parameters == "")
      throw EParserException("TIMER GET: No variable name given");

    if (!syntaxCheck) {
      ExpressionNum time;
      if (!isPaused)
        time = timer.getCurrentValue() - initialTime - pauseTime;
      else
        time = pauseStartTime - initialTime - pauseTime;
      script->getNumTables()->setValue(parameters, time);
    }
    return;
  }

  if (subCommand == "pause") {
    if (!syntaxCheck && !isPaused) {
      pauseStartTime = timer.getCurrentValue();
      isPaused = true;
    }
    return;
  }

  if (subCommand == "resume") {
    if (!syntaxCheck && isPaused) {
      pauseTime = timer.getCurrentValue() - pauseStartTime;
      isPaused = false;
    }
    return;
  }

  throw EParserException("TIMER " + strUCase(subCommand) + " is not a valid command.");
}

/****************************************************************************/
//  StringCommand - execute
//  STRING SET/LENGTH/COPY/INSERT/DELETE/FROMNUM/TONUM command
/****************************************************************************/
void StringCommand::execute(std::string parameters, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("STRING: No parameters");

  string subCommand = strLCase(extractToChar(parameters, ' '));

  // STRING SET <String> = <String expression>
  if (subCommand == "set") {
    string strName = trimStr(extractToChar(parameters, '='));
    string strExpr = trimStr(parameters);
    if (strName == "" || strExpr == "")
      throw EParserException("STRING SET: Invalid syntax");

    string value;
    if (!script->getStringTables()->evaluate(strExpr, value))
      throw EParserException("Invalid string expression " + strExpr);

    if (!syntaxCheck) {
      if (!script->getStringTables()->setValue(strName, value))
        throw EParserException("STRING SET: Invalid variable name '" + strName + "' or setting not possible");
    }

    return;
  }

  // STRING LENGTH <String Expression>, <Number Variable>
  if (subCommand == "length") {
    string sourceStr = trimStr(extractToChar(parameters, ','));

    string s;
    bool found = script->getStringTables()->evaluate(sourceStr, s);
    if (!found)
      throw EParserException("STRING LENGTH: Invalid string " + sourceStr);

    if (!syntaxCheck) {
      if (!script->getNumTables()->setValue(trimStr(parameters), s.size()))
        throw EParserException("STRING LENGTH: Setting of variable " + parameters + " failed.");
    }

    return;
  }

  // STRING FROMNUM <Expression>, <String variable>
  if (subCommand == "fromnum") {
    string expr = trimStr(extractToChar(parameters, ','));
    ExpressionNum num = script->getExpParser()->evaluate(expr);

    if (!syntaxCheck) {
      parameters = trimStr(parameters);
      if (!script->getStringTables()->setValue(parameters, numToStr(num)))
        throw EParserException("STRING FROMNUM: Could not set string " + parameters);
    }
    return;
  }

  // STRING TONUM <String Expression>, <Number variable>
  if (subCommand == "tonum") {
    string strExpr = trimStr(extractToChar(parameters, ','));
    string strValue;
    if (!script->getStringTables()->evaluate(strExpr, strValue))
      throw EParserException("STRING TONUM: Invalid string expression " + strExpr);

    if (!syntaxCheck) {
      ExpressionNum num;
      if (!strToNum(strValue, num))
        throw EParserException("STRING TONUM: " + strValue + " is not a number.");

      parameters = trimStr(parameters);
      if (!script->getNumTables()->setValue(parameters, num))
        throw EParserException("STRING TONUM: Could not set variable " + parameters);
    }
    return;
  }

  // STRING DELETE <String variable>, <Expression>, <Expression>
  if (subCommand == "delete") {
    string strName = trimStr(extractToChar(parameters, ','));
    string fromStr = trimStr(extractToChar(parameters, ','));
    string toStr = trimStr(parameters);

    if (strName == "" || fromStr == "" || toStr == "")
      throw EParserException("STRING DELETE: Not enough parameters");

    int32 fromNum = (int32)script->getExpParser()->evaluate(fromStr);
    int32 toNum = (int32)script->getExpParser()->evaluate(toStr);
    if (fromNum < 0 || toNum < 0)
      throw EParserException("STRING DELETE: Negative numbers not allowed");

    string s;
    if (!script->getStringTables()->getValue(strName, s))
      throw EParserException("STRING DELETE: Invalid string name " + strName);

    if (!syntaxCheck) {
      s.erase(fromNum, toNum);

      if (!script->getStringTables()->setValue(strName, s))
        throw EParserException("STRING DELETE: String " + strName + " could not be set");
    }
    return;
  }

  // STRING INSERT <String expression>, <Expression>, <String variable>
  if (subCommand == "insert") {
    string insertStrExp = trimStr(extractToChar(parameters, ','));
    string posStr = trimStr(extractToChar(parameters, ','));
    string targetStr = trimStr(parameters);

    if (insertStrExp == "" || posStr == "" || targetStr == "")
      throw EParserException("STRING INSERT: Not enough parameters");

    string insertStr;
    if (!script->getStringTables()->evaluate(insertStrExp, insertStr))
      throw EParserException("STRING INSERT: Invalid string expression " + insertStrExp);

    int32 posNum = (int32)script->getExpParser()->evaluate(posStr);
    if (posNum < 0)
      throw EParserException("STRING INSERT: Negative numbers not allowed");

    string s;
    if (!script->getStringTables()->getValue(targetStr, s))
      throw EParserException("STRING INSERT: Invalid target string " + targetStr);

    if (!syntaxCheck) {
      s.insert(posNum, insertStr);
      if (!script->getStringTables()->setValue(targetStr, s))
        throw EParserException("STRING INSERT: String " + targetStr + " could not be set");
    }
    return;
  }

  throw EParserException("STRING " + strUCase(subCommand) + " is not a valid command.");
}

/****************************************************************************/
//  ObjectCommand - execute
//  OBJECT NEW/DELETE command
/****************************************************************************/
void ObjectCommand::execute(std::string parameters, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("OBJECT: No parameters");

  string subCommand = strLCase(extractToChar(parameters, ' '));

  if (subCommand == "new") {
    string metaClassName = extractToChar(parameters, ' ');
    string objectName = parameters;
    if (metaClassName == "" || objectName == "")
      throw EParserException("OBJECT NEW: Not enough parameters");

    const MetaClass* mc = MetaClass::getClassByName(metaClassName);
    if (mc == NULL) throw EParserException("OBJECT NEW: class name not found.");
    
    if (!syntaxCheck) {
      Object* obj = mc->newInstance();
      if (!script->getObjectTables()->setValue(objectName, obj)) {
        delete obj;
        throw EParserException("OBJECT NEW: Could not add " + objectName + " to object dictionary");
      }
    }
    return;
  }

  if (subCommand == "delete") {
    string objectName = parameters;
    if (objectName == "")
      throw EParserException("OBJECT DELETE: Not enough parameters");

    if (!syntaxCheck) {
       if (!script->getObjectTables()->setValue(objectName, NULL)) {
         throw EParserException("OBJECT NEW: Could not delete " + objectName + " from object dictionary");
       }
    }
    return;
  }

  throw EParserException("OBJECT " + strUCase(subCommand) + " is not a valid command.");
}

/****************************************************************************/
//  CallMethodCommand - execute
//  CALLMETHOD command
/****************************************************************************/
void CallMethodCommand::execute(std::string parameters, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("CALLMETHOD: No parameters");

  string objName = extractToChar(parameters, ' ');
  string methodName = parameters;
  
  if (objName == "" || methodName == "")
     throw EParserException("CALLMETHOD: Not enough parameters");
  
  if (!syntaxCheck) {
    Object* obj;
    if (!script->getObjectTables()->getValue(objName, obj))
      throw EParserException("CALLMETHOD: Object name " + objName + " is invalid");
    
    const MetaClass* mc = obj->getInstanceClass();
    
    const ClassMethod* method = NULL;
    for (card32 i = 0; i < mc->getNumMethods(); ++i) {
      const ClassMethod* m = mc->getMethod(i);
      if (m->getName() == methodName) {
        method = m;
        break;
      }
    }
    
    if (method == NULL)
      throw EParserException("CALLMETHOD: Method name " + methodName + " is invalid");

    method->call(obj);     
  }
}

/****************************************************************************/
//  ProgressCommand - execute
//  PROGRESS ADD/POP/PUSH/NEXT/PERCENTAGE command
/****************************************************************************/
void ProgressCommand::execute(std::string parameters, bool syntaxCheck)
{
  if (!syntaxCheck && !script->isActive()) return;

  extractToChar(parameters, ' ');

  if (parameters == "")
    throw EParserException("PROGRESS: No parameters");

  string subCommand = strLCase(extractToChar(parameters, ' '));

  if (subCommand == "add") {
    ProgressWindow::WeightElem elem;
    string availStr = trimStr(extractToChar(parameters, ','));
    string weightStr = trimStr(extractToChar(parameters, ','));
    string nameStr = trimStr(parameters);

    if (availStr == "" || weightStr == "" || nameStr == "")
      throw EParserException("PROGRESS ADD: Not enough parameters");

    elem.progressAvailable = script->getExpParser()->evaluate(availStr);
    elem.weight = (float32)script->getExpParser()->evaluate(weightStr);
    if (!script->getStringTables()->evaluate(nameStr, elem.stepName))
      throw EParserException("PROGRESS ADD: Invalid string expression");

    if (!syntaxCheck) steps.push_back(elem);
    return;
  }

  if (subCommand == "push") {
    if (!syntaxCheck) {
      if (steps.empty()) throw EParserException("PROGRESS PUSH: No steps added");
      progressWindow->pushSteps(steps);
      steps.empty();
    }
    return;
  }

  if (subCommand == "pop") {
    bool dummy;  // Cancel will be processed by ScriptEngine.execute
    if (!syntaxCheck) progressWindow->popSteps(dummy);
    return;
  }

  if (subCommand == "next") {
    bool dummy;  // Cancel will be processed by ScriptEngine.execute
    if (!syntaxCheck) progressWindow->nextStep(dummy);
    return;
  }

  if (subCommand == "percentage") {
    if (parameters == "") throw EParserException("PROGRESS PERCENTAGE: No parameter");
    card32 percentage = (card32)script->getExpParser()->evaluate(parameters);

    if (!syntaxCheck) script->setProgressPercentage(percentage);
    return;
  }

  throw EParserException("PROGRESS " + strUCase(subCommand) + " is not a valid command.");
}
