//---------------------------------------------------------------------------
#ifndef ScriptEngineH
#define ScriptEngineH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <vector>
#include <string>
#include <stack>
#include "Timer.h"
#include "PTypes.h"
#include "Persistent.h"
#include "ExpressionParser.h"
#include "PropertyParser.h"
#include "ProgressWindow.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------

namespace geoxl {


class ScriptEngine;

/// Local number symbol table for the ScriptEngine (variables %x)
class FOUNDATION_API LocalNumSymbol : public NumSymbolTable {
  private:
    std::map<std::string, ExpressionNum> variables;

  public:
    virtual bool getValue(const std::string name, ExpressionNum& value);
    virtual bool setValue(const std::string name, ExpressionNum value);
    void clear();
};

/// Local string symbol table for the ScriptEngine (variables $x)
class FOUNDATION_API LocalStringSymbol : public StringSymbolTable {
  private:
    std::map<std::string, std::string> variables;

  public:
    virtual bool getValue(const std::string name, std::string& value);
    virtual bool setValue(const std::string name, std::string value);
    void clear();
};

/// Allows to access the parameters of the object symbol tables.
/**
 * Allows access to the number or string parameters of the objects
 * in the object symbol tables that are registered in the script
 * engine.
 */
template <class T>
class ParameterSymbolTable : public SymbolTableBase<T> {
  private:
    ScriptEngine* script;
    PropertyParser propertyParser;

  public:
    ParameterSymbolTable() { script = NULL; };
    void init(ScriptEngine* script) { this->script = script; };
    virtual bool getValue(const std::string name, T& value);
    virtual bool setValue(const std::string name, T value);
};

/// Base class for script commands.
class FOUNDATION_API ScriptCommand {
  protected:
    ScriptEngine* script;

  public:
    ScriptCommand() { script = NULL; }
    /// Returns a string with the script commands that this class implements
    /** The commands are separated by semicolon and need to be in lowercase.
     * For example: "if;else;endif" */
    virtual std::string supportedCommands() = 0;
    void setScriptEngine(ScriptEngine* script) { this->script = script; }

    virtual void execute(std::string parameters, bool syntaxCheck) = 0;
};

// ----

/// Registry for script commands, used by ScriptEngine
class FOUNDATION_API ScriptCommandRegistry {
  protected:
    std::map<std::string, ScriptCommand*> commands;
  public:
    /// Registers a script command
    void registerCommand(ScriptCommand*, ScriptEngine*);
    void clear() { commands.clear(); }
    /// Executes the command given or performs a syntax check
    void execute(std::string command, bool syntaxCheck);
};

/// PRINT command for GEOXLscript
class FOUNDATION_API PrintCommand : public ScriptCommand {
  public:
    PrintCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "print"; };
    /// Prints the parameters to debugOutput
    virtual void execute(std::string command, bool syntaxCheck);
};

/// SET command for GEOXLscript
class FOUNDATION_API SetCommand : public ScriptCommand {
  public:
    SetCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "set"; };
    /// Sets the variable to the artihmetic expression
    virtual void execute(std::string command, bool syntaxCheck);
};

/// IF/ELSE/ENDIF command for GEOXLscript
class FOUNDATION_API IfCommand : public ScriptCommand {
  public:
    IfCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "if;else;endif"; };
    /// Typical IF - ELSE - ENDIF block
    virtual void execute(std::string command, bool syntaxCheck);
};

/// WHILE/ENDWHILE command for GEOXLscript
class FOUNDATION_API WhileCommand : public ScriptCommand {
  public:
    WhileCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "while;endwhile"; };
    /// WHILE - ENDWHILE block
    virtual void execute(std::string command, bool syntaxCheck);
};

/// TIMER RESET/GET/PAUSE/RESUME command for GEOXLscript
class FOUNDATION_API TimerCommand : public ScriptCommand {
  protected:
    Timer timer;
    card32 initialTime, pauseTime, pauseStartTime;
    bool isPaused;

  public:
    TimerCommand();
    virtual std::string supportedCommands() { return "timer"; };

    virtual void execute(std::string parameters, bool syntaxCheck);
    /// Resets the timer, i.e. sets its value to 0.
    void reset();
};

/// STRING SET/LENGTH/COPY/INSERT/DELETE/FROMNUM/TONUM command for GEOXLscript
class FOUNDATION_API StringCommand : public ScriptCommand {
  public:
    StringCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "string"; };

    virtual void execute(std::string parameters, bool syntaxCheck);
};

/// OBJECT NEW/DELETE command for GEOXLscript
class FOUNDATION_API ObjectCommand : public ScriptCommand {
  public:
    ObjectCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "object"; };

    virtual void execute(std::string parameters, bool syntaxCheck);
};

class FOUNDATION_API CallMethodCommand : public ScriptCommand {
  public:
    CallMethodCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "callmethod"; };

    virtual void execute(std::string parameters, bool syntaxCheck);
};

/// PROGRESS ADD/POP/PUSH/NEXT/PERCENTAGE command for GEOXLscript
class FOUNDATION_API ProgressCommand : public ScriptCommand {
  private:
    ProgressWindow::WeightList steps;

  public:
    ProgressCommand() : ScriptCommand() {};
    virtual std::string supportedCommands() { return "progress"; };

    void init() { steps.clear(); };
    virtual void execute(std::string parameters, bool syntaxCheck);
    void done() { steps.clear(); };
};

/// Class that can execute GEOXLscripts.
class FOUNDATION_API ScriptEngine : public Persistent {
  public:
    /// Element of the stack of the ScriptEngine
    typedef struct {
      std::string command;
      bool active;
      card32 line;
    } StackElement;

    /// Contains the code of the script
    std::vector<std::string> code;

    ScriptEngine();
    /// Execute the script (syntaxCheck == false) or do a syntaxCheck
    virtual void execute(bool syntaxCheck);
    /// Initializes the script engine for step-by-step execution
    virtual bool debugStart();
    /// Executes the command that line points to in debug mode
    virtual bool debugNextCommand();
    /// Interrupts script execution in debug mode
    virtual void debugEnd();

    /// For use by commands: Push an element on the script stack
    void stackPush(StackElement entry) { scriptStack.push(entry); }
    /// For use by commands: Pop an element from the stack
    void stackPop() { scriptStack.pop(); }
    /// For use by commands: Return the top of the stack
    StackElement stackTop() const { return scriptStack.top(); }
    /// For use by commands: Return whether the stack is empty
    bool stackEmpty() const { return scriptStack.empty(); }
    /// For use by commands: Is the script active?
    bool isActive() const;
    /// For use by commands: Returns the active line number
    card32 getLine() const { return line; }
    /// For use by commands: Sets the active line number
    void setLine(card32 line) { this->line = line; lineChanged = true; }

    /// For use by commands: Set the progress to the value given
    void setProgressPercentage(card32 percentage) { progressPercentage = percentage; }
    /// Set to true if the local variables should still be available the next time a script is executed
    void setKeepLocalVariables(bool value) { keepLocalVariables = value; }
    /// Clears the local variables
    void clearLocalVariables();

    /// For use by commands: Return the script's expression parser
    ExpressionParser* getExpParser() { return &expression; }
    /// For use by commands: Return the script's number tables
    NumTableHandler* getNumTables() { return &numTables; }
    /// For use by commands: Return the script's string tables
    StringTableHandler* getStringTables() { return &stringTables; }
    /// For use by commands: Return the script's object tables
    ObjectTableHandler* getObjectTables() { return &objectTables; }

    /// For use by commands: Evaluate the array expression given (see detailed descr.)
    std::string evaluateArrayExpressions(std::string propertyName);

    virtual void assign(const Object* obj, CopyContext *context = NULL);

    virtual const MetaClass *getInstanceClass() const;
    static const MetaClass *getClass();
    static void init(const MetaClass *parentClass);
    static void shutdown();

  protected:
    std::stack<StackElement> scriptStack;
    card32 line;
    bool lineChanged;

    NumTableHandler numTables;
    StringTableHandler stringTables;
    ObjectTableHandler objectTables;
    LocalNumSymbol localNums;
    LocalStringSymbol localStrings;
    ParameterSymbolTable<ExpressionNum> parametersNum;
    ParameterSymbolTable<std::string> parametersString;

    ScriptCommandRegistry scriptCommands;
    ExpressionParser expression;

    // Commands
    PrintCommand printCommand;
    SetCommand setCommand;
    IfCommand ifCommand;
    WhileCommand whileCommand;
    TimerCommand timerCommand;
    StringCommand stringCommand;
    ObjectCommand objectCommand;
    CallMethodCommand callMethodCommand;
    ProgressCommand progressCommand;

    card32 progressPercentage;
    bool keepLocalVariables;

    /// Returns the line given, fully preprocessed.
    std::string preprocessLine(card32 line) const;
    /// Initializes the command registry, symbol tables and the expression parser
    virtual void initVars();
    /// Initializes the variables on script start
    virtual void onScriptStart();
    /// Does all the necessary steps after script execution
    virtual void onScriptEnd();

  private:
    static MetaClass *ScriptEngineMetaClass;
    static Object* newInstance();

	void setNumCodes(card32 size);
	void deleteCode( card32 index );
	char * getCode( card32 index );
	mpcard getNumCodes();
	string getCodeName(card32 index );
};


//---------------------------------------------------------------------------
// IMPLEMENTATION
//---------------------------------------------------------------------------

template <class T>
bool ParameterSymbolTable<T>::getValue(const std::string name, T& value)
{
  if (name.size() == 0) return false;
  if (name[0] == '$' || name[0] == '%') return false;

  string propertyName = script->evaluateArrayExpressions(name);

  string className = extractToChar(propertyName, '.');

  Object* obj;
  if (!script->getObjectTables()->getValue(className, obj))
    return false;

  return propertyParser.getValue(obj, propertyName, value);
}

template <class T>
bool ParameterSymbolTable<T>::setValue(const std::string name, T value)
{
  if (name.size() == 0) return false;
  if (name[0] == '$' || name[0] == '%') return false;

  string propertyName = script->evaluateArrayExpressions(name);

  string className = extractToChar(propertyName, '.');
  Object* obj;
  if (!script->getObjectTables()->getValue(className, obj))
    return false;

  return propertyParser.setValue(obj, propertyName, value);
}

}

#endif