//---------------------------------------------------------------------------
#ifndef AttachedDataH
#define AttachedDataH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "HashTables.h"
#include "PAssert.h" 
#include <string>
#include <vector>
//---------------------------------------------------------------------------

namespace geoxl {
   using namespace std;

class MemoryManager;
class MemoryManagedObject;
class AttachedDataContainer;



class FOUNDATION_API AttachedData : public Persistent{
 private:
   static MetaClass *AttachedDataMetaClass;
   static Object* newInstance();

   string m_name;
   card32 m_flags;
   MemoryManagedObject* m_memManObj;

   AttachedDataContainer   *  m_MyContainer;

 protected:

   static void delCallback(string name, AttachedData *data);

 friend AttachedDataContainer;

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   enum AttachedDataFlags { ADF_PERSISTENT = 1, ADF_DELETEONUPDATE = 2, ADF_QUERY_DELETE_ON_UPDATE = 4 };

   enum ChangeType { CT_ATTRIBUTE_CHANGED,    // some attribute channel other than "position" changed
                     CT_POSITION_CHANGED,     // some attribute channel including "position" changed
                     CT_ADD_DELETE,           // objects have been added/removed
                     CT_EVERYTHING_CHANGED }; // more involved/unknown changes

   AttachedData();
   virtual ~AttachedData();
   /// setup instance: flags & ADF_PERSISTENT --> data will be saved along with container
   /// flags & ADF_DELETEONUPDATE --> data will be deleted if the object the data is attached to changes
   /// flags & ADF_QUERY_DELETE_ON_UPDATE --> attachment will be informed of changes and can decide if it should be deleted
   /// MemoryManager: if not null, the object will be deleted if memory pool runs out of resources
   /// hint: usually ADF_PERSISTENT and (ADF_DELETEONUPDATE || memManager) are used mutually exclusive
   virtual void setup(const string& name, const card32 flags, MemoryManager *memManager = NULL);

   virtual bool queryDelete(ChangeType whatChanged) {return false;}

   inline card32 getFlags() const {return m_flags;}
   inline void setFlags(card32 f) {m_flags = f;}

   inline const string& getName() const {return m_name;}

   virtual void assign(const Object* obj, CopyContext *context = NULL);
   /// size of this attachment for memory management
   /// (to-do: currently static over att. lifetime)
   /// (default = 1, allows simple instance counting)
   virtual card32 getSize() const {return 1;}


   friend MemoryManager;
   friend AttachedDataContainer;
};



class FOUNDATION_API AttachedDataContainer : public Persistent {
 private:
   static MetaClass *AttachedDataContainerMetaClass;
   static Object* newInstance();

   class FOUNDATION_API StringComp {
     public:
       static bool equals(string obj1, string obj2) { return obj1 == obj2; }
   };

   class MyHashTable : public HashTable<string, AttachedData*, StringComp> {
    protected:
      virtual max_platform_card hashFunction(string key) const;
    public:
      MyHashTable(card32 fillFactor = 1, card32 startSizeIndex = 8);

   };
   mutable MyHashTable *m_dataHashTable;

 public:
   AttachedDataContainer();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   AttachedData* getData(const string& name) const;
   void attachData(AttachedData* data);
   void deleteData(const string &name);
   /// works like deleteData, but without deleting attachment
   void removeData(const string &name);

   void getAll( vector<AttachedData*> &ret );

   void clear();

   /// deletes all attached data which is marked "ADF_DELETEONUPDATE" and possibly/checks for "ADF_QUERY_DELETE_ON_UPDATE"
   void deleteDelOnWriteData(AttachedData::ChangeType change = AttachedData::CT_EVERYTHING_CHANGED);
   /// do not use!
   AttachedData::ChangeType temp_change_for_callback;

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   virtual ~AttachedDataContainer();

	/// sum all attachment data
	mpcard getSizeInMemory();
};

class FOUNDATION_API StringAttachment : public AttachedData
{
 private:
   static MetaClass *StringAttachmentMetaClass;
   static Object* newInstance();

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   string getString() {return m_String;}
   void setString( string s ) {m_String = s;}

   virtual void assign(const Object* obj, CopyContext *context = NULL);

	string   m_String;

};



}


#endif
