//---------------------------------------------------------------------------
#ifndef ObjectH
#define ObjectH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MetaClass.h"
#include "Exceptions.h"
//---------------------------------------------------------------------------

namespace geoxl {


/// provides additional context information for copying
class FOUNDATION_API CopyContext {
 private:
   struct Entry {
      string name;
      Object *data;
   };
   vector<Entry> tagStack;

 public:
   Object *getData(string name);
   void pushData(string name, Object *obj);
   void popData();
};

class FOUNDATION_API ECannotAssign : public ETypeMismatch {
 public:
   ECannotAssign(const string msg) : ETypeMismatch(msg) {};
};

class FOUNDATION_API Object {
 private:
   static MetaClass* objectMetaClass;
 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();
   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual bool equal(const Object* obj) const;
   virtual card32 getFieldsHashValue() const;
   virtual Object* copy(CopyContext *context = NULL) const;
   virtual ~Object() {}
};

/**
	Declares an abstract class - example:
	class MyClass : public Persistent
	{
		GEOXL_ABSTRACT_CLASS( MyClass )

	public:
		...
	};
*/
#define GEOXL_ABSTRACT_CLASS( className ) \
private: \
	static geoxl::MetaClass* className##MetaClass;\
public:\
   virtual const geoxl::MetaClass *getInstanceClass() const;\
   static const geoxl::MetaClass *getClass();\
   static void init(const geoxl::MetaClass *parentClass);\
	static void initMetaClass();\
   static void shutdown();\

/**
	Implements an abstract class - example (MyClass.cpp):
	
	IMPLEMENT_GEOXL_ABTRACT_CLASS( MyClass, 3 ) // version is MetaClass::FIRST_VERSION + 3 here !
	{
		MyClassMetaClass->addProperty...
	}

	MyClass::MyClass() {}
	...

*/
#define IMPLEMENT_GEOXL_ABSTRACT_CLASS_GENERIC( className, version, newInstance ) \
geoxl::MetaClass* className##::className##MetaClass = NULL;\
const geoxl::MetaClass* className##::getInstanceClass() const {\
   return getClass();\
}\
const geoxl::MetaClass* className##::getClass() {\
   if (className##MetaClass==NULL) {\
		className##MetaClass = new geoxl::MetaClass( #className , "no description.",\
                                  newInstance, sizeof( className ));\
   }\
   return className##MetaClass;\
}\
void className##::shutdown() {\
   delete className##MetaClass;\
}\
void className##::init(const geoxl::MetaClass *parentClass) \
{\
   ((geoxl::MetaClass*)getClass())->init(parentClass, geoxl::MetaClass::FIRST_VERSION + version );\
   className##MetaClass->copyParentProperties();\
	initMetaClass();\
}\
void className##::initMetaClass()

#define IMPLEMENT_GEOXL_ABSTRACT_CLASS( className, version )  IMPLEMENT_GEOXL_ABSTRACT_CLASS_GENERIC( className, version, NULL )


/**
	Declares a class - example:
	class MyClass : public Persistent
	{
		GEOXL_CLASS( MyClass )

	public:
		...
	};
*/
#define GEOXL_CLASS( className ) \
private:\
   static geoxl::Object *newInstance();\
	GEOXL_ABSTRACT_CLASS( className )


/**
	Implements a class - example (MyClass.cpp):
	
	IMPLEMENT_GEOXL_CLASS( MyClass, 3 ) // version is MetaClass::FIRST_VERSION + 3 here !
	{
		// init properties and methods here
		BEGIN_CLASS_INIT( MyClass );

		MyClassMetaClass->addProperty...
	}

	MyClass::MyClass() {}
	...

*/
#define IMPLEMENT_GEOXL_CLASS( className, version ) \
	geoxl::Object* className##::newInstance() {\
	return new className ;\
} IMPLEMENT_GEOXL_ABSTRACT_CLASS_GENERIC( className, version, newInstance )

#define BEGIN_CLASS_INIT( MyClass ) MyClass * obj_null_pointer = NULL;
#define SET_USER_EDIT( b ) ((geoxl::MetaClass *)getClass())->actProperty_setUserEdit(b);
#define SET_EDIT_READONLY( b ) ((geoxl::MetaClass *)getClass())->actProperty_setEditReadonly(b);
#define SET_STREAM_PROPERTY( b ) ((geoxl::MetaClass *)getClass())->actProperty_setStreamProperty(b);
#define SET_NAME( name ) ((geoxl::MetaClass *)getClass())->actProperty_setPropertyName(name);
#define SET_EDITOR( name ) ((geoxl::MetaClass *)getClass())->actProperty_addPropertyEditor(name, true);

#define GEOXL_MEMBER( mType, mName ) \
private: \
	mType m_##mName ; \
public: \
	void set##mName##( mType v ) { m_##mName = v; } \
	mType get##mName##() const {return m_##mName ;}


}

#endif
