//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "ObjectDictionary.h"
#include "ObjectDictionaryEditorWidget.h"
#include "BinaryObjectStreams.h"
#include "Exceptions.h"
#include "StringHelper.h"
#include "pAssert.h"
#include "Tools.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

using namespace std;
using namespace geoxl;

MetaClass* ObjectDictionary::ObjectDictionaryMetaClass = NULL;

/****************************************************************************
 * ObjectDictionary
 ****************************************************************************/
const MetaClass* ObjectDictionary::getInstanceClass() const
{
  return getClass();
}

Object* ObjectDictionary::newInstance()
{
  return new ObjectDictionary;
}

const MetaClass* ObjectDictionary::getClass()
{
  if (ObjectDictionaryMetaClass == NULL) {
    ObjectDictionaryMetaClass = new MetaClass("ObjectDictionary", "no description.",
                                    newInstance, sizeof(ObjectDictionary));
  }
  return ObjectDictionaryMetaClass;
}

void ObjectDictionary::init(const MetaClass *parentClass)
{
  ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION);
  ObjectDictionaryMetaClass->copyParentProperties();
  ObjectDictionaryMetaClass->getGUIDescriptor()->setIcon(":/foundation/Images/Classes/ObjectDictionary.png");
}

void ObjectDictionary::shutdown()
{
  delete ObjectDictionaryMetaClass;
}

void ObjectDictionary::write(OutputObjectStream *s) const
{
  card32 numObjects = 0;
  for (card32 i = 0; i < objVector.size(); ++i) {
    Persistent* p = dynamic_cast<Persistent*>(objVector[i].obj);
    if (p != NULL) ++numObjects;
  }

  s->write(numObjects);
  for (card32 i = 0; i < numObjects; ++i) {
    Persistent* p = dynamic_cast<Persistent*>(objVector[i].obj);
    if (p != NULL) {
      s->writeObject(p);
      s->writeString(objVector[i].name);
    }
  }
}

void ObjectDictionary::read(InputObjectStream *s)
{
  card32 numObjects;
  s->read(numObjects);
  objVector.resize(numObjects);

  for (card32 i = 0; i < numObjects; ++i) {
    s->readObject((Persistent*&)objVector[i].obj);
    s->readString(objVector[i].name);
    objVector[i].owned = true;
  }
}

void ObjectDictionary::assign(const Object* obj, CopyContext *context)
{
  ObjectDictionary* other = (ObjectDictionary*)obj;
  pAssert(obj->getInstanceClass() == getInstanceClass());

  clear();

  objVector.resize(other->objVector.size());
  for (card32 i = 0; i < other->objVector.size(); ++i) {
    if (!other->objVector[i].owned)
      objVector[i].obj = other->objVector[i].obj;
    else
      objVector[i].obj = other->objVector[i].obj->copy();
    objVector[i].name = other->objVector[i].name;
    objVector[i].owned = other->objVector[i].owned;
  }
  forget(context);
}

ObjectDictionary::ObjectDictionary()
{
  rootClass = Object::getClass();
}

ObjectDictionary::~ObjectDictionary()
{
  for (unsigned i = 0; i < objVector.size(); ++i) {
    if (objVector[i].owned) delete objVector[i].obj;
  }
}

/****************************************************************************/
//  ObjectDictionary - newObject
//  Creates a new instance of the class given by the metaClassName with the
//  name given
/// \param metaClassName Class type for the instance that will be created
/// \param name Name of the instance
/// \param owned TRUE if the ObjectDictionary is allowed to delete the instance
/****************************************************************************/
std::string ObjectDictionary::newObject(std::string metaClassName, std::string name, bool owned)
{
   const MetaClass *mc = MetaClass::getClassByName(metaClassName);
   if (mc == NULL) throw ERangeCheck("ObjectDictionary::newObject - invalid class name");
   return newObject(mc, name, owned);
}

/****************************************************************************/
//  ObjectDictionary - newObject
//  Creates a new instance of the class given in metaClass
/// \param metaClass Class type for the instance that will be created
/// \param name Name of the instance
/// \param owned TRUE if the ObjectDictionary is allowed to delete the instance
/****************************************************************************/
std::string ObjectDictionary::newObject(const MetaClass* metaClass, std::string name, bool owned)
{
  ObjListElem elem;
  elem.obj = metaClass->newInstance();
  if (name != "")
    elem.name = name;
  else
    elem.name = generateName(metaClass);
  elem.owned = owned;
  objVector.push_back(elem);
  return elem.name;
}

/****************************************************************************/
//  ObjectDictionary - deleteObject
//  Deletes the instance of the object and removes it from the dictionary.
/// Returns TRUE if successful and FALSE otherwise (if not owned by the
/// object dictionary). Throws an exception if the object name does not
/// exist.
/****************************************************************************/
bool ObjectDictionary::deleteObject(std::string name)
{
  card32 objNum = objectByName(name);
  if (!objVector[objNum].owned) return false;

  delete objVector[objNum].obj;
  objVector.erase(objVector.begin() + objNum);
  return true;
}

/****************************************************************************/
//  ObjectDictionary - deleteObject
//  Deletes the instance of the object and removes it from the dictionary.
/// Returns TRUE if successful and FALSE otherwise (if not owned by the
/// object dictionary or if the object given is not in the object dictionary).
/****************************************************************************/
bool ObjectDictionary::deleteObject(Object* obj)
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].obj == obj && objVector[i].owned) {
      delete objVector[i].obj;
      objVector.erase(objVector.begin() + i);
      return true;
    }
  }

  return false;
}

/****************************************************************************/
//  ObjectDictionary - addObject
//  Adds an existing object to the dictionary
/// \param obj Object instance that will be added to the dictionary
/// \param name Name of the instance
/// \param owned TRUE if the ObjectDictionary is allowed to delete the
///              instance
/****************************************************************************/
std::string ObjectDictionary::addObject(Object* obj, std::string name, bool owned)
{
  ObjListElem elem;
  elem.obj = obj;
  if (name == "")
    elem.name = generateName(obj->getInstanceClass());
  else
  {
	  if( !nameExists( name ) )
			elem.name = name;
	  else
	  {
		  int i=1;
		  string genName = name + intToStr(i);
		  while( nameExists(genName) )
		  {
			  i++;
			  genName = name + intToStr(i);
		  }
		  elem.name = genName;
	  }
  }
  elem.owned = owned;
  objVector.push_back(elem);
  return elem.name;
}

/****************************************************************************/
//  ObjectDictionary - removeObject
//  Removes the object with the name given from the object dictionary.
/// Does NOT delete the object instance. If that is the desired behavior,
/// deleteObject can be used, but only if the object is owned by the object
/// dictionary. Otherwise you will have to remove it and then delete it
/// elsewhere. Throws an exception if the object name does not exist.
/****************************************************************************/
void ObjectDictionary::removeObject(std::string name)
{
  card32 objNum = objectByName(name);

  objVector.erase(objVector.begin() + objNum);
}

/****************************************************************************/
//  ObjectDictionary - renameObject
/// If the object (oldName) does not exist, an exception is thrown. If an
/// object with the name newName already exists, the method will return
/// false and the object will not be renamed. If the renaming process was
/// successful, the method will return true.
/****************************************************************************/
bool ObjectDictionary::renameObject(std::string oldName, std::string newName)
{
  if (nameExists(newName)) return false;
  card32 objNum = objectByName(oldName);
  objVector[objNum].name = newName;
  return true;
}

/****************************************************************************/
//  ObjectDictionary - clear
//  Clears the whole object dictionary
/****************************************************************************/
void ObjectDictionary::clear()
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].owned) delete objVector[i].obj;
    objVector[i].obj = NULL;
  }
  objVector.clear();
}

/****************************************************************************/
//  ObjectDictionary - replaceObject
/// It also renames the object in the process if the string newName is not
/// empty. Throws an exception if the object oldObject is not in the object
/// dictionary.
/****************************************************************************/
void ObjectDictionary::replaceObject(Object *oldObject, Object *newObject, string newName)
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].obj == oldObject) {
      objVector[i].obj = newObject;
      if (newName != "") objVector[i].name = newName;
      return;
    }
  }

  throw EObjDictionaryException("Object dictionary.replaceObject: Invalid object");
}

/****************************************************************************/
//  ObjectDictionary - getObjectSafe
//  Returns a pointer to the object with the name given in obj.
/// Method does not throw any exceptions, but will return false if no object
/// with the name given exists. If an object could be found, it will be
/// returned in obj and the method returns true.
/// If you know that the object exists, you can use getObject insted.
/****************************************************************************/
bool ObjectDictionary::getObjectSafe(std::string name, Object* &obj) const
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].name == name) {
      obj = objVector[i].obj;
      return true;
    }
  }

  obj = NULL;
  return false;
}

/****************************************************************************/
//  ObjectDictionary - getObject
//  Returns the pointer to the object given.
/// Throws an exception if the object does not exist. If that is not the
/// desired behavior, use getObjectSafe instead.
/****************************************************************************/
Object* ObjectDictionary::getObject(std::string name) const
{
  return objVector[objectByName(name)].obj;
}

/****************************************************************************/
//  ObjectDictionary - nameExists
//  Returns true if the given name already exists and false otherwise
/****************************************************************************/
bool ObjectDictionary::nameExists(std::string name) const
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].name == name)
      return true;
  }

  return false;
}

/****************************************************************************/
//  ObjectDictionary - objectByName [PRIVATE]
//  Returns the object number for the object name given (internal use)
/****************************************************************************/
card32 ObjectDictionary::objectByName(std::string name) const
{
  for (card32 i = 0; i < objVector.size(); ++i) {
    if (objVector[i].name == name)
      return i;
  }

  throw EObjDictionaryException("ObjectDictionary: Invalid object name " + name);
}

/****************************************************************************/
//  ObjectDictionary - generateName [PRIVATE]
//  Generates a new unique name for a new instance of the metaClass.
/// A new name for an instance of the metaClass given will be generated. This
/// name will be a valid name for the new object.
/// The name will follow the scheme "MetaClass name + number",
/// e.g.: ScriptEngine1
/****************************************************************************/
std::string ObjectDictionary::generateName(const MetaClass* metaClass)
{
  string mcName = metaClass->getClassName();
  card32 count = 1;

  string newName;
  Object* obj;
  do {
    newName = mcName + numToStr(count);
    ++count;
  } while (getObjectSafe(newName, obj));

  return newName;
}

/****************************************************************************/
//  ObjectDictionary - isOwned
//  Returns whether the object given is owned by the dictionary.
/// Throws an exception if the object name given is not assigned to an
/// object.
/****************************************************************************/
bool ObjectDictionary::isOwned(std::string name) const
{
  return (objVector[objectByName(name)]).owned;
}

card32 ObjectDictionary::getNumObjects() const
{
	return (card32)objVector.size();
}

Object * ObjectDictionary::getObject( card32 i ) const
{
	pAssert(i < objVector.size());
	return objVector[i].obj;
}

std::string ObjectDictionary::getObjectName( card32 i ) const
{
	pAssert(i < objVector.size());
	return objVector[i].name;
}

void ObjectDictionary::moveObject(card32 fromPos, card32 toPos)
{
   pAssert(fromPos < objVector.size() && toPos < objVector.size());
   if (fromPos == toPos) return;
   
   ObjListElem elem = objVector[fromPos];
   objVector.erase(objVector.begin() + fromPos);
   objVector.insert(objVector.begin() + toPos, elem);
}