#include "stdafx.h"

//---------------------------------------------------------------------------
#include "PropertyParser.h"
//---------------------------------------------------------------------------
#include "pAssert.h"
#include "ObjectClassProperty.h"
#include "EmbeddedObjectClassProperty.h"
#include "EmbeddedObjectClassProperty_templCode.h"
#include "StringHelper.h"
#include "ExpressionParser.h"
#include "NumericalClassProperty.h"
#include "NumericalClassProperty_templCode.h"
#include "BooleanClassProperty.h"
#include "StringClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "VariableArrayClassProperty.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace std;
using namespace geoxl;
//---------------------------------------------------------------------------


/****************************************************************************/
//  PropertyParser - getValue [for numbers]
//  Returns the value of the number property given
/****************************************************************************/
bool PropertyParser::getValue(Object* obj, std::string property, ExpressionNum& value)
{
  PropertyTypes properties = getProperty(obj, property);
  if (properties.numeric != NULL) {
    value = properties.numeric->getValueDouble(obj, properties.arrayOffset);
    return true;
  }

  if (properties.boolean != NULL) {
    value = properties.boolean->getValue(obj, properties.arrayOffset);
    return true;
  }      
  
  if (properties.fixedArray != NULL) {
    value = properties.fixedArray->getArraySize();
    return true;
  } 
  if (properties.varArray != NULL) {
    value = properties.varArray->getNumElements(obj);
    return true;
  }

  // no property found
  return false;
}

/****************************************************************************/
//  PropertyParser - setValue [for numbers]
//  Sets the value of the number property given to the value
/****************************************************************************/
bool PropertyParser::setValue(Object* obj, std::string property, ExpressionNum value)
{
  PropertyTypes properties = getProperty(obj, property);
  bool success = false;
  if (properties.numeric != NULL) {
    properties.numeric->setValueDouble(obj, properties.arrayOffset, value);
    success = true;
  } else {
    if (properties.boolean != NULL) {
      properties.boolean->setValue(obj, properties.arrayOffset, value);
      success = true;
    } else {
      if (properties.fixedArray != NULL) {
        throw EPropertyParserException("PropertyParser: Array size is not changeable (FixedArrayClassProperty)");
		}

      if (properties.varArray != NULL) {
         properties.varArray->setSize(obj, card32(value));
         success = true;
		}
    }
  }

  if (success)
    return true;
  else
    return false;
}

/****************************************************************************/
//  PropertyParser - getValue  [for strings]
//  Returns the value of the string property given
/****************************************************************************/
bool PropertyParser::getValue(Object* obj, std::string property, std::string& value)
{
  PropertyTypes properties = getProperty(obj, property);
  if (properties.stringProp != NULL) {
    value = properties.stringProp->toString(obj, properties.arrayOffset);
    return true;
  } else {
    // Is not a string property
    return false;
  }
}

/****************************************************************************/
//  PropertyParser - setValue  [for strings]
//  Sets the value of the string property given to the value
/****************************************************************************/
bool PropertyParser::setValue(Object* obj, std::string property, std::string value)
{
  PropertyTypes properties = getProperty(obj, property);
  if (properties.stringProp != NULL) {
    properties.stringProp->fromString(obj, properties.arrayOffset, value);
    return true;
  } else {
    return false;
  }
}

/****************************************************************************/
//  PropertyParser - getProperty [PRIVATE]
//  Returns the property of the object obj with the name given.
/// The method also returns the actual object that contains the
/// property.
/****************************************************************************/
PropertyParser::PropertyTypes PropertyParser::getProperty(Object* &obj,
                                                std::string propertyName)
{
  arrayIndex = 0;
  isArrayProperty = false;

  PropertyTypes emptyProperties;
  emptyProperties.numeric = NULL;
  emptyProperties.boolean = NULL;
  emptyProperties.stringProp = NULL;
  emptyProperties.fixedArray = NULL;
  emptyProperties.varArray = NULL;
  emptyProperties.arrayOffset = 0;

  if (!descendProperties(obj, propertyName)) {
    obj = NULL;
    return emptyProperties;
  }

  // We now either have obj pointing to the last object in the hierarchy
  // and propertyName without a .
  // If it does have a ., it is an array property (".count")

  bool isArrayProp = false;
  bool isArrayCount = false;
  string arrayIndex;
  if (propertyName.find('[') != string::npos) {
    // propertyName contains a [ -> must be an array entry
    isArrayProp = true;
    arrayIndex = extractToChar(propertyName, '[');
    swap(arrayIndex, propertyName);
  }
  if (propertyName.find('.') != string::npos) {
    // propertyName contains a . -> must be .count of an array property
    string arrayName = extractToChar(propertyName, '.');
    if (strLCase(propertyName) != "count") {
      throw EPropertyParserException("Property Parser: Array property can only have .count sub-property");
    }
    propertyName = arrayName;
    isArrayCount = true;
  }

  // Check if a property with that name exists in our object's metaclass
  const ClassProperty* cp;
  const MetaClass* mc = obj->getInstanceClass();
  bool found = false;
  for (card32 i = 0; i < mc->getNumProperties(); ++i) {
    if (mc->getProperty(i)->getCName() == propertyName) {
      cp = mc->getProperty(i);
      found = true;
      break;
    }
  }
  if (!found) {
    // Invalid property; doesn't exist in our class
    obj = NULL;
    return emptyProperties;
  }

  // Property exists -> return the property
  PropertyTypes properties = emptyProperties;

  if (isArrayProp) {
    // Resolve array property
    cp = getArrayOffset(obj, cp, arrayIndex, properties.arrayOffset);
  }

  if (isArrayCount) {
    // Return array property for count
    properties.fixedArray = dynamic_cast<const FixedArrayClassProperty*>(cp);
    properties.varArray = dynamic_cast<const VariableArrayClassProperty*>(cp);
    if (properties.fixedArray != NULL || properties.varArray != NULL)
		 return properties;
	 else {
		 obj = NULL;
		 return emptyProperties;
	 }
  }

  properties.numeric = dynamic_cast<const NumericalClassPropertyBase*>(cp);
  if (properties.numeric != NULL) {
    // Is numerical property
    return properties;
  } else {
    properties.boolean = dynamic_cast<const BooleanClassProperty*>(cp);
    if (properties.boolean != NULL) {
      // Is boolean property
      return properties;
    } else {
      properties.stringProp = dynamic_cast<const StringClassProperty*>(cp);
      if (properties.stringProp != NULL) {
        // Is string property
        return properties;
      } else {
        // Is unsupported property
        obj = NULL;
        return emptyProperties;
      }
    }
  }
}

/****************************************************************************/
//  PropertyParser - descendProperties [PRIVATE]
//  Recursive method that descends all the way to the last propertie given
/// For a string that has several dots, this method descends down the the
/// last object property in the string. Returns true if the property string
/// is valid and false otherwise. The result will be a pointer to the
/// last object in the property string and the actual property name.
/// E.g.: In: "scene.camera.position[3]", Out: "position[3]"
/****************************************************************************/
bool PropertyParser::descendProperties(Object* &obj, std::string& property)
{
  if (property.find('.') != string::npos) {
    // Get property from subclass
    string className = extractToChar(property, '.');

    const ClassProperty* cp;
    const MetaClass* mc = obj->getInstanceClass();

    // Try to find out whether this property is an array property
    bool isArray = false;
    string arrayIndex;
    if (property != "") {   // don't bother if it is the last one in the list
      if (className.find('[') != string::npos) {
        isArray = true;
        arrayIndex = extractToChar(className, '[');
        swap(arrayIndex, className);
      }
    }

    bool found = false;
    for (card32 i = 0; i < mc->getNumProperties(); ++i) {
      if (mc->getProperty(i)->getCName() == className) {
        cp = mc->getProperty(i);
        found = true;
      }
    }
    if (!found) {
      return false;
    }

    // Resolve object array
    if (isArray) 
	 {
/*      const GeneralArrayClassProperty* acp;
      acp = dynamic_cast<const GeneralArrayClassProperty*>(cp);
      if (acp == NULL)
        throw EPropertyParserException("PropertyParser: " + className + " is not an array.");

      cp = getArrayElementProperty(obj, cp, arrayIndex);

      isArrayProperty = false;  // only used for setValue*/
    }

    const ObjectClassProperty* classProp = dynamic_cast<const ObjectClassProperty*>(cp);

    if (classProp == NULL) {
      // Embedded object class property
      const GeneralEmbeddedObjectClassProperty* embClassProp;
      embClassProp = dynamic_cast<const GeneralEmbeddedObjectClassProperty*>(cp);

      if (embClassProp == NULL) {
        // Array property with .count
        const FixedArrayClassProperty* facp;
        facp = dynamic_cast<const FixedArrayClassProperty*>(cp);
        const VariableArrayClassProperty* vacp;
        vacp = dynamic_cast<const VariableArrayClassProperty*>(cp);
        if (facp == NULL && vacp == NULL) {
          return false;
        } else {
          property = className + ".count";
          return true;
        }
      }
      
      obj = dynamic_cast<Object*>(embClassProp->getSubObjectAdress(obj,NULL));

      if (obj != NULL)
        return descendProperties(obj, property);
      else
        return false;

    } else {
      // Object class property
      obj = classProp->getValue(obj,NULL);
      return descendProperties(obj, property);
    }
  } else {
    return true;
  }
}

/****************************************************************************/
//  PropertyParser - getArrayElementProperty [PRIVATE]
//  Returns the element class property
/// This method returns the class property that is at the array position given
/// \param cp Array class property that is to be evaluated
/// \param arrayIndexStr index of the element that is to be returned, has to
///                      be of the form "[123]"
/// \param arrayOffset returns the offset for the array element
/****************************************************************************/
const ClassProperty* PropertyParser::getArrayOffset(Object* &obj, 
                       const ClassProperty* cp, string arrayIndexStr,
                       pointer_diff &arrayOffset)
{
  if (arrayIndexStr.size() == 0) {
    throw EPropertyParserException("Property Parser: Array index syntax error");
  }
  if (arrayIndexStr[arrayIndexStr.size()-1] != ']') {
    throw EPropertyParserException("Property Parser: Array index syntax error");
  }

  arrayIndexStr.erase(arrayIndexStr.size()-1, 1);
  ExpressionNum arrayNumDouble = 0;
  if (!strToNum(arrayIndexStr, arrayNumDouble)) {
    throw EPropertyParserException("Property Parser: Invalid number for array index (" + arrayIndexStr + ")");
  }

  const FixedArrayClassProperty* facp = NULL;
  facp = dynamic_cast<const FixedArrayClassProperty*>(cp);
  const VariableArrayClassProperty* vacp = NULL;
  vacp = dynamic_cast<const VariableArrayClassProperty*>(cp);
  if (facp == NULL && vacp == NULL) {
    // cp is not an array property
    return NULL;
  }

  arrayIndex = (card32)arrayNumDouble;
  isArrayProperty = true;

  ClassProperty* elemCP = NULL;
  if (facp != NULL) {
    // cp is fixed array class property
    if (facp->getArraySize() <= arrayIndex)
      throw EPropertyParserException("Property Parser: Fixed array index out of bounds");

     elemCP = facp->getElementProperty();
    arrayOffset = facp->getOffset() + arrayIndex * facp->getStride();
  } else {
    // cp is variable array class property
    if (vacp->getNumElements(obj) <= arrayIndex)
      throw EPropertyParserException("Property Parser: Variable array index out of bounds");
    
    elemCP = vacp->getElementClassProperty();
    arrayOffset = (pointer_diff)vacp->getElement(obj, arrayIndex);
    obj = NULL;
  }
  
  return elemCP;
}
