//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "KinectCapture.h"
#include "PTypes.h"
typedef geoxl::card8 uint8_t;
typedef geoxl::card16 uint16_t;
#include <OpenNI.h>
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS(KinectCapture, 1) {
	BEGIN_CLASS_INIT(KinectCapture);
}

KinectCapture::KinectCapture()
{
   device = NULL;
   depthStream = NULL;
   colorStream = NULL;
   depthFrame = NULL;
   colorFrame = NULL;
   depthHistogramm = new float32[MAX_DEPTH_HISTO];
   currentFrame = new PointSet();
   VertexDescriptor vd;
   vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("depth",    1, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("color",    3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("flags",    1, VAD::DATA_FORMAT_INT32));
   currentFrame->clearAndSetup(640, 480, &vd);
}

void KinectCapture::startCapture()
{
	openni::Status rc = openni::STATUS_OK;

	const char* deviceURI = openni::ANY_DEVICE;
	static bool alreadyInitialized = false;
	if (!alreadyInitialized) {
		rc = openni::OpenNI::initialize();
		debugOutput << "After initialization: " << openni::OpenNI::getExtendedError() << "\n";
		alreadyInitialized = true;
	}

	if (!device) device = new openni::Device();
	rc = device->open(deviceURI);
	if (rc != openni::STATUS_OK)
	{
		debugOutput << "KinectCapture::startCapture() -- Device open failed: " << openni::OpenNI::getExtendedError() << "\n";
		openni::OpenNI::shutdown();
		throw EInvalidState("Device open failed");
	}

	if (!depthStream) depthStream   = new openni::VideoStream();
	rc = depthStream->create(*device, openni::SENSOR_DEPTH);
	if (rc == openni::STATUS_OK)
	{
		rc = depthStream->start();
		if (rc != openni::STATUS_OK)
		{
			debugOutput << "KinectCapture::startCapture() -- Couldn't start depth stream: " << openni::OpenNI::getExtendedError() << "\n";
			depthStream->destroy();
			throw EInvalidState("Could not start depth stream.");
		}
	}
	else
	{
		debugOutput << "KinectCapture::startCapture() -- Couldn't find depth stream: " << openni::OpenNI::getExtendedError() << "\n";
	}

	if (!colorStream) colorStream   = new openni::VideoStream();
	rc = colorStream->create(*device, openni::SENSOR_COLOR);
	if (rc == openni::STATUS_OK)
	{
		rc = colorStream->start();
		if (rc != openni::STATUS_OK)
		{
			debugOutput << "KinectCapture::startCapture() -- Couldn't start color stream:\n%s\n" << openni::OpenNI::getExtendedError() << "\n";
			colorStream->destroy();
		}
	}
	else
	{
		debugOutput << "KinectCapture::startCapture() -- Couldn't find color stream: " << openni::OpenNI::getExtendedError() << "\n";
	}

	if (!depthStream->isValid() || !colorStream->isValid())
	{
		debugOutput << "KinectCapture::startCapture() -- No valid streams." << "\n";
		openni::OpenNI::shutdown();
		throw EInvalidState("no valid streams");
	}
	
	if (!depthFrame) depthFrame = new openni::VideoFrameRef();
	if (!colorFrame) colorFrame = new openni::VideoFrameRef();

	openni::VideoMode depthVideoMode;
	openni::VideoMode colorVideoMode;

	if (depthStream->isValid() && colorStream->isValid())
	{
		depthVideoMode = depthStream->getVideoMode();
		colorVideoMode = colorStream->getVideoMode();

	}
	else if (depthStream->isValid())
	{
		depthVideoMode = depthStream->getVideoMode();
	}
	else if (colorStream->isValid())
	{
		colorVideoMode = colorStream->getVideoMode();
	}
	else
	{
		throw EInvalidState("Error - expects at least one of the streams to be valid...");
	}

}



typedef openni::VideoStream *StreamPtr;


//static void calculateHistogram(float* pHistogram, int histogramSize, const openni::VideoFrameRef& frame)
//{
//	const openni::DepthPixel* pDepth = (const openni::DepthPixel*)frame.getData();
//	// Calculate the accumulative histogram (the yellow display...)
//	memset(pHistogram, 0, histogramSize*sizeof(float));
//	int restOfRow = frame.getStrideInBytes() / sizeof(openni::DepthPixel) - frame.getWidth();
//	int height = frame.getHeight();
//	int width = frame.getWidth();
//
//	unsigned int nNumberOfPoints = 0;
//	for (int y = 0; y < height; ++y)
//	{
//		for (int x = 0; x < width; ++x, ++pDepth)
//		{
//			if (*pDepth != 0)
//			{
//				pHistogram[*pDepth]++;
//				nNumberOfPoints++;
//			}
//		}
//		pDepth += restOfRow;
//	}
//	for (int nIndex=1; nIndex<histogramSize; nIndex++)
//	{
//		pHistogram[nIndex] += pHistogram[nIndex-1];
//	}
//	if (nNumberOfPoints)
//	{
//		for (int nIndex=1; nIndex<histogramSize; nIndex++)
//		{
//			pHistogram[nIndex] = (256 * (1.0f - (pHistogram[nIndex] / nNumberOfPoints)));
//		}
//	}
//}



const PointSet * KinectCapture::getNextFrame(ChangeType &whatChanged)
{
   if (!currentFrame) {
      throw EInvalidState("expecting currentFrame != 0");
   }
   AAT colorAAT    = currentFrame->getAAT("color");
   AAT depthAAT    = currentFrame->getAAT("depth");
   AAT positionAAT = currentFrame->getAAT("position");
   AAT flagsAAT    = currentFrame->getAAT("flags");

	int changedIndex;
   StreamPtr streams[2] = {colorStream, depthStream};
	openni::Status rc = openni::OpenNI::waitForAnyStream(streams, 2, &changedIndex);
	if (rc != openni::STATUS_OK || changedIndex < 0 || changedIndex >= 2)
	{
		debugOutput << "Wait failed\n";
      throw EInvalidState("OpenNI::waitForAnyStream failed.");
	}

	if (changedIndex == 0) {
	   colorStream->readFrame(colorFrame); 
	   if (colorFrame->isValid()) {
	   
	      const openni::RGB888Pixel* pImageRow = (const openni::RGB888Pixel*)colorFrame->getData();
         const int offsetY = colorFrame->getCropOriginY();
	      int rowSize = colorFrame->getStrideInBytes() / sizeof(openni::RGB888Pixel);

	      for (int y = 0; y < colorFrame->getHeight(); ++y) {
	         if (y >= 480) break;
		      const openni::RGB888Pixel* pImage = pImageRow;
		      int offsetX = colorFrame->getCropOriginX();

		      for (int x = 0; x < colorFrame->getWidth(); ++x) {
		         if (x >= 640) break;
               currentFrame->set3f(x+offsetX, y+offsetY, colorAAT, makeVector3f(pImage->r,pImage->g,pImage->b)*(1.0f/255.0f));
               ++pImage;
		      }

		      pImageRow += rowSize;
	      }
	   } else {
	      throw EInvalidState("invalid frame");
	   }
   } else if (changedIndex==1) {
	   depthStream->readFrame(depthFrame); 
	   if (depthFrame->isValid()) {
	      const openni::DepthPixel* pDepthRow = (const openni::DepthPixel*)depthFrame->getData();
         const int offsetY = depthFrame->getCropOriginY();
	      int rowSize = depthFrame->getStrideInBytes() / sizeof(openni::DepthPixel);

	      for (int y = 0; y < depthFrame->getHeight(); ++y) {
	         if (y >= 480) break;
		      const openni::DepthPixel* pDepth = pDepthRow;
		      int offsetX = depthFrame->getCropOriginX();

		      for (int x = 0; x < depthFrame->getWidth(); ++x, ++pDepth) {
		         if (x >= 640) break;
               currentFrame->set1f(x+offsetX, y+offsetY, depthAAT, *pDepth);
               
               Vector3f worldPos = NULL_VECTOR3F;
               openni::Status result = openni::CoordinateConverter::convertDepthToWorld(
                  *depthStream, x, y, *pDepth,
                  &worldPos[0], &worldPos[1], &worldPos[2]
               );
               if (result != openni::STATUS_OK) throw PException("conversion failed");
               currentFrame->set3f(x+offsetX, y+offsetY, positionAAT, worldPos);

		         if (*pDepth != 0) {
		            currentFrame->set1i(x+offsetX, y+offsetY, flagsAAT, 0);
		         } else {
		            currentFrame->set1i(x+offsetX, y+offsetY, flagsAAT, PF_FLAG_INVALID);
		         }
		      }

		      pDepthRow += rowSize;
	      }
	   } else {
	      throw EInvalidState("invalid frame");
	   }
	}

   if (changedIndex == 0) whatChanged = NewColor;
   if (changedIndex == 1) whatChanged = NewDepth;
   return currentFrame;
}

void KinectCapture::stopCapture()
{
   delete device;
   delete depthStream;
   delete colorStream;
   delete depthFrame;
   delete colorFrame;
   device = NULL;
   depthStream = NULL;
   colorStream = NULL;
   depthFrame = NULL;
   colorFrame = NULL;
}

KinectCapture::~KinectCapture()
{
   delete[] depthHistogramm;
}
