//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "GLContext.h"
//---------------------------------------------------------------------------
#include "GLObject.h"
#include "GLExtensions.h"
#include <glh/glh_extensions.h>
#include "Exceptions.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace std;
using namespace geoxl;
//---------------------------------------------------------------------------

#ifdef WIN32
	#define getCurrentContext() wglGetCurrentContext()
	#define makeContextCurrent(dc, rc) wglMakeCurrent(dc, rc)
#else
	#include <QtOpenGL/qgl.h>
	#include <glh/glh_genext.h>
	#define getCurrentContext() glXGetCurrentContext()
	#define makeContextCurrent(dc, rc) glXMakeCurrent(m_qglWidget->context()->device()->x11Display(), dc, rc)
#endif

card64 GLContext::globalContextNumber = 1;


CGcontext GLContext::m_CGContext = 0;

#ifdef WIN32
GLContext::GLContext(HDC dc, HGLRC rc,GLContext * parentContext, QGLWidget * glWidget)
#else
GLContext::GLContext(GLXDrawable dc, GLXContext rc, GLContext *parentContext, QGLWidget *glWidget)
#endif
{
   m_IsGlobalActive = false;
	m_UseVertexBufferObjects = true;

   myContextNumber = globalContextNumber++;
   m_ParentContext = parentContext;
   if (parentContext != NULL) {
      parentContext->addChildContext(this);
   }
   m_DC = dc;
   m_RC = rc;
   if( (m_CGContext == 0) && (parentContext == NULL) )
   {
      m_CGContext = cgCreateContext();

	  // pj, debugging 
	  //CGprofile cgVertexProfile = cgGLGetLatestProfile( CG_GL_VERTEX );
      initGLExtensions();
   }
	m_qglWidget = glWidget;
}

GLContext::~GLContext()
{
   onDeleteContext();
}

GLContext * GLContext::getRootContext()
{
   if( m_ParentContext )
      return m_ParentContext->getRootContext();
   else
      return this;
}

bool GLContext::isCompatible( GLContext * context )
{
   if( this->myContextNumber == context->myContextNumber )
      return true;

   if( m_ParentContext )
      return m_ParentContext->isCompatible(context);

   for( list<GLContext*>::iterator iter = m_ChildContexts.begin();
      iter != m_ChildContexts.end(); ++iter )
   {
      if( *iter == context )
         return true;
   }
   
   return false;
}

bool GLContext::activate(){
	if(getCurrentContext() != m_RC){
		setGlobalActive(makeContextCurrent(m_DC, m_RC));
		return isGlobalActive();
	} else {
		setGlobalActive( true );
		return true;
	}
}

void GLContext::deactivate()
{
   if( getCurrentContext() )
   {
      makeContextCurrent(0,0);
   }
   setGlobalActive( false );
}

bool GLContext::isGlobalActive() const
{
   if( m_ParentContext )
      return m_ParentContext->isGlobalActive();
   else
      return m_IsGlobalActive;
}

void GLContext::setGlobalActive(bool active)
{
   if( m_ParentContext )
      m_ParentContext->setGlobalActive(active);
   else
   {
      m_IsGlobalActive = active;
      if( m_IsGlobalActive )
      {
         delayedDelete();
      }
   }
}

void GLContext::delayedDelete()
{
   if( m_ParentContext )
   {
      m_ParentContext->delayedDelete();
      return;
   }

   if( m_DelayedDeleteTextures.size() )
      glDeleteTextures( (GLsizei)m_DelayedDeleteTextures.size(), &m_DelayedDeleteTextures[0] );
   if( m_DelayedDeleteBuffers.size() )
      glDeleteBuffersARB( (GLsizei)m_DelayedDeleteBuffers.size(), &m_DelayedDeleteBuffers[0] );
   if( m_DelayedDeleteShaders.size() )
      glDeleteProgramsARB( (GLsizei)m_DelayedDeleteShaders.size(), &m_DelayedDeleteShaders[0] );

   m_DelayedDeleteTextures.clear();
   m_DelayedDeleteBuffers.clear();
   m_DelayedDeleteShaders.clear();
}

void GLContext::deleteTexture( GLuint texID )
{
   if( m_ParentContext )
   {
      m_ParentContext->deleteTexture( texID );
      return;
   }

   if( isGlobalActive() )
      glDeleteTextures(1,&texID );
   else
      m_DelayedDeleteTextures.push_back(texID);
}

void GLContext::deleteShader( GLuint programID )
{
   if( m_ParentContext )
   {
      m_ParentContext->deleteShader( programID );
      return;
   }

   if( isGlobalActive() )
      glDeleteProgramsARB(1,&programID );
   else
      m_DelayedDeleteShaders.push_back(programID);
}

void GLContext::deleteBuffer( GLuint bufferID )
{
   if( m_ParentContext )
   {
      m_ParentContext->deleteBuffer( bufferID );
      return;
   }

   if( isGlobalActive() )
      glDeleteBuffersARB(1,&bufferID );
   else
      m_DelayedDeleteBuffers.push_back(bufferID);
}

list<GLObject *>::iterator GLContext::addObject( GLObject * obj )
{
   if( m_ParentContext != NULL )
      return m_ParentContext->addObject( obj );
   else
   {
      m_Objects.push_back( obj );
      list<GLObject *>::iterator ret = m_Objects.end();
      ret--;
      return ret;
   }
}

void GLContext::removeObject( list<GLObject *>::iterator pos )
{
   if( m_ParentContext != NULL )
      m_ParentContext->removeObject( pos );
   else
      m_Objects.erase( pos );
}

void GLContext::onDeleteContext()
{
   if( m_ParentContext == NULL )
   {
      for(  list<GLObject*>::iterator iter = m_Objects.begin();
            iter != m_Objects.end();   ++iter )
      {
         (*iter)->callOnDeleteContext();
      }
      m_Objects.clear();
      for( list<GLContext*>::iterator iter = m_ChildContexts.begin();
            iter != m_ChildContexts.end();   ++iter )
      {
         (*iter)->m_ParentContext = NULL;
      }

   }
   else
   {
      m_ParentContext->removeChildContext( this );
   }
}

void GLContext::removeChildContext( GLContext * context )
{
   if( m_ParentContext != NULL )
      m_ParentContext->removeChildContext( context );
   else
   {
      for( list<GLContext*>::iterator iter = m_ChildContexts.begin();
            iter != m_ChildContexts.end();   ++iter )
      {
         if( *iter == context )
         {
            (*iter)->m_ParentContext = NULL;
            m_ChildContexts.erase( iter );
            return;
         }
      }
      throw PException( "GLContext::removeChildContext - context ist not a child of this parentcontext!" );
   }
}

// (mw 13.12.04)
void GLContext::addChildContext(GLContext *context)
{
   m_ChildContexts.push_back(context);
}

void GLContext::debug( string msg )
{
   GLuint error = glGetError();
   if( error != GL_NO_ERROR )
   {
      string errorTypeMsg = "unknown gl error";
      switch( error )
      {
      case GL_INVALID_ENUM:
         errorTypeMsg = "invalid enum";
      break;
      case GL_INVALID_VALUE:
         errorTypeMsg = "invalid value";
      break;
      case GL_INVALID_OPERATION:
         errorTypeMsg = "invalid operation";
      break;
      case GL_STACK_OVERFLOW:
         errorTypeMsg = "stack overflow";
      break;
      case GL_STACK_UNDERFLOW:
         errorTypeMsg = "stack underflow";
      break;
      case GL_OUT_OF_MEMORY:
         errorTypeMsg = "out of memory";
      break;
      }

      debugOutput << msg << " errortype = " << errorTypeMsg << "\n";
   }
}





