//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "GLGeometryList.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
#include "GLBuffer.h"
#include "GLContext.h"
#include "pAssert.h"
#include "DebugOutput.h"
#include "LinearAlgebra_templCode.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

GLGeometryList::GLGeometryList( GLContext * context )
{
   m_Buffer = NULL;
   m_Context = context;
   m_ActiveBlock = false;
   m_NumPositionAttributes = 0;
   m_NumNormalAttributes = 0;
   m_NumColorAttributes = 0;
   m_NumTexcoordAttributes = 0;
}

GLGeometryList::~GLGeometryList()
{
   delete m_Buffer;
}

void GLGeometryList::compile(GLContext * context,bool useDefaultBindings)
{
   pAssert( context == m_Context );
   pAssert( m_Buffer == NULL );
   pAssert( !m_ActiveBlock );

   if( m_NumPositionAttributes == 0 )
      return;  // no vertices recorded

   VertexDescriptor vd;
   vd.pushAttrib( VAD( "position", (card8)m_NumPositionAttributes, VAD::DATA_FORMAT_FLOAT32 ) );
   if( m_NumNormalAttributes > 0 )
      vd.pushAttrib( VAD( "normal", (card8)m_NumNormalAttributes, VAD::DATA_FORMAT_FLOAT32 ) );
   if( m_NumColorAttributes > 0 )
      vd.pushAttrib( VAD( "color", (card8)m_NumColorAttributes, VAD::DATA_FORMAT_FLOAT32 ) );
   if( m_NumTexcoordAttributes > 0 )
      vd.pushAttrib( VAD( "texcoord", (card8)m_NumTexcoordAttributes, VAD::DATA_FORMAT_FLOAT32 ) );


   m_Buffer = new GLVertexBuffer( context, vd );
   m_Buffer->init( (GLuint)m_RenderData.size() );
   m_Buffer->setMappingFlag( useDefaultBindings );    // <- bind as GL_VERTEX_ARRAY, GL_NORMAL_ARRAY, ...

   VertexDescriptor vdSource;
   vdSource.pushAttrib( VAD("position", 4, VAD::DATA_FORMAT_FLOAT32 ) );
   vdSource.pushAttrib( VAD("normal", 4, VAD::DATA_FORMAT_FLOAT32 ) );
   vdSource.pushAttrib( VAD("color", 4, VAD::DATA_FORMAT_FLOAT32 ) );
   vdSource.pushAttrib( VAD("texcoord", 4, VAD::DATA_FORMAT_FLOAT32 ) );

   vector<card8> tmpVert;
   vd.convertVertices( (const char*)&m_RenderData[0], m_RenderData.size(), vdSource,
      tmpVert );
   m_Buffer->writeToBuffer( &tmpVert[0], 0, (GLuint)m_RenderData.size() );

}

void GLGeometryList::execute(GLContext * context)
{
   pAssert( context == m_Context );
   pAssert( !m_ActiveBlock );
   pAssert( m_Buffer != NULL );

   m_Buffer->bind();

   for( card32 i=0;i<m_RenderBlocks.size();i++ )
   {
      m_Buffer->drawArrays( m_RenderBlocks[i].mode, m_RenderBlocks[i].startVertex,
         m_RenderBlocks[i].numVertices );
   }

   m_Buffer->unbind();
}

void GLGeometryList::begin( GLenum mode )
{
   pAssert( !m_ActiveBlock );
   m_ActiveBlock = true;
   m_ActiveBlockMode = mode;
}

void GLGeometryList::end()
{
   pAssert( m_ActiveBlock );
   m_ActiveBlock = false;
   TRenderBlock block;
   block.mode = m_ActiveBlockMode;
   if( m_RenderBlocks.size() == 0 )
   {
      block.startVertex = 0;
      block.numVertices = m_RenderData.size();
   }
   else
   {
      block.startVertex = m_RenderBlocks.back().startVertex + m_RenderBlocks.back().numVertices;
      block.numVertices = m_RenderData.size() - block.startVertex;
   }
   m_RenderBlocks.push_back( block );
}

void GLGeometryList::vertex3f( float x, float y, float z )
{
   pAssert( m_ActiveBlock );
   m_NumPositionAttributes = max(m_NumPositionAttributes,3);
   m_ActiveVertex.position = makeVector4f( x,y,z,1.0f );
   m_RenderData.push_back( m_ActiveVertex );
}

void GLGeometryList::vertex3fv( float *x )
{
   pAssert( m_ActiveBlock );
   m_NumPositionAttributes = max(m_NumPositionAttributes,3);
   m_ActiveVertex.position = makeVector4f( x[0],x[1],x[2],1 );
   m_RenderData.push_back( m_ActiveVertex );
}

void GLGeometryList::texCoord2f( float u, float v )
{
   pAssert( m_ActiveBlock );
   m_NumTexcoordAttributes = max(m_NumTexcoordAttributes,2);
   m_ActiveVertex.texcoord = makeVector4f( u,v,0,0 );
}

void GLGeometryList::texCoord2fv( float *u )
{
   pAssert( m_ActiveBlock );
   m_NumTexcoordAttributes = max(m_NumTexcoordAttributes,2);
   m_ActiveVertex.texcoord = makeVector4f( u[0],u[1],0,0 );
}

void GLGeometryList::texCoord4f( float u, float v, float w, float s )
{
   pAssert( m_ActiveBlock );
   m_NumTexcoordAttributes = max(m_NumTexcoordAttributes,4);
   m_ActiveVertex.texcoord = makeVector4f( u,v,w,s );
}

void GLGeometryList::texCoord4fv( float *u )
{
   pAssert( m_ActiveBlock );
   m_NumTexcoordAttributes = max(m_NumTexcoordAttributes,4);
   m_ActiveVertex.texcoord = makeVector4f( u[0],u[1],u[2],u[3] );
}


void GLGeometryList::normal3f( float x, float y, float z )
{
   pAssert( m_ActiveBlock );
   m_NumNormalAttributes = max(m_NumNormalAttributes,3);
   m_ActiveVertex.normal = makeVector4f( x,y,z,0 );
}

void GLGeometryList::normal3fv( float *x )
{
   pAssert( m_ActiveBlock );
   m_NumNormalAttributes = max(m_NumNormalAttributes,3);
   m_ActiveVertex.normal = makeVector4f( x[0],x[1],x[2],0 );
}

void GLGeometryList::color3f( float r, float g, float b )
{
   pAssert( m_ActiveBlock );
   m_NumColorAttributes = max(m_NumColorAttributes,3);
   m_ActiveVertex.color = makeVector4f( r,g,b,1 );
}

void GLGeometryList::color3fv( float * rgb )
{
   pAssert( m_ActiveBlock );
   m_NumColorAttributes = max(m_NumColorAttributes,3);
   m_ActiveVertex.color = makeVector4f( rgb[0],rgb[1],rgb[2],1 );
}

void GLGeometryList::color4f( float r, float g, float b, float a )
{
   pAssert( m_ActiveBlock );
   m_NumColorAttributes = max(m_NumColorAttributes,4);
   m_ActiveVertex.color = makeVector4f( r,g,b,a );
}

void GLGeometryList::color4fv( float * rgba )
{
   pAssert( m_ActiveBlock );
   m_NumColorAttributes = max(m_NumColorAttributes,4);
   m_ActiveVertex.color = makeVector4f( rgba[0],rgba[1],rgba[2],rgba[3] );
}

