//---------------------------------------------------------------------------
#ifndef GLGeometryListH
#define GLGeometryListH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <vector>
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl
{
   using namespace std;


class GLVertexBuffer;
class GLContext;

/**
   This Class is slightly similar to GL's DisplayLists.
   It provide a mechanism to record geometry blocks(once) and to draw those.
   It doesn't record GL-States like GL DisplayLists.
   USAGE:

      - create a new GLGeometryList on startup
      - start a block with begin
      - send texcoords, normals, colors, positions with memberfunctions
      - end that block
      - record other blocks ...
      - call compile()
      - then u can use execute() to draw all blocks

   DO NOT:
    - call vertex, normal, ... methods while u aren't between a begin()/end() block
    - call end() without calling a begin() before
    - call begin() within a block
    - call execute() without compile()
    - call compile() within a begin()/end() block
    - call compile() twice


*/

class GLTOOLS_API GLGeometryList
{
private:

   struct TRenderBlock
   {
      GLenum   mode;
      card32   startVertex;
      mpcard   numVertices;
   };

   struct TVertex
   {
      Vector4f    position;
      Vector4f    color;
      Vector4f    normal;
      Vector4f    texcoord;
   };

   GLVertexBuffer          *  m_Buffer;
   GLContext         *  m_Context;
   vector<TRenderBlock> m_RenderBlocks;
   vector<TVertex>      m_RenderData;


   bool                 m_ActiveBlock;
   TVertex              m_ActiveVertex;
   GLenum               m_ActiveBlockMode;

   card32               m_NumPositionAttributes;
   card32               m_NumNormalAttributes;
   card32               m_NumColorAttributes;
   card32               m_NumTexcoordAttributes;

public:


   GLGeometryList(GLContext * context);
   ~GLGeometryList();

   void compile(GLContext * context,bool useDefaultBindings = true);
   void execute(GLContext * context);

   void begin( GLenum mode );
   void end();

   void vertex3f( float x, float y, float z );
   void vertex3fv( float *x );

   void texCoord2f( float u, float v );
   void texCoord2fv( float *u );
   void texCoord4f( float u, float v, float w, float s );
   void texCoord4fv( float *u );

   void normal3f( float x, float y, float z );
   void normal3fv( float *x );

   void color3f( float r, float g, float b );
   void color3fv( float * rgb );
   void color4f( float r, float g, float b, float a );
   void color4fv( float * rgba );



};

}

#endif
 