//---------------------------------------------------------------------------
#ifndef GLTextureH
#define GLTextureH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include <GL/gl.h>
#include "GLObject.h"
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl {

class RGBAImage;
class RGBACubeMap;
class GLObject;
class GLContext;
class GLRenderTarget;


class GLTOOLS_API GLTextureSettings : public Persistent
{
private:
   static MetaClass *GLTextureSettingsMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

public:
   GLTextureSettings();

   static const card32 FILTER_LINEAR = 1;
   static const card32 FILTER_NEAREST = 2;

   static const card32 WRAP_REPEAT = 1;
   static const card32 WRAP_CLAMP = 2;
   static const card32 WRAP_CLAMP_TO_EDGE = 3;
   static const card32 WRAP_CLAMP_TO_BORDER = 4;

   /// initial = Linear
   card32      m_MinFilter,
               m_MagFilter,
               m_MipmapFilter;

   /// inital = GL_REPEAT
   card32       m_WrapS,
                m_WrapT,
                m_WrapR;

   void set(GLenum target,bool mipmap) const;
   static void setDefaults(GLenum target, bool mipmap);
   static void setCubeDefaults();

   /// filter: wrap s/t = NEAREST
   /// wrap: CLAMP_TO_EDGE
   static void setFloatDefaults(GLenum target );

};

class GLTOOLS_API GLTexture : public GLObject
{
public:
   GLTexture( GLContext * );
   ~GLTexture();

   // virtual void create() = 0;
   void bind( GLuint stage = 0);
   void bind( const GLTextureSettings &settings, GLuint stage = 0 );
   void unbind( GLuint stage = 0 );

   virtual void onDeleteContext();

   inline GLenum getTextureTarget() const {return m_TextureTarget;}
   inline GLuint getTextureID() const {return m_TextureID;}

   /// used for MemoryManagment
   inline GLuint getSizeInBytes() const {return m_TextureSize;};
   inline GLenum getFormat() const {return m_Format;}
   inline GLuint getNumComponents() const {return m_NumComponents;}

   bool isFloatTexture() const;

   friend GLRenderTarget;

	static void setStage(int stage);

protected:


   GLenum   m_TextureTarget;
   GLuint   m_TextureID;
   bool     m_IsMipmapTexture;
   GLuint   m_TextureSize;
   GLuint   m_NumComponents; // per pixel
   GLenum   m_Format;

};

/// POT - texture ( power-of-two )
/// means: width = pow(2,i), height = pow(2,j), i,j >= 1
class GLTOOLS_API GLTexture2D : public GLTexture
{
public:
   GLTexture2D( GLContext * );

   /// creates texture from RGBAImage with defaultsettings
   void createFromImage( RGBAImage &,bool mipmap=true );

   /// creates texture from Memory with defaultsettings
   /// (dataPtr == NULL -> uninitialized texels )
   /// one component = 1 Byte
   /// numComponents element of {1,2,3,4}
   void createFromMemory( GLuint width, GLuint height, GLuint numComponents,
      char * dataPtr, bool mipmap = true );

   friend GLRenderTarget;

protected:

   void setDimensions( GLuint width, GLuint height,
      GLuint bytesPerTexel, bool mipmap = true );
   void calcSize();

   GLuint   m_Width, m_Height;
   GLuint   m_BytesPerTexel;
   bool     m_HasMipmaps;
   GLenum   m_Format;

};

/// NPOT - texture ( non-power-of-two )
/// width & height may npot
/// mipmapping is not allowed!
class GLTOOLS_API GLTextureRect : public GLTexture
{
public:
   GLTextureRect( GLContext * );

   /// creates texture from RGBAImage with defaultsettings
   void createFromImage( RGBAImage & );

   /// creates texture from memory with defaultsettings
   /// (dataPtr == NULL -> uninitialized texels )
   /// numComponents element of {1,2,3,4}
   /// one component = 1 Byte
   void createFromMemory( GLuint width, GLuint height, GLuint numComponents,
      char * dataPtr );

   /// creates float-texture from memory settings
   /// (dataPtr == NULL -> uninitialized texels )
   /// numComponents element of {1,2,3,4}
   /// useFloat32 == false : internal format is float16
   /// useFloat32 == true : internal format is float32
   /// texturesettings:
   ///  - filtering: GL_NEAREST ONLY!
   ///  - wrap: GL_CLAMP_TO_EDGE <- GL_CLAMP maybe not hw-acc
   void createFloatTextureFromMemory( GLuint width, GLuint height, GLuint numComponents,
      float * dataPtr, bool useFloat32 = true );

   friend GLRenderTarget;

protected:

   void setDimensions( GLuint width, GLuint height,
      GLuint bytesPerTexel );
   void calcSize();

   GLuint   m_Width, m_Height;
   GLuint   m_BytesPerTexel;
   GLenum   m_Format;

};

class GLTOOLS_API GLTextureCube : public GLTexture
{
public:
	GLTextureCube(GLContext *);

   void createFromImage( RGBACubeMap & );

   /// creates a cubemaptexture
   /// use setPositive/Negative/XYZ - methods to specify each side
	void create( GLuint size, GLuint numComponents );
	void setPositiveX( char * dataPtr, bool BGRA = false );
	void setPositiveY( char * dataPtr, bool BGRA = false  );
	void setPositiveZ( char * dataPtr, bool BGRA = false  );
	void setNegativeX( char * dataPtr, bool BGRA = false  );
	void setNegativeY( char * dataPtr, bool BGRA = false  );
	void setNegativeZ( char * dataPtr, bool BGRA = false  );

private:

	void calcSize();

	GLuint	m_Size;
	GLuint	m_BytesPerTexel;
	GLuint	m_NumComponents;
	GLenum	m_Format;
};


class GLTOOLS_API GLTexture3D : public GLTexture
{
public:
   GLTexture3D(GLContext *);

   void createFromMemory( GLuint width, GLuint height, GLuint depth, GLuint numComponents,
      char * dataPtr, bool mipmap = true, GLenum sourceDataFormat = GL_UNSIGNED_BYTE );

   void createFloatTextureFromMemory( GLuint width, GLuint height, GLuint depth, GLuint numComponents,	   
	   char * dataPtr, bool use16bit = true, bool mipmap = false );

   void createPerlinNoise( GLuint size, GLuint octaves, float persistance );

private:

   GLuint   m_Width, m_Height, m_Depth;
   GLuint   m_NumComponents;
   GLenum   m_Format;
	GLuint	m_BytesPerTexel;
   bool     m_HasMipmaps;
};


}; // namespace GL


#endif
