#ifndef BucketSplitPointsAddingAdaptorH
#define BucketSplitPointsAddingAdaptorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PointCloudIterators.h"
//---------------------------------------------------------------------------

namespace geoxl
{

/**
	This class can be applied to a DynamicPointCloudIterator (as an adpator) to enforce a more spatially coherent
	pattern for adding points. The points are split into spatial buckets (uniform grid), the points in the buckets are
	stored in different files and then each file's points are added to the base DynamicPointCloudIterator one after another.
	The iterator can only be applied if the points contain a position(3f) channel.
*/
class HIERARCHIES_API BucketSplitPointsAddingAdaptor : public PointCloudIterator,
																		 public DynamicPointCloudIterator
{
	GEOXL_CLASS( BucketSplitPointsAddingAdaptor )

 private:
	struct Bucket {
		string filename;
		FILE *file;
	};
	struct Key {
		Vector3i bucket;
		Key(Vector3i bucket) {this->bucket = bucket;}
		inline bool operator<(const Key &other) const {
			if (other.bucket[0] < bucket[0]) return true;
			if (other.bucket[0] > bucket[0]) return false;
			if (other.bucket[1] < bucket[1]) return true;
			if (other.bucket[1] > bucket[1]) return false;
			if (other.bucket[2] < bucket[2]) return true;
			return false;
		}
	};
	float32 cubeSideLength;
	string baseFileName;
	map<Key, Bucket> fileMapper;	
	DynamicPointCloudIterator *baseIt;
	card32 position3fOffset;
	card32 vertexSize;

	FILE *getFile(Vector3i bucket);
	
	BucketSplitPointsAddingAdaptor();
	
 public:
	/// constructor - specify base iterator, side length of the uniform grid and a base file name (and directory) for the intermediate files
	/// additionally, the AAT for the position channel (must be float[3]) and the overall size of one vertex must be provided
	BucketSplitPointsAddingAdaptor(DynamicPointCloudIterator *baseIt, float32 cubeSideLength, 
	                               string baseFileName, AAT position3fOffset, card32 vertexSize);

   virtual void add(const char *newPoint);
   virtual void add(const PointSet *newPoints);
	/// deletion is not supported - throws an exception
   virtual void deleteCurrent();

	~BucketSplitPointsAddingAdaptor(void);
};

}

#endif
