//---------------------------------------------------------------------------
#ifndef MHHierarchyH
#define MHHierarchyH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
#include "BoundingBox.h"
//---------------------------------------------------------------------------

// includes (very) expensive debug code if defined
#undef DEBUG_HIERARCHY


namespace geoxl {



// ------------------------------ ModularHierarchy ------------------------------



/** Generic spatial hierarchy. Templated implementation. */
template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
class HIERARCHIES_API ModularHierarchy_TemplBase : public HierarchyBaseClass {
 public:
   typedef typename NodeType::NodeAllocator NodeAllocator;
   typedef typename NodeType::NodePointer NodePointer;

 protected:
   SettingsType settings;
   NodePointer root;
   BoundingBox<FloatType, Dimension> rootBB;
   NodeAllocator *nAlloc;
   card64 numObjects;

   /// split node if too full, called by check update
   void checkSplit(NodePointer node);
   /// combine subtree if children are not full enough, called by check update
   void checkCombine(NodePointer node);
   /// used internally by checkCombine
   void collectRecursively(NodePointer node, NodePointer targetNode);
   /// perform lazy updates necessary before read access - recursive method that does all the work
   void checkUpdate(NodePointer node);
   /// recursively setting all node flags to modified
   void invalidateNodes(NodePointer node);

 public:
   ModularHierarchy_TemplBase();

   /// setup parameters after construction
   void clearAndSetup(NodeAllocator *nAlloc);
   /// just remove all points
   void clear();

   /// add object to hierarchy.
   void addObject(const ObjectType obj);
   /// remove object again (node and index within node must be known)
   void deleteObject(NodePointer node, max_platform_card index);
   /// find object in hierarchy. returns node == NULL if the object has not been found
   void localizeObject(ObjectType obj, NodePointer &resultNode, max_platform_card &index);
   /// change object (and update flag counters, multi-res)
   /// note: spatial properties which determine placemant in hierarchy must not be changed
   void changeObjectAttributes(ObjectType newObj, NodePointer node, max_platform_card index);

   /// build new hierarchy top-down from object list
   /// not implemented yet
   void createHierarchy(vector<ObjectType> objects);

   /// perform lazy updates necessary before read access
   void checkUpdate();
   /// set modified flag for all nodes, forces "checkUpdate" to visit all nodes
   void invalidateNodes();
   /// access root node; checkUpdate() is called implicitly before access
   void getRoot(NodePointer &root);
   /// retrieve current root bounding box
   BoundingBox<FloatType, Dimension> getRootBB();
   /// get number of objects in the hierarchy
   card64 getNumObjects();
   /// set number of objects in the hierarchy (in case of external modifications)
   void setNumObjects(card64 numObjs);
   /// get hierarchy settings
   void getSettings(SettingsType &settings);
   /// set hierarchy settings
   void setSettings(const SettingsType &settings);
   /// retrieve allocator
   void getAllocator(NodeAllocator *&alloc);
   /// alias for convinience
   inline NodeAllocator *getAllocator() {return nAlloc;}

   /// for debug purpose: check for inconsistencies (O(n(subtree)) running time)
   void checkTreeStructure(NodePointer p, int level = 0);

   void setNullAllocator() {nAlloc = NULL;}

   ~ModularHierarchy_TemplBase();
};


}

#endif
