//---------------------------------------------------------------------------
#ifndef MHHierarchy_templCodeH
#define MHHierarchy_templCodeH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MHHierarchy.h"
#include "BoundingBox_templCode.h"
#include "linearalgebra_templcode.h"
#include "AttachedData.h"
#include "PointFlags.h"
//---------------------------------------------------------------------------


namespace geoxl {


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//----------------------------
::ModularHierarchy_TemplBase()
//----------------------------
{
   root = NULL;
   nAlloc = NULL;
   numObjects = 0;
      
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//------------------------------------
::clearAndSetup(NodeAllocator *nAlloc)
//------------------------------------
{
   if (this->root != NULL && this->nAlloc != NULL) {
      this->nAlloc->deleteNode(root);
      root = NULL;
      delete nAlloc;
      nAlloc = NULL;
   }
   this->nAlloc = nAlloc;
   numObjects = 0;
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-------
::clear()
//-------
{
   if (this->root != NULL && this->nAlloc != NULL) {
      this->nAlloc->deleteNode(root);
      root = NULL;
      delete nAlloc;
      nAlloc = NULL;
   }
   numObjects = 0;
}



template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-------------------------------
::addObject(const ObjectType obj)
//-------------------------------
{
   numObjects++;
   const card32 objectFlags = obj.getFlags();
   NodePointer targetNode = root;
   // no node yet?
   if (root == NULL) {
      NodeType::createRoot(nAlloc, obj, settings, root);
      rootBB = root->getBoundingBox();
      targetNode = root;
      root->objsInSubtree = 1;
      root->nodeData.incFlagsCounters(objectFlags);
      root->setDirty(true);
      nAlloc->lruTouch(root);
   } else {
      // point outside root ?
      while (!obj.inBoundingBox(rootBB)) {
         NodePointer newRoot;
         root->enlargeRoot(nAlloc, obj, root, newRoot);
         rootBB = root->getBoundingBox();
         newRoot->objsInSubtree = root->objsInSubtree;
         newRoot->nodeData.copyFlagsCounters(root->nodeData);
         newRoot->multiRes.topDownAddMultiRes(root->multiRes, newRoot->getBoundingBox(), settings);
         const max_platform_card numNodePoints = root->nodeData.getNumObjects();
         for (card32 i=0; i<numNodePoints; i++) {
            newRoot->multiRes.topDownAddObject(root->nodeData.getObject(i), newRoot->getBoundingBox(), settings);
         }
			newRoot->setDirty(true);
	      nAlloc->lruTouch(newRoot);
         root = newRoot;
      }
      // search for matching node, create leaf if necessary
      NodePointer currentNode = root;
      do {
         currentNode->setFlags(currentNode->getFlags() | MH_NODE_FLAG_MODIFIED);
         currentNode->objsInSubtree++;
         currentNode->nodeData.incFlagsCounters(objectFlags);
         currentNode->setDirty(true);
			nAlloc->lruTouch(currentNode);


         int childIndex = NodeType::getChildIndex(obj, currentNode->getBoundingBox());
         NodePointer nextNode;
         currentNode->getChild(childIndex, nextNode);
         // search terminates...
         if (nextNode == NULL) {
            // in non-leaf node --> create new leaf node
            if (currentNode->getNumNonNullChildren() > 0) {
               currentNode->multiRes.topDownAddObject(obj, currentNode->getBoundingBox(), settings);
               currentNode->nodeData.touchData(AttachedData::CT_ADD_DELETE);
               currentNode->createChild(nAlloc, childIndex, currentNode);
               currentNode->setDirty(true);
					nAlloc->lruTouch(currentNode);
               NodePointer tempNode;
               currentNode->getChild(childIndex, tempNode);
               currentNode = tempNode;
               currentNode->objsInSubtree = 1;
               currentNode->nodeData.incFlagsCounters(objectFlags);
               currentNode->setDirty(true);
					nAlloc->lruTouch(currentNode);
               break;
            // in leaf node --> currentNode, currentBB are pointing to the desired node
            } else {
               break;
            }
         // continue search
         } else {
            currentNode->multiRes.topDownAddObject(obj, currentNode->getBoundingBox(), settings);
            currentNode->nodeData.touchData(AttachedData::CT_ADD_DELETE);
            currentNode->setDirty(true);
				nAlloc->lruTouch(currentNode);
            currentNode = nextNode;
         }
         currentNode->setDirty(true);
			nAlloc->lruTouch(currentNode);
      } while (true);
      // node found: now currentNode and currentBB are pointing to the leaf node the data has to be inserted into
      targetNode = currentNode;
   }

   targetNode->nodeData.addObject(obj, targetNode->getBoundingBox());
	targetNode->setDirty(true);
	nAlloc->lruTouch(targetNode);
   #ifdef DEBUG_HIERARCHY
    checkTreeStructure(root);
   #endif
}




template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-------------------------------------------------------
::deleteObject(NodePointer node, max_platform_card index)
//-------------------------------------------------------
{
   pAssert(node != NULL);
   pAssert(node->nodeData.getNumObjects() > index);

   pAssert(numObjects > 0);
   numObjects--;

   ObjectType obj = node->nodeData.getObject(index);
   card32 objFlags = obj.getFlags();
   node->setFlags(node->getFlags() | MH_NODE_FLAG_MODIFIED);
	node->setDirty(true);
	nAlloc->lruTouch(node);

   NodePointer delNode = node;
   NodePointer t;

   while (node != NULL) {
      pAssert(node->objsInSubtree > 0);
      node->objsInSubtree--;
      node->nodeData.decFlagsCounters(objFlags);
      node->setFlags(node->getFlags() | MH_NODE_FLAG_MODIFIED);
      if (node != delNode) {
         node->multiRes.topDownRemoveObject(obj, node->getBoundingBox(), settings);
      }
      /// call for each node (rebuild vertex buffers for multi-res, too; attachments are stored in node data)
      node->nodeData.touchData(AttachedData::CT_ADD_DELETE);
		node->setDirty(true);
		nAlloc->lruTouch(node);
      node->getParent(t);
      node = t;
   }

   delNode->nodeData.deleteObject(index);
   delNode->setDirty(true);
	nAlloc->lruTouch(delNode);
   #ifdef DEBUG_HIERARCHY
    checkTreeStructure(root);
   #endif
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//---------------------------------------------------------------------------------
::localizeObject(ObjectType obj, NodePointer &resultNode, max_platform_card &index)
//---------------------------------------------------------------------------------
{
   NodePointer currentNode = root;
   do {
		nAlloc->lruTouch(currentNode);
      if (!obj->inBoundingBox(root->getBoundingBox())) {
         resultNode = NULL;
         return;
      } else {

         if (!currentNode->isLeaf()) {
            card32 numChildren = currentNode->getMaxNumChildren();

            bool childFound = false;
            for (card32 i=0; i<numChildren; i++) {
               NodePointer child;
               currentNode->getChild(i, child);
               if (child != NULL) {
                  if (obj->inBoundingBox(child->getBoundingBox())) {
                     currentNode = child;
                     childFound = true;
                     break;
                  }
               }
            }

            if (!childFound) {
               resultNode = NULL;
               return;
            }

         } else {
            max_platform_card numObjects = currentNode->nodeData.getNumObjects();
            for (max_platform_card i=0; i<numObjects; i++) {
               if (obj == currentNode->nodeData.getObject(i)) {
                  resultNode = currentNode;
                  index = i;
                  return;
               }
            }
            resultNode = NULL;
            return;
         }
		}

   } while (true);
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//------------------------------------------------------------------------------------
::changeObjectAttributes(ObjectType newObj, NodePointer node, max_platform_card index)
//------------------------------------------------------------------------------------
{
   pAssert(node != NULL);
   pAssert(index < node->nodeData.getNumObjects());
   pAssert(node->isLeaf());

   ObjectType oldObj = node->nodeData.getObject(index);

   pAssert(newObj.getCenter() == oldObj.getCenter());

   int32 newFlags = newObj.getFlags();
   int32 oldFlags = oldObj.getFlags();

   NodePointer originalNode = node;

   while (node != NULL) {
      node->nodeData.decFlagsCounters(oldFlags);
      node->nodeData.incFlagsCounters(newFlags);
      node->setDirty(true);
		nAlloc->lruTouch(node);

      if (node != originalNode) {
         node->multiRes.topDownChangeObject(newObj, node->getBoundingBox(), settings);
      }

      node->nodeData.touchData(AttachedData::CT_ATTRIBUTE_CHANGED);

      NodePointer t;
      node->getParent(t);
      node = t;
   }

   originalNode->nodeData.setObject(index, newObj);
	originalNode->setDirty(true);
	nAlloc->lruTouch(originalNode);
   #ifdef DEBUG_HIERARCHY
    checkTreeStructure(root);
   #endif
}





template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-------------
::checkUpdate()
//-------------
{
   if (root == NULL) return;

   /// check modified nodes for split / recombination
   checkUpdate(root);

   /// trim root node (remove linear chain at root node)
   while (root->getNumNonNullChildren() == 1) {
		nAlloc->lruTouch(root);		
      NodePointer child = NULL;
      card32 index = 1234567;
      const card32 n = root->getMaxNumChildren();
      for (card32 i=0; i<n; i++) {
         root->getChild(i, child);
         if (child != NULL) {
            index = i;
            break;
         }
      }
      pAssert(child != NULL && index != 1234567);
      root->removeChild(index);
      pAssert(root->getNumNonNullChildren() == 0);
      nAlloc->deleteNode(root);
      root = child;
      root->setParent(NULL);
      root->setDirty(true);
      rootBB = root->getBoundingBox();
   }
	nAlloc->lruTouch(root);
   #ifdef DEBUG_HIERARCHY
    checkTreeStructure(root);
   #endif
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//---------------------------------
::invalidateNodes(NodePointer node)
//---------------------------------
{
   node->setFlags(node->getFlags() | MH_NODE_FLAG_MODIFIED);
	node->setDirty(true);
	nAlloc->lruTouch(node);
   const card32 n = node->getMaxNumChildren();
   for (card32 i=0; i<n; i++) {
      NodePointer child;
      node->getChild(i, child);
      if (child != NULL) {
         invalidateNodes(child);
      }
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-----------------
::invalidateNodes()
//-----------------
{
   if (root != NULL) {
      invalidateNodes(root);
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-----------------------------
::checkUpdate(NodePointer node)
//-----------------------------
{
	nAlloc->lruTouch(node);
   if ((node->getFlags() & MH_NODE_FLAG_MODIFIED) != 0) {
      if (node->isLeaf()) {
         checkSplit(node);
      } else {
         checkCombine(node);
      }
      const int n = node->getMaxNumChildren();
      for (int i=0; i<n; i++) {
         NodePointer childNode;
         node->getChild(i, childNode);
         if (childNode != NULL) {
            if (childNode->objsInSubtree == 0) {
               node->deleteChild(nAlloc, i);
               node->setDirty(true);
            } else {
               checkUpdate(childNode);
            }
         }
      }
      node->setFlags(node->getFlags() & (0xFFFFFFFF ^ MH_NODE_FLAG_MODIFIED));
      node->setDirty(true);
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//----------------------------
::checkSplit(NodePointer node)
//----------------------------
{
	nAlloc->lruTouch(node);
	if (node->nodeShouldBeSplit(settings)) {
		const max_platform_card n = node->nodeData.getNumObjects();
      for (max_platform_card i=0; i<n; i++) {
         ObjectType obj = node->nodeData.getObject(i);

         node->multiRes.topDownAddObject(obj, node->getBoundingBox(), settings);
         node->nodeData.touchData(AttachedData::CT_EVERYTHING_CHANGED);
			node->setDirty(true);

         int childIndex = node->getChildIndex(obj, node->getBoundingBox());
         NodePointer childNode;
         node->getChild(childIndex, childNode);
         if (childNode == NULL) {
            node->createChild(nAlloc, childIndex, node);
            node->setDirty(true);
            NodePointer tempNode;
            node->getChild(childIndex, tempNode);
            childNode = tempNode;
         }
         childNode->objsInSubtree++;
         childNode->nodeData.incFlagsCounters(obj.getFlags());
         childNode->nodeData.addObject(obj, childNode->getBoundingBox());
         childNode->setDirty(true);
      }
      node->nodeData.clear(false);
      node->setDirty(true);
      const int numChildren = node->getMaxNumChildren();
      for (int i=0; i<numChildren; i++) {
         NodePointer child;
         node->getChild(i, child);
         if (child != NULL) {
            checkSplit(child);
         }
      }
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//------------------------------
::checkCombine(NodePointer node)
//------------------------------
{
	nAlloc->lruTouch(node);
   if (node->subtreeShouldBeMerged(settings)) {
      pAssert(node->getNumNonNullChildren() > 0);
      const int n = node->getMaxNumChildren();
      for (int i=0; i<n; i++) {
         NodePointer childNode;
         node->getChild(i, childNode);
         if (childNode != NULL) {
            collectRecursively(childNode, node);
            node->deleteChild(nAlloc, i);
		      node->setDirty(true);
         }
      }
      node->multiRes.clear();
      node->setDirty(true);
      pAssert(node->isLeaf());
      pAssert(node->nodeData.getNumObjects() == node->objsInSubtree);
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//------------------------------------------------------------
::collectRecursively(NodePointer node, NodePointer targetNode)
//------------------------------------------------------------
{
   pAssert(node != NULL && targetNode != NULL);
	nAlloc->lruTouch(node);
   if (node->isLeaf()) {
      const max_platform_card n = node->nodeData.getNumObjects();
      for (max_platform_card i=0; i<n; i++) {
         targetNode->nodeData.addObject(node->nodeData.getObject(i), targetNode->getBoundingBox());
      }
		targetNode->setDirty(true);
   } else {
      const int n = node->getMaxNumChildren();
      for (int i=0; i<n; i++) {
         NodePointer childNode;
         node->getChild(i, childNode);
         if (childNode != NULL) {
            collectRecursively(childNode, targetNode);
            node->deleteChild(nAlloc, i);
            node->setDirty(true);
         }
      }
   }
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//--------------------------
::getRoot(NodePointer &root)
//--------------------------
{
   root = this->root;
	nAlloc->lruTouch(root);
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
BoundingBox<FloatType, Dimension>  ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-----------
::getRootBB()
//-----------
{
   return rootBB;
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
card64  ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//---------------
::getNumObjects()
//---------------
{
   return numObjects;
}


template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void  ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//--------------------------------
::setNumObjects(card64 numObjects)
//--------------------------------
{
   this->numObjects = numObjects;
}



template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//--------------------------------
::getSettings(SettingsType &settings)
//--------------------------------
{
   settings = this->settings;
}



template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//--------------------------------
::setSettings(const SettingsType &settings)
//--------------------------------
{
   this->settings = settings;
}



template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-----------------------------------
::getAllocator(NodeAllocator *&alloc)
//-----------------------------------
{
   alloc = nAlloc;
}





template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
void ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//--------------------------------------------
::checkTreeStructure(NodePointer p, int level)
//--------------------------------------------
{
   if (p == NULL) return;
   NodePointer parent;
   p->getParent(parent);
   if (parent != NULL) {
      card32 index = NodeType::getChildIndex(p->getBoundingBox().getCenter(),
                                             parent->getBoundingBox());
      NodePointer child;
      parent->getChild(index, child);
      pAssert(child == p);
      pAssert(p->getBoundingBox() == NodeType::getChildBoundingBox(parent->getBoundingBox(), index));
   }
   card64 numChildObjs = 0;
   bool hasChildren = false;
   card64 flagsCounters[MAX_POINT_FLAGS_TRACKED_IN_HIERARCHY];
   for (card32 i=0; i<MAX_POINT_FLAGS_TRACKED_IN_HIERARCHY; i++) {
      flagsCounters[i] = 0;
   }
   for (int i=0; i<p->getMaxNumChildren(); i++) {
      NodePointer child;
      p->getChild(i, child);
      if (child != NULL) {
         numChildObjs += child->objsInSubtree;
         hasChildren = true;
         int32 j = 1;
         for (card32 i=0; i<MAX_POINT_FLAGS_TRACKED_IN_HIERARCHY; i++) {
            card64 nFlags = child->nodeData.getFlagsCounter(j);
            flagsCounters[i] += nFlags;
            j <<= 1;
         }
      }
   }
   pAssert(hasChildren == !p->isLeaf());
   if (hasChildren) {
      pAssert(p->objsInSubtree == numChildObjs);
      int32 j = 1;
      for (card32 i=0; i<MAX_POINT_FLAGS_TRACKED_IN_HIERARCHY; i++) {
         card64 nFlags = p->nodeData.getFlagsCounter(j);
         pAssert(flagsCounters[i] == nFlags);
         j <<= 1;
      }
   }
   for (int i=0; i<p->getMaxNumChildren(); i++) {
      NodePointer child;
      p->getChild(i, child);
      if (child != NULL) {
         checkTreeStructure(child, level+1);
      }
   }
}



template <typename FloatType, int Dimension, typename ObjectType,
          typename NodeType, typename SettingsType, class HierarchyBaseClass>
ModularHierarchy_TemplBase<FloatType, Dimension, ObjectType, NodeType, SettingsType, HierarchyBaseClass>
//-----------------------------
::~ModularHierarchy_TemplBase()
//-----------------------------
{
   if (nAlloc != NULL) {
      nAlloc->deleteNode(root);
   }
}																				



}

#endif
