//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "MHMultiRes.h"
//---------------------------------------------------------------------------
#include "PointSet.h"
#include "MHObjects.h"
#include "MHObjects_templCode.h"
#include "MHSettings.h"
#include "ObjectStreams.h"
#include "SparseGrid_templCode.h"
#ifndef _DEBUG
#include "PointSet_templCode.h"
#endif
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



MHSparseGridMultiRes::MHSparseGridMultiRes() {
   gridPoints = NULL;
}

void MHSparseGridMultiRes::write(OutputObjectStream *s) const {
   s->writeObject(gridPoints);
   grid.write(s);
}

void MHSparseGridMultiRes::read(InputObjectStream *s) {
   clear();
   Persistent *obj;
   s->readObject(obj);
   gridPoints = (PointSet*)obj;
   if (gridPoints != NULL) {
      const VertexDescriptor *vd = gridPoints->getDescr();
      aatPosition = vd->getToken("position");
      if (vd->providesAttribute("flags")) {
         aatFlags = vd->getToken("flags");
      }
   }
   grid.read(s);
}

void MHSparseGridMultiRes::assign(const Object* obj, CopyContext *context) {
   const MHSparseGridMultiRes *other = dynamic_cast<const MHSparseGridMultiRes*>(obj);
   pAssert(obj != NULL);
   clear();
   this->aatFlags = other->aatFlags;
   this->aatPosition = other->aatPosition;
   if (other->gridPoints != NULL) {
      gridPoints = (PointSet*)(other->gridPoints->copy(context));
   }
   this->grid = other->grid;
}


void MHSparseGridMultiRes::clear() {
   grid.clear();
   delete gridPoints;
   gridPoints = NULL;
}

void MHSparseGridMultiRes::topDownAddObject(
   const MHPointSetPointObject &obj,
   const BoundingBox3f &bb,
   MHStandardSettings &settings)
{
   aatPosition = obj.getAATPosition();
   aatFlags = obj.getAATFlags();
   if (gridPoints == NULL) {
      gridPoints = new PointSet();
      gridPoints->clearAndSetup(1,0, obj.getVertexDescriptor());
   }
   Vector3i gridCell = SparseGrid<3, GridEntryType, Persistent>::calcIndex<float32>(
      obj.getCenter(), bb.lowerCorner, (bb.upperCorner[0] - bb.lowerCorner[0]) / settings.multiResGridSize
   );
   GridEntryType e;
   if (grid.getObject(gridCell, e)) {
      e.weight++;
   } else {
      gridPoints->addPoint(obj.getData());
      e.index = gridPoints->getNumEntries() - 1;
      e.weight = 1;
   }
   grid.setObject(gridCell, e);
}

void MHSparseGridMultiRes::topDownAddMultiRes(
   const MHSparseGridMultiRes &other,
   const BoundingBox3f &bb,
   MHStandardSettings &settings)
{
   const max_platform_card numOtherEntries = other.grid.getNumElements();
   if (numOtherEntries > 0) {
      aatPosition = other.aatPosition;
      aatFlags = other.aatFlags;
      if (gridPoints == NULL) {
         gridPoints = new PointSet();
         gridPoints->clearAndSetup(1,0, other.gridPoints->getDescr());
      }

      for (card32 i=0; i<numOtherEntries; i++) {
         GridEntryType otherE = other.grid.getEntry(i);
         Vector3f pos = other.gridPoints->get<float32, 3>(otherE.index, aatPosition);
         Vector3i gridCell = SparseGrid<3, GridEntryType, Persistent>::calcIndex<float32>(
            pos, bb.lowerCorner, (bb.upperCorner[0] - bb.lowerCorner[0]) / settings.multiResGridSize
         );

         GridEntryType myE;
         if (grid.getObject(gridCell, myE)) {
            myE.weight += otherE.weight;
         } else {
            gridPoints->addPoint(other.gridPoints->getDataPointer(otherE.index));
            myE.index = gridPoints->getNumEntries() - 1;
            myE.weight = otherE.weight;
         }
         grid.setObject(gridCell, myE);

      }
   }
}


void MHSparseGridMultiRes::topDownRemoveObject(
   const MHPointSetPointObject &obj,
   const BoundingBox3f &bb,
   MHStandardSettings &settings)
{
   pAssert(gridPoints != NULL);
   Vector3i gridCell = SparseGrid<3, GridEntryType, Persistent>::calcIndex<float32>(
      obj.getCenter(), bb.lowerCorner, (bb.upperCorner[0] - bb.lowerCorner[0]) / settings.multiResGridSize
   );
   GridEntryType e;
   if (!grid.getObject(gridCell, e)) {
     pAssert(false); // mw - 27.4.05 - does not work properly. no idea why...
   } else {
      pAssert(e.weight >= 1);
      e.weight--;
      if (e.weight > 0) {
         grid.setObject(gridCell, e);
      } else {

         /// deletion of points in PointSet means exchange with last point in point set
         /// --> index of last point must be corrected
         pAssert(gridPoints->getNumEntries() >= 1);
         if (e.index != gridPoints->getNumEntries()-1) {

            Vector3f pos = gridPoints->get<float32, 3>(gridPoints->getNumEntries()-1, aatPosition);
            Vector3i lastPointGridCell = SparseGrid<3, GridEntryType, Persistent>::calcIndex<float32>(
               pos, bb.lowerCorner, (bb.upperCorner[0] - bb.lowerCorner[0]) / settings.multiResGridSize
            );
            GridEntryType eLast;
            if (grid.getObject(lastPointGridCell, eLast)) {
               pAssert(eLast.index == gridPoints->getNumEntries()-1);
               eLast.index = e.index;
               grid.setObject(lastPointGridCell, eLast);
            } else {
               pAssert(false);
            }
         }

         grid.deleteObject(gridCell);
         gridPoints->deletePoint(e.index);
      }
   }
}

void MHSparseGridMultiRes::topDownChangeObject(
   const MHPointSetPointObject &obj,
   const BoundingBox3f &bb,
   MHStandardSettings &settings)
{
   pAssert(gridPoints != NULL);
   Vector3i gridCell = SparseGrid<3, GridEntryType, Persistent>::calcIndex<float32>(
      obj.getCenter(), bb.lowerCorner, (bb.upperCorner[0] - bb.lowerCorner[0]) / settings.multiResGridSize
   );
   GridEntryType e;
   if (!grid.getObject(gridCell, e)) {
      pAssert(false); 
   } else {
      pAssert(gridPoints != NULL);
      gridPoints->setVertex(e.index, obj.getData());
   }
}

max_platform_card MHSparseGridMultiRes::getNumObjects() const {
   if (gridPoints == NULL) {
      return 0;
   } else {
      return gridPoints->getNumEntries();
   }
}

MHPointSetPointObject MHSparseGridMultiRes::getObject(max_platform_card index) {
   pAssert(gridPoints != 0);
   MHPointSetPointObject result(gridPoints->getDescr(), aatPosition, aatFlags, gridPoints->getDataPointer(index));
   return result;
}

void MHSparseGridMultiRes::appendFields(const VertexDescriptor *newAttribs) {
   PointSet *newps = gridPoints->appendFields(newAttribs);
   delete gridPoints;
   gridPoints = newps;
}


MHSparseGridMultiRes::~MHSparseGridMultiRes() {
   clear();
}
