//---------------------------------------------------------------------------
#ifndef MHPointSetOctreeIterator_TemplH
#define MHPointSetOctreeIterator_TemplH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "MHIterators.h"
#include "MHObjects.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
//---------------------------------------------------------------------------

namespace geoxl {


template <class OctreeType, class NodeType>
class HIERARCHIES_API MHPointSetOctreeIterator_Templ : public MHIterator,
                                               public BasicPointCloudIterator,
                                               public IndexedPointCloudIterator,
                                               public RandomAccessPointCloudIterator,
                                               public ModifyablePointCloudIterator,
                                               public DynamicPointCloudIterator,
                                               public HierarchicalIterator,
                                               public SecondaryDataStructureIterator {
 protected:
   /// this class represents a temporary set of points
   /// it is used to store recently added / modified points
   /// this functionality must be differentiated because out-of-core
   /// storage may lead to extremly large temporary buffers
   /// the class is used in "modifiedPoints", instantiated by createTempBuffer
   /// the buffer must always provide access to the last point that has been added
   /// in case modification operations (set/delete) are called after modifying points
   class TemporaryPointsBuffer {
    public:
      /// this method adds a point to the buffer
      virtual void add(const char *point) = 0; 
      /// this mothod adds several points to the buffer
      virtual void addPointSet(const PointSet *points) {
         const max_platform_card n = points->getNumEntries();
         for (card32 i=0; i<n; i++) {
            add(points->getDataPointer(i));
         }
      }
      /// this method must return a point set that contains the last point which has been added as last element (point with largest index)
      virtual void getLastModified(PointSet *&ps) = 0;
      /// this method must add the content of the buffer to the given dynamic hierarchy and clear the buffer
      virtual void addPointsToHierarchyAndClear(OctreeType *dest) = 0;
      /// this method must delete the last point that has been added (will be called only once before adding new points)
      virtual void deleteLastModified() = 0;
      virtual ~TemporaryPointsBuffer() {}

   };

 private:
   typedef typename NodeType::NodePointer NodePointer;

   // --------- member variables ---------
   OctreeType *hierarchy;

	// current node (smart pointer)
   NodePointer currentNode;
   // shortcut to speed up access in ooc instantiation
   NodeType *currentNodeDirectPointer;
   //	to avoid moving the iterator while a ps is locked (node may be swapped otherwise)
   bool pointSetLocked;

   max_platform_card currentIndex;
   
   mutable bool atModifiedPoint;
   AAT aatPosition;
   AAT aatFlags;

   /// temporary buffer for added/modified points (see embedded class TemporaryPointsBuffer")
   mutable TemporaryPointsBuffer *modifiedPoints;
   PointSet *rwPointSet;
   // -----------------------------------

   /// helper for implementing getX()
   template <class Type, int dim>
   inline StaticVector<Type, dim> genericGet(AAT token) const;
   /// helper for implementing getX()
   template <class Type>
   inline Type genericGetS(AAT token) const;

   /// helper for implementing setX()
   template <class Type, int dim>
   inline void genericSet(AAT token, StaticVector<Type, dim> v) const;
   /// helper for implementing setX()
   template <class Type>
   inline void genericSetS(AAT token, Type x) const;

   void addModifiedPoints();
   void checkCreateModifiedPtsBuffer();

 protected:
   MHPointSetOctreeIterator_Templ();
   MHPointSetOctreeIterator_Templ(OctreeType *hierarchy);

	OctreeType *getHierarchy() {return hierarchy;}

   /// override this method to control handling of temporary storage for added/position-modified points
   /// (see embedded class TemporaryPointsBuffer")
   virtual TemporaryPointsBuffer *createTempBuffer(OctreeType *hierarchy) = 0;

   // anything changed? if yes, the hierarchy is synced in the end...
	mutable bool needSync;
	
 public:
   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual bool equal(const Object* obj) const;

   virtual void reset();
   virtual void next();
   virtual bool atEnd();

   virtual card8 get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32 get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual float32 get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   virtual void getVertex(char *buffer) const;
   virtual bool nextBlock();
   virtual const PointSet * lockBlockRead();
   virtual void unlock();

   virtual card64 getNumPointsInBlock() const; 
   virtual card64 getCurrentPointIndex() const;
   virtual void goToIndex(card64 i);

   virtual void set1ub(const AAT &token, const card8 &x)     const;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const;

   virtual void set1i(const AAT &token, const int32 &x)    const;
   virtual void set2i(const AAT &token, const Vector2i &v) const;
   virtual void set3i(const AAT &token, const Vector3i &v) const;
   virtual void set4i(const AAT &token, const Vector4i &v) const;

   virtual void set1f(const AAT &token, const float32 &x)  const;
   virtual void set2f(const AAT &token, const Vector2f &v) const;
   virtual void set3f(const AAT &token, const Vector3f &v) const;
   virtual void set4f(const AAT &token, const Vector4f &v) const;

   virtual void setVertex(const char *data) const;
   virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING);

   virtual void add(const char *newPoint);
   virtual void add(const PointSet *newPoints);
   virtual void deleteCurrent();

   virtual card32 getNumChildren();
   virtual bool goToChild(card32 index);
   virtual bool goToParent();
   virtual BoundingBox3f getBoundingBox();
   virtual bool isLeaf() const;
   virtual bool getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data);

	virtual AttachedDataContainer *lockAttachments();
	virtual void unlockAttachments();

   virtual ~MHPointSetOctreeIterator_Templ();
};


}


#endif
