#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "CachedRangeQuerry.h"
#include "HierarchicalKNNIterator.h"
#include "PointCloud.h"
#include "PointSet.h"
#include "VertexDescriptor.h"
#include "AttachedIndexedOctree.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
using namespace geoxl;

static int cacheHits = 0;
static int cacheQuerries = 0;

CachedRangeQuerry::CachedRangeQuerry(PointCloud * pc)
{
	m_Cache = NULL;
	m_PointCloud = pc;
	m_PointSet = NULL;
	m_CacheType = CT_None;
	m_hIt = new HierarchicalKNNIterator(m_PointCloud, 32, NULL);
	cacheHits = 0;
	cacheQuerries = 0;
	m_SphereRadius = 0;
	m_AIOctree = NULL;
	m_PSKNNQuerry = NULL;
	m_PosAAT = pc->getAAT("position");
}


geoxl::CachedRangeQuerry::CachedRangeQuerry( PointSet * ps )
{
	m_Cache = NULL;
	m_PointCloud = NULL;
	m_PointSet = ps;
	m_CacheType = CT_None;
	m_hIt = NULL;
	cacheHits = 0;
	cacheQuerries = 0;
	m_SphereRadius = 0;
	m_PosAAT = ps->getDescr()->getToken("position");
	m_AIOctree = PointSetKNNQuery::createOctree( ps, 32 );
	m_PSKNNQuerry = new PointSetKNNQuery(ps,m_AIOctree, NULL_VECTOR3F, m_PosAAT);
}

void geoxl::CachedRangeQuerry::setSeekPoint( const Vector3f &point )
{
	if( m_hIt )
		m_hIt->setSeekPointAndReset(point);
	else
	{
		m_PSKNNQuerry->setTargetPoint(point);
		m_PSKNNQuerry->getNextNearestPointIndex(m_CurrentIndex,m_AtEnd);
		m_AtEnd = !m_AtEnd;
	}
}

unsigned geoxl::CachedRangeQuerry::getCurrentIndex()
{
	if( m_hIt )
		return m_hIt->getCurrentPointIndex();
	else
	{
		return m_CurrentIndex;
	}
}

geoxl::Vector3f geoxl::CachedRangeQuerry::getPosition()
{
	if( m_hIt )
		return m_hIt->get3f(m_PosAAT);
	else
		return m_PointSet->get3f(m_CurrentIndex,m_PosAAT);
}

void geoxl::CachedRangeQuerry::getVertex( char * buffer )
{
	if( m_hIt )
		m_hIt->getVertex(buffer);
	else
		m_PointSet->get3f(m_CurrentIndex,m_PosAAT);
}

void geoxl::CachedRangeQuerry::nextPoint()
{
	if( m_hIt )
		m_hIt->next();
	else
	{
		m_PSKNNQuerry->getNextNearestPointIndex(m_CurrentIndex,m_AtEnd);
		m_AtEnd = !m_AtEnd;
	}
}

bool geoxl::CachedRangeQuerry::atEnd()
{
	if( m_hIt )
		return m_hIt->atEnd();
	else
		return m_AtEnd;
}
CachedRangeQuerry::~CachedRangeQuerry(void)
{
	delete m_Cache;
	delete m_hIt;
	delete m_AIOctree;
	delete m_PSKNNQuerry;
//	debugOutput << "Cachehits: " << cacheHits << "\n";
//	debugOutput << "Cachequerries: " << cacheQuerries << "\n";
}

PointSet *CachedRangeQuerry::getSpherePoints( const Vector3f &p, float radius, float absolute_overlap, int minNumberOfPoints ) 
{
	if( m_CacheType == CT_Sphere )
	{
		if( norm(p-m_SpherePoint) + radius <= m_SphereRadius )
			return m_Cache;
	}
	else
		if( m_CacheType != CT_None )
			throw "Don't mix cache querries!";


	m_CacheType = CT_Sphere;
	m_SpherePoint = p;
	m_SphereRadius = radius + absolute_overlap;

	if( m_Cache == NULL )
	{
		m_Cache = new PointSet;
		m_Cache->clearAndSetup(1,0,(VertexDescriptor*)m_PointCloud->getDescr()->copy());
	}
	m_Cache->clear();

	setSeekPoint(p);
	vector<char> buffer;
	vector<char> vertex(m_PointCloud->getDescr()->getSize());
	card32 numVertices = 0;
	while( !atEnd() )
	{
		Vector3f pos = getPosition();
		if( norm(p-pos) > m_SphereRadius )
			break;
		getVertex( &vertex[0] );
		m_Cache->addPoint( &vertex[0] );
		nextPoint();
	}

	return m_Cache;
}

vector<Vector3f> * CachedRangeQuerry::getSpherePositions(const Vector3f &p, float radius, 
																			float absolute_overlap, int minNumberOfPoints,
																			unsigned ** indices)
{
	if( m_CacheType == CT_SpherePosOnly )
	{
		if( norm(p-m_SpherePoint) + radius <= m_SphereRadius )
		{
			cacheHits++;
			if( indices )
				*indices = &m_IndexCache[0];
			return &m_PosCache;
		}
	}
	else
	{
		if( m_CacheType != CT_None )
			throw "Don't mix cache querries!";
	}

	cacheQuerries++;

	m_CacheType = CT_SpherePosOnly;
	m_SphereRadius = radius + absolute_overlap;
	m_SpherePoint = p;

	m_PosCache.clear();
	m_IndexCache.clear();

	setSeekPoint(p);
	int numVertices = 0;
	while( !atEnd() )
	{
		Vector3f pos = getPosition();
		if( norm(p-pos) > m_SphereRadius && numVertices >= minNumberOfPoints)
			break;
		m_PosCache.push_back(pos);
		m_IndexCache.push_back(getCurrentIndex());
		numVertices++;
		nextPoint();
	}

	if( indices )
		*indices = &m_IndexCache[0];
	return &m_PosCache;
}
