//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "GeometricRange.h"
#include "DebugOutput.h"
#include "LinearAlgebra_templCode.h"
#include "BoundingBox_templCode.h"

#include "math.h"

//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

AABBRange::AABBRange(Vector3f lowerCor, Vector3f upperCor)
{
	lowerCorner = lowerCor;
	upperCorner = upperCor;
}

void AABBRange::setTransformation(const geoxl::Matrix4f& m)
{
	lowerCorner = transformVector3f(m, lowerCorner);
	upperCorner = transformVector3f(m, upperCorner);
}

bool AABBRange::intersects(const geoxl::BoundingBox3f& bb) const
{	
	const Vector3f u = bb.upperCorner;
	const Vector3f l = bb.lowerCorner;
	for (int d=0; d < 3; d++) {
		if (u[d] < lowerCorner[d] || l[d] > upperCorner[d]) {
			return false;
		}
	}
	return true;
}

bool AABBRange::intersects(const Vector3f& p) const
{
	for (int d=0; d < 3; d++) {
		if (p[d] < lowerCorner[d] || p[d] > upperCorner[d]) {
			return false;
		}
	}
	return true;
}

SphereRange::SphereRange(Vector3f cent, float32 sqrrad){
   Center = cent;
   TransCenter = cent; //fixed - ab 24.01.
   SqrRadius = sqrrad;
   TransSqrRadius = sqrrad;
   Radius = sqrt(sqrrad);
   TransRadius = Radius;
   scaleFactor = 1.0f;
}


void SphereRange::setTransformation(const Matrix4f& ctm) {
   Matrix4f inv = invertMatrix(ctm);
   TransCenter = transformVector3f(inv,Center);

   Matrix4f tmpM = inv;
   tmpM[3][0] = 0;
   tmpM[3][1] = 0;
   tmpM[3][2] = 0;
   Vector3f tmpV = makeVector3f(1,0,0);
   tmpV = transformVector3f(tmpM,tmpV);

   TransSqrRadius = SqrRadius * normQuad(tmpV);
   TransRadius = sqrt(SqrRadius);

   scaleFactor = fabs(ctm[0][0]);
}


bool SphereRange::intersects(const BoundingBox3f& bb) const {

// to-do: box, not sphere

   Sphere<float32,3> bs = bb.getBoundingSphere();
   Vector3f bsCenter = bs.center;
   float32 bsRad = bs.radius;
   bsRad = bsRad * scaleFactor;

   return ((normQuad(TransCenter - bsCenter)) < (bsRad*bsRad+2*bsRad*TransRadius+TransSqrRadius));
}


bool SphereRange::intersects(const Vector3f& Point) const {
   return ((normQuad(TransCenter - Point)) < TransSqrRadius);
}


// ----------


RayConeRange::RayConeRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sra1){
   RayRight = right;
   RayUp = up;
   RayDir = dir;
   RayOrigin = origin;
   SqrRadiusAt1 = sra1;
   RadiusAt1 = sqrt(sra1);

   Matrix3f RaySystem2;                                                                         
   RaySystem2[0] = RayRight;
   RaySystem2[1] = RayUp;
   RaySystem2[2] = RayDir;
   RaySystem2 = RaySystem2.transpose();
   StaticMatrix<float,3,3> m1, m2, m3;
   m1 = m2 + m3;
   RaySystem = expand3To4(RaySystem2) * makeTranslation4f(-RayOrigin);

   TransRaySystem = RaySystem;

   scaleFactor = 1.0f;
}


void RayConeRange::setTransformation(const Matrix4f& ctm) {
   TransRaySystem = RaySystem * ctm;
   scaleFactor = fabs(ctm[0][0]);
}


bool RayConeRange::intersects(const BoundingBox3f& bb) const {

// to-do: box, not sphere

   Sphere<float32,3> bs = bb.getBoundingSphere();
   Vector3f bsCenter = bs.center;
   float32 bsRad = bs.radius;
   Vector3f bsCenterPosRay = transformVector3f(TransRaySystem, bsCenter);
   bsRad = bsRad * scaleFactor;

   float32 c = sqrt(SqrRadiusAt1 + 1.0f);
   if (RadiusAt1==0.0f) return false;
   float32 coneSinRecip = c/RadiusAt1;

   float32 coneSinSqr = SqrRadiusAt1/(c*c);
   float32 coneCosSqr = 1.0f/(c*c);

   Vector3f rayDir2 = makeVector3f(0.0f, 0.0f, -1.0f);
   Vector3f rayOrigin2 = makeVector3f(0.0f, 0.0f, 0.0f);

   Vector3f U = rayOrigin2 - rayDir2*(bsRad*coneSinRecip);
   Vector3f D = bsCenterPosRay - U;
   float32 DSqr = D*D;
   float32 e = rayDir2*D;

   if ((e>0) && (e*e >= DSqr*coneCosSqr))
   {
      D = bsCenterPosRay - rayOrigin2;
      DSqr = D*D;
      e = -(rayDir2*D);
      if ((e>0) && (e*e >= DSqr*coneSinSqr))
         return (DSqr<=(bsRad*bsRad));
      else
         return true;
   };

   return false;
}


bool RayConeRange::intersects(const Vector3f& Point) const {
   // transform point into ray coordinate system
   Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
   // only points in front of the 'camera' (which have negative z)
   if (PointRayPos[2]<0.0f) {
      if ((PointRayPos[1]*PointRayPos[1]+PointRayPos[0]*PointRayPos[0]) < (PointRayPos[2]*PointRayPos[2]*SqrRadiusAt1)) return true;
   }
   return false;
}


/**
 * returns the "normalized" squared distance from footpoint on the ray axis to input point (SqrDist to the point divided by ConeRadius at this z-value results in a float between 0.0 and 1.0).
 */
float32 RayConeRange::getDistNorm(const Vector3f& Point) const {
   // transform point into ray coordinate system
   Vector4f PointRayPos = TransRaySystem * (expand3To4(Point));
   // only points in front of the 'camera' (which have negative z)
   if (PointRayPos[2]<0.0f) {
      float32 SqrDist = (PointRayPos[1]*PointRayPos[1]+PointRayPos[0]*PointRayPos[0]);
      float32 sqrRad  = (PointRayPos[2]*PointRayPos[2]*SqrRadiusAt1);
      if ( SqrDist < sqrRad)
      {
         return SqrDist/sqrRad;
      };
   };
   return -1.0f;
}


// ----------


RayPyramidalRange::RayPyramidalRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sqrhw, float32 sqrhh){
   RayRight = right;
   RayUp = up;
   RayDir = dir;
   RayOrigin = origin;
   SqrHalfWidthAt1 = sqrhw;
   SqrHalfHeightAt1 = sqrhh;
   HalfWidthAt1 = sqrt(sqrhw);
   HalfHeightAt1 = sqrt(sqrhh);

   Matrix3f RaySystem2;
   RaySystem2[0] = RayRight;
   RaySystem2[1] = RayUp;
   RaySystem2[2] = RayDir;
   RaySystem2 = RaySystem2.transpose();
   RaySystem = expand3To4(RaySystem2) * makeTranslation4f(-RayOrigin);

   TransRaySystem = RaySystem;
   scaleFactor = 1.0f;

}


void RayPyramidalRange::setTransformation(const Matrix4f& ctm) {
   TransRaySystem = RaySystem * ctm;
   scaleFactor = fabs(ctm[0][0]);
}


bool RayPyramidalRange::intersects(const BoundingBox3f& bb) const {

   // to-do: do correct test, not cone test

   float32 SqrRadiusAt1 = SqrHalfWidthAt1 + SqrHalfHeightAt1;
   float32 RadiusAt1 = sqrt(SqrRadiusAt1);


   Sphere<float32,3> bs = bb.getBoundingSphere();
   Vector3f bsCenter = bs.center;
   float32 bsRad = bs.radius;
   bsRad = bsRad * scaleFactor;

   Vector3f bsCenterPosRay = transformVector3f(TransRaySystem, bsCenter);

//   float32 RayRadiusAtBsSphereCenter = -bsCenterPosRay[2]*RadiusAt1;
//   float32 sqrDist1 = bsCenterPosRay[1]*bsCenterPosRay[1]+bsCenterPosRay[0]*bsCenterPosRay[0];
//   float32 sqrDist2 = bsRad*bsRad + 2*bsRad*RayRadiusAtBsSphereCenter + RayRadiusAtBsSphereCenter*RayRadiusAtBsSphereCenter;
//   return (sqrDist1 < sqrDist2);

   float32 c = sqrt(SqrRadiusAt1 + 1.0f);
   if (RadiusAt1==0.0f) return false;
   float32 coneSinRecip = c/RadiusAt1;

   float32 coneSinSqr = SqrRadiusAt1/(c*c);
   float32 coneCosSqr = 1.0f/(c*c);

   Vector3f rayDir2 = makeVector3f(0.0f, 0.0f, -1.0f);
   Vector3f rayOrigin2 = makeVector3f(0.0f, 0.0f, 0.0f);

   Vector3f U = rayOrigin2 - rayDir2*(bsRad*coneSinRecip);
   Vector3f D = bsCenterPosRay - U;
   float32 DSqr = D*D;
   float32 e = rayDir2*D;

   if ((e>0) && (e*e >= DSqr*coneCosSqr))
   {
      D = bsCenterPosRay - rayOrigin2;
      DSqr = D*D;
      e = -(rayDir2*D);
      if ((e>0) && (e*e >= DSqr*coneSinSqr))
         return (DSqr<=(bsRad*bsRad));
      else
         return true;
   };

   return false;

}


bool RayPyramidalRange::intersects(const Vector3f& Point) const {
   // transform point into ray coordinate system
   Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
   // only points in front of the 'camera' (which have negative z)
   if (PointRayPos[2]<0.0f) {
      return (((PointRayPos[0]*PointRayPos[0])<(PointRayPos[2]*PointRayPos[2]*SqrHalfWidthAt1))
           && ((PointRayPos[1]*PointRayPos[1])<(PointRayPos[2]*PointRayPos[2]*SqrHalfHeightAt1)));
   }
   return false;
}


// ----------


RayClippedPyramidalRange::RayClippedPyramidalRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sqrhw, float32 sqrhh, float32 nClip, float32 fClip)
      :RayPyramidalRange(right,up,dir,origin,sqrhw,sqrhh) {

   nearClip = nClip;
   farClip = fClip;

}


bool RayClippedPyramidalRange::intersects(const BoundingBox3f& bb) const {

   // to-do: correct test, not cone test

   float32 SqrRadiusAt1 = SqrHalfWidthAt1 + SqrHalfHeightAt1;
   float32 RadiusAt1 = sqrt(SqrRadiusAt1);

   Sphere<float32,3> bs = bb.getBoundingSphere();
   Vector3f bsCenter = bs.center;
   float32 bsRad = bs.radius;
      bsRad = bsRad * scaleFactor;
   Vector3f bsCenterPosRay = transformVector3f(TransRaySystem, bsCenter);

   float32 c = sqrt(SqrRadiusAt1 + 1.0f);
   float32 coneSinRecip = c/RadiusAt1;

   float32 coneSinSqr = SqrRadiusAt1/(c*c);
   float32 coneCosSqr = 1.0f/(c*c);

   Vector3f rayDir2 = makeVector3f(0.0f, 0.0f, -1.0f);
   Vector3f rayOrigin2 = makeVector3f(0.0f, 0.0f, 0.0f);

   Vector3f U = rayOrigin2 - rayDir2*(bsRad*coneSinRecip);
   Vector3f D = bsCenterPosRay - U;
   float32 DSqr = D*D;
   float32 e = rayDir2*D;

   if ((e>0) && (e*e >= DSqr*coneCosSqr))
   {
      D = bsCenterPosRay - rayOrigin2;
      DSqr = D*D;
      e = -(rayDir2*D);
      if ((e>0) && (e*e >= DSqr*coneSinSqr))
         return (DSqr<=(bsRad*bsRad));
      else
         return true;
   };

   return false;

}


bool RayClippedPyramidalRange::intersects(const Vector3f& Point) const {
   // transform point into ray coordinate system
   Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
   // only points in front of the 'camera' (which have negative z)
   if (PointRayPos[2]<0.0f) {
      return (((PointRayPos[0]*PointRayPos[0])<(PointRayPos[2]*PointRayPos[2]*SqrHalfWidthAt1))
           && ((PointRayPos[1]*PointRayPos[1])<(PointRayPos[2]*PointRayPos[2]*SqrHalfHeightAt1))
           && (PointRayPos[2]<=nearClip) && (PointRayPos[2]>=farClip)
           );
   }
   return false;
}

