//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "HierarchicalRangeQueryIterator.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "DebugOutput.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra_templCode.h"
#include "BoundingBox_templCode.h"
#include "MathTools.h"
#include "MathTools_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------
HierarchicalRangeQueryIterator::HierarchicalRangeQueryIterator( PointCloud* pc, card32 subTreeSize, MemoryManager *attachedOctreeMemManager ):
  PointCloudIterator(),
  BasicPointCloudIterator()
{
	m_pc = pc;
	m_buffer = NULL;
	m_seekPoint = makeVector3f(0,0,0);

	// Test for hierarchical point cloud
	native = m_pc->createIterator( SOIT( SOIT::CAP_HIERARCHICAL ) );
	m_hr = dynamic_cast<HierarchicalIterator*>( native );
	nextPointReady = false;

	if ( ( m_hr->getNumChildren() > 0 ) && !m_hr->isLeaf() ) {
		//m_hr->getNumChildren()==0 means no hr
		hierachyMode = true; // put root box on Queue

		debugOutput << "HierarchicalRangeQueryIterator: hierarchical point clouds not supported yet.\n"; debugOutput.flush();

		//putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
	} else{
		hierachyMode = false;
		// Create/Check for octree (native, subTreeSize, m_attachedOctreeMemManager, modifyable)
		m_basic = dynamic_cast<BasicPointCloudIterator*>( native );
		pntSet = m_basic->lockBlockRead();
		POSITION = pntSet->getDescr()->getToken("position");
		m_sds = dynamic_cast<SecondaryDataStructureIterator*>( native );
		pAssert(m_sds);
		// OOC lock
		AttachedDataContainer* attachments = m_sds->lockAttachments();
		m_octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));
		if( m_octree == NULL ){
			m_octree = new AttachedIndexedOctree();
			m_octree->setup( "SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  attachedOctreeMemManager );
			m_octree->octreesetup( pntSet, subTreeSize );
			attachments->attachData( m_octree );
		}
		q.clear();
		nextPointReady = false;
	}
}

HierarchicalRangeQueryIterator::~HierarchicalRangeQueryIterator()
{
}

void HierarchicalRangeQueryIterator::setSeekPointAndReset( const Vector3f& seekPoint, const float32 epsilon )
{
	m_seekPoint = seekPoint;
	m_sqrEpsilon = sqr( epsilon );
	if( hierachyMode ){
		debugOutput << "HierarchicalRangeQueryIterator: hierarchical point clouds not supported yet.\n"; debugOutput.flush();
		/*
		bool showPQs =false;
		//noChaching=true;
		bool newPointinOldCacheBox= cacheBox.inBoundingBox(m_seekPoint);
		if(noChaching|| !newPointinOldCacheBox){
			clearQueues();
			putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
			noChaching=false;
			getNewCacheBox=true;
			cacheMiss++;
		}else{
			cacheRefresh();
			cacheHit++;
			showPQs=true;
		}
		seekNextPoint();
		*/
	} else {
		reset();
	}
}

void HierarchicalRangeQueryIterator::reset()
{
	if ( hierachyMode ){
		debugOutput << "HierarchicalRangeQueryIterator: hierarchical point clouds not supported yet.\n"; debugOutput.flush();
		/*
		clearQueues();
		putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
		noChaching=false;
		getNewCacheBox=true;
		seekNextPoint();
		*/
	} else {
		q.clear();
		q.push_back( m_octree->getOctree() );
		atPoint = -1;
		atIndex = -1;
		seekNextPoint();
	}
}

void HierarchicalRangeQueryIterator::next()
{
	if ( hierachyMode ){
		debugOutput << "HierarchicalRangeQueryIterator: hierarchical point clouds not supported yet.\n"; debugOutput.flush();
		//seekNextPoint();
	}else{
		seekNextPoint();
	}
}

void HierarchicalRangeQueryIterator::seekNextPoint()
{
	// Find next point
	nextPointReady = false;
	while( !nextPointReady && ( q.size() > 0 ) ){
		AIOctreeNode* node = q[0];
		if ( node->isLeaf() ){
			// Node is leaf
			atIndex++;
			IndexedNode& content = node->getContent();
			while(  (int)content.m_index.size() > atIndex ){
				// Look for next point in current node within range
				atPoint = content.m_index[atIndex];
				Vector3f p = pntSet->get3f( atPoint, POSITION );
				if ( ( p - m_seekPoint ) * ( p - m_seekPoint ) <= m_sqrEpsilon ) {
					nextPointReady = true;
					break;
				}
				atIndex++;
			} 
			if ( !nextPointReady ){
				// No more points in node
				q.erase( q.begin() );
				atIndex = -1;
				atPoint = -1;
			}
		} else {
			// Handle children
			for ( unsigned int childIndex = 0; childIndex < 8; childIndex++ ){
				AIOctreeNode* childNode = node->getChild( childIndex ); 
				BoundingBox3f& bbox = childNode->getContent().m_bb;
				float32 bboxDiameter = norm( bbox.upperCorner - bbox.lowerCorner );
				float32 dBox = norm( bbox.getCenter() - m_seekPoint ) - bboxDiameter / 2.0f;
				if ( dBox < 0 ) dBox = 0;
				if ( sqr( dBox ) <= m_sqrEpsilon ){
					q.push_back( childNode );
				}
			}
			q.erase( q.begin() );
			atIndex = -1;
			atPoint = -1;
		}
	}
	if ( nextPointReady ) m_buffer = pntSet->getDataPointer( atPoint );
}

bool HierarchicalRangeQueryIterator::atEnd()
{
	return ( !nextPointReady );
}

card8 HierarchicalRangeQueryIterator::get1ub(const AAT &token) const 
{
	return genericGetS<card8>(token);
}

Vector2ub HierarchicalRangeQueryIterator::get2ub(const AAT &token) const 
{
	return genericGet<card8, 2>(token);
}

Vector3ub HierarchicalRangeQueryIterator::get3ub(const AAT &token) const 
{
	return genericGet<card8, 3>(token);
}

Vector4ub HierarchicalRangeQueryIterator::get4ub(const AAT &token) const 
{
	return genericGet<card8, 4>(token);
}

int32 HierarchicalRangeQueryIterator::get1i(const AAT &token) const 
{
	return genericGetS<int32>(token);
}

Vector2i HierarchicalRangeQueryIterator::get2i(const AAT &token) const 
{
	return genericGet<int32, 2>(token);
}

Vector3i HierarchicalRangeQueryIterator::get3i(const AAT &token) const 
{
	return genericGet<int32, 3>(token);
}

Vector4i HierarchicalRangeQueryIterator::get4i(const AAT &token) const 
{
	return genericGet<int32, 4>(token);
}

float32 HierarchicalRangeQueryIterator::get1f(const AAT &token) const 
{
	return genericGetS<float32>(token);
}

Vector2f HierarchicalRangeQueryIterator::get2f(const AAT &token) const 
{
	return genericGet<float32, 2>(token);
}

Vector3f HierarchicalRangeQueryIterator::get3f(const AAT &token) const 
{
	return genericGet<float32, 3>(token);
}

Vector4f HierarchicalRangeQueryIterator::get4f(const AAT &token) const
{
	return genericGet<float32, 4>(token);
}

void HierarchicalRangeQueryIterator::getVertex(char *buffer) const {
	card32 elementSize= m_pc->getDescr()->getSize();
	for (card32 i=0; i<elementSize; i++) {
		buffer[i] = m_buffer[i];
	};
}

bool HierarchicalRangeQueryIterator::nextBlock()
{
	pAssert(false); 
	return false;
}

const PointSet* HierarchicalRangeQueryIterator::lockBlockRead()
{
	pAssert(false); 
	return NULL;
}

void HierarchicalRangeQueryIterator::unlock()
{
	pAssert(false);
}