//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PointSet.h"
#include "PointSet_templCode.h"
//---------------------------------------------------------------------------
#include "BoundingBox_templCode.h"
#include "LinearAlgebra_templCode.h"
#include "Tools.h"
#include "Exceptions.h"
#include "NumericalClassProperty.h"
#include "NumericalClassProperty_templCode.h"
#include "StringClassProperty.h"
#include "RGBAImage.h"
#include "DebugOutput.h"
#include "MathTools.h"
#include "MathTools_templCode.h"
#include "TemplatedImage_templCode.h"
//---------------------------------------------------------------------------
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
//---------------------------------------------------------------------------
using namespace std;
using namespace geoxl;
//---------------------------------------------------------------------------



MetaClass* PointSet::PointSetMetaClass = NULL;

const MetaClass* PointSet::getInstanceClass() const {
   return getClass();
}

Object* PointSet::newInstance() {
   return new PointSet;
}

const MetaClass* PointSet::getClass() {
   if (PointSetMetaClass==NULL) {
      PointSetMetaClass = new MetaClass("PointSet", "no description.",
                                  newInstance, sizeof(PointSet));
   }
   return PointSetMetaClass;
}

string PointSet::getDescrString() const {
   return vertDescr->getDescription();
}


void PointSet::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 1); // Version +1 is the first 64bit version
   PointSetMetaClass->copyParentProperties();
   BEGIN_CLASS_INIT( PointSet );
   ADD_MAX_PLATFORM_CARD_PROP( width, 0 );
   ADD_MAX_PLATFORM_CARD_PROP( height, 0 );
   ADD_STRING_PROP_METHOD( description, 0, &getDescrString, NULL );
	( (MetaClass *)getClass())->actProperty_setEditReadonly( true );
		SET_EDIT_READONLY( true );
		SET_STREAM_PROPERTY( false );
	ADD_MAX_PLATFORM_CARD_PROP( allocatedHeight, 0 );
		SET_USER_EDIT( false );

}

void PointSet::shutdown() {
   delete PointSetMetaClass;
}

// ----

PointSet::PointSet() {
   vertDescr = new VertexDescriptor();
   data = NULL;
   width = 0;
   height = 0;
   numEntries = 0;
   elementSize = 0;
   lastFlaggedValid = false;
   allocatedHeight=0;
   boundingBoxValid = false;
}

max_platform_int PointSet::getNumFlagged(card32 flag) {
   int32 index = simpleLog2(flag);
   pAssert(index >= 0);
   if (!lastFlaggedValid) {
      for (card32 bit=0; bit<MAX_NUM_POINT_FLAGS; bit++) {
         lastFlagged[bit] = 0;
      }
      if (!providesAttribute("flags")) return 0;
      AAT FLAGS = getAAT("flags");
      for (mpcard i=0; i<numEntries; i++) {
         int32 f = get<int32, 1>(i, FLAGS)[0];
         for (card32 bit=0; bit<MAX_NUM_POINT_FLAGS; bit++) {
            if (((1 << bit) & f) != 0) {
               lastFlagged[bit]++;
            }
         }
      }
      lastFlaggedValid = true;
   }
   return lastFlagged[index];
}


void PointSet::assign(const Object* obj, CopyContext *context) {
   const PointSet *other = dynamic_cast<const PointSet*>(obj);
   pAssert(other != NULL);
   free(data);
   delete vertDescr;
   *this = *other;

   const mpcard numBytes = other->elementSize * other->numEntries;
   this->data = (char*)malloc(numBytes);
   if (this->data == NULL && numBytes > 0) throw PException("PointSet::assign() - cannot allocate memory.");
   memcpy(this->data, other->data, numBytes);
   this->allocatedHeight = height;

   this->vertDescr = (VertexDescriptor*)(other->vertDescr->copy());
   this->boundingBoxValid = false;
   forget(context);
}

void PointSet::write(OutputObjectStream *s) const {
   s->writeObject(vertDescr);
   // write data
   s->writeMaxPlatformCard(width);
   s->writeMaxPlatformCard(height);
   pAssert(numEntries == width * height);
   s->writeMaxPlatformCard(elementSize); // for checking only!
   s->writeBuffer(data, numEntries*elementSize);
}

void PointSet::read(InputObjectStream *s) {
   delete vertDescr;
   s->readObject((Persistent*&)vertDescr);
   if(vertDescr)
      elementSize = vertDescr->getSize();
   else
      elementSize = 0;

   // retrieve extends
   bool version32bit = (s->getClassVersion(PointSet::getClass()) <= MetaClass::FIRST_VERSION);
   if (version32bit) {
      card32 temp;
      s->read(temp);
      width = temp;
   } else {
      s->readMaxPlatformCard(width);
   }
   if (version32bit) {
      card32 temp;
      s->read(temp);
      height = temp;
   } else {
      s->readMaxPlatformCard(height);
   }
   numEntries = width*height;
   mpcard elementSizeTest;

   // only to check!
   if (version32bit) {
      card32 temp;
      s->read(temp);
      elementSizeTest = temp;
   } else {
      s->readMaxPlatformCard(elementSizeTest);
   }
   pAssert(elementSize == elementSizeTest);

   // read data
   free(data);
   data = (char*)malloc(numEntries * elementSize);
   if (data == NULL && numEntries * elementSize > 0) throw PException("PointSet::read() - cannot allocate memory.");
   s->readBuffer(data, numEntries*elementSize);

   this->allocatedHeight = height;
   this->boundingBoxValid = false;
}

void PointSet::clear() {
   width = 0;
   height = 0;
   numEntries = 0;
   free(data);
   data = NULL;
   allocatedHeight = height;
   boundingBoxValid = false;
}

void PointSet::clearAndSetup(mpcard width, mpcard height, const VertexDescriptor *vertDescr) {
   this->width = width;
   this->height = height;
   this->numEntries = width * height;
   if (this->vertDescr != vertDescr) {
      delete this->vertDescr;
      this->vertDescr = (VertexDescriptor*)(vertDescr->copy());
   }
   this->elementSize = vertDescr->getSize();
   free(data);
   data = (char*)malloc(numEntries * elementSize);
   if (data == NULL && numEntries * elementSize > 0) throw PException("PointSet::clearAndSetup() - cannot allocate memory.");
   memset(data, 0, numEntries * elementSize);

   this->allocatedHeight = height;
   boundingBoxValid = false;
}


const VertexDescriptor *PointSet::getDescr() const {
   return vertDescr;
}

PointSet *PointSet::appendFields(const VertexDescriptor *newAttribs) const {
   PointSet *result = new PointSet();
   VertexDescriptor *newDescr = (VertexDescriptor*)(vertDescr->copy());
   newDescr->mergeAttributes(newAttribs);
   result->clearAndSetup(width, height, newDescr);

   const mpint sourceSize = vertDescr->getSize();
   const mpint destSize = newDescr->getSize();
   const mpint addDestOffset = destSize - sourceSize;
   pAssert(addDestOffset >= 0);
   char *dataDest = result->data;
   char *dataSource = data;
   for (mpcard vPos=0; vPos<numEntries; vPos++){
      for (mpint b=0; b<sourceSize; b++) {
         *dataDest = *dataSource;
         dataDest++;
         dataSource++;
      }
      for (mpint b=sourceSize; b<destSize; b++) {
         *dataDest = 0;
         dataDest++;
      }
      // dataDest += addDestOffset;
   }

   delete newDescr;
   return result;
}


void PointSet::changeHeight(const mpcard newheight) {

   // change ad lib
   mpcard factor = 2;

   if (newheight==0)
   {
      clear();
      return;
   }

   if (this->width==0) this->width=1;

   this->elementSize = vertDescr->getSize();
   mpcard allocatedMem = width*allocatedHeight*elementSize;
   mpcard actualMem = width*height*elementSize;
   mpcard neededMem = width*newheight*elementSize;
   
   // got larger && large enough --> exact reallocation (exactly the number of bytes demanded)
   if ((neededMem > allocatedMem*factor) || (this->allocatedHeight==0))
   {
      data = (char*)realloc(data, neededMem);
      if (data == NULL && neededMem > 0) throw PException("PointSet::changeHeight() - cannot reallocate memory.");
      pAssert(data);
      allocatedHeight = newheight;
   }
   // got only a bit larger --> double memory (but no more than 50MB more than needed)
   else if (neededMem > allocatedMem) {
      mpcard incByFactorMem = allocatedMem*factor;
      pAssert(incByFactorMem >= neededMem);

      if (incByFactorMem - neededMem > 50*1024*1024) {
         incByFactorMem = ((neededMem + 50*1024*1024) / elementSize) * elementSize; // round to proper height (integer div!)
      }
      data = (char*)realloc(data, incByFactorMem);
      if (data == NULL && incByFactorMem > 0) throw PException("PointSet::changeHeight() - cannot allocate memory.");

      pAssert(data);

      allocatedHeight = incByFactorMem / elementSize;
   }
   // got smaller && small enough --> exact reallocation (exactly the number of bytes demanded)
   else if (neededMem <= allocatedMem/factor)
   {
      data = (char*)realloc(data, neededMem);
      if (data == NULL && neededMem > 0) throw PException("PointSet::changeHeight() - cannot reallocate memory.");
      pAssert(data);
      allocatedHeight = newheight;                                                                    
   // got only a bit smaller --> zero remaining area
   }

   // zero new mem
   if (neededMem > actualMem) {
      memset(((char*)data)+actualMem, 0, neededMem - actualMem);
   }

   this->height = newheight;
   this->numEntries = width * height;
   boundingBoxValid = false;
}


void PointSet::addPoint(const char *data) {

   pAssert(width == 1 || (width == 0 && height == 0));

   changeHeight(height+1);
   setVertex(height-1, data);
   boundingBoxValid = false;
}

void PointSet::deletePoint(const mpcard index) {
   pAssert(width == 1);
   pAssert(numEntries > 0);
   pAssert(index < numEntries);

   if (index < height-1) {
      setVertex(index, getDataPointer(height-1));
   }
   changeHeight(height-1);
   boundingBoxValid = false;
}

PointSet *PointSet::copyHorizontalSlice(mpcard y) {
   PointSet *result = new PointSet();
   const mpcard width = getWidth();
   result->clearAndSetup(width, 1, getDescr());
   for (mpcard i=0; i<width; i++) {
      result->setVertex(i, 0, getDataPointer(i, y));
   }
   return result;
}

PointSet *PointSet::copyVerticalSlice(mpcard x) {
   PointSet *result = new PointSet();
   const mpcard height = getHeight();
   result->clearAndSetup(1, height, getDescr());
   for (mpcard i=0; i<height; i++) {
      result->setVertex(0, i, getDataPointer(x, i));
   }
   return result;
}


BoundingBox3f PointSet::calcBoundingBox() const {
   pAssert(providesAttribute("position"));
   const AAT POSITION = getAAT("position");
   const mpcard n = getNumEntries();

   BoundingBox3f result;

   if (n<1) {
      result = BoundingBox3f(makeVector3f(0,0,0));
   } else {
      bool resultSet = false;
      for (mpcard i=0; i<n; i++) {
         Vector3f pos = get<float32, 3>(i, POSITION);
         if (!resultSet) {
            result = BoundingBox3f(pos);
            resultSet = true;
         } else {
            result.addPoint(pos);
         }
      }
   };
   return result;
}

BoundingBox3f PointSet::getBoundingBox() const {
   if (!boundingBoxValid) {
      lastBoundingBox = calcBoundingBox();
      boundingBoxValid = true;
   }
   return lastBoundingBox;
}



RGBAImage *PointSet::getRGBAImage(string attributeChannel) const {
   if (!providesAttribute(attributeChannel)) return NULL;

   RGBAImage *result = new RGBAImage(width, height);
   VertexDescriptor RGBAVD;
   RGBAVD.pushAttrib(mVAD("color", 4, VAD::DATA_FORMAT_CARD8));
   vector<card8> conv;
   RGBAVD.convertVertices(const_cast<PointSet*>(this), conv);

   if (conv.size() != width*height*4) {
      delete result;
      return NULL;
   }

   mpcard i = 0;
   for (mpcard y=0; y<height; y++) {
      for (mpcard x=0; x<width; x++) {
         Vector4ub pix;
         pix[0] = conv[i];
         pix[1] = conv[i+1];
         pix[2] = conv[i+2];
         pix[3] = conv[i+3];
         result->setPixel(x,y,pix);
         i += 4;
      }
   }
   return result;
}


PointSet::~PointSet() {
   free(data);
   delete vertDescr;
}


//=========================================================================================




