//---------------------------------------------------------------------------
#ifndef PointSetIteratorH
#define PointSetIteratorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "AttachedData.h"
#include <vector>
#include "VariableArrayClassProperty.h"
//---------------------------------------------------------------------------


namespace geoxl {

class AttachedDataContainer;


class POINTCLOUDS_API PointSetIterator : public PointCloudIterator, public BasicPointCloudIterator,
											public IndexedPointCloudIterator, public RandomAccessPointCloudIterator,
                                 public ModifyablePointCloudIterator, public ParametricPointCloudIterator,
                                 public DynamicParametricPointCloudIterator, public HierarchicalIterator
{
 private:

   static MetaClass *PointSetIteratorMetaClass;
   static Object* newInstance();

   template <class Type, int dim>
   inline StaticVector<Type, dim> genericGet(AAT token) const;
   template <class Type>
   inline Type genericGetS(AAT token) const;
   template <class Type, int dim>
   inline void genericSet(AAT token, StaticVector<Type, dim> v) const;
   template <class Type>
   inline void genericSetS(AAT token, Type x) const;

   void initSetup(PointSet *ps);

 protected:
   mutable bool attachedDataAttribChange;
   mutable bool attachedDataLargeChange;

   PointSetIterator() {ps = NULL; attachedDataLargeChange = false; attachedDataAttribChange = false;}

   inline void invalidateLastNumFlagged() {ps->invalidateLastNumFlagged();}

   PointSet *ps;
   AAT aatPosition;
   card32 x,y, width, height;

  void retarget(PointSet *ps);
  
 public:
   // constructor
   PointSetIterator(PointSet *ps);

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   // BasicPointCloudIterator
   virtual void reset();
   virtual void next();
   virtual bool atEnd();

   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   virtual void     getVertex(char *buffer) const;
   virtual bool nextBlock();
   virtual const PointSet *lockBlockRead();
   virtual void unlock();

	// indexed (random access) iterator
   virtual card64 getNumPointsInBlock() const; 
   virtual card64 getCurrentPointIndex() const;
   virtual void goToIndex(card64 i);

   // ModifyablePointCloudIterator
   /// set attributes in different formats
   virtual void set1ub(const AAT &token, const card8 &x)     const;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const;

   virtual void set1i(const AAT &token, const int32 &x)     const;
   virtual void set2i(const AAT &token, const Vector2i &v) const;
   virtual void set3i(const AAT &token, const Vector3i &v) const;
   virtual void set4i(const AAT &token, const Vector4i &v) const;

   virtual void set1f(const AAT &token, const float32 &x)  const;
   virtual void set2f(const AAT &token, const Vector2f &v) const;
   virtual void set3f(const AAT &token, const Vector3f &v) const;
   virtual void set4f(const AAT &token, const Vector4f &v) const;


   virtual void setVertex(const char *buffer) const;

   virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING);

   
   // DynamicPointIterator
   // see UnstructuredIncorePointCloud


   // DynamicParametricPointCloudIterator
   virtual void setSize(card32 width, card32 height);


   // ParametricPointCloudIterator
   virtual void goTo(int32 x, int32 y);
   virtual card32 getWidth();
   virtual card32 getHeight();
   /// helper functions
   inline card32 getCurrentX() {return x;}
   inline card32 getCurrentY() {return y;}

   // HierarchicalIterator
   virtual card32 getNumChildren();
   virtual bool goToChild(card32 index);
   virtual bool goToParent();
   virtual BoundingBox3f getBoundingBox();
   virtual bool isLeaf() const;
   virtual bool getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data);

	
};

class POINTCLOUDS_API AttachedDataPointSetIterator : public PointSetIterator, public SecondaryDataStructureIterator {
 private:
   AttachedDataContainer *attData;

 protected:
   AttachedDataPointSetIterator() {attData = NULL; attachedDataLargeChange = false; attachedDataAttribChange = false;}
   void retarget(PointSet *ps, AttachedDataContainer *attData);
   
 public:
   // constructor
   AttachedDataPointSetIterator(PointSet *ps, AttachedDataContainer *attData);

   // SecondaryDataStructureIterator
   virtual AttachedDataContainer *lockAttachments();
	virtual void unlockAttachments();

   ~AttachedDataPointSetIterator();

};

class PointCloud;


class POINTCLOUDS_API FixedKNNTopologyAttachment : public AttachedData
{
	GEOXL_CLASS( FixedKNNTopologyAttachment );

private:
	vector<card32>		m_Graph;
public:
	
	void resize( card32 size ) { m_Graph.resize(size); }
	
	card32& operator[](card32 i) {return m_Graph[i];}
	const card32 operator[](card32 i) const {return m_Graph[i];}

	IMPLEMENT_VAR_ARRAY_METHODS_STL( m_Graph );
};


class POINTCLOUDS_API FixedKNNTopologyIterator : public AttachedDataPointSetIterator, 
																public TopologyIterator
{

private:

	void calculate();

	card32			m_NumNeighbours;
	FixedKNNTopologyAttachment	* m_Graph;
	PointCloud	*				m_PointCloud;
	AttachedDataContainer*	m_LockedAttachments;

public:

	FixedKNNTopologyIterator(PointSet *ps, AttachedDataContainer *attData, 
		PointCloud * pc, card32 numNeighbours=6);
	~FixedKNNTopologyIterator();

//	virtual void assign(const Object* obj, CopyContext *context = NULL);

	/// returns number of neighbours for current point
	virtual card32 getNumNeighbours();

	/// returns an index of a neighbour point
	virtual PointIndex getNeighbour( card32 nIndex );

	void setNumNeighbours(card32 n) {m_NumNeighbours=n;}
	card32 getNumNeighbours() const {return m_NumNeighbours;}

	/// Return topology graph
	FixedKNNTopologyAttachment* getGraph(){ return m_Graph; }


};


}

#endif
