//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
#include "BoundingBox_templCode.h"
#include "Persistent.h"
#include "Tools.h"
#include "DebugOutput.h"
#include "PointSet_templCode.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



AttachedIndexedOctree *PointSetKNNQuery::createOctree(
	const PointSet *ps,			// Point set
	card32 maxPointsPerNode	// Number of points per node
	)
{
	AttachedIndexedOctree *octree = new AttachedIndexedOctree();
	octree->setup( "SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE, NULL );
	octree->octreesetup( ps, maxPointsPerNode );
	return octree;
}


PointSetKNNQuery::PointSetKNNQuery(const PointSet *ps, AttachedIndexedOctree *octree, const Vector3f &targetPoint, const AAT &positionAAT) {
	if (ps == NULL) throw PException("PointSetKNNQuery - ps must not be NULL.");
	if (octree == NULL) throw PException("PointSetKNNQuery - octree must not be NULL.");
	this->ps = ps;
   this->targetPoint = targetPoint;
   this->octree = octree;
   this->POSITION = positionAAT;

   rootNode = octree->getOctree();
   putBoxOnQueue(rootNode);
}

void PointSetKNNQuery::setTargetPoint(const Vector3f& targetPoint) {
   this->targetPoint = targetPoint;

   static const std::priority_queue<PointEntry> EMPTY_POINTS;
   static const std::priority_queue<BoxEntry> EMPTY_BOXES;

   pqBoxes = EMPTY_BOXES;
   pqPoints = EMPTY_POINTS;
   putBoxOnQueue(rootNode);
}

void PointSetKNNQuery::putBoxOnQueue(AIOctreeNode* box) {
   // calc box-point distance
   float32 qDistance;

   Sphere3f sp=box->getContent().m_bb.getBoundingSphere();
   qDistance=(sp.center-targetPoint)*(sp.center-targetPoint);
   if (qDistance<sp.radius*sp.radius) {
      qDistance=-1.0f;
   } else {
      qDistance=sqrt(qDistance);
      qDistance-=sp.radius;
      qDistance*=qDistance;
   }

   BoxEntry bEntry(box, qDistance);
   pqBoxes.push(bEntry);
}

void PointSetKNNQuery::putPointOnQueue(const card32 &indexOfPoint) {
   // calc point-point distance
   Vector3f pos = ps->get<float32,3>(indexOfPoint, POSITION);
   float32 qDistance = (pos-targetPoint)*(pos-targetPoint);
   PointEntry pnt(indexOfPoint, qDistance);
   pqPoints.push(pnt);
}

void PointSetKNNQuery::getNextNearestPointIndex(card32 &nextIndex, bool &found) {
   // check for end of points
   if (pqBoxes.empty() && pqPoints.empty()) {   // no more points
      nextIndex = 0xFFFFFFFF;
      found = false;
      return;
   }
   
	// start search
   while (true) {
      if (!pqPoints.empty()) {
         if (pqBoxes.empty() || pqPoints.top().quDist <= pqBoxes.top().quDist) { // point nearer than any box
            card32 i = pqPoints.top().index;
            pqPoints.pop();
				nextIndex = i;
				found = true;
            return;
         }
      }

      // step into node
      if (pqBoxes.empty()) {found = false; return;}
      AIOctreeNode *thisBox = pqBoxes.top().node;
      pqBoxes.pop();

      if(thisBox->isLeaf()){  //Put Boxes Points on Queue
         size_t size = thisBox->getContent().m_index.size();
         for(card32 i=0; i<size; i++){
            putPointOnQueue(thisBox->getContent().m_index[i]);
         }

      } else { // put child boxes on queue
         for (card32 i=0; i<8; i++){
            putBoxOnQueue(thisBox->getChild(i));
         }
      }
   } //while(true)
}



