#ifndef RECONSTRUCTION_GRAPH_H
#define RECONSTRUCTION_GRAPH_H

// System
#include <vector>

// Xrt
#include "LinearAlgebra.h"
#include "DebugOutput.h"

/// Represention on node in the ReconstructionGraph containing a region frontier
struct FrontierNode
{
   vector<int32> opposites;
   int32 id;
   Vector3f color;
   card32 numPoints;
};

/// Represention on node in the ReconstructionGraph containing a region
struct RegionNode
{
   vector<FrontierNode*> frontiers;
   int32 id;
   Vector3f color;
   card32 numPoints;
   int32 addPoint(){ numPoints++; return id; }
   void removePoint(){ numPoints--; }
};

/// Graph structure representing the reconstructed object
class POINTCLOUDS_API ReconstructionGraph
{
   private:
      /// Regions in the graph
      vector<RegionNode*> regions;
      /// Frontiers in the graph
      vector<FrontierNode*> frontiers;

      /// Create new region object with unique id
      RegionNode* createRegion();
      /// Create new frontier with unique id
      FrontierNode* createFrontier();

   public:
      /// Constructor
      ReconstructionGraph();
      /// Reset
      void clear();
      /// Dump state to console
      void print();

      /*--- REGIONS ---*/
      
      /// Add region node to graph, return new node
      RegionNode* addRegion();
      /// Return the default region
      RegionNode* getDefaultRegion(){ return regions[0]; }
      /// Return region by id;
      RegionNode* getRegionNode( int32 id );
      /// Delete region
      void deleteRegion( int32 id );

      // Direct node access (performance)
      void deleteRegionNodeByIndex( card32 index );
      // Direct node access (performance)
      inline RegionNode* getRegionNodeByIndex( card32 index ){ return regions[index]; };
      // Direct node access (performance)
      inline card32 getNumRegions(){ return regions.size(); }


      /*--- FRONTIERS ---*/

      /// Add frontier to region; return reference to new frontier
      FrontierNode* addFrontier( int32 regionNodeID );
      /// Return frontier by id;
      FrontierNode* getFrontierNode( int32 id );

      // Direct node access (performance)
      void deleteFrontierNodeByIndex( card32 index );
      // Direct node access (performance)
      inline FrontierNode* getFrontierNodeByIndex( card32 index ){ return frontiers[index]; };
      // Direct node access (performance)
      inline card32 getNumFrontiers(){ return frontiers.size(); }
};

/// Unique id generator
static card32 uniqueID = 0;

/**
 * Create new object
 */
ReconstructionGraph::ReconstructionGraph()
{
   clear();
}

/**
 * Clear all nodes
 */
void ReconstructionGraph::clear()
{
   for ( unsigned int n = 0; n < regions.size(); n++ ){
      RegionNode* rn = regions[n];
      for ( unsigned int m = 0; m < rn->frontiers.size(); m++ ){
         delete rn->frontiers[m];
      }
      rn->frontiers.clear();
      delete rn;
   }
   regions.clear();

   // Create default region
   RegionNode* defaultRegion = addRegion();
   defaultRegion->color = makeVector3f( 0,1,0 );
}

/**
 * Dump state to console
 */
void ReconstructionGraph::print()
{
   debugOutput << "ReconstructionGraph (" << regions.size() << " nodes):\n";
   for ( unsigned int n = 0; n < regions.size(); n++ ){
      debugOutput << " RegionNode(id=" << regions[n]->id <<"), numPoints: " << regions[n]->numPoints << "";
      FrontierNode* fn;
      for ( unsigned int m = 0; m < regions[n]->frontiers.size(); m++ ){
         fn = regions[n]->frontiers[m];
         debugOutput << "  FrontierNode(id=" << fn->id << "), numPoints: " << fn->numPoints << "\n";
      }
      debugOutput << "\n";
   }
   unsigned int numPoints = 0;
   for ( unsigned int m = 0; m < regions.size(); m++ )
      numPoints += regions[m]->numPoints;
   debugOutput << "Number of points in graph: " << numPoints << ".\n";
}

/**
 * Create new region with unique id
 */
RegionNode* ReconstructionGraph::createRegion()
{
   RegionNode* regionNode = new RegionNode();
   regionNode->id = uniqueID;
   uniqueID++;
   regionNode->color = makeVector3f( rand()%70/100.0+0.3, rand()%70/100.0+0.3, rand()%70/100.0+0.3 );
   regionNode->frontiers.clear();
   regionNode->numPoints = 0;
   regions.push_back( regionNode );
   return regionNode;
}

/**
 * Create new frontier with unique id
 */
FrontierNode* ReconstructionGraph::createFrontier()
{
   FrontierNode* frontierNode = new FrontierNode();
   frontierNode->id = uniqueID;
   frontierNode->opposites.clear();
   frontierNode->numPoints = 0;
   uniqueID++;
   frontierNode->color = makeVector3f( rand()%70/100.0+0.3, rand()%70/100.0+0.3, rand()%70/100.0+0.3 );
   frontiers.push_back( frontierNode );
   return frontierNode;
}

/**
 * Add region node to graph, return reference to new node
 */
RegionNode* ReconstructionGraph::addRegion()
{
   RegionNode* newNode = createRegion();
   return newNode;
}

/**
 * Add frontier to region; return reference to new frontier 
 */
FrontierNode* ReconstructionGraph::addFrontier( int32 regionNodeID )
{
   unsigned int n;
   for ( n = 0; n < regions.size(); n++ ){
      if ( regions[n]->id == regionNodeID ){
         FrontierNode* frontierNode = createFrontier();
         regions[n]->frontiers.push_back( frontierNode );
         return frontierNode;
      }
   }
   debugOutput << "Could not find region with id " << regionNodeID << "!\n";
   return NULL;
}

/**
 * Return region by id;
 */
RegionNode* ReconstructionGraph::getRegionNode( int32 id )
{
   unsigned int n;
   for ( n = 0; n < regions.size(); n++ ){
      if ( regions[n]->id == id )
         return regions[n];
   }
   return NULL;
}

/**
 * Return frontier by id;
 */
FrontierNode* ReconstructionGraph::getFrontierNode( int32 id )
{
   unsigned int n;
   for ( n = 0; n < frontiers.size(); n++ ){
      if ( frontiers[n]->id == id )
         return frontiers[n];
   }
   return NULL;
}

/**
 * Direct node access (performance)
 */
void ReconstructionGraph::deleteFrontierNodeByIndex( card32 index )
{
   unsigned int n, m;
   // Find reference
   FrontierNode* fn = getFrontierNodeByIndex( index );
   if ( fn == NULL ) return;
   // Remove from region
   for ( n = 0; n < regions.size(); n++ ){
      RegionNode* rn = regions[n];
      for ( m = 0; m < rn->frontiers.size(); m++ ){
         if ( rn->frontiers[m] == fn ){
            rn->frontiers.erase( rn->frontiers.begin() + m );
            break;
         }
      }
   }
   // Remove from frontiers array
   frontiers.erase( frontiers.begin() + index );
   // Free memory
   delete fn;
}

/**
 * Direct node access (performance)
 */
void ReconstructionGraph::deleteRegionNodeByIndex( card32 index )
{
   RegionNode* region = regions[index];
   if ( region->id == getDefaultRegion()->id ){
      debugOutput << "Trying to delete default region!\n";
      return;
   }
   regions.erase( regions.begin() + index );
   delete region;
}

/**
 * Delete region
 */
void ReconstructionGraph::deleteRegion( int32 id )
{
   for ( unsigned int n = 0; n < regions.size(); n++ ){
      RegionNode* rn = regions[n];
      if ( rn->id == id ){
         if ( rn->id == getDefaultRegion()->id ){
            debugOutput << "Trying to delete default region!\n";
            return;
         }
         regions.erase( regions.begin() + n );
         delete rn;
         return;
      }
   }
   debugOutput << "Could not find region with id " << id << "!\n"; 
}

#endif