//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RegularVolume.h"
#include "RegularVolume_templCode.h"
//---------------------------------------------------------------------------
#include "ObjectClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "NumericalClassProperty.h"
#include "PointSet_templCode.h"
#include <fstream>
#include "Stringhelper.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------


IMPLEMENT_GEOXL_CLASS( RegularVolume, 0 ) {
	BEGIN_CLASS_INIT( RegularVolume )
	ADD_OBJECT_PROP( voxels, 0, PointSet::getClass(), true )
	ADD_CARD32_PROP( width, 0 )
	ADD_CARD32_PROP( height, 0 )
	ADD_CARD32_PROP( depth, 0 )
	ADD_VECTOR3F_PROP( bb.lowerCorner, 0 )
	ADD_VECTOR3F_PROP( bb.upperCorner, 0 )
}

// --

RegularVolume::RegularVolume() {
	width = 0;
	height = 0;
	depth = 0;
	voxels = 0;
	bb = BoundingBox3f(NULL_VECTOR3F);
}

void RegularVolume::clearAndSetup(card32 width, card32 height, card32 depth, const VertexDescriptor *descr, const BoundingBox3f &bb) {
	this->width = width;
	this->height = height;
	this->depth = depth;
	this->bb = bb;
	delete voxels;
	voxels = new PointSet();
	voxels->clearAndSetup(1, width*height*depth, descr);
}

void RegularVolume::computeMinMax1f(string channel1f, float32 &min, float32 &max) {
	const card32 n = voxels->getNumEntries();
	AAT aat = voxels->getAAT(channel1f);
	min = voxels->get<float32, 1>(0, aat)[0];
	max = min;
	for (card32 i=1; i<n; i++) {
		float32 v = voxels->get<float32, 1>(i, aat)[0];
		if (v>max) max = v;
		if (v<min) min = v;
	}
}

void RegularVolume::setScalarValue(const AAT channel, card32 x, card32 y, card32 z, float32 value) {
	card32 pos = x + y*width + z*width*height;
	pAssert(pos < voxels->getNumEntries());
	voxels->set<float32>(pos, channel, value);
}

void RegularVolume::createAttribute(VAD attribute) 
{
	if (!voxels->providesAttribute(attribute.name, attribute.numComponents, attribute.dataFormat)) {
		if (voxels->providesAttribute(attribute.name)) {
			throw PException("RegularInCoreVolume::createAttribute() - channel exists already with different parameters (numComponents/dataFormat)");
		}
		VertexDescriptor newAttribs;
		newAttribs.pushAttrib(attribute);
		PointSet *newV = voxels->appendFields(&newAttribs);
		delete voxels;
		voxels = newV;
	}
}

void RegularVolume::setVoxelData(PointSet *ps, card32 width, card32 height, card32 depth) 
{
   if (voxels != ps) {
      delete voxels;
      voxels = ps;
      this->width = width;
      this->height = height;
      this->depth = depth;
      pAssert(ps == NULL || width*height*depth == ps->getNumEntries());
   }
}

RegularVolume *RegularVolume::appendFields(const VertexDescriptor *newAttribs) const 
{
   RegularVolume *result = new RegularVolume();
   PointSet *newPS = voxels->appendFields(newAttribs);
   result->clearAndSetup(width, height, depth, newPS->getDescr(), bb);
   result->setVoxelData(newPS, width, height, depth);
   return result;
}

bool RegularVolume::loadFromMHDRaw( const char * filename ) 
{
	ifstream input( filename );
	if( !input.is_open() )
	{
		return false;
	}

	string line;
	int dimX,dimY,dimZ;
	dimX = dimY = dimZ = 0;
	Vector3f elementSpacing = NULL_VECTOR3F;
	string rawFilename;
	while( getline( input, line) )
	{
		vector<string> words = splitString( line, " \t" );
		if( words.size() < 3 )
			continue;

		if( words[0] == "NDims")
			if( words[2] != "3" )
				return false;	// no volume data set!

		if( words[0] == "DimSize" )
		{
			if( words.size() < 5)
				return false;

			if( !strToInt( words[2], dimX ) )
				return false;
			if( !strToInt( words[3], dimY ) )
				return false;
			if( !strToInt( words[4], dimZ ) )
				return false;

		}

		if( words[0] == "ElementType")
		{
			if( (words[2] != "MET_USHORT") && (words[2] != "MET_SHORT"))
				return false;	
		}

		if( words[0] == "ElementSpacing" )
		{
			if( words.size() < 5)
				return false;

			if( !strToFloat( words[2], elementSpacing[0] ) )
				return false;
			if( !strToFloat( words[3], elementSpacing[1] ) )
				return false;
			if( !strToFloat( words[4], elementSpacing[2] ) )
				return false;

		}

		if( words[0] == "ElementDataFile")
			rawFilename = words[2];
	}
	input.close();
	width = dimX;
	height = dimY;
	depth = dimZ;
	bb.lowerCorner = elementSpacing * -0.5f;
	bb.upperCorner = elementSpacing *  0.5f;
	bb.upperCorner[0] *= dimX;
	bb.upperCorner[1] *= dimY;
	bb.upperCorner[2] *= dimZ;
	bb.lowerCorner[0] *= dimX;
	bb.lowerCorner[1] *= dimY;
	bb.lowerCorner[2] *= dimZ;

	if( rawFilename.length() > 0 && dimX*dimY*dimZ > 0)
	{
		string tmpFilename = removeFilenameFromPath(string(filename)) + rawFilename;
		ifstream rawFile( tmpFilename.c_str(), ios::binary);
		if( !rawFile.is_open())
			return false;

		VertexDescriptor * vd = new VertexDescriptor;
		vd->pushAttrib( VAD("position",3,VAD::DATA_FORMAT_FLOAT32));
		vd->pushAttrib( VAD("intensity",1,VAD::DATA_FORMAT_FLOAT32) );
		delete voxels;
		voxels = new PointSet;
		voxels->clearAndSetup( 1, dimX*dimY*dimZ, vd );

		vector<card16> tmpBuffer(dimX*dimY*dimZ);
		rawFile.read( (char*)&tmpBuffer[0], dimX*dimY*dimZ * sizeof(card16) );

		AAT intAAT = voxels->getAAT( "intensity" );
		AAT posAAT = voxels->getAAT( "position" );

		for(card32 x = 0; x < (card32)dimX; x++)
		{
			for(card32 y = 0; y < (card32)dimY; y++)
			{
				for(card32 z = 0; z < (card32)dimZ; z++)
				{
					card32 i = x+dimX*y+dimX*dimY*z;
					card16 test = tmpBuffer[i];
					test = (test >> 8) + ((test&0xff) << 8);
					voxels->set1f(i, intAAT, (float)test / 4096.0f);
					float32 xf = bb.getCorner(0)[0]+(((x+0.5)*bb.getSideLength(0))/(float32)dimX);
					float32 yf = bb.getCorner(0)[1]+(((y+0.5)*bb.getSideLength(1))/(float32)dimY);
					float32 zf = bb.getCorner(0)[2]+(((z+0.5)*bb.getSideLength(2))/(float32)dimZ);
					voxels->set3f(i, posAAT, makeVector3f(xf, yf, zf));
				}
			}
		}
	}
	return true;
}

RegularVolume *RegularVolume::resample( card32 dimX, card32 dimY, card32 dimZ ) 
{
	RegularVolume * ret = new RegularVolume;
	ret->clearAndSetup(dimX,dimY,dimZ, (VertexDescriptor*)voxels->getDescr()->copy(), bb);
	
	vector<AAT> tokens;
	vector<VAD> attributes;
	for( card32 i=0;i<voxels->getDescr()->getNumAttributes();i++ )
	{
		tokens.push_back( voxels->getDescr()->getToken(i));
		attributes.push_back( voxels->getDescr()->getAttribute(i) );
	}

	
	FOR_XYZ( ret )
	{
		Vector3f sourcePos = makeVector3f( (float)x/(float)dimX*(float)width, 
			(float)y/(float)dimY*(float)height,
			(float)z/(float)dimZ*(float)depth);
		card32 index = (z*dimY + y)*dimX + x;
		for( card32 i=0;i<attributes.size();i++)
		{
			if( attributes[i].dataFormat == VAD::DATA_FORMAT_FLOAT32 && attributes[i].numComponents == 1)
				ret->voxels->set1f(index,tokens[i], interpolateAttribute1f(sourcePos,tokens[i]));
		}

	}
	END_FOR_XYZ();

	return ret;
}

void RegularVolume::assign(const Object* obj, CopyContext *context) 
{
	forget(context);
	const RegularVolume *other = dynamic_cast<const RegularVolume*>(obj);
	if (!other) {
		pAssert(false);
		return;
	}
	this->width = other->width;
	this->height = other->height;
	this->depth = other->depth;
	this->bb = other->bb;
	if (this->voxels) {
		this->voxels->assign(other->voxels);
	} else {
		this->voxels = dynamic_cast<PointSet*>(other->voxels->copy(context));
	}
}

geoxl::RegularVolume::~RegularVolume()
{
	delete voxels;
}