//---------------------------------------------------------------------------
#ifndef SceneObjectIteratorH
#define SceneObjectIteratorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Object.h"
#include "BoundingBox.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------
#include <string>
//---------------------------------------------------------------------------


namespace geoxl {
   using namespace std;

class AttachedDataContainer;


/// Base class of all iterator objects that are used to browse the contents of scene objects
class POINTCLOUDS_API SceneObjectIterator : public Object {
 private:
   static MetaClass *SceneObjectIteratorMetaClass;
   static Object* newInstance();

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();
};


/// Descriptor to identify the scene object iterator type
class POINTCLOUDS_API SceneObjectIteratorType {
 protected:
   card32 flags;

   static const card32 PC_MAX_BIT_MASK = 0xFFF;

 public:
   // general caps
   static const card32 CAP_ATTACH_SECONDARY = 0x1;
   static const card32 CAP_HIERARCHICAL = 0x2;
   // point cloud caps
   static const card32 CAP_BASIC_PC = 0x4;
   static const card32 CAP_MODIFYABLE_PC = 0x8;
   static const card32 CAP_DYNAMIC_PC = 0x10;
   static const card32 CAP_PARAMETRIC_PC = 0x20;
   static const card32 CAP_DYNAMIC_PARAMETRIC_PC = 0x40;
   static const card32 CAP_TRIANGLE_MESH = 0x80;
   static const card32 CAP_REGULAR_VOLUME = 0x100;
   static const card32 CAP_INDEXED_PC = 0x200;
   static const card32 CAP_RANDOM_ACCESS_PC = 0x200;
	static const card32 CAP_TOPOLOGY = 0x400;
	static const card32 CAP_VOLUMETRIC = 0x800;

   SceneObjectIteratorType() {flags = 0;}
   SceneObjectIteratorType(card32 caps) {pAssert(caps <= PC_MAX_BIT_MASK); flags = caps;}

   void operator&=(const SceneObjectIteratorType &caps) {flags &= caps.flags;}
   void operator|=(const SceneObjectIteratorType &caps) {flags |= caps.flags;}
   void removeCaps(const SceneObjectIteratorType &caps) {flags &= (PC_MAX_BIT_MASK ^ caps.flags);}

   bool exactMatch(const SceneObjectIteratorType &other) {return this->flags == other.flags;}

   bool providesCapsFor(const SceneObjectIteratorType &requirements) {
      return flags == (requirements.flags | flags);
   }

   bool attachSecondary() {return (flags & CAP_ATTACH_SECONDARY) != 0;}
   bool hierarchical() {return (flags & CAP_HIERARCHICAL) != 0;}
   bool basicPC() {return (flags & CAP_BASIC_PC) != 0;}
   bool indexedPC() {return (flags & CAP_INDEXED_PC) != 0;}
   bool randomAccessPC() {return (flags & CAP_INDEXED_PC) != 0;}
   bool modifyablePC() {return (flags & CAP_MODIFYABLE_PC) != 0;}
   bool dynamicPC() {return (flags & CAP_DYNAMIC_PC) != 0;}
   bool parametricPC() {return (flags & CAP_PARAMETRIC_PC) != 0;}
   bool dynamicParametric() {return (flags & CAP_DYNAMIC_PARAMETRIC_PC) != 0;}
   bool triangleMesh() {return (flags & CAP_TRIANGLE_MESH) != 0;}
	bool topology() {return (flags & CAP_TOPOLOGY) != 0;}

};

typedef SceneObjectIteratorType SOIT;




/** Allows to attach secondary data structures to the current block / hierarchy node.
    For hierarchical iterators, attachments are always attached to the current hierarchy node.
    For non-hierarchical iterators, attachements are always attached to the current block (which
    may be a hierarchy node, internally, and appear at this node, too) */
class POINTCLOUDS_API SecondaryDataStructureIterator {
 public:
	/// mw - May 4. 2006 - interface changed; more flexibility and out-of-core safe

	/** returns a pointer to attached data, the data will be available unless the corresponding unlock 
	    operation is called. multiple nested locks must be supported (lock counting). */
	virtual AttachedDataContainer *lockAttachments() = 0;
	
	/** unlocks the data object (might be unavailable afterwards). multiple nested locks must 
	    be supported (lock counting). */
	virtual void unlockAttachments() = 0;
};




/** Adds the capability of accessing multiple levels in a multi-resolution hierarchy */
class POINTCLOUDS_API HierarchicalIterator { /* INTERFACE CLASS */
 public:

   /** get (potential) number of child nodes in hierarchy */
   virtual card32 getNumChildren() = 0;

   /** go to child node number "index"; the function returns false if the child node slot is empty. */
   virtual bool goToChild(card32 index) = 0;

   /** go to parent node; the function returns false if the node has no parent node. */
   virtual bool goToParent() = 0;

   /** get bounding box of current hierarchy node */
   virtual BoundingBox3f getBoundingBox() = 0;

   /** is current hierarchy node leaf */
   virtual bool isLeaf() const = 0;

   /** no points in node (see below) */
   static const float32 MAX_POINT_SPACING; // == 1E35

   enum NodeInfoType {      // parameter      data
      niPointSpacing,       // -ignored-      float*
      niFlaggedInSubtree,   // card32 *flag   card64*     ('flag' is the flag itself, not its index. it must be a power of two.)
      niNumPointsInNode,    // -ignored-      card64*
      niNumPointsInSubtree, // -ignored-      card64*
      niChildAvailable,		 // card32 *index	 bool*       (out-of-core only: returns true if child is in memory, *index == 0xFFFFFFFF queries for all children)
      niPrefetchChild		 // card32 *index  -no result- (like niChildAvailable, asks to prefetch the specified (or all) child nodes)
   };
   // Example: how many Points are selected -> card64 count; getNodeInfo(niFlaggedInSubtree, &PF_FLAG_SELECTED, &count);

   /** get information about current node. whichInfo determins type, parameter refers to parameter, data is written to *data.
       result is true if info is available (and thus has been written to *data) */
   virtual bool getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data) = 0;

};



}


#endif
