#ifndef SuperIteratorH
#define SuperIteratorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PointCloudIterators.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------

namespace geoxl
{
class PointCloud;

/**
	Adaptor for a point cloud iterator - 
	Every get/set method operates on a local buffer.
	You have to call flush() to write the buffer into the point cloud.
	The name "SuperIterator" is very important and scientifically proven!
*/
class POINTCLOUDS_API SuperIterator
{
private:

	BasicPointCloudIterator				* bIt;
	IndexedPointCloudIterator			* iIt;
	RandomAccessPointCloudIterator	* rIt;
	ModifyablePointCloudIterator		* mIt;
	DynamicPointCloudIterator			* dIt;
	TopologyIterator						* tIt;
	ParametricPointCloudIterator		* pIt;
	VolumetricIterator					* vIt;

	SceneObjectIterator					* m_Iter;
	vector<char>							m_Buffer; // local vertex buffer
	char*										m_BufferPtr;
	const VertexDescriptor			*	m_VD;

	// constants for conversion routines
	float		m_ConversionConstant_Card8;
	float		m_ConversionConstant_Card16;
	float		m_ConversionConstant_Card32;

public:

	/// becomes owner of iterator!
	SuperIterator(SceneObjectIterator * iter, const VertexDescriptor * vd);
	~SuperIterator(void);

	/// writes local buffer into point cloud
	void flush();

	inline card32 getVertexBufferSize() const
	{
		return m_VD->getSize();
	}


// ------------ basic interface --------------------------

	/// go to first entry
   void reset();
   /// go to next entry
   void next();
   /// true if end-of-set
   bool atEnd();
   
	/// get local buffer, call getVertexBufferSize() for requiered buffer size
   void     getVertex(char *buffer) const;

	template<class DataType>
	DataType get(const AAT &token) const;

// ------------ modifiy interface --------------------------

	template<class DataType>
	void set(const AAT &token, const DataType &data) const;

// -------------- generic interface -----------------
// this interface is designed for volumetric data in order to provide
//	generic access to arbituary data formats (card8,card16,float,double,...)
// default conversion of data types:
// dtype		|		get		|		set
// card8		|		/255		|		*255			clamped to [0,1]
// card16	|		/2^12		|		*2^12			most data sets provide only 12 bits
// card32	|		/2^32		|		*2^32

	float getConversionConstant_Card8() const {return m_ConversionConstant_Card8;}
	void setConversionConstant_Card8(float v) {m_ConversionConstant_Card8=v;}

	float getConversionConstant_Card16() const {return m_ConversionConstant_Card16;}
	void setConversionConstant_Card16(float v) {m_ConversionConstant_Card16=v;}

	float getConversionConstant_Card32() const {return m_ConversionConstant_Card32;}
	void setConversionConstant_Card32(float v) {m_ConversionConstant_Card32=v;}

	/// 
	float gget1( const AAT &token, const VAD &vad ) const;

	void gset1( const AAT &token, const VAD &vad, const float &d);


// --------------- other interfaces -----------------

   /// get the index of the current point within the current block (current node in the hierarchical case)
	card64 getCurrentPointIndex() const;

	/// move iterator position to a certain position (within current block/node)
	void goToIndex(card64 i);

   /** adds single point. data points to vertex data with size according to the
       vertex descriptor of the point cloud. */
   void add(const char *newPoint);

   /** adds set of points */
   void add(const PointSet *newPoints);

   /** deletes point at current position of the iterator */
   void deleteCurrent();

   /** go to parametric position */
   void goTo(int32 x, int32 y);

   /** get width of parametric grid */
   card32 getWidth();

   /** get height of parametric grid */
   card32 getHeight();

	/// returns number of neighbours for current point
	card32 getNumNeighbours();

	/// returns an index of a neighbour point
	PointIndex getNeighbour( card32 nIndex );

	/// goto voxel (x,y,z)
	void goTo(card32 x, card32 y, card32 z);

	/// get extends
	Vector3i getVolumeSize();

};


}

#endif