#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TopologicalKNNIterator.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "PointCloudIterators.h"
#include "SceneObjectIterator.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
//#include 
//---------------------------------------------------------------------------
using namespace geoxl;


TopologicalKNNIteratorBase::TopologicalKNNIteratorBase(PointCloud * pc, PointCloudIterator * topologyIter)
{
	m_Pc = pc;
	m_Buffer = NULL;
	m_posAAT = pc->getAAT( "position" );

	m_Iter = topologyIter;

	m_bIt = dynamic_cast<BasicPointCloudIterator*>(m_Iter);
	m_tIt = dynamic_cast<TopologyIterator*>(m_Iter);
	m_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_Iter);

	pAssert( m_bIt != NULL );
	pAssert( m_tIt != NULL );
	pAssert( m_rIt != NULL );

	m_PointSet = m_bIt->lockBlockRead();
	m_StartIndex = 0;
}

TopologicalKNNIteratorBase::~TopologicalKNNIteratorBase()
{
	m_bIt->unlock();
}

void TopologicalKNNIteratorBase::setStartPointAndReset(const PointIndex &index) 
{
	m_StartIndex = index;
	reset();
}

bool TopologicalKNNIteratorBase::atEnd() 
{
	return !m_NextPointExists;
}

card8 TopologicalKNNIteratorBase::get1ub(const AAT &token) const {
   return genericGetS<card8>(token);
}
Vector2ub TopologicalKNNIteratorBase::get2ub(const AAT &token) const {
   return genericGet<card8, 2>(token);
}
Vector3ub TopologicalKNNIteratorBase::get3ub(const AAT &token) const {
   return genericGet<card8, 3>(token);
}
Vector4ub TopologicalKNNIteratorBase::get4ub(const AAT &token) const {
   return genericGet<card8, 4>(token);
}
int32 TopologicalKNNIteratorBase::get1i(const AAT &token) const {
   return genericGetS<int32>(token);
}
Vector2i TopologicalKNNIteratorBase::get2i(const AAT &token) const {
   return genericGet<int32, 2>(token);
}
Vector3i TopologicalKNNIteratorBase::get3i(const AAT &token) const {
   return genericGet<int32, 3>(token);
}
Vector4i TopologicalKNNIteratorBase::get4i(const AAT &token) const {
   return genericGet<int32, 4>(token);
}
float32 TopologicalKNNIteratorBase::get1f(const AAT &token) const {
   return genericGetS<float32>(token);
}
Vector2f TopologicalKNNIteratorBase::get2f(const AAT &token) const {
   return genericGet<float32, 2>(token);
}
Vector3f TopologicalKNNIteratorBase::get3f(const AAT &token) const {
   return genericGet<float32, 3>(token);
}
Vector4f TopologicalKNNIteratorBase::get4f(const AAT &token) const {
   return genericGet<float32, 4>(token);
}

void TopologicalKNNIteratorBase::getVertex(char *buffer) const {
     card32 elementSize= m_Pc->getDescr()->getSize();
     for (card32 i=0; i<elementSize; i++) {
         buffer[i] = m_Buffer[i];
     };
}

PointIndex TopologicalKNNIteratorBase::getPointSetPointNumber() 
{
	return m_CurrentIndex;
}

//----------------------------------------------------------------------------------
TopologicalKNNIterator::TopologicalKNNIterator(PointCloud* pc, PointCloudIterator * topologyIter ) : 
   TopologicalKNNIteratorBase(pc, topologyIter)
{
}

TopologicalKNNIterator::~TopologicalKNNIterator() 
{
}

void TopologicalKNNIterator::reset() 
{
   m_Visited.clear();
   m_Queue.clear();
   m_NextPointExists = true;
   visit( m_StartIndex, 0.0f );
}

void TopologicalKNNIterator::next() 
{
   bool found = false;
   while( !found )
   {
      if( m_Queue.isEmpty() )
         break;

      float dist = m_Queue.getFirstKey();
      PointIndex nextIndex = m_Queue.pop_first();

      if( m_Visited.find(nextIndex) == m_Visited.end() )
      {
         visit( nextIndex, dist );
         found = true;
      }
   }
   m_NextPointExists = found;
}

void TopologicalKNNIterator::visit( PointIndex index,float distance) 
{
   m_CurrentIndex = index;
   m_CurrentTopologicalDistance = distance;
   m_rIt->goToIndex( index );
   m_Buffer = m_PointSet->getDataPointer( index );
   m_Visited[index] = distance;

   Vector3f pos = m_bIt->get3f( m_posAAT );

   // put neighbours on queue
   for( card32 i=0;i<m_tIt->getNumNeighbours();i++ )
   {
      PointIndex nIndex = m_tIt->getNeighbour(i);
      if( m_Visited.find( nIndex ) == m_Visited.end() )
      {
         Vector3f nPos = m_PointSet->get3f( nIndex, m_posAAT );
         float nDist = distance + norm(pos - nPos);

         m_Queue.insert( nDist, nIndex );
      }
   }
}


//----------------------------------------------------------------------------------
TopologicalKNNIterator_Fast::TopologicalKNNIterator_Fast(PointCloud* pc, PointCloudIterator * topologyIter ) : 
   TopologicalKNNIteratorBase(pc, topologyIter)
{
   m_Visited.resize(pc->getNumPoints(), -1.0);
}

TopologicalKNNIterator_Fast::~TopologicalKNNIterator_Fast() 
{
}

void TopologicalKNNIterator_Fast::reset() 
{
   m_Visited.clear();
   m_Visited.resize(m_Pc->getNumPoints(), -1.0);
   m_Queue.clear();
   m_NextPointExists = true;
   visit( m_StartIndex, 0.0f );
}

void TopologicalKNNIterator_Fast::next() 
{
   bool found = false;
   while( !found )
   {
      if( m_Queue.isEmpty() )
         break;

      float dist = m_Queue.getFirstKey();
      PointIndex nextIndex = m_Queue.pop_first();

      if( m_Visited[nextIndex] < 0 )
      {
         visit( nextIndex, dist );
         found = true;
      }
   }
   m_NextPointExists = found;
}

void TopologicalKNNIterator_Fast::visit( PointIndex index,float distance) 
{
   m_CurrentIndex = index;
   m_CurrentTopologicalDistance = distance;
   m_rIt->goToIndex( index );
   m_Buffer = m_PointSet->getDataPointer( index );
   m_Visited[index] = distance;

   Vector3f pos = m_bIt->get3f( m_posAAT );

   // put neighbours on queue
   card32 numNeighbours = m_tIt->getNumNeighbours();
   for( card32 i=0;i<numNeighbours;i++ )
   {
      PointIndex nIndex = m_tIt->getNeighbour(i);
      if( m_Visited[nIndex] < 0 )
      {
         Vector3f nPos = m_PointSet->get3f( nIndex, m_posAAT );
         float nDist = distance + norm(pos - nPos);

         m_Queue.insert( nDist, nIndex );
      }
   }
}


//----------------------------------------------------------------------------------
geoxl::TopologicalRangeQuerry::TopologicalRangeQuerry( PointCloud * pc )
{
	UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if( !upc )
		throw "this querry is suitable for UnstructuredInCorePointCloud only!";

	m_PosAAT = pc->getAAT("position");
	m_hIt = new HierarchicalKNNIterator(pc,32,NULL);
	m_tIt = new TopologicalKNNIterator(pc,
		new FixedKNNTopologyIterator(upc->getPointSet(), upc->getAttachments(), upc, 10));
	m_PointSet = new PointSet;
	VertexDescriptor * vd = (VertexDescriptor*)pc->getDescr()->copy();
	m_PointSet->clearAndSetup(1,0,vd);
	m_NumPoints = 0;
}

void geoxl::TopologicalRangeQuerry::rangeQuerry( const Vector3f &pos, float topologicalRadius )
{
	m_hIt->setSeekPointAndReset(pos);
	card64 index = m_hIt->getCurrentPointIndex();
	m_NumPoints = 0;
	vector<char> buffer(m_PointSet->getDescr()->getSize());

	m_tIt->setStartPointAndReset(index);
	while( !m_tIt->atEnd())
	{
		float currentDistance = m_tIt->getTopologicalDistance();
		if( currentDistance > topologicalRadius )
			break;

		m_NumPoints++;

		if( m_NumPoints >= m_PointSet->getNumEntries() )
		{
			m_PointSet->changeHeight(m_NumPoints * 2);
			m_TopologicalDistances.resize(m_NumPoints * 2);
		}

		m_TopologicalDistances[m_NumPoints-1] = currentDistance;
		m_tIt->getVertex( &buffer[0] );
		m_PointSet->setVertex(m_NumPoints-1, &buffer[0] );

		m_tIt->next(); 
	}

}

geoxl::Vector3f geoxl::TopologicalRangeQuerry::getPosition( unsigned i )
{
	return m_PointSet->get3f(i, m_PosAAT );
}

float geoxl::TopologicalRangeQuerry::getDistance( unsigned i )
{
	return m_TopologicalDistances[i];
}

geoxl::Vector3f geoxl::TopologicalRangeQuerry::getAttribute3f( unsigned i, AAT attribute )
{
	return m_PointSet->get3f(i, attribute );
}

geoxl::TopologicalRangeQuerry::~TopologicalRangeQuerry()
{
	delete m_hIt;
	delete m_tIt;
	delete m_PointSet;
}