#ifndef TopologicalKNNIteratorH
#define TopologicalKNNIteratorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PointCLoudIterators.h"
#include <map>
#include "PointSet.h"
#include "VertexDescriptor.h"
#include "PriorityQueue.h"
//---------------------------------------------------------------------------

namespace geoxl
{
class PointCloud;
class HierarchicalKNNIterator;

/**
	topological k-nearest-neighbour iterator base class
*/
class POINTCLOUDS_API TopologicalKNNIteratorBase : public PointCloudIterator,
                                               public BasicPointCloudIterator
{

protected:

	virtual void visit( PointIndex index, float distance ) = 0;

	PointCloudIterator		*	m_Iter;
	BasicPointCloudIterator	*	m_bIt;
	RandomAccessPointCloudIterator	*	m_rIt;
	TopologyIterator			*	m_tIt;


	PointCloud				*	m_Pc;
	AAT							m_posAAT;

	PointIndex					m_StartIndex;
	PointIndex					m_CurrentIndex;
	PriorityQueue<float,PointIndex>	m_Queue;
	bool							m_NextPointExists;
	float							m_CurrentTopologicalDistance;

	const char				*	m_Buffer;
	const PointSet			*	m_PointSet;

public:


   TopologicalKNNIteratorBase(PointCloud*, PointCloudIterator * topologyIter );
	virtual ~TopologicalKNNIteratorBase();

   /// initialize iterator to start point
   void setStartPointAndReset(const PointIndex &index);

   /// go to starting point
   virtual void reset() = 0;
   /// go to next nearest point
   virtual void next() = 0;
   /// true if end-of-set
   virtual bool atEnd();

   template <class Type, int dim>
   inline StaticVector<Type, dim> genericGet(AAT token) const {
      const char *pos =  m_Buffer + token.getOffset();
      return *((StaticVector<Type, dim>*)(pos));
   }

   template <class Type>
   inline Type genericGetS(const AAT& token) const {
      const char *pos =  m_Buffer + token.getOffset();
      const StaticVector<Type, 1> v = *((StaticVector<Type, 1>*)(pos));
      return v[0];
   }


   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   /// ask VertexDescriptor for buffer size
   virtual void     getVertex(char *buffer) const;

   PointIndex getPointSetPointNumber();

	/// returns approimate topological distance to start point
	float getTopologicalDistance() {return m_CurrentTopologicalDistance;}


   // --- blockwise access NOT useable here!
   virtual bool nextBlock(){pAssert(false); return false;}
   virtual const PointSet * lockBlockRead(){pAssert(false); return NULL;}
   virtual void unlock(){pAssert(false);}
};


/**
topological k-nearest-neighbour iterator
*/
class POINTCLOUDS_API TopologicalKNNIterator : public TopologicalKNNIteratorBase
{

private:

   void visit( PointIndex index, float distance );

   map<PointIndex,float>	         m_Visited;		// holds distances from visited points

public:

   TopologicalKNNIterator(PointCloud*, PointCloudIterator * topologyIter );
   ~TopologicalKNNIterator();

   void reset();
   void next();
};


/**
topological k-nearest-neighbour iterator
*/
class POINTCLOUDS_API TopologicalKNNIterator_Fast : public TopologicalKNNIteratorBase
{

private:

   void visit( PointIndex index, float distance );
   
   vector<float>	         m_Visited;		// holds distances from visited points

public:

   TopologicalKNNIterator_Fast(PointCloud*, PointCloudIterator * topologyIter );
   ~TopologicalKNNIterator_Fast();

   void reset();
   void next();
};


class POINTCLOUDS_API TopologicalRangeQuerry
{
public:

	TopologicalRangeQuerry( PointCloud * pc);
	~TopologicalRangeQuerry();

	void rangeQuerry( const Vector3f &pos, float topologicalRadius );

	unsigned getNumPoints() {return m_NumPoints;}
	Vector3f getPosition( unsigned i);
	float getDistance( unsigned i);
	Vector3f getAttribute3f( unsigned i, AAT attribute);

private:

	TopologicalKNNIterator *	m_tIt;
	HierarchicalKNNIterator *	m_hIt;
	PointSet *	m_PointSet;
	vector<float> m_TopologicalDistances;
	unsigned		m_NumPoints;	// != m_PointSet.getNumElements() (number of actually querried points)
	AAT			m_PosAAT;
};

}

#endif