//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "TriangleMeshTools.h"
#include "AttachedIndexedOctree.h"
#include "PointSetKNNQuery.h"
#include "Mathtools.h"
#include "Mathtools_templCode.h"
#include "PCA.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

/** 
* Snap close vertices (attention naive implementation O(n^2))
*/
void geoxl::snapTriangleMeshMesh( UnstructuredInCoreTriangleMesh* mesh, const float32 maxDist )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "smoothTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) || !vertices->providesAttribute( "color"  ) || !vertices->providesAttribute( "flags") ) {
		debugOutput << "smoothTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT POSITION = vertices->getAAT( "position" );
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT COLOR = vertices->getAAT( "color" );
	AAT FLAGS = vertices->getAAT( "flags" );
	card32 nIndex;
	bool found;
	vector<int32> map; map.resize( vertices->getNumEntries() );
	vector<card32> numMerged; numMerged.resize( vertices->getNumEntries() );
	for( unsigned int n = 0; n < map.size(); n++ ) { map[n] = -1; numMerged[n] = 1; }
	card32 numDiscarded = 0;

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( vertices );
	PointSetKNNQuery knn( vertices, octree, NULL_VECTOR3F, POSITION );	
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		if ( map[pointIndex] < 0 ){
			Vector3f pos = vertices->get3f( pointIndex, POSITION );
			knn.setTargetPoint( pos );
			knn.getNextNearestPointIndex( nIndex, found );
			float32 d = norm( vertices->get3f( nIndex, POSITION ) - pos );;
			while( found && ( d < maxDist ) ){
				if( ( map[nIndex] < 0 ) && ( nIndex > pointIndex ) ){
					map[nIndex] = pointIndex;
					vertices->set3f( pointIndex, POSITION, vertices->get3f( pointIndex, POSITION ) + vertices->get3f( nIndex, POSITION ) );
					numMerged[pointIndex]++;
					numDiscarded++;
				}
				knn.getNextNearestPointIndex( nIndex, found );
				d = norm( vertices->get3f( nIndex, POSITION ) - pos );
			}
		}
	} 	
	delete octree;

	// Create new vertex point set
	PointSet* newVertices = new PointSet();
	card32 numNewVertices = vertices->getNumEntries() - numDiscarded;
	vector<int32> deletionMap; deletionMap.resize( vertices->getNumEntries() );
	newVertices->clearAndSetup( 1, numNewVertices, vertices->getDescr() );
	card32 newPointIndex = 0;
	for ( card32 pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		if ( map[pointIndex] < 0 ){
			// Copy
			Vector3f newPos = vertices->get3f( pointIndex, POSITION ) / float32( numMerged[pointIndex] );
			newVertices->set3f( newPointIndex, POSITION, newPos );
			newVertices->set3f( newPointIndex, NORMAL, vertices->get3f( pointIndex, NORMAL ) );
			newVertices->set3f( newPointIndex, COLOR, vertices->get3f( pointIndex, COLOR ) );
			newVertices->set1i( newPointIndex, FLAGS, vertices->get1i( pointIndex, FLAGS ) );
			deletionMap[pointIndex] = newPointIndex;
			newPointIndex++;
		} else {
			deletionMap[pointIndex] = -1;
		}
	}

	// Update triangle point set
	vector<mpcard> flippedTriangles; flippedTriangles.clear();
	vector<Vector3i> newTriangleContainer; newTriangleContainer.clear();
	PointSet* triangles = mesh->getTriangles();
	AAT TRIANGLE = triangles->getAAT( "index" );
	for ( unsigned int triangleIndex = 0; triangleIndex < triangles->getNumEntries(); triangleIndex++ ){
		Vector3i t = triangles->get<int32,3>( triangleIndex, TRIANGLE );
		Vector3f oldNormal = ( vertices->get3f( t[2], POSITION ) - vertices->get3f( t[0], POSITION ) ). crossProduct( vertices->get3f( t[1], POSITION ) - vertices->get3f( t[0], POSITION ) );
		// Mapping of discarded points
		if ( map[t[0]] >= 0 ) t[0] = map[t[0]];
		if ( map[t[1]] >= 0 ) t[1] = map[t[1]];
		if ( map[t[2]] >= 0 ) t[2] = map[t[2]];
		// Mapping of points moved due to deletion of other points
		if ( deletionMap[t[0]] >= 0 ) t[0] = deletionMap[t[0]];
		if ( deletionMap[t[1]] >= 0 ) t[1] = deletionMap[t[1]];
		if ( deletionMap[t[2]] >= 0 ) t[2] = deletionMap[t[2]];

		// Check if still valid
		if ( ( t[0] != t[1] ) && ( t[0] != t[2] ) && ( t[1] != t[2] ) ){
			// Check if triangle is valid
			int32 newNumVertices = int32( newVertices->getNumEntries() );
			if( ( t[0] >= newNumVertices ) || ( t[1] >= newNumVertices ) || ( t[2] >= newNumVertices ) ){
				debugOutput << "index too large.\n"; debugOutput.flush();
			} else if ( (t[0] < 0) || (t[1] < 0) || (t[2] < 0) ){
				debugOutput << "negative index.\n"; debugOutput.flush();
			} else {
				Vector3f newNormal = ( newVertices->get3f( t[2], POSITION ) - newVertices->get3f( t[0], POSITION ) ). crossProduct( newVertices->get3f( t[1], POSITION ) - newVertices->get3f( t[0], POSITION ) );
				if ( newNormal * oldNormal < 0 ){
					flippedTriangles.push_back( newTriangleContainer.size() );
				}
				newTriangleContainer.push_back( t );
			}
		}
	}
	PointSet* newTriangles = new PointSet();
	newTriangles->clearAndSetup( 1, newTriangleContainer.size(), triangles->getDescr() );
	for ( unsigned int triangleIndex = 0; triangleIndex < newTriangleContainer.size(); triangleIndex++ ){
		newTriangles->set<int32,3>( triangleIndex, TRIANGLE, newTriangleContainer[triangleIndex] );
	}

	debugOutput << "#vertices: " << vertices->getNumEntries() << "->" << newVertices->getNumEntries() << "\n"; debugOutput.flush();
	debugOutput << "#triangles: " << triangles->getNumEntries() << "->" << newTriangles->getNumEntries() << "\n"; debugOutput.flush();
	//debugOutput << "flipped triangles: " << flippedTriangles.size() << "\n"; debugOutput.flush();

	mesh->setPointSet( newVertices );
	mesh->setTriangles( newTriangles );
}

/** 
* Smooth out normals
*/
void geoxl::smoothTriangleMeshNormals( UnstructuredInCoreTriangleMesh* mesh, float32 sigma )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "smoothTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	card32 numNormalSmoothings = 1;
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "smoothTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT POSITION = vertices->getAAT( "position" );
	AAT NORMAL = vertices->getAAT( "normal" );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( vertices );
	PointSetKNNQuery knn( vertices, octree, NULL_VECTOR3F, POSITION );	
	card32 nIndex;
	bool found;
	for ( unsigned int iteration = 0; iteration < numNormalSmoothings; iteration++ ){
		for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
			Vector3f pos = vertices->get3f( pointIndex, POSITION );
			knn.setTargetPoint( pos );
			// Skip self
			knn.getNextNearestPointIndex( nIndex, found );
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = vertices->get3f( nIndex, POSITION );
			Vector3f avgNormal = makeVector3f( 0,0,0 );
			float32 d = norm( pos - nPos );
			card32 numNeighbors = 0;
			float32 summedWeight = 0;
			card32 numFound = 0;
			while( ( numFound < 1 ) || ( d < sigma ) ){
				float32 w = exp( - sqr( d ) / sqr( sigma ) );
				summedWeight += w;
				avgNormal += vertices->get3f( nIndex, NORMAL ) * w;
				numNeighbors++;

				knn.getNextNearestPointIndex( nIndex, found );
				nPos = vertices->get3f( nIndex, POSITION );
				d = norm( pos - nPos );
				numFound++;
			}
			avgNormal /= summedWeight;
			vertices->set3f( pointIndex, NORMAL, normalize( avgNormal ) );
		}
	}
	delete octree;
}

/** 
* Compute normals for the mesh
*/
void geoxl::computeTriangleMeshNormals( UnstructuredInCoreTriangleMesh* mesh )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "computeTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* triangles = mesh->getTriangles();
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "computeTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	if ( !triangles->providesAttribute( "index" ) ){
		debugOutput << "computeTriangleMeshNormals: triangle attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT POSITION = vertices->getAAT( "position" );
	AAT TRIANGLE_INDEX = triangles->getAAT( "index" );
	vector<Vector3f> newNormals; newNormals.resize( vertices->getNumEntries() );
	vector<card32> numNewNormals; numNewNormals.resize( vertices->getNumEntries() );

	// Init container
	for ( unsigned int normalIndex = 0; normalIndex < vertices->getNumEntries(); normalIndex++ ){
		newNormals[normalIndex] = makeVector3f( 0,0,0 );
		numNewNormals[normalIndex] = 0;
	}

	// Compute normals based on triangles
	for ( unsigned int triangleIndex = 0; triangleIndex < mesh->getNumTriangles(); triangleIndex++ ){
		Vector3i t = triangles->get<int32,3>( triangleIndex, TRIANGLE_INDEX );
		Vector3f v0 = vertices->get3f( t[0], POSITION );
		Vector3f v1 = vertices->get3f( t[1], POSITION );
		Vector3f v2 = vertices->get3f( t[2], POSITION );
		Vector3f tNormal = ( v2 - v0 ).crossProduct( v1 - v0 );
		float32 n = norm( tNormal );
		if ( n > 1e-7 ) {
			tNormal /= n;
			for ( unsigned int vIndex = 0; vIndex < 3; vIndex++ ){
				newNormals[ t[vIndex] ] += tNormal;
				numNewNormals[ t[vIndex] ]++;
			}
		}
	}

	// Set newly calculated normals
	for ( unsigned int normalIndex = 0; normalIndex < vertices->getNumEntries(); normalIndex++ ){
		if ( numNewNormals[normalIndex] > 0 ){
			vertices->set3f( normalIndex, NORMAL, -newNormals[normalIndex] / float32( numNewNormals[normalIndex] ) );
			//debugOutput << numNewNormals[normalIndex] << "\n"; debugOutput.flush();
		} else {
			vertices->set3f( normalIndex, NORMAL, makeVector3f( 1,0,0 ) );
		}
	}
}

Vector2f toEuler( Vector3f p )
{
	Vector2f coord;
	p = normalize( p );
	Vector3f x = makeVector3f( 1,0,0 );
	Vector3f y = makeVector3f( 0,1,0 );
	Vector3f z = makeVector3f( 0,0,1 );
	coord[0] = acos( y * p );
	// Project p into xz plane
	p[1] = 0; 
	float32 n = norm( p );
	if ( n < 1e-9 ){
		coord[1] = float32( M_PI );
	} else {
		p /= n;
		coord[1] = acos( x * p );
		if ( p * z < 0 ){
			//coord[1] = - coord[1];
		}
		coord[1] += float32( M_PI );
	}

	return coord;
}

/** 
 * Parameterize a triangle mesh
 */
void geoxl::parameterizeTriangleMesh( 
	UnstructuredInCoreTriangleMesh* mesh, 
	const card32 mode, 
	const float32 textureRepeat, 
	bool pcaROT, 
	card32 coordIndex0, 
	card32 coordIndex1 )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "computeTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "texCoord" ) ){
		debugOutput << "parameterizeTriangleMesh: vertices attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT TEX_COORD = vertices->getAAT( "texCoord" );
	AAT POSITION = vertices->getAAT( "position" );

	// Compute center of object
	PCA<float32,3> pca; pca.clear();
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		pca.addPoint( pos );
	}
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	pca.analyze( eigenValues, eigenVectors, centroid );

	Vector2f u = makeVector2f( 10000, -100000 );
	Vector2f v = makeVector2f( 10000, -100000 );
	Matrix3f T = eigenVectors.transpose();
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		// Projection to plane
		Vector3f proj = T * ( pos - centroid );
		// u parameters
		if ( proj[0] < u[0] ) u[0] = proj[0];	// min
		if ( proj[0] > u[1] ) u[1] = proj[0];	// max
		// v parameters
		if ( proj[1] < v[0] ) v[0] = proj[1];	// min
		if ( proj[1] > v[1] ) v[1] = proj[1];	// max
	}
	float32 scalePlane = max( u[1] - u[0], v[1] - v[0] );
	Vector2f tPlane = makeVector2f( u[0], v[0] );

	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		Vector2f texCoord = makeVector2f( 0,0 );

		if ( mode == MESH_PARAMETERIZATION_MODE_SPHERE ){
			texCoord = toEuler( pos - centroid );
			texCoord[0] *= float32( textureRepeat / M_PI );
			texCoord[1] *= float32( textureRepeat / ( 2.0f * M_PI ) );
		} else if ( mode == MESH_PARAMETERIZATION_MODE_PLANE ){
			Vector3f p3D = (pcaROT?T:IDENTITY3F) * ( pos - centroid );
			Vector2f p2D = makeVector2f( p3D[coordIndex0], p3D[coordIndex1] );
			texCoord = ( p2D - tPlane ) * scalePlane * textureRepeat;
		} else if ( mode == MESH_PARAMETERIZATION_MODE_CYLINDER ){
			// Direction = eigenVectors[0]
			pos = normalize( pos - centroid );
			texCoord = toEuler( pos - centroid );
			texCoord[0] = ( pos * eigenVectors[0] ) * textureRepeat;
		}
		
		vertices->set2f( pointIndex, TEX_COORD, texCoord );
	}
}