//---------------------------------------------------------------------------
#ifndef VertexArrayH
#define VertexArrayH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class PointSet;
class PointCloud;
class SceneObjectIterator;

/**
	provides easier access common vertex-formats
	special methods:
		get / setPosition3f
		get / setNormal3f
		get / setColor3f
		get / setFlags
      get / setIndex
	fell free to add other common attributes :-)
	you may also use the templated get/set methods
	Example:	
		for( card32 i=0;i<... )
		{
			Vector3f pos = vertexarray.get<float,3>(i, "position" );	// slower !
			set( i, COLOR, pos );
		}

*/
class POINTCLOUDS_API VertexArray
{
private:

   void initVD( const VertexDescriptor * desc );
   bool initAAT( AAT &aat, string name, card32 format, card32 numComponents );

	VertexDescriptor	*	m_VD;
	card8				*	m_Buffer;
	card32					m_NumElements;

	AAT						POSITION, COLOR, NORMAL, FLAGS, INDEX, TYPEID, TYPEFLAG;
	bool					m_HasPosition3f,
							m_HasColor3f,
							m_HasNormal3f,
							m_HasFlags,
                     m_HasIndex3i,
                     m_HasTypeID,
                     m_HasTypeFlag;
	SceneObjectIterator * m_Iter;

public:
	VertexArray( card8 * buffer, card32 numElements, const VertexDescriptor * desc);
	VertexArray( PointSet * ps );
	VertexArray( PointCloud * pc);
	~VertexArray(void);

	void setNumElements(card32 numElements ) {m_NumElements = numElements;}
	card32 getNumElements() {return m_NumElements;}

	void setBuffer(card8 *buffer){ m_Buffer = buffer; }
	card8 * getBuffer(){ return m_Buffer; }

   card32 getBufferSize() const;

   VertexDescriptor * getVertexDescriptor() const {return m_VD;}

// --------------------------------------------------------
//                complete vertex access methods

   /// vertex[destIndex] = vertex[sourceIndex]
   void copyVertex( card32 sourceIndex, card32 destIndex );

   /// copies all vertices from source to this array
   /// vertexdescriptor and num elements must be equal !
   void copyVertices( const VertexArray &source);

   /// copy vertex from sourceArray[sourceIndex] to this[destIndex]
   /// vertexdescriptor must be identical
   void copyVertex( const VertexArray &sourceArray, card32 sourceIndex,
      card32 destIndex );

// --------------------------------------------------------
//                attribute access methods
	/// vertices[index].attr = vec
	template<class DataType, int dim> 
	inline void set( card32 index, const string &attr, const StaticVector<DataType,dim> &vec )
	{
		AAT aat = m_VD->getToken( attr );
		set( index, aat, vec );
	}

	template<class DataType, int dim>
	inline void set( card32 index, AAT token, const StaticVector<DataType,dim> &vec )
	{
		pAssert( index < m_NumElements );
        StaticVector<DataType,dim> * ptr = (StaticVector<DataType,dim>*)&m_Buffer[index*m_VD->getSize() + token.getOffset()];
		*ptr = vec;
	}

	template<class DataType, int dim>
	inline StaticVector<DataType,dim> get( card32 index, const string &attr ) const
	{
		AAT aat = m_VD->getToken( attr );
		return get( index, aat );
	}

	template<class DataType, int dim>
	inline StaticVector<DataType,dim> get( card32 index, AAT token ) const
	{
		pAssert( index < m_NumElements );
        StaticVector<DataType,dim> * ptr = (StaticVector<DataType,dim>*)&m_Buffer[index*m_VD->getSize() + token.getOffset()];
		return *ptr;
	}

	inline Vector3f getPosition3f( card32 index ) const
	{
		pAssert( m_HasPosition3f );
		return get<float,3>( index, POSITION );
	}

	inline void setPosition3f( card32 index, const Vector3f &vec )
	{
		pAssert( m_HasPosition3f );
		set<float32, 3>( index, POSITION, vec );
	}

	inline Vector3f getColor3f( card32 index ) const
	{
		pAssert( m_HasColor3f );
		return get<float,3>( index, COLOR );
	}

	inline void setColor3f( card32 index, const Vector3f &vec )
	{
		pAssert( m_HasColor3f );
		set<float32, 3>( index, COLOR, vec );
	}

	inline Vector3f getNormal3f( card32 index ) const
	{
		pAssert( m_HasNormal3f );
		return get<float,3>( index, NORMAL );
	}

	inline void setNormal3f( card32 index, const Vector3f &vec )
	{
		pAssert( m_HasNormal3f );
		set<float32, 3>( index, NORMAL, vec );
	}

	inline int32 getFlags( card32 index ) const
	{
		pAssert( m_HasFlags );
		return get<int32,1>( index, FLAGS )[0];
	}

	inline void setFlags( card32 index, const int32 &f )
	{
		pAssert( m_HasFlags );
      StaticVector<int32,1> tmp;
      tmp[0] = f;
		set<int32, 1>( index, FLAGS, tmp );
	}

	inline Vector3i getIndex3i( card32 index ) const
	{
		pAssert( m_HasIndex3i );
		return get<int32,3>( index, INDEX );
	}

	inline void setIndex3i( card32 index, const Vector3i &vec )
	{
		pAssert( m_HasIndex3i );
		set<int32, 3>( index, INDEX, vec );
	}

   inline int32 getTypeID( card32 index) const
   {
      pAssert( m_HasTypeID );
      return get<int32,1>( index, TYPEID )[0];
   }

   inline void setTypeID( card32 index, int32 typeID )
   {
      pAssert( m_HasTypeID );
      StaticVector<int32,1> vec;
      vec[0] = typeID;
      set<int32,1>( index, TYPEID, vec );
   }

   inline int32 getTypeFlag( card32 index) const
   {
      pAssert( m_HasTypeFlag );
      return get<int32,1>( index, TYPEFLAG )[0];
   }

   inline void setTypeFlag( card32 index, int32 typeFlag )
   {
      pAssert( m_HasTypeFlag );
      StaticVector<int32,1> vec;
      vec[0] = typeFlag;
      set<int32,1>( index, TYPEFLAG, vec );
   }

   char * getAttributePointer( card32 index, string attr )
   {
	   return (char*)m_Buffer + m_VD->getAttributeOffset( m_VD->getIndex(attr) );
   }

};


}

#endif