	//---------------------------------------------------------------------------
#ifndef VertexDescriptorH
#define VertexDescriptorH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PAssert.h"
#include "LinearAlgebra.h"
#include "VertexDescriptor.h"
#include "PointFlags.h"
//---------------------------------------------------------------------------
#include <string>
//---------------------------------------------------------------------------

namespace geoxl {

class PointSet;
class VertexAttribDescriptor;
class VertexDescriptor;
class PointCloudIterator;


class POINTCLOUDS_API AttributeAccessToken {
 private:                                                     
   card32 offset;
   #ifdef _DEBUG
   card32 size;
   #endif
 protected:
   AttributeAccessToken(card32 size, card32 offset) {
      #ifdef _DEBUG
      this->size = size;
      #endif
      this->offset = offset;
   }

 public:
   AttributeAccessToken() {
      #ifdef _DEBUG
      size = 0xFFFFFFFF; // invalid ("NULL_AAT") token by default
      #endif
      offset = 0xFFFFFFFF; // invalid ("NULL_AAT") token by default
   }

   bool operator==(const AttributeAccessToken &aat) const {
      bool result = (offset == aat.offset);
      #ifdef _DEBUG
         result &= (size == aat.size);
      #endif
      return result;
   }
   bool operator!=(const AttributeAccessToken &aat) const {
      bool result = (offset != aat.offset);
      return result;
   }

   inline card32 getOffset() const {return offset;}

   template<class Format, int dim>
   inline StaticVector<Format, dim> getAttrib(const char *data) const {
      #ifdef _DEBUG
      pAssert(sizeof(StaticVector<Format, dim>) == size);
      #endif
      const char *pos = data + offset;
      return *((StaticVector<Format, dim>*)(pos));
   }
   template<class Format, int dim>
   inline void setAttrib(char *data, const StaticVector<Format, dim> &v) {
      #ifdef _DEBUG
      pAssert(sizeof(StaticVector<Format, dim>) == size);
      #endif
      char *pos = data + offset;
      *((StaticVector<Format, dim>*)(pos)) = v;
   }


   static AttributeAccessToken makeInvalidToken();

 friend PointSet;
 friend VertexAttribDescriptor;
 friend VertexDescriptor;
 friend PointCloudIterator;
};

typedef AttributeAccessToken AAT;

/// undefined AAT (e.g. to mark non-present attribute channels)
extern POINTCLOUDS_API const AttributeAccessToken NULL_AAT;

class POINTCLOUDS_API VertexAttribDescriptor {
 private:
	template <typename T>
	inline void incPtr(const char* &v) {
		const T* temp = (T*)v;
		temp++;
		v = (const char*)temp;
	}
	template <typename T>
	inline void incPtr(char* &v) {
		T* temp = (T*)v;
		temp++;
		v = (char*)temp;
	}

 protected:
   card32 getAttribElementSize() const;

 public:
   string name;
   card32 numComponents;
   card32 dataFormat;

	// first 5 bits = number of bytes
	// 16 = signed int
	// 32 = unsigned int
	// card8 is a special case (implementation relict)
	static const card32 DATA_FORMAT_CARD8 = 1;
	static const card32 DATA_FORMAT_CARD16 = 32 | 2;
	static const card32 DATA_FORMAT_CARD32 = 32 | 4;
	static const card32 DATA_FORMAT_CARD64 = 32 | 8;
   static const card32 DATA_FORMAT_FLOAT32 = 4;
	static const card32 DATA_FORMAT_FLOAT64 = 8;
	static const card32 DATA_FORMAT_INT8 = 16 | 1;
	static const card32 DATA_FORMAT_INT16 = 16 | 2;
	static const card32 DATA_FORMAT_INT32 = 16 | 4;
	static const card32 DATA_FORMAT_INT64 = 16 | 8;

   VertexAttribDescriptor();

   inline VertexAttribDescriptor(const VertexAttribDescriptor &other) {
      name = other.name;
      numComponents = other.numComponents;
      dataFormat = other.dataFormat;
   }

   inline void operator=(const VertexAttribDescriptor &other) {
      name = other.name;
      numComponents = other.numComponents;
      dataFormat = other.dataFormat;
   }

   inline bool operator==(const VertexAttribDescriptor &other) {
      return    name == other.name
             && numComponents == other.numComponents
             && dataFormat == other.dataFormat;
   }

   VertexAttribDescriptor(string name, card32 numComponents, card32 dataFormat);
   card32 getSize() const;
   string getDescription() const;
   void write(OutputObjectStream *s) const;
   void read(InputObjectStream *s);

   template<class Format, int dim>
   inline StaticVector<Format, dim> getAttrib(const card8 *data, AAT offset) const {
      #ifdef _DEBUG
      pAssert(sizeof(StaticVector<Format, dim>) == getSize());
      #endif
      return *((StaticVector<Format, dim>*)(data + offset.getOffset()));
   }
   template<class Format, int dim>
   inline void setAttrib(card8 *data, const StaticVector<Format, dim> &v, AAT offset) const {
      #ifdef _DEBUG
      pAssert(sizeof(StaticVector<Format, dim>) == getSize());
      #endif
      *((StaticVector<Format, dim>*)(data + offset.getOffset())) = v;
   }
   inline Vector4d getAnyAttrib(const card8 *data, AAT offset) const {
      Vector4d result = NULL_VECTOR4D;
      const card8 *pos = data + offset.getOffset();
      for (card32 i=0; i<numComponents; i++) {
         if (dataFormat == DATA_FORMAT_CARD8) {
            result[i] = *pos; ++pos;
         } else if (dataFormat == DATA_FORMAT_FLOAT32) {
            result[i] = *((float32*)pos); pos += sizeof(float32);
         } else {
            result[i] = *((int32*)pos); pos += sizeof(int32);
         }
      }
      return result;
   }
   inline void setAnyAttrib(card8 *data, Vector4d values, AAT offset) const {
      card8 *pos = data + offset.getOffset();
      for (card32 i=0; i<numComponents; i++) {
         if (dataFormat == DATA_FORMAT_CARD8) {
            *pos = (card8)(values[i]); ++pos;
         } else if (dataFormat == DATA_FORMAT_FLOAT32) {
            *((float32*)pos) = (float32)values[i]; pos += sizeof(float32);
         } else {
            *((int32*)pos) = (int32)(values[i]); pos += sizeof(int32);
         }
      }
   }
   /// write the number zero at an attribute, the location is "mem"
   void zero(char *mem) {
      for (card32 i=0; i<numComponents; i++) {
         switch (dataFormat) {
            case DATA_FORMAT_CARD8:   *((card8*)mem)   = 0;    incPtr<card8>(mem);   break;
            case DATA_FORMAT_FLOAT32: *((float32*)mem) = 0.0f; incPtr<float32>(mem); break;
            case DATA_FORMAT_INT32:   *((int32*)mem)   = 0;    incPtr<int32>(mem);   break;
         }
      }
   }
   void add(char *dest, const char *op, float32 scale = 1.0f) {
      for (card32 i=0; i<numComponents; i++) {
         switch (dataFormat) {
            case DATA_FORMAT_CARD8:   *((card8*)dest)   += (card8)(*((card8*)op) * scale);   
					                       incPtr<card8>(dest); incPtr<card8>(op); break; 
            case DATA_FORMAT_FLOAT32: *((float32*)dest) += (float32)(*((float32*)op) * scale); 
					                       incPtr<float32>(dest); incPtr<float32>(op); break;
            case DATA_FORMAT_INT32:   *((int32*)dest)   += (int32)(*((int32*)op) * scale);   
												  incPtr<int32>(dest); incPtr<int32>(op); break;
         }
      }
   }
   void scale(char *dest, float32 scale) {
      for (card32 i=0; i<numComponents; i++) {
         switch (dataFormat) {
            case DATA_FORMAT_CARD8:   *((card8*)dest)   = (card8)(*((card8*)dest)   * scale);   incPtr<card8>(dest);   break;
            case DATA_FORMAT_FLOAT32: *((float32*)dest) = (float32)(*((float32*)dest) * scale); incPtr<float32>(dest); break;
            case DATA_FORMAT_INT32:   *((int32*)dest)   = (int32)(*((int32*)dest)   * scale);   incPtr<int32>(dest);   break;
         }
      }
   }

};

inline bool operator==(const VertexAttribDescriptor &v, const VertexAttribDescriptor &other) {
   return v.name == other.name && v.numComponents == other.numComponents && v.dataFormat == other.dataFormat;
}

inline VertexAttribDescriptor mVAD(string name, card32 numComponents, card32 dataFormat) {
   return VertexAttribDescriptor(name, numComponents, dataFormat);
}

typedef VertexAttribDescriptor VAD;




class POINTCLOUDS_API VertexDescriptor : public Persistent {
 private:
   static MetaClass *VertexDescriptorMetaClass;
   static Object* newInstance();

   vector<VertexAttribDescriptor> attributes;
   /// offset in memory to element i (contains attributes.size() + 1 offsets, first element always == 0)
   vector<card32> offsets;

   inline int32 convertCard8ToInt32( const card8 * ptr ) const;
   inline int32 convertFloat32ToInt32( const card8 * ptr ) const;

   inline card8 convertInt32ToCard8( const card8 * ptr ) const;
   inline card8 convertFloat32ToCard8( const card8 * ptr ) const;

   inline float32 convertCard8ToFloat32( const card8 * ptr ) const;
   inline float32 convertInt32ToFloat32( const card8 * ptr ) const;

 public:
   VertexDescriptor();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual bool equal(const Object* obj) const;
   bool operator==(VertexDescriptor *vd) const;
   bool operator!=(VertexDescriptor *vd) const;

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   void   pushAttrib(VertexAttribDescriptor a);
   void   clear();
   mpcard getNumAttribs() const;
   mpcard getNumAttributes() const;
   VertexAttribDescriptor getAttribute(card32 i) const;
   VertexAttribDescriptor getAttribute(string name) const {return getAttribute(getIndex(name));}
   inline const VertexAttribDescriptor operator[](card32 i) const {return getAttribute(i);}
   inline const VertexAttribDescriptor operator[](const char* name) const {return getAttribute(getIndex(name));}
   card32 getIndex(string name) const;
   bool   providesAttribute(string name) const;
   bool   providesAttribute(string name, card32 numComponents) const;
   bool   providesAttribute(string name, card32 numComponents, card32 dataFormat) const;
   bool   providesAttribute(VAD att) const;
   /// returns size of a single vertex in memory (in bytes)
   card32 getSize() const;
   string getDescription() const;

   /// incorporate new attributes into current descriptor,
   /// duplicate fields (field with the same name) are avoided,
   /// an exception is thrown if type of duplicate fields does not match
   void mergeAttributes(const VertexDescriptor *newAttribs);

   AttributeAccessToken getToken(string name) const;
	/// gets aat with given name; also checks for other attributes and throws exception if they do not match
   AttributeAccessToken getToken(string name, card32 numComponents, card32 dataFormat) const;
   AttributeAccessToken getToken(card32 i) const;
   inline card32 getAttributeOffset(card32 i) const {return getToken(i).getOffset();}


   /// all convert routines try to convert an incoming VD "source" into VD this
   /// check if one ore more attribute name exists in both descriptors
   bool isCompatible( const VertexDescriptor &source) const;

   /// check if all attribute channels of this can be filled with attributes of
   /// descriptor "source"
   bool isFullCompatible( const VertexDescriptor &source) const;
   
   /// convertes vertices assuming that attributes have been added at the end ONLY
   static void appendAttribs(const card8 *source, const VertexDescriptor *sourceVD, card8 *dest, const VertexDescriptor *destVD);

   // /// convertes between arbitrary vertex formats - not yet implemented
   // static void convertAttributes(const card8 *source, VertexDescriptor *sourceVD, card8 *dest, VertexDescriptor *destVD);

   /// convert source_vertices with format source to format this
   /// channels are identified with attribute names
   /// non-existing source attributes are filled with zeros
   /// non-existing dest attributes are ignored
   /// result is returned in ret_vertices
   /// vertices with excludeFlags set or includeFlags not set are ignored (default = ignore invalid-flag)
   void convertVertices( const char * source_vertices, mpcard numVertices, const VertexDescriptor &sourceVD,
                         vector<card8> &ret_vertices, const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID ) const;
   void convertVertices( const vector<card8> &source_vertices, const VertexDescriptor &sourceVD,
                         vector<card8> &ret_vertices, const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID) const;
   void convertVertices( const PointSet * ps, vector<card8> &ret_vertices,
                         const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID ) const;

   /// copy every vertex of source_vertices n times in ret_vertices
   /// source_vertices'format have to be *this
   void expandVertices( const card8 * source_vertices, mpcard numVertices, card32 n, vector<card8> &ret_vertices ) const;
   void expandVertices( const vector<card8> &source_vertices, card32 n, vector<card8> &ret_vertices ) const;

   static const card32 MODIFIER_SET = 1;  /// vertex.attr[channel] = param
   static const card32 MODIFIER_ADD = 2;  /// vertex.attr[channel] += param
   static const card32 MODIFIER_MULT = 3; /// vertex.attr[channel] *= param
   static const card32 MODIFIER_INVERT = 4; // vertex.attr[channel] = -vertex.attr[channel]
   /// modifies channel of every Vertex[(i % period)==phase] with op(vertex, param )
   /// float32-types only
   /// ops: MODIFIER_SET, MODIFIER_ADD, MODIFIER_MULT, MODIFIER_INVERT
   void modifyVerticesf( card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const string &channel, card32 op, float param ) const;
   void modifyVertices4fv( card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const string &channel, card32 op, float * param ) const;

   /// modifes dest_channel of every Vertex[(i % period)==phase] with op(vertex.dest, vertex.src * param)
   /// float32-types only
   /// number of components have to be equal in dest_channel & src_channel
   /// ops: SET   vertex.attr[dest_channel] = vertex.attr[src_channel] * param
   /// ops: ADD   vertex.attr[dest_channel] += vertex.attr[src_channel] * param
   /// ops: MULT   vertex.attr[dest_channel] *= vertex.attr[src_channel] * param
   /// ops: INVERT   vertex.attr[dest_channel] = -vertex.attr[dest_channel]
   void modifyVerticesf( card8* dest_vertices, mpcard numVertices, card32 period, card32 phase,
            const string &dest_channel, card32 op,const string &src_channel, float param ) const;
   void modifyVertices4fv( card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const string &dest_channel, card32 op,const string &src_channel, float * param ) const;

   /// builds tangential quads with tangents from source_tangents
   /// source_vertices points to an array of vertices with this format
   /// source_tangents must have "tangent_u" & "tangent_v" channel! ( float32 both )
   /// source_vertices must have "position" channel (float32)
   void buildTangentialQuads( const char * source_vertices, const char * source_tangents, card32 numVertices,
         const VertexDescriptor &sourceTangentVD, vector<card8> &ret_vertices, float scale=1.0f ) const;
   void buildTangentialQuads( PointSet * ps, vector<card8> & ret_vertices, float scale=1.0f ) const;

   /// builds splatting quads with tangents from source_tangents
   /// this VD and sourceVD must have "tangent_u", "tangent_v", all 3 - float32
   /// additionally this VD must have a float32 "pcoord" attribute with at least 2 components( parameter coords )
   void buildSplattingQuads( const char * source_vertices, card32 numVertices,
      const VertexDescriptor * sourceVD, vector<card8> &ret_vertices,
      const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID ) const;
   void buildSplattingQuads( const PointSet * ps, vector<card8> & ret_vertices,
      const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID ) const;

   /// attribute interpolation:
   /// @param source: source vertex data, pointer to array of pointers to the vertices
   /// @param numVertices: number of source vertices
   /// @param dest: destination vertex data, simple pointer to some allocated space
   void interpolate(const char **source, float32 *factors, card32 numVertices, char *dest) const; 
   
   /// sets all values to zero (numerically, not zero-memory)
   void zero(char *data) const;
};



}

#endif
