#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
#include "Pointset.h"
#include "VertexArray.h"
#include "RenderingTool.h"
#include "DebugOutput.h"
#include "linearalgebra_templcode.h"
#include "MathTools.h"
#include "scene.h"
#include "SceneGraphState.h"
#include "Camera.h"
//---------------------------------------------------------------------------

using namespace geoxl;
DebugRenderer * geoxl::debugRenderer = new DebugRenderer;
static Vector3f DebugRenderer_CurrentCameraPosition;

DebugRenderer::DebugRenderer(void)
{
	m_CurrentRenderJobPtr = NULL;
	m_CurrentRenderJob = "";

}

DebugRenderer::~DebugRenderer(void)
{
	clearDebugData();
}

void DebugRenderer::clearDebugData( const std::string& id) 
{
	if( id != "" )
	for( card32 i=0;i<m_RenderJobs.size();i++ )
	{
		if( m_CurrentRenderJob == id)
		{
			endRenderJob();
		}
		if( m_RenderJobs[i]->id == id)
		{
			delete m_RenderJobs[i];
			m_RenderJobs.erase( m_RenderJobs.begin() + i );
			break;
		}
	}
	else
	{
		for( card32 i=0;i<m_RenderJobs.size();i++ )
		{
			delete m_RenderJobs[i];
		}
		m_RenderJobs.clear();
		m_CurrentRenderJob = "";
		m_CurrentRenderJobPtr = NULL;
		m_RenderJobStack.clear();
	}
}

void DebugRenderer::beginRenderJob( string id, bool active, int time ) 
{
	if( m_CurrentRenderJobPtr != NULL )
	{
		m_RenderJobStack.push_back(m_CurrentRenderJob);
	}
	for( card32 i=0;i<m_RenderJobs.size();i++ )
	{
		if( m_RenderJobs[i]->id == id)
		{
			m_CurrentRenderJob = id;
			m_CurrentRenderJobPtr = m_RenderJobs[i];
			m_CurrentRenderJobPtr->active = active;
			return;
		}
	}
	
	TDebugRenderJob * job = new TDebugRenderJob;
	job->id = id;
	job->active = active;
	job->time = time;
	m_RenderJobs.push_back(job);
	m_CurrentRenderJob = id;
	m_CurrentRenderJobPtr = job;

	// create triangle mesh
	VertexDescriptor vd;
	vd.pushAttrib( VAD( "position",3,VAD::DATA_FORMAT_FLOAT32) );
	vd.pushAttrib( VAD( "normal",3,VAD::DATA_FORMAT_FLOAT32) );
	vd.pushAttrib( VAD( "color",3,VAD::DATA_FORMAT_FLOAT32) );
	job->mesh = new UnstructuredInCoreTriangleMesh;
	job->mesh->clearAndSetup(&vd, 0);

	// create surfel point cloud
	job->surfelCloud = new UnstructuredInCorePointCloud;
	job->surfelCloud->clearAndSetup(&vd, 0);
	
	// setup default matrix
	job->objectMatrix.setIdentity();
	
	// setup default p[roperties
	job->enableBlending = false;
	job->smoothLines = false;
	job->pointSize = 4.0f;
}

void DebugRenderer::endRenderJob() 
{
	if( m_RenderJobStack.size() > 0)
	{
		string tmpRenderJob = m_RenderJobStack.back();
		m_RenderJobStack.pop_back();
		m_CurrentRenderJob = "";
		m_CurrentRenderJobPtr = NULL;
		beginRenderJob(tmpRenderJob);
	}
	else
	{
		m_CurrentRenderJob = "";
		m_CurrentRenderJobPtr = NULL;
	}
}

void DebugRenderer::addTransformationMatrix(const Matrix4f& mat)
{
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->objectMatrix = m_CurrentRenderJobPtr->objectMatrix * mat;
   }
}

void DebugRenderer::setTransformationMatrix(const Matrix4f& mat)
{
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->objectMatrix = mat;
   }
}

void DebugRenderer::setSmoothLine(bool s)
{
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->smoothLines = s;
   }
}

void DebugRenderer::setEnableBlending(bool b)
{
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->enableBlending = b;
   }
}

void DebugRenderer::setPointSize(float32 s)
{
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->pointSize = s;
   }
}

void DebugRenderer::addLine( const Vector3f &point1, const Vector3f &point2) 
{
	addLine( point1, point2, makeVector3f(1,1,1), makeVector3f(1,1,1), 2 );
}

void DebugRenderer::addLine( const Vector3f &point1, const Vector3f &point2, const Vector3f &color1, const Vector3f &color2, float linewidth) 
{
	if( m_CurrentRenderJobPtr )
	{
		TRJLine line;
		line.linewidth = linewidth;
		line.p1 = point1;
		line.p2 = point2;
		line.c1 = color1;
		line.c2 = color2;
		m_CurrentRenderJobPtr->lines.push_back( line );
	}
}

void DebugRenderer::addSceneObject(Scene* scene, SceneObject* obj, GLMaterial* material)
{
   TSceneObject ts;
   
   ts.scene = scene;
   ts.object = obj;
   ts.material = material ? material : scene->getGLMaterial(obj->getMaterialIndex());
   
   if (m_CurrentRenderJobPtr)
   {
      m_CurrentRenderJobPtr->object.push_back(ts);
   }   
}


void DebugRenderer::draw( GLContext * glContext, Scene * scene, bool drawSceneObjects ) 
{
	DebugRenderer_CurrentCameraPosition = scene->getRootState()->staticState->camera.getPosition();
	int currentTime = scene->getRootState()->dynamicState->time;
	for( card32 i=0;i<m_RenderJobs.size();i++ )
	{
		if( m_RenderJobs[i]->active )
		{
			if( m_RenderJobs[i]->time == DRAW_EVERYTIME || m_RenderJobs[i]->time == currentTime )
				drawJob( glContext, scene, m_RenderJobs[i],drawSceneObjects );
		}
	}
}

void DebugRenderer::drawJob(GLContext * glContext, Scene * scene, TDebugRenderJob* job, bool drawSceneObjects) 
{
   // enable matrix tranformation
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glMultMatrixf(job->objectMatrix.data());

   glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT);  
	for( card32 i=0;i<job->lines.size();i++ )
	{
	   if (job->smoothLines)
	   {
	      glEnable(GL_BLEND);
	      glEnable(GL_LINE_SMOOTH);
         glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
 	   }
	   
		TRJLine line = job->lines[i];
		glLineWidth( line.linewidth);
		glBegin( GL_LINES );
			glColor3fv(line.c1.data());
			glNormal3f(1,0,0);
			glVertex3fv(line.p1.data());
			glColor3fv(line.c2.data());
         glNormal3f(1,0,0);
			glVertex3fv(line.p2.data());
		glEnd();
	}
   glPopAttrib();
   
	glPointSize( job->pointSize );
	glBegin( GL_POINTS );
	for( card32 i=0;i<job->points.size(); i+=2)
	{
		glColor3fv( job->points[i+1].data());
		glVertex3fv( job->points[i].data());
	}
	glEnd();


	if( m_RTools.find(glContext) == m_RTools.end())
	{
		m_RTools[glContext] = new RenderingTool;
		m_RTools[glContext]->init();
	}

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	RenderingTool * rt = m_RTools[glContext];
	rt->enableLighting();

	for( card32 i=0;i<job->spheres.size();i++ )
	{
		TRJSphere s = job->spheres[i];
		rt->setRenderColor(s.color);
		rt->renderSphere( s.pos, s.radius );
	}
	
	glPopAttrib();
	

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(true);
	glEnable( GL_LIGHTING );
	glEnable( GL_COLOR_MATERIAL);
	enableBrightLight();

	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHT1 );

	UnstructuredInCoreTriangleMesh * mesh = job->mesh;	
	if( mesh->getNumTriangles() > 0 )
	{
		glEnable( GL_VERTEX_ARRAY );
		glEnable( GL_NORMAL_ARRAY );
		glEnable( GL_COLOR_ARRAY );
		VertexArray va_pts( mesh->getPointSet() );

		glVertexPointer( 3, GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "position") );
		glNormalPointer( GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "normal") );
		glColorPointer( 3, GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "color") );

		glDrawElements( GL_TRIANGLES, mesh->getTriangles()->getNumEntries()*3, GL_UNSIGNED_INT, mesh->getTriangles()->getDataPointer());

		glDisable(GL_COLOR_ARRAY );
		glDisable(GL_NORMAL_ARRAY );
		glDisable(GL_VERTEX_ARRAY );
	}

	UnstructuredInCorePointCloud * surfelCloud = job->surfelCloud;
	if( surfelCloud->getNumPoints() > 0 )
	{
		glEnable( GL_VERTEX_ARRAY );
		glEnable( GL_NORMAL_ARRAY );
		glEnable( GL_COLOR_ARRAY );
		VertexArray va_pts( surfelCloud->getPointSet() );

		glVertexPointer( 3, GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "position") );
		glNormalPointer( GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "normal") );
		glColorPointer( 3, GL_FLOAT, va_pts.getVertexDescriptor()->getSize(), va_pts.getAttributePointer(0, "color") );

		glPointSize(2.0f);

		glDrawArrays( GL_POINTS, 0, surfelCloud->getNumPoints());

		glDisable(GL_COLOR_ARRAY );
		glDisable(GL_NORMAL_ARRAY );
		glDisable(GL_VERTEX_ARRAY );
	}


	glPopAttrib();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	rt->enableLighting();
	glEnable(GL_COLOR_MATERIAL);
	for( card32 i=0;i<job->polygons.size();i++)
	{
		glBegin( GL_POLYGON );
		glColor3fv( job->polygons[i].color.data());
		glNormal3fv(job->polygons[i].normal.data());

		for( card32 j=0;j<job->polygons[i].points.size();j++ )
			glVertex3fv( job->polygons[i].points[j].data());
		glEnd();
	}
	rt->disableLighting();
	glPopAttrib();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	GLHelper::begin2D(1,1);
	glBegin( GL_POINTS );
	for( card32 i=0;i<job->points_2D.size();i+=2)
	{
		glColor3fv( job->points_2D[i+1].data());
		glVertex3fv( job->points_2D[i].data());
	}
	
	glEnd();
	GLHelper::end2D();
	glPopAttrib();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	rt->enableLighting();
	
	for( card32 i=0;i<job->ellipsoids.size();i++ )
	{
		rt->setRenderColor( job->ellipsoids[i].color );
		rt->renderEllipse( job->ellipsoids[i].pos, job->ellipsoids[i].orientation, job->ellipsoids[i].scale );
	}

	glPopAttrib();
   
	if( drawSceneObjects)
	{
		// Render scene object 
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		if (job->object.size() > 0)
		{
			rt->enableLighting();
			for (unsigned i=0; i < job->object.size(); i++)
			{
				TSceneObject& ts = job->object[i];
				SceneObjectIterator* sIt = ts.object->createIterator(SOIT::CAP_BASIC_PC);
				ts.material->draw(ts.scene, ts.scene->getRootState(), ts.object, sIt, glContext);
				delete sIt;
			}      
		}
		glPopAttrib();
	}

	glPopMatrix();

   
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	if( job->glowingNormalDistributions.size() > 0)
	{
		rt->disableLighting();

		// setup blend & depth buffer settings
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE,GL_ONE);
		glDepthMask(0);
				
		// setup texture
		GLTexture2D tex2D(glContext);
		unsigned size=64;
		vector<unsigned> texture(size*size);
		float sigma = size/2.0f * 0.45f;
		for( unsigned x=0;x<size;x++)
		{
			for( unsigned y=0;y<size;y++)
			{
				float xf = (float)x - (float)size/2.0f;
				float yf = (float)y - (float)size/2.0f;
				float v = exp( -(xf*xf + yf*yf) / sigma / sigma ) * 255.0f;
				int vi = v;
				texture[y*size+x] = vi + (vi << 8) + (vi << 16) + (255 << 24);
			}
		}
		
		tex2D.createFromMemory(size,size,4, (char*)&texture[0]);
		tex2D.bind();
		glEnable(GL_TEXTURE_2D);

		glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		
		glBegin(GL_QUADS);
		for( card32 i=0;i<job->glowingNormalDistributions.size();i++ )
		{
			TNormalDist2D nd = job->glowingNormalDistributions[i];

			glColor3fv(nd.color.data());
			glTexCoord2f(1,1);
			glVertex3fv(nd.x[0].data());
			glTexCoord2f(0,1);
			glVertex3fv(nd.x[1].data());
			glTexCoord2f(0,0);
			glVertex3fv(nd.x[2].data());
			glTexCoord2f(1,0);
			glVertex3fv(nd.x[3].data());
		}
		glEnd();
		tex2D.unbind();

		glDepthMask(1);


	}
	glPopAttrib();

}

void DebugRenderer::renderJob(GLContext * glContext, Scene * scene, string id)
{
	for( unsigned i=0;i<m_RenderJobs.size();i++ )
	{
		if( m_RenderJobs[i]->id == id)
		{
			drawJob(glContext, scene, m_RenderJobs[i]);
		}
	}
}

void DebugRenderer::addSphere( const Vector3f &point, const float radius, const Vector3f &color) 
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	TRJSphere sphere;
	sphere.pos = point;
	sphere.radius = radius;
	sphere.color = color;
	m_CurrentRenderJobPtr->spheres.push_back( sphere );
}

void DebugRenderer::addPatch( const Vector3f &point, const Matrix3f &paramToWorld, const StaticVector<float,6> &coeffs, int numRSegments, int numCSegments, const Vector3f &color,
							 float parametricScale, float patchSize) 
{

	if( m_CurrentRenderJobPtr == NULL )
		return;

	vector<Vector3f> positions, normals;
	for( int a = 0;a<numCSegments;a++ )
	{
		float angle = (float)a / (float)numCSegments * M_PI * 2.0f;
		for( int b = 0;b<numRSegments;b++ )
		{
			if( a > 0 && b == 0 )
				continue;

			Vector2f paramPos = makeVector2f( sin(angle),cos(angle) ) * (float)b / (float)(numRSegments-1) *  parametricScale;
			float u = paramPos[0];
			float v = paramPos[1];
			float w = coeffs[0]
				+ coeffs[1] * u
				+ coeffs[2] * v
				+ coeffs[3] * u*u
				+ coeffs[4] * u*v
				+ coeffs[5] * v*v;
			Vector3f samplePos = makeVector3f(u,v, w) * patchSize;
			//				Vector3f samplePos = makeVector3f(u,v, 0) * scalePatch;
			positions.push_back( paramToWorld * samplePos + point);

			float derivateU = 2.0f * coeffs[3] * u + coeffs[4] * v + coeffs[1];
			float derivateV = 2.0f * coeffs[5] * v + coeffs[4] * u + coeffs[2];
			Vector3f normal = -normalize( makeVector3f(1,0, derivateU).crossProduct( makeVector3f(0,1,derivateV) ) );
			normal = paramToWorld * normal;

			normals.push_back( -normal);

		}
	}

	PointSet * ps = m_CurrentRenderJobPtr->mesh->getPointSet();
	card32 currentVertexIndex = ps->getNumEntries();

	ps->changeHeight( currentVertexIndex + positions.size() );
	VertexArray points(ps);

	for( card32 j=0;j<positions.size();j++ )
	{
		points.setPosition3f(currentVertexIndex + j, positions[j]);
		points.setNormal3f(currentVertexIndex + j, normals[j]);
		points.setColor3f(currentVertexIndex + j, color );
	}

	Vector3i start = makeVector3i(1,1,1) * currentVertexIndex;
	vector<Vector3i> indices;

	// triangles
	for( card32 j=0;j<(card32)numCSegments;j++ ) // inner ring
	{
		indices.push_back( makeVector3i(0, j*(numRSegments-1)+1, ((j+1)%numCSegments)*(numRSegments-1)+1) + start );

	}

	for( card32 j=1;j<(card32)numRSegments-1;j++)
	{
		for( card32 k=0;k<(card32)numCSegments;k++ )
		{
			card32 i1 = 1 + k * (numRSegments-1) + j-1;
			card32 i2 = 1 + ((k+1)%numCSegments) * (numRSegments-1) + j-1;
			card32 i3 = 1 + ((k+1)%numCSegments) * (numRSegments-1) + j;
			card32 i4 = 1 + k * (numRSegments-1) + j;
			indices.push_back( makeVector3i(i1,i2,i3) + start );			
			indices.push_back( makeVector3i(i1,i3,i4) + start );			
		}
	}

	PointSet * tri_ps = m_CurrentRenderJobPtr->mesh->getTriangles();
	card32 currentTriangleIndex = tri_ps->getNumEntries();
	tri_ps->changeHeight( currentTriangleIndex + indices.size());
	VertexArray triangles( tri_ps );
	for( card32 i=0;i<indices.size();i++)
		triangles.setIndex3i(currentTriangleIndex + i, indices[i]);

}

void DebugRenderer::addCoordSystem( const Vector3f &pos, const Matrix3f &mat, float size )
{
	addLine( pos, pos + mat * makeVector3f(1,0,0) * size, makeVector3f(1,0,0), makeVector3f(1,0,0), 2 );
	addLine( pos, pos + mat * makeVector3f(0,1,0) * size, makeVector3f(0.5,0,0), makeVector3f(0.5,0,0), 2 );
	addLine( pos, pos + mat * makeVector3f(0,0,1) * size, makeVector3f(0,0,1), makeVector3f(0,0,1), 2 );
}

void DebugRenderer::addPoint( const Vector3f &point, const Vector3f &color) 
{
	if( !m_CurrentRenderJobPtr )
		return;
	m_CurrentRenderJobPtr->points.push_back(point);
	m_CurrentRenderJobPtr->points.push_back(color);
}

void DebugRenderer::addBoundingBox( const BoundingBox3f &bb, const Vector3f &color, const float32 &linewidth) 
{
	addLine( makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]),makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]),makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.lowerCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.lowerCorner[2]),makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.lowerCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.lowerCorner[2]),makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]), color, color, linewidth );

	addLine( makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.upperCorner[2]),makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.upperCorner[2]),makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.upperCorner[2]),makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.upperCorner[2]),makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.upperCorner[2]), color, color, linewidth );

	addLine( makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]),makeVector3f(bb.lowerCorner[0],bb.lowerCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.lowerCorner[2]),makeVector3f(bb.lowerCorner[0],bb.upperCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.lowerCorner[2]),makeVector3f(bb.upperCorner[0],bb.upperCorner[1],bb.upperCorner[2]), color, color, linewidth );
	addLine( makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.lowerCorner[2]),makeVector3f(bb.upperCorner[0],bb.lowerCorner[1],bb.upperCorner[2]), color, color, linewidth );
}

void DebugRenderer::add2DPoint( const Vector2f &point, const Vector3f &color) 
{
	if( !m_CurrentRenderJobPtr )
		return;

	m_CurrentRenderJobPtr->points_2D.push_back( makeVector3f(point[0],point[1],0.5f) );
	m_CurrentRenderJobPtr->points_2D.push_back( color );
}

void DebugRenderer::addPolygon( const Vector3f * points, card32 numPoints, const Vector3f &color )
{
	if( !m_CurrentRenderJobPtr )
		return;
	
	if( numPoints < 3 )
	{
		debugOutput << "Polygon must have at least 3 points!\n";
		return;
	}

	Vector3f normal = normalize((points[1] - points[0]).crossProduct(points[2] - points[0]));

	TRJPolygon poly;
	for( card32 i=0;i<numPoints;i++ )
		poly.points.push_back( points[i]);
	poly.color = color;
	poly.normal = normal;
	m_CurrentRenderJobPtr->polygons.push_back(poly);
}

void DebugRenderer::addPolygon( const vector<Vector3f> &points, const Vector3f &color )
{
	addPolygon( &points[0], (card32)points.size(), color);
}
void DebugRenderer::addQuad( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &p4, const Vector3f &color)
{
	static Vector3f tmpPoints[4];
	tmpPoints[0] = p1;
	tmpPoints[1] = p2;
	tmpPoints[2] = p3;
	tmpPoints[3] = p4;
	addPolygon( tmpPoints, 4, color);
}

void DebugRenderer::addTriangle( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &color)
{
	static Vector3f tmpPoints[3];
	tmpPoints[0] = p1;
	tmpPoints[1] = p2;
	tmpPoints[2] = p3;
	addPolygon( tmpPoints, 3, color);
}

/*void DebugRenderer::addPlane( const Plane &plane, const BoundingBox3f &bb)
{
	
}*/


void DebugRenderer::addEllipsoid( const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color )
{
	if( !m_CurrentRenderJobPtr )
		return;

	TRJEllipsoid tmp;
	tmp.pos = pos;
	tmp.orientation = orientation;
	tmp.scale = scale;
	tmp.color = color;
	m_CurrentRenderJobPtr->ellipsoids.push_back(tmp);
}

void DebugRenderer::disconnect( GLContext * context )
{
	delete m_RTools[context];
	m_RTools.erase(context);
}

void DebugRenderer::addBox( const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color )
{
	Vector3f c[8];
	c[0] = makeVector3f(0,0,0);
	c[1] = makeVector3f(1,0,0);
	c[2] = makeVector3f(1,0,1);
	c[3] = makeVector3f(0,0,1);
	c[4] = makeVector3f(0,1,0);
	c[5] = makeVector3f(1,1,0);
	c[6] = makeVector3f(1,1,1);
	c[7] = makeVector3f(0,1,1);

	for(int i=0;i<8;i++)
	{
		c[i] = orientation * (c[i].componentProduct(scale)) + pos;
	}

	addQuad( c[0],c[1],c[2],c[3], color );
	addQuad( c[7],c[6],c[5],c[4], color );
	addQuad( c[0],c[4],c[5],c[1], color );
	addQuad( c[3],c[2],c[6],c[7], color );
	addQuad( c[2],c[1],c[5],c[6], color );
	addQuad( c[0],c[3],c[7],c[4], color );
}

void DebugRenderer::addCenteredBox( const Vector3f &center, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color )
{
	addBox( center - scale, orientation, scale, color );
}

void DebugRenderer::addCoordSystem( const Matrix4f &mat, float size )
{
	addCoordSystem(shrink4To3(mat[3]),shrink4To3(mat),size);
}

void DebugRenderer::addFastSphere( const Vector3f &pos, float radius, const Vector3f &color, bool highRes )
{
	addFastEllipse(pos, IDENTITY3F, makeVector3f(1,1,1)*radius, color,highRes );
}

void geoxl::DebugRenderer::addFastEllipse( const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color, bool highRes )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	// create a unitsphere centerd at 0
	float numXSegments=8;
	float numYSegments=5;
	if( highRes )
	{
		numXSegments = 15;
		numYSegments = 12;
	}
	vector<Vector3f> positions, normals;
	// poles
	positions.push_back(makeVector3f(0,1,0));
	positions.push_back(makeVector3f(0,-1,0));

	for( int x=0;x<numXSegments;x++)
	{
		float xAngle = (float)x/numXSegments*2.0f*M_PI;
		for( int y=0;y<numYSegments-1;y++)
		{
			float yAngle = (float)(y+1.0f) / (numYSegments) * M_PI;
			positions.push_back(makeVector3f( sin(xAngle)*sin(yAngle),cos(yAngle),cos(xAngle)*sin(yAngle) ));
		}
	}


	// transform and scale unit sphere to ellipsoid
	for( unsigned i=0;i<positions.size();i++ )
	{
		Vector3f normal = orientation * positions[i];
		normals.push_back(normal);
		positions[i] = pos + orientation * (positions[i].componentProduct(scale));
	}

	PointSet * ps = m_CurrentRenderJobPtr->mesh->getPointSet();
	card32 currentVertexIndex = ps->getNumEntries();

	ps->changeHeight( currentVertexIndex + positions.size() );
	VertexArray points(ps);

	for( card32 j=0;j<positions.size();j++ )
	{
		points.setPosition3f(currentVertexIndex + j, positions[j]);
		points.setNormal3f(currentVertexIndex + j, normals[j]);
		points.setColor3f(currentVertexIndex + j, color );
	}

	Vector3i start = makeVector3i(1,1,1) * currentVertexIndex;
	vector<Vector3i> indices;

	// triangles
	// poles
	for( card32 j=0;j<(card32)numXSegments;j++ ) // inner ring
	{
		indices.push_back( makeVector3i(0,j*(numYSegments-1)+2,((j+1)%(int)numXSegments)*(numYSegments-1)+2) + start );
		indices.push_back( makeVector3i(1,((j+1)%(int)numXSegments)*(numYSegments-1)+2 + numYSegments-2,j*(numYSegments-1)+2 + numYSegments-2) + start );
	}

	// normals segments
	for( unsigned x=0;x<numXSegments;x++ )
	{
		int x2 = (x+1)%(int)numXSegments;
		for( unsigned y=1;y<numYSegments-1;y++ )
		{
			int i1 = x*(numYSegments-1) + y-1 +2;
			int i2 = x2*(numYSegments-1) + y-1 +2;
			int i3 = x2*(numYSegments-1) + y +2;
			int i4 = x*(numYSegments-1) + y +2;
			indices.push_back( makeVector3i(i2,i1,i4) + start );
			indices.push_back( makeVector3i(i2,i4,i3) + start );
		}
	}

	PointSet * tri_ps = m_CurrentRenderJobPtr->mesh->getTriangles();
	card32 currentTriangleIndex = tri_ps->getNumEntries();
	tri_ps->changeHeight( currentTriangleIndex + indices.size());
	VertexArray triangles( tri_ps );
	for( card32 i=0;i<indices.size();i++)
		triangles.setIndex3i(currentTriangleIndex + i, indices[i]);
	
}

void geoxl::DebugRenderer::addFineBoundingBox( const Matrix4f &frame, const BoundingBox3f &bb, const Vector3f &color, bool highRes /*= true */ )
{
	float sphereRad = bb.getMinSideLength() * 0.05f;
	float cylinderRad = sphereRad * 0.5f;

	Vector3f corners[8];
	for( unsigned i=0;i<8;i++ )
	{
		for( unsigned j=0;j<3;j++ )
		{
			corners[i][j] = (i & (1 << j)) != 0 ? bb.upperCorner[j] : bb.lowerCorner[j];
		}
	}

	for( unsigned i=0;i<8;i++ )
	{
		corners[i] = transformVector3f( frame, corners[i]);
		if( i == 0)
			addFastSphere(corners[i], sphereRad*1.6f, color * 0.5f, highRes);
		else
			addFastSphere(corners[i], sphereRad, color, highRes);
	}

	addFastCylinder( corners[0], corners[1]-corners[0],norm(corners[1]-corners[0]), cylinderRad, color*0.85f, false, highRes );
	addFastCylinder( corners[0], corners[2]-corners[0],norm(corners[2]-corners[0]), cylinderRad, color*0.7f, false, highRes );
	addFastCylinder( corners[0], corners[4]-corners[0],norm(corners[4]-corners[0]), cylinderRad, color, false, highRes );

	addFastCylinder( corners[7], corners[3]-corners[7],norm(corners[3]-corners[7]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[7], corners[5]-corners[7],norm(corners[5]-corners[7]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[7], corners[6]-corners[7],norm(corners[6]-corners[7]), cylinderRad, color, false, highRes ); 

	addFastCylinder( corners[3], corners[1]-corners[3],norm(corners[1]-corners[3]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[4], corners[5]-corners[4],norm(corners[5]-corners[4]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[5], corners[1]-corners[5],norm(corners[1]-corners[5]), cylinderRad, color, false, highRes );

	addFastCylinder( corners[2], corners[3]-corners[2],norm(corners[3]-corners[2]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[4], corners[6]-corners[4],norm(corners[6]-corners[4]), cylinderRad, color, false, highRes );
	addFastCylinder( corners[2], corners[6]-corners[2],norm(corners[6]-corners[2]), cylinderRad, color, false, highRes );
}

void geoxl::DebugRenderer::enableBrightLight()
{
	Vector3f lightAmbient0 = makeVector3f( 0.1f, 0.1f, 0.1f );
	Vector3f lightDiffuse0 = makeVector3f( 0.5f, 0.5f, 0.5f );
	Vector3f lightSpecular0 = makeVector3f( 0.15f, 0.15f, 0.15f );
	Vector4f lightPos0 = makeVector4f( 0, -1000, -1000, 1000 );
	Vector3f lightAmbient1 = makeVector3f( 0.1f, 0.1f, 0.1f );
	Vector3f lightDiffuse1 = makeVector3f( 0.5f, 0.5f, 0.5f );
	Vector3f lightSpecular1 = makeVector3f( 0.15f, 0.15f, 0.15f );
	Vector4f lightPos1 = makeVector4f( 0, 1000, 1000, 1000 );

	lightPos0 = expand3To4(DebugRenderer_CurrentCameraPosition);

	glEnable( GL_LIGHTING );
	// Light 0
	float ambient0[] = { lightAmbient0[0], lightAmbient0[1], lightAmbient0[2], 1.0f };
	float diffuse0[] = { lightDiffuse0[0], lightDiffuse0[1], lightDiffuse0[2], 1.0f };
	float specular0[] = { lightSpecular0[0], lightSpecular0[1], lightSpecular0[2], 1.0f };
	float light_pos0[] = { lightPos0[0], lightPos0[1], lightPos0[2], lightPos0[3] };
	glLightfv( GL_LIGHT0, GL_AMBIENT, ambient0 );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse0 );
	glLightfv( GL_LIGHT0, GL_SPECULAR, specular0 );
	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	// Light 1
	float ambient1[] = { lightAmbient1[0], lightAmbient1[1], lightAmbient1[2], 1.0f };
	float diffuse1[] = { lightDiffuse1[0], lightDiffuse1[1], lightDiffuse1[2], 1.0f };
	float specular1[] = { lightSpecular1[0], lightSpecular1[1], lightSpecular1[2], 1.0f };
	float light_pos1[] = { lightPos1[0], lightPos1[1], lightPos1[2], lightPos1[3] };
	glLightfv( GL_LIGHT1, GL_AMBIENT, ambient1 );
	glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse1 );
	glLightfv( GL_LIGHT1, GL_SPECULAR, specular1 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHT1 );
}

struct TDRSurfel
{
	Vector3f pos;
	Vector3f normal;
	Vector3f color;
};

void geoxl::DebugRenderer::addSurfel( const Vector3f &point, const Vector3f &normal, const Vector3f &color )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;
	PointSet * ps = m_CurrentRenderJobPtr->surfelCloud->getPointSet();
	TDRSurfel surfel;
	surfel.pos = point;
	surfel.normal = normal;
	surfel.color = color;
	ps->addPoint((char*)&surfel);
}

void geoxl::DebugRenderer::addFineCoordinateFrame( const Vector3f &pos, const Matrix3f &frame, float size, const Vector3f &color, bool highRes /*= true*/ )
{
	float cylinderRadius=size/40.0f;
	float coneLength = size / 7.0f;
	float coneRadius = size / 14.0f;
	addFastCylinder(pos, frame[0], size, cylinderRadius, makeVector3f(1,1,1)*0.6f,false,highRes);
	addFastCone(pos + frame[0]*size, frame[0], coneLength, coneRadius, color, highRes);
	addFastCylinder(pos, frame[1], size, cylinderRadius, makeVector3f(1,1,1)*0.25f,false,highRes);
	addFastCone(pos + frame[1]*size, frame[1], coneLength, coneRadius, color, highRes);
	addFastCylinder(pos, frame[2], size, cylinderRadius, color,false,highRes);
	addFastCone(pos + frame[2]*size, frame[2], coneLength, coneRadius, color, highRes);
	addFastSphere(pos, coneRadius, color, highRes);
}

void geoxl::DebugRenderer::addFineCoordinateFrame( const Matrix4f &frame, float size, const Vector3f &color, bool highRes /*= true*/ )
{
	addFineCoordinateFrame(shrink4To3(frame[3]), shrink4To3(frame),size,color,highRes);
}

void geoxl::DebugRenderer::addFastCone( const Vector3f &head, const Vector3f &direction, float length, float radius, const Vector3f &color, bool highRes /*= true*/ )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	// create a unitcylinder centerd at 0
	float numSegments=5;
	if( highRes )
	{
		numSegments = 15;
	}
	vector<Vector3f> positions, normals;
	Vector3f tmp = direction;
	tmp.normalize();
	Matrix3f frame = calcTangentSystem(tmp);

	for( int x=0;x<numSegments;x++)
	{
		float xAngle = (float)x/numSegments*2.0f*M_PI;
		positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),0 ));
		Vector3f normal = normalize( positions.back() + makeVector3f(0,0,1));
		normals.push_back(normal);
		positions.push_back(makeVector3f( 0,0,1.0f ));
		normals.push_back(normal);
	}

	for( int x=0;x<numSegments;x++)
	{
		float xAngle = (float)x/numSegments*2.0f*M_PI;
		positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),0 ));
		normals.push_back(makeVector3f(0,0,-1.0f));
	}

	positions.push_back( makeVector3f(0,0,0));
	normals.push_back(makeVector3f(0,0,-1));

	// transform and scale unit cone
	for( unsigned i=0;i<positions.size();i++ )
	{
		normals[i] = frame * normals[i];
		positions[i] = head + frame * (positions[i].componentProduct(makeVector3f(radius,radius,length)));
	}

	PointSet * ps = m_CurrentRenderJobPtr->mesh->getPointSet();
	card32 currentVertexIndex = ps->getNumEntries();

	ps->changeHeight( currentVertexIndex + positions.size() );
	VertexArray points(ps);

	for( card32 j=0;j<positions.size();j++ )
	{
		points.setPosition3f(currentVertexIndex + j, positions[j]);
		points.setNormal3f(currentVertexIndex + j, normals[j]);
		points.setColor3f(currentVertexIndex + j, color );
	}

	Vector3i start = makeVector3i(1,1,1) * currentVertexIndex;
	vector<Vector3i> indices;

	// triangles
	for( card32 j=0;j<(card32)numSegments;j++ ) 
	{
		int j2 = (j+1) % (card32)numSegments;
		int i1 = j*2;
		int i2 = j2*2;
		int i3 = j2*2+1;
		int i4 = j*2+1;
		indices.push_back( makeVector3i(i2,i1,i4) + start );
		indices.push_back( makeVector3i(i2,i4,i3) + start );
	}

	start += makeVector3i(1,1,1) * numSegments * 2;
	for( card32 j=0;j<(card32)numSegments;j++ ) 
	{
		int j2 = (j+1) % (card32)numSegments;
		indices.push_back( makeVector3i(numSegments,j,j2) + start );
	}

	PointSet * tri_ps = m_CurrentRenderJobPtr->mesh->getTriangles();
	card32 currentTriangleIndex = tri_ps->getNumEntries();
	tri_ps->changeHeight( currentTriangleIndex + indices.size());
	VertexArray triangles( tri_ps );
	for( card32 i=0;i<indices.size();i++)
		triangles.setIndex3i(currentTriangleIndex + i, indices[i]);
}

void geoxl::DebugRenderer::addSphereWiremesh( const Vector3f &point, const float radius, const float linewidth, const Vector3f &color )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	addPlanarEllipse( point, makeVector3f(1,0,0)*radius, makeVector3f(0,1,0)*radius,color,linewidth);
	addPlanarEllipse( point, makeVector3f(0,0,1)*radius, makeVector3f(0,1,0)*radius,color,linewidth);
	addPlanarEllipse( point, makeVector3f(0,0,1)*radius, makeVector3f(1,0,0)*radius,color,linewidth);
}

void geoxl::DebugRenderer::addGlowingNormalDitribution2D( const Vector3f &pos, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	TNormalDist2D nd;
	nd.color = color;
	nd.x[0] = pos + dir1 + dir2;
	nd.x[1] = pos - dir1 + dir2;
	nd.x[2] = pos - dir1 - dir2;
	nd.x[3] = pos + dir1 - dir2;

	m_CurrentRenderJobPtr->glowingNormalDistributions.push_back(nd);
}

void geoxl::DebugRenderer::addFineArrow( const Vector3f &start, const Vector3f &end, const Vector3f &color, float radius, bool highRes /*= true*/ )
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	Vector3f dir = normalize(end-start);
	float l = norm(end-start);
	addFastCylinder(start, dir, l-radius*2.0f,radius,color,true,highRes);
	addFastCone(end-dir*radius*2.0f, dir, radius*2.0f,radius*2.0f,color * 0.75f,highRes);
}

void geoxl::DebugRenderer::addFastCylinder( const Vector3f &pos, const Vector3f &pos2, float radius, const Vector3f &color, bool caps /*= true*/, bool highRes /*= true*/ )
{
	addFastCylinder(pos,(pos2-pos),norm(pos2-pos),radius,color,caps,highRes);
}

void geoxl::DebugRenderer::renderJobs( GLContext * glContext, Scene * scene, GLShaderMaterial * material )
{
	for( unsigned i=0;i<m_RenderJobs.size();i++ )
	{
		TDebugRenderJob * rj = m_RenderJobs[i];
		// enable matrix tranformation
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glMultMatrixf(rj->objectMatrix.data());
		
		for( unsigned j=0;j<rj->object.size();j++ )
		{
			TSceneObject * so = &rj->object[j];
			SceneObjectIterator * iter = so->object->createIterator(SOIT::CAP_BASIC_PC);
			material->draw(scene, scene->getRootState(),so->object,iter,glContext);
			delete iter;
		}

		glPopMatrix();
	}
}

void geoxl::DebugRenderer::beginRenderJob_OneFrame( string id, int time, bool clear )
{
	string rjname = id + intToStr(time);
	if( clear )
		clearDebugData(rjname);
	beginRenderJob(rjname,true,time);
}

void DebugRenderer::addFastCylinder( const Vector3f &pos, const Vector3f &orientation, float length, float radius, const Vector3f &color, bool caps, bool highRes ) 
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	// create a unitcylinder centerd at 0
	float numSegments=5;
	if( highRes )
	{
		numSegments = 15;
	}
	vector<Vector3f> positions, normals;
	Vector3f tmp = orientation;
	tmp.normalize();
	Matrix3f frame = calcTangentSystem(tmp);

	for( int x=0;x<numSegments;x++)
	{
		float xAngle = (float)x/numSegments*2.0f*M_PI;
		positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),0 ));
		Vector3f normal = positions.back();
		normals.push_back(normal);
		positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),1.0f ));
		normals.push_back(normal);
	}

	if( caps )
	{
		for( int x=0;x<numSegments;x++)
		{
			float xAngle = (float)x/numSegments*2.0f*M_PI;
			positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),0 ));
			normals.push_back(makeVector3f(0,0,-1.0f));
		}
		for( int x=0;x<numSegments;x++)
		{
			float xAngle = (float)x/numSegments*2.0f*M_PI;
			positions.push_back(makeVector3f( sin(xAngle),cos(xAngle),1.0f ));
			normals.push_back(makeVector3f(0,0,+1.0f));
		}

		positions.push_back( makeVector3f(0,0,0));
		normals.push_back(makeVector3f(0,0,-1));
		positions.push_back( makeVector3f(0,0,1));
		normals.push_back(makeVector3f(0,0,1));
	}

	// transform and scale unit cylinder
	for( unsigned i=0;i<positions.size();i++ )
	{
		normals[i] = frame * normals[i];
		positions[i] = pos + frame * (positions[i].componentProduct(makeVector3f(radius,radius,length)));
	}

	PointSet * ps = m_CurrentRenderJobPtr->mesh->getPointSet();
	card32 currentVertexIndex = ps->getNumEntries();

	ps->changeHeight( currentVertexIndex + positions.size() );
	VertexArray points(ps);

	for( card32 j=0;j<positions.size();j++ )
	{
		points.setPosition3f(currentVertexIndex + j, positions[j]);
		points.setNormal3f(currentVertexIndex + j, normals[j]);
		points.setColor3f(currentVertexIndex + j, color );
	}

	Vector3i start = makeVector3i(1,1,1) * currentVertexIndex;
	vector<Vector3i> indices;

	// triangles
	for( card32 j=0;j<(card32)numSegments;j++ ) 
	{
		int j2 = (j+1) % (card32)numSegments;
		int i1 = j*2;
		int i2 = j2*2;
		int i3 = j2*2+1;
		int i4 = j*2+1;
		indices.push_back( makeVector3i(i2,i1,i4) + start );
		indices.push_back( makeVector3i(i2,i4,i3) + start );
	}

	if( caps )
	{
		start += makeVector3i(1,1,1) * numSegments * 2;
		for( card32 j=0;j<(card32)numSegments;j++ ) 
		{
			int j2 = (j+1) % (card32)numSegments;
			indices.push_back( makeVector3i(2*numSegments,j,j2) + start );
			indices.push_back( makeVector3i(2*numSegments+1,j2+numSegments,j+numSegments) + start );
		}
	}

	PointSet * tri_ps = m_CurrentRenderJobPtr->mesh->getTriangles();
	card32 currentTriangleIndex = tri_ps->getNumEntries();
	tri_ps->changeHeight( currentTriangleIndex + indices.size());
	VertexArray triangles( tri_ps );
	for( card32 i=0;i<indices.size();i++)
		triangles.setIndex3i(currentTriangleIndex + i, indices[i]);
	
}
void DebugRenderer::addPositionConstraint( const Vector3f &refPoint, const Vector3f &center, const Matrix3f &covariance ) 
{
	Matrix3f eigenVectors;
	Vector3f eigenValues;
	covariance.computeEigenStructure( eigenValues, eigenVectors);

	for( unsigned i=0;i<3;i++)
		eigenValues[i] = sqrt( eigenValues[i]);

	float scale = 0.25f;

	if( eigenValues[1] > 1e-5f )
	{
		addPlanarEllipse( center, eigenVectors[0] * eigenValues[0]*scale,
			eigenVectors[1] * eigenValues[1]*scale, makeVector3f(0.5f,1.0f,0), 2.0f);
	}
	else
	{
		addLine( center + eigenVectors[0]*eigenValues[0]*scale, center - eigenVectors[0]*eigenValues[0]*scale,
			makeVector3f(1.0f,0.5f,0),makeVector3f(1.0f,0.5f,0),2.0f);
	}
}

void DebugRenderer::addPlanarEllipse( const Vector3f &center, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color, float linewidth) 
{
	if( m_CurrentRenderJobPtr == NULL )
		return;

	int numSegments = 15;
	Vector3f pointA;
	for( int32 i=0;i<numSegments;i++ )
	{
		float angle = (float)i / (numSegments-1) * 2.0f * M_PI;
		Vector3f pointB = center + dir1 * sin(angle) + dir2 * cos(angle);
		if( i > 0)
			addLine(pointA, pointB,color,color,linewidth);
		pointA = pointB;
	}
}
