#ifndef DebugRendererH
#define DebugRendererH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCoreTriangleMesh.h"
//---------------------------------------------------------------------------

namespace geoxl
{

class GLContext;
class Scene;
class RenderingTool;
class GLMaterial;
class SceneObject;

/**
	this class provides visualization tools for debugging
	(doesn't work for multiple windows, there's only one instance)
	Usage Example:
	
	string myid = "what the ... is happening in my code?";
	debugRenderer->clearDebugData( myid );	// clear data from last call
	debugRenderer->beginRenderJob( myid );	// specify job name 
	
	// add some lines
		debugRenderer->addLine( p1,p2 );
		debugRenderer->addLine( p3,p4 );

	debugRenderer->endRenderJob();
*/
class RENDERING_API DebugRenderer 
{


private:

	struct TRJLine
	{
		float32	linewidth;
		Vector3f p1,p2,c1,c2;
	};

	struct TRJSphere
	{
		Vector3f pos;
		float radius;
		Vector3f color;
	};

	struct TRJPolygon
	{
		vector<Vector3f> points;
		Vector3f color;
		Vector3f normal;
	};

	struct TRJEllipsoid
	{
		Vector3f pos;
		Matrix3f orientation;
		Vector3f scale;
		Vector3f color;
	};

	// display 2d normal distribution as textured rectangle
	struct TNormalDist2D
	{
		Vector3f color;
		Vector3f x[4];
	};

   struct TSceneObject
   {
      Scene* scene;
      SceneObject* object;
      GLMaterial* material;
   };
   
	struct TDebugRenderJob
	{
		~TDebugRenderJob() {delete mesh;delete surfelCloud;}
		string id;
		bool active;
		int time;
		vector<TRJLine> lines;
		vector<TRJSphere> spheres;
		vector<Vector3f> points;
		vector<Vector3f> points_2D;
		vector<TRJPolygon> polygons;
		vector<TRJEllipsoid> ellipsoids;
		UnstructuredInCoreTriangleMesh * mesh;
		UnstructuredInCorePointCloud * surfelCloud;
		Matrix4f objectMatrix;
		bool smoothLines;
		bool enableBlending;
		float32 pointSize;
		vector<TNormalDist2D> glowingNormalDistributions;
		
		vector<TSceneObject> object;
	};

	vector<string>	m_RenderJobStack;
	string	m_CurrentRenderJob;
	TDebugRenderJob * m_CurrentRenderJobPtr;
	vector<TDebugRenderJob*>	m_RenderJobs;

	void drawJob(GLContext * glContext, Scene * scene, TDebugRenderJob* job, bool drawSceneObjects = true);
	map< GLContext*, RenderingTool *> m_RTools;

	void enableBrightLight();

public:
	DebugRenderer(void);
	~DebugRenderer(void);

	void disconnect( GLContext * context );


  
   /// deletes render job with id (if exists)
   /// when id is not specified, all render jobs are deleted
   void clearDebugData( const std::string& id = std::string(""));

   static const int DRAW_EVERYTIME = -10000;
	
	/// specify id of current render job, is active is true then the job is automatically rendered
	/// if time is specified, then the render job is only rendered for the given time
   void beginRenderJob( string id, bool active = true, int time = DRAW_EVERYTIME );
   void endRenderJob();

	/// convenience: creates a renderjob name with id + intToStr(time)
	void beginRenderJob_OneFrame( string id, int time, bool clear = true);

	/// manual job rendering
	void renderJob( GLContext * glContext, Scene * scene, string id);

	/// manual rendering special case for high quality renderer
	void renderJobs( GLContext * glContext, Scene * scene, GLShaderMaterial * material );

   void addTransformationMatrix(const Matrix4f& mat);
   void setTransformationMatrix(const Matrix4f& mat);
   
   void setSmoothLine(bool s);
   void setEnableBlending(bool b);
   void setPointSize(float32 s);
   
   /// render scene object (this will automaticaly render through the material of the obejct)
   /// if no material specified then the attached to the given object will be used
   void addSceneObject(Scene* scene, SceneObject* obj, GLMaterial* material = NULL);
   
   void addPoint( const Vector3f &point, const Vector3f &color);
	void addSurfel( const Vector3f &point, const Vector3f &normal, const Vector3f &color);

   void addLine( const Vector3f &point1, const Vector3f &point2);
   void addLine( const Vector3f &point1, const Vector3f &point2, const Vector3f &color1, const Vector3f &color2, float linewidth);
	void addPlanarEllipse( const Vector3f &center, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color, float linewidth);
	void addSphereWiremesh( const Vector3f &point, const float radius, const float linewidth, const Vector3f &color);

   void addSphere( const Vector3f &point, const float radius, const Vector3f &color);
   void addCoordSystem( const Vector3f &pos, const Matrix3f &mat, float size );
	void addCoordSystem( const Matrix4f &mat, float size );
   void addBoundingBox( const BoundingBox3f &bb, const Vector3f &color, const float32 &linewidth=2);

	/// adds 2d point, viewport is (0,0) to (1,1)
	void add2DPoint( const Vector2f &point, const Vector3f &color);

   void addPatch( const Vector3f &point, const Matrix3f &paramToWorld, const StaticVector<float,6> &coeffs, int numRSegments, int numCSegments, const Vector3f &color, 
	   float parametricScale, float patchSize);

	void addPolygon( const Vector3f * points, card32 numPoints, const Vector3f &color );
	void addPolygon( const vector<Vector3f> &points, const Vector3f &color );
	void addQuad( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &p4, const Vector3f &color);
	void addTriangle( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &color);
	void addBox(const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color);
	void addCenteredBox(const Vector3f &center, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color);
	void addEllipsoid( const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color );
	void addPositionConstraint( const Vector3f &refPoint, const Vector3f &center, const Matrix3f &covariance );

	/// uses texture mapping to visualize a normal distribution (blend mode: GL_ONE,GL_ONE)
	void addGlowingNormalDitribution2D( const Vector3f &pos, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color);

	/// creates triangles (much faster when rendering)
	void addFastSphere(const Vector3f &pos, float radius, const Vector3f &color, bool highRes = true);
	void addFastEllipse(const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color, bool highRes = true);
	void addFastCylinder(const Vector3f &pos, const Vector3f &orientation, float length, float radius, const Vector3f &color, bool caps = true, bool highRes = true);
	void addFastCylinder(const Vector3f &pos, const Vector3f &pos2, float radius, const Vector3f &color, bool caps = true, bool highRes = true);
	void addFastCone(const Vector3f &head, const Vector3f &direction, float length, float radius, const Vector3f &color, bool highRes = true);
	void addFineBoundingBox( const Matrix4f &frame, const BoundingBox3f &bb, const Vector3f &color, bool highRes = true );
	void addFineCoordinateFrame( const Vector3f &pos, const Matrix3f &frame, float size, const Vector3f &color, bool highRes = true);
	void addFineCoordinateFrame( const Matrix4f &frame, float size, const Vector3f &color, bool highRes = true);
	void addFineArrow(const Vector3f &start, const Vector3f &end, const Vector3f &color, float radius, bool highRes = true);
	
	/// plane is visible only inside the bb
	//void addPlane( const Plane &plane, const BoundingBox3f &bb);

   void draw( GLContext * glContext, Scene * scene, bool drawSceneObjects = true );

	card32 getNumRenderJobs() const {return (card32)m_RenderJobs.size();}
	string getRenderJobName(card32 index) {return m_RenderJobs[index]->id;}
	bool getRenderJobActiveFlag(card32 index) {return m_RenderJobs[index]->active;}
	void setRenderJobActiveFlag(card32 index, bool b) {m_RenderJobs[index]->active = b;}

};

extern RENDERING_API DebugRenderer * debugRenderer;

}

#endif