//---------------------------------------------------------------------------
#ifndef LCTraversalH
#define LCTraversalH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Camera.h"
#include "ViewFrustum.h"
#include "Frustum.h"
//---------------------------------------------------------------------------

// ---- Legacy Code Hierarchy Renderer (old SHT-Renderer)

namespace geoxl {

class GLContext;
class LCState;
class LCNodeFilter;
class LCNodeProcessor;
class LCStateProcessor;
class SceneObjectIterator;
class SceneGraphState;
class Renderer;
class Scene;
class SceneObject;
class MemoryManager;
class GLShaderMaterial;


class RENDERING_API LCState : public Persistent {
 private:
   static MetaClass *LCStateMetaClass;
   static Object* newInstance();

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   // ---- basic

   card32 hierarchyLevel;
   SceneGraphState *sceneGraphState;
   Renderer *renderer;

   // Profiling
   card32 numPoints;
   card32 numBoxes;

   /** drawing order for hierarchy traversal */
   card32 PARAM_drawOrder;
   static const card32 DRAW_ORDER_ANY;
   static const card32 DRAW_ORDER_FRONT_TO_BACK;
   static const card32 DRAW_ORDER_BACK_TO_FRONT;


   // ----- projective


   Camera camera, saveCamera;
   ViewFrustum viewFrustum;
   BoundedFrustum3f viewFrustumPlanes;
   card32 viewportWidth;
   card32 viewportHeight;
   card32 thisFrameFetches;

   bool PARAM_lockCamera;


   // ----- point rendering


   /** maximum tree level shown for tree visualization */
   card32  PARAM_maxHierarchyLevel;
   /** if false, large Pointsamples are not shown. */
   bool    PARAM_drawPoints;
   /** if true, the octree is showed as white boxes. */
   bool    PARAM_drawHierarchy;
   /** if true, the octree is showed as white boxes. */
   bool    PARAM_drawHierarchyInnerNodes;
   /** if true, selected point clouds are highlighted with transparent green color. */
   bool    PARAM_drawCloudSelection;
   /** if true, selected point clouds are highlighted with transparent red color. */
   bool    PARAM_drawPointSelection;
   /** if true: call the renderNodeGL methods of SceneGraphNodes */
   bool    PARAM_drawNodeGLRenderings;
   /** if true: call the renderObjGL methods of SceneObjects */
   bool    PARAM_drawObjectGLRenderings;
   /** maximum spacing of point samples on the screen (pixels) */
   bool    PARAM_drawAttachmentGLRenderings;
   /** set state variables in shaders? */
   bool    PARAM_passStateVariablesToShaders;
   /** do not try to display non-resident ooc nodes */
   bool    PARAM_lazySwapping;
	/** maximum number of nodes to be prefetched in the same frame */
   card32  PARAM_maxPrefetch;
   /** maximum spacing of point samples on the screen (pixels) */
   float32 PARAM_maxPointSampleSpacing;
   /** tan(half of vertical view angle) for internal calculations (projective classification) */
   float32 DERIVED_PARAM_tanOfHalfVerticalViewAngle;
   /** half vertical resolution divided by sample spacing, for internal calculations */
   float32 DERIVED_PARAM_halfVerticalSamplingResolution;

   /** show exactly one hierarchy level */
   bool   PARAM_useExactRecursionDepth;
   /** show exactly one hierarchy level - level number */
   card32 PARAM_exactRecursionDepth;

   /** scale view frustum for debugging purpose */
   float32 PARAM_viewFrustumScaleFactor;
   /** output statistics on debug console */
   bool PARAM_outputStatistics;


   // ----- new stuff


   Scene *scene;
   SceneObject *sceneObject;
   GLContext *glContext;
   MemoryManager *memManager;
   bool currentCloudIsSelected;
   GLShaderMaterial *cloudSelectionMaterial;
   GLShaderMaterial *pointSelectionMaterial;
	bool useBBSelection;	// draws selection as bounding box

   LCState();

   /** prepare state for traversal. the SceneGraphState can be used for setup. */
   virtual void resetState(SceneGraphState *sceneGraphState, Renderer *renderer, card32 viewportWidth, card32 viewportHeight,
                           Scene *scene, SceneObject *sceneObject, GLContext *glContext, MemoryManager *memManager);

   ~LCState();
};


class RENDERING_API LCStateProcessor : public Persistent {
 public:
   virtual void processState(LCState *state) = 0;
};


class RENDERING_API LCNodeFilter : public Persistent {
 public:
   virtual void filterNode(card32 &currentClassification, LCState *state, SceneObjectIterator *node) = 0;
};


class RENDERING_API LCNodeProcessor : public Persistent {
 public:
   virtual void processNode(LCState *state, SceneObjectIterator *node) = 0;
};



class RENDERING_API LCTraversal : public Persistent {
 private:
   // ---- configuration
   bool processAnimations;

   // ---- state
   LCState *traversalState;

   // ---- actions
   vector<LCNodeFilter*> nodeFilters;
   vector<LCNodeProcessor*> nodeProcessors;
   vector<LCNodeProcessor*> subdividedNodesProcessors;
   vector<LCNodeProcessor*> unSubdividableNodesProcessors;
   vector<LCStateProcessor*> statePreProcessors;
   vector<LCStateProcessor*> statePostProcessors;
   vector<LCNodeProcessor*> rootNodePreProcessors;
   vector<LCNodeProcessor*> rootNodePostProcessors;

   void doStatePreProcessing();
   void doStatePostProcessing();
   void doRootNodePreProcessing(SceneObjectIterator *rootNode);
   void doRootNodePostProcessing(SceneObjectIterator *rootNode);
   void doProcess(SceneObjectIterator *node);
   void doProcessSubdivided(SceneObjectIterator *node);
   void doProcessUnSubdividable(SceneObjectIterator *node);
   void doRekTraversal(SceneObjectIterator *node);

   static MetaClass *LCTraversalMetaClass;
   static Object* newInstance();

   LCTraversal();

 protected:
  virtual void beforeTraversal() {}
   virtual void afterTraversal() {}

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   inline LCState *getTraversalState() {return traversalState;}
   void setTraversalState(LCState *shtst);

   /** node filter result: process node */
   static const card32 NFR_PROCESS;
   /** node filter result: subdivide node */
   static const card32 NFR_DESCEND;
   /** node filter result: do not process or subdivide node */
   static const card32 NFR_DROP;

   LCTraversal(LCState *state);

   void addNodeFilters(LCNodeFilter *nf);
   void addNodeProcessors(LCNodeProcessor *np);
   void addSubdividedNodeProcessors(LCNodeProcessor *sp);
   void addUnSubdividableNodesProcessors(LCNodeProcessor *usnp);
   void addStatePreProcessors(LCStateProcessor *prep);
   void addStatePostProcessors(LCStateProcessor *postp);
   void addRootNodePreProcessors(LCNodeProcessor *prep);
   void addRootNodePostProcessors(LCNodeProcessor *postp);

   virtual void performTraversal(SceneObjectIterator *node, SceneGraphState *state, Renderer *renderer,
                                 card32 viewportWidth, card32 viewportHeight,
                                 Scene *scene, SceneObject *sceneObject, GLContext *glContext, MemoryManager *memManager);

   ~LCTraversal();
};


}


#endif
 