//---------------------------------------------------------------------------
#ifndef MaterialParameterH
#define MaterialParameterH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace geoxl {
   using namespace std;

class GLContext;
class AttachedData;
class AttachedDataContainer;
class GLMaterial;
class Scene;
class SceneObject;
class SceneObjectIterator;
class SecondaryDataStructureIterator;
class AttachedData;
class MemoryManager;
class TextureAttachment;


class RENDERING_API MaterialParameter : public Persistent
{
private:
   static MetaClass *MaterialParameterMetaClass;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();


private:

   // properties
   string   m_Name;  /// identifier
   bool     m_UseGlobalFirst;   /// if true, search in global datacontainers first
   bool     m_UseDefault;  /// if true & no container knows 'Name' -> use default value

public:

   MaterialParameter()
   :Persistent()
   {
      m_Name = "unnamed";
      m_UseGlobalFirst = false;
      m_UseDefault = true;
   }

   string getName() const {return m_Name;}
   void setName( const string &name ) {m_Name = name;}
   bool getUseGlobalFirst() const {return m_UseGlobalFirst;}
   void setUseGlobalFirst( bool b) {m_UseGlobalFirst = b;}
   bool getUseDefault() const {return m_UseDefault;}
   void setUseDefault( bool b) {m_UseDefault = b;}

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject *obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm )=0;

protected:

    virtual AttachedData * getData( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SecondaryDataStructureIterator * iter );

};


class RENDERING_API MPMatrix4x4 : public MaterialParameter
{
private:
   static MetaClass *MPMatrix4x4MetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:

   card32   m_Index; /// modelview[Index] for ARB shaders
   float32  m_DefaultMatrix[4*4];

public:


   MPMatrix4x4() : MaterialParameter()
   {
      // identity
      for( card32 i=0;i<4*4;i++ )
         m_DefaultMatrix[i] = (i % 5) == 0 ? 1.0f : 0.0f;
      m_Index = 0;
   }

   virtual void assign(const Object* obj, CopyContext *context=NULL );

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );


};

class RENDERING_API MPVector4 : public MaterialParameter
{
private:
   static MetaClass *MPVector4MetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:

   card32   m_Index; /// parameter # for ARB shaders
   float32  m_DefaultVector[4];
   float32  m_MinVector[4];
   float32  m_MaxVector[4];
   bool     m_ClampMin;
   bool     m_ClampMax;

public:
   MPVector4() : MaterialParameter()
   {
      for( card32 i=0;i<4;i++ )
      {
         m_DefaultVector[i] = 0.0f;
         m_MinVector[i] = 0.0f;
         m_MaxVector[i] = 1.0f;
      }
      m_ClampMin = false;
      m_ClampMax = false;
      m_Index = 0;
   }
   virtual void assign(const Object* obj, CopyContext *context=NULL );

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

};

class RENDERING_API MPTexture : public MaterialParameter
{
private:
   static MetaClass *MPTextureMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();


private:

   //AttachedDataContainer * m_AttachmentContainer;
   card32   m_TextureStage;
   string   m_DefaultTextureName;
	TextureAttachment	*	m_DefaultTexture;

public:

   MPTexture() : MaterialParameter()
   {
      m_TextureStage = 0;
      m_DefaultTextureName = "texture white 2D";
		m_DefaultTexture = NULL;
   }
	~MPTexture();

   virtual void assign(const Object* obj, CopyContext *context = NULL);

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );
};

class RENDERING_API MPSingleFloat : public MaterialParameter
{
private:
   static MetaClass *MPSingleFloatMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:
   float32  m_Minimum, m_Maximum;
   float32  m_Interpolator;

public:

   MPSingleFloat()
   {
      m_Minimum = 0.0f;
      m_Maximum = 1.0f;
      m_Interpolator = 0.0f;
   }

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

};


}

#endif
