//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "RenderingTool.h"
#include "DebugOutput.h"
#include "LinearAlgebra_templCode.h"
#include "ClassMethods.h"
#include "FixedArrayClassProperty.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS( RenderingTool, 1 )
{
	BEGIN_CLASS_INIT( RenderingTool )
	ADD_FLOAT32_PROP( pointSplatSize, 0 )
	ADD_FLOAT32_PROP( lineThickness, 1 )
	ADD_CARD32_PROP( resolution, 0 )
	ADD_SEPARATOR( "Light0" )
	ADD_VECTOR3F_PROP( lightAmbient0, 0 )
	ADD_VECTOR3F_PROP( lightDiffuse0, 0 )
	ADD_VECTOR3F_PROP( lightSpecular0, 0 )
	ADD_VECTOR4F_PROP( lightPos0, 0 )
	ADD_SEPARATOR( "Light1" )
	ADD_VECTOR3F_PROP( lightAmbient1, 0 )
	ADD_VECTOR3F_PROP( lightDiffuse1, 0 )
	ADD_VECTOR3F_PROP( lightSpecular1, 0 )
	ADD_VECTOR4F_PROP( lightPos1, 0 )
	ADD_NOARGS_METHOD( init );
}


/**
 * Create object
 */
RenderingTool::RenderingTool()
{
	color = makeVector4f( 1, 0, 0, 1 );
	lightAmbient0 = makeVector3f( 0.1f, 0.1f, 0.1f );
	lightDiffuse0 = makeVector3f( 0.5f, 0.5f, 0.5f );
	lightSpecular0 = makeVector3f( 0.5f, 0.5f, 0.5f );
	lightPos0 = makeVector4f( -100, -100, -100, 1 );
	lightAmbient1 = makeVector3f( 0.1f, 0.1f, 0.1f );
	lightDiffuse1 = makeVector3f( 0.5f, 0.5f, 0.5f );
	lightSpecular1 = makeVector3f( 0.5f, 0.5f, 0.5f );
	lightPos1 = makeVector4f( 100, 100, 100, 1 );
	dlArrowSpike = dlDisc = dlFilledDisc = dlSphere = -1;
	pointSplatSize = 1.0f;
	lineThickness = 1.0f;
	resolution = 20;
}

/**
 * Delete object
 */
RenderingTool::~RenderingTool()
{
}

/**
 * Init depth buffer
 */
void RenderingTool::init()
{
   glEnable( GL_DEPTH_TEST );
   glEnable( GL_NORMALIZE );
   // Create display lists
   createArrowDL();
   createDiscDL();
   createSphereDL();
   createCylinderDL();
   checkErrors();
   //debugOutput << "Display lists built.\n"; debugOutput.flush();
}

/// Check for errors
void RenderingTool::checkErrors()
{
   GLenum errCode;
   errCode = glGetError();
   if ( errCode != GL_NO_ERROR ){
      debugOutput << "GL error: " << gluErrorString( errCode ) << ", " << errCode << "\n";
   } 
}

/* --------------- OpenGL setup ----------------------*/

/**
 * Switch on lighting, use standard material
 */
void RenderingTool::enableLighting()
{
	glEnable( GL_LIGHTING );
	// Light 0
	float ambient0[] = { lightAmbient0[0], lightAmbient0[1], lightAmbient0[2], 1.0f };
	float diffuse0[] = { lightDiffuse0[0], lightDiffuse0[1], lightDiffuse0[2], 1.0f };
	float specular0[] = { lightSpecular0[0], lightSpecular0[1], lightSpecular0[2], 1.0f };
	float light_pos0[] = { lightPos0[0], lightPos0[1], lightPos0[2], lightPos0[3] };
	glLightfv( GL_LIGHT0, GL_AMBIENT, ambient0 );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse0 );
	glLightfv( GL_LIGHT0, GL_SPECULAR, specular0 );
	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	// Light 1
	float ambient1[] = { lightAmbient1[0], lightAmbient1[1], lightAmbient1[2], 1.0f };
	float diffuse1[] = { lightDiffuse1[0], lightDiffuse1[1], lightDiffuse1[2], 1.0f };
	float specular1[] = { lightSpecular1[0], lightSpecular1[1], lightSpecular1[2], 1.0f };
	float light_pos1[] = { lightPos1[0], lightPos1[1], lightPos1[2], lightPos1[3] };
	glLightfv( GL_LIGHT1, GL_AMBIENT, ambient1 );
	glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse1 );
	glLightfv( GL_LIGHT1, GL_SPECULAR, specular1 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHT1 );
}

/**
 * Turn off lighting
 */
void RenderingTool::disableLighting()
{
   glDisable( GL_LIGHTING );
}

void RenderingTool::setRenderColor( float r, float g, float b )
{
   color[0] = r;
   color[1] = g;
   color[2] = b;
   glColor3f( color[0], color[1], color[2] );
   computeMaterial();
}

void RenderingTool::setRenderColor( float r, float g, float b, float32 alpha )
{
	color[0] = r;
	color[1] = g;
	color[2] = b;
	color[3] = alpha;
	glColor4fv( color.data() );
	computeMaterial();
}

void RenderingTool::setRenderColor( Vector3f c )
{
   color = makeVector4f( c[0], c[1], c[2], 1.0f );
   glColor4fv( color.data() );
   computeMaterial();
}

void RenderingTool::setRenderColor( Vector4f c )
{
	color = c;
	glColor4fv( color.data() );
	computeMaterial();
}

/**
 * Use current color to set up an appropriate material
 */
void RenderingTool::computeMaterial()
{
	float ambient[] = { color[0], color[1], color[2], 1.0 };
	float diffuse[] = { color[0], color[1], color[2], 1.0 };
	float specular[] = { 1.0, 1.0, 1.0, 1.0 };
	float shininess[] = { 20.0 };
	float light_pos[] = { 10.0, 10.0, 10.0, 0.0 };

	glMaterialfv( GL_FRONT, GL_AMBIENT, ambient );
	glMaterialfv( GL_FRONT, GL_DIFFUSE, diffuse );
	glMaterialfv( GL_FRONT, GL_SPECULAR, specular );
	glMaterialfv( GL_FRONT, GL_SHININESS, shininess );
	glLightfv( GL_LIGHT0, GL_POSITION, light_pos );
}

/**
 * Render a parametric plane
 */
/*
void RenderingTool::renderPlane( ParametricPlane p )
{
	Vector3f point;
	float sMin, sMax, tMin, tMax;
	p.getParamIntervals( sMin, sMax, tMin, tMax );
	glBegin( GL_LINE_STRIP );
		point = p.getPoint() + p.getU() * sMin + p.getV() * tMin; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + p.getU() * sMax + p.getV() * tMin; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + p.getU() * sMax + p.getV() * tMax; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + p.getU() * sMin + p.getV() * tMax; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + p.getU() * sMin + p.getV() * tMin; 
		glVertex3f( point[0], point[1], point[2] );
	glEnd();

	// Render normal
   renderArrow( p.getPoint(), 0.3f, p.getNormal(), 1.0f );
}
*/

/**
 * 'size' describes the extends of the plane quad from the center point
 */
void RenderingTool::renderPlane( Plane p, float size )
{
	Vector3f v0 = ( fabs( p.getNormal() * XAXIS_VECTOR3F ) < 1-1e-4 )? 
				  ( p.getNormal().crossProduct( XAXIS_VECTOR3F ) ) :
				  ( p.getNormal().crossProduct( YAXIS_VECTOR3F ) );
	Vector3f v1 = p.getNormal().crossProduct( v0 );
	v0.normalize();
	v1.normalize();
	Vector3f point;
	disableLighting();
	glBegin( GL_LINE_STRIP );
		point = p.getPoint() - v0 * size - v1 * size; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() - v0 * size + v1 * size; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + v0 * size + v1 * size; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() + v0 * size - v1 * size; 
		glVertex3f( point[0], point[1], point[2] );
		point = p.getPoint() - v0 * size - v1 * size; 
		glVertex3f( point[0], point[1], point[2] );
	glEnd();
	enableLighting();

	// Render normal
	renderArrow( p.getPoint(), 2.0f * size, p.getNormal(), 1.0f );
}

/**
 * Render a simple point
 */
void RenderingTool::renderPoint( Vector3f p ) const
{
	glBegin( GL_POINTS );
		glVertex3f( p[0], p[1], p[2] );
	glEnd();
}

/**
 * Disc is centered at 'center' with normal direction 'normal' and radius 'radius'
 */
void RenderingTool::renderDisc( Vector3f center, Vector3f normal, float radius, bool filled )
{
   // Compute local coordinate system
   Matrix3f m;
   m[1] = normal;
   m[0] = ( fabs( XAXIS_VECTOR3F * normal ) > 0.9 ) ? m[1].crossProduct( YAXIS_VECTOR3F ) : m[1].crossProduct( XAXIS_VECTOR3F );
   m[2] = m[1].crossProduct( m[0] );
   m[0].normalize();
   m[1].normalize();
   m[2].normalize();
   float rotMatrix[] = { m[0][0], m[0][1], m[0][2], 0,
						 m[1][0], m[1][1], m[1][2], 0,
					     m[2][0], m[2][1], m[2][2], 0,
					     0,       0,       0,       1 };

   /// Render disc
   glPushMatrix();
      glTranslatef( center[0], center[1], center[2] );
	  glMultMatrixf( rotMatrix );
      glScalef( radius, radius, radius );
      if ( filled )
         glCallList( dlFilledDisc );
      else
         glCallList( dlDisc );
   glPopMatrix();
}

/**
 * Render a parametric plane
 */
void RenderingTool::renderSphere( Vector3f pos, float radius, bool wireframe )
{
	glPushMatrix();
		glTranslatef( pos[0], pos[1], pos[2] );
		glScalef( radius, radius, radius );
		if ( wireframe ){ 
			disableLighting();
			glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
		}
		glCallList( dlSphere );
		if ( wireframe ){
			enableLighting();
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
		}
   glPopMatrix();
}

/// Render a line
void RenderingTool::renderLine( const Vector3f start, const Vector3f end, const Vector3f normalStart, const Vector3f normalEnd ) const
{
	glLineWidth( lineThickness );
	glBegin( GL_LINES );
		glColor4fv( color.data() );
		glNormal3f( normalStart[0], normalStart[1], normalStart[2] );
		glVertex3f( start[0], start[1], start[2] );
		glNormal3f( normalEnd[0], normalEnd[1], normalEnd[2] );
		glVertex3f( end[0], end[1], end[2] );
	glEnd();
}

/// Render a line
void RenderingTool::renderLine( const Line& line, float32 scale ) const
{
	glLineWidth( lineThickness );
	Vector3f start = line.getPoint() - normalize( line.getDirection() ) * scale;
	Vector3f end = line.getPoint() + normalize( line.getDirection() ) * scale;
	glBegin( GL_LINES );
		glColor3f( color[0], color[1], color[2] );
		glVertex3f( start[0], start[1], start[2] );
		glVertex3f( end[0], end[1], end[2] );
	glEnd();
}

/**
 * Starting at position 'start' towards the position 'end'.
 * 'spikeLength' describes the size of the spike.
 */
void RenderingTool::renderArrow( Vector3f start, Vector3f end, float spikeScale )
{
	// Compute local coordinate system
	Matrix3f m;
	m[1] = end - start;
	float32 length = norm( m[1] );
	m[1].normalize();
	m[0] = ( fabs( XAXIS_VECTOR3F * m[1] ) > 0.9 ) ? m[1].crossProduct( YAXIS_VECTOR3F ) : m[1].crossProduct( XAXIS_VECTOR3F );
	m[2] = m[1].crossProduct( m[0] );
	m[0].normalize();
	m[1].normalize();
	m[2].normalize();
	float rotMatrix[] = { m[0][0], m[0][1], m[0][2], 0,
						  m[1][0], m[1][1], m[1][2], 0,
						  m[2][0], m[2][1], m[2][2], 0,
						  0,       0,       0,       1 };
	// Render arrow Spike
	glPushMatrix();
		glTranslatef( start[0], start[1], start[2] );
		glScalef( spikeScale * length, spikeScale * length, spikeScale * length );
		glMultMatrixf( rotMatrix );
		glCallList( dlArrowSpike );
	glPopMatrix();
}

void RenderingTool::renderCylinder( 
								   const Vector3f& center, 
								   const Vector3f& direction, 
								   const float32 radius, 
								   bool wireframe )
{
	Matrix3f m;
	Vector3f d = direction;
	d.normalize();
	m[0] = d;
	m[1] = ( fabs( XAXIS_VECTOR3F * d ) > 0.9 ) ? d.crossProduct( YAXIS_VECTOR3F ) : d.crossProduct( XAXIS_VECTOR3F );
	m[2] = m[1].crossProduct( m[0] );
	m[0].normalize();
	m[1].normalize();
	m[2].normalize();
	float rotMatrix[] = { m[0][0], m[0][1], m[0][2], 0,
		m[1][0], m[1][1], m[1][2], 0,
		m[2][0], m[2][1], m[2][2], 0,
		0,       0,       0,       1 };

	glPushMatrix();
	glTranslatef( center[0], center[1], center[2] );
	glMultMatrixf( rotMatrix );
	glScalef( radius, radius, radius );
	// Align glut cylinder: center + point to y-direction
	float alignMatrix[] = { 
		0,0,1,0,
		0,1,0,0,
		1,0,0,0,
		0,0,0,1 };
	glMultMatrixf( alignMatrix );
	glTranslatef( 0,0,-0.5f );
	if ( wireframe ){ 
		disableLighting();
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	}
	glCallList( dlCylinder );
	if ( wireframe ){
		enableLighting();
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	}
	glPopMatrix();
}

void RenderingTool::renderCylinder( const Vector3f& start, const Vector3f& end, float radius )
{
	// Compute local coordinate system
	Matrix3f m;
	m[1] = end - start;
	float32 length = norm( m[1] );
	m[1].normalize();
	m[0] = ( fabs( XAXIS_VECTOR3F * m[1] ) > 0.9 ) ? m[1].crossProduct( YAXIS_VECTOR3F ) : m[1].crossProduct( XAXIS_VECTOR3F );
	m[2] = m[1].crossProduct( m[0] );
	m[0].normalize();
	m[1].normalize();
	m[2].normalize();
	float rotMatrix[] = { m[0][0], m[0][1], m[0][2], 0,
		m[1][0], m[1][1], m[1][2], 0,
		m[2][0], m[2][1], m[2][2], 0,
		0,       0,       0,       1 };
	float alignMatrix[] = { 0,0,1,0, 0,1,0,0, 1,0,0,0, 0,0,0,1 };
	// Render cylinder
	glPushMatrix();
	glTranslatef( start[0], start[1], start[2] );
	glScalef( radius * length, radius * length, radius * length );
	glMultMatrixf( rotMatrix );
	glMultMatrixf( alignMatrix );
	glTranslatef( 0,0,-0.5f );
	glCallList( dlCylinder );
	glPopMatrix();
}

/**
 * Starting at position 'start' in direction 'normal' with length 'length'
 * 'spikeLength' describes the size of the spike.
 */
void RenderingTool::renderArrow( Vector3f start, float length, Vector3f normal, float spikeScale )
{
   renderArrow( start, start + normal * length, spikeScale  );
}

/* --------------- Create display lists ----------------------*/

/// Create display list for the arrow spike
void RenderingTool::createArrowDL()
{
	int resolution = 15;
	float alpha;
	Vector3f p, normal;
	float spikeRadius = 0.05f, shaftRadius = 0.02f, shaftLength = 0.8f;

	if ( dlArrowSpike > 0 ){
	   glDeleteLists( dlArrowSpike, 1 );
	   debugOutput << "Deleted display list with id " << dlArrowSpike << " (arrow).\n"; debugOutput.flush();
	}
	dlArrowSpike = glGenLists( 1 );
	//debugOutput << "Creating display list with id " << dlArrowSpike << " (arrow).\n"; debugOutput.flush();
	glNewList( dlArrowSpike, GL_COMPILE );
      // Spike
      glBegin( GL_TRIANGLE_FAN );
         glNormal3f( 0,1,0 );
         glVertex3f( 0,1,0 );
         for ( int n = 0; n <= resolution; n++ ){
            alpha = (float)n / (float) resolution * 2.0 * M_PI;
            p = makeVector3f( sin( alpha ), 0, cos( alpha ));
            glNormal3f( p[0], 0, p[2] );
            glVertex3f( spikeRadius * p[0], shaftLength, spikeRadius * p[2] );
         }
      glEnd();

      // Bottom of spike
      glBegin( GL_TRIANGLE_FAN );
         glNormal3f( 0, -1, 0 );
         glVertex3f( 0, shaftLength, 0 );
         for ( int n = 0; n <= resolution; n++ ){
            alpha = (float)n / (float) resolution * 2.0 * M_PI;
            p = makeVector3f( sin( alpha ), 0, cos( alpha ));
            glNormal3f( p[0], 0, p[2] );
            glVertex3f( spikeRadius * p[0], shaftLength, spikeRadius * p[2] );
         }
      glEnd();

      // Shaft
      Vector3f lastTop = makeVector3f( 0, shaftLength, shaftRadius );
      Vector3f lastBottom = makeVector3f( 0, 0, shaftRadius );
      glBegin( GL_QUADS );
         glNormal3f( 0, 0, 1 );
         for ( int n = 0; n <= resolution; n++ ){
            alpha = (float)n / (float) resolution * 2.0 * M_PI;
            normal = makeVector3f( sin( alpha ), 0, cos( alpha ) );
            p = makeVector3f( shaftRadius * normal[0], 0, shaftRadius * normal[2] );

            glVertex3f( lastTop[0], shaftLength, lastTop[2] );
            glVertex3f( lastBottom[0], 0, lastBottom[2] );

            glNormal3f( normal[0], normal[1], normal[2] );
            glVertex3f(  p[0], 0, p[2] );
            glVertex3f( p[0], shaftLength, p[2] );

            lastTop = p; lastTop[1] = shaftLength;
            lastBottom = p; lastBottom[1] = 0;
         }
      glEnd();

      // Bottom
      glBegin( GL_TRIANGLE_FAN );
         glNormal3f( 0, -1, 0 );
         glVertex3f( 0, 0, 0 );
         for ( int n = 0; n <= resolution; n++ ){
            alpha = (float)n / (float) resolution * 2.0 * M_PI;
            p = makeVector3f( sin( alpha ), 0, cos( alpha ));
            glNormal3f( p[0], 0, p[2] );
            glVertex3f( shaftRadius * p[0], 0, shaftRadius * p[2] );
         }
      glEnd();

   glEndList();
}

/// Create display list for the disc
void RenderingTool::createSphereDL()
{
   if ( dlSphere > 0 ){
	   glDeleteLists( dlSphere, 1 );
	   debugOutput << "Deleted display list with id " << dlSphere << " (sphere).\n"; debugOutput.flush();
   }
   dlSphere = glGenLists( 1 );
   //debugOutput << "Creating display list with id " << dlSphere << " (sphere).\n"; debugOutput.flush();
   glNewList( dlSphere, GL_COMPILE );
      GLUquadricObj* sphereObj = gluNewQuadric();
      gluSphere( sphereObj, 1.0, resolution, resolution );
   glEndList();
}

void RenderingTool::createCylinderDL()
{
	if ( dlCylinder > 0 ){
		glDeleteLists( dlCylinder, 1 );
		debugOutput << "Deleted display list with id " << dlCylinder << " (cylinder).\n"; debugOutput.flush();
	}
	dlCylinder = glGenLists( 1 );
	glNewList( dlCylinder, GL_COMPILE );
		GLUquadricObj* cylinderObj = gluNewQuadric();
		gluCylinder( cylinderObj, 1, 1, 1, resolution, resolution );
	glEndList();
}

/// Create display list for the disc
void RenderingTool::createDiscDL()
{
   float alpha;
   Vector3f p;
   if ( dlFilledDisc > 0 ){
	   glDeleteLists( dlFilledDisc, 1 );
	   debugOutput << "Deleted display list with id " << dlFilledDisc << " (disc).\n"; debugOutput.flush();
   }
   dlFilledDisc = glGenLists( 1 );
   //debugOutput << "Creating display list with id " << dlFilledDisc << " (disc).\n"; debugOutput.flush();
   glNewList( dlFilledDisc, GL_COMPILE );
      glBegin( GL_TRIANGLE_FAN );
         glNormal3f( 0,1,0 );
         glVertex3f( 0,0,0 );
         for ( card32 n = 0; n <= resolution; n++ ){
            alpha = (float)n/(float)resolution * 2.0 * M_PI;
            p = makeVector3f( sin( alpha ), 0, cos( alpha ) );
            glNormal3f( 0,1,0 );
            glVertex3f( p[0], 0, p[2] );
         }
      glEnd();
   glEndList();

   if ( dlDisc > 0 ){
	   glDeleteLists( dlDisc, 1 );
	   debugOutput << "Deleted display list with id " << dlDisc << " (disc).\n"; debugOutput.flush();
   }
   dlDisc = glGenLists( 1 );
   //debugOutput << "Creating display list with id " << dlDisc << " (disc).\n"; debugOutput.flush();
   glNewList( dlDisc, GL_COMPILE );
      glBegin( GL_LINE_STRIP );
         glNormal3f( 0,1,0 );
         for ( card32 n = 0; n <= resolution; n++ ){
            alpha = (float)n/(float)resolution * 2.0 * M_PI;
            p = makeVector3f( sin( alpha ), 0, cos( alpha ) );
            glNormal3f( 0,1,0 );
            glVertex3f( p[0], 0, p[2] );
         }
      glEnd();
   glEndList();
}

/**
 * Return color between red (0) - yellow (0.5) and white (1); parameter is clampled to [0;1]
 */
Vector3f RenderingTool::getRedYellowWhiteColor( float32 v )
{
   if ( v < 0.0f ) v = 0.0f;
   else if ( v > 1.0f ) v = 1.0f;

   if ( v < 0.5f )
      return makeVector3f( 1.0f, v * 2.0f, 0.0f );
   else
      return makeVector3f( 1.0f, 1.0f, 2.0f * ( v - 0.5f ) );
}

/**
 * Return color from teh whole rainbow spectrum; parameter is clamped to [0;1]
 */
Vector3f RenderingTool::getRainbowColor( float32 v )
{
   float32 t = 0;
   Vector3f color0, color1;
   Vector3f c0 = makeVector3f( 0,0,0 );
   Vector3f c1 = makeVector3f( 0,0,1 );
   Vector3f c2 = makeVector3f( 0,1,1 );
   Vector3f c3 = makeVector3f( 0,1,0 );
   Vector3f c4 = makeVector3f( 1,1,0 );
   Vector3f c5 = makeVector3f( 1,0,0 );
   Vector3f c6 = makeVector3f( 1,0,1 );
   Vector3f c7 = makeVector3f( 1,1,1 );
   if ( v > 1 ) v = 1.0;
   if ( v < 0 ) v = 0.0;
   if ( v < 0.143 )      { t = v * 7;            color0 = c0; color1 = c1; }
   else if ( v < 0.286 ) { t = ( v - 0.143) * 7; color0 = c1; color1 = c2; }
   else if ( v < 0.429 ) { t = ( v - 0.286) * 7; color0 = c2; color1 = c3; }
   else if ( v < 0.572 ) { t = ( v - 0.429) * 7; color0 = c3; color1 = c4; }
   else if ( v < 0.715 ) { t = ( v - 0.572) * 7; color0 = c4; color1 = c5; }
   else if ( v < 0.858 ) { t = ( v - 0.715) * 7; color0 = c5; color1 = c6; }
   else                  { t = ( v - 0.858) * 7; color0 = c6; color1 = c7; }

   return color0 * ( 1 - t ) + color1 * t;
}

/** 
 * Render an ellipsoid
 */
void RenderingTool::renderEllipse( const Vector3f& pos,  
								   const Matrix3f& orientation, 
								   const Vector3f& scale )
{
	float32 rotationMatrix[] = { orientation[0][0], orientation[0][1], orientation[0][2], 0,
								 orientation[1][0], orientation[1][1], orientation[1][2], 0,
								 orientation[2][0], orientation[2][1], orientation[2][2], 0,
								 0,				    0,				   0,	              1 };
	enableLighting();
	glPushMatrix();
		glTranslatef( pos[0], pos[1], pos[2] );
		glMultMatrixf( rotationMatrix );
		glScalef( scale[0], scale[1], scale[2] );
		glCallList( dlSphere );
	glPopMatrix();
	disableLighting();
}

/** 
 * Render a box
 */
void RenderingTool::renderBox( const BoundingBox3f& bbox ) const
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);	// mb - works when lighting was already turned off
	glDisable( GL_LIGHTING );
	//disableLighting();
	Vector3f p0 = bbox.lowerCorner;
	Vector3f p1 = p0 + makeVector3f( bbox.getSideLength( 0 ), 0, 0 );
	Vector3f p2 = p0 + makeVector3f( bbox.getSideLength( 0 ), bbox.getSideLength( 1 ), 0 );
	Vector3f p3 = p0 + makeVector3f( 0, bbox.getSideLength( 1 ), 0 );
	Vector3f p4 = p0 + makeVector3f( 0, 0, bbox.getSideLength( 2 ) );
	Vector3f p5 = p1 + makeVector3f( 0, 0, bbox.getSideLength( 2 ) );
	Vector3f p6 = p2 + makeVector3f( 0, 0, bbox.getSideLength( 2 ) );
	Vector3f p7 = p3 + makeVector3f( 0, 0, bbox.getSideLength( 2 ) );
	renderLine( p0, p1 );
	renderLine( p1, p2 );
	renderLine( p2, p3 );
	renderLine( p3, p0 );
	renderLine( p0, p4 );
	renderLine( p1, p5 );
	renderLine( p2, p6 );
	renderLine( p3, p7 );
	renderLine( p4, p5 );
	renderLine( p5, p6 );
	renderLine( p6, p7 );
	renderLine( p7, p4 );
	glPopAttrib();
	//enableLighting();
}

void RenderingTool::enableBlend()
{
	glDisable( GL_LIGHTING );
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
}

void RenderingTool::disableBlend()
{
	glDisable( GL_BLEND );
	glPopAttrib();
	glEnable( GL_LIGHTING );
}

void geoxl::RenderingTool::renderBox( const Matrix4f &coordinateFrame, const BoundingBox3f &bb ) const
{
	glPushMatrix();
	glMultMatrixf( coordinateFrame.data());
	renderBox(bb);
	glPopMatrix();
}