//---------------------------------------------------------------------------
#ifndef RenderingToolH
#define RenderingToolH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------


// CGTool
#include "LinearAlgebra.h"
#include "Plane.h"
#include "BivariatePolynomial.h"
#include "Polynomial.h"
//#include "SmoothCurvatureRegression.h"
#include "BoundingBox.h"

// OpenGL
#include "GL/gl.h"
#include "GL/glu.h"

namespace geoxl {

/// Provides basic rendering functionality for simple rendering primitives
class RENDERING_API RenderingTool : public Persistent {

	GEOXL_CLASS( RenderingTool )

	private:
		Vector4f color;							/// Current render color
		Vector3f lightAmbient0, lightAmbient1;	/// Ambient: light0, light1
		Vector3f lightDiffuse0, lightDiffuse1;	/// Diffuse: light0, light1
		Vector3f lightSpecular0, lightSpecular1;
		Vector4f lightPos0, lightPos1;			/// Pos: light0, light1
		card32 resolution;						/// Resolution of objects
		GLint dlArrowSpike, dlDisc, dlFilledDisc, dlSphere, dlCylinder;
		float32 lineThickness;					/// Thickness of lines
		float32 pointSplatSize;					/// Splatting size of points

		void computeMaterial();
		void createArrowDL();
		void createDiscDL();
		void createSphereDL();
		void createCylinderDL();

   public:
      /// Constructor
      RenderingTool();
      /// Destructor
      ~RenderingTool();
      /// Initialize (has to be called before proper usage of class)
      void init();
      /// Check for errors
      void checkErrors();

	  inline void setResolution( const card32 resolution ){ this->resolution = resolution; }

      void enableLighting();
      void disableLighting();
	  void setAmbientLight(Vector3f a1, Vector3f a2) {lightAmbient0 = a1; lightAmbient1 = a2;}
	  void setDiffuseLight(Vector3f d1, Vector3f d2) {lightDiffuse0 = d1; lightDiffuse1 = d2;}
	  void setSpecularLight(Vector3f s1, Vector3f s2) {lightSpecular0 = s1; lightSpecular1 = s2;}

	  void enableBlend();
	  void disableBlend();
      /// Set the rendering color for the render primitives
      void setRenderColor( float r, float g, float b );
	  void setRenderColor( float r, float g, float b, float32 alpha );
      void setRenderColor( Vector3f );
	  void setRenderColor( Vector4f );
	   /// Set splatting size for point primitives
	   inline void setPointSplatSize( float s ){ pointSplatSize = s; glPointSize( pointSplatSize ); }
	   inline void setLineThickness( float32 s ){ lineThickness = s; glLineWidth( s ); }


      /// Render a disc
      void renderDisc( Vector3f center, Vector3f normal, float radius, bool filled = true );
      /// Render an arrow
      void renderArrow( Vector3f start, Vector3f end, float spikeLength );
      /// Render an arrow
      void renderArrow( Vector3f start, float length, Vector3f normal, float spikeLength );
      /// Render a line
      void renderLine( const Vector3f start, const Vector3f end, const Vector3f normalStart = makeVector3f( 1,0,0 ), const Vector3f normalEnd = makeVector3f( 1,0,0 ) ) const;
      /// Render a line
      void renderLine( const Line& line, float32 scale ) const;
	  /// Render a simple point
	  void renderPoint( Vector3f p ) const;
	  /// Render a plane
	  void renderPlane( Plane p, float size );
     /// Render a convex Hull
	  //void renderConvexHull2D( ConvexHull2D ch, float32 scale );
     /// Return color between red (0) - yellow (0.5) and white (1); parameter is clampled to [0;1]
     static Vector3f getRedYellowWhiteColor( float32 v );
     /// Return color from the whole rainbow spectrum; parameter is clamped to [0;1]
     static Vector3f getRainbowColor( float32 v );
     /// Render a sphere
     void renderSphere( Vector3f pos, float radius, bool wireframe = false );
	 /// Render a cylinder
	 void renderCylinder( const Vector3f& center, const Vector3f& direction, const float32 radius, bool wireframe );
	 /// Render an cylinder
	 void renderCylinder( const Vector3f& start, const Vector3f& end, const float radius );
	 /// Render an ellipsoid
	 void renderEllipse( const Vector3f& pos, 
		                 const Matrix3f& orientation, 
						 const Vector3f& scale );
	 /// Render a box
	 void renderBox( const BoundingBox3f& bbox ) const;

	 /// Render a bounding box aligned to a given coordinate system
	 void renderBox( const Matrix4f &coordinateFrame, const BoundingBox3f &bb ) const;
};

}

#endif
