//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "GLVolumeDisplayer.h"
//---------------------------------------------------------------------------
#include "GLEventDisplayerWidget.h"
#include "DebugOutput.h"
#include "PointSet.h"
#include "PointSet_templCode.h"
#include "VertexDescriptor.h"
#include "RegularVolume.h"
#include "StringHelper.h"
#include "RegularData_templCode.h"
#include "GLVolumeRenderer2DSlices.h"
#include "Tex2dVolumeRenderer.h"
#include "GLPointCloudDisplayer.h"
#include "Random.h"
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------

IMPLEMENT_GEOXL_CLASS(GLVolumeDisplayer, 0) 
{
}

// ----

void GLVolumeDisplayer::initialSetup() {
   renderer = new GLVolumeRenderer2DSlices;
	textureStack = new Tex2dVolumeRenderer();
}

GLVolumeDisplayer::GLVolumeDisplayer(RegularVolume *vol, string xAxisLabel, string yAxisLabel, string zAxisLabel) {
	initialSetup();
	this->xAxisLabel = xAxisLabel;
	this->yAxisLabel = yAxisLabel;
	this->zAxisLabel = zAxisLabel;
	volume = vol;
	if (vol !=0) {
		if (vol->getVoxelData()->getDescr()->getNumAttribs() > 0) {
		   string channel = vol->getVoxelData()->getDescr()->getAttribute(0).name;
			renderer->setChannel(channel);
			volume->computeMinMax1f(channel, minValue, maxValue);
		}
	}
}

GLVolumeDisplayer::GLVolumeDisplayer(const RegularDataF *vol, string xAxisLabel, string yAxisLabel, string zAxisLabel) {
	initialSetup();
	this->xAxisLabel = xAxisLabel;
	this->yAxisLabel = yAxisLabel;
	this->zAxisLabel = zAxisLabel;
	pAssert(vol != NULL);
	if (vol == NULL) return;
	DVectorI extends = vol->getExtends();
	pAssert(extends.getDim() == 3);
	if (extends.getDim() != 3) return;
	
	volume = new RegularVolume();
	VertexDescriptor vd; 
	vd.pushAttrib(VAD("intensity",1, VAD::DATA_FORMAT_FLOAT32));
	BoundingBox3f bb = BoundingBox3f(NULL_VECTOR3F, makeVector3f(extends[0], extends[1], extends[2]));
	volume->clearAndSetup(extends[0], extends[1], extends[2], &vd, bb);
	AAT aat = vd.getToken("intensity");
	PointSet *ps = volume->getVoxelData();
	card32 numVoxels = extends[0]*extends[1]*extends[2];
	const float32 *source = vol->dataPtr();
	for (card32 i=0; i<numVoxels; i++) {
		StaticVector<float32, 1> v;
		v[0] = *source;
		ps->set<float32, 1>(i, aat, v);
		source++;
	}
	volume->computeMinMax1f("intensity", minValue, maxValue);
   renderer->setChannel("intensity");	
}

GLVolumeDisplayer::GLVolumeDisplayer(
	const PointSet *ps, card32 width, card32 height, card32 depth, 
	string xAxisLabel, string yAxisLabel, string zAxisLabel)
{
	initialSetup();
	this->xAxisLabel = xAxisLabel;
	this->yAxisLabel = yAxisLabel;
	this->zAxisLabel = zAxisLabel;
	pAssert(ps != NULL);
	if (ps == NULL) return;
	volume = new RegularVolume();
	BoundingBox3f bb = BoundingBox3f(NULL_VECTOR3F, makeVector3f(width, height, depth));
	volume->clearAndSetup(width, height, depth, ps->getDescr(), bb);
	PointSet *vps = volume->getVoxelData();
	vps->assign(ps);
	if (ps->getDescr()->getNumAttribs() > 0) {
   	string channel;
		channel = ps->getDescr()->getAttribute(0).name;
   	volume->computeMinMax1f(channel, minValue, maxValue);
	}
}


BoundingBox3f GLVolumeDisplayer::getBoundingBox() {
	if (volume == NULL) return BoundingBox3f(makeVector3f(0,0,0), makeVector3f(1,1,1));
	return volume->getBoundingBox();
}

GLEventConfiguration *GLVolumeDisplayer::createConfigWindow(QWidget *parent) {
	return new GLVolumeDisplayerConfig(parent, this, xAxisLabel.c_str(), yAxisLabel.c_str(), zAxisLabel.c_str());
}

void GLVolumeDisplayer::updateData(GLEventObject *other) {
	GLVolumeDisplayer *cOther = dynamic_cast<GLVolumeDisplayer*>(other);
	if (cOther != NULL) {
		delete volume;
		volume = cOther->volume;
		minValue = cOther->minValue;
		maxValue = cOther->maxValue;		
		cOther->volume = NULL;
      renderer->invalidate();
	}
}

void GLVolumeDisplayer::drawGL(QGLWidget *widget, Camera &cam) {
   renderer->drawGL(volume, textureStack, cam, minValue, maxValue);
}


GLVolumeDisplayer::~GLVolumeDisplayer() {
	delete renderer;
	delete textureStack;
}

// --


GLVolumeDisplayerConfig::GLVolumeDisplayerConfig(QWidget *parent, GLEventObject *obj, QString xAxisLabel, QString yAxisLabel, QString zAxisLabel) 
: GLEventConfiguration(parent, obj)
{
	ui.setupUi(this);
	ui.lbXAxis->setText(xAxisLabel);
	ui.lbYAxis->setText(yAxisLabel);
	ui.lbZAxis->setText(zAxisLabel);
	GLVolumeDisplayer *vObj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (vObj != NULL) {
		if (vObj->getVolume() != NULL) {
			deactivateChannelChoice = true;
			ui.cbChannel->clear();
			const VertexDescriptor *vd = vObj->getVolume()->getVoxelData()->getDescr();
			for (card32 i=0; i<vd->getNumAttribs(); i++) {
				ui.cbChannel->addItem(QString(vd->getAttribute(i).name.c_str()));
			}
			ui.cbChannel->setCurrentIndex(0);
			deactivateChannelChoice = false;
		}
	}
}	

void GLVolumeDisplayerConfig::on_cbShowCoords_stateChanged(int) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) pcobj->renderer->setShowCoords(ui.cbShowCoords->isChecked());
	emit changed();
}

void GLVolumeDisplayerConfig::on_rbOpacityWhite_clicked(bool) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setTransferFunction(0);
	}
	emit changed();
}
void GLVolumeDisplayerConfig::on_rbOpacityIntensity_clicked(bool) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setTransferFunction(1);
	}
	emit changed();
}
void GLVolumeDisplayerConfig::on_rbOpacityBlueRed_clicked(bool) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setTransferFunction(2);
	}
	emit changed();
}

void GLVolumeDisplayerConfig::on_dsbOffset_valueChanged(double v) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setOffset(v);
	}
	emit changed();
}

void GLVolumeDisplayerConfig::on_dsbScale_valueChanged(double v) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setScale(v);
	}
	emit changed();
}

void GLVolumeDisplayerConfig::on_dsbGamma_valueChanged(double v) {
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setGamma(v);
	}
	emit changed();
}

void GLVolumeDisplayerConfig::on_cbAutomatic_clicked(bool a) {
	ui.dsbOffset->setEnabled(!a);
	ui.dsbScale->setEnabled(!a);
	GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
	if (pcobj != NULL) {
		pcobj->renderer->setAutomatic(a);
		emit changed();
	}
}

void GLVolumeDisplayerConfig::on_cbChannel_currentIndexChanged(const QString &text) {
	if (!deactivateChannelChoice) {
		GLVolumeDisplayer *pcobj = dynamic_cast<GLVolumeDisplayer*>(obj);
		if (pcobj != NULL) {
			pcobj->renderer->setChannel(qString2STLString(text));
		}
	}
}



void geoxl::glEventDisplayer_unitTest() {
	{
		PointSet *ps = new PointSet();
		VertexDescriptor vd;
		vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
		ps->clearAndSetup(1,0, &vd);
		for (card32 i=0; i<500; i++) {
			Vector3f v = makeVector3f(rnd01(), rnd01(), rnd01());
			ps->addPoint((char*)&v);
		}
		GLPointCloudDisplayer *pcdisp = new GLPointCloudDisplayer(ps);
		GLEventDisplayer::SetNewGLEventObject ev(pcdisp);
		debugOutput.displayEvent(GLEventDisplayer::getClass(), "glEventDispTest", &ev);
	}
	{
		RegularVolume *testVol = new RegularVolume();
		const card32 SIZE = 64;
		VertexDescriptor vd;
		vd.pushAttrib(VAD("intensity", 1, VAD::DATA_FORMAT_FLOAT32));
		testVol->clearAndSetup(SIZE, SIZE, SIZE, &vd, BoundingBox3f(makeVector3f(0,0,0), makeVector3f(1,1,1)));
		
		PointSet *ps = testVol->getVoxelData();
		AAT INTEN = ps->getAAT("intensity");
		
		for (card32 z=0; z<SIZE; z++) {
			for (card32 y=0; y<SIZE; y++) {
				for (card32 x=0; x<SIZE; x++) {
					StaticVector<float32, 1> v;
					float32 xf = (x - SIZE/2.0f) / (SIZE/4.0f);
					float32 yf = (y - SIZE/2.0f) / (SIZE/5.0f);
					float32 zf = (z - SIZE/2.0f) / (SIZE/6.0f);
					v[0] = exp(-xf*xf-yf*yf-zf*zf)*0.5*(sin(4*xf)+1)*0.5*(sin(5*yf)+1)*0.5*(sin(6*zf)+1)
					       + ((float32)z/SIZE)*0.01;
					ps->set<float32, 1>(x + SIZE*y + SIZE*SIZE*z, INTEN, v);
				}
			}
		}
															
		GLVolumeDisplayer *voldisp = new GLVolumeDisplayer(testVol);
		GLEventDisplayer::SetNewGLEventObject ev2(voldisp);
		debugOutput.displayEvent(GLEventDisplayer::getClass(), "glVolDispTest", &ev2);
	}
}


#include "GLVolumeDisplayer_moc.cpp"
