//---------------------------------------------------------------------------
#ifndef POINT_CLOUD_TOOLS_H
#define POINT_CLOUD_TOOLS_H
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "UnstructuredInCorePointCloud.h"
#include "RegressionPlane.h"
#include "HierarchicalKNNIterator.h"

#include "Scene.h"
#include "SGLegacyPCSceneAdaptor.h"
//---------------------------------------------------------------------------

#define NUM_NEIGHBORS_LOCAL_CONTROL_PLANE 12

namespace geoxl {

   /// Find point cloud in scene
   SCENEGRAPH_API PointCloud*  getPointCloud( Scene* scene, string pcID );

   /// Find any scene object
   SCENEGRAPH_API SceneObject*  getSceneObject( Scene* scene, string pcID );

	/// Get transformation of scene graph node by name, along with the node itself; parameter node is NULL if not found
   SCENEGRAPH_API void getSGNodeTransformation(Scene *scene, string nodeID, Matrix4f &transformation, SceneGraphNode *&node);

   /// Replace point cloud 'pcID' with 'newPC'
   SCENEGRAPH_API void  replacePointCloud( Scene* scene, string pcID, PointCloud* newPC );

   /// Replace point cloud 'pcID' with 'newPC' or add if it doesn't exist
   SCENEGRAPH_API void  addOrReplacePointCloud( Scene* scene, string pcID, PointCloud* newPC );

   /// Add point cloud 'pc' as a child of the root node; set node name to 'name'.
   SCENEGRAPH_API void  addPointCloud( Scene* scene, PointCloud* pc, string name );

	/// Add scene object 'so' as a child of the root node; set node name to 'name'. Returns created ObjectNode.
	SCENEGRAPH_API SceneGraphNode* addSceneObject( Scene * scene, SceneObject *so, string name );


   /// Add point cloud 'pc' as a child of the root node; set node name to 'name'; returns really set name because this can differ from the requested name. always use this returned name afterwards!
   SCENEGRAPH_API string  addPointCloudAndReturnNodeName( Scene* scene, PointCloud* pc, string name );

   /// Check for attribute: Search attribute 'attrib' in point cloud 'id' in scene 'pcscn'. If it does not exist, create attribute as (numComponents, dataType)
   SCENEGRAPH_API void  checkAttribute( Scene* pcscn, string id, string attrib, card8 numComponents, card8 dataType ) ;

   /// Check for attribute: Search attribute 'attrib' in point cloud pc. If it does not exist, create attribute as (numComponents, dataType)
   SCENEGRAPH_API void  checkAttribute( PointCloud * pc, string attrib, card8 numComponents, card8 dataType ) ;

	/// CheckAttribute: Check for (position, normal, flags, color)
   SCENEGRAPH_API void  checkSceneAllAttributes( Scene* scene );

   /// Convert point cloud to unstructured point cloud, whatever the type was before
   SCENEGRAPH_API UnstructuredInCorePointCloud*  convertToUnstructuredInCorePointCloud( PointCloud* inPC );

   /// Return transformation that automatically aligns the point cloud
   SCENEGRAPH_API Matrix4f  getAutoAlignMatrix( PointCloud* pc );

   /// Returns resampled point cloud
   SCENEGRAPH_API PointCloud*  createResampled( PointCloud *input, card32 x, card32 y );


   /// returns NULL if the scene does not containe a node named "name".
   SCENEGRAPH_API SceneGraphNode*  getSceneGraphNodeByName(Scene *scene, string name);

   /// Return point cloud from node; returns NULL if the node does not contain a point cloud
   SCENEGRAPH_API PointCloud*  getPointCloud( SceneGraphNode *node );

   /// Creates a random permutation of points in a point set
   SCENEGRAPH_API void randomlyPermutePointSet(PointSet *ps);

   /// Wwaps two points from a PointSet
   SCENEGRAPH_API void swapPSPoints(PointSet *ps, mpcard i1, mpcard i2);


   /// Return the bounding box of all scene point clouds
   SCENEGRAPH_API BoundingBox3f  getScenePCBBox( Scene* scene );

   /// Return the bounding box of a point cloud
   SCENEGRAPH_API BoundingBox3f  getPCBBox( PointCloud* pc );

	/// Return the average distance between two points of a point cloud
	SCENEGRAPH_API float getAveragePointDistance( PointCloud *pc, card32 numDirectNeighbours = 5 );

	/// returns median distance between two points (numSamples == 0 -> all points)
   SCENEGRAPH_API float getMedianPointDistance( PointCloud *pc, card32 numDirectNeighbours = 5, card32 numSamples = 1000 );
	/// returns median distance between two points (stores result in attachments, reuses if attachment exists)
	SCENEGRAPH_API float getCachedMedianPointDistance( PointCloud * pc, card32 numDirectNeighbours = 5, card32 numSamples = 1000 );

	/// Delete node with name and path 'id' from scene. Returns true on success.
   SCENEGRAPH_API bool  deleteSceneGraphNode( Scene* scene, string id);
   /// Delete point cloud 'pcID' from scene. Returns true on success.
   SCENEGRAPH_API bool  deletePointCloud( Scene* scene, string pcID);


   /// Compute the regression plane from closest points
   SCENEGRAPH_API void  getLocalPlane( RegressionPlane& plane, Vector3f pos, PointCloud* pc, HierarchicalKNNIterator* pcNNIt );

   /// Compute the regression plane from closest points
   SCENEGRAPH_API void  getLocalPlane( RegressionPlane& plane, Vector3f pos, float32 lfs, PointCloud* pc, HierarchicalKNNIterator* pcNNIt );
   
   /// retrieve transformation matrix of a scene graph node
   SCENEGRAPH_API Matrix4f getTransformation(Scene *scene, string node);

	/// smoothes attribute "inputChannel" and stores the (normalized?) result in "outputChannel"
	/// uses a gaussian filter kernel exp( -distance^2/ h^2), distance will be scaled by 1/averagePointDistance
	/// averagePointDistance will be estimated when not specified 
	/// this function doesn't perform smoothing in a two step way, therefore: inputChannel != outputChannel
	/// (otherwise there will be different results)
	SCENEGRAPH_API void smoothAttribute1f( PointCloud * pc, string inputChannel, string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	SCENEGRAPH_API void smoothAttribute2f( PointCloud * pc, string inputChannel, string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	SCENEGRAPH_API void smoothAttribute3f( PointCloud * pc, string inputChannel, string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	SCENEGRAPH_API void smoothAttribute4f( PointCloud * pc, string inputChannel, string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );


	/// create a vertexdescriptor
	SCENEGRAPH_API VertexDescriptor * createDefaultVertexDescriptor( bool position, bool color, bool normal, bool flags=true, bool tangents=false );

	/// merge two point clouds
	/// vertex descriptors should match!
	SCENEGRAPH_API void merge( PointCloud * pc1, const PointCloud * pc2);
}

#endif