#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ProcessPointsCmdObj.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "StringClassProperty.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "PointCloudIterators.h"
#include "HierarchicalKNNIterator.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "ProgressWindow.h"
//---------------------------------------------------------------------------
using namespace geoxl;

IMPLEMENT_GEOXL_ABSTRACT_CLASS_GENERIC( ProcessPointsCmdObj ,0, NULL )
{
	BEGIN_CLASS_INIT( ProcessPointsCmdObj );

	ADD_BOOLEAN_PROP( m_CreateHierarchicalIterator, 0 );
	ADD_BOOLEAN_PROP( m_InvalidateAttachments, 0 );
	ADD_CARD32_PROP( m_SelectionMode, 0 );

	ADD_STRING_PROP( m_SourceChannel, 0 );
	ADD_CARD32_PROP( m_SourceCompIndex, 0 );

	ADD_STRING_PROP( m_DestChannel, 0 );
	ADD_CARD32_PROP( m_DestCompIndex, 0 );
}

ProcessPointsCmdObj::ProcessPointsCmdObj(void)
{
	m_CreateHierarchicalIterator = true;
	m_InvalidateAttachments = true;
	m_SelectionMode = SM_SelectedOrAll;

	m_SourceCompIndex = 0;
	m_DestCompIndex = 0;
}

ProcessPointsCmdObj::~ProcessPointsCmdObj(void)
{
}

void ProcessPointsCmdObj::assign(const Object* obj, CopyContext *context)
{
	CommandObject::assign(obj, context);

	const ProcessPointsCmdObj * o = dynamic_cast<const ProcessPointsCmdObj*>(obj);
	pAssert( o != NULL );

	m_CreateHierarchicalIterator = o->m_CreateHierarchicalIterator;
	m_InvalidateAttachments = o->m_InvalidateAttachments;
	m_SelectionMode = o->m_SelectionMode;

	m_SourceChannel = o->m_SourceChannel;
	m_SourceCompIndex = o->m_SourceCompIndex;
	m_DestChannel = o->m_DestChannel;
	m_DestCompIndex = o->m_DestCompIndex;

}
void ProcessPointsCmdObj::processAllPoints( Scene * scene, void * context, PrepareProcessMethod prepare, ProcessPointMethod process ) 
{
	vector<TPCInfo> pcs = getCurrentPointClouds(scene);
	if( pcs.size() == 0 )
		return;

	bool aborted = false;

	// setup progress window
	ProgressWindow::WeightList steps;
	float numPoints = 0;
	for( unsigned i=0;i<pcs.size();i++ )
		numPoints += pcs[i].pc->getNumPoints();
	for( unsigned i=0;i<pcs.size();i++ )
	{
		steps.push_back( ProgressWindow::WeightElem(true, (float)pcs[i].pc->getNumPoints()/numPoints));
	}
	progressWindow->pushSteps( steps );


	// loop over point clouds
	for( unsigned i=0;i<pcs.size() && !aborted;i++ )
	{

		// -----------------------------
		// setup scene and pc info
		PointCloud * pc = pcs[i].pc;
		PointInfo info;
		info.pc = pc;
		info.scene = scene;
		info.pointCloudName = pcs[i].pcName;
		info.index = 0;

		// -----------------------------
		// setup iterators
		SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
																		SceneObjectIteratorType::CAP_MODIFYABLE_PC );
		pAssert( iter );

		info.bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		info.mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

		if( m_CreateHierarchicalIterator )
			info.hIt = new HierarchicalKNNIterator( pc, 32, NULL );
		else
			info.hIt = NULL;

		if( pc->providesAttribute( "position" ) )
			info.posAAT = pc->getAAT( "position" );
		if( pc->providesAttribute( "normal" ) )
			info.normalAAT = pc->getAAT( "normal" );
		if( pc->providesAttribute( "color" ) )
			info.colorAAT = pc->getAAT( "color" );
		if( pc->providesAttribute( "flags" ) )
			info.flagsAAT = pc->getAAT( "flags" );

		info.sourceCompIndex = m_SourceCompIndex;
		if( m_SourceChannel.length() > 0 )
			if( pc->providesAttribute( m_SourceChannel ) )
			{
				info.sourceAAT = pc->getAAT( m_SourceChannel );
				info.sourceVAD = pc->getDescr()->getAttribute( m_SourceChannel );
			}
		info.destCompIndex = m_DestCompIndex;
		if( m_DestChannel.length() > 0 )
			if( pc->providesAttribute( m_DestChannel ) )
			{
				info.destAAT = pc->getAAT( m_DestChannel );
				info.destVAD = pc->getDescr()->getAttribute( m_DestChannel );
			}

		// -----------------------------
		// call prepare method
		if( prepare != NULL )
			(this->*prepare)(&info,context);

		// -----------------------------
		// point loop
		while( !info.bIt->atEnd() && !aborted )
		{
			(this->*process)(&info,context);

			info.bIt->next();
			aborted = !progressWindow->progress( (card32)((float)(info.index)/(float)(pc->getNumPoints()-1)*100.0f ));
			info.index++;
		}

		if( m_InvalidateAttachments )
			pc->clearAttachments();

		if( i + 1 < pcs.size() )
			progressWindow->nextStep(aborted);
		delete info.hIt;
	}
	progressWindow->popSteps(aborted);

}

vector<ProcessPointsCmdObj::TPCInfo>	ProcessPointsCmdObj::getCurrentPointClouds(Scene * scene)
{
	vector<TPCInfo> ret;
   vector<string> selection;
	if( m_SelectionMode == SM_Selected || m_SelectionMode == SM_SelectedOrAll )
		selection = scene->getRootState()->staticState->getSelectedNodes();

	if( selection.size() == 0 )
	{
		if( m_SelectionMode == SM_SelectedOrAll || m_SelectionMode == SM_All ) 
		{
			// collect all point clouds
			StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
			it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
			while ( !it->atEnd() )
			{
				SGObjectNode* sgon = dynamic_cast<SGObjectNode*>( it->getNode() );
				if( sgon )
				{
					PointCloud * pc = dynamic_cast<PointCloud*>( sgon->getSceneObject() );
					if( pc )
					{
						selection.push_back( it->getNameWithPath() );
					}
				}
				it->next();
			}
		}
	}

	// check for required attributes
	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		for( unsigned i=0;i<m_RequiredChannels.size();i++ )
			checkAttribute( scene, selection[m], m_RequiredChannels[i].name,
			m_RequiredChannels[i].numComponents, m_RequiredChannels[i].dataFormat );
	}


	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		TPCInfo info;
		info.pc = getPointCloud( scene, selection[m] );
		info.pcName = selection[m];
		ret.push_back( info );
	}

	return ret;
}
