//---------------------------------------------------------------------------
#ifndef SceneH
#define SceneH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "CommandObject.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace geoxl {

class SceneGraphState;
class SceneGraphNode;
class AttachedDataContainer;
class GLMaterial;
class LegacyPointCloudSceneReadHandler;


/// A scene described by a scene graph.
/// The root node is the root of the scene graph (typically a SGListNode).
/// The root state contains global settings (the dynamic part can be changed for subnodes).
/// In addition, any attachments can be attached to the scene.
class SCENEGRAPH_API Scene : public Persistent,
                      public CommandObjectContextInterface {
 private:
   static MetaClass *SceneMetaClass;
   static Object* newInstance();

   SceneGraphState *rootState;
   SceneGraphNode *rootNode;
   vector<GLMaterial*> glMaterials;
   AttachedDataContainer *attachments;
   CommandHistory *commandHistory;

   // ---property access methods---
   GLMaterial *getGLMat(card32 index) const;
   void setGLMat(card32 index, GLMaterial *m);
   mpcard getNumGLMats() const;
   void setNumGLMats(card32 size);
   void pushGLMat();
   void popGLMat();
   // -----------------------------

   void addMaterial( Object * obj );
   void deleteMaterials( card32 begin, card32 end );
   Object * getMaterial( card32 index );
   void moveMaterial( card32 oldPos, card32 newPos );
   card32 getNumMaterials();
   string getMaterialName(card32 index );

	// high quality renderer material
	GLMaterial * m_HQRMaterial;


 public:
   Scene();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();
	
	void read(InputObjectStream *s);

   /// setup new scene graph; ownership of rootState and rootNode will be transfered to the scene object.
   virtual void clearAndSetup(SceneGraphState *rootState, SceneGraphNode *rootNode);

   /// get root state of scene graph
   SceneGraphState *getRootState();
   /// get root node of scene graph
   SceneGraphNode *getRootNode();
   /// set new root state, old state will be deleted
   void setRootState(SceneGraphState *r);
   /// set new root node (old node will be reference-decremented)
   void setRootNode(SceneGraphNode *r);

   /// add material to gl-materials library; ownership will be transfered to this scene
   void addGLMaterial(GLMaterial *newMat);
   /// get number of materials
   mpcard getNumGLMaterials();
   /// get material with index i
   GLMaterial *getGLMaterial(card32 i);
   /// delete material with given index
   void deleteGLMaterial(card32 i);
   /// get material index based on its name, returns first occurence 
   /// (hence if 2 same materials the secondf one will never returned, until first one exists)
   /// returns MAX_INT if no such material found
   card32 getGLMaterialIndex(const std::string& name);
   
	GLMaterial * getHQRMaterial() {return m_HQRMaterial;}
	void setHQRMaterial( GLMaterial * mat ) {m_HQRMaterial=mat;}

   /// get attachments
   inline AttachedDataContainer *getAttachments() {return attachments;}

   // --- command history interface
   virtual void clear();
   virtual CommandHistory *getCommandHistory();
   virtual void addMacro(CommandObjMacro* macro, string name);

   virtual ~Scene();

};


}

#endif
