//---------------------------------------------------------------------------
#ifndef SceneGraphNodeH
#define SceneGraphNodeH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "Tools.h"
//---------------------------------------------------------------------------

namespace geoxl {


class SGStateChanger;
class SceneGraphState;

/// A scene graph node. provides abstract interface for enumerating descenendent nodes.
/// Additionally, reference counting is implemented (accessible through protected methods add/releaseReference).
/// Override getNumChildNodes(), getChildNode() in concrete subclasses.
class SCENEGRAPH_API SceneGraphNode : public Persistent {
 private:
   static MetaClass *SceneGraphNodeMetaClass;
   static Object* newInstance();

   bool findNameInParents(const string name) const;

 protected:
   string name;
   card32 flags;
   vector<SceneGraphNode*> parents;

   string makeUniqueName(string name) const;

 public:
   SceneGraphNode();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   // --- reference counting

   /// put this object in another container (e.g. list); use NULL for root nodes (multiple entries are allowed).
   void addReference(SceneGraphNode *parent);
   /// remove this object from another container. deletes this if count is zero; use NULL for root nodes (multiple entries are allowed).
   void releaseReference(SceneGraphNode *parent);
   /// query number of references, e.g. to identify instantiated nodes
   max_platform_card getNumReferences();

   // --- naming

   /// get a name / identifier. the name is kept unique among all other siblings in parent nodes. numbers are added if necessary.
   virtual string getName() const;
   /// set a new name.
   virtual void setName(const string name);


   // --- hierarchy  

   /// get number of child nodes. might be zero.
   /// might depend on state - if traversalMode is set to tmAllOperands all nodes are reported.
   /// the same action (report all nodes) is also performed if state == NULL.
   virtual max_platform_card getNumChildNodes(const SceneGraphState *state) const = 0;
   /// get child node. might be NULL if node is not available (or >= max)
   /// might depend on state - if traversalMode is set to tmAllOperands all nodes are reported.
   /// the same action (report all nodes) is also performed if state == NULL.
   virtual SceneGraphNode *getChildNode(const SceneGraphState *state, max_platform_card i) = 0;
   /// remove child node. Indexed by name to be independent of scene graph state.
   virtual void removeChildNode(string name) = 0;
   /// if true, another child node can be added to this node
   virtual bool canAddChildNode() = 0;
   /// add a child node to this node; return value is true if node has been accepted (e.g. due to its type)
   virtual bool addChildNode(SceneGraphNode *node) = 0;



   // --- state management

   /// use this method to change for state changes associated with this node
   /// if there are any, the method creates and returns a corresponding state changer object (NULL otherwise)
   /// override in descendants to define a certain state change for this node
   /// this state change takes effect _after_ retrieving the child node from the parent node by get[Num]ChildNode[s]()
   virtual SGStateChanger *getNodeEntryStateChange(const SceneGraphState *state) {return NULL;}

   /// use this method to change for state changes associated with entering a certain child node of this node
   /// if there are any, the method creates and returns a corresponding state changer object (NULL otherwise)
   /// override in descendants to define state changes for descendant nodes
   /// this state change takes effect _after_ retrieving the child node by get[Num]ChildNode[s]()
   virtual SGStateChanger *getChildNodeEntryStateChange(const SceneGraphState *state, max_platform_card i) {forget(i); return NULL;}



   // --- flags

   /// flag indicating that the current node / subtree is visible
   static const card32 FLAG_VISIBLE = 1;
   /// flag indicating that the current node has been traversed before (use only temporary,
   /// must be unset after a operation has taken place)
   static const card32 FLAG_VISIT = 2;

   /// set node flags
   void setFlags(card32 flags) {this->flags = flags;}
   /// get node flags
   card32 getFlags() const {return flags;}
   /// remove flags from current state
   void excludeFlags(card32 mask) {this->flags &= (0xFFFFFFFF ^ mask);}
   /// include flags in addition to current state
   void includeFlags(card32 mask) {this->flags |= mask;}

   /// special case - test for visibility flag
   bool visible() const {return (flags & FLAG_VISIBLE) != 0;}
   /// special case - set visibility flag
   void setVisible(bool v);
   
   
   
   // --- visualization
   /// this method can be redefined to provide special, node type (and state) dependent visualizations
   /// renderers may call this method (typically if configured so by the user)
   virtual void renderNodeGL(const SceneGraphState *state) {}

   /// --- auxiliary routines
   static bool nodeReachable(SceneGraphNode *node, SceneGraphNode *searchFrom);

   virtual ~SceneGraphNode() {}

};


}

#endif
