//---------------------------------------------------------------------------
#ifndef SceneGraphStateH
#define SceneGraphStateH
//---------------------------------------------------------------------------
#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "Camera.h"
#include "ViewFrustum.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------
#include <set>
#include <string>
#include <hash_map>
//---------------------------------------------------------------------------


namespace geoxl {
   using namespace std;
   using namespace stdext;
   
extern SCENEGRAPH_API Vector4f sgStateStandardBackgroundColor;
   
   
class SGAllocatorList;

/// Type for describing the static scene graph state (global parameters which may not be changed from node to node).
class SCENEGRAPH_API SGStaticState : public Persistent {
 private:
   static MetaClass *SGStaticStateMetaClass;
   static Object* newInstance();

   string getSelectionString();
   
 public:
   // --- state variables

   /// current camera
   Camera camera;
   /// current view frustum
   ViewFrustum viewFrustum;
   /// background color
   Vector4f backgroundColor;
   /// current selection
   set<string> selectedNodes;
   /// list of allocators for storing data out-of-core (todo)
   SGAllocatorList *allocators;
   /// flags that control node render actions
   bool rfShowLocalCoords;

   /// Traversal mode Type: see SGDynamicState::TraversalModeType for description
	/// please update SGDynamicState::TraversalModeType also if you add a new type here
   enum TraversalModeType {tmAllNodes = 0,
                           tmRendering = 1};
   /// default traversal mode - affects the behavior of scene graph nodes during traversal.
	/// here the default value for scene graph traversion can be set, e.g. if tools should work on all frames of an animation node or just the actual frame.
	/// remember that this value is only used, if SGDynamicState is set to tmDefault!
   TraversalModeType defaultTraversalMode;

   // --- methods

   SGStaticState();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual bool equal(const Object* obj) const;

   void write(OutputObjectStream *s) const;
   void read(InputObjectStream *s);

   /// Return vector of names of selected nodes
   vector<string> getSelectedNodes();

   ~SGStaticState();
};



/// Type for describing the dynamic scene graph state (local parameters which might be changed hierarchically from node to node).
class SCENEGRAPH_API SGDynamicState : public Persistent {
 private:
   static MetaClass *SGDynamicStateMetaClass;
   static Object* newInstance();

   // --- private implementation 
   typedef hash_map<string, Vector4f> StrHashTable;
   StrHashTable stateVariables4F;
   
 public:
   // --- state variables

   /// current transformation matrix
   Matrix4f ctm;
   /// current time in seconds
   float32 time;

   /// Traversal mode Type:
   /// tmAllNodes enforces all operands of each scene graph node to be displayed (i.e. all nodes referenced by the current node)
   /// tmRendering shows the nodes that are meant to be rendered for the current frame according to the current state.
   /// These could be more or less nodes than the referenced nodes. (more --> e.g. in instantiation, less --> e.g. in animation nodes)
	/// tmDefault says that the (global) defaultTraversalMode located in SGStaticState should be used
	/// tmDefault should always be the last value for compatibility purposes
	/// please update SGStaticState::TraversalModeType also if you add a new type here
   enum TraversalModeType {tmAllNodes = 0,
                           tmRendering = 1,
									tmDefault = 2};
   /// traversal mode - affects the behavior of scene graph nodes during traversal. chose one before traversing the graph.
   /// mw - traversal mode now in dynamic state to avoid changes in iteterators create global inconsistencies (static state is shared)
   TraversalModeType traversalMode;

   // -- indirect access
   void setStateVariable(string name, Vector4f vectValue);
   void deleteStateVariable(string name);
   bool getStateVariable(string name, Vector4f &vectValue);
   vector< pair<string, Vector4f> > getStateVariables4F();

   // --- methods

   SGDynamicState();

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, CopyContext *context = NULL);
   virtual bool equal(const Object* obj) const;

   void write(OutputObjectStream *s) const;
   void read(InputObjectStream *s);

   ~SGDynamicState() {}
};



/// Scene graph state object - describes parameters to the scene graph nodes.
/// Gobal parameters that are the same for all nodes are described by the staticState.
/// Local parameters that can be changed for subtrees in the graph are described by the dynamicState.
class SCENEGRAPH_API SceneGraphState : public Persistent {
 private:
   static MetaClass *SceneGraphStateMetaClass;
   static Object* newInstance();

   bool traversalCopy;

 public:
   SceneGraphState();

   /// global parameters, may not change from node to node
   SGStaticState *staticState;
   /// local parameters, might be changed hierarchically from node to node
   SGDynamicState *dynamicState;

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// assignes a copy to this node (use createTraversalState to create traversal copy)
   virtual void assign(const Object* obj, CopyContext *context = NULL);
   /// Creates a copy of this state object; the dynamic state is copied, the static state is referenced.
   /// The dynamic state can thus be altered ad lib to track traversal state changes
   virtual SceneGraphState *createTraversalState();
   virtual bool equal(const Object* obj) const;

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   virtual ~SceneGraphState();

};

}

#endif
