//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "SelectionTools.h"
#include "SGLegacyPCSceneAdaptor.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "MHGeoRangeIteratorAdaptor.h"
#include "MHSelectionIteratorAdaptor.h"
#include "LinearAlgebra_templCode.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace geoxl;
//---------------------------------------------------------------------------



/**
 * check if the given point cloud is selected
 */
bool geoxl::testPointCloudIsSelected(Scene *scene, string PCName)
{
   set<string>::iterator start = scene->getRootState()->staticState->selectedNodes.begin();
   set<string>::iterator end = scene->getRootState()->staticState->selectedNodes.end();
   for (; start != end; ++start) {
      if ((*start)==PCName) return true;
   };
   return false;
}


// in the following "closest point" will be used for the point with shortest distance to ray axis.
// "nearest point" will be used for the point being nearest to the camera measured in ray-z-direction coordinates.


/**
 * returns the name with path of the cloud the user has clicked on or "" if no cloud was clicked
 * input: x and y define the mouse position, w and h the width and heigth of the viewport-window, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * invisible pointclouds are ignored
 */
string geoxl::getClickedCloud(int32 x, int32 y, card32 w, card32 h, Camera *cam, ViewFrustum *viewfrust, Scene *scene) {
   float32 tmpZ = 1.0f;
	string returnCloudName = "";
	Vector3f returnPointPosition = NULL_VECTOR3F;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, 5, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, returnPointPosition, returnCloudName, tmpZ);
   delete It;
   delete range;
   if (tmpZ==1.0f) return "";
      else return returnCloudName;
}


/**
 * returns true if the closest point relative to mouse x/y on screen (within a defined circle) is found. also returns point position and name of corresponding cloud. runs only through selected clouds!
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: in returnPointPosition the position of the found point and in returnCloudName the name of the corresponding cloud is returned
 *  returns false if no point is found, invisible pointclouds are ignored
 */
bool geoxl::getClosestPointPositionFromSelectedClouds(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, Vector3f& returnPointPosition, string& returnCloudName) {
   float32 tmpZ = 1.0f;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(true, false, false, range, scene, returnPointPosition, returnCloudName, tmpZ);
   delete It;
   delete range;
   if (tmpZ==1.0f) return false;
      else return true;
}


/**
 * returns true if the closest point relative to mouse x/y on screen (within a defined circle) is found. also returns point position and name of corresponding cloud. runs only through selected points!
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: in returnPointPosition the position of the found point and in returnCloudName the name of the corresponding cloud is returned
 *  returns false if no point is found, invisible pointclouds are ignored
 */
bool geoxl::getClosestPointPositionFromSelectedPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, Vector3f& returnPointPosition, string& returnCloudName) {
   float32 tmpZ = 1.0f;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, true, false, range, scene, returnPointPosition, returnCloudName, tmpZ);
   delete It;
   delete range;
   if (tmpZ==1.0f) return false;
      else return true;
}


/**
 * returns true if the closest point relative to mouse x/y on screen (within a defined circle) is found. also returns point position and name of corresponding cloud. runs through all points.
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: in returnPointPosition the position of the found point and in returnCloudName the name of the corresponding cloud is returned
 *  returns false if no point is found, invisible pointclouds are ignored
 */
bool geoxl::getClosestPointPositionFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, Vector3f& returnPointPosition, string& returnCloudName) {
   float32 tmpZ = 1.0f;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, returnPointPosition, returnCloudName, tmpZ);
   delete It;
   delete range;
   if (tmpZ==1.0f) return false;
      else return true;
}


/**
 * returns copied iterator pointing to and name of corresponding cloud of the nearest point within a defined circle around mouse x/y on screen. runs through all points.
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: returns an PointCloudIterator pointing to the found point. you have to delete the returned iterator seperately after use!
 *  in returnCloudName the name of the corresponding cloud is returned.
 *  returns NULL if no point is found, invisible pointclouds are ignored
 */
PointCloudIterator* geoxl::getNearestPointIteratorFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, string& returnCloudName) {
   Vector3f tmpPos;
   float32 tmpZ;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* pcIt = getClosestPointAndNearestZ(false, false, true, range, scene, tmpPos, returnCloudName, tmpZ, true);
   delete range;
   return pcIt;
}


/**
 * returns copied iterator pointing to and name of corresponding cloud of the closest point relative to mouse x/y on screen (within a defined circle). runs through all points.
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: returns an PointCloudIterator pointing to the found point. you have to delete the returned iterator seperately after use!
 *  in returnCloudName the name of the corresponding cloud is returned.
 *  returns NULL if no point is found, invisible pointclouds are ignored
 */
PointCloudIterator* geoxl::getClosestPointIteratorFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, string& returnCloudName,
																					Vector3f * retPointPtr) {
   Vector3f tmpPos;
   float32 tmpZ;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* pcIt = getClosestPointAndNearestZ(false, false, true, range, scene, tmpPos, returnCloudName, tmpZ,true);
   delete range;
	if( retPointPtr != NULL )
		*retPointPtr = tmpPos;
   return pcIt;
}


/**
 * returns the nearest found z-value (negative) in ray coordinates inside a defined circle around mouse x/y on screen. runs through all points.
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel search-circle around mouse position, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: returns ray-z-value (negative) of found point or 1.0f if no point is found, invisible pointclouds are ignored
 */
float32 geoxl::getNearestZValueFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene) {
   Vector3f tmpPos;
   string tmpCloud;
   float32 nZ = 1.0f;
   GeometricRange *range = getRayConeRangeFromMouseClick(x, y, w, h, tolerance, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, tmpPos, tmpCloud, nZ);
   delete It;
   delete range;
   return nZ;
}


/**
 * returns the nearest found z-value (negative) in ray coordinates inside a defined box around mouse x/y on screen. runs through all points.
 * input: x1,y1 and x2,y2 define the two corner points on 2D-screen, w and h the width and heigth of the window, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: returns ray-z-value (negative) of found point or 1.0f if no point is found, invisible pointclouds are ignored
 */
float32 geoxl::getNearestZValueFromAllPoints(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera *cam, ViewFrustum *viewfrust, Scene *scene) {
   Vector3f tmpPos;
   string tmpCloud;
   float32 nZ = 1.0f;
   GeometricRange *range = getRayPyramidalRangeFromMouseClick(x1, y1, x2, y2, w, h, cam, viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, tmpPos, tmpCloud, nZ);
   delete It;
   delete range;
   return nZ;
}


/**
 * returns copied iterator (or NULL if returnIterator=false) pointing to the closest point to ray-axis (or nearest) inside a RayCone- or RayPyramidal-Range. also returns point position, name of corresponding cloud and the nearest found z-value in the range in ray coordinates. you can filter the search-range by different switches. returned iterator has to be deleted separately after use!
 * input: with onlySelectedClouds and onlySelectedPoints you can enable to search only on selected clouds or selected points
 *  returnIterator defines if you want to return an PointCloudIterator pointing to the found point
 *  range may be an RayConeRange or an RayPyramidalRange which controls the search range and defines the search coordinate system, scene is the scene
 *  if returnNearest is true, then the nearest point instead of the closest point is returned!
 * output: in returnedPointPosition the position of the found point and in returnedCloudName the name of the corresponding cloud is returned
 *  in returnedNearestZ the nearest found ray-z-value in range according to the camera position is returned, this is abolutely independent of the found closest point(!), because the closest point to ray-axis usually won't simultaneously be the nearest to camera
 *  if returnIterator=true a PointCloudIterator pointing to the found point or NULL if no point is found is returned. the returned iterator has to be deleted after use!
 *  if returnIterator=false then NULL is returned in every case, so you have to do another test if no point was found: returnedNearestZ==1.0f is strongly suggested
 *  so it is possible to differentiate between "no point found" (NULL or returnedNearestZ=1.0f) and "point found" (legit Iterator or returnedNearestZ!=1.0f).
 * invisible pointclouds are ignored
 * (may easily be extended in future to optionally return nearest instead of closest point if needed) - MH,05/07/07,done
 */
PointCloudIterator* geoxl::getClosestPointAndNearestZ(bool onlySelectedClouds, bool onlySelectedPoints, bool returnIterator, GeometricRange* range, Scene *scene, Vector3f& returnedPointPosition, string& returnedCloudName, float32& returnedNearestZ, bool returnNearest) {

   float32 SqrDist = 0.0f;
   float32 OldDist = -1.0f;
   float32 NearestFoundZ = 1.0f;
   Vector3f FoundPointPosition = NULL_VECTOR3F;
   string FoundPointCloudName = "";
   PointCloudIterator* FoundPointIterator = NULL;

   // restrict the search region with a range
   if (range==NULL) return NULL;
   const RayConeRange *ConeRange = dynamic_cast<const RayConeRange*>(range);
   const RayPyramidalRange *PyramidRange = dynamic_cast<const RayPyramidalRange*>(range);
   if ((ConeRange==NULL) && (PyramidRange==NULL)) return NULL;
   pAssert(! ((ConeRange!=NULL) && (PyramidRange!=NULL)));

   StandardSceneGraphIterator* sceneIterator = new StandardSceneGraphIterator();
   sceneIterator->clearAndSetup(scene->getRootNode(), scene->getRootState());

   while (!sceneIterator->atEnd()) {

      // if desired search only on selected clouds
      if (onlySelectedClouds) {
         if (!testPointCloudIsSelected(scene, sceneIterator->getNameWithPath())) { sceneIterator->next(); continue; }
      };

      SGObjectNode* SGN = dynamic_cast<SGObjectNode*>(sceneIterator->getNode());
      if ((SGN==NULL) || !SGN->visible()) { sceneIterator->next(); continue; }
      PointCloud* PC = dynamic_cast<PointCloud*>(SGN->getSceneObject());
      if ((PC==NULL) || (!PC->providesAttribute("position")) || (!PC->providesAttribute("flags"))) { sceneIterator->next(); continue; }
      AAT POSITION = PC->getAAT("position");
      AAT FLAGS = PC->getAAT("flags");

      SceneObjectIterator *newItSO = PC->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_HIERARCHICAL));
      PointCloudIterator *newIt = dynamic_cast<PointCloudIterator*>(newItSO);
      if ((newIt==NULL)||(newItSO==NULL)) { delete newItSO; sceneIterator->next(); continue; };

      // set cloud transformation matrix and make range iterator adaptor
      range->setTransformation(sceneIterator->getState()->dynamicState->ctm);
      IteratorAdaptor *rIt = getGeoRangeIteratorAdaptor(PC, newIt, range);
      if (rIt == NULL) {delete newItSO; sceneIterator->next(); continue;}
      Vector3f tmpPos;
      float32 pointRayZ;

      while(!rIt->atEnd()) {

         // if desired search only on selected points
         if (onlySelectedPoints)
            if (!((rIt->get1i(FLAGS)) & PF_FLAG_SELECTED)) { rIt->next(); continue; };

         // get position of point, squared distance and z
         tmpPos = rIt->get3f(POSITION);
         if (ConeRange!=NULL) ConeRange->getSqrDistanceAndZ(tmpPos, SqrDist, pointRayZ);
            else if (PyramidRange!=NULL) PyramidRange->getSqrDistanceAndZ(tmpPos, SqrDist, pointRayZ);
               else return NULL;

         // save the nearest-to-camera z-value found in range (independent of closest point!)
			if ((pointRayZ>NearestFoundZ) || (NearestFoundZ==1.0f)) {
				NearestFoundZ = pointRayZ;
				if (returnNearest) {
	            FoundPointPosition = tmpPos;
		         FoundPointCloudName = sceneIterator->getNameWithPath();
			      if (returnIterator) {
				      delete FoundPointIterator;
					   FoundPointIterator = (PointCloudIterator*)(newIt->copy());
					};
				}
			}

         // test if point is closer to mouse x/y as any point before (independent of nearest z!)
         if ((SqrDist<OldDist) || (OldDist==-1.0f)) {
            OldDist = SqrDist;
				if (!returnNearest) {
	            FoundPointPosition = tmpPos;
		         FoundPointCloudName = sceneIterator->getNameWithPath();
			      if (returnIterator) {
				      delete FoundPointIterator;
					   FoundPointIterator = (PointCloudIterator*)(newIt->copy());
					};
				};
         };
         rIt->next();
      };

      delete rIt;
      delete newItSO;
      sceneIterator->next();
   }; // end while sceneIterator

   delete sceneIterator;

   if (OldDist==-1.0f) NearestFoundZ=1.0f;

   returnedNearestZ = NearestFoundZ;
   returnedPointPosition = FoundPointPosition;
   returnedCloudName = FoundPointCloudName;
   return FoundPointIterator;
}



/**
 * returns a cone GeometricRange pointing from camera origin in the direction of the mouse click on screen with a defined aperture angle
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel circle around mouse position giving the cone aperture angle, cam and viewfrust the actual camera and viewing frustum
 * output: a new RayConeRange with origin at the camera position, RayDir as cone axis, RayUp aligned to the screen y-axis and the _squared_ radius at distance 1 from ray origin (k1sqr) defining the aperture angle
 * the returned GeometricRange has to be deleted sperately after use!
 */
RayConeRange* geoxl::getRayConeRangeFromMouseClick(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust) {

   // size of one pixel in world coordinates
   float32 PixelSize = (tan(viewfrust->getVerticalFieldOfView()*0.5f*((float32)M_PI/180.0f))) / (h*0.5f);

   // mouse click in world coordinates
   float32 Click3D_X =  (x-w*0.5f) * PixelSize;
   float32 Click3D_Y =  (y-h*0.5f) * PixelSize;

   // squared radius of cone at z-distance 1 from ray origin
   float32 k1sqr = (tolerance*PixelSize*tolerance*PixelSize) / (1 + Click3D_X*Click3D_X + Click3D_Y*Click3D_Y);

   // calculate cone coordinate system
   Vector3f RayDir = makeVector3f(Click3D_X, Click3D_Y, -1.0f);
   RayDir = normalize(RayDir);
   RayDir = (cam->getLocalCoordsMatrix()) * RayDir;
   Vector3f RayUp = (cam->getOrthoNormRightDirection()).crossProduct(RayDir);
   RayUp = normalize(RayUp);
   Vector3f RayRight = RayDir.crossProduct(RayUp);
   RayRight = normalize(RayRight);

   return new RayConeRange(RayRight, RayUp, RayDir, cam->getPosition(), k1sqr);
};


/**
 * returns a sphere GeometricRange centered near to the nearest point towards camera in a defined circle around mouse click with squared radius adapted to the tolerance circle on screen
 * input: x and y define the mouse position, w and h the width and heigth of the window, tolerance the pixel circle around mouse position giving the sphere radius, cam and viewfrust the actual camera and viewing frustum
 * output: a new SphereRange with center near to the nearest point found in the and squared radius adapted to the tolerance circle on screen.
 * the returned GeometricRange has to be deleted sperately after use!
 */
GeometricRange* geoxl::getNearestSphereRangeFromMouseClick(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene) {

   // z-value in ray-coordinates of the nearest point in tolerance circle
   Vector3f tmpPos;
   string tmpCloud;
   float32 nearestZ = 1.0f;
   RayConeRange *range = (RayConeRange*)getRayConeRangeFromMouseClick(x,y,w,h,tolerance,cam,viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, tmpPos, tmpCloud, nearestZ);
   delete It;
   if (nearestZ==1.0f) { delete range; return NULL; };

   // use ray coordinate system
   float32 tmpSqrRadAt1= range->getSqrRadiusAt1();
   Vector3f tmpRayDir = range->getRayDir();
   Vector3f tmpRayOrigin = range->getRayOrigin();

   return new SphereRange(tmpRayOrigin+(tmpRayDir*nearestZ), (nearestZ*nearestZ*tmpSqrRadAt1));
};


/**
 * returns a pyramidal GeometricRange pointing from camera origin in the direction of the mouse click on screen with defined horizontal and vertical aperture angles
 * input: x1,y1 and x2,y2 define the two corner points on 2D-screen, w and h the width and heigth of the window, cam and viewfrust the actual camera and viewing frustum
 * output: a new RayPyramidalRange with origin at the camera position, RayDir as pyramid axis, RayUp aligned to the screen y-axis and the _squared_ half-width and half-height at distance 1 from ray origin (SqrRightAt1 and SqrUpAt1) defining the horizontal and vertical aperture angles
 * the returned GeometricRange has to be deleted sperately after use!
 */
GeometricRange* geoxl::getRayPyramidalRangeFromMouseClick(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera *cam, ViewFrustum *viewfrust) {

   // find middle of 2D-box on screen
   int32 x = (x1+x2)/2;
   int32 y = (y1+y2)/2;

   // calculate vertical and horizontal "opening angle"
   int32 sqrTolX = (x-x1)*(x-x1);
   int32 sqrTolY = (y-y1)*(y-y1);

   // size of one pixel in world coordinates
   float32 PixelSize = (tan(viewfrust->getVerticalFieldOfView()*0.5f*((float32)M_PI/180.0f))) / (h*0.5f);

   // mouse click in world coordinates
   float32 Click3D_X = (x-w*0.5f) * PixelSize;
   float32 Click3D_Y = (y-h*0.5f) * PixelSize;

   // squared half-width and half-height of pyramid at z-distance 1 from ray origin
   float32 SqrRightAt1 = (sqrTolX*PixelSize*PixelSize) / (1 + Click3D_X*Click3D_X);
   float32 SqrUpAt1 = (sqrTolY*PixelSize*PixelSize) / (1 + Click3D_Y*Click3D_Y);

   // calculate ray coordinate system
   Vector3f RayDir = makeVector3f(Click3D_X, Click3D_Y, -1.0f);
   RayDir = normalize(RayDir);
   RayDir = (cam->getLocalCoordsMatrix()) * RayDir;
   Vector3f RayUp = (cam->getOrthoNormRightDirection()).crossProduct(RayDir);
   RayUp = normalize(RayUp);
   Vector3f RayRight = RayDir.crossProduct(RayUp);
   RayRight = normalize(RayRight);

   return new RayPyramidalRange(RayRight, RayUp, RayDir, cam->getPosition(), SqrRightAt1, SqrUpAt1);
};


/**
 * returns a clipped pyramidal GeometricRange pointing from camera origin in the direction of the mouse click on screen with defined horizontal and vertical aperture angles and a near and far clipping plane defined by the nearest point towards camera (this at all makes a frustum)
 * input: x1,y1 and x2,y2 define the two corner points on 2D-screen, w and h the width and heigth of the window, cam and viewfrust the actual camera and viewing frustum, scene the scene
 * output: a new RayClippedPyramidalRange (frustum) with origin at the camera position, RayDir as pyramid axis, RayUp aligned to the screen y-axis and the _squared_ half-width and half-height at distance 1 from ray origin (SqrRightAt1 and SqrUpAt1) defining the horizontal and vertical aperture angles. additionally this pyramid has a near and a far clipping distance restricting the interior.
 * the returned GeometricRange has to be deleted sperately after use!
 */
GeometricRange* geoxl::getNearestClippedRayPyramidalRangeFromMouseClick(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera *cam, ViewFrustum *viewfrust, Scene *scene) {

   // z-value in ray-coordinates of the nearest point in tolerance box
   Vector3f tmpPos;
   string tmpCloud;
   float32 nearestZ = 1.0f;
   RayPyramidalRange *range = (RayPyramidalRange*)getRayPyramidalRangeFromMouseClick(x1,y1,x2,y2,w,h,cam,viewfrust);
   PointCloudIterator* It = getClosestPointAndNearestZ(false, false, false, range, scene, tmpPos, tmpCloud, nearestZ);
   delete It;
   if (nearestZ==1.0f) { delete range; return NULL; };

   // squared half-width and half-height of pyramid at z-distance 1 from ray origin
   float32 SqrHalfWidthAt1 = range->getSqrHalfWidthAt1();
   float32 SqrHalfHeightAt1 = range->getSqrHalfHeightAt1();

   // calculate near and far clipping distances
   float32 zTol = nearestZ*sqrt(0.5f*(SqrHalfWidthAt1+SqrHalfHeightAt1));
   float32 nClip = nearestZ - zTol;
   float32 fClip = nearestZ + zTol;

   // ray coordinate system
   Vector3f RayRight = range->getRayRight();
   Vector3f RayUp = range->getRayUp();
   Vector3f RayDir = range->getRayDir();

   delete range;

   return new RayClippedPyramidalRange(RayRight, RayUp, RayDir, cam->getPosition(), SqrHalfWidthAt1, SqrHalfHeightAt1, nClip, fClip);
}



/**
 * returns new MHGeoRangeIteratorAdaptor, which has to be deleted separately after use.
 * use this function whenever possible, due to code must only be modified here for implementation changes
 */
IteratorAdaptor* geoxl::getGeoRangeIteratorAdaptor(PointCloud* pc,PointCloudIterator* it, GeometricRange *r) {
   pAssert(it!=NULL);
   return new MHGeoRangeIteratorAdaptor(it,pc->getDescr(), r);
   //return new GeoRangeIteratorAdaptor(it,pc->getDescr(), r);
}


/**
 * returns new MHSelectionIteratorAdaptor, which has to be deleted separately after use.
 * use this function whenever possible, due to code must only be modified here for implementation changes
 */
IteratorAdaptor* geoxl::getSelectionIteratorAdaptor(PointCloud* pc ,PointCloudIterator* it) {
   pAssert(it!=NULL);
   return new MHSelectionIteratorAdaptor(it, pc->getDescr());
   // return new SelectionIteratorAdaptor(it, pc->getDescr());
}



/**
 *
 */
Matrix3f geoxl::createCovarianceAndMean(PointCloud* pc, IteratorAdaptor* it, Vector3f &mean, card32 &numPoints, const Matrix4f ctm ){

   Matrix3f sigma;

   Matrix4f curtransmatr = ctm;
   AAT POSITION = pc->getAAT("position");

   mean = makeVector3f(0.0f, 0.0f, 0.0f);

   int32 count=0;

   while(!it->atEnd()){
     //calculate mean Position

      Vector4f hPos = expand3To4(it->get3f(POSITION));
      hPos = curtransmatr*hPos;
      float32 w = hPos[3];
      if (fabs(w) > 0.0001f) {
         Vector3f pos = makeVector3f(hPos[0]/w, hPos[1]/w,hPos[2]/w);
         mean+=pos;
         count++;
      }
      it->next();
   }

   numPoints=count;
   if (count==0) return sigma;
   mean/=(float32)count;
   //end calculate mean Position

   //calculate sigma
   it->reset();
   count=0;

   float covXX ; float covXY ; float covXZ ;
   float covYY ; float covYZ ; float covZZ ;

   while(!it->atEnd()){
      Vector4f hPos = expand3To4(it->get3f(POSITION));
      hPos = curtransmatr*hPos;
      float32 w = hPos[3];
      if (fabs(w) > 0.01f) {
         Vector3f pos = makeVector3f(hPos[0]/w, hPos[1]/w,hPos[2]/w);

         covXX=(pos[0]-mean[0])*(pos[0]-mean[0]);
         covXY=(pos[0]-mean[0])*(pos[1]-mean[1]);
         covXZ=(pos[0]-mean[0])*(pos[2]-mean[2]);
         covYY=(pos[1]-mean[1])*(pos[1]-mean[1]);
         covYZ=(pos[1]-mean[1])*(pos[2]-mean[2]);
         covZZ=(pos[2]-mean[2])*(pos[2]-mean[2]);

         sigma[0]+=makeVector3f(covXX,covXY,covXZ);
         sigma[1]+=makeVector3f(covXY,covYY,covYZ);
         sigma[2]+=makeVector3f(covXZ,covYZ,covZZ);
      }

      count++;
      it->next();
   };

   sigma/=(float32)count;

   return sigma;
}

/**
 *
 */
Matrix3f geoxl::createCovariance(PointCloud* pc, const Matrix4f ctm ){

   Matrix3f sigma;

   SceneObjectIterator *tmpit2SO = pc->createIterator(SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_HIERARCHICAL));
   PointCloudIterator *tmpit2 = dynamic_cast<PointCloudIterator*>(tmpit2SO);
   pAssert(tmpit2);
   IteratorAdaptor *it = getSelectionIteratorAdaptor(pc, tmpit2);
   // replaces: SelectionIteratorAdaptor *sit = new SelectionIteratorAdaptor(newIt, pc->getDescr());

   Matrix4f curtransmatr = ctm;
   AAT POSITION = pc->getAAT("position");

   Vector3f mean = makeVector3f(0.0f, 0.0f, 0.0f);

   int32 count=0;

   while(!it->atEnd()){
     //calculate mean Position

      Vector4f hPos = expand3To4(it->get3f(POSITION));
      hPos = curtransmatr*hPos;
      float32 w = hPos[3];
      if (fabs(w) > 0.0001f) {
         Vector3f pos = makeVector3f(hPos[0]/w, hPos[1]/w,hPos[2]/w);
         mean+=pos;
         count++;
      }
      it->next();
   }

   if (count==0) return sigma;
   mean/=(float32)count;
   //end calculate mean Position

   //calculate sigma
   it->reset();
   count=0;

   float covXX ; float covXY ; float covXZ ;
   float covYY ; float covYZ ; float covZZ ;

   while(!it->atEnd()){
      Vector4f hPos = expand3To4(it->get3f(POSITION));
      hPos = curtransmatr*hPos;
      float32 w = hPos[3];
      if (fabs(w) > 0.01f) {
         Vector3f pos = makeVector3f(hPos[0]/w, hPos[1]/w,hPos[2]/w);

         covXX=(pos[0]-mean[0])*(pos[0]-mean[0]);
         covXY=(pos[0]-mean[0])*(pos[1]-mean[1]);
         covXZ=(pos[0]-mean[0])*(pos[2]-mean[2]);
         covYY=(pos[1]-mean[1])*(pos[1]-mean[1]);
         covYZ=(pos[1]-mean[1])*(pos[2]-mean[2]);
         covZZ=(pos[2]-mean[2])*(pos[2]-mean[2]);

         sigma[0]+=makeVector3f(covXX,covXY,covXZ);
         sigma[1]+=makeVector3f(covXY,covYY,covYZ);
         sigma[2]+=makeVector3f(covXZ,covYZ,covZZ);
      }

      count++;
      it->next();
   };

   sigma/=(float32)count;

   return sigma;
}



/**
 * Rotate u and v from local coordSystem to match cam up and right  //TODO fix bug
 */
void geoxl::considerCamPos( Scene* scene, Vector3f &u,Vector3f &v,Vector3f &n ){

   Vector3f camRight = (scene->getRootState()->staticState->camera).getOrthoNormRightDirection();
   Vector2f camRiUV;
   camRiUV[0]=u*camRight;
   camRiUV[1]=v*camRight;

   Vector3f camUp   = (scene->getRootState()->staticState->camera).getOrthoNormUpDirection();
   Vector3f camView = (scene->getRootState()->staticState->camera).getOrthoNormViewDirection();

   Vector3f texU = u * camRiUV[0] + v * camRiUV[1];
   //  Vector3f camUpOnPlane = u * camUiUV[0] + v * camUiUV[1];

   if (norm(texU) >= 1E-7) {
      texU = normalize(texU);
   } else {
      texU = camRight;
   }

   Vector3f texV = n.crossProduct(texU);
   texV=normalize(texV);

   /* if (norm(camUpOnPlane) >= 1E-7) {
      camUpOnPlane = normalize(camUpOnPlane);
   }*/


   u=texU;
   v=texV;
   //n=n

    /*
   debugOutput << "cam up: " << camUp << " r: " << camRight << " v: " <<camView;
   debugOutput.cr();

   debugOutput << "my   u: " << u << " v: " << v << " w: " <<n;
   debugOutput.cr();
     */
}

